/*!
 * @file autosave.h
 * @brief This file defines the AutoSave class(es).
 * @author Anthony L. Montiel
 * @date 2013-02-13
 *
 * Copyright (C) 2021 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_AUTOSAVE_H_
#define _FASTEC_AUTOSAVE_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <limits.h>
#include <condition_variable>
#include <deque>
#include <mutex>

#include <fastec/fas_thread.h>

namespace AUTO_SAVE
{
//--------------------------------------------------------------------
// Constants, Macros and Type definitions
//--------------------------------------------------------------------

//! Define a queue type comprised of session entries. Each session
//! entry is nothing more than a partition number.
typedef std::deque<int> SESS_QUEUE;

} /* End of namespace AUTO_SAVE */

//--------------------------------------------------------------------
//! @brief The AutoSave class.
//--------------------------------------------------------------------
class AutoSave : public FAS_THREAD::Thread
{
private:
    static const int MAX_QUEUED_PARTS;      //!< The maximum number of partitions we can queue to the capture pipeline.
    static const int MAX_RETRIES;           //!< The maximum number of times we will retry an operation.
    static const uint32_t SLEEP_MS;         //!< The maximum delay we will wait for an operation to complete before checking status again.

    static std::mutex               work_mutex;
    static std::condition_variable  work_cond;

    volatile bool threadRunning;            //!< Flag to indicate if we are running or not.
    uint32_t traceControl;                  //!< Trace control word for class.

    uint32_t sessionCount;                  //!< Number of sessions we are supporting.

    uint32_t saveStartNum;                  //!< The user specified starting frame number.
    uint32_t saveEndNum;                    //!< The user specified ending frame number.

    AUTO_SAVE::SESS_QUEUE  availQueue;      //!< Our queue of available sessions.
    AUTO_SAVE::SESS_QUEUE  busyQueue;       //!< Our queue of sessions posted to the FPGA FIFO.
    AUTO_SAVE::SESS_QUEUE  dirtyQueue;      //!< Our queue of dirty sessions waiting to be saved.
    std::mutex  queueLock;                  //!< Control access to the queues.

    //! @name FasFireLR support variables.
    //! @{
    uint32_t total_fflr_frames;             //!< The total number of FFLR frames written.
    int fflr_file_count;                    //!< Count of the FFLR files.
    //
    //! @}

private:
    void    adjustSave(int part_num);
    void    cleanup();
    void    fflrDiscard();
    int     getDirtyQueueFront();
    bool    makeDirty();
    int     queueNextAvail();
    void    releaseQueues();
    void    updateRegisterValue();

public:
    AutoSave();
    virtual ~AutoSave();
    void run() override;
    void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;
    void stop() override;

    bool doOverlap();
    bool armOK();
    bool isRunning()    { return threadRunning; }
    bool isEnabled();

    bool init(uint32_t session_count);
    void process();
    void queueUpdate();
    void reset();

    uint32_t availSessions() const      { return availQueue.size(); }
    uint32_t busySessions()  const      { return busyQueue.size(); }
    uint32_t dirtySessions()  const     { return dirtyQueue.size(); }

    int getNextPartition()  const       { return busyQueue.front(); }
    int getDirty()  const               { return dirtyQueue.front(); }

    constexpr int getFflrFileCount()    { return fflr_file_count; }
    constexpr uint32_t getTotalFflrFrames() { return total_fflr_frames; }
};

#endif /* !_FASTEC_AUTOSAVE_H_ */

