/*!
 * @file camera_server.h
 * @brief This is the header file for the camera server class.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __CAMERA_SERVER_H__
#define __CAMERA_SERVER_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>

#include <list>
#include <memory>

#include <fas_trace.h>
#include <fas_sharedmem.h>

#include "cs_network.h"
#include "cs_utils.h"
#include "cs_worker.h"
#include "fpga.h"
#include "gige_reg.h"
#include "lic_manager.h"

namespace CAMERA_SERVER
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! @name Starting values for messages specific to a particular function.
//! @{
const int CS_MSG_BASE = 20000;
const int FCP_MSG_BASE = 30000;
const int FPGA_MSG_BASE = 40000;
const int STREAM_MSG_BASE = 50000;
//! @}

//! @name Trace Control Selectors.
//! @{
const uint32_t FCP              = 0x00000010 | TRACE_2;     //!< Trace selector for FCP major events.
const uint32_t FCP_HEARTBEAT    = 0x00000020 | TRACE_2;     //!< Trace selector for FCP Heartbeat timer.
const uint32_t FCP_REG_ACCESS   = 0x00000040 | TRACE_2;     //!< Trace selector for FCP Register access.
const uint32_t FCP_REG_CONFIG   = 0x00000080 | TRACE_2;     //!< Trace selector for FCP Register setup/configuration.
const uint32_t FCP_DISCOVERY    = 0x00000100 | TRACE_2;     //!< Trace selector for FCP Discovery handshake.
const uint32_t FCP_IP           = 0x00000200 | TRACE_2;     //!< Trace selector for FCP IP events.
const uint32_t FCP_MESSAGE      = 0x00000400 | TRACE_2;     //!< Trace selector for FCP message events.
const uint32_t FCP_EVENT_MSG    = 0x00000800 | TRACE_2;     //!< Trace selector for FCP event messages (message channel).
const uint32_t CS_CTOR_DTOR     = 0x00001000 | TRACE_2;     //!< Trace selector for object constructor/destructor.
const uint32_t CS_STATE         = 0x00002000 | TRACE_2;     //!< Trace selector for camera server state transitions.
const uint32_t CS_LINCE_ACCESS  = 0x00004000 | TRACE_2;     //!< Trace selector for accesses to the Lince sensor.
const uint32_t CS_LINCE_ALIGN   = 0x00008000 | TRACE_2;     //!< Trace selector for Lince sensor alignment logic.
const uint32_t FPGA_OPS         = 0x00010000 | TRACE_2;     //!< Trace selector for FPGA operations.
const uint32_t STREAMER_OPS     = 0x00010000 | TRACE_2;     //!< Trace selector for streaming operations.
//! @}

} /* CAMERA_SERVER namespace */

//--------------------------------------------------------------------
// Class declarations
//--------------------------------------------------------------------

class AutoSave;
class Sensor;
class PartMgr;

class CameraServer
{
private:
    static CameraServer *camserv_ptr;       //!< Our single global instance.
    static std::mutex               work_mutex;
    static std::condition_variable  work_cond;
    static std::list<CS_NETWORK::CS_NETWORK_MESSAGE>    msg_queue;
    static FSM_HANDLE shmem_handle;

    std::unique_ptr<LicManager> lic_ptr;    //!< Pointer to the licensing information.
    std::unique_ptr<csWorker>   worker;     //!< Our worker thread.
    std::unique_ptr<cs_poll>    dvr_poller; //!< Our polling thread looking for the FPGA to finish DVR playback.
    std::unique_ptr<cs_watchdog> wd_petter; //!< Our watchdog petting thread.
    std::unique_ptr<csAutoTrigger> auto_trig;   //!< Our auto-trigger thread.

    std::unique_ptr<Sensor>     sensor_ptr; //!< Our sensor control class.
    std::unique_ptr<AutoSave>   as_mgr;     //!< Our autosave management class.
    GIGE_REG_CACHE  reg_cache;          //!< Our register cache.

    uint8_t     camera_state;           //!< What state is the camera in.
    uint32_t    cs_tid;                 //!< The thread ID of the camera server thread.

    uint32_t    as_shot_red;            //!< The AsShotNeutral value for the red channel.
    uint32_t    as_shot_grn;            //!< The AsShotNeutral value for the green channel.
    uint32_t    as_shot_blu;            //!< The AsShotNeutral value for the blue channel.

    bool    do_factory_reset;       //!< True if we're in the middle of a factory reset.
    bool    init_complete;

    void    integrity_check();

    void    cleanup();
    void    do_update(bool net_update);
    void    process_fpga_events(uint32_t fpga_val);
    void    process_internal(CS_NETWORK::CS_INTERNAL *p_msg);
    bool    instance_set_camera_state(uint8_t new_state);

    uint8_t set_idle_state();
    uint8_t set_live_state();
    uint8_t set_armed_state();
    uint8_t set_triggered_state();
    uint8_t set_dearming_state();
    uint8_t set_review_state();
    uint8_t set_autosave_state();
    uint8_t set_snapshot_state();

public:
    CameraServer();
    virtual ~CameraServer();

    bool    init();
    void    run();
    void    reset_registers();

    static bool init_is_done();
    static void init_done();
    static void add_work(CS_NETWORK::CS_NETWORK_MESSAGE *msg_ptr);

    static CameraServer *get_cs();
    static FiFPGA   *get_fpga();
    static PartMgr  *get_partmgr();
    static Sensor   *get_sensor();
    static csWorker *get_worker();
    static AutoSave *get_autosave();
    static GIGE_REG_CACHE& get_regs();
    static const LicManager *get_license();
    static FSM_HANDLE get_shmem();

    static void exit();
    static void reboot();
    static void restart();
    static void reopen();
    static void shutdown();

    static bool async_set_camera_state(uint8_t new_state, bool wait, volatile int *p_done);
    static bool set_camera_state(uint8_t new_state);
    static bool user_set_camera_state(uint8_t new_state);
    static uint8_t get_camera_state();
    static const char *board_rev();
    static bool rev_is_ok(const char *rev_ptr);
    static void watchdog();

    static bool in_factory_reset();
    static void factory_reset(bool do_reset);
    static void saw_signal(int which);

public:
    // Translate from coordinates that are relative to the sensor's licensed area
    // into coordinates that are relative to the sensor's full size.
    uint16_t get_true_startx(bitmode_t bitmode, uint32_t sensor_opts, uint16_t startx);
    uint16_t get_true_starty(bitmode_t bitmode, uint32_t sensor_opts, uint16_t starty);

    // Get the size of the licensed area.
    uint16_t get_lic_width();
    uint16_t get_lic_height();

    // Convert the floating point frame rate, to the register value.
    uint32_t frame_rate_to_reg(float frame_rate);

    // Convert the register frame rate value to the floating point value.
    float reg_to_frame_rate(uint32_t reg_value);

    // Get the licensed maximum frame rate.
    uint32_t get_max_licensed_frame_rate();

    // Get the max frame rate for ltr mode, which actually comes from the license file.
    uint32_t get_max_ltr_frame_rate();

    // Get the minimum allowed width as determined by the current LTR mode of the camera
    // and any FPGA or sensor restrictions.
    uint32_t get_min_width();
    uint32_t get_min_height();

    // Get the minimum and maximum shutter speeds in microseconds.
    // In general, these can depend on the sensor and the FPGA.
    uint32_t get_max_shutter_speed(float fps);
    uint32_t get_min_shutter_speed();

    void set_idle_behavior();
    void set_live_behavior();

    bool auto_trigger();
    bool dvr_playback();
    bool integrated();
    bool optical();
    void poll_for_dvr_finish(bool do_poll = true);
    bool do_polling_for_dvr();
    bool check_ltr_recording();

    void get_user_bit_mode(bitmode_t *rtn_bit_mode);
    void get_frame_rate(float *rtn_frame_rate);
    void get_user_fpn(fpnmode_t *rtn_fpn);
    void get_sensor_opts(uint32_t *rtn_sensor_opts);
    void get_sensor_gain(uint32_t *rtn_sensor_gain);
    void get_shutter_speed(uint32_t *rtn_shutter_speed);
    bool get_user_roi(uint16_t *rtn_start_x, uint16_t *rtn_start_y,
                      uint16_t *rtn_width, uint16_t *rtn_height);
    void get_user_settings(uint16_t *rtn_start_x, uint16_t *rtn_start_y,
                           uint16_t *rtn_width, uint16_t *rtn_height,
                           bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                           uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts);
    bool get_sensor_start(uint16_t *rtn_start_x, uint16_t *rtn_start_y);
    void get_physical_settings(uint16_t *rtn_start_x, uint16_t *rtn_start_y,
                               uint16_t *rtn_width, uint16_t *rtn_height,
                               bitmode_t *rtn_bitmode, float *rtn_frame_rate,
                               uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts);
    bool make_capture_regs_valid();

    // Calculate the frame buffer size for current settings.
    uint32_t calc_frame_buf_size();

    // Calculate the frame buffer size for passed settings.
    uint32_t calc_frame_buf_size(uint16_t width, uint16_t height, bitmode_t bit_mode)
    {
        return calc_src_dma_size(width, height, bit_mode);
    }

    // Is this ROI valid? If so, return true and provide the maximum frame rate this ROI supports.
    bool is_user_roi_valid(uint16_t startx, uint16_t starty, uint16_t width, uint16_t height,
                           bitmode_t bitmode, float framerate, uint32_t shutter_speed,
                           uint32_t sensor_opts, float *rtn_max_framerate);
    bool make_capture_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                            uint16_t *rtn_width, uint16_t *rtn_height,
                            bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                            uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts);
    bool make_shutter_valid(uint32_t fps, uint32_t *shutter_speed);

    bool setup_fpga_and_sensor();
    bool modify_capture(uint16_t startx, uint16_t starty, uint16_t width, uint16_t height,
                        bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed,
                        uint32_t sensor_opts);
    bool setup_capture_from_regs();
    bool start_lince_calibration(bool factory, bitmode_t cal_bit_mode, fpnmode_t fpn,
                                 float frame_rate, uint32_t shutter_speed);
    bool restart_normal_after_calibration();
    bool generate_custom_white_balance(uint32_t reg_value);
    bool load_ct_tables();
    void refresh_ct_status();

    bool set_fpn(uint32_t fpn_option);
    bool set_frame_rate(uint32_t frame_rate);

    // Return the AsShotNeutral values.
    // JUL2014 We used to keep these in the sensor object, probably just because there happened
    // JUL2014 to be other static data stored in it.
    uint32_t    get_as_shot_red() { return as_shot_red; }
    uint32_t    get_as_shot_grn() { return as_shot_grn; }
    uint32_t    get_as_shot_blu() { return as_shot_blu; }
    void    set_as_shot_red(uint32_t r_val) { as_shot_red = r_val; }
    void    set_as_shot_grn(uint32_t g_val) { as_shot_grn = g_val; }
    void    set_as_shot_blu(uint32_t b_val) { as_shot_blu = b_val; }
};

#endif /* !__CAMERA_SERVER_H__ */
