/*!
 * @file constants.h
 * @brief Common constants used within the Camera Server
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __CONSTANTS_H__
#define __CONSTANTS_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

// 24 frames per second for low light, since we support that normally.
const int LOW_LIGHT_FRAME_RATE = 24;

// Minimum ROI width for LTR mode.
const int LTR_MIN_WIDTH = 320;

// Minimum ROI height for LTR mode.
const int LTR_MIN_HEIGHT = 32;

//---------------------------------------------------
// DMA Backpressure Threshold and Request Size Values
//---------------------------------------------------

#define IPM_SRC_DMA_BP_THRESHOLD_VAL    0x03E08010
#define IPM_SRC_DMA_REQ_SIZE_VAL        0x08
#define IPM_DEST_DMA_BP_THRESHOLD_VAL   0x8010
#define IPM_DEST_DMA_REQ_SIZE_VAL       0x08
#define IPM_JPEG_DMA_BP_THRESHOLD_VAL   0x8010
#define IPM_JPEG_DMA_REQ_SIZE_VAL       0x08

#define IM2SSD_DMA_BP_THRESHOLD_VAL     0x8010
#define IM2SSD_DMA_REQ_SIZE_VAL         0x08

#define IM2PCIE_DMA_BP_THRESHOLD_VAL    0x8004
#define IM2PCIE_DMA_BURST_SIZE          0x40

#define PCIE2IM_DMA_BURST_SIZE          0x20

#define CSI_DMA_BP_THRESHOLD_VAL        0x03E0D006
#define CSI_DMA_DDR2_REQ_SIZE_VAL       0x10
#define HDMI_DMA_BP_THRESHOLD_VAL       0x03E0C008
#define HDMI_DMA_DDR2_REQ_SIZE_VAL      0x20

#define GOE_RD_DMA_BP_THRESHOLD_VAL        0x8010
#define GOE_RD_DMA_DDR2_REQ_SIZE_VAL         0x10

#define AUX_WR_DMA_BP_THRESHOLD_VAL        0x8004
#define AUX_WR_DMA_DDR2_REQ_SIZE_VAL         0x08

#define SATA_RD_DMA_BP_THRESHOLD_VAL       0x8010
#define SATA_WR_DMA_BP_THRESHOLD_VAL       0xA010
// DMA request size for LTR-mode operations.
#define SATA_DDR2_DMA_REQ_SIZE_VAL_LTR       0x10   // 32 octwords or 1024

// DMA request size for non-LTR I/O operations (CAP read/write).
#define SATA_DDR2_DMA_REQ_SIZE_VAL_IO        0x08   // 8 octwords or 256

//! Maximum bandwidth we permit when we came up in SATA-2 mode.
#define SATA2_SSD_BANDWIDTH 240

//! FPGA appends 64 bytes (1 HEXWORDS) of Metadata onto the end of each frame.
#define FRAME_METADATA_SIZE_HW      1
#define FRAME_METADATA_HW_BYTES     64

//! Of the 64 bytes the FPGA appends, 32 bytes contain actual Metadata.
#define FRAME_METADATA_SIZE_BYTES   32

//! Maximum ROI to permit HDMI to be used.
#define MAX_WIDTH_FOR_HDMI  1920
#define MAX_HEIGHT_FOR_HDMI 1080

//! @anchor IMGPROC_TYPES
//! @name Image processing specifications for the hardware.
//! @{
static const uint32_t IMGPROC_TYPE_MASK = 0x0f;
static const uint32_t IMGPROC_TYPE_NONE = 0x00;
static const uint32_t IMGPROC_TYPE_MONO8 = 0x01;
static const uint32_t IMGPROC_TYPE_RGB24 = 0x02;
static const uint32_t IMGPROC_TYPE_RAW8  = 0x03;
static const uint32_t IMGPROC_TYPE_RAW16 = 0x04;
static const uint32_t IMGPROC_TYPE_JPEG  = 0x05;
static const uint32_t IMGPROC_TYPE_NV12  = 0x06;
static const uint32_t IMGPROC_TYPE_Y8    = 0x07;
static const uint32_t IMGPROC_ROW_WD_ALIGN = 0x10;
static const uint32_t IMGPROC_BIG_ENDIAN = 0x20;
static const uint32_t IMGPROC_USE_FPN    = 0x40;
//! @}

//! System or config files we know about on our file systems don't need the full Linux 4K length.
#define FI_PATH_MAX     512

// Directory for LUT files. Since it is based in our home folder,
// that will have to be added at runtime.
#define LUT_FILES_DIR       "%s/.config/Fastec/settings/"

// Format of the filename for a user lookup table
#define LUT_FILENAME_FMT    LUT_FILES_DIR "%u.lut"

// Name of temporary file while writing user lookup table.
#define LUT_TEMP_FILENAME   LUT_FILES_DIR "temp.lut"

//! The largest buffer we'll ever need in this file.
const int   MAX_BUFFER_LEN = 128;

//! @name Memory values and macros
//! @{
static const uint32_t QUADWORD_SIZE = 16;   //!< Number of bytes in a quad word.
static const uint32_t OCTWORD_SIZE  = 32;   //!< Number of bytes in a octal word.
static const uint32_t HEXWORD_SIZE  = 64;   //!< Number of bytes in a hex word.

static const int QUADWORD_SHIFT = 4;        //!< Shift to get in units of QUADWORDS.
static const int OCTWORD_SHIFT = 5;         //!< Shift to get in units of OCTWORDS.
static const int HEXWORD_SHIFT = 6;         //!< Shift to get in units of HEXWORDS.

// Define macros for OCTWORD (32-byte) alignment.
static const uint32_t QUADWORD_MASK = ((1 << QUADWORD_SHIFT) - 1);
static const uint32_t OCTWORD_MASK  = ((1 << OCTWORD_SHIFT) - 1);
static const uint32_t HEXWORD_MASK  = ((1 << HEXWORD_SHIFT) - 1);
//! @}

//! @name Time values and macros
//! @{
const uint32_t NSEC_PER_MSEC    = (1000 * 1000); //!< The number of nanoseconds in a millisecond, useful for struct timespec.
const uint32_t USEC_CONVERSION  = 1000000;      //! Conversion to use when going to uSec.
//! @}

// 4096 alignment in memory needed for DMA buffers.
const int   DMA_PAGE_ALIGN      = 4096;

//! The maximum number of images per folder. Any bigger than this
//! dramatically slows down save performance.
const uint32_t  MAX_FOLDER_COUNT    = 60000;

//! Threshold for deciding to split into subfolders.
const uint32_t  FOLDER_COUNT_THRESHOLD  = (MAX_FOLDER_COUNT + (MAX_FOLDER_COUNT >> 1));

//! Size of download chunks. Make the "chunks" as big as our biggest single-image file
//! plus "a skosh."
static const uint32_t DL_CHUNK_SIZE = (2560 * 2048 * 4);
#define MAX_METADATA_XFER_LEN DL_CHUNK_SIZE

//--------------------------------------------------------------------
// Some common typedefs, here to avoid header file interdependencies
//--------------------------------------------------------------------
// Use these during compilation to verify that calls are correct.
// typedef enum { bitmode_zero=0 } bitmode_t;
// typedef enum { fpnmode_zero=0 } fpnmode_t;
// Use these at runtime to avoid confusing the compiler.
typedef uint8_t bitmode_t;
typedef uint8_t fpnmode_t;

#endif /* !__CAMERA_SERVER_H__ */
