/*!
 * @file cs_network.h
 * @brief This file defines the Camera Server network (and associated) class(es).
 * @author Anthony L. Montiel
 * @date 2012-02-16
 *
 * Copyright (C) 2012 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_CS_NETWORK_H_
#define FASTEC_CS_NETWORK_H_

//--------------------------------------------------------------------
// Include files
//--------------------------------------------------------------------

#include <stdint.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/route.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <condition_variable>
#include <memory>
#include <mutex>

#include <fastec/gige_msg.h>

#include <cs_socket.h>

namespace CS_NETWORK
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------
#define CS_NET_SUCCESS              0
#define CS_NET_ERROR_FD             -1
#define CS_NET_ERROR_SOCKET         -1
#define CS_NET_ERROR_IOCTL          -2
#define CS_NET_ERROR_BIND           -2
#define CS_NET_ERROR_RECVFROM       -2
#define CS_NET_ERROR_SENDTO         -2
#define CS_NET_INVALID_PARAM        -3

//! Message types
enum
{
    MSG_TYPE_IPC = 1,               //!< A IPC message
    MSG_TYPE_GRPC = 2,              //!< A GRPC message
    MSG_TYPE_UDP = 3,               //!< A UDP message
    MSG_TYPE_FPGA = 4,              //!< An FPGA message
    MSG_TYPE_INTERNAL = 5           //!< An internal message to defer work
};

//! Internal messages
enum
{
    CS_INTERNAL_DEARM,              //!< Internal message to dearm.
    CS_INTERNAL_STATE,              //!< Internal message to change states.
    CS_INTERNAL_UPDATE,             //!< Internal message to perform a system update.
    CS_INTERNAL_WATCHDOG            //!< Internal message to tickle the watchdog.
};

const int LINUX_PATH_LEN    = 255;      //!< Must be bigger than UNIX_PATH_MAX define in un.h
const int NET_ADDR_LEN      = 20;       //!< Size enough to hold xxx.xxx.xxx.xxx
const int MAX_PACKET_SIZE   = 9000;     //!< Maximum size of a packet.
const int MAC_ADDR_LEN      = 6;        //!< Number of bytes in a MAC address.

//--------------------------------------------------------------------
// Type definitions
//--------------------------------------------------------------------

#pragma pack(1)
//! Define the format of an IPC message. Part of the fcp_message union
typedef struct CS_IPC_MESSAGE
{
    char    from[ LINUX_PATH_LEN + 1 ]; //!< Null terminated who is the message from.
} CS_IPC_MESSAGE;
#pragma pack(0)

#pragma pack(1)
//! Define the format of an UDP message. Part of fcp_mesage union.
typedef struct CS_UDP_MESSAGE
{
    char    from[ NET_ADDR_LEN + 1 ];   //!< Network address xxx.xxx.xxx.xxx,inet_ntoa run on it.
    uint16_t    from_port;          //!< Port of the client.
} CS_UDP_MESSAGE;
#pragma pack(0)

//! Define the structure of an internal message. Go ahead and pack just like the others.
typedef struct CS_INTERNAL
{
    uint16_t type;  //!< The type of internal message.
    uint16_t flags; //!< Flag argument.
    void  *p_arg;   //!< A pointer argument.
} CS_INTERNAL;

#pragma pack(1)
//! Define the format of a network message, ipc and udp and fcp_msg are sub types
typedef struct CS_NETWORK_MESSAGE
{
    unsigned char   msg_type;       //!< One of the MSG_TYPE_xxx enums.
    struct timespec rcv_time;       //!< When we received the message.
    union
    {
        CS_IPC_MESSAGE  ipc_msg;    //!< If this is a ipc message.
        CS_UDP_MESSAGE  udp_msg;    //!< If this is a udp message.
        uint32_t    fpga_val;       //!< If it's an fpga message.
        CS_INTERNAL internal;       //!< If it's an internal message.
    };

    fcp_msg message;            //!< The actual message including header
} CS_NETWORK_MESSAGE;
#pragma pack(0)

//! Define information about our network structure. This is stored for later when
//! a remote user wants to know.
typedef struct NETWORK_INFO
{
    char if_name[IFNAMSIZ];             //!< Network interface name i.e. eth0.
    unsigned char   mac_addr[MAC_ADDR_LEN];     //!< MAC address of the network, binary.
    struct in_addr  ip_addr;            //!< IP address as binary i.e. 0xffffffff=255.255.255.255
    struct in_addr  subnet_mask;        //!< Subnet mask for the network, binary
    struct in_addr  def_gateway;        //!< Default gateway, binary
} NETWORK_INFO;

} /* End of namespace CS_NETWORK */

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
//! @brief The CsNetwork class.
//--------------------------------------------------------------------

class CsNetwork
{
protected:
    static int addEnetConnection(uint32_t ip_cfg, uint32_t ip_addr = 0, uint32_t subnet_addr = 0, uint32_t gateway_addr = 0);
    static int addWifiConnection(uint32_t ip_cfg, uint32_t ap_cfg, uint32_t ip_addr = 0, uint32_t subnet_addr = 0, uint32_t gateway_addr = 0);
    static int toCIDR(const char *ip_addr);

    static bool connectionExists(const char *con_name);

    uint32_t    findGateway(CS_NETWORK::NETWORK_INFO *net_info);
    int     refreshIpAddr(CS_NETWORK::NETWORK_INFO *net_info);
    int     refreshSubnet(CS_NETWORK::NETWORK_INFO *net_info);

protected:
    std::unique_ptr<CsSocket> sock;       //!< The socket descriptor.

public:

    //! The various network types we support.
    enum NET_TYPE
    {
        ETHER,          //!< Ethernet, aka "wired network".
        WIFI,           //!< WIFI.
        LOCAL,          //!< Local loopback interface.
    };

    explicit CsNetwork();
    virtual ~CsNetwork() {}

    void close();

    //! @name Pure Virtual Methods
    //! @{

    /*!
     * @brief The method used to populate the 'msg' parameter
     *        with a message received over a particular network.
     *
     * @param p_msg The CS_NETWORK_MESSAGE structure to be filled in by the method.
     * @return The size of the message if successful, 0 if no message, <0 if error.
     */
    virtual int     getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg) = 0;

    /*!
     * @brief The method used to transmit the 'msg' parameter data
     *        over a particular network.
     *
     * @param p_msg The CS_NETWORK_MESSAGE structure to be filled in by the method.
     * @param num_bytes The number of bytes in the CS_NETWORK_MESSAGE.
     * @return The size of the message if successful, 0 if no message, <0 if error.
     */
    virtual int     sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes) = 0;

    //! Setupt the network specific object.
    virtual int     setup() = 0;
    //
    //! @}

    //! @name Inline Utility Methods
    //! @{
    int   getSocket()     { return ((sock.get() == nullptr) ? -1 : sock->getFD()); }
    bool  isValid()       { return ((sock.get() == nullptr) ? false : (sock->getFD() >= 0)); }
    //
    //! @}

    //! @name Virtual Methods For Network Specific Access
    //! @{
    virtual int     discoverNetwork();

    virtual uint32_t    getDefGateway();
    virtual uint32_t    getIpAddress();
    virtual void        getMacAddress(uint8_t *p_mac) const;
    virtual uint32_t    getSubnetMask();

    virtual uint32_t    getWifiDefGateway();
    virtual uint32_t    getWifiIpAddress();
    virtual uint32_t    getWifiSubnetMask();
    //
    //! @}

    static int      exportCameraName(const char *new_name);
    static bool     getActiveConnection(std::string& con_name, NET_TYPE type);
    static bool     getDevName(NET_TYPE type, std::string& dev_name);
    static int      getHostname(char *hostname_buf, int buflen);
    static int      getEth0MacAddress(uint8_t *p_mac);
    static bool     getPreferredDhcpMode(NET_TYPE type);
    static uint32_t getPreferredIpAddress(NET_TYPE type);
    static uint32_t getPreferredDefGateway(NET_TYPE type);
    static uint32_t getPreferredSubnetMask(NET_TYPE type);
    static bool     isEthConnected();
    static bool     isNfsMount();
    static int      startDhcpConnection(NET_TYPE type);
    static int      startStaticConnection(NET_TYPE type, uint32_t ip_addr = 0, uint32_t subnet_addr = 0, uint32_t gateway_addr = 0);
    static int      updateIpCfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

    static void     getSsidFromConf(char *ssid, int max_len);
    static uint32_t getWifiConfigFromConf();
    static void     getWifiPassphraseFromConf(char *pw, int max_len);
    static bool     getWifiScanResults(char *buffer, int buffsize);
    static int      getWifiSignalStrength();
    static uint32_t getWifiStatus();
    static bool     isWifiDevicePresent();
    static bool     isWifiConnected();
    static int      updateWifiIpcfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
    static int      updateWifiApCfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
    static int      wifiStartScan(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
};

//--------------------------------------------------------------------
//! @brief The IPC class.
//!
//! The Inter-Process Communications (IPC) class, allows for network-like
//! communications to other processes on the controller.
//--------------------------------------------------------------------
class CsIPC : public CsNetwork
{
private:
    static const char *CS_KEY_PATH;

public:
    CsIPC();
    virtual ~CsIPC()    { close(); }

    int     setup() override;
    void    close();

    int     getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg) override;
    int     sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes) override;
};

//--------------------------------------------------------------------
//! @brief The UDP class.
//--------------------------------------------------------------------
class CsUDP : public CsNetwork
{
public:
    CsUDP();
    virtual ~CsUDP()    { close(); }

    void close();

    virtual int     setup() override;

    virtual int     getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg) override;
    virtual int     sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes) override;
};

//--------------------------------------------------------------------
//! @brief The loopback class.
//--------------------------------------------------------------------
class CsLocal : public CsUDP
{
public:
    CsLocal();
    virtual ~CsLocal()  { close(); }

    void close();

    virtual int setup() override;

    virtual int discoverNetwork() override;
    virtual uint32_t    getDefGateway() override;
    virtual uint32_t    getIpAddress() override;
    virtual uint32_t    getSubnetMask() override;
};

//--------------------------------------------------------------------
//! @brief The FSP class.
//!
//! The Fastec Streaming Protocol (FSP) class is a TCP network. It is
//! used for streaming video and data off the controller.
//--------------------------------------------------------------------

class CsFSP : public CsNetwork
{
private:
    static std::mutex serial_stream;            //!< Protect against reentrancy.

    std::unique_ptr<CsSocket> listen_sock;
    bool dnf_flag;                              //!< Not used. Fragments aren't an issue with TCP.
    std::unique_ptr<uint8_t[]> test_buffer;     //!< An aligned pointer used for test messages.

    volatile bool   stream_pause;               //!< True to pause the stream.
    std::mutex      pause_mutex;                //!< Used to control changes in pause status.
    std::condition_variable pause_status;       //!< Condition variable we can block on.

    //! Incremented (wrapping) for each block sent.
    uint16_t fspBlockId;

    //! Frame number used as the "file name".
    int fspFrameNum;

    //! We need to be able to take 24-bits of packet ID for our streaming
    //! header. To do that, it is easiest if we create a union of a word
    //! and overlay that with 4 bytes.
    union
    {
        uint32_t word;
        uint8_t  byte[4];
    } fspPacketId;

private:
    //! @name FSP helper functions
    //! @{
    bool sendLeader(int32_t data_size, uint16_t payload_type);
    bool sendTrailer();
    bool sendData(const uint8_t *body_ptr, int32_t body_size, const uint8_t *head_ptr, int32_t head_size);
    //! @}

    void checkPause();

public:
    CsFSP();
    virtual ~CsFSP();

    int     setup() override;
    void    close();
    bool    waitForClient(uint32_t msecs);

    void    setBlockID(int block_id)  { fspBlockId = block_id; }
    void    setFrameNum(int frame_no) { fspFrameNum = frame_no; }

    // The streaming class doesn't do messages.
    int     getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg) override
    {
        (void)p_msg;
        return CS_NET_ERROR_FD;
    }
    int     sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes) override
    {
        (void)p_msg;
        (void)num_bytes;
        return CS_NET_ERROR_FD;
    }

    size_t  sendfile(const int file_fd, off_t offset, size_t data_size);
    int32_t streamData(const uint8_t *data_ptr, int32_t data_size, uint16_t payload_type, const uint8_t *head_ptr = nullptr, int32_t head_size = 0);
    int32_t testFire(int32_t packet_size);

    void    setDNF(bool set);
    bool    getDNF()    { return dnf_flag; }

    void pause();
    void resume();

    int getListenFD()
    {
        return ((listen_sock.get() == nullptr) ? -1 : listen_sock->getFD());
    }
    bool isListenValid()
    {
        return (listen_sock.get() == nullptr) ? false : (listen_sock->getFD() >= 0);
    }
};

#endif /* !FASTEC_CS_NETWORK_H_ */
