/*!
 * @file cs_utils.h
 * @brief Synopsis of module functionality
 *
 * This is the header file for utility function used within the CameraServer Daemon.
 *
 * Copyright 2010 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CS_UTILS_H
#define CS_UTILS_H

#include <stdint.h>
#include "constants.h"

#include <fastec/fas_thread.h>
#include <fastec/fas_log.h>

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Debug macros and defines, levels can be added to.
#define PDEBUG_VL       0x00000001  // just used for specific testing.
#define PDEBUG_LOW      0x00000002  // low noise level
#define PDEBUG_HIGH     0x00000004  // high noise level
#define PDEBUG_FUNCTRACE    0x00000008  // Trace function calls.
#define PDEBUG_GIGEWRITE    0x00000010  // Trace GigE register writes.
#define PDEBUG_GIGEREAD     0x00000020  // Trace GigE register writes.
#define PDEBUG_DOWNLOAD     0x00000040  // Trace download activities.
#define PDEBUG_SENSOR       0x00000080  // Trace sensor activities.
#define PDEBUG_FPGAWRITE    0x00000100  // Trace register writes.
#define PDEBUG_FPGALUTWRITE 0x00000200  // Trace writes to IPM LUTs.
#define PDEBUG_FPGAREAD     0x00001000  // Trace register writes.
#define PDEBUG_FPGACRCREAD  0x00002000  // Trace read of ICM CRC channels 0-11
#define PDEBUG_FPGATIMEREAD 0x00004000  // Trace read of time registers SBS_COUNTER & SBUS_COUNTER.
#define PDEBUG_FPGASPIWRITE 0x00010000  // Trace sensor register writes.
#define PDEBUG_FPGASPIREAD  0x00020000  // Trace sensor register reads.
#define PDEBUG_FPN          0x00040000  // Trace FPN operations.
#define PDEBUG_WSS_IF       0x00080000  // Trace Web Socket Server interfaces.
#define PDEBUG_TRACE PDEBUG_FUNCTRACE // For backward compatibility.

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

// main.cxx
extern bool debug_to_log;
extern int  cs_verbose;
extern bool sensor_test;
extern uint32_t user_ssd_bandwidth;

//--------------------------------------------------------------------
// Function prototypes
//--------------------------------------------------------------------

extern const char *get_home_path();
extern uint64_t get_current_usecs(void);
extern uint32_t calc_src_dma_size(uint32_t row_width, uint32_t num_rows, bitmode_t bit_mode);
extern bool     ipm_p1_is_busy(void);
extern uint32_t get_imgproc_type(uint32_t stream_type, bitmode_t bit_mode, bool is_color);
extern uint32_t get_processed_size(uint32_t imgproc_type, uint32_t width, uint32_t height);
extern uint32_t get_stream_dma_mod(uint32_t stream_type, bitmode_t bit_mode, bool is_color);
extern uint32_t get_tid(void);
extern uint64_t measuretime(int start, struct timespec *tv1_ptr);
extern void sbs_to_tm(uint32_t sbs, struct tm *tm_ptr);
extern void timespec_add_ns(struct timespec *a, uint64_t ns);

//--------------------------------------------------------------------
// Inline functions
//--------------------------------------------------------------------

/*!
 * @brief Since we can't reliably compare floating point numbers,
 *        we have this routine to determine if we are "close enough".
 *
 * @param x     The first number to compare.
 * @param y     The second number to compare.
 * @return True if x is pretty much equal to y, false otherwise.
 */
inline bool approx_equals(float x, float y)
{
    return (std::abs(x - y) < 1e-7f);
}

/*!
 * @brief Converts the passed in byte address or value to
 *    QWORD 16 byte aligned. The result is rounded up to
 *    the next QWORD that is on a 32-byte boundary
 *    i.e. 0x31 -> 0x4.
 *
 * @param byte_addr The byte address or byte value.
 * @return The qword address or value.
 */
inline uint32_t convert_to_QWORD(uint64_t byte_addr)
{
    uint64_t qword_addr64;

    // Get the byte address down to a QWORD value.
    qword_addr64 = byte_addr / QUADWORD_SIZE;

    // Check to see if we need to round up.
    if ((byte_addr % QUADWORD_SIZE) != 0)
        qword_addr64++;

#if 0
    // Now, round it up again to make sure it's a
    // multiple of 32 bytes.
    if ((qword_addr64 & 1) != 0)
        qword_addr64++;
#endif

    // Sanity check - we should never overflow a 32-bit value,
    // but let's make sure.
    if (qword_addr64 > ((((uint64_t)1) << 32) - 1))
    {
        fi_log(LOG_ERR, "%s: QWORD calculation is too big!", __FUNCTION__);
        qword_addr64 = 0;
    }

    // Return the value.
    return ((uint32_t)qword_addr64);
}

/*!
 * @brief Converts the passed in byte address or value to
 *    QWORD 16 byte aligned. The result is rounded up to
 *    the next even QWORD so it is on a 32-byte boundary,
 *    i.e. 0x31 -> 0x4, 0x41 -> 0x6.
 *
 * @param byte_addr The byte address or byte value.
 * @return The qword address or value.
 */
inline uint32_t convert_to_QWORD(uint32_t byte_addr)
{
    uint32_t qword_addr;

    // Get the byte address down to a QWORD value.
    qword_addr = byte_addr / QUADWORD_SIZE;

    // Check to see if we need to round up.
    if ((byte_addr % QUADWORD_SIZE) != 0)
        qword_addr++;

#if 0
    // Now, round it up again to make sure it's a
    // multiple of 32 bytes.
    if ((qword_addr & 1) != 0)
        qword_addr++;
#endif

    // Return the value.
    return qword_addr;
}

/*!
 * @brief Converts the passed in byte address or value to
 *    octal word 32 byte aligned. The result is rounded
 *    up, i.e. 0x32 -> 0x1.
 *
 * @param byte_addr The byte address or byte value
 * @return The octal address or value.
 */
inline uint32_t convert_to_OCTWORD(uint32_t byte_addr)
{
    uint32_t oword_addr;

    // Divide the byte value down.
    oword_addr = byte_addr / OCTWORD_SIZE;

    // Check to see if we need to round it up.
    if ((byte_addr % OCTWORD_SIZE) != 0)
        oword_addr++;

    // Return the value.
    return oword_addr;
}

/*!
 * @brief Converts the passed in byte address or value to
 *    octal word 32 byte aligned. The result is rounded
 *    up, i.e. 0x32 -> 0x1.
 *
 * @param byte_addr The byte address or byte value
 * @return The octal address or value.
 */
inline uint32_t convert_to_OCTWORD(uint64_t byte_addr)
{
    uint64_t oword_addr;

    // Divide the byte value down.
    oword_addr = byte_addr / OCTWORD_SIZE;

    // Check to see if we need to round it up.
    if ((byte_addr % OCTWORD_SIZE) != 0)
        oword_addr++;

    // Sanity check - we should never overflow a 32-bit value,
    // but let's make sure.
    if (oword_addr > ((((uint64_t)1) << 32) - 1))
    {
        fi_log(LOG_ERR, "%s: Calculation is too big!", __FUNCTION__);
        oword_addr = 0;
    }

    // Return the value.
    return ((uint32_t)oword_addr);
}

/*!
 * @brief Converts the passed in byte address or value to
 *    hex word 64 byte aligned. The result is rounded
 *    up, i.e. 0x65 -> 0x2.
 *
 * @param byte_addr The byte address or byte value
 * @return The hex address or value.
 */
inline uint32_t convert_to_HEXWORD(uint64_t byte_addr)
{
    uint64_t hword_addr;

    // Divide the byte value down.
    hword_addr = byte_addr / HEXWORD_SIZE;

    // Check to see if we need to round it up.
    if ((byte_addr % HEXWORD_SIZE) != 0)
        hword_addr++;

    // Sanity check - we should never overflow a 32-bit value,
    // but let's make sure.
    if (hword_addr > ((((uint64_t)1) << 32) - 1))
    {
        fi_log(LOG_ERR, "%s: Calculation is too big!", __FUNCTION__);
        hword_addr = 0;
    }

    // Return the value.
    return ((uint32_t)hword_addr);
}

//--------------------------------------------------------------------
// csAutoTrigger class declarations
//--------------------------------------------------------------------

class csAutoTrigger : public FAS_THREAD::Thread
{
public:
    csAutoTrigger();
    virtual ~csAutoTrigger();

    virtual void run();

    bool is_running()   { return thread_running; }

private:
    volatile bool   thread_running;     //!< Flag to indicate if we are running or not.
};


//--------------------------------------------------------------------
// cs_poll class declarations
//--------------------------------------------------------------------

class cs_poll : public FAS_THREAD::Thread
{
public:
    cs_poll();
    virtual ~cs_poll();

    virtual void run();

    bool is_running()   { return thread_running; }

private:
    volatile bool   thread_running;     //!< Flag to indicate if we are running or not.
};

//--------------------------------------------------------------------
// cs_watchdog class declarations
//--------------------------------------------------------------------

class cs_watchdog : public FAS_THREAD::Thread
{
public:
    cs_watchdog();
    virtual ~cs_watchdog();

    virtual void run();

    bool is_running()   { return thread_running; }

private:
    volatile bool   thread_running;     //!< Flag to indicate if we are running or not.
};

#endif // CS_UTILS_H

