/*!
 * @file cs_worker.h
 * @brief This file defines the CSWorker (and associated) class(es).
 * @author Anthony L. Montiel
 * @date 2012-02-16
 *
 * Copyright (C) 2012 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_CS_WORKER_H_
#define _FASTEC_CS_WORKER_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <fastec/fas_thread.h>
#include "cs_network.h"

//--------------------------------------------------------------------
//! @brief The csWorker class.
//!
//! This class manages the incoming communications sockets. By running
//! as a background thread, we can block waiting for messages to arrive.
//!
//! @sa csWorker::run()
//--------------------------------------------------------------------
class csWorker : public FAS_THREAD::Thread
{
public:
    csWorker();
    virtual ~csWorker();

    void run() override;
    void stop() override;

    virtual void reopen()           { reopenRequested = true; }

    // Because the worker holds the socket(s), we need
    // to expose some of the socket specific stuff here.
    int         discoverNetwork(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg = nullptr);

    uint32_t    getDefaultGateway(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg = nullptr);
    uint32_t    getIpAddress(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg = nullptr);
    uint32_t    getSubnetMask(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg = nullptr);
    void        getMacAddress(uint8_t *p_mac) const     { udp_socket.getMacAddress(p_mac); }
    uint32_t    getWifiDefaultGateway()                 { return udp_socket.getWifiDefGateway(); }
    uint32_t    getWifiIpAddress()                      { return udp_socket.getWifiIpAddress(); }
    uint32_t    getWifiSubnetMask()                     { return udp_socket.getWifiSubnetMask(); }

    int sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes);

private:
    static std::mutex send_mutex;   //!< Mutex to prevent re-entrant behavior in the sendMsg() method.

    bool    reopenRequested;
    void    doReopen();
    bool    shutdown_done;

    CsIPC  ipc_socket;             //!< UNIX domain socket.
    CsLocal   local_socket;           //!< Loopback socket.
    CsUDP  udp_socket;             //!< UDP socket.

    void    cleanup();
    int     initializeReadSet(fd_set& readset, std::vector<CsNetwork *>& activeSockets);
    void    initializeNetwork();
    void    doNetworkOverride();
    bool    msgIsUDP(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
};

#endif /* !_FASTEC_CS_WORKER_H_ */
