/*!
 * @file ct_file.h
 * @brief Format of the calibration correction and gain table files.
 *
 * Copyright 2012-2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef __CT_FILE_H__
#define __CT_FILE_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <fastec/fi_sensor_types.h>
#include <sensor.h>

namespace CT_GT_FILE
{
// The FIXED_FRACTIONAL is used to adjust a number to the proper
// fixed-point format: fixed_format = (number << FIXED_FRACTIONAL).
//
// The implied integer part is represented by how many bits are left
// over in the byte (plus "1.0"). So, with a value of '8', this implies
// the number will be a maximum of "'1'.xxxxxxxx". With a value of '7',
// the implication is that the number will be a max of "'2'.xxxxxxx".
//
// Because the gain is represented as a byte, the maximum value is
// 0xFF, regardless of the fixed-point format. Note that this is
// the "adjust" value - there is alwasy an implied "+1.0".
const int FIXED_FRACTIONAL = 7;

// Sensor ID in case we don't know what it is in an old cal file.
const uint32_t LINCE_SENSOR_ANY     = SENSOR_ID('A', 'L', '5', 'x');

// Paths for the factory and user specified FPN correction tables.
const char FACTORY_GAIN_FILE_RO[]   = ".config/Fastec/calibration/factory_gain";
const char FACTORY_BLKREF_FILE_RO[] = ".config/Fastec/calibration/factory_blkref";
const char FACTORY_BLKREF_FILE_RW[] = ".config/Fastec/scratch/factory_blkref";
const char USER_BLKREF_FILE[]       = ".config/Fastec/scratch/user_blkref";
const char FACTORY_BLKREF_SCRIPT[]  = "cp ~/.config/Fastec/scratch/factory_blkref ~/.config/Fastec/calibration/.";
const char BADPIXEL_MAP_SCRIPT[]    = "mv ~/.config/Fastec/scratch/factory_blkref ~/.config/Fastec/scratch/badpixel.map";

// Size of the sensor-specific parameter area.
const int SENSOR_PARAM_BYTES = 16;

// Sensor ID. The format is "part_no-serial_no\0".
const int SENSOR_ID_LEN = (SENSOR::FI_PART_NO_LEN + SENSOR::FI_SERIALNO_LEN + 2);

// Our old "magic" number, for version 1 files.
const uint32_t CT_MAGIC_1 = 1;

// Our version 2 files, when we added the Lince5M sensor options information.
const uint32_t CT_MAGIC_2 = 2;

// The version 3 files contain the unique identifier to tie the correction & gain tables
// to the sensor.
const uint32_t CT_MAGIC = 3;

// Structures etc to support correction tables
typedef struct
{
    int x;
    int y;
} CT_POINT;

typedef struct
{
    CT_POINT    origin;
    int     width;
    int     height;
} CT_ROI;

// The header information in version 1 correction table files.
typedef struct CT_FILE_HDR_1
{
    uint32_t    magic;          // Our magic number.
    uint32_t    length;         // How many bytes of data are in the correction table.
    uint32_t    dac_vref;       // The DACVref value associated with this table. If bits 16:31 are non-zero, they are sensor gain.
    CT_ROI  roi;                // NOTE: Width is 32-bit aligned!
//
// Correction table data immediately follows in the file.
//
} CT_FILE_HDR_1;

// The header information in version 2 correction table files.
typedef struct CT_FILE_HDR_2
{
    uint32_t    magic;          // Our magic number.
    uint32_t    length;         // How many bytes of data are in the correction table.
    uint32_t    sensor_type;    // Sensor type, if known.
    uint32_t    sensor_params[SENSOR_PARAM_BYTES / sizeof(uint32_t)]; // Sensor-specific parameters.
    CT_ROI  roi;                // NOTE: Width is 32-bit aligned!
//
// Correction table data immediately follows in the file.
//
} CT_FILE_HDR_2;

// The header information in version 3 correction table files.
typedef struct CT_FILE_HDR_3
{
    uint32_t    magic;          // Our magic number.
    uint32_t    length;         // How many bytes of data are in the correction table.
    uint32_t    sensor_type;    // Sensor type, if known.
    uint32_t    sensor_params[SENSOR_PARAM_BYTES / sizeof(uint32_t)]; // Sensor-specific parameters.

    CT_ROI  roi;                // NOTE: Width is 32-bit aligned!

    char sensor_id[SENSOR_ID_LEN];  // Associate the table with a particular sensor.
//
// Correction table data immediately follows in the file.
//
} CT_FILE_HDR;

typedef CT_FILE_HDR GT_FILE_HDR;

} /* End of namespace CT_GT_FILE */

#endif /* __CT_FILE_H__ */

