/*!
 * @file fi_internal.h
 *
 * @brief A class that manages access to the Fastec Internal
 *        Proprietary class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_INTERNAL_H
#define FI_INTERNAL_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <dlfcn.h>
#include <memory>
#include <mutex>
#include <string>

#include <fastec/fas_internal.h>

//! The FiInternal class is a singleton object to control access to the
//! Fastec Internal Proprietary library.
class FiInternal
{
public:
    // Define the function "signature".
    typedef uint32_t (* GET_PACKED_SIZE_FN)(uint32_t width, uint32_t height, int bitsPerPixel);
    typedef FAS_INTERNAL::FIP_HANDLE(* GET_HANDLE_FN)(const std::string& streamName, const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo);
    typedef FAS_INTERNAL::FIP_STATUS(* CLOSE_PACK_STREAMER_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* IS_PACK_STREAMER_FINISHED_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* ADD_TIFF_IMAGE_DATA_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, const uint8_t *imagePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* ADD_PLACEHOLDER_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* REPLACE_FRAME_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t seqNumber, const uint8_t *imagePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* GET_PACK_STATISTICS_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t& frameCount, uint32_t& lastFrame, uint32_t& encodedFrames);
    typedef FAS_INTERNAL::FIP_STATUS(* CONVERT_TIFF_FN)(uint8_t *imageInPtr, uint8_t *imageOutPtr, int frameSeqNum, uint32_t width, uint32_t height, int bitsPerPixel);

    typedef FAS_INTERNAL::FIP_HANDLE(* MA_GET_HANDLE_FN)(const FAS_INTERNAL::FIP_MA_CONFIGURATION& configInfo);
    typedef FAS_INTERNAL::FIP_STATUS(* MA_CLOSE_FN)(FAS_INTERNAL::FIP_HANDLE maHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* MA_ADD_FRAME_FN)(FAS_INTERNAL::FIP_HANDLE maHandle, const uint8_t *framePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* MA_GET_FRAME_FN)(FAS_INTERNAL::FIP_HANDLE maHandle, const uint8_t *amplifiedFrame);
    typedef FAS_INTERNAL::FIP_STATUS(* MA_GET_REF_FRAME_FN)(FAS_INTERNAL::FIP_HANDLE maHandle, const uint8_t *referenceFrame);
    typedef bool (* MA_IS_AVAIL_FN)(FAS_INTERNAL::FIP_HANDLE maHandle);
    typedef int (* MA_GET_WINDOW_SIZE_FN)(FAS_INTERNAL::FIP_HANDLE maHandle);

public:
    ~FiInternal();

    // FiInternal should not be cloneable.
    FiInternal(FiInternal& other) = delete;

    // FiInternal should not be assignable.
    void operator=(const FiInternal&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static FiInternal *factory();

    GET_PACKED_SIZE_FN getPackedSizeFuncPtr();
    GET_HANDLE_FN getHandleFuncPtr();
    CLOSE_PACK_STREAMER_FN getClosePackFuncPtr();
    IS_PACK_STREAMER_FINISHED_FN geIsFinishedFuncPtr();
    ADD_TIFF_IMAGE_DATA_FN getAddTiffFuncPtr();
    ADD_PLACEHOLDER_FN getAddPlaceholderFuncPtr();
    REPLACE_FRAME_FN getReplaceFrameFuncPtr();
    GET_PACK_STATISTICS_FN getGetStatisticsFuncPtr();
    CONVERT_TIFF_FN getConvertTiffFuncPtr();

    MA_GET_HANDLE_FN getMaHandleFuncPtr();
    MA_CLOSE_FN getMaCloseFuncPtr();
    MA_ADD_FRAME_FN getMaAddFrameFuncPtr();
    MA_GET_FRAME_FN getMaFrameFuncPtr();
    MA_GET_REF_FRAME_FN getMaRefFrameFuncPtr();
    MA_IS_AVAIL_FN getMaIsAvailFuncPtr();
    MA_GET_WINDOW_SIZE_FN getMaWinSizeFuncPtr();

private:
    FiInternal();

private:
    static std::unique_ptr<FiInternal> instance;
    static std::mutex instanceMutex;

    static std::string LIBRARY_NAME;
    static std::string GET_PACKED_SIZE;
    static std::string GET_HANDLE;
    static std::string CLOSE_PACK_STREAMER;
    static std::string IS_PACK_STREAMER_FINISHED;
    static std::string ADD_TIFF_IMAGE_DATA;
    static std::string ADD_PLACEHOLDER;
    static std::string REPLACE_FRAME;
    static std::string GET_PACK_STATISTICS;
    static std::string CONVERT_TIFF;

    static std::string MA_GET_HANDLE;
    static std::string MA_CLOSE;
    static std::string MA_ADD_FRAME;
    static std::string MA_GET_FRAME;
    static std::string MA_GET_REF_FRAME;
    static std::string MA_IS_AVAIL;
    static std::string MA_GET_WINDOW_SIZE;

    static void *fasLib;
};

#endif // FI_INTERNAL_H

