/*!
 * @file formatter.h
 * @brief This file defines the background formatter class.
 *
 * Copyright (C) 2020 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FORMATTER_H_
#define _FORMATTER_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <fastec/fas_thread.h>

//--------------------------------------------------------------------
//! @brief The background formatter class follows a singleton pattern.
//!
//! This class is a threaded object that does the actual work of
//! formatting a local device via the "format_and_build" script.
//--------------------------------------------------------------------
class formatter : public FAS_THREAD::Thread
{
public:
    virtual ~formatter();

    // A singleton is not cloneable.
    formatter(formatter& other) = delete;

    // A singleton is not assignable.
    void operator=(const formatter&) = delete;

    virtual void run();

    static bool isSsdFormatInProgress();
    static bool formatStart(uint32_t media_sel, uint32_t fs_type, const char *dev_path, const char *label);
    static bool formatDone(int& status);

private:
    // As a singleton, the constructor is not public.
    formatter();

    bool initiate_format();
    bool check_format_status();
    bool reread_id_block(uint8_t *buf);
    const char *fs_name(uint32_t fs_type);

    void cleanup();

private:
    volatile bool thread_running;   //!< Flag to indicate if we are running or not.

    uint32_t fs_type;   //!< File system type
    int      status;    //!< Status of format operation.

    static std::unique_ptr<formatter> fmt_instance;
    static uint32_t format_process_media_sel;
    static std::string fmt_block_dev;
    static std::string volume_label;
    static int format_process_info_pid;
    static FILE *format_process_info_fp;

    static const char FAT32_FORMAT[];
    static const char EXFAT_FORMAT[];
    static const char EXT4_FORMAT[];
    static const char FILR_FORMAT[];
    static const char NTFS_FORMAT[];
    static const char EXEC_FILE[];
};

#endif /* !_FORMATTER_H_ */

