/*!
 * @file fpga.h
 * @brief This is the header file for the FPGA class.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_FPGA_H_
#define FASTEC_FPGA_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <atomic>
#include <memory>
#include <mutex>

#include <fastec/fas_thread.h>
#include <fastec/gige_msg.h>

#include "constants.h"
#include "cs_utils.h"

namespace FI_FPGA
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

enum FPGA_TRACE
{
    FT_FUNC     = 0x00000010,
    FT_FPN      = 0x00000020,
    FT_ICM      = 0x00000040,
    FT_IPM      = 0x00000080,
    FT_LTR      = 0x00000100,
    FT_SENSOR   = 0x00000200,
    FT_TIME     = 0x00000400,
    FT_REG      = 0x00000800,
    FT_I2C      = 0x00001000,
    FT_INT      = 0x00002000,
    FT_FUNC_NOISY   = 0x00004000,
    FT_MUTEX_COUNT  = 0x00008000,
    FT_ICM_NOISY    = 0x00010000
};

// Number of microseconds the 1PPS signal can be late
// before we declare a fault.
const uint32_t TIME_SYNC_LATE_LIMIT     = 20;

// Strictly a guess as to what we should adjust our tuning
// word so that our clock is slightly fast.
const uint32_t SYNC_IN_SLAVE_TUNING     = 256;

// Strictly a guess as to what we should adjust our tuning
// word so that our clock is slightly slow.
const uint32_t SLAVE_TUNING_KLUDGE  = 1000;

// Initialization value for LAST_FRAME_WRITTEN register.
const uint32_t LAST_FRAME_INIT  = 0x1fffffff;

// Define our blink rates for the LED.
const int LED_OFF       = 0;
const int LED_SOLID_ON      = 1;
const int LED_FAST_BLINK    = 2;
const int LED_SLOW_BLINK    = 3;

// Bytes per pixel in correction table.
const uint32_t FPGA_CORR_TABLE_BPP  = 1;

// Internal clock used by the FPGA, in hertz.
const int FPGA_CLOCK_SPEED      = 125000000;

// Internal clock used by the FPGA, in microseconds.
const int FPGA_CLOCK_SPEED_IN_USEC  = 125;

//! Enumeration defining the various I2C busses we support.
typedef enum
{
    I2C_MAIN,           //!< Talking to the main board I2C bus.
    I2C_IO,             //!< Talking to the IO board I2C bus.
    I2C_SENSOR          //!< Talking to the sensor board I2C bus.
} I2C_BUS_ID;

} /* End of namespace FI_FPGA */

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------
namespace CS_NETWORK
{
struct CS_NETWORK_MESSAGE;
}
class ltr_worker;

//--------------------------------------------------------------------
// FiFPGA class declarations
//--------------------------------------------------------------------

/*!
 * @brief The FiFPGA class provides the interface to the FPGA register
 *        space. This class is a singleton object.
 */
class FiFPGA
{
private:
    static std::unique_ptr<FiFPGA> instance;
    static std::mutex instance_mutex;

    bool        initialized;    //!< True when the FPGA object is initialized.
    uint32_t    traceControl;   //!< The FPGA class trace control word.
    std::mutex  class_mutex;    //!< Protect access to the class.

    uint32_t    fpga_ver;       //!< FPGA version.
    bool        wd_timeout;     //!< Flag indicating if we had a watchdog timeout on boot.
    uint32_t    shadow_irq;     //!< A shadow copy of the IRQ mask register.

    bool        low_light_enabled;  //!< TRUE if low light mode is enabled.
    uint32_t    low_light_shutter;  //!< Shutter speed to use for low light mode. The value "0" uses normal shutter speed.
    uint32_t    fpga_shutter;       //!< Shutter speed we're currently set to.

    //! @name The Brightness, Contrast and Gain values for adjusting the image.
    //!
    //! These values are all adjustable by the user to "tweak" the image, as opposed
    //! to the calculated values for white balance.
    //!
    //! @{
    uint32_t    brightness;     //!< Brightness setting - 0-200 centered around 100.
    uint32_t    contrast;       //!< Contrast setting - 0-200 centered around 100.
    uint32_t    gamma;          //!< The gamma setting 0-500 centered around 220.
    uint32_t    default_gamma;  //!< The user selected default gamma.

    uint16_t    red_gain;       //!< Red value of RGB gain.
    uint16_t    grn_gain;       //!< Green value of RGB gain.
    uint16_t    blu_gain;       //!< Blue value of RGB gain.
    //
    //! @}

    //! @name The matrix coefficients for the red, green, blue channels.
    //!
    //! These are the entries for the Color Correction Matrix (CCM) and are loaded
    //! into the FPGA for use by the IPM (image processing module) to generate color
    //! images.
    //!
    //! @{
    uint32_t    matrix_coeff1_red;
    uint32_t    matrix_coeff2_red;
    uint32_t    matrix_coeff3_red;
    uint32_t    matrix_coeff1_grn;
    uint32_t    matrix_coeff2_grn;
    uint32_t    matrix_coeff3_grn;
    uint32_t    matrix_coeff1_blu;
    uint32_t    matrix_coeff2_blu;
    uint32_t    matrix_coeff3_blu;
    //
    //! @}

    //! @name The white balance gains for the red, green, blue channels.
    //! @{
    uint32_t    wb_red;
    uint32_t    wb_green;
    uint32_t    wb_blue;
    //
    //! @}

    //! @name The dc_offsets for the red, green, blue channels.
    //! @{
    uint32_t    dc_offset_red;
    uint32_t    dc_offset_grn;
    uint32_t    dc_offset_blu;
    //
    //! @}

    //! @name LTR related information.
    //! @{
    bool        ltr_mode_selected;          //!< True if LTR is the user's choice or the license-forced choice, false otherwise.
    uint32_t    ltr_save_error;             //!< Error status for LTR mode. So far, only ENOSPC.
    std::unique_ptr<ltr_worker> ltr_thread; //!< Our LTR working thread.
    //
    //! @}

private:
    FiFPGA();

    //! @name General Support Functions
    //! @{
    bool    calibrateArm();
    bool    calibrateDeArm();
    void    dumpOneReg(void *, uint32_t, const char *);
    uint32_t imPeek(uint64_t peek_addr);
    void    imPoke(uint64_t poke_addr, uint32_t poke_value);
    bool    regReadWait(uint32_t status_reg, uint32_t mask, uint32_t ok_value, uint32_t reg_addr, uint32_t *value_ptr);
    bool    startupStopAll();
    uint32_t syncinGetIrigTiming();
    bool    trigoutSetPulseWidth(uint32_t pulse_usecs);
    int     waitForBit(uint32_t reg_addr, uint32_t bit, bool set, int max_retries, int sleep_usecs, uint32_t *value_ptr = nullptr);
    //
    //! @}

    //! @name ICM Control Functions
    //! @{
    void    icmAbortLrOnOverflow();
    void    icmEmptyEventFIFO();
    void    icmEndCaptureEventHappened(uint32_t frame_info, uint64_t last_frame_addr, bool underflow);
    void    icmContinuationEventHappened(uint32_t frame_info);
    int     icmEventFIFOCount();
    int     icmEventHappened();
    void    icmTriggerEventHappened(uint32_t frame_info, const uint32_t *extra_info);
    //
    //! @}

    //! @name IPM Control Functions
    //! @{
    uint32_t startP1Frame(bitmode_t bit_mode, uint32_t width, uint32_t height, uint32_t src_addr, uint32_t dest_addr, uint32_t image_type);
    //
    //! @}

    //! @name Per-Frame Metadata Extraction (PME) support.
    //! @{
    uint32_t pmeSetup(uint16_t part_num, uint32_t start_frame, uint32_t end_frame, uint32_t im_buf_addr);
    //
    //! @}

    //! @name External IO Mux support.
    //! @{
    void extioMuxDisable(int channel, int func);
    void extioMuxEnable(int channel, int func);
    //
    //! @}

public:
    static const uint32_t HISTOGRAM_RAW_SIZE    = 0x1000;   //!< Size of a raw histogram.
    static const uint32_t HISTOGRAM_RGB_SIZE    = 0x0C00;   //!< Size of a rgb histogram 1/3 red, then green, then blue.

public:
    virtual ~FiFPGA();

    // Singetons are not coneable.
    FiFPGA(FiFPGA& other) = delete;

    // Singletons are not assignable.
    void operator=(const FiFPGA&) = delete;

    // The method to get a pointer to the singleton object.
    static FiFPGA *getInstance();

    //! @name General Support Functions
    //! @{
    bool interfaceInit();
    void shutdown();
    void setup();
    void startupInit();

    constexpr uint32_t  getFpgaVersion()    { return fpga_ver; }
    constexpr bool      isInitialized()     { return initialized; }
    //
    //! @}

    void dump_regs();

    //! @name Image Capture Module (ICM) methods.
    //! @{
    void eventHappened(uint32_t fpga_val);
    uint32_t irqDisable(uint32_t irq_event);
    uint32_t irqEnable(uint32_t irq_event);

    bool icmArm(bool continuous_mode);
    bool icmDearm();
    bool icmDisable();
    bool icmHwTriggerEnable();
    bool icmSwTriggerFire();

    bool icmIsArmed();
    bool icmIsPipelineIdle(uint32_t *p_rtn = nullptr);
    bool icmIsPipelineReady(uint32_t *p_rtn = nullptr);

    uint64_t icmCurrentFrameAddr();
    uint32_t icmPostTriggerCount();

    constexpr bool icmIsLowLightEnabled()           { return low_light_enabled; }
    constexpr uint32_t icmLowLightShutterSpeed()    { return low_light_shutter; }
    constexpr uint32_t icmShutterSpeed()            { return fpga_shutter; }


    bool icmCalibrateCapture(uint32_t num_frames);
    uint32_t icmCaptureFrameCount();
    int icmCaptureFifoCount();
    bool icmGetCaptureBuffer(uint64_t *buf_start_addr, uint32_t *buf_size, uint32_t *buf_cnt);
    bool icmPipelineDisable();
    void icmResetCaptureFIFO();
    bool icmSensorReadoutEnable(bool enable);
    bool icmSetCaptureBuffer(uint64_t buf_start_addr);
    bool icmSetFrameRate(float frame_rate, uint32_t clock_speed = 0);
    bool icmSetLiveArgs(bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed);
    bool icmSetLowLightShutterSpeed(bool enable, uint32_t shutter_speed);
    bool icmSetMetadataTag(uint32_t tag);
    bool icmSetPostTriggerFrames(uint32_t num_frames);
    bool icmSetShutterSpeed(uint32_t shutter_speed, uint32_t clks_per_usec = FI_FPGA::FPGA_CLOCK_SPEED_IN_USEC);
    bool icmSetToLiveVideo();
    bool icmSetULVBuffers(uint64_t buf_start_addr, uint32_t buf_length, int num_bufs);
    bool icmStopCapture(bool triggered);
    bool icmStartup(bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed);
    //
    //! @}

    //! @name Image Processing Module (IPM) methods.
    //! @{
    void ipmInit();
    bool ipmIsIdle();
    void ipmReset();
    bool ipmLiveStreamInit(uint16_t true_startx, uint16_t true_starty, uint16_t width, uint16_t height, bitmode_t sensor_bit_mode, fpnmode_t fpn_mode);

    uint32_t ipmLiveFrameProcess(uint32_t image_type, uint32_t src_addr, uint32_t dest_addr);
    uint32_t ipmFrameProcess(uint32_t image_type, uint32_t src_addr, uint32_t dest_addr);

    static int  ipmRgbHistoSetRoiX(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t value);
    static int  ipmRgbHistoSetRoiY(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t value);

    bool ipmIsRgbHistoIdle();
    void ipmRgbHistoReset();
    bool ipmRgbHistoGenerate(uint8_t channel);
    bool ipmRgbHistogramRead(uint8_t *buffer);
    uint32_t ipmRgbHistoPeak();
    uint32_t ipmRgbHistoSumX_Red();
    uint32_t ipmRgbHistoSumX_Green();
    uint32_t ipmRgbHistoSumX_Blue();
    uint32_t ipmRgbHistoSumXSqrd_RedH();
    uint32_t ipmRgbHistoSumXSqrd_RedL();
    uint32_t ipmRgbHistoSumXSqrd_GreenH();
    uint32_t ipmRgbHistoSumXSqrd_GreenL();
    uint32_t ipmRgbHistoSumXSqrd_BlueH();
    uint32_t ipmRgbHistoSumXSqrd_BlueL();

    constexpr uint16_t ipmGainRed()                     { return red_gain; }
    constexpr uint16_t ipmGainGreen()                   { return grn_gain; }
    constexpr uint16_t ipmGainBlue()                    { return blu_gain; }
    constexpr uint32_t ipmBrightness()                  { return brightness; }
    constexpr uint32_t ipmContrast()                    { return contrast; }
    constexpr uint32_t ipmDefaultGamma()                { return default_gamma; }
    constexpr uint32_t ipmGamma()                       { return gamma; }
    constexpr void ipmSetGainRed(uint16_t new_red)      { red_gain = new_red; }
    constexpr void ipmSetGainGreen(uint16_t new_green)  { grn_gain = new_green; }
    constexpr void ipmSetGainBlue(uint16_t new_blue)    { blu_gain = new_blue; }
    constexpr void ipmSetBrightness(uint32_t new_br)    { brightness = new_br; }
    constexpr void ipmSetContrast(uint32_t new_cont)    { contrast = new_cont; }
    constexpr void ipmSetGamma(uint32_t new_gamma)      { gamma = new_gamma; }
    void ipmSetDefaultGamma(uint32_t new_default_gamma);

    void ipmSendLutTableToFpga(uint32_t gain_channel);
    void ipmSetLutValues(uint32_t brightness, uint32_t contrast, uint32_t gamma, uint32_t default_gamma);
    bool ipmLoadUserLut(unsigned lut_number);

    constexpr uint32_t ipmMatrixCoeff1Red()         { return matrix_coeff1_red; }
    constexpr uint32_t ipmMatrixCoeff2Red()         { return matrix_coeff2_red; }
    constexpr uint32_t ipmMatrixCoeff3Red()         { return matrix_coeff3_red; }
    constexpr uint32_t ipmMatrixCoeff1Green()       { return matrix_coeff1_grn; }
    constexpr uint32_t ipmMatrixCoeff2Green()       { return matrix_coeff2_grn; }
    constexpr uint32_t ipmMatrixCoeff3Green()       { return matrix_coeff3_grn; }
    constexpr uint32_t ipmMatrixCoeff1Blue()        { return matrix_coeff1_blu; }
    constexpr uint32_t ipmMatrixCoeff2Blue()        { return matrix_coeff2_blu; }
    constexpr uint32_t ipmMatrixCoeff3Blue()        { return matrix_coeff3_blu; }
    bool ipmSetMatrixCoeff1Red(uint32_t value);
    bool ipmSetMatrixCoeff2Red(uint32_t value);
    bool ipmSetMatrixCoeff3Red(uint32_t value);
    bool ipmSetMatrixCoeff1Green(uint32_t value);
    bool ipmSetMatrixCoeff2Green(uint32_t value);
    bool ipmSetMatrixCoeff3Green(uint32_t value);
    bool ipmSetMatrixCoeff1Blue(uint32_t value);
    bool ipmSetMatrixCoeff2Blue(uint32_t value);
    bool ipmSetMatrixCoeff3Blue(uint32_t value);

    constexpr uint32_t ipmWbRed()               { return wb_red; }
    constexpr uint32_t ipmWbGreen()             { return wb_green; }
    constexpr uint32_t ipmWbBlue()              { return wb_blue; }
    bool ipmSetWbRed(uint32_t value);
    bool ipmSetWbGreen(uint32_t value);
    bool ipmSetWbBlue(uint32_t value);

    constexpr uint32_t ipmDcOffsetRed()         { return dc_offset_red; }
    constexpr uint32_t ipmDcOffsetGreen()       { return dc_offset_grn; }
    constexpr uint32_t ipmDcOffsetBlue()        { return dc_offset_blu; }
    bool ipmSetDcOffsetRed(uint32_t value);
    bool ipmSetDcOffsetGreen(uint32_t value);
    bool ipmSetDcOffsetBlue(uint32_t value);

    int ipmCtUpdate(uint32_t frame_size, uint64_t ct_addr, uint64_t gt_addr);
    //
    //! @}

    //! @name Register Access Functions.
    //! @{
    bool regModify(uint32_t reg_addr, uint32_t value, uint32_t mask);
    bool regRead(uint32_t reg_addr, uint32_t *value_ptr);
    bool regWrite(uint32_t reg_addr, uint32_t value);
    //
    //! @}

    //! @name I2C helper methods.
    //! @{
    bool i2cRead(uint8_t dev_addr, uint32_t cmd_data, uint8_t *reg_data, int bus = FI_FPGA::I2C_MAIN, int byte_count = 1, bool readonly = false, int cmd_count = 1);
    bool i2cWrite(uint8_t dev_addr, uint32_t cmd_data, uint32_t reg_data, int bus = FI_FPGA::I2C_MAIN, int byte_count = 1, int cmd_count = 1);
    //
    //! @}

    //! @name Image Handling methods.
    //! @{
    int imageDownload(uint32_t rd_dma_mod, uint32_t rd_dma_dims, uint64_t src_addr, uint64_t dst_addr, uint32_t src_size, uint32_t dst_size = 0);
    int imageUpload(uint32_t dma_mod, uint64_t src_addr, uint64_t dst_addr, uint32_t dma_size);

    uint32_t imageMemRead(uint32_t rd_dma_mod, uint32_t frame_dims, uint64_t src_addr, uint8_t *buf, uint32_t buf_size, uint32_t img_size = 0);
    uint32_t imageMemWrite(uint64_t dest_addr, const uint8_t *buf, uint32_t buf_size, bool do_swap);
    //
    //! @}

    //! @name LED Support methods.
    //! @{
    void ledSetBlink(int command);
    //
    //! @}

    //! @name Shutter methods.
    //! @{
    bool isShutterOpen();
    void shutterClose(bool close);
    //
    //! @}

    //! @name External I/O methods.
    //! @{
    bool isExtio_5V();
    void extioSetVoltage(bool use_5V);
    uint32_t extioGetActivity();
    //
    //! @}

    //! @name Arm I/O methods
    //! @{
    void armInInitialize();
    void armInPause();
    void armInResume();
    static int  armInEnable(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t do_enable);
    static int  armInPolarityUpdate(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

    bool armOutSetPulseWidth(uint32_t period_ms);
    static int  armOutConfigure(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t syncout_settings);
    //
    //! @}

    //! @name Sync I/O methods.
    //! @{
    void syncInDelay(uint32_t delay_amount);
    bool syncInEnable(uint32_t type);
    uint32_t syncInGetTiming();
    static int syncInSetPolarity(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t rising_edge);

    bool syncOutConfigure(uint32_t syncout_settings);
    bool syncOutSetPulseWidth(uint32_t dc_percent);
    //
    //! @}

    //! @name Trigger I/O methods.
    //! @{
    // Configure the trigger output.
    static int  triggerOutConfigure(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t trigout_settings);

    // Configure the external trigger.
    static int  externalTriggerEnable(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t enable);
    static int  externalTriggerSetPolarity(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t rising_edge);
    //
    //! @}

    //! @name Time support methods.
    //! @{
    void timeInit();
    void timeStop();

    uint32_t timeReadSBS();
    uint32_t timeReadSBUS();

    uint32_t irigStatus();
    static int  irigConfigure(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t control_value);
    //
    //! @}

    //! @name Long Record (LR or LTR, Long Time Record) support methods.
    //! @{
    uint32_t ltrCalcFrameRate(uint32_t width, uint32_t height, bitmode_t bit_mode);
    void ltrComplete();
    void ltrEnable(bool do_enable);
    void ltrEnableROC(uint32_t mode);
    void ltrFasFireDisable();
    bool ltrFasFireEnabled();
    void ltrInit();
    uint32_t ltrMode();
    void ltrModeSelect(bool do_ltr);
    bool ltrReadImFIFO(uint8_t *imfifo_buffer, int byte_count);
    uint32_t ltrSectorCount();

    constexpr void ltrSetSaveError(uint32_t status)     { ltr_save_error = status; }
    constexpr uint32_t ltrSaveError()                   { return ltr_save_error; }
    //
    //! @}

    //! @name Per-frame Metadata (PFMD) Extraction (PME) support.
    //! @{
    constexpr uint32_t pmeSizeOfPFMD()                  { return FRAME_METADATA_HW_BYTES; }

    void pmeAbort();
    bool pmeIsDone();
    uint32_t pmeRead(uint16_t part_num, uint32_t start_frame, uint32_t end_frame, uint8_t *buf, uint32_t bufsize, bool (*abort_fn)(void *), void *abort_arg);
    uint32_t pmeStart(uint16_t part_num, uint32_t start_frame, uint32_t end_frame, uint32_t im_buf_addr, uint32_t im_buf_size);
    int pmeWaitDone(uint32_t frames, bool (*abort_fn)(void *), void *abort_arg);
    //
    //! @}

    //! @name Watchdog support methods.
    //! @{
    void watchdog();
    constexpr bool watchdogTimeout()           { return wd_timeout; }
    //
    //! @}
};

#endif /* !FASTEC_FPGA_H_ */
