/*!
 * @file fpga_io.h
 * @brief This file defines interface to the FPGA I/O class, fpgaIOWorker.
 * @author Anthony L. Montiel
 * @date 2022-02-18
 *
 * Copyright (C) 2022 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FPGA_IO_H_
#define FPGA_IO_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <mutex>
#include <vector>

#include <fastec/fas_log.h>
#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>

namespace FI_FPGA_IO
{
//--------------------------------------------------------------------
// Constant, Typedef and Macros
//--------------------------------------------------------------------

//! Message types for the IO operations.
enum FCTL_MESSAGE_TYPES
{
    IO_FPGA_GET_REG = 0x30000,      //!< Read an FPGA register.
    IO_FPGA_GET_MULTIREG,           //!< Read many FPGA registers in one "transaction".
    IO_FPGA_SET_REG,                //!< Write an FPGA register.
    IO_FPGA_SET_MULTIREG,           //!< Write many FPGA registers in one "transaction".
    IO_FPGA_MODIFY_REG,             //!< Modify (read, modify, write) an FPGA register.
    IO_I2C_ACCESS,                  //!< Read or write an I2C register.
    IO_IMG_DNLOAD,                  //!< Download an image/buffer from the FPGA.
    IO_IMG_UPLOAD,                  //!< Upload an image/buffer to the FPGA.
    IO_READ_HISTOGRAM,              //!< Read the histogram from the FPGA.
    IO_READ_IM_FIFO,                //!< Read the image memory FIFO.
    IO_MFT_ACCESSORY_CHECK,         //!< Check for an accessory connect to the camera.
    IO_MFT_LENS_CHECK,              //!< Check to see if an MFT lens is attached.
    IO_MFT_READ_DATA,               //!< Read data from the MFT bus.
    IO_MFT_RESET_LENS,              //!< Reset the lens.
    IO_MFT_SEND_COMMAND,            //!< Send a micro-fourthirds command.
};

//! Message class for IO_FPGA_GET_REG messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioFpgaGetRegisterMessage : public FAS_THREAD::SyncMessage
{
public:
    ioFpgaGetRegisterMessage();
    virtual ~ioFpgaGetRegisterMessage();

    int status;                     //!< The ioctl return status.
    FI_REG_ACCESS access;           //!< The access structure.
};

//! Message class for IO_FPGA_GET_MULTIREG messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioFpgaGetMultiRegisterMessage : public FAS_THREAD::SyncMessage
{
public:
    ioFpgaGetMultiRegisterMessage();
    virtual ~ioFpgaGetMultiRegisterMessage();

    int status;                         //!< The ioctl return status.
    std::vector<FI_REG_ACCESS> access;  //!< The access structure.
};

//! Message class for IO_FPGA_SET_REG messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioFpgaSetRegisterMessage : public FAS_THREAD::SyncMessage
{
public:
    ioFpgaSetRegisterMessage();
    virtual ~ioFpgaSetRegisterMessage();

    int status;                     //!< The ioctl return status.
    FI_REG_ACCESS access;           //!< The access structure.
};

//! Message class for IO_FPGA_SET_MULTIREG messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioFpgaSetMultiRegisterMessage : public FAS_THREAD::SyncMessage
{
public:
    ioFpgaSetMultiRegisterMessage();
    virtual ~ioFpgaSetMultiRegisterMessage();

    int status;                         //!< The ioctl return status.
    std::vector<FI_REG_ACCESS> access;  //!< The access structure.
};

//! Message class for IO_FPGA_MODIFY_REG messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioFpgaModifyRegisterMessage : public FAS_THREAD::SyncMessage
{
public:
    ioFpgaModifyRegisterMessage();
    virtual ~ioFpgaModifyRegisterMessage();

    int status;                     //!< The ioctl return status.
    FI_REG_MODIFY modify;           //!< The access structure.
};

//! Message class for IO_I2C_ACCESS messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioI2CAccessMessage : public FAS_THREAD::SyncMessage
{
public:
    ioI2CAccessMessage();
    virtual ~ioI2CAccessMessage();

    int status;                     //!< The ioctl return status.
    FI_I2C_ACCESS access;           //!< The access structure.
};

//! Message class for IO_IMG_DNLOAD messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioImgDownloadMessage : public FAS_THREAD::SyncMessage
{
public:
    ioImgDownloadMessage();
    virtual ~ioImgDownloadMessage();

    int status;                     //!< The ioctl return status.
    FI_IMAGE_ACCESS access;         //!< The access structure.
};

//! Message class for IO_IMG_UPLOAD messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioImgUploadMessage : public FAS_THREAD::SyncMessage
{
public:
    ioImgUploadMessage();
    virtual ~ioImgUploadMessage();

    int status;                     //!< The ioctl return status.
    FI_IMAGE_ACCESS access;         //!< The access structure.
};

//! Message class for IO_READ_HISTOGRAM messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioHistogramMessage : public FAS_THREAD::SyncMessage
{
public:
    ioHistogramMessage();
    virtual ~ioHistogramMessage();

    int status;                     //!< The ioctl return status.
    uint8_t *dest_ptr;              //!< Pointer to the buffer to hold the histogram data.
};

//! Message class for IO_READ_IM_FIFO messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioReadImFifoMessage : public FAS_THREAD::SyncMessage
{
public:
    ioReadImFifoMessage();
    virtual ~ioReadImFifoMessage();

    int status;                     //!< The ioctl return status.
    int byte_count;                 //!< The number of bytes to read into the destination pointer.
    uint8_t *dest_ptr;              //!< Pointer to the buffer to hold the histogram data.
};

//! Message class for IO_MFT_SEND_COMMAND messages. Typically used for
//! Type-A and Type-6 messages, as these messages are either command
//! only, or (type-6) they also write data.
//!
//! Note that this is a SYNC message (i.e. blocking).
class ioMftSendCommandMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftSendCommandMessage();
    virtual ~ioMftSendCommandMessage();

    bool isType6;                   //!< True if this is a type6 command, false otherwise.
    const uint8_t *cmd_val;         //!< The 4-byte command sequence.

    // The following are type-6 specific.
    const uint8_t *data_ptr;        //!< TYPE6: Pointer to the data portion of the command.
    int data_count;                 //!< TYPE6: Number of bytes in the data portion of the command.
};

//! Message class for IO_MFT_READ_DATA messages. Typically used for
//! Type-B and Type-C, as these messages expect to read back data.
//!
//! Note that this is a SYNC message (i.e. blocking).
class ioMftReadDataMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftReadDataMessage();
    virtual ~ioMftReadDataMessage();

    bool isTypeB;                   //!< True if this is a typeB read request, false otherwise.
    int max_count;                  //!< The maximum number of bytes expected.
    const uint8_t *cmd_val;         //!< The 4-byte command sequence.
    std::vector<uint8_t> data_ptr;  //!< Data read from the MFT bus.

    // The following are type-B specific.
    bool typeB_Fault;
};

//! Message class for IO_MFT_RESET_LENS messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioMftResetLensMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftResetLensMessage();
    virtual ~ioMftResetLensMessage();
};

//! Message class for IO_MFT_ACCESSORY_CHECK messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioMftAccessoryCheckMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftAccessoryCheckMessage();
    virtual ~ioMftAccessoryCheckMessage();

    bool accessoryFound;
};

//! Message class for IO_MFT_LENS_CHECK messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioMftLensCheckMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftLensCheckMessage();
    virtual ~ioMftLensCheckMessage();

    bool lensFound;
};

//! Message class for IO_MFT_LENS_SETTLE messages.
//! Note that this is a SYNC message (i.e. blocking).
class ioMftLensSettleMessage : public FAS_THREAD::SyncMessage
{
public:
    ioMftLensSettleMessage();
    virtual ~ioMftLensSettleMessage();

    bool lensSettled;
};

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------
class fpgaEventWorker;

//--------------------------------------------------------------------
// fpgaIOWorker class declarations
//--------------------------------------------------------------------
class fpgaIOWorker : public FAS_THREAD::MsgThread
{
private:
    static std::unique_ptr<fpgaIOWorker> io_proc;   //!< Our singleton object.
    static std::mutex io_mutex;                     //!< Mutex to protect our singleton object.
    static const char *FPGA_DRV;                    //!< FPGA driver file device path.
    static int fpga_fd;                             //!< File descriptor to the FPGA driver.

    std::unique_ptr<FI_FPGA_IO::fpgaEventWorker> event_worker;  //!< Our event thread.

    fpgaIOWorker();
    void startEventWorker();

    //! @name Register access messsage handlers.
    //! @{
    void handleFpgaModifyRegisterRequest(ioFpgaModifyRegisterMessage *msg);
    void handleFpgaReadRegisterRequest(ioFpgaGetRegisterMessage *msg);
    void handleFpgaReadMultiRegisterRequest(ioFpgaGetMultiRegisterMessage *msg);
    void handleFpgaWriteRegisterRequest(ioFpgaSetRegisterMessage *msg);
    void handleFpgaWriteMultiRegisterRequest(ioFpgaSetMultiRegisterMessage *msg);

    bool regModify(FI_REG_MODIFY& modify);
    //
    //! @}

    //! @name I2C access message handlers.
    //! @{
    void handleI2CAccessRequest(ioI2CAccessMessage *msg);
    //
    //! @}

    //! @name Memory access message handlers.
    //! @{
    void handleImgDownloadRequest(ioImgDownloadMessage *msg);
    void handleImgUploadRequest(ioImgUploadMessage *msg);
    void handleHistogramRequest(ioHistogramMessage *msg);
    void handleReadImFifoRequest(ioReadImFifoMessage *msg);
    //
    //! @}

    //! @name Micro Four Thirds (MFT) routines
    //! @{
    void handleMftAccessoryCheckRequest(ioMftAccessoryCheckMessage *msg);
    void handleMftLensCheckRequest(ioMftLensCheckMessage *msg);
    void handleMftLensSettleRequest(ioMftLensSettleMessage *msg);
    void handleMftReadDataRequest(ioMftReadDataMessage *msg);
    void handleMftResetLensRequest(ioMftResetLensMessage *msg);
    void handleMftSendCommandRequest(ioMftSendCommandMessage *msg);

    void mftControlReset();
    bool mftWait();
    bool sendCommand(const uint8_t *cmd_val);
    bool sendCommandType6(const uint8_t *cmd_val, const uint8_t *data_val, int data_count);

    int readBytes(std::vector<uint8_t>& data_ptr, int max_count, int tdl_len = 2, bool *faulted = nullptr);
    bool writeMftByte(uint8_t reg_data);
    //
    //! @}

    void handleHalt(FAS_THREAD::Message *message);

public:
    virtual ~fpgaIOWorker();
    virtual void messageHandle(FAS_THREAD::Message *message) override;

    // Singletons should not be cloneable.
    fpgaIOWorker(fpgaIOWorker& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const fpgaIOWorker&) = delete;

    static fpgaIOWorker *factory();
    static void close()
    {
        fpgaIOWorker *thread_ptr;

        // Release our singleton.
        thread_ptr = fpgaIOWorker::factory();
        if (thread_ptr != nullptr)
        {
            // Tell the thread that we are stopping.
            FAS_THREAD::SyncMessage msg;
            msg.req = FAS_THREAD::HALT;

            // Pass it on.
            msg.doSync(thread_ptr);

            // Try to join the thread.
            thread_ptr->join();
        }

        LOG("%s--", __PRETTY_FUNCTION__);
    }

    static bool isRunning()
    {
        std::lock_guard<std::mutex> lock(io_mutex);
        return (io_proc.get() != nullptr);
    }
};
} /* End of namespace FI_FPGA_IO */

#endif /* !FPGA_IO_H_ */

