/*!
 * @file gige_reg.h
 * @brief This file defines the GIGE_REG (and associated) class(es).
 * @author Anthony L. Montiel
 * @date 2013-02-06
 *
 * Copyright (C) 2013 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_GIGE_REG_H_
#define _FASTEC_GIGE_REG_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <map>
#include <mutex>

#include <fastec/gige_msg.h>

//--------------------------------------------------------------------
// Forward Delcarations
//--------------------------------------------------------------------

namespace CS_NETWORK
{
struct CS_NETWORK_MESSAGE;
}

//--------------------------------------------------------------------
// Constants, Macros and Type definitions
//--------------------------------------------------------------------

#define REG_TYPE_BIN            0   //!< register type is binary (4 bytes)
#define REG_TYPE_STR            1   //!< register type is string

#define REG_ACCESS_RO           0   //!< register is readonly
#define REG_ACCESS_RW           1   //!< register is read write
#define REG_ACCESS_WO           2   //!< register is write only

#define REG_SWAP_NONE           0   //!< no swapping on register
#define REG_SWAP_HTONL          1   //!< use system htonl
#define REG_SWAP_BIT_REV        2   //!< use bit reversal function

typedef int (* REG_UPDATE_FN)(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

#pragma pack(1)
//! Structure to hold the definition of a GIGE (aka, FCP) register.
typedef struct GIGE_REG
{
    uint32_t reg_num;       //!< Register number (0x0000 - 0xFFFF).
    uint8_t reg_type;       //!< 0 for binary, 1 for string.
    uint8_t reg_access;     //!< 0 for read only, 1 for read write, 2 for write only.
    uint8_t reg_size;       //!< Register size, in bytes.
    uint8_t reg_swap_mode;      //!< Registers swapping (0=none, 1=htonl, 2=bit reverse).
    union
    {
        uint32_t reg_val_l;     //!< Current value of the register expressed as a long.
        uint16_t reg_val_w[2];  //!< Value of the register, expressed in words.
        uint8_t reg_val_b[GIGE_REG_STR_MAX_SIZE]; //!< Register value in bytes, maximum size.
    };
    REG_UPDATE_FN   reg_update; //!< Update function specific for this register (may be NULL).
} GIGE_REG;
#pragma pack(0)

typedef std::map<uint32_t, GIGE_REG>    GIGE_MAP;

//--------------------------------------------------------------------
//! @brief The FCP register cache class.
//--------------------------------------------------------------------
class GIGE_REG_CACHE
{
private:
    uint32_t        last_index;     //!< The active register address from the array operator.
    GIGE_MAP        gige_regs;      //!< Our register cache table.
    std::timed_mutex    cache_lock;     //!< Mutex to control access to the register cache.

    void    validate_reg_file();
    void    setup_serial_number(uint8_t *mac_addr);
    bool    validate_serial_number(const char *buf);
    bool    reflect_register(uint32_t reg_addr);

public:
    //! Filename to hold register values.
    static const char *REGISTER_FILE;

    //! Filename for md5 checksum of the register file.
    static const char *REGISTER_CHECKSUM_FILE;

public:
    GIGE_REG_CACHE();
    virtual ~GIGE_REG_CACHE();

    int     init_cache();
    int     load();
    void    reset();
    void    save();
    bool    set_loaded_registers();
    void    set_mac_address(uint8_t *mac_addr);
    bool    validate_id(uint32_t new_id);

    // Operator overloads.
    GIGE_REG_CACHE& operator[](uint32_t reg_id);
    GIGE_REG_CACHE& operator[](int reg_id);

    GIGE_REG_CACHE& operator= (uint32_t reg_value);
    GIGE_REG_CACHE& operator= (int reg_value);
    GIGE_REG_CACHE& operator&=(uint32_t reg_value);
    GIGE_REG_CACHE& operator&=(int reg_value);
    GIGE_REG_CACHE& operator|=(uint32_t reg_value);
    GIGE_REG_CACHE& operator|=(int reg_value);

    bool operator==(uint32_t reg_value);
    bool operator==(int reg_value);
    bool operator!=(uint32_t reg_value);
    bool operator!=(int reg_value);

    uint32_t operator&(uint32_t reg_value);
    int operator&(int reg_value);
    uint32_t operator|(uint32_t reg_value);
    int operator|(int reg_value);

    operator uint16_t *();
    operator uint8_t *();
    operator char *();
    operator GIGE_REG *();

    uint32_t get_value();
    uint32_t set_value(uint32_t reg_value);

    uint32_t reg_addr();
    uint8_t reg_type();
    uint8_t reg_access();
    uint8_t reg_size();
    uint8_t reg_swap_mode();
    int reg_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

    REG_UPDATE_FN update_is_registered();
};

#endif /* !_FASTEC_GIGE_REG_H_ */

