/*!
 * @file image_correction.h
 * @brief This is the header file for the FiImageCorrection class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_IMAGE_CORRECTION_H_
#define FASTEC_IMAGE_CORRECTION_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <memory>
#include <string>

namespace FI_IMAGE_CORRECTION
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------

} /* End of namespace FI_IMAGE_CORRECTION */

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------
class FiFPGA;

//--------------------------------------------------------------------
// FiImageCorection class declarations
//--------------------------------------------------------------------
class FiImageCorrection
{
private:
    static bool fpn_on;                     //!< The FPGA FPN correction is to be turned on.

    static std::string prnu_red_stack;      //!< Path to the stack of "red" raw TIFFs to use for color PRNU calibration.
    static std::string prnu_green_stack;    //!< Path to the stack of "green" raw TIFFs to use for color PRNU calibration.
    static std::string prnu_blue_stack;     //!< Path to the stack of "blue" raw TIFFs to use for color PRNU calibration.

public:
    FiImageCorrection();
    virtual ~FiImageCorrection();

    //! @name Fixed Pixel Noise (FPN) Support
    //! These are the routines which are used for Fixed Pixel Noise (FPN) calculation
    //! and application.
    //! @{
    static int fpnApplyOffset(uint8_t bit_mode, uint16_t black_offset);
    static int fpnCalibrate(bool do_factory, bool manual_params);
    static bool fpnCalibrateActive();
    static int fpnDisable();
    static int fpnEnable();
    static bool fpnIsEnabled()    { return fpn_on; }
    static int fpnLoad();
    static bool fpnNeedsCal();
    static void fpnResetOffset();
    static uint8_t fpnGetOffset();
    static int fpnPropagate();
    //
    //! @}

    //! @name Photo Response Non-Uniformity (PRNU) Support
    //! These are the routines which are used for Photo Response Non-Uniformity (PRNU) calculation
    //! and application.
    //! @{
    static int prnuCalibrate();
    static bool prnuGenerate(const uint16_t *avg_image, const uint8_t *fpn_table, uint8_t *gain_table);
    static bool prnuIsAvailable();
    static uint32_t prnuSetup(const char *settings_str, int str_length);
    static bool prnuColorSum(int capture_count, uint16_t *avg_image);
    static bool prnuImageAvg(int capture_count, const std::string& img_path, uint16_t *img_avg_ptr);
    static bool prnuMonoSum(int capture_count, uint8_t bitmode, uint64_t cal_frame_addr, uint32_t *sum_table);
    //
    //! @}
};

#endif /* !FASTEC_IMAGE_CORRECTION_H_ */
