/*!
 * @file image_format.h
 * @brief This file defines the image formatting functionality.
 * @author Anthony L. Montiel
 * @date 2012-05-16
 *
 * Copyright (C) 2012 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_IMAGE_FORMAT_H_
#define _FASTEC_IMAGE_FORMAT_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>

#include <deque>
#include <filesystem>
#include <string>

#include <fastec/gige_msg.h>
#include <fastec/fas_video.h>
#include <fastec/fas_internal.h>

namespace IMAGE_FORMAT
{
//--------------------------------------------------------------------
// Constants, Macros and Type Definitions
//--------------------------------------------------------------------

typedef bool (*SSAVE_STATE)(void *user_data);

typedef enum
{
    IMAGE_STATE_START,
    IMAGE_STATE_IN_PROGRESS,
    IMAGE_STATE_FINISH,
    IMAGE_STATE_DONE,
    IMAGE_STATE_COUNT,
} IMAGE_STATES;

//! The IMG_PROCESSING_INFO structure holds the information used when saving a capture.
typedef struct IMG_PROCESSING_INFO
{
    uint32_t    start_frame = 0;        //!< Starting frame number (inclusive).
    uint32_t    end_frame = 0;          //!< Ending frame number (inclusive).
    uint32_t    current_frame = 0;      //!< Current frame being processed.
    uint64_t    total_bytes = 0;        //!< Total number of bytes transferred.

    std::string target_path;            //!< Path name for the target file.
    bool    do_subfolders = false;      //!< True if we're doing subfolders.
    std::string::size_type subfolder_index = 0; //!< Index in the target_path string where the subfolder name starts.
    uint32_t    subfolder_count = 0;    //!< If we have too many images, we need to break into subfolders.

    uint32_t    state = 0;              //!< Processing state.
    uint32_t    status = 0;             //!< Status information once we are done. 0=OK.

} IMG_PROCESSING_INFO;

//! The FFLR_CONTEXT structure holds the information to track the save state of
//! a FasFireLR save operation.
typedef struct FFLR_CONTEXT
{
    int partition;                      //!< The partition number for this information.
    bool is_trigger_partition;          //!< True if this is the trigger partition, false otherwise.
    int trigger_frame;                  //!< The trigger frame number in a particular capture, 0 to frame_count-1. Or -1 if none.
    uint32_t frame_count;               //!< The number of frames in the capture.
    int file_number;                    //!< The number of the CAP file.
    std::filesystem::path cap_file;     //!< The CAP file fully qualified name.
} FFLR_CONTEXT;

//! A handy name for a pointer to the context information.
typedef std::unique_ptr<FFLR_CONTEXT>   FFLR_CONTEXT_PTR;

} /* End of namespace IMAGE_FORMAT */

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class JpegUtil;

//--------------------------------------------------------------------
//! @brief The stream_save class.
//--------------------------------------------------------------------
class StreamSave
{
public:
    StreamSave();
    StreamSave(uint32_t dest_info);
    virtual ~StreamSave();

    void    stop();
    bool    doSave();

    int getCompletionPercent();
    const IMAGE_FORMAT::IMG_PROCESSING_INFO *getInfo() const      { return &format_info; }
    uint32_t getStatus() const                  { return format_info.status; }
    const char *getTargetPath() const           { return format_info.target_path.c_str(); }
    uint64_t getTotalBytes() const              { return format_info.total_bytes; }
    uint32_t getTotalFrames() const             { return ((format_info.end_frame - format_info.start_frame) + 1); }

    void setDestInfo(uint32_t dest_info)            { destination_info = dest_info; }
    void setH264Quality(int quality)                { h264_quality = quality; }
    void setH264Speed(FAS_VIDEO::H264_SPEEDS speed) { h264_speed = speed; }
    void setH264EnableAllKeyframes(bool enable)     { h264_enable_all_keyframes = enable; }
    void setStartFrame(uint32_t start_frame_num)    { format_info.current_frame = format_info.start_frame = start_frame_num; }
    void setEndFrame(uint32_t end_frame_num)        { format_info.end_frame = end_frame_num; }
    void setVideoCodec(uint32_t codec)              { video_codec = codec; }
    void setTargetPath(const char *save_path)       { format_info.target_path.assign(save_path); }

    static int checkSpace(uint64_t total_required, const char *target_path);

private:
    static bool startupState(void *user_data);
    static bool progressState(void *user_data);
    static bool finishState(void *user_data);
    static bool doneState(void *user_data);

    void initClass();
    bool checkSubfolder();
    void saveMetadata();
    bool savePfmd();

    // image_fmt_cap.cxx
    int capCompletionPercent();
    bool capFullSave();
    bool capSubpartSave();
    bool capSave();
    uint64_t capMultiWrite(int cap_fd, uint64_t im_source_addr, std::vector<uint64_t>& dma_addr, std::vector<uint8_t *>& xfer_buffer, uint64_t total_length);

    // image_fiv.cxx
    void getDmaMod(uint32_t stream_type, pm_metadata *meta_ptr);
    void getProcessingInfo(uint32_t stream_type, pm_metadata *meta_ptr);
    bool isStreamVideo(uint32_t streamType);
    int ltrFrameLoad(uint32_t frame_num);
    int preparePackedWriter();
    bool processFivImage();
    void saveFivMetadata(const char *dest_path, const char *meta_buffer, uint32_t meta_length);
    bool startFivSave();

private:
    volatile bool stop_request;         //!< Flag to indicate the user wants to stop this thread.

    static std::string dest_name;       //!< Generic filename buffer.

    std::shared_ptr<FAS_VIDEO::FiVideoStream> streamWriter;  //!< Our video stream writer.
    FAS_INTERNAL::FIP_HANDLE streamHandle; //!< Handle to the packed streamer.

    uint32_t destination_info;          //!< GIGE_REG_IMAGE_DESTINATION information.
    IMAGE_FORMAT::IMG_PROCESSING_INFO format_info;

    int      bb_index;                  //!< Bounce buffer index.
    uint64_t bb_addr;                   //!< Address for the bounce buffer.

    uint32_t h264_quality;              //!< If saving MP4, the quality factory is H.264 crf option.
    FAS_VIDEO::H264_SPEEDS h264_speed;  //!< If saving MP4, the compression speed is H.264 preset options.
    bool h264_enable_all_keyframes;     //!< If saving MP4, this will set all frames as keyframes if true.
    uint32_t video_codec;               //!< Codec used for the video stream (i.e. JPEG, BMP, Planar).

    int ltr_bb_index;                   //!< Bounce buffer index for FPGA processing (LR mode).
    uint64_t ltr_bb_addr;               //!< The address for the LTR bounce buffer (LR mode).
    uint64_t ltr_frame_addr;            //!< The address in IM for the LTR frame.

    IMAGE_FORMAT::SSAVE_STATE save_states[IMAGE_FORMAT::IMAGE_STATE_COUNT];
};

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

#endif /* !_FASTEC_IMAGE_FORMAT_H_ */

