/*!
 * @file image_io.h
 * @brief This file defines the image I/O class(es).
 * @author Anthony L. Montiel
 * @date 2012-05-16
 *
 * Copyright (C) 2012 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_IMAGE_IO_H_
#define _FASTEC_IMAGE_IO_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <limits.h>
#include <fastec/fas_thread.h>
#include <fastec/fas_video.h>

#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>

#include <fastec/fas_video.h>

#include "image_format.h"

//--------------------------------------------------------------------
//! @brief The imgio class.
//! The imgio class is a pure static class.
//--------------------------------------------------------------------
class imgio
{
public:
    static const char   *get_codec_str(uint32_t stream_type);
    static const char   *get_suffix(uint32_t stream_type);
    static FAS_VIDEO::VIDEO_CODEC get_codec(uint32_t dest_info);

    static std::unique_ptr<FAS_VIDEO::HEADER_DATA> get_image_header(uint32_t stream_type, pm_metadata *pm_ptr);
    static uint32_t get_image_type(uint32_t stream_type, pm_metadata *pm_ptr);

    // A static class is not cloneable.
    imgio(imgio& other) = delete;

    // A static class is not assignable.
    void operator=(const imgio&) = delete;

private:
    imgio() {}
    virtual ~imgio() {}
};

//--------------------------------------------------------------------
//! @brief The imgSaveWorker class follows a singleton pattern.
//!
//! This class is a threaded object that does the actual work of saving
//! image data to the local drive. For saving across the network, the
//! fspWorker class is used.
//!
//! Use the static imgSaveWorker::getInstance() method to retrive the
//! singleton object.
//!
//! The entry for the save process is normally via FCP_LOCAL::fcp_parse_save_json().
//--------------------------------------------------------------------
class imgSaveWorker : public FAS_THREAD::Thread
{
public:
    virtual ~imgSaveWorker();

    // A singleton is not cloneable.
    imgSaveWorker(imgSaveWorker& other) = delete;

    // A singleton is not assignable.
    void operator=(const imgSaveWorker&) = delete;

    // This is the method that controls access to our singleton.
    static imgSaveWorker *getInstance();

    void run() override;
    void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;
    void stop() override;
    void reset();

    bool isRunning()                                    { return thread_running; }
    constexpr bool needToRemount()                      { return was_mounted; }

    void clearError();
    uint32_t getCompletion();
    int getLiveSnapshot();

    constexpr bool saveCfgMetadata()                    { return cfg_metadata_save; }
    constexpr void setSaveCfgMetadata(bool do_save)     { cfg_metadata_save = do_save; }
    constexpr bool savePFMD()                           { return pfmd_save; }
    constexpr void setSavePFMD(bool do_save)            { pfmd_save = do_save; }

    constexpr uint32_t getVideoCodec()                  { return video_codec; }
    constexpr void setVideoCodec(uint32_t reg_value)    { video_codec = reg_value; }
    constexpr void setPartNum(int part_num)             { partition_num = part_num; }
    uint32_t getDestInfo()                              { return dest_info; }
    constexpr void setDestInfo(uint32_t reg_value)      { dest_info = reg_value; }
    constexpr void setH264Quality(int quality)          { h264_quality = quality; }
    constexpr void setH264Speed(FAS_VIDEO::H264_SPEEDS speed)   { h264_speed = speed; }
    constexpr void setH264EnableAllKeyframes(bool enable)       { h264_enable_all_keyframes = enable; }
    const char *getMountDir()                           { return mount_dir.c_str(); }
    void setMountDir(const std::string& dir_str)        { mount_dir.assign(dir_str); }

    constexpr bool getUsingShmem()                      { return using_sharedmem; }
    void setUsingShmem(bool use_shmem)                  { using_sharedmem = use_shmem; }

    bool parseSaveSettings(const char *save_json, int str_length, bool start_saver = true);
    void restoreSaveSettings(bool enable_as);
    void updateSaveSettings(const char *save_json);

    // Sleazy hijack of image saver thread to extract a marker list, taking advantage of existing
    // checks for the resources that it needs (e.g. PME engine, GOE buffer).
    // We really need a better way to manage the ownership of resources.
    void    requestMarkers(uint32_t pattern, uint8_t *buf, int buf_size, int *p_rtnsize);
    bool    isDoingMarkers();

    //! @name FasFireLR support methods.
    //! @{
    uint32_t getLastFflrSaveInfo(char *dest_ptr, uint32_t max_len);
    void resetLastFflrSaveInfo()                            { fflr_info.clear(); }
    constexpr uint32_t getRecordingFrameCount()             { return recording_frame_count; }
    constexpr void setRecordingFrameCount(uint32_t count)   { recording_frame_count = count; }
    void saveFflrMetadata();
    IMAGE_FORMAT::FFLR_CONTEXT_PTR popFflrQueue();
    constexpr void setCapStackIndex(int index)              { capstk_index = index; }
    constexpr void setFflrSaveError(uint32_t status)        { save_error = status; }
    constexpr bool fflrRecovery()                           { return saveMetadata; }
    //
    //! @}

private:
    imgSaveWorker();

    void    initMarkerList();
    void    cleanupMarkerList();
    void    stopMarkerList();
    void    createMarkerList();
    uint32_t    createMarkerList(uint32_t pattern, uint32_t *p_dest, uint32_t max_count,
                                 uint16_t part_num, uint32_t start, uint32_t end);
    // End of temporary marker list extraction support.

    void    cleanup();

    void    recoverFflr();
    void    saveVideo();
    bool    copyFile(const char *target_ptr, const char *source_ptr);
    const char *buildTarget();

    int32_t getStartFrameNumber();
    int32_t getEndFrameNumber();

    bool    saveFrameMetadata(void *xmlwriter, uint32_t start, uint32_t end, uint16_t part_num);
    bool    saveXmlMetadata(const char *dest_ptr, const pm_metadata *meta_ptr, uint16_t part_num);

private:
    static std::unique_ptr<imgSaveWorker> instance;
    static std::mutex instance_mutex;

    static const std::string FFLR_METAFILE_NAME;
    static const std::string SAVE_SETTINGS_FILE;

    volatile bool thread_running;           //!< Flag to indicate if we are running or not.
    std::mutex thread_mutex;                //!< Used to control changes to IPM status.
    std::condition_variable thread_status;  //!< Condition variable we can block on.
    volatile uint32_t   dest_info;          //!< GIGE_REG_IMAGE_DESTINATION value.
    volatile uint32_t   last_status;        //!< Last status as reported by the imem_manager.
    std::string mount_dir;

    bool    cfg_metadata_save;              //!< True if we're saving the "CFG" metadata file, false otherwise.
    bool    pfmd_save;                      //!< True if we're saving the PFMD file, false otherwise.
    bool    need_end_notification;          //!< Whether to send the IDLE notification when done.
    bool    was_mounted;                    //!< True if our device (SSD, etc.) was mounted.
    bool    raw_image;                      //!< True if this is a raw image.
    std::string target_file;                //!< Fully qualified name to destination file.
    int     partition_num;                  //!< Partition we want to save.
    uint32_t video_codec;                   //!< The codec to use when saving AVIs.

    std::mutex  img_lock;                   //!< Control access to data.
    std::unique_ptr<StreamSave> saver;      //!< Our video saving class.
    bool saver_starting;                    //!< Remove ambiguity about 0% vs 100%.

    uint32_t    h264_quality;               //!< If saving MP4, the quality factory is H.264 crf option.
    FAS_VIDEO::H264_SPEEDS h264_speed;      //!< If saving MP4, the compression speed is H.264 preset options.
    bool h264_enable_all_keyframes;         //!< If saving MP4, this will set all frames as keyframes if true.

    bool using_sharedmem;                   //!< Flag indicating if we're supposed to use the shared memory ring buffer.

    //! @name Variables associated with the marker list extraction.
    //! @{
    bool doing_marker_list;
    uint8_t *p_marker_buf;
    int marker_buf_size;
    int *p_rtn_marker_size;
    uint32_t marker_pattern;
    //
    //! @}

    //! @name FasFireLR support variables.
    //! @{
    std::string fflr_info;                  //!< JSON formatted information on the last FFLR save.
    int capstk_index;                       //!< Index to use for the CAP stack file name.
    std::string capstk_base;                //!< The saved base path used for saving a CAP stack.
    uint32_t recording_frame_count;         //!< The total number of frames in a FFLR recording.
    std::deque<IMAGE_FORMAT::FFLR_CONTEXT_PTR> fflr_queue;  //!< The context queue for the FFLR save.
    uint32_t save_error;                    //!< Set if something in the save process caused an error.

    bool writeBuffers;                      //!< Recovery flag, true to save captured partitions.
    bool saveMetadata;                      //!< Recovery flag, true to save the FFLR metadata.
    //
    //! @}
};

//--------------------------------------------------------------------
//! @brief The imgLoadWorker class.
//! The imgLoadWorker class follows a singleton pattern.
//--------------------------------------------------------------------
class imgLoadWorker : public FAS_THREAD::Thread
{
public:
    virtual ~imgLoadWorker();

    // A singleton is not cloneable.
    imgLoadWorker(imgLoadWorker& other) = delete;

    // A singleton is not assignable.
    void operator=(const imgLoadWorker&) = delete;

    // This is the method that controls access to our singleton.
    static imgLoadWorker *getInstance();

    void run() override;
    void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;
    void stop() override;

    bool isRunning()        { return thread_running; }

    void setFilename(const char *filename, uint32_t max_length);
    void getFilename(char *filename, uint32_t max_length);

    uint32_t getStatus();
    uint32_t getCompletion();

private:
    imgLoadWorker();

    void cleanup();

    bool capIsOk(const char *cap_file);
    bool capfmtImageLoad();
    void capfmtResetProgress();
    bool updateAfterCapLoad();

    void loadVideo();
    bool validPartition();

    bool loadFromMedia();

private:
    static std::unique_ptr<imgLoadWorker> instance;
    static std::mutex instance_mutex;

    volatile bool   thread_running;         //!< Flag to indicate if we are running or not.
    volatile bool   load_progress_valid;    //!< Is false until progress info is initialized.
    volatile uint32_t   last_status;        //!< Last status from the load operation.
    std::string target_file;                //!< Fully qualified name to source file.
    pm_metadata file_pm_data;               //!< The partition metadata read from the file.

    std::mutex img_lock;                    //!< Control access to data.
    IMAGE_FORMAT::IMG_PROCESSING_INFO format_info;  //!< Our format information structure.
};

#endif /* !_FASTEC_IMAGE_IO_H_ */

