/*!
 * @file jpeg_proc.h
 * @brief Interface for coordinating access to and using the JPEG encoder.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __JPEG_PROC_H__
#define __JPEG_PROC_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

namespace JPEG_PROC
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Organize all the input and output parameters for the request to
//! encode a pre-existing buffer.
typedef struct BUF_REQUEST
{
    //! @name Input values.
    //! @{
    uint8_t *src_pixels_Y;  //!< Pointer to the Y-plane pixel data.
    uint8_t *src_pixels_UV; //!< Pointer to the UV-plane pixel data.
    uint32_t width;     //!< Number of pixels per row.
    uint32_t height;        //!< Number of rows per image.
    uint32_t pixel_format;  //!< One of JPEG_UTIL::PIX_FORMAT.
    uint32_t quality;       //!< JPEG quality factor.
    //! @}

    //! @name Output values.
    //! @{
    uint8_t *jpeg_buffer;   //!< Pointer to the encoded data. NULL on error.
    int jpeg_len;       //!< Number of bytes in the encoded data.
    //! @}
} BUF_REQUEST;

//! Organize all the input and output parameters for the IPM encoding
//! request into one structure.
typedef struct IPM_REQUEST
{
    //! @name Input values.
    //! @{
    uint64_t image_addr;    //!< Address in Image Memory for the frame to encode.
    uint32_t dma_mod;       //!< DMA modification settings.
    uint32_t proc_info;     //!< IPM processing information.
    uint32_t width;         //!< Number of pixels per row.
    uint32_t height;        //!< Number of rows per image.
    uint32_t quality;       //!< JPEG quality factor.
    uint32_t time_limit;    //!< For image reuse: How many microseconds old can the image be?
    uint32_t ma_handle;     //!< If we are doing motion amplification, the handle to the engine.
    //! @}

    //! @name Output values.
    //! @{
    uint8_t *jpeg_buffer;   //!< Pointer to the encoded data. NULL on error.
    int jpeg_len;       //!< Number of bytes in the encoded data.
    //! @}
} IPM_REQUEST;

//! Organize all the input and output parameters for the IPM encoding
//! request into one structure.
typedef struct REUSE_REQUEST
{
    //! @name Input values.
    //! @{
    uint32_t time_limit;    //!< How many microseconds old can the image be?
    //! @}

    //! @name Output values.
    //! @{
    uint8_t *jpeg_buffer;   //!< Pointer to the encoded data. NULL on error.
    int jpeg_len;       //!< Number of bytes in the encoded data.
    //! @}
} REUSE_REQUEST;

} /* End of namespace JPEG_PROC */

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

extern void jpeg_init();
extern void jpeg_close();

extern void jpeg_dump(FILE *dump_stream);
extern bool jpeg_buf_request(JPEG_PROC::BUF_REQUEST *request);
extern bool jpeg_ipm_request(JPEG_PROC::IPM_REQUEST *request);
extern void jpeg_release_frame(uint8_t *jpeg_buffer);
extern bool jpeg_reuse(JPEG_PROC::REUSE_REQUEST *request);

#endif /* !__JPEG_PROC_H__ */
