/*!
 * @file lince5m.h
 * @brief This is the header file providing the hardware abstraction layer
 * for the Lince5M and its sensor-specific FPGA interface.
 *
 * Copyright 2010, 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef __LINCE5M_H__
#define __LINCE5M_H__

#include <memory>
#include "sensor.h"

namespace LINCE5M_SENSOR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const int ICM_LINCE_WR_FIFO_SIZE_WORDS = 256;    //!< Size of Write FIFO in 32-bit words.
const int ICM_LINCE_RD_FIFO_SIZE_WORDS = 256;    //!< Size of Read FIFO in 32-bit words.

} /* End of namespace LINCE5M_SENSOR */

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------
class FiFPGA;

//! Actual definition of the base class
class Lince5M : public Sensor
{
private:

    //! A defective pixel location.
    typedef struct BAD_PIXEL
    {
        uint16_t row;
        uint16_t col;
    } BAD_PIXEL;

    unsigned    current_gain;
    unsigned    current_framing_mode;
    unsigned    num_defects;
    std::unique_ptr<BAD_PIXEL[]> defect_map;

    bool sensor_test_on;
    uint32_t vtroi_x;
    uint32_t vtroi_y;
    uint32_t vtroi_width;
    uint32_t vtroi_height;

private:
    static constexpr int sbpCompare(const void *a, const void *b);

    bool spiReadCmd(uint8_t cmd, uint8_t *data, unsigned data_count);
    bool spiReadData(uint16_t cmd, uint16_t addr, uint16_t *data, unsigned data_count);
    bool spiReadReg(uint16_t reg_addr, uint16_t *rtn_data);
    bool spiReadStatus(uint8_t *status);
    bool spiWriteCmd(uint8_t cmd, uint8_t *data, unsigned data_count);
    bool spiWriteCmd1(uint8_t cmd, uint8_t data)      { return spiWriteCmd(cmd, &data, 1); }
    bool spiWriteData(uint16_t cmd, uint16_t addr, uint16_t *data, unsigned data_count);
    bool spiWriteReg(uint16_t reg_addr, uint16_t data);
    bool spiWaitIdle();
    bool spiWriteMemoryFile(uint16_t memaddr, uint16_t count, const char *file_name);
    bool spiWriteRegFile(const char *file_name);

    bool readDefectMap();
    bool writeDefectMap(uint32_t sensor_opts, uint16_t *p_mapsize);

    bool adjust_analog_offset(unsigned analog_gain);
    bool commit_memory_changes();
    bool disable_vfpn_correction();
    bool doHardStop();
    bool enable_vfpn_correction();
    bool ensure_stopped();
    bool bandwidth_control(uint16_t *p_w, uint16_t *p_h, bitmode_t bit_mode);
    bool sensor_to_user(unsigned factor, uint16_t sensor_startx, uint16_t sensor_starty,
                        uint16_t sensor_width, uint16_t sensor_height, bitmode_t bit_mode,
                        uint16_t *rtn_startx, uint16_t *rtn_starty,
                        uint16_t *rtn_width, uint16_t *rtn_height);
    bool force_legal_roi(unsigned factor, uint16_t *p_x, uint16_t *p_y, uint16_t *p_width,
                         uint16_t *p_height);
    bool framingModeConfig(int framing_mode);
    bool hardRestart(uint32_t sensor_opts);
    bool io_power_up();
    bool parameters_config(unsigned t_row_min, unsigned analog_gain, uint32_t sensor_opts);
    bool sensor_config(unsigned t_row_min, bool bin_fix);
    bool set_fpga_roi(unsigned factor, uint16_t sensor_startx, uint16_t sensor_width,
                      uint16_t readout_startx, uint16_t readout_width, uint16_t height);
    bool set_sensor_roi_regs(uint16_t startx, uint16_t starty, uint16_t width, uint16_t height);
    bool set_sensor_readout(bool enable) override;
    bool start_from_config(uint16_t startx, uint16_t starty, uint16_t width, uint16_t height,
                           bool sensor_fpn_on, bitmode_t bit_mode, float frame_rate, uint32_t integration_usecs,
                           uint32_t sensor_opts, unsigned analog_gain, bool is_mini);
    bool update_reg_cmd();
    bool update_roi_cmd();
    bool vfpn_offset_calibrate(uint32_t exposure, unsigned analog_gain, bool is_mini);

    bool    toggleReset();
    bool    alignLVDS(uint16_t idle_pattern);
    bool    setIcmFrequency(uint32_t sensor_opts);
    bool    icmSensorRoi(uint32_t first_ppw_ofs, uint32_t ppw_per_row,
                         uint32_t last_ppw_count, uint32_t height);
    void    dump_range(void *p, char tag, uint8_t cmd, unsigned low_addr, unsigned high_addr,
                       uint16_t *buf, unsigned buf_words);


public:

    /// Constructor and destructor
    explicit Lince5M(SENSOR::EEPROM_INFO *eeprom);
    ~Lince5M();

    bool class_init(FiFPGA *fpga_ptr) override;
    bool enable_sensor_test() override;
    void shutdown() override;

    std::string defect_filename(const char *dirname);
    void dump_sensor_regs(void *fp) override;

    uint32_t  get_default_frame_rate() override;
    bitmode_t get_default_bit_mode() override;
    uint32_t  get_default_gain() override;
    uint32_t  get_frame_rate() override;
    bitmode_t get_max_bit_mode()  override { return (bitmode_t)GIGE_BITMODE_11_0; }
    uint32_t  get_sensor_min_shutter() override;
    uint16_t  get_reported_max_width(uint32_t sensor_opts, bitmode_t bit_mode) override;
    uint16_t  get_reported_max_height(uint32_t sensor_opts, bitmode_t bit_mode) override;
    double    get_temperature() override;

    bool factory_needs_reset(bool manual_params) override { return !manual_params; }
    bool has_factory_blkref() const noexcept override { return false; }
    bool supports_pixel_fpn() const noexcept override { return false; }
    bool use_legacy_layout() const noexcept override { return false; }

    bool bitmode_needs_restart(bitmode_t old_bit_mode, bitmode_t new_bit_mode) override;
    void set_loaded_registers() override;
    bool fpn_needs_restart(fpnmode_t old_fpn, fpnmode_t new_fpn) override;
    bool frame_rate_needs_restart(float old_fps, float new_fps, uint16_t startx,
                                  uint16_t starty, uint16_t width, uint16_t height,
                                  bitmode_t bit_mode, uint32_t sensor_opts) override;

    uint16_t get_cal_offset(uint32_t sensor_type, void *params) override;
    bool     is_cal_compatible(uint32_t sensor_type, void *params, SENSOR::SENSOR_ROI *calfile_roi,
                               SENSOR::SENSOR_ROI *user_roi) override;
    void     get_ct_offset(bitmode_t bit_mode, uint32_t sensor_opts, uint16_t *rtn_ofs_x,
                           uint16_t *rtn_ofs_y) override;

    // Return the extra time, not reflected in the frame rate, that it takes to complete a frame.
    virtual uint32_t get_frame_overhead() override;

    uint32_t IPM_bayer_pattern_start(uint16_t startx, uint16_t starty) override;
    bool is_bit_mode_supported(bitmode_t bit_mode) override;
    bool is_option_supported(uint32_t options) override;
    bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                           uint16_t *rtn_width, uint16_t *rtn_height,
                           bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                           uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts) override;
    bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) override;
    bool sensor_startup(uint16_t start_x, uint16_t start_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t integration, uint32_t sensor_opts, uint32_t sensor_gain) override;
    bool run_calibration(bool is_factory, bool manual_params, bool *p_rtn_restart) override;

    bool set_fpga_bit_mode(bitmode_t bitmode) override;
    bool set_frame_rate(float frame_rate, double clock_factor) override;
    bool set_integration_time(uint32_t shutter_speed) override;
    void set_sync_delay(uint32_t delay_amount) override;
    void set_syncin_enable(uint32_t type) override;
    int set_sensor_gain(uint32_t new_sensor_gain) override;


    bool calibration_startup(bool is_factory, uint16_t startx, uint16_t starty,
                             uint16_t width, uint16_t height,
                             fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                             uint32_t integration_usecs, uint32_t sensor_opts, uint32_t sensor_gain);

    // Say whether this sensor supports video trigger.
    bool has_video_trigger() const noexcept override { return true; }

    // Control the video trigger module.
    uint32_t vtrig_control(const char *settings_str, int str_length) override;

    // Return the current video trigger settings and status, json encoded.
    uint32_t vtrig_info_json(char *dest_ptr, uint32_t max_len) override;

    int get_capture_fifo_count() override;
    void reset_capture_fifo() override;
    void reset_event_fifo() override;

    bool run_prnu(const uint8_t *ct_ptr) override;
    bool start_prnu_capture();
};

#endif // __LINCE5M_H__
