/*!
 * @file ltr_worker.h
 * @brief This file defines the background LTR mode working class.
 * @author Anthony L. Montiel
 * @date 2017-06-14
 *
 * Copyright (C) 2017 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _LTR_WORKER_H_
#define _LTR_WORKER_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <condition_variable>
#include <mutex>

#include <fastec/fas_thread.h>

//--------------------------------------------------------------------
//! @brief The LTR mode working class.
//--------------------------------------------------------------------
class ltr_worker : public FAS_THREAD::Thread
{
private:
    volatile bool thread_running;           //!< Flag to indicate if we are running or not.
    volatile int error_cond;                //!< An error indicator.

    std::mutex xfer_mutex;                  //!< Used to control changes in transfer status.
    std::condition_variable xfer_status;    //!< Condition variable we can block on.

    uint64_t last_im_address;               //!< The last image memory address we used from the FPGA.
    uint64_t last_fpga_written;             //!< The last image memory address we read from the FPGA.

    uint64_t ssd_start_offset;              //!< The starting offset on the SSD for the recording.
    uint64_t ssd_end_offset;                //!< One byte passed the last offset for the recording.
    off64_t ssd_pos;                        //!< "next" frame address on the SSD.

    int pci_fd;                             //!< File descriptor for the PCI BAR.
    int ssd_fd;                             //!< File descriptor for the SSD.
    int xfer_state;                         //!< Our current state.

    void *xfer_buffer;                      //!< Our transfer buffer.
    uint32_t xfer_len;                      //!< Number of 4K-aligned bytes in the transfer buffer.
    uint32_t leftover;                      //!< The slop over from the last 4K aligned write.
    uint32_t padding_count;                 //!< Number of bytes to pad the "wrapping" frame so it aligns to 4K.

    uint32_t af_level;                      //!< FIFO Almost Full level.
    uint32_t bar_length;                    //!< The size of our DMA BAR.

    bool need_to_splice;                    //!< True if we need to splice an image into an SSD page.
    volatile bool final_write;              //!< True when we are performing the last write out to the SSD.
    volatile bool flush_the_pipe;           //!< For ROC/BROC modes, when we need to flush out the captured frames to the SSD.

    FILE *stats_fp;                         //!< Used to save the LR mode statistics, if enabled. See LR_STATS usage.
    double delay_for_data;                  //!< Time waiting for the data to be ready.

private:
    void cleanup();
    void errorHandler();

    bool checkForEnoughData();
    bool saveImageData();
    bool setup();
    bool spliceAtEnd(uint8_t *write_ptr);
    bool spliceAtStart();
    ssize_t write(int fd, uint8_t *buf_ptr, ssize_t nbyte);

public:
    ltr_worker();
    virtual ~ltr_worker();

    void dataReady();
    constexpr uint64_t getCurAddr()     { return (ssd_pos + leftover); }
    constexpr void flushSSD()           { flush_the_pipe = true; }

    bool is_running()       { return thread_running; }

    virtual void run() override;
    virtual void stop() override;
};

#endif /* !_LTR_WORKER_H_ */

