/*!
 * @file luxima_hs.h
 * @brief This is the header file providing the hardware abstraction layer
 * for the Luxima sensors and its sensor-specific FPGA interface.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __LUXIMA_HS_H__
#define __LUXIMA_HS_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "sensor.h"

#include <memory>

namespace LUXIMA_DAC
{
//! @name Definitions for DACs.
//! @{

// Our devices' address.
#define DAC_U12_I2C_ADDR        0x10
#define DAC_U14_I2C_ADDR        0x11
#define DAC_GLOBAL_I2C_ADDR     0x73

// Global address for DACs on a chip.
#define DAC_GLOBAL      0x0F

// These are the address for the U12 DAC.
#define DAC_VREF        0   /* VOUTA */
#define DAC_VLNC        1   /* VOUTB */
#define DAC_VLN         2   /* VOUTC */
#define DAC_VADH        3   /* VOUTD */
#define DAC_VOFF        4   /* VOUTE */
#define DAC_VAD4        5   /* VOUTF */
#define DAC_VLNA        6   /* VOUTG */
#define DAC_VCAS        7   /* VOUTH */

// These are the address for the U14 DAC.
#define DAC_VMUX1       0   /* VOUTA */
#define DAC_VTX2L_VTXM      1   /* VOUTB */
#define DAC_VABL_TST        2   /* VOUTC */
#define DAC_VPIX_VRSTH      3   /* VOUTD */
#define DAC_VRSTH_VRSTL     4   /* VOUTE */
#define DAC_VTXH        5   /* VOUTF */
#define DAC_VTXL        6   /* VOUTG */
#define DAC_VMUX2       7   /* VOUTH */

//! @}

// Command codes.
enum COMMAND_CODES
{
    CC_WRITE_ONE = 0,       // Write to Input register n.
    CC_UPDATE_ONE,      // Update (power-up) DAC register n.
    CC_WRITE_ONE_UPDATE_ALL,    // Write to input register n, update all.
    CC_WRITE_UPDATE_ONE,    // Write to and update DAC register n.
    CC_POWERDOWN_ONE,       // Power down register n.
    CC_POWERDOWN_ALL,       // Power down chip.
    CC_SEL_INT_REF,     // Select internal reference.
    CC_SEL_EXT_REF      // Select external reference.
};

const int CC_SHIFT = 4;
const uint8_t CC_MASK = 0x0F;

typedef struct DAC_TRIPLES
{
    uint8_t chip;       // The chip address
    uint8_t reg;        // The register address.
    uint16_t val;       // The value to be written
} DAC_TRIPLES;

const int DT_SHIFT = 6;     // 10-bit values need to be left justified.

} /* End of namespace LUXIMA_DAC */

//--------------------------------------------------------------------
// Forward Delcarations
//--------------------------------------------------------------------
class FiFPGA;
class BadPixels;

//--------------------------------------------------------------------
// LuximaHS Class
//--------------------------------------------------------------------
class LuximaHS : public Sensor
{
public:
    /// Constructor and destructor
    explicit LuximaHS(uint16_t max_width, uint16_t max_height,
                      uint16_t min_width, uint16_t min_height,
                      SENSOR::EEPROM_INFO *eeprom);
    ~LuximaHS();

    virtual bool class_init(FiFPGA *fpga_ptr) override;
    virtual bool enable_sensor_test() override;

    bool force_legal_roi(unsigned factor, uint16_t *p_x, uint16_t *p_y, uint16_t *p_width,
                         uint16_t *p_height);
    virtual uint32_t  get_default_frame_rate() override;
    virtual bitmode_t get_default_bit_mode() override;
    virtual uint32_t  get_default_gain() override;
    virtual uint32_t  get_frame_rate() override;
    virtual std::string get_defect_filename(const char *dirname);
    virtual uint32_t  get_frame_overhead() override;
    virtual uint32_t  get_readout_delay();
    virtual uint32_t  get_sensor_min_shutter() override;
    virtual int       get_sensor_height_mod() override { return 4; }
    virtual double    get_temperature() override;

    // Get the maximum bit mode that this sensor supports.
    virtual bitmode_t get_max_bit_mode() override { return (bitmode_t)GIGE_BITMODE_10_BITS; }

    virtual int get_readout_delay_min();
    virtual int get_readout_delay_max();

    // Not all sensors use the FPGA's bad pixel correction in the IPM. Override if available.
    virtual void IPM_bad_pixel_correction(int port, uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height) override;
    virtual uint32_t IPM_bayer_pattern_start(uint16_t startx, uint16_t starty) override;
    virtual bool is_bit_mode_supported(bitmode_t bit_mode) override = 0;
    virtual bool is_option_supported(uint32_t options) override = 0;
    virtual bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                                   uint16_t *rtn_width, uint16_t *rtn_height,
                                   bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                   uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts) override;
    virtual bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                                fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                                uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) override;
    virtual bool sensor_startup(uint16_t start_x, uint16_t start_y, uint16_t width, uint16_t height,
                                fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                                uint32_t integration, uint32_t sensor_opts, uint32_t sensor_gain) override;

    virtual void set_readout_delay(int delay) { (void)delay; }
    virtual bool set_fpga_bit_mode(bitmode_t bitmode) override;
    virtual bool set_frame_rate(float frame_rate, double clock_factor) override;
    virtual bool set_integration_time(uint32_t shutter_speed) override;
    virtual void set_sync_delay(uint32_t delay_amount) override;
    virtual void set_syncin_enable(uint32_t type) override;
    virtual int set_sensor_gain(uint32_t new_sensor_gain) override;
    virtual bool set_sensor_readout(bool enable) override;

    virtual bool fpn_needs_restart(fpnmode_t old_fpn, fpnmode_t new_fpn) override;
    virtual uint16_t get_cal_offset(uint32_t sensor_type, void *params) override;
    virtual bool run_calibration(bool is_factory, bool use_manual, bool *rtn_restart) override;
    virtual bool run_prnu(const uint8_t *ct_ptr) override;
    virtual bool start_calibration_capture(bool factory, bitmode_t cal_bit_mode);
    virtual bool start_prnu_capture();

    virtual void shutdown() override;

protected:
    std::unique_ptr<BadPixels> bad_pixels;      //!< Bad pixel list for using FPGA correction in the IPM.
    uint32_t config_word;                       //!< Sensor configuration word.
    bool sensor_test_on;                        //!< Enable sensor test pattern. Use sensor_test_type to determine pattern.

    //! Pure virtual method, as each sensor has unique DAC values.
    virtual bool dac_init(FiFPGA *fpga_ptr) = 0;

    //! Pure virtual method, returns the sensor clock rate (HZ).
    virtual int sensor_clock_rate() const = 0;

    //! Pure virtual method to enable the "clpout" signal (the clamp for pixel driver).
    virtual void enable_clpout() = 0;

    //! Reset the sensor.
    virtual void hard_reset();

    //! Load bad pixel information.
    virtual void load_bad_pixels();

    //! Set the readout start.
    virtual void set_readout_start();

    //! The clock rate in microseconds.
    int sensor_clock_rate_usec() const { return (sensor_clock_rate() / 1000000); }
};

//--------------------------------------------------------------------
// Luxima13HS Class
//--------------------------------------------------------------------
class Luxima13HS : public LuximaHS
{
public:
    /// Constructor and destructor
    Luxima13HS(SENSOR::EEPROM_INFO *eeprom);
    ~Luxima13HS();

    virtual bool class_init(FiFPGA *fpga_ptr) override;
    virtual bool enable_sensor_test() override;
    virtual bool factory_needs_reset(bool /*manual_params*/) override { return false; }

    virtual uint32_t get_frame_overhead() override;
    virtual uint32_t get_readout_delay() override;
    virtual uint32_t get_sensor_min_shutter() override;

    virtual bool is_bit_mode_supported(bitmode_t bit_mode) override;
    virtual bool is_option_supported(uint32_t options) override;

    virtual bool set_fpga_bit_mode(bitmode_t bitmode) override;
    virtual bool set_frame_rate(float frame_rate, double clock_factor) override;
    virtual bool set_integration_time(uint32_t shutter_speed) override;
    virtual void set_readout_delay(int delay) override;
    virtual bool set_sensor_readout(bool enable) override;
    virtual void set_sync_delay(uint32_t delay_amount) override;
    virtual void set_syncin_enable(uint32_t type) override;
    virtual void shutdown() override;

    virtual bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                                   uint16_t *rtn_width, uint16_t *rtn_height,
                                   bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                   uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts) override;
    virtual bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                                fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                                uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) override;
protected:
    virtual bool dac_init(FiFPGA *fpga_ptr) override;
    virtual void enable_clpout() override;
    virtual void hard_reset() override;
    virtual int sensor_clock_rate() const override;
    virtual void set_readout_start() override;
};

//--------------------------------------------------------------------
// Luxima160 Class
//--------------------------------------------------------------------
class Luxima160 : public LuximaHS
{
public:
    /// Constructor and destructor
    Luxima160(SENSOR::EEPROM_INFO *eeprom);
    ~Luxima160();

    bool is_bit_mode_supported(bitmode_t bit_mode) override;
    bool is_option_supported(uint32_t options) override;

private:
    bool dac_init(FiFPGA *fpga_ptr) override;
    void enable_clpout() override;
    int sensor_clock_rate() const override;
};

//--------------------------------------------------------------------
// Luxima19HS Class
//--------------------------------------------------------------------
class Luxima19HS : public Luxima13HS
{
public:
    /// Constructor and destructor
    Luxima19HS(SENSOR::EEPROM_INFO *eeprom);
    ~Luxima19HS();

    bool class_init(FiFPGA *fpga_ptr) override;
    bool enable_sensor_test() override;

    uint32_t get_frame_overhead() override;
    uint32_t get_readout_delay() override;
    int get_readout_delay_min() override;
    int get_readout_delay_max() override;
    uint32_t get_sensor_min_shutter() override;

    bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                           uint16_t *rtn_width, uint16_t *rtn_height,
                           bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                           uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts) override;
    bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) override;

    bool set_integration_time(uint32_t shutter_speed) override;


    // Say whether this sensor supports video trigger.
    bool has_video_trigger() const noexcept override { return true; }

    // Control the video trigger module.
    uint32_t vtrig_control(const char *settings_str, int str_length) override;

    // Return the current video trigger settings and status, json encoded.
    uint32_t vtrig_info_json(char *dest_ptr, uint32_t max_len) override;

    int get_capture_fifo_count() override;
    void reset_capture_fifo() override;
    void reset_event_fifo() override;

private:
    void bit_slip();
    bool dac_init(FiFPGA *fpga_ptr) override;
    void enable_clpout() override;
    int sensor_clock_rate() const override;

private:
    uint32_t vtroi_x;
    uint32_t vtroi_y;
    uint32_t vtroi_width;
    uint32_t vtroi_height;
};

//--------------------------------------------------------------------
// Luxima2100 Class
//--------------------------------------------------------------------
class Luxima2100 : public LuximaHS
{
public:
    /// Constructor and destructor
    Luxima2100(SENSOR::EEPROM_INFO *eeprom);
    ~Luxima2100();

    bool class_init(FiFPGA *fpga_ptr) override;
    bool enable_sensor_test() override;

    uint32_t get_frame_overhead() override;
    uint32_t get_readout_delay() override;
    int get_readout_delay_min() override;
    int get_readout_delay_max() override;
    uint32_t get_sensor_min_shutter() override;

    void IPM_bad_pixel_correction(int port, uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height) override;
    bool is_bit_mode_supported(bitmode_t bit_mode) override;
    bool is_option_supported(uint32_t options) override;

    bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                           uint16_t *rtn_width, uint16_t *rtn_height,
                           bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                           uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts) override;
    bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) override;
    bool sensor_startup(uint16_t start_x, uint16_t start_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t integration, uint32_t sensor_opts, uint32_t sensor_gain) override;

    bool set_integration_time(uint32_t shutter_speed) override;
    int set_sensor_gain(uint32_t new_sensor_gain) override;

protected:
    void bit_slip();
    bool dac_init(FiFPGA *fpga_ptr) override;
    void enable_clpout() override;
    float get_max_frame_rate(int start_x, int start_y, int width, int height);
    void hard_reset() override;
    int sensor_clock_rate() const override;

    bool read_reg(uint8_t reg_addr, uint16_t *rtn_data);
    bool write_reg(uint8_t reg_addr, uint16_t reg_data);
    bool write_wavetable(const std::string& file_name);

    bool shutdown_sensor(FiFPGA *fpga_ptr);

private:
    uint32_t current_gain;      //!< The current gain setting.
};

#endif // __LUXIMA_HS_H__
