/*!
 * @file mft.h
 * @brief Interface for coordinating access to and using the micro
 *        four thirds lens.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __FI_MFT_H__
#define __FI_MFT_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <string>

namespace FI_MFT
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Lens information structure.
typedef struct LENS_INFO
{
    bool connected;             //!< True if we're connected to an MFT lens.
    bool powered;               //!< True if the lens has power.
    bool talking;               //!< True if we're communicating with the lens.

    int focal_len_wide;         //!< The closest zoom focal length. [ASK_REF_DATA_M_RESP.focal_length_wide]
    int focal_len_tele;         //!< The telephoto focal length. [ASK_REF_DATA_M_RESP.focal_length_tele]

    int focus_divisions;        //!< The number of focus steps. [ASK_REF_DATA_M_RESP.focus_divisions]
    int zoom_divisions;         //!< The number of zoom steps. [ASK_REF_DATA_M_RESP.zoom_divisions]
    bool ez_enable;             //!< True if electric zoom is enabled. [ASK_REF_DATA_M_RESP.lens_attribute_data.LAD_ELECTRIC_ZOOM]

    std::string name;           //!< Name of the lens. [ASK_REF_DATA_M_RESP.lens_name]

} LENS_INFO;

//! Lens status structure.
typedef struct LENS_STATUS
{
    bool connected;             //!< True if we're connected to an MFT lens.
    bool powered;               //!< True if the lens has power.
    bool talking;               //!< True if we're communicating with the lens.

    uint16_t apex;              //!< The current aperture setting.
    float f_number;             //!< The current f-stop/f-number.
    float f_minimum;            //!< Minimum f-stop given the current zoom.
    float f_maximum;            //!< Maximum f-stop given the current zoom.

    int object_distance;        //!< Currently calculated object distance, in millimeters. -1 means infinity.
    int focus_pos;              //!< Currently selected focal position.
    int focus_min;              //!< Closest photography distance, given the current zoom setting.
    int focus_near;             //!< Closest distance absolute pulse.
    int focus_inf;              //!< Infinite distance absolute pulse.
    bool manual_focus;          //!< Are we in manual focus mode?

    int zoom_pos;               //!< Currently selected zoom encoder position.
    int focal_len;              //!< Focal length, given the current zoom encoder position.
} LENS_STATUS;

} /* End of namespace FI_MFT */

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

extern void mft_init();
extern void mft_close();
extern void mft_irq(uint32_t irq_status);

extern void mft_dump(FILE *dump_stream);
extern bool mft_info_request(FI_MFT::LENS_INFO& request);
extern uint32_t mft_info_json(char *dest_ptr, uint32_t max_len);
extern bool mft_status_request(FI_MFT::LENS_STATUS& request, bool wait = false);
extern uint32_t mft_status_json(char *dest_ptr, uint32_t max_len);

extern bool mft_set_aperture(float f_number);
extern bool mft_set_focus(uint32_t focus_dist);
extern bool mft_set_power(bool enable);
extern bool mft_set_zoom(int index);

extern void mft_autofocus();

#endif /* !__FI_MFT_H__ */
