/*!
 * @file null_sensor.h
 * @brief This is the header file providing the hardware abstraction layer
 * for a non-existent sensor so the camera can boot without one.
 *
 * Copyright 2010, 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef NULL_SENSOR_H
#define NULL_SENSOR_H

#include "sensor.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

class FiFPGA;

//! Actual definition of the base class
class NullSensor : public Sensor
{
private:
    static const uint32_t FOT_OVERHEAD = 6;       // Default frame overhead in microseconds.

public:
    /// Constructor and destructor
    NullSensor(SENSOR::EEPROM_INFO *eeprom);
    ~NullSensor();
    bool class_init(FiFPGA *fpga_ptr);

    uint32_t  get_default_frame_rate(void);
    bitmode_t get_default_bit_mode(void);
    uint32_t  get_default_gain(void);
    uint32_t  get_frame_rate();

    // Get the maximum bit mode that this sensor supports.
    bitmode_t get_max_bit_mode(void) { return (bitmode_t)GIGE_BITMODE_7_0; }
    uint32_t  get_sensor_min_shutter(void) { return 0; }

    // Return the extra time, not reflected in the frame rate, that it takes to complete a frame.
    uint32_t get_frame_overhead(void) { return FOT_OVERHEAD; }

    uint32_t IPM_bayer_pattern_start(uint16_t startx, uint16_t starty);
    bool is_bit_mode_supported(bitmode_t bit_mode);
    bool is_option_supported(uint32_t options);
    bool sensor_startup(uint16_t start_x, uint16_t start_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain);
    bool set_fpga_bit_mode(bitmode_t bit_mode);
    bool set_frame_rate(float frame_rate, double clock_factor) { (void)frame_rate; (void)clock_factor; return true; }
    bool set_integration_time(uint32_t shutter_speed) { (void)shutter_speed; return true; }
    void set_sync_delay(uint32_t delay_amount) { (void)delay_amount; }
    void set_syncin_enable(uint32_t type) { (void)type; }
    int set_sensor_gain(uint32_t new_sensor_gain);
    bool set_sensor_readout(bool enable);
    bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                        fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                        uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain);
};


//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

#endif // NULL_SENSOR
