/*!
 * @file part_mgr.h
 * @brief This is the header file for the partition manager.
 *
 * Copyright 2010 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PARTMGR_H_
#define PARTMGR_H_

//--------------------------------------------------------------------
// Include files
//--------------------------------------------------------------------

#include <memory>
#include <mutex>
#include <vector>

#include <fastec/fas_filesystem.h>
#include <fastec/gige_msg.h>
#include <fastec/fi_lutfile.h>

#include "cs_utils.h"
#include "lic_manager.h"
#include "sensor.h"

namespace PART_MGR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Define to display the per-frame meadata.
// #define PRINT_PFMD   1

// Developer note: The primary difference between const and constexpr variables is that
// the initialization of a const variable can be deferred until run time. A constexpr
// variable must be initialized at compile time.
constexpr int NUM_IPM_PP_BUFFERS            = 2;                    //!< Number of ping pong buffers for the IPM channels.
constexpr int NUM_ULV_BUFFERS               = 1;                    //!< Number of unarmed live view buffers.
constexpr int PME_BUFFER_SIZE               = (5 * 1024 * 1024);    //!< Size of the extraction buffer.
constexpr uint32_t PARTITION_METADATA_SIZE  = 4096;                 //!< Bytes at the begining of each partition for the metadata table.
constexpr uint32_t SESSION_INCR             = (256 * 1024 * 1024);  //!< 1/4 of a GiB.

//! The various trace levels for the partition manager logic. We put this information
//! in the header because the partition manager object itself is split across a number
//! of files, all of which need access to these defines.
enum PART_MGR_TRACE
{
    PM_FUNC     = 0x00000010,
    PM_FPGA     = 0x00000020,
    PM_MEMORY   = 0x00000040,
    PM_LTR      = 0x00000080,
    PM_PFMD     = 0x00000100,
    PM_STATS    = 0x00000200,
    PM_VERBOSE  = 0x00000400
};

} /* End of namespace PART_MGR */

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------

namespace PART_MGR_WORKER
{
struct PartInfo;
}

//--------------------------------------------------------------------

/*!
 * @brief The partition manager object is responsible for managing the partitions
 * in Image Memory captures, as well as the Long Record captures.
 *
 * The partition manager is a "singleton" object. Access to the object is done via the static
 * PartMgr::getInstance() method.
 *
 * <table>
 *      <caption id="PartMgrGlossary">Glossary</caption>
 *      <tr><th>Name <th>Description
 *      <tr><td>Capture <td>A "capture" is a recording of images. It can be in image memory or on SSD.
 *      <tr><td>LR <td>Long Record mode recordings are writing images directly to the SSD.
 *      <tr><td>LTR <td>Synonymous with LR. Long Term Recording, or Long Time Recording.
 *      <tr><td>Partition <td>A "partition" is a recording in Image Memory.
 *      <tr><td>Session <td>Anytime you see "session", this is referencing an LR mode recording.
 * </table>
 */
class PartMgr
{
private:
    static std::unique_ptr<PartMgr> instance;       //!< Our singleton instance.
    static std::mutex instance_mutex;               //!< Guard access to the singleton.

    //! The maximum depth of the IM2SSD descriptor FIFO.
    static constexpr int MAX_IM2SSD_DESCRIPTOR_FIFO_DEPTH = 4096;


    bool        gainTableIsAllowed;                 //!< Flag indicating if we support a gain table (PRNU).
    uint32_t    traceControl;                       //!< Trace control word for PartMgr class.

    //! @name Capture space information.
    //! @{
    uint64_t    imageMemorySize;                    //!< Size of the image memory, in bytes.
    uint64_t    imageMemorySizeReserved;            //!< Amount of image memory reserved for non capture purposes, such as metadata, buffers, etc.
    uint64_t    partitionSize;                      //!< How big is each partition.
    uint32_t    partitionSizeReserved;              //!< How much memory is reserved in each partition for metadata, etc.
    uint32_t    sessionLength;                      //!< How big is a LR mode session, in SESSION_INCR increments.
    uint32_t    sessionLengthMax;                   //!< Maximum session length, in SESSION_INCR "chunks".

    // The size of a captured image in image memory changes when the ROI or bits per pixel changes.
    uint32_t    imageBufferSize;                    //!< Size of the current capture buffer per frame (i.e. frame size).

    // The number of image buffers changes when either the partition size changes or the
    // capture frame size changes.
    uint32_t    imageBufferCount;                   //!< Number of image capture buffers in a partition.

    // The number of capture buffers for LTR can vary as well so we minimize our chance
    // of buffer overflow. Images are captured to buffers in image memory first, then read
    // from there and written to the SSD. The maximum number is limited by the size of
    // the IM2SSD module descriptor FIFO.
    uint32_t    ltrImageBufferCount;                //!< Number of image memory capture buffers in LR mode.
    //
    //! @}

    //! @name Buffer information
    //! @{
    uint32_t    correctionTableSize;                //!< Correction (and gain) table size. The table itself immediately follows the metadata.
    uint64_t    ulvBufferAddr;                      //!< Unarmed live view buffer address base.
    uint32_t    ulvBufferSize;                      //!< Size of the unarmed live view buffer.
    uint64_t    ipmOutputBufferAddress;             //!< Address of the first buffer to hold the output of the IPM.
    uint32_t    ipmOutputBufferSize;                //!< Size of each of the IPM output buffers.
    uint64_t    pmeBufferAddr;                      //!< Extraction buffer for per-frame metadata.
    uint32_t    pmeBufferSize;                      //!< Size of the extraction buffer.
    //
    //! @}

    uint32_t    pfmdFrameNum;                       //!< The frame number where the per-frame metadata is that we will want to read.
    uint32_t    triggerReqPos;                      //!< The trigger request position. can be % or frame based.
    int fflr_trigger_frame;                         //!< The trigger frame relative to the recording space.
    int trigger_part;                               //!< The partition that had the trigger event.
    char capture_comment[GIGE_COMMENT_MAX_LEN];         //!< User-supplied comment saved with capture metadata.
    struct spline_knot knots[LUT_MAX_KNOT_COUNT * 3];   //!< User LUT knots are saved with capture. Even if we're mono, we have space for R, G, and B.
    uint64_t    finished_part_count;                //!< The number of partitions completed.

    // The LTR Filesystem information.
    FAS_FILESYSTEM::PARTITION_INFO_VEC ltr_info;

    // The LTR volume name.
    std::string ltr_volume_name;

private:
    PartMgr();

    uint64_t convertFrameNumToAddr(int part_num, uint32_t frame_num, bool absolute = false);
    uint32_t convertToAbsFrameNum(int part_num, uint32_t frame_num, bool absolute = false);

    bool    layoutMemory(uint16_t width, uint16_t height, bitmode_t bit_mode, uint32_t sess_len, uint32_t num_parts);

    int     calcSessionCount(uint32_t sess_length);
    void    cleanup_metadata_version(pm_metadata *meta_ptr);
    bool    readImMetadata(int part_num);
    bool    set_next_partition(Sensor *p_sensor);
    bool    updateMetadataSettings(int part_num, const pm_metadata *meta_ptr);
    uint32_t computePostTrigCount(uint32_t cap_count, uint32_t requested_pos);
    constexpr uint64_t fmetaTime(uint32_t *metadata_buf);

    //! @name LTR support functions
    //! @{
    void ltrDisplayPartInfo(const FAS_FILESYSTEM::PARTITION_INFO_VEC& part_info);
    bool ltrArmSetup(uint32_t recording_id, uint32_t mode);
    uint32_t getLtrRecordingID();
    bool ltrLayoutDrive(bool really_write, uint64_t num_sectors, uint32_t max_ct_size, const std::string& volume_label);
    bool ltrRocArmAppend();
    bool ltrUpdateFromPartMetadata();
    uint32_t ltrRecoverCorrectionTable(uint8_t *ct_ptr, uint32_t ct_bufsize);
    uint32_t ltrRecoverGainTable(uint8_t *gt_ptr, uint32_t gt_bufsize);
    bool ltrRecoverPartMetadata(pm_metadata *pm_ptr);
    void ltrSetCaptureBufferCount(uint32_t buf_count);
    bool ltrSetupCaptureHD(uint32_t recording_id, uint32_t frame_size, uint32_t frame_count, uint32_t pm_bytes, uint32_t ct_bytes);
    bool ltrStartupAnalyze();
    bool ltrNilCapture();
    bool ltrWriteCaptureTables(const pm_metadata *pm_ptr, uint64_t ct_addr, uint32_t ct_bytes, uint64_t gt_addr);

    bool ltrFinalizePartitionDone(uint16_t part_num, uint32_t num_frames, uint32_t first_frame_num, uint32_t trig_frame_num,
                                  uint32_t trig_s, uint32_t trig_ns);

    bool ltrFinalizeRocPartitionDone(uint16_t part_num, uint32_t num_frames);
    void ltrPartitionEventRocDone(uint32_t mode, uint64_t last_frame_addr);
    bool ltrSimulateRocCompletion(uint32_t mode, uint64_t last_frame_addr);

    std::string flagsToStr(uint8_t flags);
    std::string stateToStr(uint8_t state);
    std::string typeToStr(uint32_t type);
    //
    //! @}

public:
    ~PartMgr();

    // Singletons should not be cloneable.
    PartMgr(PartMgr& other) = delete;

    // Singletons should not be assignable.
    void operator=(const PartMgr&) = delete;

    static PartMgr *getInstance();
    static int setTriggerPos(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

    bool classInit(FiFPGA *fpga_ptr, Sensor *sensor_ptr, const LicManager *p_lic);
    bool layoutMemoryFromRegs();
    void resetFflr();
    void shutdown();

    int32_t convertCaptureAddrToFrameNum(uint64_t frame_addr);
    void getCalibrationBuf(uint64_t *rtn_addr, uint32_t *rtn_size, uint32_t *rtn_count);
    uint64_t getCaptureFrameAddr(uint32_t frame_num);
    uint32_t get_im_buf_count() { return imageBufferCount; }

    bool    arm();
    int32_t convertAddrToFrameNum(uint64_t frame_addr);
    bool    simulateCompletion(bool triggered, uint32_t post_trigger_frames,
                               uint64_t last_frame_addr, uint32_t icm_count,
                               uint32_t *trig_info);
    void    partitionEventTrigger(uint64_t trigger_frame, const uint32_t *extra_info);
    void    partitionEventDone(uint64_t last_frame_addr, uint32_t post_trigger_frames,
                               uint32_t icm_count);
    void    partitionMetadataInit(pm_metadata *meta_ptr, uint8_t trigger_setting, uint32_t trigger_frames);
    bool    useLoadedPartitionMetadata(const pm_metadata *meta_ptr);

    constexpr bool isGainTableAllowed()     { return gainTableIsAllowed; }
    void reserveGainTableSpace(bool reserve_space);
    bool writeCorrectionTable(const uint8_t *table_ptr, uint32_t table_size);
    bool writeGainTable(uint8_t *table_ptr, uint32_t table_size);

    // Info on memory available
    constexpr uint32_t get_im_size()            { return imageMemorySize; }
    constexpr uint32_t get_im_size_reserved()   { return imageMemorySizeReserved; }
    uint32_t    get_im_size_qw()        { return convert_to_QWORD(imageMemorySize); }
    uint32_t    get_im_size_reserved_qw()   { return convert_to_QWORD(imageMemorySizeReserved); }

    uint64_t    get_ulv_buf_addr()      { return ulvBufferAddr; }
    uint32_t    get_ulv_buf_size()      { return ulvBufferSize; }

    uint64_t    get_pme_buf_addr()      { return pmeBufferAddr; }
    uint32_t    get_pme_buf_size()      { return pmeBufferSize; }
    int     pme_fpga_fill_buffer(uint32_t start_frame, uint32_t end_frame, int bb_index);
    int     pme_ltr_fill_buffer(uint32_t start_frame, uint32_t end_frame, uint8_t *pfmd_buffer);

    // Find out information about partitions
    uint64_t    get_active_ct_addr();
    uint32_t    get_active_ct_addr_qw();
    uint64_t    get_active_gt_addr();
    uint64_t    get_active_frame_addr(uint32_t frame_num, bool absolute = false) { return convertFrameNumToAddr(getActivePartition(), frame_num, absolute); }
    uint32_t    get_active_frame_num()      { return pfmdFrameNum; }
    int32_t     getActivePartition(PART_MGR_WORKER::PartInfo *part_ptr = nullptr);
    int32_t     getCurrentFrameNum();
    uint32_t    getTriggerFrame();
    int32_t     getTriggerPart()            { return trigger_part; }
    uint32_t    getNumFrames();
    uint64_t    getPartitionStart();
    uint8_t     getPartitionState();

    std::unique_ptr<pm_metadata> getPartitionMetadata(int part_num);
    std::unique_ptr<pm_metadata> getActivePartMetadata()            { return getPartitionMetadata(getActivePartition()); }
    bool    readActiveImMetadata()                                  { return readImMetadata(getActivePartition()); }
    bool    updateActiveMetadata(const pm_metadata *meta_ptr = nullptr) { return updateMetadataSettings(getActivePartition(), meta_ptr); }
    bool    updatePartitionMetadata(int part_num)                   { return updateMetadataSettings(part_num, nullptr); }

    void    get_comment(char *buf, unsigned buf_len);
    void    set_comment(const char *buf);
    void    set_knots(const LUT_RGB_FILE *lut_ptr);

    int getFramePFMD(uint32_t frame_num, uint8_t *dest_ptr);
    int getActiveFramePFMD(uint8_t *dest_ptr)               { return getFramePFMD(pfmdFrameNum, dest_ptr); }
    uint32_t getFrameCount();
    uint32_t getFrameSize_QW()                              { return convert_to_QWORD(imageBufferSize); }
    uint32_t getNumPartitions();
    uint32_t getPmSize();
    uint64_t getPartitionFrameAddr(int part_num, uint32_t frame_num, bool absolute = false)   { return convertFrameNumToAddr(part_num, frame_num, absolute); }
    uint32_t getPartitionNumFrames(int part_num);

    constexpr uint32_t getMaxPartitions()                   { return (uint32_t)0x1000; }
    constexpr uint32_t getFrameSize()                       { return imageBufferSize; }
    constexpr uint64_t getPartitionSize()                   { return partitionSize; }
    constexpr uint32_t getPartitionSizeReserved()           { return partitionSizeReserved; }

    constexpr uint32_t getSessionLengthMax()                { return sessionLengthMax; }
    constexpr uint32_t getSessionLength()                   { return sessionLength; }

    constexpr uint32_t getIpmOutputBufferSize()             { return ipmOutputBufferSize; }
    constexpr uint64_t getIpmOutputBufferAddr(int which)    { return ipmOutputBufferAddress + (which * ipmOutputBufferSize); }

    bool adjustReviewPartition(uint32_t adj_length, uint32_t num_frames = 0);
    bool isReviewPartWrapped();
    bool setReviewPartNum(int part_num);

    uint64_t    getReviewPartLength(uint32_t *capture_session);
    uint32_t    getReviewPartTrigger();
    uint32_t    getReviewPartNumFrames();
    uint64_t    getReviewPartStartAddr();
    int32_t     getReviewPartition(PART_MGR_WORKER::PartInfo *part_ptr = nullptr);
    std::unique_ptr<pm_metadata> getReviewPartMetadata()            { return getPartitionMetadata(getReviewPartition()); }

    bool setActiveFrameNum(uint32_t frame_num);
    bool setActivePartition(int part_num);
    void setPartitionState(int part_num, uint32_t part_state);
    void queueCapturePartition(int part_num);

    // Access to a partition's information.
    bool getPartInfo(int part_index, PART_MGR_WORKER::PartInfo& part_info);
    bool updatePartInfo(int part_index, const PART_MGR_WORKER::PartInfo& part_info);

    constexpr uint64_t getFinishedPartCount()       { return finished_part_count; }

    //! @name LTR support functions.
    //! @{
    int ltrReadBytes(uint8_t *dest_ptr, uint64_t src_offset, int byte_count, int fd = -1);

    bool ltrStartup();
    bool ltrFormat(bool really_write, const std::string& volume_label);
    bool ltrClearPartition();
    bool ltrTruncateFrom(uint32_t free_from_frame);
    bool ltrUpdateCapture(FAS_FILESYSTEM::CAPTURE_DATA *cap_info, const pm_metadata *pm_ptr, uint64_t ct_addr, uint32_t ct_bytes, uint64_t gt_addr);

    uint32_t *ltrGetPFMD(uint32_t *sector_buf, uint32_t abs_frame);
    uint32_t ltrReadPFMD(int part_num, uint32_t start_frame, uint32_t end_frame, uint8_t *buf, uint32_t buf_size, bool use_hexwords = false);

    bool ltrIsRocFull();
    uint64_t ltrGetCaptureBufferAddr(uint32_t *rtn_buf_count);
    uint64_t ltrGetActiveStartLba();
    uint64_t ltrGetActiveImageSectors();
    uint32_t ltrGetSession();
    uint64_t ltrGetSessionSize();
    uint64_t ltrGetSessionStart(uint32_t session);
    uint64_t ltrGetActiveStartFMetaLBA();
    uint32_t ltrGetActiveFrameCount();
    uint64_t ltrGetCaptureLBA();
    FAS_FILESYSTEM::LR_CAPTURE_HDR *ltrGetCaptureHeader();
    FAS_FILESYSTEM::CAPTURE_DATA *ltrGetCaptureData();
    uint64_t ltrGetRocStart();
    uint32_t ltrGetRocFrameCount();

    constexpr std::string& ltrGetVolumeName()       { return ltr_volume_name; }
    //
    //! @}
};

#endif // PARTMGR_H_
