/*!
 * @file pb_control.h
 * @brief Playback control thread. Several sources can request a
 *    playback (or single step) while in REVIEW mode. Those
 *    requests are handled by our control thread.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef __PB_CONTROL_H__
#define __PB_CONTROL_H__

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include "cs_network.h"

namespace PB_CONTROL
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//-------------------------------------
// PB_CONTROL register bit definitions
//-------------------------------------
#define PBC_ENABLE          (1 << 0)
#define PBC_FORWARD         (0 << 1)
#define PBC_REVERSE         (1 << 1)
#define PBC_LTR_MODE            (1 << 2)
#define PBC_SINGLE_START        (1 << 4)
#define PBC_SINGLE_ABORT        (1 << 5)
#define PBC_INTERVAL_SHIFT      16
#define PBC_INTERVAL_MASK       0xFFF
#define PBC_INTERVAL(x)         (((x) & PBC_INTERVAL_MASK) << PBC_INTERVAL_SHIFT )

//! Organize all the input and output parameters for the request for
//! a playback frame.
typedef struct FRAME_REQUEST
{
    //! @name Input/Output values.
    //! @{

    //! Pointer to the JPEG encoded frame. If the consumer passes a
    //! non-nullptr value, that address will be used to store the frame
    //! data. Otherwise, if a nullptr is passed, the playback controller
    //! will allocate a buffer and it is the responsibility of the
    //! consumer to call pbc_release().
    uint8_t *frame_buffer;

    //! The absolute frame number. 0 is at "partition_start" and increments
    //! from there. If you supply a number from 0..N, the request will only
    //! be satisfied with a different frame number. A -1 means to return
    //! the most recent frame.
    int frame_num;
    //! @}

    //! @name Output values.
    //! @{

    //! The number of bytes in the JPEG encoded frame. Note that this
    //! may be zero on error.
    int frame_len;

    //! @}
} FRAME_REQUEST;

} /* End of namespace PB_CONTROL */

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

extern void pbc_init();
extern void pbc_close();

extern bool pbc_busy();
extern void pbc_dump(std::ostream& dump_stream);
extern bool pbc_request(PB_CONTROL::FRAME_REQUEST *request);
extern void pbc_release_frame(uint8_t *frame_buffer);

extern int pbc_update_control(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
extern int pbc_update_frame_period(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
extern int pbc_update_cut_in(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
extern int pbc_update_cut_out(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);
extern int pbc_update_start_frame(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

#endif /* !__PB_CONTROL_H__ */
