/*!
 * @file pm_worker.h
 * @brief This file defines interface to the Partition Manager's worker class, pmWorker.
 * @author Anthony L. Montiel
 * @date 2022-03-25
 *
 * Copyright (C) 2022 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PM_WORKER_H_
#define PM_WORKER_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <mutex>

#include <fastec/fas_log.h>
#include <fastec/fas_thread.h>
#include <fastec/gige_msg.h>

namespace PART_MGR_WORKER
{
//--------------------------------------------------------------------
// Constant, Typedef and Macros
//--------------------------------------------------------------------

//! Structure used to encapsulate the information for a partition.
struct PartInfo
{
    uint32_t    frame_size;         //!< How big is a frame in this partition.
    uint64_t    partition_start;    //!< Byte address of where the partition starts.
    uint64_t    capture_buf_start;  //!< Address of where the capture buffer starts.
    uint8_t     partition_state;    //!< What state is the partition (see gige_msg.h).
    bool        is_trig_part;       //!< True if this partition holds the trigger frame.
    uint32_t    trig_frame_num;     //!< Trigger frame number for each partition.
    uint32_t    first_frame_num;    //!< First frame number for each partition.
    uint32_t    num_frames;         //!< How many frames are recorded in each partition.
    pm_metadata metadata;           //!< Copy of metadata that we also keep in image memory.
};

//! Message types for the worker operations.
enum PMW_MESSAGE_TYPES
{
    PMW_INITIALIZE_MESSAGE = 0x30000,   //!< Initialize the partition information.
    PMW_GET_INFO_MESSAGE,               //!< Read the partition information for a specified partition.
    PMW_SET_INFO_MESSAGE,               //!< Set the partition information for a specified partition.
    PMW_GET_ACTIVE_PART_MESSAGE,        //!< Get the active partition.
    PMW_SET_ACTIVE_PART_MESSAGE,        //!< Set the active partition.
    PMW_GET_REVIEW_PART_MESSAGE,        //!< Get the review partition.
    PMW_SET_REVIEW_PART_MESSAGE,        //!< Set the review partition.
};

//! Message class for PMW_INITIALIZE_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwInitMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwInitMessage();
    virtual ~pmwInitMessage();

    int num_parts;         //!< [IN] The number of partitions we will support.
};

//! Message class for PMW_GET_INFO_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwGetInfoMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwGetInfoMessage();
    virtual ~pmwGetInfoMessage();

    int part_index;         //!< [IN] The partition we want to work with.
    PartInfo info;          //!< [OUT] The information we requested.
};

//! Message class for PMW_SET_INFO_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwSetInfoMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwSetInfoMessage();
    virtual ~pmwSetInfoMessage();

    int part_index;         //!< [IN] The partition we want to work with.
    PartInfo info;          //!< [IN] The updated information we want to save.
};

//! Message class for PMW_GET_ACTIVE_PART_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwGetActivePartMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwGetActivePartMessage();
    virtual ~pmwGetActivePartMessage();

    int active_index;       //!< [OUT] The active partition index.
    PartInfo active_info;   //!< [OUT] The information associated with the active partition.
};

//! Message class for PMW_SET_ACTIVE_PART_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwSetActivePartMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwSetActivePartMessage();
    virtual ~pmwSetActivePartMessage();

    int active_index;       //!< [IN] The new active partition index.
};

//! Message class for PMW_GET_REVIEW_PART_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwGetReviewPartMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwGetReviewPartMessage();
    virtual ~pmwGetReviewPartMessage();

    int review_index;       //!< [OUT] The review partition index.
    PartInfo review_info;   //!< [OUT] The information associated with the review partition.
};

//! Message class for PMW_SET_REVIEW_PART_MESSAGE messages.
//! Note that this is a SYNC message (i.e. blocking).
class pmwSetReviewPartMessage : public FAS_THREAD::SyncMessage
{
public:
    pmwSetReviewPartMessage();
    virtual ~pmwSetReviewPartMessage();

    int review_index;       //!< [IN] The new review partition index.
};

//--------------------------------------------------------------------
// Class forward declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// fpgaIOWorker class declarations
//--------------------------------------------------------------------
class pmWorker : public FAS_THREAD::MsgThread
{
private:
    static std::unique_ptr<pmWorker> worker_proc;   //!< Our singleton object.
    static std::mutex worker_mutex;                 //!< Mutex to protect our singleton object.

    int reviewPartition;                            //!< The partition which is being reviewed, processed, etc.
    int activePartition;                            //!< The partition which is being recorded into, -1 means none.

    // This array holds information about each of the partitions.
    std::vector<PART_MGR_WORKER::PartInfo> part_info;

    // Our constructor.
    pmWorker();

    //! @name PartInfo messsage handlers.
    //! @{
    void handleInitRequest(pmwInitMessage *msg);
    void handleGetInfoRequest(pmwGetInfoMessage *msg);
    void handleSetInfoRequest(pmwSetInfoMessage *msg);
    void handleGetActiveRequest(pmwGetActivePartMessage *msg);
    void handleSetActiveRequest(pmwSetActivePartMessage *msg);
    void handleGetReviewRequest(pmwGetReviewPartMessage *msg);
    void handleSetReviewRequest(pmwSetReviewPartMessage *msg);
    //
    //! @}

public:
    virtual ~pmWorker();
    virtual void messageHandle(FAS_THREAD::Message *message) override;

    // Singletons should not be cloneable.
    pmWorker(pmWorker& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const pmWorker&) = delete;

    static pmWorker *factory();
    static void close()
    {
        pmWorker *thread_ptr;

        // Release our singleton.
        thread_ptr = pmWorker::factory();
        if (thread_ptr != nullptr)
        {
            // Tell the thread that we are stopping.
            FAS_THREAD::SyncMessage msg;
            msg.req = FAS_THREAD::HALT;

            // Pass it on.
            msg.doSync(thread_ptr);

            // Try to join the thread.
            thread_ptr->join();
        }

        LOG("%s--", __PRETTY_FUNCTION__);
    }

    static bool isRunning()
    {
        std::lock_guard<std::mutex> lock(worker_mutex);
        return (worker_proc.get() != nullptr);
    }

    // Return the number of partitions we are working with.
    int getNumParts()       { return (int)part_info.size(); }
};
} /* End of namespace FI_FPGA_IO */

#endif /* !PM_WORKER_H_ */

