/*!
 * @file sensor.h
 * @brief This is the header file for the base class sensor. It provides the
 *  hardware abstraction layer for interfacing with a specific sensor and
 *  the FPGA. Any specific sensors should derive from this class.
 *
 * Copyright 2010-2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef SENSOR_H
#define SENSOR_H

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include <memory>

#include <fastec/gige_msg.h>
#include <fastec/fi_sensor_types.h>

#include "camera_server.h"

namespace SENSOR
{
//--------------------------------------------------------------------
// Constant and Type Definitions
//--------------------------------------------------------------------

// Part number defines.
const int FI_PART_NO_LEN = 16;

// Fabrication date length (YYWW).
const int FI_FAB_DATE_LEN = 4;

// Assembly date length (YYMMDD).
const int FI_ASSY_DATE_LEN = 6;

// Sensor board serial number.
const int FI_SERIALNO_LEN = 4;

// Padding bytes.
const int FI_PADDING_LEN = 6;

// Maximum number of bad pixels we can correct.
const int MAX_BAD_PIXELS = 32;

typedef struct SENSOR_BAD_PIXEL
{
    uint16_t    x;               //!< X coord of bad pixel
    uint16_t    y;               //!< Y coord of bad pixel
} SENSOR_BAD_PIXEL;

// Manufacturer codes for EEPROM data.
const uint32_t LUPA_SENSORS = 0;
const uint32_t LINCE_SENSORS = 1;
const uint32_t LUXIMA_SENSORS = 2;

//! This is an "overlay" of the information stored in the sensor board EEPROM.
typedef struct EEPROM_INFO
{
    uint32_t size;                                  //!< How big is this structure?
    uint32_t sensor_type;                           //!< The device tag.
    uint32_t manufacturer;                          //!< The manufacturer ID code.
    uint8_t part_no[ SENSOR::FI_PART_NO_LEN ];      //!< The board part number.
    uint8_t fab_date[ SENSOR::FI_FAB_DATE_LEN ];    //!< The fabrication date (YYWW).
    uint8_t assy_date[ SENSOR::FI_ASSY_DATE_LEN];   //!< The assembly date (YYMMDD).
    uint8_t serial_no[ SENSOR::FI_SERIALNO_LEN ];   //!< Serial number.
    uint8_t padding[ SENSOR::FI_PADDING_LEN ];
    SENSOR::SENSOR_BAD_PIXEL    bad_pixels[SENSOR::MAX_BAD_PIXELS];     //!< array of bad pixels
} EEPROM_INFO;

//! A sensor ROI
typedef struct SENSOR_ROI
{
    uint16_t x;
    uint16_t y;
    uint16_t width;
    uint16_t height;
} SENSOR_ROI;

} /* End of namespace SENSOR */

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------

//! Actual definition of the base class
class Sensor
{
protected:
    // Characteristics of the sensor
    bool    sensor_is_color;    //!< Whether this is a color sensor
    uint16_t    sensor_max_width;   //!< Width of the sensor in pixels
    uint16_t    sensor_max_height;  //!< Height of the sensor in pixels
    uint16_t    sensor_min_width;   //!< Minimum width ROI of the sensor in pixels
    uint16_t    sensor_min_height;  //!< Minimum height ROI of the sensor in pixels

    bool    sensor_test_on;     //!< True if sensor test mode is on, if supported.
    SENSOR::EEPROM_INFO eeprom;

    static const char *DEFECT_MAP_FILE_FMT;

public:
    explicit Sensor(uint16_t max_width, uint16_t max_height,
                    uint16_t min_width, uint16_t min_height,
                    SENSOR::EEPROM_INFO *eeprom);
    virtual ~Sensor();
    virtual bool class_init(FiFPGA *fpga_ptr) = 0;

    // Sensor object factory to create an instance corresponding to the installed sensor.
    static std::unique_ptr<Sensor> create_sensor();

    // Get the type of sensor we are. 'L13C' or 'L13M' for Lupa1300
    static uint32_t get_sensor_type();

    int get_sensor_ver(char *sensor_ver, int buflen);

    // Is this a color sensor?
    bool is_sensor_color() { return sensor_is_color; }

    // Return the sensor characteristics. These are set at startup.
    uint16_t get_sensor_max_width()  { return sensor_max_width; }
    uint16_t get_sensor_max_height() { return sensor_max_height; }
    uint16_t get_sensor_min_width()  { return sensor_min_width; }
    uint16_t get_sensor_min_height() { return sensor_min_height; }

    virtual int get_sensor_height_mod() { return 2; }
    virtual int get_sensor_width_mod() { return 4; }

    // The black reference name can have sensor settings encoded into it as
    // an optimization so we can keep multiple black reference files around.
    // What settings are important depend on the sensor.
    virtual const char *get_blkref_name();

    // Not all sensors have CRC protection on their data interface.
    // Override if the function exists and return true if there are errors.
    virtual bool icm_has_crc_errors() { return false; }

    // Not all sensors have a realignment function. Override if it exists. Return true for success.
    virtual bool icm_realign() { return true; }

    // Say whether this sensor supports a factory black reference.
    virtual bool has_factory_blkref() const noexcept  { return true; }

    // Not all sensors support raw histograms. Override and return true if available.
    virtual bool get_raw_histogram(uint8_t *buf) { (void)buf; return false; }

    // The default for factory calibration is to reset parameters to their defaults,
    // but a sensor may want the application to set up its own parameters first.
    // For example, we might have one set of parameters for a light-field capture and
    // different parameters for a dark-field capture, or if manual_params is set.
    // To avoid resetting to defaults before capturing, override this and return false.
    virtual bool factory_needs_reset(bool manual_params) { (void)manual_params; return true; }

    // Override to support dump of sensor registers to a file.
    virtual void dump_sensor_regs(void *fp) {(void)fp;}

    // Say whether we should use the legacy image memory layout. Override to return false.
    virtual bool use_legacy_layout() const noexcept { return true; }

    // Say whether pixel FPN is supported. Override to return false;
    virtual bool supports_pixel_fpn() const noexcept { return true; }

    // Override if there are sensor parameters that need to be initialized with registers
    // that are stored in our GigE register save file.
    virtual void set_loaded_registers() {}

    // Say whether this fpn change requires that the sensor be restarted.
    virtual bool fpn_needs_restart(fpnmode_t old_fpn, fpnmode_t new_fpn)
    {
        if ((old_fpn & GIGE_FPN_SENSOR_ENABLE_MASK) != (new_fpn & GIGE_FPN_SENSOR_ENABLE_MASK))
            return true;
        else
            return false;
    }

    // Say whether this bit_mode change requires that the sensor have an opportunity to update
    // via update_capture.
    virtual bool bitmode_needs_restart(bitmode_t old_bit_mode, bitmode_t new_bit_mode)
    {
        (void)old_bit_mode;
        (void)new_bit_mode;
        return false;
    }

    // Say whether this frame rate change requires that the sensor be restarted via update_capture.
    virtual bool frame_rate_needs_restart(float old_fps, float new_fps, uint16_t startx,
                                          uint16_t starty, uint16_t width, uint16_t height,
                                          bitmode_t bit_mode, uint32_t sensor_opts)
    {
        (void)old_fps;
        (void)new_fps;
        (void)startx;
        (void)starty;
        (void)width;
        (void)height;
        (void)bit_mode;
        (void)sensor_opts;

        return false;
    }

    // Get what we report as the sensor max ROI.
    // Override if the sensor can change based on options or bit mode.
    virtual uint16_t get_reported_max_width(uint32_t sensor_opts, bitmode_t bit_mode)
    { (void)sensor_opts; (void)bit_mode; return sensor_max_width; }
    virtual uint16_t get_reported_max_height(uint32_t sensor_opts, bitmode_t bit_mode)
    { (void)sensor_opts; (void)bit_mode; return sensor_max_height; }

    // Return the default values to use with this particular sensor.
    virtual uint32_t get_default_frame_rate() = 0;
    virtual bitmode_t get_default_bit_mode() = 0;
    virtual uint32_t get_default_gain() = 0;
    virtual uint32_t get_frame_rate() = 0;

    // Get the maximum bit mode that this sensor supports.
    virtual bitmode_t get_max_bit_mode() = 0;

    // Get the sensor's minimum shutter speed (in usecs).
    virtual uint32_t get_sensor_min_shutter() = 0;

    // Not all sensors support a test mode. Override and return true if it was enabled.
    virtual bool enable_sensor_test() { return false; }

    // Not all sensors use the FPGA's bad pixel correction in the IPM. Override if available.
    virtual void IPM_bad_pixel_correction(int port, uint16_t true_x, uint16_t true_y,
                                          uint16_t width, uint16_t height)
    {
        (void)port;
        (void)true_x;
        (void)true_y;
        (void)width;
        (void)height;
    }

    // Allow a sensor to decline calibration, at least for now.
    virtual bool run_calibration(bool is_factory, bool use_manual, bool *rtn_restart)
    {
        (void)is_factory;
        (void)use_manual;
        (void)rtn_restart;
        return true;
    }
    virtual bool run_prnu(const uint8_t *ct_ptr)
    {
        (void)ct_ptr;
        return true;
    }
    virtual bool set_cal_params(uint32_t sensor_type, void *params)
    {
        (void)sensor_type;
        (void)params;
        return true;
    }
    virtual uint16_t get_cal_offset(uint32_t sensor_type, void *params)
    {
        (void)sensor_type;
        (void)params;
        return 0;
    }
    virtual bool is_cal_compatible(uint32_t sensor_type, void *params, SENSOR::SENSOR_ROI *calfile_roi,
                                   SENSOR::SENSOR_ROI *user_roi)
    {
        (void)sensor_type;
        (void)params;
        (void)calfile_roi;
        (void)user_roi;
        return true;
    }

    // Allow a sensor to override the capture validity with more stringent criteria.
    virtual bool make_sensor_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                                   uint16_t *rtn_width, uint16_t *rtn_height,
                                   bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                   uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts)
    {
        (void)rtn_startx;
        (void)rtn_starty;
        (void)rtn_width;
        (void)rtn_height;
        (void)rtn_bit_mode;
        (void)rtn_frame_rate;
        (void)rtn_shutter_speed;
        (void)rtn_sensor_opts;
        return false;
    }

    // Make sensor-specific adjustments in the ROI offset of the CT.
    virtual void get_ct_offset(bitmode_t bit_mode, uint32_t sensor_opts,
                               uint16_t *rtn_ofs_x, uint16_t *rtn_ofs_y)
    {
        (void)bit_mode; (void)sensor_opts;
        *rtn_ofs_x = 0; *rtn_ofs_y = 0;
    }

    // Say where the Bayer pattern starts for this sensor.
    virtual uint32_t IPM_bayer_pattern_start(uint16_t startx, uint16_t starty) = 0;

    // Each sensor has to say what bit_mode values it supports.
    virtual bool is_bit_mode_supported(bitmode_t bit_mode) = 0;

    // Each sensor has to say what options it supports.
    virtual bool is_option_supported(uint32_t sensopts) = 0;

    // The way we set the bit mode in the FPGA depends on the sensor.
    virtual bool set_fpga_bit_mode(bitmode_t bitmode) = 0;

    // Set the sensor frame rate.
    virtual bool set_frame_rate(float frame_rate, double clock_factor) = 0;

    // Set the integration time (shutter_speed in usec).
    virtual bool set_integration_time(uint32_t shutter_speed) = 0;

    // Set the sync delay value (delay_amount in usec).
    virtual void set_sync_delay(uint32_t delay_amount) = 0;

    // Set the sync-in information.
    virtual void set_syncin_enable(uint32_t type) = 0;

    // Set the gain in the sensor (GIGE_REG_SENSOR_GAIN), possibly deferred until next bringup.
    virtual int set_sensor_gain(uint32_t new_sensor_gain) = 0;

    // Tell the FPGA to start or stop capturing data from the sensor.
    virtual bool set_sensor_readout(bool enable) = 0;

    // Update the sensor's ROI, which might involve restarting the sensor.
    virtual bool update_capture(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                                fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                                uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) = 0;

    virtual bool sensor_startup(uint16_t true_x, uint16_t true_y, uint16_t width, uint16_t height,
                                fpnmode_t fpn_mode, bitmode_t bit_mode, float frame_rate,
                                uint32_t shutter, uint32_t sensor_opts, uint32_t sensor_gain) = 0;

    // Return Frame Overhead Time FOT, the extra time per frame that restricts maximum exposure.
    // This value is read by GUIs so they can compute a maximum exposure via ( 1/rate - FOT ).
    // The return value is in microseconds.
    // FIXME Move this to CameraServer class and accommodate all sources of frame overhead.
    virtual uint32_t get_frame_overhead() = 0;

    // Return the temperature on the sensor board, in Celsius.
    virtual double get_temperature() {return 0;}

    // Shutdown the sensor.
    virtual void shutdown() {}

    // Say whether this sensor supports video trigger.
    virtual bool has_video_trigger() const noexcept  { return false; }

    // Control the video trigger module.
    virtual uint32_t vtrig_control(const char *settings_str, int str_length) {(void)settings_str; (void)str_length; return GEV_STATUS_NOT_IMPLEMENTED;}

    // Return the current video trigger settings and status, json encoded.
    virtual uint32_t vtrig_info_json(char *dest_ptr, uint32_t max_len) {(void)dest_ptr; (void)max_len; return 0;}

    virtual int get_capture_fifo_count() { return 0; }
    virtual void reset_capture_fifo() {}
    virtual void reset_event_fifo() {}

};

#endif // SENSOR_H

