/*!
 * @file storage.h
 * @brief This file defines the storage access class.
 * @author Anthony L. Montiel
 * @date 2012-06-22
 *
 * Copyright (C) 2012 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef _FASTEC_STORAGE_H_
#define _FASTEC_STORAGE_H_

//--------------------------------------------------------------------
// Include Files
//--------------------------------------------------------------------

#include <stdint.h>

#include <json/json.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <filesystem>
#include <memory>
#include <mutex>
#include <string>

#include "cs_network.h"

namespace STORAGE_LOCAL
{
//--------------------------------------------------------------------
// Forward declaration
//--------------------------------------------------------------------

class file_action_thread;
}

//--------------------------------------------------------------------
// Storage class declaration
//--------------------------------------------------------------------

class storage
{
private:
    static std::unique_ptr<storage> local_storage;   //!< Our singleton object.
    static std::mutex storage_mutex;

    std::unique_ptr<STORAGE_LOCAL::file_action_thread> fa_task;

    bool capstk_transfer;
    ssize_t total_transfer;

    char *m_xmlOutputBuffer;
    int m_xmlOutputBufferLen;

private:
    // This is a singleton object.
    storage();

    bool add_jhead_info(xmlNodePtr f_node, const char *f_path);
    bool add_capstack_info(xmlNodePtr f_node, std::filesystem::path& file_path);
    bool add_metadata_info(xmlNodePtr f_node, std::string file_path, const std::string *file_ext = nullptr);
    uint64_t calc_stack_size(const char *dir_path, uint32_t num_frames);
    void get_file_stats(xmlNodePtr f_node, std::string file_path, std::string file_ext);
    void get_folder_stats(xmlNodePtr f_node, std::string file_path);
    bool parse_image_metadata(FILE *meta_file, xmlNodePtr f_node, bool capstack = false);

    static uint64_t calc_dir_size(const char *dir_path);
    static void get_internal_ssd_devpath(std::string& devpath);
    static void get_internal_nvme(Json::Value& value);
    static bool get_label(const char *dev_name, Json::Value& value);
    static uint32_t media_format(const char *device_name, const char *vol_label, uint32_t format_info);
    static bool nvme_mount();
    static bool nvme_umount();
    static bool device_is_busy(const char *device_name);

    static void nvme_udev(Json::Value& root, bool full_tree = false);
    static void sdcard_udev(Json::Value& root);
    static void usb_udev(Json::Value& root);

public:
    virtual ~storage();

    // Singletons should not be cloneable.
    storage(storage& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const storage&) = delete;

    int buildXmlTree();
    int buildLutXmlTree();
    constexpr void reset_total_transfer()   { total_transfer = 0; }
    constexpr bool get_capstk_transfer()    { return capstk_transfer; }
    constexpr void set_capstk_transfer(bool transfer_enabled)   { capstk_transfer = transfer_enabled; }

    static storage *factory();
    static void stop();

    static bool copy_space_available(const char *file_path, const char *media_ptr);
    static int do_mkdir(const char *path, mode_t mode);
    static bool exists(uint32_t media_sel);
    static bool force_umount_ssd();
    static int fs_supports_large_file(const std::string target_path);

    static uint32_t get_available_media();
    static int get_capacity(const char *mount_path, uint64_t *p_total, uint64_t *p_avail, uint64_t *p_inodes);
    STORAGE_LOCAL::file_action_thread *get_fa_task();
    static void get_ltr_device(std::string& dev_path);
    static void get_listing(const std::string& dir_path, const std::string& suffix, std::vector<std::string>& files);
    static void get_media_device(uint32_t media_sel, std::string& device, int index = 0);
    static uint32_t get_media_json(char *dest_ptr, uint32_t max_len);
    static uint32_t get_media_type(const std::string device);
    static void get_mount_dir(const std::string device, std::string& mountdir);
    static uint32_t get_ssd_capabilities();
    static int get_ssd_count();
    static uint64_t get_ssd_licensed_sector_count();
    static uint32_t get_status();
    static int  get_xml_tree_len();
    static const char *get_xml_tree_ptr();
    static bool is_mounted(uint32_t media_sel);

    static void md5sum_file(const char *file_path, const char *md5_path);
    static uint32_t media_action(const char *action_str, int str_length);
    static int mkpath(const char *path, mode_t mode);
    static bool operation_in_progress();

    static int reset_media_json(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value);

    static bool save_path(const char *path_str, uint32_t path_len);
    static bool start_action(uint32_t file_action);
    static uint32_t start_json_fa(const char *action_str, int str_length);
    static void startup_mounts();
    static bool stor_copy_file(const char *target_ptr, const char *source_ptr, uint8_t *buffer = nullptr);
};

#endif /* !_FASTEC_STORAGE_H_ */
