/*!
 * @file autosave.cxx
 *
 * @brief This file contains the implementation of the AutoSave
 *    class(es).
 *
 * This file contains all the routines to handle overlapped
 * saving while capturing.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <stdint.h>
#include <unistd.h>
#include <linux/types.h>

#include <chrono>

#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>
#include <fastec/fi_fpga_regs.h>

#include "autosave.h"
#include "camera_server.h"
#include "fcp.h"
#include "image_io.h"
#include "part_mgr.h"
#include "storage.h"

namespace AUTO_SAVE
{
//--------------------------------------------------------------------
// Constant definitions
//--------------------------------------------------------------------

enum AS_TRACE
{
    AS_FUNC     = 0x00000010,
    AS_WORK     = 0x00000020,
    AS_DEBUG    = 0x00000040,
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------


} /* End of namespace AUTO_SAVE */

//--------------------------------------------------------------------
// Static Constant Declarations
//--------------------------------------------------------------------
const int       AutoSave::MAX_QUEUED_PARTS = 255;
const int       AutoSave::MAX_RETRIES = 10;
const uint32_t  AutoSave::SLEEP_MS = 50;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

std::mutex               AutoSave::work_mutex;
std::condition_variable  AutoSave::work_cond;

// Make things easier on ourselves.
using namespace AUTO_SAVE;

//--------------------------------------------------------------------
// AutoSave class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. The autosave worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
AutoSave::AutoSave() :
    FAS_THREAD::Thread(true),
    sessionCount(0),
    total_fflr_frames(0)
{
    // Initialize our member variables.
    threadRunning = false;

    traceControl = 0 |
                   AS_FUNC |
                   AS_WORK |
                   AS_DEBUG |
                   TRACE_0;        // Set our trace level so we'll trace.
//                   TRACE_3;        // Set our trace level to turn off tracing.
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 *
 * @param arg   Not used.
 */
AutoSave::~AutoSave()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Release our allocated queues, if necessary.
    releaseQueues();
}

/*!
 * @brief Get the thread running.
 */
void AutoSave::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure we're not running yet.
    if (threadRunning)
    {
        fi_log(LOG_DEBUG, "%s: Already running", __PRETTY_FUNCTION__);
        return;
    }

    // Now start the thread.
    stop_request = false;
    FAS_THREAD::Thread::start(threadFn);

    TP(AS_FUNC, "%s: Done.", __PRETTY_FUNCTION__);
}

/*!
 * @brief Ask the thread to stop.
 */
void AutoSave::stop()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Tell the thread to stop.
    FAS_THREAD::Thread::stop();

    // Make sure the worker thread knows we want to exit.
    work_cond.notify_one();
}

/*!
 * @brief Update the contents of the autosave register.
 */
void AutoSave::updateRegisterValue()
{
    uint32_t    reg_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the current value.
    reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();

    // Preserve the autosave enable bit.
    reg_value &= (GIGE_AUTOSAVE_ENABLE | GIGE_AUTOSAVE_USER_ENABLE);

    // Set the total/dirty counts into the register.
    reg_value |= (((sessionCount & GIGE_AUTOSAVE_COUNT_MASK) << GIGE_AUTOSAVE_TOTAL_SHIFT) |
                  ((availSessions() + busySessions()) & GIGE_AUTOSAVE_COUNT_MASK));

    // Update our overlapping flag, only if autosave is enabled.
    if ((reg_value & GIGE_AUTOSAVE_ENABLE) && doOverlap())
        reg_value |= GIGE_AUTOSAVE_OVERLAP;

    // Save the new value.
    regs[ GIGE_REG_AUTOSAVE ].set_value(reg_value);
}

/*!
 * @brief Release the allocated queues.
 */
void AutoSave::releaseQueues()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    {
        std::unique_lock<std::mutex> locker(queueLock);

        // Clear out the available queue.
        availQueue.clear();

        // Clear out the busy queue.
        busyQueue.clear();

        // Clear out the dirty queue.
        dirtyQueue.clear();

        // Reset our session count.
        sessionCount = 0;
    }

    // Update our register value.
    updateRegisterValue();
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void AutoSave::cleanup()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // We're not running anymore.
    threadRunning = false;
    stop_request = false;
    notify_done();
}

void AutoSave::fflrDiscard()
{
    int part_count;
    uint32_t images_per_part;
    PartMgr *pm_ptr;

    // Sanity check.
    if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_STREAM_FORMAT_MASK) != GIGE_STREAM_FORMAT_CAPSTK)
    {
        fi_log(LOG_DEBUG, "%s: Not in FFLR mode!", __PRETTY_FUNCTION__);
        return;
    }

    TP(AS_DEBUG, "%s: total_fflr_frames=%u, amt to record=%u", __PRETTY_FUNCTION__, total_fflr_frames, imgSaveWorker::getInstance()->getRecordingFrameCount());

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Get the number of images per partition.
    images_per_part = pm_ptr->get_im_buf_count();
    if (images_per_part == 0)
    {
        // Strange value. Something must be wrong.
        fi_log(LOG_ERR, "%s: get_im_buf_count() returned 0!?", __PRETTY_FUNCTION__);
        part_count = 1;
    }
    else
    {
        // Sanity check.
        if (imgSaveWorker::getInstance()->getRecordingFrameCount() == 0)
        {
            fi_log(LOG_ERR, "%s: The recording frame count is 0!", __PRETTY_FUNCTION__);
            imgSaveWorker::getInstance()->setRecordingFrameCount(pm_ptr->getPartitionNumFrames(pm_ptr->getActivePartition()));
        }

        // Calculate the number of partitions are required for the requested number of frames, round up.
        part_count = (imgSaveWorker::getInstance()->getRecordingFrameCount() + (images_per_part / 2)) / images_per_part;
    }

    // We need one extra partition for "slop".
    part_count++;
    TP(AS_DEBUG, "%s: part_count=%d", __PRETTY_FUNCTION__, part_count);

    // Do we need to worry about discarding the oldest partition?
    if (total_fflr_frames >= (((uint64_t)part_count) * images_per_part))
    {
        IMAGE_FORMAT::FFLR_CONTEXT_PTR oldest;

        // Take the oldest off the queue. The record will be deleted once
        // we exit scope.
        oldest = imgSaveWorker::getInstance()->popFflrQueue();
        if (oldest.get() == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: There were no entries in the queue?! total_fflr_frames=%" PRIu64, __PRETTY_FUNCTION__, total_fflr_frames);
            total_fflr_frames = 0;
            return;
        }

        // Delete the file.
        TP(AS_DEBUG, "%s: Remove '%s'", __PRETTY_FUNCTION__, oldest->cap_file.c_str());
        if (!storage::save_path(oldest->cap_file.c_str(), oldest->cap_file.generic_string().size()))
        {
            fi_log(LOG_DEBUG, "%s: storage::save_path() failed!", __PRETTY_FUNCTION__);
            return;
        }
        if (!storage::start_action(GIGE_FA_DELETE))
        {
            fi_log(LOG_DEBUG, "%s: storage::start_action(GIGE_FA_DELETE) failed!", __PRETTY_FUNCTION__);
            return;
        }

        // Update our counter.
        total_fflr_frames -= oldest->frame_count;
    }

    TP(AS_DEBUG, "%s: total_fflr_frames=%" PRIu64, __PRETTY_FUNCTION__, total_fflr_frames);
}

/*!
 * @brief Safely get the first entry in the dirty queue. If there are no
 *        entryies, return -1.
 * @return The first entry in the dirtyQueue, or -1 if it is empty.
 */
int AutoSave::getDirtyQueueFront()
{
    int dirty_num = -1;

    std::unique_lock<std::mutex> locker(queueLock);
    if (!dirtyQueue.empty())
    {
        // Get the first entry from the dirty queue.
        dirty_num = dirtyQueue.front();
        dirtyQueue.pop_front();
        TP(AS_DEBUG, "%s: %d", __PRETTY_FUNCTION__, dirty_num);
    }

    // Return our entry.
    return dirty_num;
}

/*!
 * @brief Move the next available session to the dirty queue.
 */
bool AutoSave::makeDirty()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure there are available sessions to move.
    if (busyQueue.empty())
        return false;

    // Lock access to the queues.
    {
        std::unique_lock<std::mutex> locker(queueLock);

        // Move the first available session into the dirty queue.
        dirtyQueue.push_back(busyQueue.front());
        busyQueue.pop_front();
    }

    // Let folks know there's work to be done.
    work_cond.notify_one();

    // Update our register value.
    updateRegisterValue();

    // Make sure the worker thread is running.
    if (!threadRunning)
        start();

    return true;
}

int AutoSave::queueNextAvail()
{
    int part_num = -1;

    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Do we have any to queue?
    if (!availQueue.empty())
    {
        std::unique_lock<std::mutex> locker(queueLock);

        // Move the first available session into the busy queue.
        part_num = availQueue.front();
        busyQueue.push_back(part_num);
        availQueue.pop_front();
    }

    // Did we get a partition to queue?
    if (part_num != -1)
    {
        PartMgr *pm_ptr;

        // Get a pointer to our support classes.
        pm_ptr = CameraServer::get_partmgr();

        TP(AS_DEBUG, "%s: queueing %d", __PRETTY_FUNCTION__, part_num);
        pm_ptr->queueCapturePartition(part_num);
    }

    // Return the partition number just queued.
    return part_num;
}

/*!
 * @brief Setup the session queues based on the session
 *    length parameter.
 *
 * @param num_sessions  Number of sessions to use
 *
 * @return true on success, false otherwise.
 */
bool AutoSave::init(uint32_t num_sessions)
{
    TP(AS_FUNC, "%s: num_sessions=%d", __PRETTY_FUNCTION__, num_sessions);

    // Check to see if the thread is running.
    if (threadRunning)
    {
        int retry;

        // Ask the thread to stop.
        stop();

        // Wait for the thread to stop.
        for (retry = 0; retry < MAX_RETRIES; ++retry)
        {
            int rc;

            // Give it a chance to stop.
            rc = wait_for_done(100);
            if (rc == 0)
                break;
        }
        if (retry == MAX_RETRIES)
        {
            fi_log(LOG_ERR, "%s: Unable to stop our thread.", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // Release our queues.
    releaseQueues();

    // Save the session count.
    if (num_sessions > MAX_QUEUED_PARTS)
    {
        fi_log(LOG_ERR, "%s: Too many sessions are being requested [%d]!", num_sessions);
        sessionCount = MAX_QUEUED_PARTS;
        return false;
    }
    sessionCount = num_sessions;

    // All of the sessions are available.
    for (int i = 0; i < (int)sessionCount; ++i)
        availQueue.push_back(i);

    // Initialize the total number of frames we've saved.
    total_fflr_frames = 0;
    fflr_file_count = 0;
    imgSaveWorker::getInstance()->reset();

    // Make sure we're syncronized to our available queue.
    TP(AS_DEBUG, "%s: OVERLAP=%d ENABLED=%d", __PRETTY_FUNCTION__, doOverlap(), isEnabled());

    // Update our register value.
    updateRegisterValue();

    return true;
}

/*!
 * @brief Reset the autosave manager to a known/beginning state.
 */
void AutoSave::reset()
{
    FiFPGA  *fpga_ptr;
    PartMgr *pm_ptr;

    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    pm_ptr = CameraServer::get_partmgr();
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null FiFPGA pointer!?", __PRETTY_FUNCTION__);
        return;
    }
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null PartMgr pointer!?", __PRETTY_FUNCTION__);
        return;
    }

    // Have the FPGA flush the FIFO.
    fpga_ptr->icmResetCaptureFIFO();

    // Have the partition manager reset the active partition.
    pm_ptr->setActivePartition(0);

    // Lock access to the queues. This will block if
    // the thread is in the middle of something and has
    // the lock.
    {
        std::unique_lock<std::mutex> locker(queueLock);

        // Clear out the available queue.
        availQueue.clear();

        // Clear out the busy queue.
        busyQueue.clear();

        // Clear out the dirty queue.
        dirtyQueue.clear();

        // Initialize our available queue with information about
        // each of the sessions.
        for (int i = 0; i < (int)sessionCount; ++i)
        {
            // Add the partition number to the available queue.
            availQueue.push_back(i);
        }
    }

    TP(AS_DEBUG, "%s: OVERLAP=%d ENABLED=%d", __PRETTY_FUNCTION__, doOverlap(), isEnabled());
    if (isEnabled() && doOverlap())
    {
        // Load up the FIFO until we can't load any more.
        while (queueNextAvail() != -1)
        {
            // Check to see if we've hit the limit.
            if (fpga_ptr->icmCaptureFifoCount() == MAX_QUEUED_PARTS)
            {
                fi_log(LOG_DEBUG, "%s: Hit the capture fifo limit!", __PRETTY_FUNCTION__);
                break;
            }
        }
    }
    else
    {
        // Make sure we're syncronized to our available queue.
        queueNextAvail();
    }

    // Reset our FFLR variables for a new save.
    total_fflr_frames = 0;
    fflr_file_count = 0;
    imgSaveWorker::getInstance()->reset();

    // Update our register value.
    updateRegisterValue();
}

/*!
 * @brief Determine if we want to do an overlapped autosave or not.
 *
 * @return true if we're OK for overlapped mode, false otherwise.
 */
bool AutoSave::doOverlap()
{
    bool rc = false;

    // Only do overlapped if we have enough sessions.
    if (sessionCount < 2)
        goto do_bail;

    // Is there a partition on either the available or busy queues?
    if (availQueue.empty() && busyQueue.empty())
    {
        TP(AS_DEBUG, "%s: availQueue.empty && busyQueue.empty", __PRETTY_FUNCTION__);
        goto do_bail;
    }

    // If our destination is off-camera, then we're not doing overlapped.
    if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
        goto do_bail;

    // Looks good for overlapped autosave.
    rc = true;

do_bail:
    TP(AS_FUNC, "%s: %d", __PRETTY_FUNCTION__, rc);
    if (!rc)
    {
        FiFPGA  *fpga_ptr;
        fpga_ptr = CameraServer::get_fpga();
        TP(AS_DEBUG, "%s: Capture FIFO count=%d", __PRETTY_FUNCTION__, fpga_ptr->icmCaptureFifoCount());
    }

    return rc;
}

bool AutoSave::armOK()
{
    bool rc = false;
    FiFPGA  *fpga_ptr;

    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get our FPGA pointer.
    fpga_ptr = CameraServer::get_fpga();

    // We can arm if we are saving off camera.
    if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
    {
        TP(AS_FUNC, "%s: Off camera, OK to ARM.", __PRETTY_FUNCTION__);
        rc = true;
    }

    // We can arm if we have sessions we can save into.
    if ((availSessions() || busySessions()) && fpga_ptr->icmCaptureFifoCount())
    {
        TP(AS_FUNC, "%s: We have sessions to save into, OK to ARM.", __PRETTY_FUNCTION__);
        rc = true;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Utility to return the state of the autosave enable.
 * @return True if autosave is enabled, false otherwise.
 */
bool AutoSave::isEnabled()
{
    bool rc;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // If autosave is not enabled, then we're done.
    rc = ((regs[ GIGE_REG_AUTOSAVE ].get_value() & ((uint32_t)GIGE_AUTOSAVE_ENABLE)) == ((uint32_t)GIGE_AUTOSAVE_ENABLE));
    TP(AS_DEBUG, "%s: GIGE_AUTOSAVE_ENABLE=%d", __PRETTY_FUNCTION__, rc);

    // Return our result.
    return rc;
}

/*!
 * @brief Process a completed partition.
 */
void AutoSave::process()
{
    TP(AS_FUNC, "%s", __PRETTY_FUNCTION__);

    // If autosave is not enabled, then we're done.
    if (!isEnabled())
        return;

    // If our destination is off-camera, then we're not doing overlapped.
    if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
        return;

    // We are doing autosave, so queue up the partition for the worker thread.
    if (!makeDirty())
    {
        fi_log(LOG_ERR, "%s: Unable to queue up the dirty partition!", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Requests that the first element of the dirty queue be
 *    moved to the available queue. Make the new top of the
 *    dirty queue be the active partition.
 *
 *    This is used to handshake with off-camera saves. When the save is complete,
 *    the remote app needs some way for us to transition to the next partition.
 */
void AutoSave::queueUpdate()
{
    PartMgr *pm_ptr;

    TP(AS_FUNC, "%s: isEnabled=%d", __PRETTY_FUNCTION__, isEnabled());

    // Make sure we have something in the dirty queue.
    if (dirtyQueue.empty())
    {
        TP(AS_FUNC, "%s: dirtyQueue empty, leaving.", __PRETTY_FUNCTION__);
        return;
    }

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Lock access to our queues before manipulating them.
    {
        int     dirty_part_num;         // Dirty partition.
        std::unique_lock<std::mutex> locker(queueLock);

        // Get the first partition off the dirty queue.
        dirty_part_num = dirtyQueue.front();

        // Pop it off the dirty queue.
        dirtyQueue.pop_front();

        // Make sure it's the active partition, so we
        // can mark it empty.
        pm_ptr->setPartitionState(dirty_part_num, GIGE_PART_STATE_EMPTY);

        // Put the partition back on to the available queue.
        availQueue.push_back(dirty_part_num);
    }

    // Queue up the next avail partition if we can.
    queueNextAvail();

    // Update our register.
    updateRegisterValue();

    // If we are not enabled, make the next dirty partition the active partition.
    if (!isEnabled() && !dirtyQueue.empty())
    {
        TP(AS_FUNC, "%s: setActivePartition=%d", __PRETTY_FUNCTION__, dirtyQueue.front());

        // Make the failed partition our active partition.
        pm_ptr->setActivePartition(dirtyQueue.front());
        pm_ptr->readActiveImMetadata();

        // Notify folks of the changes.
        fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, pm_ptr->getFrameCount());
        fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, pm_ptr->getTriggerFrame());
        fcp_notify_apps(GIGE_REG_PART_STATE, pm_ptr->getPartitionState());
    }
}

void AutoSave::adjustSave(int part_num)
{
    uint32_t    start_offset;
    uint32_t    stop_offset;
    uint32_t    trigger_frame;
    uint32_t    num_frames;
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;
    PartMgr *pm_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(AS_FUNC, "%s: part_num=%d", __PRETTY_FUNCTION__, part_num);

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();
    fpga_ptr = CameraServer::get_fpga();

    // Make sure the saver knows the part number we're interested in.
    imgSaveWorker::getInstance()->setPartNum(part_num);

    // For now, we can hijack the P1 partition. It's safe
    // 'cuz we're working in serial with the save threads.
    pm_ptr->setReviewPartNum(part_num);

    // Read and save away the starting and ending frame numbers.
    // We do this so there will be something valid when we restore
    // later in the process. Also, in the case where the session
    // length is changed, we want to catch that change.
    saveStartNum = regs[ GIGE_REG_SAVE_START_NUM ].get_value();
    saveEndNum = regs[ GIGE_REG_SAVE_END_NUM ].get_value();

    // When saving in FFLR, we always save the entire partition except for the
    // last partition. That's the only one which will be (most likely) a partial.
    if (fpga_ptr->ltrFasFireEnabled())
    {
        // We always start at the first frame.
        start_offset = 0;

        // How many frames do we still want to save?
        stop_offset = pm_ptr->getReviewPartNumFrames() - 1;

        // Make sure we make sense. We can't have a stop that is bigger than the partition.
        if (stop_offset >= pm_ptr->getFrameCount())
            stop_offset = pm_ptr->getFrameCount() - 1;
        TP(AS_DEBUG, "%s: start_offset=%d, stop_offset=%d", __PRETTY_FUNCTION__, start_offset, stop_offset);
    }
    else
    {
        // What did we configure the FPGA to have for a number
        // of post-trigger frames?
        reg_value = fpga_ptr->icmPostTriggerCount();

        // What frame was the trigger supposed to be on?
        trigger_frame = pm_ptr->getFrameCount() - (reg_value + 1);

        // What was the offset to the start & stop frames?
        start_offset = saveStartNum - trigger_frame;
        stop_offset = saveEndNum - trigger_frame;

        // Now calculate what these would be for what was actually captured.
        trigger_frame = pm_ptr->getReviewPartTrigger();
        start_offset = trigger_frame + start_offset;
        stop_offset = trigger_frame + stop_offset;
        TP(AS_DEBUG, "%s: trigger_frame=%d/%d", __PRETTY_FUNCTION__, part_num, trigger_frame);

        // Get the maximum number of frames we've allocated for capturing.
        num_frames = pm_ptr->getPartitionNumFrames(part_num);

        // How many frames does the user want us to save?
        reg_value = (stop_offset - start_offset) + 1;
        if (reg_value > num_frames)
            reg_value = num_frames;

        // Make sure they make sense.
        if (start_offset > num_frames)
            start_offset = 0;
        if (stop_offset >= (start_offset + reg_value))
            stop_offset = (start_offset + reg_value) - 1;
        if (stop_offset >= num_frames)
            stop_offset = num_frames - 1;
    }

    // Save the adjusted values.
    regs[ GIGE_REG_SAVE_START_NUM ].set_value(start_offset);
    regs[ GIGE_REG_SAVE_END_NUM ].set_value(stop_offset);
}

/*!
 * @brief The worker thread.
 */
void AutoSave::run()
{
    int dirty_part_num;             // Dirty partition.
    uint32_t save_status;           // Status of the save operation.
    PartMgr *pm_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (threadRunning)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We're running.
    threadRunning = true;
    notify_ready();

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Loop until we are asked to stop.
    while (!stop_request)
    {
        FiFPGA  *fpga_ptr = CameraServer::get_fpga();

        // Wait for something to do, or we get a stop request.
        while (!stop_request)
        {
            // Determine if our dirty queue has anything in it.
            dirty_part_num = getDirtyQueueFront();
            if (dirty_part_num != -1)
                break;

            // If there are no more dirty partitions and we are now in Live Video state,
            // we are all done and this thread can exit.
            if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_LIVE_VIDEO)
            {
                // This must be the end of the recording. Are we in FasFireLR Mode?
                if (fpga_ptr->ltrFasFireEnabled())
                {
                    TP(AS_DEBUG, "%s: FasFireLR, no dirty partitions and we're in LIVE", __PRETTY_FUNCTION__);

                    // Are there any sanity checks to make?
                    // Need to tell the saver object to write out the super metadata.
                    imgSaveWorker::getInstance()->saveFflrMetadata();
                }

                // Set our flag and immediately bail out of the loop.
                TP(AS_DEBUG, "%s: No dirty buffers and state is LIVE.", __PRETTY_FUNCTION__);
                stop_request = true;
            }
            else
            {
                std::unique_lock<std::mutex> locker(work_mutex);

                // Delay for a bit before checking the queue and state again.
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                work_cond.wait_until(locker, now + ms(SLEEP_MS));
            }
        }

        // Bail if we got a stop request.
        if (stop_request)
        {
            TP(AS_WORK, "%s: Got a stop request!", __PRETTY_FUNCTION__);
            break;
        }

        TP(AS_DEBUG, "%s: Partition %d is dirty...", __PRETTY_FUNCTION__, dirty_part_num);

        // Are we in FasFireLR mode?
        if (fpga_ptr->ltrFasFireEnabled())
        {
            // Discard the oldest capture if we need to.
            fflrDiscard();
        }

        // Make sure the save thread isn't already running.
        while (imgSaveWorker::getInstance()->isRunning())
        {
            fi_log(LOG_ERR, "%s: Cannot allow multiple instances of imgsave to run!", __PRETTY_FUNCTION__);

            // Are we being asked to stop?
            if (stop_request)
                break;

            {
                std::unique_lock<std::mutex> locker(work_mutex);

                // Delay for a bit before checking the status again.
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                work_cond.wait_until(locker, now + ms(SLEEP_MS));
            }
        }

        // Are we being asked to stop?
        if (stop_request)
            break;

        // Adjust the save parameters based on what we ACTUALLY
        // saved into the partition.
        adjustSave(dirty_part_num);

        // Start the save.
        imgSaveWorker::getInstance()->setCapStackIndex(fflr_file_count);
        imgSaveWorker::getInstance()->start();

        // Wait for the save to finish.
        while (imgSaveWorker::getInstance()->isRunning())
        {
            // Are we being asked to stop?
            if (stop_request)
            {
                // Stop the save operation.
                imgSaveWorker::getInstance()->stop();
            }

            {
                std::unique_lock<std::mutex> locker(work_mutex);

                // Delay for a bit before checking the status again.
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                work_cond.wait_until(locker, now + ms(SLEEP_MS));
            }
        }

        // If the save was successful, then add the partition back to
        // the available queue. Otherwise, on an error or a stop request,
        // push the dirty partition back on the dirty queue.
        save_status = imgSaveWorker::getInstance()->getCompletion();

        TP(AS_DEBUG, "%s: imgsave must be done, status:0x%x. --------------", __PRETTY_FUNCTION__, save_status);

        // Lock access to our queues before manipulating them.
        {
            std::unique_lock<std::mutex> locker(queueLock);

            if (!stop_request && ((save_status & GIGE_SAVE_STATE_ERR) != GIGE_SAVE_STATE_ERR))
            {
                // Are we in FasFireLR mode?
                if (fpga_ptr->ltrFasFireEnabled())
                {
                    // Update our saved frame counter.
                    total_fflr_frames += pm_ptr->getPartitionNumFrames(dirty_part_num);

                    // Update our file count.
                    fflr_file_count++;

                    TP(AS_DEBUG, "%s: part=%d, frameCount=%d/%d, total_fflr_frames=%" PRIu64 ", fflr_file_count=%d", __PRETTY_FUNCTION__, dirty_part_num, pm_ptr->getPartitionNumFrames(dirty_part_num), pm_ptr->getFrameCount(), total_fflr_frames, fflr_file_count);
                }

                // Mark it as empty.
                pm_ptr->setPartitionState(dirty_part_num, GIGE_PART_STATE_EMPTY);

                // Put the partition back on to the available queue.
                TP(AS_DEBUG, "%s: Push part %d onto the availQueue", __PRETTY_FUNCTION__, dirty_part_num);
                availQueue.push_back(dirty_part_num);
            }
            else
            {
                // Put the partition back on to the front of the dirty queue.
                fi_log(LOG_DEBUG, "%s: Error while saving [0x%x], dirty_part_num=%d", __PRETTY_FUNCTION__, save_status, dirty_part_num);
                dirtyQueue.push_front(dirty_part_num);
            }
        }

        // Restore the user specified starting/ending frames.
        regs[ GIGE_REG_SAVE_START_NUM ].set_value(saveStartNum);
        regs[ GIGE_REG_SAVE_END_NUM ].set_value(saveEndNum);

        // Update our register value.
        updateRegisterValue();

        // If there was an error, then we need to bail out of autosave.
        if ((save_status & GIGE_SAVE_STATE_ERR) == GIGE_SAVE_STATE_ERR)
        {
            uint32_t reg_value;

            fi_log(LOG_DEBUG, "%s: There was an error while saving [save_status=0x%08x]", __PRETTY_FUNCTION__, save_status);

            // Disable autosave. Keep the count information.
            reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
            reg_value &= ~(GIGE_AUTOSAVE_ENABLE | GIGE_AUTOSAVE_OVERLAP | GIGE_AUTOSAVE_SAVING);
            regs[ GIGE_REG_AUTOSAVE ].set_value(reg_value);

            // Let folks know it's changed.
            fcp_notify_apps(GIGE_REG_AUTOSAVE, reg_value);

            // Make the failed partition our active partition.
            pm_ptr->setActivePartition(dirtyQueue.front());
            pm_ptr->readActiveImMetadata();

            // Request that our camera state go to review, but don't wait for it to happen.
            CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_REVIEW, false, NULL);
            stop_request = true;
        }

        else if (stop_request)
        {
            TP(AS_DEBUG, "%s: Stopping, so setting part%d as active.", __PRETTY_FUNCTION__, dirtyQueue.front());

            // Make the first dirty partition our active partition.
            pm_ptr->setActivePartition(dirtyQueue.front());
            pm_ptr->readActiveImMetadata();
        }

        // If we have available partitions, and we are in the AUTOSAVE state
        // then we probably need to kick the FPGA to start another capture.
        else if ((availQueue.empty() == false) && (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_AUTOSAVE))
        {
            TP(AS_DEBUG, "%s: Kicking the FPGA.", __PRETTY_FUNCTION__);

            // Setup the next available partition.
            queueNextAvail();
            pm_ptr->setActivePartition(busyQueue.front());

            // Request that our camera state go to ARMed, but don't wait for it to happen.
            CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_ARMED, false, NULL);
        }
        else
        {
            // All is well. See if we can put another partition on the queue.
            TP(AS_DEBUG, "%s: Done with dirty_part_num=%d", __PRETTY_FUNCTION__, dirty_part_num);
            queueNextAvail();
        }
    }

    // Clean up on the way out.
    cleanup();

    // All done.
    TP(AS_FUNC, "%s: done", __PRETTY_FUNCTION__);
}


