/*!
 * @file bad_pixels.cpp
 *
 * @brief This file encapsulates the bad_pixels class
 *        used to tell the FPGA how to correct for detected
 *        bad pixels on the sensor
 * @author Tony Beville
 * @date 2012-10-03
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <fastec/fi_fpga_regs.h>
#include <fastec/fas_log.h>

#include "bad_pixels.h"
#include "sensor.h"

/*!
 * @brief Constructor: loads in the set of bad pixels from the sensor EEPROM via fi_sensor driver
 */
BadPixels::BadPixels(SENSOR::SENSOR_BAD_PIXEL *sensor_bp_ptr)
{
    int i;

    roi_x = 0;
    roi_y = 0;
    roi_w = 0;
    roi_h = 0;

    // Clear the array
    bad_pixels.clear();

    // Sanity check.
    if (sensor_bp_ptr == nullptr)
        return;

    // Read in each line of coordinates and save it
    for (i = 0; i < SENSOR::MAX_BAD_PIXELS; ++i)
    {
        // Is this a coordinate?
        if (sensor_bp_ptr[i].x != 0xFFFF)
        {
            // Save the information.
            BAD_PIXEL bp;
            bp.x = sensor_bp_ptr[i].x;
            bp.y = sensor_bp_ptr[i].y;
            bp.offset = 0;

            // Push it on the vector.
            bad_pixels.emplace_back(bp);
// LOG("%s: %d, [%d, %d]", __PRETTY_FUNCTION__, i, bp.x, bp.y);
        }
    }

// LOG("%s: bad_pixels.size() = %d", __PRETTY_FUNCTION__, bad_pixels.size());
}


BadPixels::~BadPixels()
{
}

/*!
 * @brief Set the ROI information that the correction will use.
 *
 * Sensor functions use the "true" coordinates relative to the sensor's actual size.
 * They might have been adjusted since the user's ROI is relative to the licensed area.
 *
 * @param true_x    current ROI "true" x coordinate, relative to the sensor's actual size
 * @param true_y    current ROI "true" y coordinate, relative to the sensor's actual size
 * @param w     current ROI width
 * @param h     current ROI height
 */
void BadPixels::set_roi(int true_x, int true_y, int w, int h)
{
    roi_x = true_x;
    roi_y = true_y;
    roi_w = w;
    roi_h = h;
}

/*!
 * @brief Write pixel corrections to the FPGA's Image Processing Module.
 */
void BadPixels::correct()
{
    // Get support class pointer(s).
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    for (int i = 0; i < (int)bad_pixels.size(); i++)
    {
        uint32_t reg_a_addr;
        uint32_t reg_b_addr;
        uint32_t reg_a_value;
        uint32_t reg_b_value;
        uint8_t new_offset;

        // for each bad pixel, calculate what the new offset is, based on the ROI settings
        new_offset = calculate_offset(i);
        bad_pixels[i].offset = new_offset;

        if (new_offset == 0)
        {
            // offset not needed, so clear the reg value to disable
            reg_a_value = 0;
            reg_b_value = 0;
        }
        else
        {
            uint32_t    new_x;
            uint32_t    new_y;

            // Adjust the x & y coordinate based on the ROI.
            new_x = bad_pixels[i].x - roi_x;
            new_y = bad_pixels[i].y - roi_y;

            //put together the reg value with enable bit row (Y), column (X), and offset
            reg_a_value = BP_CONFIG_A_VALUE(new_offset, new_y, new_x);
            reg_b_value = BP_CONFIG_B_VALUE(new_offset, new_y, new_x);
        }

        // which register to write to?
        reg_a_addr = IPM_BP00_CONFIG_A_ADDR + (8 * i);
        reg_b_addr = IPM_BP00_CONFIG_B_ADDR + (8 * i);

// LOG("%s: reg_a[0x%08x]=0x%08x, reg_b[0x%08x]=0x%08x", __PRETTY_FUNCTION__, reg_a_addr, reg_a_value, reg_b_addr, reg_b_value);

        // Write our calculated reg_value to FPGA reg_addr
        fpga_ptr->regWrite(reg_a_addr, reg_a_value);
        fpga_ptr->regWrite(reg_b_addr, reg_b_value);
    }
}


// calculate what the the offset correction of the indicated bad pixel should be
uint8_t BadPixels::calculate_offset(int index)
{
    int8_t signed_offset;
    uint8_t twoscomp_offset;

    uint16_t pixel_x = bad_pixels[index].x;
    uint16_t pixel_y = bad_pixels[index].y;

    // list of offset candidates to test, in order.  the last is zero on purpose.
    int8_t offset_test_candidates[] = { 2, -2, 4, -4, 6, -6, -8, 0 };

    // !!! I think I see a potential problem:  4 bad pixels on left edge with X positions 0, 2, 4, 6.
    // We cannot specify a good replacement, since the max offset is postive 7.
    // Of course the odds of this are astronomical...
    // The following algorithm will simply ignore this case.

    // First see if it's even in range of the ROI.
    if ((pixel_x < roi_x) || (pixel_y < roi_y) || (pixel_x >= (roi_x + roi_w)) || (pixel_y >= (roi_y + roi_h)))
    {
        // out of range of ROI, the offset is 0
        signed_offset = 0;
    }
    else
    {
        // go through the array of offset candidates and see if we can use one
        for (int o = 0; o < (int)sizeof(offset_test_candidates); o++)
        {
            bool bad_offset = false;

            signed_offset = offset_test_candidates[o];

            // Check right edge.
            if ((pixel_x + signed_offset) >= (roi_x + roi_w))
            {
                bad_offset = true;
            }

            // Check left edge.
            else if ((pixel_x + signed_offset) < roi_x)
            {
                bad_offset = true;
            }

            // Now make sure it doesn't land on other bad pixels
            else
            {
                for (int i = 0; i < (int)bad_pixels.size(); i++)
                {
                    if (i == index)
                    {
                        // Skip this one.
                        continue;
                    }

                    if ((pixel_x + signed_offset) == bad_pixels[i].x)
                    {
                        bad_offset = true;
                        break;
                    }
                }
            }

            // if we made it through the tests without it being being bad, we're done
            if (!bad_offset)
            {
                break;
            }
        }
    }

    // change to 4-bit 2's complement (range: -8 to +7 )
    if (signed_offset < 0)
    {
        twoscomp_offset = (~(abs(signed_offset) - 1)) & 0x0F;
    }
    else
        twoscomp_offset = (signed_offset & 0x0F);

    return twoscomp_offset;
}

