/*!
 * @file battery.cpp
 *
 * @brief This file contains all the routines to deal with battery
 *    data for the camera.
 *
 * Copyright 2024 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <algorithm>
#include <deque>
#include <memory>
#include <mutex>
#include <string>

#include <fastec/fas_log.h>
#include <fastec/fi_fpga_regs.h>

#include "battery.h"

#include "camera_server.h"
#include "cs_trace.h"
#include "fpga.h"

namespace BATTERY_LOCAL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Default polling time in microseconds.
const uint32_t DEFAULT_POLLING_TIME = 5000000;

// The maximum battery voltage we would expect to see.
// Anything over this can be considered as wall power.
const double MAX_BATTERY_VOLTAGE = 16.40;

// The minimum battery voltage, below which we need to shutdown or we'll crash.
const double CRITICAL_LOW_VOLTAGE = 11.7;

// How many readings are in our sample buffer.
const int SAMPLE_SIZE = 15;

// Address for the battery gas gauge device.
// 0xC6 >> 1 = 0x63
#define BGG_I2C_ADDR            0x63

// The default RSENSE value (from the latest schematic), in milliamps.
#define RSENSE                  5

// Register address
#define BGG_STATUS              0x00
#define BGG_ADC_CONTROL         0x01
#define BGG_CC_CONTROL          0x02
#define BGG_ACC_CHARGE_31_24    0x03
#define BGG_ACC_CHARGE_23_16    0x04
#define BGG_ACC_CHARGE_15_08    0x05
#define BGG_ACC_CHARGE_08_00    0x06
#define BGG_CT_LOW_31_24        0x07
#define BGG_CT_LOW_23_16        0x08
#define BGG_CT_LOW_15_08        0x09
#define BGG_CT_LOW_07_00        0x0a
#define BGG_CT_HIGH_31_24       0x0b
#define BGG_CT_HIGH_23_16       0x0c
#define BGG_CT_HIGH_15_08       0x0d
#define BGG_CT_HIGH_07_00       0x0e
#define BGG_VOLTAGE_MSB         0x0F
#define BGG_VOLTAGE_LSB         0x10
#define BGG_VOLT_THRESH_H_MSB   0x11
#define BGG_VOLT_THRESH_H_LSB   0x12
#define BGG_VOLT_THRESH_L_MSB   0x13
#define BGG_VOLT_THRESH_L_LSB   0x14
#define BGG_MAX_VOLT_MSB        0x15
#define BGG_MAX_VOLT_LSB        0x16
#define BGG_MIN_VOLT_MSB        0x17
#define BGG_MIN_VOLT_LSB        0x18
#define BGG_CURRENT_MSB         0x19
#define BGG_CURRENT_LSB         0x1a
#define BGG_CURR_THRESH_H_MSB   0x1b
#define BGG_CURR_THRESH_H_LSB   0x1c
#define BGG_CURR_THRESH_L_MSB   0x1d
#define BGG_CURR_THRESH_L_LSB   0x1e
#define BGG_MAX_CURR_MSB        0x1f
#define BGG_MAX_CURR_LSB        0x20
#define BGG_MIN_CURR_MSB        0x21
#define BGG_MIN_CURR_LSB        0x22
#define BGG_TEMPERATURE_MSB     0x23
#define BGG_TEMPERATURE_LSB     0x24
#define BGG_TEMP_THRESH_H_MSB   0x25
#define BGG_TEMP_THRESH_H_LSB   0x26
#define BGG_TEMP_THRESH_L_MSB   0x27
#define BGG_TEMP_THRESH_L_LSB   0x28
#define BGG_GPIO_VOLTAGE_MSB    0x29
#define BGG_GPIO_VOLTAGE_LSB    0x2a
#define BGG_GPIO_THRESH_H_MSB   0x2b
#define BGG_GPIO_THRESH_H_LSB   0x2c
#define BGG_GPIO_THRESH_L_MSB   0x2d
#define BGG_GPIO_THRESH_L_LSB   0x2e

// BGG_STATUS bits
#define BGS_GPIO_ALERT          (1 << 7)
#define BGS_CURR_ALERT          (1 << 6)
#define BGS_CHRG_OVR_ALERT      (1 << 5)
#define BGS_TEMP_ALERT          (1 << 4)
#define BGS_CHRG_HI_ALERT       (1 << 3)
#define BGS_CHRG_LOW_ALERT      (1 << 2)
#define BGS_VOLT_ALERT          (1 << 1)
#define BGS_UVLO_ALERT          (1 << 0)

// BGG_ADC_CONTROL bits
#define BGC_MODE_SLEEP          (0 << 5)
#define BGC_MODE_SMART          (1 << 5)
#define BGC_MODE_CONT_V         (2 << 5)
#define BGC_MODE_CONT_I         (3 << 5)
#define BGC_MODE_CONT_ALT       (4 << 5)
#define BGC_MODE_SS_VIT         (5 << 5)
#define BGC_MODE_CONT_VIT       (6 << 5)
#define BGC_MODE_MASK           (7 << 5)

#define BGC_GPIO_ALERT          (0 << 3)
#define BGC_GPIO_CC_MODE        (1 << 3)
#define BGC_GPIO_ANALOG_1       (2 << 3)
#define BGC_GPIO_ANALOG_2       (3 << 3)

#define BGC_USE_VDD             (0 << 2)
#define BGC_USE_SENSEN          (1 << 2)

/*!
 * @brief A thread to poll battery voltage & current.
 */
class Battery: public FAS_THREAD::Ticker
{
private:
    // This is a singleton object.
    Battery();

    static std::unique_ptr<Battery> instance;
    static std::mutex instance_mutex;

    static int noDeviceFound;
    static bool isClosed;

public:
    ~Battery();

    // Singletons should not be cloneable.
    Battery(Battery& other) = delete;

    // Singletons should not be assignable.
    void operator=(const Battery&) = delete;

    // The one and only instance.
    static Battery *factory();

    static void close()
    {
        // Make sure we have a device.
        if (!isClosed && (noDeviceFound == 0))
        {
            std::lock_guard<std::mutex> lock(instance_mutex);

            // Sanity check.
            if (instance.get() != nullptr)
            {
                // Stop the thread.
                instance->stop();
                instance->join();

                // Reset the instance.
                instance.reset();
            }

            // We are now closed.
            isClosed = true;
        }
    }

    // Our timer "tick" handler.
    void tick() override;

    void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;

    // Get the GIGE_REG_BATTERY_CHARGE
    uint32_t get_info();

    // Get the battery's voltage reading.
    double get_voltage(bool *okPtr);

    // Get the battery's current (amperage) reading.
    double get_current(bool *okPtr);

    // Return the raw voltage last read.
    double raw_voltage() { return battVoltage; }

public:
    bool isStarted;

private:
    double battVoltage;                 //!< The latest voltage reading.
    double battCurrent;                 //!< The latest current reading.
    uint32_t battChargeInfo;            //!< The GIGE_REG_BATTERY_CHARGE value.

    std::deque<double> samples;
    double sampleSum;
};

} /* End of namespace BATTERY_LOCAL */

using namespace BATTERY_LOCAL;
using namespace CS_TRACE;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<Battery> Battery::instance;
std::mutex Battery::instance_mutex;
int Battery::noDeviceFound(-1);

bool Battery::isClosed(false);

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Class implementation: Battery
//--------------------------------------------------------------------

Battery::Battery() :
    FAS_THREAD::Ticker(DEFAULT_POLLING_TIME),
    isStarted(false),
    battVoltage(0.0),
    battCurrent(0.0),
    battChargeInfo(GIGE_BATTERY_NO_INFO),
    sampleSum(0.0)
{
    // Turn on tracing.
    // traceControl |= TRACE_SEL_BATT_PROC;
    LOG("%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Telemetry class destructor.
 */
Battery::~Battery()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Create our singleton.
 */
Battery *Battery::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if ((instance.get() == nullptr) && (noDeviceFound < 0) && !isClosed)
    {
        // Create our one and only instance of the Battery object.
        instance = std::unique_ptr<Battery>(new Battery());
        if (instance.get() == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to allocate an instance of the Battery object", __PRETTY_FUNCTION__);
            return nullptr;
        }

        // Attempt to start the thread. This might fail, and that
        // is OK. It indicates that our hardware doesn't have a gas gauge.
        instance->start();

        // Did the instance start?
        if (!instance->isStarted)
        {
            // Reset our instance. We don't have to stop/join because the thread never
            // got started.
            instance.reset();

            // Rememberr that there is no device.
            noDeviceFound = 1;
        }
        else
        {
            // Remember that we have a device and it is operational.
            noDeviceFound = 0;
        }
    }

    return instance.get();
}

/*!
 * @brief Read the battery current from the gas gauge.
 */
double Battery::get_current(bool *okPtr = nullptr)
{
    uint8_t reg_value;
    uint16_t curr_dig;
    bool curr_is_neg;
    double result;

    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null.", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }

    // Read the MSB.
    if (!fpga_ptr->i2cRead(BGG_I2C_ADDR, BGG_CURRENT_MSB, &reg_value, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }
    curr_dig = ((uint16_t)reg_value) << 8;

    // Read the LSB.
    if (!fpga_ptr->i2cRead(BGG_I2C_ADDR, BGG_CURRENT_LSB, &reg_value, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }
    curr_dig |= ((uint16_t)reg_value);

    curr_is_neg = (curr_dig & 0x8000) == 0x8000;
    curr_dig &= 0x7FFF;

    // Convert the ADC value to something we can understand.
    if (curr_is_neg)
        result = ((double)curr_dig) / 32768.0;
    else
        result = ((double)(-32768 + curr_dig)) / 32768.0;
    result = (97.5 / RSENSE) * result;

    // For some reason, the current is about 10x too big.
    result = result / 10.0;

    if (okPtr != nullptr)
        *okPtr = true;

    return result;
}

/*!
 * @brief Read the battery voltage from the gas gauge.
 */
double Battery::get_voltage(bool *okPtr = nullptr)
{
    uint8_t reg_value;
    uint16_t volt_dig;
    double result;

    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null.", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }

    // Read the MSB.
    if (!fpga_ptr->i2cRead(BGG_I2C_ADDR, BGG_VOLTAGE_MSB, &reg_value, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }
    volt_dig = ((uint16_t)reg_value) << 8;

    // Read the LSB.
    if (!fpga_ptr->i2cRead(BGG_I2C_ADDR, BGG_VOLTAGE_LSB, &reg_value, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        if (okPtr != nullptr)
            *okPtr = false;
        return 0;
    }
    volt_dig |= ((uint16_t)reg_value);

    // Convert the ADC value to something we can understand.
    result = 62.6 * (((double)volt_dig) / 65536.0);

    if (okPtr != nullptr)
        *okPtr = true;

    return result;
}

void Battery::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_BATT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null.", __PRETTY_FUNCTION__);
        return;
    }

    // Attempt to read the status register. We don't care about the value, we just want
    // to see if the device is on the bus. Not all hardware versions have the gas gauge.
    uint8_t reg_value;
    if (!fpga_ptr->i2cRead(BGG_I2C_ADDR, BGG_STATUS, &reg_value, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to write the control information. This is OK, and likely due to early hardware (no gas gauge).", __PRETTY_FUNCTION__);
        fi_log(LOG_DEBUG, "%s: Battery gas gauge function will not be available.", __PRETTY_FUNCTION__);
        isStarted = false;

        // We are running, so we have power. We must be on hardware that doesn't
        // have a battery gas gauge.
        battChargeInfo = GIGE_BATTERY_NO_BATTERY;
        return;
    }

    // Configure the gas gauge to monitor the voltage and current.
    // - We want the Voltage ADC to digitize the voltage between
    //   pin 5 (SENSEN) and GND.
    // - Have the GPIO pin be an alert (unused).
    // - Have the ADC continuously convert, alternating between voltage and current.
    if (!fpga_ptr->i2cWrite(BGG_I2C_ADDR, BGG_ADC_CONTROL, (BGC_MODE_CONT_ALT | BGC_GPIO_ALERT | BGC_USE_SENSEN), FI_FPGA::I2C_IO))
    {
        fi_log(LOG_DEBUG, "%s: I2C Write failed, unable to setup the gas gauge.", __PRETTY_FUNCTION__);
        isStarted = false;

        // We are running, so we have power. We must be on hardware that doesn't
        // have a battery gas gauge.
        battChargeInfo = GIGE_BATTERY_NO_BATTERY;
        return;
    }

    // Pass on the call.
    FAS_THREAD::Ticker::start(threadFn);

    // We have officially started.
    isStarted = true;
}

void Battery::tick()
{
    bool okV(false);
    bool okI(false);
    double current(0.0);

    // Read the voltage from the gas gauge.
    battVoltage = get_voltage(&okV);
    if (okV)
        current = get_current(&okI);
    TP(TRACE_SEL_BATT_PROC, "%s: voltage=%f, okV=%d, current=%f, okI=%d", __PRETTY_FUNCTION__, battVoltage, okV, current, okI);

    // Was the read successful?
    if (!okV || !okI)
    {
        // We are running, so we have power. We must be on hardware that doesn't
        // have a battery gas gauge.
        battChargeInfo = GIGE_BATTERY_NO_BATTERY;

        // Stop the timer.
        stop();
    }

    // Is the voltage high enough that we must be on wall power?
    else if (battVoltage > MAX_BATTERY_VOLTAGE)
    {
        // If we are on wall power, then show no battery.
        battChargeInfo = GIGE_BATTERY_NO_BATTERY;
    }
    else
    {
        // Add the latest sample to our vector.
        samples.push_back(battVoltage);

        // Add the latest to our sum.
        sampleSum += battVoltage;

        // Have we reached our window size?
        double adjusted_sum;
        int win_size;
        if (samples.size() > SAMPLE_SIZE)
        {
            // Remove the oldest element from the queue and the sum.
            double old_sample = samples.front();
            samples.pop_front();
            sampleSum -= old_sample;

            // Now that we have our full window, remove the minimum and maximum values.
            auto min_value = std::min_element(samples.begin(), samples.end());
            auto max_value = std::max_element(samples.begin(), samples.end());
            adjusted_sum = sampleSum - (*min_value + *max_value);

            // Adjust our window size accordingly.
            win_size = SAMPLE_SIZE - 2;
        }
        else
        {
            // We haven't filled our window yet, so our adjust sum is our sample sum.
            adjusted_sum = sampleSum;
            win_size = samples.size();
        }

        // Make the voltage the running average.
        double voltage_avg = adjusted_sum / win_size;

        // Clip the voltage values to stay in range.
        if (voltage_avg < CRITICAL_LOW_VOLTAGE)
        {
            // This is reflecting 0% charge
            voltage_avg = CRITICAL_LOW_VOLTAGE;
        }
        if (voltage_avg > MAX_BATTERY_VOLTAGE)
        {
            // This is reflecting 100% charge.
            voltage_avg = MAX_BATTERY_VOLTAGE;
        }

        // Convert the voltage into a "percentage".
        int chargePct = (int)(((voltage_avg - CRITICAL_LOW_VOLTAGE) * 100) / (MAX_BATTERY_VOLTAGE - CRITICAL_LOW_VOLTAGE));

        // Sanity checks.
        if (chargePct < 0)
            chargePct = 0;
        if (chargePct > 100)
            chargePct = 100;

        // Output the data in a CSV-style format for easier inclusion in a spreadsheet.
        fi_log(LOG_DEBUG, "%s: [raw, avg, pct] %f, %f, %d", __PRETTY_FUNCTION__, battVoltage, voltage_avg, chargePct);

        // Set the information in the proper format.
        battChargeInfo = (((uint32_t)chargePct) << GIGE_BATTERY_PERCENT_SHIFT) | GIGE_BATTERY_DISCHARGING;
    }

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    regs[GIGE_REG_BATTERY_CHARGE].set_value(battChargeInfo);
}

uint32_t Battery::get_info()
{
    return battChargeInfo;
}

//--------------------------------------------------------------------
// External Function Implementation
//--------------------------------------------------------------------

void batteryStart()
{
    // Create the instance will automatically start the thread.
    Battery::factory();
}

void batteryClose()
{
    TP(TRACE_SEL_BATT_PROC, "%s", __PRETTY_FUNCTION__);

    // Pass it on.
    Battery::close();
}

double batteryVoltage()
{
    Battery *batt_ptr = Battery::factory();
    if (batt_ptr == nullptr)
        return 0.0;

    return batt_ptr->raw_voltage();
}
