/*!
 * @file bounce.cpp
 *
 * @brief Manage PC bounce buffer memory.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <mutex>

#include <fastec/fas_log.h>

#include "bounce.h"

namespace PC_BOUNCE
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Physical base address of the PC bounce buffer memory.
static const uint64_t PC_BOUNCE_BUFFER_BASE_ADDR = 0x300000000ULL;

//! Total number of bytes in the PC bounce buffer memory.
static const uint64_t PC_BOUNCE_BUFFER_LENGTH = 0x100000000ULL;

//! The number of bounce buffers to carve out of the memory.
static const int   NUM_BUFFERS = 16;

//! Type definition for our bounce buffers.
typedef struct BB_TYPE
{
    bool avail;         //!< Buffer is available.
    int mmap_fd;        //!< The MMAP file descriptor of a buffer.
    uint32_t mmap_len;  //!< The MMAP length.

    uint64_t addr;      //!< Physical base address of buffer.
    void *mmap_addr;    //!< The MMAP address of a buffer.
} BB_TYPE;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//! Protect access to the bounce buffer control structure(s).
static std::mutex bb_mutex;

static BB_TYPE bb_buf[NUM_BUFFERS];     //!< Our bounce buffers.
static uint32_t bb_size;                //!< Length (bytes) in each buffer.

static int xfer_mmap_fd;
static void *xfer_mmap_addr;

}   // End of PC_BOUNCE

// Make things easier for ourselves.
using namespace PC_BOUNCE;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief One-time initialization of the PC bounce buffer access logic.
 * return True on success, false otherwise.
 */
bool bb_init()
{
    uint32_t page_size = sysconf(_SC_PAGESIZE);

    // Given the number of buffers we want, and the amount of buffer size,
    // calculate the size of each buffer.
    bb_size = PC_BOUNCE_BUFFER_LENGTH / NUM_BUFFERS;

    // Make sure the size is multiple of pages.
    bb_size &= ~(page_size - 1);

//    fi_log(LOG_DEBUG, "%s: %d buffers @ %u [0x%08x] bytes each", __PRETTY_FUNCTION__, NUM_BUFFERS, bb_size, bb_size);

    // Initialize our information.
    for (int i = 0; i < NUM_BUFFERS; ++i)
    {
        bb_buf[i].avail = true;
        bb_buf[i].mmap_fd = -1;
        bb_buf[i].mmap_addr = MAP_FAILED;
        bb_buf[i].mmap_len = 0;
        bb_buf[i].addr = PC_BOUNCE_BUFFER_BASE_ADDR + (i * bb_size);
//        fi_log(LOG_DEBUG, "%s: index=%d, addr=0x%" PRIx64, __PRETTY_FUNCTION__, i, bb_buf[i].addr);
    }

    // Initialize our transfer buffer information.
    xfer_mmap_fd = -1;
    xfer_mmap_addr = MAP_FAILED;

    // All is well.
    return true;
}

int bb_get_avail()
{
    int avail;
    std::unique_lock<std::mutex> locker(bb_mutex);

    // Initialize our available index to not found.
    avail = -1;

    // See if we can find one.
    for (int i = 0; i < NUM_BUFFERS; ++i)
    {
        // Is this buffer available?
        if (bb_buf[i].avail == true)
        {
            // Remember the index.
            avail = i;

            // Mark it as busy.
            bb_buf[i].avail = false;

            // We're done.
            break;
        }
    }

    // Return what we've found.
    return avail;
}

uint64_t bb_get_avail_addr(int index)
{
    uint64_t ret_addr = 0;

    // Sanity check the index.
    if ((index >= 0) && (index < NUM_BUFFERS))
        ret_addr = bb_buf[index].addr;

    // Return what we've found.
    return ret_addr;
}

uint32_t bb_read(int index, uint8_t *dest_ptr, uint32_t dest_len)
{
    uint8_t *bb_buf;
    uint32_t read_count = 0;

    // Sanity check
    if ((index < 0) || (index >= NUM_BUFFERS))
    {
        fi_log(LOG_DEBUG, "%s: Invalid index=%d", __PRETTY_FUNCTION__, index);
        goto bail;
    }

    // Map the memory.
    // fi_log(LOG_DEBUG, "%s: size: 0x%x, address: 0x%" PRIx64 "", __PRETTY_FUNCTION__, dest_len, bb_buf[index].addr);
    bb_buf = (uint8_t *)bb_mmap(index, dest_len);
    if (bb_buf == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto mem_bail;
    }

    // Copy the memory into our buffer.
    memcpy(dest_ptr, bb_buf, dest_len);

    // All done!
    read_count = dest_len;

mem_bail:
    // Unmap the memory.
    bb_unmap(index);

bail:
    return read_count;
}

uint32_t bb_write(int index, const uint8_t *src_ptr, uint32_t src_len)
{
    uint8_t *bb_buf;
    uint32_t write_count = 0;

    // Sanity check
    if ((index < 0) || (index >= NUM_BUFFERS))
    {
        fi_log(LOG_DEBUG, "%s: Invalid index=%d", __PRETTY_FUNCTION__, index);
        goto bail;
    }

    // Map the memory.
    // fi_log(LOG_DEBUG, "%s: size: 0x%x, address: 0x%" PRIx64 "", __PRETTY_FUNCTION__, src_len, bb_buf[index].addr);
    bb_buf = (uint8_t *)bb_mmap(index, src_len);
    if (bb_buf == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto mem_bail;
    }

    // Copy the memory into our bounce buffer.
    memcpy(bb_buf, src_ptr, src_len);

    // All done!
    write_count = src_len;

mem_bail:
    // Unmap the memory.
    bb_unmap(index);

bail:
    return write_count;
}

void bb_release(int index)
{
    std::unique_lock<std::mutex> locker(bb_mutex);

    // Sanity check the index.
    if ((index >= 0) && (index < NUM_BUFFERS))
    {
        bb_buf[index].avail = true;

        // As a safety measure, check to see if the user had this
        // buffer MMAP'd. If so, unmap the buffer. Note that we
        // are duplicating bb_unmap(). This is because we are already
        // "bb_mutex" safe.
        if (bb_buf[index].mmap_fd > 0)
        {
            // Was the memory mapped?
            if (bb_buf[index].mmap_addr != MAP_FAILED)
            {
                // Unmap the memory.
                munmap(bb_buf[index].mmap_addr, bb_buf[index].mmap_len);
                bb_buf[index].mmap_addr = MAP_FAILED;
                bb_buf[index].mmap_len = 0;
            }

            // Close the mmap file descriptor.
            close(bb_buf[index].mmap_fd);
            bb_buf[index].mmap_fd = -1;
        }
    }
}

uint64_t bb_get_base()
{
    return PC_BOUNCE_BUFFER_BASE_ADDR;
}

void *bb_mmap(int index, uint32_t dest_len, int flags)
{
    int prot_flags;
    std::unique_lock<std::mutex> locker(bb_mutex);

    // Sanity check the index.
    if ((index >= 0) && (index < NUM_BUFFERS))
    {
        // Do we already have a map?
        if (bb_buf[index].mmap_addr != MAP_FAILED)
            goto bail;

        // Memory map the destination address.
        bb_buf[index].mmap_fd = open("/dev/mem", flags);
        if (bb_buf[index].mmap_fd < 1)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            goto bail;
        }

        // Are we doing the entire buffer?
        if (dest_len == 0)
            dest_len = bb_size;

        // Figure out which protection flags we need.
        if ((flags & O_ACCMODE) == O_RDONLY)
            prot_flags = PROT_READ;
        else if ((flags & O_ACCMODE) == O_WRONLY)
            prot_flags = PROT_WRITE;
        else
            prot_flags = PROT_READ | PROT_WRITE;

        // Now map the memory.
        // fi_log(LOG_DEBUG, "%s: size: 0x%x, address: 0x%" PRIx64 "", __PRETTY_FUNCTION__, dest_len, bb_buf[index].addr);
        bb_buf[index].mmap_addr = mmap(NULL, dest_len, prot_flags, MAP_SHARED | MAP_NORESERVE, bb_buf[index].mmap_fd, bb_buf[index].addr);
        if (bb_buf[index].mmap_addr == MAP_FAILED)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            close(bb_buf[index].mmap_fd);
            bb_buf[index].mmap_fd = -1;
        }
        else
        {
            // Save the length we've just mmapped.
            bb_buf[index].mmap_len = dest_len;
        }
    }

bail:
    return bb_buf[index].mmap_addr;
}

uint32_t bb_get_mmap_len(int index)
{
    uint32_t map_len = 0;

    // Sanity check the index.
    if ((index >= 0) && (index < NUM_BUFFERS))
    {
        // Make sure we're mapped!
        if (bb_buf[index].mmap_addr != MAP_FAILED)
        {
            map_len = bb_buf[index].mmap_len;
        }
    }

    return map_len;
}

void bb_unmap(int index)
{
    std::unique_lock<std::mutex> locker(bb_mutex);

    // Sanity check the index.
    if ((index >= 0) && (index < NUM_BUFFERS))
    {
        // Make sure the file descriptor was opened.
        if (bb_buf[index].mmap_fd > 0)
        {
            // Was the memory mapped?
            if (bb_buf[index].mmap_addr != MAP_FAILED)
            {
                // Unmap the memory.
                munmap(bb_buf[index].mmap_addr, bb_buf[index].mmap_len);
                bb_buf[index].mmap_addr = MAP_FAILED;
                bb_buf[index].mmap_len = 0;
            }

            // Close the mmap file descriptor.
            close(bb_buf[index].mmap_fd);
            bb_buf[index].mmap_fd = -1;
        }
    }
}

