﻿/*!
 * @file camera_server.cxx
 * @brief This file implements the CameraServer class.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <pwd.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <fmt/format.h>
#include <fmt/printf.h>

#include <chrono>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <list>
#include <mutex>
#include <vector>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "camera_server.h"

#include "autosave.h"
#include "battery.h"
#include "cs_trace.h"
#include "cs_update.h"
#include "fan_proc.h"
#include "fcp.h"
#include "formatter.h"
#include "fsp.h"
#include "image_correction.h"
#include "image_io.h"
#include "jpeg_proc.h"
#include "kernel_timezone.h"
#include "mft.h"
#include "part_mgr.h"
#include "pb_control.h"
#include "storage.h"
#include "wb_file.h"

namespace CAMERA_SERVER
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

static const int ERROR_WAIT_LONG = 10;              //!< Number of seconds to wait after an error before trying again.
static const int SENSOR_RESET_COUNT = 5;            //!< Number of retries to get sensor to re-align.
static const uint32_t MSGQ_WAIT_SECONDS = 30;       //!< Number of seconds to wait on the msg_queue before timing out.

static const int SHMEM_DEFAULT_DESCRIPTOR_COUNT = 1000; //!< Number of descriptors in the shared memory ring buffer.

static const char BOARD_REV_ENV_VAR[] = "BOARDREV"; //!< The environment variable used to store the board version information.
static const char DEFAULT_HS_BOARD_REV[] = "A0";    //!< Default board revision.

//! Location for the update command file.
static const char UPDATE_CMD_FILE[] = ".config/Fastec/scratch/update.sh";

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

volatile bool   exitRequested;      //!< A signal or other event requested a close
volatile bool   rebootRequested;    //!< Someone has requested we reboot.
volatile bool   restartRequested;   //!< Someone has requested we restart.
volatile bool   shutdownRequested;  //!< Someone has requested we shutdown.

//! Set to trigger a dump of FPGA registers at first opportunity.
static volatile bool do_register_dump = false;

//! Set to toggle the sensor test mode. Actually right now we only turn it on, not off.
static volatile bool sensor_test_toggle = false;

} /* End of namespace CAMERA_SERVER */

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

// Mutexes to protect items used by main thread, fpga event, and select threads.
std::mutex               CameraServer::work_mutex;
std::condition_variable  CameraServer::work_cond;
std::list<CS_NETWORK::CS_NETWORK_MESSAGE>    CameraServer::msg_queue;

FSM_HANDLE CameraServer::shmem_handle = nullptr;    //!< Shared memory ring buffer handle.
CameraServer *CameraServer::camserv_ptr = nullptr;  //!< Pointer to ourselves.

TRACE_GLOBAL;           //!< Use the global trace control word.

using namespace CAMERA_SERVER;
using namespace CS_TRACE;
using namespace CS_NETWORK;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Flag that a USR1 or USR2 signal was seen.
 *
 * @param which     1 for USR1 or 2 for USR2
 */
void CameraServer::saw_signal(int which)
{
    if (which == 1)
        do_register_dump = true;
    else if (which == 2)
        sensor_test_toggle = true;
}

/*!
 * @brief This function pushes work on to the work queu in a thread safe manner.
 *
 * @param msg_ptr   Pointer to the received message.
 */
void CameraServer::add_work(CS_NETWORK::CS_NETWORK_MESSAGE *msg_ptr)
{
    std::unique_lock<std::mutex> locker(work_mutex);

    // Push the new message on to the work queue.
    msg_queue.push_back(*msg_ptr);

    // Let folks know there's work to be done.
    work_cond.notify_one();
}

/*!
 * @brief Function used to retrieve the FPGA class pointer.
 */
CameraServer *CameraServer::get_cs()
{
    return camserv_ptr;
}

/*!
 * @brief Function used to retrieve the FPGA class pointer.
 */
FiFPGA *CameraServer::get_fpga()
{
    return FiFPGA::getInstance();
}

/*!
 * @brief Function used to retrieve the partition manager class pointer.
 */
PartMgr *CameraServer::get_partmgr()
{
    return PartMgr::getInstance();
}

/*!
 * @brief Function used to retrieve the sensor class pointer.
 */
Sensor *CameraServer::get_sensor()
{
    if ((camserv_ptr == nullptr) || (camserv_ptr->sensor_ptr.get() == nullptr))
        return nullptr;

    return camserv_ptr->sensor_ptr.get();
}

/*!
 * @brief Function used to retrieve the autosave management class pointer.
 */
AutoSave *CameraServer::get_autosave()
{
    if (camserv_ptr == nullptr)
        return nullptr;

    return camserv_ptr->as_mgr.get();
}

/*!
 * @brief Function used to retrieve the register cache class pointer.
 */
GIGE_REG_CACHE& CameraServer::get_regs()
{
    if (camserv_ptr == nullptr)
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s: cs=NULL caller=%p!", __PRETTY_FUNCTION__, (void *)caller);
        // We will crash now.
    }

    return camserv_ptr->reg_cache;
}

/*!
 * @brief Function used to retrieve the worker class pointer.
 */
csWorker *CameraServer::get_worker()
{
    if (camserv_ptr == nullptr)
        return nullptr;

    return camserv_ptr->worker.get();
}

uint8_t CameraServer::get_camera_state()
{
    if (camserv_ptr == nullptr)
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s: cs=NULL caller=%p!", __PRETTY_FUNCTION__, (void *)caller);
        return GIGE_CAMERA_STATE_IDLE;
    }
    return camserv_ptr->camera_state;
}

const char *CameraServer::board_rev()
{
    char *env_ptr;

    // Sanity check.
    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_WARNING, "%s: NULL camserv_ptr!", __PRETTY_FUNCTION__);
        return nullptr;
    }

    // Is there an environment variable?
    env_ptr = getenv(BOARD_REV_ENV_VAR);
    if ((env_ptr != nullptr) && (env_ptr[0] != '\0'))
    {
        fi_log(LOG_DEBUG, "%s: Found %s = %s", __PRETTY_FUNCTION__, BOARD_REV_ENV_VAR, env_ptr);
        return env_ptr;
    }

    // Return the default value.
    return DEFAULT_HS_BOARD_REV;
}

/*!
 * @brief Check the boardrev environment variable against the
 *        specified string.
 *
 * @param rev_ptr   [UNUSED] String to compare against.
 *
 * @return TRUE if OK, FALSE otherwise.
 */
bool CameraServer::rev_is_ok(const char *rev_ptr)
{
    (void)rev_ptr;

    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_WARNING, "%s: NULL camserv_ptr!", __PRETTY_FUNCTION__);
        return false;
    }

    return true;
}

/*!
 * @brief Function used to retrieve the license pointer.
 */
const LicManager *CameraServer::get_license()
{
    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_WARNING, "%s: NULL camserv_ptr!", __PRETTY_FUNCTION__);
        return nullptr;
    }
    return camserv_ptr->lic_ptr.get();
}

/*!
 * @brief Function used to retrieve the shared memory ring buffer handle.
 */
FSM_HANDLE CameraServer::get_shmem()
{
    return shmem_handle;
}

/*!
 * @brief Handle events coming back from the FPGA.
 *
 * @param fpga_val  The value we read for the FPGA event.
 */
void CameraServer::process_fpga_events(uint32_t fpga_val)
{
    // Let the sensor know an event has happened.
    FiFPGA::getInstance()->eventHappened(fpga_val);
}

/*!
 * @brief Allow the application to ask the Camera Server to terminate.
 */
void CameraServer::exit()
{
    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);

    // Make sure the loop knows we want to exit.
    exitRequested = true;

    // Wake up the loop.
    work_cond.notify_one();
}

void CameraServer::reboot()
{
    // Make sure we remember to reboot!
    rebootRequested = true;
    CameraServer::exit();
}

void CameraServer::restart()
{
    // Make sure we remember to restart!
    restartRequested = true;
    CameraServer::exit();
}

void CameraServer::shutdown()
{
    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);

    // Make sure the loop knows we want to exit.
    exitRequested = true;
    shutdownRequested = true;

    // Wake up the loop.
    work_cond.notify_one();
}

/*!
 * @brief Allow the application to ask the Camera Server to reopen
 *    the IPC and UDP sockets.
 */
void CameraServer::reopen()
{
    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);

    // Pass the request on.
    if (camserv_ptr != nullptr)
        camserv_ptr->worker->reopen();
}

/*
 * @brief When in LIVE_VIDEO state, set the hardware as if we were IDLE
 */
void CameraServer::set_idle_behavior()
{
    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);
    FiFPGA::getInstance()->icmDisable();
}

uint8_t CameraServer::set_idle_state()
{
    bool    disable_autosave = false;

    // We want to go to idle, decide what to do by the
    // state that the camera is in.
    switch (CameraServer::get_camera_state())
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        FiFPGA::getInstance()->icmDisable();
        break;
    case GIGE_CAMERA_STATE_ARMED:
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;
        FiFPGA::getInstance()->icmDisable();
        break;
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_TRIGGERED:
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;
        FiFPGA::getInstance()->icmDisable();
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        FiFPGA::getInstance()->icmDisable();
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        FiFPGA::getInstance()->icmDisable();
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        FiFPGA::getInstance()->icmDisable();
        break;
    default:
        break;
    }

    // Decide what needs to be done based upon what state we are transitioning into.
    // Are we supposed to disable "auto-save"?
    if (disable_autosave)
    {
        // Mark the active partition as empty.
        PartMgr::getInstance()->setPartitionState(PartMgr::getInstance()->getActivePartition(), GIGE_PART_STATE_EMPTY);

        // Make sure autosave is enabled.
        if (as_mgr->isEnabled())
        {
            // Disable autosave.
            fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);
        }
    }
    return GIGE_CAMERA_STATE_IDLE;
}

/*
 * @brief When in IDLE via set_idle_behavior(), set the hardware back to LIVE_VIDEO
 */
void CameraServer::set_live_behavior()
{
    FiFPGA::getInstance()->icmSetToLiveVideo();
}

uint8_t CameraServer::set_live_state()
{
    bool    disable_autosave = false;

    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);

    // Sanity check our pointers.
    if (sensor_ptr.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Null sensor pointer!", __PRETTY_FUNCTION__);
        // We don't have any provision for this state change failing.
        return GIGE_CAMERA_STATE_LIVE_VIDEO;
    }

    // We want to go to live video, decide what to do by the
    // state that the camera is in.
    switch (CameraServer::get_camera_state())
    {
    case GIGE_CAMERA_STATE_IDLE:
        FiFPGA::getInstance()->icmSetToLiveVideo();
        break;
    case GIGE_CAMERA_STATE_ARMED:
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;
        // We already disabled the ARM-IN input if it was level-sensitive.
        break;
    case GIGE_CAMERA_STATE_DEARMING:
        // The dearm was done already when we stopped capture
        // and the current partion was already either enqueued for save or
        // marked empty, so there is little left to do.
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;
        // We already disabled the ARM-IN input if it was level-sensitive.
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        // We already have Unarmed Live Video.
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        FiFPGA::getInstance()->icmSetToLiveVideo();
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        FiFPGA::getInstance()->icmSetToLiveVideo();
        break;
    default:
        break;
    }

    // Decide what needs to be done based upon what state we are transitioning into.
    // Are we supposed to disable "auto-save"?
    if (disable_autosave)
    {
        // Mark the active partition as empty.
        PartMgr::getInstance()->setPartitionState(PartMgr::getInstance()->getActivePartition(), GIGE_PART_STATE_EMPTY);

        // If autosave is enabled, make sure we disable it.
        if (as_mgr->isEnabled())
        {
            // Disable autosave.
            fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);
        }
    }

    // Clear out any queued up partitions in the capture FIFO.
    FiFPGA::getInstance()->icmResetCaptureFIFO();

    // Update our LED.
    FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_OFF);

    // We can accept an ARM transition now if so configured.
    FiFPGA::getInstance()->armInResume();

    return GIGE_CAMERA_STATE_LIVE_VIDEO;
}

uint8_t CameraServer::set_armed_state()
{
    uint8_t new_state;
    uint8_t cur_state;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    bool    in_ltr_mode;
    bool    fflr_mode;
    uint32_t    cap_mode = 0;

    // Get the current state.
    new_state = cur_state = CameraServer::get_camera_state();

    in_ltr_mode = ((FiFPGA::getInstance()->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    fflr_mode = FiFPGA::getInstance()->ltrFasFireEnabled();

    cap_mode = regs[ GIGE_REG_LTR_MODE ].get_value();
    cap_mode &= (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

    // If we are in the LTR mode, don't allow arming if we are formatting the drive.
    if (in_ltr_mode)
    {
        // No state change while SSD format is in progress.
        if (formatter::isSsdFormatInProgress())
            return cur_state;
    }

    // Check to see if we are in auto-trigger. If we are, then we need to
    // make sure the trigger position is set appropriately.
    else if (auto_trigger())
    {
        uint32_t pref;

        // See if auto playback is enabled.
        pref = regs[GIGE_REG_PREFERENCES].get_value();
        if ((pref & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY)
        {
            // Force a start trigger.
            syslog(LOG_DEBUG, "%s: Forcing start trigger for auto-trigger", __PRETTY_FUNCTION__);
            PartMgr::setTriggerPos(nullptr, 0);
        }
    }

    // Make sure there is an entry in the capture FIFO. If we are already
    // armed, then the FPGA has likely already grabbed the FIFO entry.
    if (!FiFPGA::getInstance()->icmIsArmed())
    {
        // If we are not doing overlap mode (FasFire), then make sure we reset the
        // capture FIFO to a known condition.
        if (!as_mgr->isEnabled())
        {
            // Reset the capture FIFO. Even if there is nothing in the FIFO, it won't hurt to reset it.
            FiFPGA::getInstance()->icmResetCaptureFIFO();

            // Now the capture FIFO is back to a known position, put partition 0 in.
            TP(TRACE_SEL_CAMERA_STATE, "%s: Nothing in the capture FIFO, Not in autosave, so queuing up the active partition 0.", __PRETTY_FUNCTION__);
            PartMgr::getInstance()->setActivePartition(0);
            PartMgr::getInstance()->queueCapturePartition(PartMgr::getInstance()->getActivePartition());
        }

        // Are we in FFLR mode?
        if (fflr_mode)
        {
            TP(TRACE_SEL_CAMERA_STATE, "%s: FFLR Enabled, not armed, cur_state=%d", __PRETTY_FUNCTION__, cur_state);

            // Make sure we restore any previously saved "save settings". Enable autosave
            // if it is not enabled.
            imgSaveWorker::getInstance()->restoreSaveSettings(!as_mgr->isEnabled());
        }
    }

    // We want to go to ARMED. Decide what to do by the
    // state that the camera is in.
    switch (cur_state)
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        // If in LIVE_VIDEO state, the display controller is running with unarmed
        // live view enabled, the ICM pipeline is enabled, and the sensor readout is enabled.
        // We might have to stop and restart to change the frame rate if we were in low light mode.
        // But the DCTL is set for both unarmed live view and armed live view so as soon as it
        // gets the Armed signal from the ICM, it will switch modes on its own.
        if (FiFPGA::getInstance()->icmArm(true))
            new_state = GIGE_CAMERA_STATE_ARMED;
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        TP(TRACE_SEL_CAMERA_STATE, "%s: FFLR (%d) or LTR (%d), ROC/BROC (0x%08x)", __PRETTY_FUNCTION__, fflr_mode, in_ltr_mode, cap_mode);
        if ((fflr_mode || in_ltr_mode) && (cap_mode != 0))
        {
            new_state = GIGE_CAMERA_STATE_ARMED;
        }
        // Other than ROC/BROC, we don't allow a transition from triggered to armed.
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        if (FiFPGA::getInstance()->icmArm(true))
        {
            new_state = GIGE_CAMERA_STATE_ARMED;
        }
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        // This enables the ICM pipeline if necessary.
        FiFPGA::getInstance()->icmSetToLiveVideo();
        if (FiFPGA::getInstance()->icmArm(true))
            new_state = GIGE_CAMERA_STATE_ARMED;
        else
        {
            // If we couldn't ARM, try to undo and get back to review.
            FiFPGA::getInstance()->icmSensorReadoutEnable(false);

            // We have to stop the stream if it's running.
            fcp_stop_stream_thread();
        }
        break;
    // Don't allow these transitions.
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_IDLE:
        break;
    default:
        break;
    }

    if (new_state == GIGE_CAMERA_STATE_ARMED)
    {
        // Update our LED.
        FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_SLOW_BLINK);

        // If the ARM-in is level sensitive, it can unarm us but it might need to be reenabled.
        if ((regs[ GIGE_REG_ARM_IN_POLARITY ] & GIGE_ARM_IN_EDGE_MASK) == GIGE_ARM_IN_LEVEL_TYPE)
            FiFPGA::getInstance()->armInResume();
    }

    return new_state;
}

uint8_t CameraServer::set_triggered_state()
{
    uint8_t new_state;
    uint8_t cur_state;

    // We want to go to triggered, decide what to do by the
    // state that the camera is in.
    new_state = cur_state = CameraServer::get_camera_state();
    switch (cur_state)
    {
    case GIGE_CAMERA_STATE_TRIGGERED:
    case GIGE_CAMERA_STATE_ARMED:
        new_state = GIGE_CAMERA_STATE_TRIGGERED;
        break;
    // None of these is a legal transition.
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_AUTOSAVE:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    default:
        break;
    }

    if (new_state == GIGE_CAMERA_STATE_TRIGGERED)
    {
        // Update our LED.
        FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_FAST_BLINK);
    }

    return new_state;
}

uint8_t CameraServer::set_dearming_state()
{
    uint8_t new_state;
    uint8_t cur_state;
    CS_NETWORK::CS_NETWORK_MESSAGE msg;

    // We want to end up in review (or maybe autosave) but it can take a while
    // to flush in LTR mode. We will queue up an internal work request if it
    // makes sense.
    cur_state = CameraServer::get_camera_state();
    switch (cur_state)
    {
    case GIGE_CAMERA_STATE_ARMED:
        msg.msg_type = MSG_TYPE_INTERNAL;
        msg.internal.type = CS_INTERNAL_DEARM;
        msg.internal.flags = 0;   // No trigger yet.
        add_work(&msg);
        new_state = GIGE_CAMERA_STATE_DEARMING;
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        msg.msg_type = MSG_TYPE_INTERNAL;
        msg.internal.type = CS_INTERNAL_DEARM;
        msg.internal.flags = 1;   // We saw the trigger already.
        add_work(&msg);
        new_state = GIGE_CAMERA_STATE_DEARMING;
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        fi_log(LOG_DEBUG, "%s: got DEARM in AUTOSAVE state", __FUNCTION__);
        new_state = cur_state;
        break;
    // None of the others are valid transitions.
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_DEARMING:
    default:
        new_state = cur_state;
        break;
    }
    return new_state;
}

uint8_t CameraServer::set_review_state()
{
    uint8_t new_state;
    uint8_t cur_state;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t trigger;

    new_state = cur_state = CameraServer::get_camera_state();
    TP(TRACE_SEL_CAMERA_STATE, "%s: state=%d", __PRETTY_FUNCTION__, cur_state);

    // Make sure this even makes sense. Can't go to REVIEW if there's no recording.
    if (PartMgr::getInstance()->getPartitionState() != GIGE_PART_STATE_COMPLETE)
    {
        // Don't change state.
        fi_log(LOG_ERR, "%s: Can't %u->REVIEW without COMPLETE partition (%d/%d)",
               __PRETTY_FUNCTION__, cur_state, PartMgr::getInstance()->getActivePartition(), PartMgr::getInstance()->getPartitionState());
        return cur_state;
    }

    // We want to go to review, decide what to do by the
    // state that the camera is in.
    switch (cur_state)
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();

        // Shutdown the ICM.
        FiFPGA::getInstance()->icmPipelineDisable();
        FiFPGA::getInstance()->icmSensorReadoutEnable(false);
        new_state = GIGE_CAMERA_STATE_REVIEW;
        break;
    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_TRIGGERED:
        // Transition the FPGA from the ARMed state.
        FiFPGA::getInstance()->icmDearm();

        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();

        // Shutdown the ICM.
        FiFPGA::getInstance()->icmPipelineDisable();
        FiFPGA::getInstance()->icmSensorReadoutEnable(false);
        new_state = GIGE_CAMERA_STATE_REVIEW;
        break;
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_AUTOSAVE:
        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        new_state = GIGE_CAMERA_STATE_REVIEW;
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
    default:
        break;
    }

    // Nothing else to do if we can't change states.
    if (new_state == cur_state)
    {
        return cur_state;
    }

    // Make sure autosave is enabled, so we can disable it.
    if (as_mgr->isEnabled())
    {
        // Disable autosave.
        fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);
    }

    // This marks the active partition as the one used by P1.
    PartMgr::getInstance()->setReviewPartNum(PartMgr::getInstance()->getActivePartition());

    // To "update metadata" means that the current FPGA color processing settings
    // are updated in the pm_metadata structure and copied to image memory.
    // Now that structure is current and ready to be saved.
    PartMgr::getInstance()->updateActiveMetadata();

    // Set up playback registers for the current partition. If we are in LTR mode,
    // addresses will be LBA sector numbers, otherwise they will be QWORD pointers
    // in image memory.

    // Update our LED.
    FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_OFF);

    // Set the current frame on the trigger in case anyone asks.
    // We won't send a notification because the UIs don't expect to see
    // this register change until after they've seen the state change.
    trigger = PartMgr::getInstance()->getTriggerFrame();
    regs[ GIGE_REG_CSI_RVW_FRAME_NUM ].set_value(trigger);

    // Pause the ARM-in signal processing if not in append mode but reenable it in append mode
    // or if we are allowing overwrites of recordings.
    if (((FiFPGA::getInstance()->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE) &&
        ((regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0))
    {
        FiFPGA::getInstance()->armInResume();
    }
    else
    {
        uint32_t prefs;

        prefs = regs[ GIGE_REG_PREFERENCES ].get_value();
        if ((prefs & GIGE_PREF_ARM_IN_OVERWRITE) != 0)
            FiFPGA::getInstance()->armInResume();
        else
            FiFPGA::getInstance()->armInPause();
    }

    // Update our playback control information.
    pbc_update_cut_in(nullptr, 0);
    pbc_update_cut_out(nullptr, PartMgr::getInstance()->getNumFrames() - 1);

    return new_state;
}

uint8_t CameraServer::set_autosave_state()
{
    uint8_t new_state;
    uint8_t cur_state;
    bool    disable_autosave = false;

    new_state = cur_state = CameraServer::get_camera_state();
    TP(TRACE_SEL_CAMERA_STATE, "%s: camera_state=%d", __PRETTY_FUNCTION__, cur_state);

    // We want to go to autosave, decide what to do by the
    // state that the camera is in.
    switch (cur_state)
    {
    case GIGE_CAMERA_STATE_TRIGGERED:
        // Are we in autosave mode? If so, do not dearm the FPGA..
        if (!as_mgr->doOverlap())
        {
            TP(TRACE_SEL_CAMERA_STATE, "%s: Not overlapping, so go ahead and dearm...", __PRETTY_FUNCTION__);
            FiFPGA::getInstance()->icmDearm();
        }
        else
        {
            TP(TRACE_SEL_CAMERA_STATE, "%s: Overlapping - do NOT dearm!!", __PRETTY_FUNCTION__);
        }

        // Maintain the unarmed live view while saving.
        as_mgr->process();
        new_state = GIGE_CAMERA_STATE_AUTOSAVE;
        break;
    case GIGE_CAMERA_STATE_ARMED:
        // When we are in FasFireLR mode, we handle the transition to AUTOSAVE much the
        // same as the "normal" TRIGGERED state transition.
        if (FiFPGA::getInstance()->ltrFasFireEnabled())
        {
            TP(TRACE_SEL_CAMERA_STATE, "%s: We are armed, and in FFLR, switching to AUTOSAVE.", __PRETTY_FUNCTION__);

            // Are we in autosave mode?
            if (as_mgr->doOverlap())
            {
                TP(TRACE_SEL_CAMERA_STATE, "%s: Overlapping - do NOT dearm!!", __PRETTY_FUNCTION__);

                // Maintain the unarmed live view while saving.
                as_mgr->process();
                new_state = GIGE_CAMERA_STATE_AUTOSAVE;
            }
            else
            {
                TP(TRACE_SEL_CAMERA_STATE, "%s: Not overlapping, so ignoring...", __PRETTY_FUNCTION__);
            }
        }
        break;
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_DEARMING:
    default:
        break;
    }

    // Nothing to do if we can't change states.
    if (new_state == cur_state)
    {
        return cur_state;
    }

    // Decide what needs to be done based upon what state we are transitioning into.
    // Are we supposed to disable "auto-save"?
    if (disable_autosave)
    {
        // Mark the active partition as empty.
        PartMgr::getInstance()->setPartitionState(PartMgr::getInstance()->getActivePartition(), GIGE_PART_STATE_EMPTY);

        // Make sure autosave is enabled.
        if (as_mgr->isEnabled())
        {
            // Disable autosave.
            fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);
        }
    }

    // Update our LED.
    FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_OFF);

    // Make sure we pause the ARM-in signal processing.
    FiFPGA::getInstance()->armInPause();

    return GIGE_CAMERA_STATE_AUTOSAVE;
}

uint8_t CameraServer::set_snapshot_state()
{
    bool    disable_autosave = false;

    // we want to go to single frame snapshot
    switch (CameraServer::get_camera_state())
    {
    case GIGE_CAMERA_STATE_IDLE:
        // to get a snapshot we first have to get live images flowing, once
        // they are flowing then we can put us into snapshot mode
        FiFPGA::getInstance()->icmSetToLiveVideo();

        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    case GIGE_CAMERA_STATE_ARMED:
        // I'm not sure this make sense (armed -> snapshot), but
        // give it a try.
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;

        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_TRIGGERED:
        // I'm not sure this make sense (triggered -> snapshot), but
        // give it a try.
        FiFPGA::getInstance()->icmDearm();
        disable_autosave = true;

        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        // to get a snapshot we first have to get live images flowing, once
        // they are flowing then we can put us into snapshot mode
        FiFPGA::getInstance()->icmSetToLiveVideo();

        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();
        break;
    default:
        break;
    }

    // Decide what needs to be done based upon what state we are transitioning into.
    // Are we supposed to disable "auto-save"?
    if (disable_autosave)
    {
        // Mark the active partition as empty.
        PartMgr::getInstance()->setPartitionState(PartMgr::getInstance()->getActivePartition(), GIGE_PART_STATE_EMPTY);

        // Make sure autosave is enabled.
        if (as_mgr->isEnabled())
        {
            // Disable autosave.
            fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);
        }
    }

    // Update our LED.
    FiFPGA::getInstance()->ledSetBlink(FI_FPGA::LED_OFF);

    // Make sure we pause the ARM-in signal processing.
    FiFPGA::getInstance()->armInPause();

    return GIGE_CAMERA_STATE_SNAPSHOT;
}

static const char *state_to_str(uint8_t state)
{
    const char *p;

    switch (state)
    {
    case GIGE_CAMERA_STATE_ARMED:
        p = "ARMED";
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        p = "TRIGGERED";
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        p = "REVIEW";
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        p = "AUTOSAVE";
        break;
    case GIGE_CAMERA_STATE_DEARMING:
        p = "DEARMING";
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        p = "SNAPSHOT";
        break;
    case GIGE_CAMERA_STATE_IDLE:
        p = "IDLE";
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        p = "LIVE";
        break;
    default:
        p = "confused";
        break;
    }
    return p;
}

/*!
 * @brief Handle an external request to update the state of the camera.
 *
 * We sometimes need to distinguish a camera state change ordered by an external agent
 * from one generated inside the camera.
 *
 * @param new_state State requested via FCP.
 *
 * @return true if changing the state was successful.
 */
bool CameraServer::user_set_camera_state(uint8_t new_state)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_CAMERA_STATE, "%s: %d", __FUNCTION__, new_state);

    // New handling is to treat setting the camera state to triggered as different from hitting
    // the software trigger, hoping that the GUIs won't notice that the state doesn't change
    // immediately -- We will now wait until the FPGA event to make the state transition.
    // This avoids triggering loops in ROC/BROC mode. We used to set the TRIGGERED state when we
    // got the trigger event from the FPGA, but the code didn't distinguish that from an external
    // command to set the camera state. If it was the FPGA event caused by a trigger button press
    // (on an A3+ camera), it would still set the software trigger just in case it was needed.
    // But in ROC/BROC that would generate another FPGA event.
    //
    // Setting the state to TRIGGERED means that we need to hit the software trigger.
    // We can do it in the TRIGGERED state if we are in ROC/BROC mode and that's not harmful
    // if we happen to not be in ROC/BROC.
    if ((new_state == GIGE_CAMERA_STATE_TRIGGERED) &&
        ((CameraServer::get_camera_state() == GIGE_CAMERA_STATE_ARMED) ||
         (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_TRIGGERED)))
    {
        return FiFPGA::getInstance()->icmSwTriggerFire();
    }
    // Setting the state to ARMED when we are TRIGGERED means that we need to hit
    // the software trigger.
    else if ((new_state == GIGE_CAMERA_STATE_ARMED) &&
             (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_TRIGGERED))
    {
        return FiFPGA::getInstance()->icmSwTriggerFire();
    }

    // If we get an external request to leave capture state and the external ARM-in is set to
    // level sensitive, we don't want to immediately jump back into ARMed state if it happens
    // to be active. To avoid that, we will just disable that input. (Caveat emptor, make up
    // your mind whether your arming is via GUI or by hardware input.)
    // But if the state is being set internally, such as on an UNARM event of the ARM input,
    // we don't want to disable it; we will want to ARM again on the next transition.
    if (((new_state == GIGE_CAMERA_STATE_LIVE_VIDEO) ||
         (new_state == GIGE_CAMERA_STATE_DEARMING)) &&
        ((CameraServer::get_camera_state() == GIGE_CAMERA_STATE_TRIGGERED) ||
         (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_ARMED)))
    {
        if ((regs[ GIGE_REG_ARM_IN_POLARITY ] & GIGE_ARM_IN_EDGE_MASK) == GIGE_ARM_IN_LEVEL_TYPE)
        {
            uint32_t reg_value;

            FiFPGA::armInEnable(nullptr, false);
            reg_value = regs[ GIGE_REG_ARM_IN_ENABLE ].get_value();
            reg_value &= ~GIGE_ARM_TYPE_EXTERNAL;
            regs[ GIGE_REG_ARM_IN_ENABLE ].set_value(reg_value);
            fcp_notify_apps(GIGE_REG_ARM_IN_ENABLE, reg_value);
        }
    }

    return set_camera_state(new_state);
}

/*!
 * @brief Request a camera state from a thread other than the main CS thread.
 *
 * @param new_state requested new state
 * @param wait      true to wait until the state transition completes.
 * @param p_done    pointer to integer used for returning status, which must
 *          be non-null if wait is true. This allows the the function
 *          to be reentered, which could conceivably happen if the
 *          main thread got stalled too long. It is safest for this
 *          to point to a static so it is still valid after a
 *          failure due to timeout.
 *
 * @return true if changing the state was successful.
 */
bool CameraServer::async_set_camera_state(uint8_t new_state, bool wait, volatile int *p_done)
{
    CS_NETWORK::CS_NETWORK_MESSAGE msg;     // Structure for either an ipc or udp message.
    bool rc = true;
    uint32_t tid;

    PDEBUG(PDEBUG_TRACE, "%s(%d,%d)", __FUNCTION__, new_state, wait ? 1 : 0);

    // Sanity check.
    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL camserv_ptr!", __PRETTY_FUNCTION__);
        return false;
    }

    // We should not be making this call from the main thread.
    tid = get_tid();
    if (camserv_ptr->cs_tid == tid)
    {
        fi_log(LOG_ERR, "%s called from PID=%u CS main thread!", __FUNCTION__, tid);
        return camserv_ptr->instance_set_camera_state(new_state);
    }

    msg.msg_type = MSG_TYPE_INTERNAL;
    msg.internal.type = CS_INTERNAL_STATE;
    msg.internal.flags = new_state;
    if (wait)
    {
        msg.internal.p_arg = (void *)p_done;
        *p_done = -1;
    }
    else
        msg.internal.p_arg = nullptr;
    add_work(&msg);

    if (wait)
    {
        int retries;
        int max_retries = 400;  // 5 msec * 1000 = 5 seconds, timeout is disastrous.

        for (retries = 0; retries < max_retries; ++retries)
        {
            // Wait 5 milliseconds, long enough for a context switch.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(5));

            if (*p_done != -1)
                break;
        }
        if (retries == max_retries)
        {
            fi_log(LOG_ERR, "%s: timed out waiting for message to be processed", __FUNCTION__);
            rc = false;
        }
        else if (*p_done == 0)
        {
            fi_log(LOG_ERR, "%s: state change to %u failed, now %u", __FUNCTION__, new_state,
                   get_camera_state());
            rc = false;
        }
    }

    return rc;
}

/*!
 * @brief Update the state of the camera. This will also update
 *    the state LED.
 *
 * @param new_state State we are going into.
 *
 * @return true if changing the state was successful.
 */
bool CameraServer::set_camera_state(uint8_t new_state)
{
    uint32_t tid;

    TP(TRACE_SEL_CAMERA_STATE, "%s(%d)", __FUNCTION__, new_state);

    if (camserv_ptr == nullptr)
        return false;

    tid = get_tid();
    if (camserv_ptr->cs_tid != tid)
    {
        int done;

        fi_log(LOG_DEBUG, "%s called from PID=%u, calling async_set_camera_state()", __FUNCTION__, tid);
        return async_set_camera_state(new_state, true, &done);
    }

    return camserv_ptr->instance_set_camera_state(new_state);
}

/*!
 * @brief private entry point to update the state of the camera.
 *
 * This will also update the state LED.
 *
 * @param new_state State we are going into.
 *
 * @return true if changing the state was successful.
 */
bool CameraServer::instance_set_camera_state(uint8_t new_state)
{
    uint8_t actual_new_state;

    TP(TRACE_SEL_CAMERA_STATE, "%s(%d)", __FUNCTION__, new_state);

    if (camserv_ptr == nullptr)
        return false;

    // Is it the same state that we're already in?
    if (camserv_ptr->camera_state == new_state)
        return true;

    switch (new_state)
    {
    case GIGE_CAMERA_STATE_ARMED:
        actual_new_state = set_armed_state();
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        actual_new_state = set_triggered_state();
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        actual_new_state = set_review_state();
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        actual_new_state = set_autosave_state();
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        actual_new_state = set_snapshot_state();
        break;
    case GIGE_CAMERA_STATE_DEARMING:
        actual_new_state = set_dearming_state();
        break;
    case GIGE_CAMERA_STATE_IDLE:
        actual_new_state = set_idle_state();
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    default:
        actual_new_state = set_live_state();
        break;
    }

    TP(TRACE_SEL_CAMERA_STATE, "%s: old=%s new=%s actual_new=%s%s", __FUNCTION__,
       state_to_str(camera_state), state_to_str(new_state),
       state_to_str(actual_new_state),
       (camera_state != actual_new_state) ? ", notifying" : "");

    if (camera_state != actual_new_state)
    {
        // Save the new state.
        camera_state = actual_new_state;

        // Notify any internal apps that something of importance has changed.
        fcp_notify_apps(GIGE_REG_CAMERA_STATE, new_state);
    }

    if (camera_state == new_state)
        return true;
    else
        return false;
}

/*!
 * @brief Reinitialize the GigE registers to factory defaults.
 */
void CameraServer::reset_registers()
{
    // Initialize the GigE register cache.
    // Then make sure the registers are all mutually compatible and legal.
    reg_cache.init_cache();
    fcp_validate_registers();

    // Reserve space for the gain table.
    PartMgr::getInstance()->reserveGainTableSpace(true);

    // Reinitialize the layout for partition memory.
    PartMgr::getInstance()->layoutMemoryFromRegs();

    // Layout of memory could change the number of frames, so that could affect
    // our trigger position and save limits.
    fcp_validate_frame_limits(PartMgr::getInstance()->getFrameCount() - 1);

    // Set any hardware registers that we reinitialized.
    reg_cache.set_loaded_registers();
}

/*!
 * @brief Request a watchdog timer "pet".
 */
void CameraServer::watchdog()
{
    CS_NETWORK::CS_NETWORK_MESSAGE msg;     // Structure for either an ipc or udp message.

    // Sanity check.
    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL camserv_ptr!", __PRETTY_FUNCTION__);
        return;
    }

    msg.msg_type = MSG_TYPE_INTERNAL;
    msg.internal.type = CS_INTERNAL_WATCHDOG;
    msg.internal.flags = 0;
    msg.internal.p_arg = nullptr;
    add_work(&msg);
}

/*!
 * @brief This is the main entry point into the camera server.
 */
bool CameraServer::init()
{
    uint32_t fpga_rev, fpga_icm;

    fi_log(LOG_INFO, "%s: VERSION (%d.%d.%d)", __PRETTY_FUNCTION__, FV_MAJ, FV_MIN, FV_PATCH);

    // Create the license manager object which other classes can
    // access to find out what features our license allows us to
    // support.
    lic_ptr = std::unique_ptr<LicManager>(new LicManager());
    if (lic_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to connect with the LM.", __PRETTY_FUNCTION__);
        return false;
    }

    // Validate the license information.
    lic_ptr->licenseValidate();

    // The FAT filesystem code in the kernel needs to convert file times from
    // UTC to local time since that is what is stored. In a grown-up (non-busybox,
    // or newer busybox) system, we would do this via "hwclock --systz".
    set_kernel_timezone();

    // Do any class initializations that could fail. We don't do those in the constructor because
    // we aren't using exceptions so have no way to catch failures.
    if (!FiFPGA::getInstance()->interfaceInit())
    {
        fi_log(LOG_ERR, "%s: FPGA interfaceInit() failed!", __PRETTY_FUNCTION__);
        return false;
    }
    FiFPGA::getInstance()->regRead(FPGA_REVISION, &fpga_rev);
    FiFPGA::getInstance()->regRead(ICM_MODULE_REVISION, &fpga_icm);
    fi_log(LOG_INFO, "%s: FPGA VERSION (0x%08x) ICM (0x%08x)", __PRETTY_FUNCTION__, fpga_rev, fpga_icm);

    // Get a handle to the correct sensor for what kind of camera we have.
    sensor_ptr = Sensor::create_sensor();
    if (sensor_ptr.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate a sensor!", __PRETTY_FUNCTION__);
        return false;
    }

    // Do any class initializations.
    if (!sensor_ptr->class_init(FiFPGA::getInstance()))
    {
        fi_log(LOG_ERR, "%s: Sensor class_init() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Have the sensor reset the event and capture FIFOs.
    sensor_ptr->reset_capture_fifo();
    sensor_ptr->reset_event_fifo();

    // Create the autosave manager
    as_mgr = std::unique_ptr<AutoSave>(new AutoSave());
    if (as_mgr.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Couldn't create an AutoSave object!", __FUNCTION__);
        return false;
    }

    // Pass class pointers to make sure everyone understands the initialization order.
    // If you get the pointer, then you know that class is safe to use.
    if (!PartMgr::getInstance()->classInit(FiFPGA::getInstance(), sensor_ptr.get(), lic_ptr.get()))
    {
        fi_log(LOG_ERR, "%s: PartMgr class_init() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Did the user want to turn on the sensor test mode?
    if (sensor_test)
        sensor_ptr->enable_sensor_test();

    // Once we have the sensor available, we can load the correct white balance coefficients.
    load_wb_file(sensor_ptr.get());

    // Initialize FPGA registers that don't have to be sequenced with the sensor.
    // This ensures that the FPGA has been cleaned up from any prior run, as if we
    // had just booted, then does any other one-time initialization that does not
    // depend on our current settings.
    // It also sets up some defaults that may be overwritten by our settings.
    FiFPGA::getInstance()->startupInit();

    // Once the reserved area is set up, the PartMgr can tell the FPGA where all the reserved
    // buffers are.
    // FIXME we left that in PartMgr::class_init() for now.
    //  FiFPGA::getInstance()->set_reserved_buffers();

    // Initialize the GigE register cache.
    // Once all the registers have default values, restore the ones from Flash if they exist.
    reg_cache.init_cache();
    reg_cache.load();

    // Most of the code depends on asking the FiFPGA class whether LTR mode is enabled
    // instead of using the register so we need to set that flag early on.
    fcp_ltr_mode_init();

    // Make sure the registers are all mutually compatible and legal.
    fcp_validate_registers();

    // Create the layout for partition memory.
    // Reserve space for the gain table.
    PartMgr::getInstance()->reserveGainTableSpace(true);

    // This depends on our ROI, session length, and whether we are in LTR mode, which were already
    // initialized from the defaults or the settings.
    PartMgr::getInstance()->layoutMemoryFromRegs();

    // Initialize the autosave object.
    as_mgr->init(PartMgr::getInstance()->getNumPartitions());

    // Layout of memory could change the number of frames, so that could affect
    // our trigger position and save limits.
    fcp_validate_frame_limits(PartMgr::getInstance()->getFrameCount() - 1);

    // Some sensor parameters (e.g. Lupa DACVREF, GAIN, BLACKCAL2) are saved in registers.
    // If we have sensor FPN on, we need to use those saved values.
    // If pixel FPN is on, we will load DACVREF and GAIN from the correction table later.
    // This is also the last chance to validate and update sensor-related GigE registers
    // before starting the sensor.
    sensor_ptr->set_loaded_registers();

    // Set up sensor and FPGA so they are consistent with the IDLE state.
    // When we later go into LIVE_VIDEO, we will be using the standard procedure for
    // changing from IDLE to LIVE_VIDEO, so we need to have everything ready for that.
    // If the user gave us a "sensor test" command line switch, we will use it now.
    // If sensor FPN is enabled, it will be turned on now.
    setup_fpga_and_sensor();

    // Now that all the registers are loaded and mutually consistent, and the rest of the
    // hardware is initialized to be consistent with the IDLE state, make sure any other
    // static hardware registers match the GigE registers.
    // JUL2014 This was originally done as we loaded the registers from the save file, which
    // JUL2014 was why we couldn't load the save file until most other init was done, which
    // JUL2014 forced some startup actions to be done in a strange and confusing order.
    reg_cache.set_loaded_registers();

    // Set up FPGA FPN if that's what was enabled. If we later discover that there is
    // an LTR recording on the drive, we will need to use that instead.
    load_ct_tables();

    // Startup the secondary thread to keep a watch over the two sockets (IPC, UDP). We need
    // this because there is also an FPGA thread giving us work.
    worker->start();

    // Now that the "base" system is good to go, let's startup our fan controller.
    fan_init();

    // Initialize the len control object.
    mft_init();

    // Startup the battery monitor.
    batteryStart();

    // If we're enabled, start up the auto-trigger thread.
    if (auto_trigger())
    {
        auto_trig = std::unique_ptr<csAutoTrigger>(new csAutoTrigger());
        if (auto_trig.get() != nullptr)
            auto_trig->start();
    }

    // Assume things will go badly.
    shmem_handle = nullptr;

    // Initialize the shared memory control structures.
    shmem_handle = fsmServerInit(0, 0, SHMEM_DEFAULT_DESCRIPTOR_COUNT);
    if (shmem_handle == nullptr)
    {
        fi_log(LOG_ERR, "%s: fsmServerInit() failed!", FI_FUNCTION);
    }

    return true;
}

bool CameraServer::in_factory_reset()
{
    if (camserv_ptr == nullptr)
        return false;

    return camserv_ptr->do_factory_reset;
}

void CameraServer::factory_reset(bool do_reset)
{
    if (camserv_ptr == nullptr)
        return;

    camserv_ptr->do_factory_reset = do_reset;
}

void CameraServer::cleanup()
{
    TP(TRACE_SEL_CAMERA_STATE, "%s", __PRETTY_FUNCTION__);

    // All done.
    if (shmem_handle != nullptr)
    {
        fsmServerCleanup(shmem_handle);
        shmem_handle = nullptr;
    }

    // Tell the FCP module we are shutting down
    fcp_exit();

    // Tell the streamer that we are shutting down.
    fsp_exit();

    // Shutdown the lens controller.
    mft_close();

    // Shutdown the fan controller.
    fan_close();

    // Shutdown the battery monitor.
    batteryClose();

    // Have the autosave manager stop.
    if (as_mgr.get() != nullptr)
    {
        as_mgr->stop();
        as_mgr.reset();
    }

    // Shutdown the auto-trigger thread
    if (auto_trig.get() != nullptr)
        auto_trig->stop();

    // Close down the polling thread.
    dvr_poller->stop();

    // Stop petting the watchdog. We don't disable the watchdog logic
    // in the FPGA because we want the FPGA to indicate a disconnect.
    wd_petter->stop();

    // Close down the secondary thread handing messages.
    worker->stop();

    // Close down the storage object.
    storage::stop();

    // Close down the playback object.
    pbc_close();

    // Close down the JPEG encoder.
    jpeg_close();

    PartMgr::getInstance()->shutdown();

    // Ask the sensor object to shutdown.
    sensor_ptr->shutdown();

    // Closing down the FPGA.
    FiFPGA::getInstance()->shutdown();

    // Get rid of the class pointer.
    camserv_ptr = nullptr;
    TP(TRACE_SEL_CAMERA_STATE, "%s: Done.", __PRETTY_FUNCTION__);
}

bool CameraServer::check_ltr_recording()
{
    bool        rc = false;
    PartMgr     *pm_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Are we in LTR mode?
    if ((regs[GIGE_REG_LTR_MODE] & GIGE_LTR_MODE_ENABLE) == GIGE_LTR_MODE_ENABLE)
    {
        // Initialize for LTR. If there is a recording, we will show COMPLETE instead of EMPTY.
        if (pm_ptr->ltrStartup())
        {
            CameraServer    *cs_ptr = CameraServer::get_cs();

            if (pm_ptr->getPartitionState() == GIGE_PART_STATE_COMPLETE)
            {
                uint32_t fpn_option;

                fi_log(LOG_DEBUG, "%s: found COMPLETE partition", __FUNCTION__);

                // Stop the ICM in case the sensor startup needs to adjust the ICM frequency
                // in the clock generator.
                cs_ptr->set_idle_behavior();

                // We already did this earlier but we might have changed some settings when
                // we loaded the LTR recording so we have to do it again. If sensor FPN was
                // enabled, this will turn it on now.
                cs_ptr->setup_fpga_and_sensor();

                // If FPGA pixel FPN is to be used, just enable it now. The table was
                // already loaded into memory by the LTR startup analysis.
                fpn_option = regs[ GIGE_REG_FPN ].get_value();
                if (fpn_option & GIGE_FPN_FPGA_ENABLE_MASK)
                    FiImageCorrection::fpnEnable();

                // Let folks know about the size of the active sector.
                fcp_notify_apps(GIGE_REG_LTR_SECTOR_COUNT, pm_ptr->ltrGetActiveImageSectors());

                // All is well.
                rc = true;
            }
        }
    }

    return rc;
}

/*!
 * @brief Run a quick system integrity check.
 */
void CameraServer::integrity_check()
{
    // No sanity checking if we're in the wrong state.
    if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
        return;

    // Make sure we have a sensor pointer.
    if (sensor_ptr.get() == nullptr)
        return;

    // Are we seeing CRC errors on the sensor data channels?
    if (sensor_ptr->icm_has_crc_errors())
    {
        fi_log(LOG_WARNING, "%s: Sensor CRC errors detected, attempting re-alignment.", __FUNCTION__);

        // Request the FPGA to perfrom a runtime realignment.
        if (!sensor_ptr->icm_realign())
        {
            fi_log(LOG_WARNING, "%s: Realignment failed - forcing sensor restart.", __FUNCTION__);

            // Try restarting the sensor interface.
            set_camera_state(GIGE_CAMERA_STATE_IDLE);
            setup_fpga_and_sensor();
            set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
    }
}

void CameraServer::do_update(bool net_update)
{
    int sys_rc;
    std::string update_cmd;
    std::string pacman_opts;
    std::string cmd_path;
    std::ofstream cmd_file;

    // Open our command file.
    cmd_path = fmt::format("{0}/{1}", get_home_path(), UPDATE_CMD_FILE);
    cmd_file.open(cmd_path, std::ios::out | std::ios::trunc);
    if (!cmd_file.is_open())
    {
        fi_log(LOG_ERR, "%s: Unable to open '%s' [%d: %s]", __PRETTY_FUNCTION__, UPDATE_CMD_FILE, errno, strerror(errno));

        // Make sure we close the object.
        cmd_file.close();
        return;
    }

    // Get any optional parameters the update package may have specified.
    update_options(pacman_opts);

    // Check for a network update vs. a local package update.
    if (net_update)
    {
        // Figure out who we are running as.
        char *uname = getenv("USER");
        if (!uname)
        {
            fi_log(LOG_DEBUG, "%s: Unknown USER", __PRETTY_FUNCTION__);

            // Make sure we close the object.
            cmd_file.close();
            return;
        }

        // Do a network update. Note that yay needs to run as the user.
        //
        // Are there any options provided?
        if (pacman_opts.empty())
        {
            // No options provided, so do the default update.
            update_cmd = fmt::sprintf("sudo -u %s yay -Syu --needed --noconfirm", uname);
        }
        else
        {
            // Use the update options provided.
            update_cmd = fmt::sprintf("sudo -u %s yay ", uname) + pacman_opts;
        }
    }
    else
    {
        // Do a local update of the "fastec" group only. Since we're running pacman,
        // we don't need to switch from root.
        if (pacman_opts.empty())
            update_cmd = std::string("pacman -S fastec --needed --noconfirm");
        else
            update_cmd = std::string("pacman ") + pacman_opts;
    }

    // Build the command file.
    cmd_file << "#!/bin/bash" << std::endl;
    cmd_file << "set -x" << std::endl;                                              // I want to see what's going on.
    cmd_file << "logger Update Starting" << std::endl;
    cmd_file << "pacman -S --needed --noconfirm archlinux-keyring" << std::endl;    // Update the keyring.
    cmd_file << "paccache -r" << std::endl;                                         // Prune the package cache so it doesn't grow indefinitely.
    cmd_file << update_cmd << std::endl;                                            // Issue our update commands.
    cmd_file << "logger Update Complete, rebooting..." << std::endl;
    cmd_file << "sync; sync; sync" << std::endl;                                    // Make sure the drive(s) are settled.
    cmd_file << "reboot" << std::endl;                                              // Rebooting...

    // Close the command file.
    cmd_file.close();

    // Build up the command line.
    update_cmd = fmt::format("/usr/bin/chmod a+x {0}; systemd-run {0}", cmd_path);
    fi_log(LOG_DEBUG, "%s: update_cmd[%s]", __PRETTY_FUNCTION__, update_cmd.c_str());

    // Start processing the update request.
    // ----------------------------------------------------------
    // ALM_TODO: We might need to move this out to the csrv script file, sort of like reboot/reset logic.
    // ----------------------------------------------------------
    sys_rc = system(update_cmd.c_str());
    if ((sys_rc < 0) || (WEXITSTATUS(sys_rc) != 0))
    {
        fi_log(LOG_ERR, "%s: Error running pacman [%d]!", __PRETTY_FUNCTION__, WEXITSTATUS(sys_rc));
    }
    else
    {
        // Shutdown the camera server.
        fi_log(LOG_INFO, "%s: shutdown the camera server.", __PRETTY_FUNCTION__);
        CameraServer::exit();
    }
}

/*!
 * @brief Process an internal message in the main CS thread.
 *
 * The only MSG_TYPE_INTERNAL we handle is when the CS is told to
 * change the state to GIGE_CAMERA_STATE_DEARMING. This means
 * that the current capture should be terminated but frames captured so
 * far should be retained. We don't want to finalize the recording
 * while we are still processing a FCP message because it can take a
 * while for frames to flush out to the drive in LTR mode. So this is a
 * way to defer the work so it can be handled in the same context as a
 * normal completion initiated by an ICM event.
 *
 * @param internal_arg  1 if we entered DEARMING from TRIGGERED,
 *          0 if we entered DEARMING from ARMED
 */
void CameraServer::process_internal(CS_INTERNAL *p_msg)
{
#ifdef FCP_DEBUG
    fi_log(LOG_DEBUG, "%s: type=%d", __FUNCTION__, p_msg->type);
#endif /* FCP_DEBUG */

    // We are stopping capture at user request but retaining the recording.
    if (p_msg->type == CS_INTERNAL_DEARM)
    {
        bool any_frames;

        // Dearm and simulate trigger and completion.
        any_frames = FiFPGA::getInstance()->icmStopCapture(p_msg->flags ? true : false);

        // If we are in autosave mode, make sure the partition we just cancelled
        // gets queued up to be saved, then go to LIVE state. The autosave will
        // continue in background. If we are not in autosave mode, just REVIEW now.
        if (as_mgr->isEnabled())
        {
            if (any_frames)
                as_mgr->process();

            // Disable autosave.
            fcp_autosave_update(nullptr, GIGE_AUTOSAVE_DISABLE);

            set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
        else if (any_frames)
        {
            // We have to stop the stream if it's running.
            fcp_stop_stream_thread();
            FiFPGA::getInstance()->icmSensorReadoutEnable(false);
            set_camera_state(GIGE_CAMERA_STATE_REVIEW);
        }
        else
        {
            // Partition was already marked empty.
            set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
    }
    // We are requested to change the camera state.
    else if (p_msg->type == CS_INTERNAL_STATE)
    {
        bool rc = set_camera_state(p_msg->flags);
        if (p_msg->p_arg != nullptr)
        {
            volatile int *p_rtn;

            p_rtn = (volatile int *)p_msg->p_arg;
            if (rc)
                *p_rtn = 1;
            else
                *p_rtn = 0;
        }
    }
    // We are requested to pet the watchdog.
    else if (p_msg->type == CS_INTERNAL_WATCHDOG)
    {
        // Do we have our watchdog object allocated yet?
        if (wd_petter.get() == nullptr)
        {
            // Allocate our watchdog petting thread.
            wd_petter = std::unique_ptr<cs_watchdog>(new cs_watchdog());
            wd_petter->start();
        }
        else
        {
            // Have the FPGA object update the watchdog.
            FiFPGA::getInstance()->watchdog();
        }
    }
    // We are requested to perform a system update.
    else if (p_msg->type == CS_INTERNAL_UPDATE)
    {
        // The flags tell us if we're doing a network (true or non-zero) or local (false or zero) update.
        do_update((p_msg->flags == 1));
    }
    else
    {
        fi_log(LOG_ERR, "%s: unexpected internal message type=%u", __FUNCTION__, p_msg->type);
    }
}

/*
 * @brief CameraServer main thread
 */
void CameraServer::run()
{
    std::cv_status  rc;     // Result codes returned by function calls.
    CS_NETWORK::CS_NETWORK_MESSAGE msg;     // Structure for either an ipc or udp message.

    cs_tid = get_tid();
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, cs_tid);

    // We run forever until we are killed by a signal
    while (!exitRequested)
    {
        // Clear our message structure.
        memset(&msg, 0, sizeof(msg));

        // Check to see if there is any work for us from local client
        // remote client or the fpga. When we return from the "wait_until"
        // we either have a bad status or we already have the lock on the
        // mutex
        while (msg_queue.empty())
        {
            std::unique_lock<std::mutex> locker(work_mutex);

            // See if we need to turn on sensor test mode.
            if (sensor_test_toggle)
            {
#if 0
                fi_log(LOG_DEBUG, "%s: enabling sensor test mode", __FUNCTION__);
                set_camera_state(GIGE_CAMERA_STATE_IDLE);
                sensor_ptr->enable_sensor_test();
                sensor_test_toggle = false;
                setup_fpga_and_sensor();
                set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
#else
                fi_log(LOG_DEBUG, "%s: OVERRIDE --- Writing bitslips!", __FUNCTION__);
                sensor_test_toggle = false;
                FiFPGA::getInstance()->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_BL, 0xFFFFF);
                FiFPGA::getInstance()->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_BR, 0xFFFFF);
                FiFPGA::getInstance()->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_TL, 0xFFFFF);
                FiFPGA::getInstance()->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_TR, 0xFFFFF);
#endif
            }

            // See if we need to dump registers.
            if (do_register_dump)
            {
                FiFPGA::getInstance()->dump_regs();
                do_register_dump = false;
            }

            // Wait.
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            using sec = std::chrono::seconds;
            rc = work_cond.wait_until(locker, now + sec(MSGQ_WAIT_SECONDS));

            // Are we supposed to bail?
            if (exitRequested)
                break;

            // Why did we exit?
            if (rc != std::cv_status::no_timeout)
            {
                // Check for a timeout.
                if (rc != std::cv_status::timeout)
                {
                    // Got an error waiting on the cond event.
                    fi_log(LOG_ERR, "%s: Couldn't wait on work threads, %d: %s", __FUNCTION__,
                           errno, strerror(errno));

                    using sec = std::chrono::seconds;
                    std::this_thread::sleep_for(sec(ERROR_WAIT_LONG));

                    continue;
                }
            }
        }

        // Keep looping while there are messages in our queue.
        // Not having any messages is pefectly legal - could be a
        // result of a "kill" signal, for example.
        //
        // NOTE: At this point, we do NOT have a lock on the msg_queue!
        while (!exitRequested && !msg_queue.empty())
        {
            // See if we need to dump registers.
            if (do_register_dump)
            {
                FiFPGA::getInstance()->dump_regs();
                do_register_dump = false;
            }

            // See if there is any work to clean up before we take new work. This will guarantee
            // that we have no locks outstanding when a new request comes in.
            fcp_cleanup();

            // Get the first element in the list. Grab a lock to be safe.
            {
                std::unique_lock<std::mutex> locker(work_mutex);
                msg = msg_queue.front();
            }

#ifdef FCP_DEBUG
            {
                uint16_t cmd = ntohs(msg.message.header.req.command);
                uint16_t len = ntohs(msg.message.header.req.length);
                uint16_t id = ntohs(msg.message.header.req.req_id);

                fi_log(LOG_INFO, "Processing msg, type=%d cmd=0x%04x, len=0x%04x, req_id=0x%04x", msg.msg_type, cmd, len, id);
            }
#endif /* FCP_DEBUG */

            // Process the message based on it's type.
            switch (msg.msg_type)
            {
            case MSG_TYPE_FPGA:
                process_fpga_events(msg.fpga_val);
                break;
            case MSG_TYPE_GRPC:
            case MSG_TYPE_IPC:
            case MSG_TYPE_UDP:
                fcp_parse(&msg);
                break;
            case MSG_TYPE_INTERNAL:
                process_internal(&msg.internal);
                break;
            default:
                fi_log(LOG_ERR, "%s: Unexpected message type (0x%x)", __FUNCTION__, msg.msg_type);
                break;
            }

            // Delete the entry we just finished processing. Grab a lock to be safe.
            {
                std::unique_lock<std::mutex> locker(work_mutex);
                msg_queue.pop_front();
            }
        }

        // If we're not exiting...
        if (!exitRequested)
        {
            // Check system integrity.
            integrity_check();
        }
    }

    // Notify folks we are shutting down.
    fcp_notify_apps(GIGE_REG_CAMERA_STATE, GIGE_CAMERA_STATE_SHUTDOWN);

    // On the way out. We need to clean up after ourselves.
    cleanup();

    // Are we REALLY supposed to reboot?
    if (rebootRequested)
        fcp_system_reboot();
    if (restartRequested)
        fcp_system_restart();
    if (shutdownRequested)
        fcp_system_exit();
}

CameraServer::CameraServer() :
    cs_tid(0),
    init_complete(false)
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_CAMERA_STATE;

    // Create our storage object.
    storage::factory();

    // Allocate our worker thread.
    worker = std::unique_ptr<csWorker>(new csWorker());

    // Allocate our polling thread.
    dvr_poller = std::unique_ptr<cs_poll>(new cs_poll());

    // Initialize our flags.
    exitRequested = false;
    shutdownRequested = false;
    do_factory_reset = false;

    // Initialize our state.
    camera_state = GIGE_CAMERA_STATE_IDLE;

    // Legal enough for now, will be set for real when we init the GigE registers.
    // We just need a static place to store these values so we will have them for the
    // metadata.
    as_shot_red = as_shot_grn = as_shot_blu = 0x10000000;

    // Save away a pointer to our class object.
    camserv_ptr = this;

    // Create the FPGA singleton now.
    FiFPGA::getInstance();

    // Create the JPEG encoder object.
    jpeg_init();

    // Create the playback control object.
    pbc_init();
}

/*!
 * @brief CameraServer object destructor.
 *
 * This needed to exist, because we declare std::unique_ptrs on
 * incomplete classes (e.g. Sensor). If you don't instantiate a
 * destructor in the code (where Sensor is fully declared) you get
 * a compile-time error.
 */
CameraServer::~CameraServer()
{
    camserv_ptr = nullptr;
}

void CameraServer::init_done()
{
    static volatile int async_done;

    // Sanity check.
    if (camserv_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: NULL camserv_ptr?!", __PRETTY_FUNCTION__);
        return;
    }

    // Initialize the FCP interface.
    fcp_init();

    // Now that the settings are loaded, we know whether we are in LTR mode for real.
    // If this returns true, we could enter Review mode (but we won't for now).
    camserv_ptr->check_ltr_recording();

    // Send out notifications about registers that changed on startup.
    fcp_startup_notify();

    // Ask the storage object to mount any/all possible media.
    storage::startup_mounts();

    // On init complete, we can go to the LIVE state.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO, true, &async_done);

    // If the ARM input should be enabled, we are able to handle the event now so enable it.
    FiFPGA::getInstance()->armInInitialize();

    // Check to see if we are in either auto-trigger or DVR playback mode. If we are, then we need to
    // switch to the ARMed state.
    if (camserv_ptr->auto_trigger() || camserv_ptr->dvr_playback())
    {
        uint32_t pref;
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        // See if auto playback is enabled.
        pref = regs[ GIGE_REG_PREFERENCES ].get_value();
        if ((pref & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY)
        {
            // Arm the camera.
            CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_ARMED, false, nullptr);
        }
    }

    // All done with initialization.
    camserv_ptr->init_complete = true;
}

bool CameraServer::init_is_done()
{
    bool rc = false;

    if (camserv_ptr != nullptr)
        rc = camserv_ptr->init_complete;

    return rc;
}
