/*!
 * @file clock_gen.cxx
 *
 * @brief This file contains the implementation of the clock_gen class.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>

#include <chrono>

#include <fastec/fas_log.h>

#include "sensor.h"
#include "camera_server.h"
#include "clock_gen.h"
#include "si5338.h"

#define code
namespace CLOCK_GEN_LUX13HS
{
#include "si5338_register_map_nuc_lux13hs.h"

const int NUM_REGS = NUM_REGS_MAX;
#undef NUM_REGS_MAX
}

namespace CLOCK_GEN_LUX19HS
{
#include "si5338_register_map_nuc_lux19hs.h"

const int NUM_REGS = NUM_REGS_MAX;
#undef NUM_REGS_MAX
}

namespace CLOCK_GEN_LUX2100
{
#include "si5338_register_map_nuc_lux2100.h"

const int NUM_REGS = NUM_REGS_MAX;
#undef NUM_REGS_MAX
}

namespace CLOCK_GEN_LINCE5M
{
#include "si5338_register_map_nuc_lince5m.h"

const int NUM_REGS = NUM_REGS_MAX;
#undef NUM_REGS_MAX
}

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Our clock gen device's address.
const int SI5338_I2C_ADDR = 0x70;

//! Number of times to wait for an event.
const int RETRY_COUNT = 100;

//! Generic register data structure for the Si5338.
typedef struct REG_DATA
{
    unsigned char Reg_Addr;
    unsigned char Reg_Val;
    unsigned char Reg_Mask;

} REG_DATA;

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// clock_gen class implementation
//--------------------------------------------------------------------

/*!
 * @brief The clock_gen class constructor. In here we initialize the interface
 *        to the ClockGen driver in preperation for the control methods.
 */
clock_gen::clock_gen(void)
{
}

/*!
 * @brief The clock_gen class destructor. Here we shutdown and cleanup the
 *        interface to the ClockGen driver.
 */
clock_gen::~clock_gen()
{
}

/*!
 * @brief Initialize the clock gen.
 * @return True on success, false otherwise.
 */
bool clock_gen::clock_init()
{
    bool    status = false;
    int     count;
    int     num_regs;
    uint8_t curr_chip_val;
    uint8_t clear_curr_val;
    uint8_t clear_new_val;
    uint8_t combined;
    uint32_t sensor_type;
    REG_DATA const *reg_store;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // Following the flow chart in the datasheet, we now want to
    // disable outputs.
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_OUTPUT_ENABLE, OEB_ALL))
        goto err_exit;

    // Pause Loss Of Lock detection.
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_LOL, DIS_LOL | LOL_DEFAULT))
        goto err_exit;

    // What is the sensor type we have?
    sensor_type = Sensor::get_sensor_type();
    switch (sensor_type)
    {
    case FI_SENSOR_LINCE5M_MONO:
    case FI_SENSOR_LINCE5M_COLOR:
        num_regs = CLOCK_GEN_LINCE5M::NUM_REGS;
        reg_store = (REG_DATA *)&CLOCK_GEN_LINCE5M::Reg_Store[0];
        break;
    case FI_SENSOR_LUX13HS_MONO:
    case FI_SENSOR_LUX13HS_COLOR:
        num_regs = CLOCK_GEN_LUX13HS::NUM_REGS;
        reg_store = (REG_DATA *)&CLOCK_GEN_LUX13HS::Reg_Store[0];
        break;
    case FI_SENSOR_LUX19HS_MONO:
    case FI_SENSOR_LUX19HS_COLOR:
        num_regs = CLOCK_GEN_LUX19HS::NUM_REGS;
        reg_store = (REG_DATA *)&CLOCK_GEN_LUX19HS::Reg_Store[0];
        break;
    case FI_SENSOR_LUX2100_MONO:
    case FI_SENSOR_LUX2100_COLOR:
        num_regs = CLOCK_GEN_LUX2100::NUM_REGS;
        reg_store = (REG_DATA *)&CLOCK_GEN_LUX2100::Reg_Store[0];
        break;
    default:
        fi_log(LOG_ERR, "%s: Unknown sensor type! [0x%08x]", __PRETTY_FUNCTION__, sensor_type);
        goto err_exit;
    }

    // Send the configuration commands to the device.
    for (count = 0; count < num_regs; count++)
    {
        REG_DATA const *reg_ptr;
        reg_ptr = &reg_store[count];

        // Check the mask to see if we can do a straight write,
        // or if we need to do a read-modify-write.
        if (reg_ptr->Reg_Mask == 0xFF)
        {
            // Just write the raw value out, since the mask is all ones.
            if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, reg_ptr->Reg_Addr, reg_ptr->Reg_Val))
                goto err_exit;
        }
        // Skip registers where the mask is zero.
        else if (reg_ptr->Reg_Mask != 0)
        {
            // We need to do a read-modify-write in order to
            // utilize the required mask.
            if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, reg_ptr->Reg_Addr, &curr_chip_val))
                goto err_exit;

            clear_curr_val = curr_chip_val & ~reg_ptr->Reg_Mask;
            clear_new_val = reg_ptr->Reg_Val & reg_ptr->Reg_Mask;
            combined = clear_new_val | clear_curr_val;

            if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, reg_ptr->Reg_Addr, combined))
                goto err_exit;
        }
    }

    // Check for valid input clock status. Read the LOS alarm for
    // crystal inputs on IN3 (we are using an 25 Mhz oscillator on IN3).
    // Note that LOS_CLKIN is for IN1, 2 OR 3.
    for (count = 0; count < RETRY_COUNT; count++)
    {
        // Read the value and if it's clear we can continue.
        if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, I2C_LOS, &curr_chip_val))
            goto err_exit;

        if ((curr_chip_val & LOS_CLKIN) == 0)
            break;
    }
    if (count == RETRY_COUNT)
        fi_log(LOG_ERR, "%s: LOS Alarm never cleared!\n", __FUNCTION__);

    // Configure the PLL for locking. Clear FCAL_OVRD_EN.
    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, I2C_PLL_CONFIG, &curr_chip_val))
        goto err_exit;

    curr_chip_val = curr_chip_val & ~FCAL_OVRD_EN;
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_PLL_CONFIG, curr_chip_val))
        goto err_exit;

    // Initiate lockin of PLL.
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_SOFT_RESET, SOFT_RESET))
        goto err_exit;

    // Delay 100 milliseconds.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(100));

    // Restart LOL. DIS_LOL = 0.
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_LOL, LOL_DEFAULT))
        goto err_exit;

    // Delay 25 milliseconds.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(25));

    // Confirm PLL Lock. Pll is locked when PLL_LOL, SYS_CAL and
    // all other alarms are cleared.
    for (count = 0; count < RETRY_COUNT; count++)
    {
        // Read the value and if it's clear we can continue.
        if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, I2C_LOS, &curr_chip_val))
            goto err_exit;

        if ((curr_chip_val & LOCK_MASK) == 0)
            break;

        // Delay 25 milliseconds.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(25));
    }
    if (count == RETRY_COUNT)
        fi_log(LOG_ERR, "%s: No device lock!\n", __FUNCTION__);

    // Copy the FCAL values to the active registers.
    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, 235, &curr_chip_val))
        goto err_exit;
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, 45, curr_chip_val))
        goto err_exit;

    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, 236, &curr_chip_val))
        goto err_exit;
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, 46, curr_chip_val))
        goto err_exit;

    // Clear bits 0:1 from 47 and combine with
    // bits 0:1 from 237.
    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, 47, &curr_chip_val))
        goto err_exit;
    clear_curr_val = (curr_chip_val & 0xFC);

    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, 237, &curr_chip_val))
        goto err_exit;
    clear_curr_val |= (curr_chip_val & 3);

    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, 47, clear_curr_val))
        goto err_exit;

    // Set PLL to use FCAL values.
    if (!fpga_ptr->i2cRead(SI5338_I2C_ADDR, I2C_PLL_CONFIG, &curr_chip_val))
        goto err_exit;

    curr_chip_val |= FCAL_OVRD_EN;

    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_PLL_CONFIG, clear_curr_val))
        goto err_exit;

    // Enable outputs.
    if (!fpga_ptr->i2cWrite(SI5338_I2C_ADDR, I2C_OUTPUT_ENABLE, 0))
        goto err_exit;

    // All is well.
    status = true;

err_exit:
    if (!status)
        fi_log(LOG_ERR, "%s: I/O error during init", __PRETTY_FUNCTION__);

    return status;
}

void clock_gen::set_lince_icm_frequency(uint32_t sensor_opts)
{
    if ((sensor_opts & GIGE_SENSOPT_INTERNAL_0) != 0)
    {
        fi_log(LOG_CRIT, "%s: We do NOT support FM6! ==============", __PRETTY_FUNCTION__);
    }
}

