/*!
 * @file cs_network.cxx
 * @brief This file contains all the miscelenous routines to deal with
 *    network issues. Such as discoverying the network interfaces
 *    on the camera.
 *
 * Copyright 2010 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stddef.h>
#include <unistd.h>

#include <memory>
#include <string>
#include <fmt/printf.h>

#include <fastec/fas_log.h>

#include "camera_server.h"
#include "cs_network.h"
#include "cs_trace.h"

namespace CS_NETWORK
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#define MAX_STATUS_LEN  10

//! The command to get the name of the active connection.
const char *CMD_GET_ACTIVE = "nmcli -t -f NAME,TYPE con show --active";

//! The command to get the default static gateway being used.
const char *CMD_GET_GATEWAY = "nmcli -t -f IP4.GATEWAY con show \"%s\" | cut -d':' -f2";

//! The command to get the connection method (DHCP, etc.) being used on the specified connection.
const char *CMD_GET_METHOD = "nmcli -t -f IPV4.METHOD con show \"%s\" | cut -d':' -f2";

//! The command to get the default static subnet mask being used.
const char *CMD_GET_SUBNET_MASK = "nmcli -t -f IP4.ADDRESS con show \"%s\" | cut -d':' -f2 | cut -d'/' -f2";

//! The command to get the default static IP address being used.
const char *CMD_GET_IP_ADDR   = "nmcli -t -f IP4.ADDRESS con show \"%s\" | cut -d':' -f2 | cut -d'/' -f1";

//! The command to get the WIFI passphrase. Note, the '30' is the strlen(802-11-wireless-security.psk).
const char *CMD_GET_PASSPHRASE  = "nmcli -t -f 802-11-wireless-security.psk con show --show-secrets \"%s\" | cut -c 30-";

//! The command to get the SSID.
const char *CMD_GET_SSID    = "nmcli -t -f 802-11-wireless.ssid con show \"%s\" | cut -d':' -f2";

//! The command to get the WIFI mode (infrastructure, adhoc, open).
const char *CMD_GET_MODE    = "nmcli -t -f 802-11-wireless.mode con show \"%s\" | cut -d':' -f2";

//! The command to get the WIFI key-management field (expecting 'wpa-psk').
const char *CMD_GET_KEY_MGMT    = "nmcli -t -f 802-11-wireless-security.key-mgmt con show \"%s\" | cut -d':' -f2";

//! The Network Manager string for an Ethernet type connection. The string is actually something
//! like "802-3-ethernet", but just looking for the following in the return value should be fine.
const char *NM_ETHERNET_TYPE    = "ethernet";

//! The Network Manager string for a WIFI type connection. The string is actually something
//! like "802-11-wireless", but just looking for the following in the return value should be fine.
const char *NM_WIFI_TYPE    = "wireless";

//! The Network Manager string for a Hotspot type connection.
const char *NM_HOTSPOT_TYPE    = "Hotspot";

//! The Network Manager string for a STATIC (vs. DHCP/LLA) connection method.
const char *NM_STATIC_METHOD    = "manual";

//! The name for the ethernet DHCP connection, as reported by the network manager (nmcli).
const char *ENET_DHCP_CLIENT    = "enet-dhcp";

//! The name for the ethernet static connection, as reported by the network manager (nmcli).
const char *ENET_STATIC_CLIENT  = "enet-static";

//! The name for the WIFI DHCP connection, as reported by the network manager (nmcli).
const char *WIFI_DHCP_CLIENT    = "wifi-dhcp";

//! The name for the WIFI static connection, as reported by the network manager (nmcli).
const char *WIFI_STATIC_CLIENT  = "wifi-static";

//! A default hostname value.
const char *HOSTNAME_DEF    = "Fastec-NC1";

//! Location of route table which will be parsed.
const char *ROUTE_TABLE     = "/proc/net/route";

//! Location of the ethernet connection status
const char *NET_CONNECT_STATUS  = "/sys/class/net/%s/carrier";

//! Location of the wifi connection status
const char *WIFI_CONFIG_FILE    = "/etc/wpa_supplicant.conf";

//! Number of bytes we'll need for the NFS check logic.
const int NFS_CHECK_BUF_SIZE    = 1024;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
static NETWORK_INFO eth_net_info;       //!< Saved info about our ethernet network.
static NETWORK_INFO wifi_net_info;      //!< Saved info about our wifi network.

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Validate that a user specified name adheres to accepted host
 *    naming conventions.
 *
 * @param name_ptr  Pointer to a null terminated character string.
 *
 * @return true if name is OK, false otherwise.
 */
static bool is_name_valid(const char *name_ptr)
{
    // First rule, first character must be alpha-numeric.
    if (!isalnum(*name_ptr))
        return 0;

    // Move up a character.
    ++name_ptr;

    // Now, loop through the rest of the string.
    while (*name_ptr != '\0')
    {
        // Make sure only valid characters exist.
        if (!isalnum(*name_ptr) && (*name_ptr != '-'))
        {
            // We're done. Bad name.
            return 0;
        }

        // Keep on looking.
        ++name_ptr;
    }

    // Must be good.
    return 1;
}

static void null_terminate(char *src_str)
{
    char    *nl_ptr;        // Pointer to the newline  in the string.

    // Check for a new-line character.
    nl_ptr = strrchr(src_str, '\n');
    if (nl_ptr != nullptr)
    {
        // Wipe out the newline, if it's there
        *nl_ptr = '\0';
    }
}

} /* End of namespace CS_NETWORK */

TRACE_GLOBAL;           //!< Use the global trace control word.

using namespace CS_TRACE;
using namespace CS_NETWORK;

/*!
 * @brief The cs_network class constructor.
 */
CsNetwork::CsNetwork()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_NM_PROC;
}

/*!
 * @brief Convert an "dotted" IP network mask to a Classless Inter-Domain Routing (CIDR) notation value.
 */
int CsNetwork::toCIDR(const char *ip_addr)
{
    int netmask_cidr;
    int ipbytes[4];

    netmask_cidr = 0;
    sscanf(ip_addr, "%d.%d.%d.%d", &ipbytes[0], &ipbytes[1], &ipbytes[2], &ipbytes[3]);

    for (int i = 0; i < 4; i++)
    {
        switch (ipbytes[i])
        {
        case 0x80:
            netmask_cidr += 1;
            break;
        case 0xC0:
            netmask_cidr += 2;
            break;
        case 0xE0:
            netmask_cidr += 3;
            break;
        case 0xF0:
            netmask_cidr += 4;
            break;
        case 0xF8:
            netmask_cidr += 5;
            break;
        case 0xFC:
            netmask_cidr += 6;
            break;
        case 0xFE:
            netmask_cidr += 7;
            break;
        case 0xFF:
            netmask_cidr += 8;
            break;
        default:
            return netmask_cidr;
            break;
        }
    }

    return netmask_cidr;
}

/*!
 * @brief Create a network manager connection for the Ethernet.
 *
 * This will create either a DHCP connection or a STATIC connection for use by
 * the Network Manager. While it is certainly possible to use libnm to do this
 * work, using the "nmcli" command line tool seems easier/more intuitive to me.
 *
 * @param ip_cfg    The Ethernet configuration settings register value [GIGE_REG_IP_CUR_CFG].
 * @returns CS_NET_SUCCESS on success, otherwise an error condition.
 */
int CsNetwork::addEnetConnection(uint32_t ip_cfg, uint32_t ip_addr, uint32_t subnet_addr, uint32_t gateway_addr)
{
    int status = CS_NET_SUCCESS;
    std::string dev_name;
    std::string command;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_NM_PROC, "%s: ip_cfg=0x%08x, ip_addr=0x%08x, subnet_addr=0x%08x, gateway_addr=0x%08x", __PRETTY_FUNCTION__,
       ip_cfg, ip_addr, subnet_addr, gateway_addr);

    // Remove all enet-XXX connections.
//    LOG("%s: nmcli con del enet-dhcp enet-static", __PRETTY_FUNCTION__);
    system("nmcli con del enet-dhcp enet-static");

    // Get the device name we're connecting to.
    CsNetwork::getDevName(ETHER, dev_name);

    // Are we using DHCP or a static address?
    if ((ip_cfg & GIGE_IFCONFIG_DHCP_BE) == GIGE_IFCONFIG_DHCP_BE)
    {
        // We need to create the DHCP connection before we activate it.
        command = fmt::sprintf("nmcli con add type ethernet con-name enet-dhcp ifname \"%s\"", dev_name.c_str());
    }
    else
    {
        struct in_addr ip;      // IP address structure - used to convert to ASCII from uint32_t.
        struct in_addr subnet;      // Subnet mask structure - used to convert to ASCII from uint32_t.
        struct in_addr gateway;     // Gateway structure - used to convert to ASCII from uint32_t.

        if (ip_addr == 0)
            ip_addr = regs[GIGE_REG_IP_ADDR].get_value();
        if (subnet_addr == 0)
            subnet_addr = regs[GIGE_REG_SUBNET_MASK].get_value();
        if (gateway_addr == 0)
            gateway_addr = regs[GIGE_REG_DEF_GATEWAY].get_value();

        ip.s_addr = ip_addr;
        subnet.s_addr = subnet_addr;
        gateway.s_addr = gateway_addr;

        // We need to create the static connection before we activate it.
        command = fmt::sprintf("nmcli con add type ethernet con-name enet-static ifname \"%s\"", dev_name.c_str());
        command += fmt::sprintf(" ip4 %s/%d", inet_ntoa(ip), CsNetwork::toCIDR(inet_ntoa(subnet)));
        command += fmt::sprintf(" gw4 %s", inet_ntoa(gateway));
    }

    // Run the command to create the connection.
//    LOG("%s: '%s'", __PRETTY_FUNCTION__, command.c_str());
    status = system(command.c_str());
    if ((status < 0) || (WEXITSTATUS(status) != 0))
    {
        // Problem running the script. Log the error.
        fi_log(LOG_ERR, "%s: System cannot execute the command '%s'.", __FUNCTION__, command.c_str());
        status = CS_NET_ERROR_FD;
    }

    // Return our status.
    return status;
}

/*!
 * @brief Create a network manager connection for the WIFI.
 *
 * This will create either a DHCP connection or a STATIC connection for use by
 * the Network Manager. While it is certainly possible to use libnm to do this
 * work, using the "nmcli" command line tool seems easier/more intuitive to me.
 *
 * @param ip_cfg    The Ethernet configuration settings register value [GIGE_REG_WIFI_IP_CUR_CFG].
 * @param ap_cfg    The WIFI configuration settings register value [GIGE_REG_WIFI_AP_CFG].
 * @returns CS_NET_SUCCESS on success, otherwise an error condition.
 */
int CsNetwork::addWifiConnection(uint32_t ip_cfg, uint32_t ap_cfg, uint32_t ip_addr, uint32_t subnet_addr, uint32_t gateway_addr)
{
    int status = CS_NET_SUCCESS;
    std::string dev_name;
    std::string command;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Remove any other wifi-XXX connections. We may not have either of these
    // so any error messages are acceptable.
//    LOG("%s: nmcli con del wifi-dhcp wifi-static", __PRETTY_FUNCTION__);
    system("nmcli con del wifi-dhcp wifi-static");

    // If the SSID is empty, then we're done.
    if (((char *)regs[GIGE_REG_WIFI_SSID])[0] == 0)
    {
        fi_log(LOG_INFO, "%s: Just delete the WIFI connections and finish.", __PRETTY_FUNCTION__);
        return CS_NET_SUCCESS;
    }

    // Get the device name we're connecting to.
    CsNetwork::getDevName(WIFI, dev_name);
    if (dev_name.empty())
    {
        fi_log(LOG_DEBUG, "%s: Unable to get a dev_name!", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_FD;
    }

    // Are we using DHCP or a static address?
    if ((ip_cfg & GIGE_IFCONFIG_DHCP_BE) == GIGE_IFCONFIG_DHCP_BE)
    {
        // We need to create the DHCP connection before we activate it.
        command = fmt::sprintf("nmcli con add type wifi con-name wifi-dhcp ifname %s ssid \"%s\"",
                               dev_name.c_str(), (char *)regs[GIGE_REG_WIFI_SSID]);
    }
    else
    {
        struct in_addr ip;      // IP address structure - used to convert to ASCII from uint32_t.
        struct in_addr subnet;      // Subnet mask structure - used to convert to ASCII from uint32_t.
        struct in_addr gateway;     // Gateway structure - used to convert to ASCII from uint32_t.

        if (ip_addr == 0)
            ip_addr = regs[GIGE_REG_WIFI_IP_ADDR].get_value();
        if (subnet_addr == 0)
            subnet_addr = regs[GIGE_REG_WIFI_SUBNET_MASK].get_value();
        if (gateway_addr == 0)
            gateway_addr = regs[GIGE_REG_WIFI_DEF_GATEWAY].get_value();

        ip.s_addr = ip_addr;
        subnet.s_addr = subnet_addr;
        gateway.s_addr = gateway_addr;

        // We need to create the static connection before we activate it.
        command = fmt::sprintf("nmcli con add type wifi con-name wifi-static ifname %s ssid \"%s\"", dev_name.c_str(), (char *)regs[GIGE_REG_WIFI_SSID]);
        command += fmt::sprintf(" ip4 %s/%d", inet_ntoa(ip), CsNetwork::toCIDR(inet_ntoa(subnet)));
        command += fmt::sprintf(" gw4 %s", inet_ntoa(gateway));
    }

    // Are we setting up an ADHOC network?
    if ((ap_cfg & GIGE_WIFI_AP_CFG_ADHOC) == GIGE_WIFI_AP_CFG_ADHOC)
    {
        // Set the WIFI mode appropriately.
        command.append(" mode adhoc");
    }

    // Figure out what the security info needs to be.
    if ((ap_cfg & GIGE_WIFI_AP_CFG_SEC_MASK) == GIGE_WIFI_AP_CFG_SEC_WPA)
    {
        command += fmt::sprintf(" wifi-sec.key-mgmt wpa-psk wifi-sec.psk \"%s\"",
                                (char *)regs[ GIGE_REG_WIFI_PASSPHRASE ]);
    }

    // Run the command to create the connection.
//    LOG("%s: '%s'", __PRETTY_FUNCTION__, command.c_str());
    status = system(command.c_str());
    if ((status < 0) || (WEXITSTATUS(status) != 0))
    {
        // Problem running the script. Log the error.
        fi_log(LOG_ERR, "%s: System cannot execute the command '%s'.", __FUNCTION__, command.c_str());
        status = CS_NET_ERROR_FD;
    }

    // Return our status.
    return status;
}

/*!
 * @brief Get our Ethernet MAC address
 *
 * We need the MAC address early during init, before other network classes are set up.
 * So this cannot rely on other things being initialized.
 *
 * @return CS_NET_SUCCESS on success, otherwise an error condition.
 */
int CsNetwork::getEth0MacAddress(uint8_t *p_mac)
{
    int rtn = -1;
    std::string dev_name;
    CsSocket eth0_sock(AF_INET, SOCK_DGRAM);

    // Do we have our socket?
    if (eth0_sock.getFD() < 0)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create our socket!", __PRETTY_FUNCTION__);
        goto ma_bail;
    }

    // Get the device we want the MAC address of.
    CsNetwork::getDevName(ETHER, dev_name);
    if (dev_name.empty())
    {
        fi_log(LOG_DEBUG, "%s: Unable to get a dev_name!", __PRETTY_FUNCTION__);
        goto ma_bail;
    }

    // Get the MAC address.
    if (!eth0_sock.getMACAddress(dev_name, p_mac))
    {
        fi_log(LOG_DEBUG, "%s: getMACAddress() failed!", __PRETTY_FUNCTION__);
        goto ma_bail;
    }

    // All is well.
    rtn = CS_NET_SUCCESS;

ma_bail:
    return rtn;
}

/*!
 * @brief Get the current ip address from the OS.
 *
 * @return CS_NET_SUCCESS if successful or an error condition.
 */
int CsNetwork::refreshIpAddr(CS_NETWORK::NETWORK_INFO *net_info)
{
    CsSocket::ADDR ip_addr;

    // Sanity check.
    if (!isValid())
    {
        // fi_log(LOG_ERR, "%s: '%s' not opened.", __PRETTY_FUNCTION__, net_info->if_name);
        return CS_NET_ERROR_FD;
    }

    // Get the IP address, if no network cable is hooked up then we
    // won't be able to get the network address
    if (sock->getIPAddress(net_info->if_name, ip_addr))
    {
        // Save the information.
        net_info->ip_addr = ip_addr.u.sin.sin_addr;
        // fi_log(LOG_DEBUG, "%s: '%s' IP addr: 0x%08x", __PRETTY_FUNCTION__, net_info->if_name, net_info->ip_addr);
    }

    // All done.
    return CS_NET_SUCCESS;
}

/*!
 * @brief Get the current subnet mask from the OS.
 *
 * @return CS_NET_SUCCESS if successful or an error condition.
 */
int CsNetwork::refreshSubnet(CS_NETWORK::NETWORK_INFO *net_info)
{
    CsSocket::ADDR sn_addr;

    // Sanity check.
    if (!isValid())
    {
        // fi_log(LOG_ERR, "%s: '%s' not opened.", __PRETTY_FUNCTION__, net_info->if_name);
        return CS_NET_ERROR_FD;
    }

    // Subnet Mask, if there isn't a network cable plugged in then this really isn't
    // an error
    if (sock->getSubNet(net_info->if_name, sn_addr))
    {
        net_info->subnet_mask = sn_addr.u.sin.sin_addr;
        // fi_log(LOG_DEBUG, "%s: '%s' netmask: 0x%08x", __PRETTY_FUNCTION__, net_info->if_name, net_info->subnet_mask);
    }

    // All done.
    return CS_NET_SUCCESS;
}

bool CsNetwork::connectionExists(const char *con_name)
{
    bool found_it = false;      // True if we found our client.
    FILE *fp;               // Our pipe descriptor.
    std::string sys_cmd;
    const char *NMCLI_COMMAND = "nmcli -t con show \"%s\""; // Note that we want to see ALL connections, not just the active!

    // What connection are we looking for?
    sys_cmd = fmt::sprintf(NMCLI_COMMAND, con_name);

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get something?
            if (temp_buf[0] != 0)
            {
                // If we got something, then the connection exists.
                found_it = true;
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

    // Return our result.
    return found_it;

}

/*!
 * @brief Find the gateway we are attached to by looking through the routing
 *    tables in the proc file system.
 *
 * @return The gateway handle sucessful, 0 if not
 */
uint32_t CsNetwork::findGateway(CS_NETWORK::NETWORK_INFO *net_info)
{
    uint32_t gateway = 0;       // The gateway value from the route table.
    uint32_t dest;          // The destination value from the route table.
//    char interface[IFNAMSIZ + 1];   // The interface name from the route table.
    std::string interface;  // The interface name from the route table.
    FILE *route_fd;         // File descriptor for the route table.
    char temp_buf[MAX_BUFFER_LEN];  // Temporary buffer.

    // Open up the routing table.
    route_fd = fopen(ROUTE_TABLE, "r");
    if (!route_fd)
    {
        fi_log(LOG_ERR, "(%s) unable to open route table %s %d(%s)",
               __FUNCTION__, ROUTE_TABLE, errno, strerror(errno));
        return 0;
    }

    // Read the header line
    if (!fgets(temp_buf, MAX_BUFFER_LEN, route_fd))
    {
        fi_log(LOG_ERR, "(%s) cannot read route table %s %d(%s)",
               __FUNCTION__, ROUTE_TABLE, errno, strerror(errno));
        goto cleanup;
    }

    // Now read read the whole table, the layout of each row is
    // interface<tab>gateway<tab>... we are only interested when
    // the gateway is not all zeros
    while (fgets(temp_buf, MAX_BUFFER_LEN, route_fd))
    {
        std::stringstream str_stream(temp_buf);

        // Parse the line.
        str_stream >> interface;
        str_stream >> std::hex >> dest;
        str_stream >> std::hex >> gateway;

        // If gateway is non-zero, and interface name matches
        if ((gateway != 0) && (strcmp(net_info->if_name, interface.c_str()) == 0))
        {
            break;
        }
        else
        {
            // Initialize our gateway value in case things go south.
            gateway = 0;
        }
    }

cleanup:

    // Close the file.
    fclose(route_fd);

    // Return our gateway information.
    return gateway;
}

/*!
 * @brief Discover information about the network,
 *
 * The information is stored in a static variable to be returned later
 * as requested.
 *
 * @returns CS_NET_SUCCESS on success negative numbers on error
 */
//
int CsNetwork::discoverNetwork()
{
    int i;                              // Generic loop variable.
    struct if_nameindex *ifni;          // Pointer to the network interface index/name structure.
    std::string enet_name;
    std::string wifi_name;

    // Sanity check.
    if (!isValid())
    {
        // Perfectly OK if we don't have a network.
        return CS_NET_ERROR_FD;
    }

    // Get the name of the network interfaces.
    CsNetwork::getDevName(ETHER, enet_name);
    CsNetwork::getDevName(WIFI, wifi_name);

    // Clear out our network information structure.
    memset(&eth_net_info, 0, sizeof(eth_net_info));
    memset(&wifi_net_info, 0, sizeof(wifi_net_info));

    // Find out the interfaces which are available.
    ifni = if_nameindex();
    if (ifni == nullptr)
    {
        fi_log(LOG_ERR, "%s: Couldn't discover network interfaces", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_FD;
    }

    for (i = 0; ifni[i].if_index; i++)
    {
        TP(TRACE_SEL_NM_PROC, "%s: Interface found index: %d interface: '%s'.", __PRETTY_FUNCTION__, ifni[i].if_index, ifni[i].if_name);

        // Is this the ethernet interface?
        if ((enet_name.compare(ifni[i].if_name) == 0) && isEthConnected())
        {
            // fi_log(LOG_DEBUG, "%s: Saving enet_name [%s]", __PRETTY_FUNCTION__, ifni[i].if_name);

            // Yup, so save it away.
            strncpy(eth_net_info.if_name, ifni[i].if_name, sizeof(eth_net_info.if_name));

            // Get the MAC address.
            if (!sock->getMACAddress(eth_net_info.if_name, eth_net_info.mac_addr))
            {
                fi_log(LOG_ERR, "%s: Unable to get the MAC address for interface '%s'", __PRETTY_FUNCTION__, eth_net_info.if_name);
                return CS_NET_ERROR_IOCTL;
            }

            // Get the IP address
            refreshIpAddr(&eth_net_info);

            // Subnet Mask
            refreshSubnet(&eth_net_info);

            // Find out our default gateway
            eth_net_info.def_gateway.s_addr = findGateway(&eth_net_info);
        }

        // How about the WIFI?
        else if (wifi_name.compare(ifni[i].if_name) == 0)
        {
            // LOG("%s: Interface found index: %d interface: '%s'.", __PRETTY_FUNCTION__, ifni[i].if_index, ifni[i].if_name );

            // Yup, so save it away.
            strncpy(wifi_net_info.if_name, ifni[i].if_name, sizeof(wifi_net_info.if_name));

            // Get the MAC address.
            if (!sock->getMACAddress(wifi_net_info.if_name, wifi_net_info.mac_addr))
            {
                fi_log(LOG_ERR, "%s: Unable to get the MAC address for interface '%s'", __PRETTY_FUNCTION__, eth_net_info.if_name);
                return CS_NET_ERROR_IOCTL;
            }

            // Get the IP address
            refreshIpAddr(&wifi_net_info);

            // Subnet Mask
            refreshSubnet(&wifi_net_info);

            // Find out our default gateway
            wifi_net_info.def_gateway.s_addr = findGateway(&wifi_net_info);
        }
    }

    // We want the UDP port (socket) to be bound to the hard wired
    // network if it exists. Otherwise, allow transactions across WIFI.
    if (eth_net_info.if_name[0] == '\0')
    {
        if (wifi_net_info.if_name[0] == '\0')
        {
            // Close the socket.
            sock.reset();
        }
        else
        {
            // Set the socket to the interface.
            if (!sock->setInterface(wifi_net_info.if_name))
            {
                fi_log(LOG_ERR, "%s: Unable to bind to '%s', errno:%d(%s)", __PRETTY_FUNCTION__, wifi_net_info.if_name, errno, strerror(errno));
                sock.reset();
                return CS_NET_ERROR_SOCKET;
            }
        }
    }
    else
    {
        // Set the socket to the interface.
        if (!sock->setInterface(eth_net_info.if_name))
        {
            fi_log(LOG_ERR, "%s: Unable to bind to '%s', errno:%d(%s)", __PRETTY_FUNCTION__, eth_net_info.if_name, errno, strerror(errno));
            sock.reset();
            return CS_NET_ERROR_SOCKET;
        }
    }

    // Release the interface record(s).
    if_freenameindex(ifni);

    // Return all is well.
    return CS_NET_SUCCESS;
}

/*!
 *
 * @brief Get the MAC address for the first interface
 *
 * @param p_mac The returned MAC address. 6 bytes.
 */
void CsNetwork::getMacAddress(uint8_t *p_mac) const
{
    // Sanity check.
    if (p_mac != nullptr)
    {
        // Always give preference to the eth0 interface!
        if (eth_net_info.if_name[0] != '\0')
        {
            // Don't need to refresh from the OS since this shouldn't change.
            memcpy(p_mac, eth_net_info.mac_addr, MAC_ADDR_LEN);
        }
        else if (wifi_net_info.if_name[0] != '\0')
        {
            // Don't need to refresh from the OS since this shouldn't change.
            memcpy(p_mac, wifi_net_info.mac_addr, MAC_ADDR_LEN);
        }
    }
}

/*!
 *
 * @brief Get the subnet mask for the first interface.
 *
 * @return The subnet mask as a binary
 */
uint32_t CsNetwork::getSubnetMask()
{
    uint32_t subnet = 0;

    // Always give preference to the eth0 interface!
    if (eth_net_info.if_name[0] != '\0')
    {
        refreshSubnet(&eth_net_info);
        subnet = (uint32_t)eth_net_info.subnet_mask.s_addr;
    }
    else if (wifi_net_info.if_name[0] != '\0')
    {
        refreshSubnet(&wifi_net_info);
        subnet = (uint32_t)wifi_net_info.subnet_mask.s_addr;
    }

    return subnet;
}

/*!
 *
 * @brief Get the IP address for the first interface.
 *
 * @return the IP address as a binary.
 */
uint32_t CsNetwork::getIpAddress()
{
    uint32_t ip_addr = 0;

    // Always give preference to the eth0 interface!
    if (eth_net_info.if_name[0] != '\0')
    {
        // This ip address might have changed out from under us,
        // so make sure we get the latest from the OS
        if (refreshIpAddr(&eth_net_info) == CS_NET_SUCCESS)
        {
            // Everything is fine. Return the IP address.
            ip_addr = (uint32_t)eth_net_info.ip_addr.s_addr;
        }
    }
    else if (wifi_net_info.if_name[0] != '\0')
    {
        // This ip address might have changed out from under us,
        // so make sure we get the latest from the OS
        if (refreshIpAddr(&wifi_net_info) == CS_NET_SUCCESS)
        {
            // Everything is fine. Return the IP address.
            ip_addr = (uint32_t)wifi_net_info.ip_addr.s_addr;
        }
    }

    return ip_addr;
}

/*!
 *
 * @brief Get the default gateway for the first interface.
 *
 * @return The default gateway.
 */
uint32_t CsNetwork::getDefGateway()
{
    uint32_t gateway = 0;

    // Always give preference to the eth0 interface!
    if (eth_net_info.if_name[0] != '\0')
    {
        eth_net_info.def_gateway.s_addr = findGateway(&eth_net_info);
        gateway = (uint32_t)eth_net_info.def_gateway.s_addr;
    }
    else if (wifi_net_info.if_name[0] != '\0')
    {
        wifi_net_info.def_gateway.s_addr = findGateway(&wifi_net_info);
        gateway = (uint32_t)wifi_net_info.def_gateway.s_addr;
    }

    return gateway;
}

/*!
 *
 * @brief Get the subnet mask for the WIFI interface.
 *
 * @return The subnet mask as a binary.
 */
uint32_t CsNetwork::getWifiSubnetMask()
{
    refreshSubnet(&wifi_net_info);
    return (uint32_t)wifi_net_info.subnet_mask.s_addr;
}

/*!
 * @brief Get the IP address for the WIFI interface.
 *
 * @return The IP address as a binary.
 */
uint32_t CsNetwork::getWifiIpAddress()
{
    refreshIpAddr(&wifi_net_info);
    return (uint32_t)wifi_net_info.ip_addr.s_addr;
}

/*!
 * @brief Get the default gateway for the WIFI interface.
 *
 * @return The default gateway.
 */
uint32_t CsNetwork::getWifiDefGateway()
{
    wifi_net_info.def_gateway.s_addr = findGateway(&wifi_net_info);
    return (uint32_t)wifi_net_info.def_gateway.s_addr;
}

/*!
 * @brief Return the name of the active connection from the
 *    Network Mangaer. If there is not an active connection
 *    for the type specified, return an empty string.
 *
 * @param con_name      [out] The name of the active connection.
 * @param type      [in] The type of connection to look for.
 *
 * @returns True if a connection is found, false otherwise.
 */
bool CsNetwork::getActiveConnection(std::string& con_name, NET_TYPE type)
{
    bool rc = false;            // Our return code, assuming failure.
    FILE *fp;               // Our pipe descriptor.

    TP(TRACE_SEL_NM_PROC, "%s: type=%d", __PRETTY_FUNCTION__, (int)type);

    // Reset our return string.
    con_name.clear();

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, CMD_GET_ACTIVE);
    fp = popen(CMD_GET_ACTIVE, "r");
    if (fp != nullptr)
    {
        char temp_buf[MAX_BUFFER_LEN]; // Our response buffer.

        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                char *type_ptr;

                // Find the separator character.
                type_ptr = strchr(temp_buf, ':');
                if (type_ptr != nullptr)
                {
                    // Null terminate the string here.
                    *type_ptr = '\0';

                    // Move up to the type string.
                    ++type_ptr;

                    // Is this one we're looking for?
                    if (((type == ETHER) && (strstr(type_ptr, NM_ETHERNET_TYPE) != nullptr)) ||
                        ((type == WIFI) && ((strstr(type_ptr, NM_WIFI_TYPE) != nullptr) || (strstr(type_ptr, NM_HOTSPOT_TYPE) != nullptr))))
                    {
                        TP(TRACE_SEL_NM_PROC, "%s: Got it [%s]", __PRETTY_FUNCTION__, temp_buf);

                        // We found it! Copy over the connection name.
                        con_name.append(temp_buf);
                        rc = true;
                        break;
                    }
                }
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

    // Return our result.
    return rc;
}

/*!
 * @brief See if the we are using DHCP (vs a static IP address).
 *
 * We need this early during init, before other network classes are set up.
 * So this cannot rely on other things being initialized.
 *
 * @returns True if using DHCP, false if not.
 */
bool CsNetwork::getPreferredDhcpMode(NET_TYPE type)
{
    bool rc = true;         // Our return code, assuming DHCP mode.
    FILE *fp;               // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.
    char temp_buf[MAX_BUFFER_LEN]; // Our response buffer.

    TP(TRACE_SEL_NM_PROC, "%s: type=%d", __PRETTY_FUNCTION__, (int)type);

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, type))
    {
        // Couldn't get an active connection. Default to DHCP.
        goto dm_bail;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_METHOD, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                TP(TRACE_SEL_NM_PROC, "%s: Found it [%s]", __PRETTY_FUNCTION__, temp_buf);

                // Is this one we're looking for?
                if (strncmp(temp_buf, NM_STATIC_METHOD, strlen(NM_STATIC_METHOD)) == 0)
                {
                    // Return that we found a static connection.
                    rc = false;
                }

                // Done.
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

dm_bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Get the IP address from the network manager.
 *
 * We need this early during init, before other network classes are set up.
 * So this cannot rely on other things being initialized.
 *
 * @returns IP address (network byte order) if successful or 0 if it couldn't be read or
 *      we couldn't make sense of what we read.
 */
uint32_t CsNetwork::getPreferredIpAddress(NET_TYPE type)
{
    uint32_t ip_addr = 0;
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

    TP(TRACE_SEL_NM_PROC, "%s: type=%d", __PRETTY_FUNCTION__, (int)type);

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, type))
    {
        // COuldn't get an active connection.
        fi_log(LOG_WARNING, "%s: Unable to find an active connection of type [%d].", __PRETTY_FUNCTION__, (int)type);
        goto ip_bail;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_IP_ADDR, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Convert to our hex value.
                ip_addr = inet_addr(temp_buf);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

ip_bail:
    // Return our result.
    return ip_addr;
}

/*!
 * @brief Get the subnet mask from the network manager.
 *
 * We need this early during init, before other network classes are set up.
 * So this cannot rely on other things being initialized.
 *
 * @returns Subnet mask (network byte order) if successful or 0 if it failed.
 */
uint32_t CsNetwork::getPreferredSubnetMask(NET_TYPE type)
{
    uint32_t subnet_mask = 0;
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

    TP(TRACE_SEL_NM_PROC, "%s: type=%d", __PRETTY_FUNCTION__, (int)type);

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, type))
    {
        // COuldn't get an active connection.
        fi_log(LOG_WARNING, "%s: Unable to find an active connection of type [%d].", __PRETTY_FUNCTION__, (int)type);
        goto sm_bail;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_SUBNET_MASK, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Convert to our hex value.
                uint32_t prefix = strtoul(temp_buf, nullptr, 0);
                if (prefix > 0)
                {
                    subnet_mask = htonl(~((1U << (32U - prefix)) - 1U));
                    break;
                }
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

sm_bail:
    // Return our result.
    return subnet_mask;
}

/*!
 * @brief Get the gateway IP address from the network manager.
 *
 * We need this early during init, before other network classes are set up.
 * So this cannot rely on other things being initialized.
 *
 * @returns Gateway address (network byte order) if successful or 0 if it failed.
 */
uint32_t CsNetwork::getPreferredDefGateway(NET_TYPE type)
{
    uint32_t gate_addr = 0;
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

    TP(TRACE_SEL_NM_PROC, "%s: type=%d", __PRETTY_FUNCTION__, (int)type);

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, type))
    {
        // COuldn't get an active connection.
        fi_log(LOG_WARNING, "%s: Unable to find an active connection of type [%d].", __PRETTY_FUNCTION__, (int)type);
        goto gw_bail;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_GATEWAY, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Convert to our hex value.
                gate_addr = inet_addr(temp_buf);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

gw_bail:
    // Return our result.
    return gate_addr;
}

/*!
 * @brief Return true if the camera is running from NFS mounted system
 */
bool CsNetwork::isNfsMount()
{
    bool rc = false;
    FILE *fp;            // file pointer for the /proc interface

    // Check the cmdline to see if the there is a nfsroot param
    fp = fopen("/proc/cmdline", "r");
    if (fp != nullptr)
    {
        char buf[NFS_CHECK_BUF_SIZE];

        // Go through each line, look for the mount point in question
        while (fgets(buf, NFS_CHECK_BUF_SIZE, fp) != nullptr)
        {
            // Get rid of any trailing '\n'.
            null_terminate(buf);

            // Any nfsroot string?
            if (strstr(buf, "nfsroot=") != nullptr)
            {
                // Once is enough.
                rc = true;
                break;
            }
        }

        // Close the pipe.
        fclose(fp);
    }

    // Return our findings.
    return rc;
}

/*!
 * @brief Client has requested that DHCP be used on the network
 * interface specified (ETHER or WIFI).
 *
 * @returns 0 if everything went well, otherwise a negative number
 */
int CsNetwork::startDhcpConnection(NET_TYPE type)
{
    int status = -1;
    std::string sys_cmd;
    std::string active_con; // The name of our active connection.
    const char *con_name = ((type == ETHER) ? ENET_DHCP_CLIENT : WIFI_DHCP_CLIENT);

    TP(TRACE_SEL_NM_PROC, "%s: con_name=[%s]", __PRETTY_FUNCTION__, con_name);

    // If our connection does NOT exist, then we want to add it.
    if (!connectionExists(con_name))
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        // Get the correct register and add the appropriate connection.
        if (type == ETHER)
        {
            status = CsNetwork::addEnetConnection(GIGE_IFCONFIG_DHCP_BE);
        }
        else
        {
            uint32_t ap_cfg_reg = regs[GIGE_REG_WIFI_AP_CFG].get_value();
            status = CsNetwork::addWifiConnection(GIGE_IFCONFIG_DHCP_BE, ap_cfg_reg);
        }

        // Check the status.
        if (status < 0)
        {
            // Problem running the script. Log the error.
            fi_log(LOG_ERR, "%s: Error adding the connection '%s'", __PRETTY_FUNCTION__, ((type == ETHER) ? ENET_DHCP_CLIENT : WIFI_DHCP_CLIENT));
            status = -1;
            goto dhcp_bail;
        }
        else if (((char *)regs[GIGE_REG_WIFI_SSID])[0] == 0)
        {
            goto dhcp_bail;
        }
    }

    // Is there an existing active connection?
    if (CsNetwork::getActiveConnection(active_con, type))
    {
        // Make sure it is not our connection.
        if (active_con.compare(con_name) != 0)
        {
            // Shutdown this connection.
            TP(TRACE_SEL_NM_PROC, "%s: Found an active WIFI connection [%s].", __PRETTY_FUNCTION__, active_con.c_str());
            sys_cmd = fmt::sprintf("nmcli con down \"%s\"; nmcli con del \"%s\"", active_con.c_str(), active_con.c_str());
//            LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
            status = system(sys_cmd.c_str());
            if ((status < 0) || (WEXITSTATUS(status) != 0))
            {
                // Problem running the script. Log the error.
                fi_log(LOG_ERR, "%s: System cannot execute the command '%s'.", __PRETTY_FUNCTION__, sys_cmd.c_str());
            }
        }
        else
        {
            status = 0;
            goto dhcp_bail;
        }
    }

    // Build the system command as appropriate.
    sys_cmd = fmt::sprintf("nmcli con up \"%s\" &", con_name);
    TP(TRACE_SEL_NM_PROC, "%s: '%s'", __PRETTY_FUNCTION__, sys_cmd.c_str());

    // Run the command to switch over to the DHCP connection.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    status = system(sys_cmd.c_str());
    if ((status < 0) || (WEXITSTATUS(status) != 0))
    {
        // Problem running the script. Log the error.
        fi_log(LOG_ERR, "%s: System cannot execute the command '%s'.", __PRETTY_FUNCTION__, sys_cmd.c_str());
        status = -1;
        goto dhcp_bail;
    }

    // All is well.
    status = 0;

dhcp_bail:
    // Return our status.
    return status;
}

/*!
 * @brief Client wants to start a static address network connection.
 *
 * @returns 0 if everything went well, negative number otherwise.
 */
int CsNetwork::startStaticConnection(NET_TYPE type, uint32_t ip_addr, uint32_t subnet_addr, uint32_t gateway_addr)
{
    int status = -1;
    std::string sys_cmd;
    const char *con_name = ((type == ETHER) ? ENET_STATIC_CLIENT : WIFI_STATIC_CLIENT);

    TP(TRACE_SEL_NM_PROC, "%s: %s", __PRETTY_FUNCTION__, con_name);

    // If our connection does NOT exist, then we want to add it.
    if (!connectionExists(con_name))
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        // Get the correct register and add the appropriate connection.
        if (type == ETHER)
        {
            status = CsNetwork::addEnetConnection(GIGE_IFCONFIG_PERSISTENT_BE, ip_addr, subnet_addr, gateway_addr);
        }
        else
        {
            uint32_t ap_cfg_reg = regs[GIGE_REG_WIFI_AP_CFG].get_value();
            status = CsNetwork::addWifiConnection(GIGE_IFCONFIG_PERSISTENT_BE, ap_cfg_reg, ip_addr, subnet_addr, gateway_addr);
        }

        // Check the status.
        if (status < 0)
        {
            // Problem running the script. Log the error.
            fi_log(LOG_ERR, "%s: Error adding the connection '%s'", __PRETTY_FUNCTION__, ((type == ETHER) ? ENET_STATIC_CLIENT : WIFI_STATIC_CLIENT));
            status = -1;
            goto static_bail;
        }
        else if (((char *)regs[GIGE_REG_WIFI_SSID])[0] == 0)
        {
            goto static_bail;
        }
    }

    // Build the system command as appropriate.
    sys_cmd = fmt::sprintf("nmcli con up \"%s\"", con_name);
    TP(TRACE_SEL_NM_PROC, "%s: '%s'", __PRETTY_FUNCTION__, sys_cmd.c_str());
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());

    // Run the command to switch over to the static connection.
    status = system(sys_cmd.c_str());
    if ((status < 0) || (WEXITSTATUS(status) != 0))
    {
        // Problem running the script. Log the error.
        fi_log(LOG_ERR, "%s: System cannot execute the command '%s'.", __PRETTY_FUNCTION__, sys_cmd.c_str());
        status = -1;
        goto static_bail;
    }

    // All is well.
    status = 0;

static_bail:

    // Return our status.
    return status;
}

/*!
 * @brief The camera name has been changed. Send the new hostname to the OS.
 *
 * @param   The new hostname (camera name)
 *
 * @returns 0 if everything went well, otherwise a negative number
 */
int CsNetwork::exportCameraName(const char *new_name)
{
    int status;
    std::string temp_buf;         // Temporary buffer to hold whatever we want it to.

    // Make sure the name is valid.
    if (!is_name_valid(new_name))
    {
        fi_log(LOG_ERR, "(%s): Invalid camera name '%s'", __PRETTY_FUNCTION__, new_name);
        return -1;
    }

    // Build up the command string.
    temp_buf = fmt::sprintf("/usr/bin/hostnamectl set-hostname \"%s\"", new_name);

    // all the real logic is contained in a shell script. All we need to do is execute it.
    status = system(temp_buf.c_str());
    if ((status < 0) || (WEXITSTATUS(status) != 0))
    {
        // Had a problem. Log it.
        fi_log(LOG_ERR, "%s: cannot execute the command '%s'.", __PRETTY_FUNCTION__, temp_buf.c_str());
        status = -1;
    }
    else
        status = 0;

    return status;
}

bool CsNetwork::getDevName(NET_TYPE type, std::string& dev_name)
{
    bool found_it(false);
    FILE *fp;                       // Our pipe descriptor.
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.
    const char *GET_DEVICES = "nmcli -t -f DEVICE,TYPE dev";

    static std::string ether_label;     // The ethernet device name or label.
    static std::string wifi_label;      // The wifi device name or label.
    static std::string local_label;     // The local network interface name or label.

    // Which one is the user asking for?
    dev_name.clear();
    switch (type)
    {
    case ETHER:
        if (!ether_label.empty())
            dev_name = ether_label;
        break;
    case WIFI:
        if (!wifi_label.empty())
            dev_name = wifi_label;
        break;
    case LOCAL:
        if (!local_label.empty())
            dev_name = local_label;
        break;
    }

    // Did we already have the name?
    if (!dev_name.empty())
    {
        found_it = true;
        goto gdn_bail;
    }

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, GET_DEVICES);
    fp = popen(GET_DEVICES, "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (!found_it && (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr))
        {
            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                char *str;

                // Find the separating ':' character.
                str = strchr(temp_buf, ':');
                if (str)
                {
                    std::string dev_type;

                    // Null terminate the device name.
                    *str++ = '\0';

                    // Our str pointer is now at the type. See if we have
                    // a match.
                    dev_type.assign(str);
                    switch (type)
                    {
                    case ETHER:
                        if (dev_type.compare("ethernet") == 0)
                        {
                            // We have a match!
                            found_it = true;
                            dev_name.assign(temp_buf);
                            ether_label = dev_name;
                        }
                        break;
                    case WIFI:
                        if (dev_type.compare("wifi") == 0)
                        {
                            // We have a match!
                            found_it = true;
                            dev_name.assign(temp_buf);
                            wifi_label = dev_name;
                        }
                        break;
                    case LOCAL:
                        if (dev_type.compare("loopback") == 0)
                        {
                            // We have a match!
                            found_it = true;
                            dev_name.assign(temp_buf);
                            local_label = dev_name;
                        }
                        break;
                    default:
                        fi_log(LOG_DEBUG, "%s: Unexpected type [%d]!", __PRETTY_FUNCTION__, (int)type);
                        break;
                    }
                }
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

gdn_bail:
    // Did we find it?
    return found_it;
}

/*!
 * @brief The camera name has been changed. Call the script
 *    which updates the hostname to the OS.
 *
 * @param hostname_buf  Buffer to hold the hostname.
 * @param buflen    How big is the buffer to hold the hostname.
 *
 * @returns The length of the hostname found, 0 is none found.
 */
int CsNetwork::getHostname(char *hostname_buf, int buflen)
{
    int     num_bytes(0);       // How many bytes have been found.
    FILE    *fp;                // File pointer for the information returned from fopen
    char    temp_buf[ MAX_BUFFER_LEN ]; // Temporary buffer.

    // Initialize our temporary buffer.
    memset(temp_buf, 0, MAX_BUFFER_LEN);

    // Check hostname file to get the current hostname i.e. camera name.
    fp = fopen("/etc/hostname", "r");
    if (fp != nullptr)
    {
        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Look for a line with "HOSTNAME"
            if (temp_buf[0] != 0)
            {
                // Get rid of any trailing '\n'.
                null_terminate(temp_buf);

                // All done.
                break;
            }
        }

        // Close the file
        fclose(fp);
    }

    // If we didn't find a name correctly then just use the default name
    if (strlen(temp_buf) == 0)
    {
        // Copy the default name over
        strncpy(hostname_buf, HOSTNAME_DEF, buflen);
    }
    else
    {
        strncpy(hostname_buf, temp_buf, buflen);
    }

    // set the length
    if (buflen > (int)strlen(hostname_buf))
    {
        num_bytes = strlen(hostname_buf);
    }
    else
    {
        num_bytes = buflen;
    }

    // Return the number of bytes read.
    return num_bytes;
}

/*!
 * @brief Close down the socket.
 */
void CsNetwork::close()
{
    LOG("%s", __PRETTY_FUNCTION__);
    sock.reset();
}

/*!
 * @brief Fetch the SSID from the network manager.
 */
void CsNetwork::getSsidFromConf(char *ssid, int max_len)
{
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.

    TP(TRACE_SEL_NM_PROC, "%s", __PRETTY_FUNCTION__);

    // Clear the return string.
    ssid[0] = 0;

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, WIFI))
    {
        // COuldn't get an active connection.
        fi_log(LOG_WARNING, "%s: Unable to find an active WIFI connection.", __PRETTY_FUNCTION__);
        return;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_SSID, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Copy the string over.
                strncpy(ssid, temp_buf, max_len - 1);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }
}

void CsNetwork::getWifiPassphraseFromConf(char *pw, int max_len)
{
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.

    TP(TRACE_SEL_NM_PROC, "%s", __PRETTY_FUNCTION__);

    // Clear the return string.
    pw[0] = 0;

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, WIFI))
    {
        // COuldn't get an active connection.
        fi_log(LOG_WARNING, "%s: Unable to find an active WIFI connection.", __PRETTY_FUNCTION__);
        return;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_PASSPHRASE, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Copy the string over.
                strncpy(pw, temp_buf, max_len - 1);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Fetch the WIFI config info from the network manager.
 */
uint32_t CsNetwork::getWifiConfigFromConf()
{
    uint32_t config;
    FILE *fp;                   // Our pipe descriptor.
    std::string active_con; // The name of our active connection.
    std::string sys_cmd;    // Command to query the connection.
    std::string mode_str;
    std::string key_mgmt;
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

    TP(TRACE_SEL_NM_PROC, "%s", __PRETTY_FUNCTION__);

    // Asumme open security, not Ad Hoc.
    config = GIGE_WIFI_AP_CFG_SEC_OPEN;

    // Get the active connection.
    if (!CsNetwork::getActiveConnection(active_con, WIFI))
    {
        // Couldn't get an active connection.
        goto config_bail;
    }

    // Build up our command.
    sys_cmd = fmt::sprintf(CMD_GET_MODE, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Copy the string over.
                mode_str.append(temp_buf);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

    // Are we doing AdHoc mode?
    if (mode_str.compare("adhoc") == 0)
    {
        config |= GIGE_WIFI_AP_CFG_ADHOC;
    }

    // Build up our command line.
    sys_cmd = fmt::sprintf(CMD_GET_KEY_MGMT, active_con.c_str());

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    fp = popen(sys_cmd.c_str(), "r");
    if (fp != nullptr)
    {
        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            TP(TRACE_SEL_NM_PROC, "%s: Read [%s]", __PRETTY_FUNCTION__, temp_buf);

            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                // Copy the string over.
                key_mgmt.append(temp_buf);
                break;
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

    // Get our WIFI security settings.
    if (key_mgmt.compare("wpa-psk") == 0)
    {
        // We DO have security settings! The only security we support is WPA-PSK, so this must be it.
        config &= ~GIGE_WIFI_AP_CFG_SEC_MASK;
        config |= GIGE_WIFI_AP_CFG_SEC_WPA;
    }
    else
    {
        fi_log(LOG_WARNING, "%s: unexpected key-management='%s'", __PRETTY_FUNCTION__, key_mgmt.c_str());
    }

config_bail:
    // Return the configuration.
    return config;
}

bool CsNetwork::isWifiDevicePresent()
{
    const LicManager *lic_ptr = nullptr;   // Pointer to the licensing information.
    bool licensed = false;

    lic_ptr = CameraServer::get_license();
    if (lic_ptr != nullptr)
        licensed = (bool)lic_ptr->wireless();

    // Return true if we are licensed for WIFI.
    if (licensed)
    {
        // Make sure our WIFI info is initialized.
        if (wifi_net_info.if_name[0] == '\0')
        {
            csWorker *p_worker;

            // Get our worker pointer.
            p_worker = CameraServer::get_worker();

            // Refresh the network just in case something has changed.
            p_worker->discoverNetwork();
        }

        // Good to go if we found a WIFI device.
        return (wifi_net_info.if_name[0] != '\0');
    }

    // Make sure we reset the WIFI data structure.
    memset(&wifi_net_info, 0, sizeof(wifi_net_info));

    // No WIFI here.
    return false;
}

int CsNetwork::getWifiSignalStrength()
{
    int sig_strength = 0;
    FILE *fp;               // Our pipe descriptor.
    char ssid[MAX_BUFFER_LEN];      // The SSID we're connected to.
    const char *GET_SIG_STRENGTH = "nmcli -t -f SSID,SIGNAL dev wifi";

    // Get our SSID.
    CsNetwork::getSsidFromConf(ssid, MAX_BUFFER_LEN);

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, GET_SIG_STRENGTH);
    fp = popen(GET_SIG_STRENGTH, "r");
    if (fp != nullptr)
    {
        char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.

        // Make sure our buffer is empty.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
        {
            // Get rid of any trailing '\n'.
            null_terminate(temp_buf);

            // Did we get anything?
            if (temp_buf[0] != 0)
            {
                char *save_ptr;
                char *str;

                str = strtok_r(temp_buf, ":", &save_ptr);
                if (str && (strncmp(str, ssid, strlen(ssid)) == 0))
                {
                    str = strtok_r(nullptr, ":", &save_ptr);
                    sig_strength = (int)strtoul(str, nullptr, 0);
                    break;
                }
            }
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
    }

    // Return the signal strength read.
    return sig_strength;
}

uint32_t CsNetwork::getWifiStatus()
{
    uint32_t status = 0;

    // Is there even a WIFI device enabled?
    if (CsNetwork::isWifiDevicePresent())
    {
        // Good to go.
        status |= GIGE_WIFI_STATUS_DONGLE;

        // Since it's present and allowed...are we connected?
        if (CsNetwork::isWifiConnected())
        {
            // We are, indeed, connected.
            status |= GIGE_WIFI_STATUS_CONNECTED;

            // Normally we'd read the signal strength for the active WIFI SSID.
            // However, it can take a REALLY long time, so we'll start the measurement
            // and issue another status asynchronously.

#ifdef ALM_TODO
            // throw in signal strength now
            int signal = cs_network::get_wifi_signal_strength();
            status |= (signal << GIGE_WIFI_STATUS_SIGNAL_SHIFT);
#endif
        }
    }

    return status;
}

bool CsNetwork::isEthConnected()
{
    std::string dev_name;
    std::string connect_path;
    char temp_buf[MAX_BUFFER_LEN];
    FILE *fp;

    // ALM_TODO_NET
    // Replace reading the carrier pseudo-file with reading the active connections
    // from the network manager (nm_client_get_active_connections).

    // Build up the path.
    CsNetwork::getDevName(ETHER, dev_name);
    connect_path = fmt::sprintf(NET_CONNECT_STATUS, dev_name.c_str());

    // Open the net connection status file.
    fp = fopen(connect_path.c_str(), "r");
    if (fp == nullptr)
        return false;

    // Initialize our buffer.
    temp_buf[0] = '\0';

    // Read the file.
    fgets(temp_buf, MAX_BUFFER_LEN, fp);
    fclose(fp);

    // Return our status. A '1' means we're connected.
    return (temp_buf[0] == '1');
}

bool CsNetwork::isWifiConnected()
{
    std::string dev_name;
    std::string connect_path;
    char temp_buf[MAX_BUFFER_LEN];
    FILE *fp;

    // Build up the path.
    CsNetwork::getDevName(WIFI, dev_name);
    connect_path = fmt::sprintf(NET_CONNECT_STATUS, dev_name.c_str());

    // Open the net connection status file.
    fp = fopen(connect_path.c_str(), "r");
    if (fp == nullptr)
        return false;

    // Initialize our buffer.
    temp_buf[0] = '\0';

    // Read the file.
    fgets(temp_buf, MAX_BUFFER_LEN, fp);
    fclose(fp);

    // Return our status. A '1' means we're connected.
    return (temp_buf[0] == '1');
}

/*!
 * @brief Update the state of the IP configuration. This could be changing from
 *    DHCP to static or vs.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns 0 on success, <0 on failure.
 */
int CsNetwork::updateIpCfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int status = 0;
    bool dhcp_mode_set;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    TP(TRACE_SEL_NM_PROC, "%s: reg_value=0x%08x", __PRETTY_FUNCTION__, reg_value);

    // See if we're already using DHCP.
    dhcp_mode_set = getPreferredDhcpMode(CsNetwork::ETHER);

    // If they are trying to switch to DHCP mode, if it's already running then ignore.
    if (reg_value & GIGE_IFCONFIG_DHCP_BE)
    {
        fi_log(LOG_INFO, "%s: ifconfig:0x%08x switching to DHCP Mode", __PRETTY_FUNCTION__, reg_value);

        // Switching to DHCP mode
        if (!dhcp_mode_set)
        {
            // Start the connection.
            status = startDhcpConnection(CsNetwork::ETHER);
        }
    }
    else if (reg_value & GIGE_IFCONFIG_LLA_BE)
    {
        fi_log(LOG_INFO, "%s: ifconfig:0x%08x switching to Link Local Mode", __PRETTY_FUNCTION__, reg_value);
    }
    else
    {
        uint32_t    ip_addr;
        uint32_t    subnet_addr;
        uint32_t    gateway_addr;

        fi_log(LOG_INFO, "%s: ifconfig:0x%08x switching to static IP Mode", __PRETTY_FUNCTION__, reg_value);

        // We are setting up a static network, so we need to make sure the
        // user has already supplied us with the information we require.
        ip_addr = regs[GIGE_REG_PERSIST_IP].get_value();
        subnet_addr = regs[GIGE_REG_PERSIST_SUBNET].get_value();
        gateway_addr = regs[GIGE_REG_PERSIST_GATEWAY].get_value();
        if ((ip_addr == 0) || (subnet_addr == 0) || (gateway_addr == 0))
        {
            fi_log(LOG_ERR, "%s: cannot change to a static IP, either the IP address, the subnet mask, or the gateway haven't been set", __PRETTY_FUNCTION__);
            return -1;
        }

        // Start the connection.
        status = startStaticConnection(CsNetwork::ETHER);
    }

    // If we're successful, then save this in our register space.
    if (status == 0)
        regs[GIGE_REG_IP_CUR_CFG].set_value(reg_value);

    return status;
}

/*!
 * @brief update the state of the wifi ip configuration. This could be changing from dhcp
 * to static or vs.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns 0 if everything went well, <0 otherwise.
 */
int CsNetwork::updateWifiIpcfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int     status = 0;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    // Make sure we have WIFI before jumping into this.
    if (!CsNetwork::isWifiDevicePresent())
    {
        fi_log(LOG_DEBUG, "%s: No WIFI device is present!", __PRETTY_FUNCTION__);
        return -1;
    }

    // If they are trying to switch to DHCP mode, if it's already running then ignore.
    if (reg_value & GIGE_IFCONFIG_DHCP_BE)
    {
        fi_log(LOG_INFO, "%s: ifconfig:0x%x switching to DHCP Mode ", __PRETTY_FUNCTION__, reg_value);

        // Get the connection started.
        status = startDhcpConnection(CsNetwork::WIFI);
    }
    else
    {
        uint32_t ip_addr;
        uint32_t subnet_addr;
        uint32_t gateway_addr;

        fi_log(LOG_INFO, "%s: ifconfig:0x%x switching to static IP Mode", __PRETTY_FUNCTION__, reg_value);

        // We are setting up a static network, so we need to make sure the
        // user has already supplied us with the information we require.
        ip_addr = regs[ GIGE_REG_WIFI_IP_ADDR ].get_value();
        subnet_addr = regs[ GIGE_REG_WIFI_SUBNET_MASK ].get_value();
        gateway_addr = regs[ GIGE_REG_WIFI_DEF_GATEWAY ].get_value();
        if ((ip_addr == 0) || (subnet_addr == 0) || (gateway_addr == 0))
        {
            fi_log(LOG_ERR, "%s: Cannot change to a static ip either the ip address, the subnet mask, or the gateway haven't been set", __PRETTY_FUNCTION__);
            return -1;
        }

        // Get the connection started.
        status = startStaticConnection(CsNetwork::WIFI);
    }

    // If we're successful, then save this in our register space.
    if (status == 0)
        regs[GIGE_REG_WIFI_IP_CUR_CFG].set_value(reg_value);

    return status;
}

/*!
 * @brief Update the state of the WIFI AP configuration.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns 0 if everything went well, <0 otherwise.
 */
int CsNetwork::updateWifiApCfg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    uint32_t ip_cfg;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    // Make sure we have WIFI before jumping into this.
    if (!CsNetwork::isWifiDevicePresent())
    {
        fi_log(LOG_DEBUG, "%s: No WIFI device is present!", __PRETTY_FUNCTION__);
        return -1;
    }

    // Get the WIFI IP configuration value.
    ip_cfg = regs[GIGE_REG_WIFI_IP_CUR_CFG].get_value();

    // Create our new connection.
    if (CsNetwork::addWifiConnection(ip_cfg, reg_value) < 0)
    {
        fi_log(LOG_ERR, "%s: add_wifi_connection() failed!", __PRETTY_FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Save this in our register space.
    regs[GIGE_REG_WIFI_AP_CFG].set_value(reg_value);

    // All is well.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief initiate a wifi AP scan
 *
 * @param p_msg     Not used.
 * @param reg_value Not used.
 *
 * @returns GEV_STATUS_SUCCESS if everything goes well, GEV_STATUS_ERROR otherwise.
 */
int CsNetwork::wifiStartScan(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;
    (void)reg_value;

    // Request a WIFI scan. Sometimes this works, sometimes it doesn't. Why
    // it doesn't can be varied: WIFI isn't enabled, there's already a scan
    // going, etc. Just give it a go, and assume life is good.
//    LOG("%s: nmcli dev wifi rescan", __PRETTY_FUNCTION__);
    system("nmcli dev wifi rescan");

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Get a set of scan results.
 *
 * @returns true if everything went well, otherwise false
 */
bool CsNetwork::getWifiScanResults(char *buffer, int buffsize)
{
    int i;
    FILE *fp;               // Our pipe descriptor.
    char temp_buf[MAX_BUFFER_LEN];  // Our response buffer.
    const char *GET_SCAN_INFO = "nmcli -t -f SSID,SIGNAL,MODE,SECURITY dev wifi";

    // Set up a pointer to the buffer
    wifi_scan_results *results = (wifi_scan_results *)buffer;

    // Figure out how many entries we put into the buffer.
    int max_entries = buffsize / sizeof(wifi_scan_results);

    // Clear the buffer.
    memset(buffer, 0, buffsize);

    // Create a pipe to read the result of the command.
//    LOG("%s: [%s]", __PRETTY_FUNCTION__, GET_SCAN_INFO);
    fp = popen(GET_SCAN_INFO, "r");
    if (fp == nullptr)
    {
        fi_log(LOG_WARNING, "%s: cannot open nmcli pipe.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure our buffer is empty.
    temp_buf[0] = 0;

    // Read data, if any.
    i = 0;
    while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != nullptr)
    {
        // Get rid of any trailing '\n'.
        null_terminate(temp_buf);

        // LOG("%s: got [%s]", __PRETTY_FUNCTION__, temp_buf);

        // Did we get anything?
        if (temp_buf[0] != 0)
        {
            char *save_ptr;
            char *str;

            // The first string is the SSID.
            str = strtok_r(temp_buf, ":", &save_ptr);
            if (str && (str[0] != 0))
            {
                // Save the SSID.
                strncpy(results->ssid, str, GIGE_WIFI_SSID_MAX);
            }
            else
            {
                continue;
            }

            // Now parse the signal strength.
            str = strtok_r(nullptr, ":", &save_ptr);
            if (str && (str[0] != 0))
            {
                // Convert the signal strength.
                results->signal_level = (uint8_t)(strtoul(str, nullptr, 0) & 0xFF);
            }
            else
            {
                continue;
            }

            // Now parse the mode information.
            str = strtok_r(nullptr, ":", &save_ptr);
            if (str && (str[0] != 0))
            {
                results->adhoc = (strncasecmp(str, "adhoc", 5) == 0);
            }
            else
            {
                continue;
            }

            // Now parse the security information.
            str = strtok_r(nullptr, ":", &save_ptr);
            if (str && (str[0] != 0) && (strstr(str, "WPA2") != nullptr))
            {
                // It supports WPA-PSK, our favorite.
                results->security = GIGE_WIFI_AP_CFG_SEC_WPA;
            }
            else
            {
                // No security is indicated.
                results->security = GIGE_WIFI_AP_CFG_SEC_OPEN;
            }

            // Check if we're done.
            ++i;
            if (i == max_entries)
                break;

            // Move to the next record.
            results++;
        }
    }

    // Close the pipe.
    pclose(fp);

    return true;
}
