/*
 * cs_socket.cpp
 *
 *  Created on: Oct 28, 2015
 *      Author: eric
 */

#if defined(WIN32)
#include <windows.h>
#elif defined(LINUX) || defined(linux)
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#endif

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include <cstring>
#include <memory>

#include <fas_log.h>

#include "cs_socket.h"
#include "camera_server.h"

TRACE_GLOBAL;

/*!
 * @brief Constructor for the CsSocket class.
 *
 * @param family    [IN] The network "family", one of AF_INET or AF_UNIX.
 * @param s_type    [IN] The socket type, one of SOCK_STREAM or SOCK_DGRAM.
 */
CsSocket::CsSocket(sa_family_t family, __socket_type s_type) noexcept
{
    int rc;

    TP(CAMERA_SERVER::CS_CTOR_DTOR, "CsSocket: ctor");

    // Assume all is well.
    rc = true;

    // Initialize our address structure. If something goes wrong, this
    // field will be set to a known value.
    memset(&address, 0, sizeof(address));
    address.fd = -1;

    // Sanity check.
    switch (family)
    {
    case AF_INET:
    case AF_UNIX:
        // Save the family information.
        address.u.s.sa_family = family;
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unsupported address family: %d", __PRETTY_FUNCTION__, family);
        return;
    }

    switch (s_type)
    {
    case SOCK_STREAM:
    case SOCK_DGRAM:
        // Open up a socket.
        if ((address.fd = ::socket(family, s_type, 0)) < 0)
        {
            // Log the error.
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            rc = false;
        }
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unsupported socket type: %d", __PRETTY_FUNCTION__, (int)s_type);
        return;
    }

    // Setup family specific behavior we want.
    if (rc && (family == AF_INET))
    {
        int opt;

        // Try to avoid the "already in use" error. This error can happen
        // as a result of the "handshake" state machine sitting in a timeout
        // waiting for an ACK that will never come. This time out is quite
        // long (minutes), and so the port can be unusable for that period.
        //
        // Setting the SO_REUSEADDR option allows that port to be re-used,
        // although I've read that it can't be reused by the same remote IP
        // address!
        opt = 1;
        if (::setsockopt(address.fd, SOL_SOCKET, SO_REUSEADDR, (const char *)&opt, sizeof(int)) < 0)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            ::close(address.fd);
            address.fd = -1;
            rc = false;
        }

        // Setup the type specific behavior.
        if (rc && (s_type == SOCK_STREAM))
        {
            // Set socket to be nonblocking. All of the sockets for
            // the incoming connections will also be nonblocking since
            // they will inherit that state from the listening socket.
            if (ioctl(address.fd, FIONBIO, (char *)&opt) < 0)
            {
                LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
                ::close(address.fd);
                address.fd = -1;
            }
        }
    }
}

/*!
 * @brief Assignment, or copy, constructor for the CsSocket class.
 * @param sock  [IN] The socket object to copy from.
 */
CsSocket::CsSocket(const CsSocket& sock) noexcept
{
    // Initialize our information from the past in object.
    memcpy(&address, &sock.address, sizeof(address));

    // Do not carry over the file descriptor though.
    address.fd = -1;
}

/*!
 * @brief The CsSocket class destructor. Just making sure that
 *        the socket descriptor is closed.
 */
CsSocket::~CsSocket()
{
    TP(CAMERA_SERVER::CS_CTOR_DTOR, "CsSocket: dtor");

    if (address.fd != -1)
    {
        ::close(address.fd);
        address.fd = -1;
    }
}

/*!
 * @brief Accept connections to this socket.
 *
 * This is typically on used by SOCK_STREAM type sockets. The accept() method
 * takes the first connection request, creates a new connected socket, and
 * returns a new CsSocket object referring to that socket.
 *
 * @return The socket object that carries the data.
 */
std::unique_ptr<CsSocket> CsSocket::accept()
{
    int fd;
    std::unique_ptr<CsSocket> accept_sock;

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        goto accept_bail;
    }

    // We must have an accept pending on our socket.
    fd = ::accept(address.fd, NULL, NULL);
    if (fd >= 0)
    {
        // Build up our return socket.
        accept_sock = std::make_unique<CsSocket>(*this);
        if (accept_sock.get() == nullptr)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        }
        else
        {
            // Assign the file descriptor.
            accept_sock->address.fd = fd;
        }
    }
    else
    {
        fi_log(LOG_ERR, "%s: accept() failed: %d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
    }

accept_bail:
    // Return our "accepted" socket.
    return accept_sock;
}

/*!
 * @brief Bind our socket to the address and port specified in the parameters.
 *
 * @note This only works for AF_INET sockets (TCP, UDP, etc.). If you are using an
 *       AF_UNIX socket, you need the other CsSocket::bind specifying the file system
 *       path.
 *
 * @param ip_addr   [IN] The address we want to bind to.
 * @param port      [IN] The port we want to bind to.
 * @return True on success, false otherwise.
 */
bool CsSocket::bind(uint32_t ip_addr, uint16_t port)
{
    int rc;
    size_t addrSize = 0;

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Differentiate between an "Network" socket and a "Unix" socket.
    switch (address.u.s.sa_family)
    {
    case AF_INET:
        addrSize = sizeof(struct sockaddr_in);
        LOG("Opening %s:%d", inet_ntoa(address.u.sin.sin_addr), ntohs(address.u.sin.sin_port));
        break;

    case AF_UNIX:
        fi_log(LOG_DEBUG, "%s: You are using the wrong bind! AF_UNIX sockets require a path!", __PRETTY_FUNCTION__);
        return false;

    default:
        fi_log(LOG_DEBUG, "%s: Unsupported address family: %d", __PRETTY_FUNCTION__, address.u.s.sa_family);
        return false;
    }

    // Save the addressing information.
    address.u.sin.sin_addr.s_addr = htonl(ip_addr);
    address.u.sin.sin_port = htons(port);

    // Bind to that socket.
    rc = ::bind(address.fd, &address.u.s, addrSize);
    if (rc < 0)
    {
        int saveErrno = errno;

        // Close the socket.
        close(address.fd);
        address.fd = -1;

        // Throw an exception.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, saveErrno);
    }

    return (rc >= 0);
}

/*!
 * @brief Bind our socket to the address and port specified in the parameters.
 *
 * @note This only works for AF_UNIX sockets. If you are using an AF_INET socket
 *       (TCP, UDP, etc.), you need the other CsSocket::bind IP and port values.
 *
 * @param path      [IN] The path to link the socket to.
 * @return True on success, false otherwise.
 */
bool CsSocket::bind(const std::string& path)
{
    int rc;
    size_t addrSize = 0;

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Differentiate between an "Network" socket and a "Unix" socket.
    switch (address.u.s.sa_family)
    {
    case AF_UNIX:
        addrSize = sizeof(struct sockaddr_un);
        LOG("Opening %s", address.u.sun.sun_path);
        break;

    case AF_INET:
        fi_log(LOG_DEBUG, "%s: You are using the wrong bind! AF_INET sockets require an IP address and port!", __PRETTY_FUNCTION__);
        return false;

    default:
        fi_log(LOG_DEBUG, "%s: Unsupported address family: %d", __PRETTY_FUNCTION__, address.u.s.sa_family);
        return false;
    }

    // Save the addressing information.
    strncpy(address.u.sun.sun_path, path.c_str(), sizeof(address.u.sun.sun_path));

    // Bind to that socket.
    rc = ::bind(address.fd, &address.u.s, addrSize);
    if (rc < 0)
    {
        int saveErrno = errno;

        // Close the socket.
        close(address.fd);
        address.fd = -1;

        // Throw an exception.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, saveErrno);
    }

    return (rc >= 0);
}

/*!
 * @brief Mark the socket as a socket that will be used to accept incoming
 *        connection requests using CsSocket::accept. This is typically used
 *        with SOCK_STREAM type sockets.
 *
 * @return True on successs, false otherwise.
 */
bool CsSocket::listen()
{
    bool rc(true);

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Indicate that we want to accept connections. The '10' is the "listen backlog" value.
    // In reality, we only accept one connection at a time...
    if (::listen(address.fd, 10) < 0)
    {
        fi_log(LOG_ERR, "%s: Error listening to FSP socket, errno:%d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
        rc = false;
    }

    return rc;
}

/*!
 * @brief Get the IP address associated with the interface, if_name.
 *
 * @note If no network cable is attached, then we will not be able
 *       to get the address.
 *
 * @param if_name   [IN] The name of the interface we are interested in.
 * @param ip_addr   [OUT] The IP address found, if successful.
 * @return True on success, false otherwise.
 */
bool CsSocket::getIPAddress(const std::string& if_name, ADDR& ip_addr)
{
    bool rc(false);
    struct ifreq ifr;       // Interface request structure we use for socket ioctls.

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Clear the return structure.
    memset(&ip_addr, 0, sizeof(ip_addr));

    // Setup the request structure.
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, if_name.c_str(), IFNAMSIZ - 1);

    // Get the IP address, if no network cable is hooked up then we
    // won't be able to get the network address
    if (ioctl(address.fd, SIOCGIFADDR, &ifr) == 0)
    {
        // Copy the information over.
        ip_addr.u.sin.sin_addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;

        // All is well.
        rc = true;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Read the MAC address for the socket we're connected to.
 *
 * @param if_name   [IN] The name of the interface we're interested in.
 * @param mac_addr  [OUT] The 6-byte MAC address.
 * @return True on success, false for failure.
 */
bool CsSocket::getMACAddress(const std::string& if_name, uint8_t *mac_addr)
{
    bool rc(false);

    // Sanity check.
    if (mac_addr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: mac_addr pointer is null!", __PRETTY_FUNCTION__);
        return false;
    }
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Clear out the buffer.
    memset(mac_addr, 0, CS_NETWORK::MAC_ADDR_LEN);

#if defined(_WIN32) || defined(WIN32)
    UUID uuid;
    if (UuidCreateSequential(&uuid) == RPC_S_OK)
    {
        memcpy(mac_addr, (char *)(uuid.Data4 + 2), 6);
        rc = true;
    }
#elif defined(LINUX) || defined(linux)
    struct ifreq    ifr;                // Interface request structure we use for socket ioctls.

    // Clear out our structure.
    memset(&ifr, 0, sizeof(ifr));

    // Now find out information about this network.
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, if_name.c_str(), IFNAMSIZ - 1);

    // Get the MAC address
    if (ioctl(address.fd, SIOCGIFHWADDR, &ifr))
    {
        fi_log(LOG_ERR, "%s: ioctl SIOCGIFHWADDR failed interface: '%s'", __PRETTY_FUNCTION__, if_name.c_str());
    }
    else
    {
        // Save the information.
        memcpy(mac_addr, ifr.ifr_hwaddr.sa_data, CS_NETWORK::MAC_ADDR_LEN);
        rc = true;
    }
#endif

    // Return our result.
    return rc;
}

/*!
 * @brief Get the IP address associated with the interface, if_name.
 *
 * @note If no network cable is attached, then we will not be able
 *       to get the address.
 *
 * @param if_name   [IN] The name of the interface we are interested in.
 * @param ip_addr   [OUT] The IP address found, if successful.
 * @return True on success, false otherwise.
 */
bool CsSocket::getSubNet(const std::string& if_name, ADDR& sn_addr)
{
    bool rc(false);
    struct ifreq ifr;       // Interface request structure we use for socket ioctls.

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Clear the return structure.
    memset(&sn_addr, 0, sizeof(sn_addr));

    // Setup the request structure.
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, if_name.c_str(), IFNAMSIZ - 1);

    // Get the IP address, if no network cable is hooked up then we
    // won't be able to get the network address
    if (ioctl(address.fd, SIOCGIFNETMASK, &ifr) == 0)
    {
        // Copy the information over.
        sn_addr.u.sin.sin_addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;

        // All is well.
        rc = true;
    }

    // Return our result.
    return rc;

}

/*!
 * @brief Associate the socket with a particular interface (i.e., network device).
 *
 * @param if_name   [IN] The interface device we want associated with the socket.
 * @return True on success, false otherwise.
 */
bool CsSocket::setInterface(const std::string& if_name)
{
    struct ifreq    ifr;                // Interface request structure we use for socket ioctls.

    // Sanity check.
    if (address.fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Socket has not been opened!", __PRETTY_FUNCTION__);
        return false;
    }

    // Clear out our structure.
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, if_name.c_str(), IFNAMSIZ - 1);

    // Bind the socket as appropriate.
    if (setsockopt(address.fd, SOL_SOCKET, SO_BINDTODEVICE, (void *)&ifr, sizeof(ifr)) < 0)
    {
        fi_log(LOG_ERR, "%s: Unable to bind to '%s', errno:%d(%s)", __PRETTY_FUNCTION__, if_name.c_str(), errno, strerror(errno));
        return false;
    }

    // All is well.
    return true;
}
