/*!
 * @file cs_update.cpp
 * @brief Manage the camera server update process.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fmt/format.h>
#include <fmt/printf.h>

#include <fastec/fas_thread.h>
#include <fastec/fas_log.h>

#include <chrono>
#include <fstream>
#include <iostream>
#include <memory>
#include <vector>

#include "camera_server.h"
#include "cs_trace.h"
#include "cs_update.h"
#include "override.h"

namespace CS_UPDATE
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Amount of time to wait for a state change.
const int POLL_DELAY = 100;

//! A script file to check for changes between our local scripts and the
//! system repository.
const char CHECK_SCRIPTS_FILE[] = "/tmp/check_scripts.sh";

//! Location of the system scripts folder.
const char SYS_SCRIPTS[] = "/etc/skel/.config/Fastec/scripts";

//! Our scripts folder. It's assumed you'll prepend the home folder path.
const char CAM_SCRIPTS[] = ".config/Fastec/scripts";

//! Command to create a single MD5SUM for a folder. Assumes usage of the "fmt"
//! libary. The arguments are:
//! {0}: What to call the folder checksum value, e.g. "MD5_HOME".
//! {1}: The path to the folder to checksum, e.g. SYS_SCRIPTS.
const char MD5FOLDER_CMD[] = "{0}=$(find {1} -type f -exec md5sum {{}} + | awk '{{print $1}}' | sort | md5sum | awk '{{print $1}}')";

//! The default URL for the Fastec update server. This may be overwritten
//! by a settings value.
const char UPDATE_SERVER_URL[] = "http://upd.fastecimaging.com";

//! Location of the seriesNC camera update package. Assumes usage of the "fmt"
//! libary. The "{0}" reference the first argument, the version.
const char CAM_NC_PP_UPDATE[] = "/linux/cameraNC/rev{0}/pp_update_{0}.tar.xz";

//! Our scratch folder is where we'll download stuff. It's assumed you'll prepend
//! the home folder path.
const char CAM_SCRATCH[] = ".config/Fastec/scratch";

//! Command to extract the pp_update files from the update tarball.
//! arg0: the device path [see dev_path].
//! arg1: the package name [see package].
const char EXTRACT_FROM_TARBALL[] = "tar xf {0}/{1} -C /tmp --wildcards --no-anchored 'pp_update_*'";

//! Our pacman options file path.
//! arg0: The home path [see get_home_path()].
//! arg1: The path to scratch [see CAM_SCRATCH].
const char PACOPT_FILE[] = "{0}/{1}/fi_update/pacopts.txt";

//! Command to extract the update package to scratch.
//! arg0: The version information [see version].
//! arg1: The home path [see get_home_path()].
//! arg2: The path to scratch [see CAM_SCRATCH].
const char PP_EXTRACT[] = "tar xf /tmp/pp_update_{0}.tar.xz -C {1}/{2}";

//! File containing the pp_update urls to fetch.
const char PP_URLS_FILE[] = "/tmp/pp_urls.txt";

//! GPG command to validate our update package.
//! arg0: the version information [see version].
const char PP_VALIDATE[] = "/usr/bin/gpg --verify-files /tmp/pp_update_{0}.tar.xz.sig &> /dev/null";

//! The pre-update entry point.
//! arg0: The home path [see get_home_path()].
//! arg1: The path to scratch [see CAM_SCRATCH].
const char PRE_UPDATE_ENTRY[] = "{0}/{1}/fi_update/pre/fi_update.sh";

//! The post-update entry point.
//! arg0: The home path [see get_home_path()].
//! arg1: The path to scratch [see CAM_SCRATCH].
const char POST_UPDATE_ENTRY[] = "{0}/{1}/fi_update/post/fi_update.sh";


//! The states of the update object.
enum UP_STATE
{
    US_SETUP,       //!< Check for pre-/post-update package.
    US_PRE,     //!< Run the pre-update package (if it exists).
    US_UPDATE,      //!< Run the update (network or from package).
    US_DONE     //!< All done.
};

/*!
 * @brief The camera server update object. This object is
 *    a singleton and is "local" to this file.
 */
class csUpdate : public FAS_THREAD::Thread
{
public:
    virtual ~csUpdate();
    virtual void run();

    static csUpdate *factory();

    void getOptions(std::string& opt_str) const { opt_str = pacman_options; }

    bool is_running()   { return thread_running; }

    void setParams(bool use_network, const std::string& update_version, const std::string& device_path, const std::string& package_name);
    void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr);

private:
    csUpdate();

    int do_setup();
    int do_precheck();
    int do_update();

    void read_options();

    volatile bool thread_running;   //!< Flag to indicate if we are running or not.
    int state;                      //!< Current upate state.

    bool network;
    std::string version;
    std::string dev_path;
    std::string package;

    std::string pacman_options;

    //! The one and only.
    static std::unique_ptr<csUpdate> updater;
};

} /* End of namespace CS_UPDATE */

using namespace CS_TRACE;
using namespace CS_UPDATE;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<csUpdate> csUpdate::updater;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

//! @name Public facing functions.
//! @{
void update_camera(bool use_network, const std::string& update_version, const std::string& device_path, const std::string& package_name)
{
    LOG("%s", __PRETTY_FUNCTION__);

    if (update_in_progress())
    {
        fi_log(LOG_DEBUG, "%s: Already in progress!", __PRETTY_FUNCTION__);
        return;
    }

    // Create our object. This will get things rolling.
    if (csUpdate::factory() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate the updater object!", __PRETTY_FUNCTION__);
        return;
    }

    // Copy over the parameters.
    csUpdate::factory()->setParams(use_network, update_version, device_path, package_name);

    // Tell the updater to start.
    csUpdate::factory()->start();
}

bool update_in_progress()
{
    // Can only be true if there is an updater. If there is an updater,
    // then the assumption is that we are updating.
    return csUpdate::factory()->is_running();
}

void update_options(std::string& opt_str)
{
    LOG("%s", __PRETTY_FUNCTION__);

    if (csUpdate::factory() != nullptr)
        csUpdate::factory()->getOptions(opt_str);
}

void update_postops()
{
    std::string update_str;
    struct stat up_stat_buf;

    LOG("%s", __PRETTY_FUNCTION__);

    // Build up our post-update command string.
    update_str = fmt::format(POST_UPDATE_ENTRY, get_home_path(), CAM_SCRATCH);
    LOG("%s: [%s]", __PRETTY_FUNCTION__, update_str.c_str());

    // Does this exist?
    if (stat(update_str.c_str(), &up_stat_buf) == 0)
    {
        int up_status;

        // Invoke the post-update command.
        up_status = system(update_str.c_str());
        if ((up_status < 0) || !WIFEXITED(up_status))
        {
            // There's a problem, log the error.
            fi_log(LOG_ERR, "%s: Cannot run the post-update [%d/%d]!", __PRETTY_FUNCTION__,
                   errno, WEXITSTATUS(up_status));
        }
        else
        {
            up_status = WEXITSTATUS(up_status);
            fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, up_status, strerror(up_status));
        }
    }

    // Entry point doesn't exist.
    LOG("%s: Checking for script updates", __PRETTY_FUNCTION__);
    {
        int sys_rc;
        std::string cmd_line;
        std::string sh_line;
        std::ofstream check_file;

        // Build our update script.
        check_file.open(CHECK_SCRIPTS_FILE, std::ios::out | std::ios::trunc);
        if (!check_file.is_open())
        {
            fi_log(LOG_ERR, "%s: Unable to open '%s' [%d: %s]", __PRETTY_FUNCTION__, PP_URLS_FILE, errno, strerror(errno));

            // Make sure we close the object.
            check_file.close();
            return;
        }

        // Build up our script.
        check_file << "#!/bin/bash" << std::endl;
//        check_file << "set -x" << std::endl;

        // MD5 sum the system folder. Make sure this agrees with the body
        // of the script below.
        sh_line = fmt::format(MD5FOLDER_CMD, "MD5_SKEL", SYS_SCRIPTS);
        check_file << sh_line << std::endl;

        // MD5 sum our home folder. Make sure this agrees with the body of
        // the script below.
        cmd_line = fmt::format("{0}/{1}", get_home_path(), CAM_SCRIPTS);
        sh_line = fmt::format(MD5FOLDER_CMD, "MD5_HOME", cmd_line);
        check_file << sh_line << std::endl;

        // Now output the rest of the script.
        check_file << "if [ \"$MD5_SKEL\" != \"$MD5_HOME\" ]; then" << std::endl;
        check_file << fmt::format("  cp -a /etc/skel/.config/Fastec/scripts {0}/.config/Fastec/.", get_home_path()) << std::endl;
        check_file << fmt::format("  chown -R fascam:fascam {0}/.config/Fastec", get_home_path()) << std::endl;
        check_file << "fi" << std::endl;

        // Close the check file.
        check_file.close();

        // Build up the command line.
        cmd_line = fmt::format("/usr/bin/chmod a+x {0}; systemd-run {0}", CHECK_SCRIPTS_FILE);
        sys_rc = system(cmd_line.c_str());
        if ((sys_rc < 0) || (WEXITSTATUS(sys_rc) != 0))
        {
            fi_log(LOG_ERR, "%s: Error running check_script.sh [%d]!", __PRETTY_FUNCTION__, WEXITSTATUS(sys_rc));
        }
    }
}

//! @}
// End of public facing functions.

//============================================================================
// csUpdate class implementation.
//============================================================================

/*!
 * @brief The update thread constructor. The update thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
csUpdate::csUpdate() :
    FAS_THREAD::Thread(true),
    thread_running(false),
    state(US_SETUP),
    network(true)
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    pacman_options.erase();
}

csUpdate::~csUpdate()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief The singleton creation method.
 *
 * Only this factory method can create the singleton object, and its
 * purpose is to ensure there is only one update control object.
 */
csUpdate *csUpdate::factory()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // We only allow one thread object.
    if (updater.get() == nullptr)
    {
        // Create our thread.
        updater = std::unique_ptr<csUpdate>(new csUpdate());
    }

    // Return the result.
    return updater.get();
}

/*!
 * @brief Set the parameters to be used by the update object.
 *
 * @param use_network    True if update is using network, false otherwise.
 * @param update_version Version to update to.
 * @param device_path    [OPTIONAL] The path to the device with the update package.
 * @param package_name   [OPTIONAL] The name of the update package.
 */
void csUpdate::setParams(bool use_network, const std::string& update_version, const std::string& device_path, const std::string& package_name)
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Set the parameters.
    network = use_network;
    version = update_version;
    dev_path = device_path;
    package = package_name;
}

/*!
 * @brief Initiate the update process.
 */
void csUpdate::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Now start the thread.
    FAS_THREAD::Thread::start(threadFn);

    // Transition to the setup state.
    state = US_SETUP;
}

/*!
 * @brief Setup for the update process. What we do here depends on
 *    if a network update is being done or not.
 */
int csUpdate::do_setup()
{
    int new_state;
    int tar_status;
    int sys_status;
    std::string tar_cmd;
    std::string sys_cmd;

    LOG("%s", __PRETTY_FUNCTION__);

    // Assume no pp_update package.
    new_state = US_UPDATE;

    // Are we doing a network based update?
    if (network)
    {
        int wget_status;
        std::string pp_url;
        std::string wget_cmd;
        std::ofstream url_file;
        std::string update_url;

        // Open our URLs file.
        url_file.open(PP_URLS_FILE, std::ios::out | std::ios::trunc);
        if (!url_file.is_open())
        {
            fi_log(LOG_ERR, "%s: Unable to open '%s' [%d: %s]", __PRETTY_FUNCTION__, PP_URLS_FILE, errno, strerror(errno));

            // Make sure we close the object.
            url_file.close();
            goto setup_exit;
        }

        // Query the update server and check for an update package.
        if (!override_string(OVR_UPDATE_SERVER, update_url))
            update_url = std::string(UPDATE_SERVER_URL);
        pp_url = update_url + fmt::format(CAM_NC_PP_UPDATE, version);
        url_file << pp_url << std::endl;

        // Append the signature file extension.
        url_file << pp_url << ".sig" << std::endl;

        // Close the file.
        url_file.close();

        // Build up the command.
        wget_cmd = std::string("/usr/bin/wget -q --directory-prefix=/tmp -i ") + PP_URLS_FILE;
        LOG("%s: wget_cmd [%s]", __PRETTY_FUNCTION__, wget_cmd.c_str());

        wget_status = system(wget_cmd.c_str());
        if ((wget_status < 0) || !WIFEXITED(wget_status))
        {
            // problem log the error
            fi_log(LOG_ERR, "%s: Cannot get the pp_update [%d/%d]!", __PRETTY_FUNCTION__,
                   errno, WEXITSTATUS(wget_status));
            goto setup_exit;
        }
        else
        {
            wget_status = WEXITSTATUS(wget_status);
            fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, wget_status, strerror(wget_status));
            if (wget_status)
            {
                // The wget failed. Assume no pre-update package.
                goto setup_exit;
            }
        }
        LOG("%s: Files retrieved OK", __PRETTY_FUNCTION__);
    }
    else
    {
        // We are doing a local package update. See if there is a
        // pp_update package in the tarball.
        tar_cmd = fmt::format(EXTRACT_FROM_TARBALL, dev_path, package);
        LOG("%s: tar_cmd [%s]", __PRETTY_FUNCTION__, tar_cmd.c_str());

        tar_status = system(tar_cmd.c_str());
        if ((tar_status < 0) || !WIFEXITED(tar_status))
        {
            // problem log the error
            fi_log(LOG_ERR, "%s: Cannot get the pp_update [%d/%d]!", __PRETTY_FUNCTION__,
                   errno, WEXITSTATUS(tar_status));
            goto setup_exit;
        }
        else
        {
            tar_status = WEXITSTATUS(tar_status);
            fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, tar_status, strerror(tar_status));
            if (tar_status)
            {
                // The tar command failed. Assume no pre-update package.
                goto setup_exit;
            }
        }
        LOG("%s: Files extracted OK", __PRETTY_FUNCTION__);
    }

    // At this point, we've put the pp_update package and signature into "/tmp".
    // Validate the update package.
    sys_cmd = fmt::format(PP_VALIDATE, version);
    LOG("%s: validation [%s]", __PRETTY_FUNCTION__, sys_cmd.c_str());
    sys_status = system(sys_cmd.c_str());
    if ((sys_status < 0) || !WIFEXITED(sys_status))
    {
        // problem log the error
        fi_log(LOG_ERR, "%s: Cannot validate the pp_update [%d/%d]!", __PRETTY_FUNCTION__,
               errno, WEXITSTATUS(sys_status));
        goto setup_exit;
    }
    else
    {
        sys_status = WEXITSTATUS(sys_status);
        fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, sys_status, strerror(sys_status));
        if (sys_status)
        {
            // The system command failed. Assume no valid pre-update package.
            goto setup_exit;
        }
    }
    LOG("%s: all is validated OK", __PRETTY_FUNCTION__);

    // Extract the files to scratch.
    tar_cmd = fmt::format(PP_EXTRACT, version, get_home_path(), CAM_SCRATCH);
    LOG("%s: tar_cmd [%s]", __PRETTY_FUNCTION__, tar_cmd.c_str());
    tar_status = system(tar_cmd.c_str());
    if ((tar_status < 0) || !WIFEXITED(tar_status))
    {
        // problem log the error
        fi_log(LOG_ERR, "%s: Cannot extract the pp_update [%d/%d]!", __PRETTY_FUNCTION__,
               errno, WEXITSTATUS(tar_status));
        goto setup_exit;
    }
    else
    {
        tar_status = WEXITSTATUS(tar_status);
        fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, tar_status, strerror(tar_status));
        if (tar_status)
        {
            // The tar command failed. Assume no pre-update package.
            goto setup_exit;
        }
    }
    LOG("%s: Files extracted OK", __PRETTY_FUNCTION__);

    // Ready to look for a pre-update package.
    new_state = US_PRE;

setup_exit:
    // All done. Return the state we should transition to.
    LOG("%s: All done. new_state=%d", __PRETTY_FUNCTION__, new_state);
    return new_state;
}

/*!
 * @brief Check to see if there is a pre-update package available. If
 *    we have one, then run the pre-updater.
 */
int csUpdate::do_precheck()
{
    int new_state(US_DONE);
    int up_status;
    std::string update_str;
    struct stat up_stat_buf;

    LOG("%s", __PRETTY_FUNCTION__);

    // Build up our pre-update command string.
    update_str = fmt::format(PRE_UPDATE_ENTRY, get_home_path(), CAM_SCRATCH);
    LOG("%s: [%s]", __PRETTY_FUNCTION__, update_str.c_str());

    // Does this exist?
    if (stat(update_str.c_str(), &up_stat_buf) != 0)
    {
        // Entry point doesn't exist.
        LOG("%s: no pre-update entry", __PRETTY_FUNCTION__);
        new_state = US_UPDATE;
        goto pre_done;
    }

    // Invoke the pre-update command.
    up_status = system(update_str.c_str());
    if ((up_status < 0) || !WIFEXITED(up_status))
    {
        // problem log the error
        fi_log(LOG_ERR, "%s: Cannot run the pre-update [%d/%d]!", __PRETTY_FUNCTION__,
               errno, WEXITSTATUS(up_status));
        goto pre_done;
    }
    else
    {
        up_status = WEXITSTATUS(up_status);
        fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, up_status, strerror(up_status));
        if (up_status)
        {
            // The up command failed. Assume no pre-update package.
            goto pre_done;
        }
    }

    // Everything is good. Move on.
    new_state = US_UPDATE;

pre_done:
    // All done. Return the next state.
    LOG("%s: done, new_state=%d", __PRETTY_FUNCTION__, new_state);
    return new_state;
}

/*!
 * @brief Perform the update. Which update is done depends on if we're
 *    doing a network update or not.
 */
int csUpdate::do_update()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Get any options the update package might include.
    read_options();

    // Are we updating from the network?
    if (!network)
    {
        int sys_rc = 0;
        std::string update_cmd;

        // Build up the update command. The package we get is a tarball of the
        // update "repository".
        // 1. Extract the tarball.
        // 2. Issue updates based on the package found.
        update_cmd = fmt::format("/usr/bin/tar xf {0}/{1} -C /tmp", dev_path, package);
        LOG("%s: Extract the tarball [%s]", __PRETTY_FUNCTION__, update_cmd.c_str());
        sys_rc = system(update_cmd.c_str());
        if ((sys_rc < 0) || (WEXITSTATUS(sys_rc) != 0))
        {
            fi_log(LOG_ERR, "%s: Unable to extract the update tarball", __PRETTY_FUNCTION__);
            goto update_exit;
        }

        // Copy all the tarballs to the cache.
        update_cmd = std::string("/usr/bin/cp -n /tmp/fastec/*.pkg.tar.xz /var/cache/pacman/pkg/.");
        LOG("%s: Copy the packages to /var/cache/pacman/pkg", __PRETTY_FUNCTION__);
        sys_rc = system(update_cmd.c_str());
        if ((sys_rc < 0) || (WEXITSTATUS(sys_rc) != 0))
        {
            fi_log(LOG_ERR, "%s: Unable to copy the packages to /var/cache/pacman/pkg", __PRETTY_FUNCTION__);
            goto update_exit;
        }
    }

    // Build up our work message to have the main server thread perform the actual update.
    CS_NETWORK::CS_NETWORK_MESSAGE msg;
    msg.msg_type = CS_NETWORK::MSG_TYPE_INTERNAL;
    msg.internal.flags = network;
    msg.internal.type = CS_NETWORK::CS_INTERNAL_UPDATE;
    CameraServer::add_work(&msg);

update_exit:
    // All done.
    return US_DONE;
}

/*!
 * @brief Read in the pacopts.txt file, if it exists.
 */
void csUpdate::read_options()
{
    std::string file_path;
    std::ifstream opt_file;
    std::ifstream::pos_type file_size;

    // Clear the exist string.
    pacman_options.erase();

    // Build up the file path.
    file_path = fmt::format(PACOPT_FILE, get_home_path(), CAM_SCRATCH);

    // See if we can open the file.
    opt_file.open(file_path, std::ios::in | std::ios::binary | std::ios::ate);
    if (opt_file.is_open())
    {
        file_size = opt_file.tellg();
        opt_file.seekg(0, std::ios::beg);

        std::vector<char> file_bytes(file_size);
        opt_file.read(file_bytes.data(), file_size);

        pacman_options = std::string(file_bytes.data(), file_size);
    }
    opt_file.close();

    LOG("%s: pacman_options=[%s]", __PRETTY_FUNCTION__, pacman_options.c_str());
}

/*!
 * @brief Main function to handle update events:
 *    setup, pre-update, update.
 */
void csUpdate::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We are now running.
    thread_running = true;

    // Enter our state machine.
    while (state != US_DONE)
    {
        LOG("%s: state=%d", __PRETTY_FUNCTION__, state);

        switch (state)
        {
        case US_SETUP:
            state = do_setup();
            break;
        case US_PRE:
            state = do_precheck();
            break;
        case US_UPDATE:
            state = do_update();
            break;
        default:
        case US_DONE:
            break;
        }
    }

    // We are no longer running.
    thread_running = false;
}
