/*!
 * @file cs_utils.cxx
 * @brief These are top-level operations within the camera server.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <stdarg.h>
#include <pwd.h>

#include <sys/time.h>
#include <sys/syscall.h>
#include <sys/types.h>

#include <chrono>

#include "cs_utils.h"
#include "camera_server.h"
#include "sensor.h"
#include "part_mgr.h"
#include "pb_control.h"
#include "fcp.h"
#include "fpga.h"
#include "image_correction.h"
#include "image_io.h"
#include "lince5m.h"

#include "override.h"

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#define FPGA_PEAK_RETRY     10

const uint64_t NSEC_PER_SEC = 1000000000;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Return the fully qualified path to the home folder.
 *
 * I'm not too worried about threading here, because there's only
 * going to be one user, so that data itself will be static.
 */
const char *get_home_path()
{
    const char *home_path;

    // Get the HOME environment variable.
    home_path = getenv("HOME");
    if (home_path == nullptr)
    {
        // There's no environment variable, let's check the password file.
        struct passwd *pw = getpwuid(getuid());
        if (pw != nullptr)
        {
            // Return the path from the passwd file.
            home_path = pw->pw_dir;
        }
    }

    // Return what we've found.
    return home_path;
}

/*!
 * @brief Measure the duration from start to end.
 *
 * Call twice, once with start=1 and then again with start=0.
 * If start==1, it writes the current time to *tv1_ptr.
 * If start==0, it computes the difference and returns the elapsed time.
 *
 * @param   start   whether this is the first or second call
 * @param   tv1_ptr if start==1, where to write the current time
 *          if start==0, the time from the previous call
 *
 * @return the elapsed time in microseconds.
 */
uint64_t measuretime(int start, struct timespec *tv1_ptr)
{
    struct timespec tv2;

    clock_gettime(CLOCK_MONOTONIC, start ? tv1_ptr : &tv2);
    if (!start)
    {
        uint64_t t1, t2;
        t1 = (uint64_t)tv1_ptr->tv_sec * 1000000 + (uint64_t)(tv1_ptr->tv_nsec / 1000);
        t2 = (uint64_t)tv2.tv_sec * 1000000 + (uint64_t)(tv2.tv_nsec / 1000);
        return (t2 - t1);
    }
    return 0;
}

/*!
 * @brief Convert straight-binary-seconds to a struct tm.
 */
void sbs_to_tm(uint32_t sbs, struct tm *tm_ptr)
{
    int year;
    int isLeap;
    time_t timeNow;
    struct tm temp_tm;

    static const int mon_yday[2][13] =
    {
        // Normal years
        { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
        // Leap years
        { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
    };

    // Clear out the structure.
    memset(&temp_tm, 0, sizeof(struct tm));

    // Initialize unknown elements of our structure by using the current time.
    timeNow = time(NULL);
    memcpy(&temp_tm, localtime(&timeNow), sizeof(struct tm));

    // Get years.
    temp_tm.tm_year = sbs / (365 * 24 * 60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_year * (365 * 24 * 60 * 60));
    temp_tm.tm_yday = sbs / (24 * 60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_yday * (24 * 60 * 60));
    temp_tm.tm_hour = sbs / (60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_hour * (60 * 60));
    temp_tm.tm_min = sbs / 60;
    temp_tm.tm_sec = (int)sbs - (temp_tm.tm_min * 60);

    // We may need to add 100 to the years, because we're older than 2000.
    if (temp_tm.tm_year < 100)
        temp_tm.tm_year += 100;

    // Are we in a leap year?
    year = temp_tm.tm_year + 1900;

    // Divisible by 400?
    if ((year % 400) == 0)
        isLeap = 1;

    // Divisible by 100?
    else if ((year % 100) == 0)
        isLeap = 0;

    // Divisible by 4?
    else if ((year % 4) == 0)
        isLeap = 1;

    // Others are not.
    else
        isLeap = 0;

    int dayOfYear = temp_tm.tm_yday;
    for (temp_tm.tm_mon = 0; temp_tm.tm_mon < 12; ++temp_tm.tm_mon)
    {
        if (dayOfYear <= mon_yday[isLeap][temp_tm.tm_mon])
        {
            // Back up our month.
            temp_tm.tm_mon -= 1;
            if (temp_tm.tm_mon < 0)
                temp_tm.tm_mon = 0;

            // Figure out day of the month.
            temp_tm.tm_mday = (dayOfYear - mon_yday[isLeap][temp_tm.tm_mon]) + 1;
            break;
        }
    }

    // Sanity check.
    if (tm_ptr != nullptr)
    {
        // Copy the structure over before mktime goes messing around with it.
        memcpy(tm_ptr, &temp_tm, sizeof(struct tm));

        // Now see if DST is on.
        (void)mktime(&temp_tm);

        // Update the return values.
        tm_ptr->tm_isdst = temp_tm.tm_isdst;

        // Update the offset.
        tm_ptr->tm_gmtoff = temp_tm.tm_gmtoff;
    }
}

/*!
 * @brief Adds nanoseconds to a timespec.
 *
 * We need something like this for any thread that does a condition wait
 * so we might as well leave it here with other thread stuff.
 *
 * Taken from time.h.
 */
void timespec_add_ns(struct timespec *a, uint64_t ns)
{
    uint64_t    temp_ns;

    // Add in the nano-seconds.
    temp_ns = ns + (uint64_t)(a->tv_nsec);
    while (temp_ns >= NSEC_PER_SEC)
    {
        // We slopped over, so adjust.
        temp_ns -= NSEC_PER_SEC;
        a->tv_sec++;
    }

    // Now, update our structure's field.
    a->tv_nsec = temp_ns;
}

/*!
 * @brief Get current time on the monotonic clock
 *
 * @return the current time in microseconds.
 */
uint64_t get_current_usecs()
{
    struct timespec tv;
    uint64_t t;

    clock_gettime(CLOCK_MONOTONIC, &tv);

    t = (uint64_t)tv.tv_sec * 1000000 + (uint64_t)(tv.tv_nsec / 1000);

    return t;
}

uint32_t get_tid()
{
    uint32_t    tid;
    tid = (uint32_t)syscall(SYS_gettid);

    return tid;
}

/*!
 * @brief  Calculate the size of a frame.
 *
 * The value returned is in bytes, but each row is rounded up into
 * a round number of words based upon the bit depth.
 *
 * @param width     Row width in bytes.
 * @param height    Number of rows.
 * @param bit_mode  The bit depth 8 or 10 bits, see gige_msg.h for definitions.
 *
 * @return The buffer size or 0 if an error.
 */
uint32_t calc_src_dma_size(uint32_t row_width, uint32_t num_rows, bitmode_t bit_mode)
{
    uint32_t value;

    if (IS_10_BIT_MODE(bit_mode))
    {
        // For 10 bits, 3 pixels fit in a 32 bit word with two bits left over.
        // Calculate how many bytes this be will rounded up to the next word on a row.

        value = row_width; // figure out how many words
        if (value % 3)
            value = (value / 3) + 1;
        else
            value = value / 3;
        value = (value * 4); // now make it bytes
        value = value * num_rows; // now figure out the frame size
    }
    else if (IS_12_BIT_MODE(bit_mode))
    {
        // For 12-bit mode, we have 5 pixels in each 64-bit double word with 4 bits left over.
        value = row_width;      // Figure out how many words.
        if (value % 5)
            value = (value / 5) + 1;
        else
            value = value / 5;
        value = (value * 8);        // Now make it bytes.
        value = value * num_rows;   // Now compute the frame size.
    }
    else
    {
        // If the bit width is 8 bits, then 4 pixels fit in a 32 bit word.
        // Calculate the number of words per row.
        value = row_width;
        if ((value % 4) == 0)
            value = value / 4;
        else
            value = (value / 4) + 1;
        value = value * 4;  // put it back in bytes
        value = value * num_rows; // now calculate number per frame
    }

    return value;
}

/*!
 * @brief Compute size of an IPM-processed frame
 *
 * The imgproc_type argument used to set up IPM processing determines the
 * size of the processed image out of the IPM. (If the type is JPEG we
 * compute the size of the RGB that is used as input to the JPEG engine.)
 *
 * @param stream_type       The stream type the user wants.
 * @param sensor_is_color   true if the capture is color, false if it is monochrome
 *
 * @return size of image if computable or zero for JPEG.
 */
uint32_t get_processed_size(uint32_t imgproc_type, uint32_t width, uint32_t height)
{
    uint32_t processed_size = 0;
    bool     row_wd_align;

//    fi_log(LOG_DEBUG, "%s: type=0x%x, width=%u, height=%u", __PRETTY_FUNCTION__, imgproc_type, width, height);

    row_wd_align = (imgproc_type & IMGPROC_ROW_WD_ALIGN) ? true : false;

    switch (imgproc_type & IMGPROC_TYPE_MASK)
    {
    case IMGPROC_TYPE_MONO8:
    case IMGPROC_TYPE_Y8:
        // Calculate the frame size. Make sure the rows are 32-bit aligned if required.
        processed_size = width;
        if (row_wd_align && ((processed_size & 3) != 0))
            processed_size = ((processed_size + 3) / 4) * 4;
        processed_size *= height;
        break;
    case IMGPROC_TYPE_NV12:
        // This is the Y-plane size plus the UV-plane size. The FPGA requires the UV plane start
        // on a HEXWORD (64 byte) boundary in image memory.
        //
        // Reference FiFPGA::start_p1_frame() and the setting of IPM_P1_FRAME_REQ_IC_Y_BUF_START
        // and IPM_P1_FRAME_REQ_IC_C_BUF_START.
        processed_size = (convert_to_HEXWORD(width * height) << HEXWORD_SHIFT) + ((width * height) >> 1);
        break;
    case IMGPROC_TYPE_JPEG:
    case IMGPROC_TYPE_RGB24:
        // Calculate the frame size. Make sure the rows are 32-bit aligned if required.
        processed_size = width * 3;
        if (row_wd_align && ((processed_size & 3) != 0))
            processed_size = ((processed_size + 3) / 4) * 4;
        processed_size *= height;
        break;
    case IMGPROC_TYPE_RAW8:
        processed_size = width * height;
        break;
    case IMGPROC_TYPE_RAW16:
        processed_size = width * height * 2;
        break;
    default:
        fi_log(LOG_ERR, "%s: Invalid image type (0x%x)!", __FUNCTION__, imgproc_type);
        break;
    }

//    fi_log(LOG_DEBUG, "%s: processed_size=0x%x/%d", __PRETTY_FUNCTION__, processed_size, processed_size);
    return processed_size;
}

/*!
 * @brief Determine low-level hardware details from stream format.
 *
 * Stream format is a file format, from which we determine how the FPGA should
 * process the image. This function sets up the IPM processing that goes from
 * a captured image to an IPM-processed buffer. Then once that is done, there
 * can be further alignment-forcing or byte swapping as the data is transferred
 * from the buffer holding the IPM/JPEG output.
 *
 * @param stream_type       The stream type the user wants.
 * @param bit_mode      The bit mode of the capture
 * @param sensor_is_color   true if the capture is color, false if it is monochrome
 *
 * @return IMGPROC_TYPE_NONE for failure, other IMGPROC_TYPE_... for success
 */
uint32_t get_imgproc_type(uint32_t stream_type, bitmode_t bit_mode, bool sensor_is_color)
{
    uint32_t    image_type = IMGPROC_TYPE_NONE;

    PDEBUG(PDEBUG_TRACE, "%s: type:0x%x", __FUNCTION__, stream_type);

    // Initialize our control variables.
    // Figure out image type from the following information.
    switch (stream_type & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_JPEG:
        image_type = IMGPROC_TYPE_JPEG;
        break;

    case GIGE_STREAM_FORMAT_AVI:
    case GIGE_STREAM_FORMAT_MP4:
        switch (imgSaveWorker::getInstance()->getVideoCodec())
        {
        case GIGE_STREAM_FORMAT_JPEG:
            // Saving AVI/JPEGs.
            image_type = IMGPROC_TYPE_JPEG;
            break;

        case GIGE_STREAM_FORMAT_BMP:
            // We are saving AVI/BMPs, now we need to determine if
            // we're mono or color.
            if (sensor_is_color)
            {
                // Color sensor means color processing.
                image_type = IMGPROC_TYPE_RGB24;
            }
            else
            {
                // Mono sensor means mono processing.
                image_type = IMGPROC_TYPE_MONO8;
            }
            break;
        case GIGE_STREAM_FORMAT_PLANAR:
            // Planar format does not support RAW images, so the
            // choice is limited to color vs. mono.
            if (sensor_is_color)
            {
                // Color sensor means color processing.
                image_type = IMGPROC_TYPE_NV12;
            }
            else
            {
                // Mono sensor means mono processing.
                image_type = IMGPROC_TYPE_Y8;
            }
            break;
        default:
            // Whatever this was, it isn't legal.
            fi_log(LOG_ERR, "%s: Invalid video codec type (0x%x) for stream (0x%x)!", __PRETTY_FUNCTION__, imgSaveWorker::getInstance()->getVideoCodec(), stream_type);
            break;
        }
        break;

    case GIGE_STREAM_FORMAT_BMP:
        // BMP format does not support RAW images, so the
        // choice is limited to color vs. mono.
        if (sensor_is_color)
        {
            // Color sensor means color processing.
            image_type = IMGPROC_TYPE_RGB24;
        }
        else
        {
            // Mono sensor means mono processing.
            image_type = IMGPROC_TYPE_MONO8;
        }
        break;

    case GIGE_STREAM_FORMAT_PACK:       // TIF for now, but maybe the hardware will support packed later.
        // Did we capture 8 or 10/12 bits?
        if (IS_10_BIT_MODE(bit_mode) || IS_12_BIT_MODE(bit_mode))
        {
            // 10-bit or 12-bit captures mean 16 bit raw.
            image_type = IMGPROC_TYPE_RAW16;
        }
        else
        {
            // 8 bit captures mean 8 bit raw.
            image_type = IMGPROC_TYPE_RAW8;
        }
        break;

    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
    case GIGE_STREAM_FORMAT_DNG:
        // Are we raw? DNG is always raw, TIF may be or not.
        if (((stream_type & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_DNG) ||
            ((stream_type & GIGE_STREAM_FORMAT_RAW) == ((uint32_t)GIGE_STREAM_FORMAT_RAW)))
        {
            // Raw color images. Did we capture 8 or 10/12 bits?
            if (IS_10_BIT_MODE(bit_mode) || IS_12_BIT_MODE(bit_mode))
            {
                // 10-bit or 12-bit captures mean 16 bit raw.
                image_type = IMGPROC_TYPE_RAW16;
            }
            else
            {
                // 8 bit captures mean 8 bit raw.
                image_type = IMGPROC_TYPE_RAW8;
            }
        }
        else if (sensor_is_color)
        {
            // Color sensor processed images means color processing.
            image_type = IMGPROC_TYPE_RGB24;
        }
        else
        {
            // Mono sensor processed images means mono processing.
            // We know this is going to be mono8 because we can't
            // do mono-16 (our only other option with 10+ bits).
            image_type = IMGPROC_TYPE_MONO8;
        }
        break;
    case GIGE_STREAM_FORMAT_CAPTURE:
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        // Planar format does not support RAW images, so the
        // choice is limited to color vs. mono.
        if (sensor_is_color)
        {
            // Color sensor means color processing.
            image_type = IMGPROC_TYPE_NV12;
        }
        else
        {
            // Mono sensor means mono processing.
            image_type = IMGPROC_TYPE_Y8;
        }
        break;

    default:
        // Whatever this was, it isn't legal.
        fi_log(LOG_ERR, "%s: Invalid stream type (0x%x)!", __PRETTY_FUNCTION__, stream_type);
        break;
    }

    return image_type;
}

/*!
 * @brief Determine low-level hardware details from stream format.
 *
 * Stream format is a file format, from which we determine how the FPGA should
 * process the image. Once we set up the IPM processing that goes from
 * a captured image to an IPM-processed buffer, we may need to do further
 * alignment-forcing or byte swapping as the data is transferred from the
 * buffer holding the IPM output to processor memory. This sets up the
 * DMA modifier that controls any processing done during that DMA step.
 *
 * @param stream_type       The stream type the user wants.
 * @param bit_mode      The bit mode of the capture
 * @param sensor_is_color   true if the capture is color, false if it is monochrome
 *
 * @return DMA modifier suitable for passing to the FPGA's DMA driver.
 */
uint32_t get_stream_dma_mod(uint32_t stream_type, bitmode_t bit_mode, bool sensor_is_color)
{
    bool    do_raw;
    uint32_t    dma_mod = 0;

    fi_log(LOG_DEBUG, "%s: type=0x%x", __PRETTY_FUNCTION__, stream_type);

    // Initialize our DMA modifier.
    dma_mod = IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;

    // Figure out what format we're creating.
    do_raw = ((stream_type & GIGE_STREAM_FORMAT_RAW) == (uint32_t)GIGE_STREAM_FORMAT_RAW);

    switch (stream_type & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_DNG:
    case GIGE_STREAM_FORMAT_PACK:
        do_raw = true;
        break;

    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
        if (!do_raw)
        {
            // Setup for processed images.
            if (sensor_is_color)
                dma_mod |= IM2PCIE_DMA_RGB_FORMAT | IM2PCIE_DMA_IMAGE_DATA;
            else
                dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_IMAGE_DATA;
        }
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
    case GIGE_STREAM_FORMAT_JPEG:
        // Color JPEGs go through the IPMs Image Converter DMA. This produces
        // the NV12 format for us, and it is packed appropriately. The monochrome
        // images go through the IPM producing Y8 data which is row word aligned.
        // We need to pack these, so use the IMAGE DMA, not the GENERIC DMA for
        // the Y8 data.
        if (sensor_is_color)
            dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_GENERIC_DATA;
        else
            dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_IMAGE_DATA;
        break;
    case GIGE_STREAM_FORMAT_AVI:
    case GIGE_STREAM_FORMAT_MP4:
        switch (imgSaveWorker::getInstance()->getVideoCodec())
        {
        case GIGE_STREAM_FORMAT_PLANAR:
        case GIGE_STREAM_FORMAT_JPEG:
            // Saving AVI/JPEGs. See above comment for JPEGs. Applies here too.
            if (sensor_is_color)
                dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_GENERIC_DATA;
            else
                dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_IMAGE_DATA;
            break;
        case GIGE_STREAM_FORMAT_BMP:
            // We are saving AVI/BMPs, now we need to determine if
            // we're mono or color.
            if (sensor_is_color)
                dma_mod |= IM2PCIE_DMA_BGR_FORMAT | IM2PCIE_DMA_IMAGE_DATA;
            else
                dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_IMAGE_DATA;
            dma_mod |= IM2PCIE_DMA_WORD_ALIGN;
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected video codec: 0x%08x", __PRETTY_FUNCTION__, imgSaveWorker::getInstance()->getVideoCodec());
            break;
        }
        break;
    case GIGE_STREAM_FORMAT_BMP:
        if (sensor_is_color)
            dma_mod |= IM2PCIE_DMA_BGR_FORMAT | IM2PCIE_DMA_IMAGE_DATA;
        else
            dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_IMAGE_DATA;
        dma_mod |= IM2PCIE_DMA_WORD_ALIGN;
        break;
    default:
        break;
    }

    // All raw processing is handled the same.
    if (do_raw)
    {
        // Setup for raw processing.
        dma_mod |= IM2PCIE_DMA_IMAGE_DATA;
        if (IS_10_BIT_MODE(bit_mode) || IS_12_BIT_MODE(bit_mode))
            dma_mod |= IM2PCIE_DMA_16BIT_RAW_FMT;
        else
            dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT;
    }

    return dma_mod;
}

/*!
 * @brief  Get the starting offset relative to sensor coordinates without any license restrictions.
 *
 * The sensor starting position is needed for the IPM to determine the Bayer starting color,
 * even though that is actually moot now because we always start on even boundaries,
 * and to set up FPGA bad pixel processing. These x-y offsets might need to be adjusted
 * if there is a license restriction.
 *
 * Skip any values for which the return-value pointer is NULL.
 *
 * @param rtn_true_x    Starting 0-based column number in pixels.
 * @param rtn_true_y    Starting 0-based row number.
 *
 * @return true for success, false for failure.
 */
bool CameraServer::get_sensor_start(uint16_t *rtn_true_x, uint16_t *rtn_true_y)
{
    uint16_t  start_x;
    uint16_t  start_y;
    bitmode_t bit_mode;
    uint32_t  sensor_opts;

    get_user_roi(&start_x, &start_y, nullptr, nullptr);
    get_user_bit_mode(&bit_mode);
    get_sensor_opts(&sensor_opts);

    // Since the size and offset we keep in GigE registers is restricted to what is licensed,
    // we need to undo the offset adjustment so it is based on the sensor's actual dimensions.
    if (rtn_true_x != nullptr)
        *rtn_true_x = get_true_startx(bit_mode, sensor_opts, start_x);
    if (rtn_true_y != nullptr)
        *rtn_true_y = get_true_starty(bit_mode, sensor_opts, start_y);

    return true;
}

/*!
 * @brief  Get the user-specified ROI.
 *
 * Skip any values for which the return-value pointer is NULL.
 *
 * @param rtn_start_x   Return starting 0-based column number in pixels, or NULL to not return it.
 * @param rtn_start_y   Return starting 0-based row number, or NULL.
 * @param rtn_width Return row width in pixels, or NULL.
 * @param rtn_height    Return height in rows, or NULL.
 *
 * @note The ROI visible to the user might be restricted by the license.
 *   It can also be restricted by the current bit mode or sensor options.
 *
 * @return true for success, false for failure.
 */
bool CameraServer::get_user_roi(uint16_t *rtn_start_x, uint16_t *rtn_start_y,
                                uint16_t *rtn_width, uint16_t *rtn_height)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint16_t start_x;
    uint16_t start_y;
    uint16_t width;
    uint16_t height;

    // Star with the user's settings for the ROI.
    // These GigE values must be legal, that is, they must be within the range for
    // the license, the mode (LTR vs. standard), and the sensor and they must honor
    // any multiple restrictions, such as even rows or columns.
    start_x = regs[ GIGE_REG_OFFSET_X ].get_value();
    start_y = regs[ GIGE_REG_OFFSET_Y ].get_value();
    width = regs[ GIGE_REG_WIDTH ].get_value();
    height = regs[ GIGE_REG_HEIGHT ].get_value();

    // Apply overrides. Choosing flexibility over safety for what is basically an engineer's tool,
    // we allow these to be illegal so buyer beware.
    {
        uint32_t ovr_xoffset;
        uint32_t ovr_yoffset;
        uint32_t ovr_height;
        uint32_t ovr_width;
        bool     got_height, got_width;

        if ((got_height = override_u32(OVR_ROI_HEIGHT, &ovr_height)) == true)
            height = (uint16_t)ovr_height;
        if ((got_width = override_u32(OVR_ROI_WIDTH, &ovr_width)) == true)
            width = (uint16_t)ovr_width;

        if (override_u32(OVR_ROI_X_OFFSET, &ovr_xoffset))
            start_x = (uint16_t)ovr_xoffset;
        else if (got_width)
            start_x = (sensor_ptr->get_sensor_max_width() / 2) - (ovr_width / 2);

        if (override_u32(OVR_ROI_Y_OFFSET, &ovr_yoffset))
            start_y = (uint16_t)ovr_yoffset;
        else if (got_height)
            start_y = (sensor_ptr->get_sensor_max_height() / 2) - (ovr_height / 2);
    }

    if (rtn_start_x != nullptr)
        *rtn_start_x = start_x;
    if (rtn_start_y != nullptr)
        *rtn_start_y = start_y;
    if (rtn_width != nullptr)
        *rtn_width = width;
    if (rtn_height != nullptr)
        *rtn_height = height;

    return true;
}

/*!
 * @brief  Get the user-specified bit selection mode.
 *
 * So the function can be called from others that get a bunch of parameters,
 * don't write if the return-value pointer is NULL.
 *
 * @param rtn_bit_mode  Return bit mode here.
 */
void CameraServer::get_user_bit_mode(bitmode_t *rtn_bit_mode)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    bitmode_t bit_mode = (bitmode_t)regs[ GIGE_REG_BIT_MODE ].get_value();

    // Allow for an override of the bit selection mode.
    {
        uint32_t ovr_bit_mode;
        if (override_u32(OVR_BIT_SELECT, &ovr_bit_mode))
            bit_mode = (bitmode_t)ovr_bit_mode;
    }

    if (rtn_bit_mode != nullptr)
        *rtn_bit_mode = bit_mode;
}

/*!
 * @brief  Get the user-specified sensor binning and subsampling options
 *
 * So the function can be called from others that get a bunch of parameters,
 * don't write if the return-value pointer is NULL.
 *
 * @param rtn_sensor_opts   Return options here.
 */
void CameraServer::get_sensor_opts(uint32_t *rtn_sensor_opts)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t sensor_opts = regs[ GIGE_REG_SENSOR_OPTIONS ].get_value();

    // We could allow for an override of the sensor options here.

    if (rtn_sensor_opts != nullptr)
        *rtn_sensor_opts = sensor_opts;
}

/*!
 * @brief  Get the current sensor gain setting
 */
void CameraServer::get_sensor_gain(uint32_t *rtn_sensor_gain)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t sensor_gain = regs[ GIGE_REG_SENSOR_GAIN ].get_value();

    if (rtn_sensor_gain != nullptr)
        *rtn_sensor_gain = sensor_gain;
}

/*!
 * @brief  Get the user-specified FPN choice.
 *
 * @param rtn_fpga_on   Return FPGA FPN true/false here if non-NULL.
 * @param rtn_sensor_on Return sensor FPN true/false here if non-NULL.
 */
void CameraServer::get_user_fpn(fpnmode_t *rtn_fpn_mode)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpnmode_t fpn_mode = (fpnmode_t)regs[ GIGE_REG_FPN ].get_value();

    // Allow for an override of the FPN choice.
    {
        bool force_pixel;
        if (override_bool(OVR_CORRECTION_MODE, &force_pixel))
        {
            if (force_pixel)
                fpn_mode = (fpnmode_t)GIGE_FPN_FPGA_ENABLE_MASK;
            else
                fpn_mode = (fpnmode_t)GIGE_FPN_SENSOR_ENABLE_MASK;
        }
    }

    if (rtn_fpn_mode != nullptr)
        *rtn_fpn_mode = fpn_mode;
}

/*!
 * @brief  Get user's setting for frame rate.
 *
 * So the function can be called from others that get a bunch of parameters,
 * don't write if the return-value pointer is NULL.
 *
 * @param rtn_frame_rate    Return frame rate here (frames per second).
 */
void CameraServer::get_frame_rate(float *rtn_frame_rate)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t frame_rate;

    frame_rate = regs[ GIGE_REG_FRAME_RATE ].get_value();
    override_u32(OVR_FRAMERATE, &frame_rate);

    if (rtn_frame_rate != nullptr)
        *rtn_frame_rate = reg_to_frame_rate(frame_rate);
}

/*!
 * @brief  Get user's setting for shutter speed, aka integration time.
 *
 * So the function can be called from others that get a bunch of parameters,
 * don't write if the return-value pointer is NULL.
 *
 * @param rtn_shutter_speed Return integration time here (microseconds).
 */
void CameraServer::get_shutter_speed(uint32_t *rtn_shutter_speed)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t shutter_speed;

    shutter_speed = regs[ GIGE_REG_SHUTTER_SPEED ].get_value();
    override_u32(OVR_INTEGRATION, &shutter_speed);

    if (rtn_shutter_speed != nullptr)
        *rtn_shutter_speed = shutter_speed;
}

/*!
 * @brief  Convenience function to get sensor-relative capture settings
 *
 * Skip any values for which the return-value pointer is NULL.
 *
 * @param rtn_true_x        For return of 0-based starting column number in pixels.
 * @param rtn_true_y        For return of 0-based starting row number.
 * @param rtn_width     For return of width in pixels.
 * @param rtn_height        For return of height in rows.
 * @param rtn_bit_mode      For return of bit mode here.
 * @param rtn_frame_rate    For return of frame rate (frames per second).
 * @param rtn_shutter_speed For return of integration time (microseconds).
 * @param rtn_sensor_opts   For return of sensor options for binning and subsampling.
 */
void CameraServer::get_physical_settings(uint16_t *rtn_true_x, uint16_t *rtn_true_y,
                                         uint16_t *rtn_width, uint16_t *rtn_height,
                                         bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                         uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts)
{
    uint16_t start_x;
    uint16_t start_y;
    bitmode_t bit_mode;
    uint32_t  sensor_opts;

    get_user_settings(&start_x, &start_y, rtn_width, rtn_height, &bit_mode,
                      rtn_frame_rate, rtn_shutter_speed, &sensor_opts);

    // Since the size and offset we keep in GigE registers is restricted to what is licensed,
    // we need to undo the offset adjustment so it is based on the sensor's actual dimensions.
    if (rtn_true_x != nullptr)
        *rtn_true_x = get_true_startx(bit_mode, sensor_opts, start_x);
    if (rtn_true_y != nullptr)
        *rtn_true_y = get_true_starty(bit_mode, sensor_opts, start_y);

    // Return to the caller other values that we needed to have in local variables.
    if (rtn_bit_mode != nullptr)
        *rtn_bit_mode = bit_mode;
    if (rtn_sensor_opts != nullptr)
        *rtn_sensor_opts = sensor_opts;
}

/*!
 * @brief  Convenience function to get user's capture settings
 *
 * @param rtn_start_x       For return of 0-based starting column number in pixels.
 * @param rtn_start_y       For return of 0-based starting row number.
 * @param rtn_width     For return of width in pixels.
 * @param rtn_height        For return of height in rows.
 * @param rtn_bit_mode      For return of bit mode here.
 * @param rtn_frame_rate    For return of frame rate (frames per second).
 * @param rtn_shutter_speed For return of integration time (microseconds).
 * @param rtn_sensor_opts   For return of sensor options for binning and subsampling.
 */
void CameraServer::get_user_settings(uint16_t *rtn_start_x, uint16_t *rtn_start_y,
                                     uint16_t *rtn_width, uint16_t *rtn_height,
                                     bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                     uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts)
{
#if 1
    get_user_roi(rtn_start_x, rtn_start_y, rtn_width, rtn_height);
    get_frame_rate(rtn_frame_rate);
    get_shutter_speed(rtn_shutter_speed);
    get_user_bit_mode(rtn_bit_mode);
    get_sensor_opts(rtn_sensor_opts);
#else
    *rtn_start_x = 0;
    *rtn_start_y = 0;
    *rtn_width = 1280;
    *rtn_height = 864;
    *rtn_bit_mode = GIGE_BITMODE_9_0;
    *rtn_frame_rate = 250;
    *rtn_shutter_speed = 200;
    *rtn_sensor_opts = 0;
#endif
}

/*!
 * @brief  Set up hardware registers so they match the user GigE registers.
 *
 * Does not start capture but sets up the hardware registers with new settings.
 * After this, going into LIVE_VIDEO can be done by changing the camera state.
 */
bool CameraServer::setup_capture_from_regs()
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    bool    rc = true;
    uint16_t    true_x;
    uint16_t    true_y;
    uint16_t    width;
    uint16_t    height;
    bitmode_t   bit_mode;
    fpnmode_t   fpn_mode;
    float    frame_rate;
    uint32_t    shutter_speed;
    uint32_t    sensor_opts;
    uint32_t    sensor_gain;

    LOG("%s", __PRETTY_FUNCTION__);

    get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode, &frame_rate,
                          &shutter_speed, &sensor_opts);
    get_user_fpn(&fpn_mode);
    get_sensor_gain(&sensor_gain);

    // Any time our ROI changes, we will need to update IPM bad pixel information
    // if that's what our sensor is using.
    sensor_ptr->IPM_bad_pixel_correction(0, true_x, true_y, width, height);

    // Set up the sensor. This sets up sensor registers but we won't actually
    // start it running here. It also sets up the FPGA ICM ROI since the way
    // that is done depends on the specific sensor.
    if (!sensor_ptr->sensor_startup(true_x, true_y, width, height, fpn_mode, bit_mode,
                                    frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: sensor_startup failed.", __FUNCTION__);
        rc = false;
    }

    // The FPGA FPN is in the IPM so we can change it on the fly. This will only change it
    // for the P0 IPM port anyway, which is for Live Video.
    if ((fpn_mode & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
    {
        if (FiImageCorrection::fpnEnable() < 0)
            rc = false;
    }

    FiFPGA::getInstance()->icmSetFrameRate(frame_rate);
    FiFPGA::getInstance()->icmSetShutterSpeed(shutter_speed);

    return rc;
}

/*!
 * @brief Update capture GigE register settings and restart the sensor if necessary.
 *
 * This is used to change OFFSET_X, OFFSET_Y, WIDTH, HEIGHT, BIT_MODE, or SENSOR_OPTIONS.
 * It might also involve changing FRAME_RATE or SHUTTER_SPEED but there are simpler ways to
 * change those if they are the only things changing; usually they only changed as a side effect
 * of other changes.
 *
 * A precondition of this function is that the hardware settings match the GigE registers. We
 * use the changed values to decide what needs to be modified.
 *
 * Changing any of those involves stopping capture, changing the parameter, then restarting capture.
 * They can result in a different number of frames so the partition manager is involved.
 * If the change results in an invalid frame rate, it is corrected automatically.
 *
 * @returns true for success, false for error
 */
bool CameraServer::modify_capture(
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t shutter_speed,
    uint32_t sensor_opts)
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();

    uint8_t save_camera_state;
    bitmode_t   save_bit_mode;
    uint16_t    save_startx;
    uint16_t    save_starty;
    uint16_t    save_width;
    uint16_t    save_height;
    float       save_frame_rate;
    uint32_t    save_shutter_speed;
    uint32_t    save_sensor_opts;
    uint32_t    sensor_gain;
    uint16_t    new_sensor_width;
    uint16_t    new_sensor_height;
    uint16_t    true_x;
    uint16_t    true_y;
    bool    need_fps_restart;

    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure we're in a good state.
    save_camera_state = CameraServer::get_camera_state();
    if ((save_camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (save_camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
    {
        fi_log(LOG_DEBUG, "%s: Invalid state [%d]", __PRETTY_FUNCTION__, save_camera_state);
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // Disable capture in fpga before we make these changes.
    // We don't stop the sensor; it just pauses because it is getting no integration pulses.
    if (save_camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        cs_ptr->set_idle_behavior();

    // Get the current settings so we can see what is changing.
    cs_ptr->get_user_settings(&save_startx, &save_starty, &save_width, &save_height,
                              &save_bit_mode, &save_frame_rate, &save_shutter_speed,
                              &save_sensor_opts);
    get_sensor_gain(&sensor_gain);

    //fi_log(LOG_DEBUG, "%s: attempting (%u,%u)%ux%u bits=0x%x %f/s %u-usec opts=0x%x", __FUNCTION__, startx, starty, width, height, bit_mode, frame_rate, shutter_speed, sensor_opts);
    // Force ROI limits into range, then make sure frame rate works for that ROI, then
    // make sure shutter works for that frame rate.
    cs_ptr->make_capture_valid(&startx, &starty, &width, &height, &bit_mode, &frame_rate,
                               &shutter_speed, &sensor_opts);

    regs[ GIGE_REG_OFFSET_X ].set_value(startx);
    regs[ GIGE_REG_OFFSET_Y ].set_value(starty);
    regs[ GIGE_REG_WIDTH ].set_value(width);
    regs[ GIGE_REG_HEIGHT ].set_value(height);
    regs[ GIGE_REG_BIT_MODE ].set_value(bit_mode);
    regs[ GIGE_REG_FRAME_RATE ].set_value(cs_ptr->frame_rate_to_reg(frame_rate));
    regs[ GIGE_REG_SHUTTER_SPEED ].set_value(shutter_speed);
    regs[ GIGE_REG_SENSOR_OPTIONS ].set_value(sensor_opts);

    // See if this change will require that we restart the sensor.
    true_x = cs_ptr->get_true_startx(bit_mode, sensor_opts, startx);
    true_y = cs_ptr->get_true_starty(bit_mode, sensor_opts, starty);
    need_fps_restart = sensor_ptr->frame_rate_needs_restart(save_frame_rate, frame_rate,
                                                            true_x, true_y, width, height,
                                                            bit_mode, sensor_opts);

    // Size changes can affect partition parameters.
    if ((save_width != width) || (save_height != height) || (save_bit_mode != bit_mode) ||
        (save_sensor_opts != sensor_opts))
    {
        // Reserve space for the gain table.
        PartMgr::getInstance()->reserveGainTableSpace(true);

        // Make sure the partition manager layout matches our current ROI.
        p_partmgr->layoutMemoryFromRegs();

        // If the ROI changed, the frame range could change, which could result in
        // a bad trigger position if it is specified by frame number.
        fcp_limit_trigger_position(p_partmgr->getFrameCount() - 1);

        // Any time the size changes, reset the save range to the full range so it
        // wasn't derived from an old notion of the session length.
        fcp_reset_save_limits(p_partmgr->getFrameCount() - 1);
    }

    // If the ROI changed or the sensor options changed, or we are restarting the sensor due
    // to a frame rate change, reload the calibration tables.
    // The bit mode matters because it could have caused a memory layout change or
    // may require that the FPN offset be shifted to match the capture bits.
    if ((save_startx != startx) || (save_starty != starty) ||
        (save_width != width) || (save_height != height) ||
        (save_bit_mode != bit_mode) || (save_sensor_opts != sensor_opts) || need_fps_restart)
    {
        // Load the correction tables.
        cs_ptr->load_ct_tables();
        cs_ptr->refresh_ct_status();
    }

    // Adjust the bit mode selection if it changed.
    if (save_bit_mode != bit_mode)
        sensor_ptr->set_fpga_bit_mode(bit_mode);

    // Allow for ROI changes or other reasons why we need to restart the sensor.
    if ((save_startx != startx) || (save_starty != starty) ||
        (save_width != width) || (save_height != height) ||
        sensor_ptr->bitmode_needs_restart(save_bit_mode, bit_mode) ||
        (save_sensor_opts != sensor_opts) || need_fps_restart)
    {
        fpnmode_t fpn_mode;

        // These are needed in case the sensor has to be restarted for this ROI change.
        cs_ptr->get_user_fpn(&fpn_mode);

        sensor_ptr->update_capture(true_x, true_y, width, height, fpn_mode, bit_mode,
                                   frame_rate, shutter_speed, sensor_opts, sensor_gain);

        // Any time our ROI changes, update IPM bad pixel information.
        sensor_ptr->IPM_bad_pixel_correction(0, true_x, true_y, width, height);
    }

    // See if we had to adjust any timing for this ROI.
    if (save_shutter_speed != shutter_speed)
        FiFPGA::getInstance()->icmSetShutterSpeed(shutter_speed);

    LOG("%s: save_frame_rate=%f, new_frame_rate=%f", __PRETTY_FUNCTION__, save_frame_rate, frame_rate);
    if (save_frame_rate != frame_rate)
    {
        uint32_t new_max;

        // Update the FPGA with the new frame rate.
        FiFPGA::getInstance()->icmSetFrameRate(frame_rate);

        // Update our maximum shutter (longest exposure) based on the new framerate.
        new_max = get_max_shutter_speed(frame_rate);
        regs[ GIGE_REG_MAX_SHUTTER_SPEED ].set_value(new_max);
        fcp_notify_apps(GIGE_REG_MAX_SHUTTER_SPEED, new_max);
    }

    // Restart capture.
    if (save_camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        cs_ptr->set_live_behavior();

    // If we were in low-light mode but changed frame_rate or shutter_speed, restore it.
    if ((!approx_equals(save_frame_rate, frame_rate) || (save_shutter_speed != shutter_speed)) &&
        FiFPGA::getInstance()->icmIsLowLightEnabled())
    {
        FiFPGA::getInstance()->icmSetFrameRate((float)LOW_LIGHT_FRAME_RATE);
        FiFPGA::getInstance()->icmSetShutterSpeed(FiFPGA::getInstance()->icmLowLightShutterSpeed());
    }

    new_sensor_width = sensor_ptr->get_reported_max_width(sensor_opts, bit_mode);
    new_sensor_height = sensor_ptr->get_reported_max_height(sensor_opts, bit_mode);
    if (new_sensor_width > get_lic_width())
        new_sensor_width = get_lic_width();
    if (new_sensor_height > get_lic_height())
        new_sensor_height = get_lic_height();
    if (new_sensor_width != regs[ GIGE_REG_SENSOR_WIDTH ].get_value())
    {
        regs[ GIGE_REG_SENSOR_WIDTH ].set_value(new_sensor_width);
        fcp_notify_apps(GIGE_REG_SENSOR_WIDTH, new_sensor_width);
    }
    if (new_sensor_height != regs[ GIGE_REG_SENSOR_HEIGHT ].get_value())
    {
        regs[ GIGE_REG_SENSOR_HEIGHT ].set_value(new_sensor_height);
        fcp_notify_apps(GIGE_REG_SENSOR_HEIGHT, new_sensor_height);
    }

    // Notify apps of any changes.
    if (save_startx != startx)
        fcp_notify_apps(GIGE_REG_OFFSET_X, startx);
    if (save_starty != starty)
        fcp_notify_apps(GIGE_REG_OFFSET_Y, starty);
    if (save_width != width)
        fcp_notify_apps(GIGE_REG_WIDTH, width);
    if (save_height != height)
        fcp_notify_apps(GIGE_REG_HEIGHT, height);
    if (save_bit_mode != bit_mode)
        fcp_notify_apps(GIGE_REG_BIT_MODE, bit_mode);
    if (save_frame_rate != frame_rate)
        fcp_notify_apps(GIGE_REG_FRAME_RATE, cs_ptr->frame_rate_to_reg(frame_rate));
    if (save_shutter_speed != shutter_speed)
        fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, shutter_speed);
    if (save_sensor_opts != sensor_opts)
        fcp_notify_apps(GIGE_REG_SENSOR_OPTIONS, sensor_opts);

    return true;
}

/*!
 * @brief Change the FPN option.
 *
 * Changing the FPN option in the sensor is a big deal, basically requiring that we take the
 * capture path down and bring it up again. But for Pixel FPN done by the FPGA, we can just
 * turn it on and off as long as we have the appropriate correction table loaded.
 *
 * Careful here, FPN is one of the few situations where the hardware state does not necessarily
 * match the GigE register settings. We can change the FPN setting in the hardware while we are
 * calibrating without modifying the user settings in the GigE registers.
 *
 * @param fpn_option    The new FPN to use.
 *
 * @return true if everything is good, otherwise false.
 */
bool CameraServer::set_fpn(uint32_t fpn_option)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t        old_fpn;
    uint8_t     saved_camera_state;
    bool        rtn = true;

    PDEBUG(PDEBUG_TRACE, "%s(0x%x)", __PRETTY_FUNCTION__, fpn_option);

    // If nothing changed, just return success.
    old_fpn = regs[ GIGE_REG_FPN ].get_value();
    if (old_fpn == fpn_option)
        return true;

    // This should only be done if we are unarmed.
    saved_camera_state = CameraServer::get_camera_state();
    if ((saved_camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (saved_camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
        return false;

    // Even then, let's set it to IDLE.
    if (saved_camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        set_idle_behavior();

    // If FPGA FPN is on in the hardware but we need to turn it off, do that first.
    if (((fpn_option & GIGE_FPN_FPGA_ENABLE_MASK) == 0) && (FiImageCorrection::fpnIsEnabled()))
    {
        FiImageCorrection::fpnDisable();
        FiImageCorrection::fpnResetOffset();
    }

    // Set the register because the down-up code wants to get the setting.
    regs[ GIGE_REG_FPN ].set_value(fpn_option);
    fcp_notify_apps(GIGE_REG_FPN, fpn_option);

    // If the FPN change requires that the sensor be restarted, do the full down-up sequence.
    if (sensor_ptr->fpn_needs_restart((fpnmode_t)old_fpn, (fpnmode_t)fpn_option))
        setup_fpga_and_sensor();

    // The FPGA FPN is in the IPM so we can change it on the fly. This will only change it
    // for the P0 IPM port anyway, which is for Live Video.
    if ((fpn_option & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
    {
        // This might be redundant but we need to make sure the CT is a match.
        if (FiImageCorrection::fpnLoad() < 0)
            rtn = false;
        else if (FiImageCorrection::fpnEnable() < 0)
            rtn = false;
    }

    // Restore the camera state.
    if (saved_camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        set_live_behavior();

    return rtn;
}

/*!
 * @brief Completely initialize the sensor and related FPGA logic.
 *
 * Set the FPGA and sensor registers so they are consistent with the IDLE state,
 * so the standard state transition code IDLE->LIVE_VIDEO will work.
 *
 * This is a major hardware bringup procedure. It is not normally needed if the only change
 * is frame rate, exposure, or ROI, but it is needed to change the sensor FPN option or
 * the sensor bit mode.
 *
 * This "big deal" procedure is used at startup, when turning on the sensor test mode,
 * after ICM runtime realignment fails (which was initiated due to CRC errors), when
 * changing the sensor FPN setting, or when restarting the sensor after calibration.
 * We can also restart the sensor when modifying the capture settings, depending on what
 * changed and which sensor it is; the Lupa and Lince have different restart requirements
 * and scenarios.
 *
 * The capture buffer should have already been set up before calling this function.
 *
 * @return true if everything is good, otherwise false.
 */
bool CameraServer::setup_fpga_and_sensor()
{
    uint16_t true_x;
    uint16_t true_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;
    fpnmode_t fpn_mode;
    uint32_t sensor_gain;
    Sensor *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_TRACE, "%s", __PRETTY_FUNCTION__);
    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA revision is set properly.
    regs[GIGE_REG_FPGA_VER].set_value(FiFPGA::getInstance()->getFpgaVersion());

    // Get the settings and make sure they are usable by the sensor-FPGA system.
    // The ROI that comes back is relative to the full sensor size, not adjusted by license.
    get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode,
                          &frame_rate, &shutter_speed, &sensor_opts);
    get_user_fpn(&fpn_mode);
    get_sensor_gain(&sensor_gain);

    // Make sure bad pixel correction is set up for live video whenever
    // we start up the sensor.
    sensor_ptr->IPM_bad_pixel_correction(0, true_x, true_y, width, height);

    // Set up the sensor. This sets up sensor registers but we won't actually
    // start it running yet. It also sets up the FPGA ICM ROI since the way
    // that is done depends on the specific sensor.
    if (!sensor_ptr->sensor_startup(true_x, true_y, width, height, fpn_mode, bit_mode,
                                    frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: sensor_startup failed.", __FUNCTION__);
        return false;
    }

    // Set up the ICM pipeline. Nothing will happen until we turn on the sensor readout bit.
    if (!FiFPGA::getInstance()->icmStartup(bit_mode, frame_rate, shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: icm_startup failed.", __FUNCTION__);
        return false;
    }

    // We can turn on the sensor readout here or we can wait until the display controller
    // is set up to request frames. On bringup the order doesn't really matter so much.
    // If for symmetry, we want the take-down order to be the reverse of the bringup order,
    // we would wait until after the display controller is set up for armed/unarmed live view.

    // Turn on the SENSOR READOUT ENABLE bit in the slave mode control
    // register. The sensor needs to be pushing out data before enabling the ICM pipeline,
    // otherwise the FPGA state machine will fail to transition from IDLE to CAPTURE_READY.
    if (!FiFPGA::getInstance()->icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __FUNCTION__);
        return false;
    }

    return true;
}

/*!
 * @brief Start capture for calibration of a Lince sensor
 *
 * @param   factory     true if this is factory calibration, currently unused
 * @param   cal_bit_mode    the bit mode to use for this capture
 * @param   fpn_mode        the FPN mode to use for this capture
 * @param   cal_frame_rate  frame rate, only used in factory calibration
 * @param   cal_shutter_speed   exposure time, only used in factory calibration
 *
 * @return true if successful, false otherwise
 */
bool CameraServer::start_lince_calibration(
    bool factory,
    bitmode_t cal_bit_mode,
    fpnmode_t fpn_mode,
    float  cal_frame_rate,
    uint32_t  cal_shutter_speed)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    bool    rc;
    uint16_t    true_x;
    uint16_t    true_y;
    uint16_t    width;
    uint16_t    height;
    bitmode_t   bit_mode;
    float    frame_rate;
    uint32_t    shutter_speed;
    uint32_t    sensor_opts;
    uint32_t    sensor_gain;

    LOG("%s", __PRETTY_FUNCTION__);

    // Assume failure.
    rc = false;

    // Disable capture in fpga.
    set_idle_behavior();

    // Make sure FPN correction is off in case we are collecting a black frame.
    FiImageCorrection::fpnDisable();

    // Get current settings.
    get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode, &frame_rate,
                          &shutter_speed, &sensor_opts);
    get_sensor_gain(&sensor_gain);
    bit_mode = cal_bit_mode;

    // For factory calibrations the sensor code might want to override these.
    if (factory)
    {
        frame_rate = cal_frame_rate;
        shutter_speed = cal_shutter_speed;
    }

    // Close the shutter.
    FiFPGA::getInstance()->shutterClose(true);

    // Set up the sensor. This sets up sensor registers but we won't actually
    // start it running here. It also sets up the FPGA ICM ROI since the way
    // that is done depends on the specific sensor.
    if (!(static_cast<Lince5M *>(sensor_ptr))->calibration_startup(
            factory, true_x, true_y, width, height, fpn_mode, bit_mode,
            frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: calibration_startup failed.", __FUNCTION__);
        goto cal_exit;
    }

    // Set up the ICM pipeline. Nothing will happen until we turn on the sensor readout bit.
    if (!FiFPGA::getInstance()->icmStartup(bit_mode, frame_rate, shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: icm_startup failed.", __FUNCTION__);
        goto cal_exit;
    }

    if (!FiFPGA::getInstance()->icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __FUNCTION__);
        goto cal_exit;
    }

    // All good.
    rc = true;

cal_exit:
    // Make sure the shutter is open.
    FiFPGA::getInstance()->shutterClose(false);

    // No live view for factory calibration.
    if (!factory)
        FiFPGA::getInstance()->icmSetLiveArgs(bit_mode, frame_rate, shutter_speed);

    // Return our result.
    return rc;
}

/*!
 * @brief Return to normal capture once a calbration finishes
 *
 * We might have set up some non-standard parameters for purposes of calibration.
 * This restores the memory layout, sensor settings, and ICM capture to "normal."
 *
 * @return true if successful, false otherwise
 */
bool CameraServer::restart_normal_after_calibration(void)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();

    bool    rtn = true;
    uint32_t    fpn_option;

    // Reserve space for the gain table.
    PartMgr::getInstance()->reserveGainTableSpace(true);

    // Make sure any memory layout we changed is restored to normal.
    p_partmgr->layoutMemoryFromRegs();

    // Set up using the user parameters.
    setup_fpga_and_sensor();

    fpn_option = regs[ GIGE_REG_FPN ].get_value();
    if ((fpn_option & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
    {
        if (FiImageCorrection::fpnLoad() < 0)
            rtn = false;
        else
            FiImageCorrection::fpnEnable();
    }

    // Restart the normal display and restart capturing.
    FiFPGA::getInstance()->icmSetToLiveVideo();

    return rtn;
}

/*!
 * @brief Generate the custom white balance values. This routine runs a RGB histogram on the current
 * frame running through live video, adjusts the grn value to nominal based upon the
 * peak values, then adjusts the other colors. If it is sucessful the wb values are
 * updated.
 *
 * @param reg_value The Correlated Color Temperature value.
 * @return true if the wb was generated, otherwise false
 */
bool CameraServer::generate_custom_white_balance(uint32_t reg_value)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int     i;
    uint32_t    value;
    uint32_t    histo_x, histo_y;
    uint32_t    red_mean, grn_mean, blu_mean;
    uint32_t    max_value;
    uint32_t    fractional;
    float   f_gain;
    float       normalize_factor;
    bool        normalize_wb;
    float   r_gain, g_gain, b_gain; // For debug message.

    PDEBUG(PDEBUG_TRACE, "%s", __PRETTY_FUNCTION__);
    LOG("%s: reg_value=%d", __PRETTY_FUNCTION__, reg_value);

    // Make sure we are in unarmed live view.
    if (CameraServer::get_camera_state() != GIGE_CAMERA_STATE_LIVE_VIDEO)
        return false;

    // Determine the number of pixels we're sampling our histogram over.
    FiFPGA::getInstance()->regRead(IPM_HISTO_ROI_X_CONFIG, &histo_x);
    FiFPGA::getInstance()->regRead(IPM_HISTO_ROI_Y_CONFIG, &histo_y);

    // Try to get the peak histogram values.
    for (i = 0; i < FPGA_PEAK_RETRY; ++i)
    {
        // Read the peak values of the histogram to use in our calculations.
        FiFPGA::getInstance()->regRead(IPM_HISTO_PEAK_VALUE, &value);
        if (value != 0)
            break;

        // Wait a bit for the histogram to go idle. Sometimes a histogram capture
        // gets kicked off just before coming in here. We need to wait for that to
        // finish.
        FiFPGA::getInstance()->ipmIsRgbHistoIdle();
    }
    if (i == FPGA_PEAK_RETRY)
        return false;

    red_mean = (value >> 16) & 0xFF;
    grn_mean = (value >> 8) & 0xFF;
    blu_mean = value & 0xFF;
    if (red_mean == 0)
    {
        fi_log(LOG_ERR, "%s: red_mean is 0?!", __PRETTY_FUNCTION__);
        red_mean = 128;
    }
    if (grn_mean == 0)
    {
        fi_log(LOG_ERR, "%s: grn_mean is 0?!", __PRETTY_FUNCTION__);
        grn_mean = 128;
    }
    if (blu_mean == 0)
    {
        fi_log(LOG_ERR, "%s: blu_mean is 0?!", __PRETTY_FUNCTION__);
        blu_mean = 128;
    }

    // Determine which value is the largest.
    max_value = red_mean;
    if (grn_mean > max_value)
        max_value = grn_mean;
    if (blu_mean > max_value)
        max_value = blu_mean;
    if (max_value == 0)
    {
        fi_log(LOG_ERR, "%s: max_value is 0?!", __PRETTY_FUNCTION__);
        max_value = 128;
    }

    normalize_factor = 1.0;
    normalize_wb = false;
    if (override_bool(OVR_WHITE_BALANCE_NORM, &normalize_wb) && normalize_wb)
    {
        double factor;

        factor = (double)(grn_mean * grn_mean + red_mean * red_mean + blu_mean * blu_mean);
        factor /= (double)(3 * max_value * max_value);
        normalize_factor = sqrt(factor);
        fi_log(LOG_DEBUG, "%s: override normalize white balance gains %.3f", __FUNCTION__,
               normalize_factor);
    }

    // Now build up the gains for the individual channels.
    // Also, calculate the AsShotNeutral values.
    f_gain = (((float)max_value) / ((float)red_mean));
    if (normalize_wb)
        f_gain *= normalize_factor;
    r_gain = f_gain;    // Save a copy for debug message.
    value = floor(f_gain);
    fractional = (uint32_t)(((f_gain - ((float)value)) * 4096.0) + 0.5);
    value = (value << 12) | fractional;
    FiFPGA::getInstance()->ipmSetWbRed(value);
    regs[ GIGE_REG_WHITE_BALANCE_RED ] = value;

    f_gain = (((float)red_mean) / ((float)max_value));
    value = floor(f_gain);
    fractional = (uint32_t)(((f_gain - ((float)value)) * ((float)(1 << AS_SHOT_SHIFT))) + 0.5);
    as_shot_red = (value << AS_SHOT_SHIFT) | fractional;

    f_gain = (((float)max_value) / ((float)grn_mean));
    if (normalize_wb)
        f_gain *= normalize_factor;
    g_gain = f_gain;    // Save a copy for debug message.
    fractional = (uint32_t)(((f_gain - ((float)value)) * 4096.0) + 0.5);
    value = (value << 12) | fractional;
    FiFPGA::getInstance()->ipmSetWbGreen(value);
    regs[ GIGE_REG_WHITE_BALANCE_GRN ] = value;

    f_gain = (((float)grn_mean) / ((float)max_value));
    value = floor(f_gain);
    fractional = (uint32_t)(((f_gain - ((float)value)) * ((float)(1 << AS_SHOT_SHIFT))) + 0.5);
    as_shot_grn = (value << AS_SHOT_SHIFT) | fractional;

    f_gain = (((float)max_value) / ((float)blu_mean));
    if (normalize_wb)
        f_gain *= normalize_factor;
    b_gain = f_gain;    // Save a copy for debug message.
    fractional = (uint32_t)(((f_gain - ((float)value)) * 4096.0) + 0.5);
    value = (value << 12) | fractional;
    FiFPGA::getInstance()->ipmSetWbBlue(value);
    regs[ GIGE_REG_WHITE_BALANCE_BLU ] = value;

    f_gain = (((float)blu_mean) / ((float)max_value));
    value = floor(f_gain);
    fractional = (uint32_t)(((f_gain - ((float)value)) * ((float)(1 << AS_SHOT_SHIFT))) + 0.5);
    as_shot_blu = (value << AS_SHOT_SHIFT) | fractional;

    fi_log(LOG_DEBUG, "%s: wb gains r=%.3f g=%.3f b=%.3f (F=%.3f)", __FUNCTION__, r_gain, g_gain, b_gain, normalize_factor);

    // All done.
    return true;
}

uint32_t CameraServer::get_min_width(void)
{
    uint32_t min_width;

    min_width = sensor_ptr->get_sensor_min_width();
    if ((FiFPGA::getInstance()->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        if (min_width < LTR_MIN_WIDTH)
            min_width = LTR_MIN_WIDTH;
    }
    return min_width;
}

uint32_t CameraServer::get_min_height(void)
{
    uint32_t min_height;

    min_height = sensor_ptr->get_sensor_min_height();
    if ((FiFPGA::getInstance()->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        if (min_height < LTR_MIN_HEIGHT)
            min_height = LTR_MIN_HEIGHT;
    }
    return min_height;
}

/*!
 * @brief Calculate the buffer size based upon the current settings
 *
 * @param none
 *
 * @return the frame size required or 0 if an error
 */
uint32_t CameraServer::calc_frame_buf_size(void)
{
    bitmode_t bit_mode;
    uint16_t width;
    uint16_t height;

    get_user_roi(nullptr, nullptr, &width, &height);
    get_user_bit_mode(&bit_mode);
    return calc_src_dma_size(width, height, bit_mode);
}

/*
 * @brief Load the FPN correction tables into all the partitions if necessary
 *
 * This needs to be done when the memory layout changes if pixel FPN is in use.
 * If it fails, it is somewhat catastrophic. The only recovery is to switch to sensor FPN.
 *
 * @return true for success, false for failure.
 */
bool CameraServer::load_ct_tables(void)
{
    bool rtn = true;
    fpnmode_t fpn_mode;

    LOG("%s", __PRETTY_FUNCTION__);

    // Get the user choice, which is not necessarily how the hardware is currently set up.
    get_user_fpn(&fpn_mode);

    // If we aren't doing Pixel FPN, we're all done, having succeeded trivially.
    if ((fpn_mode & GIGE_FPN_FPGA_ENABLE_MASK) == 0)
        return true;

    // Make sure FPN is not already active before we update the tables.
    if (FiImageCorrection::fpnIsEnabled())
        FiImageCorrection::fpnDisable();

    // Load the black reference and copy it to all the partitions,
    // then go ahead and turn it on. It's in IPM so it doesn't matter
    // whether the ICM is running now but it shouldn't be.
    if (FiImageCorrection::fpnLoad() < 0)
        rtn = false;
    else if (FiImageCorrection::fpnEnable() < 0)
        rtn = false;

    return rtn;
}

/*
 * @brief Refresh the UNCALIBRATED status in the FPN register
 *
 * This needs to be done when the memory layout changes if pixel FPN is in use.
 * If it fails, it is somewhat catastrophic. The only recovery is to switch to sensor FPN.
 *
 * @return true for success, false for failure.
 */
void CameraServer::refresh_ct_status()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t reg_value;
    uint32_t orig_reg;

    orig_reg = reg_value = regs[GIGE_REG_FPN].get_value();
    if (((reg_value & GIGE_FPN_FPGA_ENABLE_MASK) != 0) && FiImageCorrection::fpnNeedsCal())
    {
        reg_value |= GIGE_FPN_FPGA_UNCALIBRATED;
    }
    else if ((reg_value & GIGE_FPN_FPGA_UNCALIBRATED) != 0)
    {
        reg_value &= ~GIGE_FPN_FPGA_UNCALIBRATED;
    }

    if (orig_reg != reg_value)
    {
        regs[GIGE_REG_FPN].set_value(reg_value);
        fcp_notify_apps(GIGE_REG_FPN, reg_value);
    }
}

/*!
 * @brief Is this ROI valid? With this framerate?
 *
 * @param startx        The desired starting x position in pixels (relative to licensed area)
 * @param starty        The desired starting y position in pixels
 * @param width         The desired width in pixels
 * @param height        The desired height in pixels
 * @param frame_rate        The desired frame rate
 * @param shutter_speed     The desired shutter speed, aka integration time, aka exposure time
 * @param bit_mode      The desired bit mode
 * @param max_frame_rate    If false is returned but the ROI is valid, the maximum allowed
 *              frame rate is returned here.
 *
 * The starting x,y position is what the user sees, i.e., it is relative to the licensed area.
 * We will have to convert it to sensor-relative to set up the hardware.
 *
 * @return false if the parameters are not valid, true if they are.
 */
bool CameraServer::is_user_roi_valid(uint16_t startx, uint16_t starty,
                                     uint16_t width, uint16_t height,
                                     bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed,
                                     uint32_t sensor_opts, float *max_frame_rate)
{
    uint16_t my_startx = startx;
    uint16_t my_starty = starty;
    uint16_t my_width = width;
    uint16_t my_height = height;
    bitmode_t  my_bit_mode = bit_mode;
    float my_frame_rate = frame_rate;
    uint32_t my_shutter_speed = shutter_speed;
    uint32_t my_sensor_opts = sensor_opts;
    bool any_change;

    if (max_frame_rate != nullptr)
        *max_frame_rate = frame_rate;

    PDEBUG(PDEBUG_LOW, "%s: attempting (%u,%u)%ux%u bits=0x%x %f/s %u-usec opts=0x%x",
           __FUNCTION__, startx, starty, width, height, bit_mode, frame_rate, shutter_speed,
           sensor_opts);

    any_change = make_capture_valid(&my_startx, &my_starty, &my_width, &my_height, &my_bit_mode,
                                    &my_frame_rate, &my_shutter_speed, &my_sensor_opts);

    // If there were no changes, the settings are good as they are.
    if (!any_change)
        return true;

    // We had to make a change to get the settings to be valid.
    // See if the ROI changed. If so, they can't fix it by just changing frame rate.
    if ((startx != my_startx) || (starty != my_starty) ||
        (width != my_width) || (height != my_height) || (bit_mode != my_bit_mode) ||
        (my_sensor_opts != sensor_opts))
    {
        LOG("%s: FAIL,      (%u,%u)%ux%u bits=0x%x %f/s %u-usec opts=0x%x",
            __FUNCTION__, my_startx, my_starty, my_width, my_height, my_bit_mode, my_frame_rate, my_shutter_speed,
            my_sensor_opts);

        if (max_frame_rate != nullptr)
            *max_frame_rate = 0;
        return false;
    }

    // The ROI remained the same so we just fixed the frame rate. Return what we changed it to.
    if (max_frame_rate != nullptr)
        *max_frame_rate = my_frame_rate;

    return false;
}

bool CameraServer::make_capture_valid(uint16_t *rtn_startx, uint16_t *rtn_starty,
                                      uint16_t *rtn_width, uint16_t *rtn_height,
                                      bitmode_t *rtn_bit_mode, float *rtn_frame_rate,
                                      uint32_t *rtn_shutter_speed, uint32_t *rtn_sensor_opts)
{
    // Get our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Sensor   *sensor_ptr = CameraServer::get_sensor();

    uint16_t true_x;
    uint16_t true_y;
    uint32_t reg_value;
    uint16_t startx = *rtn_startx;
    uint16_t starty = *rtn_starty;
    uint16_t width = *rtn_width;
    uint16_t height = *rtn_height;
    bitmode_t  bit_mode = *rtn_bit_mode;
    float frame_rate = *rtn_frame_rate;
    uint32_t shutter_speed = *rtn_shutter_speed;
    uint32_t sensor_opts = *rtn_sensor_opts;
    bool     any_change = false;

    // Check parameters that might change as a result of e.g. LTR mode change
    // or reading the settings file, and adjust them so they are within legal limits.
    reg_value = regs[ GIGE_REG_MIN_WIDTH ].get_value();
    if (width < reg_value)
    {
        width = reg_value;
        any_change = true;
    }
    // Restrict against the licensed max width.
    reg_value = get_lic_width();
    if (width > reg_value)
    {
        width = reg_value;
        any_change = true;
    }
    // Remember, we retain reg_value for a while as max reported width.
    reg_value = sensor_ptr->get_reported_max_width(sensor_opts, bit_mode);
    if (width > reg_value)
    {
        width = reg_value;
        any_change = true;
    }
    if ((width % sensor_ptr->get_sensor_width_mod()) != 0)
    {
        width = (width / sensor_ptr->get_sensor_width_mod()) * sensor_ptr->get_sensor_width_mod();
        any_change = true;
    }
    if ((startx % sensor_ptr->get_sensor_width_mod()) != 0)
    {
        startx = (startx / sensor_ptr->get_sensor_width_mod()) * sensor_ptr->get_sensor_width_mod();
        any_change = true;
    }
    if ((startx + width) > reg_value)
    {
        // Center the width as best we can.
        startx = (reg_value - width) / 2;
        if ((startx % sensor_ptr->get_sensor_width_mod()) != 0)
            startx = (startx / sensor_ptr->get_sensor_width_mod()) * sensor_ptr->get_sensor_width_mod();

        any_change = true;
    }

    reg_value = regs[ GIGE_REG_MIN_HEIGHT ].get_value();
    if (height < reg_value)
    {
        height = reg_value;
        any_change = true;
    }
    // Restrict against the licensed max height.
    reg_value = get_lic_height();
    if (height > reg_value)
    {
        height = reg_value;
        any_change = true;
    }
    // Remember, we retain reg_value for a while as max reported height.
    reg_value = sensor_ptr->get_reported_max_height(sensor_opts, bit_mode);
    if (height > reg_value)
    {
        height = reg_value;
        any_change = true;
    }
    if ((height % sensor_ptr->get_sensor_height_mod()) != 0)
    {
        height = (height / sensor_ptr->get_sensor_height_mod()) * sensor_ptr->get_sensor_height_mod();
        any_change = true;
    }
    if ((starty % sensor_ptr->get_sensor_height_mod()) != 0)
    {
        starty = (starty / sensor_ptr->get_sensor_height_mod()) * sensor_ptr->get_sensor_height_mod();
        any_change = true;
    }
    if ((starty + height) > reg_value)
    {
        // Center the ROI as best we can.
        starty = (reg_value - height) / 2;
        if ((starty % sensor_ptr->get_sensor_height_mod()) != 0)
            starty = (starty / sensor_ptr->get_sensor_height_mod()) * sensor_ptr->get_sensor_height_mod();

        any_change = true;
    }

    // GIGE_REG_MAX_FRAME_RATE is the maximum allowed by the sensor or the license.
    reg_value = regs[ GIGE_REG_MAX_FRAME_RATE ].get_value();
    if (frame_rate > (float)reg_value)
    {
        frame_rate = (float)reg_value;
        any_change = true;
    }

    // Check whether the specific sensor has any restrictions we need to enforce.
    // This will check the sensor-specific frame rate for the ROI and sensor options.
    true_x = get_true_startx(bit_mode, sensor_opts, startx);
    true_y = get_true_starty(bit_mode, sensor_opts, starty);

    // PDEBUG( PDEBUG_SENSOR, "%s: attempting sensor (%u,%u)%ux%u bits=0x%x %u/s %u-usec opts=0x%x",
    fi_log(LOG_DEBUG, "%s: attempting sensor (%u,%u)%ux%u bits=0x%x %f/s %u-usec opts=0x%x",
           __PRETTY_FUNCTION__, true_x, true_y, width, height, bit_mode, frame_rate, shutter_speed,
           sensor_opts);
    if (sensor_ptr->make_sensor_valid(&true_x, &true_y, &width, &height, &bit_mode, &frame_rate, &shutter_speed, &sensor_opts))
    {
        any_change = true;
    }

    // Are we in LTR mode? If so, we have a different maximum frame rate.
    if ((FiFPGA::getInstance()->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        uint32_t    ltr_framerate;

        // Calculate the framerate for the suggest ROI.
        ltr_framerate = FiFPGA::getInstance()->ltrCalcFrameRate(width, height, bit_mode);

        // Adjust our maximum value appropriately.
        if (frame_rate > (float)ltr_framerate)
        {
            frame_rate = (float)ltr_framerate;
            any_change = true;
        }
        // Adjust our maximum for the license.
        ltr_framerate = get_max_ltr_frame_rate();
        if (frame_rate > (float)ltr_framerate)
        {
            frame_rate = (float)ltr_framerate;
            any_change = true;
        }
    }

    // Restrict the frame rate based on license.
    if (frame_rate > (float)get_max_licensed_frame_rate())
    {
        frame_rate = (float)get_max_licensed_frame_rate();
        any_change = true;
    }

    // Check our shutter speed.
    reg_value = get_max_shutter_speed(frame_rate);
    if (shutter_speed > reg_value)
    {
        shutter_speed = reg_value;
        any_change = true;
    }
    reg_value = regs[ GIGE_REG_MIN_SHUTTER_SPEED ].get_value();
    if (shutter_speed < reg_value)
    {
        shutter_speed = reg_value;
        any_change = true;
    }

    if (any_change)
    {
        fi_log(LOG_DEBUG, "%s: was %u:%u(%ux%u) 0x%x %f/s %u 0x%x invalid", __FUNCTION__,
               *rtn_startx, *rtn_starty, *rtn_width, *rtn_height, *rtn_bit_mode,
               *rtn_frame_rate, *rtn_shutter_speed, *rtn_sensor_opts);
    }

    *rtn_startx = startx;
    *rtn_starty = starty;
    *rtn_width = width;
    *rtn_height = height;
    *rtn_bit_mode = bit_mode;
    *rtn_frame_rate = frame_rate;
    *rtn_shutter_speed = shutter_speed;
    *rtn_sensor_opts = sensor_opts;

    if (any_change)
    {
        fi_log(LOG_DEBUG, "%s: now %u:%u(%ux%u) 0x%x %f/s %u 0x%x fixed", __FUNCTION__,
               *rtn_startx, *rtn_starty, *rtn_width, *rtn_height, *rtn_bit_mode,
               *rtn_frame_rate, *rtn_shutter_speed, *rtn_sensor_opts);
    }

    return any_change;
}

/*!
 * @brief Validate and possibly adjust a shutter speed
 *
 * This is used when we load a CAP file or get settings from an LTR capture.
 * If the shutter speed is outside the valid range, adjust it so it is acceptable.
 *
 * @param   fps         The frame rate
 * @param   shutter_speed   On entry, contains the shutter speed to check.
 *              On exit, if false is returned, is updated to a valid setting.
 *
 * @return true if the shutter speed had to be changed to get a valid setting.
 */
bool CameraServer::make_shutter_valid(uint32_t fps, uint32_t *shutter_speed)
{
    bool    rtn = true;
    uint32_t    shutter_temp;

    shutter_temp = get_max_shutter_speed(fps);
    if (*shutter_speed > shutter_temp)
    {
        *shutter_speed = shutter_temp;
        rtn = false;
    }
    shutter_temp = get_min_shutter_speed();
    if (*shutter_speed < shutter_temp)
    {
        *shutter_speed = shutter_temp;
        rtn = false;
    }
    return rtn;
}

/*!
 * @brief Validate and possibly adjust capture settings
 *
 * @note This is used both when we load new values from the settings file
 *   and when the LTR_MODE is changed. We make sure that the new settings
 *   are all mutually consistent and valid for the current mode.
 *
 * @return true if any changes were made to get valid settings.
 */
bool CameraServer::make_capture_regs_valid()
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    uint16_t    startx;
    uint16_t    starty;
    uint16_t    width;
    uint16_t    height;
    bitmode_t   bit_mode;
    float       frame_rate;
    uint32_t    shutter_speed;
    uint32_t    sensor_opts;
    bool    any_change = false;
    uint16_t    new_sensor_width;
    uint16_t    new_sensor_height;

    // Get the settings from the GigE register cache, allowing for overrides.
    get_user_settings(&startx, &starty, &width, &height, &bit_mode, &frame_rate, &shutter_speed,
                      &sensor_opts);

    any_change = make_capture_valid(&startx, &starty, &width, &height, &bit_mode,
                                    &frame_rate, &shutter_speed, &sensor_opts);
    if (any_change)
    {
        regs[ GIGE_REG_OFFSET_X ].set_value(startx);
        regs[ GIGE_REG_OFFSET_Y ].set_value(starty);
        regs[ GIGE_REG_WIDTH ].set_value(width);
        regs[ GIGE_REG_HEIGHT ].set_value(height);
        regs[ GIGE_REG_BIT_MODE ].set_value(bit_mode);
        regs[ GIGE_REG_FRAME_RATE ].set_value(frame_rate_to_reg(frame_rate));
        regs[ GIGE_REG_SHUTTER_SPEED ].set_value(shutter_speed);
        regs[ GIGE_REG_SENSOR_OPTIONS ].set_value(sensor_opts);
    }

    new_sensor_width = sensor_ptr->get_reported_max_width(sensor_opts, bit_mode);
    new_sensor_height = sensor_ptr->get_reported_max_height(sensor_opts, bit_mode);
    if (new_sensor_width > get_lic_width())
        new_sensor_width = get_lic_width();
    if (new_sensor_height > get_lic_height())
        new_sensor_height = get_lic_height();
    if (new_sensor_width != regs[ GIGE_REG_SENSOR_WIDTH ].get_value())
    {
        regs[ GIGE_REG_SENSOR_WIDTH ].set_value(new_sensor_width);
        fcp_notify_apps(GIGE_REG_SENSOR_WIDTH, new_sensor_width);
    }
    if (new_sensor_height != regs[ GIGE_REG_SENSOR_HEIGHT ].get_value())
    {
        regs[ GIGE_REG_SENSOR_HEIGHT ].set_value(new_sensor_height);
        fcp_notify_apps(GIGE_REG_SENSOR_HEIGHT, new_sensor_height);
    }

    if (regs[GIGE_REG_SESSION_LENGTH].get_value() > PartMgr::getInstance()->getSessionLengthMax())
    {
        fi_log(LOG_DEBUG, "%s: GIGE_REG_SESSION_LENGTH [%d] is wrong (max=%d)", __PRETTY_FUNCTION__, regs[GIGE_REG_SESSION_LENGTH].get_value(), PartMgr::getInstance()->getSessionLengthMax());
        regs[GIGE_REG_SESSION_LENGTH].set_value(PartMgr::getInstance()->getSessionLengthMax());
        fcp_notify_apps(GIGE_REG_SESSION_LENGTH, PartMgr::getInstance()->getSessionLengthMax());
    }

    return any_change;
}

/*!
 * @brief Translate coordinates for license restrictions
 *
 * Translate from coordinates that are relative to the sensor's licensed area
 * into coordinates that are relative to the unrestricted region. Since some
 * sizes are restricted by bit mode or sensor options, we might not need to
 * make any adjustment. Also, we might be licensed for a size bigger than
 * what we are supporting on this sensor, so be sure to avoid negative
 * adjustments.
 *
 * @param bit_mode  the capture bit mode
 * @sensor_opts     the sensor options
 * @startx      the user-visible starting offset
 *
 * @return new offset with license restrictions removed
 */
uint16_t CameraServer::get_true_startx(bitmode_t bit_mode, uint32_t sensor_opts, uint16_t startx)
{
    uint16_t max_width;

    // The sensor maximum dimensions can vary depending on the bit mode and the
    // sensor options. If the sensor's max reported dimensions are less than the
    // license dimensions, we don't need to do any adjustments.
    // For the Lupa, this always reports the full-size sensor but it varies for Lince.
    max_width = sensor_ptr->get_reported_max_width(sensor_opts, bit_mode);

    if ((lic_ptr == nullptr) || (sensor_ptr == nullptr) ||
        (max_width <= lic_ptr->widthMax()))
        return startx;

    return startx + ((max_width - lic_ptr->widthMax()) / 2);
}

/*!
 * @brief Translate coordinates for license restrictions
 *
 * Translate from coordinates that are relative to the sensor's licensed area
 * into coordinates that are relative to the unrestricted region. Since some
 * sizes are restricted by bit mode or sensor options, we might not need to
 * make any adjustment. Also, we might be licensed for a size bigger than
 * what we are supporting on this sensor, so be sure to avoid negative
 * adjustments.
 *
 * @param bit_mode  the capture bit mode
 * @sensor_opts     the sensor options
 * @starty      the user-visible starting offset
 *
 * @return new offset with license restrictions removed
 */
uint16_t CameraServer::get_true_starty(bitmode_t bit_mode, uint32_t sensor_opts, uint16_t starty)
{
    uint16_t max_height;

    max_height = sensor_ptr->get_reported_max_height(sensor_opts, bit_mode);

    if ((lic_ptr == nullptr) || (sensor_ptr == nullptr) ||
        (max_height <= lic_ptr->heightMax()))
        return starty;

    return starty + ((max_height - lic_ptr->heightMax()) / 2);
}

uint16_t CameraServer::get_lic_width()
{
    uint16_t width;

    if (sensor_ptr == nullptr)
        return 320;
    else if (lic_ptr == nullptr)
        width = sensor_ptr->get_sensor_max_width();
    else
        width = lic_ptr->widthMax();

    return width;
}

uint16_t CameraServer::get_lic_height()
{
    uint16_t height;

    if (camserv_ptr->sensor_ptr == nullptr)
        return 256;
    else if (lic_ptr == nullptr)
        height = sensor_ptr->get_sensor_max_height();
    else
        height = lic_ptr->heightMax();

    return height;
}

uint32_t CameraServer::frame_rate_to_reg(float frame_rate)
{
    uint32_t reg_value;

    // Grab the integer value.
    reg_value = (uint32_t)frame_rate;

    // Is there anything else to worry about?
    if (!approx_equals((frame_rate - (float)reg_value), 0))
    {
        // We have a true floating point frame rate. Do the conversion.
        reg_value = (uint32_t)(frame_rate * 65536.0);
        reg_value |= GIGE_FRAME_RATE_FRACTIONAL;
    }

    // Return the register value for frame rate.
    return reg_value;
}

float CameraServer::reg_to_frame_rate(uint32_t reg_value)
{
    float ret_value;

    if ((reg_value & GIGE_FRAME_RATE_FRACTIONAL) == GIGE_FRAME_RATE_FRACTIONAL)
        ret_value = (float)(reg_value & ~GIGE_FRAME_RATE_FRACTIONAL) / 65536.0;
    else
        ret_value = (float)reg_value;

    return ret_value;
}

uint32_t CameraServer::get_max_ltr_frame_rate()
{
    if (lic_ptr == nullptr)
        return 500;
    else
        return lic_ptr->fpsLtrMax();
}

uint32_t CameraServer::get_max_licensed_frame_rate()
{
    if (lic_ptr == nullptr)
        return 500;
    else
        return lic_ptr->fpsMax();
}

bool CameraServer::auto_trigger()
{
    if (lic_ptr == nullptr)
        return false;
    else
        return lic_ptr->autoTrigger();  // TRUE iff auto_trigger is non-zero.
}

bool CameraServer::dvr_playback()
{
#ifdef ALM_TODO
    // The current seriesNC camera does NOT support DVR playback mode.
    if (lic_ptr == nullptr)
        return false;
    else
        return lic_ptr->dvrPlayback();   // TRUE iff dvr_playback is non-zero.
#else
    return false;
#endif /* ALM_TODO */
}

bool CameraServer::integrated()
{
    if (lic_ptr == nullptr)
        return false;
    else
        return lic_ptr->integrated();
}

bool CameraServer::optical()
{
    if (lic_ptr == nullptr)
        return true;
    else
        return lic_ptr->optical();
}

void CameraServer::poll_for_dvr_finish(bool do_poll)
{
    if (dvr_poller == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: poller is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    if (do_poll)
    {
        // Tell the poller to start.
        dvr_poller->start();
    }
    else
    {
        // Tell the poller to stop.
        dvr_poller->stop();
    }
}

bool CameraServer::do_polling_for_dvr()
{
    if (dvr_poller == nullptr)
        return false;
    return dvr_poller->is_running();
}

// Assume max and minimum exposures are a combination of FPGA and sensor restrictions
// so the determination is in the CS instead of either FPGA or sensor.
#define COMMON_MIN_SHUTTERSPEED_USECS 3 // This should never kick in.
uint32_t CameraServer::get_min_shutter_speed()
{
    if (sensor_ptr == nullptr)
        return COMMON_MIN_SHUTTERSPEED_USECS;
    else
        return sensor_ptr->get_sensor_min_shutter();
}

uint32_t CameraServer::get_max_shutter_speed(float fps)
{
    if (sensor_ptr == nullptr)
        return COMMON_MIN_SHUTTERSPEED_USECS;
    else if (approx_equals(fps, 0))
        return sensor_ptr->get_sensor_min_shutter();
    else
    {
        uint32_t ret_value;

        ret_value = (uint32_t)((1000000 / fps) - sensor_ptr->get_frame_overhead());
        if (ret_value < get_min_shutter_speed())
        {
            LOG("%s: Invalid max exposure (%u)", __PRETTY_FUNCTION__, ret_value);
            ret_value = get_min_shutter_speed();
        }

        // Return our value.
        return ret_value;
    }
}

//============================================================================
// csAutoTrigger class implementation.
//============================================================================

/*!
 * @brief The auto-trigger thread constructor. The camera server auto-trigger thread
 *    is a "detached" thread. That is, it runs to completion (once started)
 *    and does not expect to "join" with any other thread.
 */
csAutoTrigger::csAutoTrigger() :
    FAS_THREAD::Thread(true),
    thread_running(false)
{
}

csAutoTrigger::~csAutoTrigger()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Main thread function to handle issuing a periodic auto-trigger.
 */
void csAutoTrigger::run()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We are now running.
    thread_running = true;

    // Do we need to poll our DVR playback status?
    while (!stop_request)
    {
        uint32_t pref;

        // Wait for a little before checking again.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(2000));

        // Are we doing auto-play? Need both auto-trigger and auto-play.
        pref = regs[ GIGE_REG_PREFERENCES ].get_value();
        if ((pref & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY)
        {
            // Are we armed?
            if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_ARMED)
            {
                // Force a trigger.
                if (!CameraServer::user_set_camera_state(GIGE_CAMERA_STATE_TRIGGERED))
                {
                    fi_log(LOG_DEBUG, "%s: Unable to force trigger.", __PRETTY_FUNCTION__);
                }
            }
        }
    }

    // We are no longer running.
    thread_running = false;
}

//============================================================================
// cs_poll class implementation.
//============================================================================

/*!
 * @brief The polling thread constructor. The camera server polling thread
 *    is a "detached" thread. That is, it runs to completion (once started)
 *    and does not expect to "join" with any other thread.
 */
cs_poll::cs_poll() :
    FAS_THREAD::Thread(true),
    thread_running(false)
{
}

cs_poll::~cs_poll()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Main thread function to handle waiting for FPGA events:
 *    camera armed, trigger assert, trigger deassert, partition complete.
 */
void cs_poll::run()
{
    CameraServer *cs_ptr = CameraServer::get_cs();

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We are now running.
    thread_running = true;

    // Do we need to poll our DVR playback status?
    while (!stop_request)
    {
        // We are done if the playback controller is not busy.
        if (pbc_busy())
        {
            // Poll faster when we know it's busy.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(50));
        }
        else
        {
            // We are done. Reset our flag.
            cs_ptr->poll_for_dvr_finish(false);

            // Go back into being armed. We do not need to wait for this one.
            CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_ARMED, false, nullptr);

            // Wait for a little before checking again.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(250));
        }
    }

    // We are no longer running.
    thread_running = false;
}

//============================================================================
// cs_watchdog class implementation.
//============================================================================

/*!
 * @brief The wathcdog petting thread constructor.
 */
cs_watchdog::cs_watchdog() :
    FAS_THREAD::Thread(true),
    thread_running(false)
{
}

cs_watchdog::~cs_watchdog()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Simple thread to issue request to pet the watchdog every so often.
 *
 * One of the things we check, other than time passing, is the PCI register
 * interface. If the head goes away, we will likely not have valid data in
 * the FPGA's register space.
 */
void cs_watchdog::run()
{
// #define WD_TESTER 1
#ifdef WD_TESTER
    int test_counter = 6;   // pet for 1/2 minute.
#endif

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We are now running.
    thread_running = true;

    // Keep petting the dog until we're told to knock it off.
    while (!stop_request)
    {
        bool pet_the_dog;
        uint32_t reg_value;
        FiFPGA *fpga_ptr = FiFPGA::getInstance();

        if (fpga_ptr == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Null fpga_ptr!", __PRETTY_FUNCTION__);
            pet_the_dog = false;
        }

        // Read the FPGA's version information.
        else if (!fpga_ptr->regRead(FPGA_REVISION, &reg_value))
        {
            fi_log(LOG_WARNING, "%s: Error reading FPGA_REVISION!", __PRETTY_FUNCTION__);

            // Don't pet the watchdog.
            pet_the_dog = false;
        }

        // Are they a match?
        else if (reg_value != fpga_ptr->getFpgaVersion())
        {
            fi_log(LOG_WARNING, "%s: FPGA revision mismatch! Did the head go away?", __PRETTY_FUNCTION__);

            // Don't pet the watchdog.
            pet_the_dog = false;
        }
        else
        {
            // Pet the watchdog.
            pet_the_dog = true;
        }

        // Pet the watchdog.
        if (pet_the_dog)
        {
            // Pet the dog.
            if (!fpga_ptr->regWrite(WATCHDOG_TICKLE, WD_TICKLE_MAGIC))
            {
                fi_log(LOG_ERR, "%s: Unable to write WATCHDOG_TICKLE register.", __FUNCTION__);
                return;
            }

            // Update the watchdog.
            CameraServer::watchdog();
        }
        else
        {
            // Let folks know we're having issues.
            fcp_notify_apps(GIGE_REG_BOOT_STATUS, GIGE_BOOT_STATUS_WDTIMEOUT);
        }

        // Wait for a little before checking again.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(5000));
    }

    // We are no longer running.
    thread_running = false;
    fi_log(LOG_DEBUG, "%s: exiting.", __PRETTY_FUNCTION__);
}
