/*!
 * @file cs_worker.cxx
 * @brief This file implements the csWorker class.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <chrono>
#include <mutex>

#include <fastec/fas_log.h>

#include "cs_worker.h"
#include "camera_server.h"
#include "fcp.h"
#include "fsp.h"
#include "grpc_server.h"
#include "override.h"
#include "part_mgr.h"

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::mutex csWorker::send_mutex;

/*!
 * @brief The worker thread constructor. The camera server worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
csWorker::csWorker() :
    FAS_THREAD::Thread(true),
    reopenRequested(false),
    shutdown_done(false)
{
}

csWorker::~csWorker()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Find out information about the specified network.
 * @param p_msg Pointer to a message for a network interface we're interested in discovering.
 * @return CS_NET_SUCCESS on success, one of CS_NET_ERROR_xxx on error.
 */
int csWorker::discoverNetwork(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int rc = CS_NET_ERROR_FD;

    // A null p_msg defaults to looking at the UDP interface.
    if (p_msg == nullptr)
    {
        rc = udp_socket.discoverNetwork();
    }
    else if (msgIsUDP(p_msg))
    {
        rc = udp_socket.discoverNetwork();
    }
    else if (local_socket.isValid())
    {
        rc = local_socket.discoverNetwork();
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: Nothing here?!", __FUNCTION__);
    }

    return rc;
}

/*!
 * @brief Override network parameters.
 */
void csWorker::doNetworkOverride()
{
    int status;
    uint32_t    ip_addr;
    uint32_t    netmask;
    uint32_t    gateway;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Initialize our variables.
    ip_addr = netmask = gateway = 0;

    // Do we have overrides?
    override_u32(OVR_STATIC_IP_ADDR, &ip_addr);
    override_u32(OVR_STATIC_NET_MASK, &netmask);
    override_u32(OVR_STATIC_GATEWAY, &gateway);

    // Make sure we've got everything we need.
    if ((ip_addr == 0) || (netmask == 0) || (gateway == 0))
        return;

    // Switch to a static IP address.
    status = CsNetwork::startStaticConnection(CsNetwork::ETHER, ip_addr, netmask, gateway);
    if (status == 0)
    {
        uint32_t old_value;

        // Set all the persisent registers
        regs[GIGE_REG_PERSIST_IP].set_value(ip_addr);
        regs[GIGE_REG_PERSIST_SUBNET].set_value(netmask);
        regs[GIGE_REG_PERSIST_GATEWAY].set_value(gateway);

        // Since we've change the IP address we are now in static mode so we need to
        // change the IP current configuration register
        old_value = regs[GIGE_REG_IP_CUR_CFG].get_value();
        old_value = (old_value & ~GIGE_IFCONFIG_DHCP_BE) | GIGE_IFCONFIG_PERSISTENT_BE;
        regs[GIGE_REG_IP_CUR_CFG].set_value(old_value);
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to change the network address (%d)", __PRETTY_FUNCTION__, status);
    }
}

/*!
 * @brief Find the default gateway associated with p_msg.
 * @param p_msg Pointer to the message we're interested in.
 * @return The IPv4 address of the default gateway, or 0.
 */
uint32_t csWorker::getDefaultGateway(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    uint32_t rc = 0;

    // A null p_msg defaults to looking at the UDP interface.
    if (p_msg == nullptr)
    {
        rc = udp_socket.getDefGateway();
    }
    else if (msgIsUDP(p_msg))
    {
        rc = udp_socket.getDefGateway();
    }
    else if (local_socket.isValid())
    {
        rc = local_socket.getDefGateway();
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: Nothing here?!", __FUNCTION__);
    }

    return rc;
}

/*!
 * @brief Get the IP address associated with p_msg.
 * @param p_msg Pointer to the message we're interested in.
 * @return The IPv4 address of the interface, or 0.
 */
uint32_t csWorker::getIpAddress(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    uint32_t rc = 0;

    // A null p_msg defaults to looking at the UDP interface.
    if (p_msg == nullptr)
    {
        rc = udp_socket.getIpAddress();
    }
    else if (msgIsUDP(p_msg))
    {
        rc = udp_socket.getIpAddress();
    }
    else if (local_socket.isValid())
    {
        rc = local_socket.getIpAddress();
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: Nothing here?!", __FUNCTION__);
    }

    return rc;
}

/*!
 * @brief Get the subnet mask associated with p_msg.
 * @param p_msg Pointer to the message we're interested in.
 * @return The IPv4 subnet mask of the interface, or 0.
 */
uint32_t csWorker::getSubnetMask(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    uint32_t rc = 0;

    // A null p_msg defaults to looking at the UDP interface.
    if (p_msg == nullptr)
    {
        rc = udp_socket.getSubnetMask();
    }
    else if (msgIsUDP(p_msg))
    {
        rc = udp_socket.getSubnetMask();
    }
    else if (local_socket.isValid())
    {
        rc = local_socket.getSubnetMask();
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: Nothing here?!", __FUNCTION__);
    }

    return rc;
}

/*!
 * @brief The camera server worker thread.
 *
 * The purpose of this thread is to service the many network interfaces we support.
 * The network interfaces are:
 * + csWorker::ipc_socket: This takes commands from other, local, applications such as the Web Socket Server.
 * + csWorker::local_socket: This takes commands from the loopback socket interface.
 * + csWorker::udp_socket: This interface is used for TCP/UDP commands.
 *
 * The code loops around a select() call, waiting for traffic from one of the opened sockets. This
 * is shown below:
 * @image html csWorker.svg
 */
void csWorker::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Initialize the network interface(s).
    initializeNetwork();

    // Loop until we are shut down.
    while (!stop_request)
    {
        int maxfd;
        fd_set readset;         // Set of file descriptors to read.
        std::vector<CsNetwork *> activeSockets;
        CS_NETWORK::CS_NETWORK_MESSAGE msg; // Structure for either an IPC or UDP message.

        // Are we supposed to re-open the sockets?
        if (reopenRequested)
            doReopen();

        // Initialize the file descriptor readset and the active sockets vector.
        maxfd = initializeReadSet(readset, activeSockets);
        if (maxfd < 0)
        {
            fi_log(LOG_ERR, "(%s) no file descriptor!", __FUNCTION__);

            // Something is wrong with the file descriptor, set the flag to close and reopen them.
            reopenRequested = true;

            using sec = std::chrono::seconds;
            std::this_thread::sleep_for(sec(1));
        }

        else
        {
            int nActive;            // Number of active descriptors.

            // Check to see if there is any work for us from local or remote.
            nActive = select(maxfd + 1, &readset, NULL, NULL, NULL);
            if (stop_request)
                break;

            if (nActive < 0)
            {
                fi_log(LOG_ERR, "(%s) select error: %d(%s)", __FUNCTION__, errno, strerror(errno));

                // If we've gotten a signal to close the main loop will take care of this because
                // the close flag will be set otherwise queue the select up again.
                if (errno != EINTR)
                {
                    // Something is wrong with the file descriptor, set the flag to close and reopen them.
                    reopenRequested = true;
                }
            }
            else
            {
                std::vector<CsNetwork *>::iterator i;
                for (i = activeSockets.begin(); i != activeSockets.end(); i++)
                {
                    // Prepare our message buffer.
                    memset(&msg, 0, sizeof(msg));

                    // Is the object the active one?
                    if (FD_ISSET((*i)->getSocket(), &readset))
                    {
                        int status;

                        // Get the message.
                        status = (*i)->getMsg(&msg);

                        // Everything OK?
                        if (status > 0)
                        {
                            // Add the message to the work queue.
                            CameraServer::add_work(&msg);
                        }
                        else
                        {
                            // Error on the socket or it's been closed, neither of which should
                            // happen. An error should have been logged by the called routine, so
                            // just set the flag to try and reopen the socket
                            reopenRequested = true;
                        }
                    }
                }
            }
        }
    }

    // Don't expect we'll ever get here. But if we do, let's clean things up.
    fi_log(LOG_DEBUG, "%s: done", __PRETTY_FUNCTION__);
    cleanup();
}

/*!
 * @brief Something has happened with the file descriptor, try closing and reopening them.
 *    If we can't reopen them, wait a few seconds and try again.
 */
void csWorker::doReopen()
{
    static volatile int async_done;

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Switch to the IDLE state.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_IDLE, true, &async_done);

    // Reset the locks.
    fcp_lock_reset();

    // Now switch to the LIVE state.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO, true, &async_done);

    // Reset the flag.
    reopenRequested = false;
}

void csWorker::stop()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Pass it on.
    FAS_THREAD::Thread::stop();

    // Make sure we clean things up.
    cleanup();
}

/*!
 * @brief Determine if p_msg is an UDP message or not.
 * @param p_msg Pointer to the message we're interested in.
 * @return True if the message is from the UDP interface, false otherwise.
 */
bool csWorker::msgIsUDP(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    bool rc = false;
    uint32_t from_addr;

    // Is the UDP socket valid?
    if (!udp_socket.isValid())
        return false;

    // Get the IP address in a form we can use.
    inet_aton(p_msg->udp_msg.from, (in_addr *)&from_addr);
    // fi_log(LOG_DEBUG, "%s: msg='%s', from_addr=0x%08x", __FUNCTION__, p_msg->udp_msg.from, from_addr);

    // Check the message to see if it came from the UDP interface.
    if ((from_addr & udp_socket.getSubnetMask()) == (udp_socket.getIpAddress() & udp_socket.getSubnetMask()))
    {
        // This is from the same UDP subnet.
        rc = true;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Send the message via the appropriate interface.
 *
 * @param p_msg     Pointer to the message we want to send.
 * @param num_bytes Number of bytes in the message.
 *
 * @return The size of the message if successful, 0 if no message, <0 if error.
 */
int csWorker::sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes)
{
    int rc;

    std::lock_guard<std::mutex> lock(send_mutex);

    // Process based on the which network the message came
    // in from.
    if (p_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC)
        rc = grpcSendMsg(p_msg, num_bytes);
    else if (p_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC)
        rc = ipc_socket.sendMsg(p_msg, num_bytes);
    else if (msgIsUDP(p_msg))
        rc = udp_socket.sendMsg(p_msg, num_bytes);
    else if (local_socket.isValid())
        rc = local_socket.sendMsg(p_msg, num_bytes);
    else
        rc = 0;

    // Return our result.
    return rc;
}

/*!
 * @brief Clean up the thread on our way out.
 */
void csWorker::cleanup()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    if (shutdown_done)
        return;

    // Close down the GRPC server.
    grpcServerClose();

    // Close our sockets.
    udp_socket.close();
    ipc_socket.close();
    local_socket.close();

    shutdown_done = true;
}

/*!
 * @brief Initialize a readset value and populate a vector with all the network interfaces
 *        which are active.
 *
 * @param readset           [OUT] The initialized readset of file descriptors.
 * @param activeSockets     [OUT] The initialized vector of active network interfaces.
 * @return The maximum file (socket) descriptor value.
 */
int csWorker::initializeReadSet(fd_set& readset, std::vector<CsNetwork *>& activeSockets)
{
    int ipc_fd;
    int lo_fd;
    int udp_fd;
    int maxfd = -1;              // Maximum file descriptor, used by select.

    // Initialize our file descriptor set.
    FD_ZERO(&readset);
    activeSockets.clear();

    // Get our IPC socket setup.
    ipc_fd = ipc_socket.getSocket();
    if (ipc_fd > 0)
    {
        FD_SET(ipc_fd, &readset);
        activeSockets.push_back(&ipc_socket);
        maxfd = ipc_fd;
    }

    // Get our loopback socket.
    lo_fd = local_socket.getSocket();
    if (lo_fd > 0)
    {
        FD_SET(lo_fd, &readset);
        activeSockets.push_back(&local_socket);
        if (lo_fd > maxfd)
            maxfd = lo_fd;
    }

    // Make sure we have a socket descriptor and we
    // are connected to either the ethernet or WIFI...
    udp_fd = udp_socket.getSocket();
    bool eth_conn = CsNetwork::isEthConnected();
    bool wifi_conn = CsNetwork::isWifiConnected();
    bool something_is_connected = eth_conn || wifi_conn;
    if ((udp_fd > 0) && something_is_connected)
    {
        FD_SET(udp_fd, &readset);
        activeSockets.push_back(&udp_socket);
        if (udp_fd > maxfd)
            maxfd = udp_fd;
    }
    else
    {
        // Make sure we close the socket.
        udp_socket.close();

        // See if anything has changed. Maybe the user has plugged
        // the network cable in.
        if (something_is_connected)
        {
            // Try opening the UDP socket.
            udp_socket.setup();
            udp_fd = udp_socket.getSocket();
            if (udp_fd > 0)
            {
                if (udp_socket.discoverNetwork() != CS_NET_SUCCESS)
                {
                    // Log the error.
                    fi_log(LOG_ERR, "%s: TID=%u, discover_network() failed (eth:%d, wifi:%d)",
                           __PRETTY_FUNCTION__, get_tid(),
                           udp_socket.isEthConnected(),
                           udp_socket.isWifiConnected());

                    // Make sure we close the socket.
                    udp_socket.close();
                }
                else
                {
                    // Pay attention to this socket.
                    FD_SET(udp_fd, &readset);
                    activeSockets.push_back(&udp_socket);
                    if (udp_fd > maxfd)
                        maxfd = udp_fd;
                }
            }
        }
    }

    // Return the maximum file descriptor found.
    return maxfd;
}

void csWorker::initializeNetwork()
{
    // Get our GRPC server running.
    grpcServerInit();

    // Initialize IPC to communicate with local processes.
    ipc_socket.setup();

    // Intiialize the loopback socket.
    local_socket.setup();

    // Initialize UDP sockets to communicate with remote processes
    udp_socket.setup();

    // Find out stuff about the network we are on, i.e. MAC, gateway, ip, subnet
    udp_socket.discoverNetwork();

    // Startup our FSP object.
    fsp_startup();

    // Check to see if we need to override the network settings.
    doNetworkOverride();

    // Pass the word that we are initialized.
    CameraServer::init_done();
}
