/*!
 * @file dcraw_util.cpp
 *
 * @brief Dcraw utility class implementation.
 *
 * Copyright 2008-2017 LibRaw LLC (info@libraw.org)
 * Created: Sat Mar  8, 2008
 *
 * dcraw.c -- Dave Coffin's raw photo decoder
 * Copyright 1997-2014 by Dave Coffin, dcoffin a cybercom o net
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <fastec/fas_log.h>

#include "dcraw_util.h"
#include "cs_trace.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

enum DCRAW_TRACE
{
    DCRAW_FUNC     = 0x00000010
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl = DCRAW_FUNC |
//                     TRACE_0;        // Set our trace level to TRACE_0 to trace.
                               TRACE_3;        // Set our trace level to TRACE_3 to not trace.


DcrawUtil::DcrawUtil() :
    image_ptr(nullptr)
{
    TP(DCRAW_FUNC, "%s", __PRETTY_FUNCTION__);
}

DcrawUtil::~DcrawUtil()
{
    TP(DCRAW_FUNC, "%s", __PRETTY_FUNCTION__);

    release();
}

void DcrawUtil::release()
{
    TP(DCRAW_FUNC, "%s", __PRETTY_FUNCTION__);

    if (image_ptr != nullptr)
    {
        LibRaw::dcraw_clear_mem(image_ptr);
        image_ptr = nullptr;
    }

    rawProcessor.recycle();
}

uint8_t *DcrawUtil::raw_process(uint8_t *dng_ptr, int dng_length)
{
    int ret;
    uint8_t *image_buffer = nullptr;

    TP(DCRAW_FUNC, "%s: dng_ptr=%p, dng_length=%d", __PRETTY_FUNCTION__, (void *)dng_ptr, dng_length);

    // Load it into our image processor.
    ret = rawProcessor.open_buffer(dng_ptr, dng_length);
    if (ret != LIBRAW_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: Cannot open_buffer(): %s", __PRETTY_FUNCTION__, libraw_strerror(ret));
        goto rp_done;
    }

    // Unpack the image.
    if ((ret = rawProcessor.unpack()) != LIBRAW_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: Cannot unpack(): %s", __PRETTY_FUNCTION__, libraw_strerror(ret));
        goto rp_done;
    }

    // We want to decode the image using the "-4" option of dcraw. From
    // the man page, that's equivalent to the options "-6 -W -g 1 1".
    // That means:
    // -W   Don't automatically brighten the image.
    // -6   Write 16-bit instead of 8 bit. This turns out to not work
    //      with the JPEG encoder.
    // -g 1 1   Use a linear gamma curve.
// #define ALM_USE_AUTO_WB 1
#ifdef ALM_USE_AUTO_WB
    rawProcessor.imgdata.params.use_auto_wb = 1;
#else
    rawProcessor.imgdata.params.gamm[0] = rawProcessor.imgdata.params.gamm[1] = 1;
    rawProcessor.imgdata.params.no_auto_bright = 1;
    rawProcessor.imgdata.params.use_camera_matrix = 1;
    rawProcessor.imgdata.params.use_camera_wb = 1;
    // rawProcessor.imgdata.params.output_bps = 16;
#endif /* !ALM_USE_AUTO_WB */

    // Process the data, converting it to an image.
    ret = rawProcessor.dcraw_process();
    if (ret != LIBRAW_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: Cannot do postpocessing: %s", __PRETTY_FUNCTION__, libraw_strerror(ret));
        if (LIBRAW_FATAL_ERROR(ret))
            goto rp_done;
    }

    // Get the image data.
    image_ptr = rawProcessor.dcraw_make_mem_image(&ret);
    if (ret != LIBRAW_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: dcraw_make_mem_image() failed: %s", __PRETTY_FUNCTION__, libraw_strerror(ret));
        goto rp_done;
    }

    // Get a pointer to the data.
    image_buffer = (uint8_t *)image_ptr->data;

rp_done:
    return image_buffer;
}

