/*!
 * @file fan_proc.cpp
 * @brief Manage shared usage of the fan controller.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fastec/fas_thread.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>
#include <fastec/fi_fpga_regs.h>

#include <algorithm>
#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>
#include <vector>

#include "camera_server.h"
#include "cs_trace.h"
#include "fan_proc.h"
#include "override.h"
#include "sensor.h"
#include "telemetry.h"

namespace FAN_PROC
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const uint32_t FAN_ON = 0;
const uint32_t FAN_OFF = 1;

const uint32_t DEFAULT_POLLING_TIME = 5000000;  //!< Default polling time in microseconds.

const uint8_t EMC2113_I2C_ADDR = 0x2E;          //!< I2C address of the fan controller on the IO board.

const uint8_t EMC2113_INT_TEMP_HIGH = 0x00;
const uint8_t EMC2113_INT_TEMP_LOW = 0x01;
const uint8_t EMC2113_EXT_DIODE1_HIGH = 0x02;
const uint8_t EMC2113_EXT_DIODE1_LOW = 0x03;

const uint8_t EMC2113_PUSHED_TEMP2 = 0x0D;
const uint8_t EMC2113_PWM_VALUE = 0x40;
const uint8_t EMC2113_TACH_TARGET_LOW = 0x4C;
const uint8_t EMC2113_TACH_TARGET_HIGH = 0x4D;
const uint8_t EMC2113_TACH_VALUE_HIGH = 0x4E;
const uint8_t EMC2113_TACH_VALUE_LOW = 0x4F;

const uint8_t EMC2113_CONFIG2 = 0x21;
const uint8_t EMC2113_FAN_STATUS = 0x27;
const uint8_t EMC2113_FAN_CONFIG1 = 0x42;
const uint8_t EMC2113_LUT_CONFIG = 0x50;

// Configuration 2 register bit settings.
const uint8_t CONFIG2_DIS_TO   = (1 << 5);
const uint8_t CONFIG2_DEFAULT  = (0x0E | CONFIG2_DIS_TO);

// Fan configuration 1 register bit settings.
const uint8_t FC1_UPDATE_400MS = (3 << 0);
const uint8_t FC1_EDGES_4POLE  = (3 << 3);
const uint8_t FC1_RANGE_1000   = (1 << 5);
const uint8_t FC1_EN_ALGO      = (1 << 7);
const uint8_t FC1_DEFAULT      = (FC1_UPDATE_400MS | FC1_EDGES_4POLE | FC1_RANGE_1000 | FC1_EN_ALGO);
const uint8_t FC1_DECONFIG     = (FC1_UPDATE_400MS | FC1_EDGES_4POLE | FC1_RANGE_1000);

// Default PWM value
const uint8_t PWM_PUSH_DEFAULT = 0xE0;

// PWM off value.
const uint8_t PWM_PUSH_OFF = 0;

// Look up table configuration register bit settings.
const uint8_t LUTC_PUSH2  = (1 << 0);
const uint8_t LUTC_TEMP3  = (1 << 1);
const uint8_t LUTC_TEMP1  = (1 << 2);
const uint8_t LUTC_PWM_DC = (1 << 3);
const uint8_t LUTC_PWM    = (1 << 4);
const uint8_t LUTC_LOCK   = (1 << 5);
const uint8_t LUTC_DTS_P2 = (1 << 6);
const uint8_t LUTC_DTS_P1 = (1 << 7);
const uint8_t LUTC_DECONFIG = 0;

//! For the configuration table, we setup the system:
//! LUTC_PWM_DC (b3=1, PUSH1_CFG) - If the...TEMP3_CFG bit is set (it is), the PWM Input Duty Cycle will be used.
//! LUTC_TEMP3 (b1=1, TEMP3_CFG) - Use the data in the PWM Input Duty Cycle register (because of LUTC_PWM_DC being set).
//! LUTC_PUSH2 (b0=1, TEMP4_CFG) - The data written into the Pushed Temperature 2 register is used by the Fan LUT.
//!
//! The FanTempPollThread::tick() method polls a number of different temperatures, and the maximum is written to the
//! Pushed Temperature 2 register (via the fan_set_temp method).
//!
//! The various inputs are:
//! T1: Fan controller's external diode.
//! T2: Software maximum value as described above.
//! T3: FPGA supplied value.
//! T4: Fan controller's internal temperature.
const uint8_t LUT_CONFIG[] =
{
    (LUTC_PWM_DC | LUTC_TEMP3 | LUTC_PUSH2),                // LUT Configuriation, UNLOCK.

    // Drive, T1, T2, T3, T4
    0x80, 0x32, 0x32, 0x32, 0x32,   // Setting 1: PWM = 128 (slow), T1:4=0x32 -> 50C/122F
    0x60, 0x38, 0x38, 0x38, 0x38,   // Setting 2: PWM = 96 (faster), T1:4=0x38 -> 56C/133F
    0x01, 0x3C, 0x3C, 0x3C, 0x3C,   // Setting 3: PWM = 1 (fastest), T1:4=0x3C -> 60C/140F
    0x00, 0xFF, 0xFF, 0xFF, 0xFF,   // Setting 4: Not used
    0x00, 0xFF, 0xFF, 0xFF, 0xFF,   // Setting 5: Not used
    0x00, 0xFF, 0xFF, 0xFF, 0xFF,   // Setting 6: Not used
    0x00, 0xFF, 0xFF, 0xFF, 0xFF,   // Setting 7: Not used
    0x00, 0xFF, 0xFF, 0xFF, 0xFF,   // Setting 8: Not used

    // Change the hysteresis for our set points.
    0x02, 0x02, 0x02, 0x02,         // Temperatures +/- 2 degrees C.

    (LUTC_LOCK | LUTC_PWM_DC | LUTC_TEMP3 | LUTC_PUSH2)     // LUT Configuriation, LOCK.
};

//! Message types for the fan controller.
enum FCTL_MESSAGE_TYPES
{
    FC_GET_TEMP = 0x30000,      //!< Read the temperature from the fan controller.
    FC_SET_TEMP,                //!< Send a temperature to the fan controller.
    FC_GET_STATS,               //!< Read the statistics from the fan controller.
    FC_CONTROL                  //!< Enable/disable the fan(s).
};

//! Message class for FC_CONTROL messages.
//! Note that this is an ASYNC message (i.e. non-blocking).
class fanControlMessage : public FAS_THREAD::Message
{
public:
    fanControlMessage();
    explicit fanControlMessage(bool stopFans);
    virtual ~fanControlMessage();

    bool disable;       //!< True to disable, false to enable.
};

//! Message class for FC_GET_TEMP messages.
//! Note that this is a SYNC message (i.e. blocking).
class fanGetTempMessage : public FAS_THREAD::SyncMessage
{
public:
    fanGetTempMessage();
    virtual ~fanGetTempMessage();

    double degrees_c;       //!< Temperature in degrees C.
};

//! Message class for FC_SET_TEMP messages.
//! Note that this is an ASYNC message (i.e. non-blocking).
class fanSetTempMessage : public FAS_THREAD::Message
{
public:
    fanSetTempMessage();
    explicit fanSetTempMessage(int temperature);
    virtual ~fanSetTempMessage();

    uint8_t degrees_c;      //!< Temperature in degrees C.
};

//! Message class for FC_GET_STATS messages.
//! Note that this is a SYNC message (i.e. blocking).
class fanGetStatsMessage : public FAS_THREAD::SyncMessage
{
public:
    fanGetStatsMessage();
    virtual ~fanGetStatsMessage();

    FAN_PROC::STATS_REQUEST request;
};

/*!
 * @brief A thread to poll temperature.
 */
class fanTempPollThread : public FAS_THREAD::Ticker
{
public:
    fanTempPollThread();
    ~fanTempPollThread();

    void tick() override;
};

/*!
 * @brief The fan controller object. This object is
 *    a singleton and is "local" to this file.
 */
class fanControlThread : public FAS_THREAD::MsgThread
{
private:
    static std::unique_ptr<fanControlThread> fan_proc;
    static std::mutex fct_mutex;
    fanControlThread();

public:
    virtual ~fanControlThread();
    virtual void messageHandle(FAS_THREAD::Message *message) override;

    // Singletons should not be cloneable.
    fanControlThread(fanControlThread& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const fanControlThread&) = delete;

    static fanControlThread *factory();
    static void close()
    {
        fanControlThread *thread_ptr;

        // Release our singleton.
        thread_ptr = fanControlThread::factory();
        if (thread_ptr != nullptr)
        {
            // Tell the thread that we are stopping.
            FAS_THREAD::SyncMessage msg;
            msg.req = FAS_THREAD::HALT;

            // Pass it on.
            msg.doSync(thread_ptr);

            // Try to join the thread.
            thread_ptr->join();
        }
    }

    static bool isRunning()
    {
        std::lock_guard<std::mutex> lock(fct_mutex);
        return ((fan_proc.get() != nullptr) && !fan_proc.get()->halted);
    }

    bool isDisconnected() { return disconnected; }

private:
    void configure();
    void disconnect();
    void disable();
    void enable();

    void handleControlRequest(fanControlMessage *msg);
    void handleGetTempRequest(fanGetTempMessage *msg);
    void handleSetTempRequest(fanSetTempMessage *msg);
    void handleGetStatsRequest(fanGetStatsMessage *msg);

    std::unique_ptr<fanTempPollThread> poller;
    bool disconnected;
};

} /* End of namespace FAN_PROC */

using namespace CS_TRACE;
using namespace FAN_PROC;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<fanControlThread> fanControlThread::fan_proc;
std::mutex fanControlThread::fct_mutex;

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

void fan_dump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Sanity check.
    if (fanControlThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return;
    }

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(fanControlThread::factory());
}

void fan_init()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_FAN_PROC;

    // By calling the factory, we'll create our singleton.
    fanControlThread::factory();
}

void fan_close()
{
    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Pass it on.
    fanControlThread::close();
}

bool fan_get_temp(double *degrees_c_ptr)
{
    bool rc(false);
    fanGetTempMessage msg;

    // Sanity check.
    if (degrees_c_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Nullptr!? No where to put the temperature!", __PRETTY_FUNCTION__);
        return false;
    }
    if (fanControlThread::factory() == nullptr)
    {
        // Maybe got a request before we're initialized or after we've shutdown.
        return false;
    }

    // Send the request.
    msg.doSync(fanControlThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the temperature we read.
        *degrees_c_ptr = msg.degrees_c;
        rc = true;
        TP(TRACE_SEL_FAN_PROC, "%s: Finished, got %dC", __PRETTY_FUNCTION__, *degrees_c_ptr);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

void fan_set_temp(int degrees_c)
{
    fanSetTempMessage *msg;

    // Sanity check.
    if (fanControlThread::factory() == nullptr)
    {
        // Maybe got a request before we're initialized or after we've shutdown.
        return;
    }

    // Allocate our message.
    msg = new fanSetTempMessage(degrees_c);
    if (msg == nullptr)
        return;

    // Add the message to the fan control thread.
    fanControlThread::factory()->messageAdd(msg);
}

bool fan_get_stats(FAN_PROC::STATS_REQUEST *request)
{
    bool rc(false);
    fanGetStatsMessage msg;

    // Sanity check.
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request?!", __PRETTY_FUNCTION__);
        return false;
    }
    if (fanControlThread::factory() == nullptr)
    {
        // Maybe got a request before we're initialized or after we've shutdown.
        return false;
    }

    // Build up our message.
    memcpy(&msg.request, request, sizeof(FAN_PROC::STATS_REQUEST));

    // Send the request.
    msg.doSync(fanControlThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the temperature we read.
        request->internal_temp_C = msg.request.internal_temp_C;
        request->external_temp_C = msg.request.external_temp_C;
        request->tach_reading = msg.request.tach_reading;
        request->pwm_setting = msg.request.pwm_setting;
        rc = true;
        TP(TRACE_SEL_FAN_PROC, "%s: Finished, internal=%6.2fC", __PRETTY_FUNCTION__, request->internal_temp_C);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

void fan_control(bool stopFans)
{
    fanControlMessage *msg;

    // Sanity check.
    if (fanControlThread::factory() == nullptr)
    {
        // Maybe got a request before we're initialized or after we've shutdown.
        return;
    }

    // Allocate our message.
    msg = new fanControlMessage(stopFans);
    if (msg == nullptr)
        return;

    // Add the message to the fan control thread.
    fanControlThread::factory()->messageAdd(msg);
}

uint32_t fan_get_status()
{
    if (fanControlThread::factory() == nullptr)
    {
        // Maybe got a request before we're initialized or after we've shutdown.
        return GIGE_FAN_DISABLE;
    }

    return (fanControlThread::factory()->isDisconnected() ? GIGE_FAN_DISABLE : GIGE_FAN_ENABLE);
}

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

fanGetTempMessage::fanGetTempMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)FC_GET_TEMP;
    degrees_c = 0;
}

fanGetTempMessage::~fanGetTempMessage()
{
}

fanControlMessage::fanControlMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)FC_CONTROL;
    disable = false;
}

fanControlMessage::fanControlMessage(bool stopFans) :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)FC_CONTROL;
    disable = stopFans;
}

fanControlMessage::~fanControlMessage()
{
}

fanSetTempMessage::fanSetTempMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)FC_SET_TEMP;
    degrees_c = 0;
}

fanSetTempMessage::fanSetTempMessage(int temperature) :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)FC_SET_TEMP;
    degrees_c = (uint8_t)temperature;
}

fanSetTempMessage::~fanSetTempMessage()
{
}

fanGetStatsMessage::fanGetStatsMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)FC_GET_STATS;
    memset(&request, 0, sizeof(STATS_REQUEST));
}

fanGetStatsMessage::~fanGetStatsMessage()
{
}

//--------------------------------------------------------------------
// fanControlThread Implementation
//--------------------------------------------------------------------

fanControlThread::fanControlThread() :
    FAS_THREAD::MsgThread(),
    disconnected(true)
{
    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Configure the fan controller.
    configure();

    // Create our polling thread.
    poller = std::unique_ptr<fanTempPollThread>(new fanTempPollThread());
    if (poller.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate polling thread!", __PRETTY_FUNCTION__);
    }
    else
    {
        // Start our poller.
        poller->start();
    }
}

fanControlThread::~fanControlThread()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // If the fan control thread is running, close it.
    if (fanControlThread::isRunning())
        fanControlThread::close();
}

fanControlThread *fanControlThread::factory()
{
    fanControlThread *rc(nullptr);

    std::lock_guard<std::mutex> loock(fct_mutex);

    // We only allow one thread object.
    if (fan_proc.get() == nullptr)
    {
        // Create our thread.
        fan_proc = std::unique_ptr<fanControlThread>(new fanControlThread());
        if (fan_proc.get() != nullptr)
            fan_proc.get()->start();

        // Return the pointer to the singleton.
        rc = fan_proc.get();
    }

    // Check to see if we were halted.
    else if (fan_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = fan_proc.get();
    }

    // Return our pointer.
    return rc;
}

void fanControlThread::disable()
{
    // Check to see if we've already disconnected because we only want to do
    // this the one time.
    if (!disconnected)
    {
        uint32_t reg_addr;
        FiFPGA *fpga_ptr = CameraServer::get_fpga();

        // Sanity check.
        if (fpga_ptr == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
            return;
        }

        // Disconnect the fan from the system.
        disconnect();

        // Tell the fan to turn off.
        reg_addr = EMC2113_PWM_VALUE;
        if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, PWM_PUSH_OFF, FI_FPGA::I2C_IO, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
            return;
        }

        // Tell the FPGA to turn off the controller's fan too.
        if (!fpga_ptr->regWrite(FAN_CONTROL, FAN_OFF))
        {
            fi_log(LOG_DEBUG, "%s: Error telling the FPGA's fan control to turn off.!", __PRETTY_FUNCTION__);
            return;
        }

        // This is important enough to let folks know it happened.
        fi_log(LOG_WARNING, "%s: All fans OFF.", __PRETTY_FUNCTION__);
    }
}

void fanControlThread::enable()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        return;
    }

    // Connect the fan controller.
    configure();

    // Tell the FPGA to turn on the controller's fan too.
    if (!fpga_ptr->regWrite(FAN_CONTROL, FAN_ON))
    {
        fi_log(LOG_ERR, "%s: Error telling the FPGA's fan control to turn on.!", __PRETTY_FUNCTION__);
        return;
    }

    // This is important enough to let folks know it happened.
    fi_log(LOG_INFO, "%s: All fans ON.", __PRETTY_FUNCTION__);
}

void fanControlThread::configure()
{
    int i;
    uint8_t reg_addr;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        return;
    }

#if 0
    // ALM: Not sure we want to disable the timeout. If something goes wrong, this could
    //      recover the bus...maybe?
    //
    // Disable the SMBus timeout function to be fully I2C compliant. Keep the other
    // default values.
    reg_addr = EMC2113_CONFIG2;
    if (!fpga_ptr->i2c_write(EMC2113_I2C_ADDR, reg_addr, (CONFIG2_DEFAULT | CONFIG2_DIS_TO), FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
        return;
    }
#endif

    // Configure the fan controller to use a 4-pole fan.
    reg_addr = EMC2113_FAN_CONFIG1;
    if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, FC1_DEFAULT, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
        return;
    }

    // Configure the LUT registers.
    for (reg_addr = EMC2113_LUT_CONFIG, i = 0; i < (int)sizeof(LUT_CONFIG); reg_addr++, i++)
    {
        if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, LUT_CONFIG[i], FI_FPGA::I2C_IO, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
            return;
        }
    }

    // Clear the status bits.
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_FAN_STATUS, &reg_addr, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error reading/clearing fan status!", __PRETTY_FUNCTION__);
        return;
    }

    // Let's make sure we're not in test mode.
    if (!fpga_ptr->regModify(NC1_FPGA_CONTROL, NC1FC_FAN_PWM_TEST, -1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=NC1_FPGA_CONTROL)!", __PRETTY_FUNCTION__);
        return;
    }

    // Wait a little bit to give things a chance to settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(100));

    if (!fpga_ptr->regModify(NC1_FPGA_CONTROL, 0, ~NC1FC_FAN_PWM_TEST))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=NC1_FPGA_CONTROL)!", __PRETTY_FUNCTION__);
        return;
    }

    // We are no longer disconnected.
    disconnected = false;
}

void fanControlThread::disconnect()
{
    uint8_t reg_addr;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        return;
    }

    // Unlock the lookup table.
    reg_addr = EMC2113_LUT_CONFIG;
    if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, LUTC_DECONFIG, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
        return;
    }

    // Configure the fan controller to use a 4-pole fan, but not the TACH algorithm.
    reg_addr = EMC2113_FAN_CONFIG1;
    if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, FC1_DECONFIG, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
        return;
    }

    // Tell the fan to run.
    reg_addr = EMC2113_PWM_VALUE;
    if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, reg_addr, PWM_PUSH_DEFAULT, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C (reg=0x%02X)!", __PRETTY_FUNCTION__, reg_addr);
        return;
    }

    // We are disconnected.
    disconnected = true;
}

/*!
 * @brief Process the messages for the thread.
 */
void fanControlThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case FC_GET_TEMP:
            handleGetTempRequest(static_cast<fanGetTempMessage *>(message));
            break;
        case FC_SET_TEMP:
            handleSetTempRequest(static_cast<fanSetTempMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;
        case FC_GET_STATS:
            handleGetStatsRequest(static_cast<fanGetStatsMessage *>(message));
            break;
        case FC_CONTROL:
            handleControlRequest(static_cast<fanControlMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case FAS_THREAD::DUMP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            // Shutdown our polling thread.
            if (poller.get() != nullptr)
            {
                poller->stop();
                poller->join();
                poller.reset();
            }

            // Disconnect from the fan controller.
            disconnect();

            // All done.
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Do we still have a message?
        if (message != nullptr)
        {
            // Forward on the messsage.
            message->forward();
        }
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in fanControlThread:\n%s", e.what().c_str());

        if (message != nullptr)
        {
            message->resp = FAS_THREAD::FAULT;
            message->forward();
        }
    }
}

/*!
 * @brief Read the temperature from the fan controller's sensor.
 */
void fanControlThread::handleGetTempRequest(fanGetTempMessage *msg)
{
    uint8_t temp[2];
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Read the high and low bytes for the external diode1.
    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_HIGH, &temp[0], FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_LOW, &temp[1], FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Convert to the full temperature.
    msg->degrees_c = ((double)(temp[0] & 0x7F)) + (((double)temp[1]) / 256.0);
    if ((temp[0] & 0x80) == 0x80)
        msg->degrees_c = -msg->degrees_c;

    // All done.
    msg->resp = FAS_THREAD::OK;
}

void fanControlThread::handleControlRequest(fanControlMessage *msg)
{
    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Are we disabling or enabling?
    if (msg->disable && !disconnected)
    {
        // Turn off the fan controller(s).
        disable();
    }
    else if (!msg->disable && disconnected)
    {
        // Turn on the fan controller(s).
        enable();
    }
}

/*!
 * @brief Push a temperature to the fan controller.
 */
void fanControlThread::handleSetTempRequest(fanSetTempMessage *msg)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    uint8_t degrees_c;
    uint32_t fan_temp;

    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Check for an override.
    if (override_u32(OVR_FAN_TEMPERATURE, &fan_temp) == true)
    {
        // Set the value to the override.
        degrees_c = (uint8_t)fan_temp;

        // Is the override telling us to stop the fan?
        if (fan_temp == 0)
        {
            LOG("%s: OVERRIDE and DISCONNECT! ====================", __PRETTY_FUNCTION__);
            disable();
        }
    }
    else
    {
        // Not an override, so use the past in value.
        degrees_c = (uint8_t)msg->degrees_c;
    }

    // Pass the temperture on to the fan controller.
    if (!fpga_ptr->i2cWrite(EMC2113_I2C_ADDR, EMC2113_PUSHED_TEMP2, degrees_c, FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Read the statistics from the fan controller.
 */
void fanControlThread::handleGetStatsRequest(fanGetStatsMessage *msg)
{
    uint8_t temp[2];
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_FAN_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Read the high and low bytes for the internal temperature reading.
    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_INT_TEMP_HIGH, temp, FI_FPGA::I2C_IO, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Convert to the full temperature.
    msg->request.internal_temp_C = ((double)(temp[0] & 0x7F)) + (((double)temp[1]) / 256.0);
    if ((temp[0] & 0x80) == 0x80)
        msg->request.internal_temp_C = -msg->request.internal_temp_C;

    // Read the high and low bytes for the external diode1.
    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_HIGH, temp, FI_FPGA::I2C_IO, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Convert to the full temperature.
    msg->request.external_temp_C = ((double)(temp[0] & 0x7F)) + (((double)temp[1]) / 256.0);
    if ((temp[0] & 0x80) == 0x80)
        msg->request.external_temp_C = -msg->request.external_temp_C;

    // Read the high and low bytes for the tach value.
    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_TACH_VALUE_HIGH, temp, FI_FPGA::I2C_IO, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }
    if ((temp[0] == 0xFF) && ((temp[1] & 0xF0) == 0xF0))
    {
        msg->request.tach_reading = 0;
    }
    else
    {
        double temp_val;
        temp_val = (double)((((uint32_t)temp[0]) << 5) | ((uint32_t)temp[1] >> 3));
        temp_val = (3932160.0e0 * 2.0e0) / temp_val;
        msg->request.tach_reading = (int)temp_val;
    }

    // Read the current PWM value.
    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(EMC2113_I2C_ADDR, EMC2113_PWM_VALUE, &temp[0], FI_FPGA::I2C_IO, 1))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }
    msg->request.pwm_setting = ((int)temp[0]);

    // All done.
    msg->resp = FAS_THREAD::OK;
}

//--------------------------------------------------------------------
// fanTempPollThread Implementation
//--------------------------------------------------------------------
fanTempPollThread::fanTempPollThread() :
    FAS_THREAD::Ticker(DEFAULT_POLLING_TIME)
{
    // traceControl |= TRACE_SEL_TIMER;
//    LOG("%s", __PRETTY_FUNCTION__);
}

fanTempPollThread::~fanTempPollThread()
{
//    LOG("%s", __PRETTY_FUNCTION__);
}

void fanTempPollThread::tick()
{
    std::vector<int> temps;
    Sensor *sensor_ptr = CameraServer::get_sensor();

    // Sanity check.
    if (sensor_ptr == nullptr)
        return;

    int em2120_temp = (int)telemetry::get_em2120_temperature();
    int fpga_temp = (int)telemetry::get_fpga_temperature();
    int firefly_temp = (int)telemetry::get_firefly_temperature();
    int sensor_temp = (int)sensor_ptr->get_temperature();
    int nvme_temp = telemetry::get_nvme_temp();
//    LOG("%s: em2120=%d, fpga=%d, firefly=%d, sensor=%d, nvme=%d", __PRETTY_FUNCTION__,
//        em2120_temp, fpga_temp, firefly_temp, sensor_temp, nvme_temp);

    // Read the temperatures we're interested in.
    temps.push_back(em2120_temp);
    temps.push_back(fpga_temp);
    temps.push_back(firefly_temp);
    temps.push_back(sensor_temp);
    temps.push_back(nvme_temp);

    // Pass the highest temperature on to the fan controller.
    fan_set_temp(*std::max_element(temps.begin(), temps.end()));
}
