/*!
 * @file fcp_trace.cpp
 *
 * @brief This file contains functions for tracing GigE FCP messages.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <ctype.h>
#include <time.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>
#include <fastec/port.h>

#include "cs_network.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// #define to turn on FCP transaction tracing.
#undef TRACE_FCP
//#define TRACE_FCP 1

#ifdef TRACE_FCP

////////////////////////////////////////////////////////////////////////////////
static bool use_read_includes = false;
static uint16_t include_reads[] =
{
    GIGE_REG_WIFI_IP_ADDR,  // 0xC314
    GIGE_REG_WIFI_SUBNET_MASK,  // 0xC318
    GIGE_REG_WIFI_DEF_GATEWAY,  // 0xC31C
    GIGE_REG_WIFI_IP_CUR_CFG,   // 0xC310
    GIGE_REG_WIFI_AP_CFG,   // 0xC588
    GIGE_REG_WIFI_STATUS,   // 0xC380
    GIGE_REG_WIFI_SSID,     // 0xC320
    GIGE_REG_WIFI_PASSPHRASE,   // 0xC340
    GIGE_REG_WIFI_START_SCAN,   // 0xC384

    GIGE_REG_AVAIL_MEDIA,       // 0xc068
    GIGE_REG_BATTERY_CHARGE,    // 0xc098
    GIGE_REG_FORMAT_MEDIA,      // 0xc06c
    GIGE_REG_GAMMA,             // 0xc03c
    GIGE_REG_JPEG_QUALITY,      // 0xc0ac
    GIGE_REG_AUTOSAVE,          // 0xa050
    GIGE_REG_SENSOR_GAIN,       // 0xe844
    GIGE_REG_USER_SAVE_START_NUM,   // 0xa054
    GIGE_REG_USER_SAVE_END_NUM,     // 0xa058
    GIGE_REG_SAVE_STATE,        // 0xc298
};
#define NUM_READ_INCLUDES ( sizeof( include_reads ) / sizeof( include_reads[0] ))

////////////////////////////////////////////////////////////////////////////////

static bool use_read_excludes = false;
static uint16_t exclude_reads[] =
{
    GIGE_REG_USER_DEF_NAME, // 0x00E8
    GIGE_REG_CCP,       // 0x0A00
    GIGE_REG_AUTOSAVE,      // 0xA050
    GIGE_REG_TRIGGER_REQ_POS,   // 0xC080
    GIGE_REG_PART_STATE,    // 0xC100
    GIGE_REG_IRIG_CONTROL,  // 0xC7A0
    GIGE_REG_IRIG_STATUS,   // 0xC7A4
    GIGE_REG_WIDTH,     // 0xD300
    GIGE_REG_HEIGHT,        // 0xD300
    GIGE_REG_FRAME_RATE,    // 0xE824
    GIGE_REG_SHUTTER_SPEED, // 0xE82C
    GIGE_REG_CAMERA_STATE,  // 0xE854
    GIGE_REG_SAVE_STATE,    // 0xE854
};
#define NUM_READ_EXCLUDES ( sizeof( exclude_reads ) / sizeof( exclude_reads[0] ))

////////////////////////////////////////////////////////////////////////////////

static bool use_write_excludes = true;
static uint16_t exclude_writes[] =
{
    GIGE_REG_CCP,       // 0x0A00
};
#define NUM_WRITE_EXCLUDES ( sizeof( exclude_writes ) / sizeof( exclude_writes[0] ))

////////////////////////////////////////////////////////////////////////////////

static bool use_write_includes = false;
static uint16_t include_writes[] =
{
    GIGE_REG_WIFI_IP_ADDR,  //
    GIGE_REG_WIFI_SUBNET_MASK,  //
    GIGE_REG_WIFI_DEF_GATEWAY,  //
    GIGE_REG_WIFI_IP_CUR_CFG,   //
    GIGE_REG_WIFI_AP_CFG,   //
    GIGE_REG_WIFI_STATUS,   //
    GIGE_REG_WIFI_SSID,     //
    GIGE_REG_WIFI_PASSPHRASE,   //
    GIGE_REG_WIFI_START_SCAN,   //
};
#define NUM_WRITE_INCLUDES ( sizeof( include_writes ) / sizeof( include_writes[0] ))

////////////////////////////////////////////////////////////////////////////////

// @brief Filter an address for a read operation
//
// Applies the read-include list first, if enabled, then applies the exclude
// list, if it is enabled. It's possible but not very useful to have both enabled.
//
// @param reg_addr  The address of the GigE register.
// @return true if it should be logged, false to ignore.
//
static bool read_filter(uint16_t reg_addr)
{
    bool    show(true);
    unsigned    i;

    if (use_read_includes)
    {
        show = false;
        for (i = 0; i < NUM_READ_INCLUDES; ++i)
            if (reg_addr == include_reads[i])
                break;

        // If found in the include list, do show it.
        if (i < NUM_READ_INCLUDES)
            show = true;
    }

    if (use_read_excludes)
    {
        for (i = 0; i < NUM_READ_EXCLUDES; ++i)
            if (reg_addr == exclude_reads[i])
                break;

        // If found in the exclude list, don't show it.
        if (i < NUM_READ_EXCLUDES)
            show = false;
    }

    return show;
}

// @brief Filter an address for a write operation
//
// Applies the include list first, if enabled, then applies the exclude
// list, if it is enabled. It's possible but not very useful to have both enabled.
//
// @param reg_addr  The address of the GigE register.
// @return true if it should be logged, false to ignore.
//
static bool write_filter(uint16_t reg_addr)
{
    bool show(true);
    unsigned i;

    if (use_write_includes)
    {
        show = false;
        for (i = 0; i < NUM_WRITE_INCLUDES; ++i)
            if (reg_addr == include_writes[i])
                break;
        if (i < NUM_WRITE_INCLUDES)
            show = true;
    }

    if (use_write_excludes)
    {
        for (i = 0; i < NUM_WRITE_EXCLUDES; ++i)
            if (reg_addr == exclude_writes[i])
                break;

        // If found in the exclude list, don't show it.
        if (i < NUM_WRITE_EXCLUDES)
            show = false;
    }

    return show;
}
#endif

/*!
 * @brief Trace FCP messages for debug.
 *
 * @param p_msg The message received.
 */
void fcp_trace(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    (void)p_msg;

// If not enabled, we will just return immediately.
#ifdef TRACE_FCP
    static bool show_next_ack = true;
    uint16_t    cmd;
    struct timespec t;
    char timestamp[32]; // 'YYYYMMDD HHMMSS.ssssss' (22 total characters in timestamp string)

    // Get the current time.
    fi_clock_gettime(CLOCK_REALTIME, &t);
    strftime(timestamp, 24, "%Y%m%d %H%M%S.", localtime(&t.tv_sec));
    sprintf(timestamp + 16, "%06ld", t.tv_nsec / 1000);

    // Even if it is a response, the command code is still in the same position.
    cmd = ntohs(p_msg->message.header.req.command);

    // First make sure it really is a FCP message.
    if ((p_msg->message.header.req.code != GIGE_REQ_CODE) && (cmd != GIGE_MSG_READREG_ACK))
    {
        fi_log(LOG_DEBUG, "%s: FCP: req.code=0x%x cmd=0x%x", timestamp, p_msg->message.header.req.code, cmd);
        return;
    }

    switch (cmd)
    {
    case GIGE_MSG_DISCOVERY_CMD:
        fi_log(LOG_DEBUG, "%s: FCP: DISCOVERY", timestamp);
        break;
    case GIGE_MSG_FORCEIP_CMD:
        fi_log(LOG_DEBUG, "%s: FCP: FORCEIP", timestamp);
        break;
    case GIGE_MSG_READREG_CMD:
    {
        unsigned    num_regs;
        unsigned    length;
        unsigned    i;
        uint32_t    reg_addr;

        length = (unsigned)ntohs(p_msg->message.header.req.length);
        num_regs = length / FCP_REG_SIZE; // each register is 4 long.
        if (num_regs == 1)
        {
            reg_addr = ntohl(p_msg->message.data_l[ 0 ]);
            if (read_filter(reg_addr))
            {
                fi_log(LOG_DEBUG, "%s: FCP: READREG 0x%04x", timestamp, reg_addr);
                show_next_ack = true;
            }
            else
            {
                show_next_ack = false;
            }
        }
        else
        {
            char buf[(7 * 4) + 1]; // " 0x0000" * 4 + 1
            unsigned j = 0;
            buf[0] = '\0';
            show_next_ack = false;
            for (i = 0; i < num_regs; i++)
            {
                reg_addr = ntohl(p_msg->message.data_l[ i ]);
                if (read_filter(reg_addr))
                    show_next_ack = true;
            }
            if (show_next_ack)
            {
                for (i = 0; i < num_regs; i++)
                {
                    reg_addr = ntohl(p_msg->message.data_l[ i ]);
                    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " 0x%04x",
                             reg_addr);
                    ++j;
                    if ((((j % 4) == 0) || (i == (num_regs - 1))) && (j > 0))
                    {
                        fi_log(LOG_DEBUG, "%s: FCP: READREGS[%d]%s", timestamp, num_regs, buf);
                        buf[0] = '\0';
                    }
                }
            }
        }
    }
    break;
    case GIGE_MSG_READREG_ACK:
        if (show_next_ack)
        {
            unsigned    num_regs;
            unsigned    length;
            unsigned    i;
            uint32_t    reg_val;

            length = (unsigned)ntohs(p_msg->message.header.resp.length);
            num_regs = length / FCP_REG_SIZE; // each register is 4 long.
            if (num_regs == 1)
            {
                reg_val = ntohl(p_msg->message.data_l[ 0 ]);
                fi_log(LOG_DEBUG, "%s: FCP: READREG_ACK 0x%08x", timestamp, reg_val);
            }
            else
            {
                char buf[(11 * 4) + 1]; // " 0x00000000" * 4 + 1
                unsigned j = 0;
                buf[0] = '\0';
                for (i = 0; i < num_regs; i++)
                {
                    reg_val = ntohl(p_msg->message.data_l[ i ]);
                    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " 0x%08x", reg_val);
                    ++j;

                    if ((((j % 4) == 0) || (i == (num_regs - 1))) && (j > 0))
                    {
                        fi_log(LOG_DEBUG, "%s: FCP: READREGS_ACK[%d]%s", timestamp, num_regs, buf);
                        buf[0] = '\0';
                    }
                }
            }
        }
        break;

    case GIGE_MSG_WRITEREG_CMD:
    {
        int     num_regs;
        int     length;
        int     i;
        uint32_t    reg_addr;
        uint32_t    reg_value;

        length = ntohs(p_msg->message.header.req.length);
        num_regs = length / (FCP_REG_SIZE + FCP_REG_DATA_SIZE);
        if (num_regs == 1)
        {
            reg_addr = ntohl(p_msg->message.data_l[ 0 ]);
            reg_value = ntohl(p_msg->message.data_l[ 1 ]);
            if (write_filter(reg_addr))
                fi_log(LOG_DEBUG, "%s: FCP: WRITEREG 0x%04x 0x%08x", timestamp, reg_addr, reg_value);
        }
        else
        {
            char buf[(16 * 4) + 1]; // " 0x0000:00000000" * 4 + 1
            unsigned j = 0;

            buf[0] = '\0';
            for (i = 0; i < num_regs; i++)
            {
                reg_addr = ntohl(p_msg->message.data_l[ i * 2 ]);
                reg_value = ntohl(p_msg->message.data_l[ i * 2 + 1 ]);
                if (write_filter(reg_addr))
                {
                    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf) - 1,
                             " 0x%04x:%08x", reg_addr, reg_value);
                    ++j;
                }
                if ((((j % 4) == 0) || (i == num_regs - 1)) && (j > 0))
                {
                    fi_log(LOG_DEBUG, "%s: FCP: WRITEREGS[%d]%s", timestamp, num_regs, buf);
                    buf[0] = '\0';
                }
            }
        }
    }
    break;
    case GIGE_MSG_READMEM_CMD:
    {
        uint32_t addr;

        addr = ntohl(p_msg->message.data_l[GIGE_READMEM_REQ_ADDR_L]);
        if (read_filter(addr))
            fi_log(LOG_DEBUG, "%s: FCP: READMEM 0x%08x", timestamp, addr);
    }
    break;
    case GIGE_MSG_WRITEMEM_CMD:
    {
        uint32_t addr;

        addr = ntohl(p_msg->message.data_l[GIGE_WRITEMEM_REQ_ADDR_L]);
        fi_log(LOG_DEBUG, "%s: FCP: WRITEMEM 0x%08x", timestamp, addr);
    }
    break;
    case GIGE_MSG_PACKETRESEND_CMD:
        fi_log(LOG_DEBUG, "%s: FCP: PACKETRESEND", timestamp);
        break;
    default:
        fi_log(LOG_DEBUG, "%s: FCP: unknown command 0x%x", timestamp, p_msg->message.header.req.command);
        break;
    }

    if (cmd != GIGE_MSG_READREG_CMD)
        show_next_ack = false;
#endif
}
