/*!
 * @file fi_display_drm.cpp
 *
 * @brief The DRM display object.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <va/va.h>
#include <va/va_drm.h>
#include <xf86drm.h>

#include <fastec/fas_log.h>

#include "fi_display.h"
#include "cs_trace.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

TRACE_GLOBAL;           //!< Use the global trace control word.

using namespace CS_TRACE;

//--------------------------------------------------------------------

FiDisplayDRM::FiDisplayDRM() :
    FiDisplay()
{
    // Turn on tracing.
    // traceControl |= TRACE_SEL_DISP_DRM;

    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);

    // Initialize our members.
    drm_fd = -1;
}

FiDisplayDRM::~FiDisplayDRM()
{
    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);

    // Make sure we clean up after ourselves.
    doClose();
}

VADisplay FiDisplayDRM::openDisplay()
{
    int i;
    static const char *drm_device_paths[] =
    {
        "/dev/dri/renderD128",
        "/dev/dri/card0",
        NULL
    };

    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);

    for (i = 0; drm_device_paths[i]; i++)
    {
        drmVersionPtr version;

        drm_fd = open(drm_device_paths[i], O_RDWR);
        if (drm_fd < 0)
            continue;

        version = drmGetVersion(drm_fd);
        if (!version)
        {
            close(drm_fd);
            continue;
        }
        if (!strncmp(version->name, "vgem", 4))
        {
            drmFreeVersion(version);
            close(drm_fd);
            continue;
        }
        drmFreeVersion(version);

        va_dpy = vaGetDisplayDRM(drm_fd);
        if (va_dpy)
        {
            TP(TRACE_SEL_DISP_DRM, "%s: Found '%s'", __PRETTY_FUNCTION__, drm_device_paths[i]);
            break;
        }

        close(drm_fd);
        drm_fd = -1;
    }

    // Return our display.
    return va_dpy;
}

void FiDisplayDRM::doClose()
{
    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);

    if (drm_fd < 0)
        return;

    va_dpy = nullptr;
    close(drm_fd);
    drm_fd = -1;

    FiDisplay::closeDisplay();
}

void FiDisplayDRM::closeDisplay()
{
    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);
    doClose();
}

VAStatus FiDisplayDRM::putSurface(VASurfaceID surface, const VARectangle *src_rect, const VARectangle *dst_rect)
{
    TP(TRACE_SEL_DISP_DRM, "%s", __PRETTY_FUNCTION__);

    (void)surface;
    (void)src_rect;
    (void)dst_rect;
    return VA_STATUS_ERROR_OPERATION_FAILED;
}

