/*!
 * @file fi_display_wayland.cpp
 *
 * @brief The Wayland display object.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include <va/va.h>
#include <va/va_wayland.h>

#include <fastec/fas_log.h>
#include <va/va_x11.h>

#include "fi_display.h"
#include "cs_trace.h"

namespace FI_DISPLAY_WAYLAND
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

typedef struct WAYLAND_DISPLAY
{
    struct wl_display          *display;
    struct wl_registry         *registry;
    struct wl_compositor       *compositor;
    struct wl_shell            *shell;
    struct wl_shell_surface    *shell_surface;
    struct wl_surface          *surface;
    unsigned int                ref_count;
    int                         event_fd;
} WAYLAND_DISPLAY;

};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

TRACE_GLOBAL;           //!< Use the global trace control word.

const std::string FiDisplayWayland::WL_COMPOSITOR("wl_compositor");
const std::string FiDisplayWayland::WL_SHELL("wl_shell");
void *FiDisplayWayland::wayland_display = nullptr;

const struct wl_registry_listener FiDisplayWayland::registry_listener =
{
    FiDisplayWayland::registry_handle_global,
    nullptr,
};

using namespace CS_TRACE;
using namespace FI_DISPLAY_WAYLAND;

//--------------------------------------------------------------------

FiDisplayWayland::FiDisplayWayland() :
    FiDisplay()
{
    // Turn on tracing.
    // traceControl |= TRACE_SEL_DISP_X11;
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);
}

FiDisplayWayland::~FiDisplayWayland()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Make sure we clean up after ourselves.
    doClose();
}

void FiDisplayWayland::registry_handle_global(void *data, struct wl_registry *registry, uint32_t id, const char *interface, uint32_t version)
{
    WAYLAND_DISPLAY *const d = (WAYLAND_DISPLAY *)data;
    std::string interface_str(interface);

    (void)version;

    if (interface_str.compare(WL_COMPOSITOR) == 0)
        d->compositor = (struct wl_compositor *)wl_registry_bind(registry, id, &wl_compositor_interface, 1);
    else if (interface_str.compare(WL_SHELL) == 0)
        d->shell = (struct wl_shell *)wl_registry_bind(registry, id, &wl_shell_interface, 1);
}

bool FiDisplayWayland::ensureWindow()
{
    WAYLAND_DISPLAY *d = (WAYLAND_DISPLAY *)wayland_display;

    if (!d->surface)
    {
        d->surface = wl_compositor_create_surface(d->compositor);
        if (!d->surface)
            return false;
    }

    if (!d->shell_surface)
    {
        d->shell_surface = wl_shell_get_shell_surface(d->shell, d->surface);
        if (!d->shell_surface)
            return false;
        wl_shell_surface_set_toplevel(d->shell_surface);
    }

    return true;
}

VADisplay FiDisplayWayland::openDisplay()
{
    WAYLAND_DISPLAY *d;

    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Do we already have our display allocated?
    if (wayland_display)
    {
        // Reuse our object.
        d = (WAYLAND_DISPLAY *)wayland_display;
        d->ref_count++;
    }
    else
    {
        // Allocate our display structure.
        d = new WAYLAND_DISPLAY;
        if (!d)
            return nullptr;
        memset(d, 0, sizeof(WAYLAND_DISPLAY));

        // Initialize the fields.
        d->event_fd = -1;
        d->display = wl_display_connect(nullptr);
        if (!d->display)
        {
            delete d;
            return nullptr;
        }

        wl_display_set_user_data(d->display, d);
        d->registry = wl_display_get_registry(d->display);
        wl_registry_add_listener(d->registry, &registry_listener, d);
        d->event_fd = wl_display_get_fd(d->display);
        wl_display_dispatch(d->display);

        d->ref_count = 1;
        wayland_display = d;
    }

    // Get a pointer to the display.
    va_dpy = vaGetDisplayWl(d->display);

    // Return our display.
    return va_dpy;
}

void FiDisplayWayland::doClose()
{
    WAYLAND_DISPLAY *d = (WAYLAND_DISPLAY *)wayland_display;

    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Make sure there is a display.
    if (!d)
        return;

    // Make sure this is the last reference to the display.
    if (--d->ref_count > 0)
        return;

    // Close out any element of the display structure that we've allocated.
    if (d->surface)
    {
        wl_surface_destroy(d->surface);
        d->surface = nullptr;
    }
    if (d->shell_surface)
    {
        wl_shell_surface_destroy(d->shell_surface);
        d->shell_surface = nullptr;
    }
    if (d->shell)
    {
        wl_shell_destroy(d->shell);
        d->shell = nullptr;
    }
    if (d->compositor)
    {
        wl_compositor_destroy(d->compositor);
        d->compositor = nullptr;
    }
    if (d->registry)
    {
        wl_registry_destroy(d->registry);
        d->registry = nullptr;
    }
    if (d->display)
    {
        wl_display_disconnect(d->display);
        d->display = nullptr;
    }

    // Release the memory.
    delete (WAYLAND_DISPLAY *)wayland_display;
    wayland_display = nullptr;

    // Pass it on up.
    FiDisplay::closeDisplay();
}

void FiDisplayWayland::closeDisplay()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);
    doClose();
}

VAStatus FiDisplayWayland::putSurface(VASurfaceID surface, const VARectangle */*src_rect*/, const VARectangle *dst_rect)
{
    WAYLAND_DISPLAY *d = (WAYLAND_DISPLAY *)wayland_display;
    VAStatus va_status;
    struct wl_buffer *buffer;

    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Make sure we are setup to operate.
    if (!ensureWindow())
        return VA_STATUS_ERROR_ALLOCATION_FAILED;

    va_status = vaGetSurfaceBufferWl(va_dpy, surface, VA_FRAME_PICTURE, &buffer);
    if (va_status != VA_STATUS_SUCCESS)
        return va_status;

    wl_surface_attach(d->surface, buffer, 0, 0);
    wl_surface_damage(d->surface, dst_rect->x, dst_rect->y, dst_rect->width, dst_rect->height);

    wl_surface_commit(d->surface);
    wl_display_flush(d->display);

    return VA_STATUS_SUCCESS;
}

