/*!
 * @file fi_display_x11.cpp
 *
 * @brief The X11 display object.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include <va/va.h>
#include <va/va_x11.h>

#include <fastec/fas_log.h>

#include "fi_display.h"
#include "cs_trace.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

TRACE_GLOBAL;           //!< Use the global trace control word.
using namespace CS_TRACE;

//--------------------------------------------------------------------

FiDisplayX11::FiDisplayX11() :
    FiDisplay()
{
    // Turn on tracing.
    // traceControl |= TRACE_SEL_DISP_X11;

    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Initialize our members.
    x11_display = nullptr;
    x11_window = None;
}

FiDisplayX11::~FiDisplayX11()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Make sure we clean up after ourselves.
    if ((x11_display != nullptr) || (va_dpy != nullptr))
        doClose();
}

bool FiDisplayX11::ensureWindow(uint32_t width, uint32_t height)
{
    Window win, rootwin;
    uint32_t black_pixel, white_pixel;
    int screen;

    if (!x11_display)
        return 0;

    if (x11_window)
    {
        XResizeWindow((Display *)x11_display, (Window)x11_window, width, height);
        return 1;
    }

    screen      = DefaultScreen((Display *)x11_display);
    rootwin     = RootWindow((Display *)x11_display, screen);
    black_pixel = BlackPixel((Display *)x11_display, screen);
    white_pixel = WhitePixel((Display *)x11_display, screen);

    win = XCreateSimpleWindow(
              (Display *)x11_display,
              rootwin,
              0, 0, width, height,
              1, black_pixel, white_pixel
          );
    if (!win)
        return 0;
    x11_window = (void *)win;

    XMapWindow((Display *)x11_display, (Window)x11_window);
    XSync((Display *)x11_display, False);
    return 1;
}

bool FiDisplayX11::validateRect(const VARectangle *rect)
{
    return (rect            &&
            rect->x >= 0    &&
            rect->y >= 0    &&
            rect->width > 0 &&
            rect->height > 0);
}

VADisplay FiDisplayX11::openDisplay()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    // Connect to the X server.
    x11_display = XOpenDisplay(NULL);
    if (!x11_display)
    {
        TP(TRACE_SEL_DISP_X11, "%s: can't connect to X server!", __PRETTY_FUNCTION__);
    }
    else
    {
        // Get a pointer to the display.
        va_dpy = vaGetDisplay((Display *)x11_display);
    }

    // Return our display.
    return va_dpy;
}

void FiDisplayX11::doClose()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    if (x11_display == nullptr)
        return;

    va_dpy = nullptr;
    if (x11_window)
    {
        XUnmapWindow((Display *)x11_display, (Window)x11_window);
        XDestroyWindow((Display *)x11_display, (Window)x11_window);
        x11_window = None;
    }
    XCloseDisplay((Display *)x11_display);
    x11_display = nullptr;

    FiDisplay::closeDisplay();
}

void FiDisplayX11::closeDisplay()
{
    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);
    doClose();
}

VAStatus FiDisplayX11::putSurface(VASurfaceID surface, const VARectangle *src_rect, const VARectangle *dst_rect)
{
    uint32_t win_width, win_height;

    TP(TRACE_SEL_DISP_X11, "%s", __PRETTY_FUNCTION__);

    if (!va_dpy)
        return VA_STATUS_ERROR_INVALID_DISPLAY;
    if (surface == VA_INVALID_SURFACE)
        return VA_STATUS_ERROR_INVALID_SURFACE;
    if (!validateRect(src_rect) || !validateRect(dst_rect))
        return VA_STATUS_ERROR_INVALID_PARAMETER;

    win_width  = dst_rect->x + dst_rect->width;
    win_height = dst_rect->y + dst_rect->height;
    if (!ensureWindow(win_width, win_height))
        return VA_STATUS_ERROR_ALLOCATION_FAILED;
    return vaPutSurface(va_dpy, surface, (Drawable)x11_window,
                        src_rect->x, src_rect->y,
                        src_rect->width, src_rect->height,
                        dst_rect->x, dst_rect->y,
                        dst_rect->width, dst_rect->height,
                        NULL, 0,
                        VA_FRAME_PICTURE);
}
