/*!
 * @file fi_internal.cpp
 * @brief This provides the functionality for the FiInternal class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <fastec/fas_log.h>

#include "camera_server.h"
#include "fi_internal.h"
#include "lic_manager.h"

// ==============================================================
// Static methods should be defined outside the class.
// ==============================================================

std::unique_ptr<FiInternal> FiInternal::instance;
std::mutex FiInternal::instanceMutex;
void *FiInternal::fasLib(nullptr);

std::string FiInternal::GET_PACKED_SIZE("fiGetPackedSize");
std::string FiInternal::GET_HANDLE("fiCreatePackStreamer");
std::string FiInternal::CLOSE_PACK_STREAMER("fiClosePackStreamer");
std::string FiInternal::IS_PACK_STREAMER_FINISHED("fiIsPackStreamerFinished");
std::string FiInternal::ADD_TIFF_IMAGE_DATA("fiAddTiffImageData");
std::string FiInternal::ADD_PLACEHOLDER("fiAddPlaceholder");
std::string FiInternal::REPLACE_FRAME("fiReplaceFrame");
std::string FiInternal::GET_PACK_STATISTICS("fiGetPackStatistics");
std::string FiInternal::CONVERT_TIFF("fiConvertTiffImageData");

std::string FiInternal::MA_GET_HANDLE("fiCreateMotionAmplifier");
std::string FiInternal::MA_CLOSE("fiCloseMotionAmplifier");
std::string FiInternal::MA_ADD_FRAME("fiMAAddFrame");
std::string FiInternal::MA_GET_FRAME("fiMAGetFrame");
std::string FiInternal::MA_GET_REF_FRAME("fiMAGetRefFrame");
std::string FiInternal::MA_IS_AVAIL("fiMAAvailable");
std::string FiInternal::MA_GET_WINDOW_SIZE("fiMAGetRefWindowSize");

std::string FiInternal::LIBRARY_NAME("libfas_internal.so");

FiInternal::FiInternal()
{
}

FiInternal::~FiInternal()
{
    // Did we open the library?
    if (fasLib != nullptr)
    {
        ::dlclose(fasLib);
        fasLib = nullptr;
    }
}

/*!
 * @brief This is the static method that controls the access to the singleton
 *        instance. On the first run, it creates a singleton object and places it
 *        into the static field. On subsequent runs, it returns the client existing
 *        object stored in the static field.
 */
FiInternal *FiInternal::factory()
{
    std::lock_guard<std::mutex> lock(instanceMutex);

    // Check to see if we have our singleton.
    if (instance.get() == nullptr)
    {
        // Attempt to open/load the library.
        fasLib = ::dlopen(LIBRARY_NAME.c_str(), RTLD_LAZY);
        if (fasLib != nullptr)
        {
            // Create the one and only instance.
            instance = std::unique_ptr<FiInternal>(new FiInternal());
            if (instance.get() == nullptr)
            {
                fi_log(LOG_ERR, "%s: Unable to allocate FiInternal object!", __PRETTY_FUNCTION__);
            }
        }
        else
        {
            fi_log(LOG_DEBUG, "%s: '%s' did not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        }
    }

    return instance.get();
}

FiInternal::GET_PACKED_SIZE_FN FiInternal::getPackedSizeFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (GET_PACKED_SIZE_FN)::dlsym(fasLib, GET_PACKED_SIZE.c_str());
}

FiInternal::GET_HANDLE_FN FiInternal::getHandleFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (GET_HANDLE_FN)::dlsym(fasLib, GET_HANDLE.c_str());
}

FiInternal::CLOSE_PACK_STREAMER_FN FiInternal::getClosePackFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (CLOSE_PACK_STREAMER_FN)::dlsym(fasLib, CLOSE_PACK_STREAMER.c_str());
}

FiInternal::IS_PACK_STREAMER_FINISHED_FN FiInternal::geIsFinishedFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (IS_PACK_STREAMER_FINISHED_FN)::dlsym(fasLib, IS_PACK_STREAMER_FINISHED.c_str());
}

FiInternal::ADD_TIFF_IMAGE_DATA_FN FiInternal::getAddTiffFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (ADD_TIFF_IMAGE_DATA_FN)::dlsym(fasLib, ADD_TIFF_IMAGE_DATA.c_str());
}

FiInternal::ADD_PLACEHOLDER_FN FiInternal::getAddPlaceholderFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (ADD_PLACEHOLDER_FN)::dlsym(fasLib, ADD_PLACEHOLDER.c_str());
}

FiInternal::REPLACE_FRAME_FN FiInternal::getReplaceFrameFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (REPLACE_FRAME_FN)::dlsym(fasLib, REPLACE_FRAME.c_str());
}

FiInternal::GET_PACK_STATISTICS_FN FiInternal::getGetStatisticsFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (GET_PACK_STATISTICS_FN)::dlsym(fasLib, GET_PACK_STATISTICS.c_str());
}

FiInternal::CONVERT_TIFF_FN FiInternal::getConvertTiffFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (CONVERT_TIFF_FN)::dlsym(fasLib, CONVERT_TIFF.c_str());
}

FiInternal::MA_GET_HANDLE_FN FiInternal::getMaHandleFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_GET_HANDLE_FN)::dlsym(fasLib, MA_GET_HANDLE.c_str());
}

FiInternal::MA_CLOSE_FN FiInternal::getMaCloseFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_CLOSE_FN)::dlsym(fasLib, MA_CLOSE.c_str());
}

FiInternal::MA_ADD_FRAME_FN FiInternal::getMaAddFrameFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_ADD_FRAME_FN)::dlsym(fasLib, MA_ADD_FRAME.c_str());
}

FiInternal::MA_GET_FRAME_FN FiInternal::getMaFrameFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_GET_FRAME_FN)::dlsym(fasLib, MA_GET_FRAME.c_str());
}

FiInternal::MA_GET_REF_FRAME_FN FiInternal::getMaRefFrameFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_GET_REF_FRAME_FN)::dlsym(fasLib, MA_GET_REF_FRAME.c_str());
}

FiInternal::MA_IS_AVAIL_FN FiInternal::getMaIsAvailFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_IS_AVAIL_FN)::dlsym(fasLib, MA_IS_AVAIL.c_str());
}

FiInternal::MA_GET_WINDOW_SIZE_FN FiInternal::getMaWinSizeFuncPtr()
{
    // Sanity check.
    if (!fasLib)
    {
        fi_log(LOG_DEBUG, "%s: '%s' is not loaded.", __PRETTY_FUNCTION__, LIBRARY_NAME);
        return nullptr;
    }

    // Return the function pointer.
    return (MA_GET_WINDOW_SIZE_FN)::dlsym(fasLib, MA_GET_WINDOW_SIZE.c_str());
}
