/*!
 * @file fi_libva.cpp
 *
 * @brief The FiLibVA class is an interface singleton to the intel/libva
 *        functionality.
 *
 * @note DRM is "Direct Rendering Manager".
 *
 * Derived extensively from intel/libva-utils package.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <va/va.h>
#include <va/va_x11.h>

#include <fastec/fas_log.h>

#include "fi_display.h"
#include "fi_libva.h"
#include "cs_trace.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

TRACE_GLOBAL;           //!< Use the global trace control word.
using namespace CS_TRACE;

std::unique_ptr<FiLibVA> FiLibVA::instance;
std::mutex FiLibVA::instance_mutex;

/*!
 * @brief The access method for the singleton object.
 * @return Pointer to the singleton object.
 */
FiLibVA *FiLibVA::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // Check to see if we have our singleton.
    if (instance.get() == nullptr)
    {
        // Create the one and only instance.
        instance = std::unique_ptr<FiLibVA>(new FiLibVA());
    }

    return instance.get();
}

//--------------------------------------------------------------------

/*!
 * @brief Constructor for the FiLibVA class.
 */
FiLibVA::FiLibVA()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_LIBVA;

    TP(TRACE_SEL_LIBVA, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief The destructor for the FiLibVA object.
 */
FiLibVA::~FiLibVA()
{
    TP(TRACE_SEL_LIBVA, "%s", __PRETTY_FUNCTION__);

    // Make sure we close everything out.
    vaCloseDisplay();
    vaCloseDRM();
}

/*!
 * @brief Get a VADisplay for a "display" device.
 * @return The VADisplay if we have a "display" device opened, nullptr otherwise.
 */
VADisplay FiLibVA::vaGetDisplay()
{
    if (va_display.get() == nullptr)
        return nullptr;
    return va_display->getDisplay();
}

/*!
 * @brief Get a VADisplay for a DRM ("Direct Rendering Manager") device.
 * @return The VADisplay if we have a DRM device opened, nullptr otherwise.
 */
VADisplay FiLibVA::vaGetDRM()
{
    if (va_drm.get() == nullptr)
        return nullptr;
    return va_drm->getDisplay();
}

/*!
 * @brief Open and initialize a "display" device to use with libva.
 * @return A VADisplay object if successful, or nullptr otherwise.
 */
VADisplay FiLibVA::vaOpenDisplay()
{
    VADisplay va_dpy;

    // VERY NOISY!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Do we already have a display?
    if (va_display.get() != nullptr)
    {
        va_dpy = va_display->getDisplay();
        goto va_bail;
    }

    // See if we can create a Wayland display.
    va_display = std::unique_ptr<FiDisplay>(static_cast<FiDisplay *>(new FiDisplayWayland));
    va_dpy = va_display->openDisplay();
    if (va_dpy == nullptr)
    {
        // See if we can create an X11 display.
        va_display = std::unique_ptr<FiDisplay>(static_cast<FiDisplay *>(new FiDisplayX11));
        va_dpy = va_display->openDisplay();
        if (va_dpy == nullptr)
        {
            TP(TRACE_SEL_LIBVA, "%s: Unable to open the display interface!", __PRETTY_FUNCTION__);
            va_display.reset();
            goto va_bail;
        }
    }

    // Sanity check.
    if (!vaDisplayIsValid(va_dpy))
    {
        fi_log(LOG_DEBUG, "%s: vaDisplayIsValid() failed!", __PRETTY_FUNCTION__);
        va_display->closeDisplay();
        va_display.reset();
        goto va_bail;
    }

    // We are good to go.
    TP(TRACE_SEL_LIBVA, "%s: Found display devvice.", __PRETTY_FUNCTION__);

va_bail:
    // Return our result.
    return va_dpy;
}

/*!
 * @brief Open a DRM (Direct Rendering Manager) context.
 *
 * This method opens a DRM device, and calls vaInitialize to get it setup to be
 * used for rendering operations.
 *
 * @return A VADisplay object to be used with libva functions.
 */
VADisplay FiLibVA::vaOpenDRM()
{
    VADisplay va_dpy;

    // VERY NOISY!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Do we already have a DRM instance?
    if (va_drm.get() != nullptr)
    {
        va_dpy = va_drm->getDisplay();
        goto va_bail;
    }

    // Allocate and try to open the DRM.
    va_drm = std::unique_ptr<FiDisplay>(static_cast<FiDisplay *>(new FiDisplayDRM));
    va_dpy = va_drm->openDisplay();
    if (va_dpy == nullptr)
    {
        TP(TRACE_SEL_LIBVA, "%s: Unable to open the DRM interface!", __PRETTY_FUNCTION__);
        va_drm.reset();
        goto va_bail;
    }

    // Sanity check.
    if (!vaDisplayIsValid(va_dpy))
    {
        fi_log(LOG_DEBUG, "%s: vaDisplayIsValid() failed!", __PRETTY_FUNCTION__);
        va_drm->closeDisplay();
        va_drm.reset();
        goto va_bail;
    }

    // We are good to go.
    TP(TRACE_SEL_LIBVA, "%s: Found DRM display", __PRETTY_FUNCTION__);

va_bail:
    // Return our result.
    return va_dpy;
}

/*!
 * @brief Close a display device, if one is open.
 */
void FiLibVA::vaCloseDisplay()
{
    TP(TRACE_SEL_LIBVA, "%s", __PRETTY_FUNCTION__);

    if (va_display.get() != nullptr)
    {
        va_display->closeDisplay();
        va_display.reset();
    }
}

/*!
 * @brief Close a DRM ("Direct Rendering Manager") device, if one is open.
 */
void FiLibVA::vaCloseDRM()
{
    TP(TRACE_SEL_LIBVA, "%s", __PRETTY_FUNCTION__);

    if (va_drm.get() != nullptr)
    {
        va_drm->closeDisplay();
        va_drm.reset();
    }
}

VAStatus FiLibVA::vaPutDisplaySurface(VASurfaceID surface, const VARectangle *src_rect, const VARectangle *dst_rect)
{
    if (va_display.get() == nullptr)
        return VA_STATUS_ERROR_INVALID_DISPLAY;

    return va_display->putSurface(surface, src_rect, dst_rect);
}

VAStatus FiLibVA::vaPutDRMSurface(VASurfaceID surface, const VARectangle *src_rect, const VARectangle *dst_rect)
{
    if (va_drm.get() == nullptr)
        return VA_STATUS_ERROR_INVALID_DISPLAY;

    return va_drm->putSurface(surface, src_rect, dst_rect);
}

