/*!
 * @file formatter.cpp
 * @brief This file implements the background formatting class.
 *
 * When a format operation or a sanitize operation completes,
 * we may have more work to do that doesn't fit well into the
 * format script. This replaces the old implementation that
 * would kick off the script but only check for completion on
 * GUI request.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>

#include <fcntl.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <chrono>
#include <memory>
#include <string>

#include <fmt/format.h>

#include "part_mgr.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "formatter.h"
#include "gige_reg.h"
#include "storage.h"
#include "override.h"

#include <fastec/fi_ioctl.h>
#include <fastec/popen2.h>
#include <fastec/fas_log.h>

namespace FI_FORMATTER
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Status of format operation.  Negative value is -errno and indicates
//! the operation completed with failure status.
enum FMT_STATUS
{
    FMT_RUNNING,        //!< Operation is running.
    FMT_SUCCESS,        //!< Operation completed successfully.
    FMT_STARTING        //!< Operation has not been initiated yet.
};

} /* End of namespace FI_FORMATTER */

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

//! @name Filesystem format names.
//! These are the names the format_and_build script uses.
//! @{
const char formatter::FAT32_FORMAT[] = "vfat";      //!< Format the device using FAT32 (or equivalent).
const char formatter::EXFAT_FORMAT[] = "exfat";     //!< Format the device using ExFAT.
const char formatter::EXT4_FORMAT[]  = "ext4";      //!< Format the device using EXT4.
const char formatter::FILR_FORMAT[]  = "filr";      //!< Format the device using the Fastec Long Record filesystem.
const char formatter::NTFS_FORMAT[]  = "ntfs";      //!< Format the device using NTFS.
// Space required.
//! @}

//! The location of the format_and_build script. This assumes the home path will be prepended.
const char formatter::EXEC_FILE[] = ".config/Fastec/scripts/format_and_build";

std::unique_ptr<formatter> formatter::fmt_instance; //!< The one and only formatter instance.
uint32_t formatter::format_process_media_sel;       //!< This is the media selector that was passed in
std::string formatter::fmt_block_dev;               //!< The block device path to the device we want to format (e.g., /dev/mmcblk0p1).
std::string formatter::volume_label;                //!< The volume label to use for the device.
int formatter::format_process_info_pid;             //!< The Process ID (PID) for the working thread. We get this from the popen2() call.
FILE *formatter::format_process_info_fp;            //!< The FILE pointer to read information from the working thread. We get this from the popen2() call.

using namespace FI_FORMATTER;

//--------------------------------------------------------------------
// Class formatter implementation
//--------------------------------------------------------------------

formatter::formatter() :
    FAS_THREAD::Thread(true)
{
    // Not running yet.
    thread_running = false;
    status = FMT_SUCCESS;

    // Initialize our values.
    format_process_media_sel = 0;
    format_process_info_pid = 0;
    format_process_info_fp = NULL;
}

formatter::~formatter()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Return the name of the file system type.
 * @param fs_type   One of the GIGE_FORMAT_XXX types.
 * @return The name used by the formatting script.
 */
const char *formatter::fs_name(uint32_t fs_type)
{
    const char *rc(nullptr);

    switch (fs_type)
    {
    case GIGE_FORMAT_FAT32:
        rc = FAT32_FORMAT;
        break;
    case GIGE_FORMAT_EXFAT:
        rc = EXFAT_FORMAT;
        break;
    case GIGE_FORMAT_EXT2:
        rc = EXT4_FORMAT;
        break;
    case GIGE_FORMAT_FILR:
        rc = FILR_FORMAT;
        break;
    case GIGE_FORMAT_NTFS:
        rc = NTFS_FORMAT;
        break;
    default:
        break;
    }

    return rc;
}

/*!
 * @brief Clean up the thread on our way out.
 */
void formatter::cleanup()
{
    // Indicate that we've stopped running.
    thread_running = false;
}

/*!
 * @brief Check the working thread to see its status.
 *
 * This will populate the formatter::status field as appropriate.
 *
 * @return True if we were successful in determining the status, false otherwise.
 */
bool formatter::check_format_status()
{
    int pstatus, rtn_pid;

    // Do we have a PID?
    if (format_process_info_pid == 0)
    {
        fi_log(LOG_WARNING, "%s: no PID!", __FUNCTION__);
        status = -ENXIO;
        return true;
    }

    // Is our process finished?
    rtn_pid = waitpid(format_process_info_pid, &pstatus, WNOHANG);
    if (rtn_pid == 0)
        return false;
    if (rtn_pid == -1)
    {
        status = -errno;
        fi_log(LOG_DEBUG, "%s: waitpid failed %d (%s)", __FUNCTION__, errno, strerror(errno));
    }
    // Extract the status information.
    else if (WIFEXITED(pstatus))
    {
        pstatus = WEXITSTATUS(pstatus);
        fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, pstatus, strerror(pstatus));

        switch (pstatus)
        {
        case 0:  // Success
            status = FMT_SUCCESS;
            break;
        case 1:  // Bad argument list
        case 2:  // Unrecognized filesystem type
            status = -EINVAL;
            break;
        case 3:  // Unrecognized device
            status = -ENXIO;
            break;
        case 5:  // fsdformat or mkfs failed
        case 6:  // mount failed after format completed
        case 7:  // mkdir hs_video failed
        case 8:  // mkdir DCIM/100fastc failed
        case 10: // Sanitize failed with error -1
        case 11: // Sanitize failed with error -2
        default: // (Most likely failure)
            status = -EIO;
            break;
        }
    }
    else if (WIFSIGNALED(pstatus))
    {
        pstatus = WTERMSIG(pstatus);
        fi_log(LOG_DEBUG, "%s: WTERMSIG=%d (%s)", __FUNCTION__, pstatus, strerror(pstatus));
        status = -EINTR;
    }

    // Close the pipe.
    if (format_process_info_fp != NULL)
    {
        pclose2(format_process_info_fp, format_process_info_pid);
        format_process_info_fp = NULL;
        format_process_info_pid = 0;
    }

    // We are done.
    fi_log(LOG_DEBUG, "%s: status=%d", __PRETTY_FUNCTION__, status);
    return true;
}

/*!
 * @brief Use popen2() to intiate a background thread running the format_and_build
 *        script.
 *
 * @return True if everything went well, false otherwise.
 */
bool formatter::initiate_format()
{
    bool rc;
    std::string temp_buf;

    // Clear our structure.
    format_process_info_pid = 0;
    format_process_info_fp = NULL;

    // Assume all will go well.
    rc = true;

    // Fire off the script to format and rebuild the device.
    temp_buf = fmt::format(FMT_STRING("{0}/{1} {2} {3} \"{4}\" {5:c}"), get_home_path(), EXEC_FILE,
                           fmt_block_dev,
                           fs_name(fs_type & ~GIGE_FORMAT_SANITIZE),
                           volume_label.c_str(),
                           (((fs_type & GIGE_FORMAT_SANITIZE) == GIGE_FORMAT_SANITIZE) ? '1' : '\0'));
    fi_log(LOG_DEBUG, "%s: %s", __PRETTY_FUNCTION__, temp_buf.c_str());

    format_process_info_fp = popen2(temp_buf.c_str(), "r", format_process_info_pid);
    if (format_process_info_fp == NULL)
    {
        format_process_info_pid = 0;
        status = -errno;
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Entry point for the formatter thread.
 *
 * This method assumes that all parameters have been setup, and only the act
 * of starting the format process is left.
 *
 * The parameters that are of concerned are:
 *
 * Name      | Description
 * --------- | ------------
 * formatter::format_process_media_sel | One of GIGE_MEDIA_xxx.
 * formatter::fs_type   | One of GIGE_FORMAT_xxx.
 * formatter::fmt_block_dev    | The device path (e.g. "/dev/mmcblk0p1").
 * formatter::volume_label     | The volume label (e.g. "FastecSD").
 *
 * @sa formatter::formatStart()
 */
void formatter::run()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Mark status as "not done yet".
    status = FMT_RUNNING;

    // We're ALIVE!
    thread_running = true;

    // Do we have a block device already configured?
    if (fmt_block_dev.empty())
    {
        std::string device;

        // Make sure the underlying device exists. This provides us with confidence
        // that everything is as it should be, and ALSO (very importantly) it
        // populates the block_dev array with the device name being used.
        storage::get_media_device(format_process_media_sel, device);
        if (device.empty())
        {
            fi_log(LOG_WARNING, "%s: Requested media (%d) doesn't exist.", __FUNCTION__, format_process_media_sel);
            status = -ENODEV;
            goto exit_format;
        }

        // Copy the block device name over.
        fmt_block_dev.assign(device);
    }

    // If we are in LTR mode and we had been running in Review mode previously,
    // the FPGA could still be left in LTR mode. We need to turn that off.
    if ((format_process_media_sel == GIGE_MEDIA_SSD) && ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE))
    {
        fpga_ptr->ltrEnable(false);
    }

    // Fire off the format operation.
    if (initiate_format())
    {
        // Try periodically to update our current status.
        while (!check_format_status())
        {
            // Sleep for a half-second.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(500));
        }
    }

    // At this point, status shows our completion result. If we were formatting
    // the SSD and we are in LTR mode, make sure we put down the partition table.
    //
    // ALM_TODO [FAS_FS]
    // Not necessarily in LR mode.
    // 1) Check to see if fs_type is FILR, then
    // 2) Check to see if we're in LR mode. If so -> ltr_format().
    // 3) Otherwise, looks like we want to create a STOR partition/capture space for CAP files from image memory.
    if ((status == FMT_SUCCESS) &&
        ((format_process_media_sel == GIGE_MEDIA_SSD) && ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)))
    {
        PartMgr *pm_ptr = CameraServer::get_partmgr();
        if (pm_ptr->ltrFormat(true, volume_label))
            status = FMT_SUCCESS;
        else
            status = -errno;
    }

exit_format:
    // Clean things up and leave.
    cleanup();
}

/*!
 * @brief Initiate a format operation on the device path specified.
 *
 * @param media_sel One of GIGE_MEDIA_xxx.
 * @param fs_type   One of GIGE_FORMAT_xxx.
 * @param device    The device path (e.g. "/dev/mmcblk0p1").
 * @param label     The volume label (e.g. "FastecSD").
 *
 * @return true if the format started OK, false otherwise.
 * @sa formatter::formatDone()
 */
bool formatter::formatStart(uint32_t media_sel, uint32_t fs_type, const char *dev_path, const char *label)
{
    int retry, max_retries;

    // Only get one instance.
    if (fmt_instance.get() == nullptr)
    {
        // Allocate our instance.
        fmt_instance = std::unique_ptr<formatter>(new formatter());
        if (fmt_instance.get() == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Unable to allocate the formatter instance!", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // We can't restart it if it's already running.
    if (fmt_instance->thread_running)
        return false;

    fmt_block_dev.clear();
    format_process_media_sel = media_sel;
    fmt_instance->fs_type = fs_type;
    fmt_instance->status = FMT_STARTING;  // Thread will set status when done or okay to wait.
    if (dev_path != nullptr)
    {
        fmt_block_dev.assign(dev_path);

        // Are we formatting an SSD?
        if (format_process_media_sel == GIGE_MEDIA_SSD)
        {
            int path_len = fmt_block_dev.length();

            // If the SSD is Fastec formatted, then there is no partition.
            // Make sure we append a partition to the device name.
            if ((fmt_block_dev.at(path_len - 2) != 'p') || (fmt_block_dev.at(path_len - 1) != '1'))
                fmt_block_dev.append("p1");
        }

    }
    volume_label.clear();
    if (label != nullptr)
        volume_label.assign(label);
    fmt_instance->start();

    // Wait for the thread to start.
    max_retries = 12;
    for (retry = 0; retry < max_retries; ++retry)
    {
        // Did the thread start yet?
        if (fmt_instance->status != FMT_STARTING)
            break;

        // Give it a chance to start.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(100));
    }
    if (retry == max_retries)
    {
        fi_log(LOG_ERR, "%s: Unable to start formatter thread.", __FUNCTION__);
        return false;
    }

    // If status shows an error, return that now. Otherwise the status
    // value will be FMT_IN_PROGRESS or FMT_SUCCESS.
    if (fmt_instance->status < 0)
        return false;
    else
        return true; // It apparently worked.
}

/*!
 * @brief Utility to determine if the format process has completed.
 *
 * The calling routine should poll this routine "every so often",
 * perhaps by using a timer event. When the Process ID of the formatting
 * task indicates that the task has completed, this routine will
 * return true.
 *
 * @param rtn_status    Pointer to a place for the status information.
 *
 * @return true if the process has finished, false otherwise.
 */
bool formatter::formatDone(int& rtn_status)
{
    bool rc = false;
    int cur_status;

    // If we don't have an instance, then we can be formatting!
    if (fmt_instance.get() == nullptr)
        return true;

    // Get the current status of the formatting process.
    cur_status = fmt_instance->status;
    if ((cur_status == FMT_RUNNING) || (cur_status == FMT_STARTING))
        goto fd_bail;

    // Did we finish with an error?
    if (cur_status < 0)
        rtn_status = -cur_status;
    else
        rtn_status = 0;

    // Let folks know that we are finished.
    rc = true;

fd_bail:
    return rc;
}

/*!
 * @brief Utility to determine if a format of the SSD is in progress.
 *
 * This is called when Arming to make sure that there is no format
 * operation going on.
 *
 * @return true if a format operation is in progress, false otherwise.
 *
 * @sa format_done
 */
bool formatter::isSsdFormatInProgress()
{
    int cur_status;

    if (!formatDone(cur_status) && (format_process_media_sel == GIGE_MEDIA_SSD))
        return true;
    else
        return false;
}

