/*!
 * @file fpga.cpp
 * @brief This source file implements the Fastec Imaging FPGA
 *    class, FiFPGA.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <atomic>
#include <chrono>
#include <mutex>

#include <fastec/interval_timer.h>
#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "bounce.h"
#include "camera_server.h"
#include "clock_gen.h"
#include "cs_trace.h"
#include "cs_utils.h"
#include "fpga.h"
#include "fpga_io.h"
#include "fcp.h"
#include "ipm_proc.h"
#include "ltr_worker.h"
#include "override.h"
#include "part_mgr.h"
#include "threads.h"

namespace FI_FPGA
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#define FPGA_DBG        (PDEBUG_HIGH)

const int   MAX_RETRY_COUNT = 10;       //!< Retry loop count.

// Number of FPGA clocks to wait for the trig_in signal to settle.
// There are 125 clocks per microsecond.
const uint32_t TRIG_IN_DEBOUNCE_DEFAULT = (1000 * 125);

// Number of FPGA clocks to wait for the trig_btn_in signal to settle.
// There are 125 clocks per microsecond.
const uint32_t TRIGBTN_IN_DEBOUNCE_DEFAULT = (8000 * 125);

// Number of FPGA clocks to wait for the sync_in signal to settle.
// There are 125 clocks per microsecond.
const uint32_t SYNC_IN_DEBOUNCE_DEFAULT = (1 * 125);

// Number of FPGA clocks to wait for the arm_in signal to settle.
// There are 125 clocks per microsecond.
const uint32_t ARM_IN_DEBOUNCE_DEFAULT  = (1000 * 125);

// Maximum duty-cycle is 99 percent.
const uint32_t MAX_DC_PERCENT = 99;

// Loop period for tracking the external sync-in signal.
const uint32_t SYNC_IN_TRACKING_PERIOD  = 2;

// Function of the external IO.
enum EXTIO_FUNC
{
    ARM_IN,
    ARM_OUT,
    SYNC_IN,
    SYNC_OUT,
    TRIG_IN,
    TRIG_OUT
};

// External IO Channels. Although it's clear what's going on here,
// I thought it might be "nicer" to have the constants.
const int EXTIO_CHANNEL_0 = (1 << 0);
const int EXTIO_CHANNEL_1 = (1 << 1);
const int EXTIO_CHANNEL_2 = (1 << 2);

//--------------------------------------------------------------------
// SyncIn worker thread
//--------------------------------------------------------------------

class syncInWorker : public FAS_THREAD::Thread
{
private:
    volatile bool thread_running;   //!< Flag to indicate if we are running or not.

    bool measure_only;
    uint32_t old_clock_speed;
    float old_frame_rate;
    uint32_t min_ticks;
    uint32_t sync_in_period;

    bool acquire();
    void track();

    void cleanup();

public:
    syncInWorker();
    virtual ~syncInWorker();

    virtual void run() override;
    virtual void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;
    virtual void stop() override;

    constexpr bool isMeasureOnly()      { return measure_only; }
    bool isRunning()                    { return thread_running; }
    uint32_t rawTiming();
    constexpr void setMeasureOnly(bool do_measure)  { measure_only = do_measure; }
};

//! The sync_tracker object is used to monitor and track the syncIn signal.
syncInWorker  sync_tracker;

} /* End of namespace FI_FPGA */

using namespace FI_FPGA;
using namespace FI_FPGA_IO;

//--------------------------------------------------------------------
// Static members
//--------------------------------------------------------------------
std::unique_ptr<FiFPGA> FiFPGA::instance;
std::mutex FiFPGA::instance_mutex;

//--------------------------------------------------------------------
// syncInWorker class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. This is NOT a
 *    detached thread, so it is expected that the
 *    parent will "join" at the end.
 */
syncInWorker::syncInWorker() :
    FAS_THREAD::Thread(false),
    old_clock_speed(0),
    old_frame_rate(0)
{
    // Initialize our member variables.
    thread_running = false;
    min_ticks = 0xFFFFFFFF;
    sync_in_period = 0;
    measure_only = false;
    stop_request = false;
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 *
 * @param arg   Not used.
 */
syncInWorker::~syncInWorker()
{
    syncInWorker::stop();
    join();
}

/*!
 * @brief Get the thread running.
 */
void syncInWorker::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Now start the thread.
    FAS_THREAD::Thread::start(threadFn);
}

/*!
 * @brief Ask the thread to stop.
 */
void syncInWorker::stop()
{
    // Tell the thread to stop.
    FAS_THREAD::Thread::stop();
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void syncInWorker::cleanup()
{
    // We're not running anymore.
    thread_running = false;
    measure_only = false;
    sync_in_period = 0;
}

bool syncInWorker::acquire()
{
    bool    signal_found;
    FiFPGA  *fpga_ptr;

    // Haven't found the signal yet.
    signal_found = false;

    // Reset our timing information.
    min_ticks = 0xFFFFFFFF;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // FIXME We don't want to do this if IRIG is enabled, since time_worker uses the measurement logic.
    // Setup the EXTIO control logic to measure the external SyncIn signal.
    fpga_ptr->regModify(EXT_IO_CONTROL, EXTIO_SYNC_PERIOD_MEAS_EXT_SYNC, ~EXTIO_SYNC_PERIOD_MEAS_MASK);

    // Read the measured minimum time. We want to throw away this first one.
    fpga_ptr->regRead(SYNC_IN_PERIOD_MIN, &sync_in_period);

    // Loop until we're asked to stop or found a signal.
    while (!stop_request)
    {
        // Give the logic a chance to run.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(500));

        // Read the register...see if we got something.
        fpga_ptr->regRead(SYNC_IN_PERIOD_MIN, &sync_in_period);
        if ((sync_in_period & SYNCIN_ACTIVE) == (uint32_t)SYNCIN_ACTIVE)
        {
            // Set our flag.
            signal_found = true;
//            LOG("%s: sync_in_period=0x%08x", __PRETTY_FUNCTION__, sync_in_period);

            // Do we want to start tracking? If we're only here to measure
            // the signal, then we don't want to break out.
            if (!measure_only)
                break;
        }
//        else
//        {
//            LOG("%s: sync-in not found.", __PRETTY_FUNCTION__);
//        }
    }

    // Return our status.
    return signal_found;
}

void syncInWorker::track()
{
    bool    do_update(false);
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;
    CameraServer *cs_ptr;

    // We shouldn't be here if we're only doing measurements.
    if (measure_only)
        return;

    // Get a pointer to our support classes.
    cs_ptr = CameraServer::get_cs();
    fpga_ptr = CameraServer::get_fpga();

    // Setup the EXTIO control logic to measure the external SyncIn signal.
    fpga_ptr->regModify(EXT_IO_CONTROL, EXTIO_SYNC_PERIOD_MEAS_EXT_SYNC, ~EXTIO_SYNC_PERIOD_MEAS_MASK);

    // Read one as a throw away.
    fpga_ptr->regRead(SYNC_IN_PERIOD_MIN, &sync_in_period);

    // Loop until we're asked to stop or the signal goes away.
    while (!stop_request)
    {
        bool signal_active(true);
        float frame_rate;
        uint32_t clock_speed;

        // Wait for a bit.
        sleep(SYNC_IN_TRACKING_PERIOD);

        // Are we supposed to stop?
        if (stop_request)
            break;

        // Do we still have a signal?
        fpga_ptr->regRead(SYNC_IN_PERIOD_MIN, &sync_in_period);
        signal_active = ((sync_in_period & SYNCIN_ACTIVE) == (uint32_t)SYNCIN_ACTIVE);

        // Do we still have a signal?
        if (!signal_active)
            break;

        // What is the desired frame rate?
        GIGE_REG_CACHE& regs = CameraServer::get_regs();
        frame_rate = cs_ptr->reg_to_frame_rate(regs[GIGE_REG_FRAME_RATE].get_value());

        // Did the frame rate change?
        if (!approx_equals(old_frame_rate, frame_rate))
        {
            // Since the frame rate changed, let's make sure we reset
            // our tracking information too.
            old_frame_rate = frame_rate;
            old_clock_speed = FPGA_CLOCK_SPEED;
        }

        // If the signal is still active, we assume our counter is
        // valid. Update our frame period value based on this new
        // measurement.

        // Get the measured clock speed.
        // NOTE: The measurement is in periods of FPGA_CLOCK_SPEED periods,
        //   NOT necessarily the same as SENSOR_CLOCK_SPEED period!
        clock_speed = sync_in_period & ((uint32_t)SYNCIN_PERIOD_MIN_MASK);

        // Make sure this is not zero!
        if (clock_speed != 0)
        {
            // Is this smaller?
            if (clock_speed < min_ticks)
            {
                // Save our new minimum tick value.
                min_ticks = clock_speed;

                // Make sure we update the FPGA.
                do_update = true;
            }

            // We need the FPGA to finish a little fast, so adjust
            // the clock speed to make things go quicker.
            clock_speed -= SYNC_IN_SLAVE_TUNING;

            // Make sure the new clock speed has slipped beyond a certain range.
            if ((clock_speed < (old_clock_speed - (SYNC_IN_SLAVE_TUNING >> 3))) ||
                (clock_speed > (old_clock_speed + (SYNC_IN_SLAVE_TUNING >> 3))))
            {
                // Everything looks OK. Adjust the frame rate and shutter
                // speed. Be careful about which to adjust first.
                if (clock_speed < old_clock_speed)
                {
                    // We are going faster, so adjust the shutter
                    // first.
                    fpga_ptr->icmSetShutterSpeed(fpga_ptr->icmShutterSpeed(), clock_speed / 1000000);
                    fpga_ptr->icmSetFrameRate(frame_rate, clock_speed);
                }
                else
                {
                    // We are going slower, so adjust the frame rate
                    // first.
                    fpga_ptr->icmSetFrameRate(frame_rate, clock_speed);
                    fpga_ptr->icmSetShutterSpeed(fpga_ptr->icmShutterSpeed(), clock_speed / 1000000);
                }

                // Remember this clock speed for the next time.
                old_clock_speed = clock_speed;
            }
        }

        // Do we want to update the FPGA?
        if (do_update)
        {
            // Sanity check our ticks.
            if (min_ticks != 0)
            {
                // We now have our minimum number of ticks between pulses.
                // Calculate our 1MHz tuning word by dividing 5e14 by the
                // number of 125MHz ticks we got.
                reg_value = (uint32_t)(0x1C6BF52634000LL / ((uint64_t)min_ticks));

                // Since we're in slave mode, we want the tuning word to be
                // "slightly" small, so our clock is "slightly" slow.
                reg_value -= SLAVE_TUNING_KLUDGE;

                // Update our slave value.
                fpga_ptr->regWrite(TIME_1MHZ_TUNING_WORD_SLAVE, reg_value);
            }
            else
            {
                fi_log(LOG_ERR, "%s: min_ticks is 0?!", __PRETTY_FUNCTION__);
            }
        }
    }
}

/*!
 * @brief Return the timing information from the SyncIn tracking measurement.
 *    If the tracker isn't running, return a 0.
 */
uint32_t syncInWorker::rawTiming()
{
    // We'll only have a timing value if we're running.
    if (!isRunning())
        return 0;

    // Make sure we have an active signal.
    if ((sync_in_period & SYNCIN_ACTIVE) != (uint32_t)SYNCIN_ACTIVE)
        return 0;

    // Return the last read period value.
    return (sync_in_period & ((uint32_t)SYNCIN_PERIOD_MIN_MASK));
}

/*!
 * @brief The worker thread.
 */
void syncInWorker::run()
{
    uint32_t    syncin_reg;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not only measuring.
    if (!measure_only)
    {
        // Only run if sync is enabled and 1-PPS.
        syncin_reg = regs[GIGE_REG_SYNC_IN_ENABLE].get_value();
        if ((syncin_reg & (GIGE_SYNC_TYPE_MASK | GIGE_SYNC_IN_PULSE_MASK)) != (GIGE_SYNC_TYPE_EXTERNAL | GIGE_SYNC_IN_PPS))
            goto bail1;
    }

    // We're running.
    thread_running = true;
    old_clock_speed = FPGA_CLOCK_SPEED;
    old_frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());

    // Run until the user wants us to stop.
    while (!stop_request)
    {
        // Acquire the sync-in signal.
        if (acquire())
        {
            // Once we've acquired the signal, track it.
            track();
        }
    }

bail1:
    // Cleanup on the way out.
    cleanup();
}

//============================================================================
// FiFPGA class implementation.
//============================================================================

/*!
 * @brief FiFPGA class constructor.
 */
FiFPGA::FiFPGA() :
    initialized(false),
    wd_timeout(false),
    brightness(GIGE_DEFAULT_BRIGHTNESS),
    contrast(GIGE_DEFAULT_CONTRAST),
    gamma(GIGE_DEFAULT_GAMMA),
    default_gamma(GIGE_DEFAULT_GAMMA),
    red_gain(GIGE_DEFAULT_RGB_GAIN),
    grn_gain(GIGE_DEFAULT_RGB_GAIN),
    blu_gain(GIGE_DEFAULT_RGB_GAIN),
    matrix_coeff1_red(0x1000), matrix_coeff2_red(0x000), matrix_coeff3_red(0x000),
    matrix_coeff1_grn(0x000), matrix_coeff2_grn(0x1000), matrix_coeff3_grn(0x000),
    matrix_coeff1_blu(0x000), matrix_coeff2_blu(0x000), matrix_coeff3_blu(0x1000),
    wb_red(0x1000), wb_green(0x1000), wb_blue(0x1000),
    dc_offset_red(0), dc_offset_grn(0), dc_offset_blu(0),
    ltr_mode_selected(false),
    ltr_save_error(0)
{
    // Initialize our version variable.
    fpga_ver = 0;
    fpga_shutter = 60; // We will set this when we actually set the hardware value.

    traceControl = 0 |
//                   FT_FUNC |
//                   FT_FPN |
                   FT_ICM |
//                   FT_ICM_NOISY |
//                   FT_IPM |
//                   FT_LTR |
//                   FT_SENSOR |
//                   FT_TIME |
//                   FT_REG |
//                   FT_I2C |
//                   FT_INT |
//                   FT_MUTEX_COUNT |
                   TRACE_0;      // Set the trace level to output information.
//                   TRACE_3;      // Turn off tracing.

    // Make sure our I/O worker thread is up.
    if (fpgaIOWorker::factory() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to start fpgaIOWorker thread!", __PRETTY_FUNCTION__);
    }

    // Intialize the bounce buffer logic.
    bb_init();
}

/*!
 * @brief This is the static method that controls access to the single
 *        instance of the FiFPGA object. On the first run, it creates
 *        our singleton object and places it into the static "instance"
 *        field. On subsequent runs, it returns the object.
 *
 * @return A pointer to the allocated object.
 */
FiFPGA *FiFPGA::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if (instance.get() == nullptr)
    {
        instance = std::unique_ptr<FiFPGA>(new FiFPGA());
    }

    // Return our instance.
    return instance.get();
}

/*!
 * @brief Initialize the interface to the FPGA hardware.
 *
 * Normally we would do this as part of the constructor, but it was
 * useful to allow for a return status. This is part one of a two
 * part initialization process, continued in startupInit().
 *
 * @return True on success, false otherwise.
 */
bool FiFPGA::interfaceInit()
{
    int i;
    CameraServer *cs_ptr = CameraServer::get_cs();
    clock_gen cg;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Start out by making sure the FPGA is in reset.
    regWrite(NC1_FPGA_CONTROL, NC1FC_EMIF_SW_RESET);

    // Initialize the clock.
    cg.clock_init();

    // Now that we have the clocks initialized, enable the DDR4 calibration by
    // taking the EMIF out of reset.
    regModify(NC1_FPGA_CONTROL, 0, ~NC1FC_EMIF_SW_RESET);
    for (i = 0; i < 10; ++i)
    {
        uint32_t reg_value;

        // Delay a little to let things time to settle down.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(500));

        regRead(NC1_FPGA_STATUS, &reg_value);
        LOG("%s: Waiting for EMIF calibration [0x%08x].", __PRETTY_FUNCTION__, reg_value);
        if ((reg_value & NC1FS_EMIF_LOCAL_CAL_SUCCESS) == NC1FS_EMIF_LOCAL_CAL_SUCCESS)
            break;
    }
    if (i == 10)
    {
        LOG("%s: EMIF CALIBRATION FAILED!!", __PRETTY_FUNCTION__);
        return false;
    }

    // Set the low light shutter speed. This will likely be a "worst case" value
    // until the sensor object gets setup.
    low_light_shutter = cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE);

    // Initialize our IPM processor thread.
    ipm_proc_init();

    // We are initialized.
    initialized = true;
    return true;
}

/*!
 * @brief FiFPGA class destructor.
 */
FiFPGA::~FiFPGA()
{
    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Check to see if we need to close things down.
    if (initialized)
    {
        shutdown();
    }
}

/*!
 * @brief Stop the FPGA interface.
 */
void FiFPGA::shutdown()
{
    LOG("%s++", __PRETTY_FUNCTION__);

    // Turn off the LEDs
    ledSetBlink(FI_FPGA::LED_OFF);

    // Ask the FPGA to shut things down.
    timeStop();
    startupStopAll();

    // Do we have our LTR worker thread going?
    if (ltr_thread.get() != nullptr)
        ltr_thread->stop();

    // Shutdown the IPM processor.
    ipm_proc_close();

    // Have the IO worker close.
    fpgaIOWorker::close();

    // We are no longer initialized.
    initialized = false;

    LOG("%s--", __PRETTY_FUNCTION__);
}

/*!
 * @brief Wait for a bit to achieve a requested state.
 *
 * @param reg_addr  [IN] The FPGA register to read.
 * @param bit       [IN] The bit to check.
 * @param set       [IN] True if we are waiting for the bit to be set, false for bit clear.
 * @param max_retries   [IN] The maximum number of test-sleep loop iterations.
 * @param sleep_usecs   [IN] The number of microseconds to sleep each loop
 * @param reg_value     [OUT, OPTIONAL] Pointer to put the value read from the register.
 *
 * @return The number of retries needed, which is max_retries if it failed
 */
int FiFPGA::waitForBit(uint32_t reg_addr, uint32_t bit, bool set, int max_retries, int sleep_usecs, uint32_t *value_ptr)
{
    int retries;
    uint32_t reg_value;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Loop upto the requested number of times.
    for (retries = 0; retries < max_retries; ++retries)
    {
        // Read the status register. If this fails, the read_reg() call will
        // log an error so we will just make it look like our standard failure.
        if (!regRead(reg_addr, &reg_value))
            return max_retries;

        // See if our bit went set or clear as requested.
        if (set && (((reg_value & bit) != 0)))
            break;
        else if (!set && (((reg_value & bit) == 0)))
            break;

        // No change, so wait for a bit.
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(sleep_usecs));
    }

    // Are we supposed to return the register value?
    if (value_ptr != nullptr)
        *value_ptr = reg_value;

    // Return the number of times we went around the loop.
    return retries;
}

/*!
 * @brief Make a best-effort attempt to clean up in case of restart.
 * @return True for success, false for failure.
 */
bool FiFPGA::startupStopAll()
{
    bool rc = true;
    uint32_t reg_value;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // This is basically what we used to do by calling dearm() at startup. It doesn't
    // seem adequate to recover if we were really armed, particularly in LTR mode.
    // But restarting the CameraServer without rebooting is a developer/lab
    // function so we won't fret over it.

    regRead(ICM_STATUS, &reg_value);
    if (reg_value & ICMS_ARMED)
    {
        int max_retries;
        int retries;

        fi_log(LOG_DEBUG, "%s: ICMC_UNARM, starting with ICM armed status=0x%08x", __FUNCTION__, reg_value);
        regModify(ICM_CONTROL, ICMC_UNARM, -1);

        // Wait to give the ICM a chance to release. 100 * 10 msecs = 1 second.
        max_retries = 100;
        retries = waitForBit(ICM_STATUS, ICMS_ARMED, true, max_retries, 10000, &reg_value);
        if (retries == max_retries)
        {
            fi_log(LOG_ERR, "%s: couldn't UNARM ICM, status=0x%08x", __FUNCTION__, reg_value);
            rc = false;
        }
    }

    // Clear the ICM LTR enable bit.
    regModify(ICM_CONTROL, 0, ~ICMC_LR_ENABLE);

    // Make sure this isnt't still set from a previous run.
    regWrite(ICM_SW_METADATA_MARKER, 0x00000000);

    // Make sure IPM is idle.
    regRead(IPM_STATUS, &reg_value);
    if (reg_value != 0)
    {
        // Not idle, so reset the module and see what happens.
        fi_log(LOG_DEBUG, "%s: starting with IPM_STATUS=0x%08x", __FUNCTION__, reg_value);
        ipmReset();

        // See if it idles...
        if (!ipmIsIdle())
        {
            // Didn't go  idle.
            fi_log(LOG_WARNING, "%s: IPM is not idle after reset.", __FUNCTION__);
        }
    }

    // Disable the ICM pipeline.
    icmPipelineDisable();

    // Turn off sensor readout.
    icmSensorReadoutEnable(false);

    return rc;
}

/*!
 * @brief Initialize the FPGA to be able to run.
 *
 * This is the second part of the initialization process which
 * was start in interfaceInit().
 */
void FiFPGA::startupInit()
{
    uint32_t reg_value;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the FPGA's version number and
    // put out an error if the version is less than the latest.
    regRead(FPGA_REVISION, &fpga_ver);
    if (fpga_ver < CURRENT_FPGA_VERSION)
    {
        fi_log(LOG_ERR, "%s: FPGA software version 0x%0x is not the latest version 0x%0x. Please upgrade.",
               __FUNCTION__, fpga_ver, CURRENT_FPGA_VERSION);
    }

    // Did we crash or otherwise cause a watchdog timeout event?
    regRead(NC1_FPGA_CONTROL, &reg_value);
    if ((reg_value & NC1FC_SW_FLAGS) == NC1FC_SW_ACTIVE)
    {
        fi_log(LOG_INFO, "%s: === WARM BOOT! ===", __PRETTY_FUNCTION__);
        wd_timeout = true;
    }

    // Remember that we've been here.
    regModify(NC1_FPGA_CONTROL, NC1FC_SW_ACTIVE, ~NC1FC_SW_FLAGS);

    // Start up our watchdog wakeup thread.
    CameraServer::watchdog();

    // Enable the watchdog timer logic. Make sure to reset the "expired" bit, in case
    // it's set. Clear by writing the expired bit.
    regWrite(WATCHDOG_CONTROL_STATUS, WD_EXPIRED);
    regWrite(WATCHDOG_CONTROL_STATUS, WD_ENABLE);

    // In case we are restarting, make sure nothing is still running.
    startupStopAll();

    // Initialize the IPM. This also resets the histogram and JPEG if they are not idle.
    ipmInit();

    // Set defaults in the IPM. We will initialize these for real once we load settings.

    // Setup the BCGG initial tables.
    ipmSendLutTableToFpga(GIGE_REG_RED);
    ipmSendLutTableToFpga(GIGE_REG_GREEN);
    ipmSendLutTableToFpga(GIGE_REG_BLUE);

    // set the matrix coefficients
    regWrite(IPM_P1_CC_MATRIX_COEFF1_RED, matrix_coeff1_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_RED, matrix_coeff2_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_RED, matrix_coeff3_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF1_GRN, matrix_coeff1_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_GRN, matrix_coeff2_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_GRN, matrix_coeff3_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF1_BLU, matrix_coeff1_blu);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_BLU, matrix_coeff2_blu);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_BLU, matrix_coeff3_blu);

    // Set the white balance (bayer gain).
    regWrite(IPM_P1_WB_GAIN_FACTOR_RED, wb_red);
    regWrite(IPM_P1_WB_GAIN_FACTOR_GRN, wb_green);
    regWrite(IPM_P1_WB_GAIN_FACTOR_BLU, wb_blue);

    // Mask out interrupts until we are ready.
    // Make sure we leave the PCIEIRQ_PWR_OFF interrupt enabled.
    shadow_irq = PCIEIRQ_FIFO_NOT_EMPTY | PCIEIRQ_IM2PCIE_DMA_DONE |
                 PCIEIRQ_PCIE2IM_DMA_DONE | PCIEIRQ_LR_DESC_FIFO_AF |
                 PCIEIRQ_EXT_ARM_INPUT | PCIEIRQ_EXT_UNARM_INPUT |
                 PCIEIRQ_MFT_LSDET | PCIEIRQ_MFT_LSLPSW |
                 PCIEIRQ_P1_FRAME_DONE | PCIEIRQ_TEST;
    regWrite(PCIE_IRQ_MASK, shadow_irq);

    // Initialize the debounce value for the trigger signals.
    regWrite(TRIG_IN_DEBOUNCE_TIME, TRIG_IN_DEBOUNCE_DEFAULT);
    regWrite(TRIGBTN_IN_DEBOUNCE_TIME, TRIGBTN_IN_DEBOUNCE_DEFAULT);

    // Debounce the ARM in signal.
    regWrite(ARM_IN_DEBOUNCE_TIME, ARM_IN_DEBOUNCE_DEFAULT);

    // Initialize the time module.
    timeInit();
}

/*!
 * @brief Disable (mask) a specified IRQ event.
 * @param irq_event [IN] The IRQ event we wish to disable.
 */
uint32_t FiFPGA::irqDisable(uint32_t irq_event)
{
    uint32_t previous;
    std::unique_lock<std::mutex> locker(class_mutex);

    TP(FT_INT, "%s: irq_event=0x%08x", __PRETTY_FUNCTION__, irq_event);

    previous = shadow_irq;
    shadow_irq |= irq_event;
    regWrite(PCIE_IRQ_MASK, shadow_irq);

    return previous;
}

/*!
 * @brief Enable (un-mask) a specified IRQ event.
 * @param irq_event [IN] The IRQ event we wish to enable.
 */
uint32_t FiFPGA::irqEnable(uint32_t irq_event)
{
    uint32_t previous;
    std::unique_lock<std::mutex> locker(class_mutex);

    TP(FT_INT, "%s: irq_event=0x%08x", __PRETTY_FUNCTION__, irq_event);

    previous = shadow_irq;
    shadow_irq &= ~irq_event;
    regWrite(PCIE_IRQ_MASK, shadow_irq);

    return previous;
}

/*!
 * @brief Logic to perform an I2C read of a register from a device.
 *
 * @param dev_addr  Address of the device we're interested in.
 * @param reg_addr  Address of the register we want to write, or other commands to be sent.
 * @param reg_data  Pointer to where to put the data.
 * @param bus       What bus we are talking to.
 * @param byte_count    Number of bytes we want to read.
 * @param readonly      True if we are only performing a read operations (some reads are actually read/modify/write).
 * @param cmd_count     Number of bytes in the command (some read operations require a command to be sent first).
 *
 * @return True if success, false otherwise.
 */
bool FiFPGA::i2cRead(uint8_t dev_addr, uint32_t reg_addr, uint8_t *reg_data, int bus, int byte_count, bool readonly, int cmd_count)
{
    bool rc;
    ioI2CAccessMessage msg;

    TP(FT_I2C, "%s: dev_addr=0x%02x, reg_addr=0x%08x/%d bus=%d byte_count=%d ro=%c", __PRETTY_FUNCTION__, dev_addr, reg_addr, cmd_count, bus, byte_count, (readonly ? 'T' : 'F'));

    // Sanity check.
    if (reg_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: reg_data is NULL!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize the structure for the write operation.
    msg.access.access = FI_I2C_RO;
    msg.access.bus = bus;
    msg.access.chip_addr = dev_addr;
    msg.access.cmd_count = cmd_count;
    msg.access.reg_addr = reg_addr;
    msg.access.byte_count = byte_count;
    msg.access.read_only = readonly;

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy the data over.
        for (int i = 0; i < byte_count; ++i)
            reg_data[i] = msg.access.reg_data.bytes[i];

        // All is well.
        rc = true;
    }
    else
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed bus:%d chip=0x%x reg:0x%x err:%d(%s)",
               __PRETTY_FUNCTION__, bus, dev_addr, reg_addr, errno, strerror(errno));
        rc = false;
    }

    TP(FT_I2C, "%s: bOK=%d, reg_data=0x%02x", __PRETTY_FUNCTION__, rc, *reg_data);

    // Return our status.
    return rc;
}

/*!
 * @brief Logic to perform an I2C write to a register of a device.
 *
 * @param dev_addr Address of the device we're interested in.
 * @param cmd_data Address of the register we want to write, or other commands to be sent.
 * @param reg_data Data we want written.
 *
 * @return True if success, false otherwise.
 */
bool FiFPGA::i2cWrite(uint8_t dev_addr, uint32_t cmd_data, uint32_t reg_data, int bus, int byte_count, int cmd_count)
{
    bool rc;
    ioI2CAccessMessage msg;

    TP(FT_I2C, "%s: dev_addr=0x%02x, cmd_data=0x%08x/%d, reg_data=0x%08x/%d, bus=%d", __PRETTY_FUNCTION__, dev_addr, cmd_data, cmd_count, reg_data, byte_count, bus);

    // Initialize the structure for the write operation.
    msg.access.access = FI_I2C_WO;
    msg.access.bus = bus;
    msg.access.chip_addr = dev_addr;
    msg.access.cmd_count = cmd_count;
    msg.access.reg_addr = cmd_data;
    msg.access.byte_count = byte_count;
    msg.access.reg_data.word = reg_data;

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // All is well.
        rc = true;
    }
    else
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed bus:%d chip=0x%x cmd:0x%x err:%d(%s)",
               __PRETTY_FUNCTION__, bus, dev_addr, cmd_data, errno, strerror(errno));
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Writes a value to a register in the FPGA.
 *
 * @param reg_addr  [IN] Register address we are writing to.
 * @param value     [IN] Value to write.
 *
 * @return True if everything went well, otherwise false.
 */
bool FiFPGA::regWrite(uint32_t reg_addr, uint32_t value)
{
    bool    rc;
    ioFpgaSetRegisterMessage msg;

    // Assume life is good.
    rc = true;

    // Initialize the structure for the write operation.
    msg.access.reg_id = reg_addr;
    msg.access.value = value;

    // See what writes we are logging.
    if (traceControl & (FT_REG | TRACE_0))
    {
        // Unless requested, we will skip the LUT writes.
        if (((IPM_P0_BCG_LUT_START <= reg_addr) && (reg_addr <= IPM_P0_BCG_LUT_END)) ||
            ((IPM_P1_BCG_LUT_START <= reg_addr) && (reg_addr <= IPM_P1_BCG_LUT_END)) ||
            ((IPM_P2_BCG_LUT_START <= reg_addr) && (reg_addr <= IPM_P2_BCG_LUT_END)))
        {
            TP(FT_REG, "%s(LUT): reg:0x%08x = value:0x%08x", __FUNCTION__, reg_addr, value);
        }
        else
        {
            void *caller = __builtin_return_address(0);
            TP(FT_REG, "%s(%p): reg:0x%08x = value:0x%08x", __FUNCTION__, (void *)caller, reg_addr, value);
        }
    }

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Was there a problem?
    if (msg.resp != FAS_THREAD::OK)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x err:%d(%s)", __PRETTY_FUNCTION__, reg_addr, value, errno, strerror(errno));
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Read the value from the specified FPGA register.
 *
 * @param reg_addr  [IN] Register address we want to read from.
 * @param value     [OUT] Pointer to where we want to write the data to.
 *
 * @return True if everything went well, otherwise false.
 */
bool FiFPGA::regRead(uint32_t reg_addr, uint32_t *value_ptr)
{
    ioFpgaGetRegisterMessage msg;

    // Make sure we've got our destination pointer.
    if (value_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Failed due to value_ptr==NULL", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize the structure for the read operation.
    msg.access.reg_id = reg_addr;
    msg.access.value = 0;

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the temperature we read.
        *value_ptr = msg.access.value;
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // See what reads we are logging.
    if ((traceControl & (FT_REG | TRACE_0)) == (FT_REG | TRACE_0))
    {
        if ((TIME_SBS_COUNTER_LATCHED == reg_addr) || (reg_addr == TIME_SBUS_COUNTER_LATCHED))
        {
            TP(FT_REG, "%s(TIME): reg:0x%08x = value:0x%08x", __FUNCTION__, reg_addr, *value_ptr);
        }
        // Handle other FPGA registers.
        else
        {
            void *caller = __builtin_return_address(0);
            TP(FT_REG, "%s(%p): reg:0x%08x = value:0x%08x", __FUNCTION__, (void *)caller, reg_addr, *value_ptr);
        }
    }

    return true;
}

/*!
 * @brief Wait for an "OK" status before reading the value from the specified FPGA register.
 *
 * @param status_reg    [IN] Register address of the status register.
 * @param mask          [IN] Mask for the bit(s) we are interested in.
 * @param ok_value      [IN] Value which means we are OK to proceed.
 * @param reg_addr      [IN] Register address we want to read from.
 * @param value_ptr     [OUT] (REQUIRED) Pointer to where we want to write the data to.
 *
 * @return True if everything went well, otherwise false.
 */
bool FiFPGA::regReadWait(uint32_t status_reg, uint32_t mask, uint32_t ok_value, uint32_t reg_addr, uint32_t *value_ptr)
{
    int check;
    bool rc = false;
    static int max_check = 0;

    TP(FT_REG, "%s: status_reg=0x%08x, mask=0x%08x, ok_value=0x%08x, reg_addr=0x%08x", __PRETTY_FUNCTION__,
       status_reg, mask, ok_value, reg_addr);

    // Sanity check.
    if (value_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: The return value_ptr must be specified!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // We will only do this for a maximum of 100 times.
    for (check = 0; check < 100; ++check)
    {
        uint32_t reg_val;

        if (!regRead(status_reg, &reg_val))
            goto bail;
        if ((reg_val & mask) == ok_value)
            break;
    }
    if (check > 100)
        goto bail;

    if (check > max_check)
    {
        max_check = check;
        fi_log(LOG_DEBUG, "%s: max_check = %d/100\n", __PRETTY_FUNCTION__, max_check);
    }

    // It's now OK to read the register. If there is no
    rc = regRead(reg_addr, value_ptr);

bail:
    return rc;
}

/*!
 * @brief Peform a Read-Modify-Write operation on the FPGA.
 *
 * Performs:
 *      new_reg = (old_reg & mask) | value
 *
 * To set a bit:
 * ~~~~~~~~~~
 * regModify(reg_name, BIT_NAME, -1);
 * ~~~~~~~~~~
 *
 * To clear a bit:
 * ~~~~~~~~~~
 * regModify(reg_name, 0, ~BIT_NAME);
 * ~~~~~~~~~~
 *
 * @param reg_addr  [IN] Our FPGA register we'll be modifying.
 * @param value     [IN] Value to set into the register.
 * @param mask      [IN] Value to mask with the existing contents of the register.
 *
 * @return True if everything went well, otherwise false.
 */
bool FiFPGA::regModify(uint32_t reg_addr, uint32_t value, uint32_t mask)
{
    bool    rc;
    ioFpgaModifyRegisterMessage msg;

    // Assume life is good.
    rc = true;

    // Initialize the structure for the modify operation.
    msg.modify.reg_id = reg_addr;
    msg.modify.value = value;
    msg.modify.mask = mask;
    TP(FT_REG, "%s: reg:0x%x value:0x%x mask:0x%x", __PRETTY_FUNCTION__, reg_addr, value, mask);

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Was there a problem?
    if (msg.resp != FAS_THREAD::OK)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x mask:0x%x err:%d(%s)", __PRETTY_FUNCTION__, reg_addr, value, mask, errno, strerror(errno));
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Read a buffer from the image memory.
 *
 * @param rd_dma_mod    The Read DMA Modifier saying how the DMA should be done.
 * @param rd_dma_dims   The frame dimensions, used only if rd_dma_mod specifies an image mode
 * @param src_addr  The Address in image memory, absolute, NOT relative to partition base.
 * @param buf       The pointer to the buffer that is to be written.
 * @param buf_size  How many bytes of the data to be read.
 * @param img_size  [OPTIONAL] The size of the image buffer. Useful when expanding to 16-bit values.
 *
 * @return The amount of data read, or 0 if an error occurred.
 */
uint32_t FiFPGA::imageMemRead(
    uint32_t rd_dma_mod,
    uint32_t rd_dma_dims,
    uint64_t src_addr,
    uint8_t *buf,
    uint32_t buf_size,
    uint32_t img_size)
{
    uint32_t src_size;
    uint32_t dst_size;

    int bb_index;
    int status;
    uint32_t read_count = 0;

    TP(FT_FUNC, "%s: dma_mod=0x%x, dma_dims=0x%x, src=0x%" PRIx64 ", buf=%p buf_size=0x%x img_size=0x%x",
       __PRETTY_FUNCTION__, rd_dma_mod, rd_dma_dims, src_addr, (void *)buf, buf_size, img_size);

    // Get an available buffer.
    bb_index = bb_get_avail();
    if (bb_index < 0)
    {
        fi_log(LOG_DEBUG, "%s: No bounce buffers available!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Need to do some switcheroo if the img_size is specified.
    if (img_size != 0)
    {
        src_size = img_size;
        dst_size = buf_size;
    }
    else
    {
        src_size = buf_size;
        dst_size = buf_size;
    }

    // DMA the data into our bounce buffer.
    status = imageDownload(rd_dma_mod, rd_dma_dims, src_addr, bb_get_avail_addr(bb_index), src_size, dst_size);
    if (status >= 0)
    {
        // Transfer the image from our bounce buffer.
        read_count = bb_read(bb_index, buf, buf_size);
    }

bail:
    // Release the buffer.
    if (bb_index >= 0)
        bb_release(bb_index);

    TP(FT_FUNC, "%s: Done", __PRETTY_FUNCTION__);
    return read_count;
}

/*!
 * @brief Write a buffer to the FPGA's DDR2 memory.
 *
 * @param part_num  The 0 based partition number or (uint16_t)-1 for address offset = 0.
 * @param dest_addr The address in DDR2 relative to the begining of the partition in QWord address.
 * @param buf       The pointer to the buffer that is to be written.
 * @param buf_size  The size in bytes of the data to be written.
 * @param do_swap   TRUE if we want to swap the data going out.
 *
 * @return The amount of data written, or 0 if an error occurred.
 */
uint32_t FiFPGA::imageMemWrite(uint64_t dest_addr, const uint8_t *buf, uint32_t buf_size, bool do_swap)
{
    int bb_index;
    int status;
    uint32_t write_count = 0;
    uint32_t dma_mod;

    TP(FT_FUNC, "%s: dest_addr=0x%" PRIx64 ", buf=%p, buf_size=0x%08x, do_swap='%c'", __PRETTY_FUNCTION__,
       dest_addr, (void *)buf, buf_size, (do_swap ? 'T' : 'F'));

#if 0
    {
        uint32_t *big_ptr = (uint32_t *)buf;

        LOG("%s: 0x%08x 0x%08x 0x%08x 0x%08x", __PRETTY_FUNCTION__, big_ptr[0], big_ptr[1], big_ptr[2], big_ptr[3]);
        LOG("%s: 0x%02x 0x%02x 0x%02x 0x%02x", __PRETTY_FUNCTION__, buf[0], buf[1], buf[2], buf[3]);
    }
#endif

    // Get an available buffer.
    bb_index = bb_get_avail();
    if (bb_index < 0)
    {
        fi_log(LOG_DEBUG, "%s: No bounce buffers available!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Transfer the image to our bounce buffer.
    write_count = bb_write(bb_index, buf, buf_size);
    if (write_count != buf_size)
    {
        fi_log(LOG_DEBUG, "%s: bb_write() failed [%u]", __PRETTY_FUNCTION__, write_count);
        goto bail;
    }

    // Setup our DMA settings.
    if (do_swap)
    {
        dma_mod = PCIE2IM_DMA_WORD_ORDER_L2R | PCIE2IM_DMA_BYTE_ORDER_R2L;
    }
    else
    {
        dma_mod = PCIE2IM_DMA_WORD_ORDER_R2L | PCIE2IM_DMA_BYTE_ORDER_R2L;
    }

    // DMA the data to image memory.
    status = imageUpload(dma_mod, bb_get_avail_addr(bb_index), dest_addr, buf_size);
    if (status < 0)
    {
        fi_log(LOG_ERR, "%s: upload_image() failed [%d]", __PRETTY_FUNCTION__, status);
        goto bail;
    }

#if 0
    {
        int i;
        int j;
        uint8_t *im_char;
        uint32_t im_data[4];

        im_char = (uint8_t *)im_data;
        for (i = 0; i < 48; i += 16)
        {
            char dump[128];

            im_data[0] = im_peek(dest_addr + i);
            im_data[1] = im_peek(dest_addr + i + 4);
            im_data[2] = im_peek(dest_addr + i + 8);
            im_data[3] = im_peek(dest_addr + i + 12);

            LOG("%s: 0x%08x 0x%08x 0x%08x 0x%08x", __PRETTY_FUNCTION__, im_data[0], im_data[1], im_data[2], im_data[3]);
            dump[0] = 0;
            for (j = 0; j < 16; j++)
                sprintf(dump, "%s 0x%02x", dump, im_char[j]);

            LOG("%s:%s", __PRETTY_FUNCTION__, dump);
        }
    }
#endif

bail:
    // Release the buffer.
    if (bb_index >= 0)
        bb_release(bb_index);

    // All is well.
    return write_count;
}

/*!
 * @brief Write the "blink" rate for the camera trigger LED.
 *
 * @param command   How to set the LED.
 */
void FiFPGA::ledSetBlink(int command)
{
    uint32_t reg_value = 0;

    // Setup the register control word based on what is being commanded.
    switch (command)
    {
    default:
    case LED_OFF:
        // Turn off the LED. All other bits are ignored.
        reg_value = LC_LED_MODE_OFF;
        break;
    case LED_SOLID_ON:
        // Turn on the LED. All other bits are ignored.
        reg_value = LC_LED_MODE_SOLID_ON;
        break;
    case LED_FAST_BLINK:
        // Tell the FPGA to blink the LED quickly.
        reg_value = LC_LED_MODE_BLINK |
                    LC_LED_BLINK_ONTIME(2) |
                    LC_LED_BLINK_OFFTIME(2);
        break;
    case LED_SLOW_BLINK:
        // Tell the FPGA to blink the LED slowly.
        reg_value = LC_LED_MODE_BLINK |
                    LC_LED_BLINK_ONTIME(5) |
                    LC_LED_BLINK_OFFTIME(5);
        break;
    }

    regWrite(LED_CONTROL, reg_value);
}

/*!
 * @brief Disable the external IO mux for the given channel/function.
 * @param channel   The channel to disable.
 * @param func      The function being disabled.
 */
void FiFPGA::extioMuxDisable(int channel, int func)
{
    bool rc(false);

    TP(FT_FUNC, "%s: channel=%d, func=%d", __PRETTY_FUNCTION__, channel, func);

    switch (func)
    {
    case ARM_IN:
        rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_ARM_IN_SEL_OFF, ~EXTIO_MUX_ARM_IN_SEL_OFF);
        break;
    case SYNC_IN:
        rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_SYNC_IN_SEL_OFF, ~EXTIO_MUX_SYNC_IN_SEL_OFF);
        break;
    case TRIG_IN:
        rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_TRIG_IN_SEL_OFF, ~EXTIO_MUX_TRIG_IN_SEL_OFF);
        break;
    case ARM_OUT:
    case SYNC_OUT:
    case TRIG_OUT:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO0_SEL_DISABLED, ~(EXTIO_MUX_IO0_SEL_DISABLED | EXTIO_MUX_IO0_DIR_OUTPUT));
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO1_SEL_DISABLED, ~(EXTIO_MUX_IO1_SEL_DISABLED | EXTIO_MUX_IO1_DIR_OUTPUT));
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO2_SEL_DISABLED, ~(EXTIO_MUX_IO2_SEL_DISABLED | EXTIO_MUX_IO2_DIR_OUTPUT));
            break;
        default:
            rc = false;
            break;
        }
        break;
    }
    if (!rc)
    {
        fi_log(LOG_ERR, "%s: Unable to disable mux for chan=%d, func=%d", __PRETTY_FUNCTION__, channel, func);
    }
}

/*!
 * @brief Enable the external IO mux for the given channel/function.
 * @param channel   The channel to enable.
 * @param func      The function being enabled.
 */
void FiFPGA::extioMuxEnable(int channel, int func)
{
    bool rc(false);

    TP(FT_FUNC, "%s: channel=%d, func=%d", __PRETTY_FUNCTION__, channel, func);

    switch (func)
    {
    case ARM_IN:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_ARM_IN_SEL_IO0, ~EXTIO_MUX_ARM_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_ARM_IN_SEL_IO1, ~EXTIO_MUX_ARM_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_ARM_IN_SEL_IO2, ~EXTIO_MUX_ARM_IN_SEL_OFF);
            break;
        default:
            rc = false;
            break;
        }
        break;
    case SYNC_IN:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_SYNC_IN_SEL_IO0, ~EXTIO_MUX_SYNC_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_SYNC_IN_SEL_IO1, ~EXTIO_MUX_SYNC_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_SYNC_IN_SEL_IO2, ~EXTIO_MUX_SYNC_IN_SEL_OFF);
            break;
        default:
            rc = false;
            break;
        }
        break;
    case TRIG_IN:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_TRIG_IN_SEL_IO0, ~EXTIO_MUX_TRIG_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_TRIG_IN_SEL_IO1, ~EXTIO_MUX_TRIG_IN_SEL_OFF);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_TRIG_IN_SEL_IO2, ~EXTIO_MUX_TRIG_IN_SEL_OFF);
            break;
        default:
            rc = false;
            break;
        }
        break;
    case ARM_OUT:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO0_SEL_ARM_OUT | EXTIO_MUX_IO0_DIR_OUTPUT, ~EXTIO_MUX_IO0_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO1_SEL_ARM_OUT | EXTIO_MUX_IO1_DIR_OUTPUT, ~EXTIO_MUX_IO1_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO2_SEL_ARM_OUT | EXTIO_MUX_IO2_DIR_OUTPUT, ~EXTIO_MUX_IO2_SEL_DISABLED);
            break;
        default:
            rc = false;
            break;
        }
        break;
    case SYNC_OUT:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO0_SEL_SYNC_OUT | EXTIO_MUX_IO0_DIR_OUTPUT, ~EXTIO_MUX_IO0_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO1_SEL_SYNC_OUT | EXTIO_MUX_IO1_DIR_OUTPUT, ~EXTIO_MUX_IO1_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO2_SEL_SYNC_OUT | EXTIO_MUX_IO2_DIR_OUTPUT, ~EXTIO_MUX_IO2_SEL_DISABLED);
            break;
        default:
            rc = false;
            break;
        }
        break;
    case TRIG_OUT:
        switch (channel)
        {
        case EXTIO_CHANNEL_0:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO0_SEL_TRIG_OUT | EXTIO_MUX_IO0_DIR_OUTPUT, ~EXTIO_MUX_IO0_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_1:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO1_SEL_TRIG_OUT | EXTIO_MUX_IO1_DIR_OUTPUT, ~EXTIO_MUX_IO1_SEL_DISABLED);
            break;
        case EXTIO_CHANNEL_2:
            rc = regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IO2_SEL_TRIG_OUT | EXTIO_MUX_IO2_DIR_OUTPUT, ~EXTIO_MUX_IO2_SEL_DISABLED);
            break;
        default:
            break;
        }
        break;
    }
    if (!rc)
    {
        fi_log(LOG_ERR, "%s: Unable to enable mux for chan=%d, func=%d", __PRETTY_FUNCTION__, channel, func);
    }
}

/*!
 * @brief Return the external I/O status value.
 */
uint32_t FiFPGA::extioGetActivity()
{
    uint32_t reg_value;

    regRead(EXT_IO_STATUS, &reg_value);
    return reg_value;
}

/*!
 * @brief Enable or disable external triggers on the FPGA. If enabled
 *    then the fpga will passthrough external triggers, if disabled it won't.
 *
 * @param enable    Should external triggers be enabled or disabled
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::externalTriggerEnable(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t user_value)
{
    uint32_t    reg_value;
    bool    enable;
    bool    rc;
    int     old_channel(-1);
    int     new_channel;
    FiFPGA  *fpga_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: user_value=0x%08x", __PRETTY_FUNCTION__, user_value);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Get the current setting, so we can reset the current mux setting.
    reg_value = regs[ GIGE_REG_EXT_TRIG_ENABLE ].get_value();

    // Is the external trigger enabled?
    if ((reg_value & GIGE_EXT_TRIG_ENABLE_MASK) == GIGE_EXT_TRIG_ENABLE_MASK)
    {
        // Get the channel that is being used and disable that mux entry.
        old_channel = (reg_value >> GIGE_EXT_TRIG_CHANNEL_SHIFT) & GIGE_EXT_TRIG_CHANNEL_MASK;
    }

    // Only interested in the enable.
    enable = ((user_value & GIGE_EXT_TRIG_ENABLE_MASK) == GIGE_EXT_TRIG_ENABLE_MASK);

#ifdef OVR_ALCON_EXT_TRIG_POL
    // Check for an override.
    {
        bool ovr_polarity;

        // If polarity was specified, that means we want to enable it.
        // It doesn't matter which polarity was chosen.
        if (override_bool(OVR_ALCON_EXT_TRIG_POL, &ovr_polarity))
            enable = true;
    }
#endif /* OVR_ALCON_EXT_TRIG_POL */

    // Assume failure.
    rc = false;

    // We want to enable or disable the TRIG_IN processing,
    // so first read the current value of the control register.
    if (fpga_ptr->regRead(EXT_IO_CONTROL, &reg_value))
    {
        // Set or clear the enable bit.
        if (enable)
        {
            // What channel are we enabling.
            new_channel = (user_value >> GIGE_EXT_TRIG_CHANNEL_SHIFT) & GIGE_EXT_TRIG_CHANNEL_MASK;

            // Before we setup the new channel, let's make sure we fixup any old channel that
            // might be hanging around.
            if (old_channel != -1)
            {
                // Disable the old MUX channel we were using for the external trigger.
                fpga_ptr->extioMuxDisable(old_channel, TRIG_IN);
                old_channel = -1;
            }

            // Now we are free to setup the new channel.
            fpga_ptr->extioMuxEnable(new_channel, TRIG_IN);

            // Set the enable.
            reg_value |= EXTIO_TRIGIN_ENABLE;
        }
        else
        {
            // Make sure there's no channel information.
            user_value &= ~(GIGE_EXT_TRIG_CHANNEL_MASK << GIGE_EXT_TRIG_CHANNEL_SHIFT);

            // Make sure to clear the enable bit.
            reg_value &= ~EXTIO_TRIGIN_ENABLE;
        }

        // Write the value back out to the control register.
        if (fpga_ptr->regWrite(EXT_IO_CONTROL, reg_value))
            rc = true;
        else
            fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);

        // If we're disabling the external trigger, we need to change the IO mux AFTER the fact.
        if (rc && (old_channel != -1))
        {
            // Disable the old MUX channel we were using for the external trigger.
            fpga_ptr->extioMuxDisable(old_channel, TRIG_IN);
        }
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
    }

    if (rc)
    {
        regs[GIGE_REG_EXT_TRIG_ENABLE].set_value(user_value);
    }

    // Return our result.
    return (rc ? GEV_STATUS_SUCCESS : GEV_STATUS_WRONG_CONFIG);
}

/*!
 * @brief Set the polarity for the external trigger.
 *
 * @param rising_edge   TRUE if the trigger is rising edge, FALSE for falling edge.
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::externalTriggerSetPolarity(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t rising_edge)
{
    bool    rc;
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: rising_edge=0x%08x", __PRETTY_FUNCTION__, rising_edge);

    // Assume failure.
    rc = false;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Only interested in the flag.
    rising_edge = rising_edge & GIGE_EXT_TRIG_POLARITY_RISING;

#ifdef OVR_ALCON_EXT_TRIG_POL
    // Check for an override.
    {
        bool ovr_polarity;

        // Did we get an override?
        if (override_bool(OVR_ALCON_EXT_TRIG_POL, &ovr_polarity))
        {
            // See if we're supposed to be active high or low.
            rising_edge = ovr_polarity ? GIGE_EXT_TRIG_POLARITY_RISING : GIGE_EXT_TRIG_POLARITY_FALLING;
        }
    }
#endif /* OVR_ALCON_EXT_TRIG_POL */

    // Are we doing a rising or falling edge?
    if (fpga_ptr->regRead(EXT_IO_CONTROL, &reg_value))
    {
        // Set or clear the rising edge bit.
        if (rising_edge)
            reg_value |= EXTIO_TRIGIN_RISING_EDGE;
        else
            reg_value &= ~EXTIO_TRIGIN_RISING_EDGE;

        // Write the value back out to the control register.
        if (fpga_ptr->regWrite(EXT_IO_CONTROL, reg_value))
            rc = true;
        else
            fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
    }

    if (rc)
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();
        regs[GIGE_REG_EXT_TRIG_POLARITY].set_value(rising_edge);
    }

    // Return our result.
    return (rc ? GEV_STATUS_SUCCESS : GEV_STATUS_WRONG_CONFIG);
}

/*!
 * @brief Determine what the system thinks the incoming frame rate
 *    is, by looking at the time between pulses on SyncIn.
 *
 * @return The number of 125MHz clocks (FPGA Clocks) between
 *     SyncIn pulses.
 */
uint32_t FiFPGA::syncInGetTiming()
{
    TP(FT_TIME, "%s", __PRETTY_FUNCTION__);

    // Return the timing information from the SyncIn tracking object.
    return sync_tracker.rawTiming();
}

/*!
 * @brief Read the incoming sync period for tracking IRIG.
 *
 * This can block trying to read the time. For IRIG tracking, the minimum-period measurement
 * logic is used for the IRIG input. We will have to try to estimate the PPF frame rate from
 * the live counter so we don't clobber IRIG measurements.
 *
 * @return The number of 125MHz clocks (FPGA Clocks) between
 *     SyncIn pulses.
 */
uint32_t FiFPGA::syncinGetIrigTiming()
{
#ifdef ALM_TODO_IRIG
    int retry_count;
    uint32_t    reg_value;
    bool    status;

    // Throw away the first one.
    status = read_reg(SYNC_IN_PERIOD_MIN, &reg_value);
    usleep(100000);

    // Loop a few times to try and get a valid reading.
    for (retry_count = 0; retry_count < MAX_RETRY_COUNT; ++retry_count)
    {
        // Read the minimum period between pulses from the FPGA.
        status = read_reg(SYNC_IN_PERIOD_MIN, &reg_value);
        if ((reg_value & SYNCIN_ACTIVE) == (uint32_t)SYNCIN_ACTIVE)
            break;

        // Pause for a little.
        usleep(100000);
    }

    // Return a obviously wrong value if we didn't get any "right" ones.
    if (retry_count == MAX_RETRY_COUNT)
        reg_value = 0;

    // Mask out the clock value. This is the number of
    // FPGA clocks per pulse.
    return (reg_value & ((uint32_t)SYNCIN_PERIOD_MIN_MASK));
#else
    LOG("%s: Under construction =================", __PRETTY_FUNCTION__);
    return 0;
#endif /* ALM_TODO_IRIG */
}

/*!
 * @brief Get the RGB histogram from the FPGA.
 *
 * @param buffer    The buffer to hold the returned histogram must be at least LUPA_HISTOGRAM_RGB_SIZE bytes long.
 *
 * @return true if success, false otherwise.
 */
bool FiFPGA::ipmRgbHistogramRead(uint8_t *buffer)
{
    ioHistogramMessage msg;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity checks.
    if (buffer == nullptr)
        return false;

    // Make sure the histogram is done generating.
    if (!ipmIsRgbHistoIdle())
    {
        fi_log(LOG_WARNING, "%s: histogram generation seems to be stuck.", __PRETTY_FUNCTION__);
        return false;
    }

    // Read the histogram out of the FPGA driver.
    msg.dest_ptr = buffer;

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Success?
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_ERR, "%s: Unable to read histogram!", __PRETTY_FUNCTION__);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Set the sync delay value.
 *
 * @param delay_amount The number of microseconds to delay the
 *             sync-in pulse.
 */
void FiFPGA::syncInDelay(uint32_t delay_amount)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    TP(FT_FUNC, "%s: delay_amount=0x%08x", __PRETTY_FUNCTION__, delay_amount);
    sensor_ptr->set_sync_delay(delay_amount);
}

/*!
 * @brief Set the SyncIn information. This information includes the type of
 *    signal used for SyncIn (internal or external), and whether
 *    it's 1-pulse/sec or 1-pulse/frame.
 *
 * @param type The SyncIn information used to configure the FPGA registers.
 *
 * @return true if success, false otherwise.
 */
bool FiFPGA::syncInEnable(uint32_t type)
{
    int channel;
    bool rc;
    uint32_t reg_value;
    uint32_t extio_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Sensor *sensor_ptr = CameraServer::get_sensor();

    TP(FT_FUNC, "%s: type=0x%08x", __PRETTY_FUNCTION__, type);

    // Assume failure.
    rc = false;

    // Is the external sync-in enabled?
    reg_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((reg_value & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_EXTERNAL)
    {
        // Get the channel that is being used and disable that mux entry.
        channel = (reg_value >> GIGE_SYNC_IN_CHANNEL_SHIFT) & GIGE_SYNC_IN_CHANNEL_MASK;
        extioMuxDisable(channel, SYNC_IN);
    }

    // Read the existing register settings.
    if (!regRead(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __FUNCTION__);
        goto sin_exit;
    }

    // If our sync tracker is running, request that it stop.
    if (sync_tracker.isRunning())
    {
        // Ask the tracker to stop, and wait for it.
        sync_tracker.stop();
        sync_tracker.join();
    }

    // Figure out what type we are.
    if ((type & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_INTERNAL)
    {
        // Using internal, then we need to disable the SyncIn processing block
        // and make sure we're using the internal 1-PPS source.
        extio_value &= ~(EXTIO_SYNCIN_ENABLE | EXTIO_SYNC_1PPS_SRC_SEL_MASK);
        extio_value |= EXTIO_SYNC_1PPS_INTERNAL;

        // Make sure the time module is in Master mode.
        regWrite(TIME_CONTROL, TIME_MODE_SET | TIME_MODE_MASTER);
    }
    else if ((type & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_EXTERNAL)
    {
        // Get the channel that is being used and enable that mux entry.
        channel = (type >> GIGE_SYNC_IN_CHANNEL_SHIFT) & GIGE_SYNC_IN_CHANNEL_MASK;
        extioMuxEnable(channel, SYNC_IN);

        // Initialize the debounce value for the SyncIn signal.
        if (!regWrite(SYNC_IN_DEBOUNCE_TIME, SYNC_IN_DEBOUNCE_DEFAULT))
        {
            fi_log(LOG_ERR, "%s: Unable to write SYNC_IN_DEBOUNCE_TIME register.", __FUNCTION__);
            goto sin_exit;
        }

        // Indicate that we're using the External Signal, and enable the SyncIn
        // processing block.
        extio_value &= ~EXTIO_SYNC_1PPS_SRC_SEL_MASK;

        // Make sure we enable the external SyncIn signal.
        extio_value |= EXTIO_SYNCIN_ENABLE;

        // Figure out if we are 1-Pulse/Second or 1-Pulse/Frame.
        if ((type & GIGE_SYNC_IN_PULSE_MASK) == GIGE_SYNC_IN_PPS)
        {
            float frame_rate;

            // For 1PPS, we also need to tell the Time module to slave.
            regWrite(TIME_CONTROL, TIME_MODE_SET | TIME_MODE_1PPS_SLAVE | (TIME_SYNC_LATE_LIMIT << TIME_SYNC_LATE_SHIFT));

            // Make sure we tell FPGA the number of intergrations pulses
            // to use (i.e. the frame rate).
            frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());
            syncInDelay((uint32_t)(frame_rate + 0.5));
        }
    }
    else
    {
        // GPS - not supported yet.
        goto sin_exit;
    }

    // Write out our new values.
    regWrite(EXT_IO_CONTROL, extio_value);

    // Have the sensor update based on type.
    sensor_ptr->set_syncin_enable(type);

    // If the sync-in is a 1-PPS, startup our tracker.
    // The others either don't use a tracker or will start their own.
    if ((type & (GIGE_SYNC_TYPE_MASK | GIGE_SYNC_IN_PULSE_MASK)) == (GIGE_SYNC_TYPE_EXTERNAL | GIGE_SYNC_IN_PPS))
    {
        // Request the tracker to start. We're actually going to track
        // the signal, so make sure we're not doing measure-only mode.
        sync_tracker.setMeasureOnly(false);
        sync_tracker.start();
    }

    // Don't perform the PPF measure-only function when IRIG is enabled because the
    // measurement logic is needed for IRIG tracking.
    else if ((type & (GIGE_SYNC_TYPE_MASK | GIGE_SYNC_IN_PULSE_MASK)) == (GIGE_SYNC_TYPE_EXTERNAL | GIGE_SYNC_IN_PPF))
    {
        // Start the tracker in measure-only mode.
        sync_tracker.setMeasureOnly(true);
        sync_tracker.start();
    }

    // All good.
    rc = true;

sin_exit:
    return rc;
}

/*!
 * @brief Set the polarity (rising/falling edge) of the SyncIn signal.
 *
 * @param rising_edge   TRUE if we want to use rising edge, FALSE for falling edge.
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::syncInSetPolarity(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t rising_edge)
{
    FiFPGA  *fpga_ptr;
    uint32_t    extio_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: rising_edge=0x%08x", __PRETTY_FUNCTION__, rising_edge);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Only interested in the flag.
    rising_edge = (rising_edge & GIGE_SYNC_IN_POL_MASK) == GIGE_SYNC_IN_RISING;

    // Read the existing register settings.
    if (!fpga_ptr->regRead(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Set or clear the control bit based on what the user wants us to do.
    if (rising_edge)
        extio_value |= EXTIO_SYNCIN_RISING_EDGE;
    else
        extio_value &= ~EXTIO_SYNCIN_RISING_EDGE;

    // Write the new value back out to the FPGA.
    if (!fpga_ptr->regWrite(EXT_IO_CONTROL, extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Save our value.
    regs[GIGE_REG_SYNC_IN_POLARITY].set_value(rising_edge);

    // All good.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Figure out how many FPGA clocks there are in the requested
 *    duty cycle period.
 *
 * @param dc_percent The duty cycle percentage.
 *
 * @return true on success, false otherwise.
 */
bool FiFPGA::syncOutSetPulseWidth(uint32_t dc_percent)
{
    float    frame_rate;
    uint32_t    frame_period_clks;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FT_FUNC, "%s: dc_percent=0x%08x", __PRETTY_FUNCTION__, dc_percent);

    if (dc_percent > MAX_DC_PERCENT)
        dc_percent = MAX_DC_PERCENT;
    if (dc_percent == 0)
        dc_percent = 1;

    // Get the current frame rate.
    frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[GIGE_REG_FRAME_RATE].get_value());

    // Calculate the number of FPGA clocks/frame.
    frame_period_clks = FI_FPGA::FPGA_CLOCK_SPEED / frame_rate;
    if (frame_period_clks == 0)
    {
        fi_log(LOG_ERR, "%s: Invalid ICM_FRAME_PERIOD value [0]!", __PRETTY_FUNCTION__);
        return false;
    }

    // Figure out how many clocks we need.
    frame_period_clks = (frame_period_clks * dc_percent) / 100;

    // Set the duty-cycle period.
    if (!regWrite(SYNC_OUT_PULSE_WIDTH, frame_period_clks))
    {
        fi_log(LOG_ERR, "%s: Unable to write SYNC_OUT_PULSE_WIDTH!", __PRETTY_FUNCTION__);
        return false;
    }

    // All good.
    return true;
}

/*!
 * @brief Use the passed in "syncout_settings" values to initialize
 *    the SyncOut controls in the FPGA.
 *
 * @param syncout_settings Our SyncOut settings to use.
 *
 * @return true if success, false otherwise.
 */
bool FiFPGA::syncOutConfigure(uint32_t syncout_settings)
{
    int chan;
    uint32_t curr_value;
    uint32_t extio_value = 0;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FT_FUNC, "%s: syncout_settings=0x%08x", __PRETTY_FUNCTION__, syncout_settings);

#ifdef OVR_ALCON_SYNC_OUT_POL
    // Check for an override.
    {
        bool ovr_polarity;

        // Did we get an override?
        if (override_bool(OVR_ALCON_SYNC_OUT_POL, &ovr_polarity))
        {
            // Because we have an override, that means we want to enable SyncOut.
            // This is the mode we will use for Alcon operations. If we need another mode,
            // the overrides will need to be updated to allow other specs.
            syncout_settings = GIGE_SYNC_OUT_ENABLE | GIGE_SYNC_OUT_PPF | GIGE_SYNC_OUT_INTEGRATION_PERIOD;

            // Are we active high or low?
            syncout_settings |= (ovr_polarity ? GIGE_SYNC_OUT_ACTIVE_HIGH : GIGE_SYNC_OUT_ACTIVE_LOW);
        }
    }
#endif /* OVR_ALCON_SYNC_OUT_POL */

    // Get the current settings for the sync out register.
    curr_value = regs[GIGE_REG_SYNC_OUT_SELECT].get_value();

    // If the sync out logic was enabled, we need to make sure to turn off the
    // mux channel it was using in case the new settings change the channel or
    // disable the output.
    if ((curr_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
    {
        // Extract the channel information.
        chan = (curr_value >> GIGE_SYNC_OUT_CHANNEL_SHIFT) & GIGE_SYNC_OUT_CHANNEL_MASK;
        extioMuxDisable(chan, SYNC_OUT);
    }

    // Read the existing register settings.
    if (!regRead(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __FUNCTION__);
        return false;
    }

    // Clear out any existing SyncOut settings.
    extio_value &= ~(EXTIO_SYNCOUT_ENABLE | EXTIO_SYNCOUT_PASSTHRU_ENABLE | EXTIO_SYNCOUT_SRC_SEL_MASK);

    // Start setting bits as appropriate.
    if ((syncout_settings & GIGE_SYNC_OUT_MASK) == GIGE_SYNC_OUT_IS_SYNC_IN)
    {
        // Setup to pass-through the SyncIn signal to the SyncOut pin.
        extio_value |= (EXTIO_SYNCOUT_PASSTHRU_ENABLE | EXTIO_SYNCOUT_ENABLE);

        // Figure out if the user wants to invert the SyncIn signal or not.
        if ((syncout_settings & GIGE_SYNC_OUT_POLARITY_MASK) == GIGE_SYNC_OUT_INVERT_SYNCIN)
            extio_value &= ~EXTIO_SYNCOUT_INVERT;
        else
            extio_value |= EXTIO_SYNCOUT_INVERT;
    }
    else if ((syncout_settings & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
    {
        // Enable the SyncOut pin as an output.
        extio_value |= EXTIO_SYNCOUT_ENABLE;

        // Are we doing 1-Pulse/Frame or 1-Pulse/Second?
        if ((syncout_settings & GIGE_SYNC_OUT_MASK) == GIGE_SYNC_OUT_PPF)
        {
            // Are we using the actual integration period, or a duty cycle value?
            if ((syncout_settings & GIGE_SYNC_OUT_DUTYCYCLE_MASK) == GIGE_SYNC_OUT_INTEGRATION_PERIOD)
            {
                extio_value = (extio_value & ~EXTIO_SYNCOUT_SRC_SEL_MASK) | EXTIO_SYNCOUT_INTERNAL_1PPF;
            }
            else
            {
                // Set the percentage for the Duty Cycle.
                syncOutSetPulseWidth((syncout_settings & GIGE_SYNC_OUT_DUTYCYCLE_MASK) >> GIGE_SYNC_OUT_DUTYCYCLE_SHIFT);
                extio_value = (extio_value & ~EXTIO_SYNCOUT_SRC_SEL_MASK) | EXTIO_SYNCOUT_PROG_1PPF;
            }
        }
        else
        {
            // 1-Pulse/Second always uses a duty cycle. Set the percentage for the Duty Cycle.
            syncOutSetPulseWidth((syncout_settings & GIGE_SYNC_OUT_DUTYCYCLE_MASK) >> GIGE_SYNC_OUT_DUTYCYCLE_SHIFT);
            extio_value = (extio_value & ~EXTIO_SYNCOUT_SRC_SEL_MASK) | EXTIO_SYNCOUT_PROG_1PPS;
        }

        // Determine if the user wants the SyncOut to be active High or Low.
        if ((syncout_settings & GIGE_SYNC_OUT_POLARITY_MASK) == GIGE_SYNC_OUT_ACTIVE_LOW)
            extio_value &= ~EXTIO_SYNCOUT_ACTIVE_HIGH;
        else
            extio_value |= EXTIO_SYNCOUT_ACTIVE_HIGH;
    }

    // If we're enabling sync-out, make sure to setup the mux. If it's getting disabled,
    // we've already turned off the mux.
    if ((extio_value & EXTIO_SYNCOUT_ENABLE) == EXTIO_SYNCOUT_ENABLE)
    {
        // Extract the channel information and enable the mux.
        chan = (syncout_settings >> GIGE_SYNC_OUT_CHANNEL_SHIFT) & GIGE_SYNC_OUT_CHANNEL_MASK;
        extioMuxEnable(chan, SYNC_OUT);
    }

    // Write the new value back out to the FPGA.
    if (!regWrite(EXT_IO_CONTROL, extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
        return false;
    }

    // All good.
    return true;
}

/*!
 * @brief Initialize the ARM input at startup.
 *
 * We don't want to enable the ARM input until our startup reaches a state where
 * we are willing to accept an ARM command.
 */
void FiFPGA::armInInitialize()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t reg_value;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    reg_value = regs[GIGE_REG_ARM_IN_ENABLE].get_value();
    armInEnable(nullptr, reg_value);
}

/*!
 * @brief Enable/disable the EXT_ARM input interrupt.
 *
 * @param p_msg     Not used
 * @param enable_val    GIGE_ARM_TYPE_EXTERNAL if we're to enable the ArmIn signal, 0 otherwise.
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::armInEnable(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t enable_val)
{
    int channel;
    bool armin_enable;
    FiFPGA *fpga_ptr;
    uint32_t reg_value;
    uint32_t extio_value = 0;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: enable_val=0x%08x", __PRETTY_FUNCTION__, enable_val);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Is the external arm-in enabled?
    reg_value = regs[ GIGE_REG_ARM_IN_ENABLE ].get_value();
    if ((reg_value & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
    {
        // Get the channel that is being used and disable that mux entry.
        channel = (reg_value >> GIGE_ARM_IN_CHANNEL_SHIFT) & GIGE_ARM_IN_CHANNEL_MASK;
        fpga_ptr->extioMuxDisable(channel, ARM_IN);
    }

    // Only interested in the flag.
    armin_enable = ((enable_val & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL);

    // Now, if we're supposed to enable or disable, go ahead and do it.
    if (armin_enable)
    {
        // Get the channel that is being used and enable that mux entry.
        channel = (enable_val >> GIGE_ARM_IN_CHANNEL_SHIFT) & GIGE_ARM_IN_CHANNEL_MASK;
        fpga_ptr->extioMuxEnable(channel, ARM_IN);

        // Enable the arm-in logic.
        extio_value = EXTIO_ARM_IN_ENABLE;
        fpga_ptr->irqEnable(PCIEIRQ_EXT_ARM_INPUT);
        fpga_ptr->irqEnable(PCIEIRQ_EXT_UNARM_INPUT);
    }
    else
    {
        // Make sure there's no channel information.
        enable_val &= ~(GIGE_ARM_IN_CHANNEL_MASK << GIGE_ARM_IN_CHANNEL_SHIFT);

        // Disable the arm-in logic.
        extio_value = 0;
        fpga_ptr->irqDisable(PCIEIRQ_EXT_ARM_INPUT);
        fpga_ptr->irqDisable(PCIEIRQ_EXT_UNARM_INPUT);
    }

    // Save our value.
    regs[GIGE_REG_ARM_IN_ENABLE].set_value(enable_val);

    // Write the new value back out to the FPGA.
    if (!fpga_ptr->regModify(EXT_IO_CONTROL, extio_value, ~EXTIO_ARM_IN_ENABLE))
        return GEV_STATUS_WRONG_CONFIG;

    // All is good.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update function for GIGE_REG_ARM_IN_POLARITY
 *
 * @param reg_value register value to write
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::armInPolarityUpdate(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    uint32_t    extio_value = 0;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: reg_value=0x%08x", __PRETTY_FUNCTION__, reg_value);

    // Set or clear the control bit based on what the user wants us to do.
    extio_value = 0;
    if ((reg_value & GIGE_ARM_IN_EDGE_TYPE) != GIGE_ARM_IN_EDGE_TYPE)
        extio_value |= EXTIO_ARMIN_LEVEL_INPUT;
    if ((reg_value & GIGE_ARM_IN_POLARITY_MASK) == GIGE_ARM_IN_RISING)
        extio_value |= EXTIO_ARMIN_RISING_EDGE;

    // Reject level input and active high.
    if ((extio_value & (EXTIO_ARMIN_LEVEL_INPUT | EXTIO_ARMIN_RISING_EDGE)) == (EXTIO_ARMIN_LEVEL_INPUT | EXTIO_ARMIN_RISING_EDGE))
    {
        fi_log(LOG_WARNING, "%s: Rejecting level input active high", __FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Write the new value back out to the FPGA.
    if (fpga_ptr->regModify(EXT_IO_CONTROL, extio_value, ~(EXTIO_ARMIN_RISING_EDGE | EXTIO_ARMIN_LEVEL_INPUT)))
    {
        regs[GIGE_REG_ARM_IN_POLARITY].set_value(reg_value);
        return GEV_STATUS_SUCCESS;
    }

    // Had a problem writing the register.
    return GEV_STATUS_ERROR;
}

/*!
 * @brief If the external ARM-In signal is enabled, we want to disable it. Note that we
 *        aren't changing the GIGE_REG_CACHE value, just changing the bit in the FPGA
 *        control register. The assumption is that this is temporary and a resume will
 *        be coming along.
 *
 * @san FiFPGA::armin_resume()
 */
void FiFPGA::armInPause()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Only worry about this if the external ARM-in signal is enabled.
    if ((regs[GIGE_REG_ARM_IN_ENABLE].get_value() & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
    {
        // Disable the armin control in the FPGA.
        regModify(EXT_IO_CONTROL, 0, ~EXTIO_ARM_IN_ENABLE);
    }
}

/*!
 * @brief If the external ARM-In signal is enabled, we want to re-enable the FPGA's control
 *        bit to allow the signal to propogate into the logic. The assumption is that this
 *        has been "paused".
 *
 * @san FiFPGA::armin_pause()
 */
void FiFPGA::armInResume()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    // Only worry about this if the external ARM-in signal is enabled.
    if ((regs[GIGE_REG_ARM_IN_ENABLE].get_value() & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
    {
        // Enable the armin control in the FPGA.
        regModify(EXT_IO_CONTROL, EXTIO_ARM_IN_ENABLE, -1);
    }
}

/*!
 * @brief Figure out how many FPGA clocks there are in the requested
 *    duty cycle period.
 *
 * @param period_ms The number of milliseconds for the pulse.
 *
 * @return true on success, false otherwise.
 */
bool FiFPGA::armOutSetPulseWidth(uint32_t period_ms)
{
    TP(FT_FUNC, "%s: period_ms=%u", __PRETTY_FUNCTION__, period_ms);

    // Are we just doing the ARM level (no pulse)?
    if (period_ms != GIGE_ARM_OUT_NO_PERIOD)
    {
        uint32_t    fpga_clocks;

        // There are these many clocks per millisecond.
        fpga_clocks = FPGA_CLOCK_SPEED_IN_USEC * 1000;

        // Figure out how many clocks we need.
        fpga_clocks = fpga_clocks * period_ms;

        // Set the duty-cycle period.
        if (!regWrite(ARM_OUT_PULSE_WIDTH, fpga_clocks))
        {
            fi_log(LOG_ERR, "%s: Unable to write ARM_OUT_PULSE_WIDTH!", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // All good.
    return true;
}

/*!
 * @brief Use the passed in "armout_settings" values to initialize
 *    the ArmOut controls in the FPGA.
 *
 * @param armout_settings Our ArmOut settings to use.
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::armOutConfigure(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t armout_settings)
{
    int chan;
    FiFPGA *fpga_ptr;
    uint32_t curr_value;
    uint32_t extio_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: armout_settings=0x%08x", __PRETTY_FUNCTION__, armout_settings);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

#ifdef OVR_ALCON_MODE
    // Check for an override saying we need to do special Alcon mode handling.
    {
        bool alcon_mode;

        // If we have the Alcon mode override, we don't want to enable the Arm Output until we are
        // initialized and ready to start capturing. We need to leave the pin in input mode, pulled
        // up by external resistors. Once we are ready to capture, we will enable the output and
        // use the test bit to force it low.
        if (override_bool(OVR_ALCON_MODE, &alcon_mode) && alcon_mode)
            armout_settings = GIGE_ARM_OUT_DISABLE;
    }
#endif /* OVR_ALCON_MODE */

    // Get the current settings for the arm out register.
    curr_value = regs[GIGE_REG_ARM_OUT_SELECT].get_value();

    // If the arm out logic was enabled, we need to make sure to turn off the
    // mux channel it was using in case the new settings change the channel or
    // disable the output.
    if ((curr_value & GIGE_ARM_OUT_ENABLE) == GIGE_ARM_OUT_ENABLE)
    {
        // Extract the channel information.
        chan = (curr_value >> GIGE_ARM_OUT_CHANNEL_SHIFT) & GIGE_ARM_OUT_CHANNEL_MASK;
        fpga_ptr->extioMuxDisable(chan, ARM_OUT);
    }

    // Read the existing register settings.
    if (!fpga_ptr->regRead(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Clear out any existing ArmOut settings.
    extio_value &= ~(EXTIO_ARMOUT_ENABLE | EXTIO_ARMOUT_PASSTHRU_ENABLE);

    // Start setting bits as appropriate.
    if ((armout_settings & GIGE_ARM_OUT_IS_ARM_IN) == GIGE_ARM_OUT_IS_ARM_IN)
    {
        // Setup to pass-through the ArmIn signal to the ArmOut pin.
        extio_value |= (EXTIO_ARMOUT_PASSTHRU_ENABLE | EXTIO_ARMOUT_ENABLE);

        // Figure out if the user wants to invert the ArmIn signal or not.
        if ((armout_settings & GIGE_ARM_OUT_POLARITY_MASK) == GIGE_ARM_OUT_INVERT_ARMIN)
            extio_value &= ~EXTIO_ARMOUT_INVERT;
        else
            extio_value |= EXTIO_ARMOUT_INVERT;
    }
    else if ((armout_settings & GIGE_ARM_OUT_ENABLE) == GIGE_ARM_OUT_ENABLE)
    {
        uint32_t    period_ms;

        // Enable the ArmOut pin as an output.
        extio_value |= EXTIO_ARMOUT_ENABLE;

        // Set the percentage for the Duty Cycle.
        period_ms = (armout_settings & GIGE_ARM_OUT_PERIOD_MASK) >> GIGE_ARM_OUT_PERIOD_SHIFT;
        if (period_ms == GIGE_ARM_OUT_NO_PERIOD)
        {
            // Use the internal state, not a pulse.
            extio_value &= ~EXTIO_ARM_OUT_PROG_PULSE;
        }
        else
        {
            fpga_ptr->armOutSetPulseWidth(period_ms);
            extio_value |= EXTIO_ARM_OUT_PROG_PULSE;
        }

        // Determine if the user wants the ArmOut to be active High or Low.
        if ((armout_settings & GIGE_ARM_OUT_POLARITY_MASK) == GIGE_ARM_OUT_ACTIVE_LOW)
            extio_value &= ~EXTIO_ARMOUT_ACTIVE_HIGH;
        else
            extio_value |= EXTIO_ARMOUT_ACTIVE_HIGH;
    }

    // Are we enabling the channel?
    if ((extio_value & EXTIO_ARMOUT_ENABLE) == EXTIO_ARMOUT_ENABLE)
    {
        // Extract the channel information.
        chan = (armout_settings >> GIGE_ARM_OUT_CHANNEL_SHIFT) & GIGE_ARM_OUT_CHANNEL_MASK;
        fpga_ptr->extioMuxEnable(chan, ARM_OUT);
    }

    // Write the new value back out to the FPGA.
    if (!fpga_ptr->regWrite(EXT_IO_CONTROL, extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // If we're just being enabled/disabled don't lose the
    // rest of the register settings.
    if ((armout_settings == GIGE_ARM_OUT_DISABLE) || (armout_settings == GIGE_ARM_OUT_ENABLE))
    {
        uint32_t reg_value;

        reg_value = regs[ GIGE_REG_ARM_OUT_SELECT ].get_value();
        reg_value &= ~GIGE_ARM_OUT_ENABLE;

        // Don't save the channel information on a disable!
        if (armout_settings == GIGE_ARM_OUT_DISABLE)
            reg_value &= ~(GIGE_ARM_OUT_CHANNEL_MASK << GIGE_ARM_OUT_CHANNEL_SHIFT);

        reg_value |= armout_settings;
        regs[GIGE_REG_ARM_OUT_SELECT].set_value(reg_value);
    }
    else
    {
        regs[GIGE_REG_ARM_OUT_SELECT].set_value(armout_settings);
    }

    // All good.
    return GEV_STATUS_SUCCESS;
}

#if 0 /* This was implemented specifically for Alcon. */
/*!
 * @brief Force the Arm Out signal high or low using the test mode.
 *
 * @param state The state to set, high or low.
 *
 * @return true if success, false otherwise.
 */
bool FiFPGA::set_armout_state(bool state)
{
    uint32_t    extio_value;

    // Make sure the external transceiver is enabled.
    if (!write_reg(FI_FPGA_XCVR_ENABLE, true))
    {
        fi_log(LOG_ERR, "%s: Unable to write FI_FPGA_XCVR_ENABLE register.", __FUNCTION__);
        return false;
    }

    // Read the existing register settings.
    if (!read_reg(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __FUNCTION__);
        return false;
    }

    // Clear the passthrough enable. We are using the Test mode.
    extio_value &= ~EXTIO_ARMOUT_PASSTHRU_ENABLE;

    // Set the polarity bit based on the desired state.
    if (state)
        extio_value |= EXTIO_ARMOUT_ACTIVE_HIGH;
    else
        extio_value &= ~EXTIO_ARMOUT_ACTIVE_HIGH;

    // It needs to be an output and the state is derived from the polarity bit.
    extio_value |= (EXTIO_ARMOUT_SIG_IS_OUTPUT | EXTIO_ARM_OUT_TEST);

    // Write the new value back out to the FPGA.
    if (!write_reg(EXT_IO_CONTROL, extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __FUNCTION__);
        return false;
    }

    // All good.
    return true;
}
#endif /* Alcon code has not been tested recently so don't assume it works. */

/*!
 * @brief Figure out how many FPGA clocks there are in the requested
 *    pulse width.
 *
 * @param period_usecs  The number of microseconds for the pulse.
 *
 * @return true on success, false otherwise.
 */
bool FiFPGA::trigoutSetPulseWidth(uint32_t pulse_usecs)
{
    uint32_t    fpga_clocks;

    TP(FT_FUNC, "%s: pulse_usecs=%u", __PRETTY_FUNCTION__, pulse_usecs);

    // Multiply clocks per microsecond by number of microseconds.
    fpga_clocks = FPGA_CLOCK_SPEED_IN_USEC * pulse_usecs;

    // Set the output pulse width.
    if (!regWrite(TRIG_OUT_PULSE_WIDTH, fpga_clocks))
    {
        fi_log(LOG_ERR, "%s: Unable to write TRIG_OUT_PULSE_WIDTH!", __PRETTY_FUNCTION__);
        return false;
    }

    // All good.
    return true;
}

/*!
 * @brief Use the passed-in "trigout_settings" values to initialize
 *    the TrigOut controls in the FPGA.
 *
 * @param trigout_settings Our TrigOut settings to use.
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::triggerOutConfigure(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t trigout_settings)
{
    FiFPGA *fpga_ptr;
    uint32_t chan;
    uint32_t reg_value;
    uint32_t extio_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s: trigout_settings=0x%08x", __PRETTY_FUNCTION__, trigout_settings);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Get the current settings for the trigger out register.
    reg_value = regs[GIGE_REG_TRIG_OUT_SELECT].get_value();

    // If the trigger out logic was enabled, we need to make sure to turn off the
    // mux channel it was using in case the new settings change the channel or
    // disable the output.
    if ((reg_value & GIGE_TRIG_OUT_ENABLE) == GIGE_TRIG_OUT_ENABLE)
    {
        // Extract the channel information.
        chan = (reg_value >> GIGE_TRIG_OUT_CHANNEL_SHIFT) & GIGE_TRIG_OUT_CHANNEL_MASK;
        fpga_ptr->extioMuxDisable(chan, TRIG_OUT);
    }

    // Read the existing register settings from the hardware.
    if (!fpga_ptr->regRead(EXT_IO_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Clear out any existing TrigOut settings.
    extio_value &= ~(EXTIO_TRIGOUT_ENABLE | EXTIO_TRIGOUT_PASSTHRU_ENABLE);

    // Start setting bits as appropriate.
    if ((trigout_settings & GIGE_TRIG_OUT_IS_TRIG_IN) == GIGE_TRIG_OUT_IS_TRIG_IN)
    {
        // Setup to pass-through the TrigIn signal to the TrigOut pin.
        extio_value |= (EXTIO_TRIGOUT_PASSTHRU_ENABLE | EXTIO_TRIGOUT_ENABLE);

        // Figure out if the user wants to invert the TrigIn signal or not.
        if ((trigout_settings & GIGE_TRIG_OUT_POLARITY_MASK) == GIGE_TRIG_OUT_INVERT_TRIGIN)
            extio_value &= ~EXTIO_TRIGOUT_INVERT;
        else
            extio_value |= EXTIO_TRIGOUT_INVERT;
    }
    else if ((trigout_settings & GIGE_TRIG_OUT_ENABLE) == GIGE_TRIG_OUT_ENABLE)
    {
        uint32_t    pulse_usec;

        // Enable the TrigOut pin as an output.
        extio_value |= EXTIO_TRIGOUT_ENABLE;

        // Set the width of the output pulse in microseconds. The hardware
        // actually has a finer resolution but 1 usec should be enough.
        pulse_usec = (trigout_settings & GIGE_TRIG_OUT_PULSE_MASK) >> GIGE_TRIG_OUT_PULSE_SHIFT;
        fpga_ptr->trigoutSetPulseWidth(pulse_usec);

        // Determine if the user wants the TrigOut to be active High or Low.
        if ((trigout_settings & GIGE_TRIG_OUT_POLARITY_MASK) == GIGE_TRIG_OUT_ACTIVE_LOW)
            extio_value &= ~EXTIO_TRIGOUT_ACTIVE_HIGH;
        else
            extio_value |= EXTIO_TRIGOUT_ACTIVE_HIGH;
    }

    // Setup the mux.
    if ((extio_value & EXTIO_TRIGOUT_ENABLE) == EXTIO_TRIGOUT_ENABLE)
    {
        // Extract the channel information.
        chan = (trigout_settings >> GIGE_TRIG_OUT_CHANNEL_SHIFT) & GIGE_TRIG_OUT_CHANNEL_MASK;

        // If the channel is not set, then we aren't REALLY enabled.
        if (chan != 0)
        {
            // We have a channel, so go ahead and enable the circuit.
            fpga_ptr->extioMuxEnable(chan, TRIG_OUT);
        }
    }

    // Write the new value back out to the FPGA.
    if (!fpga_ptr->regWrite(EXT_IO_CONTROL, extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write EXT_IO_CONTROL register.", __FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    // If we're just being enabled/disabled don't lose the rest of the register settings.
    if ((trigout_settings == GIGE_TRIG_OUT_DISABLE) || (trigout_settings == GIGE_TRIG_OUT_ENABLE))
    {
        reg_value &= ~GIGE_TRIG_OUT_ENABLE;

        // Don't save the channel information on a disable!
        if (trigout_settings == GIGE_TRIG_OUT_DISABLE)
            reg_value &= ~(GIGE_TRIG_OUT_CHANNEL_MASK << GIGE_TRIG_OUT_CHANNEL_SHIFT);

        reg_value |= trigout_settings;
        regs[GIGE_REG_TRIG_OUT_SELECT].set_value(reg_value);
    }
    else
    {
        regs[GIGE_REG_TRIG_OUT_SELECT].set_value(trigout_settings);
    }

    // All good.
    return GEV_STATUS_SUCCESS;
}

int FiFPGA::imageUpload(uint32_t dma_mod, uint64_t src_addr, uint64_t dst_addr, uint32_t dma_size)
{
    uint32_t norm_size;
    ioImgUploadMessage msg;

    TP(FT_FUNC, "%s: dma_mod=0x%x, src_addr=0x%" PRIx64 ", dst_addr=0x%" PRIx64 ", dma_size=0x%08x (%u)",
       __PRETTY_FUNCTION__, dma_mod, src_addr, dst_addr, dma_size, dma_size);

    // Make sure we have a normalized size to transfer data from
    // image memory to the PC's system memory.
    norm_size = convert_to_HEXWORD(dma_size);
    norm_size = norm_size << HEXWORD_SHIFT;

    // Tell the FPGA to transfer down the frame located at byte_offset.
    msg.access.options = 0;
    msg.access.control = dma_mod | PCIE2IM_DMA_START;
    msg.access.im_address = convert_to_HEXWORD(dst_addr);
    msg.access.im_block_size = convert_to_HEXWORD(norm_size);
    msg.access.im_burst_size = IM2PCIE_DMA_BURST_SIZE;
    msg.access.im_bp_thresh = IM2PCIE_DMA_BP_THRESHOLD_VAL;
    msg.access.pc_address = convert_to_QWORD(src_addr);
    msg.access.pc_block_size = convert_to_QWORD(norm_size);
    msg.access.pc_burst_size = PCIE2IM_DMA_BURST_SIZE;
    msg.access.pc_burst_per = 0;
    msg.access.pc_bp_thresh = IM2PCIE_WR_DMA_BP_CMD_FIFO(0x04) | IM2PCIE_WR_DMA_BP_DATA_FIFO(0x80);
    msg.access.frame_dim = 0;

#if 0
    {
        LOG("control       = 0x%08x\nim_address    = 0x%08x\nim_block_size = 0x%08x\n"
            "im_burst_size = 0x%08x\nim_bp_thresh  = 0x%08x",
            msg.access.control, msg.access.im_address, msg.access.im_block_size,
            msg.access.im_burst_size, msg.access.im_bp_thresh);
        LOG("pc_address    = 0x%08x\npc_block_size = 0x%08x\npc_burst_size = 0x%08x\n"
            "pc_burst_per  = 0x%08x\npc_bp_thresh  = 0x%08x",
            msg.access.pc_address, msg.access.pc_block_size, msg.access.pc_burst_size,
            msg.access.pc_burst_per, msg.access.pc_bp_thresh);
        LOG("frame_dim     = 0x%08x", msg.access.frame_dim);
    }
#endif

    // Send the request.
    if (fpgaIOWorker::isRunning())
        msg.doSync(fpgaIOWorker::factory());

    // Success?
    if (msg.resp != FAS_THREAD::OK)
    {
        // Something didn't work...
        fi_log(LOG_DEBUG, "%s: Image upload failed! [status=%d]", __PRETTY_FUNCTION__, msg.status);
    }

    // Return the status.
    return msg.status;
}

int FiFPGA::imageDownload(uint32_t rd_dma_mod,
                          uint32_t rd_dma_dims,
                          uint64_t src_addr,
                          uint64_t dst_addr,
                          uint32_t src_size,
                          uint32_t dst_size)
{
    uint32_t pc_block_size;
    uint32_t im_block_size;
    ioImgDownloadMessage msg;

#ifdef DO_STATISTICS
    Fastec::IntervalTimer frame_timer;
    static double avg_time(0.0);
    static int avg_count(0);

    // Statistics
    frame_timer.start();
    // Statistics
#endif /* DO_STATISTICS */

    TP(FT_FUNC, "%s: dma_mod=0x%x, dma_dims=0x%08x, src_addr=0x%" PRIx64 ", dst_addr=0x%" PRIx64 ", src_size=0x%08x (%u), dst_size=0x%08x (%u)",
       __PRETTY_FUNCTION__, rd_dma_mod, rd_dma_dims, src_addr, dst_addr, src_size, src_size, dst_size, dst_size);

    // If no destination size was specified, then we assume it is the same as the source.
    if (dst_size == 0)
        dst_size = src_size;

    // If we are doing IMAGE data transfers, then we need to have a different
    // PC block size.
    if ((rd_dma_mod & IM2PCIE_DMA_IMAGE_DATA) == IM2PCIE_DMA_IMAGE_DATA)
    {
        // The size is different for the PC side and IM side.
        pc_block_size = convert_to_QWORD(dst_size) << QUADWORD_SHIFT;
        im_block_size = convert_to_HEXWORD(src_size) << HEXWORD_SHIFT;
    }
    else
    {
        // The size is the same on both ends of the DMA.
        pc_block_size = convert_to_HEXWORD(src_size) << HEXWORD_SHIFT;
        im_block_size = pc_block_size;
    }

    // Tell the FPGA to transfer down the frame located at byte_offset.
    msg.access.options = 0;
    msg.access.control = rd_dma_mod | IM2PCIE_DMA_START;
    msg.access.im_address = convert_to_HEXWORD(src_addr);
    msg.access.im_block_size = convert_to_HEXWORD(im_block_size);
    msg.access.im_burst_size = IM2PCIE_DMA_BURST_SIZE;
    msg.access.im_bp_thresh = IM2PCIE_DMA_BP_THRESHOLD_VAL;
    msg.access.pc_address = convert_to_QWORD(dst_addr);
    msg.access.pc_block_size = convert_to_QWORD(pc_block_size);
    msg.access.pc_burst_size = PCIE2IM_DMA_BURST_SIZE;
    msg.access.pc_burst_per = 0x8000;
    msg.access.pc_bp_thresh = IM2PCIE_WR_DMA_BP_CMD_FIFO(0x04) | IM2PCIE_WR_DMA_BP_DATA_FIFO(0x80);
    msg.access.frame_dim = rd_dma_dims;

#if 0
    {
        LOG("control       = 0x%08x\nim_address    = 0x%08x\nim_block_size = 0x%08x\n"
            "im_burst_size = 0x%08x\nim_bp_thresh  = 0x%08x",
            msg.access.control, msg.access.im_address, msg.access.im_block_size,
            msg.access.im_burst_size, msg.access.im_bp_thresh);
        LOG("pc_address    = 0x%08x\npc_block_size = 0x%08x\npc_burst_size = 0x%08x\n"
            "pc_burst_per  = 0x%08x\npc_bp_thresh  = 0x%08x",
            msg.access.pc_address, msg.access.pc_block_size, msg.access.pc_burst_size,
            msg.access.pc_burst_per, msg.access.pc_bp_thresh);
        LOG("frame_dim     = 0x%08x", msg.access.frame_dim);
    }
#endif

    // Send the request.
    if (fpgaIOWorker::isRunning())
    {
        msg.doSync(fpgaIOWorker::factory());
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: fpgaIOWorker is not running. In shutdown?", __PRETTY_FUNCTION__);
        msg.resp = FAS_THREAD::FAULT;
        msg.status = -EPERM;
    }

    // Success?
    if (msg.resp != FAS_THREAD::OK)
    {
        // Something didn't work...
        fi_log(LOG_DEBUG, "%s: Image download failed! [status=%d]", __PRETTY_FUNCTION__, msg.status);
    }

#ifdef DO_STATISTICS
    // Statistics
    frame_timer.stop();

    // Accumulate our average time.
    if (avg_count < 10)
    {
        ++avg_count;
        avg_time += frame_timer.lastMeasurement;
        if (avg_count == 10)
        {
            avg_time /= avg_count;
            LOG("%s: avg_time=%lf", __PRETTY_FUNCTION__, avg_time);
        }
    }
    else if (avg_count == 10)
    {
        if (frame_timer.lastMeasurement > (10 * avg_time))
        {
            LOG("%s: %lf secs. !!!!!!!!!!!!!!!!!!!", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
        }
    }
    // Statistics
#endif /* DO_STATISTICS */

    // Return the status.
    TP(FT_FUNC, "%s--: status=%d", __PRETTY_FUNCTION__, msg.status);
    return msg.status;
}

void FiFPGA::dumpOneReg(void *p, uint32_t addr, const char *name)
{
    FILE *fp = (FILE *)p;
    uint32_t reg_value;

    if (regRead(addr, &reg_value))
        fprintf(fp, "%s[%08x] = 0x%08x = %u\n", name, addr, reg_value, reg_value);
}

/*!
 * @brief Dump out all the FPGA registers.
 */
void FiFPGA::dump_regs()
{
    FILE     *fp;
    uint32_t base;
    uint32_t reg;
    char temp_path[PATH_MAX];

#if 1
    snprintf(temp_path, PATH_MAX, "%s/%s", get_home_path(), ".config/Fastec/scratch/registers.txt");
    fp = fopen(temp_path, "a");
    if (fp == NULL)
    {
        fi_log(LOG_DEBUG, "%s: can't open .config/Fastec/scratch/registers.txt, errno=%d", __FUNCTION__, errno);
        return;
    }

#if 1
#if 1
    base = FI_FPGA_MAIN_OFFSET;
    for (reg = base; reg <= base + 0x00fc; reg += 4)
        dumpOneReg(fp, reg, "MAIN");
    for (reg = base + 0xa00; reg <= (base + 0x0a0c); reg += 4)
        dumpOneReg(fp, reg, "P1");
#endif
#if 1
    base = FI_FPGA_ICM_OFFSET;
    for (reg = base; reg <= base + 0x007c; reg += 4)
        dumpOneReg(fp, reg, "ICM");
    for (reg = base + 0x200; reg <= base + 0x02a0; reg += 4)
        dumpOneReg(fp, reg, "LINCE");
#endif
#if 1
    base = FI_FPGA_IPM0_OFFSET;
    for (reg = base; reg <= (base + 0x004c); reg += 4)
        dumpOneReg(fp, reg, "IPM");
    for (reg = base + 0x0100; reg <= (base + 0x0178); reg += 4)
        dumpOneReg(fp, reg, "IPM");
#endif

#else
    dump_one_reg(fp, FPGA_CONTROL,    "FPGA_CONTROL");
    dump_one_reg(fp, FPGA_CONTROL,    "FPGA_STATUS");

    dump_one_reg(fp, FPGA_IRQ_STATUS, "FPGA_IRQ_STATUS");
    dump_one_reg(fp, FPGA_IRQ_MASK,   "FPGA_IRQ_MASK");
    dump_one_reg(fp, FPGA_REVISION,   "FPGA_REVISION");

    dump_one_reg(fp, ICM_CONTROL,                "ICM_CONTROL");
    dump_one_reg(fp, ICM_STATUS,                 "ICM_STATUS");
    dump_one_reg(fp, ICM_SENSOR_WIN_X,           "ICM_SENSOR_WIN_X");
    dump_one_reg(fp, ICM_SENSOR_WIN_Y,           "ICM_SENSOR_WIN_Y");
    dump_one_reg(fp, ICM_MEM_WIN_X,              "ICM_MEM_WIN_X");
    dump_one_reg(fp, ICM_MEM_WIN_Y,              "ICM_MEM_WIN_Y");
    dump_one_reg(fp, ICM_CAPTURE_START_ADDR,     "ICM_CAPTURE_START_ADDR");
    dump_one_reg(fp, ICM_CAPTURE_FRAME_BUF_SIZE, "ICM_CAPTURE_FRAME_BUF_SIZE");
    dump_one_reg(fp, ICM_CAPTURE_FRAME_BUF_CNT,  "ICM_CAPTURE_FRAME_BUF_CNT");
    dump_one_reg(fp, ICM_POST_TRIG_CNT,          "ICM_POST_TRIG_CNT");
    dump_one_reg(fp, ICM_ULV_BUF_START_ADDR,     "ICM_ULV_BUF_START_ADDR");
    dump_one_reg(fp, ICM_SLAVE_MODE_CTRL,        "ICM_SLAVE_MODE_CTRL");
    dump_one_reg(fp, ICM_FRAME_PERIOD,           "ICM_FRAME_PERIOD");
    dump_one_reg(fp, ICM_INTEGRATION_TIME,       "ICM_INTEGRATION_TIME");

    dump_one_reg(fp, IPM_CONTROL, "IPM_CONTROL");
    dump_one_reg(fp, IPM_STATUS,  "IPM_STATUS");

    dump_one_reg(fp, IPM_SRC_DMA_REQ_SIZE,      "IPM_SRC_DMA_REQ_SIZE");
    dump_one_reg(fp, IPM_SRC_DMA_BP_THRESHOLD,  "IPM_SRC_DMA_BP_THRESHOLD");
    dump_one_reg(fp, IPM_DEST_DMA_REQ_SIZE,     "IPM_DEST_DMA_REQ_SIZE");
    dump_one_reg(fp, IPM_DEST_DMA_BP_THRESHOLD, "IPM_DEST_DMA_BP_THRESHOLD");
    dump_one_reg(fp, IPM_P0_CONFIG,             "IPM_P0_CONFIG");
    dump_one_reg(fp, IPM_P0_STATUS,             "IPM_P0_STATUS");
    dump_one_reg(fp, IPM_P0_SRC_DMA_BLOCK_SIZE, "IPM_P0_SRC_DMA_BLOCK_SIZE");
    dump_one_reg(fp, IPM_P0_FRAME_DIM,          "IPM_P0_FRAME_DIM");
    dump_one_reg(fp, IPM_P0_RGB_GAIN_RED,       "IPM_P0_RGB_GAIN_RED");
    dump_one_reg(fp, IPM_P0_RGB_GAIN_GRN,       "IPM_P0_RGB_GAIN_GRN");
    dump_one_reg(fp, IPM_P0_RGB_GAIN_BLU,       "IPM_P0_RGB_GAIN_BLU");

    dump_one_reg(fp, DCM_CONTROL, "DCM_CONTROL");
    dump_one_reg(fp, DCM_STATUS, "DCM_STATUS");
    dump_one_reg(fp, DCM_TICK_PERIOD, "DCM_TICK_PERIOD");
    dump_one_reg(fp, DCM_DISP_BUF_START_ADDR, "DCM_DISP_BUF_START_ADDR");
    dump_one_reg(fp, DCM_DISP_BUF_MAN_UP_SRC_PTR, "DCM_DISP_BUF_MAN_UP_SRC_PTR");

    dump_one_reg(fp, CSI_CONTROL,           "CSI_CONTROL");
    dump_one_reg(fp, CSI_STATUS,            "CSI_STATUS");
    dump_one_reg(fp, CSI_DMA_BLOCK_SIZE,    "CSI_DMA_BLOCK_SIZE");
    dump_one_reg(fp, CSI_DMA_FRAME_SIZE,    "CSI_DMA_FRAME_SIZE");
    dump_one_reg(fp, CSI_DMA_DDR2_REQ_SIZE, "CSI_DMA_DDR2_REQ_SIZE");
    dump_one_reg(fp, CSI_DMA_BP_THRESHOLDS, "CSI_DMA_BP_THRESHOLDS");
    dump_one_reg(fp, CSI_DISPLAY_CONFIG_H,  "CSI_DISPLAY_CONFIG_H");
    dump_one_reg(fp, CSI_DISPLAY_CONFIG_V,  "CSI_DISPLAY_CONFIG_V");

    dump_one_reg(fp, IPM_P0_WB_GAIN_FACTOR_RED, "IPM_P0_WB_GAIN_FACTOR_RED");
    dump_one_reg(fp, IPM_P0_WB_GAIN_FACTOR_GRN, "IPM_P0_WB_GAIN_FACTOR_RED");
    dump_one_reg(fp, IPM_P0_WB_GAIN_FACTOR_BLU, "IPM_P0_WB_GAIN_FACTOR_RED");

    // Matrix coefficients
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF1_RED, "IPM_P0_CC_MATRIX_COEFF1_RED");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF2_RED, "IPM_P0_CC_MATRIX_COEFF2_RED");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF3_RED, "IPM_P0_CC_MATRIX_COEFF3_RED");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF1_GRN, "IPM_P0_CC_MATRIX_COEFF1_GRN");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF2_GRN, "IPM_P0_CC_MATRIX_COEFF2_GRN");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF3_GRN, "IPM_P0_CC_MATRIX_COEFF3_GRN");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF1_BLU, "IPM_P0_CC_MATRIX_COEFF1_BLU");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF2_BLU, "IPM_P0_CC_MATRIX_COEFF2_BLU");
    dump_one_reg(fp, IPM_P0_CC_MATRIX_COEFF3_BLU, "IPM_P0_CC_MATRIX_COEFF3_BLU");
#endif

#if 0
    Sensor *sensor_ptr = CameraServer::get_sensor();
    sensor_ptr->dump_sensor_regs(fp);
#endif

    fclose(fp);
#endif
}

/*!
 * @brief Utility method to read a 32-bit value from image memory.
 *
 * @param peek_addr     The address in image memory to read from.
 * @return The value read from image memory.
 */
uint32_t FiFPGA::imPeek(uint64_t peek_addr)
{
    uint32_t peek_value = 0;

    // The address coming in is a byte address. Need to convert to a HEXWORD address for the FPGA.
    if (!regWrite(IM_ARB_PIO_ADDRESS, convert_to_HEXWORD(peek_addr)))
        goto bail;
    if (!regWrite(IM_ARB_PIO_RD_DATA, 0))
        goto bail;
    if (!regReadWait(IM_ARB_CONFIG_STATUS, CONFIG_STATUS_PIO_BUSY, 0, IM_ARB_PIO_RD_DATA, &peek_value))
    {
        fi_log(LOG_DEBUG, "%s: readwait_reg() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

bail:
    return peek_value;
}

/*!
 * @brief Utility method to write a 32-bit value to image memory.
 *
 * @param poke_addr     The address in image memory to write to.
 * @param poke_value    The value to write.
 */
void FiFPGA::imPoke(uint64_t poke_addr, uint32_t poke_value)
{
    uint32_t dummy;

    if (!regWrite(IM_ARB_PIO_ADDRESS, convert_to_HEXWORD(poke_addr)))
        return;
    if (!regWrite(IM_ARB_PIO_WR_DATA, poke_value))
        return;
    if (!regReadWait(IM_ARB_CONFIG_STATUS, CONFIG_STATUS_PIO_BUSY, 0, IM_ARB_PIO_RD_DATA, &dummy))
    {
        fi_log(LOG_DEBUG, "%s: readwait_reg() failed", __PRETTY_FUNCTION__);
    }
}

bool FiFPGA::isShutterOpen()
{
    uint32_t reg_value(0);

    // Return true if the shutter is open (i.e. NOT closed/enabled).
    regRead(NC1_FPGA_CONTROL, &reg_value);
    return ((reg_value & NC1FC_SHUTTER_ENABLE) == 0);
}

void FiFPGA::shutterClose(bool close)
{
    // Open or close the shutter.
    regModify(NC1_FPGA_CONTROL, (close ? NC1FC_SHUTTER_ENABLE : 0), ~NC1FC_SHUTTER_ENABLE);

    // Give the shutter a chance to react.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(100));
}

bool FiFPGA::isExtio_5V()
{
    bool is5V(false);
    uint32_t extio_value;

    // Read the existing register settings.
    if (!regRead(EXT_IO_MUX_CONTROL, &extio_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read EXT_IO_CONTROL register.", __FUNCTION__);
        goto is5V_bail;
    }

    // Set our return flag.
    is5V = ((extio_value & EXTIO_MUX_IOB_EXTIO_VSEL_3_3V) != EXTIO_MUX_IOB_EXTIO_VSEL_3_3V);

is5V_bail:
    return is5V;
}

/*!
 * @brief Configure the external I/O to be at 5V or 3.3V.
 * @param use_5V        [IN] True to use 5V, False for 3.3V.
 */
void FiFPGA::extioSetVoltage(bool use_5V)
{
    // For 5V, we need to clear the bit. For 3.3V we need to set the bit.
    if (use_5V)
        regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IOB_EXTIO_VSEL_5V, ~EXTIO_MUX_IOB_EXTIO_VSEL_3_3V);
    else
        regModify(EXT_IO_MUX_CONTROL, EXTIO_MUX_IOB_EXTIO_VSEL_3_3V, -1);
}

/*!
 * @brief Tickle the FPGA watchdog timer.
 *
 * If the timer has expired, then we want to log the event and reset/re-enable the timer.
 */
void FiFPGA::watchdog()
{
    uint32_t wd_status = 0;

    // Read the status.
    if (!regRead(WATCHDOG_CONTROL_STATUS, &wd_status))
    {
        fi_log(LOG_ERR, "%s: Unable to read WATCHDOG_CONTROL_STATUS register.", __PRETTY_FUNCTION__);
        return;
    }
    if ((wd_status & WD_EXPIRED) == WD_EXPIRED)
    {
        fi_log(LOG_WARNING, "%s: Watchdog expired!", __PRETTY_FUNCTION__);

        // Reset the error.
        regWrite(WATCHDOG_CONTROL_STATUS, WD_EXPIRED);
        regWrite(WATCHDOG_CONTROL_STATUS, WD_ENABLE);
    }

    // Pet the dog.
    regWrite(WATCHDOG_TICKLE, WD_TICKLE_MAGIC);
}
