/*!
 * @file fpga_icm.cxx
 * @brief This source file implements the Fastec Imaging FPGA
 *    class, FiFPGA, Image Capture Module (ICM) methods.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "autosave.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "fpga.h"
#include "fpga_io.h"
#include "fcp.h"
#include "grpc_server.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "ltr_worker.h"
#include "mft.h"
#include "part_mgr.h"
#include "pb_control.h"

#include "override.h"

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

namespace FI_FPGA_ICM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Developer note: The primary difference between const and constexpr variables is that
// the initialization of a const variable can be deferred until run time. A constexpr
// variable must be initialized at compile time.

static constexpr int        MAX_RETRY_COUNT     = 10;           //!< Retry loop count.
static constexpr uint32_t   AUX_DMA_BUSY_WAIT   = 100000;       //!< Wait 100ms for the AUX DMA to free up.
static constexpr uint32_t   FPGA_ICM_EVENT_FIFO_LVL_MASK    = 0x3FF;     //!< Only want the level value, not the other bits.
static constexpr uint32_t   FPGA_ICM_POST_TRIG_FRAME_MASK   = 0x01FFFFFF;
static constexpr uint32_t   PIPELINE_WAIT       = 100000;       //!< Wait per read attempt for the pipeline to be ready.
static constexpr uint32_t   DMA_BURST_SIZE      = 32;
static constexpr uint32_t   UNDERFLOW_FLAG_MASK = (1U << 31);   //!< This bit is set in the end-of-capture extra info word when we had an underflow in the capture FIFO.

//! Wait per read attempts waiting for ICMS_PIPELINE_READY - 5 milliseconds.
static constexpr uint32_t   PIPE_DISABLE_READY_WAIT = 5000;

//! Maximum read attempts waiting for ICMS_PIPELINE_READY - 50 milliseconds.
static constexpr int        PIPE_DISABLE_READY_RETRIES = 10;

static constexpr uint32_t   ICM_BUSY_WAIT_SEC   = 0;            //!< Wait 100ms for the ICM to free up.
static constexpr uint32_t   ICM_BUSY_WAIT_NSEC  = 100 * 1000 * 1000;
}

using namespace FI_FPGA;
using namespace FI_FPGA_ICM;
using namespace FI_FPGA_IO;

//====================================================================

/*!
 * @brief Utility function to decode the event register. For debug only.
 *
 * @param event_val Value retrieved from the FPGA.
 *
 * @return Pointer to the decoded string.
 */
static const char *decode_event(uint32_t event_val)
{
    static std::string temp_buf;

    // Set our buffer to empty.
    temp_buf.clear();

    // Now parse out the meanings...
    if (event_val & PCIEIRQ_FIFO_NOT_EMPTY)
        temp_buf.append("FIFO ");
    if (event_val & PCIEIRQ_P1_FRAME_DONE)
        temp_buf.append("P1 ");
    if (event_val & PCIEIRQ_IM2PCIE_DMA_DONE)
        temp_buf.append("IM2PCI ");
    if (event_val & PCIEIRQ_PCIE2IM_DMA_DONE)
        temp_buf.append("PCI2IM ");
    if (event_val & PCIEIRQ_LR_DESC_FIFO_AF)
        temp_buf.append("LRAF ");
    if (event_val & PCIEIRQ_PWR_OFF)
        temp_buf.append("PWR ");
    if (event_val & PCIEIRQ_EXT_ARM_INPUT)
        temp_buf.append("ARM ");
    if (event_val & PCIEIRQ_EXT_UNARM_INPUT)
        temp_buf.append("ARM ");
    if (event_val & PCIEIRQ_MFT_LSDET)
        temp_buf.append("LSDET ");
    if (event_val & PCIEIRQ_MFT_LSLPSW)
        temp_buf.append("LSLPSW ");

    // Return a pointer to our buffer.
    return temp_buf.c_str();
}

/*!
 * @brief Stop a capture that didn't complete normally by simulating
 *    a completion event (and an end trigger if no trigger so far).
 *    This is also the only way out of ROC/BROC mode.
 *
 * @param triggered TRUE if we already processed a trigger.
 *
 * @return true if there were any frames captured yet
 */
bool FiFPGA::icmStopCapture(bool saw_trigger)
{
    PartMgr *partmgr_ptr;           // Pointer to the partition manager class.
    bool    any_frames;
    uint32_t    icm_count;
    uint32_t    extra_trigger_info[2] = {0, 0};
    uint64_t    frame_addr;
    uint32_t    post_trigger_frames;
    uint32_t    value;
    int     i;

    TP(FT_ICM, "%s: saw_trigger:%d ================================", __PRETTY_FUNCTION__, saw_trigger);

    // Get a pointer to our support classes.
    partmgr_ptr = CameraServer::get_partmgr();

    // Sanity check.
    if (!isInitialized())
    {
        fi_log(LOG_ERR, "%s: FPGA not initialized!", __PRETTY_FUNCTION__);
        return false;
    }

    // Create a block so that we UNARM and make sure that the ICM stops
    // before we or anyone else can move on.
    {
        std::unique_lock<std::mutex> locker(class_mutex);

        // Set the unarm bit.
        regModify(ICM_CONTROL, ICMC_UNARM, -1);

        // Wait for unarming to complete.
        for (i = 0; i < MAX_RETRY_COUNT; ++i)
        {
            struct timespec sleep_req;      // Delay period.

            // Wait for a bit.
            sleep_req.tv_sec = ICM_BUSY_WAIT_SEC;
            sleep_req.tv_nsec = ICM_BUSY_WAIT_NSEC;
            nanosleep(&sleep_req, NULL);

            // Check for the bit to go low.
            regRead(ICM_STATUS, &value);
            if ((value & ICMS_ARMED) == 0)
                break;
        }
    }

    // Did we exceed our retry count?
    if (i == MAX_RETRY_COUNT)
        fi_log(LOG_ERR, "%s: ICMS_ARMED never dropped [0x%x]", __FUNCTION__, value);

    // Clear the SW metadata marker bit in case we never captured any frame.
    regWrite(ICM_SW_METADATA_MARKER, 0x00000000);

    // Finish LTR hardware cleanup.
    if ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Shutdown the LTR processing.
        regModify(ICM_CONTROL, 0, ~ICMC_LR_ENABLE);
        ltrEnable(false);

        // Read the final address on the SSD.
        if (ltr_thread.get() != nullptr)
            frame_addr = ltr_thread->getCurAddr();
        else
            frame_addr = 0;
    }
    else
    {
        // Read and return the current addr from the FPGA.
        regRead(ICM_LAST_FRAME_WRITTEN_PTR, &value);
        frame_addr = ((uint64_t)value) << HEXWORD_SHIFT;
        if (value == LAST_FRAME_INIT)
            icm_count = 0;
    }

    // Create a block so that we get the time registers as close together
    // as possible.
    {
        std::unique_lock<std::mutex> locker(class_mutex);

        // We need the frame count to see if we've wrapped.
        // This register becomes valid only after unarming.
        regRead(ICM_CAPTURE_FRAME_COUNT_LATCH, &icm_count);

        // If we had not previously seen the trigger, simulate one at t=now.
        if (!saw_trigger)
        {
            FI_REG_ACCESS access;
            ioFpgaGetMultiRegisterMessage msg;

            // Get the current time and use that for the trigger time.
            access.reg_id = TIME_SBS_COUNTER_LIVE;
            msg.access.push_back(access);
            access.reg_id = TIME_SBUS_COUNTER_LIVE;
            msg.access.push_back(access);
            access.reg_id = TIME_SBS_COUNTER_LIVE;
            msg.access.push_back(access);

            // Send the request.
            if (FI_FPGA_IO::fpgaIOWorker::isRunning())
                msg.doSync(FI_FPGA_IO::fpgaIOWorker::factory());
            if (msg.resp == FAS_THREAD::OK)
            {
                // Retrieve the values read.
                extra_trigger_info[0] = msg.access[0].value;
                extra_trigger_info[1] = msg.access[1].value;
                value = msg.access[2].value;

                // Allow for an unfortunately-timed rollover.
                if (value != extra_trigger_info[0])
                {
                    regRead(TIME_SBS_COUNTER_LIVE, &extra_trigger_info[0]);
                    regRead(TIME_SBUS_COUNTER_LIVE, &extra_trigger_info[1]);
                }
            }
        }
    }

    // We need the post trigger frame count to know how many pre-trigger frames
    // to retain.
    post_trigger_frames = icmPostTriggerCount() & FPGA_ICM_POST_TRIG_FRAME_MASK;

    // Flush any events that were queued while we ignored them.
    icmEmptyEventFIFO();

    // Finish LTR hardware cleanup.
    if ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        regModify(ICM_CONTROL, 0, ~ICMC_LR_ENABLE);
        ltrEnable(false);
    }

    // Simulate completion of the recording
    any_frames = partmgr_ptr->simulateCompletion(saw_trigger, post_trigger_frames, frame_addr, icm_count, extra_trigger_info);

    // Make sure the capture FIFO is cleared.
    icmResetCaptureFIFO();

    // Return our status.
    TP(FT_ICM, "%s: any_frames:%d ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^", __PRETTY_FUNCTION__, any_frames);
    return any_frames;
}

/*!
 * @brief We received a trigger event from the FPGA.
 *
 * The FPGA issues an interrupt and puts data into the event queue on a trigger event. This data
 * is as follows:
 * <table>
 *      <caption id="TrigEventInfo">Trigger Event Information</caption>
 *      <tr><th>Word <th>Description
 *      <tr><td> 0  <td>Frame information.
 *      <tr><td> 1  <td>The time in seconds of the trigger event.
 *      <tr><td> 2  <td><table cellspacing="0" cellpadding="0">
 *          <tr><td> b31    <td>IRIG Lock state
 *          <tr><td> b24:29 <td>External marker inputs [5:0]
 *          <tr><td> b0:19  <td>The microseconds part of the trigger time.
 *      </table>
 *  </table>
 *
 * @param frame_info            [IN] The value read from the queue ANDED against ICMEF_TRIGGER_FRAME_MASK.
 * @param extra_trigger_info    [IN] Pointer to the two words of extra information (words 1 & 2 above).
 */
void FiFPGA::icmTriggerEventHappened(uint32_t frame_info, const uint32_t *extra_trigger_info)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr *partmgr_ptr;           // Pointer to the partition manager class.
    bool    in_ltr_mode;
    uint64_t    frame_addr;
    uint32_t    cap_mode;

    TP(FT_ICM, "%s: frame_info:0x%08x", __PRETTY_FUNCTION__, frame_info);

    // Get a pointer to our support class.
    partmgr_ptr = CameraServer::get_partmgr();

    in_ltr_mode = ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    cap_mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

    // In ROC or BROC mode, a trigger is just "run" event, not a real trigger.
    if ((ltrFasFireEnabled() || in_ltr_mode) && (cap_mode != 0))
    {
        TP(FT_ICM, "%s: FFLR ROC/BROC (0x%x)", __PRETTY_FUNCTION__, cap_mode);
        frame_addr = ((uint64_t)frame_info) << HEXWORD_SHIFT;
        partmgr_ptr->partitionEventTrigger(frame_addr, extra_trigger_info);
        CameraServer::set_camera_state(GIGE_CAMERA_STATE_TRIGGERED);
        return;
    }

    // We got a trigger event, but we have to handle this differently
    // if we're in LTR mode.
    if (in_ltr_mode)
    {
        uint32_t frame_num;

        // Get the frame number from the image memory capture information.
        frame_num = partmgr_ptr->convertCaptureAddrToFrameNum(((uint64_t)frame_info) << HEXWORD_SHIFT);

        // Now convert that frame number to an SSD offset for the frame. Note that this frame number
        // is an absolute frame number from the beginning of the capture space (i.e. a '0' would be
        // the first frame of the capture space).
        frame_addr = partmgr_ptr->getPartitionFrameAddr(partmgr_ptr->getActivePartition(), frame_num, true);
        TP(FT_ICM, "%s: frame_info=0x%x, frame_num=%d, frame_addr=%" PRIx64, __PRETTY_FUNCTION__, frame_info, frame_num, frame_addr);
    }
    else
    {
        // Convert the frame information into a 64-bit address.
        frame_addr = ((uint64_t)frame_info) << HEXWORD_SHIFT;
        TP(FT_ICM, "%s: frame_info=0x%x, frame_addr=%" PRIx64, __PRETTY_FUNCTION__, frame_info, frame_addr);
    }

    // Tell the partition manager that we had an event.
    partmgr_ptr->partitionEventTrigger(frame_addr, extra_trigger_info);

    // Got a trigger event from the FPGA. We MAY already be in the
    // trigger state (if a remote app set a soft trigger), or we may
    // not (if the user pressed the trigger button on the camera).
    // Make certain that we're in the trigger state.
    CameraServer::set_camera_state(GIGE_CAMERA_STATE_TRIGGERED);
}

/*!
 * @brief Handle an End of Capture event read from the ICM event queue.
 *
 * @param frame_info        [IN] The value read from the queue ANDED against ICMEF_TRIGGER_FRAME_MASK.
 * @param last_frame_addr   [IN] The value read from the queue, it is the last address written by the FPGA.
 */
void FiFPGA::icmEndCaptureEventHappened(uint32_t frame_info, uint64_t last_frame_addr, bool underflow)
{
    uint32_t icm_count;
    uint32_t post_trigger_frames;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    bool in_ltr_mode;
    uint32_t cap_mode;
    PartMgr *partmgr_ptr;           // Pointer to the partition manager class.
    AutoSave    *as_mgr;

    TP(FT_ICM, "%s: frame_info:%u, last_frame_addr:0x%" PRIx64 ", underflow=%d", __PRETTY_FUNCTION__, frame_info, last_frame_addr, underflow);

    // Get a pointer to our support classes.
    partmgr_ptr = CameraServer::get_partmgr();
    as_mgr = CameraServer::get_autosave();

    in_ltr_mode = ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    cap_mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

    // We already verified that we were in the triggered state when we got this End event.
    // In ROC or BROC mode, end capture is just a "pause" event, not a real end of capture.
    if (cap_mode != 0)
    {
        // Get the maximum number of frames we're allowed to capture.
        regRead(ICM_ROC_MAX_CAPTURE_FRM_CNT, &icm_count);

        // Are we in "normal" LR mode?
        if (in_ltr_mode)
        {
            TP(FT_ICM, "%s: Normal LR ROC/BROC processing", __PRETTY_FUNCTION__);

            // Figure out if we left capture because we've filled the capture region.
            // If so, we will need to stop the capture and enter the Review state.
            if (frame_info >= icm_count)
            {
                // Drive is full, time to stop.
                icmStopCapture(true);

                // We'll store a save error state so the user gets an error message.
                // ENOSPC is "No space left on device"
                ltrSetSaveError(GIGE_SAVE_STATE_ERR | ENOSPC);

                // Of course we captured something, we're full, so go to Review.
                CameraServer::set_camera_state(GIGE_CAMERA_STATE_REVIEW);
            }
            else
            {
                fi_log(LOG_INFO, "%s: At ROC/BROC trigger/done.", __PRETTY_FUNCTION__);

                // Have our worker thread flush the last frames to the SSD.
                if (ltr_thread.get() != nullptr)
                {
                    // Tell the thread to flush out the frames.
                    ltr_thread->flushSSD();
                }

                // Switch back to the ARMed state.
                CameraServer::set_camera_state(GIGE_CAMERA_STATE_ARMED);
            }

            // All done with LR ROC/BROC modes.
            return;
        }

        else if (ltrFasFireEnabled())
        {
            TP(FT_ICM, "%s: FFLR ROC/BROC processing", __PRETTY_FUNCTION__);

            // Check to see if we've exceeded the record space.
            if (frame_info >= icm_count)
            {
                TP(FT_ICM, "%s: FFLR ROC/BROC overflow! frame_info [0x%x] >= ICM_ROC_MAX_CAPTURE_FRM_CNT [0x%x]", __PRETTY_FUNCTION__, frame_info, icm_count);

                // Drive is full, time to stop. We definitely saw a trigger - wouldn't be here otherwise.
                icmStopCapture(true);

                // Set the LR mode save status as an error.
                ltrSetSaveError(GIGE_SAVE_STATE_ERR | ENOSPC);

                // Now that we've set the error, we want to save whatever we've recorded and terminate as normal.
                // To do that, we simply fall through into a "normal" capture mode.
            }
            else
            {

                // Figure out if we left capture because we've underflowed the available buffering.
                // If so, we will need to stop the capture and enter the Review state.
                if (underflow)
                {
                    // We underflowed the capture FIFO. Stop the capture.
                    fi_log(LOG_WARNING, "%s: FFLR ROC/BROC underflowed", __PRETTY_FUNCTION__);

                    // Stop autosaving.
                    as_mgr->stop();

                    // Set the LR mode save status as an error.
                    ltrSetSaveError(GIGE_SAVE_STATE_ERR | ENOMEM);

                    // Tell the image saver object.
                    imgSaveWorker::getInstance()->setFflrSaveError(GIGE_SAVE_STATE_ERR | ENOMEM);

                    // Not triggered, so we must be done with the capture.
                    // Notify the partition manager of this event so it can
                    // clean things up.
                    partmgr_ptr->partitionEventDone(last_frame_addr, icm_count, icm_count);

                    // Of course we captured something, we're full, so go to Review.
                    CameraServer::set_camera_state(GIGE_CAMERA_STATE_REVIEW);
                }

                else
                {
                    // Normal end of capture for ROC/BROC. Switch back to the ARMed state.
                    TP(FT_ICM, "%s: FFLR ROC/BROC trigger/done.", __PRETTY_FUNCTION__);
                    CameraServer::set_camera_state(GIGE_CAMERA_STATE_ARMED);
                }

                // All done with FFLR ROC/BROC.
                return;
            }
        }
    }

    // We got an end-of-capture event, but we have to handle this differently
    // if we're in LTR mode.
    if (in_ltr_mode)
    {
        // Stop the capture. Note that this "requests" the LTR thread to stop.
        // It can be awhile before the machine finishes and shuts down.
        ltrEnable(false);
    }

    // Read up the finalized data from the FPGA.
    regRead(ICM_CAPTURE_FRAME_COUNT_LATCH, &icm_count);

    TP(FT_ICM, "%s: ICM_CAPTURE_FRAME_COUNT_LATCH=%d, ICM_LAST_FRAME_WRITTEN_PTR=0x%" PRIx64, __PRETTY_FUNCTION__, icm_count, last_frame_addr);
    if (convert_to_HEXWORD(last_frame_addr) == LAST_FRAME_INIT)
    {
        last_frame_addr = partmgr_ptr->get_active_frame_addr(0, true);
        fi_log(LOG_DEBUG, "%s: Using calculated lat_frame_addr[0x%" PRIx64 "] instead !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", __PRETTY_FUNCTION__, last_frame_addr);
        fi_log(LOG_DEBUG, "%s: ICM_LAST_FRAME_WRITTEN_PTR==LAST_FRAME_INIT, resetting frame_info and icm_count to 0. !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", __PRETTY_FUNCTION__);
        frame_info = icm_count = 0;
    }

    // For the end-of-capture event, the "frame_info" value is actually
    // the number of frames written since start of capture.
    if (frame_info != icm_count)
    {
        fi_log(LOG_DEBUG, "%s: Found (ICM_CAPTURE_FRAME_COUNT_LATCH=%u) != (event=%u)", __FUNCTION__, icm_count, frame_info);
        icm_count = frame_info;
    }

    // Find out how many post trigger frames we had specified.
    // We're only interested in the frame count, so mask out the unused bits.
    post_trigger_frames = icmPostTriggerCount() & FPGA_ICM_POST_TRIG_FRAME_MASK;

    // Not triggered, so we must be done with the capture.
    // Notify the partition manager of this event so it can
    // clean things up.
    partmgr_ptr->partitionEventDone(last_frame_addr, post_trigger_frames, icm_count);

    // Are we in FasFireLR Mode?
    if (ltrFasFireEnabled())
    {
        TP(FT_ICM, "%s: FFLR end capture. icm_count=%d, ltrSaveError=0x%08x", __PRETTY_FUNCTION__, icm_count, ltrSaveError());

        // Make sure we actually captured frames.
        if (icm_count != 0)
        {
            // Go into the autosave. This should queue up the partition to be written.
            // Note that you can't skip this step and just go straight to live. If you do
            // that, then the last partition won't be queued up to be saved.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_AUTOSAVE);
        }

        // Did we get here because we underflowed the capture buffer FIFO? Check to see if we are still armed.
        if (underflow)
        {
            // We'll store a save error state so the user gets an error message.
            fi_log(LOG_ERR, "%s: FFLR capture buffer FIFO underflowed!", __PRETTY_FUNCTION__);

            // Stop autosaving.
            as_mgr->stop();

            // Set the LR mode save status as an error.
            ltrSetSaveError(GIGE_SAVE_STATE_ERR | ENOMEM);

            // Tell the image saver object.
            imgSaveWorker::getInstance()->setFflrSaveError(GIGE_SAVE_STATE_ERR | ENOMEM);

            // Of course we captured something, we're full, so go to Review.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_REVIEW);
        }
        else
        {
            // Normal end of capture, so transition to the live state.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
    }
    else
    {
        // If we did not capture any frames, go into live view.
        if (icm_count == 0)
        {
            // PM should already have set partition state to EMPTY.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
            fi_log(LOG_DEBUG, "%s: icm_count == 0", __PRETTY_FUNCTION__);
        }

        // Transition to either the "review" or "auto-save" state.
        else if ((regs[ GIGE_REG_AUTOSAVE ].get_value() & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
        {
            AutoSave    *as_mgr;
            as_mgr = CameraServer::get_autosave();

            // Go into autosave and queue up the partition to be written.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_AUTOSAVE);

            // Are we interested in overlapping the save operation?
            TP(FT_ICM, "%s: GIGE_AUTOSAVE_ENABLE is true", __PRETTY_FUNCTION__);
            if (as_mgr->doOverlap())
            {
                // While we were futzing around, the FPGA has already re-armed and started
                // capturing into the "busy" partition. We need to catch up.
                partmgr_ptr->setActivePartition(as_mgr->getNextPartition());

                // Switch our camera state to be ARMed.
                CameraServer::set_camera_state(GIGE_CAMERA_STATE_ARMED);
                TP(FT_ICM, "%s: doOverlap() was true", __PRETTY_FUNCTION__);
            }

            // If the autosave is not in overlapped mode, check to see if we are saving
            // off camera.
            else if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
            {
                // We are in autosave, but not overlap, saving to the PC. Set the review partition.
                TP(FT_ICM, "%s: doOverlap() was false, set review partition to %d", __PRETTY_FUNCTION__, partmgr_ptr->getActivePartition());
                partmgr_ptr->setReviewPartNum(partmgr_ptr->getActivePartition());
            }
        }

        else
        {
            // Not in FasFireLR mode so we can update our playback control information.
            pbc_update_cut_in(NULL, 0);
            pbc_update_cut_out(NULL, partmgr_ptr->getNumFrames() - 1);

            // Drop into review state.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_REVIEW);
        }
    }

    // Switch to the next state.
    TP(FT_ICM, "%s: done, state=%d", __PRETTY_FUNCTION__, CameraServer::get_camera_state());
}

/*!
 * @brief Handle a Continuation event read from the ICM event queue.
 *
 * This is quite different from an end-of-capture event, even though they both signify the completion of
 * a partition. With a continuation event we know that the entire partition has been recorded, and we know
 * that recording continues.
 *
 * @param frame_info        [IN] The value read from the queue ANDED against ICMEF_TRIGGER_FRAME_MASK.
 */
void FiFPGA::icmContinuationEventHappened(uint32_t frame_info)
{
    uint32_t icm_count;
    uint32_t post_trigger_frames;
    PartMgr *partmgr_ptr;           // Pointer to the partition manager class.
    AutoSave    *as_mgr;

    TP(FT_ICM, "%s: FFLR frame_info:%u", __PRETTY_FUNCTION__, frame_info);

    // Get a pointer to our support classes.
    partmgr_ptr = CameraServer::get_partmgr();
    as_mgr = CameraServer::get_autosave();

    // Since this is the continuation event, we captured the full partition.
    icm_count = partmgr_ptr->getFrameCount();

    // Find out how many post trigger frames we had specified.
    // We're only interested in the frame count, so mask out the unused bits.
    post_trigger_frames = icmPostTriggerCount() & FPGA_ICM_POST_TRIG_FRAME_MASK;
    TP(FT_ICM, "%s: FFLR post_trigger_frames=0x%08x", __PRETTY_FUNCTION__, post_trigger_frames);

    // Notify the partition manager of this event so it can clean things up.
    // A continuation event means that the full partition has been consumed.
    partmgr_ptr->partitionEventDone(partmgr_ptr->getCaptureFrameAddr(icm_count - 1), post_trigger_frames, icm_count);

    // Go into autosave and queue up the partition to be written.
    TP(FT_ICM, "%s: FFLR Switching to GIGE_CAMERA_STATE_AUTOSAVE", __PRETTY_FUNCTION__);
    CameraServer::set_camera_state(GIGE_CAMERA_STATE_AUTOSAVE);

    // Are we interested in overlapping the save operation?
    if (as_mgr->doOverlap())
    {
        // Since we are in overlap mode, we can be confident there is a "next" partition.
        // While we were futzing around, the FPGA has already re-armed and started
        // capturing into the "busy" partition. We need to catch up.
        partmgr_ptr->setActivePartition(as_mgr->getNextPartition());

        // Switch our camera state to be ARMed.
        CameraServer::set_camera_state(GIGE_CAMERA_STATE_ARMED);

        // In FasFireLR mode, we need to see if we've been triggered.
        if (ltrFasFireEnabled() && (partmgr_ptr->getTriggerPart() >= 0))
        {
            // Switch our camera state to be triggered.
            CameraServer::set_camera_state(GIGE_CAMERA_STATE_TRIGGERED);
        }
    }

    TP(FT_ICM, "%s: FFLR done, state=%d", __PRETTY_FUNCTION__, CameraServer::get_camera_state());
}

/*!
 * @brief Handle a Descriptor FIFO "almost full" event.
 */
void FiFPGA::icmAbortLrOnOverflow()
{
    bool    in_ltr_mode;
    bool    any_frames;
    uint8_t state;

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // Are we in LR mode?
    in_ltr_mode = ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    if (!in_ltr_mode)
    {
        fi_log(LOG_ERR, "%s: Got OVERFLOW and not in LTR mode?!", __FUNCTION__);
        return;
    }

    // We need to be armed or triggered for this to make sense.
    state = CameraServer::get_camera_state();
    if ((state != GIGE_CAMERA_STATE_ARMED) && (state != GIGE_CAMERA_STATE_TRIGGERED))
    {
        fi_log(LOG_ERR, "%s: Bad state %d for overflow abort", __FUNCTION__, state);
        return;
    }

    //Log the current temperature.
    //system("logger temperature `cat /proc/fi_ssd/temp`");

    // Stop the capture.
    any_frames = icmStopCapture((state == GIGE_CAMERA_STATE_TRIGGERED) ? true : false);

    // We'll store a save error state so the user gets an error message.
    // ENFILE is "File table overflow", close enough.
    ltrSetSaveError(GIGE_SAVE_STATE_ERR | ENFILE);

    // It doesn't make sense to get an overflow if we haven't captured anything.
    // But in case the overflow caused confusing results, allow for it.
    if (any_frames)
        CameraServer::set_camera_state(GIGE_CAMERA_STATE_REVIEW);
    else
        CameraServer::set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
}

/*
 * @brief Reading the ICM event fifo and handle the events
 *
 * @return  -1 if there is an error reading FPGA information such that further event processing
 *      is problematic, 0 otherwise.
 */
int FiFPGA::icmEventHappened()
{
    int     i;
    uint32_t events_in_queue;
    uint32_t value;
    uint32_t last_frame_addr_hw;        // The last frame address, in hexwords.

    TP(FT_ICM, "%s: Begin", __PRETTY_FUNCTION__);

    // Read the FIFO level to make sure something is in the queue
    events_in_queue = icmEventFIFOCount();
    TP(FT_ICM, "%s: events_in_queue=%d, capture_fifo_count=%d", __PRETTY_FUNCTION__, events_in_queue, icmCaptureFifoCount());

    // If there are no events in the queue, then it's probably a race condition and we got the event
    // for this interrupt on the last pass but log it anyway
    if (events_in_queue == 0)
    {
        TP(FT_ICM, "%s: Interrupt happened but no queue entries.", __PRETTY_FUNCTION__);
    }

    // Handle all events in the queue
    i = 0;
    while (i < (int)events_in_queue)
    {
        bool trigger_flag;
        bool continuation_event;
        bool end_of_capture;
        uint32_t frame_info;

        // Read the FPGA event fifo to get the next event and route it where it needs to go.
        if (!regRead(ICM_EVENT_FIFO, &value))
        {
            fi_log(LOG_ERR, "%s: Cannot read ICM_EVENT_FIFO register", __FUNCTION__);
            return -1;
        }

        // Count another event read from the queue.
        ++i;

        // Get our flag...figure out if the event is a trigger, continuation or end-of-capture.
        trigger_flag = ((value & ~ICMEF_TRIGGER_FRAME_MASK) == ICMEF_TRIGGER_FLAG);
        continuation_event = ((value & ~ICMEF_TRIGGER_FRAME_MASK) == ICMEF_CONTINUATION_FLAG);
        end_of_capture = ((value & ~ICMEF_TRIGGER_FRAME_MASK) == ICMEF_END_OF_CAPTURE_FLAG);

        frame_info = value & ICMEF_TRIGGER_FRAME_MASK;
        TP(FT_ICM, "%s: TOL ev[%d]=0x%08x, events_in_queue=%d", __PRETTY_FUNCTION__, i - 1, value, icmEventFIFOCount());

        // If the user requested that we stop capture, we'll just ignore any trigger
        // or completion events that were queued as if they'd never happened to avoid
        // getting confused about the state.
        if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_DEARMING)
        {
            fi_log(LOG_INFO, "%s: FPGA event 0x%08x while dearming", __FUNCTION__, value);
            if (trigger_flag)
            {
                // Flush trigger extra words and bump count.
                regRead(ICM_EVENT_FIFO, &value);
                regRead(ICM_EVENT_FIFO, &value);
                i += 2;
            }
        }

        // Did we get a trigger?
        else if (trigger_flag)
        {
            uint32_t    extra_info[2] = {0, 0}; // Needed if the trigger flag is set.

            // Read the two words of extra trigger event information and bump the count.
            // Word 0   SBS
            // Word 1
            //   b31    IRIG Lock state
            //   b24:29 External marker inputs [5:0]
            //   b0:19  SBUS
            regRead(ICM_EVENT_FIFO, &extra_info[0]);
            regRead(ICM_EVENT_FIFO, &extra_info[1]);
            i += 2;
            TP(FT_ICM, "%s: frame_info=0x%x, ev[%d]=0x%08x, ev[%d]=0x%08x", __PRETTY_FUNCTION__, frame_info, i - 2, extra_info[0], i - 1, extra_info[1]);
            TP(FT_ICM, "%s: trigger_event_happened-events_in_queue=%d", __PRETTY_FUNCTION__, icmEventFIFOCount());

            // Did we get a trigger event when we shouldn't have?
            icmTriggerEventHappened(frame_info, extra_info);
        }

        // Is this a continuation event?
        else if (continuation_event)
        {
            TP(FT_ICM, "%s: continuation_event_happened-events_in_queue=%d, frame_info=0x%08x", __PRETTY_FUNCTION__, icmEventFIFOCount(), frame_info);
            icmContinuationEventHappened(frame_info);
        }

        // Did we finish a capture?
        else if (end_of_capture)
        {
            uint32_t eoc_info;
            bool underflow;

            if ((CameraServer::get_camera_state() == GIGE_CAMERA_STATE_TRIGGERED) || (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_ARMED))
            {
                // Get the address of the last frame written.
                regRead(ICM_EVENT_FIFO, &eoc_info);
                i++;

                // Extract the underflow bit out of the event word.
                underflow = (eoc_info & UNDERFLOW_FLAG_MASK) == UNDERFLOW_FLAG_MASK;
                last_frame_addr_hw = eoc_info & ~UNDERFLOW_FLAG_MASK;

                TP(FT_ICM, "%s: end_capture_event_happened-events_in_queue=%d", __PRETTY_FUNCTION__, icmEventFIFOCount());
                TP(FT_ICM, "%s: GIGE_CAMERA_STATE_TRIGGERED, frame_info=0x%08x, eoc_info=0x%08x, last_frame_addr_hw, ev[%d]=0x%08x, underflow=%d", __PRETTY_FUNCTION__, frame_info, eoc_info, i - 1, last_frame_addr_hw, underflow);
                icmEndCaptureEventHappened(frame_info, ((uint64_t)(last_frame_addr_hw) << HEXWORD_SHIFT), underflow);
            }

            // Did we get an event when we shouldn't have?
            else if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_AUTOSAVE)
            {
                AutoSave    *as_mgr;
                as_mgr = CameraServer::get_autosave();

                // If doOverlap() is false, then we've run out of buffers. We still need to process
                // those we've queued.
                if (!as_mgr->doOverlap())
                {
                    uint32_t last_frame_addr_hw;        // The last frame address, in hexwords.

                    // Get the address of the last frame written.
                    regRead(ICM_EVENT_FIFO, &eoc_info);
                    i++;

                    // Extract the underflow bit out of the event word.
                    underflow = (eoc_info & UNDERFLOW_FLAG_MASK) == UNDERFLOW_FLAG_MASK;
                    last_frame_addr_hw = eoc_info & ~UNDERFLOW_FLAG_MASK;

                    TP(FT_ICM, "%s: GIGE_CAMERA_STATE_AUTOSAVE, frame_info=0x%08x, eoc_info=0x%08x, last_frame_addr_hw, ev[%d]=0x%08x, underflow=%d", __PRETTY_FUNCTION__, frame_info, eoc_info, i - 1, last_frame_addr_hw, underflow);
                    TP(FT_ICM, "%s: end_capture_event_happened-events_in_queue=%d", __PRETTY_FUNCTION__, icmEventFIFOCount());
                    icmEndCaptureEventHappened(frame_info, ((uint64_t)(last_frame_addr_hw) << HEXWORD_SHIFT), underflow);
                }
                else
                {
                    fi_log(LOG_WARNING, "%s: Unexpected event[%d]=0x%x state=%d/GIGE_CAMERA_STATE_AUTOSAVE", __PRETTY_FUNCTION__, i - 1, value, CameraServer::get_camera_state());
                }
            }
        }
        else
        {
            fi_log(LOG_WARNING, "%s: Unexpected event[%d]=0x%x state=%d", __PRETTY_FUNCTION__, i - 1, value, CameraServer::get_camera_state());
        }

        TP(FT_ICM, "%s: BOL events_in_queue=%d", __PRETTY_FUNCTION__, icmEventFIFOCount());
    }

    TP(FT_ICM, "%s: Done", __PRETTY_FUNCTION__);
    return 0;
}

/*!
 * @brief Called to process an FPGA event.
 *
 * @param fpga_val The value associated with the EVENT.
 */
void FiFPGA::eventHappened(uint32_t fpga_val)
{
    int camera_state;
    uint32_t orig_val = fpga_val;

    // Only pay attention to the events we're interested in.  Ignore all others.
    fpga_val &= ~shadow_irq;
    TP(FT_ICM_NOISY, "%s: event=0x%x&0x%x=0x%x(%s).", __PRETTY_FUNCTION__, orig_val, shadow_irq, fpga_val, decode_event(fpga_val));

    // Sanity check.
    if (!isInitialized())
    {
        fi_log(LOG_ERR, "%s: FPGA not initialized!", __PRETTY_FUNCTION__);
        return;
    }

    // First see if there are any ICM events.
    if (fpga_val & PCIEIRQ_FIFO_NOT_EMPTY)
    {
        // Process the ICM event.
        if (icmEventHappened() < 0)
        {
            fi_log(LOG_WARNING, "%s: icm_event_happend() failed!", __PRETTY_FUNCTION__);
            return;
        }
    }

    // Now see if there are any frames ready on P1, since the fpga driver masks out whatever is being
    // serviced we have to handle the case where either the status register is set or the event type
    // is passed to me.
    if (fpga_val & PCIEIRQ_P1_FRAME_DONE)
    {
        ipm_proc_done();
    }

    // Did we get an LR FIFO almost full event?
    if (fpga_val & PCIEIRQ_LR_DESC_FIFO_AF)
    {
        // Is there a worker thread?
        if (ltr_thread.get() != nullptr)
        {
            // Notify the worker thread that we've got an almost full situation here.
            ltr_thread->dataReady();
        }
        else
        {
            fi_log(LOG_ERR, "%s: PCIEIRQ_LR_DESC_FIFO_AF and no ltr_thread?!", __PRETTY_FUNCTION__);
        }
    }

    // Did we get a power button press?
    if (fpga_val & PCIEIRQ_PWR_OFF)
    {
        fi_log(LOG_INFO, "Notify GIGE_REG_POWER");

        // Notify folks that the user pressed the power button.
        fcp_notify_apps(GIGE_REG_POWER, 1);

        // Are we an integrated camera?
        CameraServer *cs_ptr = CameraServer::get_cs();
        if (cs_ptr && cs_ptr->integrated())
        {
            // See if anyone is connected.
            if ((fcp_get_lock_time() == 0) && (fcp_get_fastec_lock_time() == 0) && !grpcStreamIsConnected())
            {
                // No one is listening, so go ahead and shutdown.
                fi_log(LOG_ALERT, "%s: Forcing shutdown.", __PRETTY_FUNCTION__);
                CameraServer::shutdown();
            }
        }
    }

    // Did we get a lens event?
    if (fpga_val & (PCIEIRQ_MFT_LSDET | PCIEIRQ_MFT_LSLPSW))
    {
        mft_irq(fpga_val);
    }

    // Get the camera state.
    camera_state = CameraServer::get_camera_state();

    // Did we get an arm event?
    if (fpga_val & PCIEIRQ_EXT_ARM_INPUT)
    {
        PartMgr *partmgr_ptr;           // Pointer to the partition manager class.
        uint32_t cap_mode = 0;
        bool ok_to_arm = false;
        bool really_unexpected = true;
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        LOG("%s: PCIEIRQ_EXT_ARM_INPUT state=%d", __PRETTY_FUNCTION__, camera_state);

        // Pause further ARM-in processing until we get to a state where we can handle another.
        armInPause();

        // Get our pointer to the partion manager.
        partmgr_ptr = CameraServer::get_partmgr();

        // Are we in an LR mode? If so, figure out which LR mode.
        if (ltrFasFireEnabled() || ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE))
            cap_mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

        // In Basic modes, we only arm if we're in LIVE mode and don't have a capture,
        // unless the ARM_IN_OVERWRITE preference is set.
        if (cap_mode == 0)
        {
            if ((partmgr_ptr->getPartitionState() == GIGE_PART_STATE_EMPTY) && (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO))
            {
                LOG("%s: PCIEIRQ_EXT_ARM_INPUT - GIGE_PART_STATE_EMPTY && GIGE_CAMERA_STATE_LIVE_VIDEO", __PRETTY_FUNCTION__);
                ok_to_arm = true;
            }
            else if (partmgr_ptr->getPartitionState() != GIGE_PART_STATE_EMPTY)
            {
                uint32_t pref;

                pref = regs[GIGE_REG_PREFERENCES].get_value();

                if (((pref & GIGE_PREF_ARM_IN_OVERWRITE) == GIGE_PREF_ARM_IN_OVERWRITE) &&
                    ((camera_state == GIGE_CAMERA_STATE_REVIEW) || (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)))
                {
                    LOG("%s: PCIEIRQ_EXT_ARM_INPUT - GIGE_PREF_ARM_IN_OVERWRITE && GIGE_CAMERA_STATE_LIVE_VIDEO/_REVIEW", __PRETTY_FUNCTION__);
                    ok_to_arm = true;
                }
                else
                {
                    // If we are preventing arm because we have a dirty partition, go ahead
                    // and reenable ARM events. Otherwise we'll have to figure it out based
                    // on when the partition is emptied, and there can be multiple ways to do that.
                    // Skip the "unexpected"  message since we will be more explicit.
                    really_unexpected = false;
                    armInResume();
                    fi_log(LOG_DEBUG, "%s: Dirty partition [%d], ignoring PCIEIRQ_EXT_ARM_INPUT (state=%d)", __PRETTY_FUNCTION__, partmgr_ptr->getActivePartition(nullptr), camera_state);
                }
            }
        }
        // If we are appending, we can arm in Live or Review regardless of partition state.
        else
        {
            if ((camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) || (camera_state == GIGE_CAMERA_STATE_REVIEW))
            {
                LOG("%s: PCIEIRQ_EXT_ARM_INPUT - ROC/BROC, GIGE_CAMERA_STATE_LIVE_VIDEO/_REVIEW", __PRETTY_FUNCTION__);
                ok_to_arm = true;
            }
        }

        if (ok_to_arm)
        {
            // Try to transition the camera into the "armed" state.
            // If that failed, reset the detector for the next attempt.
            if (!CameraServer::set_camera_state(GIGE_CAMERA_STATE_ARMED))
            {
                armInResume();
            }
        }
        else if (really_unexpected)
        {
            // We shouldn't be getting this event if we can't handle it.
            fi_log(LOG_DEBUG, "%s: Unexpected PCIEIRQ_EXT_ARM_INPUT (state=%d mode=0x%x)",
                   __PRETTY_FUNCTION__, (int)CameraServer::get_camera_state(), cap_mode);
        }
    }

    // Did we get an un-arm event?
    if (fpga_val & PCIEIRQ_EXT_UNARM_INPUT)
    {
        // Pause further ARM-in processing until we get to a state where we can handle another.
        armInPause();

        // We only respond to UNARM if we're in either ARM or TRIGGERED mode.
        if (camera_state & (GIGE_CAMERA_STATE_ARMED | GIGE_CAMERA_STATE_TRIGGERED))
        {
            uint32_t cap_mode;
            GIGE_REG_CACHE& regs = CameraServer::get_regs();

            // Figure out which, if any, LR mode we're in.
            cap_mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

            if ((ltrFasFireEnabled() || ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)) && (cap_mode != 0))
                // Go through DEARMING to shutdown a ROC/BROC capture.
                CameraServer::set_camera_state(GIGE_CAMERA_STATE_DEARMING);
            else
                // Transition the camera to the "live" state.
                CameraServer::set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
        else
        {
            // We shouldn't be getting this event if we can't handle it.
            // We already paused the ARM-in signal processing above.
            fi_log(LOG_DEBUG, "%s: Unexpected PCIEIRQ_EXT_UNARM_INPUT (state=%d)",
                   __PRETTY_FUNCTION__, (int)CameraServer::get_camera_state());
        }
    }

    // Put the state of interrupts back to what they were, because the driver masks
    // them all while in the ISR.
    regWrite(PCIE_IRQ_MASK, shadow_irq);
}

/*!
 * @brief Clear out any events in the FIFO.
 *
 * Useful before the camera is armed, hardware triggers might be queued up.
 */
void FiFPGA::icmEmptyEventFIFO()
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    sensor_ptr->reset_event_fifo();
}

/*!
 * @brief Utility method to retrieve the number of entries in the event FIFO.
 * @return How many entries are in the event FIFO queue.
 */
int FiFPGA::icmEventFIFOCount()
{
    uint32_t value;
    int events_in_queue;

    // Read the FIFO level to make sure something is in the queue
    if (!regRead(ICM_EVENT_FIFO_LEVEL, &value))
    {
        fi_log(LOG_ERR, "%s: Cannot read ICM_EVENT_FIFO_LEVEL register", __PRETTY_FUNCTION__);
    }
    events_in_queue = (int)(value & FPGA_ICM_EVENT_FIFO_LVL_MASK);
    TP(FT_ICM_NOISY, "%s: events_in_queue=%d, value=0x%08x", __PRETTY_FUNCTION__, events_in_queue, value);
    return events_in_queue;
}

/*!
 * @brief Software trigger the FPGA on the current frame.
 *
 * @return true if success, false otherwise
 */
bool FiFPGA::icmSwTriggerFire()
{
    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // Tell the FPGA to software trigger
    return regModify(ICM_CONTROL, ICMC_TRIGGER, ~ICMC_TRIGGER);
}

/*!
 * @brief Enable the hardware external trigger signal.
 */
bool FiFPGA::icmHwTriggerEnable()
{
    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // Tell the FPGA to enable the trigger button.
    return regModify(EXT_IO_CONTROL, EXTIO_TRIGBTN_IN_ENABLE, ~EXTIO_TRIGBTN_IN_ENABLE);
}

/*!
 * @brief Tell the FPGA to start or stop capturing data from the sensor
 *
 * @return true for valid, otherwise false
 */
bool FiFPGA::icmSensorReadoutEnable(bool enable)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    TP(FT_ICM, "%s: enable=%d", __PRETTY_FUNCTION__, enable);
    return sensor_ptr->set_sensor_readout(enable);
}

/*!
 * @brief Stop the FPGA from recording.
 *
 * @return true for valid, otherwise false
 */
bool FiFPGA::icmDearm()
{
    bool    rc;
    uint32_t    ltr_mode_enabled;
    uint32_t    value = 0;

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // See if LTR mode is enabled.
    ltr_mode_enabled = ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);

    // Create a block so that we UNARM and make sure that the ICM stops
    // before we or anyone else can move on.
    {
        std::unique_lock<std::mutex> locker(class_mutex);
        int     i;

        // Set the UNARM bit.
        TP(FT_ICM, "%s: Set ICMC_UNARM.", __PRETTY_FUNCTION__);
        regModify(ICM_CONTROL, ICMC_UNARM, -1);

        // Wait to give the ICM a chance to release.
        for (i = 0; i < MAX_RETRY_COUNT; ++i)
        {
            struct timespec sleep_req;      // Delay period.

            // Wait for a bit.
            sleep_req.tv_sec = ICM_BUSY_WAIT_SEC;
            sleep_req.tv_nsec = ICM_BUSY_WAIT_NSEC;
            nanosleep(&sleep_req, NULL);

            // Read the register.
            regRead(ICM_STATUS, &value);

            // Did our bit go low?
            if ((value & ICMS_ARMED) == 0)
                break;
        }

        // Did we exceed our retry count?
        if (i == MAX_RETRY_COUNT)
        {
            fi_log(LOG_ERR, "%s: ICMS_ARMED never dropped [0x%x]", __FUNCTION__, value);
            rc = false;
        }
        else
        {
            // All is well.
            rc = true;
        }
    }

    // Display capture information.
    {
        uint32_t status;

        regRead(ICM_WR_DATA_FIFO_STATUS, &status);
        TP(FT_ICM, "%s: >>>> ICM_WR_DATA_FIFO_STATUS=0x%08x <<<<", __PRETTY_FUNCTION__, status);
    }

    // Clear the SW metadata marker bit in case we never captured any frames.
    // This is not really necessary right now, while we just use it to mark the first
    // captured frame after arming. But if we decide to use it later for something other
    // than ROC/BROC modes, we will want it cleared out.
    regWrite(ICM_SW_METADATA_MARKER, 0x00000000);

    // If we're in LTR mode, back out gracefully.
    if (ltr_mode_enabled)
    {
        // Shutdown the LTR processing.
        regModify(ICM_CONTROL, 0, ~ICMC_LR_ENABLE);
        ltrEnable(false);
    }

    // Reset the capture FIFO.
    icmResetCaptureFIFO();

    // Return our status.
    return rc;
}

/*!
 * @brief Check to see if the ICM is armed.
 * @return True if the ICM is armed, false otherwise.
 */
bool FiFPGA::icmIsArmed()
{
    uint32_t reg_value;

    // Read the register.
    regRead(ICM_STATUS, &reg_value);
    TP(FT_ICM, "%s: ICM_STATUS=0x%08x", __PRETTY_FUNCTION__, reg_value);

    // True if the ARM bit is set.
    return ((reg_value & ICMS_ARMED) == ICMS_ARMED);
}

void FiFPGA::icmResetCaptureFIFO()
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    sensor_ptr->reset_capture_fifo();
}

int FiFPGA::icmCaptureFifoCount()
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    return sensor_ptr->get_capture_fifo_count();
}

/*!
 * @brief Tell the FPGA to start capture, either one image or continuously.
 *
 * @param continuous_mode, are we capturing one image or alot.
 * @return true for valid, otherwise false
 */
bool FiFPGA::icmArm(bool continuous_mode)
{
    bool rc(false);

    // Get our support classes.
    AutoSave     *as_mgr = CameraServer::get_autosave();
    CameraServer *cs_ptr = CameraServer::get_cs();
    PartMgr *partmgr_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    uint32_t    value = 0;

    TP(FT_ICM, "%s: cont:%d", __PRETTY_FUNCTION__, continuous_mode);
    TP(FT_ICM, "%s: enabled=%d, dirtySessions=%d, availSessions=%d, busySessions=%d, eventQueue=%d, capture_fifo_count=%d", __PRETTY_FUNCTION__, as_mgr->isEnabled(), as_mgr->dirtySessions(), as_mgr->availSessions(), as_mgr->busySessions(), icmEventFIFOCount(), icmCaptureFifoCount());

    // Clear any saved LTR error.
    ltrSetSaveError(0);

    // Clear the overflow status.
    regWrite(ICM_WR_DATA_FIFO_STATUS, ICMWDFS_RESET);

    // Tell the partition manager to flush the metadata and do final computations
    // based on the current data.
    if (!partmgr_ptr->arm())
    {
        fi_log(LOG_DEBUG, "%s: partmgr_ptr->arm() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Are we already armed?
    if (!icmIsArmed())
    {
        bool do_arm = false;

        // Reset the last frame written value.
        regWrite(ICM_LAST_FRAME_WRITTEN_PTR, LAST_FRAME_INIT);

        // If we are in autosave mode and there are no dirty sessions, then this is likely
        // the first time through. If we are NOT in autosave mode, then this is DEFINITELY
        // the first time through.

        // We are not armed. Are we in autosave mode?
        if (as_mgr->isEnabled())
        {
            // If we are in FasFireLR mode, we need to have the autosave manager re-intialize prior to an ARM.
            if ((regs[GIGE_REG_LTR_MODE].get_value() & GIGE_LTR_MODE_FFLR) == GIGE_LTR_MODE_FFLR)
            {
                // Initialize the autosave object.
                TP(FT_ICM, "%s: FFLR Resetting autosave...", __PRETTY_FUNCTION__);
                as_mgr->reset();
                partmgr_ptr->resetFflr();
            }

            // Are there any available sessions to be queued AND are the entries in the capture FIFO?
            if (as_mgr->armOK())
            {
                // If we are NOT armed, but there are sessions which could be written to, go ahead and arm.
                TP(FT_ICM, "%s: Autosave with available sessions. Arm.", __PRETTY_FUNCTION__);
                do_arm = true;
            }
            else
            {
                TP(FT_ICM, "%s: Autosave, no available sessions. Do not arm.", __PRETTY_FUNCTION__);
                do_arm = false;
            }
        }
        else
        {
            TP(FT_ICM, "%s: Not in autosave, resetting and arming.", __PRETTY_FUNCTION__);

            // Before enabling the interrupts, make sure that the queue is empty
            icmEmptyEventFIFO();

            // We are not in autosave mode. Go ahead and arm.
            do_arm = true;
        }

        if (do_arm)
        {
            TP(FT_ICM, "%s: Not ARMed, Setup ARM <<<<<<<<<<<<<<<<<<", __PRETTY_FUNCTION__);

            // Enable event interrupt.
            irqEnable(PCIEIRQ_FIFO_NOT_EMPTY);

            // If we are in low light mode, switch to normal mode before arming the sensor.
            if (icmIsLowLightEnabled())
            {
                float frame_rate;

                cs_ptr->get_frame_rate(&frame_rate);
                icmSetFrameRate(frame_rate);

                TP(FT_ICM, "%s: low light shutter enabled switching to regular frame rate to:%f", __PRETTY_FUNCTION__, frame_rate);
                icmSetLowLightShutterSpeed(false, 0);
            }

            // Read the current ICM control setting.
            value = 0;
            if (!regRead(ICM_CONTROL, &value))
            {
                fi_log(LOG_WARNING, "%s: Unable to read the ICM_CONTROL register.", __FUNCTION__);
                goto bail;
            }

            // Make sure any latched trigger has been cleared before we arm.
            // This rarely matters so don't holler if it fails, we'll get another err later.
            fi_log(LOG_DEBUG, "%s[%d]: ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, __LINE__, value);
            value |= ICMC_LATCHED_TRIGGER_CLR;
            (void)regWrite(ICM_CONTROL, value);
            value &= ~ICMC_LATCHED_TRIGGER_CLR;

            // Clear the capture mode bits before we set them for our current mode.
            value &= ~ICMC_CAPTURE_MODE_MASK;

            // Setup the bits for single vs. continuous capture mode.
            if (continuous_mode)
                value &= ~(ICMC_CAPTURE_SINGLE);

            // Arm the ICM.
            value |= ICMC_ARM;

            // Are we in LTR mode?
            if ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
            {
                uint32_t mode;

                // Figure out our mode.
                mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

                // Use the standard enable to turn on LTR mode.
                if (mode == 0)
                {
                    ltrEnable(true);

                    value |= ICMC_CAPTURE_MODE_BASIC;
                }
                else
                {
                    ltrEnableROC(mode);

                    if ((mode & GIGE_LTR_MODE_BROC) != 0)
                        value |= ICMC_CAPTURE_MODE_BROC;
                    else
                        value |= ICMC_CAPTURE_MODE_ROC;
                }

                // Make sure the worker thread is running.
                if ((ltr_thread.get() != nullptr) && !ltr_thread->is_running())
                {
                    fi_log(LOG_DEBUG, "%s: Unable to start the LTR worker thread.", __PRETTY_FUNCTION__);
                    goto bail;
                }

                // Make sure the ICM is in LTR mode.
                value |= ICMC_LR_ENABLE;
            }
            else if ((regs[GIGE_REG_LTR_MODE].get_value() & GIGE_LTR_MODE_FFLR) == GIGE_LTR_MODE_FFLR)
            {
                // Figure out our mode - are we in ROC, BROC, or basic.
                if ((regs[GIGE_REG_LTR_MODE].get_value() & GIGE_LTR_MODE_ROC) == GIGE_LTR_MODE_ROC)
                {
                    // Sanity check.
                    if (imgSaveWorker::getInstance()->getRecordingFrameCount() == 0)
                    {
                        fi_log(LOG_ERR, "%s[%d]: The recording frame count is 0!", __PRETTY_FUNCTION__, __LINE__);
                        imgSaveWorker::getInstance()->setRecordingFrameCount(partmgr_ptr->getPartitionNumFrames(partmgr_ptr->getActivePartition()));
                    }

                    // Specify the maximum frame count.
                    regWrite(ICM_ROC_MAX_CAPTURE_FRM_CNT, imgSaveWorker::getInstance()->getRecordingFrameCount());

                    // Enable FasFireLR ROC mode.
                    value |= ICMC_CAPTURE_MODE_FFLR_ROC;
                    TP(FT_ICM, "%s: FFLR ROC mode enabled.", __PRETTY_FUNCTION__);
                }
                else if ((regs[GIGE_REG_LTR_MODE].get_value() & GIGE_LTR_MODE_BROC) == GIGE_LTR_MODE_BROC)
                {
                    uint32_t burst_cnt;

                    burst_cnt = regs[ GIGE_REG_BROC_LEN ].get_value();
                    if (burst_cnt < GIGE_MIN_BROC_LEN)
                    {
                        fi_log(LOG_WARNING, "%s: Found BROC_LEN=%d, changing to %d", __FUNCTION__, burst_cnt, GIGE_MIN_BROC_LEN);
                        burst_cnt = GIGE_MIN_BROC_LEN;
                        regs[ GIGE_REG_BROC_LEN ].set_value(burst_cnt);
                    }
                    regWrite(ICM_BROC_BURST_FRM_CNT, burst_cnt);

                    // Sanity check.
                    if (imgSaveWorker::getInstance()->getRecordingFrameCount() == 0)
                    {
                        fi_log(LOG_ERR, "%s[%d]: The recording frame count is 0!", __PRETTY_FUNCTION__, __LINE__);
                        imgSaveWorker::getInstance()->setRecordingFrameCount(partmgr_ptr->getPartitionNumFrames(partmgr_ptr->getActivePartition()));
                    }

                    // Specify the maximum frame count.
                    regWrite(ICM_ROC_MAX_CAPTURE_FRM_CNT, imgSaveWorker::getInstance()->getRecordingFrameCount());

                    // Enable FasFireLR BROC mode.
                    value |= ICMC_CAPTURE_MODE_FFLR_BROC;
                    TP(FT_ICM, "%s: FFLR BROC mode enabled, cnt=%d, max_len=%d.", __PRETTY_FUNCTION__, burst_cnt, imgSaveWorker::getInstance()->getRecordingFrameCount());
                }
                else
                {
                    // Enable FasFireLR basic mode.
                    value |= ICMC_CAPTURE_MODE_FFLR;
                    TP(FT_ICM, "%s: FFLR mode enabled.", __PRETTY_FUNCTION__);
                }
            }
            else
            {
                // Enable standard image memory capture mode.
                value |= ICMC_CAPTURE_MODE_BASIC;
            }

            // Write the new control register settings out.
            TP(FT_ICM, "%s[%d]: ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, __LINE__, value);
            if (!regWrite(ICM_CONTROL, value))
            {
                fi_log(LOG_WARNING, "%s: Unable to write the ICM_CONTROL register\n", __FUNCTION__);
                goto bail;
            }
        }
    }

    // Enable the trigger button.
    icmHwTriggerEnable();
    rc = true;

bail:
    // Do we need to perform error cleanup?
    if (!rc)
    {
        // Do we have a pointer to the partition manager?
        if (partmgr_ptr != nullptr)
            partmgr_ptr->setPartitionState(partmgr_ptr->getActivePartition(), GIGE_PART_STATE_EMPTY);
    }

    return rc;
}

/*!
 * @brief Returns true if the ICM pipeline is IDLE, false otherwise.
 */
bool FiFPGA::icmIsPipelineIdle(uint32_t *p_rtn)
{
    uint32_t value;

    // Read the ICM status register.
    if (!regRead(ICM_STATUS, &value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read ICM_STATUS.", __FUNCTION__);

        // Assign value something to return a false value.
        value = 0;
    }

    // Is there somplace to put the value?
    if (p_rtn != 0)
        *p_rtn = value;

    TP(FT_ICM_NOISY, "%s: ICM_STATUS=0x%08x", __PRETTY_FUNCTION__, value);
    return ((value & (ICMS_PIPELINE_READY | ICMS_PIPELINE_IDLE)) == ICMS_PIPELINE_IDLE);
}

/*!
 * @brief Check whether the ICM reached the ICM_PIPELINE_READY state.
 *
 * There are state transitions where we expect to see this.
 * For debug, we'll warn if we don't.
 *
 * Returns true if we see ICM pipeline go READY, false otherwise.
 */
bool FiFPGA::icmIsPipelineReady(uint32_t *p_rtn)
{
    uint32_t    value;
    int     retry;

    retry = 0;
    do
    {
        if (regRead(ICM_STATUS, &value))
        {
            // If the pipeline is IDLE, it doesn't make sense to wait for READY.
            if ((value & ICMS_PIPELINE_IDLE) == ICMS_PIPELINE_IDLE)
                break;

            if ((value & ICMS_PIPELINE_READY) == 0)
                usleep(PIPE_DISABLE_READY_WAIT);
        }
    } while (((value & ICMS_PIPELINE_READY) == 0) && (++retry < PIPE_DISABLE_READY_RETRIES));

    // Return the last status we read so it can be logged.
    if (p_rtn != nullptr)
        *p_rtn = value;

    TP(FT_ICM, "%s: %sready", __PRETTY_FUNCTION__, ((value & ICMS_PIPELINE_READY) == ICMS_PIPELINE_READY) ? "" : "NOT ");
    return ((value & ICMS_PIPELINE_READY) == ICMS_PIPELINE_READY);
}

/*!
 * @brief Disable the ICM pipeline.
 * @return True if successful, false otherwise.
 */
bool FiFPGA::icmPipelineDisable()
{
    bool    rc;
    int     retry;
    uint32_t    value;

    // -----------------------------------------------------------------------
    // Do not allow any other FPGA related operation to interrupt the following sequence.
    // -----------------------------------------------------------------------
    std::unique_lock<std::mutex> locker(class_mutex);

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // We expect to see ICMS_PIPELINE_READY here before we disable.
    // If we were in LIVE_VIDEO mode, we should already have stopped the display
    // controller from requesting frames.
    // If we had been armed and are now transitioning to REVIEW, the unarming
    // process should have stopped sending integration requests automatically.
    // In either case, the ICM pipeline should transition to its READY state.
    // HOWEVER, if we have external sync=PPF and we never got a first pulse, then
    // the ICM state machine might be waiting for the start of a frame; it won't
    // reach the ready state but we can still disable the pipeline.
    if (!icmIsPipelineReady(&value))
    {
        // Didn't see ICMS_PIPELINE_READY, but maybe we have ICMS_PIPELINE_IDLE.
        // With neither, try some recovery tactics because later on when we try to
        // disable the pipeline, it might not work.
        if ((value & ICMS_PIPELINE_IDLE) != ICMS_PIPELINE_IDLE)
        {
            // Make sure we are not sending integration pulses to the sensor then check again.
            icmSensorReadoutEnable(false);
            usleep(PIPELINE_WAIT);

            regRead(ICM_STATUS, &value);
            TP(FT_ICM, "%s: status=0x%08x", __PRETTY_FUNCTION__, value);
            if ((value & (ICMS_PIPELINE_IDLE | ICMS_PIPELINE_READY)) == 0)
            {
                uint32_t icmc = 0xffffffff;

                regRead(ICM_CONTROL, &icmc);
                fi_log(LOG_WARNING, "%s: No ICMS_PIPELINE_READY! ICM_STATUS=0x%08x ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, value, icmc);
            }
        }
    }

    // Read the ICM control register.
    if (!regRead(ICM_CONTROL, &value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read ICM_CONTROL!", __FUNCTION__);
        return false;
    }

    // Clear the PIPELINE enable bit. This will transition us to the FPGA's
    // IDLE state - effectively shutting down the IPM pipeline.
    fi_log(LOG_DEBUG, "%s[%d]: ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, __LINE__, value);
    value &= ~ICMC_PIPELINE_ENABLE;
    if (!regWrite(ICM_CONTROL, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write ICM_CONTROL!", __FUNCTION__);
        return false;
    }

    // Wait for the state machine to transition to IDLE.
    TP(FT_ICM, "%s: Wait for ICM pipeline to idle.", __PRETTY_FUNCTION__);
    for (retry = 0; retry < MAX_RETRY_COUNT; ++retry)
    {
        // Wait a little bit.
        usleep(PIPELINE_WAIT);

        // Make sure the ICM pipeline is IDLE.
        if (icmIsPipelineIdle(&value))
            break;
    }

    // Check to see if we timed out.
    if (retry == MAX_RETRY_COUNT)
    {
        uint32_t icmc = 0xffffffff;

        regRead(ICM_CONTROL, &icmc);
        fi_log(LOG_WARNING, "%s: Pipeline never went idle, ICM_STATUS=0x%x, ICM_CONTROL=0x%x, resetting", __PRETTY_FUNCTION__, value, icmc);

        regWrite(ICM_MODULE_RESET, 1);
        usleep(PIPELINE_WAIT);
        regWrite(ICM_MODULE_RESET, 0);

        // Wait a little bit.
        usleep(PIPELINE_WAIT);

        // See if we recovered.
        if (icmIsPipelineIdle(&value))
            rc = true;
        else
        {
            fi_log(LOG_WARNING, "%s: Pipeline STILL never went idle, ICMS=0x%x", __FUNCTION__,
                   value);
            rc = false;
        }
    }
    else
    {
        // All done.
        rc = true;
        TP(FT_ICM, "%s: Done.", __PRETTY_FUNCTION__);
    }

    return rc;
}

/*!
 * @brief Set the low light shutter speed. This value is used until the
 *    camera is armed, at which time things revert to the normal shutter
 *    speed and frame rate.
 *
 * @param enable_low_light  Should low light settings be enabled.
 * @param shutter_speed     The shutter speed in uSec to use for low light, 0 if none.
 *
 * @return true for valid, otherwise false
 */
bool FiFPGA::icmSetLowLightShutterSpeed(bool enable, uint32_t shutter_speed)
{
    CameraServer *cs_ptr = CameraServer::get_cs();

    TP(FT_ICM, "%s: enable:%u shutter:%u", __PRETTY_FUNCTION__,
       enable ? 1 : 0, shutter_speed);

    // Make sure the shutter speed is valid. It needs to
    // be in range for the requirements of the current sensor.
    if ((shutter_speed == 0) ||
        ((shutter_speed >= cs_ptr->get_min_shutter_speed()) &&
         (shutter_speed <= cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE))))
    {
        low_light_enabled = enable;
        low_light_shutter = shutter_speed;
    }
    else
    {
        fi_log(LOG_WARNING, "%s: Invalid low light shutter speed %d (max=%d).",
               __FUNCTION__, shutter_speed, cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE));
        return false;
    }

    if (enable)
    {
        icmSetFrameRate((float)LOW_LIGHT_FRAME_RATE);
        icmSetShutterSpeed(shutter_speed);
    }
    else
    {
        float frame_rate;
        uint32_t shutter_speed;

        cs_ptr->get_frame_rate(&frame_rate);
        cs_ptr->get_shutter_speed(&shutter_speed);
        icmSetFrameRate(frame_rate);
        icmSetShutterSpeed(shutter_speed);
    }

    return true;
}

/*!
 * @brief Internal function which actually sets the shutter speed. This function
 *    can be called with either the regular or the low light shutter speed. All
 *    validation should have been done in higher level routines.
 *
 * @param shutter_speed     In microseconds (uS)
 * @return true for valid, otherwise false
 */
bool FiFPGA::icmSetShutterSpeed(uint32_t shutter_speed, uint32_t /* clks_per_usec */)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    // Allow for an override for test purposes.
    if (!icmIsLowLightEnabled())
        override_u32(OVR_INTEGRATION, &shutter_speed);

    // Pass it on to the sensor object.
    sensor_ptr->set_integration_time(shutter_speed);

    // Only save the shutter if we're not in low-light mode.
    if (!icmIsLowLightEnabled())
        fpga_shutter = shutter_speed;

    return true;
}

/*!
 * @brief Setup the FPGA registers for a specific frame rate.
 *
 * @param frame_rate    The frame rate desired, should have already been validated.
 * @param clock_speed   The measured number of FPGA_CLOCK_SPEED ticks.
 *
 * @return true if everything is good, otherwise false.
 */
bool FiFPGA::icmSetFrameRate(float frame_rate, uint32_t clock_speed)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    uint32_t reg_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FT_ICM, "%s: frame_rate=%f, clock_speed=%u", __PRETTY_FUNCTION__, frame_rate, clock_speed);

    // Check to see if we're turning off low-light mode.
    if (!approx_equals(frame_rate, (float)LOW_LIGHT_FRAME_RATE))
        low_light_enabled = false;

    // Sanity check the frame rate value.
    if (approx_equals(frame_rate, 0))
    {
        fi_log(LOG_ERR, "%s: Invalid frame_rate (0).", __PRETTY_FUNCTION__);
        return false;
    }

    // This may pull the frame rate from an overrides file.
    if (!low_light_enabled)
    {
        if (override_u32(OVR_FRAMERATE, &reg_value))
            frame_rate = CameraServer::get_cs()->reg_to_frame_rate(reg_value);
    }

    // Set the framerate.
    if (!sensor_ptr->set_frame_rate(frame_rate, (double)(((double)clock_speed) / FPGA_CLOCK_SPEED)))
    {
        fi_log(LOG_ERR, "%s: Unable to set the framerate!", __PRETTY_FUNCTION__);
        return false;
    }

    // Update the sync out pulse width setting, if it enabled.
    reg_value = regs[GIGE_REG_SYNC_OUT_SELECT].get_value();
    if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
        syncOutConfigure(reg_value);

    return true;
}

/*!
 * @brief This routine is used to bring the ICM from the IDLE state to the CAPTURE_READY
 *    state (FPGA state names used). It is very important that the sensor is actively
 *    producing data before this state transition can occur.
 *
 * @return true if everything goes OK, false otherwise.
 */
bool FiFPGA::icmStartup(bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed)
{
    uint32_t    value;
    Sensor *sensor_ptr = CameraServer::get_sensor();

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // Make sure the ICM pipeline is IDLE.
    if (!icmIsPipelineIdle(&value))
    {
        uint32_t ctrl;

        regRead(ICM_CONTROL, &ctrl);
        fi_log(LOG_WARNING, "%s: ICM Pipeline is NOT idle ICM_STATUS=0x%08x ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, value, ctrl);

        if ((value & ICMS_PIPELINE_READY) == 0)
            return false;

        // Okay, ICM Pipeline is READY but not IDLE. This probably means we
        // restarted without resetting the FPGA so the ICM didn't stop cleanly.
    }

    // The way we set the bit mode depends on the sensor.
    sensor_ptr->set_fpga_bit_mode(bit_mode);

    // Set the fpga frame rate
    icmSetFrameRate(frame_rate);

    // Set the default (shutter speed)
    if (!icmSetShutterSpeed(shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: Unable to set_shutter_speed=%u.", __FUNCTION__, shutter_speed);
        return false;
    }

    // Setup the DMA burst size for the ICM.
    if (!regWrite(ICM_DMA_BURST_SIZE, DMA_BURST_SIZE))
    {
        fi_log(LOG_WARNING, "%s: Unable to write ICM_DMA_BURST_SIZE.", __FUNCTION__);
        return false;
    }

    // Turn on the ICM Pipeline.
    if (!regRead(ICM_CONTROL, &value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read ICM_CONTROL.", __FUNCTION__);
        return false;
    }
    fi_log(LOG_DEBUG, "%s[%d]: ICM_CONTROL=0x%08x", __PRETTY_FUNCTION__, __LINE__, value);

    value |= ICMC_PIPELINE_ENABLE;
    if (!regWrite(ICM_CONTROL, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write ICM_CONTROL.", __FUNCTION__);
        return false;
    }

    return true;
}

/*!
 * @brief Set the unarmed live view buffer start address. Make sure the sensor
 *    is dearmed before calling this method. The buffer and size are in QWORDD.
 *
 * @param buf_start_addr    Starting address for the ulv (2) buffers.
 * @param buf_length        Size of each buffer.
 * @param num_bufs          The number of buffers.
 *
 * @return true if everything is good, otherwise false.
 */
bool FiFPGA::icmSetULVBuffers(uint64_t buf_start_addr, uint32_t buf_length, int num_bufs)
{
    bool    rc;
    int     retries;
    uint32_t    reg_value;

    TP(FT_ICM, "%s: addr=0x%" PRIx64 ", length=0x%x, num_bufs=%d", __PRETTY_FUNCTION__,
       buf_start_addr, buf_length, num_bufs);

    // Make sure the AUX DMA is available.
    for (retries = 0; retries < MAX_RETRY_COUNT; retries++)
    {
        regRead(IM_AUX_WR_DMA0_STATUS, &reg_value);
        if ((reg_value & AUX_DMA_BUSY) == 0)
            break;

        usleep(AUX_DMA_BUSY_WAIT);
    }

    // Did everything go OK?
    if (retries < MAX_RETRY_COUNT)
    {
        FI_REG_ACCESS access;
        ioFpgaSetMultiRegisterMessage msg;

        // Setup the AUX DMA engine to blast the buffer with 0's.
        access.reg_id = IM_AUX_WR_DMA0_CONTROL;
        access.value = AUX_DMA_ENABLE;
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_START_ADDR;
        access.value = convert_to_HEXWORD(buf_start_addr);
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_BLOCK_SIZE;
        access.value = convert_to_HEXWORD(num_bufs * buf_length);
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_BURST_SIZE;
        access.value = AUX_WR_DMA_DDR2_REQ_SIZE_VAL;
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_BURST_INTERVAL;
        access.value = 0;
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_BP_THRESHOLDS;
        access.value = AUX_WR_DMA_BP_THRESHOLD_VAL;
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_PATTERN_VAL;
        access.value = 0x00;
        msg.access.push_back(access);

        access.reg_id = IM_AUX_WR_DMA0_CONTROL;
        access.value = AUX_DMA_ENABLE | AUX_DMA_START;
        msg.access.push_back(access);

        // Send the request.
        if (fpgaIOWorker::isRunning())
            msg.doSync(fpgaIOWorker::factory());
    }
    else
    {
        // Issue a warning - either way, we're still going to setup the buffer address.
        fi_log(LOG_WARNING, "%s: AUX_DMA is stuck busy?! [0x%x]", __FUNCTION__, reg_value);
    }

    // Configure the ICM ULV buffer location in DDR memory.
    rc = regWrite(ICM_ULV_BUF_START_ADDR, convert_to_HEXWORD(buf_start_addr));
    return rc;
}

// Set the number of frames to capture after the trigger.
bool FiFPGA::icmSetPostTriggerFrames(uint32_t num_frames)
{
    fi_log(LOG_DEBUG, "%s: num_frames=%u", __PRETTY_FUNCTION__, num_frames);

    // Set the number of post trigger frames to capture
    return regWrite(ICM_POST_TRIG_CNT, num_frames);
}

// Set the metadata tag aka recording ID.
bool FiFPGA::icmSetMetadataTag(uint32_t tag)
{
    fi_log(LOG_DEBUG, "%s: tag=0x%08x", __PRETTY_FUNCTION__, tag);

    // Set tag that will go into metadata for postmortem analysis.
    return regWrite(ICM_SW_METADATA_TAG, tag);
}

/*!
 * @brief Set the buffer start address, size, and count. Make sure the sensor
 *    is dearmed before calling this method.
 *
 * @param buf_start_addr    Starting address for the capture buffers.
 * @param buf_size      Size of a single capture buffer.
 * @param buf_cnt       How many capture buffers are there.
 *
 * @return true if everything is good, otherwise false.
 */
bool FiFPGA::icmSetCaptureBuffer(uint64_t buf_start_addr)
{
    TP(FT_ICM, "%s: start=0x%" PRIx64, __PRETTY_FUNCTION__, buf_start_addr);

    // Configure the ICM capture buffer size and location in DDR memory.
    // We don't need to log errors here because write_reg() does it.
    return regWrite(ICM_CAPTURE_START_ADDR, convert_to_HEXWORD(buf_start_addr));
}

/*!
 * @brief Get the buffer start address, size, and count so we can restore them later.
 *
 * When we set the capture information for a calibration capture, we will want to restore it
 * to what it used to be.
 *
 * FIXME This is because these registers are being set deep inside partition manager code whenever
 * it changes partition count/size, ROI, or bit depth. Eventually that will get pulled out of there
 * as part of the effort to reduce overall statefulness.
 *
 * @param &buf_start_addr   Starting address for the capture buffers in QWORDS aligned 32 byte.
 * @param &buf_size     Size of a single capture buffer in QWORDS, aligned 32 byte.
 * @param &buf_cnt      How many capture buffers are there.
 *
 * @return true if everything is good, otherwise false.
 */
bool FiFPGA::icmGetCaptureBuffer(uint64_t *buf_start_addr, uint32_t *buf_size, uint32_t *buf_cnt)
{
    bool rc = false;
    uint32_t addr_value;

    std::unique_lock<std::mutex> locker(class_mutex);

    // Get the ICM capture buffer size and location in DDR memory.
    // We don't need to log errors here because write_reg() does it.
    if (regRead(ICM_CAPTURE_START_ADDR, &addr_value) &&
        regRead(ICM_CAPTURE_FRAME_BUF_SIZE, buf_size) &&
        regRead(ICM_CAPTURE_FRAME_BUF_CNT, buf_cnt))
    {
        rc = true;

        // Convert from hexwords to bytes as appropriate.
        *buf_start_addr = ((uint64_t)addr_value) << HEXWORD_SHIFT;
        *buf_size = *buf_size << HEXWORD_SHIFT;
    }
    return rc;
}

/*!
 * @brief Return the post-trigger count register value.
 */
uint32_t FiFPGA::icmPostTriggerCount()
{
    uint32_t    reg_value;

    regRead(ICM_POST_TRIG_CNT, &reg_value);
    return reg_value;
}

/*!
 * @brief Get the current position that the FPGA is recording to.
 *
 * @return The frame position.
 */
uint64_t FiFPGA::icmCurrentFrameAddr()
{
    uint64_t frame_addr = 0;

    std::unique_lock<std::mutex> locker(class_mutex);

    // Are we in LTR mode?
    if ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Read up the number of frames from our worker thread.
        if (ltr_thread.get())
        {
            frame_addr = (uint64_t)ltr_thread->getCurAddr();
        }
    }
    else
    {
        uint32_t reg_value = 0;

        // Read and return the current addr from the FPGA.
        regRead(ICM_LAST_FRAME_WRITTEN_PTR, &reg_value);

        // If we haven't written any frames yet, report the current address as zero.
        if (reg_value == LAST_FRAME_INIT)
            reg_value = 0;

        // Convert to a byte address.
        frame_addr = (((uint64_t)reg_value) << HEXWORD_SHIFT);
    }

    // Return our frame address.
    return frame_addr;
}

/*!
 * @brief Arming logic specific to the calibration process.
 * @return True on success, false otherwise.
 */
bool FiFPGA::calibrateArm()
{
    bool    rc = true;
    uint32_t    value = 0;
    CameraServer *cs_ptr = CameraServer::get_cs();

    // If we are in low light mode, switch to normal mode before arming the sensor.
    if (icmIsLowLightEnabled())
    {
        float frame_rate;

        cs_ptr->get_frame_rate(&frame_rate);
        icmSetFrameRate(frame_rate);
        icmSetLowLightShutterSpeed(false, 0);
    }

    // Reset the last written value.
    regWrite(ICM_LAST_FRAME_WRITTEN_PTR, LAST_FRAME_INIT);

    value = 0;
    if (!regRead(ICM_CONTROL, &value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read the ICM_CONTROL register.", __FUNCTION__);
        return false;
    }

    // Use basic mode for calibration.
    value &= ~ICMC_CAPTURE_MODE_MASK;
    value |= ICMC_CAPTURE_MODE_BASIC;

    // Use continuous mode for calibration.
    value &= ~(ICMC_CAPTURE_SINGLE);

    // Arm the ICM.
    value |= ICMC_ARM;

    // Write the new control register settings out.
    if (!regWrite(ICM_CONTROL, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write the ICM_CONTROL register\n", __FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Logic to disarm the camera which is specific to the calibration process.
 * @return True on success, false otherwise.
 */
bool FiFPGA::calibrateDeArm()
{
    int i;
    bool rc = true;
    uint32_t value = 0;

    // Set the UNARM bit.
    regModify(ICM_CONTROL, ICMC_UNARM, -1);

    // Wait to give the ICM a chance to release.
    for (i = 0; i < MAX_RETRY_COUNT; ++i)
    {
        struct timespec sleep_req;      // Delay period.

        // Wait for a bit.
        sleep_req.tv_sec = ICM_BUSY_WAIT_SEC;
        sleep_req.tv_nsec = ICM_BUSY_WAIT_NSEC;
        nanosleep(&sleep_req, NULL);

        // Read the register.
        regRead(ICM_STATUS, &value);

        // Did our bit go low?
        if ((value & ICMS_ARMED) == 0)
            break;
    }

    // Did we exceed our retry count?
    if (i == MAX_RETRY_COUNT)
    {
        fi_log(LOG_ERR, "%s: ICMS_ARMED never dropped [0x%x]", __FUNCTION__, value);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Capture black frames in image memory without regard to chosen mode
 *
 * We don't want to capture black frames in ROC/BROC mode so we need to bypass
 * the normal arming setup. Also, this lets us avoid writing to the SSD
 * which could at some point be a problem.
 *
 * @return  true if num_frames were captured succesfully, false if error
 */
bool FiFPGA::icmCalibrateCapture(uint32_t num_frames)
{
    bool        rc = true;
    uint32_t        frame_count;    // Number of frames captured so far.
    uint32_t        last_frame_count;   // The last frame count we read.
    uint32_t        frame_addr;     // Address of last frame captured, or zero if none.
    struct timespec sleep_req;      // Amount of time we'll sleep between queries.

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // ARM the camera to start capturing calibration frames.
    if (!calibrateArm())
    {
        rc = false;
        fi_log(LOG_ERR, "%s: calibrate_arm() failed!", __PRETTY_FUNCTION__);
        goto bail_cc;
    }

    // Initialize.
    last_frame_count = 0;

    // Wait for the proper number of frames to be captured.
    do
    {
        // Delay a little bit before reading the count.
        sleep_req.tv_sec = 0;
        sleep_req.tv_nsec = 100 * 1000 * 1000;
        nanosleep(&sleep_req, NULL);

        // Read the current frame count and last frame written.
        // No need for an error message since read_reg() does it.
        if (!regRead(ICM_LAST_FRAME_WRITTEN_PTR, &frame_addr) ||
            !regRead(ICM_CAPTURE_FRAME_COUNT_LIVE, &frame_count))
        {
            rc = false;
            break;
        }

        // Are we stalled?
        if ((frame_count == last_frame_count) || (frame_addr == LAST_FRAME_INIT))
        {
            fi_log(LOG_ERR, "%s: frame position is stalled (count=%u last=0x%08x)!",
                   __PRETTY_FUNCTION__, frame_count, frame_addr);
            rc = false;
            break;
        }
    } while (frame_count <= num_frames);

    // Stop the ICM now that we have captured enough frames.
    calibrateDeArm();

bail_cc:
    // Return the result.
    return rc;
}

uint32_t FiFPGA::icmCaptureFrameCount()
{
    uint32_t frame_count(0);
    PartMgr *partmgr_ptr;           // Pointer to the partition manager class.

    // Get a pointer to our support classes.
    partmgr_ptr = CameraServer::get_partmgr();

    // Read where the FPGA is in the current partition.
    regRead(ICM_CAPTURE_FRAME_COUNT_LIVE, &frame_count);

    // Are we in FasFireLR mode?
    if (ltrFasFireEnabled())
    {
        uint64_t total_count;

        // Adjust by the number of finished partitions.
        total_count = ((uint64_t)frame_count) + (partmgr_ptr->getFinishedPartCount() * partmgr_ptr->get_im_buf_count());
        if (total_count > 0xFFFFFFFFULL)
            frame_count = 0xFFFFFFFFU;
        else
            frame_count = (uint32_t)total_count;
    }

    // Return our frame count.
    return frame_count;
}

/*!
 * @brief Turn on the ICM interface for display of live video.
 *
 * This uses the current register settings.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::icmSetToLiveVideo()
{
    CameraServer *cs_ptr = CameraServer::get_cs();

    uint16_t start_x;
    uint16_t start_y;
    uint16_t roi_width;
    uint16_t roi_height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    cs_ptr->get_physical_settings(&start_x, &start_y, &roi_width, &roi_height, &bit_mode,
                                  &frame_rate, &shutter_speed, &sensor_opts);

    return icmSetLiveArgs(bit_mode, frame_rate, shutter_speed);
}

/*!
 * @brief Turn on the ICM interface for display of live video with passed args.
 *
 * When we do a factory calibration, we need to pass in the parameters instead of
 * using the GigE register database.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::icmSetLiveArgs(bitmode_t bit_mode, float frame_rate, uint32_t shutter_speed)
{
    uint32_t    value;
    bool    rc = false;

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // Check to see if the ICM is in IDLE mode - possible if we've bailed out of
    // capture or something earlier.
    if (icmIsPipelineIdle(nullptr))
    {
        // Flush any events that were queued while we ignored them.
        icmEmptyEventFIFO();

        // The ICM is IDLE, so get it running again.
        // This will turn on sensor readout at the right point.
        if (!icmStartup(bit_mode, frame_rate, shutter_speed))
        {
            fi_log(LOG_WARNING, "%s: Unable to restart ICM.", __FUNCTION__);
            goto sl_bail1;
        }
    }

    // Turn on the SENSOR READOUT ENABLE bit in the slave mode control register.
    // We now enable the pipeline before this, mainly for symmetry with the sequence for
    // taking it down, in which we turn off the sensor readout after disabling
    // the ICM pipe; there was a race that caused a hang with the order reversed
    // on takedown (new ICM for 1.8 ROC/BROC changes).
    if (!icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __FUNCTION__);
        goto sl_bail1;
    }

    // We expect to see the state ICMS_PIPELINE_READY here but we won't see it
    // if we have sync-in=PPF and there's no signal. We'll watch for it briefly
    // and warn if we don't see it.
    if (!icmIsPipelineReady(&value))
    {
        // If we didn't see PIPELINE READY, it might be because it is still IDLE.
        // No need to log that case.
        if ((value & ICMS_PIPELINE_IDLE) != ICMS_PIPELINE_IDLE)
        {
            uint32_t icmc = 0xffffffff;

            regRead(ICM_CONTROL, &icmc);
            fi_log(LOG_WARNING, "%s: No ICMS_PIPELINE_READY, ICM_STATUS=0x%08x ICM_CONTROL=0x%08x", __FUNCTION__, value, icmc);

            dump_regs();
        }
    }

    // All is well.
    rc = true;

sl_bail1:

    // All done.
    return rc;
}

/*!
 * @brief Tell the FPGA to turn off the ICM interface.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::icmDisable()
{
    bool    rc = false;

    TP(FT_ICM, "%s", __PRETTY_FUNCTION__);

    // We have to stop the stream if it's running.
    fcp_stop_stream_thread();

    // Wait for the IPM to go idle.
    if (!ipmIsIdle())
    {
        uint32_t value;
        uint32_t p1_status;

        regRead(IPM_STATUS, &value);
        regRead(IPM_P1_STATUS, &p1_status);

        fi_log(LOG_WARNING, "%s: IPM appears to be stuck [status=0x%x, p1=0x%x]!", __FUNCTION__, value, p1_status);
    }

    // Turn off the ICM pipeline.
    if (!icmPipelineDisable())
    {
        fi_log(LOG_ERR, "%s: Can't disable the ICM pipeline!", __FUNCTION__);
        goto toc_bail1;
    }

    // Turn off integration pulses.
    icmSensorReadoutEnable(false);

    // All done.
    rc = true;

toc_bail1:
    return rc;
}
