/*!
 * @file fpga_io.cpp
 * @brief This source file implements the Fastec Imaging FPGA
 *        I/O class, fpgaIOWorker.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <chrono>
#include <mutex>

#include <fastec/fas_log.h>

#include "fpga_io.h"

#include "cs_trace.h"
#include "camera_server.h"
#include "fpga.h"

namespace FI_FPGA_IO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The MFT "ResetLens" command.
const uint8_t RESET_LENS[]              = { 0xB0, 0xF2, 0x00, 0x00 };

//! How long do we wait around for the FPGA to finish?
const int MFT_COMMAND_MAX_RETRIES = 300;

//--------------------------------------------------------------------
// fpgaEventWorker class declarations
//--------------------------------------------------------------------

class fpgaEventWorker : public FAS_THREAD::Thread
{
public:
    explicit fpgaEventWorker(int fd);
    virtual ~fpgaEventWorker();

    virtual void run() override;
    bool isRunning()                { return thread_running; }

private:
    const uint32_t EVENT_TIMEOUT = 5000;        //!< 5 sec timeout waiting for events.
    const uint32_t POWER_OFF_DEBOUNCE = 500;    //!< Number of milliseconds to debounce the power button.

    volatile bool thread_running;       //!< Flag to indicate if we are running or not.
    int fpga_fd;                        //!< The file descriptor used to talk with the driver.
};

} /* End of namespace FI_FPGA_IO */

using namespace CS_TRACE;
using namespace FI_FPGA;
using namespace FI_FPGA_IO;

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Static Variable Instantiations
//--------------------------------------------------------------------

std::unique_ptr<fpgaIOWorker> fpgaIOWorker::io_proc;
std::mutex fpgaIOWorker::io_mutex;
const char *fpgaIOWorker::FPGA_DRV = "/dev/fi_fpga";
int fpgaIOWorker::fpga_fd = -1;

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

//! @name Message class methods
//! @{

ioFpgaGetRegisterMessage::ioFpgaGetRegisterMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_FPGA_GET_REG;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&access, 0, sizeof(access));
}

ioFpgaGetRegisterMessage::~ioFpgaGetRegisterMessage()
{
}

ioFpgaGetMultiRegisterMessage::ioFpgaGetMultiRegisterMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_FPGA_GET_MULTIREG;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    access.clear();
}

ioFpgaGetMultiRegisterMessage::~ioFpgaGetMultiRegisterMessage()
{
}

ioFpgaSetRegisterMessage::ioFpgaSetRegisterMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_FPGA_SET_REG;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&access, 0, sizeof(access));
}

ioFpgaSetRegisterMessage::~ioFpgaSetRegisterMessage()
{
}

ioFpgaSetMultiRegisterMessage::ioFpgaSetMultiRegisterMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_FPGA_SET_MULTIREG;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    access.clear();
}

ioFpgaSetMultiRegisterMessage::~ioFpgaSetMultiRegisterMessage()
{
}

ioFpgaModifyRegisterMessage::ioFpgaModifyRegisterMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_FPGA_MODIFY_REG;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&modify, 0, sizeof(modify));
}

ioFpgaModifyRegisterMessage::~ioFpgaModifyRegisterMessage()
{
}

ioI2CAccessMessage::ioI2CAccessMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_I2C_ACCESS;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&access, 0, sizeof(access));
}

ioI2CAccessMessage::~ioI2CAccessMessage()
{
}

ioImgDownloadMessage::ioImgDownloadMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_IMG_DNLOAD;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&access, 0, sizeof(access));
}

ioImgDownloadMessage::~ioImgDownloadMessage()
{
}

ioImgUploadMessage::ioImgUploadMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_IMG_UPLOAD;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    memset(&access, 0, sizeof(access));
}

ioImgUploadMessage::~ioImgUploadMessage()
{
}

ioHistogramMessage::ioHistogramMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_READ_HISTOGRAM;
    status = -1;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    dest_ptr = nullptr;
}

ioHistogramMessage::~ioHistogramMessage()
{
}

ioReadImFifoMessage::ioReadImFifoMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_READ_IM_FIFO;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    status = -1;
    byte_count = -1;
    dest_ptr = nullptr;
}

ioReadImFifoMessage::~ioReadImFifoMessage()
{
}

ioMftSendCommandMessage::ioMftSendCommandMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_MFT_SEND_COMMAND;
    resp = FAS_THREAD::RESP_CODE::FAULT;

    isType6 = false;        // Default is a "normal" command.
    cmd_val = nullptr;      // No command just yet.
    data_ptr = nullptr;     // No data for type-6 assigned yet.
    data_count = 0;         // No data for type-6 assigned yet.
}

ioMftSendCommandMessage::~ioMftSendCommandMessage()
{
}

ioMftReadDataMessage::ioMftReadDataMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_MFT_READ_DATA;
    resp = FAS_THREAD::RESP_CODE::FAULT;

    isTypeB = false;        // Default is a "normal" data read.
    typeB_Fault = false;
    max_count = 0;          // No request, no data.
    data_ptr.clear();
}

ioMftReadDataMessage::~ioMftReadDataMessage()
{
}

ioMftResetLensMessage::ioMftResetLensMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_MFT_RESET_LENS;
    resp = FAS_THREAD::RESP_CODE::FAULT;
}

ioMftResetLensMessage::~ioMftResetLensMessage()
{
}

ioMftAccessoryCheckMessage::ioMftAccessoryCheckMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_MFT_ACCESSORY_CHECK;
    resp = FAS_THREAD::RESP_CODE::FAULT;

    accessoryFound = false;     // No accessories found yet.
}

ioMftAccessoryCheckMessage::~ioMftAccessoryCheckMessage()
{
}

ioMftLensCheckMessage::ioMftLensCheckMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IO_MFT_LENS_CHECK;
    resp = FAS_THREAD::RESP_CODE::FAULT;

    lensFound = false;     // No accessories found yet.
}

ioMftLensCheckMessage::~ioMftLensCheckMessage()
{
}

//! @}
// End of message class methods.

//--------------------------------------------------------------------
// fpgaIOWorker class implementation
//--------------------------------------------------------------------

fpgaIOWorker::fpgaIOWorker() :
    FAS_THREAD::MsgThread()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_MFT_PROC;

    // Try to open up the FPGA driver interface.
    fpga_fd = ::open(FPGA_DRV, O_RDWR, 0666);
    if (fpga_fd < 0)
    {
        fi_log(LOG_ERR, "%s: can't open fpga driver %s, errno:%d(%s)", __FUNCTION__, FPGA_DRV, errno, strerror(errno));
    }
    else
    {
        // Get the event thread running.
        startEventWorker();
    }
}

fpgaIOWorker::~fpgaIOWorker()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

fpgaIOWorker *fpgaIOWorker::factory()
{
    fpgaIOWorker *rc(nullptr);

    std::lock_guard<std::mutex> lock(io_mutex);

    // We only allow one thread object.
    if (io_proc.get() == nullptr)
    {
        // Create our thread.
        io_proc = std::unique_ptr<fpgaIOWorker>(new fpgaIOWorker());
        if (io_proc.get() != nullptr)
            io_proc.get()->start();

        // Return the pointer to the singleton.
        rc = io_proc.get();
    }

    // Check to see if we were halted.
    else if (io_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = io_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Start up the event worker thread.
 */
void fpgaIOWorker::startEventWorker()
{
    // See if we need to allocate our worker thread.
    if (event_worker.get() == nullptr)
    {
        event_worker = std::unique_ptr<fpgaEventWorker>(new fpgaEventWorker(fpga_fd));
        if (event_worker.get() == nullptr)
        {
            fi_log(LOG_ERR, "%s: Can't allocate the worker thread!", __PRETTY_FUNCTION__);
            return;
        }
    }

    // Start up our worker thread.
    event_worker->start();
}

/*!
 * @brief Process the messages for the thread.
 */
void fpgaIOWorker::messageHandle(FAS_THREAD::Message *message)
{
    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case IO_FPGA_GET_REG:
            handleFpgaReadRegisterRequest(static_cast<ioFpgaGetRegisterMessage *>(message));
            break;
        case IO_FPGA_GET_MULTIREG:
            handleFpgaReadMultiRegisterRequest(static_cast<ioFpgaGetMultiRegisterMessage *>(message));
            break;
        case IO_FPGA_SET_REG:
            handleFpgaWriteRegisterRequest(static_cast<ioFpgaSetRegisterMessage *>(message));
            break;
        case IO_FPGA_SET_MULTIREG:
            handleFpgaWriteMultiRegisterRequest(static_cast<ioFpgaSetMultiRegisterMessage *>(message));
            break;
        case IO_FPGA_MODIFY_REG:
            handleFpgaModifyRegisterRequest(static_cast<ioFpgaModifyRegisterMessage *>(message));
            break;
        case IO_I2C_ACCESS:
            handleI2CAccessRequest(static_cast<ioI2CAccessMessage *>(message));
            break;
        case IO_IMG_DNLOAD:
            handleImgDownloadRequest(static_cast<ioImgDownloadMessage *>(message));
            break;
        case IO_IMG_UPLOAD:
            handleImgUploadRequest(static_cast<ioImgUploadMessage *>(message));
            break;
        case IO_READ_HISTOGRAM:
            handleHistogramRequest(static_cast<ioHistogramMessage *>(message));
            break;
        case IO_READ_IM_FIFO:
            handleReadImFifoRequest(static_cast<ioReadImFifoMessage *>(message));
            break;

        case IO_MFT_SEND_COMMAND:
            handleMftSendCommandRequest(static_cast<ioMftSendCommandMessage *>(message));
            break;
        case IO_MFT_READ_DATA:
            handleMftReadDataRequest(static_cast<ioMftReadDataMessage *>(message));
            break;
        case IO_MFT_RESET_LENS:
            handleMftResetLensRequest(static_cast<ioMftResetLensMessage *>(message));
            break;
        case IO_MFT_ACCESSORY_CHECK:
            handleMftAccessoryCheckRequest(static_cast<ioMftAccessoryCheckMessage *>(message));
            break;
        case IO_MFT_LENS_CHECK:
            handleMftLensCheckRequest(static_cast<ioMftLensCheckMessage *>(message));
            break;

        case FAS_THREAD::DUMP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            // All done.
            handleHalt(message);
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Do we still have a message?
        if (message != nullptr)
        {
            // Forward on the messsage.
            message->forward();
        }
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in fpgaIOWorker:\n%s", e.what().c_str());

        if (message != nullptr)
        {
            message->resp = FAS_THREAD::FAULT;
            message->forward();
        }
    }
}

/*!
 * @brief Modify the FPGA's register.
 */
void fpgaIOWorker::handleFpgaModifyRegisterRequest(ioFpgaModifyRegisterMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Handle the modify request.
    if (regModify(msg->modify))
    {
        // Looking good.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}

/*!
 * @brief Read the FPGA's register.
 */
void fpgaIOWorker::handleFpgaReadRegisterRequest(ioFpgaGetRegisterMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the read.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_GETREG, &msg->access);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x error:%d(%s)", __PRETTY_FUNCTION__, msg->access.reg_id, errno, strerror(errno));
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

void fpgaIOWorker::handleFpgaReadMultiRegisterRequest(ioFpgaGetMultiRegisterMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Loop through all the entries in our acces structure.
    for (std::size_t i = 0; i < msg->access.size(); ++i)
    {
        // Request the driver to perform the read.
        msg->status = ioctl(fpga_fd, FI_FPGA_IOC_GETREG, &msg->access[i]);
        if (msg->status == -1)
        {
            // IOCTL failed, bail.
            fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x error:%d(%s)", __PRETTY_FUNCTION__, msg->access[i].reg_id, errno, strerror(errno));
            goto bail;
        }
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Read the FPGA's register.
 */
void fpgaIOWorker::handleFpgaWriteRegisterRequest(ioFpgaSetRegisterMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the write.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_SETREG, &msg->access);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x err:%d(%s)", __PRETTY_FUNCTION__, msg->access.reg_id, msg->access.value, errno, strerror(errno));
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

void fpgaIOWorker::handleFpgaWriteMultiRegisterRequest(ioFpgaSetMultiRegisterMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Loop through all the entries in our acces structure.
    for (FI_REG_ACCESS access : msg->access)
    {
        // Request the driver to perform the write.
        msg->status = ioctl(fpga_fd, FI_FPGA_IOC_SETREG, &access);
        if (msg->status == -1)
        {
            // IOCTL failed, bail.
            fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x err:%d(%s)", __PRETTY_FUNCTION__, access.reg_id, access.value, errno, strerror(errno));
            goto bail;
        }
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

bool fpgaIOWorker::regModify(FI_REG_MODIFY& modify)
{
    int status;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        return false;
    }

    // Request the driver to perform the read.
    status = ioctl(fpga_fd, FI_FPGA_IOC_MODREG, &modify);
    if (status == -1)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x mask:0x%x err:%d(%s)", __PRETTY_FUNCTION__, modify.reg_id, modify.value, modify.mask, errno, strerror(errno));
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Process the I2C access request.
 */
void fpgaIOWorker::handleI2CAccessRequest(ioI2CAccessMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the write.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_I2C_ACCESS, &msg->access);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        fi_log(LOG_ERR, "%s: IOCTL failed bus:%d chip=0x%x reg:0x%x err:%d(%s)",
               __PRETTY_FUNCTION__, msg->access.bus, msg->access.chip_addr, msg->access.reg_addr, errno, strerror(errno));
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Process the image download request.
 */
void fpgaIOWorker::handleImgDownloadRequest(ioImgDownloadMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the write.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_READ_IMAGE, &msg->access);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Process the image upload request.
 */
void fpgaIOWorker::handleImgUploadRequest(ioImgUploadMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the write.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_WRITE_IMAGE, &msg->access);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Process the histogram read request.
 */
void fpgaIOWorker::handleHistogramRequest(ioHistogramMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we have a destination pointer.
    if (msg->dest_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: destination pointer is null!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Request the driver to perform the write.
    msg->status = ioctl(fpga_fd, FI_FPGA_IOC_GET_IPM_HISTO, msg->dest_ptr);
    if (msg->status == -1)
    {
        // IOCTL failed, bail.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Process the read image memory FIFO request.
 */
void fpgaIOWorker::handleReadImFifoRequest(ioReadImFifoMessage *msg)
{
    int i;
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we have a destination pointer.
    if (msg->dest_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: destination pointer is null!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if ((msg->byte_count < 0) || ((msg->byte_count & 0xF) != 0))
    {
        fi_log(LOG_DEBUG, "%s: byte_count [%d] is invalid!", __PRETTY_FUNCTION__, msg->byte_count);
        goto bail;
    }

    // Transfer the image over in 16 byte chunks.
    for (i = 0; i < msg->byte_count; i += 16)
    {
        // Request the driver to perform the write.
        msg->status = ioctl(fpga_fd, FI_FPGA_IOC_READ_IM_FIFO, msg->dest_ptr);
        if (msg->status == -1)
        {
            // IOCTL failed, bail.
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            goto bail;
        }

        msg->dest_ptr += 16;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;
}

void fpgaIOWorker::handleMftAccessoryCheckRequest(ioMftAccessoryCheckMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    TP(TRACE_SEL_MFT_PROC, "%s: unsupported", __PRETTY_FUNCTION__);

    // All done.
    msg->resp = resp;
}

void fpgaIOWorker::handleMftLensCheckRequest(ioMftLensCheckMessage *msg)
{
    FI_REG_MODIFY modify;
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;
    ioFpgaGetRegisterMessage readMsg;

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Ask the FPGA to start the lens check procedure.
    modify.reg_id = MFT_CONTROL_STATUS;
    modify.value = MFT_CHECK_LENS_CMD;
    modify.mask = -1;
    if (!regModify(modify))
    {
        fi_log(LOG_WARNING, "%s: Unable to assert MFT_LSB2L.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Wait for the FPGA to finish.
    if (!mftWait())
    {
        fi_log(LOG_WARNING, "%s: FPGA is stuck busy.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Did we find the lens?
    readMsg.access.reg_id = MFT_CONTROL_STATUS;
    handleFpgaReadRegisterRequest(&readMsg);
    if (readMsg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_WARNING, "%s: Unable to read the MFT_CONTROL_STATUS register.", __PRETTY_FUNCTION__);
        goto bail;
    }
    msg->lensFound = ((readMsg.access.value & MFT_CHECK_LENS_PASS) == MFT_CHECK_LENS_PASS);

    // Return our status.
    TP(TRACE_SEL_MFT_PROC, "%s: lens_found=%d", __PRETTY_FUNCTION__, msg->lensFound);

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;

}

/*!
 * @brief Process the MFT read data request.
 */
void fpgaIOWorker::handleMftReadDataRequest(ioMftReadDataMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null message pointer?!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (msg->max_count <= 0)
    {
        fi_log(LOG_DEBUG, "%s: Invalid max_count: %d", __PRETTY_FUNCTION__, msg->max_count);
        goto bail;
    }
    if (msg->cmd_val == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null cmd_val?!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Send the command.
    if (!sendCommand(msg->cmd_val))
    {
        fi_log(LOG_DEBUG, "%s: sendCommand() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Read the data. TypeB messages expect a 1-byte TDL, others get 2-bytes.
    if (!readBytes(msg->data_ptr, msg->max_count, (msg->isTypeB ? 1 : 2), &msg->typeB_Fault))
    {
        fi_log(LOG_DEBUG, "%s: readBytes() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;

}

/*!
 * @brief Send a reset lens command.
 */
void fpgaIOWorker::handleMftResetLensRequest(ioMftResetLensMessage *msg)
{
    std::vector<uint8_t> data_ptr;
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null message pointer?!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Send the reset command.
    if (!sendCommand(RESET_LENS))
    {
        fi_log(LOG_WARNING, "%s: RESET_LENS failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Read the byte from the reset.
    if (readBytes(data_ptr, 1, 1) == 1)
    {
        // Looking good.
        resp = FAS_THREAD::OK;
    }

bail:
    // All done.
    msg->resp = resp;

}

/*!
 * @brief Process the MFT send command request.
 */
void fpgaIOWorker::handleMftSendCommandRequest(ioMftSendCommandMessage *msg)
{
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::FAULT;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (fpga_fd < 0)
    {
        // Make sure all failure paths have an error message so the caller doesn't have
        // to have its own.
        fi_log(LOG_ERR, "%s: Failed due to fpga_fd<0", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null message pointer?!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (msg->cmd_val == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null command pointer?!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Are we type-6?
    if (msg->isType6)
    {
        // Sanity checks.
        if (msg->data_count <= 0)
        {
            fi_log(LOG_DEBUG, "%s: Invalid Type6 data count: %d", __PRETTY_FUNCTION__, msg->data_count);
            goto bail;
        }
        if (msg->data_ptr == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Null Type6 data pointer?!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Send off our Type6 message.
        if (!sendCommandType6(msg->cmd_val, msg->data_ptr, msg->data_count))
        {
            fi_log(LOG_DEBUG, "%s: sendCommandType6() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
    }
    else
    {
        // We have a "normal" command, so send it.
        if (!sendCommand(msg->cmd_val))
        {
            fi_log(LOG_DEBUG, "%s: sendCommand() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
    }

    // Looking good.
    resp = FAS_THREAD::OK;

bail:
    // All done.
    msg->resp = resp;

}

/*!
 * @brief Process the HALT message.
 */
void fpgaIOWorker::handleHalt(FAS_THREAD::Message *message)
{
    std::lock_guard<std::mutex> lock(io_mutex);

    LOG("%s++", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (fpga_fd >= 0)
    {
        int status;
        FI_REG_ACCESS access;

        // Ask the FPGA to reset the interface.
        access.reg_id = NC1_FPGA_CONTROL;
        access.value = NC1FC_EMIF_SW_RESET;

        // Request the driver to perform the write.
        status = ioctl(fpga_fd, FI_FPGA_IOC_SETREG, &access);
        if (status == -1)
        {
            // IOCTL failed, bail.
            fi_log(LOG_ERR, "%s: IOCTL failed reg:0x%x value:0x%x err:%d(%s)", __PRETTY_FUNCTION__, access.reg_id, access.value, errno, strerror(errno));
        }
    }

    // Do we have an event worker running?
    if (event_worker.get() != nullptr)
    {
        // Is it running?
        if (event_worker->isRunning())
        {
            // Ask it to stop.
            LOG("%s: Ask event worker to stop.", __PRETTY_FUNCTION__);
            event_worker->stop();
        }

        // Release it.
        LOG("%s: Release event worker.", __PRETTY_FUNCTION__);
        event_worker.reset();
    }

    // Close the driver file.
    if (fpga_fd >= 0)
    {
        // Close down the FPGA driver interface.
        ::close(fpga_fd);
        fpga_fd = -1;
    }

    // Remember that we've stopped.
    halted = true;

    // All done.
    message->resp = FAS_THREAD::OK;
    LOG("%s--", __PRETTY_FUNCTION__);
}


/*!
 * @brief Write a command sequence to the MFT interface.
 *
 * This is appropriate for use with all commands other than type-6. That is,
 * commands which do not also have data associated with the command. For a
 * type-6 command, use sendCommandType6().
 *
 * @param cmd_val       [IN] Pointer to the command sequence bytes.
 * @return True on success, false otherwise.
 */
bool fpgaIOWorker::sendCommand(const uint8_t *cmd_val)
{
    bool rc = false;
    int i;
    uint32_t reg_value;
    ioFpgaSetRegisterMessage writeMsg;

    // Very noisy, so only uncomment when required!
    TP(TRACE_SEL_MFT_PROC, "%s: 0x%02x 0x%02x 0x%02x 0x%02x", __PRETTY_FUNCTION__, cmd_val[0], cmd_val[1], cmd_val[2], cmd_val[3]);

    // Build up the 32-bit command value.
    reg_value = (((uint32_t)cmd_val[0]) << 24) | (((uint32_t)cmd_val[1]) << 16) | (((uint32_t)cmd_val[2]) << 8) | ((uint32_t)cmd_val[3]);

    // Write the command.
    writeMsg.access.reg_id = MFT_COMMAND;
    writeMsg.access.value = (uint32_t)reg_value;
    handleFpgaWriteRegisterRequest(&writeMsg);
    if (writeMsg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_WARNING, "%s: Unable to write data!", __PRETTY_FUNCTION__);
        goto sc_bail;
    }

    // Wait for the FPGA to finish.
    for (i = 0; i < MFT_COMMAND_MAX_RETRIES; ++i)
    {
        ioFpgaGetRegisterMessage readMsg;

        // Read the status.
        readMsg.access.reg_id = MFT_CONTROL_STATUS;
        handleFpgaReadRegisterRequest(&readMsg);
        if (readMsg.resp != FAS_THREAD::OK)
        {
            fi_log(LOG_WARNING, "%s: Unable to read the command register.", __PRETTY_FUNCTION__);
            goto sc_bail;
        }

        // Is the busy bit clear?
        if ((readMsg.access.value & MFT_CONTROL_BUSY) != MFT_CONTROL_BUSY)
        {
            // We are done. Was it successful?
            if ((readMsg.access.value & MFT_CMD_BLOCK_CHECKSUM_ERR) == MFT_CMD_BLOCK_CHECKSUM_ERR)
            {
                fi_log(LOG_WARNING, "%s: FPGA shows a command block checksum error!", __PRETTY_FUNCTION__);
                goto sc_bail;
            }

            // All is well!
            rc = true;
            break;
        }

        // Wait...
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(10));
    }

sc_bail:
    // Are we bailing?
    if (!rc)
        mftControlReset();

    // Return our status.
    return rc;
}

/*!
 * @brief Write a type-6 command sequence to the MFT interface.
 *
 * Unlike other types of commands, the type-6 command has data associated with the command.
 * We need to fill the TX FIFO prior to sending the actual command sequence.
 *
 * @param cmd_val       [IN] Pointer to the command sequence bytes.
 * @param data_val      [IN] Pointer to the data bytes.
 * @param data_count    [IN] Number of bytes of data to send.
 *
 * @return True on success, false otherwise.
 */
bool fpgaIOWorker::sendCommandType6(const uint8_t *cmd_val, const uint8_t *data_val, int data_count)
{
    int i;
    bool rc = false;

    TP(TRACE_SEL_MFT_PROC, "%s: 0x%02x 0x%02x 0x%02x 0x%02x", __PRETTY_FUNCTION__, cmd_val[0], cmd_val[1], cmd_val[2], cmd_val[3]);

    // Send the number of bytes to follow (the "TDL").
    if (!writeMftByte((uint8_t)data_count))
    {
        fi_log(LOG_WARNING, "%s: Error writing length (%d)", __PRETTY_FUNCTION__, data_count);
        goto sc6_bail;
    }

    // Send the data bytes.
    for (i = 0; i < data_count; ++i)
    {
        if (!writeMftByte(data_val[i]))
        {
            fi_log(LOG_WARNING, "%s: Error writing byte %d [val=0x%02X]", __PRETTY_FUNCTION__, i, data_val[i]);
            goto sc6_bail;
        }
    }

    // Send the command.
    rc = sendCommand(cmd_val);

sc6_bail:
    return rc;
}

/*!
 * @brief Reset the FPGA's MFT module.
 */
void fpgaIOWorker::mftControlReset()
{
    FI_REG_MODIFY modify;

    TP(TRACE_SEL_MFT_PROC, "%s: Resetting the MFT logic.", __PRETTY_FUNCTION__);

    // Assert the reset.
    modify.reg_id = MFT_CONTROL_STATUS;
    modify.value = MFT_CONTROL_RESET;
    modify.mask = -1;
    if (!regModify(modify))
    {
        fi_log(LOG_WARNING, "%s: Unable to assert MFT_CONTROL_RESET.", __PRETTY_FUNCTION__);
    }

    // Wait...
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(1));

    // De-assert the reset.
    modify.value = 0;
    modify.mask = ~MFT_CONTROL_RESET;
    if (!regModify(modify))
    {
        fi_log(LOG_WARNING, "%s: Unable to de-assert MFT_CONTROL_RESET.", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Read up to "max_count" bytes from the MFT interface.
 *
 * @param data_ptr      [OUT] Where to store the bytes read.
 * @param max_count     [IN] Maximum number of bytes to read. Does NOT include the checksum.
 * @param tdl_len       [IN] Size of the TDL field (1=Type-B, 2=Type-C).
 * @param faulted       [OPTIONAL, OUT] True if max_count < TDL.
 *
 * @return The number of bytes read.
 */
int fpgaIOWorker::readBytes(std::vector<uint8_t>& data_ptr, int max_count, int tdl_len, bool *faulted)
{
    int read_count;
    int tdl_value;
    int fifo_count;
    bool read_error(true);
    uint8_t data_bytes[2];
    ioFpgaGetRegisterMessage readMsg;

    // Do we have a faulted pointer? If so, initialize to indicate no fault. If
    // we find a fault later, we can switch it.
    if (faulted != nullptr)
        *faulted = false;

    // Read the FIFO level. This should be in agreement with the TDL, which should be
    // the first byte(s) in the FIFO.
    readMsg.access.reg_id = MFT_RX_FIFO_LEVEL;
    handleFpgaReadRegisterRequest(&readMsg);
    if (readMsg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_WARNING, "%s: Unable to read the command register.", __PRETTY_FUNCTION__);
        goto rb_bail;
    }

    // How much is the lens sending? We had better have, at a minimum, the amount for the TDL.
    fifo_count = readMsg.access.value;
    if (fifo_count < tdl_len)
    {
        fi_log(LOG_DEBUG, "%s: Not enough data in FIFO (%d) for the TDL (%d)", __PRETTY_FUNCTION__, fifo_count, tdl_len);
        goto rb_bail;
    }

    // Read up the TDL byte(s).
    data_bytes[0] = data_bytes[1] = 0;
    for (read_count = 0; !stop_request && (read_count < tdl_len); ++read_count)
    {
        // Read the FIFO.
        readMsg.access.reg_id = MFT_RX_FIFO_READ;
        handleFpgaReadRegisterRequest(&readMsg);
        if (readMsg.resp != FAS_THREAD::OK)
        {
            fi_log(LOG_WARNING, "%s: Unable to read the command register.", __PRETTY_FUNCTION__);
            goto rb_bail;
        }

        // Append the byte to our data vector.
        data_bytes[read_count] = (uint8_t)readMsg.access.value;
    }
    tdl_value = ((int)data_bytes[0]) | (((int)data_bytes[1]) << 8);

    // Is there any data? For example, the reset command sends a TDL of 0.
    if (tdl_value > 0)
    {
#if 0
        if (fifo_count > (max_count + 1))
        {
            FI_REG_MODIFY modify;

            // Assert the NC1FC_LED_A10_STAT0 signal
            modify.reg_id = NC1_FPGA_CONTROL;
            modify.value = NC1FC_LED_A10_STAT0;
            modify.mask = -1;
            regModify(modify);
            fi_log(LOG_WARNING, "%s: byte count is wonky (0x%x) =========================", __PRETTY_FUNCTION__, fifo_count);
        }
#endif

        // Is this a type-B command (tdl_len is 1) or a type-C command (tdl_len is 2)?
        if (tdl_len == 2)
        {
            // Type-C TDL values include the checksum, which the FPGA does NOT put in the
            // FIFO. So we need to adjust the length.
            --tdl_value;
        }

        // Is there enough data in the FIFO?
        if (tdl_value > fifo_count)
        {
            fi_log(LOG_WARNING, "%s: TDL (%d) exceeds FIFO (%d)", __PRETTY_FUNCTION__, tdl_value, fifo_count);
            goto rb_bail;
        }

        // Did we ask for enough? Remember that max_count doesn't include the checksum byte.
        if (tdl_value > max_count)
        {
            fi_log(LOG_DEBUG, "%s: TDL length (%d) > max_count (%d)", __PRETTY_FUNCTION__, tdl_value, max_count);

            // Do we have a faulted pointer?
            if (faulted != nullptr)
                *faulted = true;

            // Reset the maximum amount we'll read.
            tdl_value = max_count;
        }

        // Read up the bytes.
        for (read_count = 0; !stop_request && (read_count < tdl_value); ++read_count)
        {
            // Read the FIFO.
            readMsg.access.reg_id = MFT_RX_FIFO_READ;
            handleFpgaReadRegisterRequest(&readMsg);
            if (readMsg.resp != FAS_THREAD::OK)
            {
                fi_log(LOG_WARNING, "%s: Unable to read the command register.", __PRETTY_FUNCTION__);
                goto rb_bail;
            }

            // Append the byte to our data vector.
            data_bytes[0] = (uint8_t)(readMsg.access.value & 0xFF);
            data_ptr.push_back(data_bytes[0]);
        }
        if (stop_request)
            goto rb_bail;
    }

    // All went well.
    read_error = false;

rb_bail:
    // Was there a read error?
    if (read_error)
    {
        mftControlReset();
        read_count = 0;
    }

    // Return the number of bytes read.
    return read_count;
}

/*!
 * @brief Wait for the FPGA's MFT control module to be "not busy".
 * @return True on success, false otherwise.
 */
bool fpgaIOWorker::mftWait()
{
    int retry;

    // Wait for the signal to go high.
    for (retry = 0; !stop_request && (retry < MFT_COMMAND_MAX_RETRIES); ++retry)
    {
        ioFpgaGetRegisterMessage readMsg;

        readMsg.access.reg_id = MFT_CONTROL_STATUS;
        handleFpgaReadRegisterRequest(&readMsg);
        if (readMsg.resp != FAS_THREAD::OK)
        {
            fi_log(LOG_WARNING, "%s: Unable to read the command register.", __PRETTY_FUNCTION__);
            return false;
        }

        // Is the FPGA not busy?
        if ((readMsg.access.value & MFT_CONTROL_BUSY) == 0)
            break;

        // Wait...
        // Of course, this won't ACTUALLY wait for 1 millisecond, but the point is that
        // we want to delay for a short amount of time.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(1));
    }

    return (!stop_request && (retry < MFT_COMMAND_MAX_RETRIES));
}

/*!
 * @brief Write a byte to the MFT interface.
 * @param reg_data      [IN] Data to write to the MFT TX FIFO.
 * @return True on success, false otherwise.
 */
bool fpgaIOWorker::writeMftByte(uint8_t reg_data)
{
    ioFpgaSetRegisterMessage writeMsg;

    // Write the data.
    writeMsg.access.reg_id = MFT_TX_FIFO_WRITE;
    writeMsg.access.value = (uint32_t)reg_data;
    handleFpgaWriteRegisterRequest(&writeMsg);
    if (writeMsg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_WARNING, "%s: Unable to write data!", __PRETTY_FUNCTION__);
        return false;
    }

    // All is well.
    return true;
}


//============================================================================
// fpgaEventWorker class implementation.
//============================================================================

/*!
 * @brief The worker thread constructor. The FPGA worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
FI_FPGA_IO::fpgaEventWorker::fpgaEventWorker(int fd) :
    FAS_THREAD::Thread(true),
    thread_running(false),
    fpga_fd(fd)
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

FI_FPGA_IO::fpgaEventWorker::~fpgaEventWorker()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Main thread function to handle waiting for FPGA events:
 *    camera armed, trigger assert, trigger deassert, partition complete.
 */
void FI_FPGA_IO::fpgaEventWorker::run()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    bool power_debounce = false;
    struct timespec start_time;
    CS_NETWORK::CS_NETWORK_MESSAGE  msg;        // Structure for either an IPC or UDP message.

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Enable the power button interrupt.
    fpga_ptr->irqEnable(PCIEIRQ_PWR_OFF);

    // We are now running.
    thread_running = true;

    // Loop until we are shut down.
    while (!stop_request)
    {
        int retval;
        FI_EVENT_INFO   event_info;
        memset(&event_info, 0, sizeof(event_info));

        if (power_debounce)
            event_info.event_timeout = POWER_OFF_DEBOUNCE;
        else
            event_info.event_timeout = EVENT_TIMEOUT;

        // Make the blocking FPGA call. I don't think we need to "protect" this
        // interface because it doesn't "change" the behavior of the underlying
        // hardware.
        retval = ioctl(fpga_fd, FI_FPGA_IOC_GETEVENT, &event_info);

        // Did we have a timeout?
        if (retval < 0)
        {
            // Are we debouncing the power button?
            if (power_debounce)
            {
                // Not debouncing anymore.
                power_debounce = false;

                // Clear any "pending" value.
                fpga_ptr->regWrite(PCIE_IRQ_STATUS, PCIEIRQ_PWR_OFF);

                // Re-enable the button's interrupt.
                fpga_ptr->irqEnable(PCIEIRQ_PWR_OFF);
            }

            // Go back around.
            continue;
        }

        // If we are debouncing the power button, and the event was the power button,
        // check to see if we should ignore the event.
        if (power_debounce)
        {
            uint64_t elapsed_usecs;

            // How much time has elapsed?
            elapsed_usecs = measuretime(0, &start_time);
            if ((elapsed_usecs / 1000) > (uint64_t)POWER_OFF_DEBOUNCE)
            {
                // Not debouncing anymore.
                power_debounce = false;

                // Clear any "pending" value.
                fpga_ptr->regWrite(PCIE_IRQ_STATUS, PCIEIRQ_PWR_OFF);

                // Re-enable the button's interrupt.
                fpga_ptr->irqEnable(PCIEIRQ_PWR_OFF);
            }

            // OK. Did we get another POWER button event?
            else if ((event_info.event_value & PCIEIRQ_PWR_OFF) == PCIEIRQ_PWR_OFF)
            {
                // Make sure we disable the interrupt.
                fpga_ptr->irqDisable(PCIEIRQ_PWR_OFF);
                continue;
            }
        }

        // We returned from the blocking call. Tell the main app that an event has happened.
        // We want to do this so that all updating is on the primary thread.
        msg.msg_type = CS_NETWORK::MSG_TYPE_FPGA;
        msg.fpga_val = event_info.event_value;
        CameraServer::add_work(&msg);

        // Do we need to debounce the power button?
        if ((event_info.event_value & PCIEIRQ_PWR_OFF) == PCIEIRQ_PWR_OFF)
        {
            // Remember that we're in debounce mode for the power button.
            power_debounce = true;

            // Get our starting time for the debounce operation.
            measuretime(1, &start_time);
        }

    }

    // We are no longer running.
    thread_running = false;
    fi_log(LOG_DEBUG, "%s: done", __PRETTY_FUNCTION__);
}
