/*!
 * @file fpga_ipm.cxx
 * @brief This source file implements the Fastec Imaging FPGA
 *    class, FiFPGA, Image Processing Module (IPM) methods.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <memory>
#include <string>
#include <vector>

#include <fmt/printf.h>

#include <fastec/gige_msg.h>
#include <fastec/fi_lutfile.h>
#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "camera_server.h"
#include "cs_utils.h"
#include "fpga.h"
#include "fcp.h"
#include "image_correction.h"
#include "ipm_proc.h"
#include "part_mgr.h"

namespace FI_FPGA_IPM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const int MAX_RETRY_COUNT   = 10;
const int IPM_BUSY_ATTEMPTS = 5;        //!< How many times should we wait for the IPM to free up.
const uint32_t IPM_BUSY_WAIT = 100000;  //!< Wait 100ms for the IPM to free up.

const uint32_t MATRIX_COEFF_MASK = 0xFFFF;  //!< Mask to be applied to matrix coefficient values.
const uint32_t WB_MASK           = 0xFFFF;  //!< Mask to be applied to white balance gain values.
const uint32_t DC_OFFSET_MASK    = 0xFFFF;  //!< Mask to be applied to dc-offset values.

// FPGA register defaults, shifts and masks.
const uint32_t FPGA_HIST_SUM_SQUARED_H_MASK = 0x1F;

} /* End of namespace FI_FPGA_IPM */

using namespace FI_FPGA;
using namespace FI_FPGA_IPM;


/*!
 * @brief Get IPM_Px_CONFIG bits describing the source buffer
 *
 * @param bit_sel   The bit mode used to capture the image
 *
 * @return configuration bits for IPMC_SRC_BUF_BIT_DEPTH... and IPMC_SRC_BUF_BIT_SELECT...
 */
static uint32_t get_ipm_src_bits(bitmode_t bit_sel)
{
    Sensor   *sensor_ptr = CameraServer::get_sensor();
    uint32_t config = 0;

    // Tell the IPM what data bits were used to capture the data so it knows how to shift
    // FPN correction values if FPGA Pixel FPN is in use. (This will be ignored if no Pixel FPN.)
    switch ((unsigned)bit_sel)
    {
    case GIGE_BITMODE_11_4:
        config |= (4 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;
    case GIGE_BITMODE_10_3:
        config |= (3 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;
    case GIGE_BITMODE_11_2:     // Low bit is bit 2.
    case GIGE_BITMODE_9_2:
        config |= (2 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;
    case GIGE_BITMODE_10_1:     // Low bit is bit 1.
    case GIGE_BITMODE_8_1:
        config |= (1 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;
    case GIGE_BITMODE_11_0:     // Low bit is bit 0.
    case GIGE_BITMODE_9_0:
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_LOWER_8_BITS:
        config |= (0 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;

    case GIGE_BITMODE_UPPER_8_BITS:
        if (IS_10_BIT_MODE(sensor_ptr->get_max_bit_mode()))
            // 9:2 of 10-bit sensor data
            config |= (2 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        else
            // 11:4 of 12-bit sensor data
            config |= (4 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;

    case GIGE_BITMODE_MIDDLE_8_BITS:
        if (IS_10_BIT_MODE(sensor_ptr->get_max_bit_mode()))
            // 8:1 of 10-bit sensor data
            config |= (1 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        else
            // 9:2 of 12-bit sensor data
            config |= (2 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;

    case GIGE_BITMODE_10_BITS:
        if (IS_10_BIT_MODE(sensor_ptr->get_max_bit_mode()))
            // 9:0 of 10-bit sensor data
            config |= (0 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        else
            // Upper 10 bits, 11:2 of 12-bit sensor data
            config |= (2 << IPMC_SRC_BUF_BIT_SELECT_SHIFT);
        break;
    }

    // Say how many bits each captured pixel takes.
    if (IS_10_BIT_MODE(bit_sel))
    {
        config |= IPMC_SRC_BIT_DEPTH_10;
    }
    else if (IS_12_BIT_MODE(bit_sel))
    {
        config |= IPMC_SRC_BIT_DEPTH_12;
    }
    else
    {
        config |= IPMC_SRC_BIT_DEPTH_8;
    }

    return config;
}


/*!
 * @brief Get IPM_Px_CONFIG bits describing the destination
 *
 * @param image_type   The type of image we are creating
 *
 * @return configuration bits for IPMC_DEST_TYPE_...
 */
static uint32_t get_ipm_dest_bits(unsigned image_type)
{
    uint32_t config = 0;

    // Tell the FPGA which image type we are producing.
    switch (image_type)
    {
    case IMGPROC_TYPE_MONO8:
    case IMGPROC_TYPE_Y8:
        config |= IPMC_DEST_TYPE_MONO8;
        break;
    case IMGPROC_TYPE_JPEG:
    case IMGPROC_TYPE_NV12:
    case IMGPROC_TYPE_RGB24:
        config |= IPMC_DEST_TYPE_RGB24;
        break;
    case IMGPROC_TYPE_RAW8:
        config |= IPMC_DEST_TYPE_RAW8;
        break;
    case IMGPROC_TYPE_RAW16:
        config |= IPMC_DEST_TYPE_RAW16;
        break;
    default:
        // Whatever this was, it isn't legal.
        fi_log(LOG_ERR, "%s: Invalid image type (0x%x)!", __PRETTY_FUNCTION__, image_type);
        break;
    }
    return config;
}

/*
 * @brief initial FPGA IPM logic at startup
 *
 * Allow for the possibility that we are restarting the CameraServer when it was
 * previously running, so some FPGA logic is still active.
 */
void FiFPGA::ipmInit(void)
{
    uint32_t value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Read the IPM status.
    if (!regRead(IPM_STATUS, &value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read IPM_STATUS.", __FUNCTION__);
        return;
    }

    // We expect the IPM to be idle.
    if (value != 0)
    {
        // If something is going on with the histogram engine, reset it.
        if ((value & (IPMS_HISTO_BUSY | IPMS_HISTO_INIT_BUSY)) != 0)
        {
            fi_log(LOG_DEBUG, "%s: IPM_STATUS=0x%08x, resetting histogram", __FUNCTION__, value);
            ipmRgbHistoReset();
        }

        // Re-read the status and see if we're idle now.
        regRead(IPM_STATUS, &value);
        if (value != 0)
        {
            // Attempt resetting the entire module.
            fi_log(LOG_DEBUG, "%s: IPM is not idle (0x%x). Resetting...", __FUNCTION__, value);
            ipmReset();

            // Once more...
            regRead(IPM_STATUS, &value);
            if (value != 0)
            {
                fi_log(LOG_WARNING, "%s: IPM is not idle (0x%x) after reset!", __FUNCTION__, value);
                return;
            }
        }
    }

    // DMA src size to max.
    if (!regWrite(IPM_SRC_DMA_BURST_SIZE, IPM_SRC_DMA_REQ_SIZE_VAL))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_SRC_DMA_REQ_SIZE.", __FUNCTION__);
        return;
    }

    // SRC DMA back pressure, using setting specified by Bob for maximum
    if (!regWrite(IPM_SRC_DMA_BP_THRESHOLD, IPM_SRC_DMA_BP_THRESHOLD_VAL))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_SRC_DMA_BP_THRESHOLD.", __FUNCTION__);
        return;
    }

    // Set DMA dest size to max, but enable timer to slow down and do periodic requests
    // The period is in 125 MHz ticks.
    // Temporarily disable timer
    //    value = IPM_DEST_DMA_REQ_SIZE_VAL | IPM_DEST_DMA_REQ_TIMER_ENA |
    //          ( 1000 << IPM_DEST_DMA_REQ_PERIOD_SHIFT ) /* 8 usecs */;
    value = IPM_DEST_DMA_REQ_SIZE_VAL;
    if (!regWrite(IPM_DEST_DMA_BURST_SIZE, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_DEST_DMA_REQ_SIZE.", __FUNCTION__);
        return;
    }

    // DEST DMA back pressure, using setting specified by Bob for maximum
    if (!regWrite(IPM_DEST_DMA_BP_THRESHOLD, IPM_DEST_DMA_BP_THRESHOLD_VAL))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_DEST_DMA_BP_THRESHOLD.", __FUNCTION__);
        return;
    }
}

bool FiFPGA::ipmIsIdle(void)
{
    int     retry;
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Wait for the IPM to go idle.
    for (retry = 0; retry < 100 * MAX_RETRY_COUNT; ++retry)
    {
        if (!regRead(IPM_STATUS, &value))
        {
            fi_log(LOG_WARNING, "%s: Unable to read IPM_STATUS.", __FUNCTION__);
            return false;
        }

        // Only care about IPM_BUSY. It's OK if the histogram is busy.
        if ((value & IPMS_IPM_BUSY) == 0)
            break;

        // Wait before looking again.
        usleep(1);
    }

    TP(FT_IPM, "%s: RETRY=%d", __PRETTY_FUNCTION__, retry);

    // Return our status.
    return (retry < (100 * MAX_RETRY_COUNT));
}

/*!
 * @brief Set all the values that involve creating a LUT
 *
 * Set brighness, constrast, and gamma, plus default gamma. This can be used at startup
 * to set all of these at once then build and load the LUT one time, instead of rebuilding
 * it for each fo the three values.
 *
 * @return true for valid, otherwise false.
 */
void FiFPGA::ipmSetLutValues(uint32_t new_brightness, uint32_t new_contrast, uint32_t new_gamma,
                             uint32_t new_default_gamma)
{
    TP(FT_IPM, "%s: br=0x%08x, con=0x%08x, gamma=0x%08x, default=0x%08x", __PRETTY_FUNCTION__,
       new_brightness, new_contrast, new_gamma, new_default_gamma);

    // Sanity checks.
    if ((new_brightness > GIGE_MAX_BRIGHTNESS) ||
        (new_contrast > GIGE_MAX_CONTRAST) ||
        (new_gamma < GIGE_MIN_GAMMA) ||
        (new_gamma > GIGE_MAX_GAMMA))
    {
        fi_log(LOG_DEBUG, "%s: Got bad LUT values: b=%u c=%u g=%u", __FUNCTION__,
               new_brightness, new_contrast, new_gamma);

        if (new_brightness > GIGE_MAX_BRIGHTNESS)
            new_brightness = brightness;
        if (new_contrast > GIGE_MAX_CONTRAST)
            new_contrast = contrast;
        if (new_gamma < GIGE_MIN_GAMMA)
            new_gamma = GIGE_MIN_GAMMA;
        if (new_gamma > GIGE_MAX_GAMMA)
            new_gamma = GIGE_MAX_GAMMA;
    }

    // Save the new values.
    ipmSetDefaultGamma(new_default_gamma);
    gamma = new_gamma;
    brightness = new_brightness;
    contrast = new_contrast;

    // Rebuild the LUT and send it to the hardware.
    ipmSendLutTableToFpga(GIGE_REG_RED);
    ipmSendLutTableToFpga(GIGE_REG_GREEN);
    ipmSendLutTableToFpga(GIGE_REG_BLUE);
}

/*!
 * @brief Load a user LUT from a number file
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::ipmLoadUserLut(unsigned lut_number)
{
    PartMgr  *part_mgr = CameraServer::get_partmgr();
    bool     rc = false;
    bool     done;
    int      i;
    int      count;
    int      fd;
    uint32_t gain_channel;
    uint32_t addr;
    uint32_t value;
    uint8_t *table_ptr;
    std::unique_ptr<LUT_RGB_FILE> lut;
    std::string filename;

    TP(FT_IPM, "%s: lut=%d", __PRETTY_FUNCTION__, lut_number);

    // Allocate our working LUT.
    lut = std::unique_ptr<LUT_RGB_FILE>(new LUT_RGB_FILE);
    if (lut.get() == nullptr)
        return false;

    // Build up the filename for the LUT we're trying to load.
    filename = fmt::sprintf(LUT_FILENAME_FMT, get_home_path(), lut_number);

    // If we can't open the file, we will set up a linear LUT.
    fd = open(filename.c_str(), O_RDONLY);
    if (fd >= 0)
    {
        int lut_len;

        // Read up the header information to determine how big the LUT is.
        count = read(fd, lut.get(), sizeof(LUT_HEADER));
        if (count != sizeof(LUT_HEADER))
        {
            fi_log(LOG_WARNING, "%s: header read %s failed, got %d", __FUNCTION__, filename.c_str(), count);
            close(fd);
            goto bail;
        }

        // What type is this file?
        if (lut->hdr.lutDataType == LUT_SERIESNC_COLOR)
            lut_len = sizeof(LUT_RGB_FILE);
        else
            lut_len = sizeof(LUT_FILE);

        // We've already read the header part, so adjust accordingly.
        lut_len -= sizeof(LUT_HEADER);

        // Read the remainder of the file. In the case of a LUT_FILE structure, this will only populate
        // the redKnots elements (and LUT). This is fine, we just need to remember it.
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstringop-overflow="
        count = read(fd, &lut.get()->redKnots[0], lut_len);
#pragma GCC diagnostic pop
        if (count != lut_len)
        {
            fi_log(LOG_WARNING, "%s: read %s failed, got %d", __FUNCTION__, filename.c_str(), count);
            close(fd);
            goto bail;
        }

        // Close the file.
        close(fd);
    }
    else
    {
        // This is fine...the LUT type will be 0, which would indicate a
        // series3 LUT, which we can deal with.
        memset(lut.get(), 0, sizeof(LUT_RGB_FILE));
        for (i = 0; i < LUT_MAX_TABLE_COUNT; i++)
        {
            lut->redTable[i] = i / 4;
        }
        // First knot is (0,0) and last is (1023,255).
        lut->redKnots[1].x = 1023;
        lut->redKnots[1].y = 255;
        fi_log(LOG_WARNING, "%s: no LUT file, using linear for #%u", __FUNCTION__, lut_number);
    }

    // Now write the data to the FPGA.
    gain_channel = GIGE_REG_RED;
    done = false;
    while (!done)
    {
        // Setup for the gain channel we are working with.
        switch (gain_channel)
        {
        case GIGE_REG_RED:
            // Setup the IPM to receive the lookup table.
            regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_RED, ~ICTL_BCG_LUT_SEL_MASK);

            // Point to the red LUT table.
            table_ptr = lut->redTable;

            // Next time through, do the GREEN channel.
            gain_channel = GIGE_REG_GREEN;
            break;
        case GIGE_REG_GREEN:
            // Setup the IPM to receive the lookup table.
            regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_GREEN, ~ICTL_BCG_LUT_SEL_MASK);

            // Do we have a green LUT?
            if (lut->hdr.lutDataType == LUT_SERIESNC_COLOR)
            {
                // Point to the green LUT table.
                table_ptr = lut->greenTable;
            }
            else
            {
                // Reuse the red LUT table.
                table_ptr = lut->redTable;
            }

            // Next time through, do the BLUE channel.
            gain_channel = GIGE_REG_BLUE;
            break;
        case GIGE_REG_BLUE:
            // Setup the IPM to receive the lookup table.
            regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_BLUE, ~ICTL_BCG_LUT_SEL_MASK);

            // Do we have a green LUT?
            if (lut->hdr.lutDataType == LUT_SERIESNC_COLOR)
            {
                // Point to the blue LUT table.
                table_ptr = lut->blueTable;
            }
            else
            {
                // Reuse the red LUT table.
                table_ptr = lut->redTable;
            }

            // After we send this channel through, we're done.
            gain_channel = 0;
            done = true;
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected gain channel [0x%x]!", __PRETTY_FUNCTION__, gain_channel);
            rc = false;
            goto bail;
        }

        // Transfer the LUT table to the FPGA.
        for (i = 0; i < LUT_MAX_TABLE_COUNT; i++)
        {
            // Get the LUT entry.
            value = table_ptr[i];

            // Calculate the address for the P0 LUT and write
            // the table entry out.
            addr = IPM_P0_BCG_LUT_START + (i * 4);
            if (!regWrite(addr, value))
                break;

            // Calculate the address for the P1 LUT and write
            // the table entry out.
            addr = IPM_P1_BCG_LUT_START + (i * 4);
            if (!regWrite(addr, value))
                break;

            // Calculate the address for the P2 LUT and write
            // the table entry out.
            addr = IPM_P2_BCG_LUT_START + (i * 4);
            if (!regWrite(addr, value))
                break;
        }
    }

    // Make sure we didn't bail out of the looop.
    if (done)
    {
        rc = true;
        part_mgr->set_knots(lut.get());
    }

bail:
    return rc;
}

/*!
 * @brief Download a new LUT table to the FPGA.
 *
 * This routine uses current class settings. All the calculations
 * in this routine originated in the spreadsheet TS3OperatingModeswMacros.
 *
 * @param gain_channel  Which gain are we updating (GIGE_REG_RED, _GREEN, or _BLUE).
 *
 * @return true for valid, otherwise false.
 */
void FiFPGA::ipmSendLutTableToFpga(uint32_t gain_channel)
{
    float   c_norm;
    float   b_norm;
    float   g_norm;
    float   chl_gain;
    int i;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (!isInitialized())
        return;

    // Setup for the gain channel we are working with.
    switch (gain_channel)
    {
    case GIGE_REG_RED:
        // Get the gain in a floating point format.
        chl_gain = ((float)ipmGainRed()) / 256.0;

        // Setup the IPM to receive the lookup table.
        regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_RED, ~ICTL_BCG_LUT_SEL_MASK);
        break;
    case GIGE_REG_GREEN:
        // Get the gain in a floating point format.
        chl_gain = ((float)ipmGainGreen()) / 256.0;

        // Setup the IPM to receive the lookup table.
        regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_GREEN, ~ICTL_BCG_LUT_SEL_MASK);
        break;
    case GIGE_REG_BLUE:
        // Get the gain in a floating point format.
        chl_gain = ((float)ipmGainBlue()) / 256.0;

        // Setup the IPM to receive the lookup table.
        regModify(IPM_CONTROL, ICTL_BCG_LUT_SEL_BLUE, ~ICTL_BCG_LUT_SEL_MASK);
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected gain channel [0x%x]!", __PRETTY_FUNCTION__, gain_channel);
        return;
    }

    // Normalize the three values. This algorithm is based on
    // trial and error - thus the "magic" numbers. Although, we
    // subtract the 100 from brightness and contast because their
    // ranges are from -100 to +100. Gamma's range is 0.10 to 5.00.
    // with a step of 0.01 - thus the division with 100.0.
    c_norm = (((float)(contrast) - 100.0) * .025) + 1.0;
    b_norm = (float)(brightness) - 100.0;
    if (gamma != 0)
    {
        g_norm = 100.0 / ((float)gamma);
    }
    else
    {
        g_norm = 1.0;
        fi_log(LOG_ERR, "%s: gamma is 0!?", __PRETTY_FUNCTION__);
    }

    // Now generate the FPGA values for every pixel 0-1023.
    for (i = 0; i < 1024; i++)
    {
        double  pix_val;
        double  pix_div;
        double  pix_pow;
        double  pix_mul;
        uint32_t reg_value;
        uint32_t reg_addr;

        // This is essential mx+b. We adjust the "center" point to
        // be 512 instead of 0. So "c_norm" is our slope (i.e. 'm'),
        // adjust our 'x' to be at the middle of our range (512), and
        // "b_norm" is our 'b'.
        pix_val = ((c_norm * ((float)i - 512.0)) + b_norm) + 512.0;

        // Make sure we keep our data in-range.
        if (pix_val < 0.0)
            pix_val = 0.0;
        if (pix_val > 1023.0)
            pix_val = 1023.0;

        // Now we want to apply gamma.
        pix_div = (double)(pix_val / 1023.0);
        pix_pow = pow(pix_div, g_norm);
        pix_mul = pix_pow * 1023.0;

        // Finally, apply the channel gain.
        pix_mul *= chl_gain;

        // Convert from the floating point value to an integer,
        // rounding up.
        reg_value = (uint32_t)((pix_mul + 2.0) / 4.0);

        // Make sure we're clipped at 255 - the maximum
        // output value. This table is converting from 10-bit
        // to 8-bit.
        if (reg_value > 0xFF)
            reg_value = 0xFF;

        // Calculate the address for the P0 LUT and write
        // the table entry out.
        reg_addr = IPM_P0_BCG_LUT_START + (i * 4);
        if (!regWrite(reg_addr, reg_value))
            break;

        // Calculate the address for the P1 LUT and write
        // the table entry out.
        reg_addr = IPM_P1_BCG_LUT_START + (i * 4);
        if (!regWrite(reg_addr, reg_value))
            break;

        // Calculate the address for the P2 LUT and write
        // the table entry out.
        reg_addr = IPM_P2_BCG_LUT_START + (i * 4);
        if (!regWrite(reg_addr, reg_value))
            break;
    }
}

/*!
 * @brief Set the default gamma in the FPGA.
 *
 * @param - new default gamma setting. Valid values are from 1-250
 *
 * @return TRUE is everything is good, otherwise FALSE
 */
void FiFPGA::ipmSetDefaultGamma(uint32_t new_default_gamma)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if ((new_default_gamma < GIGE_MIN_GAMMA) || (new_default_gamma > GIGE_MAX_GAMMA))
    {
        fi_log(LOG_WARNING, "%s: new_default_gamma (%d) is bad, using GIGE_DEFAULT_GAMMA", __FUNCTION__, new_default_gamma);
        new_default_gamma = GIGE_DEFAULT_GAMMA;
    }

    // Save the new value.
    default_gamma = new_default_gamma;
}

/*!
 * @brief Set the color correction coeffients for P0 and
 *    set the local value used for P1/P2.
 *
 * These registers are part of the 3x3 color correction matrix.
 * The format is 4.12 (2's complement format with upper 4 bits
 * integer, lower 12 bits fractional ).
 *
 * @param value The new coefficient value.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::ipmSetMatrixCoeff1Red(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff1_red = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF1_RED, matrix_coeff1_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF1_RED, matrix_coeff1_red);
    return regWrite(IPM_P2_CC_MATRIX_COEFF1_RED, matrix_coeff1_red);
}
bool FiFPGA::ipmSetMatrixCoeff2Red(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff2_red = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF2_RED, matrix_coeff2_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_RED, matrix_coeff2_red);
    return regWrite(IPM_P2_CC_MATRIX_COEFF2_RED, matrix_coeff2_red);
}
bool FiFPGA::ipmSetMatrixCoeff3Red(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff3_red = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF3_RED, matrix_coeff3_red);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_RED, matrix_coeff3_red);
    return regWrite(IPM_P2_CC_MATRIX_COEFF3_RED, matrix_coeff3_red);
}
bool FiFPGA::ipmSetMatrixCoeff1Green(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff1_grn = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF1_GRN, matrix_coeff1_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF1_GRN, matrix_coeff1_grn);
    return regWrite(IPM_P2_CC_MATRIX_COEFF1_GRN, matrix_coeff1_grn);
}
bool FiFPGA::ipmSetMatrixCoeff2Green(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff2_grn = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF2_GRN, matrix_coeff2_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_GRN, matrix_coeff2_grn);
    return regWrite(IPM_P2_CC_MATRIX_COEFF2_GRN, matrix_coeff2_grn);
}
bool FiFPGA::ipmSetMatrixCoeff3Green(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff3_grn = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF3_GRN, matrix_coeff3_grn);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_GRN, matrix_coeff3_grn);
    return regWrite(IPM_P2_CC_MATRIX_COEFF3_GRN, matrix_coeff3_grn);
}
bool FiFPGA::ipmSetMatrixCoeff1Blue(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff1_blu = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF1_BLU, matrix_coeff1_blu);
    regWrite(IPM_P1_CC_MATRIX_COEFF1_BLU, matrix_coeff1_blu);
    return regWrite(IPM_P2_CC_MATRIX_COEFF1_BLU, matrix_coeff1_blu);
}
bool FiFPGA::ipmSetMatrixCoeff2Blue(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff2_blu = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF2_BLU, matrix_coeff2_blu);
    regWrite(IPM_P1_CC_MATRIX_COEFF2_BLU, matrix_coeff2_blu);
    return regWrite(IPM_P2_CC_MATRIX_COEFF2_BLU, matrix_coeff2_blu);
}
bool FiFPGA::ipmSetMatrixCoeff3Blue(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    matrix_coeff3_blu = value & MATRIX_COEFF_MASK;
    regWrite(IPM_P0_CC_MATRIX_COEFF3_BLU, matrix_coeff3_blu);
    regWrite(IPM_P1_CC_MATRIX_COEFF3_BLU, matrix_coeff3_blu);
    return regWrite(IPM_P2_CC_MATRIX_COEFF3_BLU, matrix_coeff3_blu);
}

/*!
 * @brief Set the values for white balance (bayer gain) for each color.
 *
 * These are 4.12 values
 *
 * @param value The new white balance gain value in 4.12 format.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::ipmSetWbRed(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    wb_red = value & WB_MASK;
    regWrite(IPM_P0_WB_GAIN_FACTOR_RED, wb_red);
    regWrite(IPM_P1_WB_GAIN_FACTOR_RED, wb_red);
    return regWrite(IPM_P2_WB_GAIN_FACTOR_RED, wb_red);
}
bool FiFPGA::ipmSetWbGreen(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    wb_green = value & WB_MASK;
    regWrite(IPM_P0_WB_GAIN_FACTOR_GRN, wb_green);
    regWrite(IPM_P1_WB_GAIN_FACTOR_GRN, wb_green);
    return regWrite(IPM_P2_WB_GAIN_FACTOR_GRN, wb_green);
}
bool FiFPGA::ipmSetWbBlue(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    wb_blue = value & WB_MASK;
    regWrite(IPM_P0_WB_GAIN_FACTOR_BLU, wb_blue);
    regWrite(IPM_P1_WB_GAIN_FACTOR_BLU, wb_blue);
    return regWrite(IPM_P2_WB_GAIN_FACTOR_BLU, wb_blue);
}

/*!
 * @brief Set the values for DC-offset for each color.
 *
 * These are 4.12 values.
 *
 * @param value The new dc-offset value in 4.12 format.
 *
 * @return true for valid, otherwise false.
 */
bool FiFPGA::ipmSetDcOffsetRed(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    dc_offset_red = value & DC_OFFSET_MASK;
    regWrite(IPM_P0_CCM_RED_DC_OFFSET, dc_offset_red);
    regWrite(IPM_P1_CCM_RED_DC_OFFSET, dc_offset_red);
    return regWrite(IPM_P2_CCM_RED_DC_OFFSET, dc_offset_red);
}
bool FiFPGA::ipmSetDcOffsetGreen(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    dc_offset_grn = value & DC_OFFSET_MASK;
    regWrite(IPM_P0_CCM_GRN_DC_OFFSET, dc_offset_grn);
    regWrite(IPM_P1_CCM_GRN_DC_OFFSET, dc_offset_grn);
    return regWrite(IPM_P2_CCM_GRN_DC_OFFSET, dc_offset_grn);
}
bool FiFPGA::ipmSetDcOffsetBlue(uint32_t value)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Only use the lower 16 bits which is all the FPGA allows.
    dc_offset_blu = value & DC_OFFSET_MASK;
    regWrite(IPM_P0_CCM_BLU_DC_OFFSET, dc_offset_blu);
    regWrite(IPM_P1_CCM_BLU_DC_OFFSET, dc_offset_blu);
    return regWrite(IPM_P2_CCM_BLU_DC_OFFSET, dc_offset_blu);
}


/*!
 * @brief Set the X ROI to use for the RGB histogram.
 *
 * @param value The lower 2 bytes contain the size, upper 2 bytes the offset in the x direction.
 *
 * @return GEV_STATUS_SUCCESS on success, an error code otherwise.
 */
int FiFPGA::ipmRgbHistoSetRoiX(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t value)
{
    int rc;
    FiFPGA  *fpga_ptr;

    (void)p_msg;

    // Get our class pointer.
    fpga_ptr = CameraServer::get_fpga();

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return GEV_STATUS_INVALID_PARAMETER;

    // Pass the information on to the FPGA.
    if (fpga_ptr->regWrite(IPM_HISTO_ROI_X_CONFIG, value))
        rc = GEV_STATUS_SUCCESS;
    else
        rc = GEV_STATUS_INVALID_PARAMETER;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    regs[ GIGE_REG_HISTO_RGB_ROI_X ] = value;

    return rc;
}

/*!
 * @brief Set the Y ROI to use for the RGB histogram.
 *
 * @param value The lower 2 bytes contain the size, upper 2 bytes the offset in the y direction.
 *
 * @return GEV_STATUS_SUCCESS on success, an error code otherwise.
 */
int FiFPGA::ipmRgbHistoSetRoiY(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t value)
{
    int rc;
    FiFPGA  *fpga_ptr;

    (void)p_msg;

    // Get our class pointer.
    fpga_ptr = CameraServer::get_fpga();

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return GEV_STATUS_INVALID_PARAMETER;

    // Pass the information on to the FPGA.
    if (fpga_ptr->regWrite(IPM_HISTO_ROI_Y_CONFIG, value))
        rc = GEV_STATUS_SUCCESS;
    else
        rc = GEV_STATUS_INVALID_PARAMETER;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    regs[ GIGE_REG_HISTO_RGB_ROI_Y ] = value;

    return rc;
}

/*!
 * @brief Helper routine to check if the FPGA's IPM is busy
 *    creating an RGB histogram.
 *
 * @return true if it's idle, false otherwise.
 */
bool FiFPGA::ipmIsRgbHistoIdle(void)
{
    bool hist_in_process = false;
    uint32_t value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA has been initialized.
    if (!isInitialized())
        return false;

    //  make sure a histogram isn't already in process
    for (int i = 0; i < IPM_BUSY_ATTEMPTS; ++i)
    {
        // Read the FPGA's IPM status to determine if it's busy
        // creating an RGB histogram.
        regRead(IPM_STATUS, &value);
        if ((value & (IPMS_HISTO_BUSY | IPMS_HISTO_INIT_BUSY)) != 0)
        {
            // Histogram engine is busy.
            hist_in_process = true;
        }
        else
        {
            // All we need is one time when the histogram goes
            // "not busy" and we are good to go.
            hist_in_process = false;
            break;
        }

        // Wait for the histogram to process.
        usleep(IPM_BUSY_WAIT);
    }

    // If the histogram is in process, then we are NOT idle.
    return !hist_in_process;
}

void FiFPGA::ipmReset()
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Tell the FPGA to reset.
    regWrite(IPM_CONTROL, ICTL_IPM_SW_RESET);
    usleep(IPM_BUSY_WAIT);
    regWrite(IPM_CONTROL, 0);

    // Reset the P1 interface.
    ipm_proc_reset();
}

void FiFPGA::ipmRgbHistoReset(void)
{
    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Tell the FPGA to reset the histogram engine.
    regModify(IPM_CONTROL, ICTL_RGB_HISTO_RESET, -1);
    usleep(IPM_BUSY_WAIT);
    regModify(IPM_CONTROL, 0, ~ICTL_RGB_HISTO_RESET);
}

/*!
 * @brief Generate a RGB histogram.
 *
 * @param channel   One of the IPM channels on the FPGA.
 *
 * @return true if success, false otherwise
 */
bool FiFPGA::ipmRgbHistoGenerate(uint8_t channel)
{
    uint32_t value = 0;

    TP(FT_IPM, "%s: channel=%d", __PRETTY_FUNCTION__, channel);

    // make sure the FPGA has been initialized
    if (!isInitialized())
        return false;
    if (channel > GIGE_GENERATE_RGB_HISTO_CHAN_P2)
        return false;

    // The NucCam only has P1, so force the channel.
    channel = GIGE_GENERATE_RGB_HISTO_CHAN_P1;

    // Make sure a histogram isn't already in process.
    if (!ipmIsRgbHistoIdle())
    {
        // Try to reset the histogram engine and see if that works.
        ipmRgbHistoReset();
        if (!ipmIsRgbHistoIdle())
        {
            fi_log(LOG_WARNING, "%s: histogram generation seems to be stuck.", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // Tell the FPGA to capture a histogram with the correct channel.
    if (!regRead(IPM_CONTROL, &value))
        return false;
    value |= ICTL_HISTO_START;
    value &= ~(ICTL_HISTO_PORT_MASK);
    value |= ((channel - 1) << ICTL_HISTO_PORT_SHIFT);

    return regWrite(IPM_CONTROL, value);
}

/*!
 * @brief Get the histogram peak values for an already run RGB histogram.
 *
 * @return The peak values: 0x0RGB.
 */
uint32_t FiFPGA::ipmRgbHistoPeak(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_PEAK_VALUE, &value))
        return value = 0;

    // Return the values.
    return value;
}

/*!
 * @brief Return the sum of all red pixel values within the
 *    RGB Histogram ROI.
 */
uint32_t FiFPGA::ipmRgbHistoSumX_Red(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_RED, &value))
        return value = 0;

    // Return the values.
    return value;
}

/*!
 * @brief Return the sum of all green pixel values within the
 *    RGB Histogram ROI.
 */
uint32_t FiFPGA::ipmRgbHistoSumX_Green(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_GRN, &value))
        return value = 0;

    // Return the values.
    return value;
}

/*!
 * @brief Return the sum of all blue pixel values within the
 *    RGB Histogram ROI.
 */
uint32_t FiFPGA::ipmRgbHistoSumX_Blue(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_GRN, &value))
        return value = 0;

    // Return the values.
    return value;
}

/*!
 * @brief Get the upper 5 bits (bits 36:32) of the sum of all
 *    of the red pixels squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_RedH(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_RED_H, &value))
        return value = 0;

    // Return the values.
    return (value & FPGA_HIST_SUM_SQUARED_H_MASK);
}

/*!
 * @brief Get the upper 5 bits (bits 36:32) of the sum of all
 *    of the green pixels squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_GreenH(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_GRN_H, &value))
        return value = 0;

    // Return the values.
    return (value & FPGA_HIST_SUM_SQUARED_H_MASK);
}

/*!
 * @brief Get the upper 5 bits (bits 36:32) of the sum of all
 *    of the blue pixels squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_BlueH(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_BLU_H, &value))
        return value = 0;

    // Return the values.
    return (value & FPGA_HIST_SUM_SQUARED_H_MASK);
}

/*!
 * @brief Get the lower 32 bits of the sum of all of the green pixels
 *    squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_GreenL(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_GRN_L, &value))
        return value = 0;

    // Return the value.
    return value;
}


/*!
 * @brief Get the lower 32 bits of the sum of all of the blue pixels
 *    squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_BlueL(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_BLU_L, &value))
        return value = 0;

    // Return the value.
    return value;
}

/*!
 * @brief Get the lower 32 bits of the sum of all of the red pixels
 *    squared within the already run RGB histogram.
 */
uint32_t FiFPGA::ipmRgbHistoSumXSqrd_RedL(void)
{
    uint32_t    value;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (!isInitialized())
        return false;

    // Read the values.
    if (!regRead(IPM_HISTO_SUM_OF_X_SQRD_RED_L, &value))
        return value = 0;

    // Return the value.
    return value;
}

/*!
 * @brief Initialize the IPM P0 configuration for live video from passed settings.
 */
bool FiFPGA::ipmLiveStreamInit(
    uint16_t    true_startx,
    uint16_t    true_starty,
    uint16_t    width,
    uint16_t    height,
    bitmode_t   sensor_bit_mode,
    fpnmode_t   fpn_mode)
{
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    uint32_t        value;
    uint32_t        config;

    TP(FT_IPM, "%s", __PRETTY_FUNCTION__);

    // Set up IPM configuration.
    config = 0;

    // Are we supposed to use FPGA FPN?
    if ((fpn_mode & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
    {
        // We want to preserve the FPN control setting. Clear out everything else.
        // We need to adjust the mask based on what bit values we're using.
        regRead(IPM_P1_CONFIG, &config);
        config &= (IPMC_FPN_ENA | IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7);
    }

    // CSI always wants to be RGB24 (even if sensor is mono).
    config |= IPMC_DEST_TYPE_RGB24;

    // Let the pipeline know if we are color or mono.
    if (sensor_ptr->is_sensor_color())
        config |= IPMC_SRC_FRM_COLOR_MODE;

    // Set up IPM configuration.
    config |= get_ipm_src_bits(sensor_bit_mode);

    config |= sensor_ptr->IPM_bayer_pattern_start(true_startx, true_starty);

    // Write out the IPM configuration.
    fi_log(LOG_DEBUG, "%s: IPM_P1_CONFIG=0x%08x", __PRETTY_FUNCTION__, config);
    if (!regWrite(IPM_P1_CONFIG, config))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_P1_CONFIG!", __FUNCTION__);
        return false;
    }

    // Calculate the frame size, taking into account width, height, and bit depth.
    value = cs_ptr->calc_frame_buf_size(width, height, sensor_bit_mode);
    value = convert_to_HEXWORD(value);
    fi_log(LOG_DEBUG, "%s: IPM_P1_SRC_DMA_BLOCK_SIZE = 0x%08x", __PRETTY_FUNCTION__, value);
    if (!regWrite(IPM_P1_SRC_DMA_BLOCK_SIZE, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_P1_SRC_DMA_BLOCK_SIZE!", __FUNCTION__);
        return false;
    }

    // Set the frame size which is width & height.
    value = width | ((uint32_t)height << IPM_FRAME_DIM_Y_SHIFT);
    if (!regWrite(IPM_P1_FRAME_DIM, value))
    {
        fi_log(LOG_WARNING, "%s: Unable to write IPM_P1_FRAME_DIM!", __FUNCTION__);
        return false;
    }

    // It's all good.
    return true;
}

/*!
 * @brief Process a single image through the P1 interface.
 *
 * @param bit_mode  Bit mode that the data was recorded at.
 * @param width     Width that the data was recorded at.
 * @param height    Height that the data was recorded at.
 * @param src_addr  Address in the capture buffer to take the data from.
 * @param dest_addr Address in the ping pong buffer to store the processed image.
 * @param image_type    Image processing information needed for IPM setup, IMGPROC_... bits
 *
 * @return  size of processed image for success or 0 for failure.
 */
uint32_t FiFPGA::startP1Frame(bitmode_t bit_mode, uint32_t width, uint32_t height, uint32_t src_addr, uint32_t dest_addr, uint32_t image_type)
{
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    CameraServer    *cs_ptr = CameraServer::get_cs();

    int     retry;
    uint32_t    value;
    uint32_t    config;
    uint16_t    true_startx;
    uint16_t    true_starty;
    uint32_t    last_p1_frame_size;

    TP(FT_IPM, "%s: bit_mode=%d, width=%d, height=%d, image_type=0x%08x", __PRETTY_FUNCTION__, bit_mode, width, height, image_type);

    // If the license roi is smaller than the maximum that the sensor supports then we
    // need to know where the first pixel is so we can determine where the Bayer pattern starts.
    // We also need this to set up FPGA Pixel FPN corrections.
    cs_ptr->get_sensor_start(&true_startx, &true_starty);

    // Wait for the P1 frame processing to be idle.
    for (retry = 0; retry < IPM_BUSY_ATTEMPTS; ++retry)
    {
        if (!regRead(IPM_P1_FRAME_REQ_STATUS, &value))
            return 0;
        if ((value & IPM_P1_FRAME_BUSY) != IPM_P1_FRAME_BUSY)
            break;

        // Wait for a little bit before we look again.
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(IPM_BUSY_WAIT));
    }
    if ((value & IPM_P1_FRAME_BUSY) == IPM_P1_FRAME_BUSY)
    {
        fi_log(LOG_ERR, "%s: IPM P1 appears to be stuck!", __FUNCTION__);
        return 0;
    }

    // Set the P1 source pointer.
    regWrite(IPM_P1_FRAME_REQ_SRC_PTR, src_addr);

    // Set the destination buffer address.
    if ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
    {
        uint32_t image_size_hw;

        // How many hexwords does the image take. That's the size of
        // the Y buffer.
        image_size_hw = convert_to_HEXWORD(width * height);

        // For NV12 images, we need to specify the Y and C buffers
        regWrite(IPM_P1_FRAME_REQ_IC_Y_BUF_START, dest_addr);
        regWrite(IPM_P1_FRAME_REQ_IC_C_BUF_START, dest_addr + image_size_hw);
    }
    else
    {
        // We just have the one buffer pointer to worry about.
        regWrite(IPM_P1_FRAME_REQ_DEST_PTR, dest_addr);
    }

    // Initialize the DMA burst size and backpressure.
    regWrite(IPM_IC_DMA_BURST_SIZE, IPM_DEST_DMA_REQ_SIZE_VAL);
    regWrite(IPM_IC_DMA_BP_THRESHOLD, IPM_DEST_DMA_BP_THRESHOLD_VAL);

    // Initialize our IPM port configuration register value.
    config = get_ipm_src_bits(bit_mode);

    // Set other flags which might have been passed in.
    if ((image_type & IMGPROC_USE_FPN) != 0)
    {
        // Make sure there actually is a correction table.
        regRead(IPM_P1_CT_DMA_BLOCK_SIZE, &value);
        if (value != 0)
        {
            uint32_t gt_addr;
            PartMgr *part_mgr = CameraServer::get_partmgr();

            // Enable FPN processing.
            config |= IPMC_FPN_ENA;

            // Get the gain table address.
            regRead(IPM_P1_GT_DMA_START_ADDR, &gt_addr);

            // Do we want to turn on PRNU?
            if (FiImageCorrection::prnuIsAvailable() && part_mgr->isGainTableAllowed() && (gt_addr != 0))
            {
                // Enable PRNU.
                config |= (IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7);
            }
        }
    }

    // Or in the IPMC_DEST_TYPE_ bits.
    config |= get_ipm_dest_bits(image_type & IMGPROC_TYPE_MASK);

    // Determine if we might be color.
    if (((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_JPEG) ||
        ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12) ||
        ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_RGB24))
    {
        // For 24-bit RGB, we need to tell the IPM if we're color
        // or mono (yes...some monos get processed into 24-bit RGB).
        if (sensor_ptr->is_sensor_color())
        {
            config |= IPMC_SRC_FRM_COLOR_MODE;
            TP(FT_IPM, "%s: sensor is color, config=0x%08x", __PRETTY_FUNCTION__, config);
        }
    }

    // The IPM always does row word align.
    last_p1_frame_size = get_processed_size(image_type | IMGPROC_ROW_WD_ALIGN, width, height);
    if (last_p1_frame_size == 0)
    {
        // Invalid image type.
        fi_log(LOG_DEBUG, "%s: last_p1_frame_size == 0!", __PRETTY_FUNCTION__);
        return 0;
    }
    TP(FT_IPM, "%s: last_p1_frame_size=0x%08x", __PRETTY_FUNCTION__, last_p1_frame_size);

    // Set up how the source pixels should be interpreted as far as color
    // based upon the x and y offset. This is proforma since we always use
    // even offsets so it doesn't change with the start position.
    config |= sensor_ptr->IPM_bayer_pattern_start(true_startx, true_starty);

    // Write the value out to the P1 configuration register.
    TP(FT_IPM, "%s: IPM_P1_CONFIG=0x%08x", __PRETTY_FUNCTION__, config);
    if (!regWrite(IPM_P1_CONFIG, config))
    {
        fi_log(LOG_ERR, "%s: Unable to write IPM_P1_CONFIG!", __FUNCTION__);
        return 0;
    }

    // Set the block size which is width x height in hexword
    value = calc_src_dma_size(width, height, bit_mode);
    value = convert_to_HEXWORD((uint64_t)value);
    TP(FT_IPM, "%s: IPM_P1_SRC_DMA_BLOCK_SIZE = 0x%08x", __PRETTY_FUNCTION__, value);
    if (!regWrite(IPM_P1_SRC_DMA_BLOCK_SIZE, value))
    {
        fi_log(LOG_ERR, "%s: Unable to write IPM_P1_SRC_DMA_BLOCK_SIZE!", __FUNCTION__);
        return 0;
    }

    // Set the frame size which is width & height
    value = width | (height << IPM_FRAME_DIM_Y_SHIFT);
    if (!regWrite(IPM_P1_FRAME_DIM, value))
    {
        fi_log(LOG_ERR, "%s: Unable to write IPM_P1_FRAME_DIM!", __FUNCTION__);
        return 0;
    }

    // Set up FPGA bad pixel corrections for IPM port 1. This might do nothing since
    // a sensor can be set up to use its own internal bad pixel correction.
    sensor_ptr->IPM_bad_pixel_correction(1, true_startx, true_starty, width, height);

    // Tell the P1 to fire off.
    value = IPM_P1_FRAME_REQUEST;
    if ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
    {
        value |= IPM_P1_IC_ENABLE;
    }
    else
    {
        value |= IPM_P1_DMA_ENABLE;
    }
    regWrite(IPM_P1_FRAME_REQ_CONTROL, value);

    TP(FT_IPM, "%s: %ux%u %u bytes type=0x%x p1req=0x%x addrs:%x->%x", __FUNCTION__,
       width, height, last_p1_frame_size, image_type, value, src_addr, dest_addr);

    return last_p1_frame_size;
}

/*!
 * @brief Process a single "live" frame through the IPM.
 *
 * The important point here is that this is a "live" frame vs. a frame that is part
 * of a captured partition. The fact that it is a "live" frame defines where we
 * go to get the image settings.
 *
 * For processing of a "captured" frame, see ipmFrameProcess().
 *
 * @param image_type    [IN] The image processing information, reference the @ref IMGPROC_TYPES "IMGPROC_xxx" constants.
 * @param src_addr      [IN] Address [HEXWORD] in image memory of frame to process.
 * @param dest_addr     [IN] Address [HEXWORD] in image memory to put the processed data.
 *
 * @return The size of processed image for success or 0 for failure.
 */
uint32_t FiFPGA::ipmLiveFrameProcess(uint32_t image_type, uint32_t src_addr, uint32_t dest_addr)
{
    CameraServer *cs_ptr = CameraServer::get_cs();

    uint16_t start_x;
    uint16_t start_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;

    TP(FT_IPM, "%s: image_type=%d, src_addr=0x%08x, dest_addr=0x%08x", __PRETTY_FUNCTION__, image_type, src_addr, dest_addr);

    // Get our current settings.
    cs_ptr->get_user_settings(&start_x, &start_y, &width, &height, &bit_mode, &frame_rate,
                              &shutter_speed, &sensor_opts);

    // Process the information.
    return startP1Frame(bit_mode, width, height, src_addr, dest_addr, image_type);
}

/*!
 * @brief Process a single "captured" frame through the IPM.
 *
 * The important point here is that this is a "captured" frame vs. a "live" frame.
 * The fact that it is a "captured" frame defines where we go to get the image settings.
 *
 * For processing of a "live" frame, see ipmLiveFrameProcess().
 *
 * @param image_type    [IN] The image processing information, reference the @ref IMGPROC_TYPES "IMGPROC_xxx" constants.
 * @param src_addr      [IN] Address [HEXWORD] in image memory of frame to process.
 * @param dest_addr     [IN] Address [HEXWORD] in image memory to put the processed data.
 *
 * @return The size of processed image for success or 0 for failure.
 */
uint32_t FiFPGA::ipmFrameProcess(uint32_t image_type, uint32_t src_addr, uint32_t dest_addr)
{
    PartMgr *part_mgr = CameraServer::get_partmgr();

    uint32_t    height;
    uint32_t    width;
    bitmode_t   bit_mode;
    std::unique_ptr<const pm_metadata> meta_ptr;

    TP(FT_IPM, "%s: type=0x%x, src=0x%08x, dst=0x%08x", __PRETTY_FUNCTION__, image_type, src_addr, dest_addr);

    // Do not allow this if the camera is in the wrong state.
    if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_IDLE)
    {
        fi_log(LOG_ERR, "%s: Wrong camera state!", __FUNCTION__);
        return 0;
    }

    // Get a pointer to the per-capture metadata.
    meta_ptr = part_mgr->getReviewPartMetadata();
    if (meta_ptr.get() == nullptr)
    {
        // OK, no review partition. Go for the active partition metadata.
        fi_log(LOG_DEBUG, "%s: getReviewPartMetadata() failed, try using the active partition", __PRETTY_FUNCTION__);
        meta_ptr = part_mgr->getActivePartMetadata();
    }
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No meta_ptr!?", __PRETTY_FUNCTION__);
        return 0;
    }

    // Now get the information we need about the frame.
    bit_mode = (bitmode_t)meta_ptr->bit_sel;
    width = meta_ptr->roi_width;
    height = meta_ptr->roi_height;

    // Process the information.
    return startP1Frame(bit_mode, width, height, src_addr, dest_addr, image_type);
}

/*!
 * @brief Update the correction table information for the IPM.
 *
 * @param frame_size    Number of bytes in the correction table.
 * @param ct_addr       Where in image memory the table is located.
 *
 * @return 0 on success, <0 on failure.
 */
int FiFPGA::ipmCtUpdate(uint32_t frame_size, uint64_t ct_addr, uint64_t gt_addr)
{
    uint32_t    value;

    TP(FT_IPM, "%s: size=0x%08x, ct_addr=0x%" PRIx64 ", gt_addr=0x%" PRIx64, __PRETTY_FUNCTION__, frame_size, ct_addr, gt_addr);

    // Copy the IPM DMA register for the correction table DMAs.
    regRead(IPM_SRC_DMA_BURST_SIZE, &value);
    regWrite(IPM_CT_DMA_BURST_SIZE, value);

    regRead(IPM_SRC_DMA_BP_THRESHOLD, &value);
    regWrite(IPM_CT_DMA_BP_THRESHOLD, value);

    // Tell the FPGA where to find the correction table.
    regWrite(IPM_P0_CT_DMA_START_ADDR, convert_to_HEXWORD(ct_addr));
    regWrite(IPM_P1_CT_DMA_START_ADDR, convert_to_HEXWORD(ct_addr));
    regWrite(IPM_P2_CT_DMA_START_ADDR, convert_to_HEXWORD(ct_addr));

    // Tell the FPGA where to find the gain table.
    regWrite(IPM_P0_GT_DMA_START_ADDR, convert_to_HEXWORD(gt_addr));
    regWrite(IPM_P1_GT_DMA_START_ADDR, convert_to_HEXWORD(gt_addr));
    regWrite(IPM_P2_GT_DMA_START_ADDR, convert_to_HEXWORD(gt_addr));

    // Tell the FPGA how big the correction table is. Size is in HEXWORDS, so
    // we need to adjust accordingly.
    value = convert_to_HEXWORD(frame_size);

    regWrite(IPM_P0_CT_DMA_BLOCK_SIZE, value);
    regWrite(IPM_P1_CT_DMA_BLOCK_SIZE, value);
    regWrite(IPM_P2_CT_DMA_BLOCK_SIZE, value);

    if (gt_addr == 0)
        value = 0;
    regWrite(IPM_P0_GT_DMA_BLOCK_SIZE, value);
    regWrite(IPM_P1_GT_DMA_BLOCK_SIZE, value);
    regWrite(IPM_P2_GT_DMA_BLOCK_SIZE, value);

    // All is well.
    return 0;
}

