/*!
 * @file fpga_ltr.cxx
 * @brief This source file implements the Fastec Imaging FPGA
 *    class, FiFPGA, LTR methods.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <time.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/rtc.h>

#include <chrono>
#include <vector>

#include "camera_server.h"
#include "cs_utils.h"
#include "fpga.h"
#include "fpga_io.h"
#include "fcp.h"
#include "ltr_worker.h"
#include "part_mgr.h"
#include "threads.h"

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

namespace FI_FPGA_LTR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The maximum number of times we'll spin in a loop waiting.
const int MAX_RETRY_COUNT = 100;

//! The amount of delay waiting for the LTR to finish.
const uint32_t LTR_QUERY_DELAY_MSEC = 50;

} /* End of namespace FI_FPGA_IPM */

// Make life easier on ourselves.
using namespace FI_FPGA;
using namespace FI_FPGA_LTR;

/*!
 * @brief Set the current LTR mode.
 *
 * @note This controls whether we will use LTR mode the next time we start
 *   a capture. It must be "allowed" by our license and we must not be
 *   capturing already, but that is controlled by policy at the fcp
 *   level when the LTR_MODE register is set.
 *
 * @param do_select True to select, false to deselect.
 */
void FiFPGA::ltrModeSelect(bool do_ltr)
{
    TP(FT_FUNC, "%s: do_ltr=%d", __PRETTY_FUNCTION__, do_ltr);
    ltr_mode_selected = do_ltr;
}

/*!
 * @brief Return the LTR mode information.
 */
uint32_t FiFPGA::ltrMode(void)
{
    uint32_t ret_value = 0;

    TP(FT_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // This reports whether LTR is the currently-selected mode of operation.
    // It exactly corresponds to what we report in the GigE LTR_MODE register.
    // We also maintain the state of SATA-LTR being enabled during capture,
    // which turns on and off as we arm, de-arm, or complete capture.
    if (ltr_mode_selected)
        ret_value = GIGE_LTR_MODE_ENABLE;

    // Return our value.
    return ret_value;
}

void FiFPGA::ltrComplete()
{
    int i;

    TP(FT_FUNC, "%s", __PRETTY_FUNCTION__);

    if (ltr_thread.get() == nullptr)
        return;

    TP(FT_FUNC, "%s: waiting", __PRETTY_FUNCTION__);
    for (i = 0; i < MAX_RETRY_COUNT; ++i)
    {
        // Did our bit go low?
        if (!ltr_thread->is_running())
            break;

        // Wait for a bit.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(LTR_QUERY_DELAY_MSEC));
    }

    TP(FT_FUNC, "%s: Done after %d tries.", __PRETTY_FUNCTION__, i);
}

/*!
 * @brief Enable/disable LTR mode capture.
 *
 * @note This controls the FPGA hardware LTR mode enable and disable, which is
 *       distinct from whether LTR mode is available as a licensed option or
 *       whether the user has chosen the option of using LTR mode when capturing.
 * @param do_enable True to enable, false to disable.
 */
void FiFPGA::ltrEnable(bool do_enable)
{
    TP(FT_FUNC, "%s: do_enable=%d", __PRETTY_FUNCTION__, do_enable);

    if (do_enable)
    {
        int i;
        int rc;

        // Do we already have a thread going?
        if (ltr_thread.get() != nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Make sure to close out the old thread.", __PRETTY_FUNCTION__);
            ltrComplete();
        }

        LOG("%s: Starting an LTR worker thread.", __PRETTY_FUNCTION__);
        ltr_thread = std::unique_ptr<ltr_worker>(new ltr_worker());
        ltr_thread->start();

        for (i = 0; i < MAX_RETRY_COUNT; ++i)
        {
            // Wait for a bit.
            rc = ltr_thread->wait_for_ready(50);
            if (rc == 0)
                break;
        }

        // Did our bit go low?
        LOG("%s: thread is%s running (rc=%d)", __PRETTY_FUNCTION__, (ltr_thread->is_running() ? "" : " NOT"), rc);
    }
    else
    {
        // Do we have our LTR worker thread going?
        if (ltr_thread.get() != nullptr)
        {
            // Tell the thread to wrap it up.
            ltr_thread->stop();

            // Wait for it to shutdown.
            ltrComplete();

            {
                // ==============================================================================
                uint32_t latched_count;

                // Read the latched frame counter of the DMA.
                regRead(IM2SSD_DMA0_LR_WR_FRAME_COUNT_LATCHED, &latched_count);
                LOG("%s: Latched DMA count=%u", __PRETTY_FUNCTION__, latched_count);
            }   // ==============================================================================

            // Shut down the IM2SSD DMA.
            LOG("%s: Turn off IM2SSD DMA", __PRETTY_FUNCTION__);
            regWrite(IM2SSD_DMA0_CONTROL, 0);
        }
    }
}

/*!
 * @brief Enable LTR ROC/BROC mode capture.
 *
 * @note This controls the FPGA hardware LTR mode enable when arming in ROC/BROC mode.
 *
 * @param mode  Mode bits from GIGE_REG_LTR_MODE.
 */
void FiFPGA::ltrEnableROC(uint32_t mode)
{
    uint32_t    reg_value;
    uint32_t    buf_count;
    PartMgr *pm_ptr;

    TP(FT_FUNC, "%s: mode=0x%08x", __PRETTY_FUNCTION__, mode);

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Get our frame size, in QuadWords.
    regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(pm_ptr->getFrameSize()));

    // Set up the BROC registers.
    if ((mode & GIGE_LTR_MODE_BROC) != 0)
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        reg_value = regs[ GIGE_REG_BROC_LEN ].get_value();
        if (reg_value < GIGE_MIN_BROC_LEN)
        {
            fi_log(LOG_WARNING, "%s: Found BROC_LEN=%d, changing to %d", __FUNCTION__,
                   reg_value, GIGE_MIN_BROC_LEN);
            reg_value = GIGE_MIN_BROC_LEN;
            regs[ GIGE_REG_BROC_LEN ].set_value(reg_value);
        }
        regWrite(ICM_BROC_BURST_FRM_CNT, reg_value);

        //  reg_value = regs[ GIGE_REG_BROC_DELAY ].get_value();
        // FIXME didn't implement the delay yet and probably won't until we need it.
        reg_value = 0;
        regWrite(ICM_BROC_BURST_START_DELAY_CNT, reg_value);
    }

    // How many frames can we fit in what's left of the LTR capture region?
    buf_count = pm_ptr->ltrGetRocFrameCount();
    regWrite(ICM_ROC_MAX_CAPTURE_FRM_CNT, buf_count);

    // Adjust the ICM to capture onto our buffer.
    reg_value = pm_ptr->ltrGetCaptureBufferAddr(&buf_count);
    regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
    icmSetCaptureBuffer(reg_value);

    // Enable the LTR worker thread.
    ltrEnable(true);
}

void FiFPGA::ltrFasFireDisable()
{
    // Make sure we aren't armed.
    if (icmIsArmed())
    {
        fi_log(LOG_DEBUG, "%s: Shouldn't do this when armed!", __PRETTY_FUNCTION__);
        return;
    }

    // Turn off all FasFireLR settings.
    regModify(ICM_CONTROL, 0, ~(ICMC_CAPTURE_MODE_FFLR | ICMC_CAPTURE_MODE_FFLR_ROC | ICMC_CAPTURE_MODE_FFLR_BROC));
}

bool FiFPGA::ltrFasFireEnabled()
{
    bool rc(false);
    uint32_t reg_value;

    // Read the ICM register which has the settings we're interested in.
    if (regRead(ICM_CONTROL, &reg_value))
    {
        // Is FasFireLR mode set?
        switch (reg_value & ICMC_CAPTURE_MODE_MASK)
        {
        case ICMC_CAPTURE_MODE_FFLR:
        case ICMC_CAPTURE_MODE_FFLR_ROC:
        case ICMC_CAPTURE_MODE_FFLR_BROC:
            rc = true;
            break;
        default:
            break;
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Calculate the maximum LTR frame rate for a given width, height and bit mode.
 *
 * @param width     The width we're interested in.
 * @param height    The height we're interested in.
 * @param bit_mode  The bit mode (upper-8, lower-8, or 10-bit).
 *
 * @return The maximum frame rate we can support.
 */
uint32_t FiFPGA::ltrCalcFrameRate(uint32_t width, uint32_t height, bitmode_t bit_mode)
{
    uint32_t    frame_size;
    uint32_t    ltr_frame_rate;
    uint32_t    ssd_bandwidth;

    TP(FT_FUNC, "%s: width=%u, height=%u, bitmode=%d", __PRETTY_FUNCTION__, width, height, (int)bit_mode);

    // Use the common function to calculate frame size in bytes. The actual number
    // of bytes consumed is rounded up to a hexword boundary.
    frame_size = calc_src_dma_size(width, height, bit_mode);
    frame_size = convert_to_HEXWORD(frame_size) << HEXWORD_SHIFT;

    // For bandwidth calculations, we need to add the actual metadata size back in.
    frame_size += FRAME_METADATA_HW_BYTES;

    // Get the bandwidth allocated for the SSD.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    ssd_bandwidth = regs[ GIGE_REG_SSD_BANDWIDTH ].get_value();

    // Calculate the frame rate.
    ltr_frame_rate = ssd_bandwidth / frame_size;

    // Return the maximum rate.
    return ltr_frame_rate;
}

bool FiFPGA::ltrReadImFIFO(uint8_t *imfifo_buffer, int byte_count)
{
    FI_FPGA_IO::ioReadImFifoMessage msg;

    // Sanity check.
    if (imfifo_buffer == nullptr)
        return false;
    if ((byte_count & 0xF) != 0)
    {
        fi_log(LOG_ERR, "%s: byte_count [%d] must be multiple of 16!", __PRETTY_FUNCTION__, byte_count);
        return false;
    }

    // Read the histogram out of the FPGA driver.
    msg.dest_ptr = imfifo_buffer;
    msg.byte_count = byte_count;

    // Send the request.
    if (FI_FPGA_IO::fpgaIOWorker::isRunning())
        msg.doSync(FI_FPGA_IO::fpgaIOWorker::factory());

    // Success?
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_ERR, "%s: Unable to read image memory FIFO!", __PRETTY_FUNCTION__);
        return false;
    }

    // All is well.
    return true;
}

