/*!
 * @file fpga_pme.cxx
 * @brief This source file implements the Fastec Imaging FPGA class,
 *    FiFPGA, PME or Per-frame Metadata Extraction methods
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <unistd.h>

#include "fpga.h"
#include "part_mgr.h"

#include <fastec/fi_fpga_regs.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

namespace FI_FPGA_PME
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The maximum number of times we'll spin in a loop waiting for DONE.
//! Assume 1K operations per second, very conservative.
//! FIXME Tune the times.
const int   RETRIES_PER_UNIT    = 34;
const int   FRAMES_PER_UNIT     = 1000;

//! The amount of delay between queries of the PME_STATUS register for DONE.
const uint32_t PME_DONE_DELAY_MSEC = 30;

//! The number of times to loop waiting for the PME to stop.
const int   PME_ABORT_RETRIES   = 10;

//! The amount of delay between status queries for ABORT completion.
const uint32_t  PME_ABORT_DELAY_MSEC = 10;

//! The amount of delay to wait after first asserting the abort signal.
const uint32_t  PME_ABORT_FIRST_DELAY_MSEC  = 100;

} /* End of namespace FI_FPGA_PME */

using namespace FI_FPGA_PME;

/*!
 * @brief Set up PME registers for a capture in Image Memory.
 *
 * @param part_num      [IN] The partition containing the capture.
 * @param start_frame   [IN] The first frame whose metadata we want to read.
 * @param end_frame     [IN] The last frame whose metadata we want to read.
 * @param im_buf_addr   [IN] The address in Image Memory where the PME will write the metadata, in HEXWORDs.
 *
 * @return The number of frames requested.
 */
uint32_t FiFPGA::pmeSetup(uint16_t part_num, uint32_t start_frame, uint32_t end_frame, uint32_t im_buf_addr)
{
    uint32_t part_end_frame;
    uint32_t reg_value;
    PartMgr *pm_ptr;

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Get the frame size, in HEXWORDS.
    reg_value = convert_to_HEXWORD(pm_ptr->getFrameSize());
    regWrite(PME_FRAME_BUF_SIZE, reg_value);

    // Now calculate where the PFMD is. Everything is already in HEXWORDS, so just do the math.
    reg_value -= FRAME_METADATA_SIZE_HW;
    regWrite(PME_METADATA_MWORD_OFFSET, reg_value);

    // Get the partition start and end values.
    reg_value = convert_to_HEXWORD(pm_ptr->getPartitionFrameAddr(part_num, 0, true));
    regWrite(PME_PARTITION_START, reg_value);

    part_end_frame = pm_ptr->getPartitionNumFrames(part_num) - 1;
    reg_value = convert_to_HEXWORD(pm_ptr->getPartitionFrameAddr(part_num, part_end_frame, true));
    regWrite(PME_PARTITION_END, reg_value);

    // Get the starting/ending frame addresses.
    reg_value = convert_to_HEXWORD(pm_ptr->getPartitionFrameAddr(part_num, start_frame, false));
    regWrite(PME_BLOCK_START, reg_value);

    reg_value = convert_to_HEXWORD(pm_ptr->getPartitionFrameAddr(part_num, end_frame, false));
    regWrite(PME_BLOCK_END, reg_value);

    // Specify how many frames and where to put the data.
    regWrite(PME_OUTPUT_BUF_START_PTR, im_buf_addr);
    regWrite(PME_FRAME_COUNT, end_frame - start_frame + 1);

    return (end_frame - start_frame + 1);
}

/*!
 * @brief Wait until a PME operation completes or is cancelled.
 *
 * @param frames    Number of frames in the operation, which determines the timeout.
 * @param abort_fn  Function to check for caller abort.
 * @param abort_arg Argument to pass to abort_check_fn.
 *
 * @return 0 if done, 1 if cancelled, -1 if timeout
 */
int FiFPGA::pmeWaitDone(uint32_t frame_count, bool (*abort_fn)(void *), void *abort_arg)
{
    int     retries;
    int     max_retries;
    bool    ready;
    int     rc = 0;

    max_retries = ((frame_count + FRAMES_PER_UNIT - 1) / FRAMES_PER_UNIT) * RETRIES_PER_UNIT;
    ready = false;

    for (retries = 0; retries < max_retries; ++retries)
    {
        if (pmeIsDone())
        {
            ready = true;
            break;
        }
        else if (abort_fn && (*abort_fn)(abort_arg))
            break;

        // Wait for a bit.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(PME_DONE_DELAY_MSEC));
    }
    if (!ready)
    {
        if (retries == max_retries)
        {
            uint32_t pme_control;
            uint32_t pme_status;

            regRead(PME_CONTROL, &pme_control);
            regRead(PME_STATUS, &pme_status);
            fi_log(LOG_ERR, "%s: Timeout on PME for %u frames after %u msecs: C=0x%08x S=0x%08x!",
                   __FUNCTION__, frame_count, PME_DONE_DELAY_MSEC * max_retries,
                   pme_control, pme_status);
            rc = -1;
        }
        else
        {
            rc = 1; // Aborted by callback.
        }
    }

    return rc;
}

/*!
 * @brief Read per-frame metadata into a user-supplied buffer.
 *
 * Uses the Per-frame Metadata Extraction (PME) engine to collect the metadata into
 * one contiguous region of image memory, then DMAs it into a processor memory buffer.
 *
 * @param part_num      Partition number to use.
 * @param start_frame       The first frame whose metadata we want to read.
 * @param end_frame     The last frame whose metadata we want to read.
 * @param buf           The processor buffer into which metadata will be written.
 * @param bufsize       The size in bytes of the passed buffer.
 * @param abort_check_fn    Function to check for caller abort.
 * @param abort_check_arg   Argument to pass to abort_check_fn.
 *
 * @return  The number of bytes of data written to the buffer.
 */
uint32_t FiFPGA::pmeRead(uint16_t part_num, uint32_t start_frame, uint32_t end_frame,
                         uint8_t *buf, uint32_t buf_size,
                         bool (*abort_check_fn)(void *), void *abort_check_arg)
{
    uint64_t pme_buf;
    uint32_t pme_size;
    uint32_t frame_count;
    PartMgr *pm_ptr;

    // Get a pointer to our support classes.
    pm_ptr = CameraServer::get_partmgr();

    // Get the reserved buffer for PME operations.
    pme_buf = pm_ptr->get_pme_buf_addr();
    pme_size = pm_ptr->get_pme_buf_size();

    // Use the smaller of the two buffer sizes, source and dest, though
    // the caller probably set it up so they are the same size.
    if (pme_size < buf_size)
        buf_size = pme_size;

    pme_size = pmeStart(part_num, start_frame, end_frame, convert_to_HEXWORD(pme_buf), buf_size);
    frame_count = pme_size / pmeSizeOfPFMD();

    if (pme_size != 0)
    {
        int rc = pmeWaitDone(frame_count, abort_check_fn, abort_check_arg);

        // Non-zero return can be either cancel or error.
        if (rc != 0)
            pme_size = 0;
    }

    // DMA the metadata into the caller-supplied buffer.
    if (pme_size != 0)
    {
        uint32_t read_size;

        // We don't need to byte-swap data interpreted as 32-bit words so we use big-endian.
        read_size = imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, pme_buf, buf, pme_size);
        if (read_size != pme_size)
        {
            fi_log(LOG_ERR, "%s: Unable to DMA metadata buffer (expected %u, got %u)",
                   __FUNCTION__, pme_size, read_size);
            return 0;
        }
    }

    return pme_size;
}

/*!
 * @brief Set up to load per-frame metadata into a buffer in image memory.
 *
 * Extracts metadata for all the frames in a partition and places it in
 * a buffer where it is all consecutive and can be downloaded.
 *
 * @param part_num      [IN] Partition whose metadata we are reading.
 * @param start_frame   [IN] The first frame whose metadata we want to read.
 * @param end_frame     [IN] The last frame whose metadata we want to read.
 * @param im_buf_addr   [IN] The address in image memory where PME will write the metadata, in HEXWORDs.
 * @param im_buf_size   [IN] The size of the image memory buffer, in bytes.
 *
 * @return  The number of bytes of data requested to be loaded, which might be
 *      less than end-start+1 if the buffer is too small.
 */
uint32_t FiFPGA::pmeStart(uint16_t part_num, uint32_t start_frame, uint32_t end_frame, uint32_t im_buf_addr, uint32_t im_buf_size)
{
    uint32_t    frame_count;
    uint32_t    per_frame;

    LOG("%s: part_num=%d, start_frame=%d, end_frame=%d, ddr2_buf=0x%x, ddr2_buf_size=%d",
        __PRETTY_FUNCTION__, part_num, start_frame, end_frame, im_buf_addr, im_buf_size);

    // We don't do it this way in LTR mode.
    if ((ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // FIXME Shouldn't really need this sanity check.
        fi_log(LOG_DEBUG, "%s: trying to use PME in LTR mode, no no", __FUNCTION__);
        return 0;
    }

    // This shouldn't happen.
    if (!pmeIsDone())
    {
        pmeAbort();
        if (!pmeIsDone())
        {
            fi_log(LOG_ERR, "%s: PME still busy after abort", __FUNCTION__);
            return 0;
        }
        else
            fi_log(LOG_INFO, "%s: found PME busy but abort worked", __FUNCTION__);
    }

    // How many bytes of per-frame metadata are there?
    per_frame = FRAME_METADATA_HW_BYTES;

    // Make sure we don't overrun the buffer.
    frame_count = 1 + end_frame - start_frame;
    if ((im_buf_size / per_frame) < frame_count)
    {
        // Adjust our frame count to fit the buffer size.
        frame_count = im_buf_size / per_frame;
    }
    if (end_frame != start_frame + frame_count - 1)
    {
        fi_log(LOG_DEBUG, "%s: buffer size %u too small, changing end frame from %u to %u",
               __FUNCTION__, im_buf_size, end_frame, start_frame + frame_count - 1);
        end_frame = start_frame + frame_count - 1;
    }

    // Setup the PME module.
    frame_count = pmeSetup(part_num, start_frame, end_frame, im_buf_addr);

    // Zero return says the setup failed.
    if (frame_count == 0)
        return 0;

    // Start the operation. The caller will check back with us for done status.
    regModify(PME_CONTROL, PMEC_START, -1);

    return frame_count * per_frame;
}

/*!
 * @brief Say whether the last PME operation is done.
 *
 * @return True if it is done, false if not.
 */
bool FiFPGA::pmeIsDone()
{
    uint32_t reg_value;

    // Is it still busy?
    regRead(PME_STATUS, &reg_value);
    return ((reg_value & PMES_BUSY) == 0) ? true : false;
}

/*!
 * @brief Abort a PME operation in progress.
 *
 * This is really not needed any more, since the image memory PME engine is so fast.
 */
void FiFPGA::pmeAbort()
{
    uint32_t        reg_value;
    int         retries;

    // Check whether it is still running
    regRead(PME_STATUS, &reg_value);
    if ((reg_value & PMES_BUSY) == 0)
        return;

    // Set the abort bit.
    regModify(PME_CONTROL, PMEC_ABORT, -1);

    // The abort should happen very quickly.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(PME_ABORT_FIRST_DELAY_MSEC));

    // Wait for idle.
    for (retries = 0; retries < PME_ABORT_RETRIES; ++retries)
    {
        // Read the status register to see if the PME is still busy.
        regRead(PME_STATUS, &reg_value);
        if ((reg_value & PMES_BUSY) == 0)
            break;

        // Wait for a bit.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(PME_ABORT_DELAY_MSEC));
    }
    if (retries == PME_ABORT_RETRIES)
        fi_log(LOG_ERR, "%s: Timeout waiting for PME abort!", __FUNCTION__);
}

