/*!
 * @file fpga_time.cxx
 * @brief This source file implements the Fastec Imaging FPGA
 *    class, FiFPGA, Time managment methods.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#include <time.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/rtc.h>

#include <vector>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "camera_server.h"
#include "cs_utils.h"
#include "fpga.h"
#include "fcp.h"
#include "part_mgr.h"

namespace FI_FPGA_TIME
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Our device path.
const char *IRIG_DEV_PATH   = "/dev/fi_irig";

// This is (((( 0.500V / 8.2V ) * 4095 ) + 0.5 ) * 4 ).
//
// The DAC has a maximum value of 8.2V (5.0V/0.61), the DAC is a 12-bit
// digital interface (4095 steps), we want to round up (adding 0.5) and
// we need to left shift the 12-bit DAC value to be properly justified
// for the DAC.
const uint32_t  CMP_THRESH_500mV    = 0x03E8;
const uint32_t  CMP_THRESH_MAX  = 0x0FFF;

// The debounce is in 125MHz clock ticks. There are 125 ticks/microsecond.
const uint32_t  DEBOUNCE_1uS    = 125;

// The target pulse width for a modulated signal is 325 +/- 2us.
const uint32_t  TARGET_WIDTH_MIN = 323;
const uint32_t  TARGET_WIDTH_MAX = 327;

// The pulse width for a DCLS signal is 8000us +/- 200us.
const uint32_t  TARGET_WIDTH_DCLS_MIN   = 7800;
const uint32_t  TARGET_WIDTH_DCLS_MAX   = 8200;

// The default frame offset for a DCLS signal is 10 milliseconds.
const uint32_t  DCLS_FRAME_OFFSET   = 0;

// The frame start offset values for the minimum & maximum target widths.
// Calculated as follows:
// 1. PULSE_WIDTH_DEG = 360 * ( pulse_wid / 1000 )
// 2. PULSE_START_PHASE = 90 - ( PULSE_WIDTH_DEG / 2 )
// 3. ZERO_XING_OFFSET = ( PULSE_START_PHASE / 360 ) * 0.001
// 4. IRIG_FRAME_START_OFFSET = ( 0.01 - ZERO_XING_OFFSET ) * 1.25e8
const uint32_t  FRAME_START_MIN = 1238938;
const uint32_t  FRAME_START_MAX = 1239188;

// Number of seconds to wait when tracking time.
const uint32_t  TIME_TRACKING_DELAY = 2;

// Some useful values...
const uint32_t  SEC_PER_MINUTE  = 60;
const uint32_t  SEC_PER_HOUR    = 60 * SEC_PER_MINUTE;
const uint32_t  SEC_PER_DAY = 24 * SEC_PER_HOUR;
const uint32_t  SEC_PER_YEAR    = 365 * SEC_PER_DAY;

// Number of times we will sample the incoming signal.
const int   MAX_TIMING_TRIES    = 5;

// Create some definitions for the possible IRIG types.
enum
{
    TIME_TYPE_LINUX,
    TIME_TYPE_IRIG_UNKNOWN,
    TIME_TYPE_IRIG_DCLS,    // DC level shift, no carrier, aka unmodulated signal
    TIME_TYPE_IRIG_MOD,     // modulated signal, sine wave carrier
    TIME_TYPE_SHUTDOWN
};

#define BCD_SEC(x)      ((x) & 0x7F)
#define BCD_MIN(x)      (((x) >> 7 ) & 0x7F)
#define BCD_HOUR(x)     (((x) >> 14 ) & 0x3F)
#define BCD_DAYS(x)     (((x) >> 20 ) & 0x3FF)
#define BCD_YEAR(x)     ((x) & 0xFF )

} /* End of namespace FI_FPGA_TIME */

using namespace FI_FPGA_TIME;

//--------------------------------------------------------------------
// Time worker thread
//--------------------------------------------------------------------

class time_worker : public FAS_THREAD::Thread
{
private:
    volatile bool   thread_running;         //!< Flag to indicate if we are running or not.
    uint32_t    time_type;
    uint32_t    pulse_width;
    uint32_t    min_ticks;
    uint32_t    old_clock_speed;
    float old_frame_rate;

    bool    use_year;               //!< User supplied year.

    void    acquire();
    uint32_t    calc_frame_start();
    time_t  decode_sbs(uint32_t bcd_toy);
    time_t  decode_sbs();
    int     determine_type();
    uint32_t    get_pulse_width(int irig_fd, uint16_t threshold);
    time_t  get_year();
    void    irig_processing();
    void    irig_snapshot();
    void    set_frame_offset();
    void    track_irig_time();
    void    track_linux_time();
    bool    tune(bool do_tracking);

    void    cleanup();

    // Used to create our singleton.
    time_worker();

public:
    virtual ~time_worker();

    static time_worker *factory();

    virtual void run() override;
    virtual void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr) override;
    virtual void stop() override;

    bool    is_running()    { return thread_running; }
    uint32_t    get_type()  { return time_type; }
    void    set_type(uint32_t new_type);

    void    set_year(bool use_year);
    void    snapshot();
};

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

// Latched time from IRIG. This gets latched with the user
// requests the seconds, and the usec value is passed later.
static struct timespec  irig_time_latch;

// Latched time from the time module.
static struct timespec  time_latch;

//! The time_cal object is used to calibrate the time interface.
static time_worker  *time_cal = nullptr;

//--------------------------------------------------------------------
// time_worker class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. It is expected that the
 *    thread is NOT detached and that a "join" will be used
 *    to synchronize with it on completion.
 */
time_worker::time_worker() :
    FAS_THREAD::Thread(false),
    pulse_width(TARGET_WIDTH_MIN),
    old_clock_speed(FI_FPGA::FPGA_CLOCK_SPEED),
    old_frame_rate(0)
{
    // Initialize our member variables.
    thread_running = false;
    time_type = TIME_TYPE_LINUX;
    min_ticks = 0xFFFFFFFF;

    // Initialize to not use the year.
    set_year(false);
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 *
 * @param arg   Not used.
 */
time_worker::~time_worker()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Create our singleton.
 */
time_worker *time_worker::factory()
{
    if (time_cal != nullptr)
    {
        fi_log(LOG_DEBUG, "%s: We already have our time worker!", __PRETTY_FUNCTION__);
        return time_cal;
    }

    return new time_worker();
}

/*!
 * @brief Get the thread running.
 */
void time_worker::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Now start the thread.
    FAS_THREAD::Thread::start(threadFn);
}

/*!
 * @brief Ask the thread to stop.
 */
void time_worker::stop()
{
    // Tell the thread to stop.
    FAS_THREAD::Thread::stop();
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void time_worker::cleanup()
{
    // We're not running anymore.
    thread_running = false;
}

void time_worker::set_type(uint32_t new_type)
{
    LOG("%s: new_type=%d", __PRETTY_FUNCTION__, new_type);

    // If the thread is running, have it stop.
    if (is_running())
    {
        // Request the thread to stop.
        stop_request = true;
        FAS_THREAD::Thread::stop();

        // Wait for it to stop.
        join();
    }

    // Set the new type.
    time_type = new_type;

    // Don't start the thread if we're shutting down.
    if (time_type != TIME_TYPE_SHUTDOWN)
    {
        // Startup the thread again.
        start();
    }
}

/*!
 * @brief Acquire the IRIG signal and determine what type
 *    of IRIG we're dealing with.
 */
void time_worker::acquire()
{
    struct timespec sleep_req;  // Delay period.

    // We don't know what type of signal is attached yet.
    time_type = TIME_TYPE_IRIG_UNKNOWN;

    // Reset our timing information.
    min_ticks = 0xFFFFFFFF;

    // Drop into our loop, waiting for a signal.
    while (!stop_request && (time_type == TIME_TYPE_IRIG_UNKNOWN))
    {
        int status;

        // First thing we need to do is to set the comparator threshold
        // and determine the IRIG signal type.
        status = determine_type();

        //========================================================================
        // ALM_TODO: Note that, as of right now, determine_type returns a -1.
        //           Until we actually implement IRIG, this condition will always
        //           resolve to TRUE.
        //========================================================================
        // cppcheck-suppress knownConditionTrueFalse
        if (status < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to determine IRIG type", __PRETTY_FUNCTION__);
            break;
        }
        else if (status == TIME_TYPE_IRIG_UNKNOWN)
        {
            // Initialize our sleep time.
            sleep_req.tv_sec = 0;
            sleep_req.tv_nsec = 250000000;

            // Delay for a bit before trying again.
            sleep(&sleep_req);
            continue;
        }

        // Save our IRIG type.
        time_type = (uint32_t)status;
    }
}

/*!
 * @brief Return the number of years in seconds. If we are
 *    tracking Linux time, then we always deal with years.
 *    For IRIG time, the user has to tell us if we're
 *    interested in years or not.
 *
 *    Currently assume 365 days/year.
 *
 * @return BCD Year (IRIG) or tm_year (Linux) in seconds.
 */
time_t time_worker::get_year()
{
    uint32_t    temp_value;

    // Initialize our value.
    temp_value = 0;

    // Are we doing Linux time?
    if (time_type == TIME_TYPE_LINUX)
    {
        time_t  time_now;
        struct tm   *tm_now;

        // Get the current system time
        time_now = time(NULL);

        // Now convert Linux time to SBS.
        tm_now = localtime(&time_now);

        // We want to return the year, adjusted appropriately.
        temp_value = tm_now->tm_year - (2000 - 1900);
    }
    else
    {
#ifdef ALM_TODO_IRIG
        // Make sure we're using the "year" value.
        if (use_year)
        {
            uint32_t    reg_value;
            uint32_t    bcd_value;
            FiFPGA  *fpga_ptr;

            // Get a pointer to our support classes.
            fpga_ptr = CameraServer::get_fpga();

            // Read the CONTROL FIELD register. This is the value that
            // was latched when the TOY field was latched.
            fpga_ptr->read_reg(IRIG_CONTROL_FIELD_LATCHED, &reg_value);
            bcd_value = BCD_YEAR(reg_value);
            temp_value = (bcd_value & 0x0F) + (((bcd_value >> 4) & 0x0F) * 10);
        }
#endif
    }

    // Make an assumption of 365 days/year.
    temp_value *= SEC_PER_YEAR;

    // Return our year, in seconds.
    return temp_value;
}

/*!
 * @brief Remember if we're interested in using the IRIG year.
 *
 * @param use_year True to use the year information, False otherwise.
 */
void time_worker::set_year(bool use_year)
{
    // Remember if we are supposed to use the year or not.
    this->use_year = use_year;
}

/*!
 * @brief Tune the FPGA time module to maintain sync with the IRIG value.
 */
bool time_worker::tune(bool do_tracking)
{
    (void)do_tracking;

#ifdef ALM_TODO_IRIG
    int     retries;
    bool    do_update(false);
    bool    lost_track(false);
    uint32_t    reg_value;
    float    frame_rate;
    uint32_t    clock_speed;
    FiFPGA  *fpga_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Now that we've acquired the IRIG signal, let's measure
    // how many ticks we count between IRIG 1PPS pulses.
    fpga_ptr->modify_reg(EXT_IO_CONTROL, EXTIO_SYNC_PERIOD_MEAS_IRIG, ~EXTIO_SYNC_PERIOD_MEAS_MASK);

    // Read the measured minimum time and throw it away.
    reg_value = fpga_ptr->get_irig_syncin_timing();

    // What is the desired frame rate?
    // cppcheck-suppress unreadVariable
    frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());

    // Read our minimum value. Do this a few times to
    // make an effort at getting a minimum.
    for (retries = 0; (!stop_request && (retries < MAX_TIMING_TRIES)); retries++)
    {
        // Do we care about tracking information?
        if (do_tracking)
        {
            // See if we've bumped out of slave mode. This would mean the
            // IRIG signal went away.
            fpga_ptr->read_reg(TIME_STATUS, &reg_value);
            if (TIME_STATUS_STATE(reg_value) == TIME_MODE_MASTER)
            {
                uint32_t sync_value;

                // Log the event.
                fi_log(LOG_INFO, "%s: Time module auto-switched to MASTER!", __PRETTY_FUNCTION__);

                // Are we setup for using IRIG as our 1PPS sync source?
                sync_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
                if ((sync_value & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_IRIGB)
                {
                    // Turn off the SyncIn IRIG enable and switch the
                    // ICM to use the internal time source.
                    fcp_syncin_update(NULL, GIGE_SYNC_TYPE_INTERNAL);
                }

                // We need to re-acquire the IRIG signal, so exit the tracking state.
                lost_track = true;
                break;
            }

            // Were we late?
            if ((reg_value & TIME_STATUS_SYNC_LATE) == TIME_STATUS_SYNC_LATE)
                fi_log(LOG_INFO, "%s: IRIG was late.", __FUNCTION__);
        }

        // Wait to give the logic a chance to run.
        sleep(1);

        // What is the desired frame rate?
        frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());

        // Did the frame rate change?
        if (!approx_equals(old_frame_rate, frame_rate))
        {
            // Since the frame rate changed, let's make sure we reset
            // our tracking information too.
            old_frame_rate = frame_rate;
            old_clock_speed = FI_FPGA::FPGA_CLOCK_SPEED;
        }

        // Read the measured minimum time.
        reg_value = fpga_ptr->get_irig_syncin_timing();

        // Make sure we get a "good" value.
        if (reg_value != 0)
        {
            // Is this smaller?
            if (reg_value < min_ticks)
            {
                // Save our new minimum tick value.
                min_ticks = reg_value;

                // Make sure we update the FPGA.
                do_update = true;
            }

            // We need the FPGA to finish a little fast, so adjust
            // the clock speed to make things go quicker.
            clock_speed = reg_value - SYNC_IN_SLAVE_TUNING;

            // Make sure the new clock speed has slipped beyond a certain range.
            if ((clock_speed < (old_clock_speed - (SYNC_IN_SLAVE_TUNING >> 3))) ||
                (clock_speed > (old_clock_speed + (SYNC_IN_SLAVE_TUNING >> 3))))
            {
                // Everything looks OK. Adjust the frame rate and shutter
                // speed. Be careful about which to adjust first.
                if (clock_speed < old_clock_speed)
                {
                    // We are going faster, so adjust the shutter
                    // first.
                    fpga_ptr->set_shutter_speed(fpga_ptr->get_shutter_speed(), clock_speed / 1000000);
                    fpga_ptr->set_frame_rate(frame_rate, clock_speed);
                }
                else
                {
                    // We are going slower, so adjust the frame rate
                    // first.
                    fpga_ptr->set_frame_rate(frame_rate, clock_speed);
                    fpga_ptr->set_shutter_speed(fpga_ptr->get_shutter_speed(), clock_speed / 1000000);
                }

                // Remember this clock speed for the next time.
                old_clock_speed = clock_speed;
            }
        }
    }

    // If we're supposed to stop, or we've lost track, then leave.
    if (stop_request || lost_track)
        return false;

    // Do we want to update the FPGA?
    if (do_update)
    {
        // Sanity check our ticks.
        if (min_ticks != 0)
        {
            // We now have our minimum number of ticks between pulses.
            // Calculate our 1MHz tuning word by dividing 5e14 by the
            // number of 125MHz ticks we got.
            reg_value = (uint32_t)(0x1C6BF52634000LL / ((uint64_t)min_ticks));

            // Since we're in slave mode, we want the tuning word to be
            // "slightly" small, so our clock is "slightly" slow.
            reg_value -= SLAVE_TUNING_KLUDGE;

            // Update our slave value.
            fpga_ptr->write_reg(TIME_1MHZ_TUNING_WORD_SLAVE, reg_value);
        }
        else
        {
            fi_log(LOG_ERR, "%s: min_ticks is 0?!", __PRETTY_FUNCTION__);
        }
    }

    // All good.
    return true;
#else
    return false;
#endif

}

/*!
 * @brief Keep the FPGA's time module in sync with Linux Time-Of-Day.
 */
void time_worker::track_linux_time()
{
    uint32_t    reg_value;
    uint32_t    sbs_now;
    time_t  time_now;
    struct tm   *tm_now;
    FiFPGA  *fpga_ptr;
    struct timespec sleep_req;  // Delay period.

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null!", __PRETTY_FUNCTION__);
        return;
    }

    // Get the current system time
    time_now = time(NULL);

    // Now convert Linux time to SBS.
    tm_now = localtime(&time_now);

    // Figure out what our SBS value should be.
    sbs_now = tm_now->tm_sec + (tm_now->tm_min * SEC_PER_MINUTE) + (tm_now->tm_hour * SEC_PER_HOUR) + (tm_now->tm_yday * SEC_PER_DAY);
    if (use_year)
        sbs_now += ((tm_now->tm_year - (2000 - 1900)) * SEC_PER_YEAR);

    // Load the TS3 time with our SBS.
    fpga_ptr->regWrite(TIME_SBS_LOAD_VALUE, sbs_now);

    // Turn on the time module in master mode.
    fpga_ptr->regWrite(TIME_CONTROL, TIME_MODE_SET | TIME_MODE_MASTER);

    // Read the time status register to clear out any old status.
    fpga_ptr->regRead(TIME_STATUS, &reg_value);

    // Loop until we are told to stop.
    while (!stop_request)
    {
        uint32_t    ts3_sbs;

        // Initialize our delay period.
        sleep_req.tv_sec = TIME_TRACKING_DELAY;
        sleep_req.tv_nsec = 0;
        sleep(&sleep_req);

        // Are we supposed to stop?
        if (stop_request)
        {
            fi_log(LOG_DEBUG, "%s: Stop requested", __PRETTY_FUNCTION__);
            break;
        }

        //--------------------------------------------------------------
        // NOTE: We actually are "tuning" the time module, but it's
        //   happening in the sync_worker::track() method. I did
        //   that because it's doing all the measurements already
        //   and I didn't want to duplicate that code here.
        //--------------------------------------------------------------

        // Get the current system time and take a snapshot of the time modules
        // idea of time.
        time_now = time(NULL);
        ts3_sbs = fpga_ptr->timeReadSBS();

        // Now convert Linux time to SBS.
        tm_now = localtime(&time_now);

        // Figure out what our SBS value should be.
        sbs_now = tm_now->tm_sec + (tm_now->tm_min * SEC_PER_MINUTE) + (tm_now->tm_hour * SEC_PER_HOUR) + (tm_now->tm_yday * SEC_PER_DAY);
        if (use_year)
            sbs_now += ((tm_now->tm_year - (2000 - 1900)) * SEC_PER_YEAR);

        // Are we off? +/- 1 second is good enough.
        if (((ts3_sbs - 1) != sbs_now) && (ts3_sbs != sbs_now) && ((ts3_sbs + 1) != sbs_now))
        {
            // Load the TS3 time with the Linux time.
            fpga_ptr->regWrite(TIME_SBS_LOAD_VALUE, sbs_now);
        }
    }
}

/*!
 * @brief Keep the FPGA's time module in sync with IRIG.
 */
void time_worker::track_irig_time()
{
#ifdef ALM_TODO_IRIG
    uint32_t    irig_sbs;
    uint32_t    ts3_sbs;
    uint32_t    reg_value;
    FiFPGA      *fpga_ptr;
    struct timespec sleep_req;  // Delay period.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Tune our 1Mhz loop. Don't track the signal.
    tune(false);

    // Get the IRIG SBS.  This grabs a snapshot of all
    // the relevant time values.
    irig_sbs = decode_sbs();

    // Load the TS3 time with the IRIG SBS.
    fpga_ptr->write_reg(TIME_SBS_LOAD_VALUE, irig_sbs);

    // Enable the TS3 time module to slave to IRIG 1PPS.
    reg_value = TIME_MODE_SET | TIME_MODE_IRIG_SLAVE | (TIME_SYNC_LATE_LIMIT << TIME_SYNC_LATE_SHIFT);
    fpga_ptr->modify_reg(TIME_CONTROL, reg_value, ~(TIME_SYNC_LATE_MASK | TIME_MODE_MASK));

    // If we're supposed to be using IRIG as our SyncIn source,
    // then we want to make sure it's selected as our source and
    // to make our time control slave to the 1PPS.
    if ((regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value() & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_IRIGB)
    {
        // Enable the TS3 time module to slave to IRIG.
        fpga_ptr->modify_reg(EXT_IO_CONTROL, EXTIO_SYNC_1PPS_IRIG, ~EXTIO_SYNC_1PPS_SRC_SEL_MASK);
    }

    // Stay in track until we're requested to leave.
    while (!stop_request)
    {
        // Initialize our delay period.
        sleep_req.tv_sec = TIME_TRACKING_DELAY;
        sleep_req.tv_nsec = 0;
        sleep(&sleep_req);

        // Re-tune the time. We should be tracking, so verify that.
        if (!tune(true))
            break;

        // Decode the BCD SBS. This grabs a snapshot of all
        // the relevant time values.
        irig_sbs = decode_sbs();

        // From the snapshot taken above, grab the TS3 time.
        fpga_ptr->read_reg(TIME_SBS_COUNTER_LATCHED, &ts3_sbs);

        // Do we care about years? If not, we need to ignore any "year" information
        // provided by the FPGA.
        if (!use_year)
        {
            // Pull out any year contribution.
            while (ts3_sbs > SEC_PER_YEAR)
            {
                // Subtract out a year's worth of seconds.
                ts3_sbs -= SEC_PER_YEAR;
            }
        }

        // If the clocks are different, reload the TS3 time.
        if (ts3_sbs != irig_sbs)
        {
            // Load the TS3 time with the IRIG SBS.
            fpga_ptr->write_reg(TIME_SBS_LOAD_VALUE, irig_sbs);
        }
    }

    // Switch the TS3 time module to master on the way out.
    fpga_ptr->modify_reg(TIME_CONTROL, (TIME_MODE_SET | TIME_MODE_MASTER), ~TIME_MODE_MASK);
#endif
}

/*!
 * @brief Determine the type of IRIG signal being received.
 *
 * @return One of the TIME_TYPE values, or -1 on error.
 */
int time_worker::determine_type()
{
#ifdef ALM_TODO_IRIG
    int     irig_fd;
    int     ret_type;
    bool    done(false);
    uint16_t    thresh_l;
    uint16_t    thresh_r;
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Open the IRIG device driver.
    irig_fd = open(IRIG_DEV_PATH, O_RDWR, 0666);
    if (irig_fd == -1)
    {
        fi_log(LOG_ERR, "%s: Unable to open IRIG device, err=%d(%s)", __PRETTY_FUNCTION__,
               errno, strerror(errno));
        return -1;
    }

    // Initialize our return type.
    ret_type = TIME_TYPE_IRIG_UNKNOWN;

    // Initialize our binary searching limits.
    thresh_l = 1;
    thresh_r = CMP_THRESH_MAX;

    // Initialize our debounce period to 1us.
    fpga_ptr->write_reg(IRIG_DEBOUNCE_PERIOD, DEBOUNCE_1uS);

    // Keep looping until we're done or a stop has been requested.
    while (!done && !stop_request)
    {
        uint16_t cmp_thresh;

        // Bail if we've exhausted our possibilities.
        if (thresh_l == thresh_r)
        {
            fi_log(LOG_WARNING, "%s: Threshold search exhausted", __PRETTY_FUNCTION__);
            break;
        }

        // Split the difference.
        cmp_thresh = (thresh_l + thresh_r) / 2;

        // Measure the maximum pulse width.
        pulse_width = get_pulse_width(irig_fd, cmp_thresh);

        // Adjust our threshold based on the pulse_width value.
        if (pulse_width < TARGET_WIDTH_MIN)
        {
            // Adjust the right hand edge of our search.
            thresh_r = cmp_thresh - 1;
        }
        else if (pulse_width > TARGET_WIDTH_MAX)
        {
            // Check to see if we have a DCLS signal.
            if (pulse_width > TARGET_WIDTH_DCLS_MAX)
            {
                // Either we have a serious problem or there's no signal.
                // We'll assume no signal.
                done = true;
            }
            else if (pulse_width > TARGET_WIDTH_DCLS_MIN)
            {
                // We have a winner!
                ret_type = TIME_TYPE_IRIG_DCLS;
                done = true;
            }
            else
            {
                // Adjust the left hand edge of our search.
                thresh_l = cmp_thresh + 1;
            }
        }
        else
        {
            // We have a winner!
            ret_type = TIME_TYPE_IRIG_MOD;
            done = true;
        }
    }

    // Close the IRIG device.
    close(irig_fd);

    // All done.
    return ret_type;
#else
    return -1;
#endif
}

/*!
 * @brief Have the FPGA measure the width of the incoming IRIG pulse(s).
 *
 * @param irig_fd   Descriptor for the IRIG DAC device driver.
 * @param threshold The value we want to set the DAC to.
 *
 * @return The maximum (debounced) pulse width, in 125MHz ticks.
 */
uint32_t time_worker::get_pulse_width(int irig_fd, uint16_t threshold)
{
    (void)irig_fd;
    (void)threshold;

#ifdef ALM_TODO_IRIG
    int     rc;         // Result codes returned by function calls.
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;
    char    thresh_buf[ 10 ];   // Just need a 16-bit hex value, so this is long enough.
    struct timespec sleep_req;  // Delay period.

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Adjust the threshold to where the DAC expects it to be.
    threshold <<= 2;

    // Convert to an ASCII value.
    snprintf(thresh_buf, 10, "0x%04x", threshold);

    // Set the comparator threshold
    write(irig_fd, thresh_buf, strlen(thresh_buf) + 1);

    // Ask the FPGA for a measurement.
    fpga_ptr->write_reg(IRIG_CONTROL, (IRIG_IF_ENABLE | IRIG_PULSE_MEASURE_START));

    // Delay for a bit before trying again.
    sleep_req.tv_sec = 0;
    sleep_req.tv_nsec = 150000000;
    rc = sleep(&sleep_req);
    if ((rc != 0) && (rc != ETIMEDOUT))
        fi_log(LOG_WARNING, "%s: Odd return from wait (%d)", __PRETTY_FUNCTION__, rc);
    if (stop_request)
        return 0;

    // Tell the FPGA to stop measuring.
    fpga_ptr->write_reg(IRIG_CONTROL, (IRIG_IF_ENABLE | IRIG_PULSE_MEASURE_STOP));

    // Read the measured results.
    fpga_ptr->read_reg(IRIG_DEB_MAX_PULSE_WIDTH, &reg_value);

    // Return the value we read.
    return reg_value;
#else
    return 0;
#endif
}

/*!
 * @brief Based on the pulse width, determine our frame-start offset.
 *
 * @return The frame-start offset to send to the FPGA's time module.
 */
uint32_t time_worker::calc_frame_start()
{
    uint32_t    ret_value;
    float   calc_step;      // Used for intermediate calculation steps.

    calc_step = (float)(pulse_width - TARGET_WIDTH_MIN) / ((float)(TARGET_WIDTH_MAX - TARGET_WIDTH_MIN));
    ret_value = (uint32_t)((calc_step * (FRAME_START_MAX - FRAME_START_MIN)) + 0.5);
    ret_value += FRAME_START_MIN;

    // Return our interpolated value.
    return ret_value;
}

/*!
 * @brief Program the FPGA's time module with the frame-start offset.
 */
void time_worker::set_frame_offset()
{
#ifdef ALM_TODO_IRIG
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Different values for the different types of IRIG.
    if (time_type == TIME_TYPE_IRIG_DCLS)
    {
        // Default value.
        reg_value = DCLS_FRAME_OFFSET;
    }
    else
    {
        // Interpolate our frame offset based on the pulse width
        // we ended up getting.
        reg_value = calc_frame_start();
    }

    // Write the value out to the FPGA.
    fpga_ptr->write_reg(IRIG_FRAME_START_OFFSET, reg_value);
#endif
}

/*!
 * @brief Instruct the FPGA to take a snapshot of time information
 *    and load the time_latch variable with the Time module's
 *    SBS/SBUS information.
 */
void time_worker::snapshot()
{
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Latch the time information.
    fpga_ptr->regModify(TIME_CONTROL, TIME_STAMP_SNAPSHOT, -1);

    // Grab the time modules SBS and SBUS times.
    fpga_ptr->regRead(TIME_SBS_COUNTER_LATCHED, &reg_value);
    time_latch.tv_sec = reg_value;

    // Read the latched microseconds.
    fpga_ptr->regRead(TIME_SBUS_COUNTER_LATCHED, &reg_value);

    // Convert to nanoseconds and carry on.
    time_latch.tv_nsec = reg_value * 1000;
}

/*!
 * @brief Instruct the FPGA to take a snapshot of time information
 *    and load the irig_time_latch variable with the decoded
 *    IRIG BCD information concatenated with the Time module's
 *    SBUS information.
 */
void time_worker::irig_snapshot()
{
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;

    // Decode the BCD to be number of seconds.
    irig_time_latch.tv_sec = decode_sbs();

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Read the latched microseconds.
    fpga_ptr->regRead(TIME_SBUS_COUNTER_LATCHED, &reg_value);

    // Convert to nanoseconds and carry on.
    irig_time_latch.tv_nsec = reg_value * 1000;
}

/*!
 * @brief Decode the IRIG BCD value passed in via the bcd_toy variable.
 *
 * @param bcd_toy   The IRIG BCD Time-Of-Year value.
 *
 * @return The time in seconds.
 */
time_t time_worker::decode_sbs(uint32_t bcd_toy)
{
    time_t  ret_value;
    uint32_t    temp_val;
    uint32_t    bcd_val;

    // Check for the simple case.
    if (bcd_toy == 0)
        return 0;

    // Get the seconds and convert it from BCD to binary.
    bcd_val = BCD_SEC(bcd_toy);
    ret_value = (bcd_val & 0xF) + ((bcd_val >> 4) * 10);

    // The IRIG time is always 1 second behind. Make up
    // for that now.
    ret_value++;

    // Get the minutes, convert to hex. Add to our total.
    bcd_val = BCD_MIN(bcd_toy);
    ret_value += (((bcd_val & 0xF) + ((bcd_val >> 4) * 10)) * SEC_PER_MINUTE);

    // Get the hours, convert to hex. Add to our total.
    bcd_val = BCD_HOUR(bcd_toy);
    ret_value += (((bcd_val & 0xF) + ((bcd_val >> 4) * 10)) * SEC_PER_HOUR);

    // Get the days, convert to hex, then to seconds. Add to our total
    bcd_val = BCD_DAYS(bcd_toy);
    temp_val = (bcd_val & 0xF) + (((bcd_val >> 4) & 0xF) * 10) + ((bcd_val >> 8) * 100);

    // The BCD "Day Of Year" is one-based, but Linux time assumes the "tm_yday"
    // element (which we use) is zero-based. Convert down to zero-based.
    temp_val -= 1;

    // Convert from day of year to seconds in year and add that to our accumulator.
    temp_val *= SEC_PER_DAY;
    ret_value += temp_val;

    // Add in our decoded year value.
    ret_value += get_year();

    // Our time in seconds.
    return ret_value;
}

/*!
 * @brief Have the FPGA grab a snapshot of the IRIG BCD information
 *    and decode that to number of seconds.
 *
 * @return The number of seconds represented by the IRIG BCD Time-Of-Year.
 */
time_t time_worker::decode_sbs()
{
#ifdef ALM_TODO_IRIG
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Latch the time information.
    fpga_ptr->modify_reg(TIME_CONTROL, TIME_STAMP_SNAPSHOT, -1);

    // Read the BCD Time-Of-Year register.
    fpga_ptr->read_reg(IRIG_BCD_TOY_FIELD_LATCHED, &reg_value);

    // Return the decoded value.
    return decode_sbs(reg_value);
#else
    return 0;
#endif
}

/*!
 * @brief Process IRIG time.
 */
void time_worker::irig_processing()
{
#ifdef ALM_TODO_IRIG
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Acquire the IRIG signal.
    acquire();

    // Check to see if we're being asked to stop.
    if (stop_request)
        return;

    // Adjust our frame start offset value.
    set_frame_offset();

    // Enable our demod and decoding.
    fpga_ptr->write_reg(IRIG_CONTROL, (IRIG_IF_ENABLE | IRIG_DEMOD_ENABLE | IRIG_DECODE_ENABLE));

    // Now start tracking the signal, until we're asked to stop.
    track_irig_time();
#endif
}

/*!
 * @brief The worker thread.
 */
void time_worker::run()
{
    FiFPGA  *fpga_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // We're running.
    thread_running = true;
    old_clock_speed = FI_FPGA::FPGA_CLOCK_SPEED;
    old_frame_rate = CameraServer::get_cs()->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());

    // Initialize our 1MHz tuning word for master mode. If you take
    // 5e14 / 1.25e8 (assuming 125,000,000 ticks/sec) you get 4,000,000.
    fpga_ptr->regWrite(TIME_1MHZ_TUNING_WORD_MASTER, 4000000);

    // Run until the user wants us to stop.
    while (!stop_request)
    {
        // Figure out what time "type" we're persuing.
        if (time_type == TIME_TYPE_LINUX)
        {
            // Track our Linux time.
            track_linux_time();
        }
        else
        {
            // Perform IRIG style processing.
            irig_processing();
        }
    }

    // Clean up on the way out.
    cleanup();
}

//--------------------------------------------------------------------
// FiFPGA::IRIG interfaces
//--------------------------------------------------------------------

/*!
 * @brief Return the FPGA's IRIG status information.
 */
uint32_t FiFPGA::irigStatus()
{
#ifdef ALM_TODO_IRIG
    uint32_t    rc;
    uint32_t    reg_value;
    const fi_license_options    *lic_ptr = NULL;   //!< Pointer to the licensing information.

    lic_ptr = CameraServer::get_license();
    if (lic_ptr == NULL)
        return 0;

    // Are we licensed?
    if (lic_ptr->irig == 0)
        return 0;

    // If we're here, then we're licensed.
    rc = GIGE_IRIG_ALLOWED;

    // Read the FPGA to determine if the IRIG module is locked
    // to a signal.
    read_reg(IRIG_STATUS, &reg_value);
    if ((reg_value & IRIG_DECODE_FRAME_SYNC) == IRIG_DECODE_FRAME_SYNC)
        rc |= GIGE_IRIG_LOCKED;

    // Return our result.
    return rc;
#else
    return 0;
#endif
}

/*!
 * @brief Take a snapshot of the Time module's counters and
 *    return the seconds portion.
 */
uint32_t FiFPGA::timeReadSBS()
{
    // Sanity check.
    if (time_cal == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No time worker!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Take a snapshot.
    time_cal->snapshot();

    // Return the seconds.
    return time_latch.tv_sec;
}

/*!
 * @brief Return the previously latched SBUS value as
 *    nanoseconds (more easily used in Linux time
 *    manipulations).
 */
uint32_t FiFPGA::timeReadSBUS()
{
    // Return the straight binary microseconds.
    return (uint32_t)((time_latch.tv_nsec + 500) / 1000);
}

/*!
 * @brief Setup the IRIG interface based on the bits set in the
 *    control_value parameter.
 *
 * @param control_value Value as specified in gige_msg.h
 *
 * @return GEV_STATUS_SUCCESS if success, error code otherwise.
 */
int FiFPGA::irigConfigure(CS_NETWORK::CS_NETWORK_MESSAGE * /* p_msg */, uint32_t control_value)
{
    (void)control_value;

    // Sanity check.
    if (time_cal == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No time worker!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

#ifdef ALM_TODO_IRIG
    uint32_t    status_value;
    uint32_t    reg_value;
    FiFPGA  *fpga_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Get the current status of the IRIG module.
    status_value = fpga_ptr->irig_status();

    // Are we allowed access to the IRIG?
    if ((status_value & GIGE_IRIG_ALLOWED) != (uint32_t)GIGE_IRIG_ALLOWED)
    {
        fi_log(LOG_WARNING, "%s: IRIG is not allowed.", __FUNCTION__);
        return GEV_STATUS_ACCESS_DENIED;
    }

    // Get the current settings.
    reg_value = regs[ GIGE_REG_IRIG_CONTROL ].get_value();

    // Are we enabling or disabling the IRIG?
    if ((control_value & GIGE_IRIG_ENABLE) == (uint32_t)GIGE_IRIG_ENABLE)
    {
        uint32_t fpga_reg;

        // Turn off SyncIn, if it's enabled.
        fcp_syncin_update(NULL, GIGE_SYNC_TYPE_INTERNAL);

        // Should we enable or disable the year setting?
        if ((control_value & GIGE_IRIG_USE_YEAR) == GIGE_IRIG_USE_YEAR)
        {
            time_cal->set_year(true);
            reg_value |= GIGE_IRIG_USE_YEAR;
        }
        else
        {
            time_cal->set_year(false);
            reg_value &= ~GIGE_IRIG_USE_YEAR;
        }

        // Remember that we're enabling IRIG.
        reg_value |= GIGE_IRIG_ENABLE;

        // Check to see if the FPGA is already enabled.
        fpga_ptr->read_reg(IRIG_CONTROL, &fpga_reg);
        if (((fpga_reg & IRIG_IF_ENABLE) == 0) || (time_cal->get_type() == TIME_TYPE_LINUX))
        {
            // Enable the IRIG hardware.
            fpga_ptr->write_reg(IRIG_CONTROL, IRIG_IF_ENABLE);

            // Change the time type to IRIG.
            time_cal->set_type(TIME_TYPE_IRIG_UNKNOWN);
        }

        // The LEDs will be handled by the time_worker thread or the sync_worker thread.
    }

    // We want to disable - make sure we're enabled.
    else if ((reg_value & GIGE_IRIG_ENABLE) == (uint32_t)GIGE_IRIG_ENABLE)
    {
        // Clear the enable bit.
        reg_value &= ~GIGE_IRIG_ENABLE;

        // Linux time always uses the year.
        time_cal->set_year(true);

        // Change the time type back to Linux.
        time_cal->set_type(TIME_TYPE_LINUX);

        // Turn off the IRIG interface.
        fpga_ptr->write_reg(IRIG_CONTROL, 0);

        // Turn off SyncIn, if it's enabled.
        fcp_syncin_update(NULL, GIGE_SYNC_TYPE_INTERNAL);
    }

    // Save the value.
    regs[ GIGE_REG_IRIG_CONTROL ].set_value(reg_value);

    // It's all good.
    return GEV_STATUS_SUCCESS;
#else
    return 0;
#endif
}

/*!
 * @brief Initialize the FPGA's time module.
 */
void FiFPGA::timeInit()
{
    // Create our time object.
    if (time_cal != nullptr)
    {
        fi_log(LOG_DEBUG, "%s: We already have a time worker thread!", __PRETTY_FUNCTION__);
        return;
    }

    // Create our worker thread.
    time_cal = time_worker::factory();
    if (time_cal == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to create the time worker thread!", __PRETTY_FUNCTION__);
        return;
    }

    // Linux time always uses the year.
    time_cal->set_year(true);

    // Initialize the time to Linux time.
    time_cal->set_type(TIME_TYPE_LINUX);

#ifdef ALM_TODO_IRIG
    // Make sure the IRIG interface is off.
    write_reg(IRIG_CONTROL, 0);
#endif
}

/*!
 * @brief Shut down the time server.
 */
void FiFPGA::timeStop()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (time_cal == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No time worker!", __PRETTY_FUNCTION__);
        return;
    }

    // Tell the thread to shutdown.
    time_cal->set_type(TIME_TYPE_SHUTDOWN);

    // Release the object.
    delete time_cal;
    time_cal = nullptr;
}

