/*!
 * @file fsp.cpp
 *
 * @brief This file contains the implementation of the Fastec Streaming
 *    Protocol (FSP) logic.
 *
 * This file contains all the routines to deal with FSP messages.
 * It will stream images to the desired IP address and port using
 * parameters set with the Fastec Control Protocol (FCP).
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/sendfile.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>

#include <algorithm>
#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>
#include <string>
#include <thread>
#include <vector>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>

#include <fmt/format.h>
#include <fmt/ranges.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_network.h"
#include "cs_trace.h"
#include "cs_utils.h"
#include "dcraw_util.h"
#include "fcp.h"
#include "fsp.h"
#include "fi_internal.h"
#include "grpc_server.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "override.h"
#include "part_mgr.h"
#include "pb_control.h"
#include "sensor.h"
#include "shmem_stream.h"
#include "storage.h"

namespace FASTEC_STREAM
{
// #define ULV_BUFFER_INIT  1

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const size_t ALLOCATION_UNIT = 4096;

//! The amount of time to wait for the user to step.
const int SS_STEP_PAUSE = 50;

//! Sequence number for the TXT metadata file.
const uint32_t  METADATA_TXT_SEQ_NUM            = ((uint32_t) -1);

//! Sequence number for the CFG metadata file.
const uint32_t  METADATA_CFG_SEQ_NUM            = ((uint32_t) -2);

//! Sequence number for the XML metadata file.
const uint32_t  METADATA_XML_SEQ_NUM            = ((uint32_t) -3);

//! Sequence number for the RAW metadata file.
const uint32_t  METADATA_RAW_SEQ_NUM            = ((uint32_t) -4);

static const int UDP_OVERHEAD           = 28;               //!< IP (20 bytes)/UDP (8 bytes) overhead.
static const int FSP_PAYLOAD_OVERHEAD   = 8;            //!< How big is the FSP header overhead.
static const int STREAM_POLLING_DELAY   = 10000;        //!< Microseconds to delay while polling the GOE.
static const int STREAM_MODE_DELAY      = 200000;       //!< Microseconds to delay waiting for mode switches.

static const int MAX_SETMODE_RETRY_COUNT = 100;                 //!< Max retries when setting mode.

// To download up to 20MB over GOE from image memory, allow a 2 second timeout then
// add time for packet delays.
static const uint32_t BASE_GOE_TIMEOUT = 2 * 1000 * 100;

// To download up to 20MB over GOE from SSD, allow a 5 second timeout then add time
// for packet delays.
static const uint32_t BASE_GOE_TIMEOUT_SSD = 5 * 1000 * 100;

// To stop the walk, return a non-zero value.
typedef int (* WALK_CALLBACK)(const char *file_path, const uint8_t *entry_ptr, void *user_info);

enum FSP_TRACE
{
    FSP_FUNC        = 0x00000010,
    FSP_PACKET      = 0x00000020,
    FSP_STATS       = 0x00000040,
    FSP_DOWNLOAD    = 0x00000080,
    FSP_STATE       = 0x00000100,
    FSP_PLAYBACK    = 0x00000200,
    FSP_FILESTREAM  = 0x00000400,
    FSP_SESSSTREAM  = 0x00000800,
    FSP_LIVE        = 0x00001000,
    FSP_STATE_VERBOSE   = 0x00002000,
};

enum DOWNLOAD_STATES
{
    DNLD_SETUP,
    DNLD_LTR_FRAME_LOAD,
    DNLD_PROCESS_FRAME,
    DNLD_TRANSFER_FRAME,
    DNLD_ENCODE_FRAME,
    DNLD_STREAM_FRAME,
    DNLD_CHECK_MODE
};

enum LIVE_STATES
{
    LIVE_SETUP,
    LIVE_ACQUIRE_FRAME,
    LIVE_PROCESS_FRAME,
    LIVE_TRANSFER_FRAME,
    LIVE_STREAM_FRAME,
    LIVE_CHECK_MODE
};

enum PLAYBACK_STATES
{
    PB_SETUP,
    PB_GET_FRAME,
    PB_STREAM_FRAME
};

enum SESSION_STATES
{
    SESS_SETUP,
    SESS_TRANSFER_CHUNK,
    SESS_SUBPART_TRANSFER_HEADER,
    SESS_SUBPART_TRANSFER_FIRST_SPLIT,
    SESS_STREAM_CHUNK,
    SESS_CHECK_MODE
};

enum THREAD_RUN_STATES
{
    TR_SETUP_SOCKET,
    TR_WAIT_FOR_CONNECT,
    TR_WAIT_FOR_MODE,
    TR_STREAM
};

enum SINGLESTEP_STATES
{
    SS_SETUP,
    SS_LTR_FRAME_LOAD,
    SS_PROCESS_FRAME,
    SS_STREAM_FRAME,
    SS_CHECK_MODE
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    FSP_FUNC       |
//    FSP_PACKET     |
//    FSP_STATS      |
//    FSP_DOWNLOAD   |
    FSP_STATE      |
//    FSP_PLAYBACK   |
//    FSP_FILESTREAM |
    FSP_SESSSTREAM |
//    FSP_LIVE       |
//    FSP_STATE_VERBOSE |
//    TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

} /* End of FASTEC_STREAM */

//--------------------------------------------------------------------
// Static Class Variable Declarations
//--------------------------------------------------------------------
std::mutex CsFSP::serial_stream;

//--------------------------------------------------------------------
//! @brief The class for managing image processing.
//--------------------------------------------------------------------
class ipm_state
{
public:
    ipm_state()
    {
        busy = false;
        bb_index = -1;
        frame_num = 0;
        frame_addr_hw = 0;
        frame_counter = 0;
        image_len = 0;
        ipm_buf_addr = 0;
        bb_addr = 0;
        bb_ptr = nullptr;
    }
    ~ipm_state() {}

    bool     busy;      //!< True if we are processing an image.

    int      frame_num;         //!< Frame number being processed.
    uint32_t frame_addr_hw;     //!< Frame address in image memory (hexword addr).
    uint32_t frame_counter; //!< The last live frame capture counter we read.

    uint32_t image_len;         //!< Length of processed image.
    uint64_t ipm_buf_addr;      //!< Buffer address for processed image in image memory.

    int      bb_index;          //!< Bounce buffer index.
    uint64_t bb_addr;           //!< Address for the bounce buffer.
    uint8_t  *bb_ptr;           //!< MMAPped pointer to the bounce buffer.
};

//--------------------------------------------------------------------
//! @brief The fsp_worker class.
//!
//! This class is a singleton threaded object that does the actual work of saving
//! image data across the network. For saving to the local device(s), the
//! imgSaveWorker class is used.
//--------------------------------------------------------------------
class fspWorker final : public FAS_THREAD::Thread
{
private:
    static std::unique_ptr<fspWorker> instance;
    static std::mutex instance_mutex;

private:
    fspWorker();

public:
    ~fspWorker();

    // Singletons should not be cloneable.
    fspWorker(fspWorker& other) = delete;

    // Singletons should not be assignable.
    void operator=(const fspWorker&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static fspWorker *factory();
    static bool isRunning()
    {
        std::lock_guard<std::mutex> lock(instance_mutex);
        return (instance.get() != nullptr);
    }

    void run() override;
    void stop() override;

    int setMode(int mode, bool wait_for_change = true);
    int startThread(int mode);

    bool metadataAbortCheck();
    constexpr bool modeChangePending()          { return (requested_mode != -1); }

    constexpr int getFrameNum()                 { return ipm_info.frame_num; }
    uint32_t    getIpAddr(char *dest_ip_addr);
    bool getRawLiveFrame(uint8_t *imageData, int& imageLength);
    constexpr int getMode()                     { return current_mode; }
    uint16_t    getPort();
    constexpr uint32_t getRetryIndex()          { return retry_index; }
    constexpr int getRunState()                 { return run_state; }
    constexpr int32_t getStreamPacketSize()     { return stream_packet_size; }

    void setPacketInfo(int32_t packet_size, bool dnf);
    void setPort(uint32_t new_port);
    void setRetryIndex(uint32_t index)     { retry_index = index; }
    void setStreamPath(const char *file_path, uint32_t file_len);

    void disconnectStream();
    int fireTestPacket(uint16_t packet_size);
    void pauseStream();
    void resumeStream();

private:
    static int  direct_walk_cb(const char *fpath, const struct stat *sb, int typeflag, struct FTW *ftwbuf);

    int capture_live_frame();
    int do_download_loop();
    int do_file_stream();
    int do_live_loop();
    int do_playback_loop();
    int do_raw_loop();
    int do_session_stream();
    int do_singlestep_loop();
    int get_live_frame();
    int encode_frame(uint8_t *image_ptr, JPEG_PROC::BUF_REQUEST *request);
    int ltr_frame_load(uint32_t frame_num);

    void pb_setup();
    bool sess_setup(uint64_t& xfer_length, uint64_t& im_addr, uint32_t& frame_count, uint32_t& session_length, uint32_t& dma_mod, bool& subpartition, bool& split_transfer);

    int ss_check_mode(bool ltr_mode, bool& pb_done, int& delay_msecs);
    int ss_process_frame(JPEG_PROC::IPM_REQUEST *request, bool ltr_mode);
    int ss_setup(bool ltr_mode, uint32_t& dma_mod, uint32_t& proc_info);

    int open_file_for_xfer(const char *file_path, off_t *file_len);

    void cleanup();
    void stream_work();

    uint32_t assemble_dng(uint8_t *hdr_ptr, int hdr_length, uint8_t *raw_ptr, int frame_size);
    uint32_t expand_the_pixels(pm_metadata *pm_meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr);
    uint32_t expand_10_bits(pm_metadata *meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr);
    uint32_t expand_12_bits(pm_metadata *meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr);
    uint32_t parse_id(const char *file_path);
    int  raw_metadata_stream();

#ifdef DNG_LOAD
    uint32_t read_as_dng(uint32_t frame_num);
#endif /* DNG_LOAD */

    void download_init(bool ltr_mode, uint32_t *proc_info_ptr, uint32_t *dma_mod_ptr);
    int ipm_frame(uint32_t proc_info, bool do_live = false);
    bool stream_single_file(const char *file_path);

    void connect_socket();
    int wait_new_mode_or_timeout(uint32_t msecs);

private:
    volatile bool thread_running;           //!< Flag to indicate if we are running or not.
    std::mutex thread_mutex;                //!< Used to control changes to IPM status.
    std::condition_variable thread_status;  //!< Condition variable we can block on.
    std::mutex mode_mutex;                  //!< Used to control changes in modes.
    std::condition_variable mode_status;    //!< Condition variable we can block on.

    std::mutex last_mutex;                  //!< Used to control reading the last live frame.
    std::condition_variable last_ready;     //!< Condition variable we can block on.
    uint8_t *last_ptr;                      //!< Where to put the last frame of data.
    int last_len;                           //!< The length of the last frame (JPEG encoded).

    int         current_mode;               //!< Current mode.
    int         initial_mode;               //!< Mode at entry when thread is started.
    int         requested_mode;             //!< -1 once mode request has been seen and it's safe to request another.
    uint32_t    retry_index;                //!< Which file to send (0 for all). Typically used for retrying.
    int         run_state;

    CsFSP fsp_socket;                       //!< FSP UDP socket.
    ipm_state ipm_info;                     //!< IPM information object.
    std::unique_ptr<DcrawUtil> dcraw_ptr;   //!< Our DCRAW processing object.

    int32_t stream_packet_size;             //!< Number of bytes to stream, per packet.

    char dest_ip_addr[CS_NETWORK::NET_ADDR_LEN];      //!< IP address to send streaming data to.
    uint16_t dest_port_num;                 //!< Port number to send streaming data to.

    std::unique_ptr<FAS_VIDEO::HEADER_DATA> dnld_header_buffer;  //!< The file header information for the current download operation.

    int ltr_bb_index;                       //!< Bounce buffer index for FPGA processing (LR mode).
    uint64_t ltr_bb_addr;                   //!< The address for the LTR bounce buffer (LR mode).

    std::string stream_file;                //!< Fully qualified name to source file.

    uint32_t aligned_len;                   //!< Length (on 32-bit boundary) of the file we're transferring.
    uint8_t *bounce_base;                   //!< Allocated memory for our bounce buffer (if needed).

    Fastec::IntervalTimer pb_timer;         //!< Timer for statistics during playback.

    std::vector<std::string> walk_paths;
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

std::unique_ptr<fspWorker> fspWorker::instance;
std::mutex fspWorker::instance_mutex;

// Make things easier for us.
using namespace FASTEC_STREAM;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

static int32_t  get_end_frame_number();
static uint32_t get_frame_interval();
static uint32_t get_interframe_delay();
static uint32_t get_packet_delay();
static int32_t  get_seq_start_frame(int streaming_mode);
static int32_t  get_start_frame_number();
static bool playback_forward();

//--------------------------------------------------------------------
// cs_fsp class implementation
//--------------------------------------------------------------------

/*!
 * @brief The constructor for the cs_fsp class.
 */
CsFSP::CsFSP() :
    CsNetwork()
{
    dnf_flag = false;
    stream_pause = false;
}

CsFSP::~CsFSP()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Tell our worker thread to stop if it is running.
    fsp_exit();

    // Make sure our streaming socket is closed.
    close();

    // Close down the listening socket.
    listen_sock.reset();
}

/*!
 * @brief Initialize the streaming socket.
 *
 * @return The open socket handle if sucessful, negative numbers are errors
 *     and should be considered fatal.
 */
int CsFSP::setup()
{
    uint8_t *mem_ptr;                   // Aligned memory pointer.

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Are we already good to go?
    if (isListenValid())
    {
        fi_log(LOG_DEBUG, "%s: Already setup.", __PRETTY_FUNCTION__);
        return CS_NET_SUCCESS;
    }

    // Create the handle for the FSP streaming socket.
    listen_sock = std::make_unique<CsSocket>(AF_INET, SOCK_STREAM);
    if (!isListenValid())
    {
        fi_log(LOG_ERR, "%s: Error creating FSP socket, errno:%d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
        return CS_NET_ERROR_SOCKET;
    }

    // Bind the socket.
    if (!listen_sock->bind(INADDR_ANY, GIGE_FSP_UDP_PORT))
    {
        fi_log(LOG_ERR, "%s: Error binding to FSP socket using port %d", __PRETTY_FUNCTION__, GIGE_FSP_UDP_PORT);
        listen_sock.reset();
        close();
        return CS_NET_ERROR_BIND;
    }

    // Indicate that we want to accept connections. We'll do the actual
    // "accept" in waitForClient(). The '10' is the "listen backlog" value.
    // In reality, we only accept one connection at a time...
    if (!listen_sock->listen())
    {
        fi_log(LOG_ERR, "%s: Error listening to FSP socket", __PRETTY_FUNCTION__);
        listen_sock.reset();
        close();
        return CS_NET_ERROR_SOCKET;
    }

    // Allocate our buffer for sending test packets. These packets are used
    // to negotiate how big our payload can be when sending to the client.
    if (posix_memalign((void **)&mem_ptr, ALLOCATION_UNIT, 9000) != 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        close();
        return CS_NET_INVALID_PARAM;
    }

    // Assign the pointer to our smart pointer.
    test_buffer.reset(mem_ptr);

    // Initialize our streaming variables.
    fspBlockId = 0;
    fspFrameNum = 0;
    fspPacketId.word = 0;

    // Success!
    return CS_NET_SUCCESS;
}

/*!
 * @brief Close down the streaming socket to make sure everything is cleaned up
 *    before shutting down.
 */
void CsFSP::close()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Close the streaming socket. The "listening" socket
    // will remain open, waiting for connections.
    CsNetwork::close();
}

/*!
 * @brief Set the UDP socket to either do not fragment or allow fragmentation.
 *
 * @param set If true, sets the do-not-fragment flag, if false, set to allow fragment.
 */
void CsFSP::setDNF(bool set)
{
    TP(FSP_FUNC, "%s set=%d", __PRETTY_FUNCTION__, set);

    // Remember the flag.
    dnf_flag = set;
}

void CsFSP::checkPause()
{
    fspWorker *fsp = fspWorker::factory();

    std::unique_lock<std::mutex> locker(pause_mutex);

    // Are we supposed to pause?
    while (stream_pause && !fsp->modeChangePending())
    {
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;

        // Check the mode...don't pause if we're not streaming.
        int mode = fsp->getMode();
        if (mode == GIGE_STREAM_STOP)
        {
            fi_log(LOG_DEBUG, "%s: mode is stop. Not going to wait on pause.", __PRETTY_FUNCTION__);
            break;
        }

        TP(FSP_STATE, "%s: mode=0x%08x, pausing...", __PRETTY_FUNCTION__, mode);
        if (pause_status.wait_until(locker, now + ms(50), [this] {return stream_pause == false;}))
        {
            TP(FSP_STATE, "%s: Resume", __PRETTY_FUNCTION__);
            break;
        }
    }
}

void CsFSP::pause()
{
    std::unique_lock<std::mutex> locker(pause_mutex);

    TP(FSP_STATE, "%s", __PRETTY_FUNCTION__);

    // Set the paused flag.
    stream_pause = true;
}

void CsFSP::resume()
{
    std::unique_lock<std::mutex> locker(pause_mutex);

    TP(FSP_STATE, "%s", __PRETTY_FUNCTION__);

    // Clear the pause flag.
    stream_pause = false;

    // Wake up anyone waiting.
    pause_status.notify_one();
}

/*!
 * @brief When streaming from a file descriptor, we can make use of the optimized
 *        sendfile() c-library call. To send a memory buffer, use the cs_fsp:stream_data
 *        method.
 *
 * @param file_fd       [IN] File descriptor to read from.
 * @param offset        [IN] Offset into the file to start.
 * @param data_size     [IN] Number of bytres to stream.
 * @return Number of bytes sent.
 */
size_t CsFSP::sendfile(const int file_fd, off_t offset, size_t data_size)
{
    int32_t i;
    size_t amt_sent;
    int32_t chunk_count;
    int32_t xfer_buffer_len;
    int32_t last_chunk_len;

    TP(FSP_FUNC, "%s: offset=0x%" PRIx64", data_size=%" PRId64, __PRETTY_FUNCTION__, offset, data_size);

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: No socket!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Do we need to chunk this?
    if (data_size < (size_t)DL_CHUNK_SIZE)
    {
        // No chunking required.
        chunk_count = 1;

        // Set our transfer length.
        last_chunk_len = xfer_buffer_len = data_size;
    }
    else
    {
        // How many "chunks" do we need?
        chunk_count = data_size / DL_CHUNK_SIZE;
        last_chunk_len = (data_size % DL_CHUNK_SIZE);
        if (last_chunk_len != 0)
            ++chunk_count;
        else
            last_chunk_len = DL_CHUNK_SIZE;

        // Set our transfer length.
        xfer_buffer_len = DL_CHUNK_SIZE;
    }

    // Initialize our total sent value.
    amt_sent = 0;

    // Loop through all the chunks.
    fspWorker *fsp = fspWorker::factory();
    for (i = 0; i < chunk_count; ++i)
    {
        int cork_state;
        int32_t remainder;

        // Are we supposed to stop?
        if (fsp->modeChangePending())
        {
            LOG("%s#%d: stop, mode change pending", __PRETTY_FUNCTION__, __LINE__);
            return 0;
        }

        // When we are sending a file in multiple chunks, the frame
        // number is expected to start at '1' and increment.
        if (chunk_count > 1)
            setFrameNum(i + 1);

        // Special case the last chunk.
        if (i == (chunk_count - 1))
            remainder = last_chunk_len;
        else
            remainder = xfer_buffer_len;

        TP(FSP_FILESTREAM, "%s: chunk=%d, len=%d", __PRETTY_FUNCTION__, i, remainder);

        // Protect the transfer of the 'chunk' against reentrancy.
        {
            std::lock_guard<std::mutex> lock(serial_stream);

            // Send the leader for the file.
            if (!sendLeader(remainder, FSP_PAYLOAD_FILE))
            {
                fi_log(LOG_DEBUG, "%s: Unable to transmit the leader packet!", __PRETTY_FUNCTION__);
                return 0;
            }

            // Enable TCP_CORK to maximize payload usage.
            cork_state = 1;
            ::setsockopt(sock->getFD(), IPPROTO_TCP, TCP_CORK, &cork_state, sizeof(cork_state));

            // Loop until we get the entire bit sent.
            while (remainder > 0)
            {
                int send_len;

                // See if we can stream.
                checkPause();

                // Send the amount.
                send_len = ::sendfile(sock->getFD(), file_fd, &offset, remainder);
                if (send_len < 0)
                {
                    FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
                    break;
                }

                // Update our variables.
                remainder -= send_len;
                amt_sent += send_len;
            }

            // Disable TCP_CORK.
            cork_state = 0;
            ::setsockopt(sock->getFD(), IPPROTO_TCP, TCP_CORK, &cork_state, sizeof(cork_state));

            // Send the trailer for the chunk.
            if (!sendTrailer())
            {
                fi_log(LOG_DEBUG, "%s: Unable to transmit the trailer packet!", __PRETTY_FUNCTION__);
                return 0;
            }
        }
    }

    // Return the number of bytes sent.
    return amt_sent;
}

/*!
 * @brief Unlike cs_fsp::stream_file, this method streams data from a memory buffer and, optionally, a separate header buffer.
 *
 * @param data_ptr      [IN] Pointer to the buffer we are going to stream.
 * @param data_size     [IN] Number of bytes to stream.
 * @param payload_type  [IN] What kind of stream are we going to send. Allowed are FSP_PAYLOAD_FILE and FSP_PAYLOAD_RAW.
 * @param head_ptr      [IN, optional] Pointer to a separate header buffer.
 * @param head_size     [IN, optional] Number of bytes from the head_ptr to stream.
 * @return Number of bytes streamed.
 */
int32_t CsFSP::streamData(const uint8_t *data_ptr, int32_t data_size, uint16_t payload_type, const uint8_t *head_ptr, int32_t head_size)
{
    TP(FSP_FUNC, "%s: data_size=%d, head_size=%d", __PRETTY_FUNCTION__, data_size, head_size);

    std::lock_guard<std::mutex> lock(serial_stream);

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: Socket is not valid!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Stream the data.
    if ((sock->getFD() < 0) || !sendLeader(data_size + head_size, payload_type))
    {
        fi_log(LOG_DEBUG, "%s: Unable to transmit the leader packet!", __PRETTY_FUNCTION__);
        return 0;
    }
    if ((sock->getFD() < 0) || !sendData(data_ptr, data_size, head_ptr, head_size))
    {
        fi_log(LOG_DEBUG, "%s: Unable to transmit the data packets!", __PRETTY_FUNCTION__);
        return 0;
    }
    if ((sock->getFD() < 0) || !sendTrailer())
    {
        fi_log(LOG_DEBUG, "%s: Unable to transmit the trailer packet!", __PRETTY_FUNCTION__);
        return 0;
    }

    return (data_size + head_size);
}

int32_t CsFSP::testFire(int32_t packet_size)
{
    int32_t rc = -1;

    TP(FSP_FUNC, "%s: len=%d", __PRETTY_FUNCTION__, packet_size);

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: Socket is not valid!", __PRETTY_FUNCTION__);
        goto tf_bail;
    }

    if (packet_size > 0)
    {
        // Initialize the packet with some data.
        memset((void *)test_buffer.get(), 0xAB, packet_size);

        // Send the test packet.
        rc = write(sock->getFD(), test_buffer.get(), packet_size);
    }

tf_bail:
    return rc;
}

/*!
 * @brief This function builds and sends a LEADER packet of the appropriate type,
 *        as specified by the 'payload_type' parameter.
 *
 * @param data_size     [IN] Number of bytes in the payload to be sent.
 * @param payload_type  [IN] Type of payload. Acceptable values are FSP_PAYLOAD_FILE and FSP_PAYLOAD_RAW.
 * @return True on success, false otherwise.
 */
bool CsFSP::sendLeader(int32_t data_size, uint16_t payload_type)
{
    bool bOK = true;        // Assume success.
    fsp_leader_file pkt_file;   // Packet structure for sending a file leader.
    fsp_leader_raw pkt_live;    // Packet structure if we're sending a live view image.
    const uint8_t *data_ptr;
    int32_t remainder;

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: Socket is not valid!", __PRETTY_FUNCTION__);
        return false;
    }

    // Increment our block ID.
    fspBlockId++;
    if (fspBlockId == 0)
        fspBlockId = 1;

    TP(FSP_FUNC, "%s: blockID=%d, data_size=0x%08x", __PRETTY_FUNCTION__, fspBlockId, data_size);

    // First data packet will have this ID
    fspPacketId.word = 1;

    // Populate the appropriate leader.
    if (payload_type == FSP_PAYLOAD_FILE)
    {
        // Set up the format independent fields
        memset((void *)&pkt_file, 0, sizeof(pkt_file));
        pkt_file.leader.header.block_id   = htons(fspBlockId);
        pkt_file.leader.header.pkt_fmt    = FSP_PKT_FMT_LEADER;
        pkt_file.leader.header.status     = htons(GEV_STATUS_SUCCESS);
        pkt_file.leader.magic_id          = htons(FSP_FASTEC_ID);
        pkt_file.leader.payload_type      = htons(FSP_PAYLOAD_FILE);
        // pkt_file.leader.timestamp_high = 0 (not used; cleared above)
        // pkt_file.leader.timestamp_low  = 0 (not used; cleared above)

        // Fill in the format specific field.
        pkt_file.ps_high                  = 0;
        pkt_file.ps_low                   = htonl((uint32_t)(data_size));

        // If we're in live view, then the sequence number (aka "file_name") should
        // be 0. Otherwise it is the "frame_number + 1".
        //
        // ALM_TODO_344: Streaming files should no longer happen in "live". This should
        //               only be for streams OTHER than "live".
        switch (fspWorker::factory()->getMode())
        {
        case GIGE_STREAM_LIVE:
            snprintf(pkt_file.file_name,  FSP_LEADER_FILE_NAME_LEN - 1, "%08X", 0);
            break;
        case GIGE_STREAM_RAW_METADATA:
            snprintf(pkt_file.file_name,  FSP_LEADER_FILE_NAME_LEN - 1, "%08X", METADATA_RAW_SEQ_NUM);
            break;
        default:
            snprintf(pkt_file.file_name,  FSP_LEADER_FILE_NAME_LEN - 1, "%08X", fspFrameNum);
            break;
        }

        // Setup a byte-wise data pointer to send the packet.
        data_ptr = (const uint8_t *)&pkt_file;
        remainder = sizeof(pkt_file);
    }
    else if (payload_type == FSP_PAYLOAD_RAW)
    {
        // Streaming a "live" frame we use the RAW payload type. Build up the
        // header.
        memset((void *)&pkt_live, 0, sizeof(pkt_live));
        pkt_live.leader.header.block_id    = htons(fspBlockId);
        pkt_live.leader.header.pkt_fmt     = FSP_PKT_FMT_LEADER;
        pkt_live.leader.header.status      = htons(GEV_STATUS_SUCCESS);
        pkt_live.leader.magic_id           = htons(FSP_FASTEC_ID);
        pkt_live.leader.payload_type       = htons(FSP_PAYLOAD_RAW);
        // pkt_live.leader.timestamp_high  = 0 (not used; cleared above)
        // pkt_live.leader.timestamp_low   = 0 (not used; cleared above)

        // Fill in the format specific field.
        pkt_live.data_size_high = 0;
        pkt_live.data_size_low = htonl((uint32_t)(data_size));

        // Setup a byte-wise data pointer to send the packet.
        data_ptr = (const uint8_t *)&pkt_live;
        remainder = sizeof(pkt_live);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: Invalid payload_type [0x%04x]!", __PRETTY_FUNCTION__, payload_type);
        return false;
    }

    // Loop through the data buffer.
    while (remainder > 0)
    {
        int send_len;

        // See if we can stream.
        checkPause();

        // Send the packet.
        send_len = write(sock->getFD(), data_ptr, remainder);
        if (send_len < 0)
        {
            FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
            bOK = false;
            break;
        }

        // Update our variables.
        remainder -= send_len;
        data_ptr += send_len;
    }

    // Return our status.
    return bOK;
}

bool CsFSP::sendTrailer()
{
    bool bOK = true;            // Assume success.
    fsp_trailer pkt;        // Packet structure for the trailer.
    const uint8_t *data_ptr;
    int32_t remainder;

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: Socket is not valid!", __PRETTY_FUNCTION__);
        return false;
    }

    TP(FSP_FUNC, "%s: blockID=%d, pkt_id=%d", __PRETTY_FUNCTION__, fspBlockId, fspPacketId.word);

    memset((void *)&pkt, 0, sizeof(pkt));
    pkt.header.block_id   = htons(fspBlockId);
    pkt.header.pkt_fmt    = FSP_PKT_FMT_TRAILER;
    pkt.header.pkt_id[2]  = fspPacketId.byte[0];    // Make sure the bytes are in network order.
    pkt.header.pkt_id[1]  = fspPacketId.byte[1];
    pkt.header.pkt_id[0]  = fspPacketId.byte[2];
    pkt.header.status     = htons(GEV_STATUS_SUCCESS);
    pkt.payload_type      = htons(FSP_PAYLOAD_FILE);

    // Setup a byte-wise data pointer to send the packet.
    data_ptr = (const uint8_t *)&pkt;
    remainder = sizeof(pkt);

    // Loop through the data buffer.
    while (remainder > 0)
    {
        int send_len;

        // See if we can stream.
        checkPause();

        // Send the packet.
        send_len = write(sock->getFD(), data_ptr, remainder);
        if (send_len < 0)
        {
            FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
            bOK = false;
            break;
        }

        // Update our variables.
        remainder -= send_len;
        data_ptr += send_len;
    }

    // Return our status.
    return bOK;
}

bool CsFSP::sendData(const uint8_t *body_ptr, int32_t body_size, const uint8_t *head_ptr, int32_t head_size)
{
    bool bOK = true;        // Assume success.
    const uint8_t *data_ptr;
    int32_t remainder;
    bool head_transfer;

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: Socket is not valid!", __PRETTY_FUNCTION__);
        return false;
    }

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Do we have a header?
    if (head_ptr == nullptr)
    {
        // Go straight to the data.
        data_ptr = body_ptr;
        remainder = body_size;
        head_transfer = false;
    }
    else
    {
        // We do have an image header, so transfer that first.
        data_ptr = head_ptr;
        remainder = head_size;
        head_transfer = true;
    }

    // Loop through the data buffer.
    while (remainder > 0)
    {
        int send_len;
        extern bool sigpipe_caught;

        // See if we can stream.
        checkPause();

        // Send the packet.
        send_len = write(sock->getFD(), data_ptr, remainder);
        if ((send_len < 0) || sigpipe_caught)
        {
            FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
            bOK = false;
            break;
        }

        // Update our variables.
        remainder -= send_len;
        data_ptr += send_len;

        // Are we dealing with a header, and did we just send it?
        if (head_transfer && (remainder == 0))
        {
            // Switch over to the data part.
            data_ptr = body_ptr;
            remainder = body_size;
            head_transfer = false;
        }
    }

    // Return our status.
    return bOK;
}

bool CsFSP::waitForClient(uint32_t msecs)
{
    int rc;
    int max_fd;
    struct timeval timeout;
    fd_set master_set;
    extern bool sigpipe_caught;

    // Sanity check.
    if (!isListenValid())
    {
        fi_log(LOG_DEBUG, "%s: No socket!?", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize the master fd_set.
    FD_ZERO(&master_set);
    max_fd = listen_sock->getFD();
    FD_SET(listen_sock->getFD(), &master_set);

    // Initialize our timeout.
    timeout.tv_sec = 0;
    timeout.tv_usec = msecs * 1000; // Convert millisec to microsec.

    // Now wait for something to happen.
    rc = select(max_fd + 1, &master_set, NULL, NULL, &timeout);
    if (rc < 0)
    {
        FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
        return false;
    }
    if (rc == 0)
    {
        // Got a timeout.
        return false;
    }

    // Reset our SIGPIPE flag.
    sigpipe_caught = false;

    // We must have an accept pending on our socket.
    sock = listen_sock->accept();
    if ((sock.get() == nullptr) || (sock->getFD() < 0))
    {
        fi_log(LOG_ERR, "%s: accept() failed", __PRETTY_FUNCTION__);
        return false;
    }

    // All is well.
    return true;
}

//--------------------------------------------------------------------
// fsp_worker class implementation
//--------------------------------------------------------------------

#ifdef ULV_BUFFER_INIT
static bool fill_memory(uint64_t start_addr, uint32_t pattern, uint64_t length, int control_type = AUX_DMA_PATTERN_VAL)
{
    bool rc = false;
    uint32_t reg_value;
    int retries;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: start_addr=0x%" PRIx64 ", pattern=0x%08x, length=%" PRIu64, __PRETTY_FUNCTION__, start_addr, pattern, length);

    if (!fpga_ptr->read_reg(IM_AUX_WR_DMA0_STATUS, &reg_value))
        goto bail;

    if ((reg_value & AUX_DMA_BUSY) != 0)
    {
        printf("AUX WR DMA is busy?\n");
        goto bail;
    }

    // Initialize DMA parameters.
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_BP_THRESHOLDS, 0x8004))
        goto bail;
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_BURST_INTERVAL, 0))
        goto bail;

    // Write controls with enables cleared.
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_CONTROL, control_type))
        goto bail;

    // Tell the IM_AUX_WR_DMA0 to write the pattern to memory
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_PATTERN_VAL, pattern))
        goto bail;
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_START_ADDR, convert_to_HEXWORD(start_addr)))
        goto bail;
    if (!fpga_ptr->write_reg(IM_AUX_WR_DMA0_BLOCK_SIZE, convert_to_HEXWORD(length)))
        goto bail;

    // Start the DMA.
    if (!fpga_ptr->modify_reg(IM_AUX_WR_DMA0_CONTROL, AUX_DMA_ENABLE | AUX_DMA_START, ~(AUX_DMA_ENABLE | AUX_DMA_START)))
        goto bail;

    // Wait for write done.
    for (retries = 0; retries < 100; ++retries)
    {
        if (!fpga_ptr->read_reg(IM_AUX_WR_DMA0_STATUS, &reg_value))
            goto bail;

        if ((reg_value & AUX_DMA_BUSY) == 0)
        {
            rc = true;
            break;
        }

        // Wait for a bit.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(100));
    }
    if (!rc)
        fi_log(LOG_ERR, "%s: DMA timed out", __PRETTY_FUNCTION__);

bail:
    return rc;
}
#endif /* ULV_BUFFER_INIT */

/*!
 * @brief Return the begining frame number we want to stream.
 */
static int32_t get_start_frame_number()
{
    int32_t   start_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    start_frame = regs[GIGE_REG_STREAM_START_NUM].get_value();
    return start_frame;
}

/*!
 * @brief Return true if the IPM (specifically P1) is busy.
 */
static bool ipm_is_busy()
{
    uint32_t    status;
    FiFPGA  *fpga_ptr;

    // Get our FPGA object.
    fpga_ptr = CameraServer::get_fpga();

    // Read the status.
    if (!fpga_ptr->regRead(IPM_P1_STATUS, &status))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read IPM_P1_STATUS", __PRETTY_FUNCTION__);
        return false;
    }

    // Return the status for the IPM.
    return ((status & (uint32_t)IPMS_IPM_BUSY) == (uint32_t)IPMS_IPM_BUSY);
}

/*!
 * @brief Return the ending frame number we want to stream.
 */
static int32_t get_end_frame_number()
{
    int32_t   end_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    end_frame = regs[GIGE_REG_STREAM_END_NUM].get_value();
    return end_frame;
}

/*!
 * @brief Return the starting frame number, between GIGE_REG_STREAM_START_NUM
 *    and GIGE_REG_STREAM_END_NUM.
 */
static int32_t get_seq_start_frame(int streaming_mode)
{
    int32_t seq_start;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    seq_start = regs[GIGE_REG_STREAM_SEQ_START].get_value();

    if (streaming_mode != GIGE_STREAM_SINGLESTEP)
    {
        int32_t start_frame;
        int32_t end_frame;

        // Make sure we're in range.
        start_frame = get_start_frame_number();
        end_frame = get_end_frame_number();
        if (seq_start < start_frame)
        {
            seq_start = start_frame;
        }
        else if (seq_start > end_frame)
        {
            seq_start = end_frame;
        }
    }

    // Return our frame number.
    TP(FSP_PLAYBACK, "%s: seq_start=%u", __PRETTY_FUNCTION__, seq_start);
    return seq_start;
}

/*!
 * @brief Return the interval to skip frames. A '1' means to play every frame,
 *    whereas a '10' would mean to play every 10th frame (i.e. 0, 10, 20...).
 */
static uint32_t get_frame_interval()
{
    uint32_t   frame_interval;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    frame_interval = regs[GIGE_REG_STREAM_FRAME_INTERVAL].get_value() & GIGE_STREAM_INTERVAL_MASK;
    return frame_interval;
}

/*!
 * @brief Return the direction the user wants to "play back" the data.
 *
 * @return true if we are playing forward, false for reverse.
 */
static bool playback_forward()
{
    uint32_t   direction;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    direction = regs[GIGE_REG_STREAM_FRAME_INTERVAL].get_value() & GIGE_STREAM_REVERSE;
    return (direction == GIGE_STREAM_FORWARD);
}

/*!
 * @brief Return the number of milliseconds we should delay between frames during playback.
 */
static uint32_t get_interframe_delay()
{
    uint32_t   rate;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    rate = regs[GIGE_REG_STREAM_FRAME_DELAY].get_value();
    return rate;
}

/*!
 * @brief return the type of image we should be transmitting ending frame number we want to stream.
 */
static uint32_t get_stream_format()
{
    uint32_t   image_type;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    image_type = regs[GIGE_REG_STREAM_FORMAT].get_value();
    return image_type;
}

/*!
 * @brief Return the packet delay from the correct register
 */
static uint32_t get_packet_delay()
{
    uint32_t  packet_delay;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Get the packet delay from the SCPD register.
    //
    // FORNOW - If we are going to support a timestamp then we need to support the
    // Timestamp registers 0x944, 0x948, an 0x94c
    packet_delay = regs[GIGE_REG_SCPD0].get_value();
    return packet_delay;
}

/*!
 * @brief The FSP worker thread constructor.  The worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
fspWorker::fspWorker() :
    FAS_THREAD::Thread(true),
    last_ptr(nullptr),
    last_len(0),
    run_state(0),
    dcraw_ptr(nullptr),
    stream_packet_size(0),
    ltr_bb_index(0),
    ltr_bb_addr(0),
    aligned_len(0)
{
    // Initialize our member variables.
    thread_running = false;
    initial_mode = current_mode = -1;
    requested_mode = -1;
    dest_ip_addr[ 0] = '\0';
    dest_port_num = 0;
    ipm_info.busy = false;
    ipm_info.frame_counter = 0;
    bounce_base = nullptr;
    retry_index = 0;
    stream_file.clear();

    // Start the thread running.
    start();
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 *
 * @param arg   Not used.
 */
fspWorker::~fspWorker()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Tell the streaming thread to stop.
    stop_request = true;
}

fspWorker *fspWorker::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if (instance.get() == nullptr)
    {
        instance = std::unique_ptr<fspWorker>(new fspWorker());
    }

    return instance.get();
}

int fspWorker::startThread(int mode)
{
    int rc = GEV_STATUS_SUCCESS;

    TP(FSP_FUNC, "%s: mode=%d", __PRETTY_FUNCTION__, mode);

    // Setup our mode information.
    current_mode = -1;
    initial_mode = mode;

    // Now start the thread.
    start();

    {
        // Wait for the thread to start or 50 milliseconds. Put this in it's own scope
        // and the locker will release on the exit of the scope. The timeout is short
        // enough that we should be able to pick up a mode change just fine.
        std::unique_lock<std::mutex> locker(thread_mutex);
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (thread_status.wait_until(locker, now + ms(50)) == std::cv_status::timeout)
        {
            fi_log(LOG_ERR, "%s: timeout waiting for thread start", __PRETTY_FUNCTION__);
            rc = GEV_STATUS_ERROR;
        }
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Ask the thread to stop.
 */
void fspWorker::stop()
{
    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Tell the streaming thread to stop.
    stop_request = true;
}

/*!
 * @brief Have the FPGA capture a single frame. Wait until the
 *    FPGA is finished, or we timeout.
 *
 * @return 0 on success, -1 on failure.
 */
int fspWorker::capture_live_frame()
{
    int i;
    int rc = -1;
    uint32_t reg_value;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

#ifdef ULV_BUFFER_INIT
    // Clear the ULV buffer.
    fill_memory(pm_ptr->get_ulv_buf_addr(), 0, pm_ptr->get_ulv_buf_size());
#endif /* ULV_BUFFER_INIT */

    // Check to see if the ICM is in IDLE mode - possible if we've bailed out of
    // capture or something earlier.
    if (fpga_ptr->icmIsPipelineIdle(nullptr))
    {
        fi_log(LOG_ERR, "%s: The ICM is idle!", __PRETTY_FUNCTION__);
        goto clf_bail;
    }

    // Tell the ICM to grab a frame.
    if (!fpga_ptr->regModify(ICM_CONTROL, ICMC_CAPTURE_SINGLE, -1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_CONTROL!", __PRETTY_FUNCTION__);
        goto clf_bail;
    }

    // Wait for the capture to complete.
    for (i = 0; i < 1000; i++)
    {
        // Wait for a bit.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(1));

        // Check to see if the ICM has finished.
        if (!fpga_ptr->regRead(ICM_STATUS, &reg_value))
        {
            fi_log(LOG_DEBUG, "%s: read_reg(ICM_STATUS) failed!", __PRETTY_FUNCTION__);
            goto clf_bail;
        }

        // Is the ICM still busy?
        if ((reg_value & ICMS_CAPTURE_SINGLE_BUSY) == 0)
            break;
    }

    if (i < 1000)
    {
        // Everything is OK.
        rc = 0;
        TP(FSP_STATS, "%s: Capture in %d msec", __PRETTY_FUNCTION__, i);
    }
    else
    {
//        fi_log(LOG_DEBUG, "%s: Capture took too long! [ICM_STATUS=0x%08x]", __PRETTY_FUNCTION__, reg_value);
        goto clf_bail;
    }

    // The FPGA always puts the single capture into the ULV buffer.
    ipm_info.frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_ulv_buf_addr());
    ipm_info.image_len = (pm_ptr->getFrameSize() - FRAME_METADATA_HW_BYTES);

clf_bail:
    return rc;
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the fsp_worker class object.
 */
void fspWorker::cleanup()
{
    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);
    LOG("%s", __PRETTY_FUNCTION__);

    // Shutdown our socket.
    fsp_socket.close();

    // We're not running anymore.
    thread_running = false;
}

void fspWorker::download_init(bool ltr_mode, uint32_t *proc_info_ptr, uint32_t *dma_mod_ptr)
{
    bool is_color;
    bitmode_t bitmode;
    uint32_t stream_format;
    uint32_t proc_info;
    uint32_t dma_mod;
    uint32_t format_type;
    std::unique_ptr<pm_metadata> pm_meta_ptr;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FSP_DOWNLOAD, "%s", __PRETTY_FUNCTION__);

    // Initialize our control variables, so we know what's going on.
    pm_meta_ptr = pm_ptr->getActivePartMetadata();
    stream_format = get_stream_format();
    format_type = stream_format & GIGE_STREAM_FORMAT_MASK;
    bitmode = regs[GIGE_REG_BIT_MODE].get_value();
    is_color = sensor_ptr->is_sensor_color();

    // Get a buffer for our image processing.
    ipm_info.bb_index = bb_get_avail();
    ipm_info.bb_addr = bb_get_avail_addr(ipm_info.bb_index);

    // Map the memory.
    ipm_info.bb_ptr = (uint8_t *)bb_mmap(ipm_info.bb_index, 0);
    if (ipm_info.bb_ptr == (uint8_t *)MAP_FAILED)
    {
        FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
        return;
    }

    // Initialize our frame number and IPM buffer ID.
    ipm_info.frame_num = get_seq_start_frame(current_mode);

    // This processing info will be needed to set up the IPM.
    proc_info = get_imgproc_type(stream_format, bitmode, is_color);
    if ((proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_JPEG)
    {
        // Specify the more accurate processing type for JPEG.
        if (is_color)
            proc_info = IMGPROC_TYPE_NV12;
        else
            proc_info = IMGPROC_TYPE_Y8;
    }
    if ((regs[GIGE_REG_FPN].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        proc_info |= IMGPROC_USE_FPN;

    // If we're saving bitmaps, then we need to make allowances.
    if (format_type == GIGE_STREAM_FORMAT_BMP)
        proc_info |= IMGPROC_ROW_WD_ALIGN;

    // Get the header for our stream type.
    dnld_header_buffer = imgio::get_image_header(stream_format, pm_meta_ptr.get());
    if (dnld_header_buffer == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate the header buffer!", __PRETTY_FUNCTION__);
        return;
    }
    else if ((dnld_header_buffer->size() == 0) && (format_type != GIGE_STREAM_FORMAT_JPEG))
    {
        // Only JPEGs are allowed to not have a header.
        fi_log(LOG_ERR, "%s: BAD HEADER LENGTH [%d]!", __PRETTY_FUNCTION__, dnld_header_buffer->size());
        return;
    }

    // Get the DMA settings we want to use.
    dma_mod = get_stream_dma_mod(stream_format, bitmode, is_color);
    // dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;

    // Are we in LTR mode?
    if (ltr_mode)
    {
        // Get our buffer for the raw file heading up to the FPGA.
        ltr_bb_index = bb_get_avail();
        ltr_bb_addr = bb_get_avail_addr(ltr_bb_index);

        // Get a frame address in IM that we can use to store the raw frame.
        ipm_info.frame_addr_hw = convert_to_HEXWORD(pm_ptr->getCaptureFrameAddr(0));
        TP(FSP_PLAYBACK, "%s: LTR processing frame in IM, frame_addr_hw=0x%08x", __PRETTY_FUNCTION__, ipm_info.frame_addr_hw);
    }

    // Pass the values back up.
    *proc_info_ptr = proc_info;
    *dma_mod_ptr = dma_mod;
}

/*!
 * @brief Handle download of the images.
 *
 * In the download loop, we transfer the frames as quickly as we can.
 * The loop terminates when all the frames have transfered, or the user
 * has changed the mode (i.e. cancelled the transfer).
 *
 * @return 0 on success, or an error code otherwise.
 */
int fspWorker::do_download_loop()
{
    int rc = 0;
    int download_state;
    int frame_count(0);
    int transfer_count(0);
    bool done;
    bool ltr_mode;
    uint8_t *enc_buf(nullptr);
    int enc_size(0);
    uint32_t packed_size(0);
    uint32_t img_width(0);
    uint32_t img_height(0);
    uint32_t img_bpp(0);
    uint32_t dma_mod(0);
    uint32_t dma_dims(0);
    uint32_t proc_info(0);
    uint32_t pc_img_len(0);
    int ifg_millisec;

    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Fastec::IntervalTimer dnld_timer;
    Fastec::IntervalTimer frame_timer;

    // #define DO_STATISTICS 1
#ifdef DO_STATISTICS
    FILE *stat_fp = fopen("/tmp/fsp.csv", "a");
    std::vector<double> times_measure;  //!< An array of 'durations' for the various threads.
#endif /* DO_STATISTICS */

    TP(FSP_FUNC, "%s: pkt_delay=%u", __PRETTY_FUNCTION__, get_packet_delay());

    // Initialize our state variable.
    download_state = DNLD_SETUP;

    // Determine if we're in LTR mode or not.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);

    // Get the current time so we can calculate how much time the download takes.
    dnld_timer.start();

    // Keep going until told to stop.
    done = false;
    while (!done)
    {
        TP(FSP_STATE, "%s: state=%d", __PRETTY_FUNCTION__, download_state);

        // Process based on the current state.
        switch (download_state)
        {
        case DNLD_SETUP:
            // How many frames are we downloading.
            frame_count = (get_end_frame_number() - get_start_frame_number()) + 1;
            transfer_count = 0;

            // What is our interframe delay?
            ifg_millisec = get_interframe_delay();
            fi_log(LOG_DEBUG, "%s: ifg_millisec=%d", __PRETTY_FUNCTION__, ifg_millisec);

            // Processing info is always the same for playback images.
            download_init(ltr_mode, &proc_info, &dma_mod);

            // Get our image information.
            img_height = regs[GIGE_REG_HEIGHT].get_value();
            img_width = regs[GIGE_REG_WIDTH].get_value();
            img_bpp = regs[GIGE_REG_BIT_MODE].get_value();
            if (IS_8_BIT_MODE(img_bpp))
                img_bpp = 8;
            else if (IS_10_BIT_MODE(img_bpp))
                img_bpp = 10;
            else
                img_bpp = 12;

            // Need some special processing if we're doing a packed format.
            if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_PACK)
            {
                // Make sure we have access to the internal library.
                if (FiInternal::factory() != nullptr)
                {
                    // Calculate what the packed size is.
                    FiInternal::GET_PACKED_SIZE_FN sizePtr = FiInternal::factory()->getPackedSizeFuncPtr();
                    if (sizePtr != nullptr)
                    {
                        packed_size = (*sizePtr)(img_width, img_height, img_bpp);
                    }
                }
            }

            // How many bytes will the processed frame be in PC memory?
            pc_img_len = get_processed_size(proc_info, img_width, img_height);

            // The image dimensions will be constant.
            dma_dims = (img_height << IPM_FRAME_DIM_Y_SHIFT) | img_width;

            // Initialize pertinent registers incase someone looks.
            TP(FSP_DOWNLOAD, "%s: cutin=%u start=%d cutout=%u proc=0x%08x dma=0x%08x", __PRETTY_FUNCTION__, get_start_frame_number(), ipm_info.frame_num, get_end_frame_number(), proc_info, dma_mod);

            // Move on.
            if (ltr_mode)
                download_state = DNLD_LTR_FRAME_LOAD;
            else
                download_state = DNLD_PROCESS_FRAME;
            break;
        case DNLD_LTR_FRAME_LOAD:
            // Statistics
            frame_timer.start();
            // Statistics

            // Load the frame into the IM for FPGA processing.
            if (ltr_frame_load(ipm_info.frame_num) == 0)
                download_state = DNLD_PROCESS_FRAME;
            else
                download_state = DNLD_CHECK_MODE;

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame %u sent to FPGA: %lf secs.", __PRETTY_FUNCTION__, ipm_info.frame_num, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case DNLD_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // If we're not running in LTR mode, then we need to figure out our frame address in Image Memory.
            if (!ltr_mode)
            {
                // Get the address of the frame we're going to process.
                ipm_info.frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_active_frame_addr(ipm_info.frame_num, false));
            }
            TP(FSP_DOWNLOAD, "%s: frame_addr_hw=0x%08x", __PRETTY_FUNCTION__, ipm_info.frame_addr_hw);

            // Get the frame processed by the IPM.
            rc = ipm_frame(proc_info);
            if (rc == 0)
            {
                // Move on to transferring the frame.
                download_state = DNLD_TRANSFER_FRAME;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
                download_state = DNLD_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: IPM: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case DNLD_TRANSFER_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // memset(ipm_info.bb_ptr, 0, ipm_info.image_len);

            // Have the FPGA transfer the image into our memory.
            rc = fpga_ptr->imageDownload(dma_mod, dma_dims, ipm_info.ipm_buf_addr,
                                         ipm_info.bb_addr, ipm_info.image_len, pc_img_len);
            if (rc >= 0)
            {
                // Move on to processing the frame.
                download_state = DNLD_ENCODE_FRAME;
                rc = 0;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
                download_state = DNLD_CHECK_MODE;
            }

            // Release our P1 buffer.
            ipm_proc_release(ipm_info.ipm_buf_addr);
            ipm_info.ipm_buf_addr = 0;

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: DMA: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case DNLD_ENCODE_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Now that we have the frame in our system memory. Do we want to encode it?
            if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG)
            {
                JPEG_PROC::BUF_REQUEST request;

                // Initialize our request structure with what we know.
                if ((proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
                    request.pixel_format = JPEG_UTIL::NV12;
                else
                    request.pixel_format = JPEG_UTIL::Y8;
                request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();
                request.jpeg_buffer = nullptr;
                request.jpeg_len = 0;

                rc = encode_frame(ipm_info.bb_ptr, &request);
                if ((request.jpeg_len > 0) && (request.jpeg_buffer != nullptr))
                {
                    // Move on to streaming the frame.
                    download_state = DNLD_STREAM_FRAME;

                    // Save the encoded information.
                    enc_buf = request.jpeg_buffer;
                    enc_size = request.jpeg_len;
                }
                else
                {
                    // Go try to acquire another frame.
                    fi_log(LOG_WARNING, "%s: Unable to encode the frame! Did we change states?", __PRETTY_FUNCTION__);
                    download_state = DNLD_CHECK_MODE;
                    enc_buf = nullptr;
                    enc_size = 0;
                }
            }
            else if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_PACK)
            {
                // Map the image memory so we can stream it.
                enc_buf = (uint8_t *)ipm_info.bb_ptr;

                // Save the size.
                enc_size = (int)pc_img_len;

                // Move on to streaming the frame.
                download_state = DNLD_STREAM_FRAME;

                //===================================================================
                // ALM_TODO:
                // At some point, the FPGA might do the packing for us. But for now
                // we take a raw TIFF image and pack it in software.
                //===================================================================

                // Make sure we have access to the internal library.
                if (FiInternal::factory() != nullptr)
                {
                    // Get our function pointer.
                    FiInternal::CONVERT_TIFF_FN funcPtr = FiInternal::factory()->getConvertTiffFuncPtr();
                    if (funcPtr != nullptr)
                    {
                        // As it turns out, our bounce buffer is plenty big enough to hould the output of
                        // the conversion. Just re-use it.
                        FAS_INTERNAL::FIP_STATUS convertStatus = (*funcPtr)(enc_buf, enc_buf, transfer_count, img_width, img_height, img_bpp);
                        if (convertStatus == FAS_INTERNAL::FIP_OK)
                        {
                            // The size of the image is now the packed size.
                            enc_size = packed_size;
                        }
                        else
                        {
                            fi_log(LOG_DEBUG, "%s: ConvertTiffToPack() failed [%d]", __PRETTY_FUNCTION__, (int)convertStatus);
                        }
                    }
                }
            }
            else
            {
                // Map the image memory so we can stream it.
                enc_buf = (uint8_t *)ipm_info.bb_ptr;

                // Save the size.
                enc_size = (int)pc_img_len;

                // Move on to streaming the frame.
                download_state = DNLD_STREAM_FRAME;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: encode: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case DNLD_STREAM_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Do we have an FSP stream?
            if (fsp_socket.isValid())
            {
                // Tell the stream the frame number (+1 for our protocol needs).
                fsp_socket.setFrameNum(transfer_count + 1);

                // Send the data onwards. Ignore the header for the packed stream.
                if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_PACK)
                    rc = fsp_socket.streamData(enc_buf, enc_size, FSP_PAYLOAD_FILE);
                else
                    rc = fsp_socket.streamData(enc_buf, enc_size, FSP_PAYLOAD_FILE, dnld_header_buffer->data(), dnld_header_buffer->size());
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;

                // Populate our image structure.
                image_info.format = get_stream_format();
                image_info.width = img_width;
                image_info.height = img_height;
                image_info.frame_num = transfer_count + 1;
                image_info.image_len = enc_size;
                image_info.image_data = enc_buf;

                // We want to explicitly skip any header information for the PACKED format.
                if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_PACK)
                {
                    image_info.header_data = nullptr;
                    image_info.header_len = 0;
                }
                else
                {
                    image_info.header_data = dnld_header_buffer->data();
                    image_info.header_len = dnld_header_buffer->size();
                }

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }

            if (!rc)
                fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);
            else
                transfer_count++;

            // Release the buffer, if necessary.
            if ((get_stream_format() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG)
                jpeg_release_frame(enc_buf);

            // Check for a mode change before getting the next frame.
            download_state = DNLD_CHECK_MODE;

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: stream: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case DNLD_CHECK_MODE:
#ifdef DO_STATISTICS
            if (stat_fp != nullptr)
            {
                fprintf(stat_fp, "%d", ipm_info.frame_num);
                for (double elapsed : times_measure)
                {
                    fprintf(stat_fp, ",%lf", elapsed);
                }
                fprintf(stat_fp, "\n");
                fflush(stat_fp);
            }

            // Reset the vector.
            times_measure.clear();
#endif /* DO_STATISTICS */

            // Are we done?
            if (transfer_count == frame_count)
            {
                // We're done.
                done = true;
                rc = 0;
                break;
            }

            // Move to the next frame.
            ipm_info.frame_num++;
            if (ipm_info.frame_num > get_end_frame_number())
            {
                // Wrap our frame number back around to the start.
                ipm_info.frame_num = get_start_frame_number();
            }

            // Setup to acquire the next frame.
            if (ltr_mode)
                download_state = DNLD_LTR_FRAME_LOAD;
            else
                download_state = DNLD_PROCESS_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, download_state);
            download_state = DNLD_CHECK_MODE;
            break;
        }

        // Check for a mode change.
        if (wait_new_mode_or_timeout(0) == 0)
        {
            // We are done.
            done = true;
            rc = 0;
        }
    }

    // Now stop the timer. All per-frame processing has completed.
    dnld_timer.stop();
    TP(FSP_STATS, "%s: download done in %lf secs.", __PRETTY_FUNCTION__, dnld_timer.lastMeasurement);

#ifdef DO_STATISTICS
    fclose(stat_fp);
#endif /* DO_STATISTICS */

    // If we're in LTR mode, we need to do some more cleanup.
    if (ltr_mode)
    {
        // Release the LTR buffer.
        bb_release(ltr_bb_index);
    }

    // Release the bounce buffer.
    bb_release(ipm_info.bb_index);

    // Release the IPM buffer, if we have one.
    if (ipm_info.ipm_buf_addr != 0)
    {
        ipm_proc_release(ipm_info.ipm_buf_addr);
        ipm_info.ipm_buf_addr = 0;
    }

    // Return our status.
    TP(FSP_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Handle live view until the mode changes.
 *
 * In the live loop, we wait for a signal or a timeout. Then we check whether
 * the mode has changed or there is a new frame to process.
 *
 * @return 0 on success, or an error code otherwise.
 */
int fspWorker::do_live_loop()
{
    int rc = 0;
    int live_state;
    int ifg_millisec;
    bool live_done;
    bool stream_err(false);
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Fastec::IntervalTimer live_timer;
    Fastec::IntervalTimer frame_timer;
    JPEG_PROC::IPM_REQUEST request;

    bool liveMA_active(false);
    std::unique_ptr<uint8_t> ampFrame;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Check overrides.
    bool stream_raw = false;
    override_bool(OVR_STREAM_RAW, &stream_raw);
    if (stream_raw)
    {
        return do_raw_loop();
    }

    // Initialize our state.
    live_state = LIVE_SETUP;

    // Keep going until told to stop.
    live_done = false;
    while (!live_done)
    {
        int msecs;

        TP(FSP_STATE_VERBOSE, "%s: state=%d", __PRETTY_FUNCTION__, live_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (live_state)
        {
        case LIVE_SETUP:
            // Clear out the result structure.
            memset(&request, 0, sizeof(request));

            // We want around 20 fps, but only reuse a frame if it was captured
            // more recently than that time. 20 fps is 50 msec, so we want a frame no older
            // than, say, 15 msec. "time_limit" is in usec, so that's...
            request.time_limit = 15000;

            // Processing info is always the same for live view images.
            request.dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
            if (sensor_ptr->is_sensor_color())
            {
                request.dma_mod |= IM2PCIE_DMA_GENERIC_DATA;
                request.proc_info = IMGPROC_TYPE_NV12;
            }
            else
            {
                request.dma_mod |= IM2PCIE_DMA_IMAGE_DATA;
                request.proc_info = IMGPROC_TYPE_Y8;
            }
            if ((regs[GIGE_REG_FPN].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
                request.proc_info |= IMGPROC_USE_FPN;

            // Get our dimensions.
            request.width = regs[GIGE_REG_WIDTH].get_value();
            request.height = regs[GIGE_REG_HEIGHT].get_value();

            // Move on.
            live_state = LIVE_ACQUIRE_FRAME;
            break;
        case LIVE_ACQUIRE_FRAME:
            // Get the current time so we can calculate how much time each frame
            // takes to proceses.
            live_timer.start();

            // Statistics
            frame_timer.start();
            // Statistics

            // Is the MA stream active?
            if (!liveMA_active)
                liveMA_active = shmemMAIsActive();
            if (liveMA_active)
            {
                // Do we have our buffer already allocated?
                if (ampFrame.get() == nullptr)
                {
                    ampFrame = std::unique_ptr<uint8_t>(new uint8_t[request.width * request.height]);
                    if (ampFrame.get() == nullptr)
                    {
                        fi_log(LOG_ERR, "%s: Unable to allocate the ampFrame buffer!", FI_FUNCTION);
                        break;
                    }
                }

                // Get the amplified frame from the MA engine.
                fi_log(LOG_DEBUG, "%s: Getting amplified frame.", FI_FUNCTION);
                if (!shmemMAGetFrame(ampFrame.get()))
                {
                    fi_log(LOG_ERR, "%s: Unable to get an amplified frame!", FI_FUNCTION);

                    // Probably need to transition state.
                    live_state = LIVE_CHECK_MODE;
                    liveMA_active = false;
                }
                else
                {
                    JPEG_PROC::BUF_REQUEST buf_request;

                    // Initialize our request structure with what we know.
                    buf_request.pixel_format = JPEG_UTIL::Y8;
                    buf_request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();
                    buf_request.jpeg_buffer = nullptr;
                    buf_request.jpeg_len = 0;

                    fi_log(LOG_DEBUG, "%s: JPEG encoding amplified frame.", FI_FUNCTION);
                    rc = encode_frame(ampFrame.get(), &buf_request);
                    if ((buf_request.jpeg_len > 0) && (buf_request.jpeg_buffer != nullptr))
                    {
                        // Move the JPEG information over so the streaming logic can use it.
                        request.jpeg_len = buf_request.jpeg_len;
                        request.jpeg_buffer = buf_request.jpeg_buffer;

                        // Move on to streaming the frame.
                        live_state = LIVE_STREAM_FRAME;
                        fi_log(LOG_DEBUG, "%s: Moving to LIVE_STREAM_FRAME.", FI_FUNCTION);
                    }
                    else
                    {
                        // Go try to acquire another frame.
                        fi_log(LOG_WARNING, "%s: Unable to encode the frame! Did we change states?", __PRETTY_FUNCTION__);
                        live_state = LIVE_CHECK_MODE;
                    }
                }
            }
            else
            {
                // Get a frame from the sensor.
                rc = get_live_frame();
                if (rc == 0)
                {
                    // Figure out where the image is in image memory.
                    request.image_addr = ((uint64_t)ipm_info.frame_addr_hw) << HEXWORD_SHIFT;

                    // Move on to proceessing the frame.
                    live_state = LIVE_PROCESS_FRAME;
                }
                else
                {
                    // Probably need to transition state.
                    live_state = LIVE_CHECK_MODE;
                }
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: acquire frame: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case LIVE_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Set this as late as possible to allow the user to change at runtime.
            request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();

            // Ask for the frame from the JPEG processor.
            if (jpeg_ipm_request(&request))
            {
                // Move on to streaming the frame.
                live_state = LIVE_STREAM_FRAME;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_WARNING, "%s: Unable to encode the frame! Did we change states?", __PRETTY_FUNCTION__);
                live_state = LIVE_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: encode frame: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case LIVE_STREAM_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Send the data onwards.
            if (fsp_socket.isValid())
            {
                // Send it to the FSP stream.
                rc = fsp_socket.streamData(request.jpeg_buffer, request.jpeg_len, FSP_PAYLOAD_FILE /* ALM_TODO_344: FSP_PAYLOAD_RAW */);
            }

            {
                std::unique_lock<std::mutex> locker(last_mutex);    // Protect access to the last_ptr.

                // Is anyone interested in the frame?
                if (last_ptr != nullptr)
                {
                    // Copy the image.
                    memcpy(last_ptr, request.jpeg_buffer, request.jpeg_len);

                    // All done.
                    last_ptr = nullptr;
                    last_len = request.jpeg_len;

                    // Let the client know the data is ready.
                    last_ready.notify_one();
                }
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;

                // Populate our image structure.
                image_info.format = GIGE_STREAM_FORMAT_JPEG;
                image_info.width = request.width;
                image_info.height = request.height;
                image_info.frame_num = 0;
                image_info.image_len = request.jpeg_len;
                image_info.image_data = request.jpeg_buffer;
                image_info.header_data = nullptr;
                image_info.header_len = 0;

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }

            if (!rc)
            {
                fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);
                stream_err = true;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame stream: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics

            // Release the buffer.
            jpeg_release_frame(request.jpeg_buffer);

            // Clear out the result values.
            request.jpeg_buffer = nullptr;
            request.jpeg_len = 0;

            // Go acquire the next frame.
            live_state = LIVE_CHECK_MODE;
            break;
        case LIVE_CHECK_MODE:
            // Statistics
            // Stop the timer.
            live_timer.stop();
            // Statistics

            // If we are not in error mode, go ahead and figure out what our
            // interframe gap time should be.
            if (!stream_err)
            {
                // What is our interframe delay?
                ifg_millisec = get_interframe_delay();

                // The user has specified an interframe gap time, which will control how
                // fast we stream (or attempt to stream) frames. How much of that gap time
                // is still available?
                msecs = ifg_millisec - ((int)((live_timer.lastMeasurement + 0.0005) * 1000.0));
                if (msecs < 0)
                {
                    TP(FSP_STATS, "%s: Loop took too long (%lf)!", __PRETTY_FUNCTION__, live_timer.lastMeasurement);
                    msecs = 0;
                }
                else
                {
                    TP(FSP_STATS, "%s: total: %lf secs.", __PRETTY_FUNCTION__, live_timer.lastMeasurement);
                }
            }
            else
            {
                FI_LOG_TAG(__FILE__, __LINE__, "%s: Got a stream error.", __PRETTY_FUNCTION__);
                live_done = true;
                rc = -1;
            }

            // Assume we'll go around and get the next frame.
            live_state = LIVE_ACQUIRE_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, live_state);
            live_state = LIVE_CHECK_MODE;
            break;
        }

        // Wait for either a mode change or time to get the next frame.
        if (!live_done)
        {
            if (wait_new_mode_or_timeout(msecs) == 0)
            {
                // We are done.
                live_done = true;
                rc = 0;
            }
        }
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Expand from the packed 10-bit format our cameras capture to a format we can
 *        actually use.
 */
uint32_t fspWorker::expand_10_bits(pm_metadata *meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr)
{
    uint32_t    *long_ptr;
    bool        need_alignment_shim;
    uint16_t    *expanded_ptr;
    uint32_t    row;
    uint32_t    row_byte_count;

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // How many bytes of 16-bit pixels will we have per row.
    row_byte_count = meta_ptr->roi_width * 2;

    // Make sure the row is 32-bit aligned.
    if ((row_byte_count & 3) != 0)
    {
        // We need an alignment "shim" to get the row to 32-bit alignment.
        need_alignment_shim = true;
        row_byte_count = ((row_byte_count >> 2) + 1) << 2;
    }
    else
    {
        // We're 32-bit aligned on the row. No shim needed.
        need_alignment_shim = false;
    }

    // Setup our pointers.
    long_ptr = (uint32_t *)raw_ptr;
    expanded_ptr = (uint16_t *)dest_ptr;

    // Loop through the data, swapping the bytes.
    for (row = 0; row < meta_ptr->roi_height; ++row)
    {
        uint32_t    column;
        uint16_t pixel_short;
        uint32_t *swapword_ptr;

        // Initialize our starting column in the row.
        column = 0;
        swapword_ptr = long_ptr;

        // Swap all the words in the row.
        while (column < meta_ptr->roi_width)
        {
            uint32_t temp_word;

            // Grab a word of data.
            temp_word = htonl(*swapword_ptr);

            // Put it back into memory.
            *swapword_ptr++ = temp_word;

            // There are 3 pixels per 32 bits.
            column += 3;
        }

        // Reset our starting column in the row.
        column = 0;

        // Work through all the pixels in this row.
        while (column < meta_ptr->roi_width)
        {
            uint32_t packed_word;

            // Grab a word of data.
            packed_word = *long_ptr++;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & 0x3FF00000) >> 14);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & 0xFFC00) >> 4);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & 0x3FF) << 6);
            *expanded_ptr++ = pixel_short;

            // Update our column count.
            column++;
        }

        // Do we need an alignment word?
        if (need_alignment_shim)
        {
            // Add an extra false "pixel" to pad the row to be 32-bit aligned.
            pixel_short = 0;
            *expanded_ptr++ = pixel_short;
        }
    }

    // Return the size of the expanded data.
    return (row_byte_count * meta_ptr->roi_height);
}

/*!
 * @brief Expand from the packed 12-bit format our cameras capture to a format we can
 *        actually use.
 */
uint32_t fspWorker::expand_12_bits(pm_metadata *meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr)
{
    bool        need_alignment_shim;
    uint16_t    *expanded_ptr;
    uint32_t    row;
    uint32_t    row_byte_count;
    uint64_t    *long_ptr;

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // How many bytes of 16-bit pixels will we have per row.
    row_byte_count = meta_ptr->roi_width * 2;

    // Make sure the row is 32-bit aligned.
    if ((row_byte_count & 3) != 0)
    {
        // We need an alignment "shim" to get the row to 32-bit alignment.
        need_alignment_shim = true;
        row_byte_count = ((row_byte_count >> 2) + 1) << 2;
    }
    else
    {
        // We're 32-bit aligned on the row. No shim needed.
        need_alignment_shim = false;
    }

    // Setup our pointers.
    long_ptr = (uint64_t *)raw_ptr;
    expanded_ptr = (uint16_t *)dest_ptr;

    // Loop through the data, swapping the bytes.
    for (row = 0; row < meta_ptr->roi_height; ++row)
    {
        uint32_t    column;
        uint16_t    pixel_short;

        // Initialize our starting column in the row.
        column = 0;

        // Work through all the pixels in this row.
        while (column < meta_ptr->roi_width)
        {
            uint64_t    packed_word;
            uint64_t    swapper;

            // Grab a word of data.
            packed_word = *long_ptr++;

            // Swap the 32-bit words around.
            swapper = ((uint64_t)htonl(((uint32_t)packed_word))) << 32;
            packed_word = ((uint64_t)htonl((uint32_t)(packed_word >> 32)));
            packed_word |= swapper;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 48)) >> 44);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 36)) >> 32);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 24)) >> 20);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 12)) >> 8);
            *expanded_ptr++ = pixel_short;

            // Bump our column count and see if we're done.
            column++;
            if (column == meta_ptr->roi_width)
                break;

            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & ((uint64_t)0xFFF)) << 4);
            *expanded_ptr++ = pixel_short;

            // Bump our column count.
            column++;
        }

        // Do we need an alignment word?
        if (need_alignment_shim)
        {
            // Add an extra false "pixel" to pad the row to be 32-bit aligned.
            pixel_short = 0;
            *expanded_ptr++ = pixel_short;
        }
    }

    // Return the size of the expanded data.
    return (row_byte_count * meta_ptr->roi_height);
}

uint32_t fspWorker::expand_the_pixels(pm_metadata *pm_meta_ptr, uint8_t *dest_ptr, uint8_t *raw_ptr)
{
    uint32_t new_length = 0;

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Are we 10-bit or 12-bit?
    if (IS_10_BIT_MODE(pm_meta_ptr->bit_sel))
    {
        new_length = expand_10_bits(pm_meta_ptr, dest_ptr, raw_ptr);
    }
    else if (IS_12_BIT_MODE(pm_meta_ptr->bit_sel))
    {
        new_length = expand_12_bits(pm_meta_ptr, dest_ptr, raw_ptr);
    }

    // Return the resulting length of the expanded pixels.
    return new_length;
}

uint32_t fspWorker::assemble_dng(uint8_t *hdr_ptr, int hdr_length, uint8_t *raw_ptr, int frame_size)
{
    uint8_t *dest_ptr;
    std::unique_ptr<pm_metadata> pm_meta_ptr;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Initialize our length.
    ipm_info.image_len = 0;

    // Get our metadata information.
    pm_meta_ptr = pm_ptr->getActivePartMetadata();

    // Our destination pointer is the ipm_info bounce buffer.
    dest_ptr = (uint8_t *)bb_mmap(ipm_info.bb_index, 0);
    if (dest_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to map bb_index (%d).", __PRETTY_FUNCTION__, ipm_info.bb_index);
        goto bail;
    }

    // Copy the header information over.
    memcpy(dest_ptr, hdr_ptr, hdr_length);

    // We handle 8-bit differently than multi-byte data.
    if (IS_8_BIT_MODE(pm_meta_ptr->bit_sel))
    {
        // Now copy the image data.
        memcpy(&dest_ptr[hdr_length], raw_ptr, frame_size);

        // The final size is the header plus the frame.
        ipm_info.image_len = hdr_length + frame_size;
    }
    else
    {
        // We need to expand the pixels to be 16-bit.
        ipm_info.image_len = expand_the_pixels(pm_meta_ptr.get(), &dest_ptr[hdr_length], raw_ptr);
        if (ipm_info.image_len > 0)
        {
            // Add in the header length.
            ipm_info.image_len += hdr_length;
        }
    }

    // Unmap the buffer.
    bb_unmap(ipm_info.bb_index);

bail:
    // Return the total number of bytes in the DNG "image".
    return ipm_info.image_len;
}

#ifdef DNG_LOAD
uint32_t fsp_worker::read_as_dng(uint32_t frame_num)
{
    uint32_t dng_length = 0;

    uint8_t *raw_ptr;
    uint32_t raw_length;
    uint64_t frame_offset;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(FSP_PLAYBACK, "%s: frame_num=%u", __PRETTY_FUNCTION__, frame_num);

    // Get the offset in bytes.
    frame_offset = pm_ptr->get_active_frame_addr(frame_num, false);

    // Our destination pointer is the ipm_info bounce buffer.
    raw_ptr = (uint8_t *)bb_mmap(dng_bb_index, 0);
    if (raw_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to map dng_bb_index (%d).", __PRETTY_FUNCTION__, dng_bb_index);
        goto bail;
    }

    // The length of the image is the frame size minus the per-frame
    // metadata size.
    raw_length = pm_ptr->get_frame_size() - FRAME_METADATA_HW_BYTES;

    // Read up the image data.
    if (pm_ptr->ltr_read_bytes(raw_ptr, frame_offset, raw_length) != (int)raw_length)
    {
        fi_log(LOG_ERR, "%s: Unable to read from 0x%" PRIx64 " for %u bytes", __PRETTY_FUNCTION__, frame_offset, raw_length);
        goto bail;
    }

    // Assembly our DNG.
    dng_length = assemble_dng(dnld_header_buffer->data(), dnld_header_buffer->size(), raw_ptr, raw_length);

bail:
    // Unmap our data buffer.
    if (raw_ptr != nullptr)
    {
        bb_unmap(dng_bb_index);
        raw_ptr = nullptr;
    }

    // Return the amount of data we've read.
    return dng_length;
}
#endif /* DNG_LOAD */

int fspWorker::ss_check_mode(bool ltr_mode, bool& pb_done, int& delay_msecs)
{
    int ret_state = SS_CHECK_MODE;

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Are we at the same frame?
    if (ipm_info.frame_num == get_seq_start_frame(current_mode))
    {
        // Wait around a bit in case the user is really stepping through.
        TP(FSP_PLAYBACK, "%s: frame_num=%u, seq_start_frame=%u", __PRETTY_FUNCTION__, ipm_info.frame_num, get_seq_start_frame(current_mode));
        delay_msecs = SS_STEP_PAUSE;
        pb_done = false;
    }
    else
    {
        // We have a new frame to step to!
        ipm_info.frame_num = get_seq_start_frame(current_mode);
        delay_msecs = 0;
        pb_done = false;

        // Setup to acquire the next frame.
        ret_state = (ltr_mode ? SS_LTR_FRAME_LOAD : SS_PROCESS_FRAME);
        TP(FSP_PLAYBACK, "%s: new frame=%u, next state=%d", __PRETTY_FUNCTION__, ipm_info.frame_num, ret_state);
    }

    return ret_state;
}

int fspWorker::ltr_frame_load(uint32_t frame_num)
{
    int rc = -1;
    uint8_t *raw_ptr;
    uint32_t raw_length;
    uint32_t dma_mod;
    uint64_t frame_offset;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(FSP_PLAYBACK, "%s: frame_num=%u", __PRETTY_FUNCTION__, frame_num);

    // Get the SSD offset in bytes.
    frame_offset = pm_ptr->get_active_frame_addr(frame_num, false);
    TP(FSP_PLAYBACK, "%s: frame_offset=0x%" PRIx64, __PRETTY_FUNCTION__, frame_offset);

    // Our destination pointer is the LTR bounce buffer.
    raw_ptr = (uint8_t *)bb_mmap(ltr_bb_index, 0);
    if (raw_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to map ltr_bb_index (%d).", __PRETTY_FUNCTION__, ltr_bb_index);
        goto bail;
    }

    // Get the length of the image. This also includes the PFMD, but that's OK.
    raw_length = pm_ptr->getFrameSize();

    // Read up the image data.
    if (pm_ptr->ltrReadBytes(raw_ptr, frame_offset, raw_length) != (int)raw_length)
    {
        fi_log(LOG_ERR, "%s: Unable to read from 0x%" PRIx64 " for %u bytes", __PRETTY_FUNCTION__, frame_offset, raw_length);
        goto bail;
    }

    // Now ask the FPGA to put the data into the image memory buffer.
    dma_mod = PCIE2IM_DMA_BYTE_ORDER_R2L | PCIE2IM_DMA_WORD_ORDER_L2R;
    TP(FSP_PLAYBACK, "%s: dma_mod=0x%08x, len=%u", __PRETTY_FUNCTION__, dma_mod, raw_length);
    rc = fpga_ptr->imageUpload(dma_mod, ltr_bb_addr, ((uint64_t)ipm_info.frame_addr_hw) << HEXWORD_SHIFT, raw_length);
    if (rc < 0)
    {
        // Oops! What happened here?!
        fi_log(LOG_ERR, "%s: Unable to DMA data!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = 0;

bail:
    // Return our status.
    return rc;
}

int fspWorker::ss_process_frame(JPEG_PROC::IPM_REQUEST *request, bool ltr_mode)
{
    int rc;
    int ret_state;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(FSP_PLAYBACK, "%s: ltr_mode=%c", __PRETTY_FUNCTION__, (ltr_mode ? 'T' : 'F'));

    // Get the current time so we can calculate how much time each frame
    // takes to proceses.
    pb_timer.start();

    // Are we in LTR mode?
    if (ltr_mode)
    {
        // We already have our IM addres, we just need it in a different format.
        request->image_addr = ((uint64_t)ipm_info.frame_addr_hw) << HEXWORD_SHIFT;

        // Ask for the frame from the JPEG processor.
        rc = jpeg_ipm_request(request);
        if (rc)
        {
            // Move on to transferring the frame.
            ret_state = SS_STREAM_FRAME;
        }
        else
        {
            // Go try to acquire another frame.
            fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
            ret_state = SS_CHECK_MODE;
        }
    }
    else
    {
        // Get the address of the frame we're going to process.
        ipm_info.frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_active_frame_addr(ipm_info.frame_num, false));

        // Convert to a full address.
        request->image_addr = ((uint64_t)ipm_info.frame_addr_hw) << HEXWORD_SHIFT;

        // Get the frame processed by the IPM.
        rc = jpeg_ipm_request(request);
        if (rc)
        {
            // Move on to transferring the frame.
            ret_state = SS_STREAM_FRAME;
        }
        else
        {
            // Go try to acquire another frame.
            fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
            ret_state = SS_CHECK_MODE;
        }
    }

    // Move on.
    return ret_state;
}

int fspWorker::ss_setup(bool ltr_mode, uint32_t& dma_mod, uint32_t& proc_info)
{
    int next_state = SS_PROCESS_FRAME;
    uint32_t start_frame;
    uint32_t pb_control;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    Sensor  *sensor_ptr = CameraServer::get_sensor();

    TP(FSP_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Initialize our frame number and IPM buffer ID.
    ipm_info.frame_num = get_seq_start_frame(current_mode);

    // Processing info is always the same for playback images.
    dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
    if (sensor_ptr->is_sensor_color())
    {
        dma_mod |= IM2PCIE_DMA_GENERIC_DATA;
        proc_info = IMGPROC_TYPE_NV12;
    }
    else
    {
        dma_mod |= IM2PCIE_DMA_IMAGE_DATA;
        proc_info = IMGPROC_TYPE_Y8;
    }

    // Are we supposed to use FPN?
    if ((regs[GIGE_REG_FPN].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        proc_info |= IMGPROC_USE_FPN;

    // Set the status register to indicate we're busy.
    regs[GIGE_REG_PB_STATUS].set_value(GIGE_PBS_BUSY);
    regs[GIGE_REG_PB_LAST_FRAME].set_value(ipm_info.frame_num);
    regs[GIGE_REG_CSI_RVW_FRAME_NUM].set_value(ipm_info.frame_num);
    fcp_notify_apps(GIGE_REG_CSI_RVW_FRAME_NUM, ipm_info.frame_num);

    // Initialize pertinent registers incase someone looks.
    TP(FSP_PLAYBACK, "%s: start=%u, end=%u", __PRETTY_FUNCTION__, get_start_frame_number(), get_end_frame_number());
    start_frame = get_seq_start_frame(current_mode);
    pbc_update_start_frame(nullptr, start_frame);

    // Let folks know we're playing.
    pb_control = GIGE_PBC_ENABLE;
    if (!playback_forward())
        pb_control |= GIGE_PBC_REVERSE;
    if (get_frame_interval() == 0)
        pb_control |= (1 << GIGE_PBC_INTERVAL_SHIFT);
    else
        pb_control |= ((get_frame_interval() << GIGE_PBC_INTERVAL_SHIFT) & GIGE_PBC_INTERVAL_MASK);
    TP(FSP_PLAYBACK, "%s: GIGE_REG_PB_CONTROL=0x%08x", __PRETTY_FUNCTION__, pb_control);
    regs[GIGE_REG_PB_CONTROL].set_value(pb_control);

    // Do we need to notify users?
    if (current_mode == GIGE_STREAM_RECORDED)
        fcp_notify_apps(GIGE_REG_PB_CONTROL, pb_control);

    // Are we in LTR mode?
    if (ltr_mode)
    {
        // Get our buffer for the raw file heading up to the FPGA.
        ltr_bb_index = bb_get_avail();
        ltr_bb_addr = bb_get_avail_addr(ltr_bb_index);

        // Get a frame address in IM that we can use to store the raw frame.
        ipm_info.frame_addr_hw = convert_to_HEXWORD(pm_ptr->getCaptureFrameAddr(0));
        TP(FSP_PLAYBACK, "%s: LTR processing frame in IM, frame_addr_hw=0x%08x", __PRETTY_FUNCTION__, ipm_info.frame_addr_hw);

        // Setup for the next state.
        next_state = SS_LTR_FRAME_LOAD;
    }

    // Move on.
    return next_state;
}

/*!
 * @brief Handle single-step until the mode changes.
 *
 * In the single-step loop, we wait for a signal or a timeout. Then we check whether
 * the mode has changed or there is a new frame to process.
 *
 * @return 0 on success, or an error code otherwise.
 */
int fspWorker::do_singlestep_loop()
{
    int rc = 0;
    int ss_state;
    bool ss_done;
    bool ltr_mode;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    Fastec::IntervalTimer frame_timer;
    JPEG_PROC::IPM_REQUEST request;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    ss_state = SS_SETUP;

    // Determine if we're in LTR mode or not.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);

    // Keep going until told to stop.
    ss_done = false;
    while (!ss_done)
    {
        int msecs;

        TP(FSP_STATE_VERBOSE, "%s: state=%d", __PRETTY_FUNCTION__, ss_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (ss_state)
        {
        case SS_SETUP:
            // Clear out the result values.
            memset(&request, 0, sizeof(request));

            // Set the quality value.
            request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();

            // There's no time limit on the frame.
            request.time_limit = 0;

            // Get our dimensions.
            request.width = regs[GIGE_REG_WIDTH].get_value();
            request.height = regs[GIGE_REG_HEIGHT].get_value();

            // Initialize the rest of the system.
            ss_state = ss_setup(ltr_mode, request.dma_mod, request.proc_info);
            break;
        case SS_LTR_FRAME_LOAD:
            // Statistics
            frame_timer.start();
            // Statistics

            // Load the frame into the IM for FPGA processing.
            if (ltr_frame_load(ipm_info.frame_num) == 0)
                ss_state = SS_PROCESS_FRAME;
            else
                ss_state = SS_CHECK_MODE;

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame %u sent to FPGA: %lf secs.", __PRETTY_FUNCTION__, ipm_info.frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case SS_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            ss_state = ss_process_frame(&request, ltr_mode);

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame %u processed: %lf secs.", __PRETTY_FUNCTION__, ipm_info.frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case SS_STREAM_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Send the data onwards.
            if (fsp_socket.isValid())
            {
                // Send it to the FSP stream.
                rc = fsp_socket.streamData(request.jpeg_buffer, request.jpeg_len, FSP_PAYLOAD_FILE /* ALM_TODO_344: FSP_PAYLOAD_RAW */);
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;

                // Populate our image structure.
                image_info.format = GIGE_STREAM_FORMAT_JPEG;
                image_info.width = request.width;
                image_info.height = request.height;
                image_info.frame_num = ipm_info.frame_num;
                image_info.image_len = request.jpeg_len;
                image_info.image_data = request.jpeg_buffer;
                image_info.header_data = nullptr;
                image_info.header_len = 0;

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }

            if (!rc)
                fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame %u stream: %lf secs.", __PRETTY_FUNCTION__, ipm_info.frame_num, frame_timer.lastMeasurement);
            // Statistics

            // Release the buffer.
            jpeg_release_frame(request.jpeg_buffer);

            // Clear out the result values.
            request.jpeg_buffer = nullptr;
            request.jpeg_len = 0;

            // Update the "last frame" register.
            regs[GIGE_REG_PB_LAST_FRAME].set_value(ipm_info.frame_num);
            regs[GIGE_REG_CSI_RVW_FRAME_NUM].set_value(ipm_info.frame_num);
            fcp_notify_apps(GIGE_REG_CSI_RVW_FRAME_NUM, ipm_info.frame_num);

            // Check for a mode change before getting the next frame.
            ss_state = SS_CHECK_MODE;
            break;
        case SS_CHECK_MODE:
            ss_state = ss_check_mode(ltr_mode, ss_done, msecs);
            if (ss_done)
                rc = 0;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, ss_state);
            ss_state = SS_CHECK_MODE;
            break;
        }

        // Check for a mode change or wait the specified amount of time to get
        // the next frame.
        if (wait_new_mode_or_timeout(msecs) == 0)
        {
            // We are done.
            ss_done = true;
            rc = 0;
        }
    }

    // If we're in LTR mode, we need to do some more cleanup.
    if (ltr_mode)
    {
        // Release the LTR buffer.
        bb_release(ltr_bb_index);
    }

    // Set the status register to indicate we're idle.
    regs[GIGE_REG_PB_STATUS].set_value(0);

    // Return our status.
    TP(FSP_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

void fspWorker::pb_setup()
{
    uint32_t pb_control;

    TP(FSP_PLAYBACK, "%s: start=%u, end=%u", __PRETTY_FUNCTION__, get_start_frame_number(), get_end_frame_number());

    // Are we simply attaching to the playback stream?
    if (current_mode == GIGE_STREAM_SLAVE_ENABLE)
    {
        TP(FSP_PLAYBACK, "%s: attach to current playback stream.", __PRETTY_FUNCTION__);
        return;
    }

    // Make sure we start in the right place.
    pbc_update_start_frame(nullptr, get_seq_start_frame(current_mode));

    // Let folks know we're playing.
    pb_control = GIGE_PBC_ENABLE;
    if (!playback_forward())
        pb_control |= GIGE_PBC_REVERSE;
    if (get_frame_interval() == 0)
        pb_control |= (1 << GIGE_PBC_INTERVAL_SHIFT);
    else
        pb_control |= ((get_frame_interval() << GIGE_PBC_INTERVAL_SHIFT) & GIGE_PBC_INTERVAL_MASK);
    pbc_update_control(nullptr, pb_control);
}

/*!
 * @brief Handle playback until the mode changes.
 *
 * In the playback loop, we wait for a signal or a timeout. Then we check whether
 * the mode has changed or there is a new frame to process.
 *
 * @return 0 on success, or an error code otherwise.
 */
int fspWorker::do_playback_loop()
{
    int rc = 0;
    int pb_state;
    bool pb_done;
    PB_CONTROL::FRAME_REQUEST frame_request;
    Fastec::IntervalTimer frame_timer;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    pb_state = PB_SETUP;

    // Keep going until told to stop.
    pb_done = false;
    while (!pb_done)
    {
        int msecs;

        TP(FSP_STATE_VERBOSE, "%s: state=%d", __PRETTY_FUNCTION__, pb_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (pb_state)
        {
        case PB_SETUP:
            // Initialize the rest of the system.
            pb_setup();

            // Make sure this first time, we ask for the first available frame.
            memset(&frame_request, 0, sizeof(frame_request));
            frame_request.frame_num = -1;

            // Move to the next state.
            pb_state = PB_GET_FRAME;
            break;
        case PB_GET_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // This will block until the playback controller has
            // a new frame.
            TP(FSP_STATE_VERBOSE, "%s: PB_GET_FRAME++", __PRETTY_FUNCTION__);
            if (pbc_request(&frame_request))
            {
                // Got our frame!
                pb_state = PB_STREAM_FRAME;
            }
            else
            {
                // Failure here may be because the user cancelled playback.
                msecs = 10000;
            }
            TP(FSP_STATE_VERBOSE, "%s: PB_GET_FRAME--", __PRETTY_FUNCTION__);

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: Get frame %u: %lf secs.", __PRETTY_FUNCTION__, frame_request.frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case PB_STREAM_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Send the data onwards.
            if (fsp_socket.isValid())
            {
                // Send it to the FSP stream.
                rc = fsp_socket.streamData(frame_request.frame_buffer, frame_request.frame_len, FSP_PAYLOAD_FILE /* ALM_TODO_344: FSP_PAYLOAD_RAW */);
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;
                GIGE_REG_CACHE& regs = CameraServer::get_regs();

                // Populate our image structure.
                image_info.format = GIGE_STREAM_FORMAT_JPEG;
                image_info.width = regs[GIGE_REG_WIDTH].get_value();
                image_info.height = regs[GIGE_REG_HEIGHT].get_value();
                image_info.frame_num = frame_request.frame_num;
                image_info.image_len = frame_request.frame_len;
                image_info.image_data = frame_request.frame_buffer;
                image_info.header_data = nullptr;
                image_info.header_len = 0;

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }

            if (!rc)
                fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: Stream frame %u: %lf secs.", __PRETTY_FUNCTION__, frame_request.frame_num, frame_timer.lastMeasurement);
            // Statistics

            // Release the buffer.
            pbc_release_frame(frame_request.frame_buffer);

            // Clear out some of the request fields. Make sure we keep the frame_num
            // field intact, so that we will only get a new frame (not a repeat).
            frame_request.frame_buffer = nullptr;
            frame_request.frame_len = 0;

            // Back to waiting for a frame.
            pb_state = PB_GET_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, pb_state);
            pb_done = true;
            break;
        }

        // Check for a mode change or wait the specified amount of time to get
        // the next frame.
        if (wait_new_mode_or_timeout(msecs) == 0)
        {
            // We are done.
            pb_done = true;
            rc = 0;
        }
    }

    // Return our status.
    TP(FSP_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

void fspWorker::stream_work()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Make sure we're not running yet.
    if (current_mode != -1)
    {
        fi_log(LOG_WARNING, "%s: Start fail: current_mode=%d!", __PRETTY_FUNCTION__, current_mode);
        return;
    }

    TP(FSP_FUNC, "%s: requested_mode=%d PID=%u", __PRETTY_FUNCTION__, requested_mode, get_tid());

    {
        std::unique_lock<std::mutex> locker(mode_mutex);    // Protect access to requested_mode.

        // Set our mode to the requested value.
        current_mode = requested_mode;
        requested_mode = -1;
        mode_status.notify_one();
    }

    // Update the register and let folks know where we are.
    regs[GIGE_REG_STREAM_CONTROL].set_value(current_mode);

    // Loop until we are shut down. Everytime around this loop we have finished
    // a complete transfer whether that is an IMAGE (BMP), a JPEG or a larger file.
    while (1)
    {
        int rc;
        int previous_mode;

        TP(FSP_STATE_VERBOSE, "%s: Top; current_mode == %d", __PRETTY_FUNCTION__, current_mode);

        previous_mode = current_mode;
        rc = 0;

        // Make sure we're not paused.
        fsp_socket.resume();

        // See if we are supposed to be streaming. This is possible because the creation of the
        // thread is different than the register to start streaming. Also find out if we are
        // doing recorded video what our frame range is.
        switch (current_mode)
        {
        case GIGE_STREAM_LIVE:
            rc = do_live_loop();
            break;
        case GIGE_STREAM_RECORDED:
            rc = do_playback_loop();
            break;
        case GIGE_STREAM_SINGLESTEP:
            rc = do_singlestep_loop();
            current_mode = GIGE_STREAM_STOP;
            break;
        case GIGE_STREAM_SLAVE_ENABLE:
            rc = do_playback_loop();
            break;
        case GIGE_STREAM_DOWNLOAD:
            rc = do_download_loop();
            current_mode = GIGE_STREAM_STOP;
            break;
        case GIGE_STREAM_GET_FILE:
            rc = do_file_stream();
            current_mode = GIGE_STREAM_STOP;
            break;
        case GIGE_STREAM_RAW_METADATA:
            rc = raw_metadata_stream();
            current_mode = GIGE_STREAM_STOP;
            TP(FSP_STATE, "%s: GIGE_STREAM_RAW_METADATA done (%d)", __PRETTY_FUNCTION__, rc);
            break;
        case GIGE_STREAM_SESSION:
            rc = do_session_stream();
            current_mode = GIGE_STREAM_STOP;
            break;
        case GIGE_STREAM_STOP:
            break;
        default:
            fi_log(LOG_WARNING, "%s: Unexpected mode: 0x%x", __PRETTY_FUNCTION__, current_mode);
            break;
        }

        if (rc != 0)
        {
            fi_log(LOG_ERR, "%s: mode=0x%x failed (errno:%d [%s], rc:%d), stopping", __PRETTY_FUNCTION__, current_mode, errno, strerror(errno), rc);

            // Check to see if we had an error, and not just stopping.
            if (errno != 0)
            {
                // Shutdown the stream.
                disconnectStream();
            }

            // Stop our state machine and go back waiting for instructions.
            current_mode = GIGE_STREAM_STOP;
            requested_mode = -1;
        }
        else if (requested_mode != -1)
        {
            std::unique_lock<std::mutex> locker(mode_mutex);    // Protect access to requested_mode.

            current_mode = requested_mode;
            requested_mode = -1;
            mode_status.notify_one();
        }

        // Do we have a mode change while we are running?
        if (previous_mode != current_mode)
        {
            TP(FSP_STATE, "%s: Bottom of loop, GIGE_REG_STREAM_CONTROL from %d to %d", __PRETTY_FUNCTION__, previous_mode, current_mode);
            regs[GIGE_REG_STREAM_CONTROL].set_value(current_mode);
        }

        // Are we stopping?
        if (current_mode == GIGE_STREAM_STOP)
            break;
    }
}

/*!
 * @brief The worker thread.
 */
void fspWorker::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Initialize our FSP socket.
    if (fsp_socket.setup() != CS_NET_SUCCESS)
    {
        fi_log(LOG_ERR, "%s: Unable to setup streaming", __PRETTY_FUNCTION__);
        return;
    }

    // We're running.
    thread_running = true;

    // Start out waiting for a connection.
    run_state = TR_WAIT_FOR_CONNECT;

    // Get our connection thread running.
    std::thread fsp_connect(&fspWorker::connect_socket, this);

    // Loop until we are shut down.
    while (!stop_request)
    {
        TP(FSP_STATE_VERBOSE, "%s: run_state=%d", __PRETTY_FUNCTION__, run_state);

        switch (run_state)
        {
        case TR_SETUP_SOCKET:
            // Make sure the socket is closed.
            fsp_socket.close();

            // Starting up waiting for a mode.
            current_mode = -1;
            run_state = TR_WAIT_FOR_CONNECT;
            break;
        case TR_WAIT_FOR_CONNECT:
            // Do we have a connection?
            if (fsp_socket.isValid() || grpcStreamIsConnected())
            {
                // Got a connection!
                TP(FSP_STATE, "%s: got a connection!", __PRETTY_FUNCTION__);

                // Starting up waiting for a mode.
                current_mode = -1;
                run_state = TR_WAIT_FOR_MODE;
            }

            else if (!stop_request)
            {
                // Wait until we are given a mode, or 50 milliseconds. We could be in shutdown,
                // so I don't want to wait TOO long.
                std::unique_lock<std::mutex> locker(mode_mutex);
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                mode_status.wait_until(locker, now + ms(50));
            }
            break;
        case TR_WAIT_FOR_MODE:
            // Wait for the user to ask us to start something. I want a block
            // here, and I'm cranky with the formatter not indenting properly. Using
            // a do-while false is equivalent to empty braces, so...
            do
            {
                // Wait until we are given a mode, or 500 milliseconds.
                std::unique_lock<std::mutex> locker(mode_mutex);
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                mode_status.wait_until(locker, now + ms(500));

                // Make sure we still have a valid connection.
                if (!fsp_socket.isValid() && !grpcStreamIsConnected())
                {
                    fi_log(LOG_DEBUG, "%s: [TR_WAIT_FOR_MODE] fsp_socket is invalid (did FasMotion close?), or the GRPC stream closed.", __PRETTY_FUNCTION__);
                    FI_LOG_TAG(__FILE__, __LINE__, "%s: fsp_socket=%d, listen_fd=%d, grpc=%d", __PRETTY_FUNCTION__, fsp_socket.getSocket(), fsp_socket.getListenFD(), grpcStreamIsConnected());

                    // Setup the socket again.
                    run_state = TR_SETUP_SOCKET;
                }

                // Check to see if we've got a new state.
                else if (requested_mode != -1)
                {
                    // Setup our mode information.
                    current_mode = -1;

                    // Move on...
                    run_state = TR_STREAM;
                }
            } while (0);
            break;
        case TR_STREAM:
            // Start streaming.
            stream_work();

            // Make sure we still have a valid connection.
            if (!fsp_socket.isValid() && !grpcStreamIsConnected())
            {
                fi_log(LOG_DEBUG, "%s: [TR_STREAM] fsp_socket is invalid or the GRPC stream is not connected.", __PRETTY_FUNCTION__);

                // Setup the socket again.
                run_state = TR_SETUP_SOCKET;
            }
            else
            {
                // Wait for the a new mode selection.
                run_state = TR_WAIT_FOR_MODE;
            }

            // Reset our mode.
            current_mode = -1;
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected state=%d", __PRETTY_FUNCTION__, run_state);
            break;
        }
    }

    // Make sure the stop request is set.
    stop_request = true;

    // Wait for the FSP connection thread to quit.
    fsp_connect.join();

    TP(FSP_FUNC, "%s: Exiting", __PRETTY_FUNCTION__);

    // Clean things up on the way out.
    cleanup();
}

/*!
 * @brief Set the streaming mode.
 *
 * @param streaming_type One of the GIGE_REG_STREAM_CONTROL values.
 */
int fspWorker::setMode(int mode, bool wait_for_change)
{
    int retry;
    int rc = GEV_STATUS_SUCCESS;
    fspWorker *fsp = fspWorker::factory();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

//    TP(FSP_FUNC, "%s: Setting new mode (%d -> %d), state=%d", __PRETTY_FUNCTION__, current_mode, mode, fsp->getRunState());
    LOG("%s: Setting new mode (%d -> %d), state=%d", __PRETTY_FUNCTION__, current_mode, mode, fsp->getRunState());

    if (fsp == nullptr)
    {
        rc = GEV_STATUS_ERROR;
        goto mode_bail;
    }

    // If the thread is not running, start it with the requested mode.
    // If we are stopping it and it is not running, that's success.
    if (mode == GIGE_STREAM_STOP)
    {
        uint32_t reg_value;

        reg_value = regs[GIGE_REG_PB_CONTROL].get_value() & ~GIGE_PBC_ENABLE;
        pbc_update_control(nullptr, reg_value);

        if ((fsp->getRunState() == TR_WAIT_FOR_CONNECT) || (fsp->getRunState() == TR_WAIT_FOR_MODE))
            goto mode_bail;

        // Tell the thread we'd like to stop please.
        // stop();
    }
    else if ((current_mode == mode) && (requested_mode == -1))
    {
        TP(FSP_FUNC, "%s: Requesting duplicate mode (%d)", __PRETTY_FUNCTION__, mode);
        goto mode_bail;
    }

    // Otherwise just set the requested mode.
    // But we can't do that if the thread has not yet accepted the
    // previous request.
    for (retry = 0; retry < MAX_SETMODE_RETRY_COUNT; ++retry)
    {
        // Is the thread running?
        if (!thread_running)
        {
            // Are we supposed to stop it?
            if (mode != GIGE_STREAM_STOP)
                rc = GEV_STATUS_ERROR;
            break;
        }

        // If we can set the new mode, do so and wake up the thread.
        if (requested_mode == -1)
        {
            std::unique_lock<std::mutex> locker(mode_mutex);

            // Set the new mode we're requesting.
            requested_mode = mode;

            // Wake up anyone sleeping.
            LOG("%s: Notifying of a mode status change. requested_mode=%d", __PRETTY_FUNCTION__, requested_mode);
            mode_status.notify_one();

            // Are we going to wait?
            if (!wait_for_change)
                break;
        }

        {
            // Wait until the thread accepts the new mode, or 50 milliseconds.
            std::unique_lock<std::mutex> locker(mode_mutex);
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            using ms = std::chrono::milliseconds;
            mode_status.wait_until(locker, now + ms(50));
        }

        // Did our mode change over?
        if (requested_mode == -1)
            break;
    }
    if (retry == MAX_SETMODE_RETRY_COUNT)
    {
        fi_log(LOG_ERR, "%s: thread stuck, trying %d->%d(%d)", __PRETTY_FUNCTION__, current_mode, requested_mode, mode);
        rc = GEV_STATUS_ERROR;
    }

mode_bail:
    TP(FSP_FUNC, "%s: Done status=%d", __PRETTY_FUNCTION__, rc);
    return rc;
}

void fspWorker::setPacketInfo(int32_t packet_size, bool dnf)
{
    TP(FSP_FUNC, "%s: size=%d, dnf=%d", __PRETTY_FUNCTION__, packet_size, dnf);

    // Make sure the socket "Do Not Fragment" is set the correct way.
    fsp_socket.setDNF(dnf);

    // Sanity check our packet size.
    if (packet_size > CS_NETWORK::MAX_PACKET_SIZE)
    {
        fi_log(LOG_ERR, "%s: Packet Size is too big (%d)!", __PRETTY_FUNCTION__, packet_size);
        packet_size = 0;
    }

    if (packet_size > 0)
    {
        // Calculate the effective packet size. This is the
        // maximum number of bytes we can send (packet_size) minus
        // all the overhead stuff.
        stream_packet_size = packet_size - UDP_OVERHEAD - sizeof(fsp_header);

        // Make sure we're multiple of 32-bits.
        if (stream_packet_size & 3)
            stream_packet_size = stream_packet_size & ~3;
    }
}

void fspWorker::setPort(uint32_t new_port)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FSP_FUNC, "%s(%u)", __PRETTY_FUNCTION__, new_port);

    // Update the port register
    regs[GIGE_REG_SCP0].set_value(new_port);
}

/*!
 * @brief Return the port number from the correct register.
 *
 * @return The port number we're supposed to send the image data to. 0 on failure.
 */
uint16_t fspWorker::getPort()
{
    uint32_t  scp_reg_value;    // register with the port number
    uint16_t  dest_port_num;    // port number to send streaming data to
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // get the port number out of the register, which after byte swapping
    // is in the low 16 bits of the register.
    scp_reg_value = regs[GIGE_REG_SCP0].get_value();
    dest_port_num = (uint16_t)(scp_reg_value & 0xFFFF);

    PDEBUG(PDEBUG_HIGH, "%s: target port=%u", __PRETTY_FUNCTION__, dest_port_num);
    return dest_port_num;
}

/*!
 * @brief Return the IP address to send stream stuff to.
 *
 * @param ip_addr_ptr   (optional) IP address returned to send stuff to.
 * @return ip_addr (HOST BYTE ORDER) if success, 0 on failure.
 */
uint32_t fspWorker::getIpAddr(char *ip_addr_ptr)
{
    uint32_t    scda_reg_value;     // register with the ip address
    struct in_addr dest_ip_addr_st;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Clear out the target buffer.
    memset(dest_ip_addr, 0, CS_NETWORK::NET_ADDR_LEN);

    // Convert the register value to an ASCII IP address.
    scda_reg_value = regs[GIGE_REG_SCDA0].get_value();

    dest_ip_addr_st.s_addr = scda_reg_value;
    strncpy(dest_ip_addr, inet_ntoa(dest_ip_addr_st), CS_NETWORK::NET_ADDR_LEN - 1);

    // Is there a return parameter as well?
    if (ip_addr_ptr && (ip_addr_ptr != dest_ip_addr))
        strncpy(ip_addr_ptr, dest_ip_addr, CS_NETWORK::NET_ADDR_LEN);

    PDEBUG(PDEBUG_HIGH, "%s: target IP=%s[0x%x]", __PRETTY_FUNCTION__, dest_ip_addr, ntohl(scda_reg_value));
    return ntohl(scda_reg_value);
}

bool fspWorker::getRawLiveFrame(uint8_t *imageData, int& imageLength)
{
    bool rc;

    std::unique_lock<std::mutex> locker(last_mutex);    // Protect access to the last_ptr.

    // Set the address where we want to copy the last live image.
    last_len = 0;
    last_ptr = imageData;

    // Wait for the data to get copied over.
    rc = last_ready.wait_for(locker, std::chrono::milliseconds(250), [this] { return (last_len != 0); });
    imageLength = last_len;

    // Return the result.
    return rc;
}

/*!
 * @brief Ask the FPGA to capture a single frame.
 * @return 0 on success, negative on error.
 */
int fspWorker::get_live_frame()
{
    int rc = -1;
    bool ltr_mode;
    uint32_t mode = 0;  // ROC or BROC or 0
    uint32_t camera_state;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    camera_state = CameraServer::get_camera_state();
    TP(FSP_FUNC, "%s: camera_state=%d", __PRETTY_FUNCTION__, camera_state);

    // Check to see if the ICM is in IDLE mode - possible if we've bailed out of
    // capture or something earlier, or doing a black reference.
    if (fpga_ptr->icmIsPipelineIdle(nullptr))
    {
        LOG("%s: pipeline is idle", __PRETTY_FUNCTION__);
        return rc;
    }

    // Statistics
    Fastec::IntervalTimer frame_timer;
    frame_timer.start();
    // Statistics

    // Are we in LTR mode?
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    if (ltr_mode || fpga_ptr->ltrFasFireEnabled())
    {
        // Figure out if we're in ROC/BROC or normal.
        mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
    }

    // Are we in ARMED or UNARMED (also ROC/BROC are here) mode?
    if ((camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) ||
        ((camera_state == GIGE_CAMERA_STATE_ARMED) && mode))
    {
        // Setup to capture a frame.
        rc = capture_live_frame();
        if (rc != 0)
        {
            uint32_t sync_value;

            // We couldn't capture a frame, but if SyncIn is enabled that isn't too unusual.
            sync_value = regs[GIGE_REG_SYNC_IN_ENABLE].get_value();
            sync_value &= GIGE_SYNC_TYPE_MASK;

            // It's a problem if we're using the internal sync.
            if (sync_value == GIGE_SYNC_TYPE_INTERNAL)
                fi_log(LOG_ERR, "%s: Could not capture a live frame!", __PRETTY_FUNCTION__);
            else
                fi_log(LOG_DEBUG, "%s: Unable to capture a live frame, but SyncIn is enabled.", __PRETTY_FUNCTION__);

            // Time to leave.
            goto glf_bail;
        }
    }
    else if ((camera_state == GIGE_CAMERA_STATE_ARMED) || (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        PartMgr *pm_ptr = CameraServer::get_partmgr();
        uint32_t part_state = pm_ptr->getPartitionState();

        // Determine what the partition state is.
        if ((part_state == GIGE_PART_STATE_ARMED) || (part_state == GIGE_PART_STATE_TRIGGERED))
        {
            uint32_t reg_value;

            // Read the frame counter.
            if (!fpga_ptr->regRead(ICM_CAPTURE_FRAME_COUNT_LIVE, &reg_value))
            {
                fi_log(LOG_ERR, "%s: Unable to read ICM_CAPTURE_FRAME_COUNT_LIVE register", __PRETTY_FUNCTION__);
                goto glf_bail;
            }

            // Have we captured any frames?
            if (reg_value == ipm_info.frame_counter)
            {
                fpga_ptr->regRead(ICM_STATUS, &reg_value);
                TP(FSP_LIVE, "%s: Still at the same frame buffer [0x%08x]. ICM_STATUS=0x%08x", __PRETTY_FUNCTION__, ipm_info.frame_counter, reg_value);
                goto glf_bail;
            }

            // Remember the counter for next time.
            ipm_info.frame_counter = reg_value;

            // Figure out which buffer the ICM used.
            if (!fpga_ptr->regRead(ICM_LAST_FRAME_WRITTEN_PTR, &ipm_info.frame_addr_hw))
            {
                fi_log(LOG_ERR, "%s: Unable to read LAST_FRAME_WRITTEN register", __PRETTY_FUNCTION__);
                goto glf_bail;
            }
        }
        else
        {
            // The partition state is not in the same state we would expect. This could be normal,
            // if for example the capture just completed.
            TP(FSP_LIVE, "%s: wrong part_state [0x%x]", __PRETTY_FUNCTION__, part_state);
            goto glf_bail;
        }
    }
    else
    {
        // Not a valid state.
        goto glf_bail;
    }

    // Statistics
    frame_timer.stop();
    TP(FSP_STATS, "%s: frame capture: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
    // Statistics

    // All done.
    rc = 0;

glf_bail:
    return rc;
}

void fspWorker::disconnectStream()
{
    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Close the socket.
    fsp_socket.close();

    // Do we need to wake up the thread?
    if (getRunState() == TR_WAIT_FOR_CONNECT)
    {
        std::unique_lock<std::mutex> locker(mode_mutex);    // Protect access to requested_mode.

        // Wake the thread up so it can see we've closed
        // the socket.
        requested_mode = -1;
        mode_status.notify_one();
    }
}

/*!
 * @brief Fire a test packet to the stream receiver.
 *
 * @param packet_size Size of the packet to send
 */
int fspWorker::fireTestPacket(uint16_t packet_size)
{
    int32_t     rc;

    TP(FSP_FUNC, "%s: size=%u", __PRETTY_FUNCTION__, packet_size);

    // Send it on it's way. Make sure we backoff the length by the UDP overhead.
    rc = fsp_socket.testFire(packet_size - UDP_OVERHEAD);
    if (rc > 0)
        rc = packet_size;

    // All done.
    TP(FSP_PACKET, "%s: Finished, stat=%d", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Have the FPGA's IPM process the frame we just captured. All
 *    of the relevant information can be found in the ipm_info object.
 *
 * @param proc_info IPM specific processing information (see IMGPROC_XXX).
 * @return 0 on success, negative otherwise.
 */
int fspWorker::ipm_frame(uint32_t proc_info, bool do_live)
{
    int rc = -1;
    IPM_PROC::IPM_REQUEST ipm;

    // We are busy.
    ipm_info.busy = true;

    // Clear out our request structure.
    memset(&ipm, 0, sizeof(IPM_PROC::IPM_REQUEST));

    // Initialize the fields.
    ipm.image_addr = ((uint64_t)ipm_info.frame_addr_hw) << HEXWORD_SHIFT;
    ipm.proc_info = proc_info;

    // Get the raw pixel data processed properly through the IPM depending on
    // our current camera state.
    ipm.process_live = do_live;

    // Process the frame.
    if (!ipmProcRequest(&ipm))
    {
        fi_log(LOG_ERR, "%s: ipm_request for addr 0x%" PRIx64 " failed!", __PRETTY_FUNCTION__, ipm.image_addr);
        ipm_info.ipm_buf_addr = 0;
        ipm_info.image_len = -1;
        goto if_bail;
    }

    // Grab the output buffer address and the length.
    ipm_info.ipm_buf_addr = ipm.ipm_buf_addr;
    ipm_info.image_len = ipm.image_len;

    // We are done.
    rc = 0;

if_bail:
    // All done processing.
    ipm_info.busy = false;

    // All done.
    return rc;

}

/*!
 * @brief Parse the file ID from the file name. If it's a single file, like
 *    a snap shot, then the file name looks something like "ts3_000000.jpg".
 *    The ID in this example would be 0.
 *
 *    For image stacks, the file name IS the frame number which is the ID.
 *    So a file name might be "0000501.dng" and the ID would be 501.
 *
 *    Very large files, such as CAP files, might be split into two files,
 *    so the split file might look like "ts3_000005_1.cap". The ID of this
 *    file is 5, just like the first half of the file - they both share the
 *    same ID.
 *
 *    Metadata files such as "metadata.txt" don't have a file name, so we
 *    associate a special ID for them. -1 (0xFFFFFFFF) for the ".txt" file,
 *    -2 (0xFFFFFFFE) for the ".cfg" file, -3 (0xFFFFFFFD) for the ".xml" file,
 *    -4 (0xFFFFFFFC) for the ".pm" file.
 *
 * @param file_path The fully qualified file name.
 *
 * @return The ID, or 0 if we can't figure one out.
 */
uint32_t fspWorker::parse_id(const char *file_path)
{
    const char *file_ptr;

    // Initialize our pointer to the period leading the suffix.
    file_ptr = strrchr(file_path, '.');
    if (file_ptr == NULL)
    {
        fi_log(LOG_WARNING, "%s: Expected suffix period not found (%s)", __PRETTY_FUNCTION__, file_path);
        return 0;
    }

    // Check to see if this is a metadata file.
    if (strncmp(file_ptr, ".pm", 4) == 0)
        return -4;
    else if (strncmp(file_ptr, ".xml", 4) == 0)
        return -3;
    else if (strncmp(file_ptr, ".cfg", 4) == 0)
        return -2;
    else if (strncmp(file_ptr, ".txt", 4) == 0)
        return -1;

    // Check to see if this is a "split" file - a file that was split into 2 parts
    // because it exceeded 4Gbytes.
    if (*(file_ptr - 2) == '_')
    {
        fi_log(LOG_INFO, "%s: Split file detected (%s)", __PRETTY_FUNCTION__, file_path);

        // Back up the pointer to this spot.
        file_ptr -= 2;
    }

    // Now backup 7 characters.
    file_ptr -= 7;

    // We're either at an underscore character, or a digit. If it's an underscore
    // we need to move forward 1 character to a digit.
    if (*file_ptr == '_')
    {
        file_ptr++;
    }

    // We should be pointed at a digit now.
    if (!isdigit(*file_ptr))
    {
        fi_log(LOG_WARNING, "%s: Expected digit not found (%s)", __PRETTY_FUNCTION__, file_path);
        return 0;
    }

    // Conver to an integer and return.
    return atoi(file_ptr);
}

/*!
 * @brief Open a file and get its length
 *
 * @param   file_path   The file name
 * @param   file_len    Where to return the file's size
 *
 * @return the file descriptor if successful, otherwise a negative value.
 */
int fspWorker::open_file_for_xfer(const char *file_path, off_t *file_len)
{
    int     fd;

    // Initialize our values.
    *file_len = 0;
    aligned_len = 0;

    // Open our file.
    fd = open(file_path, O_RDONLY);
    if (fd == -1)
    {
        fi_log(LOG_ERR, "%s: Unable to open (%s), err=%d(%s)", __PRETTY_FUNCTION__, file_path, errno, strerror(errno));
        goto bail_1;
    }

    // Figure out how big the file is.
    *file_len = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);

bail_1:
    // All done.
    return fd;
}

/*!
 * @brief This function transfers the specified file out the streaming interface.
 *
 * @param file_path Fully qualified file name.
 *
 * @return TRUE on success, FALSE for an error.
 */
bool fspWorker::stream_single_file(const char *file_path)
{
    int file_fd;
    bool rc;
    off_t file_len;
    size_t total_xfer;
    size_t amount_sent;
    uint32_t file_id;

    TP(FSP_FUNC, "%s: retry_index=0x%08x, file='%s'", __PRETTY_FUNCTION__, getRetryIndex(), file_path);

    // Assume failure.
    rc = false;

    // Setup for the file transfer.
    file_fd = open_file_for_xfer(file_path, &file_len);
    if ((file_fd < 0) || (file_len < 0))
    {
        fi_log(LOG_ERR, "%s: open-file_for_xfer() failed", __PRETTY_FUNCTION__);
        goto bail_1;
    }

    // Start at the beginning and stream the whole file.
    total_xfer = file_len;

    // Get the ID for this file.
    file_id = fspWorker::factory()->parse_id(file_path);
    TP(FSP_FILESTREAM, "%s: FileID=0x%08x, file_len=%" PRId64, __PRETTY_FUNCTION__, file_id, total_xfer);

    // Make sure we don't have a 0 for an ID. Retain the IDs for the metadata files.
    if ((file_id != (uint32_t) -1) && (file_id != (uint32_t) -2) && (file_id != (uint32_t) -3) && (file_id != (uint32_t) -4))
    {
        // Bump the ID by one for any file other than the
        // metadata files - we want to be 1 based, not 0.
        ++file_id;
    }
    TP(FSP_FILESTREAM, "%s: FileID=0x%08x", __PRETTY_FUNCTION__, file_id);

    // Reset our block ID.
    fsp_socket.setBlockID(0);

    // Tell the stream the frame number.
    fsp_socket.setFrameNum(file_id);

    // Stream the file.
    amount_sent = fsp_socket.sendfile(file_fd, 0, total_xfer);
    if (amount_sent != total_xfer)
        fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);
    else
        rc = true;

    // Close the file.
    close(file_fd);

bail_1:
    // All done.
    return rc;
}

/*!
 * @brief This method is used to encapsulate the logic to allow for a client
 *        to connect to the FSP stream.
 *
 * It is intended that this be a thread, to allow the connection to happen
 * asynchronously, or in parallel to, the GRPC interface.
 */
void fspWorker::connect_socket()
{
    TP(FSP_STATE, "%s: starting.", __PRETTY_FUNCTION__);

    // Keep running until we are asked to stop.
    while (!stop_request)
    {
        // Check to see if the socket already has a connection.
        if (fsp_socket.isValid())
        {
            // Sleep for a bit.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(100));
        }
        else
        {
            // See if we get a client to connect.
            fsp_socket.waitForClient(500);
        }
    }

    TP(FSP_STATE, "%s: Done.", __PRETTY_FUNCTION__);
}

/*!
 * @brief Callback for NFTW walking of a folder.
 */
int fspWorker::direct_walk_cb(const char *fpath, const struct stat *sb, int /*typeflag*/, struct FTW * /*ftwbuf*/)
{
    fspWorker *fsp = fspWorker::factory();

    // Are we supposed to stop?
    if (fsp->modeChangePending())
    {
        LOG("%s#%d: stop, mode change pending", __PRETTY_FUNCTION__, __LINE__);
        errno = 0;
        return -1;
    }

    // We don't really care about folders, only files.
    if (S_ISREG(sb->st_mode))
    {
        // Add this file to our path vector.
        fsp->walk_paths.push_back(std::string(fpath));
    }

    // Keep on going.
    return 0;
}

/*!
 * @brief Callback function to check whether we should abort metadata loading.
 *
 * @return true if we should abort, false to continue
 */
bool fspWorker::metadataAbortCheck()
{
    if (requested_mode != -1)
        return true;
    else
        return false;
}

/*!
 * @brief Stream raw metadata over GOE
 *
 * @note    For a DDR2 capture, use the PME to extract metadata from the partition images.
 *      For LTR capture, read metadata sectors from the drive.
 *      Once the metadata is in a buffer, stream it out.
 *
 * @return 0 on success, <0 for an error.
 */
int fspWorker::raw_metadata_stream()
{
    int rc;
    int frames_transferred; // Number of frames we transferred in a loop.
    int bb_index;           // Bounce buffer index.
    uint8_t *bb_ptr;
    uint32_t cur_frame;
    uint32_t end_frame;
    uint32_t start_frame;
    FiFPGA *fpga_ptr;
    PartMgr *pm_ptr;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Assume failure.
    rc = -1;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    pm_ptr = CameraServer::get_partmgr();

    // Sanity check.
    if (pm_ptr->getPartitionState() != GIGE_PART_STATE_COMPLETE)
    {
        fi_log(LOG_ERR, "%s: partition is in the wrong state! [0x%x]", __PRETTY_FUNCTION__, pm_ptr->getPartitionState());
        return -1;
    }

    // Initialize our start and ending frame numbers.
    start_frame = get_start_frame_number();
    end_frame = get_end_frame_number();
    frames_transferred = 0;

    // Get our bounce buffer.
    bb_index = bb_get_avail();

    // Get a pointer to our buffer.
    bb_ptr = (uint8_t *)bb_mmap(bb_index, MAX_METADATA_XFER_LEN);
    if (bb_ptr == (uint8_t *)MAP_FAILED)
    {
        fi_log(LOG_ERR, "%s: Unable to map the bounce buffer [index=%d]", __PRETTY_FUNCTION__, bb_index);
        goto bail;
    }

    // Loop until we've sent all the frames of metadata.
    for (cur_frame = start_frame; cur_frame < end_frame; cur_frame += frames_transferred)
    {
        TP(FSP_FUNC, "%s: cur_frame=%u, end_frame=%u, frames_transferred=%d", __PRETTY_FUNCTION__, cur_frame, end_frame, frames_transferred);

        // Get our bounce buffer filled.
        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            frames_transferred = pm_ptr->pme_ltr_fill_buffer(cur_frame, end_frame, bb_ptr);
        }
        else
        {
            frames_transferred = pm_ptr->pme_fpga_fill_buffer(cur_frame, end_frame, bb_index);
        }

        // Did the fill go OK?
        if (frames_transferred == 0)
        {
            fi_log(LOG_ERR, "%s: Unable to fill the transfer buffer!", __PRETTY_FUNCTION__);
            goto bail;
        }

#ifdef DUMP_RAW_METADATA
        {
            int fd;
            char dump_path[128];
            snprintf(dump_path, 127, "/tmp/dump-%u.dat", cur_frame);
            fd = open(dump_path, O_CREAT | O_WRONLY, 0666);
            write(fd, bb_ptr, frames_transferred * FRAME_METADATA_HW_BYTES);
            close(fd);
        }
#endif /* DUMP_RAW_METADATA */

        // Are we supposed to stop?
        if (metadataAbortCheck())
        {
            // The user cancelled the opeation. This isn't an "error" per-se, so
            // reset our return code.
            fi_log(LOG_INFO, "%s: Cancel detected!", __PRETTY_FUNCTION__);
            rc = 0;
            goto bail;
        }

        // Send the data onwards.
        TP(FSP_FUNC, "%s: Sending %u PFMD 'frames' (%u bytes)", __PRETTY_FUNCTION__, frames_transferred, frames_transferred * FRAME_METADATA_HW_BYTES);
        if (fsp_socket.isValid())
        {
            // Send it to the FSP stream.
            rc = fsp_socket.streamData(bb_ptr, frames_transferred * FRAME_METADATA_HW_BYTES, FSP_PAYLOAD_FILE);
        }

        // Do we have a GRPC stream?
        if (grpcStreamIsConnected())
        {
            GRPC_SERVER::GRPC_IMAGE image_info;

            // Populate our image structure.
            image_info.format = GIGE_STREAM_FORMAT_META_ONLY;
            image_info.width = 0;
            image_info.height = 0;
            image_info.frame_num = cur_frame;
            image_info.image_len = frames_transferred * FRAME_METADATA_HW_BYTES;
            image_info.image_data = bb_ptr;
            image_info.header_data = nullptr;
            image_info.header_len = 0;

            // Send it on.
            if (grpcStreamImage(image_info))
                rc = image_info.header_len + image_info.image_len;
            else
                rc = 0;
        }

        if (!rc)
        {
            // Couldn't stream the data, so return an error.
            fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);
            rc = -1;
            goto bail;
        }
    }

    // Determine our status.
    // cppcheck-suppress redundantAssignment
    rc = (cur_frame >= end_frame) ? 0 : -1;

bail:
    // Release our buffer. This will also unmap the buffer, if appropriate.
    bb_release(bb_index);

    // All done.
    TP(FSP_FUNC, "%s: done (%d)", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Transfer a file from SSD, SDCard or USB to the network.
 *
 * @return 0 on success, <0 for an error.
 */
int fspWorker::do_file_stream()
{
    int rc;
    struct stat file_stat;              // Used to determine if we're file or folder.

    // Assume failure.
    rc = -1;

    // Get the statistics of the stream_file so we can determine if
    // we're dealing with a folder or file.
    if (stat(stream_file.c_str(), &file_stat))
    {
        fi_log(LOG_ERR, "%s: Unable to stat path '%s'. %d(%s)", __PRETTY_FUNCTION__, stream_file.c_str(), errno, strerror(errno));
        goto bail;
    }

    // Are we looking at a folder or a file?
    if (S_ISDIR(file_stat.st_mode))
    {
        // Reset our vector of paths.
        walk_paths.clear();

        // Walk the tree and build a list of files.
        if (nftw(stream_file.c_str(), fspWorker::direct_walk_cb, 64, FTW_PHYS) == -1)
        {
            // Make sure it's a "real" error.
            if (errno != 0)
            {
                fi_log(LOG_ERR, "%s: Unable to walk (%s), err=%d(%s)", __PRETTY_FUNCTION__, stream_file.c_str(), errno, strerror(errno));
                goto bail;
            }
        }

        // Sort the result.
        std::sort(walk_paths.begin(), walk_paths.end());

        // Walk the sorted vector and stream the files.
        fspWorker *fsp = fspWorker::factory();
        for (std::vector<std::string>::iterator it = walk_paths.begin(); it != walk_paths.end(); ++it)
        {
            // Are we supposed to stop?
            if (fsp->modeChangePending())
            {
                LOG("%s#%d: stop, mode change pending", __PRETTY_FUNCTION__, __LINE__);
                errno = 0;
                rc = -1;
                goto bail;
            }

            // Stream the specified file.
            rc = (stream_single_file(it->c_str()) ? 0 : -1);
            if (rc < 0)
            {
                fi_log(LOG_ERR, "%s: stream_single_file(%s) failed", __PRETTY_FUNCTION__, it->c_str());
                break;
            }
        }
    }
    else
    {
        // Stream the specified file.
        rc = (stream_single_file(stream_file.c_str()) ? 0 : -1);
    }

bail:
    return rc;
}

/*!
 * @brief Handle live RAW view until the mode changes.
 *
 * In the live raw loop, we wait for a signal or a timeout. Then we check whether
 * the mode has changed or there is a new frame to process.
 *
 * @return 0 on success, or an error code otherwise.
 */
int fspWorker::do_raw_loop()
{
    int rc = 0;
    int msecs;
    int live_state;
    int ifg_millisec;
    bool live_done;
    bool stream_err(false);
    bitmode_t bitmode;
    uint32_t fv_fourcc;
    uint32_t dma_mod;
    uint32_t dma_dims;
    uint32_t proc_info;
    uint32_t pc_img_len;
    std::unique_ptr<pm_metadata> pm_meta(new pm_metadata);

    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Fastec::IntervalTimer live_timer;
    Fastec::IntervalTimer frame_timer;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state.
    live_state = LIVE_SETUP;

    // Keep going until told to stop.
    live_done = false;
    while (!live_done)
    {
        TP(FSP_STATE_VERBOSE, "%s: state=%d", __PRETTY_FUNCTION__, live_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (live_state)
        {
        case LIVE_SETUP:
            // Get a buffer for our image processing.
            ipm_info.bb_index = bb_get_avail();
            ipm_info.bb_addr = bb_get_avail_addr(ipm_info.bb_index);

            // Map the memory.
            ipm_info.bb_ptr = (uint8_t *)bb_mmap(ipm_info.bb_index, 0);
            if (ipm_info.bb_ptr == (uint8_t *)MAP_FAILED)
            {
                FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
                rc = -1;
                live_state = LIVE_CHECK_MODE;
                live_done = true;
                break;
            }

            // Raw images. Did we capture 8 or 10/12 bits?
            bitmode = regs[GIGE_REG_BIT_MODE].get_value();
            if (IS_10_BIT_MODE(bitmode) || IS_12_BIT_MODE(bitmode))
            {
                // 10-bit or 12-bit captures mean 16 bit raw.
                proc_info = IMGPROC_TYPE_RAW16;
                fv_fourcc = FAS_VIDEO::FOURCC_BYR2;
            }
            else
            {
                // 8 bit captures mean 8 bit raw.
                proc_info = IMGPROC_TYPE_RAW8;
                fv_fourcc = FAS_VIDEO::FOURCC_BYR1;
            }

            // Get a copy of the metadata - we have to tweak it.
            memcpy(pm_meta.get(), pm_ptr->getActivePartMetadata().get(), sizeof(pm_metadata));
            pm_meta->roi_height = regs[GIGE_REG_HEIGHT].get_value();
            pm_meta->roi_width = regs[GIGE_REG_WIDTH].get_value();
            pm_meta->bit_sel = bitmode;

            // What will the size of the image in PC memory be?
            pc_img_len = get_processed_size(proc_info, pm_meta->roi_width, pm_meta->roi_height);

            // Get our header.
            dnld_header_buffer = FAS_VIDEO::FiVideo::CreateImageHeader(FAS_VIDEO::FI_TIFF_STACK_CODEC, fv_fourcc, pm_meta.get());
            if ((dnld_header_buffer == nullptr) || (dnld_header_buffer->size() == 0))
            {
                // Since we are in RAW LOOP, we are sending TIFF files, and TIFF files SHOULD have headers.
                fi_log(LOG_ERR, "%s: Unable to create header!", __PRETTY_FUNCTION__);
                rc = -1;
                live_state = LIVE_CHECK_MODE;
                live_done = true;
                break;
            }

            // Setup for raw processing.
            dma_mod = IM2PCIE_DMA_IMAGE_DATA;
            dma_mod |= IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
            if (IS_10_BIT_MODE(bitmode) || IS_12_BIT_MODE(bitmode))
                dma_mod |= IM2PCIE_DMA_16BIT_RAW_FMT;
            else
                dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT;

            // The image dimensions will be constant.
            dma_dims = (regs[GIGE_REG_HEIGHT].get_value() << IPM_FRAME_DIM_Y_SHIFT) | regs[GIGE_REG_WIDTH].get_value();

            // Move on to getting a frame.
            live_state = LIVE_ACQUIRE_FRAME;
            break;
        case LIVE_ACQUIRE_FRAME:
            // Get the current time so we can calculate how much time each frame
            // takes to proceses.
            live_timer.start();

            // Statistics
            frame_timer.start();
            // Statistics

            // Get a frame from the sensor. This initializes the ipm_info structure.
            rc = get_live_frame();
            if (rc == 0)
            {
                // Move on to proceessing the frame.
                live_state = LIVE_PROCESS_FRAME;
            }
            else
            {
                // Probably need to transition state.
                live_state = LIVE_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: acquire frame: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case LIVE_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Get the frame processed by the IPM.
            rc = ipm_frame(proc_info, true);
            if (rc == 0)
            {
                // Move on to transferring the frame.
                live_state = LIVE_TRANSFER_FRAME;
            }
            else
            {
                // Probably need to transition state.
                fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
                live_state = LIVE_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: IPM: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case LIVE_TRANSFER_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // memset(ipm_info.bb_ptr, 0, ipm_info.image_len);

            // Have the FPGA transfer the image into our memory.
            rc = fpga_ptr->imageDownload(dma_mod, dma_dims, ipm_info.ipm_buf_addr,
                                         ipm_info.bb_addr, ipm_info.image_len, pc_img_len);
            if (rc >= 0)
            {
                // Move on to processing the frame.
                live_state = LIVE_STREAM_FRAME;
                rc = 0;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
                live_state = LIVE_CHECK_MODE;
            }

            // Release our P1 buffer.
            ipm_proc_release(ipm_info.ipm_buf_addr);
            ipm_info.ipm_buf_addr = 0;

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: DMA: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
#ifdef DO_STATISTICS
            times_measure.push_back(frame_timer.lastMeasurement);
#endif /* DO_STATISTICS */
            // Statistics
            break;
        case LIVE_STREAM_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Send the data onwards.
            if (fsp_socket.isValid())
            {
                // Send it to the FSP stream.
                rc = fsp_socket.streamData(ipm_info.bb_ptr, pc_img_len, FSP_PAYLOAD_FILE /* ALM_TODO_344: FSP_PAYLOAD_RAW */, dnld_header_buffer->data(), dnld_header_buffer->size());
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;

                // Populate our image structure.
                image_info.format = GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW;
                image_info.width = pm_meta->roi_width;
                image_info.height = pm_meta->roi_height;
                image_info.frame_num = 0;
                image_info.image_len = pc_img_len;
                image_info.image_data = ipm_info.bb_ptr;
                image_info.header_data = dnld_header_buffer->data();
                image_info.header_len = dnld_header_buffer->size();

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }

            if (!rc)
            {
                fi_log(LOG_ERR, "%s: Unable to stream the frame!", __PRETTY_FUNCTION__);
                stream_err = true;
            }

            // Statistics
            frame_timer.stop();
            TP(FSP_STATS, "%s: frame stream: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics

            // Go acquire the next frame.
            live_state = LIVE_CHECK_MODE;
            break;
        case LIVE_CHECK_MODE:
            // Statistics
            // Stop the timer.
            live_timer.stop();
            // Statistics

            // If we are not in error mode, go ahead and figure out what our
            // interframe gap time should be.
            if (!stream_err)
            {
                // What is our interframe delay?
                ifg_millisec = get_interframe_delay();

                // The user has specified an interframe gap time, which will control how
                // fast we stream (or attempt to stream) frames. How much of that gap time
                // is still available?
                msecs = ifg_millisec - ((int)((live_timer.lastMeasurement + 0.0005) * 1000.0));
                if (msecs < 0)
                {
                    TP(FSP_STATS, "%s: Loop took too long (%lf)!", __PRETTY_FUNCTION__, live_timer.lastMeasurement);
                    msecs = 0;
                }
                else
                {
                    TP(FSP_STATS, "%s: total: %lf secs.", __PRETTY_FUNCTION__, live_timer.lastMeasurement);
                }
            }
            else
            {
                FI_LOG_TAG(__FILE__, __LINE__, "%s: Got a stream error.", __PRETTY_FUNCTION__);
                live_done = true;
                rc = -1;
            }

            // Assume we'll go around and get the next frame.
            live_state = LIVE_ACQUIRE_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, live_state);
            live_state = LIVE_CHECK_MODE;
            break;
        }

        // Wait for either a mode change or time to get the next frame.
        if (!live_done)
        {
            if (wait_new_mode_or_timeout(msecs) == 0)
            {
                // We are done.
                live_done = true;
                rc = 0;
            }
        }
    }

    // Release the bounce buffer.
    bb_release(ipm_info.bb_index);

    // Release the IPM buffer, if we have one.
    if (ipm_info.ipm_buf_addr != 0)
    {
        ipm_proc_release(ipm_info.ipm_buf_addr);
        ipm_info.ipm_buf_addr = 0;
    }

    // Return our status.
    return rc;
}

bool fspWorker::sess_setup(uint64_t& xfer_length, uint64_t& im_addr, uint32_t& frame_count,
                           uint32_t& session_length, uint32_t& dma_mod, bool& subpartition,
                           bool& split_transfer)
{
    uint32_t capture_session;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    // Are we doing a full capture save?
    if ((get_start_frame_number() == 0) && ((uint32_t)get_end_frame_number() == (pm_ptr->getReviewPartNumFrames() - 1)))
    {
        // We are doing a full capture.
        subpartition = false;
        split_transfer = false;

        // Calculate the size of the video file we're going to create. Make sure we
        // incorporate the metadata, including the correction table, if necessary.
        xfer_length = pm_ptr->getReviewPartLength(&capture_session);
        if (xfer_length == 0)
        {
            fi_log(LOG_ERR, "%s: xfer_length is 0?!!", __PRETTY_FUNCTION__);
            return false;
        }

        // Can we shorten the length? Check to see if the partition has wrapped.
        if (capture_session < pm_ptr->getSessionLength())
        {
            // Save the frame count so we can restore that.
            frame_count = pm_ptr->getReviewPartNumFrames();

            // Save the old session length and change it to be
            // our new (shorter) value.
            session_length = pm_ptr->getSessionLength();
            pm_ptr->adjustReviewPartition(capture_session);
        }
    }
    else
    {
        uint64_t im_start_addr;
        uint64_t im_stop_addr;

        // We are saving a partial capture.
        subpartition = true;
        session_length = 0;

        // How many frames are we talking about?
        frame_count = get_end_frame_number() - get_start_frame_number() + 1;

        // Find the starting location (inlusive) based on the starting frame number.
        im_start_addr = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), get_start_frame_number(), false);

        // Find the ending location (not-inclusive) based on the ending frame number.
        im_stop_addr = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), get_end_frame_number(), false) + pm_ptr->getFrameSize();

        // Are we going to have to do a split transfer?
        split_transfer = (im_stop_addr < im_start_addr);

        // How many total bytes are we going to be transferring?
        xfer_length = pm_ptr->getPartitionSizeReserved() + (pm_ptr->getFrameSize() * frame_count);
    }

    // We are going to transfer the data raw - no formatting.
    dma_mod = IM2PCIE_DMA_GENERIC_DATA;
    dma_mod |= IM2PCIE_DMA_HIGHSPEED_BYPASS;

    // Get the start address for the partition.
    im_addr = pm_ptr->getReviewPartStartAddr();

    // All is well
    return true;
}

/*!
 * @brief Transfer the session to the network.
 *
 * @return 0 on success, <0 for an error.
 */
int fspWorker::do_session_stream()
{
    int rc = -1;
    int session_state;
    int transfer_count(0);
    bool done;
    bool ltr_mode;
    bool subpartition;
    bool split_transfer;
    uint32_t dma_mod;
    uint64_t im_addr;
    int bb_index(0);
    uint64_t bb_addr(0);
    uint8_t *xfer_buffer(nullptr);
    uint64_t xfer_length;
    uint64_t total_transfer_length;
    uint64_t transferred_length;
    uint32_t frame_count;
    uint32_t session_length;

    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    Fastec::IntervalTimer sess_timer;

    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    session_state = SESS_SETUP;

    // Determine if we're in LTR mode or not.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    if (ltr_mode)
    {
        LOG("%s: LTR mode not supported!", __PRETTY_FUNCTION__);
        return -1;
    }

    // Initialize our session length variable to indicate that
    // nothing has changed.
    session_length = 0;
    frame_count = 0;
    xfer_length = 0;
    transferred_length = 0;
    subpartition = false;
    split_transfer = false;

    // Get the current time so we can calculate how much time the trasnfer takes.
    sess_timer.start();

    // Keep going until told to stop.
    done = false;
    while (!done)
    {
        TP(FSP_STATE_VERBOSE, "%s: state=%d", __PRETTY_FUNCTION__, session_state);

        // Process based on the current state.
        switch (session_state)
        {
        case SESS_SETUP:
            // Get a buffer for our image processing.
            bb_index = bb_get_avail();
            bb_addr = bb_get_avail_addr(bb_index);

            // Map the memory.
            xfer_buffer = (uint8_t *)bb_mmap(bb_index, DL_CHUNK_SIZE);
            if (xfer_buffer == (uint8_t *)MAP_FAILED)
            {
                FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
                done = true;
                break;
            }

            // Setup parameters for the transfer.
            if (!sess_setup(total_transfer_length, im_addr, frame_count, session_length, dma_mod, subpartition, split_transfer))
            {
                fi_log(LOG_DEBUG, "%s: sess_setup() failed.", __PRETTY_FUNCTION__);
                done = true;
                break;
            }
            fi_log(LOG_INFO, "%s: total_transfer_length=0x%" PRIx64 ", frame_count=%d, subpart=%d, split_transfer=%d", __PRETTY_FUNCTION__, total_transfer_length, frame_count, subpartition, split_transfer);

            // Reset our block ID.
            fsp_socket.setBlockID(0);

            // Initialize our counter.
            xfer_length = 0;

            // Move on.
            if (subpartition)
                session_state = SESS_SUBPART_TRANSFER_HEADER;
            else
                session_state = SESS_TRANSFER_CHUNK;
            break;
        case SESS_SUBPART_TRANSFER_HEADER:
            // Have the FPGA transfer the header to our transfer buffer.
            xfer_length = pm_ptr->getPartitionSizeReserved();

            // Download the capture header.
            TP(FSP_SESSSTREAM, "%s: SUBPARTHDR [%d] im_addr=0x%" PRIx64 ", header_length=%u", __PRETTY_FUNCTION__, transfer_count, im_addr, xfer_length);
            rc = fpga_ptr->imageDownload(dma_mod, 0, im_addr, bb_addr, xfer_length);
            if (rc >= 0)
            {
                pm_metadata *metadata_ptr;

                // Overlay our metadata structure on to the buffer we just wrote.
                metadata_ptr = (pm_metadata *)xfer_buffer;

                // Adjust some of the settings. Our frame range for the CAP file goes from 0 (format_info.current_frame)
                // to frame_count - 1 (format_info.end_frame).
                metadata_ptr->frame_count = frame_count;
                if (pm_ptr->getReviewPartTrigger() < (uint32_t)get_start_frame_number())
                {
                    metadata_ptr->first_frame = get_start_frame_number() - pm_ptr->getReviewPartTrigger();
                    metadata_ptr->trig_frame = 0;
                    metadata_ptr->skip_count = metadata_ptr->first_frame;
                }
                else
                {
                    metadata_ptr->first_frame = 0;
                    metadata_ptr->trig_frame = pm_ptr->getReviewPartTrigger() - get_start_frame_number();
                    metadata_ptr->skip_count = 0;
                }

                // Adjust the image address to be at the first frame.
                im_addr = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), get_start_frame_number(), false);

                // Move on to get a chunk-o-data.
                session_state = SESS_SUBPART_TRANSFER_FIRST_SPLIT;
                rc = 0;
            }

            // Did we have troubles?
            if (rc < 0)
            {
                fi_log(LOG_ERR, "%s [%d]: Unable to DMA chunk!", __PRETTY_FUNCTION__, __LINE__);
                session_state = SESS_CHECK_MODE;
            }
            break;
        case SESS_SUBPART_TRANSFER_FIRST_SPLIT:
        {
            uint32_t last_frame = pm_ptr->get_im_buf_count() - 1;
            uint64_t end_of_part = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), last_frame, true) + pm_ptr->getFrameSize();
            uint64_t remaining_len = end_of_part - im_addr;
            uint64_t split_len;

            // Do we have less than a chunk to transfer before the end of the partition?
            if ((remaining_len + xfer_length) < DL_CHUNK_SIZE)
                split_len = remaining_len;
            else
                split_len = (DL_CHUNK_SIZE - xfer_length);

            // Have the FPGA transfer the data to our transfer buffer.
            TP(FSP_SESSSTREAM, "%s: SUBPART, [%d] im_addr=0x%" PRIx64 ", xfer_length=%" PRIu64 ", split_len=%" PRIu64, __PRETTY_FUNCTION__, transfer_count, im_addr, xfer_length, split_len);
            rc = fpga_ptr->imageDownload(dma_mod, 0, im_addr, bb_addr + xfer_length, split_len);
            if (rc >= 0)
            {
                // Update our image memory address.
                im_addr += split_len;

                // Update the number of bytes we want to send.
                xfer_length += split_len;

                // Are we done with the first split?
                if (im_addr == end_of_part)
                {
                    // We are done with the split.
                    TP(FSP_SESSSTREAM, "%s: Done with the split", __PRETTY_FUNCTION__);
                    split_transfer = false;

                    // Reset our image address to the beginning of the capture.
                    im_addr = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), 0, true);
                }

                // Where do we go next?
                if (xfer_length < DL_CHUNK_SIZE)
                    session_state = SESS_TRANSFER_CHUNK;
                else
                    session_state = SESS_STREAM_CHUNK;

                // All is well.
                rc = 0;
            }

            // Did we have troubles?
            if (rc < 0)
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s [%d]: Unable to DMA chunk!", __PRETTY_FUNCTION__, __LINE__);
                session_state = SESS_CHECK_MODE;
            }
            break;
        }
        case SESS_TRANSFER_CHUNK:
        {
            uint64_t download_len;
            uint64_t remaining_len = total_transfer_length - transferred_length;

            // Do we have less than a chunk to transfer before the end of the partition?
            if ((remaining_len + xfer_length) < DL_CHUNK_SIZE)
                download_len = remaining_len;
            else
                download_len = (DL_CHUNK_SIZE - xfer_length);

            // Have the FPGA transfer the data to our transfer buffer.
            TP(FSP_SESSSTREAM, "%s: [%d] im_addr=0x%" PRIx64 ", xfer_length=%" PRIu64 ", download_len=%" PRIu64, __PRETTY_FUNCTION__, transfer_count, im_addr, xfer_length, download_len);
            rc = fpga_ptr->imageDownload(dma_mod, 0, im_addr, bb_addr + xfer_length, download_len);
            if (rc >= 0)
            {
                // Update our image memory address.
                im_addr += download_len;

                // Update the number of bytes we want to send.
                xfer_length += download_len;

                // Move on to processing the frame.
                session_state = SESS_STREAM_CHUNK;
                rc = 0;
            }
            else
            {
                fi_log(LOG_ERR, "%s [%d]: Unable to DMA chunk!", __PRETTY_FUNCTION__, __LINE__);
                session_state = SESS_CHECK_MODE;
            }
            break;
        }
        case SESS_STREAM_CHUNK:
            if (fsp_socket.isValid())
            {
                // Tell the stream the frame number (+1 for our protocol needs).
                fsp_socket.setFrameNum(transfer_count + 1);

                // Send the data onwards.
                rc = fsp_socket.streamData(xfer_buffer, xfer_length, FSP_PAYLOAD_FILE);
            }

            // Do we have a GRPC stream?
            if (grpcStreamIsConnected())
            {
                GRPC_SERVER::GRPC_IMAGE image_info;
                GIGE_REG_CACHE& regs = CameraServer::get_regs();

                // Populate our image structure.
                image_info.format = GIGE_STREAM_FORMAT_CAPTURE;
                image_info.width = regs[GIGE_REG_WIDTH].get_value();
                image_info.height = regs[GIGE_REG_HEIGHT].get_value();
                image_info.frame_num = transfer_count + 1;
                image_info.image_len = xfer_length;
                image_info.image_data = xfer_buffer;
                image_info.header_data = nullptr;
                image_info.header_len = 0;

                // Send it on.
                if (grpcStreamImage(image_info))
                    rc = image_info.header_len + image_info.image_len;
                else
                    rc = 0;
            }
            if (!rc)
                fi_log(LOG_ERR, "%s: Unable to stream the chunk %d!", __PRETTY_FUNCTION__, transfer_count);

            // Reset our transfer count.
            xfer_length = 0;

            // Update our number of transfers.
            transfer_count++;

            // Update the number of bytes we've transferred.
            transferred_length += xfer_length;

            // Check for a mode change before getting the next frame.
            session_state = SESS_CHECK_MODE;
            break;
        case SESS_CHECK_MODE:
            // Are we done?
            if (transferred_length == total_transfer_length)
            {
                // We're done.
                done = true;
                rc = 0;
                break;
            }

            // Move to the next chunk.
            if (split_transfer)
                session_state = SESS_SUBPART_TRANSFER_FIRST_SPLIT;
            else
                session_state = SESS_TRANSFER_CHUNK;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, session_state);
            session_state = SESS_CHECK_MODE;
            break;
        }

        // Check for a mode change.
        if (wait_new_mode_or_timeout(0) == 0)
        {
            // We are done.
            done = true;
            rc = 0;
        }
    }

    // Now stop the timer. All per-frame processing has completed.
    sess_timer.stop();
    TP(FSP_STATS, "%s: transfer done in %lf secs.", __PRETTY_FUNCTION__, sess_timer.lastMeasurement);

    // Do we need to adjust the session length?
    if (session_length)
        pm_ptr->adjustReviewPartition(session_length, frame_count);

    // Release the bounce buffer.
    if (bb_index >= 0)
        bb_release(bb_index);

    // Return our status.
    TP(FSP_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief JPEG encode a frame. Relevant information is found in
 *    the ipm_info object.
 *
 * @return 0 on success, negative otherwise.
 */
int fspWorker::encode_frame(uint8_t *image_ptr, JPEG_PROC::BUF_REQUEST *request)
{
    int rc = -1;
    uint32_t uv_offset;  // Number of bytes (HEXWORD alignment) to the UV-plane.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Statistics
    Fastec::IntervalTimer jpeg_timer;
    // Statistics

    TP(FSP_FUNC, "%s: image_ptr=%p", __PRETTY_FUNCTION__, (void *)image_ptr);

    // Get our dimensions.
    request->width = regs[GIGE_REG_WIDTH].get_value();
    request->height = regs[GIGE_REG_HEIGHT].get_value();

    // Calculate our length for the Y plane so we can find the start ov the UV plane.
    // We transfered the frame directly from image memory, so the UV plane will be on
    // a HEXWORD boundary.
    uv_offset = convert_to_HEXWORD(request->width * request->height) << HEXWORD_SHIFT;

    // Now setup the pointers to the data.
    request->src_pixels_Y = image_ptr;
    request->src_pixels_UV = image_ptr + uv_offset;

    // ----------------------------------------------------------------------
    // #define ALM_SAVE_DATA
#ifdef ALM_SAVE_DATA
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;
        uint32_t yuv_len;

        if (request->pixel_format == JPEG_UTIL::Y8)
            yuv_len  = request->width * request->height;
        else
            yuv_len = uv_offset + ((request->width * request->height) >> 1);

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.raw", file_no);
        // ++file_no;   // Done when we save the encoded data.
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(image_ptr, yuv_len, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_DATA */
    // ----------------------------------------------------------------------

    // Statistics
    jpeg_timer.start();

    // JPEG encode the data.
    if (jpeg_buf_request(request))
        rc = 0;

    // Statistics
    jpeg_timer.stop();
    TP(FSP_STATS, "%s: JPEG encoding: %u bytes, %lf secs.", __PRETTY_FUNCTION__, request->jpeg_len, jpeg_timer.lastMeasurement);
    // Statistics

    // ----------------------------------------------------------------------
    // #define ALM_SAVE_JPG
#ifdef ALM_SAVE_JPG
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.jpg", file_no);
        ++file_no;
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(request->jpeg_buffer, request->jpeg_len, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_JPG */
    // ----------------------------------------------------------------------

    return rc;
}

void fspWorker::pauseStream()
{
    // Pass it on.
    fsp_socket.pause();
}

void fspWorker::resumeStream()
{
    // Pass it on.
    fsp_socket.resume();
}

void fspWorker::setStreamPath(const char *file_path, uint32_t file_len)
{
    // Clear the file name buffer.
    stream_file.clear();

    // Figure out how many characters we can transfer.
    if (file_len > (PATH_MAX - 1))
        file_len = PATH_MAX - 1;

    // Copy the name over.
    stream_file.assign(file_path, file_len);
}

/*!
 * @brief Wait for a mode change or a timeout.
 * @param msecs The number of milliseconds to wait.
 * @return 0 for new mode, or 1 for timeout.
 */
int fspWorker::wait_new_mode_or_timeout(uint32_t msecs)
{
    int    rc = 1;      // Assume a timeout.
    bool   done;

    // If the 'msecs' parameter is 0, then set our done flag. We still
    // want to go through the check once, just incase the mode did change.
    done = (msecs == 0);

    do
    {
        // Is there a requested mode change (i.e. 'requested_mode' is not -1)?
        if (requested_mode != -1)
        {
            // Return to the caller, indicating a new mode is pending.
            rc = 0;
            done = true;
        }
        else if (msecs > 0)
        {
            // Hang out for the specified amount of time.
            std::unique_lock<std::mutex> locker(mode_mutex);
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            using ms = std::chrono::milliseconds;

            if (mode_status.wait_until(locker, now + ms(msecs)) == std::cv_status::timeout)
            {
                // We timed out, so assume we don't have a mode change pending.
                done = true;
            }
        }
    } while (!done);

    return rc;
}

int fsp_control_update(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    fspWorker *fsp = fspWorker::factory();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FSP_FUNC, "%s: 0x%08x", __PRETTY_FUNCTION__, reg_value);

    // If not in the Review state, can download but not playback or single step.
    if ((CameraServer::get_camera_state() != GIGE_CAMERA_STATE_REVIEW) &&
        ((reg_value == GIGE_STREAM_RECORDED) ||
         (reg_value == GIGE_STREAM_SLAVE_ENABLE) ||
         (reg_value == GIGE_STREAM_SINGLESTEP)))
    {
        fi_log(LOG_ERR, "%s: wrong state (%d) for playback", __PRETTY_FUNCTION__, CameraServer::get_camera_state());
        return GEV_STATUS_WRONG_CONFIG;
    }

    if ((CameraServer::get_camera_state() == GIGE_CAMERA_STATE_LIVE_VIDEO) &&
        (reg_value == GIGE_STREAM_SINGLESTEP))
    {
        fi_log(LOG_DEBUG, "%s: Unexpected STREAM_SINGLESTEP in LIVE_VIDEO state", __PRETTY_FUNCTION__);
        return GEV_STATUS_SUCCESS;
    }

    // Is this a stream control setting?
    // Tell the streaming thread to start or stop streaming.
    // make sure the value passed in is valid
    if ((reg_value & GIGE_STREAM_CONTROL_MASK) != 0)
    {
        if (reg_value == GIGE_STREAM_PAUSE)
        {
            TP(FSP_DOWNLOAD, "%s: PAUSE", __PRETTY_FUNCTION__);
            LOG("%s: PAUSE", __PRETTY_FUNCTION__);
            fsp->pauseStream();
        }
        else if (reg_value == GIGE_STREAM_RESUME)
        {
            TP(FSP_DOWNLOAD, "%s: RESUME", __PRETTY_FUNCTION__);
            LOG("%s: RESUME", __PRETTY_FUNCTION__);
            fsp->resumeStream();
        }
        else
        {
            TP(FSP_DOWNLOAD, "%s: Unkown [0x%08x]", __PRETTY_FUNCTION__, reg_value);
            return GEV_STATUS_INVALID_PARAMETER;
        }

        // All done.
        regs[GIGE_REG_STREAM_CONTROL].set_value(reg_value);
        return GEV_STATUS_SUCCESS;
    }

    // Sanity check setting.
    if ((reg_value != GIGE_STREAM_LIVE) &&
        (reg_value != GIGE_STREAM_RECORDED) &&
        (reg_value != GIGE_STREAM_SINGLESTEP) &&
        (reg_value != GIGE_STREAM_SLAVE_ENABLE) &&
        (reg_value != GIGE_STREAM_DOWNLOAD) &&
        (reg_value != GIGE_STREAM_GET_FILE) &&
        (reg_value != GIGE_STREAM_RAW_METADATA) &&
        (reg_value != GIGE_STREAM_SESSION) &&
        (reg_value != GIGE_STREAM_STOP))
    {
        syslog(LOG_ERR, "%s: illegal option %u", __PRETTY_FUNCTION__, reg_value);
        return GEV_STATUS_ERROR;
    }

    // We need the IPM for any activity so make sure it is available to us,
    // e.g., not doing a save operation.
    if (ipm_is_busy() && (reg_value != GIGE_STREAM_STOP))
    {
        syslog(LOG_INFO, "%s: IPM is busy, got STREAM_CONTROL=0x%08x.",
               __PRETTY_FUNCTION__, reg_value);
        return GEV_STATUS_BUSY;
    }

    //=====================================================================
    // ALM_TODO_SHMEM:
    // Test to stream live (armed live actually).
    //=====================================================================
//#define ALM_TODO_SHMEM 1
#ifdef ALM_TODO_SHMEM
    if ((reg_value == GIGE_STREAM_LIVE) || (reg_value == GIGE_STREAM_STOP))
    {
        uint32_t shm_value = (reg_value == GIGE_STREAM_LIVE) ? GIGE_LOCALSTREAM_LIVE : GIGE_LOCALSTREAM_STOP;
        shmemStreamControlUpdate(nullptr, shm_value);
    }
#endif /* ALM_TODO_SHMEM */
    //=====================================================================
    //=====================================================================

    // Change the mode { STOP, DOWNLOAD, SINGLESTEP, PLAYBACK }
    return fsp->setMode((int)reg_value);
}

void fsp_disconnect()
{
    TP(FSP_FUNC, "%s: run_state=%d", __PRETTY_FUNCTION__, fspWorker::factory()->getRunState());
    fspWorker::factory()->disconnectStream();
}

void fsp_set_download_file_retry(uint32_t retry_index)
{
    TP(FSP_FUNC, "%s: retry_index=0x%08x", __PRETTY_FUNCTION__, retry_index);
    fspWorker::factory()->setRetryIndex(retry_index);
}

void fsp_exit()
{
    // If the worker thread is running, ask it to stop.
    if (fspWorker::isRunning())
        fspWorker::factory()->stop();
}

bool fsp_last_raw_image(uint8_t *image_data, int& image_len)
{
    bool rc(false);

    // The worker thread needs to be running.
    if (fspWorker::isRunning())
    {
        // We need to be live streaming.
        if (fspWorker::factory()->getMode() == GIGE_STREAM_LIVE)
        {
            // Ask for the next live frame.
            rc = fspWorker::factory()->getRawLiveFrame(image_data, image_len);
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Allow the "client" to asynchronously set the inter-frame
 *    delay in the GOE.
 *
 * @param p_msg     Not used.
 * @param frame_delay   The new frame delay, in milliseconds.
 */
int fsp_set_interframe_delay(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t frame_delay)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(FSP_FUNC, "%s: frame_delay=%u", __PRETTY_FUNCTION__, frame_delay);

    // Check to see if we have an override.
    override_u32(OVR_FSP_FRAME_DELAY, &frame_delay);

    // Save the new value.
    regs[GIGE_REG_STREAM_FRAME_DELAY].set_value(frame_delay);

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Allow the "client" to asynchronously set the inter-packet
 *    delay in the GOE.
 *
 * @param p_msg     Not used.
 * @param packet_delay  The new packet delay, in microseconds.
 */
int fsp_set_inter_packet_delay(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t packet_delay)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_TRACE, "%s: packet_delay=%u", __PRETTY_FUNCTION__, packet_delay);

    // Check to see if we have an override.
    override_u32(OVR_FSP_PKT_DELAY, &packet_delay);

    // Save the new value.
    regs[GIGE_REG_SCPD0].set_value(packet_delay);
    return GEV_STATUS_SUCCESS;
}

int fsp_set_packet_size(uint32_t reg_value)
{
    uint16_t    packet_size;        // our return value.
    bool    dnf;
    bool    fire_tp;
    fspWorker *fsp = fspWorker::factory();

    LOG("%s: reg_value=0x%08x", __PRETTY_FUNCTION__, reg_value);

    // Get the packet size out of the register, which after
    // byte swapping is in the low 16 bits of the register.
    packet_size = (uint16_t)(reg_value & 0xFFFF);

    // See if Fire Test Packet is set.
    fire_tp = ((reg_value & GIGE_SCPSX_FTP_BE) ? true : false);

    // See if DNF should be set.
    dnf = ((reg_value & GIGE_SCPSX_DNF_BE) ? true : false);

    // Set the information in our class.
    fsp->setPacketInfo(packet_size, dnf);

    // Are we sending a test packet?
    if (fire_tp)
    {
        int     istat;

        // Fire the test packet. It's OK if we get an error.
        // That just means the packet couldn't be sent...all part
        // of the test packet process.
        istat = fsp->fireTestPacket(packet_size);
        if (istat < 0)
        {
            fi_log(LOG_DEBUG, "%s: fireTestPacket failed!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ERROR;
        }
    }

    // All is well
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Store the specified 'file_path' into our class member 'stream_file'.
 *
 * @param file_path The new file path.
 * @param file_len  The length of the path.
 */
void fsp_set_save_path(const char *file_path, uint32_t file_len)
{
    TP(FSP_FUNC, "%s: file_path='%s'", __PRETTY_FUNCTION__, file_path);
    fspWorker::factory()->setStreamPath(file_path, file_len);
}

/*!
 * @brief Specify the port number to use when streaming.
 */
void fsp_set_stream_port_num(uint32_t reg_value)
{
    TP(FSP_FUNC, "%s(%u)", __PRETTY_FUNCTION__, reg_value);

    // Pass on the port information.
    fspWorker::factory()->setPort(reg_value);
}

/*!
 * @brief Get our fspWorker singleton up and running.
 */
void fsp_startup()
{
    TP(FSP_FUNC, "%s", __PRETTY_FUNCTION__);
    (void)fspWorker::factory();
}
