/*!
 * @file gige_reg.cxx
 * @brief This file implements the GIGE_REG_CACHE class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <ctype.h>

#include <sys/utsname.h>

#include <chrono>
#include <mutex>
#include <string>
#include <fmt/printf.h>

#include "battery.h"
#include "camera_server.h"
#include "gige_reg.h"
#include "fcp.h"
#include "fsp.h"
#include "lic_manager.h"
#include "part_mgr.h"
#include "pb_control.h"
#include "storage.h"
#include "telemetry.h"
#include "shmem_stream.h"

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

const int NUM_GIGE_REG      = 159;      //!< number of registers supported, mandatory

const char *MODEL_NAME_DEF  = "FiX";    //!< Default model name.
const char *BOOTLOADER_REV  = ".config/Fastec/bootloader_version";
const char *MD5_CHECK       = ".config/Fastec/scripts/md5check";
const char *MANUFACTURERS_INFO_DEF = "Fastec Imaging, San Diego, CA";
const char *MANUFACTURERS_NAME_DEF = "FASTEC";

const char *GIGE_REG_CACHE::REGISTER_FILE   = ".config/Fastec/settings/camera_server_regs";
const char *GIGE_REG_CACHE::REGISTER_CHECKSUM_FILE = ".config/Fastec/settings/camera_server_regs.md5";

//! The enviroment variable holding the serial number.
const char *SERIAL_NUMBER_ENV_VAR = "SERIAL_NUMBER";

//! The name of the historical serial number database.
const char *SERIAL_NUMBER_FILE = ".config/Fastec/serial_number.csv";

//! Registers and their default values listed. Strings need to be done
//! during initialization.
const GIGE_REG gige_def_regs[ NUM_GIGE_REG ] =
{
#include "gige_reg_init.h"

    { 0, REG_TYPE_BIN, REG_ACCESS_RO, 4, REG_SWAP_HTONL, {0x0}, nullptr}
};

const uint32_t  LOCK_TIMEOUT_MS     = 500;
const uint32_t  MAX_RETRIES     = 100;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

static void load_bootloader_ver(char *boot_rev, int max_len);
static void load_manufacturer_name(char *mfg_name, int max_len);
static void parse_reg_line(const char *line_ptr, GIGE_REG *p_reg);

/*!
 * @brief Read the contents of the BOOTLOADER_REV file
 *    and copy the value to the boot_rev parameter.
 *
 * @param boot_rev  Pointer to put the boot loader version.
 * @param max_len   Maximum number of characters.
 */
static void load_bootloader_ver(char *boot_rev, int max_len)
{
    std::size_t index;
    std::string rev_string;
    struct utsname kern_info;

    // Clear out the return value.
    memset(boot_rev, 0, max_len);

    // Query the kernel information.
    if (uname(&kern_info) < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return;
    }

    // Move the string over to where we can work on it.
    rev_string.assign(kern_info.release);

    // We only want the version part of the release string, which is seperated
    // by the hyphen character.
    index = rev_string.find_first_of("-");
    if (index != std::string::npos)
    {
        // Safety check the length.
        if (index > (std::size_t)(max_len - 1))
            index = (std::size_t)(max_len - 1);

        // Copy the data over.
        rev_string.copy(boot_rev, index, 0);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: No hyphen in rev_str: '%s'?", __PRETTY_FUNCTION__, kern_info.release);
    }
}

/*!
 * @brief Load the manufacturer's name, either with an OEM string
 *    or the default value.
 */
static void load_manufacturer_name(char *mfg_name, int max_len)
{
    const char  *exec_file = ".config/Fastec/scripts/get_oem_name";
    const char  *arg1 = "Manufacturer";
    FILE    *fp;
    char    temp_buf[ PATH_MAX ];

    // See if there's an OEM strings file with the data we want.
    snprintf(temp_buf, PATH_MAX, "%s/%s %s", get_home_path(), exec_file, arg1);
    fp = popen(temp_buf, "r");
    if (fp != NULL)
    {
        int i;

        // Reset our buffer.
        temp_buf[0] = 0;

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != NULL)
        {
            // Did we get anything?, we need the checksum so that it can be written
            if (temp_buf[0] != 0)
                break;
        }

        // Close the pipe.
        pclose(fp);

        // Make sure we strip off any trailing whitespace.
        for (i = strlen(temp_buf); i > 0; --i)
        {
            if (isspace(temp_buf[ i - 1 ]))
                temp_buf[ i - 1 ] = '\0';
        }

        // If we got something, use the OEM string, otherwise just
        // use our default.
        if (temp_buf[0] != '\0')
            strncpy(mfg_name, temp_buf, max_len);
        else
            strncpy(mfg_name, MANUFACTURERS_NAME_DEF, max_len);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot create register checksum.", __FUNCTION__);
    }
}

/*!
 * @brief Parse a line from the registration file and initialize
 *    a GIGE_REG structure from it.
 *
 * @param line_ptr  Our string from the registration file.
 * @param p_reg     Pointer to the GIGE_REG structure we want initialized.
 */
static void parse_reg_line(const char *line_ptr, GIGE_REG *p_reg)
{
    const char  *pcomma;

    // Clear out the structure.
    memset(p_reg, 0, sizeof(GIGE_REG));

    // Parse out the line read, note that this is a two stage process because
    // sscanf will return partial strings if they have a blank in them. So first
    // get the numeric values, then get the register value manually.
    sscanf(line_ptr, "%u,%hhu\n", &p_reg->reg_num, &p_reg->reg_type);
    pcomma = strrchr(line_ptr, ',');
    if (pcomma)
    {
        // Get a pointer to make things easier.
        char *str_ptr = (char *)p_reg->reg_val_b;

        // Copy the "value" into our string buffer. Note that there may
        // not BE any "value".
        strncpy(str_ptr, pcomma + 1, GIGE_REG_STR_MAX_SIZE);
        if (str_ptr[ strlen(str_ptr) - 1] == '\n')
            str_ptr[ strlen(str_ptr) - 1] = '\0';
        if (p_reg->reg_type != REG_TYPE_STR)
        {
            uint32_t reg_val = 0;

            sscanf(str_ptr, "%u", &reg_val);
            p_reg->reg_val_l = reg_val;
        }
    }
}

//--------------------------------------------------------------------
// Class Implementation.
//--------------------------------------------------------------------

GIGE_REG_CACHE::GIGE_REG_CACHE() :
    last_index(0)
{
}

GIGE_REG_CACHE::~GIGE_REG_CACHE()
{
    // Make sure we empty our cache.
    for (auto it = gige_regs.begin(); it != gige_regs.end();)
    {
        // Clear out this entry.
        it = gige_regs.erase(it);
    }
    gige_regs.clear();
}

int GIGE_REG_CACHE::init_cache()
{
    // Get a pointer to our support classes.
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    PartMgr *p_partmgr = CameraServer::get_partmgr();
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint16_t max_width, max_height;
    uint8_t mac_addr[CS_NETWORK::MAC_ADDR_LEN];
    const LicManager *p_lic = CameraServer::get_license();

    // Reset back to defaults.
    reset();

    // All the dynamic and string registers need to be
    // set now since they couldn't be done as part of the
    // static initialization
    gige_regs[ GIGE_REG_GIGE_VER ].reg_val_w[0] = GIGE_VER_MINOR;
    gige_regs[ GIGE_REG_GIGE_VER ].reg_val_w[1] = GIGE_VER_MAJOR;

    // Get our MAC address and pass it on to the FCP.
    memset(mac_addr, 0, CS_NETWORK::MAC_ADDR_LEN);
    CsNetwork::getEth0MacAddress(mac_addr);
    set_mac_address(mac_addr);

    // Find out if DHCP is running so we know how to setup the initial register
    if (CsNetwork::getPreferredDhcpMode(CsNetwork::ETHER))
    {
        gige_regs[ GIGE_REG_IP_CUR_CFG ].reg_val_l |= GIGE_IFCONFIG_DHCP_BE;
        gige_regs[ GIGE_REG_IP_CUR_CFG ].reg_val_l &= (~GIGE_IFCONFIG_PERSISTENT_BE);
    }
    else
    {
        // Since we are not running DHCP, fill in our persistent values.
        // If we are going to use the prefs file to decide if DHCP is set, we also
        // need to use it to get our address information.
        gige_regs[ GIGE_REG_PERSIST_IP ].reg_val_l = CsNetwork::getPreferredIpAddress(CsNetwork::ETHER);
        gige_regs[ GIGE_REG_PERSIST_SUBNET ].reg_val_l = CsNetwork::getPreferredSubnetMask(CsNetwork::ETHER);
        gige_regs[ GIGE_REG_PERSIST_GATEWAY ].reg_val_l = CsNetwork::getPreferredDefGateway(CsNetwork::ETHER);

        gige_regs[ GIGE_REG_IP_CUR_CFG ].reg_val_l &= (~((uint32_t)GIGE_IFCONFIG_DHCP_BE));
        gige_regs[ GIGE_REG_IP_CUR_CFG ].reg_val_l |= GIGE_IFCONFIG_PERSISTENT_BE;
    }

    // Find out if WIFI DHCP is selected so we know how to setup the initial register
    if (CsNetwork::getPreferredDhcpMode(CsNetwork::WIFI))
    {
        gige_regs[ GIGE_REG_WIFI_IP_CUR_CFG ].reg_val_l |= GIGE_IFCONFIG_DHCP_BE;
        gige_regs[ GIGE_REG_WIFI_IP_CUR_CFG ].reg_val_l &= (~GIGE_IFCONFIG_PERSISTENT_BE);
    }
    else
    {
        // Since we are not runing dhcp fill in our persistent values.
        // If we are going to use the prefs file to decide if DHCP is set, we also
        // need to use it to get our address information.
        gige_regs[ GIGE_REG_WIFI_IP_ADDR ].reg_val_l = CsNetwork::getPreferredIpAddress(CsNetwork::WIFI);
        gige_regs[ GIGE_REG_WIFI_SUBNET_MASK ].reg_val_l = CsNetwork::getPreferredSubnetMask(CsNetwork::WIFI);
        gige_regs[ GIGE_REG_WIFI_DEF_GATEWAY ].reg_val_l = CsNetwork::getPreferredDefGateway(CsNetwork::WIFI);

        gige_regs[ GIGE_REG_WIFI_IP_CUR_CFG ].reg_val_l &= (~((uint32_t)GIGE_IFCONFIG_DHCP_BE));
        gige_regs[ GIGE_REG_WIFI_IP_CUR_CFG ].reg_val_l |= GIGE_IFCONFIG_PERSISTENT_BE;
    }

    // The manufacturer's name is a maximum of 32 characters.
    load_manufacturer_name((char *)gige_regs[ GIGE_REG_MAN_NAME ].reg_val_b, 32);

    // The manufacturer's name is a maximum of 32 characters.
    load_bootloader_ver((char *)gige_regs[ GIGE_REG_BOOTLOADER_VER ].reg_val_b, 32);

    // get the model name from the license manager
    if (p_lic && (strlen((const char *)p_lic->model()) > 0))
        strncpy((char *)gige_regs[ GIGE_REG_MODEL_NAME ].reg_val_b, (const char *)p_lic->model(), GIGE_REG_MODEL_NAME_B_LEN);
    else
        strncpy((char *)gige_regs[ GIGE_REG_MODEL_NAME ].reg_val_b, MODEL_NAME_DEF, GIGE_REG_MODEL_NAME_B_LEN);

    // See if we have an SSD bandwidth value in the license file. If so, use that as the default.
    if (p_lic && (p_lic->ssdBandwidth() != 0))
        gige_regs[ GIGE_REG_SSD_BANDWIDTH ].reg_val_l = (p_lic->ssdBandwidth() * 1000 * 1000);

    // This is from a command line switch so we will allow a crazy override for test purposes.
    if (user_ssd_bandwidth != 0)
    {
        fi_log(LOG_INFO, "%s: Override SSD Bandwidth, new value=%u", __FUNCTION__,
               user_ssd_bandwidth);
        gige_regs[ GIGE_REG_SSD_BANDWIDTH ].reg_val_l = user_ssd_bandwidth;
    }

    // We won't find out whether we are in LTR mode or not until the settings are loaded.
    // If it turns out there is no valid settings file, we will write one out with these
    // initial defaults and those will be the settings that remain in effect.
    // That means we need to use a valid default LTR mode setting here.
    // We default to non-LTR mode if that's available at all and only default to
    // LTR mode if it's the only mode licensed.
    if (p_lic && (p_lic->ssdMax() > 0) && (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_ONLY))
    {
        gige_regs[ GIGE_REG_LTR_MODE ].reg_val_l = GIGE_LTR_MODE_ENABLE;
        fpga_ptr->ltrModeSelect(true);   // This sets a flag that we hopefully can get rid of.
    }
    else
    {
        // Mode is NONE or DUAL, or else there's no license info.
        gige_regs[ GIGE_REG_LTR_MODE ].reg_val_l &= ~GIGE_LTR_MODE_ENABLE;
        fpga_ptr->ltrModeSelect(false);
    }

    // Device version which is an ascii string is generated from the firmware version
    // which is an integer, the bytes are major, minor, path
    gige_regs[ GIGE_REG_FIRMWARE_VER ].reg_val_l = CURRENT_FIRMWARE_VERSION;
    if (CURRENT_FIRMWARE_VERSION != 0)
    {
        sprintf((char *)gige_regs[ GIGE_REG_DEVICE_VER ].reg_val_b, "%hhu.%hhu.%hhu",
                (uint8_t)(gige_regs[ GIGE_REG_FIRMWARE_VER ].reg_val_b[2]),
                (uint8_t)(gige_regs[ GIGE_REG_FIRMWARE_VER ].reg_val_b[1]),
                (uint8_t)(gige_regs[ GIGE_REG_FIRMWARE_VER ].reg_val_b[0]));
    }
    else
    {
        fi_log(LOG_NOTICE, "%s: CURRENT_FIRMWARE_VERSION is 0?!", __PRETTY_FUNCTION__);
    }

    // Manufacturers info
    strncpy((char *)gige_regs[ GIGE_REG_MAN_INFO ].reg_val_b, MANUFACTURERS_INFO_DEF, GIGE_REG_MAN_INFO_B_LEN);  // FORNOW - Initial default value

    // Serial number
    setup_serial_number(mac_addr);

    max_width  = cs_ptr->get_lic_width();
    if (max_width > sensor_ptr->get_sensor_max_width())
        max_width = sensor_ptr->get_sensor_max_width();
    max_height = cs_ptr->get_lic_height();
    if (max_height > sensor_ptr->get_sensor_max_height())
        max_height = sensor_ptr->get_sensor_max_height();

    // Is the FPN register uninitialized?
    if (gige_regs[GIGE_REG_FPN].reg_val_l == 0)
    {
        // Lince sensors default to Sensor FPN
        if ((sensor_ptr->get_sensor_type() == FI_SENSOR_LINCE5M_MONO) || (sensor_ptr->get_sensor_type() == FI_SENSOR_LINCE5M_COLOR))
        {
            // We have a Lince sensor, so enable the sensor FPN by default. Note that this
            // will be overwritten when the saved settings are loaded.
            gige_regs[GIGE_REG_FPN].reg_val_l = GIGE_FPN_SENSOR_ENABLE_MASK;
        }
        else
        {
            // The other sensors default to using the FPGA (aka "Pixel") FPN.
            gige_regs[GIGE_REG_FPN].reg_val_l = GIGE_FPN_FPGA_ENABLE_MASK;
        }
    }

    // If we don't support Pixel FPN for this sensor, don't default to it.
    if ((gige_regs[GIGE_REG_FPN].reg_val_l == GIGE_FPN_FPGA_ENABLE_MASK) && (!sensor_ptr->supports_pixel_fpn()))
    {
        gige_regs[GIGE_REG_FPN].reg_val_l = GIGE_FPN_SENSOR_ENABLE_MASK;
    }

    // Initialize the sensor-specific and license-restricted registers.
    gige_regs[ GIGE_REG_SENSOR_TYPE ].reg_val_l = sensor_ptr->get_sensor_type();
    gige_regs[ GIGE_REG_SENSOR_WIDTH ].reg_val_l = max_width;
    gige_regs[ GIGE_REG_SENSOR_HEIGHT ].reg_val_l = max_height;
    gige_regs[ GIGE_REG_ABS_SENSOR_WIDTH ].reg_val_l = max_width;
    gige_regs[ GIGE_REG_ABS_SENSOR_HEIGHT ].reg_val_l = max_height;
    gige_regs[ GIGE_REG_MIN_WIDTH ].reg_val_l = sensor_ptr->get_sensor_min_width();
    gige_regs[ GIGE_REG_MIN_HEIGHT ].reg_val_l = sensor_ptr->get_sensor_min_height();

    // Get the default settings for this sensor.
    gige_regs[ GIGE_REG_WIDTH ].reg_val_l = max_width;
    gige_regs[ GIGE_REG_HEIGHT ].reg_val_l = max_height;
    gige_regs[ GIGE_REG_BIT_MODE ].reg_val_l = sensor_ptr->get_default_bit_mode();
    gige_regs[ GIGE_REG_FRAME_RATE ].reg_val_l = sensor_ptr->get_default_frame_rate();
    gige_regs[ GIGE_REG_SHUTTER_SPEED ].reg_val_l =
        cs_ptr->get_max_shutter_speed(sensor_ptr->get_default_frame_rate());

    // Get the maximum licensed frame rate.
    gige_regs[ GIGE_REG_MAX_FRAME_RATE ].reg_val_l = cs_ptr->get_max_licensed_frame_rate();

    // Initialize the FPGA-specific registers.

    // JUL2014 The defaults for these were set in the FiFPGA constructor.
    // JUL2014 Don't know why we can't just have the defaults in the table.
    gige_regs[ GIGE_REG_RED ].reg_val_l = fpga_ptr->ipmGainRed();
    gige_regs[ GIGE_REG_GREEN ].reg_val_l = fpga_ptr->ipmGainGreen();
    gige_regs[ GIGE_REG_BLUE ].reg_val_l = fpga_ptr->ipmGainBlue();

    gige_regs[ GIGE_REG_MIN_SHUTTER_SPEED ].reg_val_l = cs_ptr->get_min_shutter_speed();
    gige_regs[ GIGE_REG_LOW_LIGHT_SHUTTER ].reg_val_l =
        cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE);

    // The registers having to do with the partition manager. The initial default values
    // are derived from the amount of memory licensed.
    gige_regs[ GIGE_REG_MAX_PART ].reg_val_l = p_partmgr->getMaxPartitions();
    gige_regs[ GIGE_REG_NUM_PART ].reg_val_l = p_partmgr->getNumPartitions();
    gige_regs[ GIGE_REG_SESSION_LENGTH ].reg_val_l = p_partmgr->getSessionLength();
    gige_regs[ GIGE_REG_PART_SIZE ].reg_val_l = convert_to_QWORD(p_partmgr->getPartitionSize());
    gige_regs[ GIGE_REG_PART_SIZE_RESERVED ].reg_val_l = convert_to_QWORD(p_partmgr->getPartitionSizeReserved());
    gige_regs[ GIGE_REG_MEM_SIZE ].reg_val_l = p_partmgr->get_im_size_qw();
    gige_regs[ GIGE_REG_MEM_SIZE_RESERVED ].reg_val_l = p_partmgr->get_im_size_reserved_qw();
    gige_regs[ GIGE_REG_FRAME_SIZE ].reg_val_l = p_partmgr->getFrameSize_QW();
    gige_regs[ GIGE_REG_NUM_FRAMES ].reg_val_l = p_partmgr->getFrameCount();

    // Get the capabilities of the SSD.
    gige_regs[ GIGE_REG_SSD_CAPABILITIES ].reg_val_l = storage::get_ssd_capabilities();

    // Save the ICM version.
    fpga_ptr->regRead(ICM_MODULE_REVISION, &gige_regs[GIGE_REG_ICM_MODULE_VER].reg_val_l);

    return 0;
}

/*!
 * @brief Restores the registers from flash after a reboot are resume.
 *
 * Note: md5check is a shell script which is checked into the same directory as
 *   camera server, however this needs to reside in the same directory as
 *   md5sum otherwise popen returns a file pointer but no data, not even
 *   an error. The only thing I can think of is because the CameraServer is a
 *   daemon and not running under a regular user account.
 */
int GIGE_REG_CACHE::load()
{
    // Get a pointer to our support classes.
    FILE    *reg_file = NULL;
    char    temp_buf[PATH_MAX];
    std::string file_path;

    // Validate the register settings file.
    validate_reg_file();

    // By the time we get here, everything is going to be OK.
    file_path = fmt::sprintf("%s/%s", get_home_path(), REGISTER_FILE);
    reg_file = fopen(file_path.c_str(), "r");
    if (!reg_file)
    {
        fi_log(LOG_WARNING, "(%s): cannot open file %s to restore registers to errno:%d(%s) registers will not be restored, using default value",
               __FUNCTION__, temp_buf, errno, strerror(errno));
        return -1;
    }

    // Now read all the registers
    temp_buf[0] = 0;
    while (fgets(temp_buf, MAX_BUFFER_LEN, reg_file) != NULL)
    {
        bool        do_continue;
        GIGE_REG    my_reg;

        // LOG("(%s): fgets:%s", __FUNCTION__, temp_buf);

        // parse out the line read, note that this is a two stage process because
        // sscanf will return partial strings if they have a blank in them. So first get the
        // numeric values, then get the register value manually
        parse_reg_line(temp_buf, &my_reg);

        // Skip registers we don't want restored
        switch (my_reg.reg_num)
        {
        case GIGE_REG_MAX_FRAME_RATE:
        case GIGE_REG_MAX_SHUTTER_SPEED:
        case GIGE_REG_MIN_SHUTTER_SPEED:
        case GIGE_REG_MAX_PART:
        case GIGE_REG_ACTIVE_PART:
        case GIGE_REG_PART_SIZE_RESERVED:
        case GIGE_REG_PART_SIZE:
        case GIGE_REG_MEM_SIZE:
        case GIGE_REG_MEM_SIZE_RESERVED:
        case GIGE_REG_FRAME_SIZE:
        case GIGE_REG_NUM_FRAMES:
        case GIGE_REG_FIRMWARE_VER:
        case GIGE_REG_BOOTLOADER_VER:
        case GIGE_REG_FPGA_VER:
        case GIGE_REG_BATTERY_CHARGE:
        case GIGE_REG_MIN_WIDTH:
        case GIGE_REG_MIN_HEIGHT:
        case GIGE_REG_SENSOR_TYPE:
        case GIGE_REG_AVAIL_MEDIA:
        case GIGE_REG_TEMPERATURE:
        case GIGE_REG_MAN_NAME:
        case GIGE_REG_MODEL_NAME:
        case GIGE_REG_MAN_INFO:
        case GIGE_REG_SERIAL_NUM:
        case GIGE_REG_XML_URL_1:
        case GIGE_REG_XML_URL_2:
        case GIGE_REG_CCP:
        case GIGE_REG_MCP:
        case GIGE_REG_MCDA:
        case GIGE_REG_SCP0:
        case GIGE_REG_SCPS0:
        case GIGE_REG_SCDA0:
        case GIGE_REG_HEARTBEAT:
        case GIGE_REG_WIFI_START_SCAN:
        case GIGE_REG_SENSOR_VER:
        case GIGE_REG_SSD_BANDWIDTH:
        case GIGE_REG_ABS_SENSOR_WIDTH:
        case GIGE_REG_ABS_SENSOR_HEIGHT:
        case GIGE_REG_ICM_MODULE_VER:
            do_continue = true;
            break;
        default:
            do_continue = false;
            break;
        }

        // Are we supposed to continue?
        if (do_continue)
            continue;

        // Does this register exist?
        if (gige_regs.count(my_reg.reg_num) == 0)
        {
            fi_log(LOG_WARNING, "%s: Register 0x%x not found", __FUNCTION__, my_reg.reg_num);

            // Back around the loop.
            continue;
        }

        // Sanity check loaded values.
        switch (my_reg.reg_num)
        {
        case GIGE_REG_JPEG_QUALITY:
            if ((my_reg.reg_val_l < 1) || (my_reg.reg_val_l > 100))
            {
                LOG("%s: GIGE_REG_JPEG_QUALITY is out of range [%d]!", __PRETTY_FUNCTION__, my_reg.reg_val_l);
                my_reg.reg_val_l = 80;
            }
            break;
        default:
            break;
        }

        // Normally we don't want to update a read only register...but there
        // are exceptions.
        GIGE_REG& reg = gige_regs[ my_reg.reg_num ];
        if (reg.reg_access == REG_ACCESS_RO)
        {
            // Is this an exception?
            switch (my_reg.reg_num)
            {
            case GIGE_REG_SENSOR_RDOUT_DELAY:
                LOG("%s: GIGE_REG_SENSOR_RDOUT_DELAY=0x%08x", __PRETTY_FUNCTION__, my_reg.reg_val_l);
                reg.reg_val_l = my_reg.reg_val_l;
                break;
            default:
                break;
            }
        }
        else
        {
            if (reg.reg_type == REG_TYPE_STR)
            {
                memcpy(reg.reg_val_b, my_reg.reg_val_b, GIGE_REG_STR_MAX_SIZE);
            }
            else
            {
                // Update the register.
                reg.reg_val_l = my_reg.reg_val_l;
            }
        }
    }
    fclose(reg_file);

    return 0;
}

/*!
 * @brief Initialize hardware to be consistent with loaded settings.
 *
 * Update the hardware functions using what is now loaded in the GigE registers.
 * They might be default values or they might be what we loaded from the settings file.
 *
 * JUL2014 These actions used to be part of init_cache() and load().
 * JUL2014 Instead, we now load everything, validate it, then use it.
 * JUL2014 Not every update function is legal to call here.
 *
 * @return false in case something went wrong, which is inconceivable!
 */
bool GIGE_REG_CACHE::set_loaded_registers()
{
    // Get a pointer to our support classes.
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    uint32_t    reg_value;
    uint32_t    wb_value = 0;   // the white balance register, interested if a preset

    reflect_register(GIGE_REG_SCPD0);
    reflect_register(GIGE_REG_STREAM_FRAME_DELAY);

    // If the white balance is a preset, we have fixed values for the color correction matrix
    // and the white balance color gains. Otherwise, the value can be custom.
    wb_value = regs[ GIGE_REG_WHITE_BAL ].get_value();
    if (wb_value < GIGE_WHITE_BAL_NORMAL)
        fcp_set_wb_preset(wb_value);
    else
    {
        reflect_register(GIGE_REG_RED);
        reflect_register(GIGE_REG_GREEN);
        reflect_register(GIGE_REG_BLUE);
        reflect_register(GIGE_REG_MATRIX_COEFF1_RED);
        reflect_register(GIGE_REG_MATRIX_COEFF2_RED);
        reflect_register(GIGE_REG_MATRIX_COEFF3_RED);
        reflect_register(GIGE_REG_MATRIX_COEFF1_GRN);
        reflect_register(GIGE_REG_MATRIX_COEFF2_GRN);
        reflect_register(GIGE_REG_MATRIX_COEFF3_GRN);
        reflect_register(GIGE_REG_MATRIX_COEFF1_BLU);
        reflect_register(GIGE_REG_MATRIX_COEFF2_BLU);
        reflect_register(GIGE_REG_MATRIX_COEFF3_BLU);
        reflect_register(GIGE_REG_WHITE_BALANCE_RED);
        reflect_register(GIGE_REG_WHITE_BALANCE_GRN);
        reflect_register(GIGE_REG_WHITE_BALANCE_BLU);
    }

    // Setting brightness, contrast, and gamma involves building a LUT, so we'll only do it once.
    // We have to handle GAMMA a little differently. Its values are different from what's in the
    // register and what's sent via clients. Once it is set, we will rebuild the cached value.
    // We go ahead and do what we always did to build the LUT because that sets some data in
    // the FPGA class where it keeps it instead of using the regs cache (FIXME it shouldn't).
    reg_value = regs[ GIGE_REG_GAMMA ].get_value();
    fpga_ptr->ipmSetLutValues(regs[ GIGE_REG_BRIGHTNESS ].get_value(),
                              regs[ GIGE_REG_CONTRAST ].get_value(),
                              reg_value & 0xFFFF, (reg_value >> 16) & 0xFFFF);
    reg_value = (fpga_ptr->ipmDefaultGamma() << 16) | fpga_ptr->ipmGamma();
    regs[ GIGE_REG_GAMMA ].set_value(reg_value);

    // If we have a custom LUT, load it.
    if (regs[ GIGE_REG_LUT ].get_value() != 0)
    {
        reg_value = regs[ GIGE_REG_LUT ].get_value();

        // Don't leave the DEFER bit lying around across a config reload.
        if ((reg_value & GIGE_LUT_DEFER) != 0)
        {
            reg_value &= ~GIGE_LUT_DEFER;
            regs[ GIGE_REG_LUT ].set_value(reg_value);
        }
        if (!fpga_ptr->ipmLoadUserLut(reg_value))
        {
            // We were unable to load the LUT. Indicate that
            // no LUT is selected.
            fi_log(LOG_WARNING, "%s: Unable to load LUT (%d).", __FUNCTION__, reg_value);
            regs[ GIGE_REG_LUT ].set_value(GIGE_LUT_NONE);
        }
    }

    // Update our maximum shutter speed based on the restored and validated frame rate.
    reg_value = regs[ GIGE_REG_FRAME_RATE ].get_value();
    reg_value = cs_ptr->get_max_shutter_speed(cs_ptr->reg_to_frame_rate(reg_value));
    regs[ GIGE_REG_MAX_SHUTTER_SPEED ].set_value(reg_value);

    reflect_register(GIGE_REG_PREFERENCES);
    reflect_register(GIGE_REG_SYNC_DELAY);
    reflect_register(GIGE_REG_SYNC_IN_ENABLE);
    reflect_register(GIGE_REG_SYNC_OUT_SELECT);
    reflect_register(GIGE_REG_SYNC_IN_POLARITY);
    reflect_register(GIGE_REG_EXT_TRIG_POLARITY);
    reflect_register(GIGE_REG_EXT_TRIG_ENABLE);
    reflect_register(GIGE_REG_ARM_OUT_SELECT);
    reflect_register(GIGE_REG_TRIG_OUT_SELECT);

    reflect_register(GIGE_REG_ARM_IN_POLARITY);
    // reflect_register( GIGE_REG_ARM_IN_ENABLE ); // Wait until we can accept an ARM transition.

    reflect_register(GIGE_REG_DS_SHUTTER_SPEED);
    reflect_register(GIGE_REG_TS_SHUTTER_SPEED);

    // Set the things which rely on the ROI already having been set.
    reflect_register(GIGE_REG_HISTO_RGB_ROI_X);
    reflect_register(GIGE_REG_HISTO_RGB_ROI_Y);

    // This has the side effect of setting GIGE_REG_LOW_LIGHT_SHUTTER value as well.
    reflect_register(GIGE_REG_LOW_LIGHT_ENABLE);

    // We could have updated the trigger requested position due to changes when
    // we validated other registers.
    reflect_register(GIGE_REG_TRIGGER_REQ_POS);

    return true;
}

/*!
 * @brief Clear the cache.
 */
void GIGE_REG_CACHE::reset()
{
    int     i;

    // Do we need to release an old version of the cache?
    for (auto it = gige_regs.begin(); it != gige_regs.end();)
    {
        // Clear out this entry.
        it = gige_regs.erase(it);
    }
    gige_regs.clear();

    // It's easy to neglect to update the #define for NUM_GIGE_REG so warn if it's wrong.
    // If it's too big, the penultimate entry will be zero but should not be.
    // If it's too small, we get a compile error.
    if ((gige_def_regs[ NUM_GIGE_REG - 1 ].reg_num != 0) || (gige_def_regs[ NUM_GIGE_REG - 2 ].reg_num == 0))
    {
        fi_log(LOG_WARNING, "%s: NUM_GIGE_REG=%u is wrong 0x%04x!=0 | 0x%04x==0", __FUNCTION__,
               NUM_GIGE_REG, gige_def_regs[ NUM_GIGE_REG - 1 ].reg_num,
               gige_def_regs[ NUM_GIGE_REG - 2 ].reg_num);
    }

    // Populate our cache with the defaults.
    for (i = 0; i < NUM_GIGE_REG; ++i)
    {
        uint32_t reg_num;
        reg_num = gige_def_regs[i].reg_num;

        if (reg_num != 0)
        {
            try
            {
                gige_regs[reg_num] = gige_def_regs[i];
            }
            catch (...)
            {
                fi_log(LOG_DEBUG, "%s: Unable to initialize register 0x%x! =========================", __PRETTY_FUNCTION__, reg_num);
            }
        }
    }
}

/*!
 * @brief Saves the registers to flash so that they can be restored later after a reboot.
 *    The calling routine passes which array to save, this lets us have the option
 *    of saving the default registers.
 */
void GIGE_REG_CACHE::save()
{
    char temp_path[PATH_MAX];
    char md5_path[PATH_MAX];
    FILE *reg_file;
    GIGE_MAP::iterator  reg_it;

    LOG("%s", __PRETTY_FUNCTION__);

    // Open the register save file.
    snprintf(temp_path, PATH_MAX, "%s/%s", get_home_path(), REGISTER_FILE);
    reg_file = fopen(temp_path, "w");
    if (!reg_file)
    {
        fi_log(LOG_WARNING, "(%s): cannot open file %s to save registers to errno:%d(%s) registers will not be saved",
               __PRETTY_FUNCTION__, temp_path, errno, strerror(errno));
        return;
    }

    // Now dump all of the registers.
    for (reg_it = gige_regs.begin(); reg_it != gige_regs.end(); ++reg_it)
    {
        GIGE_REG *reg_ptr = &reg_it->second;

        if (reg_ptr->reg_num == 0)
            continue; // skip unused registers

        // Change format depending upon what we are writing out.
        if (reg_ptr->reg_type == REG_TYPE_STR)
            fprintf(reg_file, "%u,%u,%s\n", reg_ptr->reg_num, reg_ptr->reg_type, reg_ptr->reg_val_b);
        else
            fprintf(reg_file, "%u,%u,%u\n", reg_ptr->reg_num, reg_ptr->reg_type, reg_ptr->reg_val_l);
    }

    // Close the file.
    fclose(reg_file);

    // The file is finished writing, create the checksum file, which we will use to
    // determine validity on a restore of the registers.
    snprintf(md5_path, PATH_MAX, "%s/%s", get_home_path(), REGISTER_CHECKSUM_FILE);
    storage::md5sum_file(temp_path, md5_path);
}

/*!
 * @brief Validate a serial number, logging details of the problem to help troubleshoot.
 *
 * @param buf   Contains the null-terminated serial number string.
 * @return  true if the serial number is okay, false if not
 */
bool GIGE_REG_CACHE::validate_serial_number(const char *buf)
{
    bool sn_okay = true;

    // Make sure the characters we are checking are really there.
    if (strlen(buf) < 10)
    {
        fi_log(LOG_DEBUG, "%s: bad length (%lu) for '%s'", __FUNCTION__, strlen(buf), buf);
        sn_okay = false;
    }

    // Starts with TS or IL or CC.
    if (sn_okay)
    {
        if (!(((buf[0] == 'T') && (buf[1] == 'S')) ||
              ((buf[0] == 'I') && (buf[1] == 'L')) ||
              ((buf[0] == 'C') && (buf[1] == 'C'))))
        {
            fi_log(LOG_DEBUG, "%s: bad housing '%c%c'", __FUNCTION__, buf[0], buf[1]);
            sn_okay = false;
        }
    }

    // Next 8 characters must be digits.
    if (sn_okay)
    {
        if (!(isdigit(buf[2]) && isdigit(buf[3]) &&         // YY
              isdigit(buf[4]) && isdigit(buf[5]) &&         // MM
              isdigit(buf[6]) && isdigit(buf[7]) &&         // DD
              isdigit(buf[8]) && isdigit(buf[9])))          // NN
        {
            fi_log(LOG_DEBUG, "%s: not digits '%s'", __FUNCTION__, buf + 2);
            sn_okay = false;
        }
    }

    // Validate the range of the numbers.
    if (sn_okay)
    {
        int year;
        int month;
        int day;
        int seqno;

        year  = (buf[2] - '0') * 10 + (buf[3] - '0');
        month = (buf[4] - '0') * 10 + (buf[5] - '0');
        day   = (buf[6] - '0') * 10 + (buf[7] - '0');
        seqno = (buf[8] - '0') * 10 + (buf[9] - '0');

        if (!((year >= 11) &&
              ((month >= 1) && (month <= 12)) &&
              ((day >= 1) && (day <= 31)) &&
              ((seqno >= 1) && (seqno <= 99))))
        {
            fi_log(LOG_DEBUG, "%s: bad digits (y=%d m=%d d=%d n=%d)", __FUNCTION__,
                   year, month, day, seqno);
            sn_okay = false;
        }
    }

    // Leave a hook for something like '-xxx'.
    if ((strlen(buf) > 10) && (buf[10] != '-'))
    {
        fi_log(LOG_DEBUG, "%s: extra chars in '%s'", __FUNCTION__, buf);
        sn_okay = false;
    }

    return sn_okay;
}

/*!
 * @brief Set up the serial number GigE register
 *
 * Tries the environment variable and uses it if valid. Otherwise checks the "old unit"
 * database. Otherwise constructs something using the MAC address.
 */
void GIGE_REG_CACHE::setup_serial_number(uint8_t *mac_addr)
{
    const char *env_ptr;
    std::string buf;
    bool sn_okay = false;
    const LicManager *p_lic;

    // Get our license pointer.
    p_lic = CameraServer::get_license();

    // Try to get it from the environment variable first. This allows
    // for an override of the serial number.
    env_ptr = getenv(SERIAL_NUMBER_ENV_VAR);
    if ((env_ptr != NULL) && (env_ptr[0] != '\0'))
    {
        buf.assign(env_ptr);

        // Validate what we got. If it is fishy, we will use a default string.
        if (validate_serial_number(buf.c_str()))
            sn_okay = true;
    }

    // There had better be a serial number in the license.
    if (!sn_okay)
    {
        // Make sure we have the license manager and it
        // has the serial number.
        if (p_lic && (strlen(p_lic->serial_no()) > 0))
        {
            // Copy the serial number.
            buf.assign(p_lic->serial_no());

            // Make sure it'll fit.
            if (buf.size() > GIGE_REG_SERIAL_NUM_B_LEN - 4)
                buf.resize(GIGE_REG_SERIAL_NUM_B_LEN - 4);
            sn_okay = true;
        }
    }

    // We need to have something, so this will be our last resort.
    if (!sn_okay)
    {
        // First, try to use the model number if it is present.

        // get the model name from the license manager
        if (p_lic && (strlen(p_lic->model()) > 0))
        {
            buf[0] = p_lic->model()[0];
            buf[1] = p_lic->model()[1];
            buf[2] = p_lic->model()[2];
            buf[3] = '\0';
        }
        else
            buf.assign("FAS");

        buf.append(fmt::sprintf("100%02X%02X", mac_addr[4], mac_addr[5]));
    }

    // FasMotion and the Web GUI expect the serial number to end with ".An" to show what the
    // main board version is so they know the capabilities of the camera. We need to append it
    // to the serial number but we also need to make sure we don't overrun the buffer.
    {
        std::string rev_str = CameraServer::board_rev();

        // See if there is enough room for dot + boardrev (".A3"). If not, truncate.
        // We'll take no more than ( max length - 4 ) chars for the serial number, leaving
        // room for ".RR" and a null.
        if (rev_str.size() > 2)
            rev_str.resize(2);

        buf.append(".");
        buf.append(rev_str);
    }

    strncpy((char *)gige_regs[ GIGE_REG_SERIAL_NUM ].reg_val_b, buf.c_str(), GIGE_REG_SERIAL_NUM_B_LEN);
}

/*!
 * @brief Save the MAC address into the register cache.
 *
 * @param mac_addr The new MAC address.
 */
void GIGE_REG_CACHE::set_mac_address(uint8_t *mac_addr)
{
    // Try to lock access.
    if (!cache_lock.try_lock_for(std::chrono::milliseconds(LOCK_TIMEOUT_MS)))
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s(%p): =============== Couldn't LOCK! ==============",
               __PRETTY_FUNCTION__, (void *)caller);
    }

    memcpy(&(gige_regs[ GIGE_REG_MAC_HIGH ].reg_val_w[0]), mac_addr, 2);
    memcpy(&(gige_regs[ GIGE_REG_MAC_LOW ].reg_val_l), &mac_addr[2], 4);
    cache_lock.unlock();
}

/*!
 * @brief Return true if new_id matches to an entry in the register
 *    cache, false otherwise.
 */
bool GIGE_REG_CACHE::validate_id(uint32_t new_id)
{
    return (gige_regs.count(new_id) != 0);
}

/*!
 * @brief Validate our non-volatile settings file.
 */
void GIGE_REG_CACHE::validate_reg_file()
{
    int     fd;
    int     istatus = -99;
    FILE    *fp;                // file pointer for the information returned from popen
    struct stat stat_buf;
    char    temp_buf[PATH_MAX];

    // First see if the register file and checksum file exist. If they don't they can't be restored
    // or have a checksum run on them.
    snprintf(temp_buf, PATH_MAX, "%s/%s", get_home_path(), REGISTER_FILE);
    fd = stat(temp_buf, &stat_buf);
    if (fd < 0)
    {
        fi_log(LOG_WARNING, "%s: cannot stat %s, err:%d(%s)", __FUNCTION__, temp_buf, errno, strerror(errno));

        // Save the default values out to the file.
        save();
        return;
    }
    snprintf(temp_buf, PATH_MAX, "%s/%s", get_home_path(), REGISTER_CHECKSUM_FILE);
    fd = stat(temp_buf, &stat_buf);
    if (fd < 0)
    {
        fi_log(LOG_WARNING, "%s: cannot stat %s, err:%d(%s)", __FUNCTION__, temp_buf, errno, strerror(errno));

        // Save the default values out to the file.
        save();
        return;
    }

    // Check the checksum file which contains the original file name. Note that md5check is a
    // shell script.
    snprintf(temp_buf, PATH_MAX, "%s/%s %s/%s", get_home_path(), MD5_CHECK, get_home_path(), REGISTER_CHECKSUM_FILE);
    fp = popen(temp_buf, "r");
    if (fp != NULL)
    {
        // Read data, if any.
        if (fgets(temp_buf, MAX_BUFFER_LEN, fp) != NULL)
        {
            // Did we get anything?, we should have gotten a '0' otherwise we have an error
            if (temp_buf[0] == '0')
            {
                istatus = 0;
            }
            else
            {
                fi_log(LOG_WARNING, "%s: Error validating checksum of the register file.", __FUNCTION__);
                istatus = -10;
            }
        }
        else
        {
            fi_log(LOG_WARNING, "%s: popen did not run correctly. Suspect it couldn't find md5check err:%d(%s).", __FUNCTION__, errno, strerror(errno));
            istatus = -11;
        }

        // Close the pipe.
        pclose(fp);
    }
    else
    {
        fi_log(LOG_WARNING, "%s: Cannot open register checksum compare file, err:%d(%s).",
               __FUNCTION__, errno, strerror(errno));
        istatus = -12;
    }

    // The MD5 sum failed, so the settings file is corrupt. Use the
    // default values.
    if (istatus != 0)
    {
        // Save the default values out to the file.
        save();
        return;
    }

    // Now we need to read a few values out of the settings file to determine if
    // it is valid.
    snprintf(temp_buf, PATH_MAX, "%s/%s", get_home_path(), REGISTER_FILE);
    fp = fopen(temp_buf, "r");
    if (!fp)
    {
        // This is bad.
        fi_log(LOG_ERR, "%s: cannot open file %s, err:%d(%s)", __FUNCTION__, temp_buf, errno, strerror(errno));
        return;
    }

    // Initialize our status variable.
    istatus = 0;

    // While istatus is positive we're doing OK. We are only looking for 3 items,
    // so while istatus is less than 3, keep looking. If we run out of file, then
    // we're done.
    while ((istatus >= 0) && (istatus < 3) && (fgets(temp_buf, MAX_BUFFER_LEN, fp) != NULL))
    {
        GIGE_REG    my_reg;

        // Pull the information out of the string.
        parse_reg_line(temp_buf, &my_reg);

        // Check specific values for validity.
        switch (my_reg.reg_num)
        {
        case GIGE_REG_MAC_HIGH:
            // Make sure we're a match (MAC address SHOULDN'T change).
            if (my_reg.reg_val_l != gige_regs[ GIGE_REG_MAC_HIGH ].reg_val_l)
            {
                istatus = -1;
                fi_log(LOG_DEBUG, "%s: MAC_HIGH didn't match %08x!=%08x", __FUNCTION__,
                       my_reg.reg_val_l, gige_regs[ GIGE_REG_MAC_HIGH ].reg_val_l);
            }
            else
                ++istatus;
            break;
        case GIGE_REG_MAC_LOW:
#ifdef NEED_EXACT_MATCH
            // ===================================================
            // ALM: Actually, we don't need an EXACT match - the
            //      OUI match (GIGE_REG_MAC_HIGH) should be
            //      good enough. Also, doing this exact match means
            //      FasMotion can't "push" configuration to a group.
            // ===================================================
            //
            // Make sure we're a match (MAC address SHOULDN'T change).
            if (my_reg.reg_val_l != gige_regs[ GIGE_REG_MAC_LOW ].reg_val_l)
                istatus = -2;
            else
#endif  /* NEED_EXACT_MATCH */

                ++istatus;
            break;
        case GIGE_REG_MAN_NAME:
            // This had better be US!
        {
            int len = strlen((char *)my_reg.reg_val_b);
            if (len > GIGE_REG_MAN_NAME_B_LEN)
            {
                // GIGE_REG buffer has room for null byte after this length.
                len = GIGE_REG_MAN_NAME_B_LEN;
                my_reg.reg_val_b[ len ] = '\0';
            }
            if (strcmp((char *)my_reg.reg_val_b,
                       (char *)gige_regs[ GIGE_REG_MAN_NAME ].reg_val_b) == 0)
                ++istatus;  // Exactly matched manufacturer name.
            else
            {
                // Strip off trailing junk.
                while (len > 0)
                {
                    if (isspace(my_reg.reg_val_b[ len - 1 ]) ||
                        iscntrl(my_reg.reg_val_b[ len - 1 ]))
                    {
                        --len;
                        my_reg.reg_val_b[ len ] = '\0';
                    }
                    else
                        break;
                }
                if (strcmp((char *)my_reg.reg_val_b,
                           (char *)gige_regs[ GIGE_REG_MAN_NAME ].reg_val_b) == 0)
                    ++istatus;  // After cleanup, exactly matched manufacturer name.
                else
                {
                    fi_log(LOG_DEBUG, "%s: MAN_NAME didn't match '%.31s'!='%.31s'", __FUNCTION__,
                           (const char *)(my_reg.reg_val_b), (const char *)(gige_regs[ GIGE_REG_MAN_NAME ].reg_val_b));

                    istatus = -3;
                }
            }
        }
        break;
        default:
            break;
        }
    }

    fclose(fp);

    // Check our status.
    if (istatus < 3)
    {
        // Save the default values out to the file.
        fi_log(LOG_WARNING, "%s: Settings file didn't validate (istatus=%d).", __FUNCTION__, istatus);
        save();
    }
}

/*!
 * @brief Operator overload for array indexing. This is used
 *    in conjunction with other overloads (casting, etc.)
 *    to access register from the cache.
 *
 *    Note that the mutex is grabbed here, and released
 *    in the other overload which completes the operation.
 *
 * @param reg_id    The register we're interested in.
 *
 * @return Reference to the class.
 */
GIGE_REG_CACHE& GIGE_REG_CACHE::operator[](uint32_t reg_id)
{
    if (!cache_lock.try_lock_for(std::chrono::milliseconds(LOCK_TIMEOUT_MS)))
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s(%p): =============== Couldn't LOCK [0x%x]! ==============",
               __PRETTY_FUNCTION__, (void *)caller, reg_id);
    }

    // Make sure the register ID is valid.
    if (validate_id(reg_id))
        last_index = reg_id;

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator[](int reg_id)
{
    if (!cache_lock.try_lock_for(std::chrono::milliseconds(LOCK_TIMEOUT_MS)))
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s(%p): =============== Couldn't LOCK [0x%x]! ==============",
               __PRETTY_FUNCTION__, (void *)caller, reg_id);
    }

    // Make sure the register ID is valid.
    if (validate_id((uint32_t)reg_id))
        last_index = (uint32_t)reg_id;

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator= (uint32_t reg_value)
{
    gige_regs[ last_index ].reg_val_l = reg_value;
    cache_lock.unlock();

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator= (int reg_value)
{
    gige_regs[ last_index ].reg_val_l = (uint32_t)reg_value;
    cache_lock.unlock();

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator&=(uint32_t reg_value)
{
    gige_regs[ last_index ].reg_val_l &= reg_value;
    cache_lock.unlock();

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator&=(int reg_value)
{
    gige_regs[ last_index ].reg_val_l &= (uint32_t)reg_value;
    cache_lock.unlock();

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator|=(uint32_t reg_value)
{
    gige_regs[ last_index ].reg_val_l |= reg_value;
    cache_lock.unlock();

    return *this;
}
GIGE_REG_CACHE& GIGE_REG_CACHE::operator|=(int reg_value)
{
    gige_regs[ last_index ].reg_val_l |= (uint32_t)reg_value;
    cache_lock.unlock();

    return *this;
}

bool GIGE_REG_CACHE::operator==(uint32_t reg_value)
{
    bool rc;

    rc = (gige_regs[ last_index ].reg_val_l == reg_value);
    cache_lock.unlock();

    return rc;
}
bool GIGE_REG_CACHE::operator==(int reg_value)
{
    bool rc;

    rc = (gige_regs[ last_index ].reg_val_l == (uint32_t)reg_value);
    cache_lock.unlock();

    return rc;
}
bool GIGE_REG_CACHE::operator!=(uint32_t reg_value)
{
    bool rc;

    rc = (gige_regs[ last_index ].reg_val_l != reg_value);
    cache_lock.unlock();

    return rc;
}
bool GIGE_REG_CACHE::operator!=(int reg_value)
{
    bool rc;

    rc = (gige_regs[ last_index ].reg_val_l != (uint32_t)reg_value);
    cache_lock.unlock();

    return rc;
}

uint32_t GIGE_REG_CACHE::operator&(uint32_t reg_value)
{
    uint32_t rc;

    rc = (gige_regs[ last_index ].reg_val_l & reg_value);
    cache_lock.unlock();

    return rc;
}
int GIGE_REG_CACHE::operator&(int reg_value)
{
    int rc;

    rc = (int)(gige_regs[ last_index ].reg_val_l & (uint32_t)reg_value);
    cache_lock.unlock();

    return rc;
}
uint32_t GIGE_REG_CACHE::operator|(uint32_t reg_value)
{
    uint32_t rc;

    rc = (gige_regs[ last_index ].reg_val_l | reg_value);
    cache_lock.unlock();

    return rc;
}
int GIGE_REG_CACHE::operator|(int reg_value)
{
    int rc;

    rc = (int)(gige_regs[ last_index ].reg_val_l | (uint32_t)reg_value);
    cache_lock.unlock();

    return rc;
}

#if 0
GIGE_REG_CACHE::operator int()
{
    int rc;

    rc = (int)gige_regs[ last_index ].reg_val_l;
    cache_lock.unlock();

    return rc;
}
GIGE_REG_CACHE::operator uint32_t()
{
    uint32_t rc;

    rc = gige_regs[ last_index ].reg_val_l;
    cache_lock.unlock();

    return rc;
}
#endif
GIGE_REG_CACHE::operator uint16_t *()
{
    uint16_t *word_ptr;

    word_ptr = &gige_regs[ last_index ].reg_val_w[0];
    cache_lock.unlock();

    return word_ptr;
}
GIGE_REG_CACHE::operator uint8_t *()
{
    uint8_t *byte_ptr;

    byte_ptr = &gige_regs[ last_index ].reg_val_b[0];
    cache_lock.unlock();

    return byte_ptr;
}
GIGE_REG_CACHE::operator char *()
{
    char *char_ptr;

    char_ptr = (char *)(&gige_regs[ last_index ].reg_val_b[0]);
    cache_lock.unlock();

    return char_ptr;
}
GIGE_REG_CACHE::operator GIGE_REG *()
{
    GIGE_REG    *reg_ptr;

    reg_ptr = ((gige_regs.count(last_index) == 0) ? NULL : &gige_regs[ last_index ]);
    cache_lock.unlock();

    return reg_ptr;
}

uint32_t GIGE_REG_CACHE::reg_addr()
{
    uint32_t rc = last_index;

    // Release the mutex.
    cache_lock.unlock();

    return rc;
}
uint8_t GIGE_REG_CACHE::reg_type()
{
    uint8_t rc;

    rc = gige_regs[ last_index ].reg_type;
    cache_lock.unlock();

    return rc;
}
uint8_t GIGE_REG_CACHE::reg_access()
{
    uint8_t rc;

    rc = gige_regs[ last_index ].reg_access;
    cache_lock.unlock();

    return rc;
}
uint8_t GIGE_REG_CACHE::reg_size()
{
    uint8_t rc;

    rc = gige_regs[ last_index ].reg_size;
    cache_lock.unlock();

    return rc;
}
uint8_t GIGE_REG_CACHE::reg_swap_mode()
{
    uint8_t rc;

    rc = gige_regs[ last_index ].reg_swap_mode;
    cache_lock.unlock();

    return rc;
}
int GIGE_REG_CACHE::reg_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int status;
    REG_UPDATE_FN update_fn;

    // Assume failure.
    status = GEV_STATUS_INVALID_PARAMETER;

    // Make sure we have an update function registered.
    if (gige_regs[ last_index ].reg_update == NULL)
    {
        fi_log(LOG_DEBUG, "%s: No update function for addr 0x%x!?", __FUNCTION__, last_index);
        update_fn = NULL;
    }
    else
    {
        // Invoke the update function.
        update_fn = gige_regs[ last_index ].reg_update;
    }

    // Release the mutex.
    cache_lock.unlock();

    // Invoke the update function, if appropriate.
    if (update_fn != NULL)
        status = update_fn(p_msg, reg_value);

    // Return our status.
    return status;
}

REG_UPDATE_FN GIGE_REG_CACHE::update_is_registered()
{
    REG_UPDATE_FN reg_update;

    reg_update = gige_regs[last_index].reg_update;
    cache_lock.unlock();

    return reg_update;
}

uint32_t GIGE_REG_CACHE::get_value()
{
    uint32_t rc;

    rc = gige_regs[ last_index ].reg_val_l;
    cache_lock.unlock();

    return rc;
}

uint32_t GIGE_REG_CACHE::set_value(uint32_t reg_value)
{
    gige_regs[ last_index ].reg_val_l = reg_value;
    cache_lock.unlock();

    return reg_value;
}

/*!
 * @brief Set a single GigE register by calling its update function.
 *
 * This ensures that the hardware matches the cached setting by calling the update function
 * that normally writes the hardware on a setting change.
 *
 * @param reg_addr  The GigE register number to update.
 *
 * @return true if success, false on failure
 */
bool GIGE_REG_CACHE::reflect_register(uint32_t reg_addr)
{
    REG_UPDATE_FN   reg_update;
    uint32_t        reg_value;
    int         status = GEV_STATUS_INVALID_PARAMETER;

    if (!cache_lock.try_lock_for(std::chrono::milliseconds(LOCK_TIMEOUT_MS)))
    {
        void *caller = __builtin_return_address(0);
        fi_log(LOG_ERR, "%s(%p): =============== Couldn't LOCK! ==============",
               __PRETTY_FUNCTION__, (void *)caller);
    }

    reg_value = gige_regs[ reg_addr ].reg_val_l;
    reg_update = gige_regs[ reg_addr ].reg_update;

    // FIXME Change the calling convention for update functions so we pass the register address.
    // It's dangerous to depend on last_index when calling the update function
    // because we don't continue to hold the lock while we invoke it.
    last_index = (uint32_t)reg_addr;

    // Release the mutex.
    cache_lock.unlock();

    if (reg_update != NULL)
        status = (*reg_update)(NULL, reg_value);

    return (status == GEV_STATUS_SUCCESS) ? true : false;
}
