/*!
 * @file image_correction.c
 * @brief Functions to support FPN and PRNU correction.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <memory>
#include <mutex>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/fas_video.h>

#include "image_correction.h"

#include "bounce.h"
#include "camera_server.h"
#include "ct_file.h"
#include "fcp.h"
#include "fpga.h"
#include "ipm_proc.h"
#include "mmapper.h"
#include "part_mgr.h"
#include "storage.h"

namespace FI_IMAGE_CORRECTION
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// IPM flags and constants.
const uint32_t IPMC_FPN_12BIT_BACKOFF   = 0x3F7;
const uint32_t IPMC_FPN_10BIT_BACKOFF   = 0x3F7;
const uint32_t IPMC_FPN_8BIT_BACKOFF    = 0x3E0;
const uint32_t IPMC_FPN_NO_BACKOFF  = 0x3FF;

// For PRNU, we need to use 16-bit RAW images. By transferring the images
// through the IPM, we can expand our data to the format we need. Because
// of this complication, we need a state machine to walk through the steps
// of the transfer.
enum PRNU_STATES
{
    PRNU_SETUP,                     // Initialize the various members of the state machine.
    PRNU_PROCESS_FRAME,             // Push a frame through the IPM, expanding it to 16-bits.
    PRNU_TRANSFER_FRAME,            // Transfer the frame to our system memory.
    PRNU_SUM_FRAME                  // Accumulate the frame into a summation matrix.
};

// This is our in-memory representation of the correction table file.
typedef struct CT_FILE
{
    CT_GT_FILE::CT_FILE_HDR hdr;    // The header information
    uint8_t *ct_ptr;                // Pointer to the data
    uint8_t *ct_base_ptr;           // Unaligned pointer to the memory area allocated.
} CT_FILE;

/*!
 * @brief The class declaration for our FPGA FPN and PRNU calibration functions.
 *
 * The CalibrationWorker object is a singleton.
 */
class CalibrationWorker : public FAS_THREAD::Thread
{
private:
    CalibrationWorker();

    static std::unique_ptr<CalibrationWorker> instance;
    static std::mutex instance_mutex;

    static void cleanup(void *arg);

    volatile bool   thread_running;     //!< Flag to indicate if we are running or not.

    std::unique_ptr<uint8_t[]> raw_table_base;  //!< A place to store data.
    uint8_t *raw_table;                 //!< An aligned pointer to the raw data.

    bool    ct_is_null;
    bool    gt_is_null;


    uint16_t black_offset;  //!< The black offset to use with the currently-loaded black frame.
    int     fpn_part;       //!< The currently active partition information.
    bool    do_factory_cal; //!< A flag to indicate if we are doing factory calibration (true) or not (false).
    bool    do_manual;      //!< True if calibration should use manually-set params.
    bool    do_prnu_cal;    //!< A flag to indicate if we are doing a PRNU calibration (true) or not (false).

    bool    ct_is_factory;  //!< True if the currently loaded CT is the factory CT.
    CT_FILE ct_info;        //!< The correction table info structure.
    mmapper ct_mmap;        //!< Memory mapped CT information.

    CT_FILE gt_info;        //!< The gain table info structure.
    mmapper gt_mmap;        //!< Memory mapped GT information.

    void    get_user_roi(CT_GT_FILE::CT_ROI *roi);
    bool    cur_ct_compatible(CT_GT_FILE::CT_ROI *part_roi);
    bool    ct_roi_is_ok(CT_GT_FILE::CT_ROI *part_roi);
    void    create_null_ct(const CT_GT_FILE::CT_ROI& roi);
    void    create_null_gt(const CT_GT_FILE::CT_ROI& roi);
    void    reset_ct();
    void    reset_gt();

public:
    virtual ~CalibrationWorker();

    // Singletons should not be cloneable.
    CalibrationWorker(CalibrationWorker& other) = delete;

    // Singletons should not be assignable.
    void operator=(const CalibrationWorker&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static CalibrationWorker *getInstance();

    virtual void run() override;

    bool    isCtNull()      { return ct_is_null; }
    bool    isGtNull()      { return gt_is_null; }
    bool    isRunning()     { return thread_running; }

    int     ctLoad(const char *ct_file);
    int     ctBuild();
    int     ctPropagate();
    void    factoryCalibration(bool enable, bool manual_params);
    int     gtLoad(const char *gt_file);
    void    prnuCalibrate();

    // Offset access methods.
    uint16_t getBlackOffset()               { return black_offset; }
    void setBlackOffset(uint16_t offset)    { black_offset = offset; }
};

std::unique_ptr<CalibrationWorker> CalibrationWorker::instance;
std::mutex CalibrationWorker::instance_mutex;

CalibrationWorker *CalibrationWorker::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if (instance.get() == nullptr)
    {
        instance = std::unique_ptr<CalibrationWorker>(new CalibrationWorker());
    }

    return instance.get();
}

} /* End of namespace FI_IMAGE_CORRECTION */

using namespace FI_IMAGE_CORRECTION;

/*!
 * @brief The worker thread constructor.
 */
CalibrationWorker::CalibrationWorker() :
    FAS_THREAD::Thread(true),
    do_manual(false)
{
    // Initialize our member variables.
    thread_running = false;
    raw_table = nullptr;
    do_factory_cal = false;
    do_prnu_cal = false;
    fpn_part = 0;
    black_offset = 0;

    ct_is_factory = false;
    ct_is_null = false;
    gt_is_null = false;

    // Prep the info structure.
    memset(&ct_info, 0, sizeof(ct_info));
    memset(&gt_info, 0, sizeof(gt_info));
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 */
CalibrationWorker::~CalibrationWorker()
{
}

/*!
 * @brief Reset the Correction Table information structure.
 */
void CalibrationWorker::reset_ct()
{
    // Release memory associated with the correction table.
    ct_mmap.close();

    // Clear out the structure.
    memset(&ct_info, 0, sizeof(ct_info));

    // No offset now.
    black_offset = 0;
}

/*!
 * @brief Reset the Gain Table information structure.
 */
void CalibrationWorker::reset_gt()
{
    // Release memory associated with the correction table.
    gt_mmap.close();

    // Clear out the structure.
    memset(&gt_info, 0, sizeof(gt_info));
}

/*!
 * @brief Set the calibration options before starting a calibration run
 */
void CalibrationWorker::factoryCalibration(bool enable, bool manual_params)
{
    do_factory_cal = enable;
    do_manual = manual_params;
}

/*!
 * @brief Set the calibration options before starting a calibration run
 */
void CalibrationWorker::prnuCalibrate()
{
    do_prnu_cal = true;
}

/*!
 * @brief Say whether the currently-loaded CT is compatible with current settings
 *
 * If the settings have changed since the last time we loaded from a CT file,
 * the current black frame might be incompatible. It is largely up to the sensor
 * to say what is compatible but for now we will just do some common checks.
 *
 * @param roi   The ROI on the sensor
 *
 * @return true if the current CT is compatible with current settings, false if not.
 */
bool CalibrationWorker::cur_ct_compatible(CT_GT_FILE::CT_ROI *part_roi)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    SENSOR::SENSOR_ROI ct_roi, sensor_roi;

    ct_roi.x = ct_info.hdr.roi.origin.x;
    ct_roi.y = ct_info.hdr.roi.origin.y;
    ct_roi.width  = ct_info.hdr.roi.width;
    ct_roi.height = ct_info.hdr.roi.height;
    sensor_roi.x = part_roi->origin.x;
    sensor_roi.y = part_roi->origin.y;
    sensor_roi.width  = part_roi->width;
    sensor_roi.height = part_roi->height;

    return sensor_ptr->is_cal_compatible(ct_info.hdr.sensor_type, &ct_info.hdr.sensor_params,
                                         &ct_roi, &sensor_roi);
}

/*!
 * @brief The worker thread.
 */
void CalibrationWorker::run()
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    Sensor *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t    value;
    bool    need_restart = true;

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // We're running.
    thread_running = true;

    // Are we doing a PRNU calibration?
    if (do_prnu_cal)
    {
        // The calibration will write out a new file, so close the old one.
        reset_gt();

        // PRNU is a sensor specific operation, so we need to hand off the
        // control to the sensor.
        sensor_ptr->run_prnu(ct_info.ct_ptr);
    }
    else
    {
        // The calibration will write out a new file so close the old one.
        reset_ct();

        // Run the calibration algorithm.
        sensor_ptr->run_calibration(do_factory_cal, do_manual, &need_restart);

        // Make sure the shutter is open.
        fpga_ptr->shutterClose(false);
    }

    // We're not running anymore.
    thread_running = false;

    // Restart normal capture. We might need to:
    // - Restore the capture buffer addresses, e.g. return to LR mode.
    // - Restore the sensor capture area so it respects the license.
    // - Restore the bit mode to the user setting.
    // - Set up the user-selected FPN method.
    //
    // We had to do this after we cleared the "thread_running" flag because
    // reenabling FPN will call functions (fpnLoad, fpnEnable) that check
    // whether the thread is currently running.
    if (need_restart)
        cs_ptr->restart_normal_after_calibration();

    // Clear the CALIBRATE bits and send an async notification to show we are done.
    // This also clears the UNCALIBRATED bit.
    value = regs[ GIGE_REG_FPN ].get_value();
    value &= (GIGE_FPN_SENSOR_ENABLE_MASK | GIGE_FPN_FPGA_ENABLE_MASK);
    if (ct_is_null && ((value & GIGE_FPN_FPGA_ENABLE_MASK) != 0))
    {
        value |= GIGE_FPN_FPGA_UNCALIBRATED; // This should not happen!
        fi_log(LOG_ERR, "%s: Null CT found after calibration", __FUNCTION__);
    }
    regs[ GIGE_REG_FPN ].set_value(value);
    fcp_notify_apps(GIGE_REG_FPN, value);
}

/*!
 * @brief Get the current ROI settings.
 *
 * Encapsulate standard function because CT_ROI has different data types.
 *
 * @param roi   Pointer to where we want to put the ROI values.
 */
void CalibrationWorker::get_user_roi(CT_GT_FILE::CT_ROI *roi)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint16_t start_x;
    uint16_t start_y;
    uint16_t width;
    uint16_t height;

    cs_ptr->get_user_roi(&start_x, &start_y, &width, &height);

    roi->origin.x = start_x;
    roi->origin.y = start_y;
    roi->width = width;
    roi->height = height;
}

/*!
 * @brief Check to make sure that the correction table's ROI is a super-set of
 *    the new ROI we're trying to work with.
 *
 * @param part_roi  The new ROI we're trying to work with.
 *
 * @return false if it is NOT OK, true if the CT's ROI is a super-set.
 */
bool CalibrationWorker::ct_roi_is_ok(CT_GT_FILE::CT_ROI *part_roi)
{
    uint32_t    ct_value;
    uint32_t    part_value;

    // Check to see if the existing correction table origin will work.
    if ((ct_info.hdr.roi.origin.x > part_roi->origin.x) ||
        (ct_info.hdr.roi.origin.y > part_roi->origin.y))
    {
        // The loaded CT will NOT work with the partitions ROI.
        return false;
    }

    // Check to see if the width will work.
    ct_value = ct_info.hdr.roi.origin.x + ct_info.hdr.roi.width;
    part_value = part_roi->origin.x + part_roi->width;
    if (ct_value < part_value)
    {
        // The loaded CT will NOT work with the partitions ROI.
        return false;
    }

    // Check to see if the height will work.
    ct_value = ct_info.hdr.roi.origin.y + ct_info.hdr.roi.height;
    part_value = part_roi->origin.y + part_roi->height;
    if (ct_value < part_value)
    {
        // The loaded CT will NOT work with the partitions ROI.
        return false;
    }

    // This will work.
    return true;
}

/*!
 * @brief Load a correction table from a file.
 *
 * This leaves the full correction table in memory in a buffer.
 * Constructing a smaller CT for the current ROI is in ct_build.
 *
 * @param ct_file   The file we want to open and load from.
 *
 * @return 0 if everything is OK, <0 on error.
 */
int CalibrationWorker::ctLoad(const char *ct_file)
{
    uint8_t *ct_ptr;    // The returned pointer after memory mapping the file.
    int     rc = -1;    // Our return code. Assume failure.
    int     ct_offset;  // Offset to the CT data.

    // Prep the info structure.
    reset_ct();

    // Is there a user reference frame?
    ct_ptr = (uint8_t *)ct_mmap.open(ct_file);
    if (ct_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: unable to open file '%s' (#%d:%s).\n", __PRETTY_FUNCTION__, ct_file, errno, strerror(errno));
        return rc;
    }

    // Read up the header.
    memcpy(&ct_info.hdr, ct_ptr, sizeof(CT_GT_FILE::CT_FILE_HDR));
    ct_offset = sizeof(CT_GT_FILE::CT_FILE_HDR);

    // Make sure the magic value is correct. If it is the old header, convert it.
    if (ct_info.hdr.magic == CT_GT_FILE::CT_MAGIC_1)
    {
        CT_GT_FILE::CT_FILE_HDR_1 old_header;

        memcpy(&old_header, &ct_info.hdr, sizeof(old_header));
        memset(&ct_info.hdr, 0, sizeof(ct_info.hdr));

        ct_info.hdr.magic = CT_GT_FILE::CT_MAGIC;
        ct_info.hdr.length = old_header.length;
        ct_info.hdr.sensor_type = CT_GT_FILE::LINCE_SENSOR_ANY;
        ct_info.hdr.roi = old_header.roi;

        // The offset in the file is immediately following the version 1 header.
        ct_offset = sizeof(old_header);
    }
    else if (ct_info.hdr.magic == CT_GT_FILE::CT_MAGIC_2)
    {
        // Headers version 2 and 3 are the same, except for
        // the identifying values.
        ct_info.hdr.magic = CT_GT_FILE::CT_MAGIC;

        // Clear out the overwritten values.
        memset(ct_info.hdr.sensor_id, 0, CT_GT_FILE::SENSOR_ID_LEN);

        // The offset in the file is immediately following the version 2 header.
        ct_offset = sizeof(CT_GT_FILE::CT_FILE_HDR_2);
    }
    else if (ct_info.hdr.magic != CT_GT_FILE::CT_MAGIC)
    {
        fi_log(LOG_ERR, "%s: Invalid CT magic (0x%x)!", __PRETTY_FUNCTION__, ct_info.hdr.magic);
        goto ct_bail0;
    }

    // Now that we have the header all straightened out, check to see if it is tied
    // to a particular sensor.
    if (ct_info.hdr.sensor_id[0] != 0)
    {
        Sensor *sensor_ptr = CameraServer::get_sensor();
        std::string sensor_id = std::string(ct_info.hdr.sensor_id);
        std::unique_ptr<char[]> sensor_ver = std::make_unique<char[]>(GIGE_SENSOR_VER_LEN);

        // Get the sensor version number.
        memset(sensor_ver.get(), 0, GIGE_SENSOR_VER_LEN);
        sensor_ptr->get_sensor_ver(sensor_ver.get(), GIGE_SENSOR_VER_LEN);

        // The table is tied to a sensor. Let's make sure there is a match.
        if (sensor_id.compare(sensor_ver.get()) != 0)
        {
            fi_log(LOG_WARNING, "%s: '%s' does not match this sensor (%s != %s).", __PRETTY_FUNCTION__, ct_file, sensor_id.c_str(), sensor_ver.get());
            goto ct_bail0;
        }
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: '%s' is generic.", __PRETTY_FUNCTION__, ct_file);
    }

    // Initialize our pointers.
    ct_info.ct_base_ptr = ((uint8_t *)ct_mmap.get_mmap_ptr()) + ct_offset;
    ct_info.ct_ptr = ct_info.ct_base_ptr;

    // Everything looks good.
    rc = 0;

ct_bail0:
    // Reset the info structure if we are in error.
    if (rc)
        reset_ct();

    // Return our result.
    return rc;
}

void CalibrationWorker::create_null_ct(const CT_GT_FILE::CT_ROI& roi)
{
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor          *sensor_ptr = CameraServer::get_sensor();
    uint32_t        framesize;
    uint16_t        width;

    // Prep the info structure.
    reset_ct();

    width = (roi.width + 3) & ~3;   // CT width must be multiple of 4 bytes.

    framesize = QUADWORD_SIZE * cs_ptr->calc_frame_buf_size(width, roi.height,
                                                            (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    framesize += FRAME_METADATA_HW_BYTES;

    ct_info.hdr.magic = CT_GT_FILE::CT_MAGIC;
    ct_info.hdr.length = framesize;
    ct_info.hdr.sensor_type = sensor_ptr->get_sensor_type();
    ct_info.hdr.roi.origin.x = roi.origin.x;
    ct_info.hdr.roi.origin.y = roi.origin.y;
    ct_info.hdr.roi.width = width;
    ct_info.hdr.roi.height = roi.height;

    // Allocate our buffer. Over allocate so we can get a decent alignment.
    ct_info.ct_base_ptr = nullptr;
    ct_info.ct_ptr = nullptr;
    ct_is_null = true;

    // Null subtraction means null offset.
    black_offset = 0;
}

void CalibrationWorker::create_null_gt(const CT_GT_FILE::CT_ROI& roi)
{
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor          *sensor_ptr = CameraServer::get_sensor();
    uint32_t        framesize;
    uint16_t        width;

    // Prep the info structure.
    reset_gt();

    width = (roi.width + 3) & ~3;   // GT width must be multiple of 4 bytes.

    framesize = QUADWORD_SIZE * cs_ptr->calc_frame_buf_size(width, roi.height,
                                                            (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    framesize += FRAME_METADATA_HW_BYTES;

    gt_info.hdr.magic = CT_GT_FILE::CT_MAGIC;
    gt_info.hdr.length = framesize;
    gt_info.hdr.sensor_type = sensor_ptr->get_sensor_type();
    gt_info.hdr.roi.origin.x = roi.origin.x;
    gt_info.hdr.roi.origin.y = roi.origin.y;
    gt_info.hdr.roi.width = width;
    gt_info.hdr.roi.height = roi.height;

    // Allocate our buffer. Over allocate so we can get a decent alignment.
    gt_info.ct_base_ptr = nullptr;
    gt_info.ct_ptr = nullptr;
    gt_is_null = true;
}


/*!
 * @brief Load the FPN and PRNU correction tables into the active partition.
 *
 * @return 0 on success, <0 on error.
 */
int CalibrationWorker::ctBuild()
{
    int retval;
    uint16_t    ofs_x;
    uint16_t    ofs_y;
    uint32_t    sensor_opts;
    bitmode_t   bit_mode;
    uint32_t    framesize;
    uint32_t    aligned_size;
    std::string temp_path;
    CT_GT_FILE::CT_ROI      part_roi;

    uint8_t *raw_ct;
    std::unique_ptr<uint8_t[]> raw_ct_base;
    uint8_t *raw_gt;
    std::unique_ptr<uint8_t[]> raw_gt_base;

    // Initialize our pointers.
    CameraServer *cs_ptr = CameraServer::get_cs();
    Sensor *sensor_ptr = CameraServer::get_sensor();
    PartMgr *part_mgr = CameraServer::get_partmgr();

    LOG("%s", __PRETTY_FUNCTION__);

    // Assume failure.
    retval = -1;

    // Get the ROI being used for the partition.
    get_user_roi(&part_roi);

    // We might need to adjust it based on internal sensor considerations.
    cs_ptr->get_user_bit_mode(&bit_mode);
    cs_ptr->get_sensor_opts(&sensor_opts);
    sensor_ptr->get_ct_offset(bit_mode, sensor_opts, &ofs_x, &ofs_y);
    part_roi.origin.x += ofs_x;
    part_roi.origin.y += ofs_y;

    // Assume the first partition is always 0.
    fpn_part = 0;

    // If we have the factory black reference currently loaded,
    // or if we have no black frame loaded at all,
    // or if the current black frame is a constructed no-op placeholder,
    // or if the current CT is not compatible with our settings,
    // try loading the user black reference.
    if ((ct_mmap.get_file_len() == 0) || ct_is_factory || ct_is_null)
    {
        ct_is_factory = false;
        ct_is_null = false;
        ctLoad(sensor_ptr->get_blkref_name());

        // If the load fails, it will leave us with no CT loaded; if we previously had
        // a factory black frame loaded, it is now cleared out so we will have to load
        // it again.
    }

    // If we decided to maintain multiple black frame files, we would want to attempt
    // a reload when the current one is not compatible. But with only one, if what we
    // have loaded is not compatible, the file won't be compatible either.
    if (!ct_roi_is_ok(&part_roi) || !cur_ct_compatible(&part_roi))
    {
        fi_log(LOG_DEBUG, "%s: Not compatible!", __PRETTY_FUNCTION__);

        // If what we have loaded is not compatible, we will use a null CT or factory.
        reset_ct();
    }

    // If we still didn't load a CT or if the ROI used to capture the loaded CT
    // doesn't cover what we need, see if there is a factory CT we can use.
    if (((ct_mmap.get_file_len() == 0) || !ct_roi_is_ok(&part_roi)) && sensor_ptr->has_factory_blkref())
    {
        // If we're here, then we'll use the factory CT if available.
        ct_is_factory = true;
        ct_is_null = false;

        // Fall back to the factory default CT.
        fi_log(LOG_INFO, "%s: Unable to use existing CT - switching to factory default.", __FUNCTION__);
        temp_path = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_BLKREF_FILE_RO);
        if (ctLoad(temp_path.c_str()) < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to open factory blkref! %s", __FUNCTION__, temp_path);

            // Fall back to the factory default CT.
            temp_path = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_BLKREF_FILE_RW);
            if (ctLoad(temp_path.c_str()) < 0)
            {
                fi_log(LOG_ERR, "%s: Unable to open factory blkref! %s", __FUNCTION__, temp_path);
                ct_is_factory = false;
            }
        }
    }

    // We discard any old black reference whenever we try to load so if our
    // previous attempts failed, we won't have one loaded now.
    // If there's none, create an empty black frame that does nothing.
    if (ct_info.ct_base_ptr == nullptr)
    {
        create_null_ct(part_roi);
    }

    // Special processing if we're using the factory black ref frame.
    if (ct_is_factory)
    {
        CT_GT_FILE::CT_POINT    lic_offset;

        // Figure out if we need to add in an offset due to licensing restrictions.
        //
        // Don't forget, some cameras were built with a small factory black reference
        // frame. We must use the correction table header info to determine what the offsets
        // should be.
        lic_offset.x = (ct_info.hdr.roi.width - cs_ptr->get_lic_width()) / 2;
        lic_offset.y = (ct_info.hdr.roi.height - cs_ptr->get_lic_height()) / 2;

        fi_log(LOG_INFO, "%s: ct( %d:%d@%dx%d ) part( %d:%d@%dx%d ) lic_offset=%d:%d",
               __PRETTY_FUNCTION__,
               ct_info.hdr.roi.origin.x, ct_info.hdr.roi.origin.y,
               ct_info.hdr.roi.width, ct_info.hdr.roi.height,
               part_roi.origin.x, part_roi.origin.y, part_roi.width, part_roi.height,
               lic_offset.x, lic_offset.y);

        // Adjust the partition ROI according to the licensing restrictions, if any.
        //
        // We do this NOW, as opposed to before the call to ct_roi_is_ok(), because the
        // USER blackref is license restricted. We don't want to adjust until we know we're
        // using a factory black ref.
        part_roi.origin.x += lic_offset.x;
        part_roi.origin.y += lic_offset.y;
    }
    else if (ct_is_null)
    {
        fi_log(LOG_INFO, "%s: using nullptr blackref ( %dx%d )", __FUNCTION__,
               part_roi.width, part_roi.height);
    }
    else
    {
        fi_log(LOG_INFO, "%s: using USER blackref ct( %d:%d@%dx%d ) part( %d:%d@%dx%d )",
               __FUNCTION__,
               ct_info.hdr.roi.origin.x, ct_info.hdr.roi.origin.y,
               ct_info.hdr.roi.width, ct_info.hdr.roi.height,
               part_roi.origin.x, part_roi.origin.y, part_roi.width, part_roi.height);
    }

    // Load the gain table.
    temp_path = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_GAIN_FILE_RO);
    if (gtLoad(temp_path.c_str()) < 0)
    {
        // A GT doesn't exist, so create a temporary stand-in.
        fi_log(LOG_ERR, "%s: Unable to open factory PRNU! %s", __FUNCTION__, temp_path);
        create_null_gt(part_roi);
    }
    else
    {
        // Not a null GT.
        gt_is_null = false;
    }

    // Figure out how much space the correction table takes for this size frame.
    // The CT table is always 8 bits per pixel regardless of the bit mode of captured frames.
    // The size calculation forces the width to a multiple of 4.
    framesize = calc_src_dma_size(part_roi.width, part_roi.height, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    if (framesize == 0)
    {
        fi_log(LOG_ERR, "%s: Unable to compute framesize for %ux%u!", __FUNCTION__,
               part_roi.width, part_roi.height);
        goto load_bail;
    }

    // Get the aligned size.
    aligned_size = convert_to_HEXWORD(framesize) << HEXWORD_SHIFT;

    // Allocate our buffesr. Over allocate so we can get a decent pointer alignment.
    raw_ct_base = std::unique_ptr<uint8_t[]>(new uint8_t[aligned_size + HEXWORD_MASK]());
    if (raw_ct_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate correction table memory!", __FUNCTION__);
        raw_ct = nullptr;
        goto load_bail;
    }
    raw_gt_base = std::unique_ptr<uint8_t[]>(new uint8_t[aligned_size + HEXWORD_MASK]());
    if (raw_gt_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate gain table memory!", __FUNCTION__);
        raw_gt = nullptr;
        goto load_bail;
    }

    // Clear out the tables.
    memset(raw_ct_base.get(), 0, aligned_size + HEXWORD_MASK);
    memset(raw_gt_base.get(), 0, aligned_size + HEXWORD_MASK);

    // Get an aligned pointer into the table.
    raw_ct = (uint8_t *)((((uint64_t)raw_ct_base.get()) + HEXWORD_MASK) & (uint64_t)(~((uint64_t)HEXWORD_MASK)));
    raw_gt = (uint8_t *)((((uint64_t)raw_gt_base.get()) + HEXWORD_MASK) & (uint64_t)(~((uint64_t)HEXWORD_MASK)));

    // Make sure we aren't using a null CT.
    if (!ct_is_null)
    {
        uint8_t     *ct_row_ptr;
        uint32_t    row_count;
        uint32_t    row_offset;
        uint8_t     *raw_row_ptr;

        // Skip our starting point down the proper number of rows.
        ct_row_ptr = ct_info.ct_ptr + (ct_info.hdr.roi.width * (part_roi.origin.y - ct_info.hdr.roi.origin.y));

        // Our raw pointer is at the beginning of our table.
        raw_row_ptr = raw_ct;

        // Figure out how deep into a row we have to skip.
        row_offset = part_roi.origin.x - ct_info.hdr.roi.origin.x;

        LOG("%s: ct_addr=0x%" PRIx64 ", gt_addr=0x%" PRIx64 ", row_offset=%d, w=%d, h=%d", __PRETTY_FUNCTION__, part_mgr->get_active_ct_addr(), part_mgr->get_active_gt_addr(),
            row_offset, part_roi.width, part_roi.height);

        // Make sure the partition ROI reflects 32-bit alignment.
        part_roi.width = (part_roi.width + 3) & ~3;

        // Loop through the number of rows in the partition ROI, copying the data over.
        for (row_count = 0; (int)row_count < part_roi.height; ++row_count)
        {
            // Copy the column data over.
            memcpy(raw_row_ptr, &ct_row_ptr[ row_offset ], part_roi.width);

            // Increment our pointers down a row.
            ct_row_ptr += ct_info.hdr.roi.width;
            raw_row_ptr += part_roi.width;
        }
    }

    // Make sure we aren't using a null GT.
    if (!gt_is_null)
    {
        uint8_t     *gt_row_ptr;
        uint32_t    row_count;
        uint32_t    row_offset;
        uint8_t     *raw_row_ptr;

        // If the correction table we are using is NOT the factory table, then we
        // still may need to tweak the ROI origin for the gain table.
        if (!ct_is_factory)
        {
            CT_GT_FILE::CT_POINT    lic_offset;

            // Figure out if we need to add in an offset due to licensing restrictions.
            //
            // Don't forget, some cameras were built with a small factory black reference
            // frame. We must use the correction table header info to determine what the offsets
            // should be.
            lic_offset.x = (gt_info.hdr.roi.width - cs_ptr->get_lic_width()) / 2;
            lic_offset.y = (gt_info.hdr.roi.height - cs_ptr->get_lic_height()) / 2;

            // Sanity check!
            if (lic_offset.x < 0)
            {
                fi_log(LOG_WARNING, "%s: license width (%d) is larger than our ROI (%d)!", __PRETTY_FUNCTION__, cs_ptr->get_lic_width(), gt_info.hdr.roi.width);
                lic_offset.x = 0;
            }
            if (lic_offset.y < 0)
            {
                fi_log(LOG_WARNING, "%s: license height (%d) is larger than our ROI (%d)!", __PRETTY_FUNCTION__, cs_ptr->get_lic_height(), gt_info.hdr.roi.height);
                lic_offset.y = 0;
            }

            fi_log(LOG_INFO, "%s: gt( %d:%d@%dx%d ) part( %d:%d@%dx%d ) lic_offset=%d:%d",
                   __PRETTY_FUNCTION__,
                   gt_info.hdr.roi.origin.x, gt_info.hdr.roi.origin.y,
                   gt_info.hdr.roi.width, gt_info.hdr.roi.height,
                   part_roi.origin.x, part_roi.origin.y, part_roi.width, part_roi.height,
                   lic_offset.x, lic_offset.y);

            // Adjust the partition ROI according to the licensing restrictions, if any.
            //
            // We do this NOW, as opposed to before the call to ct_roi_is_ok(), because the
            // USER blackref is license restricted. We don't want to adjust until we know we're
            // using a factory black ref.
            part_roi.origin.x += lic_offset.x;
            part_roi.origin.y += lic_offset.y;
        }

        // Skip our starting point down the proper number of rows.
        gt_row_ptr = gt_info.ct_ptr + (gt_info.hdr.roi.width * (part_roi.origin.y - gt_info.hdr.roi.origin.y));

        // Our raw pointer is at the beginning of our table.
        raw_row_ptr = raw_gt;

        // Figure out how deep into a row we have to skip.
        row_offset = part_roi.origin.x - gt_info.hdr.roi.origin.x;

        // Make sure the partition ROI reflects 32-bit alignment.
        part_roi.width = (part_roi.width + 3) & ~3;

        // Loop through the number of rows in the partition ROI, copying the data over.
        for (row_count = 0; (int)row_count < part_roi.height; ++row_count)
        {
            // Copy the column data over.
            memcpy(raw_row_ptr, &gt_row_ptr[ row_offset ], part_roi.width);

            // Increment our pointers down a row.
            gt_row_ptr += gt_info.hdr.roi.width;
            raw_row_ptr += part_roi.width;
        }
    }

    // Turn off the FPGA's FPN.
    // This SHOULD be redundant 'cuz it should already be disabled.
    FiImageCorrection::fpnDisable();

    // Give the sensor its parameters for this table. This is for things like
    // the sensor gain and the Lupa DACVREF value.
    sensor_ptr->set_cal_params(ct_info.hdr.sensor_type, &ct_info.hdr.sensor_params);

    // Set up the FPN offset. This will remain the same as long as we don't
    // change the bit selection or load a different correction table.
    FiImageCorrection::fpnApplyOffset(bit_mode, sensor_ptr->get_cal_offset(ct_info.hdr.sensor_type, &ct_info.hdr.sensor_params));

    // If we write it to image memory successfully, we succeeded.
    // This writes it to all the partitions even though we set up the FPGA FPN
    // to point to the first one; this way, if we save a CAP from any of the other
    // partitions, it will save the same CT table.
    if (part_mgr->writeCorrectionTable(raw_ct, aligned_size))
    {
        if (part_mgr->writeGainTable(raw_gt, aligned_size))
        {
            // Success.
            retval = 0;
        }
    }

load_bail:

    // Return our result.
    return retval;
}

/*!
 * @brief Propagate the correction and gain tables from the active partition to all the others.
 *
 * @return 0 on success, <0 on error.
 */
int CalibrationWorker::ctPropagate()
{
    // Initialize our pointers.
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();

    int     retval = -1;
    CT_GT_FILE::CT_ROI  part_roi;
    uint32_t    table_size;
    uint64_t    ct_addr;

    LOG("%s", __PRETTY_FUNCTION__);

    // Is the FPN thread running? This shouldn't happen but everyone checks for it so we will too.
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: worker is already running!", __FUNCTION__);
        return -1;
    }

    // Find the address for the correction table in the current partition.
    fpn_part = part_mgr->getActivePartition();
    ct_addr = part_mgr->get_active_ct_addr();
    if (ct_addr == 0)
    {
        fi_log(LOG_ERR, "%s: Can't get the address of the correction table.", __FUNCTION__);
        return -1;
    }

    // Get the ROI being used for the partition.
    get_user_roi(&part_roi);

    // Make sure the partition ROI reflects 32-bit alignment.
    part_roi.width = (part_roi.width + 3) & ~3;

    // Figure out how much space the correction table takes for this size frame.
    // The correction table is always 8 bits per pixel regardless of the bit mode of captured frames.
    // The gain table immediately follows the CT and is the same size.
    table_size = calc_src_dma_size(part_roi.width, part_roi.height, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);

    // Get the aligned size.
    table_size = convert_to_HEXWORD(table_size) << HEXWORD_SHIFT;

    // Allocate our buffer. Over allocate so we can get a decent alignment.
    raw_table_base = std::unique_ptr<uint8_t[]>(new uint8_t[table_size + HEXWORD_MASK]());
    if (raw_table_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of the frame buffer failed!", __FUNCTION__);
        raw_table = nullptr;
        goto prop_bail;
    }
    raw_table = (uint8_t *)((((uint64_t)raw_table_base.get()) + HEXWORD_MASK) & ~((uint64_t)HEXWORD_MASK));

    // Read the correction table from the active partition.
    // Byte swap and word swap the data.
    if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, ct_addr, raw_table, table_size) == 0)
        retval = -1;

    // If we write the correction and gain tables to image memory successfully, we succeeded.
    else if (part_mgr->writeCorrectionTable(raw_table, table_size))
        retval = 0;

    // Is there a gain table?
    if (part_mgr->isGainTableAllowed())
    {
        uint64_t    gt_addr;

        // Get the address.
        gt_addr = part_mgr->get_active_gt_addr();

        // Read the gain table from the active partition.
        // Byte swap and word swap the data.
        if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, gt_addr, raw_table, table_size) == 0)
            retval = -1;

        // If we write the correction and gain tables to image memory successfully, we succeeded.
        else if (part_mgr->writeGainTable(raw_table, table_size))
            retval = 0;
    }

    // Don't need the buffer any more.
    raw_table = nullptr;

prop_bail:

    // Return our result.
    return retval;
}

/*!
 * @brief Load a gain table from a file.
 *
 * This leaves the full gain table in memory in a buffer.
 * Constructing a smaller GT for the current ROI is in gt_build.
 *
 * @param gt_file   The file we want to open and load from.
 *
 * @return 0 if everything is OK, <0 on error.
 */
int CalibrationWorker::gtLoad(const char *gt_file)
{
    uint8_t *gt_ptr;    // The returned pointer after memory mapping the file.
    int     rc = -1;    // Our return code. Assume failure.
    int     gt_offset;  // Offset to the CT data.

    // Prep the info structure.
    reset_gt();

    // Is there a user reference frame?
    gt_ptr = (uint8_t *)gt_mmap.open(gt_file);
    if (gt_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: unable to open file '%s' (#%d:%s).\n", __FUNCTION__, gt_file,
               errno, strerror(errno));
        return rc;
    }

    // Read up the header.
    memcpy(&gt_info.hdr, gt_ptr, sizeof(CT_GT_FILE::GT_FILE_HDR));
    gt_offset = sizeof(CT_GT_FILE::GT_FILE_HDR);

    // The gain table was introduced with a version 2 file.
    if (gt_info.hdr.magic == CT_GT_FILE::CT_MAGIC_2)
    {
        // Headers version 2 and 3 are the same, except for
        // the identifying values.
        gt_info.hdr.magic = CT_GT_FILE::CT_MAGIC;

        // Clear out the overwritten values.
        memset(gt_info.hdr.sensor_id, 0, CT_GT_FILE::SENSOR_ID_LEN);

        // The offset in the file is immediately following the version 2 header.
        gt_offset = sizeof(CT_GT_FILE::CT_FILE_HDR_2);
    }
    else if (gt_info.hdr.magic != CT_GT_FILE::CT_MAGIC)
    {
        fi_log(LOG_ERR, "%s: Invalid CT magic (0x%x)!", __PRETTY_FUNCTION__, gt_info.hdr.magic);
        goto gt_bail0;
    }

    // Now that we have the header all straightened out, check to see if it is tied
    // to a particular sensor.
    if (gt_info.hdr.sensor_id[0] != 0)
    {
        Sensor *sensor_ptr = CameraServer::get_sensor();
        std::string sensor_id = std::string(gt_info.hdr.sensor_id);
        std::unique_ptr<char[]> sensor_ver = std::make_unique<char[]>(GIGE_SENSOR_VER_LEN);

        // Get the sensor version number.
        memset(sensor_ver.get(), 0, GIGE_SENSOR_VER_LEN);
        sensor_ptr->get_sensor_ver(sensor_ver.get(), GIGE_SENSOR_VER_LEN);

        // The table is tied to a sensor. Let's make sure there is a match.
        if (sensor_id.compare(sensor_ver.get()) != 0)
        {
            fi_log(LOG_WARNING, "%s: '%s' does not match this sensor (%s != %s).", __PRETTY_FUNCTION__, gt_file, sensor_id.c_str(), sensor_ver.get());
            goto gt_bail0;
        }
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: '%s' is generic.", __PRETTY_FUNCTION__, gt_file);
    }

    // Initialize our pointers.
    gt_info.ct_base_ptr = ((uint8_t *)gt_mmap.get_mmap_ptr()) + gt_offset;
    gt_info.ct_ptr = gt_info.ct_base_ptr;

    // Everything looks good.
    rc = 0;

gt_bail0:
    // Reset the info structure if we are in error.
    if (rc)
        reset_gt();

    // Return our result.
    return rc;
}

// ====================================================================================

// Initialize our static members.
bool FiImageCorrection::fpn_on(false);
std::string FiImageCorrection::prnu_red_stack;
std::string FiImageCorrection::prnu_green_stack;
std::string FiImageCorrection::prnu_blue_stack;

/*!
 * @brief Standard constructor.
 */
FiImageCorrection::FiImageCorrection()
{
}

/*!
 * @brief Standard destructor.
 */
FiImageCorrection::~FiImageCorrection()
{
}

/*!
 * @brief Start a black reference calibration run.
 *
 * @param do_factory    true if we're doing factory blackref, false otherwise.
 *
 * @return 0 on success, <0 on failure.
 */
int FiImageCorrection::fpnCalibrate(bool do_factory, bool manual_params)
{
    // Is the FPN thread running?
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: FPN worker is already running!", __FUNCTION__);
        return -1;
    }

    // Remember if we're supposed to do factory calibration.
    CalibrationWorker::getInstance()->factoryCalibration(do_factory, manual_params);

    // Start our calibration thread working.
    CalibrationWorker::getInstance()->start();

    // All good.
    return 0;
}

/*!
 * @brief Start a PRNU calibration run.
 *
 * This is the entry point for the PRNU calibration process. A LOT of assumptions
 * are being made:
 * 1) The capture settings (ROI, frame rate, etc.) are all setup appropriately.
 * 2) A black reference calibration has been taken at these settings.
 *
 * Since PRNU makes use of FPN, it is important that everything has been setup
 * properly.
 *
 * @return 0 on success, <0 on failure.
 */
int FiImageCorrection::prnuCalibrate()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Is the calibration thread running?
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: FPN worker is already running!", __FUNCTION__);
        return -1;
    }

    // Tell the calibration worker that we would like to do a PRNU calibration.
    CalibrationWorker::getInstance()->prnuCalibrate();

    // Start our calibration thread working.
    CalibrationWorker::getInstance()->start();

    // All good.
    return 0;
}

bool FiImageCorrection::prnuGenerate(const uint16_t *avg_image, const uint8_t *fpn_table, uint8_t *gain_table)
{
    bool rc = false;
    int bad_count = 0;
    int bit_depth;
    float max_pixel;
    uint32_t col;
    uint32_t row;
    uint16_t max_fixed_gain;
    Sensor *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

//#define DUMP_VALUES 1
#ifdef DUMP_VALUES
    FILE *wb_gains_pf = nullptr;
    FILE *wb_gains_p = nullptr;

    do
    {
        const char PRNU_FRAME_FILE[] = ".config/Fastec/scratch/prnu";
        std::string file_path;

        // Open our gains file.
        file_path = fmt::sprintf("%s/%s/wb_gains_Pf.csv", get_home_path(), PRNU_FRAME_FILE);
        wb_gains_pf = fopen(file_path.c_str(), "w");
        if (wb_gains_pf == nullptr)
        {
            fprintf(stderr, "Unable to open wb_gains_pf.csv\n");
            goto gen_exit;
        }
        file_path = fmt::sprintf("%s/%s/wb_gains_P.csv", get_home_path(), PRNU_FRAME_FILE);
        wb_gains_p = fopen(file_path.c_str(), "w");
        if (wb_gains_p == nullptr)
        {
            fprintf(stderr, "Unable to open wb_gains_p.csv\n");
            goto gen_exit;
        }
    } while (0);
#endif /* DUMP_VALUES */

    if (fpn_table == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpn_table is null!", __PRETTY_FUNCTION__);
        goto gen_exit;
    }

    // Figure out the bit-depth we're using.
    switch (regs[GIGE_REG_BIT_MODE].get_value())
    {
    case GIGE_BITMODE_9_0:
        bit_depth = 10;
        break;
    case GIGE_BITMODE_11_0:
        bit_depth = 12;
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected bit-sel value [0x%x]!", __PRETTY_FUNCTION__, regs[GIGE_REG_BIT_MODE].get_value());
        goto gen_exit;
    }

    // Use the maximum allowed pixel value.
    max_pixel = powf(2, bit_depth) - 1.0;

    // The maximum fixed gain is 2^fixed-point-resolution. If we
    // are using 8-bits of resolution, then the floating point value
    // of 1.0 is equivalent to the fixed point value of 256, or in
    // hex, 0x0100.
    max_fixed_gain = pow(2, CT_GT_FILE::FIXED_FRACTIONAL);

    // Loop through each row.
    for (row = 0; row < sensor_ptr->get_sensor_max_height(); ++row)
    {
        // Run through the row.
        for (col = 0; col < sensor_ptr->get_sensor_max_width(); ++col)
        {
            int pixel_index;
            uint16_t wn_pix;
            uint16_t bn_pix;
            uint16_t u_gain;
            float f_gain ;

            // Figure out where our pixel is.
            pixel_index = (row * sensor_ptr->get_sensor_max_width()) + col;

            // Get our pixel value low-order justified.
            wn_pix = avg_image[pixel_index] >> (16 - bit_depth);
            bn_pix = (uint16_t)fpn_table[pixel_index];

            // Calculate what the floating point gain is. This should NEVER
            // happen, but let's be safe.
            if (wn_pix <= bn_pix)
            {
                fi_log(LOG_ERR, "%s: How is the wn_pix (0x%04X/0x%04X, shfit=%d) <= bn_pix (0x%04X)?!!! [pixel_index=%d]", __PRETTY_FUNCTION__, wn_pix, avg_image[pixel_index], (16 - bit_depth), bn_pix, pixel_index);
                f_gain = 0; // This is really 1.0.
            }
            else
            {
                f_gain = (max_pixel / (float)(wn_pix - bn_pix + 1)) - 1.0;
            }

            // Calculate the fixed-point gain. This has the "assumed-1" subtracted.
            // NOTE: The value is truncated, not rounded up!
            u_gain = (uint16_t)(f_gain * max_fixed_gain);

            // The maximum gain value has to fit in a byte.
            if (u_gain > 0x00FF)
            {
                // Saturate the gain.
//                fi_log(LOG_DEBUG, "%s: pixel_index=%d gain overflow [0x%04x]", __PRETTY_FUNCTION__, pixel_index, u_gain);
                bad_count++;
                u_gain = 0x00FF;
            }

#ifdef DUMP_VALUES
            {
                // Save the gains out to their files.
                if ((col + 1) == pm_meta_ptr->roi_width)
                {
                    fprintf(wb_gains_p, "%f\n", f_gain);
                    fprintf(wb_gains_pf, "0x%04x\n", u_gain);
                }
                else
                {
                    fprintf(wb_gains_p, "%f,", f_gain);
                    fprintf(wb_gains_pf, "0x%04x,", u_gain);
                }
            }
#endif /* DUMP_VALUES */

            // Save the gain to the table.
            gain_table[pixel_index] = (uint8_t)(u_gain & 0xFF);
        }
    }

    // See if we had overflows.
    if (bad_count > 0)
        fi_log(LOG_DEBUG, "%s: bad_count=%d", __PRETTY_FUNCTION__, bad_count);

    // Looks like we did it!
    rc = true;

gen_exit:
#ifdef DUMP_VALUES
    if (wb_gains_p != nullptr)
        fclose(wb_gains_p);
    if (wb_gains_pf != nullptr)
        fclose(wb_gains_pf);
#endif /* DUMP_VALUES */

    // Return our status.
    return rc;
}

/*!
 * @brief Handle downloading RAW 16-bit images and summing them.
 *
 * @return True on success, false otherwise.
 */
bool FiImageCorrection::prnuMonoSum(int capture_count, bitmode_t bitmode, uint64_t cal_frame_addr, uint32_t *sum_table)
{
    bool rc = false;
    int ret;
    int frame_count;
    int pixel_idx;
    int prnu_state;
    bool prnu_done;
    uint32_t image_len;
    uint64_t ipm_buf_addr = 0;
    uint32_t dma_mod;
    uint32_t dma_dims;
    uint32_t proc_info;
    uint32_t pc_img_len(0);
    uint16_t *raw_table_ptr;
    uint32_t *raw_sum_ptr;
    std::unique_ptr<pm_metadata> pm_meta_ptr;

    int      bb_index = -1;
    uint64_t bb_addr;
    uint16_t *bb_ptr(nullptr);

    IPM_PROC::IPM_REQUEST ipm;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Sanity checks.
    if (!IS_10_BIT_MODE(bitmode) && !IS_12_BIT_MODE(bitmode))
    {
        fi_log(LOG_DEBUG, "%s: Invalid bitmode [0x%x]", __PRETTY_FUNCTION__, bitmode);
        return false;
    }
    if (sum_table == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sum_table is null!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize our state.
    prnu_state = PRNU_SETUP;

    // Keep going until told to stop.
    prnu_done = false;
    frame_count = 0;
    while (!prnu_done)
    {
        // Process based on the current state.
        switch (prnu_state)
        {
        case PRNU_SETUP:
            // Get a buffer for our image processing.
            bb_index = bb_get_avail();
            bb_addr = bb_get_avail_addr(bb_index);

            // Map the memory.
            bb_ptr = (uint16_t *)bb_mmap(bb_index, 0);
            if (bb_ptr == (uint16_t *)MAP_FAILED)
            {
                FI_LOG_TAG(__FILE__, __LINE__, "%s: FAIL! errno=%d [%s]", __PRETTY_FUNCTION__, errno, strerror(errno));
                rc = -1;
                prnu_done = true;
                break;
            }

            // 10-bit or 12-bit captures mean 16 bit raw.
            proc_info = IMGPROC_TYPE_RAW16;

            // Get a copy of the metadata - we have to tweak it.
            pm_meta_ptr = pm_ptr->getActivePartMetadata();
            pm_meta_ptr->roi_height = regs[GIGE_REG_HEIGHT].get_value();
            pm_meta_ptr->roi_width = regs[GIGE_REG_WIDTH].get_value();
            pm_meta_ptr->bit_sel = bitmode;

            // Update the partition metadata and set our review partition to the active partition.
            pm_ptr->updateActiveMetadata(pm_meta_ptr.get());
            pm_ptr->setReviewPartNum(pm_ptr->getActivePartition());

            // What will the size of the image in PC memory be?
            pc_img_len = get_processed_size(proc_info, pm_meta_ptr->roi_width, pm_meta_ptr->roi_height);

            // Setup for raw processing.
            dma_mod = IM2PCIE_DMA_IMAGE_DATA;
            dma_mod |= IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
            dma_mod |= IM2PCIE_DMA_16BIT_RAW_FMT;

            // The image dimensions will be constant.
            dma_dims = (regs[GIGE_REG_HEIGHT].get_value() << IPM_FRAME_DIM_Y_SHIFT) | regs[GIGE_REG_WIDTH].get_value();

            // Move on to getting a frame.
            prnu_state = PRNU_PROCESS_FRAME;
            break;
        case PRNU_PROCESS_FRAME:
            // Clear out our request structure.
            memset(&ipm, 0, sizeof(IPM_PROC::IPM_REQUEST));

            // Initialize the fields.
            ipm.image_addr = cal_frame_addr + (frame_count * pm_ptr->getFrameSize());
            ipm.proc_info = proc_info;
            ipm.process_live = false;

            // Process the frame.
            if (!ipmProcRequest(&ipm))
            {
                fi_log(LOG_ERR, "%s: ipm_request for addr 0x%" PRIx64 " failed!", __PRETTY_FUNCTION__, ipm.image_addr);
                ipm_buf_addr = 0;
                image_len = -1;
                prnu_done = true;
                break;
            }

            // Grab the output buffer address and the length.
            ipm_buf_addr = ipm.ipm_buf_addr;
            image_len = ipm.image_len;

            // Move on to transferring the frame.
            prnu_state = PRNU_TRANSFER_FRAME;
            break;
        case PRNU_TRANSFER_FRAME:
            // Have the FPGA transfer the image into our memory.
            ret = fpga_ptr->imageDownload(dma_mod, dma_dims, ipm_buf_addr, bb_addr, image_len, pc_img_len);
            if (ret >= 0)
            {
                // Move on to processing the frame.
                prnu_state = PRNU_SUM_FRAME;
                rc = 0;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
                prnu_done = true;
            }

            // Release our P1 buffer.
            ipm_proc_release(ipm_buf_addr);
            ipm_buf_addr = 0;
            break;
        case PRNU_SUM_FRAME:
            // ------------------------------------------------------------------------
//#define DEBUG_DUMP_FILES
#ifdef DEBUG_DUMP_FILES
            // I'm cranky with the formatter not indenting properly. Using a do-while false is equivalent to empty braces, so...
            do
            {
                const char PRNU_FRAME_FILE[] = ".config/Fastec/scratch/prnu";

                std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), PRNU_FRAME_FILE);
                if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
                {
                    fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
                }

                // Open up the black reference file.
                std::string tmp_file = fmt::sprintf("%s/cal_%03d", tmp_dir, frame_count);
                int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
                if (fd < 0)
                {
                    fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
                }
                else
                {
                    // Save out the information.
                    write(fd, bb_ptr, pc_img_len);
                    close(fd);
                }
            } while (0);
#endif /* DEBUG_DUMP_FILES */
            // ------------------------------------------------------------------------

            // Add this frame into the summation.
            raw_table_ptr = bb_ptr;
            raw_sum_ptr = sum_table;

            // Loop through all the pixels.
            for (pixel_idx = 0; pixel_idx < (int32_t)(pc_img_len / 2); ++pixel_idx)
            {
                // Add in the new pixel value.
                *raw_sum_ptr += *raw_table_ptr;

                // Bump our pointers.
                ++raw_table_ptr;
                ++raw_sum_ptr;
            }

            // Bump our frame count.
            ++frame_count;
            if (frame_count == capture_count)
            {
                // We are done.
                prnu_done = true;
            }
            else
            {
                // Go grab the next frame.
                prnu_state = PRNU_PROCESS_FRAME;
            }
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, prnu_state);
            prnu_done = true;
            break;
        }
    }

    // Did we finish?
    if (frame_count == capture_count)
        rc = true;

    // Release the bounce buffer.
    if (bb_index >= 0)
        bb_release(bb_index);

    // Release the IPM buffer, if we have one.
    if (ipm_buf_addr != 0)
        ipm_proc_release(ipm_buf_addr);

    // Return our status.
    return rc;
}


/*!
 * @brief Set the black offset level for the current correction table.
 *
 * Depending on how the correction table was created, we may have an
 * offset that should be added back in to restore the average
 * black level to what the sensor produces. How we do this depends
 * on what bits we are retaining from the sensor.
 *
 * @return returns 0 if everything went well, otherwise -1
 */
int FiImageCorrection::fpnApplyOffset(bitmode_t bit_mode, uint16_t offset_value)
{
    uint32_t backoff;
    uint32_t raw_reg;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // If the correction method uses an offset, set that for all the IPM ports.
    // If the offset is zero, revert to the old method and set the FPN backoff
    // value; the FPGA will not subtract the FPN correction if the pixel reading
    // from the sensor is above this 10-bit value.
    backoff = IPMC_FPN_NO_BACKOFF;
    if (offset_value == 0)
    {
        if (IS_8_BIT_MODE(bit_mode))
            backoff = IPMC_FPN_8BIT_BACKOFF;
        else if (IS_10_BIT_MODE(bit_mode))
            backoff = IPMC_FPN_10BIT_BACKOFF;
        else if (IS_12_BIT_MODE(bit_mode))
            backoff = IPMC_FPN_12BIT_BACKOFF;

        // Set the offset to zero and the gain to 1.0.
        raw_reg = 0x00001000;
    }
    else
    {
        unsigned    shift;

        // Depending on which bits are being captured, line up the offset.
        // The offset is in the low bits and we need to shift it so e.g. its
        // bit 7 aligns with bit 7 of the bits retained from the sensor.
        // We also need to shift it to the upper 16 bits of the register.
        shift = 16;
        switch ((unsigned)bit_mode)
        {
        // Bits 7:0 of the sensor will be bits 15:8 of the IPM sample.
        case GIGE_BITMODE_LOWER_8_BITS:
        case GIGE_BITMODE_7_0:
            shift += 8;
            if (offset_value > 0xff)
                offset_value = 0xff;
            break;
        // Bits 8:1 of the sensor will be bits 15:8 of the IPM sample.
        case GIGE_BITMODE_MIDDLE_8_BITS:
        case GIGE_BITMODE_8_1:
            shift += 7;
            if (offset_value > 0x1ff)
                offset_value = 0x1ff;
            break;
        // Bit 9 of the sensor output will be bit 15 of the IPM sample.
        case GIGE_BITMODE_UPPER_8_BITS: // 10-bit-sensor only
        case GIGE_BITMODE_10_BITS:  // 10-bit sensor only
        case GIGE_BITMODE_9_2:
        case GIGE_BITMODE_9_0:
            shift += 6;
            if (offset_value > 0x3ff)
                offset_value = 0x3ff;
            break;
        // Bit 10 of the sensor output will be bit 15 of the IPM sample.
        case GIGE_BITMODE_10_3:
        case GIGE_BITMODE_10_1:
            shift += 5;
            if (offset_value > 0x7ff)
                offset_value = 0x7ff;
            break;
        // Bit 11 of the sensor output will be bit 15 of the IPM sample so
        // bit 7 of the offset needs to shift by 4 to line up with sensor bit 7.
        case GIGE_BITMODE_11_4:
        case GIGE_BITMODE_11_2:
        case GIGE_BITMODE_11_0:
            shift += 4;
            if (offset_value > 0xfff)
                offset_value = 0xfff;
            break;
        }
        // Set the offset for the CT with gain=1.0.
        raw_reg = ((uint32_t)offset_value << shift) | 0x1000;
    }

    if (!fpga_ptr->regWrite(IPM_P0_FPN_CORR_THRESHOLD, backoff) ||
        !fpga_ptr->regWrite(IPM_P1_FPN_CORR_THRESHOLD, backoff) ||
        !fpga_ptr->regWrite(IPM_P2_FPN_CORR_THRESHOLD, backoff))
    {
        fi_log(LOG_ERR, "%s: Can't write IPM_Px_FPN_CORR_THRESHOLD.", __FUNCTION__);
        return -1;
    }
    if (!fpga_ptr->regWrite(IPM_P0_RAW_PIXEL_OFFSET_GAIN, raw_reg) ||
        !fpga_ptr->regWrite(IPM_P1_RAW_PIXEL_OFFSET_GAIN, raw_reg) ||
        !fpga_ptr->regWrite(IPM_P2_RAW_PIXEL_OFFSET_GAIN, raw_reg))
    {
        fi_log(LOG_ERR, "%s: Can't write IPM_Px_RAW_PIXEL_OFFSET_GAIN.", __FUNCTION__);
        return -1;
    }

    LOG("%s: thresh=0x%03x offset-gain=0x%08x", __FUNCTION__, backoff, raw_reg);

    // Remember what we used.
    CalibrationWorker::getInstance()->setBlackOffset(offset_value);
    return 0;
}

/*!
 * @brief Get the black offset for the currently-loaded black frame.
 *
 * @return returns 0 if everything went well, otherwise -1
 */
uint8_t FiImageCorrection::fpnGetOffset()
{
    return CalibrationWorker::getInstance()->getBlackOffset();
}

/*!
 * @brief Reset the FPN offset register to its default value.
 */
void FiImageCorrection::fpnResetOffset()
{
    uint32_t raw_reg = 0x00001000;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fpga_ptr->regWrite(IPM_P0_RAW_PIXEL_OFFSET_GAIN, raw_reg);
    fpga_ptr->regWrite(IPM_P1_RAW_PIXEL_OFFSET_GAIN, raw_reg);
    fpga_ptr->regWrite(IPM_P2_RAW_PIXEL_OFFSET_GAIN, raw_reg);
    fi_log(LOG_DEBUG, "%s: offset-gain=0x%08x", __FUNCTION__, raw_reg);
}

/*!
 * @brief Make sure all of our partitions have the CT we just loaded.
 *
 * When we load a CAP file, we switch over to using its correction table.
 * But if we had multiple partitions, we only loaded the table into the first
 * partition. This function propagates the correction table from the first partition
 * into all the others so they satisfy our normal conventions.
 *
 * @return returns 0 if everything went well, otherwise -1
 */
int FiImageCorrection::fpnPropagate()
{
    // Is the FPN thread running? This shouldn't happen but everyone checks for it so we will too.
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: worker is already running!", __FUNCTION__);
        return -1;
    }

    return CalibrationWorker::getInstance()->ctPropagate();
}

/*!
 * @brief Disable the FPGA's pixel-based FPN correction logic.
 *
 * @return returns 0 if everything went well, otherwise -1
 */
int FiImageCorrection::fpnDisable()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // Modify what's in the IPM  control register to disable FPN and PRNU.
    if (!fpga_ptr->regModify(IPM_P1_CONFIG, 0, ~(IPMC_FPN_ENA | IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7)))
    {
        fi_log(LOG_ERR, "%s: Can't modify IPM_P1_CONFIG.", __FUNCTION__);
        return -1;
    }

    // Clear the flag to indicate that we're now disabled.
    fpn_on = false;

    // All done.
    return 0;
}

/*!
 * @brief Enable the FPGA's pixel-based FPN correction logic.
 *
 * @return returns 0 if everything went well, otherwise -1
 */
int FiImageCorrection::fpnEnable()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    CameraServer *cs_ptr = CameraServer::get_cs();

    uint32_t    value;
    uint32_t    frame_size;
    uint64_t    ct_addr;    // Address within the partition for the correction table.
    uint64_t    gt_addr;    // Address within the partition for the gain table.
    uint16_t    width;
    uint16_t    height;
    int     rc;

    // Read what's in the P1 (Live View) control register.
    if (!fpga_ptr->regRead(IPM_P1_CONFIG, &value))
    {
        fi_log(LOG_ERR, "%s: Can't read IPM_P1_CONFIG.", __FUNCTION__);
        return -1;
    }

    // We do not want to modify any of the settings if the FPGA has FPN enabled.
    // Make absolutely certain that FPN is disabled.
    if ((value & (IPMC_FPN_ENA | IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7)) != 0)
    {
        // Modify what's in the IPM  control register to disable FPN and PRNU.
        if (!fpga_ptr->regModify(IPM_P1_CONFIG, 0, ~(IPMC_FPN_ENA | IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7)))
        {
            fi_log(LOG_ERR, "%s: Can't modify IPM_P1_CONFIG.", __FUNCTION__);
            return -1;
        }
    }

    // Clear the flag to indicate that we're now disabled.
    fpn_on = false;

    // Is the FPN thread running?
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: FPN worker is running, can't set up FPN", __FUNCTION__);
        return -1;
    }

    // Set up the FPGA's correction and gain table addresses.
    // Even if we have multiple partitions, the FPGA will continue to use the table for
    // the currently-active one, which is what we set up here.

    // Find the address for the correction and gain tables in the current partition.
    ct_addr = part_mgr->get_active_ct_addr();
    if (ct_addr == 0)
    {
        fi_log(LOG_ERR, "%s: Can't get the address of the correction table.", __FUNCTION__);
        return -1;
    }

    // If PRNU is allowed, get the address for the gain table. Otherwise, set it to 0.
    if (part_mgr->isGainTableAllowed())
        gt_addr = part_mgr->get_active_gt_addr();
    else
        gt_addr = 0;

    // Get the ROI we're working with.
    cs_ptr->get_user_roi(nullptr, nullptr, &width, &height);

    // Notify the FPGA of the new correction table information. Since this is going to the FPGA,
    // make sure we give it the absolute address (not partition relative). It doesn't matter which
    // 8-bit mode we use here for the CT size, the size comes out the same.
    frame_size = calc_src_dma_size(width, height, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    rc = fpga_ptr->ipmCtUpdate(frame_size, ct_addr, gt_addr);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Can't update the FPGA CT information.", __FUNCTION__);
        return -1;
    }

    // Set the enable bits.
    value |= IPMC_FPN_ENA;
    if (prnuIsAvailable() && part_mgr->isGainTableAllowed() && (gt_addr != 0))
    {
        // Enable PRNU.
        value |= (IPMC_PRNU_ENA | IPMC_PRNU_GAIN_SEL_1_7);
    }

    // Write the control register back out.
    if (!fpga_ptr->regWrite(IPM_P1_CONFIG, value))
    {
        fi_log(LOG_ERR, "%s: Can't write IPM_P1_CONFIG.", __FUNCTION__);
        return -1;
    }

    // Set the flag to indicate that we're now enabled.
    fpn_on = true;

    // Let folks know that FPN is enabled.
    cs_ptr->refresh_ct_status();

    // All good.
    return 0;
}


/*!
 * @brief Load the FPN correction table into the active partition.
 *
 * @return 0 on success, <0 on error.
 */
int FiImageCorrection::fpnLoad()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Is the FPN thread running?
    if (CalibrationWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: worker is already running!", __FUNCTION__);
        return -1;
    }

    // Get the correction table setup.
    return CalibrationWorker::getInstance()->ctBuild();
}

/*!
 * @brief Return true if calibration is in progress.
 */
bool FiImageCorrection::fpnCalibrateActive()
{
    return CalibrationWorker::getInstance()->isRunning();
}

/*!
 * @brief Return true if the loaded CT is null
 */
bool FiImageCorrection::fpnNeedsCal()
{
    return CalibrationWorker::getInstance()->isCtNull();
}

/*!
 * @brief Return true if the gain table is loaded. False if it's empty/not available.
 */
bool FiImageCorrection::prnuIsAvailable()
{
    return !CalibrationWorker::getInstance()->isGtNull();
}

uint32_t FiImageCorrection::prnuSetup(const char *settings_str, int str_length)
{
    Json::Value requested_settings;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    bool parsingSuccessful;
    uint32_t status;

//    LOG("%s: settings [%s]", __PRETTY_FUNCTION__, settings_str);

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(settings_str, settings_str + str_length, &requested_settings, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, settings_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // Do we have the red path?
    prnu_red_stack.clear();
    if (requested_settings["red"].empty())
    {
        fi_log(LOG_DEBUG, "%s: No red path!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }
    try
    {
        prnu_red_stack = requested_settings["red"].asString();
    }
    catch (Json::Exception& e)
    {
        fi_log(LOG_DEBUG, "%s: requested_settings[\"red\"].asString() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
    }
    if (prnu_red_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: The red path is empty!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // Do we have the green path?
    prnu_green_stack.clear();
    if (requested_settings["green"].empty())
    {
        fi_log(LOG_DEBUG, "%s: No green path!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }
    try
    {
        prnu_green_stack = requested_settings["green"].asString();
    }
    catch (Json::Exception& e)
    {
        fi_log(LOG_DEBUG, "%s: requested_settings[\"green\"].asString() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
    }
    if (prnu_green_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: The green path is empty!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // Do we have the blue path?
    prnu_blue_stack.clear();
    if (requested_settings["blue"].empty())
    {
        fi_log(LOG_DEBUG, "%s: No blue path!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }
    try
    {
        prnu_blue_stack = requested_settings["blue"].asString();
    }
    catch (Json::Exception& e)
    {
        fi_log(LOG_DEBUG, "%s: requested_settings[\"blue\"].asString() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
    }
    if (prnu_blue_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: The blue path is empty!", __PRETTY_FUNCTION__);
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // All done.
    status = GEV_STATUS_SUCCESS;

bail:
    return status;
}

bool FiImageCorrection::prnuImageAvg(int capture_count, const std::string& img_path, uint16_t *img_avg_ptr)
{
    int col;
    int row;
    int width;
    int height;
    int framesize;
    int file_count;
    uint32_t *raw_sum_ptr;
    std::unique_ptr<uint32_t[]> raw_sum_table;
    std::unique_ptr<uint16_t[]> row_buffer;
    std::vector<std::string> files;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity checks.
    if (capture_count <= 0)
    {
        fi_log(LOG_DEBUG, "%s: invalid capture_count: %d", __PRETTY_FUNCTION__, capture_count);
        return false;
    }
    if (img_avg_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: img_avg_ptr is null!", __PRETTY_FUNCTION__);
        return false;
    }
    if (img_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: img_path is empty!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize our dimensions.
    height = regs[GIGE_REG_HEIGHT].get_value();
    width = regs[GIGE_REG_WIDTH].get_value();
    framesize = width * height;

    // Allocate our "summation" table.
    raw_sum_table = std::unique_ptr<uint32_t[]>(new uint32_t[framesize]());
    if (raw_sum_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Allocate our row buffer.
    row_buffer = std::unique_ptr<uint16_t[]>(new uint16_t[width]);

    // Clear the sum table.
    memset(raw_sum_table.get(), 0, sizeof(raw_sum_table[0]) * framesize);

    // Get a list of all the tiff files in the path.
    storage::get_listing(img_path, std::string(".tif"), files);
    if (files.size() < (size_t)capture_count)
    {
        fi_log(LOG_ERR, "%s: <%s> does not have %d files", __PRETTY_FUNCTION__, img_path, capture_count);
        return false;
    }

    // Walk through the first "capture_count" files found.
    for (file_count = 0; file_count < capture_count; ++file_count)
    {
        TIFF *image_tif;

        // Open the file.
        image_tif = TIFFOpen(files[file_count].c_str(), "r");
        if (image_tif == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Unable to open '%s'", __PRETTY_FUNCTION__, files[file_count].c_str());
            break;
        }

        // Read the file and add the pixels to our sum table.
        for (row = 0; row < height; ++row)
        {
            uint16_t *row_ptr;

            // Read the row.
            TIFFReadScanline(image_tif, row_buffer.get(), row);

            // Point to the beginning of the sum table row.
            raw_sum_ptr = raw_sum_table.get() + (row * width);
            row_ptr = row_buffer.get();

            // Add the pixels to the sum table.
            for (col = 0; col < width; ++col)
                raw_sum_ptr[col] += row_ptr[col];
        }

        // Close the file.
        TIFFClose(image_tif);
    }
    if (file_count != capture_count)
    {
        fi_log(LOG_DEBUG, "%s: Did not process all files, file_count=%d", __PRETTY_FUNCTION__, file_count);
        return false;
    }

    // Now average the values.
    for (row = 0; row < height; ++row)
    {
        uint16_t *avg_ptr;

        // Point to the beginning of the sum table row.
        raw_sum_ptr = raw_sum_table.get() + (row * width);
        avg_ptr = img_avg_ptr + (row * width);

        // Average the pixels.
        for (col = 0; col < width; ++col)
        {
            avg_ptr[col] = (uint16_t)(raw_sum_ptr[col] / capture_count);
        }
    }

    // All is well.
    return true;
}

bool FiImageCorrection::prnuColorSum(int capture_count, uint16_t *avg_image)
{
    int col;
    int row;
    int width;
    int height;
    int framesize;
    uint16_t *r_ptr;
    uint16_t *g_ptr;
    uint16_t *b_ptr;
    uint32_t bayer_pattern;
    std::unique_ptr<uint16_t[]> r_buffer;
    std::unique_ptr<uint16_t[]> g_buffer;
    std::unique_ptr<uint16_t[]> b_buffer;
    Sensor *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure we have the paths to our stacks.
    if (prnu_red_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: prnu_red_stack path is empty!", __PRETTY_FUNCTION__);
        return false;
    }
    if (prnu_green_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: prnu_green_stack path is empty!", __PRETTY_FUNCTION__);
        return false;
    }
    if (prnu_blue_stack.empty())
    {
        fi_log(LOG_DEBUG, "%s: prnu_blue_stack path is empty!", __PRETTY_FUNCTION__);
        return false;
    }
    if (avg_image == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: avg_image pointer is null!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize our dimensions.
    height = regs[GIGE_REG_HEIGHT].get_value();
    width = regs[GIGE_REG_WIDTH].get_value();
    framesize = width * height;

    // What is our sensor's bayer pattern?
    bayer_pattern = sensor_ptr->IPM_bayer_pattern_start(0, 0);
    if (bayer_pattern != IPMC_SRC_FRM_BAYER_GR)
    {
        fi_log(LOG_DEBUG, "%s: Unsupported bayer pattern (0x%08x)", __PRETTY_FUNCTION__, bayer_pattern);
        return false;
    }

    // Allocate our "average" red, green and blue tables.
    r_buffer = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (r_buffer.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of red buffer failed!", __PRETTY_FUNCTION__);
        return false;
    }
    g_buffer = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (g_buffer.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of green buffer failed!", __PRETTY_FUNCTION__);
        return false;
    }
    b_buffer = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (b_buffer.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of blue buffer failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Generate the average red, green and blue images.
    if (!prnuImageAvg(capture_count, prnu_red_stack, r_buffer.get()))
    {
        fi_log(LOG_DEBUG, "%s: prnu_image_avg() for red failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (!prnuImageAvg(capture_count, prnu_green_stack, g_buffer.get()))
    {
        fi_log(LOG_DEBUG, "%s: prnu_image_avg() for green failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (!prnuImageAvg(capture_count, prnu_blue_stack, b_buffer.get()))
    {
        fi_log(LOG_DEBUG, "%s: prnu_image_avg() for blue failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Get our pointers setup.
    r_ptr = r_buffer.get();
    g_ptr = g_buffer.get();
    b_ptr = b_buffer.get();

    // Build up our composite image.
    for (row = 0; row < height; ++row)
    {
        for (col = 0; col < width; ++col)
        {
            uint32_t pixel_index = (row * width) + col;

            switch (bayer_pattern)
            {
            case IPMC_SRC_FRM_BAYER_GR:
                // The Bayer matrix is "GR", so the pixels are
                //      GRGR
                //      BGBG
                if ((row & 1) == 0)
                {
                    if ((col & 1) == 0)
                    {
                        // Even row, Even col ==> GREEN
                        avg_image[pixel_index] = g_ptr[pixel_index];
                    }
                    else
                    {
                        // Even row, Odd col ==> RED
                        avg_image[pixel_index] = r_ptr[pixel_index];
                    }
                }
                else
                {
                    if ((col & 1) == 0)
                    {
                        // Odd row, Even col ==> BLUE
                        avg_image[pixel_index] = b_ptr[pixel_index];
                    }
                    else
                    {
                        // Odd row, Odd col ==> GREEN
                        avg_image[pixel_index] = g_ptr[pixel_index];
                    }
                }
                break;
            default:
                break;
            }
        }
    }

    // All is well
    return true;
}
