/*!
 * @file image_fiv.cpp
 *
 * @brief This file contains the code which implements saving
 *    GIGE_STREAM_FORMAT_xxx files.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>

#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>

#include <dirent.h>
#include <fcntl.h>

#include <memory>
#include <fmt/printf.h>

#include <fastec/fi_exception.h>
#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include <fastec/fas_video.h>
#include <fastec/fi_image.h>
#include <fastec/fi_stream.h>

#include "autosave.h"
#include "bounce.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "fcp.h"
#include "fi_internal.h"
#include "image_format.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "part_mgr.h"
#include "save_proc.h"
#include "storage.h"

namespace IMAGE_FIV
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The default playback rate for saved videos.
const int DEFAULT_PLAYBACK_RATE = 30;

//! This string is used to format a shell command which checks that the space
//! available on a mounted device is less than a specific size.
//!
//! The string parameter is the mount point and the 64bit int is the file size.
//!
//! A return value of 0 indicates the device does not have enough space for the file.
const char  *VID_GEN_CHECK_SPACE_FORMAT = "%s/.config/Fastec/scripts/check_space.sh %.*s %llu";

//--------------------------------------------------------------------
// Function prototypes
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

std::unique_ptr<FAS_VIDEO::HEADER_DATA> fiv_header_buffer;    // The file header information for the current save operation.
static uint8_t  *fiv_image_buf_ptr = nullptr;       // Where we want to actually put the data. Assigned by bb_mmap().
static uint32_t fiv_image_buf_size;                 // Size of allocated buffer in bytes
static uint32_t fiv_proc_info;
static uint32_t fiv_dma_mod;
static uint32_t fiv_dma_width;
static uint32_t fiv_dma_height;

} /* End of IMAGE_FIV */

// Make things easier for us.
using namespace IMAGE_FIV;

/*!
 * @brief Checks the space available on external storage
 *
 * @param total_required    Number of bytes we are hoping for.
 * @param target_path       Path to look for the space.
 *
 * @return 0 if there is enough space, one of the errno codes if there is not enough space
 */
int StreamSave::checkSpace(uint64_t total_required, const char *target_path)
{
    int err_code;
    uint64_t total, avail, inodes;

    // Are we talking about more than FAT32 can support, we better check it.
    if (total_required > FAS_VIDEO::FI_VIDEO_MAX_FAT32_SIZE)
    {
        // Check to see if the filesystem will support a file this big.
        if ((err_code = storage::fs_supports_large_file(std::string(target_path))) != 0)
        {
            fi_log(LOG_WARNING, "%s: File system doesn't support large files [%d] (need %" PRIu64 ")!", __PRETTY_FUNCTION__, err_code, total_required);
            goto bail;
        }
    }

    // Determine the capacity of the device.
    if ((err_code = storage::get_capacity(target_path, &total, &avail, &inodes)) != 0)
    {
        fi_log(LOG_WARNING, "%s Error checking disk space", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (total_required > avail)
    {
        err_code = ENOSPC;
        goto bail;
    }

    // Assume all is well.
    err_code = 0;

bail:
    // Return the error code.
    return err_code;
}

/*!
 * @brief Standard C interface to write a metadata buffer out.
 *
 * @param dest_path Where to write.
 * @param meta_buffer   What to write.
 * @param meta_length   How much to write.
 */
void StreamSave::saveFivMetadata(const char *dest_path, const char *meta_buffer, uint32_t meta_length)
{
    int fd;

    // Open the destination file.
    fd = open(dest_path, O_CREAT | O_RDWR | O_TRUNC, 0666);
    if (fd == -1)
    {
        fi_log(LOG_ERR, "%s: Unable to create '%s', err=%d(%s)", __FUNCTION__, dest_path, errno, strerror(errno));
        return;
    }

    // Write the data to the file.
    write(fd, meta_buffer, meta_length);

    // Close the file.
    close(fd);
}

void StreamSave::getProcessingInfo(uint32_t stream_type, pm_metadata *meta_ptr)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();

    LOG("%s: stream_type=0x%x", __PRETTY_FUNCTION__, stream_type);

    // Set up the image type, one of the IMGPROC_TYPE_... values.
    fiv_proc_info = imgio::get_image_type(stream_type, meta_ptr);

    // If we're doing JPEG processing, then we need to decide about
    // what YUV format to use.
    if (fiv_proc_info == IMGPROC_TYPE_JPEG)
    {
        if (sensor_ptr->is_sensor_color())
        {
            fiv_proc_info = IMGPROC_TYPE_NV12;
        }
        else
        {
            fiv_proc_info = IMGPROC_TYPE_Y8;
        }
    }

    // Was FPN enabled?
    if ((meta_ptr->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
        fiv_proc_info |= IMGPROC_USE_FPN;

    if ((destination_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_BMP)
        fiv_proc_info |= IMGPROC_ROW_WD_ALIGN;
}

void StreamSave::getDmaMod(uint32_t stream_type, pm_metadata *meta_ptr)
{
    // Get our class pointer.
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    bool    sensor_is_color;

    // Are we a color or mono sensor?
    sensor_is_color = sensor_ptr->is_sensor_color();
    fiv_dma_mod = get_stream_dma_mod(stream_type, meta_ptr->bit_sel, sensor_is_color);
}

/*!
 * @brief This is the public facing entry point to saving
 *    a sequence of encoded frames using the fi_video
 *    library.
 *
 * @return true if we're still working, false otherwise.
 */
bool StreamSave::doSave()
{
    uint32_t stream_format;
    bool still_working;         // Return true if we're still working on it.

    // Sanity checks.
    if (format_info.state >= IMAGE_FORMAT::IMAGE_STATE_COUNT)
    {
        fi_log(LOG_ERR, "%s: Unexpected state [0x%x]", __FUNCTION__, format_info.state);
        format_info.state = IMAGE_FORMAT::IMAGE_STATE_DONE;
        return false;
    }

    // Process the command. CAP files are handled a little differently.
    stream_format = destination_info & GIGE_STREAM_FORMAT_MASK;
    if ((stream_format == GIGE_STREAM_FORMAT_CAPTURE) || (stream_format == GIGE_STREAM_FORMAT_CAPSTK))
    {
        // Save the capture file.
        if (capSave())
        {
            // Save the metadata.
            saveMetadata();

            // Are we in FasFireLR mode?
            if ((stream_format == GIGE_STREAM_FORMAT_CAPTURE) || (stream_format == GIGE_STREAM_FORMAT_CAPSTK))
            {
                // Save the PFMD.
                savePfmd();
            }
        }

        // We are always complete at this point with captures, so
        // force "still_working" to be false.
        still_working = false;
    }
    else
    {
        // Run the state machine.
        still_working = save_states[format_info.state](this);
    }

    // Clean up if we are done.
    if (!still_working)
    {
        // Release our image buffer.
        if (fiv_image_buf_ptr != nullptr)
        {
            bb_release(bb_index);
            fiv_image_buf_ptr = nullptr;
        }

        // Finish up with the AVI writer, if we have one.
        if (streamWriter != nullptr)
        {
            streamWriter->CloseStream();
            streamWriter.reset();
        }
        else if (streamHandle)
        {
            // Make sure we have access to the internal library.
            if (FiInternal::factory() != nullptr)
            {
                // Get our function pointer.
                FiInternal::CLOSE_PACK_STREAMER_FN funcPtr = FiInternal::factory()->getClosePackFuncPtr();
                if (funcPtr != nullptr)
                {
                    // Close the pack writer.
                    FAS_INTERNAL::FIP_STATUS closeStatus = (*funcPtr)(streamHandle);
                    if (closeStatus != FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED)
                    {
                        fi_log(LOG_WARNING, "%s: Packed writer did not produce a complete video, closeStatus=0x%x", __PRETTY_FUNCTION__, (int)closeStatus);
                    }
                }
            }

            // Reset the handle.
            streamHandle = 0;
        }
    }

    // Return our "still_working" status.
    return still_working;
}

/*!
 * @brief Is the save stream type a stream video.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is a video, false if not.
 */
bool StreamSave::isStreamVideo(uint32_t streamType)
{
    bool isAVideo = false;

    switch (streamType)
    {
    case GIGE_STREAM_FORMAT_AVI:
    case GIGE_STREAM_FORMAT_MP4:
    case GIGE_STREAM_FORMAT_PACK:
        isAVideo = true;
        break;
    default:
        isAVideo = false;
        break;
    }

    return isAVideo;
}

int StreamSave::ltrFrameLoad(uint32_t frame_num)
{
    int rc = -1;
    uint8_t *raw_ptr;
    uint32_t raw_length;
    uint32_t dma_mod;
    uint64_t frame_offset;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();

    LOG("%s: frame_num=%u", __PRETTY_FUNCTION__, frame_num);

    // Get the SSD offset in bytes.
    frame_offset = part_mgr->get_active_frame_addr(frame_num, false);
    LOG("%s: frame_offset=0x%" PRIx64, __PRETTY_FUNCTION__, frame_offset);

    // Our destination pointer is the LTR bounce buffer.
    raw_ptr = (uint8_t *)bb_mmap(ltr_bb_index, 0);
    if (raw_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to map ltr_bb_index (%d).", __PRETTY_FUNCTION__, ltr_bb_index);
        goto bail;
    }

    // Get the length of the image. This also includes the PFMD, but that's OK.
    raw_length = part_mgr->getFrameSize();

    // Read up the image data.
    if (part_mgr->ltrReadBytes(raw_ptr, frame_offset, raw_length) != (int)raw_length)
    {
        fi_log(LOG_ERR, "%s: Unable to read from 0x%" PRIx64 " for %u bytes", __PRETTY_FUNCTION__, frame_offset, raw_length);
        goto bail;
    }

    // Now ask the FPGA to put the data into the image memory buffer.
    dma_mod = PCIE2IM_DMA_BYTE_ORDER_R2L | PCIE2IM_DMA_WORD_ORDER_L2R;
    LOG("%s: dma_mod=0x%08x, len=%u", __PRETTY_FUNCTION__, dma_mod, raw_length);
    rc = fpga_ptr->imageUpload(dma_mod, ltr_bb_addr, ltr_frame_addr, raw_length);
    if (rc < 0)
    {
        // Oops! What happened here?!
        fi_log(LOG_ERR, "%s: Unable to DMA data!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = 0;

bail:
    // Return our status.
    return rc;
}

int StreamSave::preparePackedWriter()
{
    FAS_INTERNAL::FIP_STREAM_CONFIGURATION configInfo;
    std::unique_ptr<pm_metadata> pm_ptr;
    PartMgr *part_mgr = CameraServer::get_partmgr();

    // Sanity checks.
    if (part_mgr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Missing part_mgr pointer!", __PRETTY_FUNCTION__);
        return EFAULT;
    }

    // Get a pointer to the metadata.
    pm_ptr = part_mgr->getActivePartMetadata();
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Missing partition metadata pointer!", __PRETTY_FUNCTION__);
        return EFAULT;
    }

    // Initialize the FAS_INTERNAL::FIP_STREAM_CONFIGURATION structure.
    configInfo.manufacturer.assign(pm_ptr->man_name);
    configInfo.cameraModel.assign(pm_ptr->model_name);
    configInfo.cameraVersion.assign(pm_ptr->app_ver);
    configInfo.macAddress.assign((const char *)pm_ptr->mac_addr);
    configInfo.userComment.assign((const char *)pm_ptr->comment);

    configInfo.roiX = pm_ptr->roi_x;
    configInfo.roiY = pm_ptr->roi_y;
    configInfo.width = pm_ptr->roi_width;
    configInfo.height = pm_ptr->roi_height;
    configInfo.frameRate = pm_ptr->fps;
    configInfo.exposure = pm_ptr->shutter_speed;
    configInfo.frameCount = getTotalFrames();        // This is the number of frames to save, not the number captured.

    if (IS_8_BIT_MODE(pm_ptr->bit_sel))
        configInfo.bitsPerPixel = 8;
    else if (IS_10_BIT_MODE(pm_ptr->bit_sel))
        configInfo.bitsPerPixel = 10;
    else
        configInfo.bitsPerPixel = 12;

    configInfo.sensorType = pm_ptr->sensor_type;
    switch ((pm_ptr->sensor_opts >> GIGE_SENSOPT_BIN_SHIFT) & GIGE_SENSOPT_BIN_MASK)
    {
    case 0:
    case 1:
    default:
        configInfo.sensorBinningOpt = FAS_INTERNAL::NO_BINNING;
        break;
    case 2:
        configInfo.sensorBinningOpt = FAS_INTERNAL::BINNING_2X;
        break;
    case 4:
        configInfo.sensorBinningOpt = FAS_INTERNAL::BINNING_4X;
        break;
    }
    switch ((pm_ptr->sensor_opts >> GIGE_SENSOPT_SUBS_SHIFT) & GIGE_SENSOPT_SUBS_MASK)
    {
    case 0:
    case 1:
    default:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::NO_SUBSAMPLING;
        break;
    case 2:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::SUBSAMPLE_2X;
        break;
    case 4:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::SUBSAMPLE_4X;
        break;
    }

    // ALM_TODO:
    // We need to gather this information. Provide defaults if unknown.
    (void)configInfo.lensManufacturer;
    (void)configInfo.lensModel;
    configInfo.focalLength = 35;        // Just a default value for now.
    (void)configInfo.aperture;

    // The time in the pm_metadata is the time of the trigger frame, not the time of the
    // first frame of the sequence.
    {
        uint32_t startTimeSec;
        uint32_t startTimeMicrosec;

        double trigTime;
        double clipStartTime;

        struct tm tm_info;

        // Convert the trigger time to a double for easier manipulation.
        trigTime = ((double)pm_ptr->trig_time_s) + (((double)pm_ptr->trig_time_ns) / 1000000000);
        if (pm_ptr->trig_frame == pm_ptr->first_frame)
            clipStartTime = trigTime;
        else if (pm_ptr->trig_frame > format_info.start_frame)
            clipStartTime = trigTime - ((pm_ptr->trig_frame - format_info.start_frame) * (1.0 / ((double)pm_ptr->fps)));
        else
            clipStartTime = trigTime + ((format_info.start_frame - pm_ptr->trig_frame) * (1.0 / ((double)pm_ptr->fps)));

        // Convert the clipStartTime in seconds and microseconds.
        startTimeSec = (uint32_t)clipStartTime;
        startTimeMicrosec = (uint32_t)((clipStartTime - ((double)startTimeSec)) * 1000000.0);

        // Convert our SBS to a time structure.
        sbs_to_tm(startTimeSec, &tm_info);

        // Convert the time struct in to time since epoch.
        configInfo.timeStartOfStreamSeconds = mktime(&tm_info);
        configInfo.timeStartOfStreamMicrosec = startTimeMicrosec;
    }

    // Create an instance of our streamer.
    streamHandle = 0;
    if (FiInternal::factory() != nullptr)
    {
        FiInternal::GET_HANDLE_FN funcPtr = FiInternal::factory()->getHandleFuncPtr();
        if (funcPtr != nullptr)
        {
            streamHandle = (*funcPtr)(format_info.target_path, configInfo);
        }
    }
    if (streamHandle == 0)
    {
        fi_log(LOG_DEBUG, "%s: fiCreatePackStreamer() failed.", __PRETTY_FUNCTION__);
        return EFAULT;
    }

    // Create the video stream writer.
    fi_log(LOG_DEBUG, "%s Creating Packed Writer for: '%s'", __PRETTY_FUNCTION__, format_info.target_path.c_str());

    // Return our result.
    return 0;
}

/*!
 * @brief Sit in as tight a loop as possible, processing each image
 *    as fast as we can.
 *
 * @return true on success, false otherwise.
 */
bool StreamSave::processFivImage()
{
    bool ltr_mode;
    bool done_with_frames = false;
    std::string temp_path;

    // Get a pointer to our support classes.
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if ((destination_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_META_ONLY)
    {
        format_info.current_frame = format_info.end_frame + 1;
        fi_log(LOG_ERR, "%s: META_ONLY format so frames are not saved!", __FUNCTION__);
        return true;
    }

    // Loop through the frames.
    format_info.current_frame = format_info.start_frame;

    // Determine if we're in LTR mode or not.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);

    try
    {
        int status;
        const char *suffix_ptr;     // Pointer to path components.
        SAVE_PROC::SAVE_REQUEST request;

        // Initialize the save processor.
        saveProcInit();

        // Initialize our request block.
        request.start_frame = format_info.start_frame;
        request.end_frame = format_info.end_frame;
        request.ltr_mode = ltr_mode;
        request.proc_info = fiv_proc_info;
        request.dma_mod = fiv_dma_mod;
        request.image_height = fiv_dma_height;
        request.image_width = fiv_dma_width;
        request.video_writer = streamWriter;
        request.pack_handle = streamHandle;
        request.video_codec = video_codec;

        // Figure out the suffix information.
        suffix_ptr = imgio::get_suffix(destination_info & GIGE_STREAM_FORMAT_MASK);
        if (suffix_ptr == nullptr)
        {
            format_info.status = EINVAL;
            fi_log(LOG_ERR, "%s: no suffix for dest_info=0x%x!", __PRETTY_FUNCTION__, destination_info);
            return false;
        }

        // Save information for creating sub-folders.
        request.do_subfolders = format_info.do_subfolders;
        request.subfolder_count = -1;

        // Pass along if we are supposed to use the shared memory interface.
        request.using_sharedmem = imgSaveWorker::getInstance()->getUsingShmem();

        // Build up the target file name. Should contain everything except the frame number so we can
        // use this string as the format string in an sprintf().
        if (format_info.do_subfolders)
            temp_path = fmt::sprintf("%s/img%%04d/%%07d.%.5s", format_info.target_path, suffix_ptr);
        else
            temp_path = fmt::sprintf("%s/%%07d.%.5s", format_info.target_path, suffix_ptr);
        request.path_str.assign(temp_path);

        // Load up our image header, if appropriate.
        if (fiv_header_buffer == nullptr)
        {
            request.header_ptr = nullptr;
            request.header_len = 0;
        }
        else
        {
            request.header_ptr = fiv_header_buffer->data();
            request.header_len = fiv_header_buffer->size();
        }

        // Add a pointer to our format information structure.
        request.info_ptr = &format_info;

        // Do the save.
        status = saveProcRequest(&request);
        if (status)
        {
            fi_log(LOG_DEBUG, "%s: save_proc_request() returned %d", __PRETTY_FUNCTION__, status);
            format_info.status = status;
        }

        // Close down the save processor.
        saveProcClose();

        // We're done.
        done_with_frames = true;
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "%s: Exception in save_proc: '%s'", __PRETTY_FUNCTION__, e.what().c_str());
    }
    catch (std::exception& e)
    {
        // Note that the standard exceptions "what()" returns a 'const char *'.
        fi_log(LOG_DEBUG, "%s: Unhandled C++ exception: '%s'", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        LOG("%s: Unhandled exception of unknown type", __PRETTY_FUNCTION__);
    }

    // Are we done?
    if (done_with_frames)
    {
        if (streamWriter != nullptr)
        {
            FAS_VIDEO::STREAM_STATUS closeStatus = streamWriter->CloseStream();
            if (closeStatus != FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
            {
                fi_log(LOG_WARNING, "%s: Closing streamWriter was not clean.  Stream status=%d", __PRETTY_FUNCTION__, (int)closeStatus);
            }
            streamWriter.reset();    // Reset calls destructor and sets the unique pointer empty.
        }
        else if (streamHandle)
        {
            if (FiInternal::factory() != nullptr)
            {
                FiInternal::CLOSE_PACK_STREAMER_FN funcPtr = FiInternal::factory()->getClosePackFuncPtr();
                if (funcPtr != nullptr)
                {
                    FAS_INTERNAL::FIP_STATUS closeStatus = (*funcPtr)(streamHandle);
                    if (closeStatus != FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED)
                    {
                        fi_log(LOG_WARNING, "%s: Packed writer did not produce a complete video, closeStatus=0x%x", __PRETTY_FUNCTION__, (int)closeStatus);
                    }
                }
            }

            // Reset the handle.
            streamHandle = 0;
        }
    }

    // Return our result.
    return done_with_frames;
}

/*!
 * @brief Return the percentage completion for the current or most recent save.
 *
 * This must be legal to call as soon as the saver has been constructed, so
 * anything it depends on has to be initialized in the constructor.
 *
 * @return percentage done 0-100
 */
int StreamSave::getCompletionPercent()
{
    int percent;

    // Are we saving CAP files?
    if (((destination_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPTURE) || ((destination_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPSTK))
    {
        percent = capCompletionPercent();
    }
    else
    {
        int total_frames;

        // How many frames are we talking about?
        total_frames = (format_info.end_frame - format_info.start_frame) + 1;

        // Have we started yet?
        if ((format_info.current_frame < format_info.start_frame) || (total_frames <= 0))
        {
            fi_log(LOG_WARNING, "%s: funny state: cur=%u start=%u total=%u!", __PRETTY_FUNCTION__,
                   format_info.current_frame, format_info.start_frame, total_frames);
            percent = 0;
        }
        else
        {
            // Calculate the percent complete. The current frame can actually be end_frame+1
            // so we could compute more than 100% done.
            if (format_info.current_frame == format_info.end_frame)
                percent = 100;
            else
                percent = ((format_info.current_frame - format_info.start_frame) * 100) / total_frames;
            if (percent > 100)
                percent = 100;
        }
    }

    return percent;
}

/*!
 * @brief Setup our state so we can get started saving our image files.
 *
 * @return true if everything started out fine, false otherwise.
 */
bool StreamSave::startFivSave()
{
    uint32_t    stream_type;
    std::unique_ptr<pm_metadata> meta_ptr;
    PartMgr *part_mgr = CameraServer::get_partmgr();

    LOG("%s: target_path='%s'", __PRETTY_FUNCTION__, format_info.target_path.c_str());

    // Get a pointer to our metadata.
    meta_ptr = part_mgr->getReviewPartMetadata();

    // Isolate our stream type.
    stream_type = destination_info & (GIGE_STREAM_FORMAT_MASK | GIGE_STREAM_FORMAT_RAW);

    // Remember the image size in case we need it for DMA processing.
    fiv_dma_width = meta_ptr->roi_width;
    fiv_dma_height = meta_ptr->roi_height;

    // Initialize our globals.
    streamHandle = 0;
    streamWriter.reset();

    // Are we doing a video (vs. a stack of stills)?
    if (stream_type == GIGE_STREAM_FORMAT_PACK)
    {
        int ppwErr;

        // Prepare our writer.
        ppwErr = preparePackedWriter();
        if (ppwErr != 0)
        {
            // Log the error.
            fi_log(LOG_ERR, "%s: Unable to create Packed Stream Writer for file %s [errno:%d(%s)]", __PRETTY_FUNCTION__,
                   format_info.target_path.c_str(), ppwErr, strerror(ppwErr));
            return false;
        }
    }
    else if (isStreamVideo(stream_type))
    {
        int     playback_rate;          // Playback rate of a video.

        // For videos we don't do subfolders.
        format_info.do_subfolders = false;
        format_info.subfolder_count = 0;

        // Pull the playback rate from the destination info.
        playback_rate = (destination_info & GIGE_PLAYBACK_RATE_MASK) >> GIGE_PLAYBACK_RATE_SHIFT;
        if (playback_rate <= 0)
            playback_rate = DEFAULT_PLAYBACK_RATE;

        // Create the video stream writer.
        //
        // NOTE: Normally we'd do a std::move() here, but with C++ 11 we have something called "copy elision"
        //       (https://en.cppreference.com/w/cpp/language/copy_elision). To quote from the reference below:
        //       "Copy elision is the general process where, when returned from a function, an object
        //       is not copied nor moved, resulting in zero-copy pass-by-value semantics. It includes
        //       both return value optimization (RVO) and named return value optimization (NRVO)."
        //       (https://source.coveo.com/2018/11/07/interaction-between-move-semantic-and-copy-elision-in-c++/)
        switch (stream_type)
        {
        case GIGE_STREAM_FORMAT_AVI:
            streamWriter = FAS_VIDEO::FiVideo::CreateAVIVideoStreamWriter(format_info.target_path.c_str(),
                                                                          playback_rate, fiv_dma_width, fiv_dma_height,
                                                                          ((format_info.end_frame - format_info.start_frame) + 1));

            break;
        case GIGE_STREAM_FORMAT_MP4:
            streamWriter = FAS_VIDEO::FiVideo::CreateMP4VideoStreamWriter(format_info.target_path.c_str(), playback_rate,
                                                                          fiv_dma_width, fiv_dma_height, ((format_info.end_frame - format_info.start_frame) + 1),
                                                                          h264_quality, h264_speed,
                                                                          (h264_enable_all_keyframes ? 1 : FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL));
            break;
        default:
            fi_log(LOG_DEBUG, "%s: No support for video streamType: %d", __PRETTY_FUNCTION__, stream_type);
            return false;
        }

        // If unable to create the video stream writer or its reusable image data object log an error.
        if (streamWriter == nullptr)
        {
            // Log the error.
            fi_log(LOG_ERR, "%s: Unable to create Video Stream Writer for file %s [errno:%d(%s)]", __PRETTY_FUNCTION__,
                   format_info.target_path.c_str(), errno, strerror(errno));
            return false;
        }

        fi_log(LOG_DEBUG, "%s: Successfully created Video Stream Writer for: %s", __PRETTY_FUNCTION__, streamWriter->GetVideoName().c_str());
    }
    else
    {
        int         status;

        // Set our subfolder information, in case we need it.
        format_info.subfolder_index = format_info.target_path.size();
        format_info.subfolder_count = 0;
        if ((format_info.end_frame - format_info.start_frame) > FOLDER_COUNT_THRESHOLD)
            format_info.do_subfolders = true;
        else
            format_info.do_subfolders = false;

        // Make sure our target folder exists.
        status = storage::mkpath(format_info.target_path.c_str(), (S_IRWXU | S_IRWXO | S_IRWXG));
        if (status == -1)
        {
            format_info.status = errno;
            fi_log(LOG_ERR, "%s: Can't create '%s', err=%d(%s)", __FUNCTION__, format_info.target_path.c_str(), errno, strerror(errno));
            return false;
        }

        // Load up our image header, if appropriate.
        fiv_header_buffer = imgio::get_image_header(stream_type, meta_ptr.get());

        // Make sure it worked.
        if (fiv_header_buffer == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to get the header information!", __FUNCTION__);
            return false;
        }
    }

    // Set the current frame count.
    format_info.current_frame = format_info.start_frame;

    // Get the information we need to process the image data.
    getProcessingInfo(stream_type, meta_ptr.get());

    // Initialize the DMA modifier we'll use.
    getDmaMod(stream_type, meta_ptr.get());

    // Get a buffer for our image processing.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Map the memory.
    fiv_image_buf_ptr = (uint8_t *)bb_mmap(bb_index, 0);
    if (fiv_image_buf_ptr == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        format_info.status = ENOMEM;
        return false;
    }

    // Remember how big our buffer is.
    fiv_image_buf_size = bb_get_mmap_len(bb_index);

    // Load up our header information.
    return true;
}
