/*!
 * @file image_fmt_cap.cxx
 *
 * @brief This file contains the code which implements saving
 *    GIGE_STREAM_FORMAT_CAPTURE files.
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>

#include <dirent.h>
#include <fcntl.h>
#include <memory.h>

#include <chrono>
#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <vector>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <linux/fs.h>
#include <linux/fiemap.h>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/gige_msg.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_trace.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_correction.h"
#include "image_format.h"
#include "image_io.h"
#include "part_mgr.h"
#include "sensor.h"

namespace IMAGE_FMT
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! How much will we move in a single file system "write".
//! NOTE: Make sure we fit within a bounce buffer! Nominally,
//!   this means 256 MBytes.
const uint32_t MAX_SINGLE_WRITE = 32 * 1024 * 1024;
const uint32_t MAX_SINGLE_READ = MAX_SINGLE_WRITE;

//! How much will we DMA from image memory to our bounce buffer.
//! NOTE: Make sure that the value is on a "page" boundary (4K should work).
const uint32_t MAX_SINGLE_DMA = MAX_SINGLE_WRITE;

//! How many DMA buffers do we want.
const int DMA_BB_COUNT = 8;

//! How big is our bounce buffer?
//! NOTE: Make sure the total size fits within the MAXIMUM bounce
//!       buffer size, becaue we're only allocating one.
const uint32_t TOTAL_DMA_BB_SIZE = DMA_BB_COUNT * MAX_SINGLE_DMA;

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

volatile int     percent;
volatile uint64_t    save_offset;
volatile uint32_t    load_offset;

} /* End of IMAGE_FMT */

// Make things easier for us.
using namespace IMAGE_FMT;
using namespace CS_TRACE;
using namespace std::chrono;

TRACE_GLOBAL;           //!< Use the global trace control word.

/*!
 * @brief Highly optimized, multi-threaded routine to DMA from image memory (im_source_addr) to system memory
 *        (dma_addr/xfer_buffer) and write to a file (cap_fd).
 *
 * @param cap_fd            File descriptor we are writing to.
 * @param im_source_addr    Starting address in image memory.
 * @param dma_addr          DMA address of the bounce buffers we'll be using.
 * @param xfer_buffer       Pointers to the bounce buffers we'll be using.
 * @param total_length      Total number of bytes to transfer.
 *
 * @return Left over bytes, i.e. 0 if all bytes were transferred and written, otherwise the number of bytes remaining.
 */
uint64_t StreamSave::capMultiWrite(int cap_fd, uint64_t im_source_addr, std::vector<uint64_t>& dma_addr, std::vector<uint8_t *>& xfer_buffer, uint64_t total_length)
{
    int i;
    uint64_t xfer_length;                   // How much is left to transfer.
    std::mutex list_mutex;                  // Control access to all of the lists.
    std::queue<int>  avail_list;
    std::queue<int>  filled_list;
    std::vector<uint32_t> amt_to_write(dma_addr.size());    // Sent from DMA thread -> WRITER thread.

    TP(TRACE_SEL_CAP_FILE, "%s: save_offset=%" PRIu64, __PRETTY_FUNCTION__, save_offset);

    // Setup for multiple transfers (assuming partition size will be larger than MAX).
    xfer_length = total_length;

    // Set up our lists.
    for (i = 0; i < (int)dma_addr.size(); ++i)
        avail_list.push(i);

    // DMA THREAD ======================================
    // Lambda Tutorial:
    // * We are defining a "lambda function" (LF) called "dma_thread".
    // * We are "capturing" local variables for use inside the LF by
    //   specifying them in the square brackets. These are (by default)
    //   "const copy-by-value" (i.e. read-only).
    //   * By specifying the keyword "mutable", these captured values are
    //     now "copy-by-value", i.e. you can modify them locally.
    // * Add a reference indicator (&) to pass by reference and allow
    //   changes to proceed to the outer scope.
    std::thread dma_thread([this,
                            im_source_addr,
                            dma_addr,
                            &amt_to_write,
                            &xfer_length,
                            &list_mutex,
                            &avail_list,
                            &filled_list
                           ]()
                           mutable
    {
        int dma_index;
        uint32_t amt_to_dma;
        uint32_t dma_mod;
        FiFPGA *fpga_ptr;

        // Get a pointer to our FPGA class.
        fpga_ptr = CameraServer::get_fpga();
        if (fpga_ptr == nullptr)
        {
            fi_log(LOG_ERR, "%s: NULL FPGA pointer?!", __PRETTY_FUNCTION__);
            format_info.status = EFAULT;
            this->stop_request = true;
            return;
        }

        // We are going to transfer the data raw - no formatting.
        dma_mod = IM2PCIE_DMA_GENERIC_DATA;
        dma_mod |= IM2PCIE_DMA_HIGHSPEED_BYPASS;

        // Loop until we're done pulling all the data down from the camera.
        while (1)
        {
            uint32_t waits_for_buf;

            // Figure out how much we can DMA this time.
            if (xfer_length > (uint64_t)MAX_SINGLE_DMA)
                amt_to_dma = MAX_SINGLE_DMA;
            else
                amt_to_dma = (uint32_t)xfer_length;

            // Make sure there is a buffer available.
            waits_for_buf = 0;
            while (avail_list.empty())
            {
                // Wait for a buffer to become available.
                std::this_thread::sleep_for(milliseconds(1));

                // Count how many times we wait.
                ++waits_for_buf;

                // Bail if there is a stop request.
                if (this->stop_request)
                    break;
            }

            // Bail if there is a stop request.
            if (this->stop_request)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                break;
            }

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);

                // Get the next available buffer index.
                dma_index = avail_list.front();
                avail_list.pop();
            }
            // Release access to the lists.

            // Have the FPGA transfer the data to our transfer buffer.
            TP(TRACE_SEL_CAP_FILE, "%s [DMA]: waited=%d, buf=%d, im_addr=0x%" PRIx64 ", amt_to_dma=%u", __PRETTY_FUNCTION__, waits_for_buf, dma_index, im_source_addr, amt_to_dma);
            fpga_ptr->imageDownload(dma_mod, 0, im_source_addr, dma_addr[dma_index], amt_to_dma);

            // Let the writer know how much data is available.
            amt_to_write[dma_index] = amt_to_dma;

            // Update our image memory address.
            im_source_addr += amt_to_dma;

            // Update our total written count.
            xfer_length -= amt_to_dma;

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);
                filled_list.push(dma_index);
            }
            // Release access to the lists.

            // Are we done?
            if (xfer_length == 0)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [DMA] DONE", __PRETTY_FUNCTION__);
                break;  // Done.
            }

            // Did the user try to cancel?
            if (this->stop_request)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                break;
            }
        }
    });
    // DMA THREAD ======================================

    // WRITING THREAD ==================================
    std::thread writer_thread([
                                  this,
                                  cap_fd,
                                  xfer_buffer,
                                  total_length,
                                  &amt_to_write,
                                  &list_mutex,
                                  &avail_list,
                                  &filled_list]()
                              mutable
    {
        int write_index = 0;
        uint64_t xfer_count = 0;

        // Loop until we've written all the data out to the destination.
        while (1)
        {
            uint32_t buffer_size;
            ssize_t bytes_written;
            uint32_t waits_for_buf;

            // Make sure there is a buffer available.
            waits_for_buf = 0;
            while (filled_list.empty())
            {
                // Wait for a buffer to become available.
                std::this_thread::sleep_for(milliseconds(1));

                // Count how many times we wait.
                ++waits_for_buf;

                // Bail if there is a stop request.
                if (this->stop_request)
                    break;
            }

            // Bail if there is a stop request.
            if (this->stop_request)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [WRITER] STOP_REQUEST.", __PRETTY_FUNCTION__);
                break;
            }

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);

                // Get the next filled buffer index.
                write_index = filled_list.front();
                filled_list.pop();
            }
            // Release access to the lists.

            // Save the amount of data in a local variable.
            buffer_size = amt_to_write[write_index];
            amt_to_write[write_index] = 0;

            // Write the "chunk". We output data in "chunks" so we can provide
            // progress, and so the user can cancel the operation.
            TP(TRACE_SEL_CAP_FILE, "%s [WRITER]: waited=%d, buf=%d, size=%u", __PRETTY_FUNCTION__, waits_for_buf, write_index, buffer_size);
            bytes_written = write(cap_fd, xfer_buffer[write_index], buffer_size);
            if (bytes_written < 0)
            {
                // Log the error and bail.
                this->format_info.status = errno;
                LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
                break;
            }
            else if (bytes_written != (ssize_t)buffer_size)
            {
                fi_log(LOG_ERR, "%s [WRITER]: bytes_written (%u) != amt_to_write (%u)!", __PRETTY_FUNCTION__, (uint32_t)bytes_written, buffer_size);
            }

            // Update our offset into the total amount we need to write. This is used for, at a minimum, progress displays.
            // We also want to know how much we've written so we know when we're done.
            xfer_count += bytes_written;
            save_offset += bytes_written;
            TP(TRACE_SEL_CAP_FILE, "%s [WRITER]: save_offset=%" PRIu64 ", xfer_count=%" PRIu64 ", total_length=%" PRIu64, __PRETTY_FUNCTION__, save_offset, xfer_count, total_length);

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);
                avail_list.push(write_index);
            }
            // Release access to the lists.

            // Are we done? We know we're done when we've saved the total amount.
            if (xfer_count == total_length)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [WRITER] DONE [save_offset=%" PRIu64 "]", __PRETTY_FUNCTION__, save_offset);
                break;  // Done.
            }

            // Did the user try to cancel?
            if (this->stop_request)
            {
                TP(TRACE_SEL_CAP_FILE, "%s [WRITER] STOP_REQUEST.", __PRETTY_FUNCTION__);
                break;
            }
        }
    });
    // WRITING THREAD ==================================

    // Wait for everyone to finish.
    dma_thread.join();
    writer_thread.join();

    // Return the number of bytes left to write. Hopefully this will be 0.
    return xfer_length;
}

/*!
 * @brief Save a complete capture partition.
 * @return True on success, false otherwise.
 */
bool StreamSave::capFullSave()
{
    bool rc;
    int i;
    int err_code;
    int cap_fd = -1;

    PartMgr *partmgr_ptr;
    uint64_t    im_addr;
    uint64_t    xfer_length;
    uint64_t    image_length;
    uint32_t    session_length;
    uint32_t    capture_session;
    uint32_t    frame_count;
    std::size_t pos;
    std::string temp_str;
    std::vector<uint64_t> dma_bb_addr(DMA_BB_COUNT);
    std::vector<uint8_t *> xfer_buffer(DMA_BB_COUNT);

    TP(TRACE_SEL_CAP_FILE, "%s", __PRETTY_FUNCTION__);

    // Get our partition mananger pointer.
    partmgr_ptr = CameraServer::get_partmgr();
    if (partmgr_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL partmgr pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

    // Initialize our session length variable to indicate that
    // nothing has changed.
    session_length = 0;
    frame_count = 0;

    // Get a buffer for our image processing.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Split the one buffer up into our bounce buffers.
    dma_bb_addr[0] = bb_addr;
    for (i = 1; i < DMA_BB_COUNT; ++i)
        dma_bb_addr[i] = dma_bb_addr[i - 1] + MAX_SINGLE_DMA;

    // Map the memory.
    xfer_buffer[0] = (uint8_t *)bb_mmap(bb_index, TOTAL_DMA_BB_SIZE);
    if (xfer_buffer[0] == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        bb_release(bb_index);
        format_info.status = ENOMEM;
        return false;
    }

    // Get pointers to our buffers. We already have the first (at 0), so now
    // get the others.
    for (i = 1; i < DMA_BB_COUNT; ++i)
    {
        xfer_buffer[i] = xfer_buffer[i - 1] + MAX_SINGLE_DMA;
    }

    // Calculate the size of the video file we're going to create. Make sure we
    // incorporate the metadata, including the correction table, if necessary.
    image_length = partmgr_ptr->getReviewPartLength(&capture_session);

    // Can we shorten the length? Check to see if the partition has wrapped.
    if (capture_session < partmgr_ptr->getSessionLength())
    {
        // Save the frame count so we can restore that.
        frame_count = partmgr_ptr->getReviewPartNumFrames();

        // Save the old session length and change it to be
        // our new (shorter) value.
        session_length = partmgr_ptr->getSessionLength();
        partmgr_ptr->adjustReviewPartition(capture_session);
    }

    // Get the start address for the partition.
    im_addr = partmgr_ptr->getReviewPartStartAddr();

    // Save the total number of bytes we'll be transferring.
    format_info.total_bytes = image_length;

    // ALM_TODO [FAS_FS]
    // + No "target_path" if we're saving to a FILR.
    // + checkSpace() has already been done, maybe?
    // + No "open" necessary, maybe?
    // + No "close" necessary, maybe?
    // + Ditto for the partial save...

    // Strip off the filename so we're only checking space on the folder.
    temp_str.assign(format_info.target_path);
    pos = temp_str.find_last_of('/');
    LOG("%s: target='%s', pos=%u", __PRETTY_FUNCTION__, format_info.target_path.c_str(), (uint32_t)pos);

    // Make sure there's enough room.
    if ((err_code = StreamSave::checkSpace(image_length, temp_str.substr(0, pos).c_str())) != 0)
    {
        // Save the status and bail.
        fi_log(LOG_ERR, "%s: Error checking for space [%d] (need %" PRIu64 ").", __FUNCTION__, err_code, image_length);
        format_info.status = err_code;
        rc = false;
        goto full_bail1;
    }

    // Open our file.
    //
    // ALM_TODO [FAS_FS]
    // If we are saving to a FILR CAP/STOR, then we won't need this open, but we do need a
    // seek to the proper "LBA" to start the writing. Once we've seeked to the proper
    // location, the standard writes "should" work as expected.
    //
    // Since we are saving the full capture, we "might" be able to use P2P, which
    // "might" give us a performance boost. But then we have to worry about things
    // like: are we saving to the local drive, custom setup and save, FPGA wasn't
    // original designed for this, etc. For now, KISS and get it to work.
    cap_fd = open(StreamSave::dest_name.c_str(), O_CREAT | O_RDWR | O_LARGEFILE | O_TRUNC, 0666);
    if (cap_fd < 0)
    {
        format_info.status = errno;
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        rc = false;
        goto full_bail1;
    }

    // Turn off cacheing.
    posix_fadvise(cap_fd, 0, image_length, POSIX_FADV_DONTNEED);

#if 0
    // "Flesh-out" our file.
    TP(TRACE_SEL_CAP_FILE, "%s: Call fallocate()", __PRETTY_FUNCTION__);
    if ((fallocate(cap_fd, /*FALLOC_FL_NO_HIDE_STALE*/ 0, 0, image_length) == -1) && (errno == EOPNOTSUPP))
    {
        // This is a problem, but for now we'll just move on.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);

        // Later, we would want to do something like write zeros to the file.
        // write_zeros(cap_fd, image_length);
    }
#endif

    // Initialize our offset.
    save_offset = 0;

    // Read the image memory and write to the file.
    xfer_length = capMultiWrite(cap_fd, im_addr, dma_bb_addr, xfer_buffer, image_length);

    // Close the file.
    close(cap_fd);

    // Did the user try to cancel or was there an error?
    if (stop_request || (xfer_length > 0))
    {
        fi_log(LOG_DEBUG, "%s: Error [0x%x], removing '%s'", __PRETTY_FUNCTION__, format_info.status, StreamSave::dest_name.c_str());
        unlink(StreamSave::dest_name.c_str());

        rc = false;
        goto full_bail1;
    }

    // All is good.
    rc = true;
    save_offset = image_length;

full_bail1:
    // Do we need to adjust the session length?
    if (session_length)
    {
        // Restore the session length. Restoring the session
        // length whacks the frame count, so restore that too.
        partmgr_ptr->adjustReviewPartition(session_length, frame_count);
    }

    if (bb_index >= 0)
        bb_release(bb_index);

    return rc;
}

/*!
 * @brief Save a sub-partition, from format_info.current_frame to format_info.end_frame, inclusive.
 * @return True on success, false otherwise.
 */
bool StreamSave::capSubpartSave()
{
    PartMgr *partmgr_ptr;
    FiFPGA *fpga_ptr;

    bool rc = false;
    int err_code;
    int i;
    int cap_fd = -1;

    uint32_t frame_count;
    uint64_t im_addr;
    uint64_t im_start_addr;
    uint64_t im_stop_addr;
    uint64_t xfer_length;
    uint64_t image_length;
    pm_metadata *metadata_ptr;

    std::size_t pos;
    std::string temp_str;
    std::vector<uint64_t> dma_bb_addr(DMA_BB_COUNT);
    std::vector<uint8_t *> xfer_buffer(DMA_BB_COUNT);

    TP(TRACE_SEL_CAP_FILE, "%s", __PRETTY_FUNCTION__);

    // Get our partition mananger pointer.
    partmgr_ptr = CameraServer::get_partmgr();
    if (partmgr_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL partmgr pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

    // Get a pointer to our FPGA class.
    fpga_ptr = CameraServer::get_fpga();
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL FPGA pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

    // Initialize our session length variable to indicate that
    // nothing has changed.
    frame_count = format_info.end_frame - format_info.current_frame + 1;

    // Get a buffer for our image processing.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Split the one buffer up into our bounce buffers.
    dma_bb_addr[0] = bb_addr;
    for (i = 1; i < DMA_BB_COUNT; ++i)
        dma_bb_addr[i] = dma_bb_addr[i - 1] + MAX_SINGLE_DMA;

    // Map the memory.
    xfer_buffer[0] = (uint8_t *)bb_mmap(bb_index, TOTAL_DMA_BB_SIZE);
    if (xfer_buffer[0] == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        bb_release(bb_index);
        format_info.status = ENOMEM;
        return false;
    }

    // Get pointers to our buffers. We already have the first (at 0), so now
    // get the others.
    for (i = 1; i < DMA_BB_COUNT; ++i)
    {
        xfer_buffer[i] = xfer_buffer[i - 1] + MAX_SINGLE_DMA;
    }

    // Calculate the size of the video file we're going to create. Make sure we
    // incorporate the metadata, including the correction table, if necessary.
    image_length = partmgr_ptr->getPartitionSizeReserved() + (partmgr_ptr->getFrameSize() * frame_count);

    // Get the start address for the partition.
    im_addr = partmgr_ptr->getReviewPartStartAddr();

    // Initialize our offset.
    save_offset = 0;

    // Save the total number of bytes we'll be transferring.
    format_info.total_bytes = image_length;

    // Strip off the filename so we're only checking space on the folder.
    temp_str.assign(format_info.target_path);
    pos = temp_str.find_last_of('/');
    LOG("%s: target='%s', pos=%u", __PRETTY_FUNCTION__, format_info.target_path.c_str(), (uint32_t)pos);

    // Make sure there's enough room.
    if ((err_code = StreamSave::checkSpace(image_length, temp_str.substr(0, pos).c_str())) != 0)
    {
        // Save the status and bail.
        fi_log(LOG_ERR, "%s: Erorr checking for space [%d] (need %" PRIu64 ").", __FUNCTION__, err_code, image_length);
        format_info.status = err_code;
        rc = false;
        goto part_bail1;
    }

    // Open our file.
    cap_fd = open(StreamSave::dest_name.c_str(), O_CREAT | O_RDWR | O_LARGEFILE | O_TRUNC, 0666);
    if (cap_fd < 0)
    {
        format_info.status = errno;
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        rc = false;
        goto part_bail1;
    }

    // Turn off cacheing.
    posix_fadvise(cap_fd, 0, image_length, POSIX_FADV_DONTNEED);

    // Read the partition header (im_addr..(im_addr+partmgr_ptr->get_partition_size_reserved())
    fpga_ptr->imageDownload(IM2PCIE_DMA_GENERIC_DATA | IM2PCIE_DMA_HIGHSPEED_BYPASS, 0, im_addr, dma_bb_addr[0], partmgr_ptr->getPartitionSizeReserved());

    // Overlay our metadata structure on to the buffer we just wrote.
    metadata_ptr = (pm_metadata *)xfer_buffer[0];

    // Adjust some of the settings. Our frame range for the CAP file goes from 0 (format_info.current_frame)
    // to frame_count - 1 (format_info.end_frame).
    metadata_ptr->frame_count = frame_count;

    // Figure out how the trigger frame relates to our range of frames.
    if (partmgr_ptr->getReviewPartTrigger() < format_info.current_frame)
    {
        metadata_ptr->first_frame = format_info.current_frame - partmgr_ptr->getReviewPartTrigger();
        metadata_ptr->trig_frame = 0;
        metadata_ptr->skip_count = metadata_ptr->first_frame;
    }
    else
    {
        metadata_ptr->first_frame = 0;
        metadata_ptr->trig_frame = partmgr_ptr->getReviewPartTrigger() - format_info.current_frame;
        metadata_ptr->skip_count = 0;
    }

    // Write the partition header out to the file.
    save_offset = (uint64_t)write(cap_fd, xfer_buffer[0], partmgr_ptr->getPartitionSizeReserved());

    // Find the starting location (inlusive) based on the starting frame number.
    im_start_addr = partmgr_ptr->getPartitionFrameAddr(partmgr_ptr->getReviewPartition(), format_info.current_frame, false);

    // Find the ending location (not-inclusive) based on the ending frame number.
    im_stop_addr = partmgr_ptr->getPartitionFrameAddr(partmgr_ptr->getReviewPartition(), format_info.end_frame, false) + partmgr_ptr->getFrameSize();

    // Determine if we wrap. That is, is the ending location before the starting
    // location.
    if (im_stop_addr < im_start_addr)
    {
        // The recording does, in fact, wrap around the end of the partition.
        // This means we need to do the transfer in two steps. The first
        // step is to transfer from the starting location to the end of the partition.
        // The second step is to transfer from the start of the partition to the
        // ending location. This second step is the same for the "no-wrap" case.

        // Where does the partition end? Make sure this is in "frame size" increments.
        uint32_t last_frame = partmgr_ptr->get_im_buf_count() - 1;
        uint64_t part_len = partmgr_ptr->getPartitionFrameAddr(partmgr_ptr->getReviewPartition(), last_frame, true) + partmgr_ptr->getFrameSize();

        // Addjust to be the number of bytes we need to transfer between start and end.
        part_len = part_len - im_start_addr;
        TP(TRACE_SEL_CAP_FILE, "%s: wrapped, first transfer from 0x%" PRIx64", %" PRIu64 " bytes", __PRETTY_FUNCTION__, im_start_addr, part_len);

        // Transfer this data to the file.
        xfer_length = capMultiWrite(cap_fd, im_start_addr, dma_bb_addr, xfer_buffer, part_len);

        // Did the user try to cancel or was there an error?
        if (stop_request || (xfer_length > 0))
        {
            fi_log(LOG_DEBUG, "%s: Removing '%s'", __PRETTY_FUNCTION__, StreamSave::dest_name.c_str());
            unlink(StreamSave::dest_name.c_str());

            rc = false;
            goto part_bail1;
        }

        // Update our start address.
        im_start_addr = partmgr_ptr->getPartitionFrameAddr(partmgr_ptr->getReviewPartition(), 0, true);
    }

    // For the no-wrap case, or the second step of the wrap case, we need to copy from
    // a start address in image memory to the ending location. Remember that we may need
    // to trim the start and/or the end of the memory space before saving it out to
    // the destination.
    TP(TRACE_SEL_CAP_FILE, "%s: transfer from 0x%" PRIx64 ", %" PRIu64 " bytes", __PRETTY_FUNCTION__, im_start_addr, im_stop_addr - im_start_addr);
    xfer_length = capMultiWrite(cap_fd, im_start_addr, dma_bb_addr, xfer_buffer, im_stop_addr - im_start_addr);

    // Did the user try to cancel or was there an error?
    if (stop_request || (xfer_length > 0))
    {
        fi_log(LOG_DEBUG, "%s: Removing '%s'", __PRETTY_FUNCTION__, StreamSave::dest_name.c_str());
        unlink(StreamSave::dest_name.c_str());

        rc = false;
        goto part_bail1;
    }

    // All is good.
    rc = true;
    save_offset = image_length;

part_bail1:
    // Close the file.
    if (cap_fd != -1)
        close(cap_fd);

    // Release our bounce buffers.
    if (bb_index >= 0)
        bb_release(bb_index);

    return rc;
}

/*!
 * @brief Save a CAP file to the location specified in format_info.target_path.
 * @return True on success, false otherwise.
 */
bool StreamSave::capSave()
{
    bool rc = false;
    PartMgr *partmgr_ptr;

    TP(TRACE_SEL_CAP_FILE, "%s", __PRETTY_FUNCTION__);

    // Initialize our save offset.
    save_offset = 0;
    percent = 0;

    // Get our partition mananger pointer.
    partmgr_ptr = CameraServer::get_partmgr();
    if (partmgr_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL partmgr pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

    // Save our destination name.
    //
    // ALM_TODO [FAS_FS]
    // If we're saving to a FILR, then we need to do things differently. There won't be a
    // "path" per-se. We need to carry that information down through the "format_info" structure.
    StreamSave::dest_name = format_info.target_path;

    // Are we saving the entire capture?
    if ((format_info.current_frame == 0) && (format_info.end_frame == (partmgr_ptr->getReviewPartNumFrames() - 1)))
    {
        // Save the entire partition.
        rc = capFullSave();
    }
    else
    {
        // Save a sub-partition.
        rc = capSubpartSave();
    }

    if (!rc)
        fi_log(LOG_DEBUG, "%s: Save error=0x%x", __PRETTY_FUNCTION__, format_info.status);

    return rc;
}

/*!
 * @brief Return progress as a percentage.
 */
int StreamSave::capCompletionPercent()
{
    // Sanity check.
    if (format_info.total_bytes == 0)
    {
        // This is perfectly normal when we are first starting a save.
        return 0;
    }

    // Figure out percent complete.
    percent = (uint32_t)((save_offset * 100) / format_info.total_bytes);
    return percent;
}

/*!
 * @brief Update metadata-driven parameters after loading a CAP file.
 */
bool imgLoadWorker::updateAfterCapLoad()
{
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    PartMgr     *partmgr_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    std::unique_ptr<pm_metadata> meta_ptr;
    bool        rc = false;
    uint32_t    shutter_temp;
    uint32_t    wb_option;
    uint32_t    gamma_default;
    uint32_t    value;
    bool        updatedPartition;

    // TICKET #437:  When loading CAP files of different record settings the bug that
    // was fixed loads the metadata after the partition is updated via readActiveImMetadata.

    // Tell the partition manager to update its partition metadata from what was loaded.
    updatedPartition = partmgr_ptr->readActiveImMetadata();

    // Now that the partition was updated with the loaded CAP partition data
    // grab a copy of it for our processing here.
    meta_ptr = partmgr_ptr->getActivePartMetadata();

    // Make sure all is ok.
    if (!updatedPartition || (meta_ptr.get() == nullptr))
    {
        fi_log(LOG_ERR, "%s: Unable to read the metadata!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        goto bail;
    }

    LOG("%s: lut_id=%d corr_off=0x%x sov_off=0x%x", __PRETTY_FUNCTION__, meta_ptr->lut_id, meta_ptr->corr_off, meta_ptr->sov_off);

    // Adjust the shutter speed if necessary. Old files may have a value that is too long
    // dating back from when we lied about the shutter speed at 1X.
    shutter_temp = meta_ptr->shutter_speed;
    if (!cs_ptr->make_shutter_valid(meta_ptr->fps, &shutter_temp))
    {
        fi_log(LOG_ERR, "%s: adjusted restored shutter %u->%u fps=%u", __PRETTY_FUNCTION__,
               meta_ptr->shutter_speed, shutter_temp, meta_ptr->fps);
        meta_ptr->shutter_speed = shutter_temp;
    }

    // Make sure the sensor options are valid in case it is an old file.
    if (meta_ptr->sensor_opts == 0)
        meta_ptr->sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);

    // If the file is so old that sensor parameters are not stored, use current settings.
    if (meta_ptr->sensor_dac == 0)
        meta_ptr->sensor_dac = regs[GIGE_REG_SENSOR_DAC].get_value();
    if (meta_ptr->sensor_vrefcol == 0)
        meta_ptr->sensor_vrefcol = regs[GIGE_REG_SENSOR_VREFCOL].get_value();
    if (meta_ptr->sensor_gain == 0)
        meta_ptr->sensor_gain = regs[GIGE_REG_SENSOR_GAIN].get_value();

    // See if we can actually use this CAP file. Maybe we swapped sensor boards but
    // we don't want to deal with pm metadata that could have invalid information.
    // Or maybe during development we changed what options we support.
    if (meta_ptr->sensor_type != (uint32_t)sensor_ptr->get_sensor_type())
    {
        fi_log(LOG_ERR, "%s: Invalid CAP file sensor=0x%08x != 0x%08x (%u,%u):%ux%u bits=%u fps=%u",
               __PRETTY_FUNCTION__, meta_ptr->sensor_type, (uint32_t)sensor_ptr->get_sensor_type(),
               meta_ptr->roi_x, meta_ptr->roi_y, meta_ptr->roi_width, meta_ptr->roi_height,
               meta_ptr->bit_sel, meta_ptr->fps);
        format_info.status = EINVAL;

        // FIXME The partition should still be marked as EMPTY at this point, but we need
        // to reinitialize anything we overwrote, like the local partition metadata structure.
        goto bail;
    }
    if (!cs_ptr->is_user_roi_valid(meta_ptr->roi_x, meta_ptr->roi_y,
                                   meta_ptr->roi_width, meta_ptr->roi_height,
                                   (bitmode_t)meta_ptr->bit_sel, meta_ptr->fps,
                                   meta_ptr->shutter_speed, meta_ptr->sensor_opts, NULL))
    {
        fi_log(LOG_ERR, "%s: Invalid ROI sensor=0x%08x != 0x%08x (%u,%u):%ux%u bits=%u fps=%u",
               __PRETTY_FUNCTION__, meta_ptr->sensor_type, (uint32_t)sensor_ptr->get_sensor_type(),
               meta_ptr->roi_x, meta_ptr->roi_y, meta_ptr->roi_width, meta_ptr->roi_height,
               meta_ptr->bit_sel, meta_ptr->fps);
        format_info.status = EINVAL;

        // FIXME The partition should still be marked as EMPTY at this point, but we need
        // to reinitialize anything we overwrote, like the local partition metadata structure.
        goto bail;
    }

    // Update the FCP register database.
    regs[ GIGE_REG_BIT_MODE ].set_value(meta_ptr->bit_sel);
    regs[ GIGE_REG_OFFSET_X ].set_value(meta_ptr->roi_x);
    regs[ GIGE_REG_OFFSET_Y ].set_value(meta_ptr->roi_y);
    regs[ GIGE_REG_WIDTH ].set_value(meta_ptr->roi_width);
    regs[ GIGE_REG_HEIGHT ].set_value(meta_ptr->roi_height);
    regs[ GIGE_REG_FRAME_RATE ].set_value(meta_ptr->fps);
    regs[ GIGE_REG_SHUTTER_SPEED ].set_value(meta_ptr->shutter_speed);
    regs[ GIGE_REG_FPN ].set_value(meta_ptr->fpn_status);
    regs[ GIGE_REG_SENSOR_OPTIONS ].set_value(meta_ptr->sensor_opts);
    regs[ GIGE_REG_SENSOR_GAIN ].set_value(meta_ptr->sensor_gain);
    regs[ GIGE_REG_SENSOR_DAC ].set_value(meta_ptr->sensor_dac);
    regs[ GIGE_REG_SENSOR_VREFCOL ].set_value(meta_ptr->sensor_vrefcol);

    // Update the partition information from what we loaded.
    // Should we reserve space for the gain table?
    partmgr_ptr->reserveGainTableSpace(meta_ptr->gain_off != 0);

    // This will lay out memory again but only the ROI dependencies will change.
    partmgr_ptr->useLoadedPartitionMetadata(meta_ptr.get());

    // Update the color processing options.
    wb_option = meta_ptr->white_bal;

    fcp_set_wb_preset(wb_option);
    regs[ GIGE_REG_WHITE_BAL ].set_value(wb_option);
    if (wb_option == GIGE_WHITE_BAL_CUSTOM)
    {
        fpga_ptr->ipmSetWbRed(meta_ptr->wb_red);
        regs[ GIGE_REG_WHITE_BALANCE_RED ].set_value(meta_ptr->wb_red);
        fpga_ptr->ipmSetWbGreen(meta_ptr->wb_grn);
        regs[ GIGE_REG_WHITE_BALANCE_GRN ].set_value(meta_ptr->wb_grn);
        fpga_ptr->ipmSetWbBlue(meta_ptr->wb_blu);
        regs[ GIGE_REG_WHITE_BALANCE_BLU ].set_value(meta_ptr->wb_blu);

        cs_ptr->set_as_shot_red(meta_ptr->as_shot_red);
        cs_ptr->set_as_shot_grn(meta_ptr->as_shot_grn);
        cs_ptr->set_as_shot_blu(meta_ptr->as_shot_blu);
    }

    // As of 2.2.2, RGB gains are 16 bits to allow for their 8.8 format.
    regs[ GIGE_REG_RED ].set_value(meta_ptr->red);
    fcp_color_update(nullptr, meta_ptr->red);
    regs[ GIGE_REG_GREEN ].set_value(meta_ptr->green);
    fcp_color_update(nullptr, meta_ptr->green);
    regs[ GIGE_REG_BLUE ].set_value(meta_ptr->blue);
    fcp_color_update(nullptr, meta_ptr->blue);

    value = meta_ptr->brightness;
    if (value <= GIGE_MAX_BRIGHTNESS)
    {
        regs[ GIGE_REG_BRIGHTNESS ].set_value(value);
        fcp_color_update(nullptr, value);
    }
    value = meta_ptr->contrast;
    if (value <= GIGE_MAX_CONTRAST)
    {
        regs[ GIGE_REG_CONTRAST ].set_value(value);
        fcp_color_update(nullptr, value);
    }
    // As of 2.2.2, gamma is 16 bits to allow for values > 2.55, which the UI allows.
    value = meta_ptr->gamma;
    if ((value >= GIGE_MIN_GAMMA) && (value <= GIGE_MAX_GAMMA))
    {
        gamma_default = fpga_ptr->ipmDefaultGamma();
        value |= (gamma_default << 16);
        regs[ GIGE_REG_GAMMA ].set_value(value);
        fcp_color_update(nullptr, value);
    }

    // Make sure the correction table is propagated to all the partitions.
    // This is our convention, to load the same CT into all partitions so if we
    // did decide to save a CAP file, it is all ready to go.
    if ((partmgr_ptr->getNumPartitions() > 1) &&
        ((meta_ptr->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) != 0))
        FiImageCorrection::fpnPropagate();

    // We didn't really "load" the CT table so our information about what CT is loaded
    // is not accurate. But we can now switch to live video and continue using the
    // black frame we loaded from the CAP file so we need to be prepared for that.
    if ((meta_ptr->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        FiImageCorrection::fpnApplyOffset((bitmode_t)meta_ptr->bit_sel, meta_ptr->black_offset);
    else
        FiImageCorrection::fpnResetOffset();

    // Set up the hardware to use the loaded ROI.
    cs_ptr->setup_capture_from_regs();

    // Clear the comment field so they need to start fresh.
    partmgr_ptr->set_comment("");

    // Load the LUT, if one is specified.
    if (meta_ptr->lut_id)
    {
        if (!fpga_ptr->ipmLoadUserLut(meta_ptr->lut_id))
        {
            // We were unable to load the LUT. Indicate that
            // no LUT is selected.
            fi_log(LOG_WARNING, "%s: Unable to load LUT (%d).", __FUNCTION__, meta_ptr->lut_id);
            meta_ptr->lut_id = GIGE_LUT_NONE;
        }
    }

    // Remember the LUT selected.
    regs[ GIGE_REG_LUT ].set_value(meta_ptr->lut_id);

    // Notify folks that things have changed.
    fcp_notify_apps(GIGE_REG_OFFSET_X, meta_ptr->roi_x);
    fcp_notify_apps(GIGE_REG_OFFSET_Y, meta_ptr->roi_y);
    fcp_notify_apps(GIGE_REG_WIDTH, meta_ptr->roi_width);
    fcp_notify_apps(GIGE_REG_HEIGHT, meta_ptr->roi_height);
    fcp_notify_apps(GIGE_REG_FRAME_RATE, meta_ptr->fps);
    fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, meta_ptr->shutter_speed);
    fcp_notify_apps(GIGE_REG_FPN, meta_ptr->fpn_status);
    fcp_notify_apps(GIGE_REG_SENSOR_OPTIONS, meta_ptr->sensor_opts);
    fcp_notify_apps(GIGE_REG_SENSOR_GAIN, meta_ptr->sensor_gain);
    fcp_notify_apps(GIGE_REG_BIT_MODE, meta_ptr->bit_sel);

    // Notify about changes in the partition.
    fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, partmgr_ptr->getNumFrames());
    fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, partmgr_ptr->getTriggerFrame());
    fcp_notify_apps(GIGE_REG_PART_STATE, partmgr_ptr->getPartitionState());

    // Notify about color changes.
    fcp_notify_apps(GIGE_REG_WHITE_BAL, wb_option);
    if (wb_option == GIGE_WHITE_BAL_CUSTOM)
    {
        fcp_notify_apps(GIGE_REG_WHITE_BALANCE_RED, meta_ptr->wb_red);
        fcp_notify_apps(GIGE_REG_WHITE_BALANCE_GRN, meta_ptr->wb_grn);
        fcp_notify_apps(GIGE_REG_WHITE_BALANCE_BLU, meta_ptr->wb_blu);
    }
    // These values were not usable before 2.2.2.
    fcp_notify_apps(GIGE_REG_RED, meta_ptr->red);
    fcp_notify_apps(GIGE_REG_GREEN, meta_ptr->green);
    fcp_notify_apps(GIGE_REG_BLUE, meta_ptr->blue);

    value = regs[ GIGE_REG_BRIGHTNESS ].get_value();
    fcp_notify_apps(GIGE_REG_BRIGHTNESS, value);
    value = regs[ GIGE_REG_CONTRAST ].get_value();
    fcp_notify_apps(GIGE_REG_CONTRAST, value);
    value = regs[ GIGE_REG_GAMMA ].get_value();
    fcp_notify_apps(GIGE_REG_GAMMA, value);
    value = regs[ GIGE_REG_LUT ].get_value();
    fcp_notify_apps(GIGE_REG_LUT, value);

    rc = true;

bail:
    return rc;
}

bool imgLoadWorker::loadFromMedia()
{
    bool    rc = false;
    bool    reader_rc = false;
    bool    dma_rc = false;
    bool    session_length_changed = false;
    int     i;
    int     bb_index;
    int     cap_fd = -1;
    PartMgr *partmgr_ptr;
    FiFPGA  *fpga_ptr;
    uint64_t    bb_addr;
    uint64_t    im_addr;
    uint8_t *xfer_buffer[DMA_BB_COUNT];
    uint64_t    dma_bb_addr[DMA_BB_COUNT];
    uint64_t    xfer_length;
    uint32_t    session_length;
    uint64_t    file_size;
    uint64_t    part_size;
    struct stat64 sb;
    std::queue<int>  avail_list;
    std::queue<int>  filled_list;
    std::string dest_name;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_CAP_FILE, "%s", __PRETTY_FUNCTION__);

    // Get our partition mananger pointer.
    partmgr_ptr = CameraServer::get_partmgr();
    if (partmgr_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: NULL partmgr pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

    // Get a pointer to our FPGA class.
    fpga_ptr = CameraServer::get_fpga();
    if (fpga_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: NULL FPGA pointer?!", __PRETTY_FUNCTION__);
        format_info.status = EFAULT;
        return false;
    }

#if 0
    // Put the camera into an idle state (if we're not already there).
    // This will freeze the video display and alleviate memory bandwidth use.
    save_state = CameraServer::get_camera_state();
    if (save_state != GIGE_CAMERA_STATE_IDLE)
    {
        fi_log(LOG_DEBUG, "%s: Need to change states to IDLE", __PRETTY_FUNCTION__);
        CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_IDLE, true, nullptr);
    }
#endif

    // Initialize our session length variable to indicate that
    // nothing has changed.
    session_length = 0;

    // Save our destination name.
    dest_name = format_info.target_path;

    // Do a sanity check on the file.
    if (!capIsOk(dest_name.c_str()))
    {
        fi_log(LOG_ERR, "%s: CAP file is not safe to load!", __PRETTY_FUNCTION__);
        format_info.status = ENODEV;
        return false;
    }

    // How big is the file?
    if (stat64(dest_name.c_str(), &sb) != 0)
    {
        fi_log(LOG_ERR, "%s: stat(%s) fail err=%d(%s)", __PRETTY_FUNCTION__, dest_name.c_str(), errno,
               strerror(errno));
        format_info.status = errno;
        return false;
    }
    file_size = sb.st_size;

    // Get a buffer for our image processing.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Split the one buffer up into our bounce buffers.
    dma_bb_addr[0] = bb_addr;
    for (i = 1; i < DMA_BB_COUNT; ++i)
        dma_bb_addr[i] = dma_bb_addr[i - 1] + MAX_SINGLE_DMA;

    // Map the memory.
    xfer_buffer[0] = (uint8_t *)bb_mmap(bb_index, TOTAL_DMA_BB_SIZE);
    if (xfer_buffer[0] == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        bb_release(bb_index);
        format_info.status = ENOMEM;
        return false;
    }

    // Get pointers to our buffers. We already have the first (at 0), so now
    // get the others.
    for (i = 1; i < DMA_BB_COUNT; ++i)
        xfer_buffer[i] = xfer_buffer[i - 1] + MAX_SINGLE_DMA;

    // Set up our lists.
    for (i = 0; i < DMA_BB_COUNT; ++i)
        avail_list.push(i);

    // Setup for multiple transfers (assuming partition size will be larger than MAX).
    xfer_length = file_size;

    // Save the total number of bytes we'll be transferring.
    format_info.total_bytes = xfer_length;

    // Disable FPN because we will be overwriting the correction table.
    if (FiImageCorrection::fpnIsEnabled())
    {
        fi_log(LOG_DEBUG, "%s: disable FPN", __PRETTY_FUNCTION__);
        FiImageCorrection::fpnDisable();
    }

    // Determine if there's enough room in the partition.
    part_size = partmgr_ptr->getPartitionSize();
    if (format_info.total_bytes != part_size)
    {
        session_length_changed = true;

        TP(TRACE_SEL_CAP_FILE, "%s: Need to adjust partition size [0x%" PRIx64 " != 0x%" PRIx64, __PRETTY_FUNCTION__, file_size, part_size);

        // Convert to session length.
        session_length = format_info.total_bytes / PART_MGR::SESSION_INCR;
        if (format_info.total_bytes > ((uint64_t)session_length * (uint64_t)PART_MGR::SESSION_INCR))
            ++session_length;

        // Make sure the value is valid.
        if ((session_length == 0) || (session_length > partmgr_ptr->getSessionLengthMax()))
        {
            fi_log(LOG_ERR, "%s: Cannot change partition size (0x%" PRIx64 ") to image size (0x%" PRIx64 ")",
                   __PRETTY_FUNCTION__, part_size, format_info.total_bytes);
            format_info.status = ENOMEM;
            rc = false;
            goto load_bail1;
        }

        // This will let us get the partition start address but the ROI will still be wrong
        // so we'll have to do it again later. The function still does all the original work
        // of both creating the partition layout and setting it up based on ROI.
        TP(TRACE_SEL_CAP_FILE, "%s: new session_length=%d", __PRETTY_FUNCTION__, session_length);
        regs[ GIGE_REG_SESSION_LENGTH ].set_value(session_length);
        partmgr_ptr->reserveGainTableSpace(file_pm_data.gain_off != 0);
        partmgr_ptr->layoutMemoryFromRegs();
    }

    // Get the start address for the partition.
    im_addr = partmgr_ptr->getPartitionStart();

    // Open our file.
    cap_fd = open(dest_name.c_str(), O_RDONLY | O_LARGEFILE);
    if (cap_fd < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        format_info.status = errno;
        rc = false;
        goto load_bail1;
    }

    // Turn off cacheing.
    posix_fadvise(cap_fd, 0, xfer_length, POSIX_FADV_DONTNEED);

//#define ALM_DEBUG_CAP_LOAD 1
#ifdef ALM_DEBUG_CAP_LOAD
    {
        int retries;
        uint32_t reg_value;

        // --------------------------------------------------------------------------
        // Make sure we have a known value in image memory. This code will use the
        // AUX DMA to fill image memory with zeros.
        // --------------------------------------------------------------------------

        fi_log(LOG_INFO, "%s: Initializing the buffers via AUXDMA to 0.", __PRETTY_FUNCTION__);

        // Make sure the AUX DMA is available.
        for (retries = 0; retries < 10; retries++)
        {
            fpga_ptr->read_reg(IM_AUX_WR_DMA0_STATUS, &reg_value);
            if ((reg_value & AUX_DMA_BUSY) == 0)
                break;

            usleep(100000);
        }

        // Did everything go OK?
        if (retries < 10)
        {
            // Setup the AUX DMA engine to blast the buffer with 0's.
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_CONTROL, AUX_DMA_ENABLE);
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_START_ADDR, convert_to_HEXWORD(im_addr));
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_BLOCK_SIZE, convert_to_HEXWORD(xfer_length));
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_BURST_SIZE, AUX_WR_DMA_DDR2_REQ_SIZE_VAL);
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_BURST_INTERVAL, 0);
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_BP_THRESHOLDS, AUX_WR_DMA_BP_THRESHOLD_VAL);
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_PATTERN_VAL, 0x00);
            fpga_ptr->write_reg(IM_AUX_WR_DMA0_CONTROL, AUX_DMA_ENABLE | AUX_DMA_START | AUX_DMA_PATTERN_VAL);
        }
        else
        {
            // Issue a warning - either way, we're still going to setup the buffer address.
            fi_log(LOG_WARNING, "%s: AUX_DMA is stuck busy?! [0x%x]", __FUNCTION__, reg_value);
        }
    }
#endif

    // Once load_offset is initialized, it is safe to start reporting the percentage.
    load_offset = 0;
    load_progress_valid = true;

    {
        uint32_t amt_to_dma[DMA_BB_COUNT];      // Sent from READER thread -> DMA thread.
        std::mutex list_mutex;                  // Control access to all of the lists.

        // READER THREAD ======================================
        // Lambda Tutorial:
        // * We are defining a "lambda function, LF" "dma_threa".
        // * We are "capturing" local variables for use inside the LF by
        //   specifying them in the square brackets. These are (by default)
        //   "const copy-by-value" (i.e. read-only).
        //   * By specifying the keyword "mutable", these captured values are
        //     now "copy-by-value", i.e. you can modify them locally.
        // * Add a reference indicator (&) to pass by reference and allow
        //   changes to proceed to the outer scope.
        std::thread reader_thread([
                                      this,
                                      cap_fd,
                                      file_size,
                                      xfer_buffer,
                                      xfer_length,
                                      &amt_to_dma,
                                      &list_mutex,
                                      &avail_list,
                                      &reader_rc,
                                      &filled_list]()
                                  mutable
        {
            int read_index = 0;
            uint32_t amt_to_read;

            // Transfer the capture data to the drive.
            while (xfer_length > 0)
            {
                ssize_t bytes_read;
                uint32_t waits_for_buf;

                // Figure out how much we can read this time.
                if (xfer_length > (uint64_t)MAX_SINGLE_READ)
                    amt_to_read = MAX_SINGLE_READ;
                else
                    amt_to_read = (uint32_t)xfer_length;

                // Make sure there is a buffer available.
                waits_for_buf = 0;
                while (avail_list.empty())
                {
                    // Wait for a buffer to become available.
                    std::this_thread::sleep_for(milliseconds(1));

                    ++waits_for_buf;

                    // Bail if there is a stop request.
                    if (this->stop_request)
                        break;
                }

                // Bail if there is a stop request.
                if (this->stop_request)
                {
                    TP(TRACE_SEL_CAP_FILE, "%s [READER] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    break;
                }

                // Lock access to the lists.
                {
                    std::unique_lock<std::mutex> locker(list_mutex);

                    // Get the next available buffer index.
                    read_index = avail_list.front();
                    avail_list.pop();
                }
                // Release access to the lists.

                // Read the "chunk". We input data in "chunks" so we can provide
                // progress, and so the user can cancel the operation.
                bytes_read = read(cap_fd, xfer_buffer[read_index], amt_to_read);
                if (bytes_read < 0)
                {
                    // Log the error and bail.
                    LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
                    this->format_info.status = errno;
                    break;
                }
                else if (bytes_read != amt_to_read)
                {
                    fi_log(LOG_DEBUG, "%s: bytes_read (%u) != amt_to_read (%u)!", __PRETTY_FUNCTION__, bytes_read, amt_to_read);
                }

#ifdef ALM_DEBUG_CAP_LOAD
                // --------------------------------------------------------------------------
                // This is test code to add "bands" to the beginning and end of the MAX_SINGLE_READ
                // "chunks" of data being DMA'ed into image memory.
                // --------------------------------------------------------------------------

                // Is this the first "chunk"? The xfer_length will match the file_size if it is,
                // otherwise the xfer_length should be getting smaller.
                if (xfer_length != file_size)
                {
                    // Cause an "easily" visible artifact in the image.
                    if (bytes_read > (2 * (128 * 1024)))
                    {
                        memset(xfer_buffer[read_index], 0x55, (128 * 1024));
                        memset(xfer_buffer[read_index] + ((bytes_read - (128 * 1024)) - 1), 0xAA, (128 * 1024));
                    }
                }
                else
                {
                    // For the first "chunk", we do not want to overwrite the partition metadata, so only
                    // put the band at the end.
                    memset(xfer_buffer[read_index] + ((bytes_read - (128 * 1024)) - 1), 0xAA, (128 * 1024));
                }
#endif

                // Did the user try to cancel?
                if (stop_request)
                {
                    TP(TRACE_SEL_CAP_FILE, "%s [READER] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    break;
                }

                // Let the DMA engine know how much data is available.
                amt_to_dma[read_index] = bytes_read;

                // Lock access to the lists.
                {
                    std::unique_lock<std::mutex> locker(list_mutex);
                    filled_list.push(read_index);
                }
                // Release access to the lists.

                // Update our total read count.
                xfer_length -= bytes_read;
                TP(TRACE_SEL_CAP_FILE, "%s [READER]: bytes_read=%u, remaining=%u", __PRETTY_FUNCTION__, bytes_read, xfer_length);

                // Update the progress indicator.
                load_offset = (100 * (file_size - xfer_length)) / file_size;

                // Don't report 100% until we are done done.
                if (load_offset > 99)
                    load_offset = 99;
            }

            // Indicate success.
            reader_rc = (xfer_length == 0);
        });
        // READER THREAD ======================================

        // DMA THREAD ======================================
        std::thread dma_thread([this,
                                im_addr,
                                dma_bb_addr,
                                xfer_length,
                                &amt_to_dma,
                                &dma_rc,
                                &list_mutex,
                                &avail_list,
                                &filled_list
                               ]()
                               mutable
        {
            int dma_index;
            uint32_t dma_mod;
            FiFPGA *fpga_ptr;

            // Get a pointer to our FPGA class.
            fpga_ptr = CameraServer::get_fpga();
            if (fpga_ptr == nullptr)
            {
                fi_log(LOG_ERR, "%s: NULL FPGA pointer?!", __PRETTY_FUNCTION__);
                this->format_info.status = EFAULT;
                this->stop_request = true;
                return;
            }

            // We are going to transfer the data raw.
            dma_mod = PCIE2IM_DMA_BYTE_ORDER_R2L | PCIE2IM_DMA_WORD_ORDER_R2L;
            // dma_mod |= IM2PCIE_DMA_HIGHSPEED_BYPASS;

            // Loop until we're done pulling all the data down from the camera.
            while (xfer_length > 0)
            {
                uint32_t waits_for_buf;
                uint32_t dma_length;

                // Make sure there is a buffer available.
                waits_for_buf = 0;
                while (filled_list.empty())
                {
                    // Wait for a buffer to become available.
                    std::this_thread::sleep_for(milliseconds(1));

                    ++waits_for_buf;

                    // Bail if there is a stop request.
                    if (this->stop_request)
                        break;
                }

                // Bail if there is a stop request.
                if (this->stop_request)
                {
                    TP(TRACE_SEL_CAP_FILE, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    break;
                }

                // Lock access to the lists.
                {
                    std::unique_lock<std::mutex> locker(list_mutex);

                    // Get the next filled buffer index.
                    dma_index = filled_list.front();
                    filled_list.pop();
                }
                // Release access to the lists.

                // Find out how much we need to DMA.
                dma_length = amt_to_dma[dma_index];
                amt_to_dma[dma_index] = 0;

                // Have the FPGA transfer the data to image memory.
                TP(TRACE_SEL_CAP_FILE, "%s [DMA] im_addr=0x%" PRIx64 ", xfer amount=%u", __PRETTY_FUNCTION__, im_addr, dma_length);
                fpga_ptr->imageUpload(dma_mod, dma_bb_addr[dma_index], im_addr, dma_length);

                // Lock access to the lists.
                {
                    std::unique_lock<std::mutex> locker(list_mutex);
                    avail_list.push(dma_index);
                }
                // Release access to the lists.

                // Update our image memory address.
                im_addr += dma_length;
                xfer_length -= dma_length;
            }

            // Indicate success.
            dma_rc = (xfer_length == 0);
        });
        // DMA THREAD ======================================

        dma_thread.join();
        reader_thread.join();
    }

    // Close the file.
    close(cap_fd);

    // Did the user try to cancel or was there an error?
    if (stop_request || !reader_rc || !dma_rc)
    {
        fi_log(LOG_DEBUG, "%s: stop_request=%d, reader_rc=%d, dma_rc=%d", __PRETTY_FUNCTION__, stop_request, reader_rc, dma_rc);
        rc = false;
        goto load_bail1;
    }

    // Update registers and hardware with what we loaded from the CAP file.
    TP(TRACE_SEL_CAP_FILE, "%s: update_after_cap_load()", __PRETTY_FUNCTION__);
    if (!updateAfterCapLoad())
        goto load_bail1;

    // Put the camera into a Review state. No need to wait for it to go into effect.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_REVIEW, false, nullptr);

    // Update our percentage.
    load_offset = 100;

    // All is good.
    rc = true;

load_bail1:
    // Let folks know that the session length changed.
    if (session_length_changed)
        fcp_notify_apps(GIGE_REG_SESSION_LENGTH, session_length);

    // Release our buffer.
    if (bb_index >= 0)
        bb_release(bb_index);

    // If it didn't work, return to the state we were in when we started entry.
    if (!rc)
    {
        static int async_done;
        CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO, true, &async_done);
    }

    // Return our status.
    return rc;
}

bool imgLoadWorker::capIsOk(const char *cap_file)
{
    bool rc = false;
    int cap_fd = -1;
    ssize_t bytes_read;

    // Open our file.
    cap_fd = open(cap_file, O_RDONLY | O_LARGEFILE);
    if (cap_fd < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto check_bail;
    }

    bytes_read = read(cap_fd, &file_pm_data, sizeof(file_pm_data));
    if (bytes_read < 0)
    {
        // Log the error and bail.
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto check_bail;
    }
    else if (bytes_read != sizeof(file_pm_data))
    {
        fi_log(LOG_DEBUG, "%s: bytes_read (%u) != amt_to_read (%u)!", __PRETTY_FUNCTION__, bytes_read, sizeof(file_pm_data));
        goto check_bail;
    }

    // The sensor types need to match.
    if (file_pm_data.sensor_type != Sensor::get_sensor_type())
    {
        fi_log(LOG_DEBUG, "%s: CAP sensor_type (0x%08x) != hardware (0x%08x)!", __PRETTY_FUNCTION__, file_pm_data.sensor_type, Sensor::get_sensor_type());
        goto check_bail;
    }

    TP(TRACE_SEL_CAP_FILE, "%s: Sensor types are a match!", __PRETTY_FUNCTION__);

    // Everything looks good.
    rc = true;

check_bail:
    // Close the file.
    close(cap_fd);

    // Return the result.
    return rc;
}

/*!
 * @brief This is the entry point for loading a partition from SSD to the
 *    image memory.
 *
 * The caller uses the returned boolean to determine if the transfer is still
 * in-progress. When the call returns a "false", then the caller needs to check
 * the information structure's status field to determine if an error occurred.
 *
 * @return true to indicate a successful load, false for an error.
 */
bool imgLoadWorker::capfmtImageLoad()
{
    TP(TRACE_SEL_CAP_FILE, "%s: format_info.target_path='%s'", __PRETTY_FUNCTION__, format_info.target_path.c_str());

    // Initialize our load offset.
    load_offset = 0;
    percent = 0;

    // Load the capture file into memory.
    return loadFromMedia();
}

/*!
 * @brief Reset progess before starting load of a CAP file
 *
 * We could be queried for completion progress before the worker thread
 * has even started. This must be called before starting the thread, but
 * AFTER verifying that it is not already running, so there is no chance
 * of conflict with that thread's progress management.
 */
void imgLoadWorker::capfmtResetProgress()
{
    load_progress_valid = false;
}

/*!
 * @brief Return our completion as a percentage.
 */
uint32_t imgLoadWorker::getCompletion()
{
    uint32_t    rc;

    if (last_status == 0)
    {
        if (load_progress_valid)
            rc = load_offset;
        else
            rc = 0;
    }
    else
    {
        // Must be an error - set the error flag.
        rc = last_status | GIGE_LOAD_ERROR;
    }

    return rc;
}
