/*!
 * @file image_io.cxx
 *
 * @brief This file contains the implementation of the image I/O
 *    class(es).
 *
 * This file contains all the routines to handle saving an image
 * to one of the local storage medias.
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>

#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>

#include <dirent.h>
#include <fcntl.h>
#include <fstream>
#include <iostream>
#include <memory.h>
#include <filesystem>
#include <sstream>
#include <string>
#include <fmt/printf.h>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>
#include <fastec/fas_video.h>

#include "autosave.h"
#include "camera_server.h"
#include "cs_network.h"
#include "cs_trace.h"
#include "cs_utils.h"
#include "fcp.h"
#include "fi_internal.h"
#include "fpga.h"
#include "image_io.h"
#include "image_format.h"
#include "jpeg_proc.h"
#include "part_mgr.h"
#include "pm_worker.h"
#include "storage.h"

namespace IMAGE_IO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Number of milliseconds to wait for the worker thread to signal it's running.
const int STARTUP_DELAY = 1000;

//! "Nice" the worker priority so that we don't bog down the entire system.
const int WORKER_PRIORITY = 10;

//! @name Strings used as suffixes for file and folder names.
//! @{
const char AVI_SUFFIX[]   = "avi";        //!< AVI folder & suffix.
const char BMP_SUFFIX[]   = "bmp";        //!< BMP folder & suffix.
const char BPP_SUFFIX[]   = "bpp";        //!< BPP folder & suffix.
const char CAP_SUFFIX[]   = "cap";        //!< CAP suffix.
const char DNG_SUFFIX[]   = "dng";        //!< DNG folder & suffix.
const char JPEG_SUFFIX[]  = "jpg";        //!< JPEG folder & suffix.
const char TIF_SUFFIX[]   = "tif";        //!< TIF folder & suffix.
const char MP4_SUFFIX[]   = "mp4";        //!< H264/MP4 suffix.
const char RDI_SUFFIX[]   = "rdi";        //!< RDI packed metadata file suffix.
// <space required by Doxygen>
//! @}

//! @name Strings used to describe the codec being used
//! @sa imgio::get_codec_str()
//! @{
const char AVI_JPEG_CODEC_STR[]   = "MJPEG";
const char AVI_BMP_CODEC_STR[]    = "DIB";
const char H264_CODEC_STR[]   = "H264";
const char BMP_CODEC_STR[]    = "BMP";
const char BPP_CODEC_STR[]    = "BPP";
const char DNG_CODEC_STR[]    = "DNG";
const char JPEG_CODEC_STR[]   = "JPEG";
const char RAW_CODEC_STR[]    = "RAW";
const char TIFF_CODEC_STR[]   = "TIFF";
const char RDI_CODEC_STR[]    = "PACK";
const char UNKNOWN_CODEC_STR[]    = "Unknown";
// <space required by Doxygen>
//! @}

//! @name Subdirectories created/populated by the camera.
//! @{
const char STILL_FOLDER[]     = "/DCIM/100fastc/";
const char VIDEO_FOLDER[]     = "/hs_video/";
const char CAPTURE_FOLDER[] = "/captures/";
// <space required by Doxygen>
//! @}

//! Standard name prefix we use if the user doesn't supply something else.
const char NAME_PREFIX[]      = "fi_";

//! A temporary file to use when we're saving a snapshot.
const char *STILL_NAME       = "/tmp/still.jpg";

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

//! String we'll use to prefix our file(s) with (imgSaveWorker::build_target()).
std::string file_prefix;

//--------------------------------------------------------------------
// Local function prototypes
//--------------------------------------------------------------------
static int find_next_index(const char *target_path);
static void get_destination_root(uint32_t dest_info, std::string& mount_dir);
static int select_name(const struct dirent *entry_ptr);
static void set_default_tag(char *buf, int max_len);

}   // End of namespace IMAGE_IO

TRACE_GLOBAL;
using namespace CS_TRACE;
using namespace IMAGE_IO;

//--------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------
std::unique_ptr<imgLoadWorker> imgLoadWorker::instance;
std::mutex imgLoadWorker::instance_mutex;

std::unique_ptr<imgSaveWorker> imgSaveWorker::instance;
std::mutex imgSaveWorker::instance_mutex;

const std::string imgSaveWorker::FFLR_METAFILE_NAME("metamap.txt");
const std::string imgSaveWorker::SAVE_SETTINGS_FILE(".config/Fastec/scratch/save_settings.json");

//--------------------------------------------------------------------
// Local function implementations
//--------------------------------------------------------------------

/*!
 * @brief Create a default tag string with the current date.
 *
 * If the destination flags include GIGE_IMAGE_USE_TAG, but the tag is
 * a null string, we fill in the tag with the current date.
 *
 * @param   buf     Where to write the tag
 * @param   max_len Maximum characters to write, including null terminator.
 */
static void IMAGE_IO::set_default_tag(char *buf, int max_len)
{
    time_t      time_now;

    time_now = time(NULL);
    strftime(buf, max_len, "%F", localtime(&time_now));
}

/*!
 * @brief Using the dest_info parameter, derive the folder that
 * device is mounted at.
 *
 * @param dest_info     Which device we're interested in (GIGE_IMAGE_DEST_SDCARD, GIGE_IMAGE_DEST_USB, GIGE_IMAGE_DEST_SSD).
 * @param mount_dir     Folder path where the device is mounted.
 */
static void IMAGE_IO::get_destination_root(uint32_t dest_info, std::string& mount_dir)
{
    std::string device;

    switch (dest_info & (GIGE_DEST_FIELD_MASK << GIGE_DEST_SHIFT))
    {
    case GIGE_IMAGE_DEST_SDCARD:
        storage::get_media_device(GIGE_MEDIA_SDCARD, device);
        break;
    case GIGE_IMAGE_DEST_USB:
        storage::get_media_device(GIGE_MEDIA_USB, device);
        break;
    case GIGE_IMAGE_DEST_SSD:
        storage::get_media_device(GIGE_MEDIA_SSD, device);
        break;
    default:
        break;
    }

    // Did we get a device?
    if (device.empty())
    {
        fi_log(LOG_ERR, "%s: Invalid destination media value (0x%x)", __PRETTY_FUNCTION__, dest_info);
    }
    else
    {
        // Get the mount point for the device.
        storage::get_mount_dir(device, mount_dir);
    }
}

/*!
 * @brief Make sure we only look at OUR files.
 *
 * This is a filter function to be used when calling scandir(). We make sure that
 * only IMAGE_IO::file_prefix named directory entries are used.
 *
 * @param entry_ptr     The directory entry we're interested in.
 * @sa The man page for scandir.
 */
static int IMAGE_IO::select_name(const struct dirent *entry_ptr)
{
    int rc;

    rc = strncasecmp(file_prefix.c_str(), entry_ptr->d_name, file_prefix.length());
    // LOG("%s: rc=%d, file_prefix='%s'/%d, entry='%s'", __PRETTY_FUNCTION__, rc, file_prefix.c_str(), file_prefix.length(), entry_ptr->d_name);
    return (rc == 0);
}

/*!
 * @brief Look through the target folder and figure out what the
 *    next index will be for our use.
 *
 * @param target_path   Where we're going to be looking.
 * @return The next index to use when saving videos.
 *
 * @sa The man page for scandir.
 */
static int IMAGE_IO::find_next_index(const char *target_path)
{
    int num_found;                  // Number of entries found.
    int last_index = 0;             // The next index to use.
    struct dirent **namelist;       // The dynamically allocated array of entries.

    // Find all matching entries in the folder.
    num_found = scandir(target_path, &namelist, select_name, alphasort);
    if (num_found > 0)
    {
        int  i;

        // Walk the list.
        for (i = 0; i < num_found; i++)
        {
            int  new_index;
            char *name_ptr;

            // Convert the very last one to get our index.
            name_ptr = namelist[ i ]->d_name;
            name_ptr += file_prefix.length();
            new_index = atoi(name_ptr) + 1;

            // Is this a better index to use?
            if (new_index > last_index)
                last_index = new_index;

            // Release the memory for this node.
            free(namelist[i]);
        }

        // Release the list.
        free(namelist);
    }

    // Return the next available index value.
    return last_index;
}

//--------------------------------------------------------------------
// imgSaveWorker class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. The imgSaveWorker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
imgSaveWorker::imgSaveWorker() :
    FAS_THREAD::Thread(true),
    last_status(0),
    save_error(0)
{
    // Enable tracing
//    traceControl |= TRACE_SEL_CAP_FILE;

    // Initialize our member variables.
    cfg_metadata_save = false;
    pfmd_save = false;
    thread_running = false;
    was_mounted = false;
    saver_starting = true;
    video_codec = GIGE_STREAM_FORMAT_JPEG;
    target_file.clear();
    saveMetadata = false;
    writeBuffers = false;

    // Initialize our marker list.
    initMarkerList();
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 *
 * @param arg   Not used.
 */
imgSaveWorker::~imgSaveWorker()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    cleanupMarkerList();
}

imgSaveWorker *imgSaveWorker::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // If we have an instance, just use it. Otherwise we
    // need to create it.
    if (instance.get() == nullptr)
    {
        // Allocate our instance.
        instance = std::unique_ptr<imgSaveWorker>(new imgSaveWorker());
    }

    // Return our instance pointer.
    return instance.get();
}

/*!
 * @brief Get the thread running.
 *
 * Only one "save" operation can happen at a time. If the worker thread is already running (imgSaveWorker::is_running())
 * this method will exit.
 *
 * @note There's no return value, so there's no indication of a failure.
 *
 * All things being equal, it is assumed that imgSaveWorker::run() will be executed. This method will hang around
 * for a little while (IMAGE_IO::STARTUP_DELAY) to make sure the worker thread signals that it is running (imgSaveWorker::thread_status).
 *
 * @param threadFn  [OPTIONAL] It's not expected that anyone would supply a different method here.
 */
void imgSaveWorker::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We can be asked for completion status before we create the saver instance.
    // We need to know whether it is null because it doesn't exist yet or because
    // it finished.
    saver_starting = true;

    // Now start the thread.
    FAS_THREAD::Thread::start(threadFn);

    {
        // Wait for the thread to start or IMAGE_IO::STARTUP_DELAY milliseconds. Put this in it's own scope
        // and the locker will release on the exit of the scope. The timeout is short
        // enough that we should be able to pick up a mode change just fine.
        std::unique_lock<std::mutex> locker(thread_mutex);
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (thread_status.wait_until(locker, now + ms(STARTUP_DELAY)) == std::cv_status::timeout)
        {
            // Check to see if the thread is actually running.
            if (!thread_running)
            {
                fi_log(LOG_DEBUG, "%s: timeout waiting for thread start", __PRETTY_FUNCTION__);
            }
        }
    }
}

/*!
 * @brief Ask the thread to stop.
 */
void imgSaveWorker::stop()
{
    // Tell the thread to stop.
    stop_request = true;

    if (doing_marker_list)
        stopMarkerList();

    // Is our saving class running?
    //
    // FIXME This is not entirely thread safe but we get away with it because there is such a low
    // probability of a context switch when there are no system calls being made. You can break
    // it by putting in a fi_log() before calling saver->stop(); then a context switch could cause
    // saver to be nulled between the test and the call to saver->stop().
    if (saver != NULL)
        saver->stop();
}

/*!
 * @brief Reset the data so we can have a fresh start.
 */
void imgSaveWorker::reset()
{
    capstk_base.clear();
    fflr_queue.clear();
    last_status = 0;
    save_error = 0;
    saveMetadata = false;
}

/*!
 * @brief Clean up the thread on our way out.
 */
void imgSaveWorker::cleanup()
{
    // Have the file system flush.
//    LOG("%s: Start of sync()", __PRETTY_FUNCTION__);
    sync();
//    LOG("%s: End of sync()", __PRETTY_FUNCTION__);

    // Notify any apps that we're done saving, at least if we had previously
    // notified them when we started saving.
    if (need_end_notification)
        fcp_notify_apps(GIGE_REG_SAVE_STATE, GIGE_SAVE_STATE_IDLE);

    // We're not running anymore.
    thread_running = false;
    LOG("%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Recover a failed FasFireLR recording.
 *
 * When a FasFireLR session ends in an underflow of buffers, there are dirty partitions
 * sitting around. Depending on what the user has requested, this routine will save or
 * discard the dirty partitions and save the metadata information for the capture.
 *
 * @sa fspWorker
 */
void imgSaveWorker::recoverFflr()
{
    int part_count;
    int dirty_count;
    uint32_t saveStartNum;
    uint32_t saveEndNum;
    uint32_t frame_count;
    int64_t num_frames;
    AutoSave *as_ptr;
    PartMgr *pm_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    LOG("%s: camera_state=%d", __PRETTY_FUNCTION__, CameraServer::get_camera_state());

    // Get our partition mananger pointer.
    pm_ptr = CameraServer::get_partmgr();
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null PartMgr pointer!?", __PRETTY_FUNCTION__);
        return;
    }
    as_ptr = CameraServer::get_autosave();
    if (as_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null AutoSave pointer!?", __PRETTY_FUNCTION__);
        return;
    }

    // Initialize our values.
    capstk_index = as_ptr->getFflrFileCount();

    // Read and save away the starting and ending frame numbers.
    // We do this so there will be something valid when we restore
    // later in the process. Also, in the case where the session
    // length is changed, we want to catch that change.
    saveStartNum = regs[ GIGE_REG_SAVE_START_NUM ].get_value();
    saveEndNum = regs[ GIGE_REG_SAVE_END_NUM ].get_value();

    // We always start at the first frame.
    regs[ GIGE_REG_SAVE_START_NUM ].set_value(0);

    // How many frames do we still want to save?
    num_frames = (int64_t)(pm_ptr->getReviewPartNumFrames() - 1);

    // Make sure we make sense. We can't have a stop that is bigger than the partition.
    if (num_frames >= (int64_t)pm_ptr->getFrameCount())
        num_frames = (int64_t)(pm_ptr->getFrameCount() - 1);

    // Sanity check the number of frames.
    if (num_frames <= 0)
    {
        fi_log(LOG_ERR, "%s: Invalid number of frames (%" PRId64 ")", __PRETTY_FUNCTION__, num_frames);
        return;
    }
    regs[ GIGE_REG_SAVE_END_NUM ].set_value((uint32_t)num_frames);

    // How many frames have we already written.
    frame_count = as_ptr->getTotalFflrFrames();

    // Get the number of images per partition.
    num_frames = (int64_t)pm_ptr->get_im_buf_count();
    if (num_frames <= 0)
    {
        // Strange value. Something must be wrong.
        fi_log(LOG_ERR, "%s: get_im_buf_count(%" PRId64 ") is bad!?", __PRETTY_FUNCTION__, num_frames);
        part_count = 1;
    }
    else
    {
        // Sanity check.
        if (getRecordingFrameCount() == 0)
        {
            fi_log(LOG_ERR, "%s: The recording frame count is 0!", __PRETTY_FUNCTION__);
            setRecordingFrameCount(pm_ptr->getPartitionNumFrames(pm_ptr->getActivePartition()));
        }

        // Calculate the number of partitions are required for the requested number of frames, round up.
        part_count = (int)(getRecordingFrameCount() + (num_frames / 2)) / num_frames;
    }

    // We need one extra partition for "slop".
    part_count++;

    // Calculate the number of frames for the entire capture.
    num_frames *= part_count;
    LOG("%s: Total number of frames allowed: %" PRId64 ", current frame count: %u", __PRETTY_FUNCTION__, num_frames, frame_count);

    // Loop through all the "dirty" partitions.
    for (dirty_count = (int)as_ptr->dirtySessions(); !stop_request && (dirty_count > 0); --dirty_count)
    {
        // Get the partition we need to work with.
        partition_num = as_ptr->getDirty();
        LOG("%s: dirty_count=%d, partition_num=%d, capstk_index=%d", __PRETTY_FUNCTION__, dirty_count, partition_num, capstk_index);

        // ALM_TODO: The following should already be done. If it is, then we can safely
        //           remove this code.
        if (pm_ptr->getActivePartition() != partition_num)
        {
            fi_log(LOG_ALERT, "%s: Expected partition %d to be active, but it is not!!", __PRETTY_FUNCTION__, partition_num);

            // Make this partition our active partition.
            pm_ptr->setActivePartition(partition_num);
        }
        else
        {
            fi_log(LOG_INFO, "%s: As expected, partition %d is setup.", __PRETTY_FUNCTION__, partition_num);
        }

        // Are we saving the dirty buffers?
        if (writeBuffers)
        {
            int64_t frames_in_part = (int64_t)pm_ptr->getPartitionNumFrames(partition_num);

            // Check to see if we need to discard an older CAP to make room.
            LOG("%s: frame_count=%u, amt to record=%u, partition_count:%" PRId64, __PRETTY_FUNCTION__, frame_count, getRecordingFrameCount(), frames_in_part);
            if (frames_in_part > num_frames)
            {
                fi_log(LOG_ERR, "%s: Invalid partition count: %u", __PRETTY_FUNCTION__, frames_in_part);
            }
            else
            {
                if ((frame_count + frames_in_part) >= num_frames)
                {
                    IMAGE_FORMAT::FFLR_CONTEXT_PTR oldest;

                    // Take the oldest off the queue. The record will be deleted once
                    // we exit scope.
                    oldest = popFflrQueue();
                    if (oldest.get() == nullptr)
                    {
                        fi_log(LOG_DEBUG, "%s: There were no entries in the queue?! frame_count=%u", __PRETTY_FUNCTION__, frame_count);
                        frame_count = 0;
                    }
                    else
                    {
                        // Delete the file.
                        fi_log(LOG_DEBUG, "%s: Remove '%s'", __PRETTY_FUNCTION__, oldest->cap_file.c_str());
                        if (!storage::save_path(oldest->cap_file.c_str(), oldest->cap_file.generic_string().size()))
                        {
                            fi_log(LOG_DEBUG, "%s: storage::save_path() failed!", __PRETTY_FUNCTION__);
                        }
                        else if (!storage::start_action(GIGE_FA_DELETE))
                        {
                            fi_log(LOG_DEBUG, "%s: storage::start_action(GIGE_FA_DELETE) failed!", __PRETTY_FUNCTION__);
                        }

                        // Update our counter.
                        frame_count -= oldest->frame_count;
                    }
                }

                // Save the partition.
                saveVideo();

                // Update the frame count.
                frame_count += (uint32_t)frames_in_part;
            }
        }

        // Release the dirty buffer.
        as_ptr->queueUpdate();

        // Increment the CAP stack file number.
        capstk_index++;
    }

    // We expect the saveMetadata flag to be set.
    if (!saveMetadata)
    {
        fi_log(LOG_DEBUG, "%s: saveMetadata is not true!?", __PRETTY_FUNCTION__);
    }

    // Save the FFLR metadata file.
    saveFflrMetadata();

    // There are no more dirty partitions, switch the camera to live.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO, false, NULL);

    // Restore the user specified starting/ending frames.
    regs[ GIGE_REG_SAVE_START_NUM ].set_value(saveStartNum);
    regs[ GIGE_REG_SAVE_END_NUM ].set_value(saveEndNum);
}

/*!
 * @brief Save a video to the SSD.
 *
 * This is the main entry for a save operation to copy from image memory or an LTR SSD capture
 * to a LOCAL target.
 *
 * @sa fspWorker
 */
void imgSaveWorker::saveVideo()
{
    bool    still_working;
    double  elapsed;
    uint64_t    elapsed_usecs;
    const char  *target_ptr(nullptr);
    PartMgr *pm_ptr;
    struct timespec start_time;

    // Get our partition mananger pointer.
    pm_ptr = CameraServer::get_partmgr();
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null PartMgr pointer!?", __PRETTY_FUNCTION__);
        return;
    }

    // Initialize our stream save class.
    saver = std::make_unique<StreamSave>(dest_info);
    if (saver.get() == nullptr)
    {
        fi_log(LOG_CRIT, "Unable to allocate the saver object!");
        return;
    }

    // Okay, the completion reporter can get info from the saver now.
    saver_starting = false;

    // Tell the partition manager about our partition.
    pm_ptr->setReviewPartNum(partition_num);

    // Make sure we update our metadata.
    pm_ptr->updatePartitionMetadata(partition_num);

    // Get the actual cut-in/-out frames for "normal" file types.
    saver->setStartFrame(getStartFrameNumber());
    saver->setEndFrame(getEndFrameNumber());

    // Are we saving an MP4?
    if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_MP4)
    {
        // Pass on the MP4 specific parameters.
        saver->setH264Speed(h264_speed);
        saver->setH264Quality(h264_quality);
        saver->setH264EnableAllKeyframes(h264_enable_all_keyframes);
    }
    saver->setVideoCodec(video_codec);

    // Are we saving a CAP stack, and do we already have a base path?
    if (((dest_info & GIGE_STREAM_FORMAT_MASK) != GIGE_STREAM_FORMAT_CAPSTK) || capstk_base.empty())
    {
        fi_log(LOG_DEBUG, "%s: Not CAPSTK [dest_info=0x%x] or capstk_base is empty", __PRETTY_FUNCTION__, dest_info);

        // The function to build the target file/directory name will get the next
        // index each time. Presumably the reason this loops is in case that function
        // creates a file/dir name that already exists, with the idea that a retry
        // will come up with a different name.
        while (1)
        {
            struct stat sb;

            target_ptr = buildTarget();
            if (target_ptr == NULL)
            {
                fi_log(LOG_WARNING, "%s: Target pointer is NULL.", __PRETTY_FUNCTION__);
                goto sv_bail1;
            }

            // Does it exist?
            if (stat(target_ptr, &sb) < 0)
                break;

            fi_log(LOG_DEBUG, "%s: '%s' already exists...keep looking...", __PRETTY_FUNCTION__, target_ptr);
        }
    }

    // Are we saving a CAP stack?
    if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPSTK)
    {
        // Does our base path need to be initialized?
        if (capstk_base.empty())
        {
            int status;

            // Make sure our "root" folder exists.
            status = storage::mkpath(target_file.c_str(), (S_IRWXU | S_IRWXO | S_IRWXG));
            if (status == -1)
            {
                fi_log(LOG_ERR, "%s: Can't create '%s', err=%d(%s)", __FUNCTION__, target_file.c_str(), errno, strerror(errno));
                goto sv_bail1;
            }

            // Save the new path for reuse later.
            capstk_base.assign(target_file);
        }
        else
        {
            // Reuse the prevviously saved path.
            target_file.assign(capstk_base);
        }

        // Build up the full path to the capstk file.
        target_file.append(fmt::sprintf("/%07d.%s", capstk_index, imgio::get_suffix(GIGE_STREAM_FORMAT_CAPSTK)));
        target_ptr = target_file.c_str();
        fi_log(LOG_DEBUG, "%s: FFLR target='%s'", __PRETTY_FUNCTION__, target_ptr);
    }

    // Set the target path.
    saver->setTargetPath(target_ptr);

    // Get our starting time for the statistics output.
    measuretime(1, &start_time);

    // Now that we've prepped the SSD properly, we can hand off the actual saving
    // of data to our image specific processor.
    still_working = true;
    while (still_working)
    {
        // Is it time to quit?
        if (stop_request)
            break;

        // Run the save process.
        still_working = saver->doSave();
    }

    // Gather our statistics for our save.
    elapsed_usecs = measuretime(0, &start_time);
    if (elapsed_usecs < 10)     // Avoid overflow or zero division.
        elapsed = saver->getTotalBytes();
    else
        elapsed = (double)elapsed_usecs / 1000000.0;

    if (stop_request)
        fi_log(LOG_DEBUG, "%s: file='%s' part_num=%d status=0x%x sec=%f (cancel)",
               __FUNCTION__, saver->getTargetPath(), partition_num, saver->getStatus(), elapsed);
    else
        fi_log(LOG_DEBUG, "%s: file='%s' part_num=%d status=0x%x frames=%d bytes=%" PRIu64 " sec=%f [%6.2f Byte/sec]",
               __FUNCTION__, saver->getTargetPath(), partition_num, saver->getStatus(), saver->getTotalFrames(),
               saver->getTotalBytes(), elapsed, (saver->getTotalBytes() / elapsed));

    // Save the status information.
    if (last_status != 0)
        fi_log(LOG_DEBUG, "%s: last_status is set [%d], refusing to overwrite with [%d]", __PRETTY_FUNCTION__, last_status, saver->getStatus());
    else
    {
        last_status = saver->getStatus();
        fi_log(LOG_DEBUG, "%s: last_status=0x%x", __PRETTY_FUNCTION__, last_status);
    }

    // Was the save successful?
    if (!stop_request && (saver->getInfo()->status == 0))
    {
        PART_MGR_WORKER::PartInfo part_info;

        // Get a pointer to our metadata.
        if (!pm_ptr->getPartInfo(partition_num, part_info))
        {
            fi_log(LOG_DEBUG, "%s: Unable to get the partition information for partition %d", __PRETTY_FUNCTION__, partition_num);
            goto sv_bail1;
        }

        // Are we in FasFireLR mode?
        if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPSTK)
        {
            // Allocate a tracking structure.
            IMAGE_FORMAT::FFLR_CONTEXT_PTR entry = std::make_unique<IMAGE_FORMAT::FFLR_CONTEXT>();

            // Fill in the structure.
            entry->partition = pm_ptr->getReviewPartition();
            entry->frame_count = part_info.metadata.frame_count;
            entry->cap_file = target_ptr;
            entry->file_number = capstk_index;

            // Is this partition the trigger partition?
            if (part_info.is_trig_part)
            {
                // This is the trigger partition!
                entry->is_trigger_partition = true;

                // Reset the entry. It can never be "the trigger partition" again.
                part_info.is_trig_part = false;
                pm_ptr->updatePartInfo(partition_num, part_info);
            }
            else
            {
                // This is NOT the trigger partition!
                entry->is_trigger_partition = false;
            }

            // Did we get trigger information for this partition?
            if ((part_info.metadata.trig_time_s == 0) && (part_info.metadata.trig_time_ns == 0))
                entry->trigger_frame = -1;
            else
                entry->trigger_frame = part_info.metadata.trig_frame;

            // Push the pointer to the structure on our queue.
            fflr_queue.push_back(std::move(entry));
        }

        // Check to see if we're supposed to create an XML metafile.
        if ((dest_info & GIGE_IMAGE_SAVE_XML) == GIGE_IMAGE_SAVE_XML)
        {
            std::string dest_name;
            uint32_t format_type = dest_info & GIGE_STREAM_FORMAT_MASK;

            // Build up the target file name.
            dest_name.assign(saver->getInfo()->target_path);
            if ((format_type == GIGE_STREAM_FORMAT_AVI) || (format_type == GIGE_STREAM_FORMAT_CAPTURE) || (format_type == GIGE_STREAM_FORMAT_MP4))
            {
                dest_name.replace(dest_name.length() - 3, 3, "xml");
            }
            else
            {
                dest_name.append("/metadata.xml");
            }

            // Get our partition manager pointer.
            PartMgr *pm_ptr;
            pm_ptr = CameraServer::get_partmgr();

            // Save the XML file.
            if (!saveXmlMetadata(dest_name.c_str(), pm_ptr->getPartitionMetadata(partition_num).get(), partition_num))
                last_status = EIO;
        }
    }

sv_bail1:
    // Release our image saver class.
    saver.reset();
}

/*!
 * @brief Save the FasFireLR metadata file for the recording.
 */
void imgSaveWorker::saveFflrMetadata()
{
    FILE *fd;
    uint32_t reg_value;
    uint32_t frame_total;
    Json::Value json_meta;
    Json::Value save_info;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;
    std::filesystem::path meta_file;
    PartMgr *pm_ptr;
    FiFPGA *fpga_ptr;

    // Make sure we have something.
    if (fflr_queue.empty())
    {
        // If there wasn't an error, just exit.
        if (save_error == 0)
        {
            fi_log(LOG_DEBUG, "%s: Empty FFLR queue...exiting.", __PRETTY_FUNCTION__);
        }
        else
        {
            // Build up our FFLR information for the user to query.
            save_info["status"] = save_error;

            // Create a string.
            fflr_info.clear();
            wr_builder["commentStyle"] = "All";
            wr_builder["indentation"] = "";
            fflr_info = Json::writeString(wr_builder, save_info);
            fi_log(LOG_DEBUG, "%s: %s", __PRETTY_FUNCTION__, fflr_info.c_str());
        }

        // All done here.
        return;
    }

    // Get our partition mananger pointer.
    pm_ptr = CameraServer::get_partmgr();
    fpga_ptr = CameraServer::get_fpga();

    // Read up the ICM control register.
    fpga_ptr->regRead(ICM_CONTROL, &reg_value);

    // Build up our FFLR information for the user to query.
    meta_file = fflr_queue.front()->cap_file;
    save_info["path"] = meta_file.parent_path().string();
    save_info["first"] = fflr_queue.front()->file_number;

    // Grab some information for the global meta file.
    json_meta["start_cap"] = fflr_queue.front()->file_number;
    json_meta["end_cap"] = fflr_queue.back()->file_number;

    // Figure out what mode the FPGA is in.
    if ((reg_value & ICMC_CAPTURE_MODE_MASK) == ICMC_CAPTURE_MODE_FFLR)
    {
        save_info["status"] = save_error & GIGE_SAVE_STATE_ERR_MASK;
        json_meta["recmode"] = std::string("basic");
    }
    else if ((reg_value & ICMC_CAPTURE_MODE_MASK) == ICMC_CAPTURE_MODE_FFLR_ROC)
    {
        save_info["status"] = fpga_ptr->ltrSaveError() & GIGE_SAVE_STATE_ERR_MASK;
        json_meta["recmode"] = std::string("roc");
    }
    else if ((reg_value & ICMC_CAPTURE_MODE_MASK) == ICMC_CAPTURE_MODE_FFLR_BROC)
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        save_info["status"] = fpga_ptr->ltrSaveError() & GIGE_SAVE_STATE_ERR_MASK;
        json_meta["recmode"] = std::string("broc");
        json_meta["blen"] = regs[GIGE_REG_BROC_LEN].get_value();
    }

    // Was this a recovery?
    if (fflrRecovery())
    {
        // Note that this was a recovered session.
        save_info["recover"] = last_status;
    }

    // Create a string.
    fflr_info.clear();
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    fflr_info = Json::writeString(wr_builder, save_info);
    fi_log(LOG_DEBUG, "%s: %s", __PRETTY_FUNCTION__, fflr_info.c_str());

    // Reset our path information.
    meta_file.clear();

    // Loop through all the entries in the queue.
    frame_total = 0;
    while (fflr_queue.size() > 0)
    {
        Json::Value json_entry;
        IMAGE_FORMAT::FFLR_CONTEXT_PTR entry;

        // Move the information over to our return value.
        entry.reset(fflr_queue.front().release());

        // Pop the entry of the queue.
        fflr_queue.pop_front();

        // Do we need to initialize the meta_file?
        if (meta_file.empty())
        {
            // Get the path to the first cap file. We'll put the metadata file
            // in the same folder.
            meta_file = entry->cap_file;

            // Get the target folder.
            meta_file.replace_filename(FFLR_METAFILE_NAME);
//            LOG("%s: meta_file='%s'", __PRETTY_FUNCTION__, meta_file.c_str());
        }

        // Build up our JSON entry.
        json_entry["name"] = entry->file_number;

        // The trigger information only matters in basic mode.
        if ((reg_value & ICMC_CAPTURE_MODE_MASK) == ICMC_CAPTURE_MODE_FFLR)
        {
            // Is this partition the trigger partition?
            if (entry->is_trigger_partition)
            {
                // Save the recording's trigger frame.
                json_entry["tframe"] = entry->trigger_frame + frame_total;

                // Save the CAP file number containing the trigger event in the global data.
                json_meta["trigger_cap"] = entry->file_number;
            }
        }

        json_entry["sframe"] = frame_total;
        json_entry["eframe"] = entry->frame_count + frame_total - 1;
        json_entry["fcount"] = entry->frame_count;

        // Update our frame count total.
        frame_total += entry->frame_count;

        // Append the entry to our metadata.
        json_meta["map"].append(json_entry);
    }

    // Finish up the global meta values.
    json_meta["tframes"] = frame_total;
    json_meta["psize"] = (int)(pm_ptr->getPartitionSize() >> 30);   // Need this in units of GBytes. 1 GByte is 2^30.

    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 2;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, json_meta);

    // Save the data out to the file.
    fd = fopen(meta_file.c_str(), "w");
    if (fd != NULL)
    {
        fwrite(json_message.c_str(), json_message.size(), 1, fd);
        fclose(fd);
    }
}

/*!
 * @brief Take the oldest entry off of the fflr_queue and return
 *        it to the caller.
 *
 * @return The oldest entry, or an empty unique_ptr.
 */
IMAGE_FORMAT::FFLR_CONTEXT_PTR imgSaveWorker::popFflrQueue()
{
    IMAGE_FORMAT::FFLR_CONTEXT_PTR entry;

    if (fflr_queue.size() > 0)
    {
        // Move the information over to our return value.
        entry = std::move(fflr_queue.front());

        // Pop the entry of the queue.
        fflr_queue.pop_front();
    }

    // Return the entry.
    return entry;
}

int imgSaveWorker::getLiveSnapshot()
{
    int rc;
    int fd;
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    JPEG_PROC::IPM_REQUEST request;

    // Clear out the request structure.
    memset(&request, 0, sizeof(request));

    // Anything within the past 100ms should be fine.
    request.time_limit = 100000;

    // Processing info is always the same for live view images.
    request.dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
    if (sensor_ptr->is_sensor_color())
    {
        request.proc_info = IMGPROC_TYPE_NV12;
    }
    else
    {
        request.proc_info = IMGPROC_TYPE_Y8;
    }
    if ((regs[ GIGE_REG_FPN ].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        request.proc_info |= IMGPROC_USE_FPN;
    request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();

    // Get our dimensions.
    request.width = regs[GIGE_REG_WIDTH].get_value();
    request.height = regs[GIGE_REG_HEIGHT].get_value();

    // Ask for the frame from the JPEG processor.
    if (!jpeg_ipm_request(&request))
    {
        // Go try to acquire another frame.
        fi_log(LOG_WARNING, "%s: Unable to encode the frame! Did we change states?", __PRETTY_FUNCTION__);
        return EFAULT;
    }

    // Open the file.
    fd = open(STILL_NAME, O_CREAT | O_RDWR | O_TRUNC, 0666);
    if (fd == -1)
    {
        rc = errno;
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return rc;
    }

    // Write out the data.
    if (write(fd, request.jpeg_buffer, request.jpeg_len) < 0)
    {
        rc = errno;
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return rc;
    }

    // All done.
    close(fd);

    // Release the buffer.
    jpeg_release_frame(request.jpeg_buffer);

    // All is well.
    return 0;
}

bool imgSaveWorker::parseSaveSettings(const char *save_json, int str_length, bool start_saver)
{
    Json::Value save_info;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    bool parsingSuccessful;
    uint32_t reg_value(0);
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return false;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(save_json, save_json + str_length, &save_info, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, save_json, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        goto parse_bail;
    }
    LOG("%s: save_info='%s'", __PRETTY_FUNCTION__, save_info.toStyledString().c_str());

    // See if we have our values.
    if (!save_info["write_buffers"].isNull() && !save_info["save_metadata"].isNull())
    {
        // Assume our parsing will be fine.
        parsingSuccessful = true;

        try
        {
            // Read the parameters from the JSON string.
            if (!save_info["write_buffers"].isNull())
                writeBuffers = save_info["write_buffers"].asBool();
            if (!save_info["save_metadata"].isNull())
                saveMetadata = save_info["save_metadata"].asBool();
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: reading FFLR recover settings asBool() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
            parsingSuccessful = false;
        }
        catch (...)
        {
            fi_log(LOG_DEBUG, "%s: reading FFLR recover settings asBool() UNKNOWN EXCEPTION", __PRETTY_FUNCTION__);
            parsingSuccessful = false;
        }

        // Continue to recovering the FFLR session.
        LOG("%s: Parsed, moving on.", __PRETTY_FUNCTION__);
        goto parse_bail;
    }
    else
    {
        // We need both values. Assume this isn't about recovering an FFLR session.
        LOG("%s: FFLR recovery values not set.", __PRETTY_FUNCTION__);
    }

    // Reset our recovery flags.
    writeBuffers = false;
    saveMetadata = false;

    // Assume we are going to fail.
    parsingSuccessful = false;

    try
    {
        // We better have the destination information field.
        if (save_info["dst_info"].isNull())
        {
            goto parse_bail;
        }

        // Parse out the information and setup the registers for a save operation.
        memset((char *)((GIGE_REG *)regs[GIGE_REG_IMAGE_TAG])->reg_val_b, 0, GIGE_IMAGE_TAG_MAX_LEN);
        if (!save_info["tag"].isNull())
        {
            std::string tag_str(save_info["tag"].asString());
            strncpy((char *)((GIGE_REG *)regs[GIGE_REG_IMAGE_TAG])->reg_val_b, tag_str.c_str(), tag_str.size());
        }
        if (!save_info["frame_start"].isNull())
            regs[GIGE_REG_SAVE_START_NUM].set_value((uint32_t)save_info["frame_start"].asUInt64());
        if (!save_info["frame_end"].isNull())
            regs[GIGE_REG_SAVE_END_NUM].set_value((uint32_t)save_info["frame_end"].asUInt64());
        if (!save_info["format"].isNull())
        {
            // Extract the format.
            regs[GIGE_REG_STREAM_FORMAT].set_value((uint32_t)save_info["format"].asUInt64());

            // Packed is special. Make sure we can support that.
            if ((regs[GIGE_REG_STREAM_FORMAT].get_value() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_PACK)
            {
                // Do we have accerss to the library?
                if (FiInternal::factory() == nullptr)
                {
                    // Not permitted.
                    save_error = EPERM;
                    fi_log(LOG_DEBUG, "%s: Unable to load the fas_internal library.", __PRETTY_FUNCTION__);
                    fcp_notify_apps(GIGE_REG_SAVE_STATE, GIGE_SAVE_STATE_ERR | EPERM);
                    goto parse_bail;
                }
            }
        }
        if (!save_info["avi_codec"].isNull())
            setVideoCodec((uint32_t)save_info["avi_codec"].asUInt64());
        if (!save_info["dstpath"].isNull())
            setMountDir(save_info["dstpath"].asString());
        if (save_info["save_metadata_cfg"].isNull())
            setSaveCfgMetadata(false);
        else
            setSaveCfgMetadata(save_info["save_metadata_cfg"].asBool());
        if (save_info["save_pfmd"].isNull())
            setSavePFMD(false);
        else
            setSavePFMD(save_info["save_pfmd"].asBool());
        if (!save_info["h264_quality"].isNull())
            setH264Quality(save_info["h264_quality"].asInt());
        if (!save_info["h264_speed"].isNull())
            setH264Speed((FAS_VIDEO::H264_SPEEDS)(save_info["h264_speed"].asInt()));
        if (!save_info["h264_enable_all_keyframes"].isNull())
            setH264EnableAllKeyframes(save_info["h264_enable_all_keyframes"].asBool());

        if (!save_info["frame_count"].isNull())
            setRecordingFrameCount(save_info["frame_count"].asUInt64());

        if (!save_info["shared_mem"].isNull())
            setUsingShmem(save_info["shared_mem"].asBool());
        else
            setUsingShmem(false);

        if (!save_info["dst_info"].isNull())
        {
            // Get our information.
            reg_value = (uint32_t)save_info["dst_info"].asUInt64();

            // If we are not supposed to immediately start, make sure we set the delay bit.
            if (!start_saver)
                reg_value |= GIGE_IMAGE_SAVE_DELAY;

            // Set the destination information.
            parsingSuccessful = (fcp_image_dest_update(NULL, reg_value) == GEV_STATUS_SUCCESS);
        }
    }
    catch (Json::Exception& e)
    {
        fi_log(LOG_DEBUG, "%s: reading JSon save settings EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        parsingSuccessful = false;
    }
    catch (...)
    {
        fi_log(LOG_DEBUG, "%s: reading JSon save settings UNKNOWN EXCEPTION", __PRETTY_FUNCTION__);
        parsingSuccessful = false;
    }

parse_bail:
    // Return our status.
    return parsingSuccessful;

}

/*!
 * @brief If appropriate, restore the previously saved JSON save settings string.
 * @param enable_as True if we are also supposed to enable autosave.
 */
void imgSaveWorker::restoreSaveSettings(bool enable_as)
{
    std::string file_path;
    std::string settings_str;
    std::ifstream settings_file;

    // Try to open the saved settings file.
    file_path = fmt::format("{0}/{1}", get_home_path(), SAVE_SETTINGS_FILE);
    settings_file.open(file_path, std::ios::in);
    if (settings_file.is_open())
    {
        FiFPGA *fpga_ptr;

        // Get our partition mananger pointer.
        fpga_ptr = CameraServer::get_fpga();

        // Are we in FasFireLR mode?
        if (fpga_ptr->ltrFasFireEnabled())
        {
            GIGE_REG_CACHE& regs = CameraServer::get_regs();

            // Only worry about this if the external ARM-in signal is enabled.
            if ((regs[GIGE_REG_ARM_IN_ENABLE].get_value() & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
            {
                std::ostringstream ss;

                // Read in the previous settings.
                ss << settings_file.rdbuf();
                settings_str = ss.str();
                settings_file.close();

                // Parse the settings values.
                if (enable_as && parseSaveSettings(settings_str.c_str(), settings_str.length()))
                {
                    // Enable autosave.
                    fcp_autosave_update(nullptr, GIGE_AUTOSAVE_ENABLE);
                }
            }
        }
    }
}

/*!
 * @brief Save the JSON string so we can reuse it later.
 * @param save_json The pointer to the JSON string we want to save.
 */
void imgSaveWorker::updateSaveSettings(const char *save_json)
{
    std::string file_path;
    std::ofstream settings_file;

    // Save the save settings JSON string for possible use later.
    file_path = fmt::format("{0}/{1}", get_home_path(), SAVE_SETTINGS_FILE);
    settings_file.open(file_path, std::ios::out | std::ios::trunc | std::ios::binary);
    if (settings_file.is_open())
    {
        settings_file.write(save_json, strlen(save_json));
        settings_file.close();
    }
}

/*!
 * @brief Return the begining frame number we want to save.
 */
int32_t imgSaveWorker::getStartFrameNumber()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int32_t   start_frame;

    start_frame = regs[ GIGE_REG_SAVE_START_NUM ].get_value();
    return start_frame;
}

/*!
 * @brief The worker thread.
 *
 * This method does the actual work the imgSaveWorker class. This thread:
 * + Collects event information (imgSaveWorker::create_marker_list())
 * + Captures snapshot images (GIGE_IMAGE_PREP_SNAP, GIGE_IMAGE_SNAPSHOT)
 * + Saves the per-frame metadata (GIGE_IMAGE_SAVE_XML, imgSaveWorker::save_xml_metadata())
 * + Saves images (imgSaveWorker::saveVideo())
 *
 * Due to the fact that this thread can generate a very large load on the system,
 * we reduce the priority (IMAGE_IO::WORKER_PRIORITY).
 *
 * @sa fspWorker for network saves.
 */
void imgSaveWorker::run()
{
    pid_t tid;

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Reduce the priority on this thread, so if other tasks need to come in they can.
    tid = syscall(SYS_gettid);
    setpriority(PRIO_PROCESS, tid, WORKER_PRIORITY);

    // Cleanup function depends on having this set.
    need_end_notification = false;

    // Notify folks that we're saving, if that's what we're doing.
    if (!doing_marker_list)
    {
        fcp_notify_apps(GIGE_REG_SAVE_STATE, GIGE_SAVE_STATE_BUSY);
        need_end_notification = true;

        // We'll clear the save status for a save but not for markers.
        last_status = 0;
    }

    // Notify anyone listening (i.e., the start() method) that we're active.
    {
        std::unique_lock<std::mutex> locker(thread_mutex);

        // We're running.
        thread_running = true;
        stop_request = 0;
        thread_status.notify_one();
    }

    // Check for the hijacked work task. Are we assembling the marker events?
    if (doing_marker_list)
    {
        PartMgr *pm_ptr;

        // Get our partition manager pointer.
        pm_ptr = CameraServer::get_partmgr();

        // Set the active partition the same way the rest of the savers do.
        pm_ptr->setActivePartition(partition_num);
        pm_ptr->queueCapturePartition(partition_num);
        pm_ptr->updateActiveMetadata();

        // Go create our list.
        createMarkerList();
    }

    // Are we doing snap shot work here?
    else if (dest_info & (GIGE_IMAGE_PREP_SNAP | GIGE_IMAGE_SNAPSHOT))
    {
        uint32_t    reg_value;
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        // Get the autosave register.
        reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();

        // Is autosave enabled? If so, disable it.
        if ((reg_value & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
        {
            reg_value &= ~GIGE_AUTOSAVE_ENABLE;
            regs[ GIGE_REG_AUTOSAVE ].set_value(reg_value);
        }

        // Are we preparing the snapshot, or saving it?
        if (dest_info & GIGE_IMAGE_PREP_SNAP)
        {
            int rc;

            // Grab a snapshot from the live stream.
            rc = getLiveSnapshot();
            if (rc != 0)
                last_status = rc;
        }
        else
        {
            const char *target_ptr;
            std::string temp_file;

            // Get our target information.
            target_ptr = buildTarget();
            if (target_ptr == NULL)
            {
                fi_log(LOG_WARNING, "%s: Target pointer is NULL.", __PRETTY_FUNCTION__);
                goto is_bail1;
            }

            // Build up our destination path.
            temp_file = fmt::sprintf("%s.jpg", target_ptr);

            // Do the deed.
            if (!copyFile(temp_file.c_str(), STILL_NAME))
                last_status = errno;
        }
    }

    // Are we ONLY saving an XML file?
    else if (dest_info == GIGE_IMAGE_SAVE_XML)
    {
        std::string temp_path;
        PartMgr *pm_ptr;

        // Get our partition manager pointer.
        pm_ptr = CameraServer::get_partmgr();

        // Build up the destination path for the XML.
        temp_path = fmt::sprintf("%s/%s", get_home_path(), DOWNLOAD_XML_METADATA);

        // Save the XML metadata to our path.
        if (!saveXmlMetadata(temp_path.c_str(), pm_ptr->getActivePartMetadata().get(), pm_ptr->getActivePartition()))
        {
            fi_log(LOG_ERR, "%s: Unable to save the XML metadata.", __FUNCTION__);
            last_status = EIO;
        }
    }

    // Are we in FFLR recovery mode?
    else if (imgSaveWorker::getInstance()->fflrRecovery())
    {
        // Process the FFLR recovery operation.
        LOG("%s: Starting FFLR recovery.", __PRETTY_FUNCTION__);
        recoverFflr();
    }

    else
    {
        // Start saving the video/stack.
        saveVideo();
    }

is_bail1:
    // Clean up on the way out.
    cleanup();
}

/*!
 * @brief Return the ending frame number we want to save.
 */
int32_t imgSaveWorker::getEndFrameNumber()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int32_t   end_frame;

    end_frame = regs[ GIGE_REG_SAVE_END_NUM ].get_value();
    return end_frame;
}

/*!
 * @brief Builds a JSon formatted status string for a FasFireLR save. It will return
 *        the null terminated C-string in the "dest_ptr" parameter specified buffer.
 *
 * If the save operation has not completed, a length of 1 and a null character ('\0')
 * will be returned. Otherwise, the JSon status C-string will be returned. The string
 * will be copied up to the "max_len" parameter. Subsequent calls will copy the remaining
 * substrings.
 *
 * @param dest_ptr  [IN, OUT] Pointer to the buffer to hold the JSon status C-string.
 * @param max_len   [IN] The maximum number of characters to copy to "dest_ptr".
 * @return The number of characters copied to "dest_ptr".
 */
uint32_t imgSaveWorker::getLastFflrSaveInfo(char *dest_ptr, uint32_t max_len)
{
    uint32_t ret_len = 0;

    // Sanity check.
    if (dest_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: dest_ptr is NULL!?", __PRETTY_FUNCTION__);
        return 0;
    }

    // Make sure we null terminate the return value.
    dest_ptr[0] = '\0';

    // Do we have a status string?
    if (fflr_info.empty())
    {
        AutoSave    *as_mgr = CameraServer::get_autosave();

        // If autosave is still running, we just haven't finished yet.
        if (as_mgr && as_mgr->isRunning())
        {
            // Need to wait for the autosave operation to finish up.
            return 1;
        }

        // Autosave isn't running...do we have any information to return?
        if (fflr_queue.empty())
        {
            fi_log(LOG_DEBUG, "%s: Empty FFLR queue, status=ENOENT[%d].", __PRETTY_FUNCTION__, ENOENT);
            save_error = ENOENT;
        }
        else
        {
            // Try building up the information, but set an error condition first.
            fi_log(LOG_DEBUG, "%s: Appears we are done, but not real happy. status=EPIPE[%d]", __PRETTY_FUNCTION__, EPIPE);
            save_error = EPIPE;
        }

        // This should generate an fflr_info string, even if it only contains status.
        saveFflrMetadata();
    }

    // Make sure we didn't get crazy.
    if (fflr_info.length() >= max_len)
    {
        // Copy over what we can.
//        fi_log(LOG_DEBUG, "%s: JSON string [%s] is too long! Only transferring %d bytes", __PRETTY_FUNCTION__, fflr_info.c_str(), max_len);
        memcpy(dest_ptr, fflr_info.c_str(), max_len - 1);
        ret_len = max_len - 1;

        // Shorten our string.
        fflr_info = fflr_info.substr((max_len - 1), std::string::npos);
    }
    else
    {
        // Copy the string over.
//        fi_log(LOG_DEBUG, "%s: JSON string copy [%s], %d bytes.", __PRETTY_FUNCTION__, fflr_info.c_str(), fflr_info.length());
        memcpy(dest_ptr, fflr_info.c_str(), fflr_info.length());
        ret_len = fflr_info.length();

        // Clear the string.
        fflr_info.clear();
    }

    // Return the length, including the null-terminator.
    return ret_len + 1;
}

/*!
 * @brief Build up the target file name and return the result.
 *
 * @return The member variable imgSaveWorker::target_file, or NULL on failure.
 * @sa FCP_LOCAL::fcp_parse_save_json()
 */
const char *imgSaveWorker::buildTarget()
{
    int status;
    int next_index;                 // Which index is available.
    bool is_still;                  // True if we're doing stills, false for videos.
    const char *suffix_ptr;         // Pointer to path components.

    // Check to see if the destination is off-camera. If it is, just return NULL.
    if ((dest_info & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
        return NULL;

    // Figure out the suffix information.
    suffix_ptr = imgio::get_suffix(dest_info & GIGE_STREAM_FORMAT_MASK);
    switch (dest_info & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
    case GIGE_STREAM_FORMAT_BMP:
    case GIGE_STREAM_FORMAT_DNG:
    case GIGE_STREAM_FORMAT_JPEG:
    case GIGE_STREAM_FORMAT_PLANAR:
        is_still = true;
        break;
    case GIGE_STREAM_FORMAT_AVI:
    case GIGE_STREAM_FORMAT_CAPTURE:
    case GIGE_STREAM_FORMAT_CAPSTK:
    case GIGE_STREAM_FORMAT_MP4:
    default:
        is_still = false;
        break;
    }

    // Build up our destination path.
    target_file.clear();

    // Is the mount folder already populated? If we were invoked due to a GIGE_REG_SAVE_JSON write,
    // the mount_dir is likely already populated.
    if (mount_dir.empty())
    {
        // Get our destination base path using the media type information.
        get_destination_root(dest_info, mount_dir);
        if (mount_dir.empty())
        {
            fi_log(LOG_ERR, "%s: Invalid destination media value (0x%x)", __PRETTY_FUNCTION__, dest_info);
            return NULL;
        }
    }

    // Get the mount point for the device.
    target_file.assign(mount_dir);
    if (is_still)
        target_file.append(STILL_FOLDER);
    else if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPTURE)
    {
        target_file.append(CAPTURE_FOLDER);
    }
    else if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPSTK)
    {
        // If we are in FFLR mode, we go into the STILL_FOLDER.
        target_file.append(STILL_FOLDER);
    }
    else
    {
        target_file.append(VIDEO_FOLDER);
    }

    // Make sure our "root" folder exists.
    status = storage::mkpath(target_file.c_str(), (S_IRWXU | S_IRWXO | S_IRWXG));
    if (status == -1)
    {
        fi_log(LOG_ERR, "%s: Can't create '%s', err=%d(%s)", __FUNCTION__, target_file.c_str(), errno, strerror(errno));
        return nullptr;
    }

    // Initialize our prefix string.
    file_prefix.clear();
    if ((dest_info & (GIGE_IMAGE_USE_TAG | GIGE_IMAGE_USE_CAMERA_NAME)) == 0)
    {
        file_prefix.assign(NAME_PREFIX);
    }
    else
    {
        // Check to see if we want to use the camera name.
        if (dest_info & GIGE_IMAGE_USE_CAMERA_NAME)
        {
            char hostname_buf[GIGE_REG_USER_DEF_NAME_B_LEN + 1];
            hostname_buf[0] = '\0';

            // Get the name from the system.
            CsNetwork::getHostname(hostname_buf, GIGE_REG_USER_DEF_NAME_B_LEN);
            file_prefix.assign(hostname_buf);
            file_prefix += '_';
        }

        // See if we want the tag name.
        if (dest_info & GIGE_IMAGE_USE_TAG)
        {
            GIGE_REG_CACHE& regs = CameraServer::get_regs();
            char tag[GIGE_IMAGE_TAG_MAX_LEN + 1];

            // Copy the tag into a buffer.
            tag[0] = 0;
            strncpy(tag, (char *)regs[GIGE_REG_IMAGE_TAG], GIGE_IMAGE_TAG_MAX_LEN);

            // GIGE_IMAGE_USE_TAG set with no text means use default tag.
            if (tag[0] == 0)
            {
                set_default_tag(tag, GIGE_IMAGE_TAG_MAX_LEN);
                file_prefix += tag;
            }
            else
            {
                file_prefix += tag;
            }

            file_prefix += '_';
        }
    }

    // Find the next available index.
    next_index = find_next_index(target_file.c_str());

    // Build up the template.
    if (is_still)
    {
        target_file.append(fmt::sprintf("%s%06d", file_prefix.c_str(), next_index));
    }
    else if ((dest_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPSTK)
    {
        // Build up the folder path.
        target_file.append(fmt::sprintf("%s%06d", file_prefix.c_str(), next_index));
    }
    else
    {
        // Now build up the rest of the filename.
        target_file.append(fmt::sprintf("%s%06d.%s", file_prefix.c_str(), next_index, suffix_ptr));
    }

    // Return our path information.
    return target_file.c_str();
}

/*!
 * @brief Clear any pending error status
 */
void imgSaveWorker::clearError()
{
    if (!thread_running)
    {
        last_status = 0;
        save_error = 0;
    }
}

/*!
 * @brief Copy a file from source to target.
 *
 * @param   source_ptr  String with full source path.
 * @param   target_ptr  String with full destination path
 *
 * @return true for success, false for failure
 */
bool imgSaveWorker::copyFile(const char *target_ptr, const char *source_ptr)
{
    // Copy the file.
    return storage::stor_copy_file(target_ptr, source_ptr);
}

/*!
 * @brief Return our completion as a percentage.
 */
uint32_t imgSaveWorker::getCompletion()
{
    uint32_t    percent;

    // Check if our destination is off camera, in which case we don't know anything.
    if ((dest_info & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
    {
        // Nothing to report.
        return 0;
    }

    // Did we report an FFLR error?
    if (save_error != 0)
    {
        return save_error;
    }

    // Make sure we're happy.
    if (last_status == 0)
    {
        if (saver != NULL)
        {
            // It can take a while to finish if we are saving metadata so
            // just lie and say we're stuck at 99%. And if we're not stalled
            // there, this will just be transient anyway.
            percent = saver->getCompletionPercent();
            if (isRunning() && (percent == 100))
                percent = 99;
        }
        // We could have saver==NULL if we already finished a save or we never started one
        // or even for a brief period before our thread has started and initialized.
        // The client could be reading the save state to check completion of a previous
        // save or just to see if there is an unreported error.
        // The old code here could report 0% or 100% depending on whatever start/end
        // information was left lying around, depending on whether any save was ever done
        // since startup. This is a little more explicit about what's going on.
        else if (saver_starting)
        {
            percent = 0;
        }
        else
        {
            // Everything seems to be OK, but our class is NULL. This can happen
            // as we're wrapping things up or long after. Assume we're finished.
            percent = 100;
        }
    }
    else
    {
        // Must have an error, so set the error flag.
        percent = last_status | GIGE_SAVE_STATE_ERR;
    }

    return percent;
}

//--------------------------------------------------------------------
// imgLoadWorker class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. The imgload worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
imgLoadWorker::imgLoadWorker() :
    FAS_THREAD::Thread(true),
    last_status(0)
{
    // Enable tracing
    // traceControl |= TRACE_SEL_CAP_FILE;

    // Initialize our member variables.
    thread_running = false;
    memset(&file_pm_data, 0, sizeof(file_pm_data));

    // Report
    load_progress_valid = false;
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 */
imgLoadWorker::~imgLoadWorker()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

imgLoadWorker *imgLoadWorker::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // If we have an instance, just use it. Otherwise we
    // need to create it.
    if (instance.get() == nullptr)
    {
        // Allocate our instance.
        instance = std::unique_ptr<imgLoadWorker>(new imgLoadWorker());
    }

    // Return our instance pointer.
    return instance.get();
}

/*!
 * @brief Get the thread running.
 */
void imgLoadWorker::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    capfmtResetProgress();

    // Now start the thread.
    FAS_THREAD::Thread::start(threadFn);
}

/*!
 * @brief Ask the thread to stop.
 */
void imgLoadWorker::stop()
{
    // Tell the thread to stop.
    FAS_THREAD::Thread::stop();
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void imgLoadWorker::cleanup()
{
    // We're not running anymore.
    thread_running = false;
}

void imgLoadWorker::setFilename(const char *filename, uint32_t max_length)
{
    (void)max_length;

    // Clear the file name buffer.
    target_file.clear();

    // Copy the name over.
    target_file.assign(filename);
}

void imgLoadWorker::getFilename(char *filename, uint32_t max_length)
{
    // Figure out how many characters we can transfer.
    if (max_length > PATH_MAX)
        max_length = PATH_MAX;

    // Clear the file name buffer.
    memset(filename, 0, max_length);

    // Copy the name over.
    strncpy(filename, target_file.c_str(), max_length);
}

uint32_t imgLoadWorker::getStatus()
{
    uint32_t ret_status;
    std::unique_lock<std::mutex> locker(img_lock);

    // Grab the latest status value.
    ret_status = last_status;
    if (ret_status)
        fi_log(LOG_DEBUG, "%s: status=0x%x", __PRETTY_FUNCTION__, ret_status);

    // Return the status we read.
    return ret_status;
}

/*!
 * @brief Load a video from the SSD to the image memory.
 */
void imgLoadWorker::loadVideo()
{
    // Initialize our info structure.
    format_info.start_frame = 0;
    format_info.end_frame = 0;
    format_info.current_frame = 0;
    format_info.total_bytes = 0;
    format_info.target_path = target_file;
    format_info.do_subfolders = false;
    format_info.subfolder_index = 0;
    format_info.subfolder_count = 0;
    format_info.state = IMAGE_FORMAT::IMAGE_STATE_START;
    format_info.status = 0;

    // Start the loading process.
    if (!capfmtImageLoad())
        fi_log(LOG_WARNING, "%s: Problems loading", __PRETTY_FUNCTION__);

    // Save the status information.
    last_status = format_info.status;
}

/*!
 * @brief Validate the specified partition, making sure it exists and is available.
 *
 * @return true if the partition is valid, false otherwise.
 */
bool imgLoadWorker::validPartition()
{
    uint8_t part_state;
    PartMgr *pm_ptr;
    FiFPGA *fpga_ptr;

    // Get our partition mananger pointer.
    pm_ptr = CameraServer::get_partmgr();
    fpga_ptr = CameraServer::get_fpga();

    // Grab the first partition.
    fpga_ptr->icmResetCaptureFIFO();
    pm_ptr->setActivePartition(0);
    pm_ptr->queueCapturePartition(0);

    // Make sure the partition is available.
    part_state = pm_ptr->getPartitionState();
    if (part_state != GIGE_PART_STATE_EMPTY)
    {
        fi_log(LOG_ERR, "%s: Partition is not empty (%d)!", __FUNCTION__, part_state);
        return false;
    }

    // Everything looks good.
    return true;
}

/*!
 * @brief The worker thread.
 */
void imgLoadWorker::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Make sure the partition we're supposed to use exists and is
    // available.
    if (!validPartition())
        goto load_bail1;

    // We're running.
    thread_running = true;
    last_status = 0;

    // Load the image into our partition.
    loadVideo();

load_bail1:
    // Clean up on the way out.
    cleanup();
}

//--------------------------------------------------------------------
// imgio class implementation
//--------------------------------------------------------------------

const char *imgio::get_suffix(uint32_t stream_type)
{
    const char *suffix_ptr;

    // Figure out the suffix information.
    switch (stream_type)
    {
    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
        suffix_ptr = TIF_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_BMP:
        suffix_ptr = BMP_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_DNG:
        suffix_ptr = DNG_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_JPEG:
        suffix_ptr = JPEG_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_CAPTURE:
    case GIGE_STREAM_FORMAT_CAPSTK:
        suffix_ptr = CAP_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_AVI:
        suffix_ptr = AVI_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        suffix_ptr = BPP_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_MP4:
        suffix_ptr = MP4_SUFFIX;
        break;
    case GIGE_STREAM_FORMAT_PACK:
        suffix_ptr = RDI_SUFFIX;
        break;
    default:
        suffix_ptr = NULL;
        fi_log(LOG_ERR, "%s: Unknown stream type (%d)", __PRETTY_FUNCTION__, stream_type);
        break;
    }

    return suffix_ptr;
}

/*!
 * @brief Return the codec we use for a particular format type.
 */
FAS_VIDEO::VIDEO_CODEC imgio::get_codec(uint32_t dest_info)
{
    FAS_VIDEO::VIDEO_CODEC codec;   // Which codec to use to generated the images.

    // Figure out which codec we want to use.
    switch (dest_info & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_BMP:
        codec = FAS_VIDEO::FI_BMP_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_JPEG:
        codec = FAS_VIDEO::FI_JPG_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
        codec = FAS_VIDEO::FI_TIFF_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_DNG:
        codec = FAS_VIDEO::FI_DNG_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        codec = FAS_VIDEO::FI_PLANAR_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_MP4:
        codec = FAS_VIDEO::FI_H264_CODEC;
        break;
#ifdef NOT_YET_READY
    case GIGE_STREAM_FORMAT_DNGMXF:
        codec = FI_DNG_CODEC;
        break;
#endif
    case GIGE_STREAM_FORMAT_AVI:
        if (imgSaveWorker::getInstance()->getVideoCodec() == GIGE_STREAM_FORMAT_JPEG)
            codec = FAS_VIDEO::FI_MJPEG_JPEG_CODEC;
        else
            codec = FAS_VIDEO::FI_MJPEG_DIB_CODEC;
        break;
    default:
        codec = FAS_VIDEO::FI_INVALID_CODEC;
        break;
    }

    return codec;
}

const char *imgio::get_codec_str(uint32_t stream_type)
{
    const char *ret_str;

    // Get our return string.
    switch (stream_type & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_META_ONLY:  // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
        ret_str = TIFF_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_BMP:
        ret_str = BMP_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        ret_str = BPP_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_DNG:
        ret_str = DNG_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_JPEG:
        ret_str = JPEG_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_AVI:
        // True for AVI/JPEGs and AVI/BMPs.
        if (imgSaveWorker::getInstance()->getVideoCodec() == GIGE_STREAM_FORMAT_JPEG)
            ret_str = AVI_JPEG_CODEC_STR;
        else
            ret_str = AVI_BMP_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_CAPTURE:
        ret_str = RAW_CODEC_STR;
        break;
    case GIGE_STREAM_FORMAT_MP4:
        ret_str = H264_CODEC_STR;
        break;
    default:
        ret_str = UNKNOWN_CODEC_STR;
        break;
    }

    return ret_str;
}

std::unique_ptr<FAS_VIDEO::HEADER_DATA> imgio::get_image_header(uint32_t stream_type, pm_metadata *pm_ptr)
{
    uint32_t    image_type;
    uint32_t    pix_fourcc;
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> image_hdr;
    FAS_VIDEO::VIDEO_CODEC fv_codec = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;

    // Get the image type.
    image_type = get_image_type(stream_type, pm_ptr);
//    LOG("%s: stream_type=0x%x, image_type=0x%x", __PRETTY_FUNCTION__, stream_type, image_type);

    // Now build up the header.
    switch (stream_type & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_JPEG:
    case GIGE_STREAM_FORMAT_AVI:
    case GIGE_STREAM_FORMAT_CAPTURE:
    case GIGE_STREAM_FORMAT_MP4:
        // No file header for these. But, we still want an "empty" header.
        pix_fourcc = FAS_VIDEO::FOURCC_NV12;
        fv_codec = FAS_VIDEO::FI_JPG_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_BMP:
        // Get the BMP header information. We'll get different "types"
        // of headers based on if we're color or mono.
        if ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_RGB24)
            pix_fourcc = FAS_VIDEO::FOURCC_RGB;
        else
            pix_fourcc = FAS_VIDEO::FOURCC_Y8;

        // Codec is a BMP stack.
        fv_codec = FAS_VIDEO::FI_BMP_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_PACK:
    // ==========================================
    // ALM_TODO_PACKED:
    // This is a bit of a hack, but there is an interaction between
    // creating the header and downloading the images that I haven't
    // tracked down yet. For now, create the header and ignore it.
    // ==========================================
    case GIGE_STREAM_FORMAT_META_ONLY:      // Pretend to be TIF.
    case GIGE_STREAM_FORMAT_TIF:
        // Get the TIFF header information. Check to see what kind
        // of header we're interested in.
        switch (image_type & IMGPROC_TYPE_MASK)
        {
        case IMGPROC_TYPE_RAW16:
            pix_fourcc = FAS_VIDEO::FOURCC_BYR2;
            break;
        case IMGPROC_TYPE_RAW8:
            pix_fourcc = FAS_VIDEO::FOURCC_BYR1;
            break;
        case IMGPROC_TYPE_RGB24:
            pix_fourcc = FAS_VIDEO::FOURCC_RGB;
            break;
        case IMGPROC_TYPE_MONO8:
            pix_fourcc = FAS_VIDEO::FOURCC_Y8;
            break;
        default:
            fi_log(LOG_ERR, "%s: Invalid pixel format!", __FUNCTION__);
            pix_fourcc = FAS_VIDEO::FOURCC_NONE;
            break;
        }

        // Get the header.
        if (pix_fourcc != FAS_VIDEO::FOURCC_NONE)
        {
            // Codec is a TIFF stack.
            fv_codec = FAS_VIDEO::FI_TIFF_STACK_CODEC;
        }
        break;
    case GIGE_STREAM_FORMAT_DNG:
        // DNG is strictly a raw format, so we only need to choose between
        // the two types.
        if ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_RAW16)
            pix_fourcc = FAS_VIDEO::FOURCC_BYR2;
        else
            pix_fourcc = FAS_VIDEO::FOURCC_BYR1;

        // Codec is a DNG stack.
        fv_codec = FAS_VIDEO::FI_DNG_STACK_CODEC;
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        // Get the planar header information. We'll get different "types"
        // of headers based on if we're color or mono.
        if (((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_RGB24) ||
            ((image_type & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12))
            pix_fourcc = FAS_VIDEO::FOURCC_NV12;
        else
            pix_fourcc = FAS_VIDEO::FOURCC_Y8;

        // Codec is a planar stack.
        fv_codec = FAS_VIDEO::FI_PLANAR_STACK_CODEC;
        break;

    default:
        // Whatever this was, it isn't legal.
        fi_log(LOG_ERR, "%s: Invalid stream type (0x%x)!", __PRETTY_FUNCTION__, stream_type);
        break;
    }

    if (fv_codec != FAS_VIDEO::FI_INVALID_CODEC)
    {
        // Get the header.
        image_hdr = FAS_VIDEO::FiVideo::CreateImageHeader(fv_codec, pix_fourcc, pm_ptr);
        if (image_hdr == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to get the header information!", __PRETTY_FUNCTION__);
        }
    }

    return image_hdr;
}

uint32_t imgio::get_image_type(uint32_t stream_type, pm_metadata *pm_ptr)
{
    // Get our support class pointers.
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    bool    sensor_is_color;
    bitmode_t   bitmode;

    sensor_is_color = sensor_ptr->is_sensor_color();
    bitmode = pm_ptr->bit_sel;

    return get_imgproc_type(stream_type, bitmode, sensor_is_color);
}
