/*!
 * @file image_io_evm.cxx
 *
 * @brief   This file contains the implementation of the event marker search functions
 *      which run in the context of the imgsave_worker thread because it uses some
 *      of the same resources.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>

#include "camera_server.h"
#include "part_mgr.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_io.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// The maximum buffer size we use to walk through frame metadata.
#define MAX_METADATA_BYTES_PER_READ (1024 * 1024)

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Local functions
//--------------------------------------------------------------------

#define EVENT_FLAG_OFFSET 0
#define EDGE_FLAG_OFFSET  8
static bool match_criteria(uint32_t pattern, uint32_t last_markers, uint32_t cur_markers)
{
    int i;
    int mask(0);
    int index_match(0);
    bool got_a_match(false);

    // Look through each of the indices and see how they compare.
    for (i = 0; i < GIGE_MARKER_NUM_INPUTS; ++i)
    {
        int filter;
        uint32_t    event_mask = 0;
        uint32_t    edge_mask = 0;

        // Build our bitmask.
        event_mask = (1 << (i + EVENT_FLAG_OFFSET));
        edge_mask = (1 << (i + EDGE_FLAG_OFFSET));

        // Look at this event.
        filter = (pattern >> (i * GIGE_MARKER_BITS_PER_FILTER)) & GIGE_MARKER_FILTER_MASK;
        switch (filter)
        {
        default:
        case GIGE_MARKER_FILTER_X:
            break;
        case GIGE_MARKER_FILTER_LOW:
            index_match |= (((cur_markers & event_mask) == 0) << i);
            mask |= (1 << i);
            break;
        case GIGE_MARKER_FILTER_FALL:
            index_match |= ((((cur_markers & edge_mask) == edge_mask) &&
                             ((cur_markers & event_mask) == 0)) << i);
            mask |= (1 << i);
            break;
        case GIGE_MARKER_FILTER_HIGH:
            index_match |= (((cur_markers & event_mask) == event_mask) << i);
            mask |= (1 << i);
            break;
        case GIGE_MARKER_FILTER_RISE:
            index_match |= ((((cur_markers & edge_mask) == edge_mask) &&
                             ((cur_markers & event_mask) == event_mask)) << i);
            mask |= (1 << i);
            break;
        case GIGE_MARKER_FILTER_CHANGE:
            index_match |= (((cur_markers & event_mask) != (last_markers & event_mask)) << i);
            mask |= (1 << i);
            break;
        case GIGE_MARKER_FILTER_GLITCH:
            index_match |= ((((cur_markers & edge_mask) == edge_mask) &&
                             ((cur_markers & event_mask) == (last_markers & event_mask))) << i);
            mask |= (1 << i);
            break;
        }
    }

    // Do we care about any bit?
    if (mask)
    {
        // Determine how we're supposed to combine.
        if ((pattern & GIGE_MARKER_OP_MASK) == GIGE_MARKER_OP_AND)
        {
            // We are AND-ing.
            // First, grab the bits we're interested in (the AND operation), then
            // check to see if ALL the interesting bits are set (1 XOR 1 = 0, 1 XOR 0 = 1).
            got_a_match = (((index_match & mask) ^ mask) == 0);
        }
        else
        {
            // We are OR-ing.
            got_a_match = ((index_match & mask) != 0);
        }
    }

    // Return our decision.
    return got_a_match;
}

//--------------------------------------------------------------------
// Member functions
//--------------------------------------------------------------------

/*!
 * @brief Set the parameters for doing marker pattern match extraction.
 *
 * @param pattern   The filter pattern to match against.
 * @param p_buf     Where to put the results.
 * @param buf_size  The size of the buffer at p_buf.
 * @param p_size    Where to return the list size when work is complete.
 *
 * @return true on success, false otherwise.
 */
void imgSaveWorker::requestMarkers(uint32_t pattern, uint8_t *p_buf, int buf_size, int *p_size)
{
    std::unique_lock<std::mutex> locker(img_lock);

    if ((p_buf != NULL) && (buf_size != 0))
    {
        doing_marker_list = true;
        p_marker_buf = p_buf;
        marker_buf_size = buf_size;
        p_rtn_marker_size = p_size;
        marker_pattern = pattern;
    }
}

/*!
 * @brief Say whether we are currently working on marker list creation.
 *
 * @return true if working on markers, false if doing something else or nothing.
 */
bool imgSaveWorker::isDoingMarkers(void)
{
    bool rtn;
    std::unique_lock<std::mutex> locker(img_lock);

    rtn = doing_marker_list;
    return rtn;
}

/*!
 * @brief initialize class variables for the marker list worker (constructor)
 */
void imgSaveWorker::initMarkerList(void)
{
    doing_marker_list = false;
    p_marker_buf = NULL;
    p_rtn_marker_size = NULL;
    marker_buf_size = 0;
}

/*!
 * @brief clean up class variables for the marker list worker (destructor)
 */
void imgSaveWorker::cleanupMarkerList(void)
{
}

/*!
 * @brief stop the marker list creation
 */
void imgSaveWorker::stopMarkerList(void)
{
    doing_marker_list = false;
}

/*!
 * @brief Start the marker list creation
 */
void imgSaveWorker::createMarkerList(void)
{
    PartMgr  *pm_ptr;
    uint32_t *p_dest;
    uint32_t max_count;
    uint32_t rtn_count;
    uint32_t start_frame;
    uint32_t end_frame;
    uint32_t pattern;
    uint16_t part_num;

    // Get a pointer to our support class.
    pm_ptr = CameraServer::get_partmgr();

    {
        std::unique_lock<std::mutex> locker(img_lock);
        p_dest = (uint32_t *)p_marker_buf;
        max_count = (uint32_t)marker_buf_size / sizeof(uint32_t);
        pattern  = marker_pattern;
    }

    part_num = pm_ptr->getActivePartition();
    start_frame = 0;
    end_frame = pm_ptr->getPartitionNumFrames(part_num) - 1;
    rtn_count = createMarkerList(pattern, p_dest, max_count, part_num, start_frame, end_frame);

    {
        std::unique_lock<std::mutex> locker(img_lock);
        if (!stop_request)
            *p_rtn_marker_size = rtn_count * sizeof(uint32_t);
        doing_marker_list = false;
    }

    fcp_notify_apps(GIGE_REG_MARKER_LIST, *p_rtn_marker_size);
}

uint32_t imgSaveWorker::createMarkerList(uint32_t pattern, uint32_t *p_dest, uint32_t max_count,
                                           uint16_t part_num, uint32_t start, uint32_t end)
{
    int     ma_status;
    uint32_t    frame_no;
    FiFPGA  *fpga_ptr;
    PartMgr *pm_ptr;
    uint32_t    *metadata_buf;
    uint32_t    *p_prev;
    uint32_t    metadata_buf_size;
    uint32_t    frames_per_pass;
    uint32_t    frames_read;
    uint32_t    cur_count;
    uint32_t    cur_start;
    uint32_t    *cur_addr;
    uint32_t    frames_this_pass;
    uint32_t    prev_frame[ FRAME_METADATA_HW_BYTES / sizeof(uint32_t) ];

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    pm_ptr = CameraServer::get_partmgr();

    metadata_buf_size = pm_ptr->get_pme_buf_size();
    if (metadata_buf_size > MAX_METADATA_BYTES_PER_READ)
        metadata_buf_size = MAX_METADATA_BYTES_PER_READ;

    // Get an aligned buffer suitable for DMA.
    ma_status = posix_memalign((void **)&metadata_buf, DMA_PAGE_ALIGN, metadata_buf_size);
    if ((metadata_buf == NULL) || (ma_status != 0))
    {
        syslog(LOG_ERR, "%s: Failed to allocate metadata buffer (status=%d)", __FUNCTION__, ma_status);
        return false;
    }

    frames_per_pass = metadata_buf_size / FRAME_METADATA_HW_BYTES;
    cur_start = start;
    if (start != 0)
        cur_start = start - 1;

    // We need the previous record for testing for changes.
    // For our very first record, this will be either start or ( start - 1 ).
    p_prev = metadata_buf;

    cur_count = 0;
    do
    {
        uint32_t bytes_read;

        if ((end + 1 - cur_start) > frames_per_pass)
            frames_this_pass = frames_per_pass;
        else
            frames_this_pass = end + 1 - cur_start;

        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
            bytes_read = pm_ptr->ltrReadPFMD(part_num, cur_start,
                                             cur_start + frames_this_pass - 1,
                                             (uint8_t *)metadata_buf,
                                             metadata_buf_size);
        else
            // If we use an abort function, pme_read_metdata can return less than we requested
            // even with no error.
            bytes_read = fpga_ptr->pmeRead(part_num, cur_start,
                                           cur_start + frames_this_pass - 1,
                                           (uint8_t *)metadata_buf, metadata_buf_size,
                                           NULL, NULL);

        frames_read = bytes_read / FRAME_METADATA_HW_BYTES;
        if (frames_read != frames_this_pass)
        {
            syslog(LOG_ERR, "%s: reading metadata failed frame_count=%u (%u/%u) but expected %u",
                   __FUNCTION__, frames_read, bytes_read, FRAME_METADATA_HW_BYTES,
                   frames_this_pass);
            last_status = EIO;
            goto bail;
        }

        // Loop through each of the frames. Maybe frames_read is zero, if we got a stop_request.
        cur_addr = metadata_buf;
        for (frame_no = cur_start; frame_no < cur_start + frames_read; frame_no++)
        {
            // If we read one extra before our real start, don't filter on it (cur_start < start).
            // It's just there so our p_prev points to it when we get to the real start.
            if (cur_start >= start)
            {
                if (match_criteria(pattern, p_prev[2], cur_addr[2]))
                {
                    if (cur_count < max_count)
                    {
                        bool handled = false;

                        // We need to add this frame to the list of matches, but squeeze each run
                        // into a sequence of { first, RUN_FLAG, last }.

                        // If it is the first or second one, just copy it, nothing to squeeze yet.
                        if (cur_count <= 1)
                        {
                            p_dest[ cur_count++ ] = frame_no;
                            handled = true;
                        }

                        // We know it is at least the third entry so cur_count-2 makes sense.
                        if (!handled && (p_dest[ cur_count - 1 ] == (frame_no - 1))
                            && (p_dest[ cur_count - 2 ] == (frame_no - 2)))
                        {
                            // Now we know we have a run of three consecutive frame numbers so
                            // it's time to switch to the run encoding.
                            p_dest[ cur_count - 1] = MARKER_LIST_RUN_FLAG;
                            p_dest[ cur_count++ ] = frame_no;
                            handled = true;
                        }

                        // See if it extends the run.
                        if (!handled && (p_dest[cur_count - 1] == (frame_no - 1))
                            && (p_dest[cur_count - 2] == MARKER_LIST_RUN_FLAG))
                        {
                            p_dest[ cur_count - 1 ] = frame_no;
                            handled = true;
                        }

                        // No conditions apply so just copy it over.
                        if (!handled)
                            p_dest[ cur_count++ ] = frame_no;
                    }
                }
            }

            p_prev = cur_addr;
            cur_addr += (FRAME_METADATA_HW_BYTES / sizeof(uint32_t));

            // Early exit if we've filled up the whole buffer already.
            if (cur_count == max_count)
                break;
        }

        // Make a copy of the last record before we overwrite the buffer.
        memcpy((void *)&prev_frame, p_prev, FRAME_METADATA_HW_BYTES);
        p_prev = &prev_frame[0];

        cur_start += frames_read;

        // Early exit if we've filled up the whole buffer already.
        if (cur_count == max_count)
            break;

    } while (cur_start <= end && !stop_request);

    // Mark the one after the last as an invalid frame number, just for fun.
    if (cur_count < max_count)
        p_dest[ cur_count ] = 0xFFFFFFFE;

bail:

    // We're finished with this buffer now.
    free(metadata_buf);

    return cur_count;
}

