/*!
 * @file image_io_xml.cxx
 *
 * @brief This file contains the implementation of the image I/O
 *    class(es) XML functions.
 *
 * This file contains all the routines to handle saving an image
 * to one of the local storage medias.
 *
 * Copyright 2009-2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>

#include <dirent.h>
#include <fcntl.h>
#include <memory.h>

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <libxml/xmlIO.h>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_network.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_io.h"
#include "image_format.h"
#include "part_mgr.h"
#include "storage.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Include the hardware frame number for debug
#define DEBUG_INCLUDE_HW_NUMBER

static const char   *FASTEC_XMLMETA_ROOT    = "FastecMetadata";

static const char   *CAP_XMLMETA_TOP    = "capture";
static const char   *CAP_XMLMETA_VER_STR    = "version";
static const char   *CAP_XMLMETA_VERION = "1.0";

static const char   *IMAGE_XMLMETA_TOP      = "image";
static const char   *IMAGE_XMLMETA_BITMODE      = "bit_mode";
static const char   *IMAGE_XMLMETA_SENSOR_OPTIONS   = "sensor_options";
static const char   *IMAGE_XMLMETA_FRAME_COUNT  = "frame_count";
static const char   *IMAGE_XMLMETA_START_FRAME  = "start_frame";
static const char   *IMAGE_XMLMETA_END_FRAME    = "end_frame";
static const char   *IMAGE_XMLMETA_COMMENT      = "comment";

static const char   *ROI_XMLMETA_TOP    = "roi";
static const char   *ROI_XMLMETA_X      = "x";
static const char   *ROI_XMLMETA_Y      = "y";
static const char   *ROI_XMLMETA_WIDTH  = "width";
static const char   *ROI_XMLMETA_HEIGHT = "height";

static const char   *TRIGGER_XMLMETA_TOP    = "trigger";
static const char   *TRIGGER_XMLMETA_FRAME  = "frame";
static const char   *TRIGGER_XMLMETA_TIME   = "time";

static const char   *CAMERA_XMLMETA_TOP = "camera";
static const char   *CAMERA_XMLMETA_MAKE    = "make";
static const char   *CAMERA_XMLMETA_MODEL   = "model";
static const char   *CAMERA_XMLMETA_FPGA    = "fpga_rev";
static const char   *CAMERA_XMLMETA_SWREV   = "software_version";
static const char   *CAMERA_XMLMETA_MAC = "mac_address";
static const char   *CAMERA_XMLMETA_NAME    = "camera_name";
static const char   *CAMERA_XMLMETA_SENSOR  = "sensor_type";

static const char   *RECORD_XMLMETA_TOP = "record";
static const char   *RECORD_XMLMETA_FPS = "fps";
static const char   *RECORD_XMLMETA_SHUTTER = "shutter_speed";
static const char   *RECORD_XMLMETA_MULTISLOPE  = "multi_slope";
static const char   *RECORD_XMLMETA_SYNC_IN = "sync_in";
static const char   *RECORD_XMLMETA_SYNC_OUT    = "sync_out";

static const char   *NORMAL_XMLMETA_TOP = "normalization";
static const char   *NORMAL_XMLMETA_RED_BAL = "red_balance";
static const char   *NORMAL_XMLMETA_GRN_BAL = "green_balance";
static const char   *NORMAL_XMLMETA_BLU_BAL = "blue_balance";
static const char   *NORMAL_XMLMETA_BRIGHTNESS  = "brightness";
static const char   *NORMAL_XMLMETA_CONTRAST    = "contrast";
static const char   *NORMAL_XMLMETA_GAMMA   = "gamma";
static const char   *NORMAL_XMLMETA_SENSOR_GAIN = "sensor_gain";
static const char   *NORMAL_XMLMETA_RED_GAIN    = "red_gain";
static const char   *NORMAL_XMLMETA_GRN_GAIN    = "green_gain";
static const char   *NORMAL_XMLMETA_BLU_GAIN    = "blue_gain";
static const char   *NORMAL_XMLMETA_RED_MATRIX  = "red_matrix";
static const char   *NORMAL_XMLMETA_GRN_MATRIX  = "green_matrix";
static const char   *NORMAL_XMLMETA_BLU_MATRIX  = "blue_matrix";

static const char   *FRAMES_XMLMETA_TOP     = "frames";
static const char   *FRAME_XMLMETA_TOP      = "frame";
static const char   *FRAME_XMLMETA_NUMBER       = "number";
#ifdef DEBUG_INCLUDE_HW_NUMBER
static const char   *FRAME_XMLMETA_HW_NUMBER    = "hw_number";
#endif
static const char   *FRAME_XMLMETA_IRIG     = "irig";
static const char   *FRAME_XMLMETA_SYNC_MODE    = "sync_mode";
static const char   *FRAME_XMLMETA_TIME_STATE   = "time_state";
static const char   *FRAME_XMLMETA_MARKERS      = "markers";
static const char   *FRAME_XMLMETA_TIME     = "time";

// Defines the old-style bit mode strings for use in the metadata text. New ones are last:first
static const char   *TS3_BIT_MODE_LOWER_8       = "low 8";
static const char   *TS3_BIT_MODE_MIDDLE_8      = "middle 8";
static const char   *TS3_BIT_MODE_UPPER_8       = "upper 8";
static const char   *TS3_BIT_MODE_FULL_10       = "full 10";
static const char   *TS3_BIT_MODE_UNKNOWN       = "unknown";

// The per-frame metadata IRIG decode lock bit.
static const uint32_t   IRIG_DECODE_LOCK         = (1U << 31);

// The per-frame metadata SYNC MODE values.
static const uint32_t   TS3_SYNC_MODE_SHIFT     = 24;
static const uint32_t   TS3_SYNC_MODE_MASK      = (3 << TS3_SYNC_MODE_SHIFT);

// The per-frame metadata TIME STATE values.
static const uint32_t   TS3_TIME_STATE_SHIFT        = 20;
static const uint32_t   TS3_TIME_STATE_MASK     = (7 << TS3_TIME_STATE_SHIFT);

// The maximum buffer size we use to walk through frame metadata.
static const uint32_t   MAX_METADATA_BYTES_PER_READ = 1024 * 1024;

// Extracting the marker information from the per-frame metadata.
// State markers are in bits 0-5.  Edge markers are now nicely placed in bits 8-13.
// We are retaining the unused bits so we have them for debug or for future enhancements;
// the FPGA normally writes zeroes there.
#define MARKERS( x ) (( x ) & 0xffff )

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

static void save_normalization_metadata(xmlTextWriterPtr writer, const pm_metadata *meta_ptr)
{
    uint32_t    red_gain;
    uint32_t    green_gain;
    uint32_t    blue_gain;
    FiFPGA  *fpga_ptr;

    // Get our FPGA pointer.
    fpga_ptr = CameraServer::get_fpga();
    if (fpga_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: Null FPGA pointer!", __FUNCTION__);
        red_gain = meta_ptr->red;
        green_gain = meta_ptr->green;
        blue_gain = meta_ptr->blue;
    }
    else
    {
        red_gain = fpga_ptr->ipmGainRed();
        green_gain = fpga_ptr->ipmGainGreen();
        blue_gain = fpga_ptr->ipmGainBlue();
    }

    // Start with our top level element.
    xmlTextWriterStartElement(writer, BAD_CAST NORMAL_XMLMETA_TOP);

    // Write out our elements
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_RED_BAL, "%hd", meta_ptr->wb_red);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_GRN_BAL, "%hd", meta_ptr->wb_grn);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_BLU_BAL, "%hd", meta_ptr->wb_blu);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_BRIGHTNESS, "%hd", meta_ptr->brightness);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_CONTRAST, "%hd", meta_ptr->contrast);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_GAMMA, "%hd", meta_ptr->gamma);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_SENSOR_GAIN, "%u", meta_ptr->sensor_gain);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_RED_GAIN, "%u.%u", (red_gain >> 8), (((((red_gain & 0xFF) * 10000) / 256) + 5) / 10));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_GRN_GAIN, "%u.%u", (green_gain >> 8), (((((green_gain & 0xFF) * 10000) / 256) + 5) / 10));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_BLU_GAIN, "%u.%u", (blue_gain >> 8), (((((blue_gain & 0xFF) * 10000) / 256) + 5) / 10));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_RED_MATRIX, "[%hd,%hd,%hd]", (int16_t)meta_ptr->matrix_coeff1_red, (int16_t)meta_ptr->matrix_coeff2_red, (int16_t)meta_ptr->matrix_coeff3_red);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_GRN_MATRIX, "[%hd,%hd,%hd]", (int16_t)meta_ptr->matrix_coeff1_grn, (int16_t)meta_ptr->matrix_coeff2_grn, (int16_t)meta_ptr->matrix_coeff3_grn);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST NORMAL_XMLMETA_BLU_MATRIX, "[%hd,%hd,%hd]", (int16_t)meta_ptr->matrix_coeff1_blu, (int16_t)meta_ptr->matrix_coeff2_blu, (int16_t)meta_ptr->matrix_coeff3_blu);

    // End the camera element.
    xmlTextWriterEndElement(writer);
}

static void save_record_metadata(xmlTextWriterPtr writer, const pm_metadata *meta_ptr)
{
    // Start with our top level element.
    xmlTextWriterStartElement(writer, BAD_CAST RECORD_XMLMETA_TOP);

    // Write out our elements
    xmlTextWriterWriteFormatElement(writer, BAD_CAST RECORD_XMLMETA_FPS, "%u", meta_ptr->fps);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST RECORD_XMLMETA_SHUTTER, "%u", meta_ptr->shutter_speed);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST RECORD_XMLMETA_MULTISLOPE, "%hu:%hu", meta_ptr->multslope[0], meta_ptr->multslope[1]);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST RECORD_XMLMETA_SYNC_IN, "0x%x", meta_ptr->sync_in_set);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST RECORD_XMLMETA_SYNC_OUT, "0x%x", meta_ptr->sync_out_set);

    // End the camera element.
    xmlTextWriterEndElement(writer);
}

static void save_camera_metadata(xmlTextWriterPtr writer, const pm_metadata *meta_ptr)
{
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta_ptr is NULL?!", __PRETTY_FUNCTION__);
        return;
    }

    // Start with our top level element.
    xmlTextWriterStartElement(writer, BAD_CAST CAMERA_XMLMETA_TOP);

    // Write out our elements
    xmlTextWriterWriteElement(writer, BAD_CAST CAMERA_XMLMETA_MAKE, BAD_CAST meta_ptr->man_name);
    xmlTextWriterWriteElement(writer, BAD_CAST CAMERA_XMLMETA_MODEL, BAD_CAST meta_ptr->model_name);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST CAMERA_XMLMETA_FPGA, "0x%08x", meta_ptr->fpga_rev);
    xmlTextWriterWriteElement(writer, BAD_CAST CAMERA_XMLMETA_SWREV, BAD_CAST meta_ptr->app_ver);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST CAMERA_XMLMETA_MAC, "%02hx:%02hx:%02hx:%02hx:%02hx:%02hx",
                                    meta_ptr->mac_addr[0], meta_ptr->mac_addr[1], meta_ptr->mac_addr[2],
                                    meta_ptr->mac_addr[3], meta_ptr->mac_addr[4], meta_ptr->mac_addr[5]);
    xmlTextWriterWriteElement(writer, BAD_CAST CAMERA_XMLMETA_NAME, BAD_CAST meta_ptr->camera_name);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST CAMERA_XMLMETA_SENSOR, "%4s", (const char *)(&meta_ptr->sensor_type));

    // End the camera element.
    xmlTextWriterEndElement(writer);
}

/*!
 * @brief Save the Image specific metadata.
 *
 * @param writer    Our XML writer.
 * @param meta_ptr  Our metadata pointer.
 * @param start     The starting frame number.
 * @param end       The ending frame number.
 */
static void save_image_metadata(xmlTextWriterPtr writer, const pm_metadata *meta_ptr, uint32_t start, uint32_t end)
{
    int first = -1;
    int last = -1;
    uint32_t    frame_count;
    uint32_t    trig_frame_num;
    const char  *bit_mode = nullptr;
    struct tm   sbs_tm;

    // Sanity check.
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta_ptr is NULL?!", __PRETTY_FUNCTION__);
        return;
    }

    // Adjust frame count and trigger frame for selected clip
    frame_count = end - start + 1;

    // Now calculate the "relative" trigger offset.
    if (meta_ptr->trig_frame < meta_ptr->first_frame)
        trig_frame_num = meta_ptr->frame_count - (meta_ptr->first_frame - meta_ptr->trig_frame);
    else
        trig_frame_num = meta_ptr->trig_frame - meta_ptr->first_frame;

    // Start with our top level element.
    xmlTextWriterStartElement(writer, BAD_CAST IMAGE_XMLMETA_TOP);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST IMAGE_XMLMETA_FRAME_COUNT, "%u", frame_count);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST IMAGE_XMLMETA_START_FRAME, "%u", start);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST IMAGE_XMLMETA_END_FRAME, "%u", end);
    xmlTextWriterWriteElement(writer, BAD_CAST IMAGE_XMLMETA_COMMENT, BAD_CAST meta_ptr->comment);

    // Make string for bit_sel
    switch (meta_ptr->bit_sel)
    {
    case GIGE_BITMODE_LOWER_8_BITS: bit_mode = TS3_BIT_MODE_LOWER_8; break;
    case GIGE_BITMODE_MIDDLE_8_BITS: bit_mode = TS3_BIT_MODE_MIDDLE_8; break;
    case GIGE_BITMODE_UPPER_8_BITS: bit_mode = TS3_BIT_MODE_UPPER_8; break;
    case GIGE_BITMODE_10_BITS: bit_mode = TS3_BIT_MODE_FULL_10; break;
    case GIGE_BITMODE_7_0:  first = 0; last = 7; break;
    case GIGE_BITMODE_8_1:  first = 1; last = 8; break;
    case GIGE_BITMODE_9_2:  first = 2; last = 9; break;
    case GIGE_BITMODE_10_3: first = 3; last = 10; break;
    case GIGE_BITMODE_11_4: first = 4; last = 11; break;
    case GIGE_BITMODE_9_0:  first = 0; last = 9; break;
    case GIGE_BITMODE_10_1: first = 1; last = 10; break;
    case GIGE_BITMODE_11_2: first = 2; last = 11; break;
    case GIGE_BITMODE_11_0: first = 0; last = 11; break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected bit_sel (%d)", __PRETTY_FUNCTION__, meta_ptr->bit_sel);
        break;
    }

    // Did we get a known format?
    if (bit_mode != nullptr)
        xmlTextWriterWriteElement(writer, BAD_CAST IMAGE_XMLMETA_BITMODE, BAD_CAST bit_mode);
    else if ((first >= 0) && (last >= 0))
        xmlTextWriterWriteFormatElement(writer, BAD_CAST IMAGE_XMLMETA_BITMODE, "%u:%u", last, first);
    else
        xmlTextWriterWriteElement(writer, BAD_CAST IMAGE_XMLMETA_BITMODE, BAD_CAST TS3_BIT_MODE_UNKNOWN);

    xmlTextWriterWriteFormatElement(writer, BAD_CAST IMAGE_XMLMETA_SENSOR_OPTIONS,
                                    "bin%ux:subs%ux",
                                    (meta_ptr->sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT,
                                    (meta_ptr->sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT);

    // Write our ROI elements.
    xmlTextWriterStartElement(writer, BAD_CAST ROI_XMLMETA_TOP);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST ROI_XMLMETA_X, "%u", meta_ptr->roi_x);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST ROI_XMLMETA_Y, "%u", meta_ptr->roi_y);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST ROI_XMLMETA_WIDTH, "%u", meta_ptr->roi_width);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST ROI_XMLMETA_HEIGHT, "%u", meta_ptr->roi_height);
    xmlTextWriterEndElement(writer);

    // Add the trigger information.  Start with the top level element.
    xmlTextWriterStartElement(writer, BAD_CAST TRIGGER_XMLMETA_TOP);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST TRIGGER_XMLMETA_FRAME, "%u", trig_frame_num);

    // Calculate time/date of trigger
    sbs_to_tm(meta_ptr->trig_time_s, &sbs_tm);
    xmlTextWriterWriteFormatElement(writer, BAD_CAST TRIGGER_XMLMETA_TIME, "%02d:%03d:%02d:%02d:%02d.%06d",
                                    sbs_tm.tm_year, sbs_tm.tm_yday + 1, sbs_tm.tm_hour, sbs_tm.tm_min, sbs_tm.tm_sec, ((meta_ptr->trig_time_ns + 500) / 1000));

    // End the trigger element.
    xmlTextWriterEndElement(writer);

    // End the image element.
    xmlTextWriterEndElement(writer);
}

/*!
 * @brief Save the partition metadata.
 *
 * @param writer    Our XML writer.
 * @param meta_ptr  Our partition metadata.
 *
 * @return true on success, false otherwise.
 */
static bool save_capture_metadata(xmlTextWriterPtr writer, const pm_metadata *meta_ptr, uint32_t start, uint32_t end)
{
    int rc;

    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta_ptr is NULL?!", __PRETTY_FUNCTION__);
        return false;
    }

    // Start with our top level element.
    rc = xmlTextWriterStartElement(writer, BAD_CAST CAP_XMLMETA_TOP);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error writing top (%d)", __FUNCTION__, rc);
        return false;
    }

    // Add a version attribute attribute.
    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST CAP_XMLMETA_VER_STR, BAD_CAST CAP_XMLMETA_VERION);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error writing version attribute (%d)", __FUNCTION__, rc);
        return false;
    }

    // Add the image data.
    save_image_metadata(writer, meta_ptr, start, end);

    // Add the camera specific metadata.
    save_camera_metadata(writer, meta_ptr);

    // Add the recording information.
    save_record_metadata(writer, meta_ptr);

    // Add the normalization information.
    save_normalization_metadata(writer, meta_ptr);

    // End the capture element.
    rc = xmlTextWriterEndElement(writer);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error closing capture element (%d)", __FUNCTION__, rc);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Write the per-frame metadata for a single frame.
 *
 * @param writer    Our XML writer.
 * @param frame_words   The raw per-frame metadata.
 * @param frame_no  The frame number.
 *
 * @return true on success, false otherwise.
 */
static bool save_one_frame_metadata(xmlTextWriterPtr writer, uint32_t *frame_words, uint32_t frame_no)
{
    int     rc;
    struct tm   sbs_tm;

    // Convert the SBS to something we can use.
    sbs_to_tm(frame_words[0], &sbs_tm);

    // Start with our top level element.
    xmlTextWriterStartElement(writer, BAD_CAST FRAME_XMLMETA_TOP);

    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_NUMBER, "%u", frame_no);
#ifdef DEBUG_INCLUDE_HW_NUMBER
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_HW_NUMBER, "%u", frame_words[3]);
#endif
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_IRIG, "%d",
                                    ((frame_words[1] & IRIG_DECODE_LOCK) == IRIG_DECODE_LOCK));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_SYNC_MODE, "%d",
                                    ((frame_words[1] & TS3_SYNC_MODE_MASK) >> TS3_SYNC_MODE_SHIFT));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_TIME_STATE, "%d",
                                    ((frame_words[1] & TS3_TIME_STATE_MASK) >> TS3_TIME_STATE_SHIFT));
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_MARKERS, "0x%04x",
                                    MARKERS(frame_words[2]));
    // Linux tm day-of-year is zero based, but we report days starting with 1=Jan 1.
    xmlTextWriterWriteFormatElement(writer, BAD_CAST FRAME_XMLMETA_TIME, "%02d:%03d:%02d:%02d:%02d.%06d",
                                    sbs_tm.tm_year, sbs_tm.tm_yday + 1, sbs_tm.tm_hour, sbs_tm.tm_min,
                                    sbs_tm.tm_sec, (frame_words[ 1 ] & 0xFFFFF));
    rc = xmlTextWriterEndElement(writer);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error closing frame element (%d)", __FUNCTION__, rc);
        return false;
    }

    return true;
}

/*!
 * @brief Extract and save the per-frame metadata for a capture.
 *
 * @param destptr   Our XML writer, cast to avoid dragging in XML header files elsewhere.
 * @param start     Starting frame number.
 * @param end       Ending frame number, inclusive.
 * @param part_num  Partition number that is being saved.
 * @param use_p2    Because partmgr functions want to know P1 vs P2.
 *
 * @return true on success, false otherwise.
 */
bool imgSaveWorker::saveFrameMetadata(void *destptr, uint32_t start, uint32_t end, uint16_t part_num)
{
    xmlTextWriterPtr writer = (xmlTextWriterPtr)destptr;
    int rc;
    int frames_transferred; // Number of frames we transferred in a loop.
    int bb_index;           // Bounce buffer index.
    uint8_t *bb_ptr;
    uint32_t *cur_addr;
    uint32_t cur_frame = 0;
    uint32_t frame_no;
    uint32_t metadata_words_per_frame;
    FiFPGA *fpga_ptr;
    PartMgr *pm_ptr;

    fi_log(LOG_DEBUG, "%s: start=%u, end=%u, part_num=%u", __PRETTY_FUNCTION__, start, end, part_num);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    pm_ptr = CameraServer::get_partmgr();

    // Initialize our frame information.
    frames_transferred = 0;
    metadata_words_per_frame = FRAME_METADATA_HW_BYTES / sizeof(uint32_t);

    // Get our bounce buffer.
    bb_index = bb_get_avail();

    // Get a pointer to our buffer.
    bb_ptr = (uint8_t *)bb_mmap(bb_index, MAX_METADATA_XFER_LEN);
    if (bb_ptr == (uint8_t *)MAP_FAILED)
    {
        fi_log(LOG_ERR, "%s: Unable to map the bounce buffer [index=%d]", __PRETTY_FUNCTION__, bb_index);
        goto bail;
    }

    // Start with our top level element.
    rc = xmlTextWriterStartElement(writer, BAD_CAST FRAMES_XMLMETA_TOP);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error writing top (%d)", __FUNCTION__, rc);
        goto bail;
    }

    // Loop until we've sent all the frames of metadata.
    for (cur_frame = start; !stop_request && (cur_frame < end); cur_frame += frames_transferred)
    {
        fi_log(LOG_DEBUG, "%s: cur_frame=%u/%u, frames_transferred=%d", __PRETTY_FUNCTION__, cur_frame, end, frames_transferred);

        // Get our bounce buffer filled.
        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            frames_transferred = pm_ptr->pme_ltr_fill_buffer(cur_frame, end, bb_ptr);
        }
        else
        {
            frames_transferred = pm_ptr->pme_fpga_fill_buffer(cur_frame, end, bb_index);
        }

        // Did the fill go OK?
        if (frames_transferred == 0)
        {
            fi_log(LOG_ERR, "%s: Unable to fill the transfer buffer!", __PRETTY_FUNCTION__);
            break;
        }

// #define DUMP_RAW_METADATA 1
#ifdef DUMP_RAW_METADATA
        {
            int fd;
            char dump_path[128];
            snprintf(dump_path, 127, "/tmp/dump-%u.dat", cur_frame);
            fd = open(dump_path, O_CREAT | O_WRONLY, 0666);
            write(fd, bb_ptr, frames_transferred * FRAME_METADATA_HW_BYTES);
            close(fd);
        }
#endif /* DUMP_RAW_METADATA */

        // Loop through each of the frames. Maybe frames_read is zero, if we got a stop_request.
        cur_addr = (uint32_t *)bb_ptr;
        for (frame_no = cur_frame; !stop_request && (frame_no < (cur_frame + frames_transferred)); frame_no++)
        {
            if (!save_one_frame_metadata(writer, cur_addr, frame_no))
            {
                fi_log(LOG_ERR, "%s: Error writing frame #%u metadata", __FUNCTION__, frame_no);
                goto bail;
            }

            // Update our address.
            cur_addr += metadata_words_per_frame;
        }
    }

bail:
    // Release our buffer. This will also unmap the buffer, if appropriate.
    bb_release(bb_index);

    // End the frame element.
    rc = xmlTextWriterEndElement(writer);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error closing frame element (%d)", __FUNCTION__, rc);
        return false;
    }

    // All is well, but we might not have saved all frames if we got a stop request.
    if (cur_frame <= end)
    {
        if (stop_request)
        {
            fi_log(LOG_INFO, "%s: Canceled frame metadata save after %u frames", __PRETTY_FUNCTION__, cur_frame);
        }
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Save the metadata information as an XML formatted file.
 *
 * @param dest_ptr  Where to put the data (Fully qualified path name).
 * @param meta_ptr  The capture metadata.
 * @param part_num  Partition number that is being saved.
 *
 * @return true on success, false otherwise.
 */
bool imgSaveWorker::saveXmlMetadata(const char *dest_ptr, const pm_metadata *meta_ptr, uint16_t part_num)
{
    int rc;
    bool rtn = true;
    xmlTextWriterPtr writer;

    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta_ptr is NULL?!", __PRETTY_FUNCTION__);
        return false;
    }

    // Create our XmlWriter for the file.
    writer = xmlNewTextWriterFilename(dest_ptr, 0);
    if (writer == NULL)
    {
        fi_log(LOG_ERR, "%s: Error creating the XML writer, err=%d(%s)!", __FUNCTION__, errno, strerror(errno));
        return false;
    }

    // Start the document with the XML defaults for the version,
    // encoding and standalone declaration.
    rc = xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error with starting the doc (%d)", __FUNCTION__, rc);
        xmlFreeTextWriter(writer);
        return false;
    }

    // Add the root element of the document.
    rc = xmlTextWriterStartElement(writer, BAD_CAST FASTEC_XMLMETA_ROOT);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error writing root (%d)", __FUNCTION__, rc);
        xmlFreeTextWriter(writer);
        return false;
    }

    // Save the capture settings metadata.
    if (!save_capture_metadata(writer, meta_ptr, getStartFrameNumber(),
                               getEndFrameNumber()))
    {
        fi_log(LOG_ERR, "%s: Error writing capture metadata", __FUNCTION__);
        xmlFreeTextWriter(writer);
        return false;
    }

    // Now save all the per-frame metadata information.
    // If we stopped on a cancel, we will still write out
    // the end elements to create a valid file even if it
    // is incomplete. If there was no stop_request, figure
    // that some other problem occurred and bail immediately.
    if (!saveFrameMetadata(writer, getStartFrameNumber(), getEndFrameNumber(), part_num))
    {
        if (!stop_request)
        {
            xmlFreeTextWriter(writer);
            return false;
        }
        rtn = false;
    }

    // End the root element.
    rc = xmlTextWriterEndElement(writer);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error closing root element (%d)", __FUNCTION__, rc);
        xmlFreeTextWriter(writer);
        return false;
    }

    // Close the document.
    rc = xmlTextWriterEndDocument(writer);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Error ending the document (%d)", __FUNCTION__, rc);
        xmlFreeTextWriter(writer);
        return false;
    }

    // Release our writer.
    xmlFreeTextWriter(writer);

    // All done.
    return rtn;
}
