/*!
 * @file image_stream_save.cxx
 *
 * @brief This file contains the code which implements saving
 *    GIGE_STREAM_FORMAT_xxx files.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>

#include <dirent.h>
#include <fcntl.h>
#include <memory.h>

#include <string>
#include <fmt/printf.h>
#include <fmt/format.h>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "fcp.h"
#include "jpeg_util.h"
#include "part_mgr.h"
#include "image_format.h"
#include "image_io.h"
#include "save_proc.h"

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

//! Generic filename buffer.
std::string StreamSave::dest_name;

/*!
 * @brief Remove a directory on a local file system if it is empty.
 *
 * This is a best-effort to clean up after an error. We don't really care
 * if it fails, except maybe for post-mortem troubleshooting.
 *
 * @param dir_path  The directory to scan and possibly delete.
 */
static void rmdir_if_empty(const char *dir_path)
{
    DIR         *dir_ptr;
    bool        found_anything;

    // Open the folder. If that fails, log it and return;
    dir_ptr = opendir(dir_path);
    if (dir_ptr == NULL)
    {
        fi_log(LOG_WARNING, "%s: opendir(%s) failed. %d(%s)", __FUNCTION__,
               dir_path, errno, strerror(errno));
        return;
    }

    // Scan the directory for any reason to keep it.
    found_anything = false;
    do
    {
        struct dirent   *dir_info;

        // Get the next directory entry.
        errno = 0;
        dir_info = readdir(dir_ptr);
        if (dir_info == NULL)
        {
            // Was there an error? If so, err on the side of caution and don't
            // delete the directory.
            if (errno != 0)
            {
                fi_log(LOG_WARNING, "%s: opendir(%s) failed. %d(%s)", __FUNCTION__,
                       dir_path, errno, strerror(errno));
                found_anything = true;
            }
            // Otherwise we are done and we didn't find any files or subdirs.
            break;
        }

        // Check for anything other than "." or "..".
        if ((strcmp(dir_info->d_name, ".") != 0) &&
            (strcmp(dir_info->d_name, "..") != 0))
            found_anything = true;

    } while (!found_anything);

    // Clean up.
    closedir(dir_ptr);

    // If the directory is empty, do a best-effort rmdir.
    if (!found_anything)
    {
        if (rmdir(dir_path) < 0)
            fi_log(LOG_WARNING, "%s: rmdir(%s) failed. %d(%s)", __FUNCTION__,
                   dir_path, errno, strerror(errno));
    }
}

/*!
 * @brief Save pm data to a .txt file and a Xcitex .cfg.
 */
void StreamSave::saveMetadata()
{
    int32_t meta_length;
    uint32_t format_type;
    FiFPGA  *fpga_ptr;
    PartMgr *part_mgr;
    FAS_VIDEO::VIDEO_GAINS gain_info;
    std::string metadataString;     // The user text friendly metadata string.
    std::unique_ptr<const pm_metadata> meta_ptr;

    // The xcite configuration record using our camera metadata in bytes.
    std::unique_ptr<FAS_VIDEO::X_CFG_METADATA_VECTOR> xcfgMetadataInBytes;

    // If we're doing subfolders, back up to the root.
    if (format_info.do_subfolders)
    {
        if (format_info.subfolder_index == 0)
        {
            fi_log(LOG_ERR, "%s: subfolder_ptr is NULL?", __PRETTY_FUNCTION__);
            return;
        }

        // Clear the path.
        format_info.target_path[format_info.subfolder_index] = '\0';
    }

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    part_mgr = CameraServer::get_partmgr();

    // Get the format of the data we're saving.
    format_type = destination_info & GIGE_STREAM_FORMAT_MASK;

    // Get the metadata information from the partition manager.
    meta_ptr = part_mgr->getReviewPartMetadata();
    if (meta_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: NULL metadata pointer?!", __PRETTY_FUNCTION__);
        return;
    }

    // Are we supposed to save out the CFG metadata?
    if (imgSaveWorker::getInstance()->saveCfgMetadata())
    {
        // Get the Xcitex metadata buffer.
        meta_length = FAS_VIDEO::FiVideo::CreateX_CFG_MetadataString(xcfgMetadataInBytes, (void *)meta_ptr.get(), sizeof(pm_metadata), format_info.start_frame, format_info.end_frame);
        if (meta_length <= 0)
        {
            fi_log(LOG_ERR, "%s: Unable to create Xcitex metadata (%d)!", __FUNCTION__, meta_length);
            format_info.status = -meta_length;
            return;
        }

        // Build up the target file name.
        StreamSave::dest_name.assign(format_info.target_path);
        if ((format_type == GIGE_STREAM_FORMAT_AVI) || (format_type == GIGE_STREAM_FORMAT_MP4) || (format_type == GIGE_STREAM_FORMAT_CAPTURE) || (format_type == GIGE_STREAM_FORMAT_CAPSTK))
        {
            StreamSave::dest_name.replace(StreamSave::dest_name.length() - 3, 3, "cfg");
        }
        else
        {
            StreamSave::dest_name.append("/metadata.cfg");
        }

        // Save the metadata out using the FIV interface.
        saveFivMetadata(StreamSave::dest_name.c_str(), reinterpret_cast<const char *>(xcfgMetadataInBytes->data()), meta_length);
    }

    // Initialize the gain information structure with the latest settings.
    gain_info.red = (uint32_t)fpga_ptr->ipmGainRed();
    gain_info.green = (uint32_t)fpga_ptr->ipmGainGreen();
    gain_info.blue = (uint32_t)fpga_ptr->ipmGainBlue();

    // Get the Fastec metadata buffer.
    meta_length = FAS_VIDEO::FiVideo::CreateMetadataString(metadataString, meta_ptr.get(), sizeof(pm_metadata), format_info.start_frame, format_info.end_frame, &gain_info);
    if (meta_length <= 0)
    {
        fi_log(LOG_ERR, "%s: Unable to create Fastec metadata (%d)!", __FUNCTION__, meta_length);
        format_info.status = -meta_length;
        return;
    }

    // Add on our "extras".
    if ((format_type == GIGE_STREAM_FORMAT_CAPTURE) || (format_type == GIGE_STREAM_FORMAT_CAPSTK))
    {
        metadataString.append("\traw=1\n\tcodec=CAP\n");
    }
    else
    {
        bool is_raw;

        is_raw = ((destination_info & GIGE_STREAM_FORMAT_RAW) != 0);
        if ((destination_info & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_DNG)
            is_raw  = true;

        metadataString.append(fmt::sprintf("\traw=%d\n\tcodec=%s",
                                           is_raw, imgio::get_codec_str(destination_info & (GIGE_STREAM_FORMAT_MASK | GIGE_STREAM_FORMAT_RAW))));
        if (format_type == GIGE_STREAM_FORMAT_MP4)
        {
            std::string h264SpeedText;
            FAS_VIDEO::FiVideo::H264SpeedSettingToText(h264SpeedText, h264_speed);
            metadataString.append(fmt::sprintf(":speed=%s:quality=%d:keyframes=%s", h264SpeedText.c_str(), h264_quality, h264_enable_all_keyframes ? "All" : "60"));
        }

        metadataString.append("\n");
    }

    // Build up the target file name.
    StreamSave::dest_name.assign(format_info.target_path);
    if ((format_type == GIGE_STREAM_FORMAT_AVI) || (format_type == GIGE_STREAM_FORMAT_MP4) || (format_type == GIGE_STREAM_FORMAT_CAPTURE) || (format_type == GIGE_STREAM_FORMAT_CAPSTK))
    {
        StreamSave::dest_name.replace(StreamSave::dest_name.length() - 3, 3, "txt");
    }
    else
    {
        StreamSave::dest_name.append("/metadata.txt");
    }

    // Save the metadata out using the FIV interface.
    saveFivMetadata(StreamSave::dest_name.c_str(), reinterpret_cast<const char *>(metadataString.data()), metadataString.size());
}

/*!
 * @brief Save the per-frame metadata.
 *
 * The data is saved to either the path in format_info.target_path, with a ".pm" suffix, or to the "metadata.pm" file.
 * @return True if successful, false otherwise.
 */
bool StreamSave::savePfmd()
{
    bool rc(false);
    int fd(-1);
    int frames_transferred(0);  // Number of frames we transferred in a loop.
    int bb_index(-1);           // Bounce buffer index.
    uint32_t start_frame;
    uint32_t end_frame;
    uint8_t *bb_ptr;
    uint32_t format_type;
    uint32_t cur_frame(0);
    FiFPGA *fpga_ptr;
    PartMgr *part_mgr;

    // Get the format of the data we're saving.
    format_type = destination_info & GIGE_STREAM_FORMAT_MASK;

    // No need to save the metadata if we're only saving a single capture.
    if (format_type == GIGE_STREAM_FORMAT_CAPTURE)
        return true;

    // If we're doing subfolders, back up to the root.
    if (format_info.do_subfolders)
    {
        if (format_info.subfolder_index == 0)
        {
            fi_log(LOG_ERR, "%s: subfolder_ptr is NULL?", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Clear the path.
        format_info.target_path[format_info.subfolder_index] = '\0';
    }

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    part_mgr = CameraServer::get_partmgr();

    // Build up the target file name.
    StreamSave::dest_name.assign(format_info.target_path);
    if ((format_type == GIGE_STREAM_FORMAT_AVI) || (format_type == GIGE_STREAM_FORMAT_MP4) || (format_type == GIGE_STREAM_FORMAT_CAPTURE) || (format_type == GIGE_STREAM_FORMAT_CAPSTK))
    {
        StreamSave::dest_name.replace(StreamSave::dest_name.length() - 3, 3, "pm");
    }
    else
    {
        StreamSave::dest_name.append("/metadata.pm");
    }

    // Open the target file.
    fd = open(StreamSave::dest_name.c_str(), O_CREAT | O_WRONLY, 0666);
    if (fd < 0)
    {
        fi_log(LOG_DEBUG, "%s: Unable to open '%s' [%d: %s]", __PRETTY_FUNCTION__, StreamSave::dest_name.c_str(), errno, strerror(errno));
        goto bail;
    }

    // Get our bounce buffer.
    bb_index = bb_get_avail();

    // Get a pointer to our buffer.
    bb_ptr = (uint8_t *)bb_mmap(bb_index, MAX_METADATA_XFER_LEN);
    if (bb_ptr == (uint8_t *)MAP_FAILED)
    {
        fi_log(LOG_ERR, "%s: Unable to map the bounce buffer [index=%d]", __PRETTY_FUNCTION__, bb_index);
        goto bail;
    }

    // Are we in FFLR mode?
    if (fpga_ptr->ltrFasFireEnabled())
    {
        start_frame = 0;
        end_frame = part_mgr->getReviewPartNumFrames() - 1;
    }
    else
    {
        start_frame = format_info.start_frame;
        end_frame = format_info.end_frame;
    }
    fi_log(LOG_DEBUG, "%s: part=%d, start_frame=%u, end_frame=%u, stop_request=%d", __PRETTY_FUNCTION__, part_mgr->getReviewPartition(), start_frame, end_frame, stop_request);

    // Loop until we've sent all the frames of metadata.
    frames_transferred = 0;
    for (cur_frame = start_frame; !stop_request && (cur_frame < end_frame); cur_frame += frames_transferred)
    {
        fi_log(LOG_DEBUG, "%s: part=%d, cur_frame=%u/%u, frames_transferred=%d", __PRETTY_FUNCTION__, part_mgr->getReviewPartition(), cur_frame, end_frame, frames_transferred);

        // Get our bounce buffer filled.
        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            frames_transferred = part_mgr->pme_ltr_fill_buffer(cur_frame, end_frame, bb_ptr);
        }
        else
        {
            frames_transferred = part_mgr->pme_fpga_fill_buffer(cur_frame, end_frame, bb_index);
        }

        // Did the fill go OK?
        if (frames_transferred == 0)
        {
            fi_log(LOG_ERR, "%s: Unable to fill the transfer buffer!", __PRETTY_FUNCTION__);
            break;
        }

        // Write out the file.
        write(fd, bb_ptr, frames_transferred * FRAME_METADATA_HW_BYTES);
    }

bail:
    // Release our buffer. This will also unmap the buffer, if appropriate.
    if (bb_index >= 0)
        bb_release(bb_index);

    // Close the file, if it is open.
    if (fd >= 0)
        close(fd);

    // All is well, but we might not have saved all frames if we got a stop request.
    if (cur_frame <= format_info.end_frame)
    {
        if (stop_request)
        {
            fi_log(LOG_INFO, "%s: Canceled frame metadata save after %u frames", __PRETTY_FUNCTION__, cur_frame);
        }
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Determine if we need to create a subfolder at this point in the
 *    save process. Create the subfolder if needed.
 *
 * @return true on success, false otherwise.
 */
bool StreamSave::checkSubfolder()
{
    uint32_t    check_val;

    // Are we doing subfolders?
    if (!format_info.do_subfolders)
        return true;

    // Do we want to create a new subfolder?
    check_val = ((format_info.current_frame - format_info.start_frame) / MAX_FOLDER_COUNT);
    if ((check_val + 1) > format_info.subfolder_count)
    {
        int status;
        std::string subfolder_str;

        // Build our new subfolder.
        subfolder_str = fmt::sprintf("/img%04d", format_info.subfolder_count);
        format_info.target_path.replace(format_info.subfolder_index, std::string::npos, subfolder_str);
        format_info.subfolder_count++;

        // Make our new folder.
        status = mkdir(format_info.target_path.c_str(), S_IRWXU | S_IRWXO | S_IRWXG);
        if (status == -1)
        {
            format_info.status = errno;
            fi_log(LOG_ERR, "%s: Can't create '%s', err=%d(%s)", __FUNCTION__, format_info.target_path.c_str(), errno, strerror(errno));
            return false;
        }
    }

    // All is well.
    return true;
}

/*!
 * @brief IMAGE_STATE_START: In this state we need to start
 *    the save operation, and then transition to the
 *    next state - IMAGE_STATE_IN_PROGRESS on success or
 *    IMAGE_STATE_DONE if there are problems.
 *
 * @return true if we're still working, false otherwise.
 */
bool StreamSave::startupState(void *user_data)
{
    StreamSave *save_ptr = (StreamSave *)user_data;
    bool still_working;

    // Initialize the system.
    still_working = save_ptr->startFivSave();
    if (still_working)
    {
        // Move to the processing state.
        save_ptr->format_info.state = IMAGE_FORMAT::IMAGE_STATE_IN_PROGRESS;
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to start the save.", __FUNCTION__);
        save_ptr->format_info.state = IMAGE_FORMAT::IMAGE_STATE_DONE;
    }

    // Return our status.
    return still_working;
}

/*!
 * @brief IMAGE_STATE_IN_PROGRESS: In this state we need to keep
 *    pushing the ball down the road. Create all the files
 *    and associated SATA commands.
 *
 * @return true if we're still working, false otherwise.
 */
bool StreamSave::progressState(void *user_data)
{
    bool    rc;
    StreamSave *save_ptr = (StreamSave *)user_data;

    // Process the "stream" of frames, until we're done.
    rc = save_ptr->processFivImage();

    // Did the process call fail?
    if (!rc)
    {
        // No error if we're told to stop.
        if (!save_ptr->stop_request)
        {
            fi_log(LOG_ERR, "%s: Error %d processing frames!", __FUNCTION__, errno);
            save_ptr->format_info.status = errno;
        }
    }

    // The "stream" is done, so do any cleanup necessary.
    save_ptr->format_info.state = IMAGE_FORMAT::IMAGE_STATE_FINISH;

    // Still working if successful so far.
    return rc;
}

/*!
 * @brief IMAGE_STATE_FINISH: In this state, we determine if
 *    we're done transferring all the frames and proceed
 *    accordingly.
 *
 * @return false. We are done.
 */
bool StreamSave::finishState(void *user_data)
{
    StreamSave *save_ptr = (StreamSave *)user_data;

    // Make sure our next state is done.
    save_ptr->format_info.state = IMAGE_FORMAT::IMAGE_STATE_DONE;

    // Check to see if we are done.
    if (save_ptr->format_info.current_frame >= save_ptr->format_info.end_frame)
    {
        // We don't save any additional metadata with the packed format.
        if ((save_ptr->destination_info & GIGE_STREAM_FORMAT_MASK) != GIGE_STREAM_FORMAT_PACK)
        {
            // Check to see if we're being asked to stop.
            if (!save_ptr->stop_request && imgSaveWorker::getInstance()->savePFMD())
            {
                // Save the PFMD.
                save_ptr->savePfmd();
            }

            // If we're not being asked to stop...
            if (!save_ptr->stop_request)
            {
                // Save the metadata information.
                save_ptr->saveMetadata();
            }
        }
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to transfer all frames [curr=%d, end=%d].",
               __PRETTY_FUNCTION__, save_ptr->format_info.current_frame, save_ptr->format_info.end_frame);

        // Delete empty directories. The target_path might include a subfolder,
        // so we delete that first.
        rmdir_if_empty(save_ptr->format_info.target_path.c_str());

        if (save_ptr->format_info.do_subfolders && (save_ptr->format_info.subfolder_index != 0))
        {
            StreamSave::dest_name = save_ptr->format_info.target_path.substr(0, save_ptr->format_info.subfolder_index);
            rmdir_if_empty(StreamSave::dest_name.c_str());
        }
    }

    // We are done.
    return false;
}

/*!
 * @brief IMAGE_STATE_DONE: This is the end state. We are done.
 *
 * @return true if we're still working, false otherwise.
 */
bool StreamSave::doneState(void *user_data)
{
    StreamSave *save_ptr = (StreamSave *)user_data;

    // Make sure we close down the save processing object.
    saveProcClose();

    // Release the bounce buffer.
    if (save_ptr->bb_index >= 0)
    {
        bb_release(save_ptr->bb_index);
        save_ptr->bb_index = -1;
    }

    // All done.
    return false;
}

void StreamSave::stop()
{
    // Ask the thread to stop.
    stop_request = true;

    // Close down the save processor.
    saveProcCancel();
}

/*!
 * @brief Initialize our class members.
 */
void StreamSave::initClass()
{
    // Setup our state machine.
    save_states[IMAGE_FORMAT::IMAGE_STATE_START] = startupState;
    save_states[IMAGE_FORMAT::IMAGE_STATE_IN_PROGRESS] = progressState;
    save_states[IMAGE_FORMAT::IMAGE_STATE_FINISH] = finishState;
    save_states[IMAGE_FORMAT::IMAGE_STATE_DONE] = doneState;

    // Initialize our other variables.
    format_info.start_frame = 0;
    format_info.end_frame = 0;
    format_info.current_frame = 0;
    format_info.total_bytes = 0;
    format_info.target_path.clear();
    format_info.do_subfolders = false;
    format_info.subfolder_index = 0;
    format_info.subfolder_count = 0;
    format_info.state = IMAGE_FORMAT::IMAGE_STATE_START;
    format_info.status = 0;

    stop_request = false;
    destination_info = 0;

    streamHandle = 0;
}

/*!
 * @brief The default constructor, no parameters.
 */
StreamSave::StreamSave()
{
    // Initialize the class structures.
    initClass();
};

/*!
 * @brief Constructor with our destination information as a parameter.
 */
StreamSave::StreamSave(uint32_t dest_info)
{
    // Do default initialization.
    initClass();

    // Save the destination information.
    destination_info = dest_info;
}

/*!
 * @brief Standard destructor.
 */
StreamSave::~StreamSave()
{
}
