/*!
 * @file ipc.cxx
 * @brief This file contains the implementation of the cs_ipc class.
 *
 * This file contains all the routines to deal with IPC
 * IPC is used to communicate with internal processes wanting
 * to integrate with the camera. Right now the IPC choosen
 * is UNIX UDP domain sockets.
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <time.h>
#include <sys/ipc.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <fas_log.h>

#include "cs_network.h"
#include "cs_utils.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Derived from time.h (for the kernel)
#define NSEC_PER_USEC       1000L
#define NSEC_PER_SEC        1000000000L
#define TIMEVAL_TO_NS(tv)   (((int64_t) tv.tv_sec * NSEC_PER_SEC ) + tv.tv_nsec )

//!< Path to the IPC communication socket.
const char *CsIPC::CS_KEY_PATH = "/var/run/CameraServer.socket";

/*!
 * @brief The constructor for the cs_ipc class.
 */
CsIPC::CsIPC()
{
}

/*!
 * @brief Initialize IPC for the daemon, currently using UNIX Domain sockets.
 * This is used for internal processes to communicate with the Camera Server.
 * Error should be be considered fatal because without other processes being
 * able to communicate with the camera server we can't do anything.
 *
 *
 * @return CS_NET_SUCCESS if sucessful, negative numbers are errors and should
 *     be considered fatal.
 */
int CsIPC::setup()
{
    // Make sure the file path we are using doesn't exist, this is if
    // someone crashed last time.
    unlink(CS_KEY_PATH);

    // Create our Linux IPC socket.
    sock = std::make_unique<CsSocket>(AF_UNIX, SOCK_DGRAM);
    if ((sock.get() == nullptr) || (sock->getFD() < 0))
    {
        fi_log(LOG_ERR, "%s: error creating UNIX domain socket", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_SOCKET;
    }

    // Bind the socket to the Unix domain.
    if (!sock->bind(CS_KEY_PATH))
    {
        fi_log(LOG_ERR, "%s: error binding to UNIX domain socket key:%s", __PRETTY_FUNCTION__, CS_KEY_PATH);
        return CS_NET_ERROR_BIND;
    }

    // It's all good.
    return CS_NET_SUCCESS;
}

/*!
 * @brief Close down the IPC channels to make sure everything is cleaned up
 *    before shutting down.
 */
void CsIPC::close()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Close the socket.
    CsNetwork::close();

    // Remove the file used for the socket.
    unlink(CS_KEY_PATH);
}


/*!
 * @brief Get the next message from any local client trying to send us a request.
 *    Since we are using datagrams the client is not waiting.
 *
 * @param p_msg Pointer to where we'll put the received message, if any.
 *
 * @return The size of the message if successful, 0 if no message, <0 if error.
 */
int CsIPC::getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     fromlen;
    ssize_t msg_size;
    struct sockaddr_un  from;           // UNIX domain socket structure

    // Sanity check.
    if (!isValid())
        return CS_NET_ERROR_FD;
    if (p_msg == NULL)
        return CS_NET_INVALID_PARAM;

    // Clear our receive socket structure. Initialize it as
    // a local (IPC) socket.
    memset(&from, 0, sizeof(from));
    from.sun_family = AF_LOCAL;

    // Initialize the size of our structure.
    fromlen = sizeof(from);

    // Receive a message from the socket.
    msg_size = ::recvfrom(sock->getFD(), (char *)&p_msg->message, sizeof(p_msg->message), 0,
                          (struct sockaddr *)&from, (socklen_t *)&fromlen);
    if (msg_size == -1)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return CS_NET_ERROR_RECVFROM;
    }

    // Timestamp when we received the message.
    clock_gettime(CLOCK_REALTIME, &p_msg->rcv_time);

//    LOG("(%s) message received from client:%s length:%d rcv:%ld:%ld", __PRETTY_FUNCTION__, from.sun_path, msg_size,
//        p_msg->rcv_time.tv_sec, p_msg->rcv_time.tv_nsec);

    // Clear out our "from" structure.
    memset(p_msg->ipc_msg.from, 0, CS_NETWORK::LINUX_PATH_LEN + 1);

    // Now copy the path over.
    strncpy(p_msg->ipc_msg.from, from.sun_path, CS_NETWORK::LINUX_PATH_LEN);

    // Make sure we note that this is an IPC message.
    p_msg->msg_type = CS_NETWORK::MSG_TYPE_IPC;

    // Return the length of the message.
    return msg_size;
}

/*!
 * @brief Send a response back to the client, for a message request we have already received.
 *
 * @param p_msg     Message which is destined for the client.
 * @param num_bytes Length of the message to transmit.
 *
 * @return the size of the message if successful, 0 if no message, <0 if error
 */
int CsIPC::sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes)
{
    int     tolen;              // Number of bytes
    ssize_t msg_size;           // Length of the entire message structure.
    struct sockaddr_un  to;         // UNIX domain socket structure.

    // Sanity check.
    if (!isValid())
    {
        fi_log(LOG_DEBUG, "%s: No socket_fd?", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_FD;
    }
    if (p_msg == NULL)
    {
        fi_log(LOG_DEBUG, "%s: p_msg == NULL?", __PRETTY_FUNCTION__);
        return CS_NET_INVALID_PARAM;
    }

    // Populate the socket. Make sure that the socket path is null-terminated.
    memset(&to, 0, sizeof(to));
    to.sun_family = AF_LOCAL;
    strncpy(to.sun_path, p_msg->ipc_msg.from, sizeof(to.sun_path));
    to.sun_path[ sizeof(to.sun_path) - 1 ] = '\0';

    // Send the message to the socket.
    tolen = sizeof(to);
    msg_size = sendto(sock->getFD(), (char *)&p_msg->message, num_bytes, 0, (struct sockaddr *)&to, tolen);
    if (msg_size == -1)
    {
        fi_log(LOG_ERR, "%s: Cannot send to client len:%d to:%s errno:%d(%s)",
               __PRETTY_FUNCTION__, num_bytes, p_msg->ipc_msg.from, errno, strerror(errno));
        return CS_NET_ERROR_SENDTO;
    }

    // Return the size of the message sent.
    return msg_size;
}
