/*!
 * @file ipm_proc.cpp
 * @brief Manage shared usage of the IPM.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>

#include <chrono>
#include <condition_variable>
#include <mutex>

#include "bounce.h"
#include "camera_server.h"
#include "cs_trace.h"
#include "ipm_proc.h"
#include "part_mgr.h"

namespace IPM_PROC
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Number of milliseconds to wait for the IPM to finish processing.
const int IPM_PROCESS_DELAY = 50;

//! Number of retries while waiting for the IPM to finish processing.
const int IPM_PROCESS_RETRY = 20;

//! Message types for the IPM processor.
enum IPM_MESSAGE_TYPES
{
    IPM_ENCODE = 0x30000,       //!< Encode a buffer in Image Memory.
    IPM_BUF_RELEASE,            //!< Buffer is no longer in use.
    IPM_BUF_RESET               //!< Reset the buffer manager.
};

//! The state of the IPM buffers.
enum IPM_BUFFER_STATE
{
    IPM_BUF_STATE_EMPTY,        //!< The buffer is open and available to be written.
    IPM_BUF_STATE_WRITING,      //!< The buffer is being written by the IPM.
    IPM_BUF_STATE_FULL,         //!< The buffer has been processed by the IPM.
    IPM_BUF_STATE_READING       //!< The buffer is in use by a reader.
};

struct IPM_BUFFER_INFO
{
    uint64_t addr;              //!< Address of the buffer, in image memory.
    uint32_t outputDataLen;     //!< The number of bytes of data in the buffer.
    uint32_t processingInfo;    //!< The IPM processing information for the buffer.
    IPM_BUFFER_STATE state;     //!< The state of the buffer.
};

//! Message class for IPM_ENCODE messages.
class IpmMessage : public FAS_THREAD::SyncMessage
{
public:
    IpmMessage();
    virtual ~IpmMessage();

    IPM_PROC::IPM_REQUEST request;
};

//! Message class to reset the IPM buffers
class IpmBufferResetMsg : public FAS_THREAD::Message
{
public:
    IpmBufferResetMsg();
    virtual ~IpmBufferResetMsg();
};

//! Message class to release an IPM buffer.
class IpmBufferReleaseMsg : public FAS_THREAD::Message
{
public:
    IpmBufferReleaseMsg();
    virtual ~IpmBufferReleaseMsg();

    uint64_t buffer_addr;
};
/*!
 * @brief The IPM processor control object. This object is
 *    a singleton and is "local" to this file.
 */
class ipmThread : public FAS_THREAD::MsgThread
{
public:
    ~ipmThread();
    virtual void messageHandle(FAS_THREAD::Message *message) override;

    static ipmThread *factory();
    static void processingFinished();

    // Singletons should not be cloneable.
    ipmThread(ipmThread& other) = delete;

    // Singletons should not be assignable.
    void operator=(const ipmThread&) = delete;

private:
    ipmThread();

    void handleBufResetRequest();
    void handleBufReleaseRequest(IpmBufferReleaseMsg *message);
    void handleIPMRequest(IpmMessage *msg);

    constexpr int getBufIndexForAddr(uint64_t buf_addr);
    int getNextBufferIndex();
    uint64_t getNextBuffer(uint32_t proc_type);
    int processFrame(IPM_PROC::IPM_REQUEST *request);
    bool releaseProcessedFrame(uint64_t free_addr);

private:
    static std::unique_ptr<ipmThread> ipm_proc;
    static std::mutex ipm_proc_mutex;

    uint32_t image_len;                 //!< Length of processed image.
    uint64_t ipm_buf_addr;              //!< Buffer address for processed image in image memory.

    std::mutex ipmMutex;                //!< Used to control changes to IPM status.
    std::condition_variable ipmStatus;  //!< Condition variable we can block on.

    IPM_BUFFER_INFO bufInfo[PART_MGR::NUM_IPM_PP_BUFFERS];  //!< Information about the IPM processing buffer(s).
};

std::unique_ptr<ipmThread> ipmThread::ipm_proc;
std::mutex ipmThread::ipm_proc_mutex;

} /* End of namespace IPM_PROC */

using namespace IPM_PROC;
using namespace CS_TRACE;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

void ipm_proc_dump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Sanity check.
    if (ipmThread::factory() == nullptr)
        return;

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(ipmThread::factory());
}

void ipm_proc_init()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_IPM_PROC;

    // Make sure we have our singleton.
    ipmThread::factory();
}

void ipm_proc_close()
{
    ipmThread *thread_ptr;

    // Release our singleton.
    thread_ptr = ipmThread::factory();
    if (thread_ptr != nullptr)
    {
        // Tell the thread that we are stopping.
        FAS_THREAD::SyncMessage msg;
        msg.req = FAS_THREAD::HALT;

        // Pass it on.
        msg.doSync(thread_ptr);

        // Try to join the thread.
        thread_ptr->join();
    }
}

void ipm_proc_done()
{
    ipmThread::processingFinished();
}

void ipm_proc_reset()
{
    IpmBufferResetMsg *msg;

    // Sanity check. Make sure we have a singleton.
    if (ipmThread::factory() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No IPM thread yet!?", __PRETTY_FUNCTION__);
        return;
    }

    // Allocate our message.
    msg = new IpmBufferResetMsg();
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to allocate an IpmBufferResetMsg!?", __PRETTY_FUNCTION__);
        return;
    }

    // Send the message onwards.
    ipmThread::factory()->messageAdd(msg);
}

/*!
 * @brief Entry to IPM processing of a raw image (i.e. one located in
 *    Image Memory).
 *
 * @note It is the caller's responsibility to release the IPM buffer
 *   once they are done with it.
 */
bool ipmProcRequest(IPM_PROC::IPM_REQUEST *request)
{
    bool rc;
    IpmMessage msg;

    // Sanity check
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request!?", __PRETTY_FUNCTION__);
        return false;
    }
    if (ipmThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    TP(TRACE_SEL_IPM_PROC, "%s: im_addr=0x%" PRIx64 ", proc_info=0x%x",
       __PRETTY_FUNCTION__, request->image_addr, request->proc_info);

    // Build up our message.
    memcpy(&msg.request, request, sizeof(IPM_PROC::IPM_REQUEST));

    // Send it on.
    msg.doSync(ipmThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        request->image_len = msg.request.image_len;
        request->ipm_buf_addr = msg.request.ipm_buf_addr;
        rc = true;
        TP(TRACE_SEL_IPM_PROC, "%s: IPM=%d bytes @ 0x%" PRIx64, __PRETTY_FUNCTION__, request->image_len, request->ipm_buf_addr);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed. Did we change states?", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Release an IPM buffer.
 */
void ipm_proc_release(uint64_t ipm_buffer)
{
    if (ipmThread::factory() != nullptr)
    {
        IpmBufferReleaseMsg *msg;

        msg = new IpmBufferReleaseMsg();
        if (msg == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Unable to allocate an IpmBufferReleaseMsg!", __PRETTY_FUNCTION__);
            return;
        }
        msg->buffer_addr = ipm_buffer;

        // Send it onwards.
        ipmThread::factory()->messageAdd(msg);
    }
}

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

IpmMessage::IpmMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)IPM_ENCODE;
    memset(&request, 0, sizeof(IPM_PROC::IPM_REQUEST));
}

IpmMessage::~IpmMessage()
{
}

IpmBufferResetMsg::IpmBufferResetMsg() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)IPM_BUF_RESET;
}

IpmBufferResetMsg::~IpmBufferResetMsg()
{
}

IpmBufferReleaseMsg::IpmBufferReleaseMsg() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)IPM_BUF_RELEASE;
    buffer_addr = 0;
}

IpmBufferReleaseMsg::~IpmBufferReleaseMsg()
{
}

//--------------------------------------------------------------------
// ipmThread Implementation
//--------------------------------------------------------------------

ipmThread::ipmThread() :
    FAS_THREAD::MsgThread(),
    image_len(0),
    ipm_buf_addr(0)
{
    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);
}

ipmThread::~ipmThread()
{
    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);
}

ipmThread *ipmThread::factory()
{
    ipmThread *rc(nullptr);

    std::lock_guard<std::mutex> lock(ipm_proc_mutex);

    // We only allow one thread object.
    if (ipm_proc.get() == nullptr)
    {
        // Create our thread.
        ipm_proc = std::unique_ptr<ipmThread>(new ipmThread());
        if (ipm_proc.get() != nullptr)
            ipm_proc.get()->start();

        // Return the pointer to the singleton.
        rc = ipm_proc.get();
    }

    // Check to see if we were halted.
    else if (ipm_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = ipm_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Process the messages for the thread.
 */
void ipmThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case IPM_ENCODE:
            handleIPMRequest(static_cast<IpmMessage *>(message));
            break;
        case IPM_BUF_RESET:
            handleBufResetRequest();

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case IPM_BUF_RELEASE:
            handleBufReleaseRequest(static_cast<IpmBufferReleaseMsg *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case FAS_THREAD::HALT:
            // Do not allow another instance to accidently get created. We are done.
            halted = true;
            message->resp = FAS_THREAD::OK;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Do we still have a message?
        if (message != nullptr)
        {
            // Forward on the messsage.
            message->forward();
        }
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in ipmThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

/*!
 * @brief Reset the buffer control structure.
 */
void ipmThread::handleBufResetRequest()
{
    PartMgr *part_mgr;

    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);

    // Get our pointer to the partition manager.
    part_mgr = PartMgr::getInstance();
    if (part_mgr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: PartMgr nullptr!?", __PRETTY_FUNCTION__);
        return;
    }

    // Initialize our buffers.
    for (int i = 0; i < PART_MGR::NUM_IPM_PP_BUFFERS; ++i)
    {
        bufInfo[i].state = IPM_BUF_STATE_EMPTY;
        bufInfo[i].addr = part_mgr->getIpmOutputBufferAddr(i);
        bufInfo[i].outputDataLen = 0;
        bufInfo[i].processingInfo = IMGPROC_TYPE_NONE;
    }
}

void ipmThread::handleBufReleaseRequest(IpmBufferReleaseMsg *message)
{
    // Pass this along to the p1 processing logic.
    releaseProcessedFrame(message->buffer_addr);
}

/*!
 * @brief Use the IPM to process a raw image from IM.
 */
void ipmThread::handleIPMRequest(IpmMessage *msg)
{
    int rc = 0;

    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);

    // Get the frame processed by the IPM.
    rc = processFrame(&msg->request);
    if (rc)
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Return the processed frame information.
    msg->request.ipm_buf_addr = ipm_buf_addr;
    msg->request.image_len = image_len;

    // All done.
    msg->resp = FAS_THREAD::OK;
}

uint64_t ipmThread::getNextBuffer(uint32_t proc_type)
{
    int buf_select;

    // Get the next availabe buffer.
    buf_select = getNextBufferIndex();
    if (buf_select < 0)
    {
        fi_log(LOG_DEBUG, "%s: unable to find an available buffer!", __PRETTY_FUNCTION__);
        return 0;
    }
    TP(TRACE_SEL_IPM_PROC, "%s: buf=%d", __PRETTY_FUNCTION__, buf_select);

    // Mark this buffer as busy.
    bufInfo[buf_select].state = IPM_BUF_STATE_WRITING;

    // Remember what kind of processing is being done.
    bufInfo[buf_select].processingInfo = proc_type;

    // Return the address of the next available P1 buffer.
    return bufInfo[buf_select].addr;
}

int ipmThread::getNextBufferIndex()
{
    int     buf_select;

    TP(TRACE_SEL_P1_PROC, "%s", __PRETTY_FUNCTION__);

    // Check to see if [0] is available.
    if (bufInfo[0].state == IPM_BUF_STATE_EMPTY)
    {
        // Use the [0] buffer.
        buf_select = 0;
    }

    // Check to see if [1] is available.
    else if (bufInfo[1].state == IPM_BUF_STATE_EMPTY)
    {
        // Use the [0] buffer.
        buf_select = 1;
    }
    else
    {
        // Oops...something bad happened.
        fi_log(LOG_WARNING, "%s: Frames are not free (%d, %d).", __PRETTY_FUNCTION__, (int)bufInfo[0].state, (int)bufInfo[1].state);

        // Try to release at least one of the buffers and then see
        // if we can make use of it.
        releaseProcessedFrame(0);
        if (bufInfo[0].state == IPM_BUF_STATE_EMPTY)
        {
            buf_select = 0;
        }
        else if (bufInfo[1].state == IPM_BUF_STATE_EMPTY)
        {
            buf_select = 1;
        }
        else
        {
            fi_log(LOG_ERR, "%s: Unable to determine available buffer...bailing.", __PRETTY_FUNCTION__);
            return -1;
        }
    }

    // Return the selected buffer index.
    return buf_select;
}

bool ipmThread::releaseProcessedFrame(uint64_t free_addr)
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    bool    rc;

    TP(TRACE_SEL_P1_PROC, "%s", __PRETTY_FUNCTION__);

    // Assume success.
    rc = true;

    // Check to see if the user wants to free all the buffers,
    // or a specific buffer.
    if (free_addr == 0)
    {
        // If the buffer is queued and the IPM is not idle, then we can't release it.
        for (int i = 0; i < PART_MGR::NUM_IPM_PP_BUFFERS; ++i)
        {
            if ((bufInfo[i].state == IPM_BUF_STATE_WRITING) && !fpga_ptr->ipmIsIdle())
            {
                fi_log(LOG_WARNING, "%s: IPM buffer %d is QUEUED - not releasing.", __PRETTY_FUNCTION__, i);
                rc = false;
            }
            else
            {
                // Go ahead and free the buffer.
                bufInfo[i].state = IPM_BUF_STATE_EMPTY;
            }
        }
    }
    else
    {
        int     index;

        index = getBufIndexForAddr(free_addr);
        if (index < 0)
        {
            fi_log(LOG_ERR, "%s: Invalid address (0x%" PRIx64 ")! Valid are (0x%" PRIx64 ") and (0x%" PRIx64 ").",
                   __PRETTY_FUNCTION__, free_addr, bufInfo[0].addr, bufInfo[1].addr);
            rc = false;
        }
        else if ((bufInfo[index].state == IPM_BUF_STATE_WRITING) && !fpga_ptr->ipmIsIdle())
        {
            fi_log(LOG_ERR, "%s: P1 buffer %d (0x%" PRIx64 ") is QUEUED - not releasing.", __PRETTY_FUNCTION__, index, free_addr);
            rc = false;
        }
        else
        {
            bufInfo[index].state = IPM_BUF_STATE_EMPTY;
        }
    }

    return rc;
}

constexpr int ipmThread::getBufIndexForAddr(uint64_t buf_addr)
{
    for (int i = 0; i < PART_MGR::NUM_IPM_PP_BUFFERS; ++i)
    {
        // Do we have a match?
        if (bufInfo[i].addr == buf_addr)
        {
            return i;
        }
    }

    // No match found.
    return -1;
}

/*!
 * @brief Callback for when the IPM interrupts indicating that it is finished.
 */
void ipmThread::processingFinished()
{
    TP(TRACE_SEL_IPM_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (ipmThread::factory() == nullptr)
        return;

    // Acquire the processing mutex.
    std::unique_lock<std::mutex> locker(ipmThread::factory()->ipmMutex);

    // Wake up anyone waiting.
    ipmThread::factory()->ipmStatus.notify_one();
}

/*!
 * @brief Have the FPGA's IPM process the frame we just captured. All
 *    of the relevant information can be found in the ipm_info object.
 *
 * @param request   Pointer to the structure containing parameter information.
 * @return 0 on success, negative otherwise.
 */
int ipmThread::processFrame(IPM_PROC::IPM_REQUEST *request)
{
    int i;
    int rc = -1;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_IPM_PROC, "%s: proc_info=0x%x", __PRETTY_FUNCTION__, request->proc_info);

    // Statistics
    Fastec::IntervalTimer frame_timer;
    frame_timer.start();
    // Statistics

    // Make sure the IPM is idle.
    if (!fpga_ptr->ipmIsIdle())
    {
        fi_log(LOG_DEBUG, "%s: The IPM is NOT idle!", __PRETTY_FUNCTION__);

        // Reset the IPM.
        fpga_ptr->ipmReset();

        // Try again.
        if (!fpga_ptr->ipmIsIdle())
        {
            fi_log(LOG_DEBUG, "%s: The IPM is STILL NOT idle!", __PRETTY_FUNCTION__);
            goto if_fatal;
        }
    }

    // Enable the IPM processing complete interrupt.
    fpga_ptr->irqEnable(PCIEIRQ_P1_FRAME_DONE);

    // Get our P1 destination buffer.
    ipm_buf_addr = getNextBuffer(request->proc_info);
    if (ipm_buf_addr == 0)
    {
        fi_log(LOG_ERR, "%s: Unable to get a buffer!", __PRETTY_FUNCTION__);
        goto if_bail;
    }

    // Get the raw pixel data processed properly through the IPM depending on
    // the callers request.
    if (request->process_live)
        image_len = fpga_ptr->ipmLiveFrameProcess(request->proc_info, convert_to_HEXWORD(request->image_addr), convert_to_HEXWORD(ipm_buf_addr));
    else
        image_len = fpga_ptr->ipmFrameProcess(request->proc_info, convert_to_HEXWORD(request->image_addr), convert_to_HEXWORD(ipm_buf_addr));
    if (image_len == 0)
    {
        fi_log(LOG_DEBUG, "%s: process_live/p1_frame() failed", __PRETTY_FUNCTION__);
        rc = -1;
        goto if_bail;
    }

    for (i = 0; i < IPM_PROCESS_RETRY; ++i)
    {
        // Wait for the IPM to finish or until we timeout. Put this in it's own scope
        // and the locker will release on the exit of the scope. The timeout is short
        // enough that we should be able to pick up a mode change just fine.
        std::unique_lock<std::mutex> locker(ipmMutex);
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (ipmStatus.wait_until(locker, now + ms(IPM_PROCESS_DELAY)) == std::cv_status::timeout)
        {
            uint32_t value(0);

            // Read the FPGA to determine if the IPM is still busy. When we get real busy,
            // we seem to be missing interrupts!
            //
            // TODO TODO TODO
            // Why are we missing interrupts!?
            // TODO TODO TODO
            fpga_ptr->regRead(IPM_P1_FRAME_REQ_STATUS, &value);
            if ((value & IPM_P1_FRAME_BUSY) != IPM_P1_FRAME_BUSY)
                break;
        }
        else
        {
            // Find the buffer that was queued.
            for (int i = 0; i < PART_MGR::NUM_IPM_PP_BUFFERS; ++i)
            {
                // Is this the one?
                if (bufInfo[i].addr == ipm_buf_addr)
                {
                    // Update the buffer state.
                    bufInfo[i].state = IPM_BUF_STATE_FULL;
                    break;
                }
            }

            // All done.
            break;
        }
    }
    if (i == IPM_PROCESS_RETRY)
    {
        fi_log(LOG_ERR, "%s: IPM took too long [%d ms]!", __PRETTY_FUNCTION__, (IPM_PROCESS_RETRY * IPM_PROCESS_DELAY));
        rc = -1;
        goto if_bail;
    }

    // We are done.
    rc = 0;

if_bail:
    // We're all done with the interrupt.
    fpga_ptr->irqDisable(PCIEIRQ_P1_FRAME_DONE);

    // Is there an error?
    if (rc != 0)
    {
        // There was a problem, so reset the interface.
        fpga_ptr->ipmReset();
        ipm_buf_addr = 0;
    }

if_fatal:
    // Statistics
    frame_timer.stop();
    TP(TRACE_SEL_IPM_PROC, "%s: frame processing: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
    // Statistics

    // All done.
    return rc;
}
