/*!
 * @file jpeg_proc.cpp
 * @brief Manage shared usage of the JPEG encoder.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>

#include <chrono>
#include <condition_variable>
#include <iomanip>
#include <memory>
#include <mutex>

#include "bounce.h"
#include "camera_server.h"
#include "cs_trace.h"
#include "fi_internal.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "part_mgr.h"

namespace JPEG_PROC
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Pick the largest sensor width we support.
const int POOL_DEFAULT_WIDTH    = 2560;

//! Pick the largest sensor height we support.
const int POOL_DEFAULT_HEIGHT   = 2048;

//! The number of microseconds for aging out a pool buffer. Assume a
//! worst case of 20 frames/sec = 0.05 sec/frame.
const uint32_t POOL_AGEOUT  = 50000;

//! Force an age-out if the buffer exceeds this time.
const uint32_t POOL_FORCE_AGEOUT = 250000;

//! Message types for the JPEG processor.
enum JPEG_MESSAGE_TYPES
{
    JM_BUF_ENCODE = 0x30000,        //!< Encode a buffer already in system memory.
    JM_IPM_ENCODE,                  //!< Encode a buffer in Image Memory.
    JM_RELEASE,                     //!< Release a buffer.
    JM_REUSE                        //!< Reuse a buffer.
};

//! Pool buffer management structure.
typedef struct BUF_POOL_REC
{
    struct timespec time_stamp;     //!< Timestamp when buffer was created/filled.
    uint64_t frame_addr;            //!< Image memory address of the frame.
    int use_count;                  //!< Number of clients using this buffer.

    uint8_t *buf_ptr;               //!< Poiner to the buffer.
    int buf_len;                    //!< Number of encoded bytes in the buffer.
} BUF_POOL_REC;

//! Message class for JM_BUF_ENCODE messages.
class jpegBufMessage : public FAS_THREAD::SyncMessage
{
public:
    jpegBufMessage();
    virtual ~jpegBufMessage();

    JPEG_PROC::BUF_REQUEST request;
};

//! Message class for JM_IPM_ENCODE messages.
class jpegIPMMessage : public FAS_THREAD::SyncMessage
{
public:
    jpegIPMMessage();
    virtual ~jpegIPMMessage();

    JPEG_PROC::IPM_REQUEST request;
};

//! Message class for JM_RELEASE messages.
class jpegReleaseMessage : public FAS_THREAD::SyncMessage
{
public:
    jpegReleaseMessage();
    virtual ~jpegReleaseMessage();

    uint8_t *buffer_ptr;    //!< The buffer we want to release.
};

//! Message class for JM_REUSE messages.
class jpegReuseMessage : public FAS_THREAD::SyncMessage
{
public:
    jpegReuseMessage();
    virtual ~jpegReuseMessage();

    JPEG_PROC::REUSE_REQUEST request;
};

/*!
 * @brief The JPEG processor control object. This object is
 *    a singleton and is "local" to this file.
 */
class jpegThread : public FAS_THREAD::MsgThread
{
public:
    virtual ~jpegThread();
    virtual void messageHandle(FAS_THREAD::Message *message);

    // Singletons should not be cloneable.
    jpegThread(jpegThread& other) = delete;

    // Singletons should not be assignable.
    void operator=(const jpegThread&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static jpegThread *factory();

private:
    jpegThread();

    void handleBufRequest(jpegBufMessage *msg);
    void handleIPMRequest(jpegIPMMessage *msg);
    void handleReleaseRequest(jpegReleaseMessage *msg);
    void handleReuseRequest(jpegReuseMessage *msg);

    int acquire_buf();
    void create_pool();
    int encode_frame(JPEG_PROC::IPM_REQUEST *request, uint8_t *image_ptr);
    int find_buf(uint64_t frame_addr);
    int ipm_frame(JPEG_PROC::IPM_REQUEST *request);
    void refresh_pool(uint32_t width, uint32_t height);
    void update_ju(int pixel_format, int quality);

private:
    static std::unique_ptr<jpegThread> jpeg_proc;
    static std::mutex jpeg_proc_mutex;

    JpegUtil ju;        //!< Our JPEG encoding object...does the actual work.

    int bb_index = -1;      //!< Bounce buffer used for DMA-ing images.
    uint64_t bb_addr;       //!< Physical address of the buffer.
    uint8_t *bb_ptr;        //!< Memory mapped pointer to the buffer.

    int pool_bb_index;      //!< The bounce buffer index for our buffer pool space.
    uint64_t pool_addr;     //!< The physical address of our buffer pool space.
    uint8_t *pool_ptr;      //!< The memory mapped base address of the buffer pool space.
    uint32_t pool_width;    //!< Width used for the pool layout.
    uint32_t pool_height;   //!< Height used for the pool layout.
    int pool_buf_len;       //!< Maximum length of the individual buffers.
    std::vector<std::unique_ptr<BUF_POOL_REC>> buf_pool;  //!< The pool of buffers.

    uint32_t image_len;         //!< Length of processed image.
    uint64_t ipm_buf_addr;      //!< Buffer address for processed image in image memory.
};

} /* End of namespace JPEG_PROC */

using namespace CS_TRACE;
using namespace JPEG_PROC;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<jpegThread> jpegThread::jpeg_proc;
std::mutex jpegThread::jpeg_proc_mutex;

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

void jpeg_dump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Sanity check.
    if (jpegThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return;
    }

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(jpegThread::factory());
}

void jpeg_init()
{
    // Turn on tracing.
    // traceControl |= TRACE_SEL_JPEG_PROC;

    // Make sure we have our singleton.
    jpegThread::factory();
}

void jpeg_close()
{
    jpegThread *thread_ptr;

    // Release our singleton.
    thread_ptr = jpegThread::factory();
    if (thread_ptr != nullptr)
    {
        // Tell the thread that we are stopping.
        FAS_THREAD::SyncMessage msg;
        msg.req = FAS_THREAD::HALT;

        // Pass it on.
        msg.doSync(thread_ptr);

        // Try to join the thread.
        thread_ptr->join();
    }
}

/*!
 * @brief Entry to JPEG encoding an existing (i.e. in system memory)
 *    NV12 frame.
 */
bool jpeg_buf_request(JPEG_PROC::BUF_REQUEST *request)
{
    bool rc;
    jpegBufMessage msg;

    // Sanity check
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request!?", __PRETTY_FUNCTION__);
        return false;
    }
    if (jpegThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    TP(TRACE_SEL_JPEG_PROC, "%s: Y=%p, UV=%p, width=%d, height=%d, pixfmt=0x%x",
       __PRETTY_FUNCTION__, (void *)request->src_pixels_Y, (void *)request->src_pixels_UV,
       request->width, request->height, request->pixel_format);

    // Build up our message.
    memcpy(&msg.request, request, sizeof(JPEG_PROC::BUF_REQUEST));

    // Send it on.
    msg.doSync(jpegThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        request->jpeg_buffer = msg.request.jpeg_buffer;
        request->jpeg_len = msg.request.jpeg_len;
        rc = true;
        TP(TRACE_SEL_JPEG_PROC, "%s: jpeg=%d bytes @ %p", __PRETTY_FUNCTION__, request->jpeg_len, (void *)request->jpeg_buffer);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed. Did we change states?", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Entry to JPEG encoding a raw image (i.e. one located in
 *    Image Memory).
 */
bool jpeg_ipm_request(JPEG_PROC::IPM_REQUEST *request)
{
    bool rc;
    jpegIPMMessage msg;

    // Sanity check
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request!?", __PRETTY_FUNCTION__);
        return false;
    }
    if (jpegThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    TP(TRACE_SEL_JPEG_PROC, "%s: im_addr=0x%" PRIx64 ", dma_mod=0x%08x, proc_info=0x%x, width=%d, height=%d, tl=%d, jpeg_buf=%p",
       __PRETTY_FUNCTION__, request->image_addr, request->dma_mod, request->proc_info,
       request->width, request->height, request->time_limit, (void *)request->jpeg_buffer);

    // Build up our message.
    memcpy(&msg.request, request, sizeof(JPEG_PROC::IPM_REQUEST));

    // Send it on.
    msg.doSync(jpegThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        request->jpeg_buffer = msg.request.jpeg_buffer;
        request->jpeg_len = msg.request.jpeg_len;
        rc = true;
        TP(TRACE_SEL_JPEG_PROC, "%s: Finished, jpeg=%d bytes @ %p", __PRETTY_FUNCTION__, request->jpeg_len, (void *)request->jpeg_buffer);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed. Did we change states?", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Release a JPEG encoded buffer.
 */
void jpeg_release_frame(uint8_t *jpeg_buffer)
{
    jpegReleaseMessage msg;

    // Sanity check.
    if (jpeg_buffer == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null jpeg_buffer pointer?", __PRETTY_FUNCTION__);
        return;
    }
    if (jpegThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return;
    }

    TP(TRACE_SEL_JPEG_PROC, "%s: %p", __PRETTY_FUNCTION__, (void *)jpeg_buffer);

    // Build up the message.
    msg.buffer_ptr = jpeg_buffer;

    // Send it on.
    msg.doSync(jpegThread::factory());
}

/*!
 * @brief See if we have a buffer that's ready to go.
 */
bool jpeg_reuse(JPEG_PROC::REUSE_REQUEST *request)
{
    bool rc = false;
    jpegReuseMessage msg;

    // Sanity check.
    if (jpegThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    TP(TRACE_SEL_JPEG_PROC, "%s: %u", __PRETTY_FUNCTION__, request->time_limit);

    // Build up our message.
    msg.request.time_limit = request->time_limit;

    // Send it on.
    msg.doSync(jpegThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        request->jpeg_buffer = msg.request.jpeg_buffer;
        request->jpeg_len = msg.request.jpeg_len;
        rc = true;
        TP(TRACE_SEL_JPEG_PROC, "%s: jpeg=%d bytes @ %p", __PRETTY_FUNCTION__, request->jpeg_len, (void *)request->jpeg_buffer);
    }

    // Return the result.
    return rc;
}

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

jpegBufMessage::jpegBufMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)JM_BUF_ENCODE;
    memset(&request, 0, sizeof(request));
}

jpegBufMessage::~jpegBufMessage()
{
}

jpegIPMMessage::jpegIPMMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)JM_IPM_ENCODE;
    memset(&request, 0, sizeof(request));
}

jpegIPMMessage::~jpegIPMMessage()
{
}

jpegReleaseMessage::jpegReleaseMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)JM_RELEASE;
    buffer_ptr = nullptr;
}

jpegReleaseMessage::~jpegReleaseMessage()
{
}

jpegReuseMessage::jpegReuseMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)JM_REUSE;
    memset(&request, 0, sizeof(request));
}

jpegReuseMessage::~jpegReuseMessage()
{
}

//--------------------------------------------------------------------
// jpegThread Implementation
//--------------------------------------------------------------------

jpegThread::jpegThread() :
    FAS_THREAD::MsgThread(),
    pool_bb_index(-1)
{
    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Create our pool of buffers.
    create_pool();

    // Get a buffer for our image processing.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Map the memory.
    bb_ptr = (uint8_t *)bb_mmap(bb_index, 0);
    if (bb_ptr == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return;
    }
    fi_log(LOG_DEBUG, "%s: index=%d, addr=0x%" PRIx64 ", ptr=%p", __PRETTY_FUNCTION__, bb_index, bb_addr, (void *)bb_ptr);
}

jpegThread::~jpegThread()
{
    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);
    if (pool_bb_index >= 0)
        bb_release(pool_bb_index);
    if (bb_index >= 0)
        bb_release(bb_index);
}

int jpegThread::acquire_buf()
{
    int i;

    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Loop through the pool, looking for an available buffer.
    for (i = 0; i < (int)buf_pool.size(); ++i)
    {
        uint64_t time_dif;

        // Determine the "age" of the buffer.
        time_dif = measuretime(0, &buf_pool[i]->time_stamp);

        // Is this buffer available?
        if (buf_pool[i]->use_count == 0)
        {
            // Was it used to compress "system memory" (vs. IPM)
            // frames? We don't need to worry about aging with these.
            if (buf_pool[i]->frame_addr == (uint64_t) -1)
                break;

            // This is an IPM buffer, has it aged out?
            if (time_dif >= POOL_AGEOUT)
                break;
        }

        // Check to see if we need to force age it.
        else if (time_dif >= POOL_FORCE_AGEOUT)
        {
            // Force the buffer back into service.
            TP(TRACE_SEL_JPEG_PROC, "%s: Force ageout buffer %d!", __PRETTY_FUNCTION__, i);
            break;
        }
    }
    if (i == (int)buf_pool.size())
    {
        // No buffers are available!
        i = -1;
        fi_log(LOG_DEBUG, "%s: No buffers are available.", __PRETTY_FUNCTION__);
    }
    else
    {
        // Clear out the buffer for use
        buf_pool[i]->frame_addr = -1;
        buf_pool[i]->use_count = 0;
        buf_pool[i]->time_stamp.tv_sec = 0;
        buf_pool[i]->time_stamp.tv_nsec = 0;
    }

    // Return our findings.
    return i;
}

void jpegThread::create_pool()
{
    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Get a bounce buffer to hold our pool.
    pool_bb_index = bb_get_avail();
    pool_addr = bb_get_avail_addr(pool_bb_index);

    // Map the memory.
    pool_ptr = (uint8_t *)bb_mmap(pool_bb_index, 0);
    if (pool_ptr == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        bb_release(pool_bb_index);
        pool_bb_index = -1;
        pool_addr = 0;
        return;
    }

    // Now build up the buffer pool.
    refresh_pool(POOL_DEFAULT_WIDTH, POOL_DEFAULT_HEIGHT);
}

jpegThread *jpegThread::factory()
{
    jpegThread *rc(nullptr);

    std::lock_guard<std::mutex> lock(jpeg_proc_mutex);

    // Check to see if we have our singleton.
    if (jpeg_proc.get() == nullptr)
    {
        // Create the one and only instance.
        jpeg_proc = std::unique_ptr<jpegThread>(new jpegThread());

        // If the creation was successful, get the thread running.
        if (jpeg_proc.get() != nullptr)
            jpeg_proc.get()->start();

        // Return the pointer to the singleton.
        rc = jpeg_proc.get();
    }

    // Check to see if we were halted.
    else if (jpeg_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = jpeg_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Process the messages for the thread.
 */
void jpegThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case JM_BUF_ENCODE:
            handleBufRequest(static_cast<jpegBufMessage *>(message));
            break;
        case JM_IPM_ENCODE:
            handleIPMRequest(static_cast<jpegIPMMessage *>(message));
            break;
        case JM_RELEASE:
            handleReleaseRequest(static_cast<jpegReleaseMessage *>(message));
            break;
        case JM_REUSE:
            handleReuseRequest(static_cast<jpegReuseMessage *>(message));
            break;

        case FAS_THREAD::DUMP:
            *message->stream << "jpegThread Pool Buffers:" << std::endl;
            for (int i = 0; i < (int)buf_pool.size(); ++i)
            {
                *message->stream << i << ": "
                                 << "tstamp=" << buf_pool[i]->time_stamp.tv_sec
                                 << "." << std::setfill('0') << std::setw(6) << (int)(buf_pool[i]->time_stamp.tv_nsec / 1000);
                *message->stream << ", frame_addr=0x" << std::setbase(16) << buf_pool[i]->frame_addr;
                *message->stream << ", use=" << buf_pool[i]->use_count
                                 << ", buf=" << buf_pool[i]->buf_ptr
                                 << ", len=" << buf_pool[i]->buf_len << std::endl;
            }
            *message->stream << std::endl;
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            halted = true;
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in jpegThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

/*!
 * @brief JPEG encode an NV12/Y8 image.
 */
void jpegThread::handleBufRequest(jpegBufMessage *msg)
{
    int rc = 0;
    int pool_index;
    BUF_REQUEST *request;

    // Statistics
    Fastec::IntervalTimer jpeg_timer;
    // Statistics

    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to the request buffer, to make things easier.
    request = &msg->request;

    // Sanity check.
    if (pool_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No pool buffers created!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Make sure our JPEG object is setup properly.
    update_ju(request->pixel_format, request->quality);

    // Do we want to use one of our buffers?
    pool_index = -1;
    if (request->jpeg_buffer == NULL)
    {
        // See if we can get a buffer.
        pool_index = acquire_buf();
        if (pool_index < 0)
        {
            TP(TRACE_SEL_JPEG_PROC, "%s: Unable to acquire a buffer.", __PRETTY_FUNCTION__);
            msg->resp = FAS_THREAD::FAULT;
            return;
        }

        // By setting the frame address to -1, we are marking this as
        // a "system" memory (vs. IPM) frame.
        buf_pool[pool_index]->frame_addr = -1;

        // Grab the current time.
        measuretime(1, &buf_pool[pool_index]->time_stamp);

        // Give over the buffer.
        request->jpeg_buffer = buf_pool[pool_index]->buf_ptr;
        buf_pool[pool_index]->use_count = 1;
        TP(TRACE_SEL_JPEG_PROC, "%s: acquired buf %d:%p", __PRETTY_FUNCTION__, pool_index, (void *)request->jpeg_buffer);
    }

    // Statistics
    jpeg_timer.start();

    // JPEG encode the data.
    rc = ju.encode(request->src_pixels_Y, request->src_pixels_UV, request->width, request->height, &request->jpeg_buffer);

    // Statistics
    jpeg_timer.stop();
    TP(TRACE_SEL_JPEG_PROC, "%s: JPEG encoding: %u bytes, %lf secs.", __PRETTY_FUNCTION__, rc, jpeg_timer.lastMeasurement);
    // Statistics

    // Did it work?
    if (rc < 0)
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Unable to encode the frame!", __PRETTY_FUNCTION__);

        // Did we grab a buffer?
        if (pool_index >= 0)
        {
            // Reset the buffer information.
            buf_pool[pool_index]->time_stamp.tv_sec = 0;
            buf_pool[pool_index]->time_stamp.tv_nsec = 0;
            buf_pool[pool_index]->use_count = 0;
        }

        // Reset the return information
        msg->request.jpeg_buffer = nullptr;

        // Make sure the length is cleared.
        msg->request.jpeg_len = 0;

        // Done.
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Fill in the missing data.
    msg->request.jpeg_len = rc;
    if (pool_index >= 0)
    {
        if (rc > pool_buf_len)
            fi_log(LOG_ERR, "%s: JPEG data exceeds buffer length!!! [%d > %d]", __PRETTY_FUNCTION__, rc, pool_buf_len);
        buf_pool[pool_index]->buf_len = rc;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Transfer a frame from the Image Memory through the
 *    IPM producing an NV12 image. DMA that down to system
 *    memory and run it through the JPEG encoder.
 */
void jpegThread::handleIPMRequest(jpegIPMMessage *msg)
{
    int rc = 0;
    int pool_index;
    uint32_t dma_dims;
    struct timespec time_stamp_start;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (pool_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No pool buffers created!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Make sure our JPEG object is setup properly.
    if ((msg->request.proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
        update_ju(JPEG_UTIL::NV12, msg->request.quality);
    else
        update_ju(JPEG_UTIL::Y8, msg->request.quality);

    // Can we look for a match?
    if (msg->request.time_limit)
    {
        TP(TRACE_SEL_JPEG_PROC, "%s: TL=%d", __PRETTY_FUNCTION__, msg->request.time_limit);

        // Do we have a match?
        pool_index = find_buf(msg->request.image_addr);
        if (pool_index >= 0)
        {
            uint64_t time_dif;

            // Is it "new" enough?
            time_dif = measuretime(0, &buf_pool[pool_index]->time_stamp);
            TP(TRACE_SEL_JPEG_PROC, "%s: time_dif=%" PRIu64 ", tl=%d, buffer %d", __PRETTY_FUNCTION__, time_dif, msg->request.time_limit, pool_index);
            if (time_dif < (uint64_t)msg->request.time_limit)
            {
                TP(TRACE_SEL_JPEG_PROC, "%s: Reusing buffer %d", __PRETTY_FUNCTION__, pool_index);

                // Reuse the buffer.
                msg->request.jpeg_buffer = buf_pool[pool_index]->buf_ptr;
                msg->request.jpeg_len = buf_pool[pool_index]->buf_len;
                buf_pool[pool_index]->use_count++;

                // We're done.
                msg->resp = FAS_THREAD::OK;
                return;
            }
            else
            {
                TP(TRACE_SEL_JPEG_PROC, "%s: Buffer %d is too old.", __PRETTY_FUNCTION__, pool_index);
            }
        }
        else
        {
            TP(TRACE_SEL_JPEG_PROC, "%s: No buffer for image_addr=0x%" PRIx64 ".", __PRETTY_FUNCTION__, msg->request.image_addr);
        }
    }

    // Grab the time right now to time stamp our buffers.
    measuretime(1, &time_stamp_start);

    // Get the frame processed by the IPM.
    rc = ipm_frame(&msg->request);
    if (rc)
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // The image dimensions will be constant.
    dma_dims = (msg->request.height << IPM_FRAME_DIM_Y_SHIFT) | msg->request.width;

    // Have the FPGA transfer the image into our memory. This is a two step
    // if we're color.
    if ((msg->request.proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
    {
        // Because we are using a color sensor, we pass the data through the IPM's
        // Image Converter. This results in packed data, so we don't have to worry
        // about "alignment byte" shenanigans.
        //
        // The FPGA stores the C plane (UV) on a HEXWORD (64 byte) aligned boundary. Reference
        // FiFPGA::start_p1_frame() and the setting of the IPM_P1_FRAME_REQ_IC_Y_BUF_START
        // and IPM_P1_FRAME_REQ_IC_C_BUF_START registers.

        // Transfer the image.
        rc = fpga_ptr->imageDownload(msg->request.dma_mod, dma_dims, ipm_buf_addr, bb_addr, image_len);
    }
    else
    {
        // Transfer the image data. The IPM always row word aligns it's output data. The "image_len" value
        // is this padded size, so we need to tell the FPGA what the destination size should be.
        TP(TRACE_SEL_JPEG_PROC, "%s: Not NV12 - is it Y8 maybe? image_len=%d", __PRETTY_FUNCTION__, image_len);
        rc = fpga_ptr->imageDownload(msg->request.dma_mod, dma_dims, ipm_buf_addr, bb_addr, image_len, (msg->request.width * msg->request.height));
    }

    // We are done with the IPM buffer so free it.
    ipm_proc_release(ipm_buf_addr);

    // Did we successfully download the image?
    if (rc < 0)
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Reset our IPM buffer.
    ipm_buf_addr = 0;

    // Do we want to use one of our buffers?
    pool_index = -1;
    if (msg->request.jpeg_buffer == nullptr)
    {
        // See if we can get a buffer.
        pool_index = acquire_buf();
        if (pool_index < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to acquire a buffer!", __PRETTY_FUNCTION__);
            msg->resp = FAS_THREAD::FAULT;
            return;
        }

        // Initialize the buffer.
        buf_pool[pool_index]->frame_addr = msg->request.image_addr;
        if (msg->request.time_limit)
        {
            // Mark the buffer with the time just prior to the IPM capture of the frame.
            buf_pool[pool_index]->time_stamp.tv_sec = time_stamp_start.tv_sec;
            buf_pool[pool_index]->time_stamp.tv_nsec = time_stamp_start.tv_nsec;
        }
        else
        {
            // Reset the time information.
            buf_pool[pool_index]->time_stamp.tv_sec = 0;
            buf_pool[pool_index]->time_stamp.tv_nsec = 0;
        }

        // Give over the buffer.
        msg->request.jpeg_buffer = buf_pool[pool_index]->buf_ptr;
        buf_pool[pool_index]->use_count = 1;
        TP(TRACE_SEL_JPEG_PROC, "%s: acquired buf %d:%p", __PRETTY_FUNCTION__, pool_index, (void *)msg->request.jpeg_buffer);
    }
    else
    {
        TP(TRACE_SEL_JPEG_PROC, "%s: request.jpeg_buffer=%p", __PRETTY_FUNCTION__, (void *)msg->request.jpeg_buffer);
    }

    // Make sure we have a motion amplification handle and access to the internal library.
    std::unique_ptr<uint8_t[]> amplified_frame;
    if ((msg->request.ma_handle > 0) && (FiInternal::factory() != nullptr))
    {
        // Add the frame to the amplifier.
        FiInternal::MA_ADD_FRAME_FN addFramePtr = FiInternal::factory()->getMaAddFrameFuncPtr();
        if (addFramePtr != nullptr)
        {
            LOG("%s: Add frame", FI_FUNCTION);
            (*addFramePtr)(msg->request.ma_handle, bb_ptr);
        }

        // Can we amplify frames?
        bool amplifierReady(false);
        FiInternal::MA_IS_AVAIL_FN isAvailPtr = FiInternal::factory()->getMaIsAvailFuncPtr();
        if (isAvailPtr != nullptr)
        {
            amplifierReady = (*isAvailPtr)(msg->request.ma_handle);
        }

        // Is the amplifier ready?
        if (amplifierReady)
        {
            // Allocate memory for the amplified frame.
            amplified_frame = std::unique_ptr<uint8_t[]>(new uint8_t[msg->request.width * msg->request.height]());

            // Try getting the amplified frame.
            FiInternal::MA_GET_FRAME_FN getFramePtr = FiInternal::factory()->getMaFrameFuncPtr();
            if (getFramePtr != nullptr)
            {
                // Get the amplified frame.
                int status = (*getFramePtr)(msg->request.ma_handle, amplified_frame.get());
                if (status != FAS_INTERNAL::FIP_OK)
                    amplified_frame.reset();
            }
        }
    }

    // Do we have an amplified frame?
    if (amplified_frame.get() == nullptr)
    {
        // Just encode the base frame.
        rc = encode_frame(&msg->request, bb_ptr);
    }
    else
    {
        // Just encode the amplified frame.
        rc = encode_frame(&msg->request, amplified_frame.get());
    }

    // How did the encoding go?
    if (rc <= 0)
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Unable to encode the frame!", __PRETTY_FUNCTION__);

        // Did we grab a buffer?
        if (pool_index >= 0)
        {
            // Reset the buffer information.
            buf_pool[pool_index]->time_stamp.tv_sec = 0;
            buf_pool[pool_index]->time_stamp.tv_nsec = 0;
            buf_pool[pool_index]->frame_addr = -1;
            buf_pool[pool_index]->use_count = 0;

            // Reset the return information
            msg->request.jpeg_buffer = nullptr;
        }

        // Make sure the length is cleared.
        msg->request.jpeg_len = 0;

        // Done.
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Fill in the missing data.
    msg->request.jpeg_len = rc;
    if (pool_index >= 0)
    {
        if (rc > pool_buf_len)
            fi_log(LOG_ERR, "%s: JPEG data exceeds buffer length!!! [%d > %d]", __PRETTY_FUNCTION__, rc, pool_buf_len);
        buf_pool[pool_index]->buf_len = rc;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Release a buffer.
 */
void jpegThread::handleReleaseRequest(jpegReleaseMessage *msg)
{
    int i;

    TP(TRACE_SEL_JPEG_PROC, "%s: buf_ptr=%p", __PRETTY_FUNCTION__, (void *)msg->buffer_ptr);

    // Loop through the pool, looking for a match.
    for (i = 0; i < (int)buf_pool.size(); ++i)
    {
        if (buf_pool[i]->buf_ptr == msg->buffer_ptr)
            break;
    }
    if (i == (int)buf_pool.size())
    {
        // Huh...
        fi_log(LOG_DEBUG, "%s: %p was not found.", __PRETTY_FUNCTION__, (void *)msg->buffer_ptr);
    }
    else
    {
        // Decrement the usage count.
        buf_pool[i]->use_count--;
        TP(TRACE_SEL_JPEG_PROC, "%s: buf=%d, use_count=%d", __PRETTY_FUNCTION__, i, buf_pool[i]->use_count);

        // Anyone still using it? Sanity check, should never
        // be less than 0...
        if (buf_pool[i]->use_count < 0)
        {
            buf_pool[i]->use_count = 0;
            fi_log(LOG_DEBUG, "%s: Use count for buffer %d went negative!", __PRETTY_FUNCTION__, i);
        }
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Reuse a buffer.
 */
void jpegThread::handleReuseRequest(jpegReuseMessage *msg)
{
    int i;

    TP(TRACE_SEL_JPEG_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (pool_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No pool buffers created!", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Assume we won't find a match.
    msg->resp = FAS_THREAD::FAULT;

    // Loop through the pool, looking for a match.
    for (i = 0; i < (int)buf_pool.size(); ++i)
    {
        uint64_t time_dif;

        // Is it "new" enough?
        time_dif = measuretime(0, &buf_pool[i]->time_stamp);
        TP(TRACE_SEL_JPEG_PROC, "%s: time_dif=%" PRIu64 ", tl=%d, buffer %d", __PRETTY_FUNCTION__, time_dif, msg->request.time_limit, i);
        if (time_dif < (uint64_t)msg->request.time_limit)
        {
            TP(TRACE_SEL_JPEG_PROC, "%s: Reusing buffer %d", __PRETTY_FUNCTION__, i);

            // Reuse the buffer.
            msg->request.jpeg_buffer = buf_pool[i]->buf_ptr;
            msg->request.jpeg_len = buf_pool[i]->buf_len;
            buf_pool[i]->use_count++;

            // We're done.
            msg->resp = FAS_THREAD::OK;
        }
    }
}

/*!
 * @brief JPEG encode a frame. Relevant information is found in
 *    the ipm_info object.
 *
 * @return >0 on success, negative otherwise.
 */
int jpegThread::encode_frame(JPEG_PROC::IPM_REQUEST *request, uint8_t *image_ptr)
{
    int rc = -1;
    uint32_t uv_offset;  // Number of bytes (HEXWORD alignment) to the UV-plane.

    // Statistics
    Fastec::IntervalTimer jpeg_timer;
    // Statistics

    TP(TRACE_SEL_JPEG_PROC, "%s: image_ptr=%p", __PRETTY_FUNCTION__, (void *)image_ptr);

    // ----------------------------------------------------------------------
// #define ALM_SAVE_RAW
#ifdef ALM_SAVE_RAW
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.raw", file_no);
        ++file_no;
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(image_ptr, request->width * request->height, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_RAW */
    // ----------------------------------------------------------------------

    // Calculate our length for the Y plane so we can find the start ov the UV plane.
    // We transfered the frame directly from image memory, so the UV plane will be on
    // a HEXWORD boundary.
    uv_offset = convert_to_HEXWORD(request->width * request->height) << HEXWORD_SHIFT;

    // Statistics
    jpeg_timer.start();

    // JPEG encode the data.
    rc = ju.encode(image_ptr, image_ptr + uv_offset, request->width, request->height, &request->jpeg_buffer);

    // Statistics
    jpeg_timer.stop();
    TP(TRACE_SEL_JPEG_PROC, "%s: JPEG encoding: %u bytes, %lf secs.", __PRETTY_FUNCTION__, rc, jpeg_timer.lastMeasurement);
    // Statistics

    // ----------------------------------------------------------------------
// #define ALM_SAVE_JPG
#ifdef ALM_SAVE_JPG
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.jpg", file_no);
        ++file_no;
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(request->jpeg_buffer, rc, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_JPG */
    // ----------------------------------------------------------------------

    return rc;
}

/*!
 * @brief Check to see if we already have a buffer for a particular
 *    image memory frame.
 */
int jpegThread::find_buf(uint64_t frame_addr)
{
    int i;

    // Loop through the pool, looking for a match.
    for (i = 0; i < (int)buf_pool.size(); ++i)
    {
        if (buf_pool[i]->frame_addr == frame_addr)
            break;
    }
    if (i == (int)buf_pool.size())
        i = -1;

    TP(TRACE_SEL_JPEG_PROC, "%s: Found 0x%" PRIx64 " at buffer %d", __PRETTY_FUNCTION__, frame_addr, i);

    // Return our findings.
    return i;
}

/*!
 * @brief Have the FPGA's IPM process the frame we just captured. All
 *    of the relevant information can be found in the ipm_info object.
 *
 * @param request   Pointer to the structure containing parameter information.
 * @return 0 on success, negative otherwise.
 */
int jpegThread::ipm_frame(JPEG_PROC::IPM_REQUEST *request)
{
    int rc = -1;
    IPM_PROC::IPM_REQUEST ipm;

    // Clear out our request structure.
    memset(&ipm, 0, sizeof(IPM_PROC::IPM_REQUEST));

    // Initialize the fields.
    ipm.image_addr = request->image_addr;
    ipm.proc_info = request->proc_info;

    // Get the raw pixel data processed properly through the IPM depending on
    // our current camera state.
    ipm.process_live = (CameraServer::get_camera_state() != GIGE_CAMERA_STATE_REVIEW);

    // Process the frame.
    if (!ipmProcRequest(&ipm))
    {
        fi_log(LOG_ERR, "%s: ipm_request for addr 0x%" PRIx64 " failed!", __PRETTY_FUNCTION__, request->image_addr);
        ipm_buf_addr = 0;
        image_len = -1;
        goto if_bail;
    }

    // Grab the output buffer address and the length.
    ipm_buf_addr = ipm.ipm_buf_addr;
    image_len = ipm.image_len;

    // All is well.
    rc = 0;

if_bail:
    // All done.
    return rc;
}

void jpegThread::refresh_pool(uint32_t width, uint32_t height)
{
    int i;
    int count;

    // Sanity check.
    if (pool_bb_index < 0)
    {
        fi_log(LOG_DEBUG, "%s: No BB allocated?", __PRETTY_FUNCTION__);
        return;
    }

    // Make sure we row-word align the width.
    pool_width = ((width + 3) / 4) * 4;
    pool_height = height;

    // How long can each pool buffer be? The "3" is for color.
    pool_buf_len = pool_width * pool_height * 3;

    // Put the size on a 4K boundary.
    pool_buf_len = ((pool_buf_len + 4095) / 4096) * 4096;

    // How many of these can we fit in our pool?
    count = bb_get_mmap_len(pool_bb_index) / pool_buf_len;
    fi_log(LOG_DEBUG, "%s: index=%d, addr=0x%" PRIx64 ", ptr=%p, count=%d, len=%d", __PRETTY_FUNCTION__, pool_bb_index, pool_addr, (void *)pool_ptr, count, pool_buf_len);

    // Clear our pool.
    buf_pool.clear();

    // Initialize our pointers.
    for (i = 0; i < count; ++i)
    {
        // Allocate our record buffer.
        std::unique_ptr<BUF_POOL_REC> rec_ptr = std::unique_ptr<BUF_POOL_REC>(new BUF_POOL_REC());
        if (rec_ptr == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: [%d] Unable to allocate the BUF_PPOOL_REC!", __PRETTY_FUNCTION__, i);
            return;
        }

        // Initialize the record.
        memset(rec_ptr.get(), 0, sizeof(BUF_POOL_REC));
        rec_ptr->buf_ptr = pool_ptr + (pool_buf_len * i);
        rec_ptr->frame_addr = -1;

        // Add the record to our array.
        buf_pool.push_back(std::move(rec_ptr));
    }
}

void jpegThread::update_ju(int pixel_format, int quality)
{
    TP(TRACE_SEL_JPEG_PROC, "%s: pixel_format=%d, quality=%d", __PRETTY_FUNCTION__, pixel_format, quality);

    // If the JPEG utility object's pixel format matches the value
    // passed in, then we are good to go. Otherwise...
    if ((ju.get_pixel_format() != pixel_format) || (ju.get_quality() != quality))
    {
        // Setup the pixel format.
        ju.set_pixel_format(pixel_format);

        // Set the quality factor.
        ju.set_quality(quality);
    }
}
