/*!
 * @file jpeg_util.cpp
 *
 * @brief JPEG utility class implementation. An abstraction for
 *    JPEG encoding using libVA.
 *
 * Derived extensively from:
 * https://github.com/intel/libva-utils/encode/jpegenc.c
 *
 * ----------------------------------------------------
 * @note Windows Media Player (WMP) has issues with monochrome
 *   MJPEG AVIs, i.e. using JPEGs with only the Y plane.
 *   Historically we've created pseudo-color JPEGs to work
 *   around this. For this generation of cameras, we've
 *   decided to stay with the smaller files and recommend
 *   people use VLC.
 * ----------------------------------------------------
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * @returns The number of bytes the header uses, or a negative number
 *      indicating an error.
 * * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <linux/limits.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <fastec/fas_log.h>

#include <memory>
#include <mutex>

#include "cs_trace.h"
#include "cs_utils.h"
#include "fi_libva.h"
#include "jpeg_util.h"

namespace JPEG_UTIL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Only for Y, U and V
#define MAX_JPEG_COMPONENTS     3

#define JPEG_Y 0
#define JPEG_U 1
#define JPEG_V 2

#define BITSTREAM_ALLOCATE_STEPPING     4096
#define NUM_QUANT_ELEMENTS          64
#define NUM_MAX_HUFFTABLE           2
#define NUM_AC_RUN_SIZE_BITS        16
#define NUM_AC_CODE_WORDS_HUFFVAL   162
#define NUM_DC_RUN_SIZE_BITS        16
#define NUM_DC_CODE_WORDS_HUFFVAL   12

enum JPEG_TRACE
{
    JPEG_FUNC       = 0x00000010,
    JPEG_FUNC_NOISY = 0x00000020
};

//As per Jpeg Spec ISO/IEC 10918-1, below values are assigned
enum jpeg_markers
{
    //Define JPEG markers as 0xFFXX if you are adding the value directly to the buffer
    //Else define marker as 0xXXFF if you are assigning the marker to a structure variable.
    //This is needed because of the little-endedness of the IA

    SOI  = 0xFFD8, //Start of Image
    EOI  = 0xFFD9, //End of Image
    SOS  = 0xFFDA, //Start of Scan
    DQT  = 0xFFDB, //Define Quantization Table
    DRI  = 0xFFDD, //Define restart interval
    RST0 = 0xFFD0, //Restart interval termination
    DHT  = 0xFFC4, //Huffman table
    SOF0 = 0xFFC0, //Baseline DCT
    APP0 = 0xFFE0, //Application Segment
    COM  = 0xFFFE  //Comment segment
};

typedef struct _YUVComponentSpecs
{
    // One of PIX_FORMAT: I420, NV12, Y8
    unsigned int yuv_type;
    // One of VA_RT_FORMAT_YUV420, VA_RT_FORMAT_YUV422, VA_RT_FORMAT_YUV400, VA_RT_FORMAT_YUV444, VA_RT_FORMAT_RGB32
    unsigned int va_surface_format;
    //One of VA_FOURCC_I420, VA_FOURCC_NV12, VA_FOURCC_UYVY, VA_FOURCC_YUY2, VA_FOURCC_Y800, VA_FOURCC_444P, VA_FOURCC_RGBA
    unsigned int fourcc_val; //Using this field to evaluate the input file type.
    //no.of. components
    unsigned int num_components;
    //Y horizontal subsample
    unsigned int y_h_subsample;
    //Y vertical subsample
    unsigned int y_v_subsample;
    //U horizontal subsample
    unsigned int u_h_subsample;
    //U vertical subsample
    unsigned int u_v_subsample;
    //V horizontal subsample
    unsigned int v_h_subsample;
    //V vertical subsample
    unsigned int v_v_subsample;
} YUVComponentSpecs;

typedef struct __bitstream
{
    unsigned int *buffer;
    int bit_offset;
    int max_size_in_dword;
} bitstream;

typedef struct _JPEGFrameHeader
{
    uint16_t SOF;    //Start of Frame Header
    uint16_t Lf;     //Length of Frame Header
    uint8_t  P;      //Sample precision
    uint16_t Y;      //Number of lines
    uint16_t X;      //Number of samples per line
    uint8_t  Nf;     //Number of image components in frame

    struct _JPEGComponent
    {
        uint8_t Ci;    //Component identifier
        uint8_t Hi: 4; //Horizontal sampling factor
        uint8_t Vi: 4; //Vertical sampling factor
        uint8_t Tqi;   //Quantization table destination selector
    } JPEGComponent[MAX_JPEG_COMPONENTS];

} JPEGFrameHeader;


typedef struct _JPEGScanHeader
{

    uint16_t SOS;  //Start of Scan
    uint16_t Ls;   //Length of Scan
    uint8_t  Ns;   //Number of image components in the scan

    struct _ScanComponent
    {
        uint8_t Csj;   //Scan component selector
        uint8_t Tdj: 4; //DC Entropy coding table destination selector(Tdj:4 bits)
        uint8_t Taj: 4; //AC Entropy coding table destination selector(Taj:4 bits)
    } ScanComponent[MAX_JPEG_COMPONENTS];

    uint8_t Ss;    //Start of spectral or predictor selection, 0 for Baseline
    uint8_t Se;    //End of spectral or predictor selection, 63 for Baseline
    uint8_t Ah: 4; //Successive approximation bit position high, 0 for Baseline
    uint8_t Al: 4; //Successive approximation bit position low, 0 for Baseline

} JPEGScanHeader;


typedef struct _JPEGQuantSection
{

    uint16_t DQT;    //Quantization table marker
    uint16_t Lq;     //Length of Quantization table definition
    uint8_t  Tq: 4;  //Quantization table destination identifier
    uint8_t  Pq: 4;  //Quatization table precision. Should be 0 for 8-bit samples
    uint8_t  Qk[NUM_QUANT_ELEMENTS]; //Quantization table elements

} JPEGQuantSection;

typedef struct _JPEGHuffSection
{

    uint16_t DHT;                            //Huffman table marker
    uint16_t Lh;                             //Huffman table definition length
    uint8_t  Tc: 4;                          //Table class- 0:DC, 1:AC
    uint8_t  Th: 4;                          //Huffman table destination identifier
    uint8_t  Li[NUM_AC_RUN_SIZE_BITS];       //Number of Huffman codes of length i
    uint8_t  Vij[NUM_AC_CODE_WORDS_HUFFVAL]; //Value associated with each Huffman code

} JPEGHuffSection;


typedef struct _JPEGRestartSection
{

    uint16_t DRI;  //Restart interval marker
    uint16_t Lr;   //Legth of restart interval segment
    uint16_t Ri;   //Restart interval

} JPEGRestartSection;


typedef struct _JPEGCommentSection
{

    uint16_t COM;  //Comment marker
    uint16_t Lc;   //Comment segment length
    uint8_t  Cmi;  //Comment byte

} JPEGCommentSection;


typedef struct _JPEGAppSection
{

    uint16_t APPn;  //Application data marker
    uint16_t Lp;    //Application data segment length
    uint8_t  Api;   //Application data byte

} JPEGAppSection;

//Luminance quantization table
//Source: Jpeg Spec ISO/IEC 10918-1, Annex K, Table K.1
static const uint8_t jpeg_luma_quant[NUM_QUANT_ELEMENTS] =
{
    16, 11, 10, 16, 24,  40,  51,  61,
    12, 12, 14, 19, 26,  58,  60,  55,
    14, 13, 16, 24, 40,  57,  69,  56,
    14, 17, 22, 29, 51,  87,  80,  62,
    18, 22, 37, 56, 68,  109, 103, 77,
    24, 35, 55, 64, 81,  104, 113, 92,
    49, 64, 78, 87, 103, 121, 120, 101,
    72, 92, 95, 98, 112, 100, 103, 99
};

//Luminance quantization table
//Source: Jpeg Spec ISO/IEC 10918-1, Annex K, Table K.2
static const uint8_t jpeg_chroma_quant[NUM_QUANT_ELEMENTS] =
{
    17, 18, 24, 47, 99, 99, 99, 99,
    18, 21, 26, 66, 99, 99, 99, 99,
    24, 26, 56, 99, 99, 99, 99, 99,
    47, 66, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99
};


//Zigzag scan order of the the Luma and Chroma components
//Note: Jpeg Spec ISO/IEC 10918-1, Figure A.6 shows the zigzag order differently.
//The Spec is trying to show the zigzag pattern with number positions. The below
//table will use the patter shown by A.6 and map the postion of the elements in the array
static const uint8_t jpeg_zigzag[] =
{
    0,   1,   8,   16,  9,   2,   3,   10,
    17,  24,  32,  25,  18,  11,  4,   5,
    12,  19,  26,  33,  40,  48,  41,  34,
    27,  20,  13,  6,   7,   14,  21,  28,
    35,  42,  49,  56,  57,  50,  43,  36,
    29,  22,  15,  23,  30,  37,  44,  51,
    58,  59,  52,  45,  38,  31,  39,  46,
    53,  60,  61,  54,  47,  55,  62,  63
};


//Huffman table for Luminance DC Coefficients
//Reference Jpeg Spec ISO/IEC 10918-1, K.3.3.1
//K.3.3.1 is the summarized version of Table K.3
static const uint8_t jpeg_hufftable_luma_dc[] =
{
    //TcTh (Tc=0 since 0:DC, 1:AC; Th=0)
    0x00,
    //Li
    0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    //Vi
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B
};

//Huffman table for Chrominance DC Coefficients
//Reference Jpeg Spec ISO/IEC 10918-1, K.3.3.1
//K.3.3.1 is the summarized version of Table K.4
static const uint8_t jpeg_hufftable_chroma_dc[] =
{
    //TcTh (Tc=0 since 0:DC, 1:AC; Th=1)
    0x01,
    //Li
    0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
    //Vi
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B
};


//Huffman table for Luminance AC Coefficients
//Reference Jpeg Spec ISO/IEC 10918-1, K.3.3.2
//K.3.3.2 is the summarized version of Table K.5
static const uint8_t jpeg_hufftable_luma_ac[] =
{
    //TcTh (Tc=1 since 0:DC, 1:AC; Th=0)
    0x10,
    //Li
    0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7D,
    //Vi
    0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
    0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xA1, 0x08, 0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0,
    0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
    0x4A, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
    0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
    0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7,
    0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5,
    0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2,
    0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8,
    0xF9, 0xFA
};

//Huffman table for Chrominance AC Coefficients
//Reference Jpeg Spec ISO/IEC 10918-1, K.3.3.2
//K.3.3.2 is the summarized version of Table K.6
static const uint8_t jpeg_hufftable_chroma_ac[] =
{
    //TcTh (Tc=1 since 0:DC, 1:AC; Th=1)
    0x11,
    //Li
    0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04, 0x00, 0x01, 0x02, 0x77,
    //Vi
    0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
    0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 0x33, 0x52, 0xF0,
    0x15, 0x62, 0x72, 0xD1, 0x0A, 0x16, 0x24, 0x34, 0xE1, 0x25, 0xF1, 0x17, 0x18, 0x19, 0x1A, 0x26,
    0x27, 0x28, 0x29, 0x2A, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4A, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
    0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5,
    0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3,
    0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA,
    0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8,
    0xF9, 0xFA
};

}

using namespace JPEG_UTIL;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static const uint32_t traceControl = JPEG_FUNC |
//                                     JPEG_FUNC_NOISY |
//                                     TRACE_0;        // Set our trace level to TRACE_0 to trace.
                                     TRACE_3;        // Set our trace level to TRACE_3 to not trace.

static std::mutex encode_mutex;                //!< Used to control access to the encoder.

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

static void bitstream_end(JPEG_UTIL::bitstream *bs);
static bool bitstream_put_ui(JPEG_UTIL::bitstream *bs, unsigned int val, int size_in_bits);
static bool bitstream_start(JPEG_UTIL::bitstream *bs);
static int build_packed_jpeg_header_buffer(unsigned char **header_buffer, JPEG_UTIL::YUVComponentSpecs *yuvComp, int picture_width, int picture_height, uint16_t restart_interval, int quality);
static void init_yuv_component(JPEG_UTIL::YUVComponentSpecs *yuvComponent, int yuv_type, int *surface_type, VASurfaceAttrib *fourcc);
static void jpegenc_hufftable_init(VAHuffmanTableBufferJPEGBaseline *hufftable_param, JPEG_UTIL::YUVComponentSpecs *yuvComp);
static void jpegenc_pic_param_init(VAEncPictureParameterBufferJPEG *pic_param, int width, int height, int quality, JPEG_UTIL::YUVComponentSpecs *yuvComp);
static void jpegenc_qmatrix_init(VAQMatrixBufferJPEG *quantization_param, JPEG_UTIL::YUVComponentSpecs *yuvComp);
static void jpegenc_slice_param_init(VAEncSliceParameterBufferJPEG *slice_param, JPEG_UTIL::YUVComponentSpecs *yuvComp);
static void populate_frame_header(JPEG_UTIL::JPEGFrameHeader *frameHdr, JPEG_UTIL::YUVComponentSpecs *yuvComp, int picture_width, int picture_height);
static void populate_huff_section_header(JPEG_UTIL::JPEGHuffSection *huffSectionHdr, int th, int tc);
static void populate_quantdata(JPEG_UTIL::JPEGQuantSection *quantVal, int type);
static void populate_scan_header(JPEG_UTIL::JPEGScanHeader *scanHdr, JPEG_UTIL::YUVComponentSpecs *yuvComp);
static void upload_yuv_to_surface(VADisplay va_dpy, uint8_t *img_y, uint8_t *img_uv, VASurfaceID surface_id, JPEG_UTIL::YUVComponentSpecs *yuvComp, int picture_width, int picture_height);
static uint32_t swap32(uint32_t val);

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

static void bitstream_end(bitstream *bs)
{
    int pos = (bs->bit_offset >> 5);
    int bit_offset = (bs->bit_offset & 0x1f);
    int bit_left = 32 - bit_offset;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    if (bit_offset)
    {
        bs->buffer[pos] = swap32((bs->buffer[pos] << bit_left));
    }
}

static bool bitstream_put_ui(JPEG_UTIL::bitstream *bs, unsigned int val, int size_in_bits)
{
    int pos;
    int bit_offset;
    int bit_left;

    TP(JPEG_FUNC_NOISY, "%s: bs=%p, val=0x%08x, size=%d", __PRETTY_FUNCTION__, (void *)bs, val, size_in_bits);

    // Sanity check.
    if (bs == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: bs pointer is NULL!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize our values.
    pos = (bs->bit_offset >> 5);
    bit_offset = (bs->bit_offset & 0x1f);
    bit_left = 32 - bit_offset;

    if (!size_in_bits)
        return false;

    if (size_in_bits < 32)
        val &= ((1U << size_in_bits) - 1);

    bs->bit_offset += size_in_bits;

    if (bit_left > size_in_bits)
    {
        bs->buffer[pos] = (bs->buffer[pos] << size_in_bits | val);
    }
    else
    {
        size_in_bits -= bit_left;
        bs->buffer[pos] = (bs->buffer[pos] << bit_left) | (val >> size_in_bits);
        bs->buffer[pos] = swap32(bs->buffer[pos]);

        if (pos + 1 == bs->max_size_in_dword)
        {
            bs->max_size_in_dword += BITSTREAM_ALLOCATE_STEPPING;
            bs->buffer = (unsigned int *)realloc(bs->buffer, bs->max_size_in_dword * sizeof(unsigned int));
            if (bs->buffer == nullptr)
                return false;
        }

        bs->buffer[pos + 1] = val;
    }

    return true;
}

static bool bitstream_start(bitstream *bs)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    bs->max_size_in_dword = BITSTREAM_ALLOCATE_STEPPING;
    bs->buffer = (unsigned int *)calloc(bs->max_size_in_dword * sizeof(unsigned int), 1);
    if (bs->buffer == nullptr)
        return false;
    bs->bit_offset = 0;
    return true;
}

// This method packs the header information which is to be sent to the driver through LibVA.
// All the information that needs to be inserted in the encoded buffer should be built and sent.
// It is the responsibility of the app talking to LibVA to build this header and send it.
// This includes Markers, Quantization tables (normalized with quality factor), Huffman tables, etc.
static int build_packed_jpeg_header_buffer(unsigned char **header_buffer, YUVComponentSpecs *yuvComp, int picture_width, int picture_height, uint16_t restart_interval, int quality)
{
    bitstream bs;
    int i;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    bitstream_start(&bs);

    //Add SOI
    bitstream_put_ui(&bs, SOI, 16);

    //Add AppData
    bitstream_put_ui(&bs, APP0, 16);  //APP0 marker
    bitstream_put_ui(&bs, 16, 16);    //Length excluding the marker
    bitstream_put_ui(&bs, 0x4A, 8);   //J
    bitstream_put_ui(&bs, 0x46, 8);   //F
    bitstream_put_ui(&bs, 0x49, 8);   //I
    bitstream_put_ui(&bs, 0x46, 8);   //F
    bitstream_put_ui(&bs, 0x00, 8);   //0
    bitstream_put_ui(&bs, 1, 8);      //Major Version
    bitstream_put_ui(&bs, 1, 8);      //Minor Version
    bitstream_put_ui(&bs, 1, 8);      //Density units 0:no units, 1:pixels per inch, 2: pixels per cm
    bitstream_put_ui(&bs, 72, 16);    //X density
    bitstream_put_ui(&bs, 72, 16);    //Y density
    bitstream_put_ui(&bs, 0, 8);      //Thumbnail width
    bitstream_put_ui(&bs, 0, 8);      //Thumbnail height

    // Regarding Quantization matrices: As per JPEG Spec ISO/IEC 10918-1:1993(E), Pg-19:
    // "applications may specify values which customize picture quality for their particular
    // image characteristics, display devices, and viewing conditions"

    // Safety check.
    if (quality == 0)
        quality = 80;

    // Normalization of quality factor
    quality = (quality < 50) ? (5000 / quality) : (200 - (quality * 2));

    //Add QTable - Y
    JPEGQuantSection quantLuma;
    populate_quantdata(&quantLuma, 0);

    bitstream_put_ui(&bs, quantLuma.DQT, 16);
    bitstream_put_ui(&bs, quantLuma.Lq, 16);
    bitstream_put_ui(&bs, quantLuma.Pq, 4);
    bitstream_put_ui(&bs, quantLuma.Tq, 4);
    for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
    {
        uint32_t temp;

        //scale the quantization table with quality factor
        temp = (quantLuma.Qk[i] * quality) / 100;
        //clamp to range [1,255]
        temp = (temp > 255) ? 255 : temp;
        temp = (temp < 1) ? 1 : temp;
        quantLuma.Qk[i] = (unsigned char)temp;
        bitstream_put_ui(&bs, quantLuma.Qk[i], 8);
    }

    //Add QTable - U/V
    if (yuvComp->fourcc_val != VA_FOURCC_Y800)
    {
        JPEGQuantSection quantChroma;
        populate_quantdata(&quantChroma, 1);

        bitstream_put_ui(&bs, quantChroma.DQT, 16);
        bitstream_put_ui(&bs, quantChroma.Lq, 16);
        bitstream_put_ui(&bs, quantChroma.Pq, 4);
        bitstream_put_ui(&bs, quantChroma.Tq, 4);
        for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
        {
            uint32_t temp;

            //scale the quantization table with quality factor
            temp = (quantChroma.Qk[i] * quality) / 100;
            //clamp to range [1,255]
            temp = (temp > 255) ? 255 : temp;
            temp = (temp < 1) ? 1 : temp;
            quantChroma.Qk[i] = (unsigned char)temp;
            bitstream_put_ui(&bs, quantChroma.Qk[i], 8);
        }
    }

    //Add FrameHeader
    JPEGFrameHeader frameHdr;
    memset(&frameHdr, 0, sizeof(JPEGFrameHeader));
    populate_frame_header(&frameHdr, yuvComp, picture_width, picture_height);

    bitstream_put_ui(&bs, frameHdr.SOF, 16);
    bitstream_put_ui(&bs, frameHdr.Lf, 16);
    bitstream_put_ui(&bs, frameHdr.P, 8);
    bitstream_put_ui(&bs, frameHdr.Y, 16);
    bitstream_put_ui(&bs, frameHdr.X, 16);
    bitstream_put_ui(&bs, frameHdr.Nf, 8);
    for (i = 0; i < frameHdr.Nf; i++)
    {
        bitstream_put_ui(&bs, frameHdr.JPEGComponent[i].Ci, 8);
        bitstream_put_ui(&bs, frameHdr.JPEGComponent[i].Hi, 4);
        bitstream_put_ui(&bs, frameHdr.JPEGComponent[i].Vi, 4);
        bitstream_put_ui(&bs, frameHdr.JPEGComponent[i].Tqi, 8);
    }

    //Add HuffTable AC and DC for Y,U/V components
    JPEGHuffSection acHuffSectionHdr, dcHuffSectionHdr;
    memset(&acHuffSectionHdr, 0, sizeof(acHuffSectionHdr));
    memset(&dcHuffSectionHdr, 0, sizeof(dcHuffSectionHdr));

    for (i = 0; (i < (int)yuvComp->num_components && (i <= 1)); i++)
    {
        int j;

        //Add DC component (Tc = 0)
        populate_huff_section_header(&dcHuffSectionHdr, i, 0);

        bitstream_put_ui(&bs, dcHuffSectionHdr.DHT, 16);
        bitstream_put_ui(&bs, dcHuffSectionHdr.Lh, 16);
        bitstream_put_ui(&bs, dcHuffSectionHdr.Tc, 4);
        bitstream_put_ui(&bs, dcHuffSectionHdr.Th, 4);
        for (j = 0; j < NUM_DC_RUN_SIZE_BITS; j++)
        {
            bitstream_put_ui(&bs, dcHuffSectionHdr.Li[j], 8);
        }

        for (j = 0; j < NUM_DC_CODE_WORDS_HUFFVAL; j++)
        {
            bitstream_put_ui(&bs, dcHuffSectionHdr.Vij[j], 8);
        }

        //Add AC component (Tc = 1)
        populate_huff_section_header(&acHuffSectionHdr, i, 1);

        bitstream_put_ui(&bs, acHuffSectionHdr.DHT, 16);
        bitstream_put_ui(&bs, acHuffSectionHdr.Lh, 16);
        bitstream_put_ui(&bs, acHuffSectionHdr.Tc, 4);
        bitstream_put_ui(&bs, acHuffSectionHdr.Th, 4);
        for (j = 0; j < NUM_AC_RUN_SIZE_BITS; j++)
        {
            bitstream_put_ui(&bs, acHuffSectionHdr.Li[j], 8);
        }

        for (j = 0; j < NUM_AC_CODE_WORDS_HUFFVAL; j++)
        {
            bitstream_put_ui(&bs, acHuffSectionHdr.Vij[j], 8);
        }

        if (yuvComp->fourcc_val == VA_FOURCC_Y800)
            break;
    }

    //Add Restart Interval if restart_interval is not 0
    if (restart_interval != 0)
    {
        JPEGRestartSection restartHdr;
        restartHdr.DRI = DRI;
        restartHdr.Lr = 4;
        restartHdr.Ri = restart_interval;

        bitstream_put_ui(&bs, restartHdr.DRI, 16);
        bitstream_put_ui(&bs, restartHdr.Lr, 16);
        bitstream_put_ui(&bs, restartHdr.Ri, 16);
    }

    //Add ScanHeader
    JPEGScanHeader scanHdr;
    populate_scan_header(&scanHdr, yuvComp);

    bitstream_put_ui(&bs, scanHdr.SOS, 16);
    bitstream_put_ui(&bs, scanHdr.Ls, 16);
    bitstream_put_ui(&bs, scanHdr.Ns, 8);

    for (i = 0; i < scanHdr.Ns; i++)
    {
        bitstream_put_ui(&bs, scanHdr.ScanComponent[i].Csj, 8);
        bitstream_put_ui(&bs, scanHdr.ScanComponent[i].Tdj, 4);
        bitstream_put_ui(&bs, scanHdr.ScanComponent[i].Taj, 4);
    }

    bitstream_put_ui(&bs, scanHdr.Ss, 8);
    bitstream_put_ui(&bs, scanHdr.Se, 8);
    bitstream_put_ui(&bs, scanHdr.Ah, 4);
    bitstream_put_ui(&bs, scanHdr.Al, 4);

    bitstream_end(&bs);
    *header_buffer = (unsigned char *)bs.buffer;

    return bs.bit_offset;
}

static void init_yuv_component(YUVComponentSpecs *yuvComponent, int yuv_type, int *surface_type, VASurfaceAttrib *fourcc)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    // Save the type.
    yuvComponent->yuv_type = yuv_type;

    // One of JPEG_UTIL::PIX_FORMAT
    switch (yuv_type)
    {
    case I420:
    case NV12:
        yuvComponent->va_surface_format = (*surface_type) = VA_RT_FORMAT_YUV420;
        if (yuv_type == I420)
        {
            yuvComponent->fourcc_val = VA_FOURCC_I420;
            fourcc->value.value.i = VA_FOURCC_NV12;
        }
        else
        {
            yuvComponent->fourcc_val = fourcc->value.value.i = VA_FOURCC_NV12;
        }
        yuvComponent->num_components = 3;
        yuvComponent->y_h_subsample = 2;
        yuvComponent->y_v_subsample = 2;
        yuvComponent->u_h_subsample = 1;
        yuvComponent->u_v_subsample = 1;
        yuvComponent->v_h_subsample = 1;
        yuvComponent->v_v_subsample = 1;
        break;
    case Y8:
        yuvComponent->va_surface_format = (*surface_type) = VA_RT_FORMAT_YUV400;
        yuvComponent->fourcc_val = fourcc->value.value.i = VA_FOURCC_Y800;
        yuvComponent->num_components = 1;
        yuvComponent->y_h_subsample = 1;
        yuvComponent->y_v_subsample = 1;
        yuvComponent->u_h_subsample = 0;
        yuvComponent->u_v_subsample = 0;
        yuvComponent->v_h_subsample = 0;
        yuvComponent->v_v_subsample = 0;
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unsupported format [0x%08x]", __PRETTY_FUNCTION__, yuv_type);
        break;
    }
}

static void jpegenc_hufftable_init(VAHuffmanTableBufferJPEGBaseline *hufftable_param, YUVComponentSpecs *yuvComp)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    hufftable_param->load_huffman_table[0] = 1; //Load Luma Hufftable
    if (yuvComp->fourcc_val == VA_FOURCC_Y800)
    {
        hufftable_param->load_huffman_table[1] = 0; //Do not load Chroma Hufftable for Y8
    }
    else
    {
        hufftable_param->load_huffman_table[1] = 1; //Load Chroma Hufftable for other formats
    }

    //Load Luma hufftable values
    //Load DC codes
    memcpy(hufftable_param->huffman_table[0].num_dc_codes, jpeg_hufftable_luma_dc + 1, 16);
    //Load DC Values
    memcpy(hufftable_param->huffman_table[0].dc_values, jpeg_hufftable_luma_dc + 17, 12);
    //Load AC codes
    memcpy(hufftable_param->huffman_table[0].num_ac_codes, jpeg_hufftable_luma_ac + 1, 16);
    //Load AC Values
    memcpy(hufftable_param->huffman_table[0].ac_values, jpeg_hufftable_luma_ac + 17, 162);
    memset(hufftable_param->huffman_table[0].pad, 0, 2);


    //Load Chroma hufftable values if needed
    if (yuvComp->fourcc_val != VA_FOURCC_Y800)
    {
        //Load DC codes
        memcpy(hufftable_param->huffman_table[1].num_dc_codes, jpeg_hufftable_chroma_dc + 1, 16);
        //Load DC Values
        memcpy(hufftable_param->huffman_table[1].dc_values, jpeg_hufftable_chroma_dc + 17, 12);
        //Load AC codes
        memcpy(hufftable_param->huffman_table[1].num_ac_codes, jpeg_hufftable_chroma_ac + 1, 16);
        //Load AC Values
        memcpy(hufftable_param->huffman_table[1].ac_values, jpeg_hufftable_chroma_ac + 17, 162);
        memset(hufftable_param->huffman_table[1].pad, 0, 2);

    }

}

static void jpegenc_pic_param_init(VAEncPictureParameterBufferJPEG *pic_param, int width, int height, int quality, YUVComponentSpecs *yuvComp)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (pic_param == nullptr)
        return;

    pic_param->picture_width = width;
    pic_param->picture_height = height;
    pic_param->quality = quality;

    pic_param->pic_flags.bits.profile = 0;      //Profile = Baseline
    pic_param->pic_flags.bits.progressive = 0;  //Sequential encoding
    pic_param->pic_flags.bits.huffman = 1;      //Uses Huffman coding
    pic_param->pic_flags.bits.interleaved = 0;  //Input format is interleaved (YUV)
    pic_param->pic_flags.bits.differential = 0; //non-Differential Encoding

    pic_param->sample_bit_depth = 8; //only 8 bit sample depth is currently supported
    pic_param->num_scan = 1;
    pic_param->num_components = yuvComp->num_components; // Supporting only upto 3 components maximum

    // set component_id Ci and Tqi
    if (yuvComp->fourcc_val == VA_FOURCC_Y800)
    {
        pic_param->component_id[0] = 0;
        pic_param->quantiser_table_selector[0] = 0;
    }
    else
    {
        pic_param->component_id[0] = pic_param->quantiser_table_selector[0] = 0;
        pic_param->component_id[1] = pic_param->quantiser_table_selector[1] = 1;
        pic_param->component_id[2] = 2;
        pic_param->quantiser_table_selector[2] = 1;
    }
}

static void jpegenc_qmatrix_init(VAQMatrixBufferJPEG *quantization_param, YUVComponentSpecs *yuvComp)
{
    int i = 0;
    quantization_param->load_lum_quantiser_matrix = 1;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    // LibVA expects the QM in zigzag order
    for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
    {
        quantization_param->lum_quantiser_matrix[i] = jpeg_luma_quant[jpeg_zigzag[i]];
    }

    // Are we monochrome?
    if (yuvComp->fourcc_val == VA_FOURCC_Y800)
    {
        quantization_param->load_chroma_quantiser_matrix = 0;
    }
    else
    {
        quantization_param->load_chroma_quantiser_matrix = 1;
        for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
        {
            quantization_param->chroma_quantiser_matrix[i] = jpeg_chroma_quant[jpeg_zigzag[i]];
        }
    }

}

static void jpegenc_slice_param_init(VAEncSliceParameterBufferJPEG *slice_param, YUVComponentSpecs *yuvComp)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    slice_param->restart_interval = 0;

    slice_param->num_components = yuvComp->num_components;

    slice_param->components[0].component_selector = 1;
    slice_param->components[0].dc_table_selector = 0;
    slice_param->components[0].ac_table_selector = 0;

    if (yuvComp->num_components > 1)
    {
        slice_param->components[1].component_selector = 2;
        slice_param->components[1].dc_table_selector = 1;
        slice_param->components[1].ac_table_selector = 1;

        slice_param->components[2].component_selector = 3;
        slice_param->components[2].dc_table_selector = 1;
        slice_param->components[2].ac_table_selector = 1;
    }
}


static void populate_quantdata(JPEGQuantSection *quantVal, int type)
{
    uint8_t zigzag_qm[NUM_QUANT_ELEMENTS];
    int i;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    quantVal->DQT = DQT;
    quantVal->Pq = 0;
    quantVal->Tq = type;
    if (type == 0)
    {
        for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
        {
            zigzag_qm[i] = jpeg_luma_quant[jpeg_zigzag[i]];
        }

        memcpy(quantVal->Qk, zigzag_qm, NUM_QUANT_ELEMENTS);
    }
    else
    {
        for (i = 0; i < NUM_QUANT_ELEMENTS; i++)
        {
            zigzag_qm[i] = jpeg_chroma_quant[jpeg_zigzag[i]];
        }
        memcpy(quantVal->Qk, zigzag_qm, NUM_QUANT_ELEMENTS);
    }
    quantVal->Lq = 3 + NUM_QUANT_ELEMENTS;
}

static void populate_frame_header(JPEGFrameHeader *frameHdr, YUVComponentSpecs *yuvComp, int picture_width, int picture_height)
{
    int i = 0;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    frameHdr->SOF = SOF0;
    frameHdr->Lf = 8 + (3 * yuvComp->num_components); //Size of FrameHeader in bytes without the Marker SOF
    frameHdr->P = 8;
    frameHdr->Y = picture_height;
    frameHdr->X = picture_width;
    frameHdr->Nf = yuvComp->num_components;

    for (i = 0; i < (int)yuvComp->num_components; i++)
    {
        frameHdr->JPEGComponent[i].Ci = i + 1;

        if (i == 0)
        {
            frameHdr->JPEGComponent[i].Hi = yuvComp->y_h_subsample;
            frameHdr->JPEGComponent[i].Vi = yuvComp->y_v_subsample;
            frameHdr->JPEGComponent[i].Tqi = 0;

        }
        else
        {
            //Analyzing the sampling factors for U/V, they are 1 for all formats except for Y8.
            //So, it is okay to have the code below like this. For Y8, we wont reach this code.
            frameHdr->JPEGComponent[i].Hi = yuvComp->u_h_subsample;
            frameHdr->JPEGComponent[i].Vi = yuvComp->u_v_subsample;
            frameHdr->JPEGComponent[i].Tqi = 1;
        }
    }
}

static void populate_huff_section_header(JPEGHuffSection *huffSectionHdr, int th, int tc)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    huffSectionHdr->DHT = DHT;
    huffSectionHdr->Tc = tc;
    huffSectionHdr->Th = th;

    if (th == 0)  //If Luma
    {
        int totalCodeWords = 0;

        //If AC
        if (tc == 1)
        {
            memcpy(huffSectionHdr->Li, jpeg_hufftable_luma_ac + 1, NUM_AC_RUN_SIZE_BITS);
            memcpy(huffSectionHdr->Vij, jpeg_hufftable_luma_ac + 17, NUM_AC_CODE_WORDS_HUFFVAL);
        }

        //If DC
        if (tc == 0)
        {
            memcpy(huffSectionHdr->Li, jpeg_hufftable_luma_dc + 1, NUM_DC_RUN_SIZE_BITS);
            memcpy(huffSectionHdr->Vij, jpeg_hufftable_luma_dc + 17, NUM_DC_CODE_WORDS_HUFFVAL);
        }

        for (int i = 0; i < NUM_AC_RUN_SIZE_BITS; i++)
        {
            totalCodeWords += huffSectionHdr->Li[i];
        }

        huffSectionHdr->Lh = 3 + 16 + totalCodeWords;

    }
    else     //If Chroma
    {
        //If AC
        if (tc == 1)
        {
            memcpy(huffSectionHdr->Li, jpeg_hufftable_chroma_ac + 1, NUM_AC_RUN_SIZE_BITS);
            memcpy(huffSectionHdr->Vij, jpeg_hufftable_chroma_ac + 17, NUM_AC_CODE_WORDS_HUFFVAL);
        }

        //If DC
        if (tc == 0)
        {
            memcpy(huffSectionHdr->Li, jpeg_hufftable_chroma_dc + 1, NUM_DC_RUN_SIZE_BITS);
            memcpy(huffSectionHdr->Vij, jpeg_hufftable_chroma_dc + 17, NUM_DC_CODE_WORDS_HUFFVAL);
        }
    }
}

static void populate_scan_header(JPEGScanHeader *scanHdr, YUVComponentSpecs *yuvComp)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    scanHdr->SOS = SOS;
    scanHdr->Ns = yuvComp->num_components;

    //Y Component
    scanHdr->ScanComponent[0].Csj = 1;
    scanHdr->ScanComponent[0].Tdj = 0;
    scanHdr->ScanComponent[0].Taj = 0;

    if (yuvComp->num_components > 1)
    {
        //U Component
        scanHdr->ScanComponent[1].Csj = 2;
        scanHdr->ScanComponent[1].Tdj = 1;
        scanHdr->ScanComponent[1].Taj = 1;

        //V Component
        scanHdr->ScanComponent[2].Csj = 3;
        scanHdr->ScanComponent[2].Tdj = 1;
        scanHdr->ScanComponent[2].Taj = 1;
    }

    scanHdr->Ss = 0;  //0 for Baseline
    scanHdr->Se = 63; //63 for Baseline
    scanHdr->Ah = 0;  //0 for Baseline
    scanHdr->Al = 0;  //0 for Baseline

    scanHdr->Ls = 3 + (yuvComp->num_components * 2) + 3;

}

/*!
 * @brief Upload the YUV image to the VASurface.
 */
static void upload_yuv_to_surface(VADisplay va_dpy, uint8_t *img_y, uint8_t *img_uv, VASurfaceID surface_id, YUVComponentSpecs *yuvComp, int picture_width, int picture_height)
{
    VAImage surface_image;
    VAStatus va_status;
    void *surface_p = NULL;
    unsigned char *y_src, *u_src, *v_src;
    unsigned char *y_dst, *u_dst;
    int y_size = picture_width * picture_height;
    int u_size = 0;

    TP(JPEG_FUNC, "%s: %dx%d", __PRETTY_FUNCTION__, picture_width, picture_height);

    //u_size is used for I420, NV12 formats only
    u_size = ((picture_width >> 1) * (picture_height >> 1));

    va_status = vaDeriveImage(va_dpy, surface_id, &surface_image);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaDeriveImage() failed!", __PRETTY_FUNCTION__);
        return;
    }

    vaMapBuffer(va_dpy, surface_image.buf, &surface_p);

    y_src = img_y;
    if (img_uv == nullptr)
    {
        u_src = img_y + y_size; /* UV offset for NV12 */
        v_src = img_y + y_size + u_size;
    }
    else
    {
        u_src = img_uv; /* UV offset for Y8 */
        v_src = img_uv + u_size;
    }

//    LOG("%s: %dx%d, num_planes: %d, offsets: [%d, %d, %d], pitches: [%d, %d, %d]", __PRETTY_FUNCTION__,
//        surface_image.width, surface_image.height, surface_image.num_planes,
//        surface_image.offsets[0], surface_image.offsets[1], surface_image.offsets[2],
//        surface_image.pitches[0], surface_image.pitches[1], surface_image.pitches[2]);
//    LOG("%s fourcc_val: 0x%x, num_components: %d", __PRETTY_FUNCTION__, yuvComp->fourcc_val, yuvComp->num_components);

    y_dst = ((uint8_t *)surface_p) + surface_image.offsets[0];
    u_dst = ((uint8_t *)surface_p) + surface_image.offsets[1]; /* UV offset for NV12 */

    if ((yuvComp->fourcc_val == VA_FOURCC_NV12) ||
        (yuvComp->fourcc_val == VA_FOURCC_I420) ||
        (yuvComp->fourcc_val == VA_FOURCC_Y800))
    {
        int row;

        /* Y plane */
        for (row = 0; row < surface_image.height; row++)
        {
            memcpy(y_dst, y_src, surface_image.width);
            y_dst += surface_image.pitches[0];
            y_src += picture_width;
        }

        if (yuvComp->num_components > 1)
        {
            switch (yuvComp->fourcc_val)
            {
            case VA_FOURCC_NV12:
                for (row = 0; row < surface_image.height / 2; row++)
                {
                    memcpy(u_dst, u_src, surface_image.width);
                    u_dst += surface_image.pitches[1];
                    u_src += (picture_width);
                }
                break;
            case VA_FOURCC_I420:
                for (row = 0; row < surface_image.height / 2; row++)
                {
                    for (int col = 0; col < surface_image.width / 2; col++)
                    {
                        u_dst[col * 2] = u_src[col];
                        u_dst[col * 2 + 1] = v_src[col];
                    }

                    u_dst += surface_image.pitches[1];
                    u_src += (picture_width / 2);
                    v_src += (picture_width / 2);
                }
                break;
            }
        }
    }

    vaUnmapBuffer(va_dpy, surface_image.buf);
    vaDestroyImage(va_dpy, surface_image.image_id);
}

static uint32_t swap32(uint32_t val)
{
    union
    {
        uint32_t    l_val;
        uint8_t     b_val[4];
    } swapper;
    swapper.l_val = val;

    //Swap the bytes around. We'll just re-use 'val'.
    val = ((swapper.b_val[0] << 24)     |
           (swapper.b_val[1] << 16)     |
           (swapper.b_val[2] << 8)      |
           (swapper.b_val[3] << 0));

    // Return the swapped value.
    return val;
}

//--------------------------------------------------------------------
// JpegUtil Implementation
//--------------------------------------------------------------------

JpegUtil::JpegUtil() :
    pixelFormat(JPEG_UTIL::NONE),
    quality(80)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);
}

JpegUtil::~JpegUtil()
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Sets the pixel format of the images to be worked with.
 *
 * @param pixelFormat    The pixel format. Can be: JPEG_UTIL::GRAY,
 *           JPEG_UTIL::I420 or JPEG_UTIL::NV12.
 */
void JpegUtil::set_pixel_format(int pixelFormat)
{
    // Remember the pixel format.
    this->pixelFormat = pixelFormat;
}

/*!
 * @brief Gets the current pixel format.
 * @return  The pixel format.
 */
int JpegUtil::get_pixel_format() const
{
    return pixelFormat;
}

/*!
 * @brief Sets a compression quality.
 * @param quality  the quality in percents [0, 100]
 */
void JpegUtil::set_quality(int quality)
{
    if (quality <= 0)
        quality = 80;
    this->quality = quality;
}


/*!
 * @brief Gets the current compression quality.
 * @return  the quality value
 */
int JpegUtil::get_quality() const
{
    return quality;
}

/*!
 * @brief Release a buffer allocated by the JpegUtil object.
 *    There's no need to call this if you supplied the buffer.
 *
 * @param enc_buf Pointer to a buffer allocated by the object.
 */
void JpegUtil::release(uint8_t *enc_buf)
{
    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);
    delete []enc_buf;
}

/*!
 * @brief JPEG encode an NV12 image. Provide the pointer to the Y and UV
 *    planes separately. This is due to how the FPGA creates the NV12
 *    image.
 *
 * @param img_y     Pointer to the Y plane.
 * @param img_uv    Pointer to the UV (CbCr) plane. Note, for Y8 this will be nullptr.
 * @param width     Width of a row. It is assumed any row word alignment has already been applied.
 * @param height    Number of rows in an image.
 * @param enc_buf   Pointer to a buffer pointer. This can be initialized to
 *          nullptr, in which case the buffer will be dynamically allocated
 *          and need to be released. Or the user can supply a buffer.
 *
 * @return The number of bytes in the encoded buffer.
 */
int JpegUtil::encode(uint8_t *img_y, uint8_t *img_uv, int width, int height, uint8_t **enc_buf)
{
    uint32_t frame_size = 0;
    uint32_t out_size = 0;
    uint32_t config_id;
    FiLibVA *display;
    VAStatus va_status;
    VADisplay va_dpy = nullptr;
    VASurfaceID surface_id;
    VAContextID context_id;
    VABufferID packed_raw_header_param_buf_id;  /* Header parameter buffer id */
    VABufferID packed_raw_header_buf_id;        /* Header buffer id */
    VABufferID slice_param_buf_id;              /* Slice parameter id, only 1 slice per frame in jpeg encode */
    VAEncSliceParameterBufferJPEG slice_param;  /* Slice parameter buffer */
    VABufferID huffmantable_buf_id;             /* Huffman table id*/
    VAHuffmanTableBufferJPEGBaseline hufftable_param; /* Huffman table buffer */
    VABufferID pic_param_buf_id;                /* Picture parameter id*/
    VAEncPictureParameterBufferJPEG pic_param;  /* Picture parameter buffer */
    VABufferID qmatrix_buf_id;                  /* Quantization Matrix id */
    VAQMatrixBufferJPEG quantization_param;     /* Quantization Matrix buffer */
    VABufferID codedbuf_buf_id;                 /* Output buffer id, compressed data */
    VASurfaceStatus surface_status;
    VACodedBufferSegment *coded_buffer_segment;
    VAEncPackedHeaderParameterBuffer packed_header_param_buffer;
    unsigned char *coded_mem;
    unsigned int length_in_bits;
    unsigned char *packed_header_buffer = nullptr;

    TP(JPEG_FUNC, "%s: img_y=%p, img_uv=%p, %dx%d", __PRETTY_FUNCTION__, (void *)img_y, (void *)img_uv, width, height);

    // ----------------------------------------------------------------------
//#define ALM_SAVE_DATA
#ifdef ALM_SAVE_DATA
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;
        uint32_t c_len;
        uint32_t y_len;

        // Number of bytes in the Y plane.
        y_len = width * height;

        // Calculate the number of bytes in the UV plane.
        c_len = y_len >> 1;

        snprintf(file_path, PATH_MAX, "/tmp/image_%0udy_%0uduv_%03d.raw", y_len, c_len, file_no);
        // ++file_no;   // Done when we save the encoded data.

        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(img_y, y_len, 1, fd);

            switch (pixelFormat)
            {
            case I420:
            case NV12:
                fwrite(img_uv, c_len, 1, fd);
                break;
            case Y8:
            default:
                break;
            }

            fclose(fd);
        }
    }
#endif /* ALM_SAVE_DATA */
    // ----------------------------------------------------------------------

    // Lock access to the routine.
    std::unique_lock<std::mutex> locker(encode_mutex);

    // Intialize our IDs.
    surface_id = VA_INVALID_ID;
    config_id = VA_INVALID_ID;
    context_id = VA_INVALID_ID;
    packed_raw_header_param_buf_id = VA_INVALID_ID;
    packed_raw_header_buf_id = VA_INVALID_ID;
    slice_param_buf_id = VA_INVALID_ID;
    huffmantable_buf_id = VA_INVALID_ID;
    pic_param_buf_id = VA_INVALID_ID;
    qmatrix_buf_id = VA_INVALID_ID;
    codedbuf_buf_id = VA_INVALID_ID;

    // Sanity check.
    if (pixelFormat == NONE)
    {
        fi_log(LOG_DEBUG, "%s: You must specify pixelFormat first!", __PRETTY_FUNCTION__);
        return 0;
    }

    switch (pixelFormat)
    {
    case I420:
    case NV12:
        frame_size = width * height + ((width * height) >> 1);
        break;
    case Y8:
        frame_size = width * height;
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Invalid pixel formate [0x%x]", __PRETTY_FUNCTION__, pixelFormat);
        frame_size = 0;
        break;
    }
    if (frame_size == 0)
        goto bail;
//    LOG("%s: pixelFormat=0x%x, frame_size=%d", __PRETTY_FUNCTION__, pixelFormat, frame_size);

    // Get our display.
    // Try to open the VA driver.
    display = FiLibVA::factory();
    if (display == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: FiLibVA::factory() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }
    va_dpy = display->vaOpenDRM();
    if (va_dpy == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: vaOpenDRM() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Initialize our compressor.
    if (!init_compressor(va_dpy, config_id))
    {
        fi_log(LOG_DEBUG, "%s: init_compressor() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Create Surface for the input picture.
    va_status = vaCreateSurfaces(va_dpy, surface_type, width, height, &surface_id, 1, fourcc.get(), 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateSurfaces() failed [0x%x]!", __PRETTY_FUNCTION__, va_status);
        goto bail;
    }

    // Map the input YUV data to the input surface created with the surface_id.
    upload_yuv_to_surface(va_dpy, img_y, img_uv, surface_id, yuvComponent.get(), width, height);

    // Create Context for the encode pipe.
    va_status = vaCreateContext(va_dpy, config_id, width, height,
                                VA_PROGRESSIVE, &surface_id, 1, &context_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateContext() failed [0x%x]!", __PRETTY_FUNCTION__, va_status);
        goto bail;
    }

    // Create buffer for Encoded data to be stored.
    va_status = vaCreateBuffer(va_dpy, context_id, VAEncCodedBufferType,
                               frame_size, 1, NULL, &codedbuf_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(encoded data) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Sanity check.
    if (quality <= 0)
    {
        fi_log(LOG_ERR, "%s[%d]: JPEG quality is zero?!", __PRETTY_FUNCTION__, __LINE__);
        quality = 80;
    }

    // Initialize the picture parameter buffer.
    pic_param.coded_buf = codedbuf_buf_id;
    jpegenc_pic_param_init(&pic_param, width, height, quality, yuvComponent.get());

    // Create buffer for the picture parameter.
    va_status = vaCreateBuffer(va_dpy, context_id, VAEncPictureParameterBufferType,
                               sizeof(VAEncPictureParameterBufferJPEG), 1, &pic_param, &pic_param_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(picture) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Load the QMatrix.
    jpegenc_qmatrix_init(&quantization_param, yuvComponent.get());

    // Create buffer for Quantization Matrix.
    va_status = vaCreateBuffer(va_dpy, context_id, VAQMatrixBufferType,
                               sizeof(VAQMatrixBufferJPEG), 1, &quantization_param, &qmatrix_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(QMatrix) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Load the Huffman Tables.
    jpegenc_hufftable_init(&hufftable_param, yuvComponent.get());

    // Create buffer for Huffman Tables.
    va_status = vaCreateBuffer(va_dpy, context_id, VAHuffmanTableBufferType,
                               sizeof(VAHuffmanTableBufferJPEGBaseline), 1, &hufftable_param, &huffmantable_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(Huffman) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Initialize the slice parameter buffer.
    jpegenc_slice_param_init(&slice_param, yuvComponent.get());

    // Create buffer for slice parameter.
    va_status = vaCreateBuffer(va_dpy, context_id, VAEncSliceParameterBufferType,
                               sizeof(slice_param), 1, &slice_param, &slice_param_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(slice) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Pack headers and send using Raw data buffer
    length_in_bits = build_packed_jpeg_header_buffer(&packed_header_buffer, yuvComponent.get(), width, height, slice_param.restart_interval, quality);
    packed_header_param_buffer.type = VAEncPackedHeaderRawData;
    packed_header_param_buffer.bit_length = length_in_bits;
    packed_header_param_buffer.has_emulation_bytes = 0;

    // Create raw buffer for header.
    va_status = vaCreateBuffer(va_dpy,
                               context_id,
                               VAEncPackedHeaderParameterBufferType,
                               sizeof(packed_header_param_buffer), 1, &packed_header_param_buffer,
                               &packed_raw_header_param_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(header param) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaCreateBuffer(va_dpy,
                               context_id,
                               VAEncPackedHeaderDataBufferType,
                               (length_in_bits + 7) / 8, 1, packed_header_buffer,
                               &packed_raw_header_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateBuffer(header data) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Begin picture.
    va_status = vaBeginPicture(va_dpy, context_id, surface_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaBeginPicture() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Render picture for all the VA buffers created.
    va_status = vaRenderPicture(va_dpy, context_id, &pic_param_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(pic_param) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaRenderPicture(va_dpy, context_id, &qmatrix_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(qmatrix) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaRenderPicture(va_dpy, context_id, &huffmantable_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(huffman) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaRenderPicture(va_dpy, context_id, &slice_param_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(slice) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaRenderPicture(va_dpy, context_id, &packed_raw_header_param_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(packed param) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaRenderPicture(va_dpy, context_id, &packed_raw_header_buf_id, 1);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaRenderPicture(header) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaEndPicture(va_dpy, context_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaEndPicture() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaSyncSurface(va_dpy, surface_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaSyncSurface() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    surface_status = (VASurfaceStatus)0;
    va_status = vaQuerySurfaceStatus(va_dpy, surface_id, &surface_status);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaQuerySurfaceStatus() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    va_status = vaMapBuffer(va_dpy, codedbuf_buf_id, (void **)(&coded_buffer_segment));
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaMapBuffer() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    coded_mem = (unsigned char *)coded_buffer_segment->buf;
    if (coded_buffer_segment->status & VA_CODED_BUF_STATUS_SLICE_OVERFLOW_MASK)
    {
        fi_log(LOG_DEBUG, "%s: ERROR......Coded buffer too small", __PRETTY_FUNCTION__);
    }
    else
    {
        uint8_t *dest_buf;

        // Copy the encoded image over to the buffer.
        out_size = coded_buffer_segment->size;
        if (*enc_buf == nullptr)
        {
            TP(JPEG_FUNC, "%s: Need to allocate buffer.", __PRETTY_FUNCTION__);
            *enc_buf = new uint8_t[out_size];
        }

        // Get our destination pointer.
        dest_buf = (uint8_t *)(*enc_buf);
        memcpy(dest_buf, coded_mem, out_size);
        TP(JPEG_FUNC, "%s: Copied JPEG data to %p", __PRETTY_FUNCTION__, (void *)dest_buf);
    }

    va_status = vaUnmapBuffer(va_dpy, codedbuf_buf_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaUnmapBuffer() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    TP(JPEG_FUNC, "%s: enc_buf=%p, length=%d", __PRETTY_FUNCTION__, (void *)*enc_buf, out_size);

bail:
    // Cleanup.
    if (va_dpy != nullptr)
    {
        vaDestroyBuffer(va_dpy, pic_param_buf_id);
        vaDestroyBuffer(va_dpy, qmatrix_buf_id);
        vaDestroyBuffer(va_dpy, slice_param_buf_id);
        vaDestroyBuffer(va_dpy, huffmantable_buf_id);
        vaDestroyBuffer(va_dpy, codedbuf_buf_id);
        vaDestroyBuffer(va_dpy, packed_raw_header_param_buf_id);
        vaDestroyBuffer(va_dpy, packed_raw_header_buf_id);
        vaDestroySurfaces(va_dpy, &surface_id, 1);
        vaDestroyContext(va_dpy, context_id);
        vaDestroyConfig(va_dpy, config_id);
        vaTerminate(va_dpy);
        display->vaCloseDRM();
    }
    if (packed_header_buffer != nullptr)
    {
        free(packed_header_buffer);
        packed_header_buffer = nullptr;
    }

    // Return the number of bytes in the output buffer.
    return out_size;
}

/*!
 * @brief Initialize the compressor, which means to initialize
 *    the VA library and context(s).
 */
bool JpegUtil::init_compressor(VADisplay va_dpy, uint32_t& config_id)
{
    int num_entrypoints;
    int enc_entrypoint;
    int major_ver, minor_ver;
    bool rc(false);
    VAStatus va_status;
    VAConfigAttrib attrib[2];
    VAConfigAttribValEncJPEG jpeg_attrib_val;
    std::unique_ptr<VAEntrypoint[]> entrypoints;

    TP(JPEG_FUNC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (pixelFormat == NONE)
    {
        fi_log(LOG_DEBUG, "%s: You must specify pixelFormat first!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Initialize the library.
    va_status = vaInitialize(va_dpy, &major_ver, &minor_ver);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaInitialize() failed! [0x%x]", __PRETTY_FUNCTION__, va_status);
        goto bail;
    }

    // Allocate memory for the entry points.
    entrypoints = std::unique_ptr<VAEntrypoint[]>(new VAEntrypoint[vaMaxNumEntrypoints(va_dpy)]());
    if (entrypoints.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: unable to allocate entrypoints!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Query for the entry points for the JPEG Baseline profile.
    va_status = vaQueryConfigEntrypoints(va_dpy, VAProfileJPEGBaseline, entrypoints.get(), &num_entrypoints);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaQueryConfigEntrypoints() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // We need picture level encoding (VAEntrypointEncPicture). Find if it is supported.
    for (enc_entrypoint = 0; enc_entrypoint < num_entrypoints; enc_entrypoint++)
    {
        if (entrypoints[enc_entrypoint] == VAEntrypointEncPicture)
            break;
    }
    if (enc_entrypoint == num_entrypoints)
    {
        // No JPEG Encode (VAEntrypointEncPicture) entry point found.
        fi_log(LOG_DEBUG, "%s: No JPEG Encode (VAEntrypointEncPicture) entry point found", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Query for the Render Target (RT) format supported.
    attrib[0].type = VAConfigAttribRTFormat;
    attrib[1].type = VAConfigAttribEncJPEG;
    vaGetConfigAttributes(va_dpy, VAProfileJPEGBaseline, VAEntrypointEncPicture, &attrib[0], 2);

    // RT should be one of below.
    if (!((attrib[0].value & VA_RT_FORMAT_YUV420) || (attrib[0].value & VA_RT_FORMAT_YUV400)))
    {
        /* Did not find the supported RT format */
        fi_log(LOG_ERR, "%s: Unable to find a support RT format!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Move the attribute value into a format that's easier to modify.
    jpeg_attrib_val.value = attrib[1].value;

    // Set JPEG profile attribs.
    jpeg_attrib_val.bits.arithmatic_coding_mode = 0;
    jpeg_attrib_val.bits.progressive_dct_mode = 0;
    jpeg_attrib_val.bits.non_interleaved_mode = 1;
    jpeg_attrib_val.bits.differential_mode = 0;

    // Move the updated value back.
    attrib[1].value = jpeg_attrib_val.value;

    // Create Config for the profile=VAProfileJPEGBaseline, entrypoint=VAEntrypointEncPicture,
    // with RT format attribute.
    va_status = vaCreateConfig(va_dpy, VAProfileJPEGBaseline, VAEntrypointEncPicture,
                               &attrib[0], 2, &config_id);
    if (va_status != VA_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: vaCreateConfig() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Initialize our fourcc and surface information.
    yuvComponent = std::unique_ptr<YUVComponentSpecs>(new YUVComponentSpecs);
    fourcc = std::unique_ptr<VASurfaceAttrib>(new VASurfaceAttrib);
    fourcc->type = VASurfaceAttribPixelFormat;
    fourcc->flags = VA_SURFACE_ATTRIB_SETTABLE;
    fourcc->value.type = VAGenericValueTypeInteger;
    init_yuv_component(yuvComponent.get(), pixelFormat, &surface_type, fourcc.get());

    // All done.
    rc = true;

bail:
    // Return our result.
    return rc;
}

