/*!
 * @file lic_manager.cpp
 * @brief Contains definitions for methods used by the license manager.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <linux/limits.h>

#include <memory>
#include <string>
#include <fmt/format.h>
#include <fmt/printf.h>

#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include <fastec/fi_sensor_types.h>

#include "camera_server.h"
#include "lic_manager.h"
#include "sensor.h"
#include "storage.h"

namespace LIC_MANAGER
{
//--------------------------------------------------------------------
// Constants, defines, Typedefs and Macros
//--------------------------------------------------------------------

#define CAL_MOUNT_FOLDER    ".config/Fastec/calibration"
#define CAL_BACKUP_FOLDER   "/tmp/calbak"

// Fastec Productions fingerprint.
const char *FASTEC_FINGERPRINT = "B3AE4C6626CE86B3D91FFD9EFE65770923F5E49E";
const char *SIG_CHECK_SCRIPT = ".config/Fastec/scripts/verify_sig.sh";

// Make sure these are all in agreement!
const char *LICENSE_FILE    = "fi_lic.conf";
const char *VALIDATED_LICENSE_FILE  = CAL_MOUNT_FOLDER"/fi_lic.conf";

const char *TS3_LEGACY_COLOR_MODEL = "TS3100SC8064";
const char *TS3_LEGACY_MONO_MODEL = "TS3100SM8064";

// Defaults for missing or invalid license
const uint32_t DEFAULT_MEMORY = 2048;
const uint32_t DEFAULT_WIDTH = 800;
const uint32_t DEFAULT_HEIGHT = 600;
const uint32_t DEFAULT_FPS = 1000;
const uint32_t DEFAULT_MAX_FPS = 5000;

//
// Major Version 1 is read via fscanf, so we need to define the number of fields to expect
//
#define VER_1_0_FIELDS      11  // This one apparently lacked the model field.
#define VER_1_1_FIELDS      12  // This one has the model field. (Rev 1.5)
#define VER_1_2_FIELDS      13  // This one has ltr and model fields. (Rev 1.6)
#define VER_1_3_FIELDS      14  // This one adds ssd_bandwidth. (Rev 1.7)
#define VER_1_4_FIELDS      15  // This one adds ltr_max_fps. (Rev 2.1.12)
#define VER_1_5_FIELDS      16  // This one adds dvr_playback.
#define VER_1_6_FIELDS      17  // This one adds auto-trigger.
#define VER_1_7_FIELDS      18  // Add the serial number field.
#define VER_1_8_FIELDS      20  // Add the integrated and optical fields.

// Maximum number of characters in a line of the license file.
const int MAX_LINE_LEN = 256;

// Bump the major rev to indicate complete incompatibility with prior structures
const uint32_t LM_STRUCTURE_MAJOR_REV = 1;

// Bump the minor rev when previous values are left alone, but new values are
// added to the end of the structure.
const uint32_t LM_STRUCTURE_MINOR_REV = 8;

//! Maximum length for the model name, including terminating null.
const int LM_CAMERA_MODEL_MAX_LEN = 32;

/*!
 * @struct LIC_OPTIONS
 * @brief Licensed capabilities for a ft camera
 *
 * If the major rev matches the define above, then the structure can be used if
 * the client minor rev is less than the structure minor rev.
 *
 * major_rev is also used to signify that the structure is initialized.
 * if major_rev == 0, then the structure is not yet initialized.  This logic is
 * handled in lm_load_options() and the client only needs to check for NULL.
 *
 * Each of the byte (on/off) values may also be used later for further
 * information.  One example would be if it was necessary to clients to know
 * the kind of wireless card in addition to knowing that wireless was licensed.
 *
 */
typedef struct LIC_OPTIONS
{
    uint32_t    major_rev;  //!< Major rev for strucure/file
    uint32_t    minor_rev;  //!< Minor rev for structure/file

    uint32_t    max_width;  //!< Maximum pixel width for the camera
    uint32_t    max_height; //!< Maximum pixed height for the camera
    uint32_t    max_fps;    //!< Maximum frame rate for the camera
    uint32_t    max_memory; //!< Maximum amount of image memory in MB
    uint32_t    max_ssd;    //!< Maximum size of the SSD in GB

    bool        is_valid;   //!< True if the license is valid.
    uint8_t     segmented_memory;//!< Set != 0 if segmented memory is licensed
    uint8_t     video_trigger;  //!< Set != 0 if video trigger is licensed
    uint8_t     auto_exposure;  //!< Set != 0 if auto exposure is licensed

    uint8_t     irig;       //!< Set != 0 if irig is licensed
    uint8_t     gps;        //!< Set != 0 if gps is licensed
    uint8_t     wireless;   //!< Set != 0 if wireless is licensed

    // Added as of minor rev 2, but strangely not at the end.
    uint8_t ltr_mode;   //!< Set to indicate LTR mode licensing, one of
    //!< LM_LTR_NONE, LM_LTR_ONLY, LM_LTR_DUAL

    // Added as of minor rev 1
    std::string model;          //!< Set to the camera model.

    // Added as of minor rev 3
    uint32_t    ssd_bandwidth;  //!< Max data rate for SSD in MB per second.

    // Added as of minor rev 4
    uint32_t    max_ltr_fps;    //!< Max fps for LTR mode (based on SSD limits).

    // Added as of minor rev 5
    uint32_t    dvr_playback;   //!< Non-zero if we are operating in DVR playback mode.

    // Added as of minor rev 6
    uint32_t    auto_trigger;   //!< Non-zero if we are operating in auto-trigger mode.

    // Add as of minor rev 7
    std::string serial_no;      //!< Serial number of the camera head.

    // Add as of minor rev 8
    uint32_t    integrated;     //!< Non-zero if we are an integrated camera, 0 for a 'split' camera.
    uint32_t    optical;        //!< Non-zero if we are using an optical interface, 0 for a copper interface.

} LIC_OPTIONS;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//! The structure holding our licensing values.
LIC_OPTIONS options;

} /* End of namespace LIC_MANAGER */

// Make things easier on us.
using namespace LIC_MANAGER;

std::string LicManager::UNLICENSED("<unlicensed>");
std::string LicManager::BAD_LICENSE("HSbadlicense");
std::string LicManager::NO_SERIAL_NUMBER("0");

//----------------------------------------------------------------------------

LicManager::LicManager()
{
    // Zero out licensed options structure.
    memset((char *)&options, 0, sizeof(LIC_OPTIONS));

    // Check the signature for the license file.
    if (!signatureCheck())
    {
        fi_log(LOG_WARNING, "(%s) Invalid signature, loading defaults", __FUNCTION__);

        // Fill structure with defaults, and mark as not from a valid license.
        defaultOptions(&options);
    }

    // Try to open the validated license.
    else if (!licenseRead(&options))
    {
        fi_log(LOG_WARNING, "(%s) Unable to load license file, loading defaults", __FUNCTION__);

        // Fill structure with defaults, and mark as not from a valid license.
        defaultOptions(&options);
    }

    if (options.is_valid && ((options.major_rev != LM_STRUCTURE_MAJOR_REV) ||
                             (options.minor_rev != LM_STRUCTURE_MINOR_REV)))
    {
        fi_log(LOG_INFO, "%s: Found license file version %u.%u", __FUNCTION__,
               options.major_rev, options.minor_rev);

        options.major_rev = LM_STRUCTURE_MAJOR_REV;
        options.minor_rev = LM_STRUCTURE_MINOR_REV;
    }
}

LicManager::~LicManager()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

bool LicManager::autoTrigger() const
{
    return (options.is_valid ? (options.auto_trigger != 0) : false);
}

bool LicManager::dvrPlayback() const
{
    return (options.is_valid ? (options.dvr_playback != 0) : false);
}

bool LicManager::integrated() const
{
    return (options.is_valid ? (options.integrated != 0) : false);
}

bool LicManager::IRIG() const
{
    return (options.is_valid ? (options.irig != 0) : false);
}

bool LicManager::optical() const
{
    return (options.is_valid ? (options.optical != 0) : true);
}

bool LicManager::wireless() const
{
    return (options.is_valid ? (options.wireless != 0) : false);
}

uint32_t LicManager::heightMax() const
{
    return (options.is_valid ? options.max_height : DEFAULT_HEIGHT);
}

uint32_t LicManager::widthMax() const
{
    return (options.is_valid ? options.max_width : DEFAULT_WIDTH);
}

uint32_t LicManager::fpsLtrMax() const
{
    return (options.is_valid ? options.max_ltr_fps : DEFAULT_FPS);
}

uint32_t LicManager::fpsMax() const
{
    return (options.is_valid ? options.max_fps : DEFAULT_MAX_FPS);
}

uint32_t LicManager::ltrMode() const
{
    return (options.is_valid ? options.ltr_mode : LIC_MANAGER::LM_LTR_NONE);
}

uint32_t LicManager::memoryMax() const
{
    return (options.is_valid ? options.max_memory : DEFAULT_MEMORY);
}

const char *LicManager::model() const
{
    return options.model.c_str();
}

uint32_t LicManager::ssdBandwidth() const
{
    return (options.is_valid ? options.ssd_bandwidth : 0);
}

uint32_t LicManager::ssdMax() const
{
    return (options.is_valid ? options.max_ssd : 0);
}

const char *LicManager::serial_no() const
{
    return options.serial_no.c_str();
}

/*!
 * @brief Compare two-character board revs a and b like strcmp.
 *
 * @param a Left-hand side of comparison, e.g. "A3".
 * @param b Right-hand side of comparison.
 *
 * @return -1 if a < b, 0 if a == b, 1 if a > b
 */
int LicManager::revCompare(const char *a, const char *b)
{
    int rc;

    // This logic doesn't care what the format is but our revisions are
    // an upper case letter followed by a number.
    if (a[0] > b[0])
        rc = 1;
    else if (a[0] < b[0])
        rc = -1;
    else if (a[1] > b[1])
        rc = 1;
    else if (a[1] < b[1])
        rc = -1;
    else
        rc = 0;

    return rc;
}

/*!
 * @brief Initialize an options structure to default values.
 */
void LicManager::defaultOptions(LIC_MANAGER::LIC_OPTIONS *options)
{
    // Fill structure with defaults, and mark as not from a valid license.
    options->major_rev = LM_STRUCTURE_MAJOR_REV;
    options->minor_rev = LM_STRUCTURE_MINOR_REV;

    options->max_width = DEFAULT_WIDTH;
    options->max_height = DEFAULT_HEIGHT;
    options->max_fps = DEFAULT_FPS;
    options->max_memory = DEFAULT_MEMORY;
    options->max_ssd = 0;

    options->is_valid = 0;
    options->segmented_memory = 0;
    options->video_trigger = 0;
    options->auto_exposure = 0;

    options->irig = 0;
    options->gps = 0;
    options->wireless = 0;

    options->ltr_mode = 0;
    options->model.assign(UNLICENSED);
    options->ssd_bandwidth = 0;
    options->max_ltr_fps = DEFAULT_FPS;
    options->dvr_playback = 0;
    options->auto_trigger = 0;
    options->integrated = 0;
    options->optical = 1;
    options->serial_no.assign(NO_SERIAL_NUMBER);
}

/*
 * @brief Read license options from a license file
 *
 * @return 1 on success 0 on failure
 */
bool LicManager::licenseRead(LIC_OPTIONS *options)
{
    std::string temp_path;
    FILE *lic_file;
    bool bOK = false;

    // See if we can open the "official" license file.
    temp_path = fmt::sprintf("%s/%s", get_home_path(), VALIDATED_LICENSE_FILE);
    lic_file = fopen(temp_path.c_str(), "r");
    if (lic_file == NULL)
    {
        fi_log(LOG_ERR, "(%s) Unable to open license files %s: %d(%s)", __FUNCTION__,
               temp_path.c_str(), errno, strerror(errno));
    }
    else
    {
        // Read the license file information.
        bOK = licenseParse(options, lic_file);
        if (!bOK)
        {
            fi_log(LOG_ERR, "(%s) Invalid license file format!", __FUNCTION__);
        }

        // Close our file and return the result.
        fclose(lic_file);
    }

    return bOK;
}

/**
 * @brief Attempt to read a license file with major_ver 1.
 *
 *      On success, the is_valid major_rev and minor_rev fields
 *      are also filled in.
 *
 *      Note that minor_rev is set to the rev we actually read from the file,
 *      so that the caller can look at that to determine if the file needs
 *      to be re-written with a newer revision file.
 *
 * @return 1 on success 0 on failure
 */
bool LicManager::licenseParse(LIC_OPTIONS *options, FILE *lic_file)
{
    int     num_fields;
    std::unique_ptr<char[]> buf;
    char    *p;

    // Sanity check.
    if (lic_file == nullptr)
        return 0;

    // TODO: Need to double check validated license.  File format needs
    // signature.  Perhaps use xml signature syntax defined in
    // http://www.w3.org/TR/xmldsig-core/

    // Double check that the max len matches our expectations
    // Since this value is compile time this is fine.  Otherwise we would
    // need to use sprintf to generate our fscanf format, which I would like
    // to avoid
    if (LM_CAMERA_MODEL_MAX_LEN != 32)
    {
        fi_log(LOG_CRIT, "(%s) camera model string length mismatch %d "
               "can not safely scan file", __FUNCTION__, LM_CAMERA_MODEL_MAX_LEN);

        return 0;
    }

    // Allocate our buffer.
    buf = std::unique_ptr<char[]>(new char[MAX_LINE_LEN + 1]());
    if (buf.get() == nullptr)
    {
        fi_log(LOG_CRIT, "(%s) buffer allocation failed", __FUNCTION__);
        return 0;
    }

    // Read and parse lines of the file.
    options->is_valid = 1;
    options->major_rev = LM_STRUCTURE_MAJOR_REV;
    options->minor_rev = 0;
    num_fields = 0;

    do
    {
        // Read a line from the file.
        p = fgets(buf.get(), MAX_LINE_LEN, lic_file);
        if (p != nullptr)
        {
            int len;
            char *endp;
            char *save_ptr;

            // Make sure we're null terminated.
            buf[ MAX_LINE_LEN ] = 0;

            // Warn if it is close, but try to use it anyway.
            len = strlen(buf.get());
            if (len >= MAX_LINE_LEN - 3)
            {
                fi_log(LOG_ERR, "(%s) WARNING license file line too long (%s)", __FUNCTION__, buf.get());
            }

            // Clear newlines, carriage returns, and spaces at the end.
            while ((len > 0) &&
                   ((*(buf.get() + len - 1) == '\n') ||
                    (*(buf.get() + len - 1) == '\r') ||
                    (*(buf.get() + len - 1) == '\t') ||
                    (*(buf.get() + len - 1) == ' ')))
            {
                *(buf.get() + len - 1) = 0;
                --len;
            }

            // Skip over initial whitespace.
            p = buf.get();
            while ((*p == ' ') || (*p == '\t'))
                ++p;

            // If there's any text left after the cleanup, process it.
            if (*p != 0)
            {
                unsigned long tempval;
                int badline;
                char *str1;
                char *str2;

                badline = 0;

                str1 = strtok_r(p, "= \t", &save_ptr);
                str2 = nullptr;
                if (str1 != nullptr)
                    str2 = strtok_r(nullptr, "= \t", &save_ptr);

                if ((str1 == nullptr) || (str2 == nullptr))
                    badline = 1;
                else if (strcmp("max_width", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->max_width = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("max_height", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->max_height = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("max_fps", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->max_fps = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("max_memory", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->max_memory = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("max_ssd", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->max_ssd = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("segmented_memory", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->segmented_memory = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("video_trigger", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->video_trigger = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("auto_exposure", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->auto_exposure = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("irig", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->irig = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("gps", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->gps = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                else if (strcmp("wireless", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->wireless = tempval;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }

                // New with 1.1
                else if (strcmp("model", str1) == 0)
                {
                    options->model.assign(str2);
                    if (options->model.size() > LM_CAMERA_MODEL_MAX_LEN)
                        options->model.resize(LM_CAMERA_MODEL_MAX_LEN);

                    ++num_fields;
                    if (options->minor_rev < 1)
                        options->minor_rev = 1;
                }
                // New with 1.2
                else if (strcmp("ltr", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->ltr_mode = tempval;
                        if (options->minor_rev < 2)
                            options->minor_rev = 2;
                        ++num_fields;
                    }
                    else
                        badline = 1;
                }
                // New with 1.3
                else if (strcmp("ssd_bandwidth", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        options->ssd_bandwidth = tempval;
                        ++num_fields;
                        if (options->minor_rev < 3)
                            options->minor_rev = 3;
                    }
                    else
                        badline = 1;
                }
                // New with 1.4. This is an optional setting
                // which may or may not be present.
                else if (strcmp("max_ltr_fps", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        fi_log(LOG_INFO, "Found max_ltr_fps=%lu", tempval);
                        options->max_ltr_fps = tempval;
                        ++num_fields;
                        if (options->minor_rev < 4)
                            options->minor_rev = 4;
                    }
                    else
                        badline = 1;
                }
                // New with 1.5
                else if (strcmp("dvr_playback", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        fi_log(LOG_INFO, "Found dvr_playback=%lu", tempval);
                        options->dvr_playback = tempval;
                        ++num_fields;
                        if (options->minor_rev < 5)
                            options->minor_rev = 5;
                    }
                    else
                        badline = 1;
                }
                // New with 1.6
                else if (strcmp("auto_trigger", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        fi_log(LOG_INFO, "Found auto_trigger=%lu", tempval);
                        options->auto_trigger = tempval;
                        ++num_fields;
                        if (options->minor_rev < 6)
                            options->minor_rev = 6;
                    }
                    else
                        badline = 1;
                }
                // New with 1.7
                else if (strcmp("serial_no", str1) == 0)
                {
                    options->serial_no.assign(str2);
                    ++num_fields;
                    if (options->minor_rev < 7)
                        options->minor_rev = 7;
                }

                // New with 1.8
                else if (strcmp("integrated", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        fi_log(LOG_INFO, "Found integrated=%lu", tempval);
                        options->integrated = tempval;
                        ++num_fields;
                        if (options->minor_rev < 8)
                            options->minor_rev = 8;
                    }
                }
                else if (strcmp("optical", str1) == 0)
                {
                    tempval = strtoul(str2, &endp, 10);
                    if ((endp != nullptr) && (*endp == '\0'))
                    {
                        fi_log(LOG_INFO, "Found optical=%lu", tempval);
                        options->optical = tempval;
                        ++num_fields;
                        if (options->minor_rev < 8)
                            options->minor_rev = 8;
                    }
                }

                // Otherwise, unrecognized.
                else
                    badline = 1;

                if (badline)
                {
                    fi_log(LOG_ERR, "(%s) bad license file line", __FUNCTION__);
                    options->is_valid = 0;
                }
            }
        }
    } while (p != nullptr);

    if (((num_fields == VER_1_0_FIELDS) && (options->minor_rev != 0)) ||
        ((num_fields == VER_1_1_FIELDS) && (options->minor_rev != 1)) ||
        ((num_fields == VER_1_2_FIELDS) && (options->minor_rev != 2)) ||
        ((num_fields == VER_1_3_FIELDS) && (options->minor_rev != 3)) ||
        ((num_fields == VER_1_4_FIELDS) && (options->minor_rev < 4)) ||
        ((num_fields == VER_1_5_FIELDS) && (options->minor_rev < 5)) ||
        ((num_fields == VER_1_6_FIELDS) && (options->minor_rev < 6)) ||
        ((num_fields == VER_1_7_FIELDS) && (options->minor_rev != 7)))
    {
        fi_log(LOG_ERR, "(%s) bad license file, has %u fields but minor=%u", __FUNCTION__,
               num_fields, options->minor_rev);
        options->is_valid = 0;
    }

    // 1.0 license file did not have a model number so we construct one.
    if (options->model.empty())
        getDefaultModel(options);

    // We didn't get max_ltr_fps until version 1.4. If our version is lower than that,
    // set the max_ltr_fps to a sane value.
    if (options->max_ltr_fps == 0)
    {
        options->max_ltr_fps = DEFAULT_MAX_FPS;
    }

    // We didn't get dvr_playback until version 1.5. If our version is lower than that,
    // set the dvr_playback to a sane value.
    if (options->minor_rev < 5)
    {
        options->dvr_playback = 0;
    }
    if (options->minor_rev < 6)
    {
        options->auto_trigger = 0;
    }

    // Did we get a serial number?
    if (options->minor_rev < 7)
    {
        options->serial_no.assign(BAD_LICENSE);
    }

    if (options->minor_rev < 8)
    {
        options->integrated = 0;
        options->optical = 1;
    }

    return options->is_valid;
}

/*!
 * @brief Validate and correct the license data.
 *
 * @note There's nothing we can do about any problems other than
 *   to log it and choose sensible alternatives to publish.
 *
 * @param options The options structure we're working on.
 */
void LicManager::licenseValidate()
{
    if ((options.ltr_mode != LM_LTR_NONE) && (options.ltr_mode != LM_LTR_ONLY) && (options.ltr_mode != LM_LTR_DUAL))
    {
        fi_log(LOG_ERR, "%s: Unknown ltr value %d, using NONE", __FUNCTION__, options.ltr_mode);
        options.ltr_mode = LM_LTR_NONE;
    }
    if (options.ltr_mode != LM_LTR_NONE)
    {
        // Do we have an SSD?
        if (storage::get_ssd_count() == 0)
        {
            fi_log(LOG_INFO, "%s: LTR enabled, but no SSD found!", __FUNCTION__);
            options.ltr_mode = LM_LTR_NONE;
        }
    }
}

void LicManager::getDefaultModel(LIC_OPTIONS *options)
{
    uint32_t sensor_type = 0;
    Sensor *sensor_ptr = CameraServer::get_sensor();

    if (sensor_ptr == NULL)
    {
        fi_log(LOG_ERR, "(%s) Unable to get the sensor type register", __FUNCTION__);
        fi_log(LOG_INFO, "(%s) Fall back to assuming a color sensor", __FUNCTION__);

        options->model.assign(TS3_LEGACY_COLOR_MODEL);
        return;
    }

    sensor_type = sensor_ptr->get_sensor_type();
    if (sensor_type == 0)
    {
        return;
    }
    switch (sensor_type)
    {
    case FI_SENSOR_LUPA1300_COLOR:
        options->model.assign(TS3_LEGACY_COLOR_MODEL);
        break;
    case FI_SENSOR_LUPA1300_MONO:
        options->model.assign(TS3_LEGACY_MONO_MODEL);
        break;
    default:
        fi_log(LOG_ERR, "(%s) Unknown sensor_type %08x (%4.4s)", __FUNCTION__,
               sensor_type, ((char *)(&sensor_type)));
        fi_log(LOG_INFO, "(%s) Fall back to assuming a color sensor", __FUNCTION__);
        options->model.assign(TS3_LEGACY_COLOR_MODEL);
        break;
    }
}

/*!
 * @brief Validate that the license file was signed by Fastec Production.
 */
bool LicManager::signatureCheck()
{
    bool rc = false;
    int verify_status;
    std::string verify_cmd;

    verify_cmd = fmt::format("{0}/{1} {2} {0}/{3}.sig", get_home_path(), SIG_CHECK_SCRIPT, FASTEC_FINGERPRINT, VALIDATED_LICENSE_FILE);
    verify_status = system(verify_cmd.c_str());
    if ((verify_status < 0) || !WIFEXITED(verify_status))
    {
        // There's a problem, log the error.
        fi_log(LOG_ERR, "%s: Cannot run the verify command [%d/%d]!", __PRETTY_FUNCTION__,
               errno, WEXITSTATUS(verify_status));
    }
    else
    {
        // An exit value of 0 is success.
        verify_status = WEXITSTATUS(verify_status);
        rc = (verify_status == 0);
        fi_log(LOG_DEBUG, "%s: WEXITSTATUS=%d (%s)", __FUNCTION__, verify_status, strerror(verify_status));
    }

    // Return our result.
    return rc;
}
