/*!
 * @file lince5m.cxx
 * @brief This source file contains the hardware abstraction layer which
 *    provides all the method implementations required for support
 *    of the Lince5M sensor and its sensor-specific FPGA interfaces.
 *
 * Copyright 2010,2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <chrono>
#include <memory>
#include <string>

#include <fmt/printf.h>
#include <json/json.h>

#include <fastec/gige_msg.h>
#include <fastec/lince5m_param.h>
#include <fastec/lince5m_utils.h>
#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "camera_server.h"
#include "cs_network.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_correction.h"
#include "lince5m.h"
#include "tmp451.h"

namespace LINCE5M_SENSOR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// The temperature sensor on the Lince5M sensor board is the TMP451.
#define TMP451_I2C_ADDR         0x4C

//! This is what we expect to read back from the sensor.
const uint8_t EXPECTED_SENSOR_ID[4] = { 0x18, 0x20, 0x55, 0xf9 };

//! Lince sensor parameters
const uint32_t DEF_FRAME_RATE = 250;

//! Lince Command Codes
const uint8_t CMD_UPDATE_REG     = 0x99;
const uint8_t CMD_UPDATE_ROI     = 0xaa;
const uint8_t CMD_START_SENS     = 0xdd;
const uint8_t CMD_STOP_HARD_SENS = 0xee;
const uint8_t CMD_READ_STATUS    = 0x21;
const uint8_t CMD_WRITE_REG      = 0x52;
const uint8_t CMD_WRITE_MEM      = 0x46;
const uint8_t CMD_READ_REG       = 0x12;
const uint8_t CMD_READ_MEM       = 0x06;
const uint8_t CMD_READ_ID        = 0x20;

#define ALL_CHANNELS 0x01ffffff //!< One bit for each channel 0-23 + sync
#define SYNC_IDLE    0x4D   //!< The IDLE pattern on the sync channel.
#define NUM_CHANS    25     //!< Number of LVDS channels.

#define PDEBUG_ALIGN PDEBUG_SENSOR  //!< So we can turn on alignment details at will.

//! Sensor Configuration Register Rg0x000
#define SC_TRIGGER_ENA      ( 1 << 1 )
#define SC_TRIG_POL_HIGH    ( 0 << 2 )
#define SC_TRIG_POL_LOW     ( 1 << 2 )
#define SC_TRIG_POL_MASK    ( 1 << 2 )
#define SC_TRIG_NONCONCUR   ( 1 << 3 )
#define SC_VFPN_CAL_ENA     ( 1 << 6 )
#define SC_MODE_MONO        ( 0 << 7 )
#define SC_MODE_COLOR       ( 1 << 7 )
#define SC_MODE_MASK        ( 1 << 7 )

//! Lince Device Status Bits (READ_STATUS or Rg0x001)
#define DS_PIXEL_BUSY       ( 1 << 0 )
#define DS_UPDATE_BUSY      ( 1 << 1 )
#define DS_ROI_UPDATE_BUSY  ( 1 << 2 )
#define DS_ROI_READOUT_BUSY ( 1 << 3 )  // aka rd_enable
#define DS_RAMP_LOCKED      ( 1 << 5 )
#define DS_SYS_LOCKED       ( 1 << 6 )
#define DS_LVDS_LOCKED      ( 1 << 7 )

//! Common registers
const uint16_t LREG_ROI_NUM_ROWS     = 0x010; // Rg0x010 Number of rows minus 1
const uint16_t LREG_ROI_FIRST_ROW    = 0x011; // Rg0x011 Zero-based first row
const uint16_t LREG_ROI_NUM_COLS     = 0x012; // Rg0x012 Number of columns minus 1
const uint16_t LREG_ROI_FIRST_COL    = 0x013; // Rg0x013 Zero-based first column
const uint16_t LREG_SUBS_BIN         = 0x014; // Rg0x014 Subsampling and binning
const uint16_t LREG_INTEGRATION_HIGH = 0x002; // Rg0x002 High word of integration time (for mini-calibration)
const uint16_t LREG_INTEGRATION_LOW  = 0x003; // Rg0x003 Low word of integration time
const uint16_t LREG_ANALOG_GAIN      = 0x05b; // Rg0x015 Analog gain, set to regfile value divided by gain.

//! Lince subsampling and binning options for Rg0x014. We set row and column the same.
#define SB_BINSUBS_MASK ( 0xff )
#define SB_BIN_X1   (( 0 << 2 ) | ( 0 << 6 ))   // 0x00
#define SB_BIN_X2   (( 1 << 2 ) | ( 1 << 6 ))   // 0x44
#define SB_BIN_X4   (( 2 << 2 ) | ( 2 << 6 ))   // 0x88
#define SB_SUBS_X1  (( 0 << 0 ) | ( 0 << 4 ))   // 0x00
#define SB_SUBS_X2  (( 1 << 0 ) | ( 1 << 4 ))   // 0x11
#define SB_SUBS_X4  (( 2 << 0 ) | ( 2 << 4 ))   // 0x22

#define IDLE_PATTERN 0x032a //!< The IDLE pattern emitted on data lines.
#define IDLE_ENABLE  0x1000 //!< Enable the IDLE pattern.

#define FIRST_IN_ROW (1 << 12)
#define LAST_IN_ROW  (1 << 13)

//! Defect map filename derived from sensor serial number.
#define DEFECT_MAP_FILE_DIR  ".config/Fastec/calibration"
#define DEFECT_MAP_FILE_TEST ".config/Fastec/scratch/bad_pixels.txt"

// Framing modes
#define FM_NONE 0   // Not yet initialized.
#define FM_1    1   // Framing Mode 1 is 86.4 Mhz.
#define FM_6    6   // Framing Mode 6 is 72 MHz.

// Directory concatenates with filename strings to form fully-qualified path.
#define CFG_DIR ".config/Fastec/lince5m/"

// These files are for Framing Mode 1, which uses 12-bit data and maximum frame rates,
// and are for Linear (non-HDR) acquisition.
// For trow=52, we only have the version for binnning because Anafocus did not find
// SNR performance acceptable without binning.
// For trow=84, the binning vs non-binning files are the same except for Rg0x07c so we
// don't actually load the file; we change the register after the "normal" version is loaded.
#define FM1_TROW_52B_FILE ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow52_HSHS_bin2x2_rev910.dat" )
#define FM1_TROW_84_FILE  ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow84_HSHS_rev910.dat" )
//#define FM1_TROW_84B_FYI  ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow84_HSHS_bin2x2_rev910.dat" )
#define FM1_TROW_124_FILE ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow124_HSHS_rev707.dat" )
#define FM1_TROW_132_FILE ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow132_HSHS_rev707.dat" )
#define FM1_TROW_165_FILE ( CFG_DIR "Lince5M_Regs_FM1_86.4MHz_trow165_HSHS_rev707.dat" )

// Mode Conversion Block (MCB) Finite State Machine (FSM) is defined by the MCB file,
// which chooses the trigger operation mode.
#define MCB_MEMADDR 0x600
#define MCB_WORDS   128
#define MCB_DEFAULT_FILE ( CFG_DIR "Lince5M_MCB_Mode5_SingleEdgeTrigger_Ignore_NoConc_rev536.dat" )
#define MCB_REAL_FILE    ( CFG_DIR "Lince5M_MCB_Mode1_PWC_Ignore_Conc_rev536.dat" )

// Global States FSM (GBLST) - so far, just support Linear and not HDR
#define GBLST_MEMADDR 0x680
#define GBLST_WORDS   128
#define GBLST_LINEAR_FILE  ( CFG_DIR "Lince5M_GBLST_Linear_rev536.dat" )

// Readout States FSM (RDOUTST)
#define RDOUTST_MEMADDR 0x800
#define RDOUTST_WORDS   512
#define RDOUTST_DEFAULT_FILE  ( CFG_DIR "Lince5M_RDOUTST_Mode0_rev536.dat" )

#if 0
// For setting skew values for the LVDS data lines.
// Each register has values for two data lines, one in lower 8 bits, one in higher.
// Each 8-bit value is a 4-bit delay from rising edge of internal LVDS clock to data change
// and a 4-bit delay from the falling edge to data change, in the range 0-12 (0xc).
#define REG_LVDS_SKEWS 0x027        // Rg0x027: start of skew registers
#define REG_LVDS_COARSE_SKEWS 0x036 // Rg0x036: coarse skew values
#define NUM_LVDS_SKEWS 26       // 24 data, sync, clock
static uint16_t skews[ NUM_LVDS_SKEWS / 2 ] =
{
    0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8),
    0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8),
    0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8), 0x00 | (0x00 << 8),
    0x00 | (0x00 << 8)
};
#endif

//! Structure to mimic defective pixel memory layout
typedef struct SENSOR_BAD_PIXEL
{
    uint16_t row;
    uint16_t gcol;
    uint16_t col_bit;
} SENSOR_BAD_PIXEL;

} /* End of namespace LINCE5M_SENSOR */

using namespace LINCE5M_SENSOR;

//--------------------------------------------------------------------
// Class Implementation
//--------------------------------------------------------------------

void Lince5M::dump_range(void *p, char tag, uint8_t cmd,
                         unsigned low_addr, unsigned high_addr,
                         uint16_t *buf, unsigned buf_words)
{
    FILE *fp = (FILE *)p;

    // Dump registers.
    for (unsigned cur_addr = low_addr; cur_addr <= high_addr; cur_addr += buf_words)
    {
        unsigned this_count;

        this_count = buf_words;
        if ((high_addr - low_addr + 1) < this_count)
            this_count = high_addr - cur_addr + 1;

        if (spiReadData(cmd, cur_addr, buf, this_count))
        {
            for (unsigned i = 0; i < this_count; ++i)
                fprintf(fp, "%c[%03x] = 0x%04x = %u\n", tag, cur_addr + i, buf[i], buf[i]);
        }
    }
}

void Lince5M::dump_sensor_regs(void *p)
{
    std::unique_ptr<uint16_t[]> buf;
    const int BUF_WORDS = 128;

    buf = std::unique_ptr<uint16_t[]>(new uint16_t[BUF_WORDS]());
    if (buf.get() == nullptr)
    {
        FILE *fp = (FILE *)p;

        fprintf(fp, "ENOMEM\n");
        return;
    }

    dump_range(p, 'R', CMD_READ_REG,     0, 0x1ff, buf.get(), BUF_WORDS);
    dump_range(p, 'M', CMD_READ_MEM, 0x600, 0x67f, buf.get(), BUF_WORDS);
    dump_range(p, 'M', CMD_READ_MEM, 0x680, 0x6ff, buf.get(), BUF_WORDS);
}

/*!
 * @brief Initiate the Dynamic Phase Alignment of the LVDS data inputs.
 *
 * There are 24 LVDS data channels and a sync channel. Although they all transmitted
 * using the same 345MHz clock, there can be skew between the channels, jitter, or noise,
 * which the receiver needs to account for somehow. The Altera LVDS megafunction does
 * Dynamic Phase Alignment (DPA) to optimize the data sampling window; it does most of
 * the heavy lifting and we just guide it through the process.
 *
 * @return true if everything went well, otherwise false
 */
bool Lince5M::alignLVDS(uint16_t idle_pattern)
{
    bool rc = false;
    int i, j, error_count;
    int usec_sleep, retries, max_retries;
    uint32_t control;
    uint32_t status;
    uint32_t value;
    union
    {
        uint32_t w[(NUM_CHANS + 3) / 4];
        uint8_t b[NUM_CHANS];
    } chan;
    uint8_t slip_count[NUM_CHANS];
    uint32_t slips, last_slipped;
    uint32_t zeroes, sevens;
    uint8_t expected = 0;
    uint8_t minval, maxval;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Turn on my debug logging and send it to the log file.
    int saved_cs_verbose = cs_verbose;
    bool saved_debug_to_log = debug_to_log;
    //cs_verbose |= PDEBUG_ALIGN;
    //debug_to_log = true;

    // Reset the LVDS_RX module and the LVDS_RX_PLL (pll_reset).
    if (!fpga_ptr->regRead(ICM_LINCE_CONTROL, &control))
        goto bail;
    control |= (ILC_LVDS_RX_RESET | ILC_LVDS_RX_PLL_ARESET);
    control &= ~ILC_LVDS_RX_PAR_OUT_VALID;
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, control))
        goto bail;

    // Put the DPA channels into reset.
    if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_DPA_RESET, ALL_CHANNELS))
        goto bail;

    // Do not be alarmed that Altera says that rx_reset should not deassert until we see rx_locked.
    // We can deassert these two together here because our RX_RESET is not the one
    // Altera is talking about.
    control &= ~(ILC_LVDS_RX_RESET | ILC_LVDS_RX_PLL_ARESET);
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, control))
        goto bail;

    // Look for rx_locked to be set. This says that the LVDS_RX PLL is locked
    // to the LVDS clock coming from the sensor.
    max_retries = 20;
    usec_sleep = 5;
    error_count = 0;
    for (retries = 0; retries < max_retries; ++retries)
    {
        int stable;

        // Delay
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(usec_sleep));

        if (!fpga_ptr->regRead(ICM_LINCE_STATUS, &status))
            goto bail;
        if ((status & ILS_LVDS_RX_PLL_LOCKED) != 0)
        {
            // FIXME Does RX_PLL_LOCKED latch? If so, we don't need to read it again.
            for (stable = 0; stable < 3; ++stable)
            {
                if (!fpga_ptr->regRead(ICM_LINCE_STATUS, &status))
                    goto bail;
                if ((status & ILS_LVDS_RX_PLL_LOCKED) == 0)
                {
                    ++error_count;
                    break;
                }
            }
            if (stable == 3)
                break;
        }
    }
    if (retries >= max_retries)
    {
        fi_log(LOG_ERR, "%s: RX_PLL_LOCKED timeout, status=0x%08x after %u usecs, unlock=%d",
               __FUNCTION__, status, usec_sleep * retries, error_count);
        goto bail;
    }
    LOG("%s: Got RX_PLL_LOCKED in %u usecs, unlock=%d, okay (ICM_LINCE_STATUS=0x%08x).", __FUNCTION__, error_count, usec_sleep * retries, status);

    // Take the DPA channels out of reset.
    if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_DPA_RESET, 0))
        goto bail;

    // Now we need to see the DPA lock on all channels (rx_dpa_locked).
    // This should take no more than 3 * 256 serial data transitions.
    // I think that means actual transitions in the data pattern so the
    // number could be different for different channels, but the worst
    // case would be one transition per byte or 6144 clocks at 345 MHz
    // so we're still talking microseconds.
    max_retries = 20;
    usec_sleep = 6;
    for (retries = 0; retries < max_retries; ++retries)
    {
        // Delay
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(usec_sleep));

        if (!fpga_ptr->regRead(ICM_LINCE_LVDS_RX_DPA_LOCKED, &status))
            goto bail;
        if ((status & ALL_CHANNELS) == ALL_CHANNELS)
            break;
    }
    if (retries >= max_retries)
    {
        fi_log(LOG_ERR, "%s: RX_DPA_LOCKED=0x%08x, timeout after %u usecs!", __FUNCTION__, status, usec_sleep * retries);
        goto bail;
    }
    LOG("%s: Got RX_DPA_LOCKED in %u usecs, okay.", __FUNCTION__, usec_sleep * (retries + 1));

    // Assert rx_fifo_reset for one clock. The FPGA toggles so we just need to write ones.
    // This does not hold the data FIFOs in reset. We would not want that because
    // rx_dpa_lock will never assert if there are no transitions in the parallel data.
    if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_FIFO_RESET, ALL_CHANNELS))
        goto bail;

    // Enable parallel logic so we can read what the alignment logic is presenting as bytes.
    control |= ILC_LVDS_RX_PAR_OUT_VALID;
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, control))
        goto bail;

    // Assert rx_cda_reset for one clock. The FPGA toggles so we just need to write ones.
    // This resets the module's Channel Data Align (CDA) logic.
    if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_BITSLIP_RESET, ALL_CHANNELS))
        goto bail;

    // Now do the word alignment. Three one-byte patterns repeat every 3 channels:
    // IDLE[0:7], IDLE[0:3]IDLE[8:11], IDLE[4:11]
    memset(slip_count, 0, NUM_CHANS * sizeof(slip_count[0]));
    last_slipped = ALL_CHANNELS;
    max_retries = 12;
    for (retries = 0; retries < max_retries; ++retries)
    {
        // We want to see stable data. Read the channel data three times and call it
        // an error if it ever changes. Once we can read all the channels consistently,
        // call it stable.
        for (j = 0; j < 10; ++j)
        {
            error_count = 0;
            for (i = 0; i < ((NUM_CHANS + 3) / 4); ++i)
            {
                if (!fpga_ptr->regRead(ICM_LINCE_LVDS_RX_CH00_03_DATA + 4 * i, &value))
                    goto bail;
                if (!fpga_ptr->regRead(ICM_LINCE_LVDS_RX_CH00_03_DATA + 4 * i, &chan.w[i]))
                    goto bail;
                if (value != chan.w[i])
                    ++error_count;
                if (!fpga_ptr->regRead(ICM_LINCE_LVDS_RX_CH00_03_DATA + 4 * i, &chan.w[i]))
                    goto bail;
                if (value != chan.w[i])
                    ++error_count;
            }
            if (error_count == 0)
                break;
        }
        // LOG("Data(%d): %08x %08x %08x %08x %08x %08x %02x", j, chan.w[0], chan.w[1], chan.w[2], chan.w[3], chan.w[4], chan.w[5], chan.w[6]);

        slips = 0;
        for (i = 0; i < NUM_CHANS; ++i)
        {
            switch (i % 3)
            {
            case 0:
                // IDLE[0:7]
                expected = idle_pattern & 0x0ff;
                break;
            case 1:
                // IDLE[0:3]IDLE[8:11]
                expected = ((idle_pattern << 4) & 0x0f0) | ((idle_pattern >> 8) & 0x00f);
                break;
            case 2:
                // IDLE[4:11]
                expected = (idle_pattern >> 4) & 0x0ff;
                break;
            }
            // Channel 25 is the sync channel, which has its own special pattern.
            if (i == (NUM_CHANS - 1))
                expected = SYNC_IDLE;

            if (chan.b[ i ] != expected)
            {
                slips |= (1 << i);
                ++slip_count[ i ];
                // LOG("chan%02d=0x%02x, expected=0x%02x, slips=0x%08x", i, chan.b[i], expected, slips);
            }
        }

        // If a channel did not slip last time (~last=1) and it slips this time (slip=1),
        // that indicates a lack of stability.
        if (((~last_slipped) & slips) != 0)
            fi_log(LOG_ERR, "%s: Unstable(#%d): last=%08x cur=%08x", __FUNCTION__, retries, last_slipped, slips);
        else
            LOG("%s: (#%d) slip=%08x", __FUNCTION__, retries, slips);
        last_slipped = slips;

        // If no more slips, we are done.
        if (slips == 0)
            break;

        if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_BITSLIP, slips))
            goto bail;

        // Delay
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(6));
    }

    // See if we got the special case where there are some channels that took 7 slips and
    // some that took zero. We need to have our channels no more than 1 bit off from each
    // other and this is 1 modulo 8 which would seem okay, but this particular case can cause
    // our parallel data words to be off by one from one another.
    zeroes = sevens = 0;
    for (i = 0; i < NUM_CHANS; ++ i)
    {
        if (slip_count[ i ] == 0)
            zeroes |= (1 << i);
        if (slip_count[ i ] == 7)
            sevens |= (1 << i);
    }
    if ((zeroes != 0) && (sevens != 0))
    {
        // If all the channels are either 0 or 7, slip all the zeros 8 times.
        if ((zeroes | sevens) == ALL_CHANNELS)
        {
            for (i = 0; i < 8; ++i)
            {
                if (!fpga_ptr->regWrite(ICM_LINCE_LVDS_RX_BITSLIP, zeroes))
                    goto bail;

                // Delay
                using us = std::chrono::microseconds;
                std::this_thread::sleep_for(us(6));
            }
            for (i = 0; i < NUM_CHANS; ++ i)
            {
                if ((zeroes & (1 << i)) != 0)
                    slip_count[i] += 8;
            }
        }
        // Otherwise, it is sevens and zeroes and something else, which is bad.
        // But we will catch it when we check the min and max slip counts.
        else
            fi_log(LOG_DEBUG, "%s: zeroes=0x%08x sevens=0x%08x", __FUNCTION__, zeroes, sevens);
    }

    // See the range of how many slips it took. These numbers should not vary by more than 1.
    minval = maxval = slip_count[0];
    for (i = 1; i < NUM_CHANS; ++ i)
    {
        if (slip_count[ i ] < minval)
            minval = slip_count[ i ];
        if (slip_count[ i ] > maxval)
            maxval = slip_count[ i ];
    }
    if (maxval > minval + 1)
    {
        fi_log(LOG_ERR, "Unusable slip combination: min=%u max=%u sync=%d", minval, maxval,
               slip_count[ NUM_CHANS - 1 ]);
        for (i = 0; i < (NUM_CHANS - 1); i += 6)
            fi_log(LOG_ERR, "Slip count(%2d:%-2d): %d %d %d %d %d %d", i, i + 5,
                   slip_count[i + 0], slip_count[i + 1], slip_count[i + 2],
                   slip_count[i + 3], slip_count[i + 4], slip_count[i + 5]);
        //  goto bail;
    }

    // Indicate whether we succeeded.
    if (retries < max_retries)
        rc = true;

    LOG("Slip count: sync=%u min=%u max=%u", slip_count[NUM_CHANS - 1], minval, maxval);
    for (i = 0; i < (NUM_CHANS - 1); i += 6)
    {
        LOG("Slip count(%2u:%-2u): %u %u %u %u %u %u", i, i + 5,
            slip_count[i + 0], slip_count[i + 1], slip_count[i + 2],
            slip_count[i + 3], slip_count[i + 4], slip_count[i + 5]);
    }

bail:
    cs_verbose = saved_cs_verbose;
    debug_to_log = saved_debug_to_log;

    // Return our result.
    return rc;
}

/*!
 * @brief Write a command and optionally read back data
 *
 * @param cmd   the sensor command code
 * @param data  NULL or pointer to buffer to hold returned data
 * @param count number of bytes of data to read
 */
bool Lince5M::spiReadCmd(uint8_t cmd, uint8_t *data, unsigned count)
{
    bool rc = false;
    unsigned i, j;
    uint32_t cur_count;
    uint32_t reg_value;
    int retries, max_retries;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return false;

    // Make sure the SPI bus is available.
    if (!spiWaitIdle())
    {
        fi_log(LOG_ERR, "%s: SPI never becomes idle.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Sanity check, make sure RD_FIFO is empty.
    if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &reg_value))
        goto bail;
    if (reg_value != 0)
    {
        fi_log(LOG_ERR, "%s: RD DATA FIFO not empty %u, resetting\n", __FUNCTION__, reg_value);
        fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_RD_FIFO_RESET);

        if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &reg_value))
            goto bail;

        if (reg_value != 0)
        {
            fi_log(LOG_ERR, "%s: RD DATA FIFO still not empty %u", __FUNCTION__, reg_value);
            goto bail;
        }
    }

    // Commands have only a small amount of data, if any, so no danger of filling FIFO.

    // Set Read direction, no address or dummy.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_RD_DIRECTION))
        goto bail;

    // Set the number of bytes of data we will read after sending the command.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_DATA_BYTE_COUNT, count))
        goto bail;

    // This starts the operation.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_COMMAND, cmd))
        goto bail;

    j = 0;
    cur_count = 0;
    i = 0;
    retries = 0;
    max_retries = 10;
    while ((i < count) && (retries < max_retries))
    {
        if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &cur_count))
            goto bail;

        for (j = 0; (j < cur_count) && (i < count); ++j)
        {
            // Read from the FIFO.
            if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO, &reg_value))
                goto bail;

            // Extract up to four bytes.
            while (i < count)
            {
                unsigned shift = (i % 4) * 8;
                data[i++] = (uint8_t)(reg_value >> shift);

                // Stop when all bytes of word are shifted out.
                if (shift == 24)
                    break;
            }
        }

        // Sleep if we didn't get any data.
        if ((cur_count == 0) && (i < count))
        {
            // Wait 2 milliseconds.
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(2));
        }

        ++retries;
    }
    if (retries == max_retries)
    {
        fi_log(LOG_WARNING, "%s: Retries exceeded!", __FUNCTION__);
        goto bail;
    }

    // This should not happen.
    if (j < cur_count)
    {
        fi_log(LOG_WARNING, "%s: Too much data: FIFO level=%u j=%u i=%u count=%u", __FUNCTION__, cur_count, j, i, count);
    }

    rc = true;

bail:
    //fi_log( LOG_DEBUG, "%s: cmd=0x%x count=%u data=0x%x", __FUNCTION__, cmd, count, ( data != NULL ) ? data[0] : 0 );

    // Return our result.
    return rc;
}

/*!
 * @brief Uses a READ_REG or READ_MEM command to read data from the sensor.
 *
 * @param cmd       The SPI command, READ_MEM or READ_REG
 * @param addr      register or memory address
 * @param data      buffer to fill with read data
 * @param word_count    number of 16-bit words of data to read
 *
 * The data is in native (little-endian) byte order but must be
 * byte-swapped to send it to the sensor in big-endian order.
 *
 * @return true if everything went well, otherwise false.
 */
bool Lince5M::spiReadData(uint16_t cmd, uint16_t addr, uint16_t *data, unsigned word_count)
{
    bool rc = false;
    uint32_t reg_value;
    uint16_t cur_addr;
    unsigned cur_count;
    uint32_t avail_count;
    unsigned i, j;
    unsigned this_op;
    int retries, max_retries = 10;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return false;

    // Make sure the SPI bus is available.
    if (!spiWaitIdle())
    {
        fi_log(LOG_ERR, "%s: SPI never becomes idle.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Sanity check, make sure RD_FIFO is empty.
    if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &reg_value))
        goto bail;
    if (reg_value != 0)
    {
        fi_log(LOG_ERR, "%s: RD DATA FIFO not empty %u, resetting\n", __FUNCTION__, reg_value);
        fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_RD_FIFO_RESET);

        if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &reg_value))
            goto bail;

        if (reg_value != 0)
        {
            fi_log(LOG_ERR, "%s: RD DATA FIFO still not empty %u", __FUNCTION__, reg_value);
            goto bail;
        }
    }

    // We must break up the operation into chunks that fit into the FIFO because
    // once the SPI operation starts, we can't guarantee we can keep the FIFO from
    // overflowing.
    cur_addr = addr;
    cur_count = word_count;
    while (cur_count > 0)
    {
        // Set the read direction and address/dummy options.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL,
                                ILSC_RD_DIRECTION | ILSC_ADDRESS_ENA | ILSC_DUMMY_ENA))
            goto bail;

        // Set the memory or register address.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_REG_ADDRESS, cur_addr))
            goto bail;

        // Set the number of bytes of data to read after sending the command.
        // Remember that we are reading words so it is 2 bytes per address.
        this_op = cur_count;
        if (this_op > (LINCE5M_SENSOR::ICM_LINCE_RD_FIFO_SIZE_WORDS * 2))
            this_op = LINCE5M_SENSOR::ICM_LINCE_RD_FIFO_SIZE_WORDS * 2;
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_DATA_BYTE_COUNT, this_op * 2))
            goto bail;

        // This starts the operation.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_COMMAND, cmd))
            goto bail;

        i = 0;
        retries = 0;
        while ((i < this_op) && (retries < max_retries))
        {
            if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO_LEVEL, &avail_count))
                goto bail;

            j = 0;
            while (j < avail_count)
            {
                // Read from the FIFO.
                if (!fpga_ptr->regRead(ICM_LINCE_SPI_RD_DATA_FIFO, &reg_value))
                    goto bail;

                // Count a word read from FIFO.
                ++j;

                // Swap the 16-bit word's two bytes since the SPI data is high-byte first.
                data[i++] = (uint16_t)(((reg_value << 8) & 0x0000ff00) |
                                       ((reg_value >> 8) & 0x000000ff));
                if (i < this_op)
                    data[i++] = (uint16_t)(((reg_value >> 24) & 0x000000ff) |
                                           ((reg_value >> 16) & 0x0000ff00));

                if (i >= this_op)
                    break;
            }

            // If we didn't read anything from the FIFO and we still need data, wait a bit.
            if ((avail_count == 0) && (i < this_op))
            {
                // Wait 2 milliseconds.
                using ms = std::chrono::milliseconds;
                std::this_thread::sleep_for(ms(2));

                ++retries;
            }

            // This should not happen, just a debug sanity check that can be removed sometime.
            if (j < avail_count)
            {
                fi_log(LOG_WARNING, "%s: Too much data: FIFO level=%u j=%u i=%u this=%u", __FUNCTION__, avail_count, j, i, this_op);
            }
        }

        cur_addr += this_op;
        cur_count -= this_op;
    }
    rc = true;

bail:
    //fi_log( LOG_DEBUG, "%s: cmd=0x%x addr=0x%x count=%u data[0]=0x%x", __FUNCTION__, cmd, addr, word_count, data[0] );

    // Return our result.
    return rc;
}

bool Lince5M::spiReadReg(uint16_t reg_addr, uint16_t *rtn_data)
{
    return spiReadData(CMD_READ_REG, reg_addr, rtn_data, 1);
}

bool Lince5M::spiReadStatus(uint8_t *rtn_status)
{
    return spiReadCmd(CMD_READ_STATUS, rtn_status, 1);
}

/*!
 * @brief Write a command optionally followed by data
 *
 * @param cmd   the sensor command code
 * @param data  NULL or pointer to buffer of data to follow command
 * @param count number of bytes of data to follow command
 */
bool Lince5M::spiWriteCmd(uint8_t cmd, uint8_t *data, unsigned count)
{
    bool rc = false;
    unsigned i;
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    //fi_log( LOG_DEBUG, "%s: cmd=0x%x count=%u data=0x%x", __FUNCTION__, cmd, count, ( data != NULL ) ? data[0] : 0 );

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return false;

    // Make sure the SPI bus is available.
    if (!spiWaitIdle())
    {
        fi_log(LOG_ERR, "%s: SPI never becomes idle.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Sanity check, make sure WR_FIFO is empty.
    if (!fpga_ptr->regRead(ICM_LINCE_SPI_WR_DATA_FIFO_LEVEL, &reg_value))
        goto bail;
    if (reg_value != 0)
    {
        fi_log(LOG_ERR, "%s: WR DATA FIFO not empty %u, resetting\n", __FUNCTION__, reg_value);
        fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_WR_FIFO_RESET);

        if (!fpga_ptr->regRead(ICM_LINCE_SPI_WR_DATA_FIFO_LEVEL, &reg_value))
            goto bail;

        if (reg_value != 0)
        {
            fi_log(LOG_ERR, "%s: WR DATA FIFO still not empty %u", __FUNCTION__, reg_value);
            goto bail;
        }
    }

    // Commands have only a small amount of data, if any, so no danger of overflowing FIFO.

    // Set the write direction and address option.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_WR_DIRECTION))
        goto bail;

    // Set the number of bytes of data that will follow the command.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_DATA_BYTE_COUNT, count))
        goto bail;

    for (i = 0; i < count; ++i)
    {
        unsigned shift = (i % 4) * 8;

        if (shift == 0)
            reg_value = 0;

        reg_value |= (((uint32_t)data[i] << shift) & (0x000000ff  << shift));

        if ((shift == 24) || (i == count - 1))
        {
            if (!fpga_ptr->regWrite(ICM_LINCE_SPI_WR_DATA_FIFO, reg_value))
                goto bail;
        }
    }

    // This starts the operation.
    if (!fpga_ptr->regWrite(ICM_LINCE_SPI_COMMAND, cmd))
        goto bail;

    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Writes a WRITE_REG or WRITE_MEM command to the sensor with data
 *
 * @param cmd       The SPI command, WRITE_MEM or WRITE_REG
 * @param addr      register or memory address
 * @param data      buffer of data to write
 * @param word_count    number of 16-bit words of data to send
 *
 * The data is in native (little-endian) byte order but must be
 * byte-swapped to send it to the sensor in big-endian order.
 *
 * @return true if everything went well, otherwise false.
 */
bool Lince5M::spiWriteData(uint16_t cmd, uint16_t addr, uint16_t *data, unsigned word_count)
{
    bool rc = false;
    uint16_t cur_addr;
    unsigned cur_count;
    unsigned i;
    unsigned this_op;
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    //fi_log( LOG_DEBUG, "%s: cmd=0x%02x addr=0x%03x count=%u data[0]=0x%04x", __FUNCTION__, cmd, addr, word_count, data[0] );

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return false;

    // Make sure the SPI bus is available.
    if (!spiWaitIdle())
    {
        fi_log(LOG_ERR, "%s: SPI never becomes idle.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Sanity check, make sure WR_FIFO is empty.
    if (!fpga_ptr->regRead(ICM_LINCE_SPI_WR_DATA_FIFO_LEVEL, &reg_value))
        goto bail;
    if (reg_value != 0)
    {
        fi_log(LOG_ERR, "%s: WR DATA FIFO not empty %u, resetting\n", __FUNCTION__, reg_value);
        fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_WR_FIFO_RESET);

        if (!fpga_ptr->regRead(ICM_LINCE_SPI_WR_DATA_FIFO_LEVEL, &reg_value))
            goto bail;

        if (reg_value != 0)
        {
            fi_log(LOG_ERR, "%s: WR DATA FIFO still not empty %u", __FUNCTION__, reg_value);
            goto bail;
        }
    }

    // We must break up the operation into chunks that fit into the FIFO because
    // once the SPI operation starts, we can't guarantee we can keep the FIFO from
    // underflowing.
    cur_addr = addr;
    cur_count = word_count;
    while (cur_count > 0)
    {
        // Set the write direction and address option.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_CONTROL, ILSC_WR_DIRECTION | ILSC_ADDRESS_ENA))
            goto bail;

        // Set the memory or register address.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_REG_ADDRESS, cur_addr))
            goto bail;

        // Set the number of bytes of data that will follow the command.
        this_op = cur_count;
        if (this_op > (LINCE5M_SENSOR::ICM_LINCE_WR_FIFO_SIZE_WORDS * 2))
            this_op = LINCE5M_SENSOR::ICM_LINCE_WR_FIFO_SIZE_WORDS * 2;
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_DATA_BYTE_COUNT, this_op * 2))
            goto bail;

        for (i = 0; i < this_op; i += 2)
        {
            // Swap bytes since the SPI is high-byte first, then write to FIFO.
            reg_value = (((uint32_t)data[i] << 8) & 0x0000ff00) |
                        (((uint32_t)data[i] >> 8) & 0x000000ff);
            if ((i + 1) < this_op)
                reg_value |= ((((uint32_t)data[i + 1] << 24) & 0xff000000) |
                              (((uint32_t)data[i + 1] << 8) & 0x00ff0000));

            if (!fpga_ptr->regWrite(ICM_LINCE_SPI_WR_DATA_FIFO, reg_value))
                goto bail;
        }

        cur_addr += this_op;
        cur_count -= this_op;

        // This starts the operation.
        if (!fpga_ptr->regWrite(ICM_LINCE_SPI_COMMAND, cmd))
            goto bail;
    }
    rc = true;

bail:
    // Return our result.
    return rc;
}

bool Lince5M::spiWriteReg(uint16_t reg_addr, uint16_t data)
{
    return spiWriteData(CMD_WRITE_REG, reg_addr, &data, 1);
}

/*!
 * @brief Write a register file to the sensor
 *
 * The code assumes that all registers are listed in increasing order in the file.
 * If this ever fails to be true, it will be necessary to update this code to break
 * up the sensor SPI writes into smaller chunks.
 *
 * @param file_name Name of register file to write, which is provided by AnaFocus
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::spiWriteRegFile(const char *file_name)
{
    int read_count;
    bool rc = false;
    FILE *fp = nullptr;
    uint16_t reg_addr;
    uint16_t last_reg = 0;
    const int REG_COUNT = 512;
    std::string temp_path;
    std::unique_ptr<uint16_t[]> buf;

    temp_path = fmt::format("{0}/{1}", get_home_path(), file_name);
    LOG("%s: file='%s'", __PRETTY_FUNCTION__, temp_path.c_str());

    buf = std::unique_ptr<uint16_t[]>(new uint16_t[REG_COUNT]());
    if (buf.get() == nullptr)
        goto bail;

    fp = fopen(temp_path.c_str(), "r");
    if (fp == nullptr)
        goto bail;

    for (read_count = 0; read_count < REG_COUNT; ++read_count)
    {
        uint16_t reg_data;

        // We should have address<sp>data<cr>.
        if (fscanf(fp, "%hx %hx\n", &reg_addr, &reg_data) != 2)
            break;

        // The first register must be zero and the rest must be consecutive.
        if (((read_count != 0) && (reg_addr != (last_reg + 1))) || ((read_count == 0) && (reg_addr != 0)))
        {
            fi_log(LOG_DEBUG, "%s: reg file has changed, need smarter code!", __FUNCTION__);
            break;
        }

        // Set the data into our buffer.
        buf[read_count] = reg_data;

        // Remember this register address so we can make sure they are consecutive.
        last_reg = reg_addr;
    }
    if (read_count == REG_COUNT)
        rc = spiWriteData(CMD_WRITE_REG, 0x000, buf.get(), REG_COUNT);
    else
        fi_log(LOG_DEBUG, "%s: read_count(%d) != REG_COUNT(%d)!!!!", __PRETTY_FUNCTION__, read_count, REG_COUNT);

bail:
    if (rc == false)
        fi_log(LOG_ERR, "%s: failed err=%d (%s)", __FUNCTION__, errno, strerror(errno));
    if (fp != nullptr)
        fclose(fp);

    return (rc);
}

/*!
 * @brief Write a memory file to the sensor
 *
 * @param memaddr   Starting memory address.
 * @param count     Maximum number of uint16_t items expected, used to allocate buffer.
 * @param file_name Name of memory file to write, which is provided by AnaFocus.
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::spiWriteMemoryFile(uint16_t memaddr, uint16_t count, const char *file_name)
{
    bool rc = false;
    FILE *fp = nullptr;
    uint16_t data, read_count;
    std::unique_ptr<uint16_t[]> buf;
    std::string temp_path;

    buf = std::unique_ptr<uint16_t[]>(new uint16_t[count]());
    if (buf.get() == nullptr)
        goto bail;

    temp_path = fmt::sprintf("%s/%s", get_home_path(), file_name);
    fp = fopen(temp_path.c_str(), "r");
    if (fp == nullptr)
        goto bail;

    for (read_count = 0; read_count < count; ++read_count)
    {
        if (fscanf(fp, "%hx\n", &data) != 1)
            break;

        buf[read_count] = data;
    }
    rc = spiWriteData(CMD_WRITE_MEM, memaddr, buf.get(), read_count);

bail:
    if (rc == false)
        fi_log(LOG_ERR, "%s: failed err=%d (%s)", __FUNCTION__, errno, strerror(errno));
    if (fp != nullptr)
        fclose(fp);
    return (rc);
}

/*!
 * @brief Read the defect map file
 *
 * @return true if we found and read the file, false otherwise.
 */
bool Lince5M::readDefectMap()
{
    FILE *fp = nullptr;
    bool rc = false;
    unsigned i;
    std::string temp_path;

    // Try the test file first, then the real one.
    temp_path = fmt::sprintf("%s/%s", get_home_path(), DEFECT_MAP_FILE_TEST);
    fp = fopen(temp_path.c_str(), "r");
    if (fp == nullptr)
    {
        std::string fname_buf;

        fname_buf = defect_filename(".config/Fastec/calibration");
        if (fname_buf.empty())
            goto bail;

        fp = fopen(fname_buf.c_str(), "r");

        // If we were not able to open a defect map file, we're done.
        if (fp == nullptr)
            goto bail;
    }

    if (fscanf(fp, "AL5Xv1:%u\n", &num_defects) != 1)
    {
        errno = EINVAL; // Invalid argument = Bad format defect map file.
        goto bail;
    }
    if (num_defects == 0)
    {
        rc = true;
        goto bail;
    }

    // Allocate space for the defect map.
    defect_map = std::unique_ptr<BAD_PIXEL[]>(new BAD_PIXEL[num_defects]());
    if (defect_map.get() == nullptr)
    {
        errno = ENOMEM; // Out of memory
        goto bail;
    }

    for (i = 0; i < num_defects; ++i)
    {
        if (fscanf(fp, "%hu,%hu\n", &defect_map[i].col, &defect_map[i].row) != 2)
            break;
    }
    if (i < num_defects)
    {
        errno = EINVAL; // Invalid argument = Bad format defect map file.
        goto bail;
    }
    fclose(fp);

    // Success.
    rc = true;

bail:
    if (rc == false)
        fi_log(LOG_ERR, "%s: failed err=%d (%s)", __FUNCTION__, errno, strerror(errno));
    return rc;
}

/*!
 * @brief Compare bad pixels.
 *
 * Used in the qsort(), this is the comparator function returning
 * 0 if equal, <0 if smaller, >0 if greater.
 *
 * @return Comparison result for qsort().
 */
constexpr int Lince5M::sbpCompare(const void *a, const void *b)
{
    const SENSOR_BAD_PIXEL *p1 = (const SENSOR_BAD_PIXEL *)a;
    const SENSOR_BAD_PIXEL *p2 = (const SENSOR_BAD_PIXEL *)b;

    if (p1->row < p2->row)
        return -1;
    else if (p1->row > p2->row)
        return 1;
    else if (p1->gcol < p2->gcol)
        return -1;
    else if (p1->gcol > p2->gcol)
        return 1;
    else
        return 0;
}

//#include "dump_short.h"
//#include "dump_hex.h"

bool Lince5M::writeDefectMap(uint32_t sensor_opts, uint16_t *p_mapsize)
{
    bool rc = false;
    int bin, subs;
    unsigned i, j, nentries;
    std::unique_ptr<uint16_t[]> wbuf;
    std::unique_ptr<SENSOR_BAD_PIXEL[]> sbp;

    // Get the binning and subsampling factors.
    subs = (sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT;
    bin  = (sensor_opts & GIGE_SENSOPT_BIN_MASK)  >> GIGE_SENSOPT_BIN_SHIFT;

    // Allocate buffer for memory writes;
    wbuf = std::unique_ptr<uint16_t[]>(new uint16_t[MAX_DEFECTS]());
    if (wbuf.get() == nullptr)
    {
        errno = ENOMEM;
        goto bail;
    }

    // Initialize MEMDEFROW with 0x0A10=2576 (addresses 0x0000-0x01ff).
    for (i = 0; i < MAX_DEFECTS; ++i)
        wbuf[i] = 0x0A10;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0000, wbuf.get(), MAX_DEFECTS))
        goto bail;

    // If there are no defects, we're done after writing null map data to memory.
    if (num_defects == 0)
    {
        rc = true;
        *p_mapsize = 0;
        goto bail;
    }

    // Allocate temporary buffer to hold munged defect list.
    sbp = std::unique_ptr<SENSOR_BAD_PIXEL[]>(new SENSOR_BAD_PIXEL[num_defects]());
    if (sbp.get() == nullptr)
    {
        errno = ENOMEM;
        goto bail;
    }

    // Copy the original defect list.
    // For now, gcol is still column number and col_bit is not set.
    for (i = 0; i < num_defects; ++i)
    {
        sbp[i].row = defect_map[i].row;
        sbp[i].gcol = defect_map[i].col;
        sbp[i].col_bit = 0;
    }
    //debug_dump_short("~/.config/Fastec/scratch/1_orig.txt", (uint16_t *)sbp, num_defects * 3, 3 );

    // For binning, we will consolidate rows and columns.
    // For subsampling, we will discard defects on skipped rows, and consolidate columns.
    j = 0;
    for (i = 0 ; i < num_defects; ++i)
    {
        uint16_t row, col;

        // For binning, map each row down to the next lower multiple of bin size.
        row = sbp[i].row & ~(bin - 1);
        //fi_log( LOG_DEBUG, "%s: bin=%u row %u->%u", __FUNCTION__, bin, sbp[i].row, row );

        // For subsampling, skip rows that are not an even multiple of the subsample size.

        // For 4x subsampling or 2x2 bin+subs, skip rows that are not multiples of 4.
        if (((subs == 4) || ((subs * bin) == 4)) && ((row  & 3) != 0))
        {
            //fi_log(LOG_DEBUG, "%s: skip (%u,%u) subs*bin=%u", __FUNCTION__, row, sbp[i].gcol, subs*bin );
            continue;
        }
        // For 2x subsampling, skip rows that are not multiples of 2.
        if ((subs == 2) && ((row & 1) != 0))
        {
            //fi_log(LOG_DEBUG, "%s: skip (%u,%u) subs=%u", __FUNCTION__, row, sbp[i].gcol, subs );
            continue;
        }

        // Save the column number.
        col = sbp[i].gcol;

        // Consolidate the columns based on the binning or subsampling factor by mapping
        // to next lower gcol that is a multiple of the factor.
        sbp[j].gcol = (col / 16) & ~((bin * subs) - 1);

        // The bit number has to be divided by the factor.
        sbp[j].col_bit = 1 << ((col / (bin * subs)) % 16);

        // Copy the row number and bump the destination count.
        sbp[j].row = row;
        //fi_log(LOG_DEBUG, "%s: sbp[j=%u]=(%u,%u) sbp[i=%u]=(%u,%u)", __FUNCTION__, j, sbp[j].row, sbp[j].gcol, i, sbp[i].row, sbp[i].gcol );
        ++j;
    }
    nentries = j;
    //debug_dump_short("~/.config/Fastec/scratch/2_bin-subs.txt", (uint16_t *)sbp, nentries * 3, 3 );

    // Sort now that some entries may have been consolidated or dropped.
    // This also means we don't require the original defect file to start out sorted.
    qsort((void *)sbp.get(), nentries, sizeof(SENSOR_BAD_PIXEL), sbpCompare);
    //debug_dump_short("~/.config/Fastec/scratch/3_sorted.txt", (uint16_t *)sbp, nentries * 3, 3 );

    // Remove duplicate (row,gcol) pairs and consolidate the column bits.
    j = 1; // Don't need to copy the first one.
    for (i = 1 ; i < nentries; ++i)
    {
        // See if this duplicates the previous entry's (row,gcol).
        if ((sbp[i].row == sbp[j - 1].row) && (sbp[i].gcol == sbp[j - 1].gcol))
        {
            // Duplicate (row,gcol) so copy the column bit.
            sbp[j - 1].col_bit |= (sbp[i].col_bit);
        }
        else
        {
            // Not a duplicate, copy it over.
            sbp[j] = sbp[i];
            ++j;
        }
    }
    nentries = j;
    //debug_dump_hex("~/.config/Fastec/scratch/4_uniq.txt", (uint8_t *)sbp, nentries * 2 * 3 );

    // Now build the data to write to the memories and write it.
    for (i = 0; i < nentries; ++i)
    {
        wbuf[i] = sbp[i].row;
        if ((i == 0) || (sbp[i - 1].row != sbp[i].row))
            wbuf[i] |= FIRST_IN_ROW;
        if ((i == (nentries - 1)) || (sbp[i].row != sbp[i + 1].row))
            wbuf[i] |= LAST_IN_ROW;
    }
    if (!spiWriteData(CMD_WRITE_MEM, 0x0000, wbuf.get(), nentries))
        goto bail;
    //debug_dump_hex("~/.config/Fastec/scratch/5_row_data.txt", (uint8_t *)wbuf, nentries * 2 );

    for (i = 0; i < nentries; ++i)
        wbuf[i] = sbp[i].gcol;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0200, wbuf.get(), nentries))
        goto bail;
    //debug_dump_hex("~/.config/Fastec/scratch/6_gcol_data.txt", (uint8_t *)wbuf, nentries * 2 );

    for (i = 0; i < nentries; ++i)
        wbuf[i] = sbp[i].col_bit;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0400, wbuf.get(), nentries))
        goto bail;
    //debug_dump_hex("~/.config/Fastec/scratch/7_col_bit_data.txt", (uint8_t *)wbuf, nentries * 2 );
    *p_mapsize = nentries;
    rc = true;

bail:
    if (rc == false)
        fi_log(LOG_ERR, "%s: failed err=%d (%s)", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Perform hard-reset and required initialization
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::hardRestart(uint32_t sensor_opts)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    bool rc = false;
    int  fm;
    uint32_t reg_value;

    LOG("%s: sensor_opts=0x%x", __PRETTY_FUNCTION__, sensor_opts);

    // Make sure the ICM is disabled.
    if (!fpga_ptr->icmDisable())
    {
        fi_log(LOG_ERR, "%s: turn_off_icm() failed!", __FUNCTION__);
        goto bail;
    }

    // First make sure the resets are set and the parallel output is disabled.
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, ILC_LVDS_RX_RESET | ILC_LVDS_RX_PLL_ARESET | ILC_SENSOR_RESET))
    {
        fi_log(LOG_ERR, "%s: ICM_LINCE_CONTROL write failed", __FUNCTION__);
        goto bail;
    }

    // ICM is supposed to be disabled right now so this should be safe.
    setIcmFrequency(sensor_opts);

    // Reset the sensor.
    if (!toggleReset())
        goto bail;

    // Still not sure if we want to leave this here.
    // Read the sensor ID after reset to make sure the interface is working.
    // This has a delay after reset to let the sensor come up, maybe should reduce that, but
    // but the reset is only used after startup, when changing framing modes, or calibrating.
    {
        struct timespec sleep_req;
        uint8_t buf[8];
        sleep_req.tv_nsec = 100 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);

        memset(buf, 0, 8);
        if (!spiReadCmd(CMD_READ_ID, buf, 4))
        {
            fi_log(LOG_ERR, "%s: could not read ID", __FUNCTION__);
            goto bail;
        }
        if (memcmp(buf, EXPECTED_SENSOR_ID, 4) != 0)
        {
            fi_log(LOG_ERR, "%s: read id %x %x %x %x", __FUNCTION__, buf[0], buf[1], buf[2], buf[3]);
            goto bail;
        }
    }

    if ((sensor_opts & GIGE_SENSOPT_INTERNAL_0) == 0)
    {
        fm = FM_1;
    }
    else
    {
        fm = FM_6;
    }
    LOG("%s: using framing mode=%u sensopts=0x%08x", __FUNCTION__, fm, sensor_opts);

    // The framing mode configuration sets up the PLLs in the sensor and then
    // sets up the LVDS lines for our framing mode. When that completes, the
    // sensor should be emitting the IDLE pattern we set.
    current_framing_mode = FM_NONE;
    if (!framingModeConfig(fm))
    {
        fi_log(LOG_ERR, "%s: frame_mode_config() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    current_framing_mode = fm;

    // Enable or disable "sensor" test mode as selected.
    fpga_ptr->regRead(ICM_LINCE_CONTROL, &reg_value);
    if (sensor_test_on)
        reg_value |= ILC_TP_ENABLE;
    else
        reg_value &= ~ILC_TP_ENABLE;
    fpga_ptr->regWrite(ICM_LINCE_CONTROL, reg_value);

    // Made it through successfully.
    rc = true;

bail:
    //Error should have been logged by failed function.
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __PRETTY_FUNCTION__);

    return rc;
}

/*
 * @brief Do a hard stop and wait for READ_STATUS<3>=0
 *
 * READ_STATUS<3> is ROI_READOUT_BUSY
 */
bool Lince5M::doHardStop()
{
    struct timespec sleep_req;
    bool rc = false;
    uint8_t status = 0xff;
    int retries, max_retries = 20;

    for (retries = 0; retries < max_retries; ++retries)
    {
        if (!spiWriteCmd(CMD_STOP_HARD_SENS, nullptr, 0))
            break;
        if (!spiReadStatus(&status))
            break;
        if ((status & DS_ROI_READOUT_BUSY) == 0)
        {
            rc = true;
            break;
        }

        // Wait 2 milliseconds.
        sleep_req.tv_nsec = 2 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }
    if (!rc)
        fi_log(LOG_ERR, "%s failed, timeout", __FUNCTION__);

    return rc;
}

/*
 * @brief Ensure the sensor is stopped
 */
bool Lince5M::ensure_stopped()
{
    uint16_t reg_value;
    bool rc = false;

    if (!doHardStop())
        goto bail;

    // Set Rg0x000<1> = 0 to disable sensor's trigger input.
    if (!spiReadReg(0x000, &reg_value))
        goto bail;

    reg_value &= ~SC_TRIGGER_ENA;
    if (!spiWriteReg(0x000, reg_value))
        goto bail;

    // Execute UPDATE_REG and UPDATE_ROI commands
    if (!update_reg_cmd() || !update_roi_cmd())
        goto bail;

    // Made it through successfully.
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*
 * @brief Send UPDATE_REG command and wait for UPDATE_BUSY to clear.
 */
bool Lince5M::update_reg_cmd()
{
    bool rc = false;
    uint8_t status = 0xFF;
    struct timespec sleep_req;
    int retries;
    int max_retries = 10, msec_wait = 2;

    if (!spiWriteCmd(CMD_UPDATE_REG, nullptr, 0))
        goto bail;

    for (retries = 0; retries < max_retries; ++retries)
    {
        if (!spiReadStatus(&status))
            break;

        // If not busy, we're done.
        if ((status & DS_UPDATE_BUSY) == 0)
        {
            rc = true;
            break;
        }

        // Sleep milliseconds.
        sleep_req.tv_nsec = msec_wait * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }

    if (retries >= max_retries)
        fi_log(LOG_ERR, "%s: timed out, status=0x%02x", __FUNCTION__, status);

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*
 * @brief Send UPDATE_ROI command and wait for UPDATE_BUSY to clear.
 */
bool Lince5M::update_roi_cmd()
{
    bool rc = false;
    uint8_t status = 0xFF;
    struct timespec sleep_req;
    int retries;
    int max_retries = 10, msec_wait = 5;

    if (!spiWriteCmd(CMD_UPDATE_ROI, nullptr, 0))
        goto bail;

    for (retries = 0; retries < max_retries; ++retries)
    {
        if (!spiReadStatus(&status))
            break;

        // If not busy, we're done.
        if ((status & DS_ROI_UPDATE_BUSY) == 0)
        {
            rc = true;
            break;
        }

        // Sleep milliseconds.
        sleep_req.tv_nsec = msec_wait * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }

    if (retries >= max_retries)
        fi_log(LOG_ERR, "%s: timed out, status=0x%02x", __FUNCTION__, status);

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Activate and enable memories
 *
 * Section 3.5.2.1 of the Operating Procedures document gives the procedure for writing
 * the memories. Writing a register file sets Rg0x0170 to 0x0000, which we do before writing
 * memory files. Then after writing, we activate the initial values and enable/lock them.
 */
bool Lince5M::commit_memory_changes()
{
    bool rc = false;

    // Activate initial values.
    if (!spiWriteReg(0x171, 0x07ff))
        goto bail;

    // Enable and lock the memories.
    if (!spiWriteReg(0x170, 0x0007))
        goto bail;

    // Now do UPDATE_REG  and UPDATE_ROI.
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Perform Framing Mode configuration for the Lince5M sensor
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::framingModeConfig(int fm)
{
    bool rc = false;
    uint8_t status;
    int retries, max_retries;
    struct timespec sleep_req;

    LOG("%s: %d", __PRETTY_FUNCTION__, fm);

    // Step I. Enable LVDS analog references
    if (!spiWriteCmd1(0x74, 0x03))        // WRITE_REF_CONFIG
    {
        goto bail;
    }

    // Step II. Disable CGU (Disable System, LVDS, ADC Ramp clocks generation)
    if (!spiWriteCmd1(0x69, 0x00) ||      // WRITE_SYS_CGU_ENA
        !spiWriteCmd1(0x68, 0x00) ||      // WRITE_LVDS_CGU_ENA
        !spiWriteCmd1(0x6a, 0x00))        // WRITE_RAMP_CGU_ENA
    {
        goto bail;
    }

    if (fm == FM_1)
    {
        // Step III. Configure on-chip PLLs (define operation frequencies)
        // 1. Set Enable registers using Framing Mode 1 line in table.
        //    This disables the PLLs before configuration.
        if (!spiWriteCmd1(0x65, 0x2a) ||      // WRITE_LVDS_PLL_ENA
            !spiWriteCmd1(0x67, 0x2a) ||      // WRITE_RAMP_PLL_ENA
            !spiWriteCmd1(0x66, 0x04))        // WRITE_SYS_PLL_ENA
        {
            goto bail;
        }

        // 2. Configure frequencies
        if (!spiWriteCmd1(0x62, 0x42) ||      // WRITE_LVDS_PLL_DIV
            !spiWriteCmd1(0x63, 0x04) ||      // WRITE_SYS_PLL_DIV
            !spiWriteCmd1(0x6b, 0x15) ||      // WRITE_SYS_CGU_DIV
            !spiWriteCmd1(0x64, 0x42))        // WRITE_RAMP_PLL_DIV
        {
            goto bail;
        }

        // Step IV. Enable on-chip PLLs
        // 1. Set Enable registers as in Step III-1, but with bit 0 = 1.
        //    This enables the PLLs.
        if (!spiWriteCmd1(0x65, 0x2b) ||      // WRITE_LVDS_PLL_ENA
            !spiWriteCmd1(0x67, 0x2b) ||      // WRITE_RAMP_PLL_ENA
            !spiWriteCmd1(0x66, 0x05))        // WRITE_SYS_PLL_ENA
        {
            goto bail;
        }
    }
    else
    {
        // Step III. Configure on-chip PLLs (define operation frequencies)
        // 1. Set Enable registers using Framing Mode 1 line in table.
        //    This disables the PLLs before configuration.
        if (!spiWriteCmd1(0x65, 0x24) ||      // WRITE_LVDS_PLL_ENA
            !spiWriteCmd1(0x67, 0x24) ||      // WRITE_RAMP_PLL_ENA
            !spiWriteCmd1(0x66, 0x04))        // WRITE_SYS_PLL_ENA
        {
            goto bail;
        }

        // 2. Configure frequencies
        if (!spiWriteCmd1(0x62, 0x38) ||      // WRITE_LVDS_PLL_DIV
            !spiWriteCmd1(0x63, 0x04) ||      // WRITE_SYS_PLL_DIV
            !spiWriteCmd1(0x6b, 0x15) ||      // WRITE_SYS_CGU_DIV
            !spiWriteCmd1(0x64, 0x38))        // WRITE_RAMP_PLL_DIV
        {
            goto bail;
        }

        // Step IV. Enable on-chip PLLs
        // 1. Set Enable registers as in Step III-1, but with bit 0 = 1.
        //    This enables the PLLs.
        if (!spiWriteCmd1(0x65, 0x25) ||      // WRITE_LVDS_PLL_ENA
            !spiWriteCmd1(0x67, 0x25) ||      // WRITE_RAMP_PLL_ENA
            !spiWriteCmd1(0x66, 0x05))        // WRITE_SYS_PLL_ENA
        {
            goto bail;
        }
    }

    // Step V. Wait for PLLs to lock
    // 100 attempts at 2 milliseconds each , or 200 milliseconds
    max_retries = 100;
    for (retries = 0; retries < max_retries; ++retries)
    {
        if (!spiReadStatus(&status))
            goto bail;

        // We are looking for all three PLL statuses to be 1.
        if ((status & 0xE0) == 0xE0)
            break;

        // Wait for update to complete.
        sleep_req.tv_nsec = 2 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }
    LOG("%s: READ_STATUS=0x%02x retries=%d", __FUNCTION__, status, retries);

    // Step VI. Enable CGUs, then UPDATE_ROI
    if (!spiWriteCmd1(0x69, 0x01) ||      // WRITE_SYS_CGU_ENA
        !spiWriteCmd1(0x68, 0x01) ||      // WRITE_LVDS_CGU_ENA
        !spiWriteCmd1(0x6a, 0x01))        // WRITE_RAMP_CGU_ENA
    {
        goto bail;
    }
    if (!spiWriteCmd(CMD_UPDATE_ROI, nullptr, 0))
        goto bail;

    // Step VII. Configure the output channels
    // Framing mode 1 configuration per Table 6:
    // Set bits per pixel and active LVDS pads
    // by writing to Rg0x025, Rg0x026, Rg0x036, Rg0x038.
    // Use the same values to make FM_6 use 24 LVDS lines and emit 12 bits.
    if (!spiWriteReg(0x025, 0x06ff) ||    // Rg0x025
        !spiWriteReg(0x026, 0xffff) ||    // Rg0x026
        !spiWriteReg(0x036, 0x0001) ||    // Rg0x036
        !spiWriteReg(0x038, 0x0000))      // Rg0x038
    {
        goto bail;
    }
    if (!update_reg_cmd())
        goto bail;

    // Step VIII. Set Analog reference.
    // Voltages: 101 100 s g = 0xb0 + serializer enable + global enable
    if (!spiWriteReg(0x024, 0x00b0))      // Rg0x024: Set voltage and disable LVDS
        goto bail;
    if (!update_reg_cmd())
        goto bail;
    if (!spiWriteReg(0x024, 0x00b1))      // Rg0x024: Global enable
        goto bail;
    if (!update_reg_cmd())
        goto bail;
    if (!spiWriteReg(0x024, 0x00b3))      // Rg0x024: Global + serializer enable
        goto bail;
    if (!update_reg_cmd())
        goto bail;

    // Step IX. LVDS skew
    // Just set idle pattern and enable output of pattern
    if (!spiWriteReg(0x034, IDLE_PATTERN | IDLE_ENABLE))       // Rg0x034
        goto bail;

#if 0
    // We don't actually need to set skew values so this code has not been tested.
    // This might be needed if our board had different length traces for the LVDS lines.
    // Rg0x027-Rg0x033 contain skew configurations.
    if (!write_data(CMD_WRITE_REG, REG_LVDS_SKEWS, skews, NUM_LVDS_SKEWS))
        goto bail;
#endif

    max_retries = 10;
    for (retries = 0; retries < max_retries; ++retries)
    {
        // Align the LVDS interface.
        if (alignLVDS(IDLE_PATTERN))
            break;

        // Wait for update to complete.
        sleep_req.tv_nsec = 2 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);
    }
    if (retries == max_retries)
    {
        LOG("%s: lince_align_lvds() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Made it through Framing Mode configuration.
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_ERR, "%s: failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Perform sensor configuration for the Lince5M sensor
 *
 * @param t_row_min     The t_row minimum timing value for this configuration
 * @param bin_fix       true to do the secret sauce fix for trow=84
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::sensor_config(unsigned t_row_min, bool bin_fix)
{
    bool rc = false;
    const char *reg_file_name;

    // Write the register file appropriate for our width.
    switch (t_row_min)
    {
    case TROW_52:
        // We only have the FM1 file. We adjust clocks but their secret sauce is the same.
        reg_file_name = FM1_TROW_52B_FILE;
        break;
    case TROW_84:
        reg_file_name = FM1_TROW_84_FILE;
        break;
    case TROW_124:
        reg_file_name = FM1_TROW_124_FILE;
        break;
    case TROW_132:
        reg_file_name = FM1_TROW_132_FILE;
        break;
    default:
    case TROW_165:
        reg_file_name = FM1_TROW_165_FILE;
        break;
    }
    if (!spiWriteRegFile(reg_file_name))
        goto bail;

    // For the stressed trow values with binning, Anafocus found that they need to
    // adjust Rg0x07c, which is the "polarization current of the pixel." Their
    // characterization showed it "needed to be raised to achieve desired performance."
    // Otherwise all the values within the two are the same. If that weren't true, we
    // would have to allow for more differences between the two files, like perhaps the
    // N_CYCLES values for states B and F or the analog gain numbers.
    // The only difference between the binnning and normal register file is this value
    // and the registers that we will overwrite anyway when we set up the ROI.
    // This is the same value for the FM1 configuration and FM6 configuration.
    if (bin_fix)
    {
        if (!spiWriteReg(0x07c, 0x0084))    // Rg0x07c
            goto bail;
    }

#ifdef USE_DUMMY_ROW_FIX
    // If we decide to fix ticket #1661 by reading out the dummy rows and
    // discarding them in the FPGA for the trow=84 configuration, we would
    // need to do this to enable the dummy rows.
    if (t_row_min == TROW_84)
    {
        if (!writeSpiReg(0x17a, 0x003e))    // Rg0x17a
            goto bail;
    }
#endif

    // Reconfigure registers we overwrote: bits per pixel, LVDS channels, IDLE pattern.
    // We rewrite Rg0x24 just as a reminder in case we decide to change it in the future;
    // the 20140626 register files already have 0x00B3 for Rg0x024 so it's redundant.
    // Rg0x025, Rg0x026, Rg0x036, and Rg0x038 are from Operating Procedures Table 6 under
    // Framing Mode Configuration.
    if (!spiWriteReg(0x025, 0x06ff) ||    // Rg0x025
        !spiWriteReg(0x026, 0xffff) ||    // Rg0x026
        !spiWriteReg(0x036, 0x0001) ||    // Rg0x036
        !spiWriteReg(0x038, 0x0000) ||    // Rg0x038
        !spiWriteReg(0x024, 0x00b3) ||    // Rg0x024
        !spiWriteReg(0x034, IDLE_PATTERN | IDLE_ENABLE))   // Rg0x034
        goto bail;

    // Now do UPDATE_REG and wait for that to complete.
    if (!update_reg_cmd())
        goto bail;

    // Write the programmable memories.
    // The register file wrote 0x0000 to Rg0x170 and Rg0x171 so the
    // memories are available to be written now.

    // Write MCB memory with Mode 5 file (Single Edge Trigger mode).
    if (!spiWriteMemoryFile(MCB_MEMADDR, MCB_WORDS, MCB_DEFAULT_FILE))
        goto bail;

    // Write GBLST for Linear mode.
    if (!spiWriteMemoryFile(GBLST_MEMADDR, GBLST_WORDS, GBLST_LINEAR_FILE))
        goto bail;

    // Write RDOUTST memory.
    if (!spiWriteMemoryFile(RDOUTST_MEMADDR, RDOUTST_WORDS, RDOUTST_DEFAULT_FILE))
        goto bail;

    if (!commit_memory_changes())
        goto bail;

    // Made it through Sensor configuration.
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_ERR, "%s: failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Perform parameters configuration for the Lince5M sensor
 *
 * @param t_row_min     The t_row minimum timing value for this configuration
 * @param analog_gain       The gain parameter (1, 2, or 4)
 * @param sensor_opts       The sensor binning/subsampling options
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::parameters_config(unsigned t_row_min, unsigned analog_gain, uint32_t sensor_opts)
{
    bool rc = false;
    uint16_t reg_value;
    int opt;

    // Configure the analog gain.
    // We start with a #defined value, which is the default analog gain taken from the regs file.
    // Then for 2x or 4x, divide by 2 or 4 and write it back to Rg0x05b.
    switch (t_row_min)
    {
    case TROW_52:
        // We don't have separate files for FM1 vs FM6.
        reg_value = FM1_ANALOG_GAIN_52 / analog_gain;
        break;
    case TROW_84:
        reg_value = FM1_ANALOG_GAIN_84 / analog_gain;
        break;
    case TROW_124:
        reg_value = FM1_ANALOG_GAIN_124 / analog_gain;
        break;
    case TROW_132:
        reg_value = FM1_ANALOG_GAIN_132 / analog_gain;
        break;
    default:
    case TROW_165:
        reg_value = FM1_ANALOG_GAIN_165 / analog_gain;
        break;
    }
    if (!spiWriteReg(LREG_ANALOG_GAIN, reg_value))
        goto bail;

    LOG("%s: t_row_min=%u analog_gain=%u value=%u", __FUNCTION__,
        t_row_min, analog_gain, reg_value);

    // Configure binning and sub-sampling in Rg0x014.
    if (!spiReadReg(LREG_SUBS_BIN, &reg_value))
        goto bail;
    reg_value &= ~SB_BINSUBS_MASK;  // Clear settable bits.
    opt = (sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT;
    switch (opt)
    {
    case 2:
        reg_value |= SB_BIN_X2;
        break;
    case 4:
        reg_value |= SB_BIN_X4;
        break;
    default:
        break;  // No binning
    }
    opt = (sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT;
    switch (opt)
    {
    case 2:
        reg_value |= SB_SUBS_X2;
        break;
    case 4:
        reg_value |= SB_SUBS_X4;
        break;
    default:
        break;  // No subsampling
    }
    if (!spiWriteReg(LREG_SUBS_BIN, reg_value))
        goto bail;
    //fi_log(LOG_DEBUG, "%s: Wrote 0x%04x to Rg0x%03x (subs/bin)", __FUNCTION__, reg_value, LREG_SUBS_BIN );

    // Configure HDR if used.
    // TODO configure HDR

    rc = true;
bail:
    if (!rc)
        fi_log(LOG_ERR, "%s: failed", __FUNCTION__);
    return rc;
}

/*
 * @brief adjust the analog offset, which sets the black level
 *
 * The sensor needs to be IDLE when starting this. If it is not part of the
 * sensor_startup() sequence, use ensure_stopped() before calling this.
 * It also requires MCB mode 5, which is set up by MCB_DEFAULT_FILE.
 * The sensor_config() sequence loaded so it is set up when get here,
 * otherwise we would need to load it.
 *
 * @param analog_gain       The gain parameter (1, 2, or 4)
 */
bool Lince5M::adjust_analog_offset(unsigned analog_gain)
{
    bool rc = false;
    struct timespec sleep_req;
    uint16_t config;
    uint16_t reg_value;
    uint16_t rows = 0x200;
    uint32_t sum, avg;
    uint32_t optimal;
    uint16_t offset;
    uint32_t integration;
    uint16_t save_rows, save_first_row, save_binning, Rg0x180_save;
    uint16_t Nrows;
    int retries, max_retries;
    uint16_t sum_high, sum_low;

    // Step I. Initialization.
    // 1-5.  We already stopped and loaded MCB mode 5
    // so Rg0x000<1>= 0, meaning trigger is disabled.

    // 6. Set Rg0x000<3>=1 for non-concurrent triggering.
    if (!spiReadReg(0x000, &config))
        goto bail;
    if (!spiWriteReg(0x000, config | SC_TRIG_NONCONCUR))
        goto bail;

    // 7. Enable statistics module of the DIP to get means of OB columns.
    if (!spiWriteReg(0x1bc, 0x0006))
        goto bail;

    // 8. Save registers.
    if (!spiReadReg(LREG_ROI_NUM_ROWS, &save_rows) ||           // Rg0x010
        !spiReadReg(LREG_ROI_FIRST_ROW, &save_first_row) ||    // Rg0x011
        !spiReadReg(LREG_SUBS_BIN, &save_binning) ||           // Rg0x014
        !spiReadReg(0x180, &Rg0x180_save))                     // Rg0x180, config hot ROI
        goto bail;

    // 9. Set registers.
    if (!spiWriteReg(LREG_ROI_NUM_ROWS, rows - 1) ||
        !spiWriteReg(LREG_ROI_FIRST_ROW, 0) ||
        !spiWriteReg(LREG_SUBS_BIN, save_binning & ~0x300) ||    // Clear Rg0x014<9:8>
        !spiWriteReg(0x180, 0x0003))                             //
        goto bail;

    // 10. Set exposure of 0 usecs:
    integration = FM1_INTEGRATION_USECS_TO_FSYS(0);
    if (!spiWriteReg(LREG_INTEGRATION_HIGH, (uint16_t)(integration >> 16)))
        goto bail;
    if (!spiWriteReg(LREG_INTEGRATION_LOW, (uint16_t)(integration & 0xffff)))
        goto bail;

    // 11. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // Set optimal value for average of optically-black columns depends on the analog gain.
    switch (analog_gain)
    {
    default:
    case 1:
        optimal = 125;
        break;
    case 2:
        optimal = 200;
        break;
    case 4:
        optimal = 340;
        break;
    }

    // Set Nrows based on binning and subsampling choices.
    Nrows = rows;
    if (((save_binning & SB_BINSUBS_MASK) == SB_BIN_X4) ||
        ((save_binning & SB_BINSUBS_MASK) == SB_SUBS_X4) ||
        ((save_binning & SB_BINSUBS_MASK) == (SB_BIN_X2 | SB_SUBS_X2)))
        Nrows = rows / 2;

    // We sometimes find that the first read after a cold boot is not consistent with
    // our second read, so just do it twice.
    for (retries = 1; retries <= 2; ++ retries)
    {
        // Step II. Capture image and check minimum value.
        // 1. Execute START_SENS command
        if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
            goto bail;

        // 2. Wait 100 milliseconds.
        sleep_req.tv_nsec = 100 * 1000 * 1000;
        sleep_req.tv_sec = 0;
        nanosleep(&sleep_req, nullptr);

        // 3. Read Rg0x1cc and Rg0x1cb, then compute average value of optically-black columns.
        if (!spiReadReg(0x1cc, &sum_low) ||
            !spiReadReg(0x1cb, &sum_high))
            goto bail;
        sum = (((uint32_t)sum_high) << 16) + (uint32_t)sum_low;
        // Compute average with rounding.
        avg = (sum + (Nrows * 8)) / (Nrows * 16);

        LOG("%s: (%d) last sum=%04x:%04x avg=%u gain=%u opt=%u", __FUNCTION__,
            retries, sum_high, sum_low, avg, analog_gain, optimal);
    }

    if (avg > optimal)
    {
        if (!spiReadReg(0x065, &reg_value))
            goto bail;
        offset = reg_value >> 8;

        max_retries = 0xE4; // Full scale of adjustment.
        for (retries = 0; retries < max_retries; ++retries)
        {
            //fi_log( LOG_DEBUG, "%s(III) avg=%u Rg0x065<15:8>=0x%02x", __FUNCTION__, avg, offset );
            if ((offset < 0xE3) && (avg > optimal))
            {
                // 1. Increase Rg0x065<15:8> by 1. We know this is safe because it is < 0xE3.
                reg_value += 0x100;
                if (!spiWriteReg(0x065, reg_value))
                    goto bail;
                offset = reg_value >> 8;

                // 2. UPDATE_REG command.
                if (! update_reg_cmd())
                    goto bail;

                // 3a. Execute START_SENS command.
                if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
                    goto bail;
                // 3b. Wait 100 milliseconds.
                sleep_req.tv_nsec = 100 * 1000 * 1000;
                sleep_req.tv_sec = 0;
                nanosleep(&sleep_req, nullptr);
                // 3c. Compute new average value of optically-black columns.
                if (!spiReadReg(0x1cc, &sum_low) ||
                    !spiReadReg(0x1cb, &sum_high))
                    goto bail;
                sum = (((uint32_t)sum_high) << 16) + (uint32_t)sum_low;
                // Compute average with rounding.
                avg = (sum + (Nrows * 8)) / (Nrows * 16);
            }
            else
            {
                if ((offset >= 0xE3) && (avg > optimal))
                {
                    fi_log(LOG_ERR, "%s: avg=%u>%u but offset=0x%02x, not good!", __FUNCTION__,
                           avg, optimal, offset);
                }
                // Either ( offset >= 0xE3 ) or ( avg <= optimal ) so we're done.
                break;
            }
        }

        // Sanity check value to allow for pathological case.
        if (offset > 0)
        {
            // 5. Undo last adjustment, decrease Rg0x065<15:8> by 1.
            reg_value -= 0x100;
            if (!spiWriteReg(0x065, reg_value))
                goto bail;
            // 6. UPDATE_REG command.
            if (! update_reg_cmd())
                goto bail;
            LOG("%s(III.6) avg=%u sum=%u Rg0x065<15:8>=0x%02x", __FUNCTION__,
                avg, sum, offset);
        }
    }
    else if (avg < optimal)
    {
        if (!spiReadReg(0x065, &reg_value))
            goto bail;
        offset = reg_value >> 8;

        max_retries = 0x41; // More than full scale of adjustment.
        for (retries = 0; retries < max_retries; ++retries)
        {
            //fi_log( LOG_DEBUG, "%s(IV) avg=%u Rg0x065<15:8>=0x%02x", __FUNCTION__, avg, offset );
            if ((offset > 0xC0) && (avg < optimal))
            {
                // 1. Decrease Rg0x065<15:8> by 1. We know this is safe because it is > 0xC0.
                reg_value -= 0x100;
                if (!spiWriteReg(0x065, reg_value))
                    goto bail;
                offset = reg_value >> 8;

                // 2. UPDATE_REG command.
                if (! update_reg_cmd())
                    goto bail;

                // 3a. Execute START_SENS command.
                if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
                    goto bail;
                // 3b. Wait 100 milliseconds.
                sleep_req.tv_nsec = 100 * 1000 * 1000;
                sleep_req.tv_sec = 0;
                nanosleep(&sleep_req, nullptr);
                // 3c. Compute new average value of optically-black columns.
                if (!spiReadReg(0x1cc, &sum_low) ||
                    !spiReadReg(0x1cb, &sum_high))
                    goto bail;
                sum = (((uint32_t)sum_high) << 16) + (uint32_t)sum_low;
                // Compute average with rounding.
                avg = (sum + (Nrows * 8)) / (Nrows * 16);
            }
            else
            {
                if ((offset <= 0xC0) && (avg < optimal))
                {
                    fi_log(LOG_ERR, "%s: avg=%u<%u but offset=0x%02x, not good!", __FUNCTION__,
                           avg, optimal, offset);
                }
                // Either ( offset <= 0xC0 ) or ( avg >= optimal ) so we're done.
                break;
            }
        }
        LOG("%s(IV.6) avg=%u sum=%u Rg0x065<15:8>=0x%02x", __FUNCTION__,
            avg, sum, offset);
    }
    else
    {
        LOG("%s: Average=%u optimal=%u, nothing to do", __FUNCTION__,
            avg, optimal);
    }

    // Step V. Restore registers.

    // 1. Turn off statistics module of the DIP.
    if (!spiWriteReg(0x1bc, 0x0000))
        goto bail;

    // 2. Restore config register.
    config &= ~SC_TRIG_NONCONCUR;
    if (!spiWriteReg(0x000, config))
        goto bail;

    if (!spiWriteReg(LREG_ROI_NUM_ROWS, save_rows) ||
        !spiWriteReg(LREG_ROI_FIRST_ROW, save_first_row) ||
        !spiWriteReg(LREG_SUBS_BIN, save_binning) ||
        !spiWriteReg(0x180, Rg0x180_save))
        goto bail;

    // Made it through successfully!
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Perform the VFPN Offset Calibration steps.
 *
 * Must do ensure_stopped() before entering this function.
 *
 * @param integration_usecs The operational setting for exposure
 * @param analog_gain       The gain parameter (1, 2, or 4)
 * @param is_mini       true for mini calibration with sensor possibly exposed to light
 *
 * @return true for success, false for any failure.
 */
bool Lince5M::vfpn_offset_calibrate(
    uint32_t integration_usecs,
    unsigned analog_gain,
    bool is_mini)
{
    struct timespec sleep_req;

    bool rc = false;
    uint32_t integration_counts;
    uint16_t config;
    uint16_t reg_value;
    uint16_t *buf = nullptr;
    unsigned count;
    uint16_t Rg0x010_save;
    uint16_t Rg0x011_save;
    uint16_t Rg0x012_save;
    uint16_t Rg0x013_save;
    uint16_t save_binning;
    uint16_t Rg0x04b_save;
    uint16_t Rg0x180_save;
    uint16_t Rg0x01c_save;
    uint16_t Rg0x015_save;
    uint16_t Rg0x172_save;
    uint16_t Rg0x14d_save;
    uint16_t Rg0x14e_save;
#ifdef ASSUME_STRONG_LIGHT
    uint16_t Rg0x048_save;
    uint16_t Rg0x049_save;
    uint16_t Rg0x0e8_save;
    uint16_t Rg0x0e9_save;
#endif
    uint32_t Nrowsx16;
    uint32_t cdoCalibration;
    uint32_t cdoFactor;
    uint32_t cdoScale;
    uint32_t cdoOffset;
    uint16_t sum_high;
    uint16_t sum_low;
    uint32_t sum;

    // Step I. Stop sensing and set operation mode
    // 1-5.  We already stopped and loaded MCB mode 5.

    // 6. Set Rg0x000<3>=1 for non-concurrent triggering.
    if (!spiReadReg(0x000, &config))
        goto bail;
    if (!spiWriteReg(0x000, config | SC_TRIG_NONCONCUR))
        goto bail;

    // 7. Unlock the VFPN offset correction memories for writing.
    if (!spiWriteReg(0x1ad, 0x0000))
        goto bail;

    // 8. Clear VFPN offset correction memory in steps:
    //    A00-BFF, C00-DFF, E00-FFF, 1000-1FF, 1200-3FF
    count = 0x200;
    buf = (uint16_t *)calloc(count, sizeof(uint16_t));
    if (buf == nullptr)
        goto bail;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0A00, buf, count))
        goto bail;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0C00, buf, count))
        goto bail;
    if (!spiWriteData(CMD_WRITE_MEM, 0x0E00, buf, count))
        goto bail;
    if (!spiWriteData(CMD_WRITE_MEM, 0x1000, buf, count))
        goto bail;
    if (!spiWriteData(CMD_WRITE_MEM, 0x1200, buf, count))
        goto bail;
    free(buf);
    buf = nullptr;

    // 9. Set exposure time. (Document is not entirely clear here.)
    if (is_mini)
    {
        integration_counts = FM1_INTEGRATION_USECS_TO_FSYS(0);
    }
    else
    {
        integration_counts = FM1_INTEGRATION_USECS_TO_FSYS(integration_usecs);
    }
    if (!spiWriteReg(LREG_INTEGRATION_HIGH, (uint16_t)(integration_counts >> 16)))
        goto bail;
    if (!spiWriteReg(LREG_INTEGRATION_LOW, (uint16_t)(integration_counts & 0xffff)))
        goto bail;
    LOG("%s: integration_usecs=%u fsys_counts=%u", __FUNCTION__,
        is_mini ? 0 : integration_usecs, integration_counts);

    // 10. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // Step II. Save registers. It doesn't matter which one fails since we would
    // have to reset anyway to recover.
    if (!spiReadReg(0x010, &Rg0x010_save) ||
        !spiReadReg(0x011, &Rg0x011_save) ||
        !spiReadReg(0x012, &Rg0x012_save) ||
        !spiReadReg(0x013, &Rg0x013_save) ||
        !spiReadReg(LREG_SUBS_BIN, &save_binning) ||   // Rg0x014
        !spiReadReg(0x015, &Rg0x015_save) ||
        !spiReadReg(0x01c, &Rg0x01c_save) ||
        !spiReadReg(0x04b, &Rg0x04b_save) ||
        !spiReadReg(0x14d, &Rg0x14d_save) ||
        !spiReadReg(0x14e, &Rg0x14e_save) ||
        !spiReadReg(0x172, &Rg0x172_save) ||
        !spiReadReg(0x180, &Rg0x180_save))
        goto bail;

    // The doc says to save these for strong light conditions, which we assume for mini caibration,
    // but we'll save them regardless.
#ifdef ASSUME_STRONG_LIGHT
    if (!readSpiReg(0x048, &Rg0x048_save) ||
        !readSpiReg(0x049, &Rg0x049_save) ||
        !readSpiReg(0x0e8, &Rg0x0e8_save) ||
        !readSpiReg(0x0e9, &Rg0x0e9_save))
        goto bail;
#endif

    // Step III. VFPN Calibration Configuration
    // 1. Write ROI, bin/subsampling with <9:8> clear, and statistics module enable
    if (!spiWriteReg(0x010, 0x07ff) ||
        !spiWriteReg(0x011, 0x0000) ||
        !spiWriteReg(0x012, 0x09ff) ||
        !spiWriteReg(0x013, 0x0000) ||
        !spiWriteReg(LREG_SUBS_BIN, save_binning & ~0x0300) ||      // Rg0x014<9:8>
        !spiWriteReg(0x1bc, 0x0006) ||
        !spiWriteReg(0x015, 0x00a7) ||
        !spiWriteReg(0x172, 0x01f8))
        goto bail;

    // 2. Write Rg0x14D and Rg0x14E.
    if (current_framing_mode == FM_6)
    {
        if (!spiWriteReg(0x14d, 0x003c) ||
            !spiWriteReg(0x14e, 0x003d))
            goto bail;
    }
    else
    {
        if (!spiWriteReg(0x14d, 0x0000) ||
            !spiWriteReg(0x14e, 0x0000))
            goto bail;
    }
    // 3. UPDATE_REG and UPDATE_ROI.
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // 4a. Execute START_SENS command.
    if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
        goto bail;
    // 4b. Wait 500 milliseconds.
    sleep_req.tv_nsec = 500 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    // 4c. Set up some parameters for computing cdoCalibration.
    Nrowsx16 = 2048 * 16;
    if (((save_binning & SB_BINSUBS_MASK) == SB_BIN_X4) ||
        ((save_binning & SB_BINSUBS_MASK) == SB_SUBS_X4) ||
        ((save_binning & SB_BINSUBS_MASK) == (SB_BIN_X2 | SB_SUBS_X2)))
        Nrowsx16 = 1024 * 16;

    switch (analog_gain)
    {
    default:
    case 1:
        cdoFactor = 1000;
        break;
    case 2:
        cdoFactor = 1080;
        break;
    case 4:
        cdoFactor = 1320;
        break;
    }
    // 4d. Read stats and compute cdoCalibration
    if (!spiReadReg(0x1cb, &sum_high) ||
        !spiReadReg(0x1cc, &sum_low))
        goto bail;
    sum = (((uint32_t)sum_high) << 16) + (uint32_t)sum_low;
    cdoCalibration = (uint32_t)(((uint64_t)sum * (uint64_t)cdoFactor) / (1000 * Nrowsx16));

    // Step IV. VFPN Calibration Configuration.
    // 1. Calculate and write cdoScale and cdoOffset.
    cdoScale = cdoCalibration / 512;
    cdoOffset = cdoCalibration / (1 << cdoScale);

    LOG("%s: sum=%u(0x%04x:0x%04x) fac=%u Nr=%u cdoCal=%u cdoScale=%u cdoOffset=%u",
        __FUNCTION__, sum, sum_high, sum_low, cdoFactor, Nrowsx16 / 16,
        cdoCalibration, cdoScale, cdoOffset);

    if (cdoScale > 3)
    {
        fi_log(LOG_ERR, "%s: Got cdoScale=%u but only have two bits!", __FUNCTION__, cdoScale);
        cdoScale = 3;
    }
    if (cdoOffset > 0x1ff)
    {
        fi_log(LOG_ERR, "%s: Got cdoOffset=0x%x but only have 9 bits!", __FUNCTION__, cdoOffset);
        cdoOffset = 0x1ff;
    }
    if (!spiReadReg(0x1a9, &reg_value))
        goto bail;
    reg_value &= ~0x1ff;    // Clear 8:0
    reg_value &= ~0xc00;    // Clear 11:10
    reg_value |= (cdoOffset | (cdoScale << 10));
    if (!spiWriteReg(0x1a9, reg_value))
        goto bail;
    LOG("%s: wrote cdoScale:cdoOffset [Rg0x1a9]=%04x", __FUNCTION__, reg_value);

    // 2. Write more registers:
    if (!spiWriteReg(0x1aa, 0x097a) ||
        !spiWriteReg(0x1ab, 0x0000))
        goto bail;

    // 3. Write more registers.
    // For mini calibration that we do automatically, assume strong light.
    if (is_mini)
    {
        if (!spiWriteReg(0x04b, 0x1740))
            goto bail;

#ifdef ASSUME_STRONG_LIGHT
        if (! writeSpiReg(0x048, 0x7fff) ||
            ! writeSpiReg(0x049, 0x3fff) ||
            ! writeSpiReg(0x0e8, 0xffff) ||
            ! writeSpiReg(0x0e9, 0xfffe))
            goto bail;
#endif
    }
    else
    {
        if (!spiWriteReg(0x04b, 0x174c))
            goto bail;
    }

    // Step V. Enable HFPN Offset Correction.
    // 1. Write the enable. (Rg0x1B5=0x001c would disable it.)
    if (!spiWriteReg(0x1b5, 0x001d))
        goto bail;
    // 2. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // Step VI. Enable VFPN Offset Calibration.
    // 1. Rg0x000<6>=1.
    if (!spiReadReg(0x000, &reg_value))
        goto bail;
    reg_value |= SC_VFPN_CAL_ENA;  // Set <6>
    if (!spiWriteReg(0x000, reg_value) ||
        !spiWriteReg(0x180, 0x0003) ||
        !spiWriteReg(0x01c, 0x0907))
        goto bail;
    // 2. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // Step VII. Execute Calibration (START_SENS Command).
    // 1. Rg0x01C<8:6>=100
    if (!spiReadReg(0x01c, &reg_value))
        goto bail;
    reg_value &= ~0x01C0;   // Clear <8:6>.
    reg_value |=  0x0100;   // <8:6> = 100
    if (!spiWriteReg(0x01c, reg_value))
        goto bail;

    // 2. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // 3. Execute START_SENS to trigger the VFPN Offset Auto-calibration process.
    if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
        goto bail;

    // 4. Wait 250 milliseconds.
    sleep_req.tv_nsec = 250 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    // 5. Write registers.
    // Rg0x01C<8:6>=101
    if (!spiReadReg(0x01c, &reg_value))
        goto bail;
    reg_value &= ~0x01C0;   // Clear <8:6>
    reg_value |=  0x0140;   // <8:6> = 101
    if (!spiWriteReg(0x01c, reg_value))
        goto bail;
    // Rg0x1AA<9>=1
    if (!spiReadReg(0x1aa, &reg_value))
        goto bail;
    reg_value |=  0x0200;   // Set <9>
    if (!spiWriteReg(0x1aa, reg_value))
        goto bail;

    // 6. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // 7. Execute START_SENS to trigger the VFPN Offset Auto-calibration process.
    if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
        goto bail;

    // 8. Wait 250 milliseconds.
    sleep_req.tv_nsec = 250 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    // 9. Rg0x01C<8:6>=100
    if (!spiReadReg(0x01c, &reg_value))
        goto bail;
    reg_value &= ~0x01C0;   // Clear <8:6>.
    reg_value |=  0x0100;   // <8:6> = 100
    if (!spiWriteReg(0x01c, reg_value))
        goto bail;

    // 10. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // 11. Execute START_SENS to trigger the VFPN Offset Auto-calibration process.
    if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
        goto bail;

    // 12. Wait 250 milliseconds.
    sleep_req.tv_nsec = 250 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    // 13. Rg0x01C<8:6>=101
    if (!spiReadReg(0x01c, &reg_value))
        goto bail;
    reg_value &= ~0x01C0;   // Clear <8:6>
    reg_value |=  0x0140;   // <8:6> = 101
    if (!spiWriteReg(0x01c, reg_value))
        goto bail;

    // 14. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // 15. Execute START_SENS to trigger the VFPN Offset Auto-calibration process.
    if (!spiWriteCmd(CMD_START_SENS, nullptr, 0))
        goto bail;

    // 16. Wait 250 milliseconds.
    sleep_req.tv_nsec = 250 * 1000 * 1000;
    sleep_req.tv_sec = 0;
    nanosleep(&sleep_req, nullptr);

    // Step VIII. Restore previous configuration.
    // 1. Restore saved registers.
    if (!spiWriteReg(0x000, config))
        goto bail;
    if (!spiWriteReg(0x010, Rg0x010_save) ||
        !spiWriteReg(0x011, Rg0x011_save) ||
        !spiWriteReg(0x012, Rg0x012_save) ||
        !spiWriteReg(0x013, Rg0x013_save) ||
        !spiWriteReg(LREG_SUBS_BIN, save_binning) ||      // Rg0x014
        !spiWriteReg(0x015, Rg0x015_save) ||
        !spiWriteReg(0x01c, Rg0x01c_save) ||
        !spiWriteReg(0x04b, Rg0x04b_save) ||
        !spiWriteReg(0x14d, Rg0x14d_save) ||
        !spiWriteReg(0x14e, Rg0x14e_save) ||
        !spiWriteReg(0x172, Rg0x172_save) ||
        !spiWriteReg(0x180, Rg0x180_save))
        goto bail;

    // If Linear and strong light (mini calibration, without lens covered), restore those registers.
    if (is_mini)
    {
#ifdef ASSUME_STRONG_LIGHT
        if (! writeSpiReg(0x048, Rg0x048_save) ||
            ! writeSpiReg(0x049, Rg0x049_save) ||
            ! writeSpiReg(0x0e8, Rg0x0e8_save) ||
            ! writeSpiReg(0x0e9, Rg0x0e9_save))
            goto bail;
#endif
    }

    // 2. Rg0x1bc=0x0000
    if (!spiWriteReg(0x1bc, 0x0000))
        goto bail;

    // 3. UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    // Made it through successfully!
    LOG("%s: VFPN %s calibration complete", __FUNCTION__,
        is_mini ? "mini" : "full");
    rc = true;

bail:
    if (buf != nullptr)
        free(buf);

    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief Enable VFPN Offset Correction
 *
 * @return true for success, false for any failure.
 */
bool Lince5M::enable_vfpn_correction()
{
    bool rc = false;

    // 1. Write registers
    if (!spiWriteReg(0x1aa, 0x0001))
        goto bail;
    if (!spiWriteReg(0x1ab, 0x0280))
        goto bail;

    // 2. Rg0x1A9
    if (!spiWriteReg(0x1a9, 0x0000))
        goto bail;

    // UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd() || ! update_roi_cmd())
        goto bail;

    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

bool Lince5M::disable_vfpn_correction()
{
    bool rc = false;

    if (!spiWriteReg(0x1aa, 0x0000))
        goto bail;
    if (!spiWriteReg(0x1ab, 0x0000))
        goto bail;
    if (!spiWriteReg(0x1a9, 0x0000))
        goto bail;

    // UPDATE_REG and UPDATE_ROI
    if (! update_reg_cmd())
        goto bail;
    if (! update_roi_cmd())
        goto bail;

    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

/*!
 * @brief This routine is used to bring the sensor up in a controlled, managed fashion.
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::sensor_startup(
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    fpnmode_t fpn_mode,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t integration_usecs,
    uint32_t sensor_opts,
    uint32_t sensor_gain)
{
    int rc;
    unsigned analog_gain = sensor_gain / 100;   // Convert from canonical form (100 -> 1, 200 -> 2).
    bool sensor_fpn_on;

    LOG("%s", __PRETTY_FUNCTION__);

    // For Lince, we use sensor FPN if either user FPN choice is enabled.
    sensor_fpn_on = ((fpn_mode & (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK)) != 0) ? true : false;

    // Start the power up sequence.
    if (!io_power_up())
    {
        fi_log(LOG_ERR, "%s: powerup failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Do a hard reset and framing mode configuration.
    if (!hardRestart(sensor_opts))
    {
        fi_log(LOG_ERR, "%s: Hard reset failed!", __PRETTY_FUNCTION__);
        return false;
    }

    rc = start_from_config(startx, starty, width, height, sensor_fpn_on, bit_mode, frame_rate,
                           integration_usecs, sensor_opts, analog_gain, true);
    if (rc)
        current_gain = analog_gain;

    if (!rc)
        fi_log(LOG_DEBUG, "%s: start_from_config() failed", __PRETTY_FUNCTION__);

    return rc;
}

/*!
 * @brief This routine is used to bring the sensor up for calibration.
 *
 * For user calibration in SENSOR_FPN mode, it is just a normal bringup with some slightly adjusted
 * parameters or procedures.
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::calibration_startup(
    bool      is_factory,
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    fpnmode_t fpn_mode,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t integration_usecs,
    uint32_t sensor_opts,
    uint32_t sensor_gain)
{
    int rc;
    unsigned analog_gain = sensor_gain / 100;   // Convert from canonical form (100 -> 1, 200 -> 2).
    bool     sensor_fpn_on;
    bool     is_mini;

    LOG("%s", __PRETTY_FUNCTION__);

    // Do a hard reset and framing mode configuration.
    if (!hardRestart(sensor_opts))
    {
        fi_log(LOG_ERR, "%s: hard_reset() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    sensor_fpn_on = ((fpn_mode & (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK)) != 0) ? true : false;

    if (is_factory)
    {
        // If we are doing a factory calibration, we need to do a mini calibration because
        // we are not necessarily capturing black frames.
        is_mini = true;

        // For a factory calibration, override the ROI and sensor options.
        startx = 0;
        starty = 0;
        width = LINCE_SENSOR_WIDTH;
        height = LINCE_SENSOR_HEIGHT;
        sensor_opts = (1 << GIGE_SENSOPT_SUBS_SHIFT) | (1 << GIGE_SENSOPT_BIN_SHIFT);

        // Force sensor correction on when capturing for defective pixel detection.
        sensor_fpn_on = true;
    }
    else
    {
        // It is always a mini calibration unless we are using only sensor FPN.
        if ((fpn_mode & (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK)) ==
            GIGE_FPN_SENSOR_ENABLE_MASK)
            is_mini = false;
        else
            is_mini = true;
    }

    rc = start_from_config(startx, starty, width, height, sensor_fpn_on, bit_mode, frame_rate,
                           integration_usecs, sensor_opts, analog_gain, is_mini);
    if (rc)
        current_gain = analog_gain;

    return rc;
}

/*!
 * @brief Wait for SPI bus to go idle
 *
 * @return  true if the bus went idle before a timeout, false otherwise.
 */
bool Lince5M::spiWaitIdle()
{
    int retry_count;
    int max_retries = 10;
    uint32_t spi_status = ILS_SPI_BUSY;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Make sure the FPGA has been initialized.
    if (!fpga_ptr->isInitialized())
        return false;

    // Read the SPI status to make sure the bus is available.
    for (retry_count = 0; retry_count < max_retries; ++retry_count)
    {
        // Check to see if the busy bit is still set.
        if (!fpga_ptr->regRead(ICM_LINCE_STATUS, &spi_status) ||
            ((spi_status & ILS_SPI_BUSY) == 0))
            break;

        // Wait 2 milliseconds.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(2));
    }

    if (spi_status & ILS_SPI_BUSY)
        fi_log(LOG_DEBUG, "%s: timeout 0x%08x", __FUNCTION__, spi_status);

    // Did we exceed our retries?
    return ((spi_status & ILS_SPI_BUSY) == 0);
}

/*!
 * @brief Bring up the sensor starting with the Sensor Configration step.
 *
 * @param user_startx       User's starting x-offset, adjusted for license restriction
 * @param user_starty       User's starting y-offset, adjusted for license restriction
 * @param user_width        Apparent width for the capture, what the user thinks the size is
 * @param user_height       Apparent height for the capture, what the user thinks the size is
 * @param sensor_fpn_on     true to use the sensor's FPN
 * @param frame_rate        Frame rate in FPS
 * @param integration_usecs Integration timme in usecs
 * @param sensor_opts       GIGE_REG_SENSOR_OPTIONS settings
 * @param analog_gain       The gain parameter (1, 2, or 4)
 * @param is_mini       true if this is a mini light-exposed calibration,
 *              false if user claims lens cap is on
 *
 * If the ROI is full-size, we must be doing a factory calibration because that's the only
 * way we support that.
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::start_from_config(uint16_t user_startx,
                                uint16_t user_starty,
                                uint16_t user_width,
                                uint16_t user_height,
                                bool     sensor_fpn_on,
                                bitmode_t bit_mode,
                                float frame_rate,
                                uint32_t integration_usecs,
                                uint32_t sensor_opts,
                                unsigned analog_gain,
                                bool     is_mini)
{
    unsigned t_row_min;
    unsigned effective_width;
    unsigned factor;
    bool rc = false;
    uint16_t config;
    uint16_t sensor_startx;
    uint16_t sensor_width;
    uint16_t sensor_starty;
    uint16_t sensor_height;
    uint16_t readout_startx;
    uint16_t readout_width;
    uint16_t defect_map_entries = 0;
    bool     bin_fix = false;   // Special handling for binned trow=84

    LOG("%s: %ux%u@(%u,%u)/%ux%u gain=%u fpn=%u", __FUNCTION__,
        user_width, user_height, user_startx, user_starty,
        ((sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT),
        ((sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT),
        analog_gain, sensor_fpn_on ? 1 : 0);

    // All we care about here is the binning/subsampling factor.
    factor = ((sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT) *
             ((sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT);
    if (factor == 0)
        factor = 1;

    // Make sure sensor is in IDLE state. Since we just reset, it should always be so.
    if (!ensure_stopped())
        goto bail;

    // If we are doing a special capture of the full sensor area, we don't want to
    // go through the standard path for getting the sensor readout coordinates.
    if ((user_width == LINCE_SENSOR_WIDTH) && (user_height == LINCE_SENSOR_HEIGHT) &&
        (factor == 1) &&
        (LINCE_SENSOR_WIDTH > LINCE_REPORTED_MAX_WIDTH) &&
        (LINCE_SENSOR_HEIGHT > LINCE_REPORTED_MAX_HEIGHT))
    {
        sensor_startx = 0;
        sensor_starty  = 0;
        sensor_width = LINCE_SENSOR_WIDTH;
        sensor_height  = LINCE_SENSOR_HEIGHT;

        readout_startx = 0;
        readout_width  = LINCE_SENSOR_WIDTH;
    }
    else
    {
        // Convert from user coordinates to sensor coordinates. Any license adjustments were done
        // before we got here. These coordinates reflect the field of view on the sensor.
        // Coordinates are generally multiplied by the binning factor but may also be offset
        // to allow for exposing less than the full sensor width.
        if (lince5m_user_to_sensor(factor, user_startx, user_starty, user_width, user_height,
                                   bit_mode, &sensor_startx, &sensor_starty, &sensor_width, &sensor_height))
        {
            fi_log(LOG_ERR, "%s: Got invalid ROIx%u (%u,%u)%ux%u", __FUNCTION__, factor,
                   user_startx, user_starty, user_width, user_height);
            goto bail;
        }

        // The readout startx and width from the sensor are reduced by the binning factor and
        // adjusted to make sure that the readout conforms to sensor restrictions
        // (64-pixel column boundaries).
        lince5m_get_readout_dims(factor, sensor_startx, sensor_width,
                                 &readout_startx, &readout_width);
    }

    // For sensor configuration, we need the width so we can pick t.row_min.
    effective_width = readout_width / factor;
    t_row_min = lince5m_get_minimum_trow(effective_width, sensor_opts);

    // If the size allows us to use a reduced t_row value, see if our frame rate actually requires
    // that we use that or if instead we can get away the cleaner configuration for width <=1920.
    // Otherwise flag that we are using the fast-mode configuration.
    if ((t_row_min == TROW_84) || (t_row_min == TROW_52))
    {
        uint32_t alt_limit;

        alt_limit = lince5m_get_max_frame_rate_trow(readout_width / factor, sensor_height / factor,
                                                    sensor_opts, TROW_124);
        if (frame_rate <= alt_limit)
        {
            t_row_min = TROW_124;
        }
    }

    // For t_row=84, if we are binning we need to "patch" one entry in the register file.
    bin_fix = false;
    if ((t_row_min == TROW_84) && (((sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT) > 1))
    {
        bin_fix = true;
    }

    // This writes MCB memory with the Mode 5 file, which we will need to fix later
    // for our actual mode.
    if (!sensor_config(t_row_min, bin_fix))
        goto bail;

    // Configure binning/subsampling, analog gain, and Linear/HDR choices.
    if (!parameters_config(t_row_min, analog_gain, sensor_opts))
        goto bail;

    // This requires MCB mode 5 and that the sensor is IDLE.
    if (!adjust_analog_offset(analog_gain))
        goto bail;

    // Make sure sensor is in IDLE state again after the offset adjustment captures.
    if (!ensure_stopped())
        goto bail;

    if (sensor_fpn_on)
    {
        // Do a calibration using the optically-black (OB) pixels.
        if (!vfpn_offset_calibrate(integration_usecs, analog_gain, is_mini))
            goto bail;

        if (!enable_vfpn_correction())
            goto bail;
    }
    else
    {
        // Or not.
        if (!disable_vfpn_correction())
            goto bail;
    }

    // Make sure the sensor automatic mode is stopped.
    if (!ensure_stopped())
        goto bail;

    // Reenable writing to the MCB memory.
    if (!spiWriteReg(0x170, 0))
        goto bail;

    // Now we can overwrite the Mode 5 memory file with the mode we really use.
    // Write MCB memory with Mode 1 file (Pulse-Width-Control + Concurrent).
    if (!spiWriteMemoryFile(MCB_MEMADDR, MCB_WORDS, MCB_REAL_FILE))
        goto bail;
    if (!writeDefectMap(sensor_opts, &defect_map_entries))
        goto bail;
    if (!commit_memory_changes())
        goto bail;

    // Write the number of GCOLs-1 and enable defect correction.
    if (defect_map_entries > 0)
    {
        // Move count to bits <9:1> and OR in the enable bit.
        defect_map_entries = ((defect_map_entries - 1) << 1) | 0x0001;
    }
    if (!spiWriteReg(0x1ae, defect_map_entries))
        goto bail;
    LOG("%s: Rg0x1AE=0x%04x", __FUNCTION__, defect_map_entries);

    // Set up the Configuration register with trigger disabled.
    if (!spiReadReg(0x000, &config))
        goto bail;

    // We never want these set at this point.
    config &= ~(SC_VFPN_CAL_ENA | SC_TRIG_NONCONCUR);
    // Clear these before setting appropriately.
    config &= ~(SC_TRIGGER_ENA | SC_TRIG_POL_MASK | SC_MODE_MASK);
    config |= SC_TRIG_POL_HIGH;
    config |= (is_sensor_color() ? SC_MODE_COLOR : SC_MODE_MONO);

    // Write that out with trigger disabled.
    if (!spiWriteReg(0x000, config))
        goto bail;

    // Adjust the digital gains and offsets.
    // Adjust offset compensation.

    if (!set_sensor_roi_regs(readout_startx, sensor_starty, readout_width, sensor_height))
        goto bail;

    // Now readout_startx will be the first pixel the FPGA sees and it will see
    // ( readout_width / factor ) pixels. Set up the FPGA for this ROI.
    if (!set_fpga_roi(factor, sensor_startx, sensor_width, readout_startx, readout_width, sensor_height))
        goto bail;

    // Now enable the trigger to the sensor.
    config |= SC_TRIGGER_ENA;
    if (!spiWriteReg(0x000, config))
        goto bail;

    LOG("%s: current_framing mode=%u sensopts=0x%08x", __FUNCTION__, current_framing_mode, sensor_opts);

    // Made it all the way through successfully.
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

bool Lince5M::sensor_to_user(
    unsigned  factor,
    uint16_t  sensor_startx,
    uint16_t  sensor_starty,
    uint16_t  sensor_width,
    uint16_t  sensor_height,
    bitmode_t bit_mode,
    uint16_t  *rtn_startx,
    uint16_t  *rtn_starty,
    uint16_t  *rtn_width,
    uint16_t  *rtn_height)
{
    *rtn_width  = sensor_width  / factor;
    *rtn_height = sensor_height / factor;

    // While we were working in sensor coordinates, we might have adjusted the user
    // ROI offset because we only show the user either the center 1920x1080 region
    // or the center 1280x1024 region. We only need to do that if factor==1.
    // Undo it now to convert back to user coordinates.
    if (factor == 1)
    {
        uint16_t max_width;
        uint16_t max_height;
        uint32_t offset;

        lince5m_max_roi_setting(bit_mode, &max_width, &max_height);

        offset = (LINCE_SENSOR_WIDTH - max_width) / 2;
        if (sensor_startx >= offset)
            *rtn_startx = sensor_startx - offset;
        else
            // This should not happen.
            *rtn_startx = 0;

        offset = (LINCE_SENSOR_HEIGHT - max_height) / 2;
        if (sensor_starty >= offset)
            *rtn_starty = sensor_starty - offset;
        else
            // This should not happen.
            *rtn_starty = 0;
    }
    else
    {
        *rtn_startx = sensor_startx / factor;
        *rtn_starty = sensor_starty / factor;
    }
    // fi_log( LOG_DEBUG, "%s: x%u (%u,%u)%ux%u -> (%u,%u)%ux%u", __FUNCTION__, factor, sensor_startx, sensor_starty, sensor_width, sensor_height, *rtn_startx, *rtn_starty, *rtn_width, *rtn_height );

    return false;
}

/*!
 * @brief Make sure the sensor ROI is legal.
 *
 * On entry, the values passed by reference are what was requested.
 * On exit, they are corrected if necessary.
 *
 * @param p_startx  Sensor-relative starting column
 * @param p_starty  Sensor-relative starting row
 * @param p_width   Sensor-relative columns
 * @param p_height  Sensor-relative rows
 * @param factor    Binning or subsampling factor
 *
 * Each coordinate is divided by factor to get the value that the user sees as ROI.
 *
 * @return true if there were any changes, false otherwise.
 */
bool Lince5M::force_legal_roi(unsigned factor, uint16_t *p_startx, uint16_t *p_starty,
                              uint16_t *p_width, uint16_t *p_height)
{
    bool any_changes = false;
    uint16_t startx = *p_startx;
    uint16_t starty = *p_starty;
    uint16_t width = *p_width;
    uint16_t height = *p_height;

    // Check the effective sizes first.
    if ((width / factor) < LINCE_ROI_MIN_WIDTH)
    {
        width = LINCE_ROI_MIN_WIDTH * factor;
        any_changes = true;
    }
    if ((height / factor) < LINCE_ROI_MIN_HEIGHT)
    {
        height = LINCE_ROI_MIN_HEIGHT * factor;
        any_changes = true;
    }

    // After dividing down the sizes by the binning/subsampling factor, we still want the size
    // to be a multiple of 2 just because the rest of the software expects it and could find it
    // confusing to sometimes allow odd values.
    factor *= 2;

    if (width > LINCE_SENSOR_WIDTH)
    {
        width = LINCE_SENSOR_WIDTH;
        any_changes = true;
    }
    if ((width % factor) != 0)
    {
        width = (width / factor) * factor;
        any_changes = true;
    }
    if ((startx % factor) != 0)
    {
        startx = (startx / factor) * factor;
        any_changes = true;
    }
    if ((startx + width) > LINCE_SENSOR_WIDTH)
    {
        startx = LINCE_SENSOR_WIDTH - width;
        any_changes = true;
    }

    if (height > LINCE_SENSOR_HEIGHT)
    {
        height = LINCE_SENSOR_HEIGHT;
        any_changes = true;
    }
    if ((height % factor) != 0)
    {
        height = (height / factor) * factor;
        any_changes = true;
    }
    if ((starty % factor) != 0)
    {
        starty = (starty / factor) * factor;
        any_changes = true;
    }
    if ((starty + height) > LINCE_SENSOR_HEIGHT)
    {
        starty = LINCE_SENSOR_HEIGHT - height;
        any_changes = true;
    }

    factor /= 2; // Undo what we did above, we don't need it any more.

    // Debug, this should never happen.
    if (((width / factor) > LINCE_REPORTED_MAX_WIDTH) || ((height / factor) > LINCE_REPORTED_MAX_HEIGHT))
    {
        fi_log(LOG_WARNING, "%s: ROI(%ux%u)x%u too big, restricting size to %ux%u", __FUNCTION__,
               *p_width, *p_height, factor, width, height);
        if (width > factor * LINCE_REPORTED_MAX_WIDTH)
            width = factor * LINCE_REPORTED_MAX_WIDTH;
        if (height > factor * LINCE_REPORTED_MAX_HEIGHT)
            height = factor * LINCE_REPORTED_MAX_HEIGHT;
    }

    if (any_changes)
    {
        fi_log(LOG_ERR, "%s: Bad sensor ROIx%u, using (%u,%u)%ux%u, was (%u,%u)%ux%u",
               __FUNCTION__, factor,
               startx, starty, width, height,
               *p_startx, *p_starty, *p_width, *p_height);

        *p_startx = startx;
        *p_starty = starty;
        *p_width = width;
        *p_height = height;
    }

    return any_changes;
}

/*!
 * @brief Enforce bandwidth limitations on the ROI.
 *
 * On exit, *rtn_width and *rtn_height have been reduced to the valid rate for bit_mode.
 *
 * @param rtn_width     ROI width, on entry the proposed width
 * @param rtn_height        ROI height, on entry the proposed height
 * @param bit_mode      Bit mode
 *
 * @return true if any changes were needed to make the capture parameters valid.
 */
bool Lince5M::bandwidth_control(
    uint16_t *rtn_width,
    uint16_t *rtn_height,
    bitmode_t  bit_mode)
{
    bool     any_changes = false;
    uint16_t max_width;
    uint16_t max_height;

    // Use common utility to get max size.
    lince5m_max_roi_setting(bit_mode, &max_width, &max_height);

    if (*rtn_width > max_width)
    {
        *rtn_width = max_width;
        any_changes = true;
    }
    if (*rtn_height > max_height)
    {
        *rtn_height = max_height;
        any_changes = true;
    }
    return any_changes;
}

/*!
 * @brief Set ICM frequency by controlling the clock generator
 *
 * We only need to reprogram the clock generator, but we want to hold the ICM in
 * reset while we do it. Hold the mutex to keep anyone else from fooling with it.
 * While the ICM is held in reset, the sensor SPI interface and the ICM registers
 * are not available.
 *
 * @param the sensor options, from which we can tell the required clock speed
 *
 * @return  true if success, false if error
 */
bool Lince5M::setIcmFrequency(uint32_t /*sensor_opts*/)
{
    bool rc = false;
    uint32_t value = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // We should never be here unless the ICM is disabled so sanity check.
    if (!fpga_ptr->regRead(ICM_CONTROL, &value))
        goto bail;

    // Sanity check. The pipeline should NOT be enabled!
    if ((value & ICMC_PIPELINE_ENABLE) != 0)
    {
        fi_log(LOG_ERR, "%s: ICM pipeline is still enabled! 0x%08x\n", __FUNCTION__, value);
        goto bail;
    }

    LOG("%s: Resetting the ICM module", __PRETTY_FUNCTION__);

    // Put the ICM in reset.
    if (!fpga_ptr->regWrite(ICM_MODULE_RESET, 1))
        goto bail;

    // Delay 10 milliseconds.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(10));

#if 0
    clock_gen cg;
    cg.set_lince_icm_frequency(sensor_opts);
#endif

    // Take the ICM out of  reset.
    if (!fpga_ptr->regWrite(ICM_MODULE_RESET, 0))
        goto bail;

    rc = true;

bail:
    return rc;
}

bool Lince5M::set_integration_time(uint32_t shutter_speed)
{
    uint32_t integration_time;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Figure out the integration time.
    integration_time = FI_FPGA::FPGA_CLOCK_SPEED_IN_USEC * shutter_speed;

    // Tell the sensor to update it's integration period.
    // Setup the integration pulse to happen at the end of the frame time.
    if (!fpga_ptr->regWrite(ICM_INTEGRATION_TIME, integration_time))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_INTEGRATION_TIME (ss=%d, time=0x%x)", __FUNCTION__, shutter_speed, integration_time);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Set the sync delay value.
 *
 * @param delay_amount The number of microseconds to delay the
 *             sync-in pulse.
 */
void Lince5M::set_sync_delay(uint32_t delay_amount)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s: delay_amount=0x%08x", __PRETTY_FUNCTION__, delay_amount);
    fpga_ptr->regWrite(ICM_1PPS_DELAY_CNT, delay_amount);
}

/*!
 * @brief Set the SyncIn information. This information includes the type of
 *    signal used for SyncIn (internal, external, GPS or IRIG), and whether
 *    it's 1-pulse/sec or 1-pulse/frame.
 *
 * @param type The SyncIn information used to configure the FPGA registers.
 */
void Lince5M::set_syncin_enable(uint32_t type)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: type=0x%08x", __PRETTY_FUNCTION__, type);

    // Are we internal?
    if ((type & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_INTERNAL)
    {
        if (!fpga_ptr->regModify(ICM_FRAME_TIMING_CONTROL, ICMFTC_SYNC_MODE_MASTER, ~ICMFTC_SYNC_MODE_MASK))
        {
            fi_log(LOG_ERR, "%s: Unable to read ICM_FRAME_TIMING_CONTROL!", __PRETTY_FUNCTION__);
            return;
        }

    }

    // We must be external. Are we 1PPS?
    else if ((type & GIGE_SYNC_IN_PULSE_MASK) == GIGE_SYNC_IN_PPS)
    {
        if (!fpga_ptr->regModify(ICM_FRAME_TIMING_CONTROL, ICMFTC_SYNC_MODE_1PPS_SLAVE, ~ICMFTC_SYNC_MODE_MASK))
        {
            fi_log(LOG_ERR, "%s: Unable to read ICM_FRAME_TIMING_CONTROL!", __PRETTY_FUNCTION__);
            return;
        }
    }

    // Must be external, 1PPF.
    else
    {
        uint32_t new_settings;
        new_settings = ICMFTC_SYNC_MODE_1PPF_SLAVE;
        new_settings |= ICMFTC_SYNC_IN_1PPF_SKIP_CNT((type & GIGE_SYNC_IN_SKIP_MASK) >> GIGE_SYNC_IN_SKIP_SHIFT);

        if (!fpga_ptr->regModify(ICM_FRAME_TIMING_CONTROL, new_settings, ~(ICMFTC_SYNC_MODE_MASK | ICMFTC_SYNC_IN_1PPF_SKIP_CNT_MASK)))
        {
            fi_log(LOG_ERR, "%s: Unable to read ICM_FRAME_TIMING_CONTROL!", __PRETTY_FUNCTION__);
            return;
        }
    }
}

bool Lince5M::set_frame_rate(float frame_rate, double clock_factor)
{
    uint32_t value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s: frame_rate=%f, clock_factor=%lf", __PRETTY_FUNCTION__, frame_rate, clock_factor);

    // Sanity check the frame rate value.
    if (approx_equals(frame_rate, 0))
    {
        fi_log(LOG_ERR, "%s: Invalid frame_rate (0).", __FUNCTION__);
        return false;
    }
    if (clock_factor == 0)
        clock_factor = 1.0;

    // Calculate the value based upon the frame rate
    value = (uint32_t)((clock_factor * FI_FPGA::FPGA_CLOCK_SPEED) / (double)frame_rate);

    // Sanity check the value.
    if (value == 0)
    {
        fi_log(LOG_ERR, "%s: Invalid frame period value!", __FUNCTION__);
        return false;
    }

    if (!fpga_ptr->regWrite(ICM_FRAME_PERIOD, value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_FRAME_PERIOD.", __FUNCTION__);
        return false;
    }

    // Make sure we tell FPGA the number of intergrations pulses
    // to use (i.e. the frame rate) if we were in 1PPS sync mode.
    fpga_ptr->regWrite(ICM_1PPS_FRAME_CNT, (uint32_t)(frame_rate + 0.5));

    return true;
}

/*!
 * @brief Set the sensor's ROI registers.
 *
 * ROI parameters passed into the sensor do not reflect any license restrictions.
 * They refer to the physical ROI of the sensor.
 *
 * @param startx    Sensor-relative starting column, multiple of 2;
 * @param starty    Sensor-relative starting row, multiple of 2.
 * @param height    Sensor-relative width, multiple of 2;
 * @param height    Sensor-relative height, multiple of 2.
 *
 * @return true if everything goes well, false otherwise.
 */
bool Lince5M::set_sensor_roi_regs(uint16_t startx, uint16_t starty,
                                  uint16_t width, uint16_t height)
{
    bool rc = false;

    if (!spiWriteReg(LREG_ROI_FIRST_COL, startx))
        goto bail;
    // NUM_COLS register wants "value-1"
    if (!spiWriteReg(LREG_ROI_NUM_COLS, width - 1))
        goto bail;

    if (!spiWriteReg(LREG_ROI_FIRST_ROW, starty))
        goto bail;
    // NUM_ROWS register wants "value-1"
    if (!spiWriteReg(LREG_ROI_NUM_ROWS, height - 1))
        goto bail;

    LOG("%s: sensor roi (%u,%u)%ux%u", __FUNCTION__, startx, starty, width, height);

    // UPDATE_REG and UPDATE_ROI
    if (!update_reg_cmd())
        goto bail;
    if (!update_roi_cmd())
        goto bail;

    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

bool Lince5M::set_sensor_readout(bool enable)
{
    bool rc = false;
    uint32_t reg_value = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: enable=%d", __PRETTY_FUNCTION__, enable);

    // Read the control register, set the readout enable bit and
    // then write the value back out to the register.
    if (fpga_ptr->regRead(ICM_FRAME_TIMING_CONTROL, &reg_value))
    {
        if (enable)
            reg_value |= ICMFTC_SENSOR_READOUT_ENA;
        else
            reg_value &= ~ICMFTC_SENSOR_READOUT_ENA;

        if (fpga_ptr->regWrite(ICM_FRAME_TIMING_CONTROL, reg_value))
            rc = true;
    }

    return rc;
}

/*!
 * @brief Reset the sensor by toggling its reset line
 *
 * @return true if everything went well, otherwise false.
 */
bool Lince5M::toggleReset()
{
    bool rc = false;
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Set and clear reset
    if (!fpga_ptr->regRead(ICM_LINCE_CONTROL, &reg_value))
        goto bail;
    reg_value |= ILC_SENSOR_RESET;
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, reg_value))
        goto bail;

    // Delay 2 microseconds.
    using us = std::chrono::microseconds;
    std::this_thread::sleep_for(us(2));

    reg_value &= ~ILC_SENSOR_RESET;
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, reg_value))
        goto bail;

    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Update the sensor parameters with as little disruption as possible.
 *
 * @param true_startx   Starting x position in pixels, sensor-relative
 * @param true_starty   Starting y position in pixels
 * @param roi_width Width in pixels to capture
 * @param roi_height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param integration   Integration time (exposure) in microseconds
 * @param sensor_opts   Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100 )
 *
 * Following the flow diagram at Figure 2 in Section 3.2 of Operating Procedures:
 * - If t_row_min changes, we have to restart from sensor_config.
 * - If analog gain or binning/subsampling changes but not t_row_min, we could restart from
 * parameters_config but it would require reloading MCB Mode 5, so it doesn't save much time
 * to start there; we'd still have to go through adjust_analog_offset and vfpn_offset_correction,
 * which is what takes all the time.
 * - The only optimization we take is if the ROI changes without needing a t_row_min change.
 *
 * @return true if everything is good, otherwise false.
 */
bool Lince5M::update_capture(
    uint16_t  true_startx,
    uint16_t  true_starty,
    uint16_t  roi_width,
    uint16_t  roi_height,
    fpnmode_t fpn_mode,
    bitmode_t bit_mode,
    float  frame_rate,
    uint32_t  integration,
    uint32_t  sensor_opts,
    uint32_t  sensor_gain)
{
    unsigned analog_gain = sensor_gain / 100; // Convert from canonical form (100 -> 1, 200 -> 2).
    bool sensor_fpn_on;

    // For Lince, we use sensor FPN if either user FPN choice is enabled.
    sensor_fpn_on = ((fpn_mode & (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK)) != 0) ? true : false;

    // Do a hard restart if the framing mode changes.
    if ((((sensor_opts & GIGE_SENSOPT_INTERNAL_0) != 0) && (current_framing_mode != 6)) ||
        (((sensor_opts & GIGE_SENSOPT_INTERNAL_0) == 0) && (current_framing_mode != 1)))
    {
        if (!hardRestart(sensor_opts))
            return false;
    }

    // Optimize the capture update so we change just the ROI if we can get away with it.
    //    if ( sensor_started ) && ( false /* fast change is okay */ ))
    //    else
    // This starts from the Sensor Configuration step without a hard reset.
    // We need to do this any time t_row changes because of a width change or when
    // binning, subsampling, or analog gain changes.
    return start_from_config(true_startx, true_starty, roi_width, roi_height,
                             sensor_fpn_on, bit_mode, frame_rate, integration, sensor_opts,
                             analog_gain, true);
}

bool Lince5M::set_fpga_bit_mode(bitmode_t bitmode)
{
    bool rc = false;
    uint32_t value = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: bit_mode=%d", __PRETTY_FUNCTION__, (int)bitmode);

    if (!fpga_ptr->regRead(ICM_LINCE_CONTROL, &value))
        goto bail;

    // Clear the bits so we can set the ones we want.
    value &= ~(ILC_BIT_DEPTH_MASK | ILC_BIT_SEL_MASK);

    // Now figure out the bit mode,
    switch ((unsigned)bitmode)
    {
    case GIGE_BITMODE_11_0:
        value |= ILC_BIT_DEPTH_12;
        break;
    case GIGE_BITMODE_9_0:
        value |= (ILC_BIT_DEPTH_10 | ILC_BIT_SEL_9_0);
        break;
    case GIGE_BITMODE_10_1:
        value |= (ILC_BIT_DEPTH_10 | ILC_BIT_SEL_10_1);
        break;
    case GIGE_BITMODE_11_2:
    case GIGE_BITMODE_10_BITS:
        value |= (ILC_BIT_DEPTH_10 | ILC_BIT_SEL_11_2);
        break;
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_LOWER_8_BITS:
        value |= (ILC_BIT_DEPTH_8 | ILC_BIT_SEL_7_0);
        break;
    case GIGE_BITMODE_8_1:
        value |= (ILC_BIT_DEPTH_8 | ILC_BIT_SEL_8_1);
        break;
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_MIDDLE_8_BITS:
        value |= (ILC_BIT_DEPTH_8 | ILC_BIT_SEL_9_2);
        break;
    case GIGE_BITMODE_10_3:
        value |= (ILC_BIT_DEPTH_8 | ILC_BIT_SEL_10_3);
        break;
    case GIGE_BITMODE_11_4:
    case GIGE_BITMODE_UPPER_8_BITS:
    default:
        value |= (ILC_BIT_DEPTH_8 | ILC_BIT_SEL_11_4);
        break;
    }

    // Write the settings back to the register.
    if (!fpga_ptr->regWrite(ICM_LINCE_CONTROL, value))
        goto bail;

    rc = true;

bail:
    return rc;
}

/*!
 * @brief Setup the FGPA registers for a specific ROI value.
 *
 * These ROI values are already adjusted to be relative to the full sensor coordinates
 * by undoing the license adjustments but they do not have any adjustments for kernel boundaries,
 * which are specific to the particular sensor. The FPGA can drop some pixels before writing
 * a line to memory to achieve the desired ROI.
 *
 * The ROI that is visible outside the sensor-specific code has no "awareness" of the
 * binning or subsampling options. It is always the number of pixels that are written
 * to memory during captures.
 *
 * @param factor    Binning/subsampling factor (1, 2, or 4)
 * @param sensor_startx Starting x position in pixels of sensor window
 * @param sensor_width  Width in pixels of sensor window
 * @param readout_startx    Starting x position in pixels of sensor window
 * @param readout_width Width in pixels of sensor window
 * @param roi_starty    Starting y position in pixels
 * @param roi_height    Height in pixels to capture
 *
 * @return true if everything is good, otherwise false.
 */
bool Lince5M::set_fpga_roi(unsigned factor, uint16_t sensor_startx, uint16_t sensor_width,
                           uint16_t readout_startx, uint16_t readout_width, uint16_t height)
{
    bool rc = false;
    unsigned discard;
    uint32_t ppw_per_row, first_ppw_ofs, last_ppw_count;

//    LOG("%s: factor:%d sensor_start:%d,%d readout_start:%d,%d height:%d", __PRETTY_FUNCTION__,
//      factor, sensor_startx, sensor_width, readout_startx, readout_width, height);

    // First PPW offset is the starting pixel (0-63) in the first block.
    first_ppw_ofs = ((sensor_startx - readout_startx) / factor) % PIXELS_PER_BLOCK;

    // Last PPW count is the number of pixels in the last block, 1-64.
    discard = (readout_startx + readout_width - (sensor_startx + sensor_width)) / factor;
    if (discard >= PIXELS_PER_BLOCK)
    {
        fi_log(LOG_ERR, "%s: shouldn't happen sensor=%u:%u readout=%u:%u factor=%u discard=%u",
               __FUNCTION__, sensor_startx, sensor_width, readout_startx, readout_width, factor, discard);
        discard = 0;
    }
    last_ppw_count = PIXELS_PER_BLOCK - discard;

    // The number of PPWs actually seen by the FPGA is divided by the binning factor.
    ppw_per_row = (readout_width / factor) / PIXELS_PER_PPW;

    // Tell the FPGA what the sensor is sending.
    if (!icmSensorRoi(first_ppw_ofs, ppw_per_row, last_ppw_count, height / factor))
        goto bail;

    // Everything worked out ok.
    rc = true;

bail:
    if (!rc)
        fi_log(LOG_DEBUG, "%s failed", __FUNCTION__);
    return rc;
}

//--------------------------------------------------------------------
// Public Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Lince5M::Lince5M(SENSOR::EEPROM_INFO *eeprom)
    : Sensor(
          LINCE_REPORTED_MAX_WIDTH,
          LINCE_REPORTED_MAX_HEIGHT,
          LINCE_ROI_MIN_WIDTH,
          LINCE_ROI_MIN_HEIGHT,
          eeprom)
{
    LOG("%s", __PRETTY_FUNCTION__);

    current_gain = 0;
    current_framing_mode = FM_NONE;
    num_defects = 0;
    sensor_test_on = false;
    vtroi_x = vtroi_y = vtroi_width = vtroi_height = 0;
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Lince5M::~Lince5M()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

bool Lince5M::class_init(FiFPGA *fpga_ptr)
{
    // Turn off power to the sensor board.
    fpga_ptr->regModify(NC1_FPGA_CONTROL, 0, ~NC1FC_ISB_PWR_ENABLE);

    // Wait a bit.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(50));

    // Tell the FPGA to assert POWER_EN for the sensor board.
    if (!fpga_ptr->regModify(NC1_FPGA_CONTROL, NC1FC_ISB_PWR_ENABLE, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable power for the sensor board!", __PRETTY_FUNCTION__);
        return false;
    }
    LOG("%s: ISB_PWR_ENABLE. ========================", __PRETTY_FUNCTION__);

    readDefectMap();
    return true;
}

/*!
 * @brief Enable the sensor test pattern, which for Lince is an ICM test pattern.
 */
bool Lince5M::enable_sensor_test()
{
    sensor_test_on = true;
    return true;
}

/*!
 * @brief Start capture for PRNU calibration of a LuximaHS sensor
 *
 * @return true if successful, false otherwise
 */
bool Lince5M::start_prnu_capture()
{
    uint16_t true_x;
    uint16_t true_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;
    uint32_t sensor_gain;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure FPN correction is off while we calibrate.
    FiImageCorrection::fpnDisable();

    // Get the user settings, adjusted to the physical sensor dimensions.
    cs_ptr->get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode,
                                  &frame_rate, &shutter_speed, &sensor_opts);
    if ((bit_mode != (bitmode_t)GIGE_BITMODE_11_0) ||
        (true_x != 0) || (true_y != 0) ||
        (width != get_sensor_max_width()) ||
        (height != get_sensor_max_height()))

    {
        fi_log(LOG_WARNING, "%s: System not setup properly for factory calibration!", __PRETTY_FUNCTION__);

        // Use what we know to be valid settings.
        true_x = 0;
        true_y = 0;
        width = get_sensor_max_width();
        height = get_sensor_max_height();
        bit_mode = (bitmode_t)GIGE_BITMODE_11_0;
    }

    // Make sure we're not binning or anything.
    sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);
    sensor_gain = get_default_gain();
    fi_log(LOG_INFO, "%s: Adjusted frame_rate=%f, shutter_speed=%d, bit_mode=%d", __PRETTY_FUNCTION__, frame_rate, shutter_speed, bit_mode);

    // Start up the sensor.  This also sets up the FPGA ICM ROI since the way that is done depends
    // on the specific sensor.
    //
    // For PRNU, we want to make sure sensor FPN is enabled. This is because, on the Lince5M,
    // "pixel" FPN includes "sensor" FPN. We don't want the FPGA to do the pixel FPN correction,
    // but we need the sensor to do it's thing.
    if (!sensor_startup(true_x, true_y, width, height, (fpnmode_t)GIGE_FPN_SENSOR_ENABLE_MASK, bit_mode,
                        frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: sensor_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Set up the ICM pipeline. Nothing will happen until we turn on the sensor readout bit.
    if (!fpga_ptr->icmStartup(bit_mode, frame_rate, shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: icm_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    if (!fpga_ptr->icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __PRETTY_FUNCTION__);
        return false;
    }

    fpga_ptr->icmSetLiveArgs(bit_mode, frame_rate, shutter_speed);

    return true;
}

/*!
 * @brief Shutdown the sensor hardware.
 */
void Lince5M::shutdown()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    if (fpga_ptr != nullptr)
    {
        // Force the reset signal.
        fpga_ptr->regWrite(ICM_LINCE_CONTROL, ILC_SENSOR_RESET);

        // Reset the external memory interface and power off the sensor board.
        if (!fpga_ptr->regWrite(NC1_FPGA_CONTROL, NC1FC_EMIF_SW_RESET))
            fi_log(LOG_ERR, "%s: Unable to power off sensor board.", __PRETTY_FUNCTION__);
        else
            fi_log(LOG_DEBUG, "%s: Sensor board powered off.", __PRETTY_FUNCTION__);
    }

    Sensor::shutdown();
}

/*!
 * @brief initialize anything that requires the GigE register cache to have been initialized.
 */
void Lince5M::set_loaded_registers()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t opts;

    // Make sure FM6 is turned off.
    opts = regs[ GIGE_REG_SENSOR_OPTIONS ].get_value();
    opts &= ~GIGE_SENSOPT_INTERNAL_0;
    regs[ GIGE_REG_SENSOR_OPTIONS ].set_value(opts);
}

// Say whether this fpn change requires that the sensor be restarted.
bool Lince5M::fpn_needs_restart(fpnmode_t old_fpn, fpnmode_t new_fpn)
{
    bool old_used_fpn;
    bool new_uses_fpn;

    if ((old_fpn & (GIGE_FPN_SENSOR_ENABLE_MASK | GIGE_FPN_FPGA_ENABLE_MASK)) != 0)
        old_used_fpn = true;
    else
        old_used_fpn = false;

    if ((new_fpn & (GIGE_FPN_SENSOR_ENABLE_MASK | GIGE_FPN_FPGA_ENABLE_MASK)) != 0)
        new_uses_fpn = true;
    else
        new_uses_fpn = false;

    if (old_used_fpn != new_uses_fpn)
        return true;
    else
        return false;
}

// Say whether this bit_mode change requires that the sensor be given an opportunity
// to restart via update_capture(). We use FM1 for 8-bit modes and FM6 for 10- or
// 12-bit modes, so we might need to restart the sensor on those changes.
// Also, we want to make adjustments based on the usable ROI of the sensor when
// we change between 12-bit mode and non-12-bit mode.
bool Lince5M::bitmode_needs_restart(bitmode_t old_bit_mode, bitmode_t new_bit_mode)
{
    if (IS_8_BIT_MODE(old_bit_mode) != IS_8_BIT_MODE(new_bit_mode))
        return true;
    else if (IS_12_BIT_MODE(old_bit_mode) != IS_12_BIT_MODE(new_bit_mode))
        return true;
    else
        return false;
}

// Say whether this frame_rate change requires that we restart via update_capture().
bool Lince5M::frame_rate_needs_restart(float old_fps,
                                       float new_fps,
                                       uint16_t startx,
                                       uint16_t starty,
                                       uint16_t width,
                                       uint16_t height,
                                       bitmode_t bit_mode,
                                       uint32_t  sensor_opts)
{
    uint32_t alt_limit;
    uint16_t sensor_x;
    uint16_t sensor_y;
    uint16_t sensor_width;
    uint16_t sensor_height;
    uint16_t readout_x;
    uint16_t readout_width;
    unsigned factor;
    unsigned t_row_min;
    unsigned effective_width;

    // Get the binning/subsampling factor.
    factor = lince5m_binsubs_factor(sensor_opts);

    // Convert to sensor coordinates.
    (void)lince5m_user_to_sensor(factor, startx, starty, width, height, bit_mode,
                                 &sensor_x, &sensor_y, &sensor_width, &sensor_height);

    // Get sensor readout coordinates
    lince5m_get_readout_dims(factor, sensor_x, sensor_width, &readout_x, &readout_width);

    // For sensor configuration, we need the width so we can pick t.row_min.
    effective_width = readout_width / factor;
    t_row_min = lince5m_get_minimum_trow(effective_width, sensor_opts);

    // This scenario only applies for the fast-mode trow values.
    // For other values we don't care about frame rate.
    if ((t_row_min != TROW_84) && (t_row_min != TROW_52))
    {
        fi_log(LOG_DEBUG, "%s: trow=%u width=%u/%u FPS:%f->%f", __FUNCTION__, t_row_min, readout_width, effective_width, old_fps, new_fps);
        return false;
    }

    // If the size allows us to use a reduced t_row value, see if our frame rate actually requires
    // that we use that or if instead we can get away the cleaner configuration for width <=1920.
    // Otherwise flag that we are using the fast-mode configuration.
    alt_limit = lince5m_get_max_frame_rate_trow(readout_width / factor, sensor_height / factor,
                                                sensor_opts, TROW_124);
    fi_log(LOG_DEBUG, "%s: trow=%u width=%u/%u FPS:%f->%f alt=%u", __FUNCTION__, t_row_min, readout_width, effective_width, old_fps, new_fps, alt_limit);

    if (((old_fps <= (float)alt_limit) && (new_fps > (float)alt_limit)) ||
        ((new_fps <= (float)alt_limit) && (old_fps > (float)alt_limit)))
        return true;
    else
        return false;
}

bool Lince5M::is_bit_mode_supported(bitmode_t bit_mode)
{
    switch ((unsigned)bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
    case GIGE_BITMODE_MIDDLE_8_BITS:
    case GIGE_BITMODE_UPPER_8_BITS:
    case GIGE_BITMODE_10_BITS:
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_8_1:
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_10_3:
    case GIGE_BITMODE_11_4:
    case GIGE_BITMODE_9_0:
    case GIGE_BITMODE_10_1:
    case GIGE_BITMODE_11_2:
    case GIGE_BITMODE_11_0:
        return true;
    default:
        return false;
    }
}

bool Lince5M::is_option_supported(uint32_t sensor_opts)
{
    bool rc = true;
    unsigned opt;

    // Check subsampling.
    opt = (sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT;
    switch (opt)
    {
    case 0:
    case 1:
    case 2:
    case 4:
        break;
    default:
        rc = false;
        break;
    }

    // Check binning.
    opt = (sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT;
    switch (opt)
    {
    case 0:
    case 1:
    case 2:
    case 4:
        break;
    default:
        rc = false;
        break;
    }

    // Check the combination of the two.
    if ((((sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT) == 4) &&
        (((sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT) == 4))
        rc = false;

    return rc;
}

/*!
 * @brief Get Lince Frame Overhead Time in microseconds
 *
 * This is used to compute the maximum exposure permitted for a particular frame rate.
 * We have it here instead of inline so that the definition is kept with the rest of the
 * constants that were derived from AnaFocus-provided register files, which need to be checked
 * and possibly updated if we get new files.
 *
 * @return overhead in usecs
 */
uint32_t Lince5M::get_frame_overhead()
{
    return LINCE_FOT_USEC;
}

/*!
 * @brief Return the default bit mode we want to use with this sensor.
 */
bitmode_t Lince5M::get_default_bit_mode()
{
    return (bitmode_t)GIGE_BITMODE_11_4;
}

/*!
 * @brief Return the default frame rate we want to use with this sensor.
 */
uint32_t Lince5M::get_default_frame_rate()
{
    return DEF_FRAME_RATE;
}

/*!
 * @brief Return the default gain for this sensor in canonical form, where 1x=100, 2x=200, 4x=400.
 */
uint32_t Lince5M::get_default_gain()
{
    return 100;
}

uint32_t Lince5M::get_frame_rate()
{
    uint32_t    reg_val = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fpga_ptr->regRead(ICM_FRAME_PERIOD, &reg_val);

    return reg_val;
}

uint16_t Lince5M::get_reported_max_width(uint32_t sensor_opts, bitmode_t bit_mode)
{
    uint16_t max_width, max_height;
    uint16_t rtn_width;
    unsigned factor;

    factor = lince5m_binsubs_factor(sensor_opts);
    rtn_width = LINCE_SENSOR_WIDTH / factor;

    lince5m_max_roi_setting(bit_mode, &max_width, &max_height);

    if (rtn_width > max_width)
        rtn_width = max_width;

    return rtn_width;
}

uint16_t Lince5M::get_reported_max_height(uint32_t sensor_opts, bitmode_t bit_mode)
{
    uint16_t max_width, max_height;
    uint16_t rtn_height;
    unsigned factor;

    factor = lince5m_binsubs_factor(sensor_opts);
    rtn_height = LINCE_SENSOR_HEIGHT / factor;

    lince5m_max_roi_setting(bit_mode, &max_width, &max_height);

    if (rtn_height > max_height)
        rtn_height = max_height;

    return rtn_height;
}

/*!
 * @brief Return the minimum shutter speed for this sensor
 */
uint32_t Lince5M::get_sensor_min_shutter()
{
    return LINCE_MIN_EXP_USEC;
}

/*!
 * @brief Set the sensor gain.
 *
 * For the Lince5M, setting the gain requires that we go through the whole sensor startup
 * sequence. But we'll just go ahead and set it anyway, then the next time we go through
 * the startup sequence it will get fixed.
 *
 * @param new_sensor_gain   The value to set in canonical form, 100 for 1x, 200 for 2x, 400 for 4x.
 *
 * @return  -1 if value is illegal
 *       0 if it was good but needs an update to go into effect
 *       1 if value is good and was put into effect. Since that requires a sensor restart,
 *         we only return 1 if it is the same as the gain we're already using.
 */
int Lince5M::set_sensor_gain(uint32_t new_sensor_gain)
{
    int rtn = -1;

    new_sensor_gain &= ~GIGE_SENSOR_GAIN_IMMEDIATE;
    switch (new_sensor_gain)
    {
    case 100:
    case 200:
    case 400:
        if ((new_sensor_gain / 100) != current_gain)
            rtn = 0;
        else
            rtn = 1;    // Same as the gain we already have.
        break;
    default:
        rtn = -1;
        break;
    }
    return rtn;
}

/*!
 * @brief Read the temperature sensor on the sensor board.
 */
double Lince5M::get_temperature()
{
    uint8_t oldh, newh, lo_byte;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    /*
     * There is a trick here. We have to read two registers to have the
     * sensor temperature, but we have to beware a conversion could occur
     * between the readings. The datasheet says we should either use
     * the one-shot conversion register, which we don't want to do
     * (disables hardware monitoring) or monitor the busy bit, which is
     * impossible (we can't read the values and monitor that bit at the
     * exact same time). So the solution used here is to read the high
     * byte once, then the low byte, then the high byte again. If the new
     * high byte matches the old one, then we have a valid reading. Else
     * we have to read the low byte again, and now we believe we have a
     * correct reading.
     */
    if (!fpga_ptr->i2cRead(TMP451_I2C_ADDR, TMP451_REG_R_LOCAL_TEMP, &oldh, FI_FPGA::I2C_SENSOR) ||
        !fpga_ptr->i2cRead(TMP451_I2C_ADDR, TMP451_REG_R_LOCAL_TEMPL, &lo_byte, FI_FPGA::I2C_SENSOR) ||
        !fpga_ptr->i2cRead(TMP451_I2C_ADDR, TMP451_REG_R_LOCAL_TEMP, &newh, FI_FPGA::I2C_SENSOR))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0;
    }
    if (oldh != newh)
    {
        if (!fpga_ptr->i2cRead(TMP451_I2C_ADDR, TMP451_REG_R_LOCAL_TEMPL, &lo_byte, FI_FPGA::I2C_SENSOR))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C on lo_byte!", __PRETTY_FUNCTION__);
            return 0;
        }
    }

    // All is well. The high byte is degrees Celsius, the low byte is the decimal value.
    return (((double)newh) + ((double)(lo_byte >> 4) / 16.0));
}

/*!
 * @brief Set the FPGA's ROI from the sensor
 *
 * The FPGA has no awareness of the sensor's binning and subsampling
 * options. It sees only rows and columns of data, whatever the sensor
 * is sending. When we are binning the rows and column counts were
 * reduced by a factor of 2 or 4 but we still need to set up the ROI
 * in multiples of PIXELS_PER_BLOCK.
 */
bool Lince5M::icmSensorRoi(
    uint32_t first_ppw_ofs,
    uint32_t ppw_per_row,
    uint32_t last_ppw_count,
    uint32_t height)
{
    bool rc = false;
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Each Pixel Processing Window is 16 pixels.
    reg_value = (ppw_per_row << ILRH_PPWS_PER_ROW_SHIFT) +
                (first_ppw_ofs << ILRH_FIRST_PPW_OFFSET_SHIFT) +
                (last_ppw_count << ILRH_LAST_PPW_COUNT_SHIFT);

    LOG("%s: height=%u ofs=%u last=%u ppws=%u reg=0x%08x%s", __FUNCTION__,
        height, first_ppw_ofs, last_ppw_count, ppw_per_row, reg_value,
        (reg_value & ~((ILRH_PPWS_PER_ROW_MASK & 0xFE) |      /* row ppws must be multiple of 4 */
                       ILRH_FIRST_PPW_OFFSET_MASK | ILRH_LAST_PPW_COUNT_MASK))
        ? " ILLEGAL" : "");

    if (!fpga_ptr->regWrite(ICM_LINCE_ROI_H_CONFIG, reg_value))
        goto bail;

    // Number of rows
    if (!fpga_ptr->regWrite(ICM_LINCE_ROI_V_CONFIG, height))
        goto bail;

    rc = true;

bail:
    return rc;
}

/*!
 * @brief Start the power up sequence.
 *
 * @return true if everything went well, otherwise false.
 */
bool Lince5M::io_power_up()
{
    bool rc = false;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    if (!fpga_ptr->regWrite(ICM_LINCE_POWER_ENABLE, 0))
        goto bail;

    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Say how the Bayer pattern starts depending on the xy position
 *
 * We can return the value based on what it would be but in fact we only support
 * even offsets so it is always the same.
 *
 * Say how the source pixels should be interpreted as far as color based upon
 * x and y offset. For the Lince5M181, the pattern starting from upper left is
 *
 *   GRGR
 *   BGBG
 *
 * bits x_offset y_offset src_frm_bayer_pattern_start
 * ---- -------- -------- ----------------------------
 *  0   even     even     GR    10
 *  1   even     odd      BG    01
 *  2   odd      even     RG    11
 *  3   odd      odd      GB    00
 */
uint32_t Lince5M::IPM_bayer_pattern_start(uint16_t startx, uint16_t starty)
{
    uint32_t value;
    int bits = ((startx & 1) << 1) | (starty & 1);

    switch (bits)
    {
    default:
    case 0: // 00
        value = IPMC_SRC_FRM_BAYER_GR;
        break;
    case 1: // 01
        value = IPMC_SRC_FRM_BAYER_BG;
        break;
    case 2: // 10
        value = IPMC_SRC_FRM_BAYER_RG;
        break;
    case 3: // 11
        value = IPMC_SRC_FRM_BAYER_GB;
        break;
    }
    return value;
}

/*!
 * @brief Enforce sensor-specific restrictions on the capture parameters
 *
 * At minimum, it must check the frame rate because that is always sensor-specific.
 * The parameters have the proposed value on entry.
 * On exit, they may be updated to make them valid.
 *
 * The start position and size are relative to the sensor's area; they have already been
 * adjusted for license restrictions. Except for license restrictions, they are what the
 * user would see as the offset and size. If there are binning or subsampling options
 * in use, the ROI fields have to be multiplied by the appropriate factor.
 * We allow a sensor ROI larger than what the user can see if binning is in effect.
 *
 * Force FM6, aka GIGE_SENSOPT_INTERNAL_0, versus FM1 depending on (1) whether we are in LTR mode
 * and (2) whether the bit mode is greater than 8 bits. We keep it in the sensor options so the
 * GUIs have access to the flag for computing frame rates.
 *
 * @param rtn_startx        ROI column start
 * @param rtn_starty        ROI row start
 * @param rtn_width     ROI width
 * @param rtn_height        ROI height
 * @param rtn_bit_mode      Bit mode
 * @param rtn_frame_rate    Frame rate valid for this ROI
 * @param rtn_shutter_speed Shutter speed valid for that frame rate
 * @param rtn_sensor_opts   Sensor binning and subsampling options
 *
 * @return true if any changes were needed to make the capture parameters valid.
 */
bool Lince5M::make_sensor_valid(
    uint16_t *rtn_startx,
    uint16_t *rtn_starty,
    uint16_t *rtn_width,
    uint16_t *rtn_height,
    bitmode_t *rtn_bit_mode,
    float *rtn_frame_rate,
    uint32_t * /*rtn_shutter_speed*/,
    uint32_t *rtn_sensor_opts)
{
    uint16_t startx = *rtn_startx;
    uint16_t starty = *rtn_starty;
    uint16_t width = *rtn_width;
    uint16_t height = *rtn_height;
    uint32_t sensor_opts = *rtn_sensor_opts;
    bitmode_t bit_mode = *rtn_bit_mode;
    float max_frame_rate;
    uint16_t sensor_startx;
    uint16_t sensor_starty;
    uint16_t sensor_width;
    uint16_t sensor_height;
    uint16_t readout_startx;
    uint16_t readout_width;
    unsigned bin_factor;
    unsigned subs_factor;
    unsigned factor;
    uint32_t any_changes = 0;

    // We support all the bit modes so this should never happen.
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Invalid bit_mode, using 0x%02x", __FUNCTION__, GIGE_BITMODE_11_4);
        bit_mode = (bitmode_t)GIGE_BITMODE_11_4;
        any_changes |= (1 << 0);
    }
    // Don't leave the bit mode setting in one of the old-style values.
    switch (bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
        bit_mode = GIGE_BITMODE_7_0;
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
        bit_mode = GIGE_BITMODE_9_2;
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
        bit_mode = GIGE_BITMODE_11_4;
        break;
    case GIGE_BITMODE_10_BITS:
        bit_mode = GIGE_BITMODE_11_2;
        break;
    default:
        break;
    }

    // Turn off FM6 and use FM1. We don't support FM6 anymore.
    if ((sensor_opts & GIGE_SENSOPT_INTERNAL_0) != 0)
    {
        fi_log(LOG_DEBUG, "%s: Forcing FM1!", __PRETTY_FUNCTION__);
        sensor_opts &= ~GIGE_SENSOPT_INTERNAL_0;
        any_changes |= (1 << 3);
    }

    // Do the bandwidth-control tests and reduce the ROI if necessary.
    if (bandwidth_control(&width, &height, bit_mode))
    {
        any_changes |= (1 << 4);
    }

    // We treat binning and subsampling options of zero the same as 1, but we still correct them.
    factor = (sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT;
    if ((factor != 1) && (factor != 2) && (factor != 4))
        factor = 1;
    bin_factor = factor;

    factor = (sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT;
    if ((factor != 1) && (factor != 2) && (factor != 4))
        factor = 1;
    subs_factor = factor;

    // They are okay individually but we can have both binning and subsampling. Make sure the
    // product is acceptable. If not, it must be 4x4 or 2x4 or 4x2 so we will bump them both
    // back to 2.
    factor = bin_factor * subs_factor;
    if (factor > 4)
    {
        bin_factor = subs_factor = 2;
        factor = 4;
    }

    sensor_opts = (sensor_opts & ~(GIGE_SENSOPT_SUBS_MASK | GIGE_SENSOPT_BIN_MASK)) |
                  (subs_factor << GIGE_SENSOPT_SUBS_SHIFT) |
                  (bin_factor << GIGE_SENSOPT_BIN_SHIFT);
    LOG("%s: now sensor_opts=0x%08x", __FUNCTION__, sensor_opts);

    if (lince5m_user_to_sensor(factor, startx, starty, width, height, bit_mode,
                               &sensor_startx, &sensor_starty, &sensor_width, &sensor_height))
    {
        any_changes |= (1 << 5);
    }

    (void)force_legal_roi(factor, &sensor_startx, &sensor_starty, &sensor_width, &sensor_height);

    if (sensor_to_user(factor, sensor_startx, sensor_starty, sensor_width, sensor_height,
                       bit_mode, &startx, &starty, &width, &height))
    {
        any_changes |= (1 << 6);
    }

    if ((startx != *rtn_startx) ||
        (starty != *rtn_starty) ||
        (width  != *rtn_width) ||
        (height != *rtn_height) ||
        (bit_mode != *rtn_bit_mode) ||
        (sensor_opts != *rtn_sensor_opts))
    {
        //fi_log(LOG_DEBUG, "%s: was: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );
        *rtn_startx = startx;
        *rtn_starty = starty;
        *rtn_width = width;
        *rtn_height = height;
        *rtn_sensor_opts = sensor_opts;
        *rtn_bit_mode = bit_mode;
        //fi_log(LOG_DEBUG, "%s: now: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );

        any_changes |= (1 << 7);
    }

    // The maximum frame rate depends on the t_row configuration, which is based on the
    // width of the readout.
    lince5m_get_readout_dims(factor, sensor_startx, sensor_width, &readout_startx, &readout_width);
    max_frame_rate = (float)lince5m_get_max_frame_rate(readout_width / factor, sensor_height / factor, sensor_opts);
    if (*rtn_frame_rate > max_frame_rate)
    {
        *rtn_frame_rate = max_frame_rate;
        any_changes |= (1 << 8);
    }

    if (any_changes != 0)
    {
        //fi_log(LOG_DEBUG, "%s: changed 0x%08x", __FUNCTION__, any_changes );
        return true;
    }
    else
        return false;
}

/*!
 * @brief Get a buffer containing the defective pixel filename.
 *
 * @param dirname   the directory in which the file will be found
 *          e.g. /scratch or /opt/fastec/calibration
 *
 * @return  std::string containing fully-qualified file name.
 */
std::string Lince5M::defect_filename(const char *dirname)
{
    std::string fname_buf;

    fname_buf = fmt::sprintf(Sensor::DEFECT_MAP_FILE_FMT, get_home_path(), dirname,
                             eeprom.serial_no[0], eeprom.serial_no[1],
                             eeprom.serial_no[2], eeprom.serial_no[3]);

    return fname_buf;
}

uint32_t Lince5M::vtrig_control(const char *settings_str, int str_length)
{
    Json::Value requested_settings;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    bool parsingSuccessful;
    bool updateROI;
    uint32_t status;
    uint32_t reg_val;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

//    LOG("%s: settings [%s]", __PRETTY_FUNCTION__, settings_str);

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(settings_str, settings_str + str_length, &requested_settings, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, settings_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // Read any and all ROI settings.
    updateROI = false;
    if (!requested_settings["roi_x"].empty())
    {
        try
        {
            uint16_t offset_x;
            CameraServer *cs_ptr = CameraServer::get_cs();

            // Get some settings we'll need to figure the ROI out.
            cs_ptr->get_sensor_start(&offset_x, nullptr);

            // Get the value sent by the user.
            reg_val = requested_settings["roi_x"].asUInt();

            // Calculate the real x-coordinate.
            vtroi_x = reg_val + offset_x;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_x].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_y"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_y"].asUInt();
            vtroi_y = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_y].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_width"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_width"].asUInt();
            vtroi_width = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_width].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_height"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_height"].asUInt();
            vtroi_height = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_height].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // Do we have width/height?
    if (updateROI)
    {
        // Convert the ROI into the FPGA settings.
        reg_val = ILVT_SET_ROI_H_CONFIG_A((vtroi_x % PIXELS_PER_PPW), (((vtroi_x + vtroi_width) % PIXELS_PER_PPW) + 1));
        fpga_ptr->regWrite(ICM_LINCE_VTRIG_ROI_H_CONFIG_A, reg_val);

        reg_val = ILVT_SET_ROI_H_CONFIG_B((vtroi_x / PIXELS_PER_PPW), ((vtroi_x + vtroi_width) / PIXELS_PER_PPW));
        fpga_ptr->regWrite(ICM_LINCE_VTRIG_ROI_H_CONFIG_B, reg_val);

        reg_val = ILVT_SET_ROI_V(vtroi_y, (vtroi_y + vtroi_height));
        fpga_ptr->regWrite(ICM_LINCE_VTRIG_ROI_V, reg_val);
    }

    // Do we have the threshold settings?
    if (!requested_settings["pix_thresh"].empty() || !requested_settings["area_thresh"].empty())
    {
        uint32_t pix_thresh;
        uint32_t area_thresh;

        // Initialize our thresholds.
        fpga_ptr->regRead(ICM_LINCE_VTRIG_THRESHOLDS, &reg_val);

        if (!requested_settings["pix_thresh"].empty())
        {
            try
            {
                pix_thresh = requested_settings["pix_thresh"].asUInt();
            }
            catch (Json::Exception& e)
            {
                fi_log(LOG_DEBUG, "%s: requested_settings[pix_thresh].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
                pix_thresh = 10;
            }

            if ((pix_thresh == 0) || (pix_thresh > 100))
            {
                fi_log(LOG_DEBUG, "%s: Illegal pix_thresh [%u]!", __PRETTY_FUNCTION__, pix_thresh);
                pix_thresh = (reg_val & ILVT_CHANGE_THRESHOLD_MASK);
            }
            else
            {
                // Calculate the actual threshold value.
                pix_thresh = (255 * pix_thresh) / 100;
            }
        }
        else
        {
            // Just reuse the existing value.
            pix_thresh = (reg_val & ILVT_CHANGE_THRESHOLD_MASK);
        }

        if (!requested_settings["area_thresh"].empty())
        {
            try
            {
                area_thresh = requested_settings["area_thresh"].asUInt();
            }
            catch (Json::Exception& e)
            {
                fi_log(LOG_DEBUG, "%s: requested_settings[area_thresh].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
                area_thresh = 10;
            }

            if ((area_thresh == 0) || (area_thresh > 100))
            {
                fi_log(LOG_DEBUG, "%s: Illegal pix_thresh [%u]!", __PRETTY_FUNCTION__, pix_thresh);
                area_thresh = (reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT;
            }
            else
            {
                // Calculate the actual area value.
                area_thresh = (vtroi_width * vtroi_height * area_thresh) / 100;
            }
        }
        else
        {
            // Just reuse the existing value.
            area_thresh = (reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT;
        }

        reg_val = ILVT_SET_THRESHOLD(pix_thresh, area_thresh);
        fpga_ptr->regWrite(ICM_LINCE_VTRIG_THRESHOLDS, reg_val);
    }

    // Do we have the interval settings?
    if (!requested_settings["interval"].empty())
    {
        try
        {
            reg_val = requested_settings["interval"].asUInt();
            fpga_ptr->regWrite(ICM_LINCE_VTRIG_ROI_BUF_UPDATE_INTERVAL, reg_val);
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[interval].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // Is the control setting available?
    if (!requested_settings["control"].empty())
    {
        try
        {
            reg_val = requested_settings["control"].asUInt();
            fpga_ptr->regWrite(ICM_LINCE_VTRIG_CONTROL, reg_val);
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[control].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // To latch these values, we need to write to the status register.
    fpga_ptr->regWrite(ICM_LINCE_VTRIG_STATUS, 0);

    // All done.
    status = GEV_STATUS_SUCCESS;

bail:
    return status;
}

uint32_t Lince5M::vtrig_info_json(char *dest_ptr, uint32_t max_len)
{
    Json::Value vtrig_json;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;
    uint32_t reg_val;
    uint32_t config_a;
    uint32_t config_b;
    int x;
    int width;
    int height;
    int thresh;
    uint16_t offset_x;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    // Get some settings we'll need to figure the ROI out.
    cs_ptr->get_sensor_start(&offset_x, nullptr);

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // Read the register settings and populate the JSon response.
    fpga_ptr->regRead(ICM_LINCE_VTRIG_ROI_H_CONFIG_A, &config_a);
    vtrig_json["roi_ha"] = config_a;
    fpga_ptr->regRead(ICM_LINCE_VTRIG_ROI_H_CONFIG_B, &config_b);
    vtrig_json["roi_hb"] = config_b;
    x = ((config_b & ILVT_FIRST_ACTIVE_PPW_MASK) * PIXELS_PER_PPW) + (config_a & ILVT_FIRST_PPW_START_MASK);
    width = (((config_b >> 8) * PIXELS_PER_PPW) + ((config_a >> 4) - 1)) - x;
    vtrig_json["roi_x"] = x - offset_x;
    vtrig_json["roi_width"] = width;

    fpga_ptr->regRead(ICM_LINCE_VTRIG_ROI_V, &reg_val);
    vtrig_json["roi_v"] = reg_val;
    vtrig_json["roi_y"] = (reg_val & ILVT_FIRST_ROW_MASK);
    height = (reg_val >> ILVT_LAST_ROW_SHIFT) - (reg_val & ILVT_FIRST_ROW_MASK);
    vtrig_json["roi_height"] = height;

    fpga_ptr->regRead(ICM_LINCE_VTRIG_THRESHOLDS, &reg_val);
    vtrig_json["threshold"] = reg_val;
    thresh = (int)((((float)(reg_val & ILVT_CHANGE_THRESHOLD_MASK) * 100.0) / 255.0) + 0.5);
    if (thresh <= 0)
        thresh = 1;
    vtrig_json["pix_thresh"] = thresh;
    thresh = (int)((((float)((reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT) * 100.0) / (float)(width * height)) + 0.5);
    if (thresh <= 0)
        thresh = 1;
    vtrig_json["area_thresh"] = thresh;

    fpga_ptr->regRead(ICM_LINCE_VTRIG_ROI_BUF_UPDATE_INTERVAL, &reg_val);
    vtrig_json["interval"] = reg_val;
    fpga_ptr->regRead(ICM_LINCE_VTRIG_CONTROL, &reg_val);
    vtrig_json["control"] = reg_val;
    fpga_ptr->regRead(ICM_LINCE_VTRIG_STATUS, &reg_val);
    vtrig_json["status"] = reg_val;

    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 2;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, vtrig_json);

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
        return 0;
    }

    // Copy the string over.
//    LOG("%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
    memcpy(dest_ptr, json_message.c_str(), json_message.length());

    // Return the length, including the null-terminator.
    return json_message.length() + 1;
}

int Lince5M::get_capture_fifo_count()
{
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fpga_ptr->regRead(ICM_LINCE_STATUS, &reg_value);
    return ((reg_value & ILS_CAPSTART_FIFO_LEVEL_MASK) >> ILS_CAPSTART_FIFO_LEVEL_SHIFT);
}

void Lince5M::reset_capture_fifo()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Toggle the FIFO reset bit.
    fpga_ptr->regModify(ICM_LINCE_CONTROL, ILC_CAPSTART_FIFO_RESET, -1);
    fpga_ptr->regModify(ICM_LINCE_CONTROL, 0, ~ILC_CAPSTART_FIFO_RESET);
}

void Lince5M::reset_event_fifo()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Toggle the FIFO reset bit.
    fpga_ptr->regModify(ICM_LINCE_CONTROL, ILC_EVENT_FIFO_RESET, -1);
    fpga_ptr->regModify(ICM_LINCE_CONTROL, 0, ~ILC_EVENT_FIFO_RESET);
}
