/*!
 * @file lince5m_fpn.c
 * @brief Functions to support FPN correction within the FPGA.
 *
 * Copyright 2012,2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <math.h>

#include <arpa/inet.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <fastec/fi_ioctl.h>
#include <fastec/lince5m_param.h>
#include <fastec/lince5m_utils.h>
#include <fastec/fas_log.h>

#include <memory>
#include <string>
#include <fmt/printf.h>

#include "fcp.h"
#include "image_correction.h"
#include "lince5m.h"
#include "part_mgr.h"
#include "storage.h"
#include "threads.h"
#include "ct_file.h"

namespace LINCE5M_SENSOR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// In case we save an average file for test or debug.
const char FULL_FRAME_FILE[] = ".config/Fastec/scratch/black_2560x2048";

const char FACTORY_DEFECT_SCRIPT[] = "cp {0} ~/.config/Fastec/calibration/.";

// Indices of Lince sensor-specific parameters in version 2 calibration file header.
const int LINCE_OPTS_IDX  = 0;  // Sensor options used to capture the frame, 4 bytes, aligned.
const int LINCE_OFS_IDX   = 4;  // The offset to use with the black frame, 2 bytes, aligned.
const int LINCE_GAIN_IDX  = 6;  // The sensor gain used to capture the black frame.
const int LINCE_TROW_IDX  = 7;  // The trow configuration when we captured the black frame.

// How many frames do we want to grab to create a black reference for FPN?
const int FPN_CAPTURE_FRAMES = 20;

// How many frames do we want to grab for factory "calibration" to find pixel defects?
const int FACTORY_CAPTURE_FRAMES = 5;

// We are capturing 8-bit data so each value is 0-255.
// The histogram has one slot for each possible value.
const int HISTOGRAM_ENTRIES = 256;

// Threshold for rejecting pixels as "white spot defects" when capturing upper 8 bits.
// This is 1024 on a scale of 0 to 4095, full-scale at 12 bits.
const int WHITESPOT_THRESHOLD_8BIT = 0x40;

// Exposure for capturing black frames to locate white spots (microseconds).
const int WHITESPOT_SHUTTER = 16000;

// Frame rate for capturing black frames to locate white spots.
const int WHITESPOT_FPS = 30;

//! Structure to hold coordinates of a defective pixel.
typedef struct DEFECT_ENTRY
{
    uint16_t x;
    uint16_t y;
} DEFECT_ENTRY;

/*!
 * @brief The class declaration for Lince5M FPN calibration function.
 */
class Lince5MCalibration
{
private:

    CT_GT_FILE::CT_ROI  ct_roi;             //!< The user ROI used for this calibration capture.
    uint32_t    sensor_opts;    //!< Sensor bin/subsample options for user calibration.
    uint32_t    sensor_gain;    //!< Sensor gain used during calibration.
    unsigned    capture_trow;   //!< Sensor trow configuration used during calibration.
    bool        simulate_gain;  //!< Simulate use of gain for trow=84 case.
    uint64_t    cal_frame_addr;
    uint32_t    cal_frame_size;

    std::unique_ptr<uint32_t[]> hist_table; //!< Pointer to a histogram buffer
    std::unique_ptr<uint8_t[]> raw_table;       //!< A DMA-aligned pointer to the raw data.

    void        getUserRoi(CT_GT_FILE::CT_ROI *roi);
    unsigned    computePixelAverage(unsigned *rtn_std_dev);
    void        saveHistogram(const char *filename, uint32_t arg);
    bool        fpnCalc(const char *filename);
    uint64_t    frameAddress(int frame_num);
    int         frameCapture();
    int         prnuCalc(const uint8_t *ct_ptr);
    bool        saveCapture(const char *filename, unsigned num_frames);
    bool        createBadPixelMap();

public:
    Lince5MCalibration();
    ~Lince5MCalibration();

    bool captureFullSizeFrames(bool manual_params);
    bool runFpnCalibration();
    bool runPrnuCalibration(const uint8_t *ct_ptr);
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

// Disable factory calibration while it is work-in-progress.
static bool allow_factory = true;

} /* End of namespace LINCE5M_SENSOR */

using namespace LINCE5M_SENSOR;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor.
 */
Lince5MCalibration::Lince5MCalibration() :
    capture_trow(0),
    simulate_gain(0),
    cal_frame_addr(0),
    cal_frame_size(0)
{
    // Initialize our member variables.
    sensor_opts = 0;
    sensor_gain = 1;
}

Lince5MCalibration::~Lince5MCalibration()
{
}


/*!
 * @brief Write out the histogram
 */
void Lince5MCalibration::saveHistogram(const char *filename, uint32_t arg)
{
    FILE *fp;

    fp = fopen(filename, "w");
    if (fp == nullptr)
        fi_log(LOG_ERR, "%s: can't open '%s': (%d) %s", __FUNCTION__, filename, errno,
               strerror(errno));
    else
    {
        unsigned i;

        fprintf(fp, "offset,%u\n", arg);
        for (i = 0; i < HISTOGRAM_ENTRIES; ++i)
            fprintf(fp, "%u,%d\n", i, hist_table[i]);
        fclose(fp);
    }
}

/*!
 * @brief Compute the average of the pixels in the image from the histogram.
 *
 * @return average pixel value if successful, <0 if histogram is empty which shouldn't happen.
 */
unsigned Lince5MCalibration::computePixelAverage(unsigned *rtn_std_dev)
{
    uint64_t sum;
    uint64_t sum_square;
    uint32_t count;
    uint32_t index;

    count = 0;
    sum = 0;
    sum_square = 0;
    for (index = 0; index <  HISTOGRAM_ENTRIES; index++)
    {
        sum += (hist_table[index] * index);
        count += hist_table[index];
        sum_square += ((uint64_t)hist_table[index] * (uint64_t)index * (uint64_t)index);
    }
    if (count == 0)
    {
        fi_log(LOG_ERR, "%s: Got count=zero sum=%" PRIu64 ".", __FUNCTION__, sum);
        return -1;
    }
    else
    {
        int      avg;
        double   f_N, f_mean, f_s;

        f_N = (double)count;
        f_mean = (double)sum / f_N;
        f_s = sqrt(((double)sum_square / f_N) - (f_mean * f_mean));
        LOG("%s: got pixel mean=%.1f std_dev=%.1f on %u pixels", __FUNCTION__,
            f_mean, f_s, count);

        avg = (unsigned)(f_mean + 0.5);
        *rtn_std_dev = (unsigned)(f_s + 0.5);

        return avg;
    }
}

/*!
 * @brief Read the current ROI from GigE registers.
 *
 * @param roi   Pointer to where we want to put the ROI values.
 */
void Lince5MCalibration::getUserRoi(CT_GT_FILE::CT_ROI *roi)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint16_t  ofs_x;
    uint16_t  ofs_y;
    uint16_t  width;
    uint16_t  height;
    bitmode_t bit_mode;
    uint32_t  sensor_opts;

    cs_ptr->get_user_roi(&ofs_x, &ofs_y, &width, &height);
    cs_ptr->get_user_bit_mode(&bit_mode);
    cs_ptr->get_sensor_opts(&sensor_opts);

    roi->origin.x = ofs_x;
    roi->origin.y = ofs_y;
    roi->width    = width;
    roi->height   = height;

    // If the bit mode is 12 bits and no binning or subsampling is used,
    // we need to adjust the offset since we hide the region outside 1280x1024.
    // THIS IS OBSOLETE since we no longer use the bandwidth restrictions on ROI.
    lince5m_get_ct_offset(bit_mode, sensor_opts, &ofs_x, &ofs_y);
    roi->origin.x += ofs_x;
    roi->origin.y += ofs_y;
}



uint64_t Lince5MCalibration::frameAddress(int frame_num)
{
    return cal_frame_addr + ((uint64_t)frame_num * (uint64_t)cal_frame_size);
}

bool Lince5MCalibration::fpnCalc(const char *ct_file)
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    Sensor *sensor_ptr = CameraServer::get_sensor();
    bool    rc = false;

    uint32_t frame_idx;
    size_t  pixel_idx;
    int     file_fd;
    uint16_t *p_sum;
    uint8_t *p_pixel;
    CT_GT_FILE::CT_FILE_HDR ct_header;
    uint8_t     min_pixel;
    uint8_t     max_pixel;
    unsigned    mean_pixel;
    unsigned    std_dev;
    unsigned    multiplier;
    uint16_t    offset;
    uint16_t    threshold;
    unsigned    count;
    uint16_t    new_max;
    int16_t signed_offset;
    uint8_t *fb_ptr;
    std::unique_ptr<uint16_t[]> sum_table;

    LOG("%s", __PRETTY_FUNCTION__);

    // Allocate memory for our "raw" histogram.
    hist_table = std::unique_ptr<uint32_t[]>(new uint32_t[HISTOGRAM_ENTRIES]());
    if (hist_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: raw histogram creation failed!", __PRETTY_FUNCTION__);
        goto bail;
    }
    memset(hist_table.get(), 0, HISTOGRAM_ENTRIES * sizeof(uint32_t));

    // Allocate our frame buffer with alignment suitable for DMA from image memory.
    if ((errno = posix_memalign((void **)&fb_ptr, DMA_PAGE_ALIGN, cal_frame_size)) != 0)
    {
        fi_log(LOG_ERR, "%s: creation of frame buffer failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Assign the pointer to our smart pointer.
    raw_table.reset(fb_ptr);

    // Allocate our "summation" table.
    sum_table = std::unique_ptr<uint16_t[]>(new uint16_t[cal_frame_size]());
    if (sum_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
        goto bail;
    }
    memset(sum_table.get(), 0, cal_frame_size * sizeof(uint16_t));

    // Loop through our frames.
    for (frame_idx = 0; frame_idx < FPN_CAPTURE_FRAMES; ++frame_idx)
    {
        uint64_t frame_addr;
        uint32_t read_count;

        // Read the frame from DDR2. We use "little-endian" to have the image data byte-swapped
        // from the big-endian native mode of the FPGA.
        frame_addr = frameAddress(frame_idx);
        read_count = fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, frame_addr, raw_table.get(), cal_frame_size);
        if (read_count != cal_frame_size)
        {
            fi_log(LOG_ERR, "%s: Can't read the raw frame at 0x%" PRIx64 " (0x%x!=0x%x).",
                   __PRETTY_FUNCTION__, frame_addr, read_count, cal_frame_size);
            goto bail;
        }

        // Add this frame into the summation.
        p_pixel = raw_table.get();
        p_sum = sum_table.get();
        for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
        {
            // Add in the new pixel value.
            *p_sum += (uint16_t) * p_pixel;

            // Bump our pointers.
            ++p_pixel;
            ++p_sum;
        }
    }

    // Now generate the average table from the summation table.
    p_pixel = raw_table.get();
    p_sum = sum_table.get();
    for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
    {
        // Calculate and save the average for each pixel.
        *p_pixel = (uint8_t)(*p_sum / FPN_CAPTURE_FRAMES);

        // Bump our pointers.
        ++p_pixel;
        ++p_sum;
    }

    // Create a histogram with the averaged frame and compute stats.
    p_pixel = raw_table.get();
    min_pixel = 0xff;
    max_pixel = 0x00;
    for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
    {
        if (*p_pixel < min_pixel)
            min_pixel = *p_pixel;
        if (*p_pixel > max_pixel)
            max_pixel = *p_pixel;

        // Update the histogram for this pixel's value.
        hist_table[*p_pixel]++;
        ++p_pixel;
    }
    mean_pixel = computePixelAverage(&std_dev);
    LOG("%s: gain=%u mean=%u (%u-%u) std_dev=%u %ux%u@%u,%u", __FUNCTION__,
        sensor_gain, mean_pixel, min_pixel, max_pixel, std_dev,
        ct_roi.width, ct_roi.height, ct_roi.origin.x, ct_roi.origin.y);

    // Assuming a normal distribution, which seems reasonable and consistent with what
    // we have seen of the sensor's behavior, the percentage of the pixels that are lower
    // than N standard deviations below the mean, or higher than that above the mean, is:
    // 2 standard deviations ~ 2.3%, 2.5 std devs ~ 0.6%, 3 std devs ~ .14%
    // We could reasonably fail to correct that many pixels without noticeable problems.
    //
    // With trow=84 or gain > 1, the black frame we captured can have pixel values above 255,
    // so we had to use higher bits (8:1 or 9:2 instead of 7:0). Then we still have to convert
    // that black frame into something that, when subtracted, will remove the per-pixel fixed
    // pattern noise. That works as long as the values we subtract fit in the low eight bits,
    // which just means that the range of say 3 standard deviations below and above the mean
    // fits in eight bits.
    //
    // If the smallest pixel value is bigger than N standard deviations below the mean,
    // use the smallest pixel value. Otherwise, use ( mean - N * std_dev ).
    threshold = (5  * std_dev) / 2;
    if ((mean_pixel > threshold) && (min_pixel < (mean_pixel - threshold)))
        offset = mean_pixel - threshold;
    else
        offset = min_pixel;

    // We'll clip at the high end too so it's not so scary when we try to offset the values.
    new_max = mean_pixel - offset + threshold;

    // Subtract that offset from all the corrections. Then the smallest value in the correction
    // table will be zero-ish and the max we care about will be about 2 * 'threshold', with the mean
    // of this correction table at 'threshold'. If we subtracted this correction table from another
    // black frame, the mean of the corrected black frame would be 'offset'.
    p_pixel = raw_table.get();
    count = 0;
    for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
    {
        if (*p_pixel > offset)
            *p_pixel -= offset;
        else
        {
            *p_pixel = 0;
            ++count;
        }

        ++p_pixel;
    }

    // It is unexpected to clip more than 2% of the corrections.
    if (count > ((cal_frame_size * 2) / 100))
    {
        fi_log(LOG_WARNING, "%s: clipped %u pixel corrections to 0 (%5.2f pct)", __FUNCTION__,
               count, (100.0 * (double)count) / (double)cal_frame_size);
    }

    // If the gain is bigger than 1, we must scale the corrections back to match the low 8 bit
    // range since what we captured was higher-order bits. The corrections are as small in value
    // as we could make them, so we expect the product to fit in 8 bits.

    // Determine the multiplier based on our knowledge of what bits we captured (7:0, 8:1, 9:2).
    multiplier = sensor_gain;
    if (simulate_gain && (sensor_gain < 4))
        multiplier *= 2;

    // Sanity check in case the corrections don't fit as expected after our scale and offset.
    if ((new_max * multiplier) > 255)
    {
        fi_log(LOG_WARNING, "%s: Corrections too big: new_max(%u)*multiplier(%u)",
               __FUNCTION__, new_max, multiplier);
    }

    // Multiply each correction to adjust for using upper bits.
    count = 0;
    p_pixel = raw_table.get();
    for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
    {
        uint16_t value;

        value = *p_pixel;

        // Clip the high end so the corrections are below new_max.
        if (value > new_max)
        {
            ++count;
            value = new_max;
        }

        value *= multiplier;
        if (value > 255)
            value = 255;

        *p_pixel++ = (uint8_t)value;
    }

    // Warn if we clipped more than 2% of the pixels.
    if (count > ((2 * cal_frame_size) / 100))
    {
        fi_log(LOG_WARNING, "%s: clipped %u pixel corrections to %u*%u (%5.2f pct)", __FUNCTION__,
               count, multiplier, new_max, (100.0 * (double)count) / (double)cal_frame_size);
    }

    // A FPN-corrected black frame's mean is now "offset" (or offset*multiplier),
    // not zero and not our original mean_pixel via ( mean - ( mean - offset )).
    // We could use the offset register to add enough to bring the mean of our post-correction
    // black frame to the same as the mean of our uncorrected black frame (or a frame with column
    // FPN); that way the average saturation levels would be about the same as for the uncorrected
    // black frame or a column FPN frame, maybe avoiding some artifacts at the high end.
    // (It is not PRNU, not a difference in pixel response, it is just moving the saturation point
    // because of the FPN subtraction.)
    // But let's try to get the mean black level to an arbitrary but pleasing target.
    // The offset could be either positive or negative, but all we will do with it is
    // write it into a hardware register where it will be interpreted as twos-complement.
    // In practice it is negative because our black frames are so bright.
    signed_offset = 48 - (multiplier * offset);

    // An offset of zero means do it old-style using the correction table backoff value.
    // Fudge the value so it's not zero.
    if (signed_offset == 0)
        signed_offset = 1;
    LOG("%s: offset=%d=0x%x threshold=%u multiplier=%u", __FUNCTION__,
        signed_offset, offset, threshold, multiplier);

    // FPGA doesn't do signed offsets! See ticket #1668.
    if (signed_offset < 0)
    {
        uint16_t addend = -signed_offset;
        uint16_t xmax = new_max * multiplier;

        if (xmax + addend > 255)
        {
            if (xmax <= 255)
            {
                LOG("%s: xmax(%u)+addend(%u) >255, use addend=%u", __FUNCTION__,
                    new_max, addend, 255 - xmax);
                addend = 255 - xmax;
            }
            else
            {
                // This shouldn't happen.
                fi_log(LOG_WARNING, "%s: New max=%u too big!", __FUNCTION__, xmax);
                addend = 0;
            }
        }
        // Add that offset to the table entries, making each subtraction bigger therefore making
        // the resulting corrected black level lower.
        p_pixel = raw_table.get();
        for (pixel_idx = 0; pixel_idx < cal_frame_size; ++pixel_idx)
        {
            uint16_t value;

            value = *p_pixel;
            value += addend;
            *p_pixel++ = (uint8_t)value;
        }
        LOG("%s: Faking negative, set offset=1, added %u", __FUNCTION__, addend);
        signed_offset = 1;
    }

    // Prep the info structure.
    memset(&ct_header, 0, sizeof(ct_header));

    // Store the CT capture parameters we need to use this CT or to check whether it
    // is compatible.
    *((uint32_t *)((uint8_t *)ct_header.sensor_params + LINCE_OPTS_IDX)) = sensor_opts;
    *((uint16_t *)((uint8_t *)ct_header.sensor_params + LINCE_OFS_IDX)) = signed_offset;
    *((uint8_t *)ct_header.sensor_params + LINCE_GAIN_IDX) = sensor_gain;
    *((uint8_t *)ct_header.sensor_params + LINCE_TROW_IDX) = capture_trow;

    {
        std::string temp_path;

        temp_path = fmt::sprintf("%s/.config/Fastec/scratch/blacklevel_hist.csv", get_home_path());
        saveHistogram(temp_path.c_str(), mean_pixel);
    }

    // Setup our CT info structure.
    ct_header.magic = CT_GT_FILE::CT_MAGIC;
    ct_header.length = cal_frame_size;

    // Get the ROI used for this calibration.
    ct_header.roi = ct_roi;

    // Make sure the partition ROI reflects 32-bit alignment.
    ct_header.roi.width = (ct_header.roi.width + 3) & ~3;

    file_fd = open(ct_file, O_CREAT | O_TRUNC | O_RDWR, 0666);
    if (file_fd < 0)
    {
        fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, ct_file,
               strerror(errno));
        goto bail;
    }

    // Associate this correction table with the sensor.
    sensor_ptr->get_sensor_ver(ct_header.sensor_id, CT_GT_FILE::SENSOR_ID_LEN);

    // Save out the information.
    write(file_fd, &ct_header, sizeof(ct_header));
    write(file_fd, raw_table.get(), cal_frame_size);

    // Close the file.
    close(file_fd);

    // Write only the correction table to image memory. If we are successful, we succeeded.
    if (part_mgr->writeCorrectionTable(raw_table.get(), cal_frame_size))
        rc = true;

bail:
    return rc;
}

/*!
 * @brief Perform the PRNU calibration.
 * @param ct_ptr    [IN] Pointer to the FPN correction table in use.
 * @return 0 on success, <0 otherwise.
 */
int Lince5MCalibration::prnuCalc(const uint8_t *ct_ptr)
{
    int     file_fd;
    int     retval;
    uint16_t    *avg_ptr;
    uint32_t    framesize;          // Size of the raw frame, in pixels.
    uint32_t    frame_bytecount;    // Size of the raw frame, in bytes.
    std::string prnu_file;
    CT_GT_FILE::GT_FILE_HDR gt_header;
    std::unique_ptr<uint16_t[]> avg_table;
    std::unique_ptr<uint8_t[]> gain_table;

    PartMgr *part_mgr = CameraServer::get_partmgr();
    Lince5M *sensor_ptr = static_cast<Lince5M *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    // Assume failure.
    retval = -1;

    // Prep the info structure.
    memset(&gt_header, 0, sizeof(gt_header));

    // Since we are dealing with 16-bit values, we don't have to worry about row alignment.
    framesize = ct_roi.width * ct_roi.height;
    frame_bytecount = framesize * 2;
    LOG("%s: frame_bytecount=%d", __PRETTY_FUNCTION__, frame_bytecount);

    // Allocate our "averages" table.
    avg_table = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (avg_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of averages buffer failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Allocate our "gain" table.
    gain_table = std::unique_ptr<uint8_t[]>(new uint8_t[framesize]());
    if (gain_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of averages buffer failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Are we a color sensor?
    if (sensor_ptr->is_sensor_color())
    {
        // Have the FPGA build up our color image. This will use the red, green, and blue
        // raw TIFF images that have previously been captured.
        if (!FiImageCorrection::prnuColorSum(FPN_CAPTURE_FRAMES, avg_table.get()))
        {
            fi_log(LOG_DEBUG, "%s: prnu_color_sum() failed!", __PRETTY_FUNCTION__);
            goto fpn_exit;
        }
    }
    else
    {
        int pixel_idx;
        uint32_t *raw_sum_ptr;
        std::unique_ptr<uint32_t[]> raw_sum_table;

        // Allocate our "summation" table.
        raw_sum_table = std::unique_ptr<uint32_t[]>(new uint32_t[framesize]());
        if (raw_sum_table.get() == nullptr)
        {
            fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
            goto fpn_exit;
        }
        memset(raw_sum_table.get(), 0, sizeof(raw_sum_table[0]) * framesize);

        // Download the capture images and sum them together.
        FiImageCorrection::prnuMonoSum(FPN_CAPTURE_FRAMES, GIGE_BITMODE_11_0, cal_frame_addr, raw_sum_table.get());

        // Now run through our summation table and generate our average table.
        avg_ptr = avg_table.get();
        raw_sum_ptr = raw_sum_table.get();
        for (pixel_idx = 0; pixel_idx < (int32_t)framesize; ++pixel_idx)
        {
            uint32_t avg_value;

            // Calculate and save the average. Make sure we saturate rather than overflow.
            avg_value = (*raw_sum_ptr / FPN_CAPTURE_FRAMES);
            if (avg_value > 0xFFFF)
                avg_value = 0xFFFF;
            *avg_ptr = (uint16_t)(avg_value & 0xFFFF);

            // Bump our pointers.
            ++avg_ptr;
            ++raw_sum_ptr;
        }
    }

    // ------------------------------------------------------------------------
#ifdef DEBUG_DUMP_FILES
    // I'm cranky with the formatter not indenting properly. Using a do-while false is equivalent to empty braces, so...
    do
    {
        std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), PRNU_FRAME_FILE);
        if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
        {
            fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
        }

        // Open up the black reference file.
        std::string tmp_file = fmt::sprintf("%s/cal_avg", tmp_dir);
        int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
        }
        else
        {
            // Save out the information.
            write(fd, avg_table.get(), frame_bytecount);
            close(fd);
        }
    } while (0);
#endif /* DEBUG_DUMP_FILES */
    // ------------------------------------------------------------------------

    // Using the averaged white frame, calculate what the gain table should be.
    if (!FiImageCorrection::prnuGenerate(avg_table.get(), ct_ptr, gain_table.get()))
    {
        fi_log(LOG_DEBUG, "%s: FiImageCorrection::prnuGenerate() failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Initialze our file header.
    memset(&gt_header, 0, sizeof(gt_header));
    gt_header.magic = CT_GT_FILE::CT_MAGIC;
    gt_header.length = ct_roi.width * ct_roi.height;
    gt_header.sensor_params[0] = CT_GT_FILE::FIXED_FRACTIONAL;
    gt_header.roi.width = ct_roi.width;
    gt_header.roi.height = ct_roi.height;

    // Associate this correction table with the sensor.
    sensor_ptr->get_sensor_ver(gt_header.sensor_id, CT_GT_FILE::SENSOR_ID_LEN);

    // Get the file name we are writing.
    prnu_file = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_GAIN_FILE_RO);

    // Open up the black reference file.
    file_fd = open(prnu_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
    if (file_fd < 0)
    {
        fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, prnu_file.c_str(), strerror(errno));
        goto fpn_exit;
    }
    else
    {
        // Save out the information.
        write(file_fd, &gt_header, sizeof(gt_header));
        write(file_fd, gain_table.get(), framesize);
    }

    // Close the file.
    close(file_fd);

    // If we write it to image memory successfully, we succeeded.
    if (part_mgr->writeGainTable(gain_table.get(), framesize))
        retval = 0;

fpn_exit:
    // Read up the frame from the image memory.
    return retval;
}

/*!
 * @brief Do the calibration work for pixel FPN calibration.
 */
bool Lince5MCalibration::runFpnCalibration()
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA       *fpga_ptr = CameraServer::get_fpga();
    PartMgr      *part_mgr = CameraServer::get_partmgr();
    Sensor       *sensor_ptr = CameraServer::get_sensor();

    const char  *filename;
    uint32_t    total_size;
    uint32_t    buf_count;
    bitmode_t   cal_bit_mode;
    float frame_rate;
    bool rc = false;

    LOG("%s", __PRETTY_FUNCTION__);

    // User correction tables use the current ROI.
    getUserRoi(&ct_roi);

    // Save options so we can store them in the CT file.
    cs_ptr->get_sensor_opts(&sensor_opts);

    // The canonical gain is 100x the gain value we tell the sensor.
    // Divide it out so it will fit in 8 bits.
    cs_ptr->get_sensor_gain(&sensor_gain);
    sensor_gain /= 100;
    cs_ptr->get_frame_rate(&frame_rate);
    capture_trow = lince5m_get_trow_from_user(ct_roi.origin.x, ct_roi.origin.y, ct_roi.width,
                                              ct_roi.height, GIGE_BITMODE_7_0, sensor_opts,
                                              frame_rate);

    // For trow=84 but not for trow=52, we need to simulate using gain
    // because the black levels are so high.
    simulate_gain = false;
    if (capture_trow == TROW_84)
        simulate_gain = true;

    if ((sensor_gain == 4) || ((sensor_gain == 2) && simulate_gain))
        cal_bit_mode = (bitmode_t)GIGE_BITMODE_9_2;
    else if ((sensor_gain == 2) || ((sensor_gain == 1) && simulate_gain))
        cal_bit_mode = (bitmode_t)GIGE_BITMODE_8_1;
    else /* sensor_gain == 1 */
        cal_bit_mode = (bitmode_t)GIGE_BITMODE_7_0;

    // Active partition is zero, since we don't calibrate with any captures in memory.
    part_mgr->setActivePartition(0);
    part_mgr->queueCapturePartition(0);

    // Set up capture parameters for calibration. This will only be writing to
    // the Unarmed Live View buffers, so the capture buffers don't need to be set yet.
    // The size of the ULV buffers is based on the maximum sensor area without regard
    // to license so we can still capture full-size images even if the license limits
    // us to a smaller size.
    // Pass 0 for frame rate and shutter speed when we don't need to override them.
    cs_ptr->start_lince_calibration(false, cal_bit_mode, (fpnmode_t)GIGE_FPN_SENSOR_ENABLE_MASK, 0, 0);

    // Set up the memory buffer and size that we will capture black frames into.
    part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);

    // The size we got from PM matches the camera settings but it might be adjusted for
    // license restrictions and our calibration captures are always low 8 bits.
    // Recompute the number of buffers that will fit in that area.
    total_size = cal_frame_size * buf_count;
    cal_frame_size = cs_ptr->calc_frame_buf_size(ct_roi.width, ct_roi.height, cal_bit_mode);
    cal_frame_size += FRAME_METADATA_HW_BYTES;
    buf_count = total_size / cal_frame_size;
    LOG("%s: %ux%u@%u,%u framesize=%u QW", __PRETTY_FUNCTION__, ct_roi.width,
        ct_roi.height, ct_roi.origin.x, ct_roi.origin.y, cal_frame_size);

    if (buf_count < FPN_CAPTURE_FRAMES)
        goto cal_exit;
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
    fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

    // Close the shutter.
    fpga_ptr->shutterClose(true);

    // Capture FPN_CAPTURE frames regardless of the currently-selected mode.
    if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
        goto cal_exit;

    // Create the user black reference file.
    filename = sensor_ptr->get_blkref_name();
    if (filename == nullptr)
    {
        std::string temp_path;

        fi_log(LOG_ERR, "%s: oops, got bad filename", __FUNCTION__);
        temp_path = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::USER_BLKREF_FILE);

        // Calculate the FPN correction table.
        if (!fpnCalc(temp_path.c_str()))
            goto cal_exit;
    }
    else
    {
        // Calculate the FPN correction table.
        if (!fpnCalc(filename))
            goto cal_exit;
    }

    rc = true;

cal_exit:
    // Make sure the shutter is open.
    fpga_ptr->shutterClose(false);

    return rc;
}


/*!
 * @brief Do the PRNU calibration work.
 * @param ct_ptr    [IN] The user black reference table to use for PRNU.
 * @returns True on success, false otherwise.
 */
bool Lince5MCalibration::runPrnuCalibration(const uint8_t *ct_ptr)
{
    bool rc = false;
    uint32_t    total_size;
    uint32_t    buf_count;
    uint16_t    width;
    uint16_t    height;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    CameraServer *cs_ptr = CameraServer::get_cs();
    Lince5M *sensor_ptr = static_cast<Lince5M *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (sensor_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sensor_ptr is NULL!? Is this a Lince5M sensor?", __PRETTY_FUNCTION__);
        goto cal_exit;
    }

    // If we are doing a PRNU calibration, we want to make sure the settings are
    // how we need them.
    width = sensor_ptr->get_sensor_max_width();
    height = sensor_ptr->get_sensor_max_height();
    ct_roi.origin.x = 0;
    ct_roi.origin.y = 0;
    ct_roi.width = width;
    ct_roi.height = height;

    // Only need to capture images for mono sensors.
    if (!sensor_ptr->is_sensor_color())
    {
        // Active partition is zero, since we don't calibrate with any captures in memory.
        part_mgr->setActivePartition(0);
        part_mgr->queueCapturePartition(0);

        // Set up capture parameters for PRNU calibration. This will only be writing to
        // the calibration buffers, so the capture buffers don't need to be set yet.
        // The size of the buffers is based on the maximum sensor area without regard
        // to license so we can still capture full-size images even if the license limits
        // us to a smaller size.
        //
        // For PRNU calibration, we enable the sensor, but not pixel, FPN.
        sensor_ptr->start_prnu_capture();

        // Set up the memory buffer and size that we will capture white frames into.
        part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);

        // Figure out the total size we have for capturing.
        total_size = cal_frame_size * buf_count;

        // The size we got from PM matches the camera settings but it might be adjusted for
        // license restrictions. Factory is always full size so we need to recompute the size.
        cal_frame_size = cs_ptr->calc_frame_buf_size(width, height, (bitmode_t)GIGE_BITMODE_11_0);
        cal_frame_size = convert_to_HEXWORD(cal_frame_size + FRAME_METADATA_HW_BYTES) << HEXWORD_SHIFT;

        // Now figure out how many buffers we can actually fit.
        buf_count = total_size / cal_frame_size;

        // Make sure we have enough space for the buffers we need to capture. It's
        // OK if we have more buffers than we need.
        if (buf_count < FPN_CAPTURE_FRAMES)
        {
            fi_log(LOG_DEBUG, "%s: Not enough space for capture buffers [buf_count=%d]?!", __PRETTY_FUNCTION__, buf_count);
            goto cal_exit;
        }
        fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
        fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
        fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

        // Capture FPN_CAPTURE frames regardless of the currently-selected mode.
        if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
        {
            fi_log(LOG_ERR, "%s: calibrate_capture() failed!", __PRETTY_FUNCTION__);
            goto cal_exit;
        }
    }

    // Calculate the PRNU table. This will create the FACTORY_GAIN_FILE_RO file.
    if (prnuCalc(ct_ptr) < 0)
        goto cal_exit;

    // All is well.
    rc = true;

cal_exit:
    // All done.
    return rc;
}

/*!
 * @brief Write out the defect map.
 *
 * @param num_defects   the number of defects found
 * @param defect_buf    pointer to defect list
 *
 * @return true on success, false if failure
 */
static bool write_defect_file(unsigned num_defects, DEFECT_ENTRY *defect_buf)
{
    std::string fname_buf;
    FILE *fd = nullptr;
    bool  rc = false;
    Lince5M *sensor_ptr = static_cast<Lince5M *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    fname_buf = sensor_ptr->defect_filename(".config/Fastec/scratch");
    if (fname_buf.empty())
        return false;

    fd = fopen(fname_buf.c_str(), "w");
    if (fd == nullptr)
    {
        fi_log(LOG_ERR, "%s: Can't open %s (%s)", __FUNCTION__, fname_buf.c_str(), strerror(errno));
        goto bail;
    }

    fprintf(fd, LINCE5M_DEFECT_HEADER, num_defects);
    for (unsigned i = 0; i < num_defects; ++i)
    {
        if (fprintf(fd, "%u,%u\n", defect_buf[i].x, defect_buf[i].y) < 0)
        {
            fi_log(LOG_ERR, "%s: Write to %s failed (%s)", __FUNCTION__, fname_buf.c_str(),
                   strerror(errno));
            goto bail;
        }
    }

    fi_log(LOG_INFO, "%s: Wrote %u defects to '%s'", __FUNCTION__, num_defects, fname_buf.c_str());
    rc = true;

bail:
    if (fd != nullptr)
        fclose(fd);

    return rc;
}

/*!
 * @brief Create an initial bad pixel map using the captured frames
 *
 * @return true if successful, false otherwise.
 */
bool Lince5MCalibration::createBadPixelMap()
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();
    bool    rc = false;
    unsigned    num_frames = FACTORY_CAPTURE_FRAMES;
    FILE        *frame_fd = nullptr;
    unsigned    num_defects;
    uint32_t    lines_per_buf;
    uint32_t    cur_line_count;
    uint32_t    cur_line;
    uint32_t    cur_addr;
    uint32_t    cur_size;
    unsigned    i, frame;
    uint8_t *s_ptr;
    std::unique_ptr<DEFECT_ENTRY[]> defect_buf;
    std::unique_ptr<uint8_t[]> sample_buf;
    std::unique_ptr<uint16_t[]> sum_buf;
    std::string temp_path;

    LOG("%s", __PRETTY_FUNCTION__);

    // Allocate a buffer to hold the defect list.
    defect_buf = std::unique_ptr<DEFECT_ENTRY[]>(new DEFECT_ENTRY[MAX_DEFECTS]());
    if (defect_buf.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of defect buffer failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // We will allocate a reasonable size sum buffer and a corresponding sample buffer.

    // Figure out how many lines to work with at a time, then cur_size is the number of
    // samples we DMA and sum on each iteration.
    cur_size = 1024 * 1024; // This is the max size for the sum buffer.
    lines_per_buf = cur_size / (LINCE_SENSOR_WIDTH * sizeof(uint16_t));
    cur_size = lines_per_buf * LINCE_SENSOR_WIDTH;  // This is the number of samples per iteration.

    // Make the sample buffer size a nice alignment multiple since we will
    // DMA from image memory to processor memory.
    // Bump the allocated size up to avoid issues at upper boundary.
    if ((errno = posix_memalign((void **)&s_ptr, DMA_PAGE_ALIGN,
                                ((cur_size + DMA_PAGE_ALIGN - 1) / DMA_PAGE_ALIGN) * DMA_PAGE_ALIGN)) != 0)
    {
        fi_log(LOG_ERR, "%s: creation of sample buffer failed!", __FUNCTION__);
        goto bail;
    }
    sample_buf.reset(s_ptr);

    // Allocate the sum buffer with 16 bits per sample.
    sum_buf = std::unique_ptr<uint16_t[]>(new uint16_t[cur_size]());
    if (sum_buf.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of sum buffer failed!", __FUNCTION__);
        goto bail;
    }

    // For test purposes, we can write the average of the frames out to a file.
    // The way we know to do this is if the file already exists.
    // "touch ~/.config/Fastec/scratch/black_2560x2048" to create an empty file.
    temp_path = fmt::sprintf("%s/%s", get_home_path(), FULL_FRAME_FILE);
    if (access(temp_path.c_str(), F_OK) == 0)
    {
        CT_GT_FILE::CT_FILE_HDR ct_header;
        uint32_t    cur_sopts;    //!< Sensor bin/subsample options for user calibration.
        uint32_t    cur_sgain;    //!< Sensor gain used during calibration.

        cs_ptr->get_sensor_opts(&cur_sopts);
        cs_ptr->get_sensor_gain(&cur_sgain);
        cur_sgain /= 100;

        // Prep the info structure.
        memset(&ct_header, 0, sizeof(ct_header));

        // Setup our CT info structure.
        ct_header.magic = CT_GT_FILE::CT_MAGIC;
        ct_header.length = LINCE_SENSOR_WIDTH * LINCE_SENSOR_HEIGHT;
        ct_header.roi = ct_roi;

        // Store CT capture parameters just on principle, even though we don't need this
        // information to locate bad pixels.
        *((uint32_t *)((uint8_t *)ct_header.sensor_params + LINCE_OPTS_IDX)) = cur_sopts;
        *((uint16_t *)((uint8_t *)ct_header.sensor_params + LINCE_OFS_IDX)) = 0;
        *((uint8_t *)ct_header.sensor_params + LINCE_GAIN_IDX) = cur_sgain;
        *((uint8_t *)ct_header.sensor_params + LINCE_TROW_IDX) = capture_trow;

        // Write the header. If it fails, just log and proceed since it's optional.
        temp_path = fmt::sprintf("%s/%s", get_home_path(), FULL_FRAME_FILE);
        frame_fd = fopen(temp_path.c_str(), "w");
        if (frame_fd == nullptr)
        {
            fi_log(LOG_ERR, "%s: Can't open %s (%s)", __FUNCTION__, temp_path.c_str(), strerror(errno));
        }
        else if (fwrite((void *)&ct_header, 1, sizeof(ct_header), frame_fd) != sizeof(ct_header))
        {
            fi_log(LOG_ERR, "%s: Can't write to %s (%s)", __FUNCTION__, temp_path.c_str(), strerror(errno));
            fclose(frame_fd);
            frame_fd = nullptr;
        }
    }

    cur_line_count = lines_per_buf;
    cur_size = lines_per_buf * LINCE_SENSOR_WIDTH;
    cur_line = 0;
    num_defects = 0;
    while (cur_line < LINCE_SENSOR_HEIGHT)
    {
        // The final buffer might not have a full complement of lines.
        if ((LINCE_SENSOR_HEIGHT - cur_line) < lines_per_buf)
        {
            cur_line_count = LINCE_SENSOR_HEIGHT - cur_line;
            cur_size = cur_line_count * LINCE_SENSOR_WIDTH;
        }

        // Sum up the frames.
        memset(sum_buf.get(), 0, cur_size * sizeof(uint16_t));
        for (frame = 0; frame < num_frames; ++frame)
        {
            LOG("%s: transfer frame #%d", __PRETTY_FUNCTION__, frame);

            cur_addr = cal_frame_addr + (frame * cal_frame_size) +
                       (cur_line * LINCE_SENSOR_WIDTH);

            // We need to have the samples byte-swapped as we copy to the processor buffer
            // so we correctly compute the horizontal position of any bad pixel, so we use
            // the "little-endian" DMA modifier.
            if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, cur_addr, sample_buf.get(), cur_size) == 0)
            {
                fi_log(LOG_ERR, "%s: read %u from 0x%x failed", __FUNCTION__, cur_size, cur_addr);
                goto bail;
            }

            for (i = 0; i < cur_size ; ++i)
                sum_buf[i] += sample_buf[i];
        }

        // Compute the averages.
        for (i = 0; i < cur_size ; ++i)
            sample_buf[i] = (sum_buf[i] / num_frames);

        // If we are writing out the frame average, save this chunk.
        if (frame_fd != nullptr)
        {
            if (fwrite((void *)sample_buf.get(), 1, cur_size, frame_fd) != cur_size)
            {
                fi_log(LOG_ERR, "%s: Write @%u failed (%s)", __FUNCTION__, cur_line,
                       strerror(errno));
                goto bail;
            }
        }

        // Now check this chunk for pixels above the rejection threshold.
        for (unsigned y = 0; y < cur_line_count; ++y)
        {
            for (unsigned x = 0; x < LINCE_SENSOR_WIDTH; ++x)
            {
                if (sample_buf[x + y * LINCE_SENSOR_WIDTH] > WHITESPOT_THRESHOLD_8BIT)
                {
                    if (num_defects >= MAX_DEFECTS)
                    {
                        fi_log(LOG_ERR, "%s: Too many defects, can't add (%u,%u)", __FUNCTION__,
                               x, y + cur_line);
                        // Write an empty defect file.
                        write_defect_file(0, defect_buf.get());
                        goto bail;
                    }
                    defect_buf[num_defects].x = x;
                    defect_buf[num_defects].y = y + cur_line;
                    ++num_defects;
                }
            }
        }

        cur_line += cur_line_count;
    }

    // Write out the defect file.
    rc = write_defect_file(num_defects, defect_buf.get());

    // Everything is good. Copy the defect file to the ro partition.
    if (rc)
    {
        int     ret;
        std::string sys_cmd;
        Lince5M *sensor_ptr = static_cast<Lince5M *>(CameraServer::get_sensor());

        sys_cmd = fmt::format(FACTORY_DEFECT_SCRIPT, sensor_ptr->defect_filename(".config/Fastec/scratch"));
        if (sys_cmd.empty())
            return false;

        errno = 0;
        ret = system(sys_cmd.c_str());
        if ((ret < 0) || !WIFEXITED(ret))
        {
            // problem log the error
            fi_log(LOG_ERR, "%s: Cannot save the defect map (%d, %d)!", __PRETTY_FUNCTION__,
                   errno, WEXITSTATUS(ret));
            rc = false;
        }
    }

bail:
    if (frame_fd != nullptr)
        fclose(frame_fd);

    return rc;
}

/*!
 * @brief Save the full-size frames we captured for factory calibration.
 *
 * @param filename  Fully-qualified path of file to save to
 * @param num_frames    Number of frames to save
 *
 * @return true if successful, false otherwise.
 */
bool Lince5MCalibration::saveCapture(const char *filename, unsigned num_frames)
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();

    pm_metadata *meta_ptr;
    int     fd = -1;
    bool    rc = false;
    uint32_t    total_size;
    uint64_t    cur_addr;
    uint32_t    cur_size;

    size_t  bounce_bufsize;
    uint8_t *bb_ptr;
    std::unique_ptr<uint8_t[]> bounce_buffer;

    // Allocate a bounce buffer with alignment suitable for DMA from image memory.
    bounce_bufsize = 1024 * 1024;
    if ((errno = posix_memalign((void **)&bb_ptr, DMA_PAGE_ALIGN, bounce_bufsize)) != 0)
    {
        fi_log(LOG_ERR, "%s: creation of bounce buffer failed!", __PRETTY_FUNCTION__);
        goto bail;
    }
    bounce_buffer.reset(bb_ptr);

    // Open our save file.
    fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fi_log(LOG_ERR, "%s: Can't open %s (%d)", __FUNCTION__, filename, errno);
        goto bail;
    }

    // Get and save the partition metadata with our adjustments to match this capture.
    cur_addr = part_mgr->getPartitionStart();
    cur_size = part_mgr->getPmSize();
    if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, cur_addr,
                               bounce_buffer.get(), cur_size) == 0)
    {
        fi_log(LOG_ERR, "%s: failed to read pm_metadata from 0x%" PRIx64, __FUNCTION__, cur_addr);
        goto bail;
    }
    meta_ptr = (pm_metadata *)bounce_buffer.get();
    meta_ptr->roi_x = 0;
    meta_ptr->roi_y = 0;
    meta_ptr->roi_width = LINCE_SENSOR_WIDTH;
    meta_ptr->roi_height = LINCE_SENSOR_HEIGHT;
    meta_ptr->frame_count = num_frames;
    meta_ptr->first_frame = 0;
    meta_ptr->frame_size = cal_frame_size;
    meta_ptr->trig_frame = 0;
    meta_ptr->trig_time_s = time(0);
    meta_ptr->trig_time_ns = 0;
    meta_ptr->fpn_status = GIGE_FPN_SENSOR_ENABLE_MASK;
    meta_ptr->corr_off = 0;
    meta_ptr->corr_len = 0;
    meta_ptr->sov_off = convert_to_QWORD(part_mgr->getPmSize());

    if (write(fd, bounce_buffer.get(), cur_size) < 0)
    {
        fi_log(LOG_ERR, "%s: write pm_metadata failed (%d)", __FUNCTION__, errno);
        goto bail;
    }

    // Convert the framesize to be a byte value.
    total_size = num_frames * cal_frame_size;
    cur_addr = cal_frame_addr;
    while (total_size > 0)
    {
        if (total_size > bounce_bufsize)
            cur_size = bounce_bufsize;
        else
            cur_size = total_size;

        if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, cur_addr,
                                   bounce_buffer.get(), cur_size) == 0)
        {
            fi_log(LOG_ERR, "%s: failed to read %u from 0x%" PRIx64, __FUNCTION__, cur_size, cur_addr);
            goto bail;
        }

        if (write(fd, bounce_buffer.get(), cur_size) < 0)
        {
            fi_log(LOG_ERR, "%s: write failed (%d)", __FUNCTION__, errno);
            goto bail;
        }

        total_size -= cur_size;
        cur_addr += cur_size;
    }
    rc = true;

bail:
    if (fd >= 0)
    {
        close(fd);
    }
    return rc;
}

/*!
 * @brief Capture full-size frames for factory calibration
 *
 * For factory calibration, we can capture both light frames and dark frames.
 * To capture light frames, the application can set up bit modes, integration time,
 * and frame rate, but has no way to set up a full-size ROI so we have to force it.
 *
 * @return true if successful, false otherwise.
 */
bool Lince5MCalibration::captureFullSizeFrames(bool manual_params)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA       *fpga_ptr = CameraServer::get_fpga();
    PartMgr      *part_mgr = CameraServer::get_partmgr();

    std::unique_ptr<pm_metadata> meta_ptr;
    uint32_t    total_size;
    uint32_t    buf_count;
    bitmode_t   bit_mode;
    uint32_t    cal_shutter_speed;
    float cal_frame_rate;
    bool rc = false;

    // Override the parameters for white spot detection but let the application use its own
    // settings for bright/black analysis (manual).
    if (manual_params)
    {
        cs_ptr->get_shutter_speed(&cal_shutter_speed);
        cs_ptr->get_frame_rate(&cal_frame_rate);
    }
    else
    {
        cal_shutter_speed = WHITESPOT_SHUTTER;
        cal_frame_rate = (float)WHITESPOT_FPS;
    }

    // If "manual parameters" were requested, the application sets everything except
    // the capture size. If not, the registers were reset to default values but we
    // still force the bit mode to 8-bit.
    if (manual_params)
        cs_ptr->get_user_bit_mode(&bit_mode);
    else
        bit_mode  = GIGE_BITMODE_11_4;
    ct_roi.origin.x = ct_roi.origin.y = 0;
    ct_roi.width =  LINCE_SENSOR_WIDTH;
    ct_roi.height = LINCE_SENSOR_HEIGHT;

    // Active partition is zero, since we don't calibrate with any captures in memory.
    part_mgr->setActivePartition(0);
    part_mgr->queueCapturePartition(0);

    // Get a "template" for metadata into image memory. This makes sure all the common
    // fields are are initialized from the camera settings, but we will adjust it later
    // ourselves with the settings for this capture. We will need this if we are saving
    // CAP files for the light frames where the analysis is more complicated.
    meta_ptr = std::unique_ptr<pm_metadata>(new pm_metadata);
    part_mgr->partitionMetadataInit(meta_ptr.get(), 0, FACTORY_CAPTURE_FRAMES);
    part_mgr->updateActiveMetadata(meta_ptr.get());

    // Set up the memory buffer and size that we will capture black frames into.
    part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);

    // The size we got from PM matches the camera settings, but we need it to match
    // what we will capture. Recompute the number of buffers that will fit in that area.
    total_size = cal_frame_size * buf_count;
    cal_frame_size = cs_ptr->calc_frame_buf_size(ct_roi.width, ct_roi.height, bit_mode);
    cal_frame_size += FRAME_METADATA_HW_BYTES;
    buf_count = total_size / cal_frame_size;
    if (buf_count < FACTORY_CAPTURE_FRAMES)
        goto cal_exit;

    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
    fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

    // Start capture with our calibration settings.
    cs_ptr->start_lince_calibration(true, bit_mode, (fpnmode_t)GIGE_FPN_SENSOR_ENABLE_MASK,
                                    cal_frame_rate, cal_shutter_speed);

    // Close the shutter.
    fpga_ptr->shutterClose(true);

    // Capture FACTORY_CAPTURE_FRAMES regardless of the currently-selected mode. In particular,
    // we don't want to write to the SSD in LTR mode and we don't want ROC/BROC.
    if (!fpga_ptr->icmCalibrateCapture(FACTORY_CAPTURE_FRAMES))
        goto cal_exit;

#ifdef LINCE_FACTORY_DEBUG
    {
        std::string usb_path;

        // Get the path to the USB drive.
        storage::get_media_device(GIGE_MEDIA_USB, usb_path);

        // Build up a filename.
        usb_path.append("/factory.cap");

        // For debug purposes, we can save the capture for offline analysis.
        rc = save_capture(usb_path.c_str(), FACTORY_CAPTURE_FRAMES);
    }
#else
    // Normally we go off and create our bad pixel map.
    rc = createBadPixelMap();
#endif

cal_exit:
    // Make sure the shutter is open.
    fpga_ptr->shutterClose(false);

    // Return our result.
    return rc;
}

/*!
 * @brief say whether parameters from a CT file are compatible with current settings
 *
 * @param sensor_type   The sensor type read from the file.
 * @param params    Pointer to buffer containing sensor-specific params.
 *
 * @return true if the parameters are compatible, false if not
 */
bool Lince5M::is_cal_compatible(uint32_t /*sensor_type*/, void *params,
                                SENSOR::SENSOR_ROI * /*calfile_roi*/, SENSOR::SENSOR_ROI *user_roi)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint32_t cur_opts;
    uint32_t cur_gain;
    uint32_t sensopt_mask;
    float frame_rate;
    uint32_t ct_opts;
    uint8_t  ct_gain;
    uint8_t  bit_mode;
    bool     rc = true;
    unsigned cur_trow;
    unsigned ct_trow;

    // Sensor options can make a black frame incompatible.
    cs_ptr->get_sensor_opts(&cur_opts);

    // Convert from canonical sensor gain to sensor-specific form, i.e., divide by 100.
    cs_ptr->get_sensor_gain(&cur_gain);
    cur_gain /= 100;

    ct_opts = *((uint32_t *)((uint8_t *)params + LINCE_OPTS_IDX));
    ct_gain = *((uint8_t *)params + LINCE_GAIN_IDX);

    // Check only the sensor option bits we care about.
    sensopt_mask = GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK | GIGE_SENSOPT_HDR_MASK |
                   GIGE_SENSOPT_INTERNAL_0;
    if ((cur_opts & sensopt_mask) != (ct_opts & sensopt_mask))
        rc = false;

    if (cur_gain != ct_gain)
        rc = false;

    // The characteristics for trow=84 are really different from the other configurations.
    // To figure out the trow config, we need to call functions that will sort out where the
    // ROI is on the sensor and determine the readout width. Those functions still take bit_mode
    // as a parameter due to the history of how we got here, but it is not actually used any more.
    // So it is safe to just pass the current bitmode for both the current settings and for the
    // correction table ROI. We already checked the sensor options to make sure there is a match
    // on GIGE_SENSOPT_INTERNAL_0, which could cause incompatibility.
    cs_ptr->get_user_bit_mode(&bit_mode);
    cs_ptr->get_frame_rate(&frame_rate);
    cur_trow = lince5m_get_trow_from_user(user_roi->x, user_roi->y, user_roi->width,
                                          user_roi->height, bit_mode, cur_opts, frame_rate);
    ct_trow = *((uint8_t *)params + LINCE_TROW_IDX);

    if (((cur_trow == TROW_84) && (ct_trow != TROW_84)) ||
        ((cur_trow != TROW_84) && (ct_trow == TROW_84)) ||
        ((cur_trow == TROW_52) && (ct_trow != TROW_52)) ||
        ((cur_trow != TROW_52) && (ct_trow == TROW_52)))
    {
        rc = false;
    }

    return rc;
}

/*!
 * @brief Get the sensor-specific black offset from the CT parameters.
 *
 * @param sensor_type   The sensor type read from the file.
 * @param params    Pointer to buffer containing sensor-specific params.
 *
 * @return true for success, false for failure
 */
uint16_t Lince5M::get_cal_offset(uint32_t /*sensor_type*/, void *params)
{
    uint16_t ct_offset;

    ct_offset = *((uint16_t *)((uint8_t *)params + LINCE_OFS_IDX));
    return ct_offset;
}

/*!
 * @brief Get the "hidden" offset for the CT.
 *
 * We can capture a black frame in a mode where we are artificially restricting the
 * size of the capture based on bandwidth considerations, e.g. 12-bit mode with no
 * binning or subsampling. In this case, the user's ROI offsets of (0,0) are actually
 * moved from the upper left corner of our reported maximum capture region. This
 * function returns that offset so the correction table ROI does not depend on what
 * bit mode it was captured in.
 *
 * @bit_mode    the bit mode
 * @sensor_opts the sensor options
 * @rtn_ofs_x   the hidden ROI X offset for the (bit mode, sensor options) setting
 * @rtn_ofs_y   the hidden ROI Y offset for the (bit mode, sensor options) setting
 *
 * THIS IS OBSOLETE because we no longer restrict the size based on bandwidth.
 */
void Lince5M::get_ct_offset(
    bitmode_t bit_mode,
    uint32_t sensor_opts,
    uint16_t *rtn_ofs_x,
    uint16_t *rtn_ofs_y)
{
    lince5m_get_ct_offset(bit_mode, sensor_opts, rtn_ofs_x, rtn_ofs_y);
}

/*!
 * @brief The calibration process for Lince5M
 *
 * Factory calibration captures full-size frames using the entire sensor area, which is
 * the only time we actually do that, but it doesn't actually do anything other than
 * saving the frames. This is for detecting bad pixels since we don't use a factory
 * black reference.
 *
 * For user calibration, we either do a pixel-FPN calibration and capture a black reference
 * fusing the current ROI or else we do a sensor-FPN calibration, which is the same as normal
 * sensor startup except that a few parameters are different. The normal sensor startup
 * always does a "mini calibration" which can be done with light on the sensor, but the
 * non-mini version requires that the lens cap be on.
 * It doesn't make any sense to restart the sensor after a non-mini calibration, since
 * that will just undo the parameters that the process tuned.
 *
 * @param is_factory    true if this is a factory calibration
 * @param manual_params true if there are parameters the user set manually, not used for Lince
 * @param p_rtn_restart set to true on exit if capture must be restarted after calibration.
 */
bool Lince5M::run_calibration(bool is_factory, bool manual_params, bool *p_rtn_restart)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    bool rc = false;

    if (is_factory)
    {
        if (allow_factory)
        {
            Lince5MCalibration fpn;

            // Our factory calibration captures full-size frames, which
            // means we can't have live view running.
            rc = fpn.captureFullSizeFrames(manual_params);
            *p_rtn_restart = true;
        }
        else
        {
            fi_log(LOG_ERR, "%s: attempt to do factory calibration for Lince", __FUNCTION__);
            return false;
        }
    }
    else
    {
        fpnmode_t fpn_mode;
        bitmode_t bit_mode;

        cs_ptr->get_user_fpn(&fpn_mode);
        cs_ptr->get_user_bit_mode(&bit_mode);

        // FPGA FPN calibration captures an average black frame.
        if ((fpn_mode & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        {
            Lince5MCalibration fpn;

            rc = fpn.runFpnCalibration();

            // If we are collecting black frames, we need to restart capture after calibration.
            *p_rtn_restart = true;
        }
        // Other calibration just starts the sensor with lens-covered parameters.
        else
        {
            // Pass 0 for frame rate and shutter speed when we are not overriding them.
            rc = cs_ptr->start_lince_calibration(false, bit_mode, fpn_mode, 0, 0);
            *p_rtn_restart = false;
        }
    }
    return rc;
}


/*!
 * @brief Entry point from the calibration worker thread, this
 *        method performs the PRNU calibration operations for the Luxima
 *        sensors.
 *
 * @param ct_ptr    [IN] Pointer to the FPN correction table, required by PRNU.
 * @returns True if successfull, false otherwise.
 */
bool Lince5M::run_prnu(const uint8_t *ct_ptr)
{
    bool rc = false;
    Lince5MCalibration calib;

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (ct_ptr != nullptr)
    {
        // Run the calibration algorithm.
        rc = calib.runPrnuCalibration(ct_ptr);
    }

    return rc;
}

