/*!
 * @file ltr_worker.cxx
 * @brief This file implements the LTR mode working class.
 *
 * This class move data from image memory the the NVME.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

//#define LR_STATS 1
//#define DO_STATS 1

#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include <fcntl.h>
#include <inttypes.h>
#include <memory.h>
#include <signal.h>
#include <termios.h>

#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <chrono>
#include <memory>

#include "bounce.h"
#include "camera_server.h"
#include "constants.h"
#include "cs_utils.h"
#include "fpga.h"
#include "gige_reg.h"
#include "ltr_worker.h"
#include "override.h"
#include "part_mgr.h"
#include "storage.h"

#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>
#include <fastec/fas_filesystem.h>

namespace LTR_WORKER
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

const uint32_t MAX_AF_LEVEL = 0x48;                     //! Imperically derived, seems like a good number.
const uint32_t MIN_TRANSFER_SIZE = (32 * 1024 * 1024);  //!< Minimum amount we want to transfer.
const uint32_t WAIT_FOR_TERMINATION = 1000;             //!< Time to wait to be shutdown.

const char *BAR_LENGTH_FILE = "/proc/fi_pci/bar_sel_len";

const int WAIT_FOR_FRAMES_TIMEOUT = 2;     //!< Amount of time to wait for frames to be captured, milliseconds.

//! The various trace levels for the LTR worker thread.
enum LTRW_TRACE
{
    LTRW_FUNC       = 0x00000010,
    LTRW_STATS      = 0x00000020,
    LTRW_STATE      = 0x00000040,
    LTRW_DEBUG      = 0x00000080,
    LTRW_VERBOSE    = 0x00000100
};

//! The various states the LTR object can hold.
enum XFER_STATES
{
    XFER_WAIT_FOR_FRAMES,       //!< Wait until we get enough frames to transfer.
    XFER_TO_SSD,            //!< Transfer the captured data to the SSD.
    XFER_END                //!< Waiting to be terminated.
};

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    LTRW_FUNC       |
//        LTRW_STATS      |
    LTRW_STATE      |
    LTRW_DEBUG      |
//    LTRW_VERBOSE    |
//    TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

} /* End of LTR_WORKER */

// Make things easier for us.
using namespace LTR_WORKER;

uint32_t total_descr_count = 0;
uint64_t total_bytes_written = 0;

//--------------------------------------------------------------------
// Class ltr_worker implementation
//--------------------------------------------------------------------

ltr_worker::ltr_worker() :
    FAS_THREAD::Thread(true),       // Run as a "detached" thread.
    last_im_address(0),
    ssd_pos(0),
    ssd_fd(-1),
    xfer_state(0),
    xfer_buffer(nullptr),
    leftover(0),
    af_level(0),
    bar_length(0),
    need_to_splice(false),
    final_write(false),
    flush_the_pipe(false)
{
    FILE *fp;
    uint32_t temp_len;

    TP(LTRW_FUNC, "%s", __PRETTY_FUNCTION__);

    total_descr_count = 0;
    total_bytes_written = 0;
    error_cond = 0;

    // Not running yet.
    thread_running = false;
    stop_request = false;

#ifdef LR_STATS
    stats_fp = fopen("lr_stats.csv", "w");
    if (stats_fp != NULL)
    {
        fprintf(stats_fp, "Time,MB/sec,Bytes,Delay,Overhead\n");
    }
#endif /* LR_STATS */

    // Determine the size of our DMA BAR.
    fp = fopen(BAR_LENGTH_FILE, "r");
    if (fp != NULL)
    {
        fscanf(fp, "%u", &temp_len);
        fclose(fp);

        // Don't need to use the whole bar.
        bar_length = temp_len >> 2;
    }

    TP(LTRW_VERBOSE, "%s: bar_length=%u", __PRETTY_FUNCTION__, bar_length);
}

ltr_worker::~ltr_worker()
{
    TP(LTRW_FUNC, "%s", __PRETTY_FUNCTION__);

    // This should already be done, but just in case...
    cleanup();

#ifdef LR_STATS
    if (stats_fp != NULL)
    {
        fclose(stats_fp);
        stats_fp = NULL;
    }
#endif /* LR_STATS */
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the LTR worker class object.
 */
void ltr_worker::cleanup()
{
    TP(LTRW_FUNC, "%s", __PRETTY_FUNCTION__);

    // Do we have an open file?
    if (ssd_fd >= 0)
    {
        // fsync(ssd_fd);
        close(ssd_fd);
        ssd_fd = -1;
    }

    // Release the transfer buffer.
    if ((xfer_buffer != nullptr) && (xfer_buffer != (void *)MAP_FAILED))
    {
        LOG("%s: Releasing the xfer_buffer", __PRETTY_FUNCTION__);
        munmap(xfer_buffer, bar_length);
        xfer_buffer = nullptr;
    }

    // Close the PCI driver descriptor, if we have one.
    if (pci_fd >= 0)
    {
        LOG("%s: Closing PCI driver.", __PRETTY_FUNCTION__);
        close(pci_fd);
        pci_fd = -1;
    }
}

void ltr_worker::errorHandler()
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // Dearm the camera.
    CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_DEARMING, false, nullptr);

    // Indicate an error.
    fpga_ptr->ltrSetSaveError(GIGE_SAVE_STATE_ERR | error_cond);
}

void ltr_worker::dataReady()
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(LTRW_VERBOSE, "%s", __PRETTY_FUNCTION__);

    // Diable interrupts.
    fpga_ptr->irqDisable(PCIEIRQ_LR_DESC_FIFO_AF);

    // Get a scope for doing the notify.
    {
        std::unique_lock<std::mutex> locker(xfer_mutex);

        // Notify our worker that there's data ready.
        xfer_status.notify_all();
    }
    TP(LTRW_VERBOSE, "%s: done", __PRETTY_FUNCTION__);
}

/*!
 * @brief The LTR state machine implementation. It will keep running until someone tells us to
 *    "stop", and then it will keep going until the FPGA says there's nothing left to transfer.
 */
void ltr_worker::run()
{
    bool done = false;
#ifdef LR_STATS
    Fastec::IntervalTimer data_ready_timer;
#endif /* LR_STATS */

    LOG("%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Initialize stuff.
    if (!setup())
    {
        fi_log(LOG_DEBUG, "%s: Unable to setup worker thread!", __PRETTY_FUNCTION__);
        done = true;
        error_cond = EFAULT;
        xfer_state = XFER_END;
    }
    else
    {
        // Initialize our state to wait for data.
        xfer_state = XFER_WAIT_FOR_FRAMES;
        error_cond = 0;

        // We're ALIVE!
        thread_running = true;
        notify_ready();

#ifdef LR_STATS
        delay_for_data = 0.0;
        data_ready_timer.start();
#endif /* LR_STATS */
    }

    // Loop until we're done.
    while (!done)
    {
        int msecs = 0;

        TP(LTRW_STATE, "%s: xfer_state=%d stop_request=%d flush=%d", __PRETTY_FUNCTION__, xfer_state, stop_request, flush_the_pipe);

        switch (xfer_state)
        {
        case XFER_WAIT_FOR_FRAMES:
            // We are in the "waiting for frames" state. We need to see if the FPGA has accumulated enough data to
            // warrant starting a DMA transfer.
            if (checkForEnoughData())
            {
                // We have enough data, so go get it.
                xfer_state = XFER_TO_SSD;
            }

            else if (error_cond)
            {
                TP(LTRW_DEBUG, "%s: wait_for_frames() failed!", __PRETTY_FUNCTION__);
                xfer_state = XFER_END;
            }

            //======================================================
            // ALM: Trial to see if we can shut it down quickly.
            //======================================================
            else if (stop_request)
            {
                // All done.
                TP(LTRW_DEBUG, "%s: stop_request is set, we're done.", __PRETTY_FUNCTION__);
                msecs = 0;
                done = true;
            }
            else
            {
                // Not enough data. Wait around.
                msecs = WAIT_FOR_FRAMES_TIMEOUT;
            }
            break;
        case XFER_TO_SSD:
#ifdef LR_STATS
            // Get the time in microseconds.
            data_ready_timer.stop();
            delay_for_data = data_ready_timer.lastMeasurement;
            data_ready_timer.start();
            LOG("%s: delay_for_data=%lf", __PRETTY_FUNCTION__, delay_for_data);
#endif /* LR_STATS */

            // We have enough data, so start a DMA transfer.
            if (saveImageData())
            {
                // Go wait for more data.
                xfer_state = XFER_WAIT_FOR_FRAMES;
            }
            else
            {
                TP(LTRW_DEBUG, "%s: save_image_data() failed!", __PRETTY_FUNCTION__);
                xfer_state = XFER_END;
            }
            break;
        case XFER_END:
        default:
            fi_log(LOG_WARNING, "%s: Error state - waiting to die...", __PRETTY_FUNCTION__);
            errorHandler();
            done = true;
            break;
        }

        // Do we want to wait around?
        if (msecs > 0)
        {
            FiFPGA  *fpga_ptr = CameraServer::get_fpga();

            TP(LTRW_VERBOSE, "%s: Waiting %d msec or AF interrupt.", __PRETTY_FUNCTION__, msecs);

            // Clear the IRQ status bit. It'll immediatly be set if we didn't drain the
            // descriptor FIFO enough.
            fpga_ptr->regWrite(PCIE_IRQ_STATUS, PCIEIRQ_LR_DESC_FIFO_AF);

            // Re-enable the IRQ.
            fpga_ptr->irqEnable(PCIEIRQ_LR_DESC_FIFO_AF);

            // Wait until we're told a transfer is ready, or 'msecs' milliseconds.
            std::unique_lock<std::mutex> locker(xfer_mutex);
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            using ms = std::chrono::milliseconds;
            if (xfer_status.wait_until(locker, now + ms(msecs)) == std::cv_status::timeout)
            {
                TP(LTRW_DEBUG, "%s: Timed out. <<<<<<<<<<<<<<<<<<", __PRETTY_FUNCTION__);
            }
            else
            {
                TP(LTRW_DEBUG, "%s: Got notification. ++++++++++++++++++", __PRETTY_FUNCTION__);
            }
        }

        TP(LTRW_VERBOSE, "%s: Bottom of loop", __PRETTY_FUNCTION__);
    }

    {
        // Diable interrupts.
        FiFPGA *fpga_ptr = CameraServer::get_fpga();
        if (fpga_ptr != nullptr)
        {
            fpga_ptr->irqDisable(PCIEIRQ_LR_DESC_FIFO_AF);

            // Shut down the IM2SSD DMA.
            fpga_ptr->regWrite(IM2SSD_DMA0_CONTROL, 0);
        }
    }

    // Clean things up and leave.
    cleanup();

    // Indicate that we've stopped running.
    thread_running = false;
    notify_done();

    LOG("%s: done", __PRETTY_FUNCTION__);
}

bool ltr_worker::spliceAtEnd(uint8_t *write_ptr)
{
    bool bOK = false;
    int rc;
    ssize_t result;
    uint8_t *mbuf_ptr;
    off64_t merge_pos;
    std::string dev_path;
    std::unique_ptr<uint8_t[]> merge_buf;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(LTRW_FUNC, "%s: merge leftover=0x%" PRIx64, __PRETTY_FUNCTION__, leftover);

    // Ensure the write has completed.
    fsync(ssd_fd);

    // Read the current position on the SSD.
    merge_pos = lseek64(ssd_fd, 0, SEEK_CUR);
    TP(LTRW_FUNC, "%s: merge merge_pos=0x%" PRIx64, __PRETTY_FUNCTION__, merge_pos);

    // We need to close the SSD so we can re-open it in read/write mode.
    LOG("%s: close the fd", __PRETTY_FUNCTION__);
    close(ssd_fd);

    // Get the path for the SSD.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
        return false;
    }

    // Now open the SSD in re-write mode for the final splice.
    LOG("%s: open read/write", __PRETTY_FUNCTION__);
    ssd_fd = open(dev_path.c_str(), O_RDWR | O_LARGEFILE | O_DIRECT, 0666);
    if (ssd_fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for write failed - (%d) %s", __PRETTY_FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Go to the location prior to closing and reopening the drive.
    lseek64(ssd_fd, merge_pos, SEEK_SET);

    // Override our FIFO empty behavior.
    fpga_ptr->regModify(IM2SSD_DMA0_CONTROL, IM2SSD_DMA0_FIFO_EMPTY_OR, -1);

    // Allocate an aligned buffer.
    rc = posix_memalign((void **)&mbuf_ptr, FAS_FILESYSTEM::SSD_PAGE_LEN, 2 * FAS_FILESYSTEM::SSD_PAGE_LEN);
    if ((mbuf_ptr == nullptr) || (rc != 0))
    {
        fi_log(LOG_ERR, "%s: Failed to allocate merge buffer (status=%d)", __FUNCTION__, rc);
        goto splice_bail;
    }
    merge_buf.reset(mbuf_ptr);

    // Read up the page currently located at this sector.
    result = read(ssd_fd, merge_buf.get(), FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't read all the data (result=%d)", __PRETTY_FUNCTION__, result);
        goto splice_bail;
    }

    // Go back to the beginning of this page.
    lseek64(ssd_fd, merge_pos, SEEK_SET);

    // Write the FPGA's data (plus some garbage).
    result = write(ssd_fd, write_ptr, FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't write all the last data (result=%d)", __PRETTY_FUNCTION__, result);
        goto splice_bail;
    }

    // Go back to the beginning of this page.
    lseek64(ssd_fd, merge_pos, SEEK_SET);

    // Read up the page currently located at this sector which
    // contains the FPGA data (plus some garbage).
    result = read(ssd_fd, &merge_buf[FAS_FILESYSTEM::SSD_PAGE_LEN], FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't read all the data (result=%d)", __PRETTY_FUNCTION__, result);
        goto splice_bail;
    }

    // Merge the two buffers.
    for (rc = 0; rc < (int)leftover; rc++)
        merge_buf[rc] = merge_buf[rc + FAS_FILESYSTEM::SSD_PAGE_LEN];

    // Now write the mreged page back to the SSD.
    lseek64(ssd_fd, merge_pos, SEEK_SET);
    result = write(ssd_fd, merge_buf.get(), FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't write all the merge data (result=%d)", __PRETTY_FUNCTION__, result);
    }

    // Restore our FIFO empty behavior.
    fpga_ptr->regModify(IM2SSD_DMA0_CONTROL, 0, ~IM2SSD_DMA0_FIFO_EMPTY_OR);

    merge_pos = lseek64(ssd_fd, 0, SEEK_CUR);
    TP(LTRW_FUNC, "%s: final merge_pos=0x%" PRIx64, __PRETTY_FUNCTION__, merge_pos);

    // Everything seems OK.
    bOK = true;

splice_bail:
    return bOK;
}

bool ltr_worker::spliceAtStart()
{
    bool bOK = false;
    int rc;
    ssize_t result;
    uint32_t existing_data;
    uint32_t fifo_data_count;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    uint8_t *mbuf_ptr;
    std::string dev_path;
    std::unique_ptr<uint8_t[]> merge_buf;

    TP(LTRW_FUNC, "%s: starting ssd_pos=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos);

    // Ensure the write has completed.
    fsync(ssd_fd);

    // We need to close the SSD so we can re-open it in read/write mode.
    LOG("%s: close the fd", __PRETTY_FUNCTION__);
    close(ssd_fd);

    // Get the path for the SSD.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
        return false;
    }

    // Now open the SSD in re-write mode for the final splice.
    LOG("%s: open read/write", __PRETTY_FUNCTION__);
    ssd_fd = open(dev_path.c_str(), O_RDWR | O_LARGEFILE | O_DIRECT, 0666);
    if (ssd_fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for write failed - (%d) %s", __PRETTY_FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Allocate an aligned buffer.
    rc = posix_memalign((void **)&mbuf_ptr, FAS_FILESYSTEM::SSD_PAGE_LEN, 2 * FAS_FILESYSTEM::SSD_PAGE_LEN);
    if ((mbuf_ptr == nullptr) || (rc != 0))
    {
        fi_log(LOG_ERR, "%s: Failed to allocate merge buffer (status=%d)", __FUNCTION__, rc);
        goto splice_bail;
    }
    merge_buf.reset(mbuf_ptr);

    // How much data is already in the page?
    existing_data = (ssd_pos & (FAS_FILESYSTEM::SSD_PAGE_LEN - 1));

    // How much data do we need from the IM FIFO?
    fifo_data_count = FAS_FILESYSTEM::SSD_PAGE_LEN - existing_data;

    // Sanity check.
    if ((fifo_data_count & 0xF) != 0)
    {
        // We MUST be in multiples of 16 bytes!
        fi_log(LOG_ERR, "%s: fifo_data_count [%d] not 16-byte aligned!", __PRETTY_FUNCTION__, fifo_data_count);
        fi_log(LOG_ERR, "%s: existing_data=%d", __PRETTY_FUNCTION__, existing_data);
        goto splice_bail;
    }
    if (fifo_data_count > xfer_len)
    {
        fi_log(LOG_ERR, "%s: fifo_data_count [%d] > xfer_len [%d]?!", __PRETTY_FUNCTION__, fifo_data_count, xfer_len);
        goto splice_bail;
    }

    // Adjust our offset to the beginning of this page.
    ssd_pos -= existing_data;

    // Go back to the beginning of this page.
    lseek64(ssd_fd, ssd_pos, SEEK_SET);

    // Read up the page currently located at this sector.
    result = read(ssd_fd, merge_buf.get(), FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't read all the data (result=%d)", __PRETTY_FUNCTION__, result);
        goto splice_bail;
    }

    // Read up the data we need from the IM FIFO.
    if (!fpga_ptr->ltrReadImFIFO(&merge_buf[FAS_FILESYSTEM::SSD_PAGE_LEN], fifo_data_count))
    {
        fi_log(LOG_ERR, "%s: ltr_read_imfifo() failed!", __PRETTY_FUNCTION__);
        goto splice_bail;
    }

    // Merge the two buffers.
    for (rc = 0; rc < (int)fifo_data_count; rc++)
        merge_buf[rc + existing_data] = merge_buf[rc + FAS_FILESYSTEM::SSD_PAGE_LEN];

    // Now write the merged page back to the SSD.
    lseek64(ssd_fd, ssd_pos, SEEK_SET);
    result = write(ssd_fd, merge_buf.get(), FAS_FILESYSTEM::SSD_PAGE_LEN);
    if (result != (ssize_t)FAS_FILESYSTEM::SSD_PAGE_LEN)
    {
        fi_log(LOG_ERR, "%s: Didn't write all the merge data (result=%d)", __PRETTY_FUNCTION__, result);
        goto splice_bail;
    }

    // Our SSD position is OK, but we need to adjust our lengths.
    xfer_len = (xfer_len + leftover) - fifo_data_count;

    // Make sure we are on a page boundary.
    if ((xfer_len & (FAS_FILESYSTEM::SSD_PAGE_LEN - 1)) != 0)
    {
        uint32_t aligned_len;

        // Round the count downwards.
        aligned_len = ((xfer_len / FAS_FILESYSTEM::SSD_PAGE_LEN) * FAS_FILESYSTEM::SSD_PAGE_LEN);

        // The difference is the amount that will go into the next
        // SSD page.
        leftover = xfer_len - aligned_len;

        // Remember the aligned length. This is what we're going to
        // want to transfer.
        xfer_len = aligned_len;

        // Sanity check.
        if (leftover >= FAS_FILESYSTEM::SSD_PAGE_LEN)
        {
            fi_log(LOG_DEBUG, "%s: leftover (%u) is too large!!!", __PRETTY_FUNCTION__);
        }
    }

    // Read the current position on the SSD.
    ssd_pos = lseek64(ssd_fd, 0, SEEK_CUR);

    // We need to close the SSD so we can re-open it in write-only mode.
    LOG("%s: close the fd", __PRETTY_FUNCTION__);
    close(ssd_fd);

    // Now open the SSD in re-write mode for the final splice.
    LOG("%s: open write-only", __PRETTY_FUNCTION__);
    ssd_fd = open(dev_path.c_str(), O_WRONLY | O_LARGEFILE | O_DIRECT, 0666);
    if (ssd_fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for write failed - (%d) %s", __PRETTY_FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Get back to where we need to be.
    lseek64(ssd_fd, ssd_pos, SEEK_SET);

    // Everything seems OK.
    bOK = true;

splice_bail:
    TP(LTRW_FUNC, "%s: bOK=%d", __PRETTY_FUNCTION__, bOK);
    return bOK;
}

/*!
 * @brief Write the data in the FPGA's FIFO to the drive. This is accomplished by issuing
 *        a normal "write" from the FPGA's BAR as the source buffer to the SSD's file descriptor.
 *
 * In addition, this handles the various cases of:
 *     - a "normal" write
 *     - aligning the data to start on a page bounder (spliceAtStart)
 *     - merge the last "chunk-o-data" into the partial page at the end (spliceAtEnd).
 *
 * @return True if everything went OK, false otherwise.
 */
bool ltr_worker::saveImageData()
{
    bool bOK = false;
    ssize_t result;
    ssize_t write_len;
    uint8_t *write_ptr;
    uint32_t total_xfer_len;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    // Statistics
#ifdef DO_STATS
    uint32_t fpga_timing_len(0);
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    Fastec::IntervalTimer write_timer;
    write_timer.start();
#endif /* DO_STATS */
    // Statistics

    // Read the current position on the SSD.
    ssd_pos = lseek64(ssd_fd, 0, SEEK_CUR);
    TP(LTRW_FUNC, "%s: starting ssd_pos=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos);

    // If we need to catch up, emit a log message.
    if (xfer_len >= bar_length)
        fi_log(LOG_WARNING, "%s[%d]: xfer_len=0x%08x bar_length=0x%08x ssd_pos=0x%" PRIx64 " <<<<<<<<<<<<", __PRETTY_FUNCTION__, __LINE__, xfer_len, bar_length, ssd_pos);

    // Loop until we're done with the total amount to write.
    total_xfer_len = xfer_len;
    while (total_xfer_len != 0)
    {
        // Transfer as much as we can...but no more.
        if (total_xfer_len >= bar_length)
            xfer_len = bar_length;
        else
            xfer_len = total_xfer_len;
        total_xfer_len -= xfer_len;

        // Set our write pointer to the start of the FIFO "memory".
        write_ptr = (uint8_t *)xfer_buffer;

        // Do we need to splice our image data?
        if (need_to_splice)
        {
            // We will splice in the image data to align ourselves on an
            // SSD page. After this routine finishes, everything will
            // be good to go.
            spliceAtStart();
            need_to_splice = false;
        }

        // Determine if we can fit the amount of data into the remaining space of the SSD.
        if ((ssd_pos + xfer_len) >= (off64_t)ssd_end_offset)
        {
            uint64_t last_addr;

            TP(LTRW_DEBUG, "%s: Need to wrap, ssd_pos=0x%" PRIx64 ", len=%u, end=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos, xfer_len, ssd_end_offset);

            // What is the address after the last frame in the partition.
            last_addr = pm_ptr->get_active_frame_addr(pm_ptr->ltrGetActiveFrameCount() - 1, true);
            last_addr += pm_ptr->getFrameSize();

            // We have configured the FPGA to padd out this last frame, so add in those bytes too.
            last_addr += padding_count;
            TP(LTRW_DEBUG, "%s: Byte beyond last frame+pad is: 0x%" PRIx64, __PRETTY_FUNCTION__, last_addr);

            // Figure out how many bytes we can fit.
            write_len = (ssize_t)(last_addr - ssd_pos);
            TP(LTRW_DEBUG, "%s: We can fit %u bytes", __PRETTY_FUNCTION__, write_len);

            // Sanity check.
            if ((ssd_pos + write_len) >= (off64_t)ssd_end_offset)
            {
                fi_log(LOG_ERR, "%s: Our calculations failed, write_len=%u", __PRETTY_FUNCTION__, write_len);
            }

            // Statistics
#ifdef DO_STATS
            fpga_ptr->regWrite(IM2SSD_DMA0_PCIE_BURST_SIZE, convert_to_QWORD((uint32_t)write_len));
            fpga_timing_len = (uint32_t)write_len;
            fpga_ptr->regModify(IM2SSD_DMA0_CONTROL, IM2SSD_DMA0_BURST_TIMER_ARM, -1);
#endif /* DO_STATS */
            // Statistics

            // Check to see if we have ANY room.
            if (write_len > 0)
            {
                // Write the data to the SSD. Our file pointer SHOULD already be good to go.
                TP(LTRW_VERBOSE, "%s: write from %p for %u bytes", __PRETTY_FUNCTION__, (void *)write_ptr, write_len);
                result = write(ssd_fd, write_ptr, write_len);
                if (result != (ssize_t)write_len)
                {
                    // Log the error, but keep going. We'll die otherwise.
                    fi_log(LOG_ERR, "%s: First write failed (result=%d, len=%u).", __PRETTY_FUNCTION__, result, write_len);
                }
            }
            else
            {
                fi_log(LOG_WARNING, "%s: No room, wrap to the beginning.", __PRETTY_FUNCTION__);
            }

            // Now move back to the beginning of the capture.
            if (lseek64(ssd_fd, (off64_t)ssd_start_offset, SEEK_SET) != (off64_t)ssd_start_offset)
            {
                fi_log(LOG_ERR, "%s: Unable to seek to the start=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_start_offset);
                goto save_bail;
            }

            // There are no partial frames left in the FIFO.
            leftover = 0;

            // Where are we now?
            ssd_pos = (off64_t)ssd_start_offset;
            TP(LTRW_FUNC, "%s: finished ssd_pos=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos);
        }
        else
        {
            // Statistics
#ifdef DO_STATS
            fpga_ptr->regWrite(IM2SSD_DMA0_PCIE_BURST_SIZE, convert_to_QWORD(xfer_len));
            fpga_timing_len = (uint32_t)xfer_len;
            fpga_ptr->regModify(IM2SSD_DMA0_CONTROL, IM2SSD_DMA0_BURST_TIMER_ARM, -1);
#endif /* DO_STATS */
            // Statistics

            // Write the data to the SSD. Our file pointer SHOULD already be good to go.
            TP(LTRW_VERBOSE, "%s: write from %p for %u bytes", __PRETTY_FUNCTION__, (void *)write_ptr, xfer_len);
            result = write(ssd_fd, write_ptr, xfer_len);
            if (result != (ssize_t)xfer_len)
            {
                fi_log(LOG_ERR, "%s: Didn't write all the data (result=%d)", __PRETTY_FUNCTION__, result);
                goto save_bail;
            }
            TP(LTRW_VERBOSE, "%s: write complete, result=%" PRIu64, __PRETTY_FUNCTION__, result);

            // If we are stopping, then we need to address the leftover amount.
            if (final_write && (total_xfer_len == 0) && (leftover > 0))
            {
                TP(LTRW_DEBUG, "%s: Final write, leftover=%u", __PRETTY_FUNCTION__, leftover);
                if (!spliceAtEnd(write_ptr))
                    goto save_bail;

                // Update the xfer_len to include the amount we just spliced in.
                xfer_len += leftover;
                leftover = 0;
                TP(LTRW_FUNC, "%s: after splice, xfer_len=0x%x", __PRETTY_FUNCTION__, xfer_len);
            }

            // Where are we now?
            ssd_pos += xfer_len;
            TP(LTRW_FUNC, "%s: finished ssd_pos=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos);

            // Check to see if we should wrap.
            if (!final_write)
            {
                uint64_t last_addr;

                // What is the address after the last frame in the partition.
                last_addr = pm_ptr->get_active_frame_addr(pm_ptr->ltrGetActiveFrameCount() - 1, true);
                last_addr += pm_ptr->getFrameSize();

                // We have configured the FPGA to padd out this last frame, so add in those bytes too.
                last_addr += padding_count;

                // Figure out how many bytes we can fit.
                if (last_addr == (uint64_t)ssd_pos)
                {
                    // Now move back to the beginning of the capture.
                    if (lseek64(ssd_fd, (off64_t)ssd_start_offset, SEEK_SET) != (off64_t)ssd_start_offset)
                    {
                        fi_log(LOG_ERR, "%s: Unable to seek to the start=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_start_offset);
                        goto save_bail;
                    }

                    // Where are we now?
                    ssd_pos = (off64_t)ssd_start_offset;
                    fi_log(LOG_INFO, "%s: pre-emptive wrap, ssd_pos=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_pos);
                }
            }
        }
    }

    // All is well.
    bOK = true;

#ifdef DO_STATS
//    if (traceControl & (LTRW_STATS | TRACE_0))
    {
        // Statistics
        uint32_t reg_value;

        TP(LTRW_STATS, "%s: total_bytes_written=%" PRIu64 ", total_descr_count=%u, leftover=%u", __PRETTY_FUNCTION__, total_bytes_written, total_descr_count, leftover);

        // Software timing.
        write_timer.stop();
        TP(LTRW_STATS, "%s: LR capture: %lf secs. (%lf MB/sec)", __PRETTY_FUNCTION__, write_timer.lastMeasurement, (((double)xfer_len) / (1024.0 * 1024.0)) / write_timer.lastMeasurement);

        // FPGA timing.
        fpga_ptr->regRead(IM2SSD_DMA0_PCIE_BURST_TIMER, &reg_value);
        if ((reg_value & IM2SSD_DMA0_PCIE_BT_READY) != IM2SSD_DMA0_PCIE_BT_READY)
        {
            LOG("%s: Timer is not ready [0x%08x], not enough data yet?", __PRETTY_FUNCTION__, reg_value);
        }
        else
        {
            double fpga_time;

            fpga_time = ((double)(fpga_timing_len)) / (double)(reg_value & IM2SSD_DMA0_PCIE_BT_MASK);
            LOG("%s: FPGA time: %lf secs. (%lf MB/sec)", __PRETTY_FUNCTION__, (double)(reg_value & IM2SSD_DMA0_PCIE_BT_MASK) / 1000000.0, fpga_time);
        }
        fpga_ptr->regModify(IM2SSD_DMA0_CONTROL, IM2SSD_DMA0_BURST_TIMER_CLR, -1);
    }   // Statistics
#endif /* DO_STATS */

save_bail:
    // Return our status.
    TP(LTRW_FUNC, "%s: Done (%d)", __PRETTY_FUNCTION__, bOK);
    if (!bOK)
    {
        // Generic error condition.
        error_cond = ENOSPC;
    }
    return bOK;
}

bool ltr_worker::setup()
{
    uint64_t last_addr;
    uint64_t aligned_addr;
    uint32_t padding_count_qw;
    uint32_t dummy;
    uint32_t mode;
    uint32_t reg_value;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    std::string dev_path;

    TP(LTRW_FUNC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (ssd_fd != -1)
    {
        fi_log(LOG_DEBUG, "%s: SSD already open?!", __PRETTY_FUNCTION__);
        close(ssd_fd);
        ssd_fd = -1;
    }

    // Find the LTR device.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
        return false;
    }

    // Open the SSD.
    TP(LTRW_DEBUG, "%s: open write-only '%s'", __PRETTY_FUNCTION__, dev_path.c_str());
    ssd_fd = open(dev_path.c_str(), O_WRONLY | O_LARGEFILE | O_DIRECT, 0666);
    if (ssd_fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for write failed - (%d) %s", __PRETTY_FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Are we ROC/BROC mode?
    mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

    // Make sure we get to the correct place on the drive to start writing image data.
    ssd_start_offset = pm_ptr->ltrGetSessionStart(pm_ptr->ltrGetSession());
    ssd_end_offset = ssd_start_offset + pm_ptr->ltrGetSessionSize();

    // If we are in ROC/BROC, then we might need to append to the end of the session.
    if (mode)
    {
        // Get the start point to be appending frames.
        ssd_start_offset = pm_ptr->ltrGetRocStart();
        if ((ssd_start_offset & (FAS_FILESYSTEM::SSD_PAGE_LEN - 1)) != 0)
        {
            // Do we really care? Emit a message until we figure out if it's a problem.
            fi_log(LOG_DEBUG, "%s: ssd_start_offset(0x%" PRIx64 ") is not page aligned!", __PRETTY_FUNCTION__, ssd_start_offset);
            need_to_splice = true;
        }
    }

    TP(LTRW_DEBUG, "%s: start_offset=0x%" PRIx64 ", end_offset=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_start_offset, ssd_end_offset);
    if (lseek64(ssd_fd, (off64_t)ssd_start_offset, SEEK_SET) != (off64_t)ssd_start_offset)
    {
        fi_log(LOG_ERR, "%s: Unable to seek to the start=0x%" PRIx64, __PRETTY_FUNCTION__, ssd_start_offset);
        goto ssd_bail;
    }

    // Open our PCI device.
    pci_fd = open("/dev/fi_pci", O_RDONLY | O_SYNC | O_LARGEFILE);
    if (pci_fd < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto ssd_bail;
    }

    // Memory map the BAR.
    TP(LTRW_DEBUG, "%s: Attempting to mmap %u bytes", __PRETTY_FUNCTION__, bar_length);
    xfer_buffer = (void *)mmap(NULL, bar_length, PROT_READ, MAP_NORESERVE | MAP_SHARED, pci_fd, 0);
    if ((xfer_buffer == nullptr) || (xfer_buffer == (void *)MAP_FAILED))
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        fi_log(LOG_INFO, "%s: Attempted to map %u bytes.", __PRETTY_FUNCTION__, bar_length);
        close(pci_fd);
        pci_fd = -1;
        goto ssd_bail;
    }

    // Initialize our last frame value. Assumption is that we're starting at the beginning.
    last_im_address = pm_ptr->getCaptureFrameAddr(0);
    TP(LTRW_DEBUG, "%s: starting last_im_address=0x%" PRIx64, __PRETTY_FUNCTION__, last_im_address);
    leftover = 0;

    // Tell the FPGA when to declare the FIFO almost full.
    af_level = MIN_TRANSFER_SIZE / pm_ptr->getFrameSize();

    // At higher rates and small ROIs, it takes some time for the NUC to swing around and
    // respond to the AF interrupt. Pad this so we have a little more time.
    if (af_level > MAX_AF_LEVEL)
        af_level = MAX_AF_LEVEL;
    fpga_ptr->regWrite(IM2SSD_DMA0_LR_DESCR_FIFO_AFULL_LEVEL, af_level);
    TP(LTRW_DEBUG, "%s: setting AF=%d", __PRETTY_FUNCTION__, af_level);

    // Initialize some statistics counters.
    total_descr_count = 0;
    total_bytes_written = 0;

    // Tell the FPGA to pad out the "wrap" frame, i.e. the last frame that fits into the partition.
    fpga_ptr->regWrite(IM2SSD_DMA0_PADDING_INTERVAL, pm_ptr->ltrGetActiveFrameCount());
    TP(LTRW_DEBUG, "%s: frame to pad=%u", __PRETTY_FUNCTION__, pm_ptr->ltrGetActiveFrameCount());

    // Sanity check.
    if (pm_ptr->ltrGetActiveFrameCount() == 0)
    {
        fi_log(LOG_DEBUG, "%s: Why is get_ltr_active_frame_count() 0?!", __PRETTY_FUNCTION__);
        goto pci_bail;
    }

    // What is the address after the last frame in the partition.
    last_addr = pm_ptr->get_active_frame_addr(pm_ptr->ltrGetActiveFrameCount() - 1, true);
    last_addr += pm_ptr->getFrameSize();

    // What is the aligned address beyond the last frame?
    aligned_addr = (((last_addr + (FAS_FILESYSTEM::SSD_PAGE_LEN - 1)) / FAS_FILESYSTEM::SSD_PAGE_LEN) * FAS_FILESYSTEM::SSD_PAGE_LEN);

    // Now how many bytes are between the two.
    padding_count = (uint32_t)(aligned_addr - last_addr);
    padding_count_qw = convert_to_QWORD(padding_count);
    TP(LTRW_DEBUG, "%s: last frame addr=0x%" PRIx64 ", aligned_addr=0x%" PRIx64 ", padding_count=%d (%d)",
        __PRETTY_FUNCTION__, last_addr, aligned_addr, padding_count, padding_count_qw);

    // Clear out the descriptor level.
    fpga_ptr->regRead(IM2SSD_DMA0_LR_DESCR_FIFO_STATUS, &dummy);

    // Tell the FPGA how many bytes (in qwords) that padding will be.
    fpga_ptr->regWrite(IM2SSD_DMA0_PADDING_QWORD_COUNT, padding_count_qw);

    // Configure the DMA.
    fpga_ptr->regWrite(IM2SSD_DMA0_IM_RD_DMA_BLOCK_SIZE, convert_to_HEXWORD((uint64_t)pm_ptr->getFrameSize()));
    fpga_ptr->regWrite(IM2SSD_DMA0_IM_RD_DMA_BURST_SIZE, IM2SSD_DMA_REQ_SIZE_VAL);
    fpga_ptr->regWrite(IM2SSD_DMA0_IM_RD_DMA_BP_THRESHOLDS, IM2SSD_DMA_BP_THRESHOLD_VAL);

    // Tell the FPGA to start.
    reg_value = IM2SSD_DMA0_WORD_ORDER_R2L | IM2SSD_DMA0_BYTE_ORDER_L2R | IM2SSD_DMA0_ENABLE | IM2SSD_DMA0_LR_MODE_ENABLE;
    if (padding_count > 0)
        reg_value |= IM2SSD_DMA0_PADDING_ENABLE;
    fpga_ptr->regWrite(IM2SSD_DMA0_CONTROL, reg_value);

    // All is well.
    return true;

pci_bail:
    close(pci_fd);
    pci_fd = -1;

ssd_bail:
    // Make sure to close the drive.
    close(ssd_fd);
    ssd_fd = -1;

    return false;
}

void ltr_worker::stop()
{
    TP(LTRW_FUNC, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    stop_request = true;
    xfer_status.notify_all();
}

/*!
 * @brief This method looks to see if the FPGA has enough data queued to need a DMA transfer.
 *
 * The FPGA maintains a count of how many "descriptors" are queued in the FIFO. When this number
 * reaches (or exceeds) the "almost full" level, @ref af_level, it is time to start a DMA.
 *
 * We also check to see:
 *   - if the descriptor FIFO overflowed, in which case we return ENOSPC
 *   - if the user canceled the operation, stop_request is true
 *
 * @return True if there's data to DMA, false otherwise.
 */
bool ltr_worker::checkForEnoughData()
{
    uint32_t descr_level;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(LTRW_FUNC, "%s: leftover=%u", __PRETTY_FUNCTION__, leftover);

    // Sanity check.
    if (ssd_fd == -1)
    {
        fi_log(LOG_ERR, "%s: SSD isn't open!?", __PRETTY_FUNCTION__);
        error_cond = ENXIO;
        return false;
    }
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL fpga ptr!?", __PRETTY_FUNCTION__);
        error_cond = ENXIO;
        return false;
    }
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL partmgr ptr!?", __PRETTY_FUNCTION__);
        error_cond = ENXIO;
        return false;
    }

    // How many descriptors are pending.
    fpga_ptr->regRead(IM2SSD_DMA0_LR_DESCR_FIFO_STATUS, &descr_level);
    TP(LTRW_DEBUG, "%s: DESCR_FIFO=0x%08x, af_level=0x%08x", __PRETTY_FUNCTION__, descr_level, af_level);

    // Did we overflow?
    if ((descr_level & IM2SSD_DMA0_LR_DESCR_FIFO_OVFLOW) != 0)
    {
        fi_log(LOG_ERR, "%s: DMA overflowed! [DSCR_FIFO_STATUS=0x%08x]", __PRETTY_FUNCTION__, descr_level);
        error_cond = ENOSPC;
        return false;
    }

    // Is there a flush or stop request pending, or do we have enough in our FIFO for a transfer?
    if (flush_the_pipe || stop_request || ((descr_level & IM2SSD_DMA0_LR_DESCR_LEVEL_MASK) >= af_level))
    {
        // How many bytes (frames * frame_size) are in the FIFO?
        xfer_len = (descr_level & IM2SSD_DMA0_LR_DESCR_LEVEL_MASK) * pm_ptr->getFrameSize();
        TP(LTRW_DEBUG, "%s: xfer_len = %u [reg=0x%08x], total_bytes_written=%" PRIu64, __PRETTY_FUNCTION__, xfer_len, descr_level, total_bytes_written);

        // Is this our final write?
        if (!final_write && stop_request)
        {
            // Remember that we've done this.
            final_write = stop_request;

            // Adjust our lengths...but not if we've never seen any data.
            if (total_bytes_written > 0)
            {
                // The descriptor is always off by 1, so add that in now.
                xfer_len += pm_ptr->getFrameSize();
                total_descr_count++;
                TP(LTRW_DEBUG, "%s: Adding in last frame, xfer_len = %u", __PRETTY_FUNCTION__, xfer_len);
            }
            else
            {
                TP(LTRW_DEBUG, "%s: Never written, so bail out.", __PRETTY_FUNCTION__);
            }
        }

        // Sanity check.
        if (xfer_len == 0)
        {
            if (!stop_request && !flush_the_pipe)
            {
                fi_log(LOG_DEBUG, "%s: xfer_len == 0?!!", __PRETTY_FUNCTION__);
                error_cond = ENOMEM;
            }

            // Wait for something to show up in the FIFO.
            return false;
        }

        // Add in the bytes left over from the last time, if any.
        xfer_len += leftover;
        TP(LTRW_VERBOSE, "%s: Adding in leftover (%u), xfer_len = %u", __PRETTY_FUNCTION__, leftover, xfer_len);

        // Assume no leftovers.
        leftover = 0;

        // Reset our flush flag.
        flush_the_pipe = false;

        // Make sure we are on a page boundary.
        if ((xfer_len & (FAS_FILESYSTEM::SSD_PAGE_LEN - 1)) != 0)
        {
            uint32_t aligned_len;

            // Round the count downwards.
            aligned_len = ((xfer_len / FAS_FILESYSTEM::SSD_PAGE_LEN) * FAS_FILESYSTEM::SSD_PAGE_LEN);

            // The difference is the amount that will go into the next
            // SSD page.
            leftover = xfer_len - aligned_len;

            // Remember the aligned length. This is what we're going to
            // want to transfer.
            xfer_len = aligned_len;

            // Sanity check.
            if (leftover >= FAS_FILESYSTEM::SSD_PAGE_LEN)
            {
                fi_log(LOG_DEBUG, "%s: leftover (%u) is too large!!!", __PRETTY_FUNCTION__);
            }
        }

        total_descr_count += (descr_level & IM2SSD_DMA0_LR_DESCR_LEVEL_MASK);
        TP(LTRW_VERBOSE, "%s: Total descriptors=%u", __PRETTY_FUNCTION__, total_descr_count);

        // Sanity check. Let folks know if we need to catch up.
        if (xfer_len >= bar_length)
        {
            fi_log(LOG_WARNING, "%s: FPGA transfer size overflowed [%u >= %u]!", __PRETTY_FUNCTION__, xfer_len, bar_length);
        }

        // Get the data moving to the SSD.
        TP(LTRW_VERBOSE, "%s: Enough frames to DMA, xfer_len = %u, leftover=%u", __PRETTY_FUNCTION__, xfer_len, leftover);
        return true;
    }

    // Not enough to worry about.
    TP(LTRW_VERBOSE, "%s: Waiting on LR_DESCR_FIFO_STATUS.", __PRETTY_FUNCTION__);
    return false;
}

ssize_t ltr_worker::write(int fd, uint8_t *buf_ptr, ssize_t nbyte)
{
    ssize_t ret_value = 0;

    // Accumulate total bytes written.
    total_bytes_written += nbyte;

#ifdef LR_STATS
    Fastec::IntervalTimer write_timer;
    write_timer.start();
#endif /* LR_STATS */

    // Write out the data.
    while (nbyte > 0)
    {
        ssize_t written;

        // Write out the data.
        written = ::write(fd, buf_ptr, nbyte);
        if (written == (ssize_t) -1)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
            fi_log(LOG_ERR, "%s: fd=%d, buf_ptr=0x%p, nbyte=%" PRIu64, __PRETTY_FUNCTION__, fd, (void *)buf_ptr, nbyte);
            break;
        }
        TP(LTRW_VERBOSE, "%s: %" PRIu64 " bytes written", __PRETTY_FUNCTION__, written);

        // Accumulate the amount written.
        ret_value += written;

        // Adjust our pointer.
        buf_ptr += written;

        // Adjust the amount left to write.
        nbyte -= written;
    }

#ifdef LR_STATS
    if (stats_fp != NULL)
    {
        static double last_time(0.0);

        // Stop the time for the write operation.
        write_timer.stop();

        // Output our statistics.
        fprintf(stats_fp, "%lf,%lf,%lu,%lf,%lf\n", write_timer.lastMeasurement, (((double)ret_value) / (1024.0 * 1024.0)) / write_timer.lastMeasurement, ret_value, delay_for_data, ((delay_for_data > 0.0000001) ? (delay_for_data - last_time) : 0.0));

        // Save the time it takes to do the write so we can figure out the overhead. Note that this
        // does NOT account for multiple consecutive writes! You can see multiple writes happening
        // because the delay_for_data value will be 0.0.
        last_time = write_timer.lastMeasurement;
        delay_for_data = 0.0;
    }
#endif /* LR_STATS */

    return ret_value;
}

