/*!
 * @file luxima_2100.cxx
 * @brief This source file implements the sensor interface(s) for the
 *    Luxima 2100 sensor.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <syslog.h>

#include <chrono>
#include <string>
#include <memory>
#include <fmt/printf.h>
#include <fmt/format.h>

#include <fastec/fi_ioctl.h>

#include "bad_pixels.h"
#include "camera_server.h"
#include "gige_reg.h"
#include "sensor.h"
#include "luxima_hs.h"

namespace LUXIMA_2100_SENSOR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! @name The Luxima2100 register addresses.
//! @{

const uint8_t CHIP_INFO = 0x00;
#define GET_REV(x)              ((x) & 0xFF)
#define GET_ID(x)               (((x) >> 8) & 0xFF)
#define REV_2                   2
#define CHIP_ID                 0x28

const uint8_t SHUTTER = 0x01;
#define TIMING_EN               0x0001
#define GLOBAL_SHUTTER          0x0010
#define FTR_EN                  0x0100
#define FT_RST_EN               0x1000

const uint8_t ROI_OPTIONS = 0x02;
#define ROI_SEL_MASK            0x00FF
#define INTER_ROI_SP            0x0100
#define DRK_COL_RD              0x0200
#define VFLIP                   0x0400
#define HFLIP                   0x0800
#define CLK_SEL_FT_FULL         0x0000
#define CLK_SEL_FT_DIV2         0x1000
#define CLK_SEL_FT_DIV4         0x2000
#define CLK_SEL_FT_DIV8         0x3000
#define CLK_SEL_FT_MASK         0x3000

const uint8_t HBLANK = 0x03;
#define HBLANK_MIN              0x0002

const uint8_t DP_SCIP_EN = 0x04;
#define MAIN_PAGE               0x0000
#define DATAPATH_PAGE           0x0001

const uint8_t SHADOW_REGS = 0x05;

const uint8_t X_START = 0x06;
const uint8_t X_END = 0x07;
const uint8_t Y_START = 0x08;
const uint8_t Y_END = 0x09;
#define ROI_SHIFT               1

// Skipping the ROI registers, as we have no plans at this time to support ROI capture.

const uint8_t DRK_ROWS_ST_ADDR_TOP = 0x2A;
const uint8_t NB_DRK_ROWS_TOP = 0x2B;
const uint8_t NEXT_ROW = 0x2C;
#define NEXT_ROW_ADDR_OVR_MASK  0x0FFE
#define NEXT_ROW_OVR_EN         0x1000

const uint8_t SOF_DELAY = 0x30;
#define SOF_SHIFT               8
#define SET_SOF_DELAY(x)        ((uint16_t)(((x) & 0xFF) << SOF_SHIFT))

const uint8_t FT_TRIG_NB_PULSE = 0x31;
const uint8_t FT_RST_NB_PULSE = 0x32;
const uint8_t RDOUT_DLY = 0x34;
#define STANDARD_RDOUT_DELAY    0x42

const uint8_t ROW_TIME = 0x35;
const uint8_t ABN_CONTROL = 0x36;
#define ABN_SEL                 0x0001
#define ABN2_EN                 0x0010
#define ABN2_ALT_PAT            0x0100
#define ABN2_LD                 0x1000

const uint8_t X_ORG = 0x40;
const uint8_t Y_ORG = 0x41;

const uint8_t PCLK_LINEVALID = 0x52;
const uint8_t PCLK_VBLANK = 0x53;
const uint8_t PCLK_HBLANK = 0x54;
const uint8_t PCLK_OPTICAL_BLACK = 0x55;
const uint8_t OPT_CONTROL = 0x56;
#define COLOR_SEL               0x0000
#define MONO_SEL                0x0001
#define ROWBIN2                 0x0010
#define ROW2EN                  0x0020
#define COLBIN2                 0x0100
#define COLBIN4                 0x0200
#define POUTSEL_MASK            0x3000
#define POUTSEL_DEFAULT         0x2000
#define INVERT_ANALOG           0x8000

const uint8_t GAIN_SEL_SAMP = 0x57;
const uint8_t GAIN_SEL_CONTROL = 0x58;
#define GAIN_SEL_FB_MASK        0x007F
#define GAIN_SERIAL_MASK        0x0700

const uint8_t LV_DELAY = 0x5B;
const uint8_t TESTPAT_CONTROL = 0x5E;
#define CUST_PAT_MASK           0x0FFF
#define TST_PAT_MASK            0x7000
#define TST_PAT_VALID_DATA      0x0000
#define TST_PAT_DESKEW          0x1000
#define TST_PAT_SYNC            0x2000
#define TST_PAT_CUSTOM          0x3000

const uint8_t CHANNEL_CONTROL = 0x5F;
#define PWR_EN_SERIALIZER_B_MASK    0x000F
#define PWR_EN_CH0mod8_DISABLE      0x0001
#define PWR_EN_CH4mod8_DISABLE      0x0002
#define PWR_EN_CH2_6mod8_DISABLE    0x0004
#define PWR_EN_CH1_3_5_7_mod8_DISABLE   0x0008
#define MUX_MODE_MASK               0x0030
#define MUX_MODE_ALL                0x0000
#define MUX_MODE_16                 0x0010
#define MUX_MODE_8                  0x0020
#define MUX_MODE_4                  0x0030

//! @}

const int X_MIN = 32;                               //!< Left-most "active" pixel. The left-most "active" pixel address is 32.
const int X_MAX = 1983;                             //!< Right-most "active" pixel.
const int WIDTH_MAX = ((X_MAX + 1) - X_MIN);        //!< Total "active" width.
const int X_STEP = 16;                              //!< The stepping in the X direction.

const int Y_MIN = 8;                                //!< Bottom-most "active" pixel.
const int Y_MAX = 1087;                             //!< Top-most "active pixel.
const int HEIGHT_MAX = ((Y_MAX + 1) - Y_MIN);       //!< Total "active" height.

const int SENSOR_PCLK = 75000000;                   //!< Sensor's clock rate, in MHz.

// From the datasheet, pg 10, 7.1 "Exposure Timing"
const double    Tclk = (1.0 / SENSOR_PCLK);         //!< 13.33 ns
const double    Thblank = (2 * Tclk);               //!< "Horizontal blank" period, value from the Luxima framerate calculator spreadsheet.
const int       Twavetable = 66;                    //!< Value from the Luxima framerate calculator spreadsheet.

const int       NUM_CHANNELS = 32;                  //!< The number of channels.
const int       LINE_READ = 1;                      //!< From the Luxima framerate calculator spreadsheet.
const double    TXN = 500e-9;                       //!< The amount of time for the TXN pulse (nsec). Value from the Luxima framerate calculator spreadsheet.
const double    TXLN = 300e-9;                      //!< Value from the Luxima framerate calculator spreadsheet.
const double    NON_OVERLAP = 100e-9;               //!< Value from the Luxima framerate calculator spreadsheet.
const double    SOF = 200e-9;                       //!< Value from the Luxima framerate calculator spreadsheet.

const int       MIN_ROW = (Twavetable + 3);         //!< From the Luxima framerate calculator spreadsheet.
const double    VBLANK = (TXN + TXLN + NON_OVERLAP + SOF);

//! Our table of chip, DAC, value triplets.
//! To make things easier to read, the values are as computed in the spreadsheet,
//! but they need to be left justified before writing them out.
const LUXIMA_DAC::DAC_TRIPLES lux2100_dac_settings[] =
{
    {DAC_GLOBAL_I2C_ADDR, (LUXIMA_DAC::CC_UPDATE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_GLOBAL, 0x0000},
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

} /* End of namespace LUXIMA_2100_SENSOR */

using namespace LUXIMA_DAC;
using namespace LUXIMA_2100_SENSOR;

//--------------------------------------------------------------------
// Luxima2100 Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima2100::Luxima2100(SENSOR::EEPROM_INFO *eeprom)
    : LuximaHS(
          1952,  /* max width */
          1080,  /* max height */
          64,  /* min width */
          64,  /* min height */
          eeprom)
{
    LOG("%s", __PRETTY_FUNCTION__);

    config_word = 0;
    current_gain = 0;
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima2100::~Luxima2100()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // Turn off the sensor board.
    shutdown_sensor(fpga_ptr);
}

void Luxima2100::bit_slip()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

bool Luxima2100::class_init(FiFPGA *fpga_ptr)
{
    uint32_t reg_value;

    LOG("%s", __PRETTY_FUNCTION__);

    // Turn off the sensor board.
    if (!shutdown_sensor(fpga_ptr))
    {
        fi_log(LOG_DEBUG, "%s: Unable to shutdown the sensor board.", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait a bit.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(50));

    // Tell the FPGA to assert POWER_EN for the sensor board.
    if (!fpga_ptr->regModify(NC1_FPGA_CONTROL, NC1FC_ISB_PWR_ENABLE, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable power for the sensor board!", __PRETTY_FUNCTION__);
        return false;
    }
    LOG("%s: ISB_PWR_ENABLE. ========================", __PRETTY_FUNCTION__);

    // Wait a little bit to give things a chance to settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(15));

    // Read the power status register to verify that the POWER_OK
    // signal is asserted.
    if (!fpga_ptr->regRead(NC1_FPGA_STATUS, &reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read the status register!", __PRETTY_FUNCTION__);
        return false;
    }
    if ((reg_value & NC1FS_ISB_PWR_OK) != NC1FS_ISB_PWR_OK)
    {
        fi_log(LOG_ERR, "%s: Unable to power up sensor board! [status = 0x%08x]!", __PRETTY_FUNCTION__, reg_value);
        return false;
    }

    // Load any bad pixel information we might have.
    load_bad_pixels();
    if (bad_pixels.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: load_bad_pixels() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize the DACs
    if (!dac_init(fpga_ptr))
    {
        fi_log(LOG_ERR, "%s: Unable to initialize the DACs on the sensor board!", __PRETTY_FUNCTION__);
        return false;
    }

    // Bring the sensor out of reset.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_SENSOR_RESET))
    {
        fi_log(LOG_ERR, "%s: Unable to deassert sensor reset!", __PRETTY_FUNCTION__);
        return false;
    }

    // Enable the PLLs.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_SENSOR_PLL_ENA, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable sensor PLLs!", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait a while to let the PLL settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(1));

    // All is well.
    return true;
}

bool Luxima2100::dac_init(FiFPGA *fpga_ptr)
{
    int i;
    bool rc(false);

    LOG("%s", __PRETTY_FUNCTION__);

    // Loop through our table of commands.
    for (i = 0; i < (int)(sizeof(lux2100_dac_settings) / sizeof(lux2100_dac_settings[0])); ++i)
    {
        uint8_t chip_addr = lux2100_dac_settings[i].chip;
        uint8_t reg_addr = lux2100_dac_settings[i].reg;
        uint16_t reg_value = lux2100_dac_settings[i].val << DT_SHIFT;

        // fi_log(LOG_DEBUG, "%02X[%02X] <= 0x%04X [0x%04X << %d]", chip_addr, reg_addr, reg_value, lux2100_dac_settings[i].val, DT_SHIFT);

        // Write the DAC setting. NOTE: This is a 2-byte write.
        if (!fpga_ptr->i2cWrite(chip_addr, reg_addr, reg_value, FI_FPGA::I2C_SENSOR, 2))
            goto err_exit;
    }

    // Everything is OK.
    rc = true;

err_exit:
    if (!rc)
        fi_log(LOG_ERR, "%s: I/O error", __PRETTY_FUNCTION__);

    // Return our status.
    return rc;

}

void Luxima2100::enable_clpout()
{
    // Not going to enable clpout for the 2100.
}

bool Luxima2100::enable_sensor_test()
{
    LOG("%s", __PRETTY_FUNCTION__);
    return false;
}

/*!
 * @brief Return the amount of time (in microseconds) the FPGA/Sensor requires to
 *    move a frame, etc. AKA "Frame Overhead".
 */
uint32_t Luxima2100::get_frame_overhead()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // VBLANK is in nanoseconds, so convert to microseconds and round up.
    return (uint32_t)ceil(VBLANK / 1000.0);
}

/*!
 * @brief Calculate the maximum frame rate for a given (X, Y)(WIDTH, HEIGHT).
 *
 * @param start_x   The starting X position [0..max_width - 1]
 * @param width     The width of the ROI.
 * @param height    The height of the ROI.
 *
 * @return The maximum frame rate for the specified ROI.
 */
float Luxima2100::get_max_frame_rate(int start_x, int /*start_y*/, int width, int height)
{
    float frame_rate;
    int num_cols;
    int readout_startx;
    int readout_width;
    double Tdata;
    double Tread;
    double Tframe;
    CameraServer *cs_ptr = CameraServer::get_cs();

    LOG("%s: start_x=%d, width=%d, height=%d", __PRETTY_FUNCTION__, start_x, width, height);

    // Where ACTUALLY is our starting X? Need to round DOWN to the nearest PPW.
    readout_startx = (start_x / X_STEP) * X_STEP;

    // Where ACTUALLY does our ROI end? Need to round UP to the nearest PPW.
    readout_width = ((start_x + width + X_STEP + 1) / X_STEP) * X_STEP;
    if (readout_width > WIDTH_MAX)
    {
        fi_log(LOG_DEBUG, "%s: readout_width [%d] > WIDTH_MAX!", __PRETTY_FUNCTION__, readout_width);
        readout_width = WIDTH_MAX;
    }

    // How many columns are we talking about?
    num_cols = readout_width - readout_startx;

    // Calculate our Tdata.
    Tdata = ceil(((double)num_cols) * ((double)LINE_READ) / ((double)NUM_CHANNELS)) * Tclk;

    // Calculate our Tread now. Make sure we're greater than or equal to the MIN_ROW value.
    Tread = Tdata + Thblank;
    if ((MIN_ROW * Tclk) > Tread)
        Tread = MIN_ROW * Tclk;

    // Calculate the time for a frame.
    Tframe = (Tread * ceil((double)height / (double)LINE_READ)) + VBLANK;

    // Calculate the maximum frame rate based on the sensor dimensions. Intentionally
    // round down, so the frame rate is within bounds.
    frame_rate = (1.0 / Tframe);

    // Make sure our maximum rate fits within the limitations of what is allowed.
    if (frame_rate > (float)cs_ptr->get_max_licensed_frame_rate())
        frame_rate = (float)cs_ptr->get_max_licensed_frame_rate();

    // Return the rate.
    return frame_rate;
}

uint32_t Luxima2100::get_readout_delay()
{
    LOG("%s", __PRETTY_FUNCTION__);
    return 0;
}

int Luxima2100::get_readout_delay_min()
{
    LOG("%s", __PRETTY_FUNCTION__);
    return 0;
}

int Luxima2100::get_readout_delay_max()
{
    LOG("%s", __PRETTY_FUNCTION__);
    return 0;
}

/*!
 * @brief Return the minimum allowed shutter, in microseconds. To say this another way,
 *    return the fastest shutter period we can support.
 */
uint32_t Luxima2100::get_sensor_min_shutter()
{
    LOG("%s", __PRETTY_FUNCTION__);
    return 0;
}

/*!
 * @brief Perform hard-reset and required initialization
 *
 * @return true if everything goes well, false otherwise.
 */
void Luxima2100::hard_reset()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure the ICM is disabled.
    if (!fpga_ptr->icmDisable())
    {
        fi_log(LOG_ERR, "%s: turn_off_icm() failed!", __FUNCTION__);
        return;
    }
}

/*!
 * @brief Set up the FPGA's IPM to do bad pixel correction.
 *
 * @param port      [unused]
 * @param true_x    current ROI "true" x coordinate, relative to the sensor's actual size
 * @param true_y    current ROI "true" y coordinate, relative to the sensor's actual size
 * @param width     current ROI width
 * @param height    current ROI height
 */
void Luxima2100::IPM_bad_pixel_correction(int port, uint16_t true_x, uint16_t true_y,
                                          uint16_t width, uint16_t height)
{
    (void)port;

    if (bad_pixels.get() != nullptr)
    {
        LOG("%s: true_x=%d, true_y=%d, width=%d, height=%d", __PRETTY_FUNCTION__, true_x, true_y, width, height);

        bad_pixels->set_roi(true_x, true_y, width, height);
        bad_pixels->correct();
    }
    else
    {
        LOG("%s: bad_pixels is NULL", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Check to see if a particular bit mode/depth is supported by this sensor.
 *
 * @param bit_mode  The bit mode/depth.
 * @return True if the sensor supports, false otherwise.
 */
bool Luxima2100::is_bit_mode_supported(bitmode_t bit_mode)
{
    bool rc = false;

    LOG("%s: bit_mode=0x%x", __PRETTY_FUNCTION__, bit_mode);

    switch ((int)bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_MIDDLE_8_BITS:
    case GIGE_BITMODE_8_1:
    case GIGE_BITMODE_UPPER_8_BITS:
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_10_3:
    case GIGE_BITMODE_11_4:
    case GIGE_BITMODE_10_BITS:
    case GIGE_BITMODE_9_0:
    case GIGE_BITMODE_10_1:
    case GIGE_BITMODE_11_2:
    case GIGE_BITMODE_11_0:
        rc = true;
        break;
    default:
        break;
    }

    return rc;
}

bool Luxima2100::is_option_supported(uint32_t /*options*/)
{
    // There are no options supported at this time.
    LOG("%s", __PRETTY_FUNCTION__);
    return false;
}

bool Luxima2100::make_sensor_valid(
    uint16_t *rtn_startx,
    uint16_t *rtn_starty,
    uint16_t *rtn_width,
    uint16_t *rtn_height,
    bitmode_t *rtn_bit_mode,
    float *rtn_frame_rate,
    uint32_t * /* rtn_shutter_speed */,
    uint32_t *rtn_sensor_opts)
{
    bitmode_t bit_mode = *rtn_bit_mode;
    uint16_t startx = *rtn_startx;
    uint16_t starty = *rtn_starty;
    uint16_t width = *rtn_width;
    uint16_t height = *rtn_height;
    uint32_t sensor_opts = *rtn_sensor_opts;
    uint32_t any_changes = 0;
    float max_frame_rate;

    LOG("%s", __PRETTY_FUNCTION__);

    // We support all the bit modes so this should never happen.
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Invalid bit_mode, using 0x%02x", __FUNCTION__, GIGE_BITMODE_11_4);
        *rtn_bit_mode = (bitmode_t)GIGE_BITMODE_11_4;
        any_changes |= (1 << 0);
    }

    // Don't leave the bit mode setting in one of the old-style values.
    switch (bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
        bit_mode = GIGE_BITMODE_7_0;
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
        bit_mode = GIGE_BITMODE_9_2;
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
        bit_mode = GIGE_BITMODE_11_4;
        break;
    case GIGE_BITMODE_10_BITS:
        bit_mode = GIGE_BITMODE_11_2;
        break;
    default:
        break;
    }

    // There are no sensor options supported at this time.
    if (sensor_opts && (sensor_opts != ((1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT))))
    {
        *rtn_sensor_opts = 0;
        any_changes |= (1 << 1);
    }

    // Make sure our ROI is on legal boundaries.
    (void)force_legal_roi(1, &startx, &starty, &width, &height);
    if ((startx != *rtn_startx) ||
        (starty != *rtn_starty) ||
        (width  != *rtn_width) ||
        (height != *rtn_height))
    {
        // fi_log(LOG_DEBUG, "%s: was: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );
        *rtn_startx = startx;
        *rtn_starty = starty;
        *rtn_width = width;
        *rtn_height = height;
        // fi_log(LOG_DEBUG, "%s: now: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );

        any_changes |= (1 << 2);
    }

    // Make sure the frame rate is OK.
    max_frame_rate = get_max_frame_rate(startx, starty, width, height);
    if (*rtn_frame_rate > max_frame_rate)
    {
        *rtn_frame_rate = max_frame_rate;
        any_changes |= (1 << 3);
    }

    if (any_changes != 0)
        fi_log(LOG_DEBUG, "%s: changed 0x%08x", __PRETTY_FUNCTION__, any_changes);

    // Return true if we had changes.
    return (any_changes != 0);
}

int Luxima2100::sensor_clock_rate() const
{
    return SENSOR_PCLK;
}

bool Luxima2100::set_integration_time(uint32_t shutter_speed)
{
    // All is well.
    LOG("%s: shutter_speed=%d", __PRETTY_FUNCTION__, shutter_speed);
    return false;
}

/*!
 * @brief Set the sensor gain.
 *
 * @param new_sensor_gain   The value to set in canonical form, 100 for 1x, 200 for 2x, 400 for 4x.
 * @return  -1 if value is illegal
 *       0 if it was good but needs an update to go into effect
 *       1 if value is good and was put into effect.
 */
int Luxima2100::set_sensor_gain(uint32_t new_sensor_gain)
{
    int rtn = 1;
    uint16_t reg_57;
    uint16_t reg_58;

    LOG("%s: new_sensor_gain=0x%08x", __PRETTY_FUNCTION__, new_sensor_gain);

    // Mask of the "immediate" action bit.
    new_sensor_gain &= ~GIGE_SENSOR_GAIN_IMMEDIATE;

    // Any change in the gain?
    if (new_sensor_gain == current_gain)
        return 1;

    // Figure out what the registers need to be set to.
    switch (new_sensor_gain)
    {
    case 100:
        // Gain 1.0, we'll go 4/4.
        reg_57 = 0x0000;    // gain_sel_samp[11:0] = 0 1s, +4 makes the numerator 4.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 150:
        // Gain 1.5, we'll go 6/4.
        reg_57 = 0x0003;    // gain_sel_samp[11:0] = 3 -> 2 1s, +4 makes the numerator 6.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 200:
        // Gain 2.0, we'll go 8/4.
        reg_57 = 0x000F;    // gain_sel_samp[11:0] = 0xF -> 4 1s, +4 makes the numerator 8.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 250:
        // Gain 2.5, we'll go 10/4.
        reg_57 = 0x003F;    // gain_sel_samp[11:0] = 0x3F -> 6 1s, +4 makes the numerator 10.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 300:
        // Gain 3.0, we'll go 12/4.
        reg_57 = 0x00FF;    // gain_sel_samp[11:0] = 0xFF -> 8 1s, +4 makes the numerator 12.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 350:
        // Gain 3.5, we'll go 14/4.
        reg_57 = 0x03FF;    // gain_sel_samp[11:0] = 0x3FF -> 10 1s, +4 makes the numerator 14.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 400:
        // Gain 4.0, we'll go 16/4.
        reg_57 = 0x0FFF;    // gain_sel_samp[11:0] = 0xFFF -> 12 1s, +4 makes the numerator 16.
        reg_58 = 0x0307;    // sel_samp_fb[6:0] = 0x07 -> 3 1s, +1 makes the denominator 4.
        break;
    case 500:
        // Gain 5.0, we'll go 5/1.
        reg_57 = 0x0001;    // gain_sel_samp[11:0] = 0x001 -> 1 1s, +4 makes the numerator 5.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 600:
        // Gain 6.0, we'll go 6/1.
        reg_57 = 0x0003;    // gain_sel_samp[11:0] = 0x003 -> 2 1s, +4 makes the numerator 6.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 700:
        // Gain 7.0, we'll go 7/1.
        reg_57 = 0x0007;    // gain_sel_samp[11:0] = 0x007 -> 3 1s, +4 makes the numerator 7.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 800:
        // Gain 8.0, we'll go 8/1.
        reg_57 = 0x000F;    // gain_sel_samp[11:0] = 0x00F -> 4 1s, +4 makes the numerator 8.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 900:
        // Gain 9.0, we'll go 9/1.
        reg_57 = 0x001F;    // gain_sel_samp[11:0] = 0x01F -> 5 1s, +4 makes the numerator 9.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1000:
        // Gain 10.0, we'll go 10/1.
        reg_57 = 0x003F;    // gain_sel_samp[11:0] = 0x03F -> 6 1s, +4 makes the numerator 10.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1100:
        // Gain 11.0, we'll go 11/1.
        reg_57 = 0x007F;    // gain_sel_samp[11:0] = 0x07F -> 7 1s, +4 makes the numerator 11.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1200:
        // Gain 12.0, we'll go 12/1.
        reg_57 = 0x00FF;    // gain_sel_samp[11:0] = 0x0FF -> 8 1s, +4 makes the numerator 12.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1300:
        // Gain 13.0, we'll go 13/1.
        reg_57 = 0x01FF;    // gain_sel_samp[11:0] = 0x1FF -> 9 1s, +4 makes the numerator 13.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1400:
        // Gain 14.0, we'll go 14/1.
        reg_57 = 0x03FF;    // gain_sel_samp[11:0] = 0x3FF -> 10 1s, +4 makes the numerator 14.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1500:
        // Gain 15.0, we'll go 15/1.
        reg_57 = 0x07FF;    // gain_sel_samp[11:0] = 0x7FF -> 11 1s, +4 makes the numerator 15.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    case 1600:
        // Gain 16.0, we'll go 16/1.
        reg_57 = 0x0FFF;    // gain_sel_samp[11:0] = 0xFFF -> 12 1s, +4 makes the numerator 16.
        reg_58 = 0x0300;    // sel_samp_fb[6:0] = 0x00, +1 makes the denominator 1.
        break;
    default:
        rtn = -1;
        break;
    }

    // Everything still OK?
    if (rtn >= 0)
    {
        // =========================================================================
        // ALM_TODO: Write out the register values for the gain.
        LOG("%s: reg_57=0x%04x, reg_58=0x%04x", __PRETTY_FUNCTION__, reg_57, reg_58);
        // =========================================================================
    }

    return rtn;
}

/*!
 * @brief This routine is used to bring the sensor up in a controlled, managed fashion.
 *
 * The assumption is that the sensor board is shut down/powered off. This follows (roughly)
 * the description in the spec, 9.0 "BRING-UP SEQUENCE".
 *
 * @param startx    Starting x position in pixels, sensor-relative
 * @param starty    Starting y position in pixels
 * @param width     Width in pixels to capture
 * @param height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param shutter   Integration time (exposure) in microseconds
 * @param opts      Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100)
 *
 * @return True if everything goes well, false otherwise.
 */
bool Luxima2100::sensor_startup(
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    fpnmode_t fpn_mode,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t shutter,
    uint32_t opts,
    uint32_t sensor_gain)
{
    uint16_t reg_data;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to get fpga_ptr!", __PRETTY_FUNCTION__);
        return false;
    }

    // Enable the power supplies was complete in class_init(). We'll do a hard reset to
    // make sure we're at the right place in the process.
    hard_reset();

    // Apply power up sequence. 7.4 "Power-Up Sequence", pg. 13.
    // ALM_TODO: This might be handled by the FPGA due the real time concerns involved?

    // Confirm register read/write.
    read_reg(CHIP_INFO, &reg_data);
    LOG("%s: CHIP_INFO=0x%04x", __PRETTY_FUNCTION__, reg_data);
    if ((GET_REV(reg_data) != REV_2) || (GET_ID(reg_data) != CHIP_ID))
    {
        // Just log this for now.
        fi_log(LOG_DEBUG, "%s: CHIP_INFO [0x%04x] isn't right!", __PRETTY_FUNCTION__, reg_data);
    }

    // Soft reset.
    write_reg(0x7E, 0x0000);                // Per data sheet, step 3.c "Perform soft reset..."

    // Gain setting.
    set_sensor_gain(sensor_gain);

    // Set up the sensor for the 66 Tclk wavetable. Straight from the datasheet.
    write_reg(RDOUT_DLY, STANDARD_RDOUT_DELAY);

    // Note: The documentation says that this must be >= lv_delay + 2. lv_delay is register 0x5B
    //       and we'll be setting that in step 6.
    write_reg(SOF_DELAY, SET_SOF_DELAY(0x0A));

    if (is_sensor_color())
        write_reg(OPT_CONTROL, COLOR_SEL);
    else
        write_reg(OPT_CONTROL, MONO_SEL);

    write_reg(X_START, X_MIN << ROI_SHIFT);
    write_reg(X_END, X_MAX << ROI_SHIFT);
    write_reg(Y_START, Y_MIN << ROI_SHIFT);
    write_reg(Y_END, Y_MAX << ROI_SHIFT);

    // Internal Control registers to fine tune the sensor.
    write_reg(LV_DELAY, 0x0008);
    write_reg(CHANNEL_CONTROL, 0);

    // 7. Training Sequence.

    // 8. Return to valid data mode.

    // 9. Program Sensor SRAM for timing control.

    // 10. Enable timing engine.

    // 11. Check if frame timing is correct.

    // 12. Check to see if there is light sensitivity.

    return update_capture(startx, starty, width, height, fpn_mode, bit_mode, frame_rate, shutter, opts, sensor_gain);
}

/*!
 * @brief Shutdown the Lux2100 sensor board.
 * @param fpga_ptr  Pointer to our FPGA object.
 * @return True if everything worked out fine, false otherwise.
 */
bool Luxima2100::shutdown_sensor(FiFPGA *fpga_ptr)
{
    uint32_t reg_value;

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is NULL?!", __PRETTY_FUNCTION__);
        return false;
    }

    // Put the sensor interface into reset.
    reg_value = ICMX13HSC_ALTLVDS_RX_RESET | ICMX13HSC_SENSOR_RESET |
                ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT | ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP;
    fpga_ptr->regWrite(ICM_LUX13HS_CONTROL, reg_value);

    // Sometimes the camera doesn't exit nicely. We need to make sure we've
    // reset the board.
    //
    // Shutdown the DACs
    fpga_ptr->i2cWrite(DAC_GLOBAL_I2C_ADDR, CC_POWERDOWN_ALL, 0, FI_FPGA::I2C_SENSOR, 2);

    // Turn off power to the sensor board.
    fpga_ptr->regModify(NC1_FPGA_CONTROL, 0, ~NC1FC_ISB_PWR_ENABLE);

    // All done.
    return true;
}

bool Luxima2100::update_capture(
    uint16_t true_startx,
    uint16_t true_starty,
    uint16_t roi_width,
    uint16_t roi_height,
    fpnmode_t /* fpn_mode */,
    bitmode_t bit_mode,
    float /*frame_rate*/,
    uint32_t /*integration*/,
    uint32_t /* sensor_opts */,
    uint32_t /* sensor_gain */)
{
    bool rc(false);

    LOG("%s: %ux%u@(%u,%u)", __PRETTY_FUNCTION__, roi_width, roi_height, true_startx, true_starty);

    // Need to make sure our startx is multiples of 2 and starty is multiples of 4.
    if ((true_startx & 1) || (true_starty & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! startx=0x%x, starty=0x%x", __PRETTY_FUNCTION__, true_startx, true_starty);
        true_startx = (true_startx + 1) & ~1;
        true_starty = (true_starty + 3) & ~3;
    }
    if ((roi_width & 1) || (roi_height & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! width=0x%x, height=0x%x", __PRETTY_FUNCTION__, roi_width, roi_height);
        roi_width = (roi_width + 1) & ~1;
        roi_height = (roi_height + 3) & ~3;
    }
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_DEBUG, "%s: Invalid bit_mode=0x%x", __PRETTY_FUNCTION__, bit_mode);
        bit_mode = GIGE_BITMODE_11_4;
    }

    return rc;
}

bool Luxima2100::read_reg(uint8_t reg_addr, uint16_t *rtn_data)
{
//    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    *rtn_data = 0;
    LOG("%s: reg_addr=0x%02x, data=0x%04x", __PRETTY_FUNCTION__, reg_addr, *rtn_data);
    return false;
}

bool Luxima2100::write_reg(uint8_t reg_addr, uint16_t reg_data)
{
    //    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    LOG("%s: reg_addr=0x%02x, data=0x%04x", __PRETTY_FUNCTION__, reg_addr, reg_data);
    return false;
}

bool Luxima2100::write_wavetable(const std::string& file_name)
{
    bool rc = false;
    FILE *fp = nullptr;
    std::string temp_path;
    std::unique_ptr<uint16_t[]> buf;
    //    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    const int MAX_TRANSFER_LEN = 512;       // ALM_TODO: Just a place holder for how big a buffer we'll need.

    // Build up the path.
    temp_path = fmt::format("{0}/{1}", get_home_path(), file_name);
    LOG("%s: file='%s'", __PRETTY_FUNCTION__, temp_path.c_str());

    // Allocate our buffer.
    buf = std::unique_ptr<uint16_t[]>(new uint16_t[MAX_TRANSFER_LEN]());
    if (buf.get() == nullptr)
        goto bail;

    // Open the wave table file.
    fp = fopen(temp_path.c_str(), "r");
    if (fp == nullptr)
        goto bail;

    // ------------------------------------------------------------------------
    // ALM_TODO:
    // Loop through the file, reading the data and writing it out to the FPGA.
    // ------------------------------------------------------------------------

bail:
    // ALM_TODO: This shows as false because we haven't implemented the logic yet.
    if (rc == false)
        fi_log(LOG_ERR, "%s: failed err=%d (%s)", __FUNCTION__, errno, strerror(errno));
    if (fp != nullptr)
        fclose(fp);

    return rc;
}

