/*!
 * @file luxima_hs.cxx
 * @brief This source file implements the sensor interface(s) for the
 *    Luxima 13HS, 19HS, and 160 sensors.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <syslog.h>

#include <chrono>
#include <memory>
#include <string>

#include <fmt/printf.h>
#include <json/json.h>

#include <fastec/fi_ioctl.h>

#include "luxima_hs.h"

#include "bad_pixels.h"
#include "camera_server.h"
#include "gige_reg.h"
#include "image_correction.h"
#include "sensor.h"

namespace LUXIMA_HS_SENSOR
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// The temperature sensor on the Luxima sensor board is the
// MCP98244, and is physically combined with the EEPROM.
#define MCP_I2C_ADDR            0x18

// Register address within the MCP98244 temperature sensor.
#define MCP_SPDREG_CAPABILITY       0
#define MCP_SPDREG_CONFIG           1
#define MCP_SPDREG_EVENT_TEMP_H     2
#define MCP_SPDREG_EVENT_TEMP_L     3
#define MCP_SPDREG_CRIT_TEMP        4
#define MCP_SPDREG_TEMPERATURE      5
#define MCP_SPDREG_MANUFACTURER     6
#define MCP_SPDREG_DEV_ID           7

//! @name Bit Definitions for LUX HS Sensors
//! @{

//! A '1' disables pixel read and enables column test with VTST-AGND as an input signal. Default: 0.
#define COL_TEST_EN         (1 << 0)

//! A '1' makes the offset negative, '0' positive. The offset value is contrlled with VOFF voltage.
#define OFFSET_SIGN         (1 << 1)

//! A '1' enables gain X2.
#define GAIN_X2_EN          (1 << 2)

//! A '1' enables another gain X2 through reducing amp feedback.
#define GAIN_BIT1_EN            (1 << 3)

//! With bit3 set to '1', this enables another gain X2 through reducing amp feedback capacitance further.
#define GAIN_BIT2_EN            (1 << 4)

//! Lux13HS: A '1' disconnects dummy differential circuit.
#define UNIPOLAR_ADC            (1 << 5)

//! Lux19HS: A '1' extends in-row TX pulse. Note that the Lux13HS has
//! the EXTEDED_TX in bit15.
#define EXTENDED_TX_19HS        (1 << 5)

//! A '1' increases time for comparator decision, decreases for signal settling.
#define LONG_LATCH          (1 << 6)

//! 13HS: A '1' enables change in row select signal; may affect FPN.
#define ROW_CUT_DIS         (1 << 7)

//! Lux160 & Lux19HS: Enables internal filter for AMP bias.
#define SH_AMP              (1 << 7)

//! 13HS: Enables 149 column read versus 162 column default. 19HS: UNUSED, keep 0.
#define SHORT_X_WINDOW          (1 << 8)

//! VLN cascode enable; '1' disables the cascoding.
#define VLN_CAS_DIS         (1 << 9)

//! Overrides internal biases.
#define VMUX1_MASK          (3 << 10)
#define     VCASNA          (0 << 10)
#define     VCASVLN         (1 << 10)
#define     VCASN2          (2 << 10)
#define     VCASN3          (3 << 10)

//! Overrides internal biases VLN3(0) and VLNT(1).
#define VMUX2               (1 << 12)

//! Enables internal filter for VLN bias.
#define SH_VLN              (1 << 13)

//! Enables the clamp for pixel driver.
#define C1_POUT             (1 << 14)

//! Lux13HS: A '1' extends in-row TX pulse. Note that the Lux19HS has
//! the EXTEDED_TX in bit5.
#define EXTENDED_TX_13HS        (1 << 15)

//! Lux160 & Lux19HS: A '1' enables chess-board test pattern.
#define TEST_PATTERN            (1 << 15)

//! A '0' is power enable for LVDS driver, a '1' disables.
#define PWREN_LVDS_DIS          (1 << 16)

//! A '0' enables current DACs for LVDS, a '1' disables.
#define PWREN_BIAS_DIS          (1 << 17)

//! A '0' enables LSB first, a '1' is MSB first.
#define MSB_FIRST           (1 << 18)

//! Specify the LVDS current value.
#define LVDS_CURRENT_MASK       (7 << 19)
#define     LVDS_2_5mA      (1 << 19)
#define     LVDS_3_5mA      (0 << 19)
#define     LVDS_4_5mA      (2 << 19)
#define     LVDS_6_0mA      (6 << 19)

//! A '0' disables on-chip termination. '1' enables.
#define ONCHIP_TERMINATION      (1 << 22)

//! Shift pixels to the next column for the bottom read.
#define SHIFT_PIX_BOT           (1 << 23)

//! Remove shift of pixels to the previous column; Top read.
#define SHIFT_PIX_TOP_N         (1 << 24)

//! @}


// Our table of chip, dac, value triplets.
// To make things easier to read, the values are as computed in the spreadsheet,
// but they need to be left justified before writing them out.
const LUXIMA_DAC::DAC_TRIPLES lux13hs_dac_settings[] =
{
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_SEL_INT_REF << LUXIMA_DAC::CC_SHIFT), 0x0000},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VREF, 0x012C},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLNC, 0x00E1},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLN, 0x0096},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VADH, 0x00E1},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VOFF, 0x007D},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VAD4, 0x0038},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLNA, 0x00ED},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VCAS, 0x015E},

    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_SEL_INT_REF << LUXIMA_DAC::CC_SHIFT), 0x0000},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VMUX1, 0x00E1},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTX2L_VTXM, 0x00AF},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VABL_TST, 0x004B},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VPIX_VRSTH, 0x01C2},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VRSTH_VRSTL, 0x0338},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTXH, 0x0338},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTXL, 0x0000},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VMUX2, 0x00E1},

    {DAC_GLOBAL_I2C_ADDR, (LUXIMA_DAC::CC_UPDATE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_GLOBAL, 0x0000},
};


const LUXIMA_DAC::DAC_TRIPLES lux19hs_dac_settings[] =
{
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_SEL_INT_REF << LUXIMA_DAC::CC_SHIFT), 0x0000},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VREF, 0x012C},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLNC, 0x00E1},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLN, 0x0089},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VADH, 0x00E1},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VOFF, 0x00AF},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VAD4, 0x0038},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VLNA, 0x00ED},
    {DAC_U12_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VCAS, 0x015E},

    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_SEL_INT_REF << LUXIMA_DAC::CC_SHIFT), 0x0000},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VMUX1, 0x00E1},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTX2L_VTXM, 0x00AF},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VABL_TST, 0x0032},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VPIX_VRSTH, 0x01C2},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VRSTH_VRSTL, 0x0338},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTXH, 0x0338},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VTXL, 0x0000},
    {DAC_U14_I2C_ADDR, (LUXIMA_DAC::CC_WRITE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_VMUX2, 0x00E1},

    {DAC_GLOBAL_I2C_ADDR, (LUXIMA_DAC::CC_UPDATE_ONE << LUXIMA_DAC::CC_SHIFT) | DAC_GLOBAL, 0x0000},
};

// From the datasheet (pg 26), we have the following recommended settings:
// UNIPOLAR_ADC
// LONG_LATCH
// ROW_CUT_DIS
// SH_VLN
// LVDS_3_5mA
// ONCHIP_TERMINATION off
// MSB_FIRST [added for Bob]
const uint32_t LUX13HS_RECOMMENDED = 0x0420E0;

// From the datasheet (pg 28), we have the following recommended settings:
// OFFSET_SIGN (b1)
// LONG_LATCH (b6)
// SH_AMP (b7)
// SH_VLN (b13)
// LVDS_3_5mA (b19:21)
// MSB_FIRST [added for Bob] (b18)
const uint32_t LUX19HS_RECOMMENDED = 0x0420C2;

// There are no recommended register settings for the Lux160 at this time.
// I just picked the same settings as the Lux19HS.
const uint32_t LUX160_RECOMMENDED = 0x080262;

// When the FPGA decides it has a start frame.
const int DEFAULT_READOUT_START = 8;
//const int PD_OVERLAP = 8;
const int PD_OVERLAP = 4;
const int PRE_READ_GAP = 4;

// Readout time for one row
const int LUX13HS_DEFAULT_QROW_TIME = 168;
//const int LUX13HS_DEFAULT_QROW_TIME = 180;
const int LUX19HS_DEFAULT_QROW_TIME = 128;

// Countdown clock value for when to end the PD_n pulse.
const int LUX13HS_PD_N_END_CLOCK = DEFAULT_READOUT_START + PRE_READ_GAP;
const int LUX19HS_PD_N_END_CLOCK = DEFAULT_READOUT_START + PRE_READ_GAP;

// Number of clocks to finish reading out all the rows. (QROW_TIME * 1.5)
const int LUX13HS_LAST_ROW_READOUT = (LUX13HS_DEFAULT_QROW_TIME * 1.5);
const int LUX19HS_LAST_ROW_READOUT = (LUX19HS_DEFAULT_QROW_TIME * 1.5);

// Minimum number of clocks between the readout and the TX_n pulse.
const int POST_READ_GAP = 82;

// Width of the TX_n pulse, in clocks. Datasheet recommends 100-150. This will
// determine when to make TX_n active (low).
const int LUX13HS_TX_N_WIDTH = 150;
//const int LUX13HS_TX_N_WIDTH = 22;
const int LUX19HS_TX_N_WIDTH = 100;

// Countdown clock value, in front of PD_N_END_CLOCK, for when to end the TX_n pulse
const int LUX13HS_TX_N_DONE = LUX13HS_PD_N_END_CLOCK + PD_OVERLAP;
const int LUX19HS_TX_N_DONE = LUX19HS_PD_N_END_CLOCK + PD_OVERLAP;

// Countdown clock value for when to start the TX_n pulse.
const int LUX13HS_TX_N_START = LUX13HS_TX_N_DONE + LUX13HS_TX_N_WIDTH;
const int LUX19HS_TX_N_START = LUX19HS_TX_N_DONE + LUX19HS_TX_N_WIDTH;

const int TX2_TX_GAP = 30;
//const int TX2_TX_GAP = 8;
const int LUX13HS_TX2_N_WIDTH = 50;
//const int LUX13HS_TX2_N_WIDTH = 22;
const int LUX19HS_TX2_N_WIDTH = 20;
const int LUX13HS_TX2_N_DONE = LUX13HS_TX_N_START + TX2_TX_GAP;
const int LUX19HS_TX2_N_DONE = LUX19HS_TX_N_START + TX2_TX_GAP;
const int LUX13HS_TX2_N_START = LUX13HS_TX2_N_DONE + LUX13HS_TX2_N_WIDTH;
const int LUX19HS_TX2_N_START = LUX19HS_TX2_N_DONE + LUX19HS_TX2_N_WIDTH;

// Time to increment the row address. This may change, but for now
// it's halfway through the QRow time.
const int LUX13HS_ROW_ADDR_INCR_TIME = (LUX13HS_DEFAULT_QROW_TIME / 2);
const int LUX19HS_ROW_ADDR_INCR_TIME = (LUX19HS_DEFAULT_QROW_TIME / 2);

// Default clock count to read/row pulses.
const int DEFAULT_ROW_CLOCK = 3;

// Number of pixels in a "block".
const int BLOCK_SIZE = 32;

// Number of overhead clocks per frame. Note that TX_N_START includes both the 'PD_OVERLAP'
// and the 'PRE_READ_GAP'.
const int LUX13HS_FRAME_OVERHEAD = LUX13HS_LAST_ROW_READOUT + POST_READ_GAP + LUX13HS_TX2_N_START;
const int LUX19HS_FRAME_OVERHEAD = LUX19HS_LAST_ROW_READOUT + POST_READ_GAP + LUX19HS_TX2_N_START;

// The default setting for the PD_N_OFFSET register. This effects the visibility
// of the shutter line.
const uint32_t LUX13HS_PD_N_OFFSET = 0x28;
const uint32_t LUX19HS_PD_N_OFFSET = 0x1A;

enum
{
    SENSOR_TEST_USER = 0,
    SENSOR_TEST_INCREMENT,
    SENSOR_TEST_DECREMENT,
    SENSOR_TEST_RANDOM,
    SENSOR_TEST_SHIFT
};

const int TEST_SHIFT_STRIDE = 7;
const int MAX_TP_ROW_INDEX = 8;
const int MAX_RETRIES = 10;

// This is in MCLK increments (nominally, 133MHz).
const uint32_t LUX13HS_READOUT_DELAY_DEFAULT = 0x14;
const int LUX13HS_MIN_READOUT_DELAY = 0x10;
const int LUX13HS_MAX_READOUT_DELAY = 0x18;

// This is in MCLK increments (nominally, 133MHz).
const uint32_t LUX19HS_READOUT_DELAY_DEFAULT = 0x1e;
const int LUX19HS_MIN_READOUT_DELAY = 0x18;
const int LUX19HS_MAX_READOUT_DELAY = 0x22;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

} /* End of namespace LUXIMA_HS_SENSOR */


using namespace LUXIMA_DAC;
using namespace LUXIMA_HS_SENSOR;

//--------------------------------------------------------------------
// LuximaHS Base Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
LuximaHS::LuximaHS(uint16_t max_width, uint16_t max_height,
                   uint16_t min_width, uint16_t min_height,
                   SENSOR::EEPROM_INFO *eeprom) :
    Sensor(max_width, max_height, min_width, min_height, eeprom),
    config_word(0),
    sensor_test_on(false)
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
LuximaHS::~LuximaHS()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

bool LuximaHS::class_init(FiFPGA *fpga_ptr)
{
    uint32_t reg_value;

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is NULL?!", __PRETTY_FUNCTION__);
        return false;
    }

    // Put the sensor interface into reset.
    reg_value = ICMX13HSC_ALTLVDS_RX_RESET | ICMX13HSC_SENSOR_RESET |
                ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT | ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP;
    fpga_ptr->regWrite(ICM_LUX13HS_CONTROL, reg_value);

    // Sometimes the camera doesn't exit nicely. We need to make sure we've
    // reset the board.
    //
    // Shutdown the DACs
    fpga_ptr->i2cWrite(DAC_GLOBAL_I2C_ADDR, CC_POWERDOWN_ALL, 0, FI_FPGA::I2C_SENSOR, 2);

    // Turn off power to the sensor board.
    fpga_ptr->regModify(NC1_FPGA_CONTROL, 0, ~NC1FC_ISB_PWR_ENABLE);

    // Wait a bit.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(50));

    // Enable the clpout signal.
    enable_clpout();

    // Tell the FPGA to assert POWER_EN for the sensor board.
    if (!fpga_ptr->regModify(NC1_FPGA_CONTROL, NC1FC_ISB_PWR_ENABLE, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable power for the sensor board!", __PRETTY_FUNCTION__);
        return false;
    }
    LOG("%s: ISB_PWR_ENABLE. ========================", __PRETTY_FUNCTION__);

    // Wait a little bit to give things a chance to settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(15));

    // Read the power status register to verify that the POWER_OK
    // signal is asserted.
    if (!fpga_ptr->regRead(NC1_FPGA_STATUS, &reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to read the status register!", __PRETTY_FUNCTION__);
        return false;
    }
    if ((reg_value & NC1FS_ISB_PWR_OK) != NC1FS_ISB_PWR_OK)
    {
        fi_log(LOG_ERR, "%s: Unable to power up sensor board! [status = 0x%08x]!", __PRETTY_FUNCTION__, reg_value);
        return false;
    }

    // Load any bad pixel information we might have.
    load_bad_pixels();

    // All is well.
    return (bad_pixels.get() == nullptr) ? false : true;
}

/*!
 * @brief Enable the sensor test pattern.
 */
bool LuximaHS::enable_sensor_test()
{
    return true;
}

int LuximaHS::get_readout_delay_min()
{
    return LUX13HS_MIN_READOUT_DELAY;
}

int LuximaHS::get_readout_delay_max()
{
    return LUX13HS_MAX_READOUT_DELAY;
}

void LuximaHS::load_bad_pixels()
{
    // The LuximaHS sensors use the FPGA's pixel correction functionality.
    bad_pixels = std::unique_ptr<BadPixels>(new BadPixels(eeprom.bad_pixels));
}

void LuximaHS::hard_reset()
{
}

/*!
 * @brief This routine is used to bring the sensor up in a controlled, managed fashion.
 *
 * @param startx    Starting x position in pixels, sensor-relative
 * @param starty    Starting y position in pixels
 * @param width     Width in pixels to capture
 * @param height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param shutter   Integration time (exposure) in microseconds
 * @param opts      Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100)
 *
 * @return true if everything goes well, false otherwise.
 */
bool LuximaHS::sensor_startup(
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    fpnmode_t fpn_mode,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t shutter,
    uint32_t opts,
    uint32_t sensor_gain)
{
    // TODO: Do we want to reset the sensor? Do we need to? This is the main difference between
    //       this method (sensor_startup) and the update_capture() method.
    // hard_reset();
    return update_capture(startx, starty, width, height, fpn_mode, bit_mode, frame_rate, shutter, opts, sensor_gain);
}

/*!
 * @brief Enforce sensor-specific restrictions on the capture parameters
 *
 * At minimum, it must check the frame rate because that is always sensor-specific.
 * The parameters have the proposed value on entry.
 * On exit, they may be updated to make them valid.
 *
 * The start position and size are relative to the sensor's area; they have already been
 * adjusted for license restrictions. Except for license restrictions, they are what the
 * user would see as the offset and size. If there are binning or subsampling options
 * in use, the ROI fields have to be multiplied by the appropriate factor.
 * We allow a sensor ROI larger than what the user can see if binning is in effect.
 *
 * Force FM6, aka GIGE_SENSOPT_INTERNAL_0, versus FM1 depending on (1) whether we are in LTR mode
 * and (2) whether the bit mode is greater than 8 bits. We keep it in the sensor options so the
 * GUIs have access to the flag for computing frame rates.
 *
 * @param rtn_startx        ROI column start
 * @param rtn_starty        ROI row start
 * @param rtn_width     ROI width
 * @param rtn_height        ROI height
 * @param rtn_bit_mode      Bit mode
 * @param rtn_frame_rate    Frame rate valid for this ROI
 * @param rtn_shutter_speed [unused] Shutter speed valid for that frame rate
 * @param rtn_sensor_opts   Sensor binning and subsampling options
 *
 * @return true if any changes were needed to make the capture parameters valid.
 */
bool LuximaHS::make_sensor_valid(
    uint16_t * /* rtn_startx */,
    uint16_t * /* rtn_starty */,
    uint16_t * /* rtn_width */,
    uint16_t * /* rtn_height */,
    bitmode_t * /* rtn_bit_mode */,
    float * /* rtn_frame_rate */,
    uint32_t * /* rtn_shutter_speed */,
    uint32_t * /* rtn_sensor_opts */)
{
    return true;
}

/*!
 * @brief Set up the sensor for a specific ROI.
 *
 * If the sensor is enabled, we disable it, set the ROI registers, then reenable it.
 *
 * @param true_startx   Starting x position in pixels, sensor-relative
 * @param true_starty   Starting y position in pixels
 * @param roi_width Width in pixels to capture
 * @param roi_height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param integration   Integration time (exposure) in microseconds
 * @param sensor_opts   Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100)
 *
 * @return true if everything is good, otherwise false.
 */
bool LuximaHS::update_capture(uint16_t /* true_startx */,
                              uint16_t /* true_starty */,
                              uint16_t /* roi_width */,
                              uint16_t /* roi_height */,
                              fpnmode_t /* fpn_mode */,
                              bitmode_t /* bit_mode */,
                              float /* frame_rate */,
                              uint32_t /* integration */,
                              uint32_t /* sensor_opts */,
                              uint32_t /* sensor_gain */)
{
    return true;
}

/*!
 * @brief Set the bit mode.
 * @returns True if successful, false otherwise.
 */
bool LuximaHS::set_fpga_bit_mode(bitmode_t /* bitmode */)
{
    return true;
}

bool LuximaHS::set_integration_time(uint32_t /* shutter_speed */)
{
    return true;
}

/*!
 * @brief Set the sync delay value.
 *
 * @param delay_amount The number of microseconds to delay the
 *             sync-in pulse.
 */
void LuximaHS::set_sync_delay(uint32_t /* delay_amount */)
{
}

void LuximaHS::set_syncin_enable(uint32_t /* type */)
{
}

/*!
 * @brief Take the desired frame rate (frames/second) and convert it to the
 *        number of sensor-specific clocks per frame.
 *
 * @param frame_rate    The number of frames per second.
 * @param clock_factor  An adjustment factor for the FPGA clock. Typically 1.0.
 * @return True if everthing went OK, false otherwise.
 */
bool LuximaHS::set_frame_rate(float frame_rate, double clock_factor)
{
    uint32_t value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check the frame rate value.
    if (approx_equals(frame_rate, 0))
    {
        fi_log(LOG_ERR, "%s: Invalid frame_rate (0).", __PRETTY_FUNCTION__);
        return false;
    }
    if (clock_factor == 0)
    {
        // Default the clock speed
        clock_factor = 1.0;
    }

    // Calculate the value based upon the frame rate
    value = (uint32_t)((clock_factor * sensor_clock_rate()) / (double)frame_rate);

    // Sanity check the value.
    if (value == 0)
    {
        fi_log(LOG_ERR, "%s: Invalid frame period value!", __PRETTY_FUNCTION__);
        return false;
    }

    if (!fpga_ptr->regWrite(ICM_SENSOR_FRAME_PERIOD, value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_SENSOR_FRAME_PERIOD.", __PRETTY_FUNCTION__);
        return false;
    }

    return true;
}

void LuximaHS::set_readout_start()
{
}

/*!
 * @brief Set the sensor gain.
 *
 * @param new_sensor_gain   The value to set in canonical form, 100 for 1x, 200 for 2x, 400 for 4x.
 *
 * @return  -1 if value is illegal
 *       0 if it was good but needs an update to go into effect
 *       1 if value is good and was put into effect.
 */
int LuximaHS::set_sensor_gain(uint32_t /*new_sensor_gain*/)
{
    LOG("%s: Under construction ============================", __PRETTY_FUNCTION__);
    return 1;
}

bool LuximaHS::set_sensor_readout(bool /* enable */)
{
    return true;
}

/*!
 * @brief Make sure the sensor ROI is legal.
 *
 * On entry, the values passed by reference are what was requested.
 * On exit, they are corrected if necessary.
 *
 * @param p_startx  Sensor-relative starting column
 * @param p_starty  Sensor-relative starting row
 * @param p_width   Sensor-relative columns
 * @param p_height  Sensor-relative rows
 * @param factor    [unused] Binning or subsampling factor
 *
 * @return true if there were any changes, false otherwise.
 */
bool LuximaHS::force_legal_roi(unsigned /* factor */, uint16_t *p_startx, uint16_t *p_starty,
                               uint16_t *p_width, uint16_t *p_height)
{
    bool any_changes = false;
    uint16_t startx = *p_startx;
    uint16_t starty = *p_starty;
    uint16_t width = *p_width;
    uint16_t height = *p_height;

    // LOG("%s", __PRETTY_FUNCTION__);

    // Check the effective sizes first.
    if (width < sensor_min_width)
    {
        width = sensor_min_width;
        any_changes = true;
    }
    if (height < sensor_min_height)
    {
        height = sensor_min_height;
        any_changes = true;
    }
    if (width > sensor_max_width)
    {
        width = sensor_max_width;
        any_changes = true;
    }
    if ((width % get_sensor_width_mod()) != 0)
    {
        width = (width / get_sensor_width_mod()) * get_sensor_width_mod();
        any_changes = true;
    }
    if ((startx % get_sensor_width_mod()) != 0)
    {
        startx = (startx / get_sensor_width_mod()) * get_sensor_width_mod();
        any_changes = true;
    }
    if ((startx + width) > sensor_max_width)
    {
        startx = sensor_max_width - width;
        any_changes = true;
    }

    if (height > sensor_max_height)
    {
        height = sensor_max_height;
        any_changes = true;
    }
    if ((height % get_sensor_height_mod()) != 0)
    {
        height = (height / get_sensor_height_mod()) * get_sensor_height_mod();
        any_changes = true;
    }
    if ((starty % get_sensor_height_mod()) != 0)
    {
        starty = (starty / get_sensor_height_mod()) * get_sensor_height_mod();
        any_changes = true;
    }
    if ((starty + height) > sensor_max_height)
    {
        starty = sensor_max_height - height;
        any_changes = true;
    }

    if (any_changes)
    {
        fi_log(LOG_ERR, "%s: Bad sensor ROI, using (%u,%u)%ux%u, was (%u,%u)%ux%u",
               __FUNCTION__,
               startx, starty, width, height,
               *p_startx, *p_starty, *p_width, *p_height);

        *p_startx = startx;
        *p_starty = starty;
        *p_width = width;
        *p_height = height;
    }

    return any_changes;
}

/*!
 * @brief Return the default bit mode we want to use with this sensor.
 */
bitmode_t LuximaHS::get_default_bit_mode()
{
    return (bitmode_t)GIGE_BITMODE_10_BITS;
}

/*!
 * @brief Return the default frame rate we want to use with this sensor.
 */
uint32_t LuximaHS::get_default_frame_rate()
{
    return 60;
}

/*!
 * @brief Return the default gain setting.
 */
uint32_t LuximaHS::get_default_gain()
{
    return 100;
}

/*!
 * @brief Return the frame rate in counts of sensor-specific clock "ticks".
 */
uint32_t LuximaHS::get_frame_rate()
{
    uint32_t    reg_val = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fpga_ptr->regRead(ICM_SENSOR_FRAME_PERIOD, &reg_val);

    return reg_val;
}

/*!
 * @brief Get a buffer containing the defective pixel filename.
 *
 * @param dirname   the directory in which the file will be found
 *          e.g. /scratch or /opt/fastec/calibration
 *
 * @return  std::string containing fully-qualified file name.
 */
std::string LuximaHS::get_defect_filename(const char *dirname)
{
    std::string fname_buf;

    fname_buf = fmt::sprintf(Sensor::DEFECT_MAP_FILE_FMT, get_home_path(), dirname,
                             eeprom.serial_no[0], eeprom.serial_no[1],
                             eeprom.serial_no[2], eeprom.serial_no[3]);

    return fname_buf;
}

/*!
 * @brief Return the amount of time (in microseconds) the FPGA/Sensor requires to
 *    move a frame, etc. AKA "Frame Overhead".
 */
uint32_t LuximaHS::get_frame_overhead()
{
    return 10;
}

uint32_t LuximaHS::get_readout_delay()
{
    return LUX13HS_READOUT_DELAY_DEFAULT;
}

/*!
 * @brief Return the minimum allowed shutter, in microseconds. To say this another way,
 *    return the fastest shutter period we can support.
 */
uint32_t LuximaHS::get_sensor_min_shutter()
{
    return 100;
}

/*!
 * @brief Read the temperature value from the ISB.
 */
double LuximaHS::get_temperature()
{
    double result;
    uint8_t temp[2];
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    temp[0] = temp[1] = 0;
    if (!fpga_ptr->i2cRead(MCP_I2C_ADDR, MCP_SPDREG_TEMPERATURE, temp, FI_FPGA::I2C_SENSOR, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Now convert the value into a float.
    if ((temp[0] & 0x10) == 0x10)
    {
        // Temperature < 0 degrees C.
        result = (double)(256 - (((temp[0] & 0xF) << 4) + ((temp[1] & 0xF0) >> 4)));
    }
    else
    {
        // Temperature > 0 degrees C.
        result = (double)(((temp[0] & 0xF) << 4) + ((temp[1] & 0xF0) >> 4));
    }

    // Return the temperature.
    return result;
}

/*!
 * @brief Given the starting X/Y coordinate, what does the Bayer pattern
 *    look like?
 */
uint32_t LuximaHS::IPM_bayer_pattern_start(uint16_t /*startx*/, uint16_t /*starty*/)
{
    return IPMC_SRC_FRM_BAYER_GR;
}


/*!
 * @brief Set up the FPGA's IPM to do bad pixel correction.
 *
 * For the LuximaHS sensors, we use the FPGA's bad pixel corrections.
 *
 * @param port      [unused]
 * @param true_x    current ROI "true" x coordinate, relative to the sensor's actual size
 * @param true_y    current ROI "true" y coordinate, relative to the sensor's actual size
 * @param width     current ROI width
 * @param height    current ROI height
 */
void LuximaHS::IPM_bad_pixel_correction(int port, uint16_t true_x, uint16_t true_y,
                                        uint16_t width, uint16_t height)
{
    (void)port;

    if (bad_pixels.get() != nullptr)
    {
        // LOG("%s: true_x=%d, true_y=%d, width=%d, height=%d", __PRETTY_FUNCTION__, true_x, true_y, width, height);

        bad_pixels->set_roi(true_x, true_y, width, height);
        bad_pixels->correct();
    }
    else
    {
        LOG("%s: bad_pixels is NULL", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Start capture for calibration of a LuximaHS sensor
 *
 * @param   factory     true if this is factory calibration, false for user calibration
 * @param   cal_bit_mode    the bit mode to use for this capture, if it is user calibration
 * @param   fpn_mode        the FPN mode to use for this capture, if it is user calibration
 *
 * @return true if successful, false otherwise
 */
bool LuximaHS::start_calibration_capture(bool factory, bitmode_t cal_bit_mode)
{
    uint16_t true_x;
    uint16_t true_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;
    uint32_t sensor_gain;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    LOG("%s: factory=%d, cal_bit_mode=0x%x", __PRETTY_FUNCTION__, factory, cal_bit_mode);

    // Make sure FPN correction is off while we calibrate.
    FiImageCorrection::fpnDisable();

    if (factory)
    {
        // Get the user settings, adjusted to the physical sensor dimensions.
        cs_ptr->get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode,
                                      &frame_rate, &shutter_speed, &sensor_opts);
        if ((bit_mode != (bitmode_t)GIGE_BITMODE_7_0) ||
            (true_x != 0) || (true_y != 0) ||
            (width != get_sensor_max_width()) ||
            (height != get_sensor_max_height()))

        {
            fi_log(LOG_WARNING, "%s: System not setup properly for factory calibration!", __PRETTY_FUNCTION__);

            // Use what we know to be valid settings.
            true_x = 0;
            true_y = 0;
            width = get_sensor_max_width();
            height = get_sensor_max_height();
            bit_mode = (bitmode_t)GIGE_BITMODE_7_0;
        }

        // Make sure we're not binning or anything.
        sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);
        sensor_gain = get_default_gain();
    }
    else
    {
        // Get the user settings, adjusted to the physical sensor dimensions.
        cs_ptr->get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode,
                                      &frame_rate, &shutter_speed, &sensor_opts);
        bit_mode = cal_bit_mode;
        cs_ptr->get_sensor_gain(&sensor_gain);
    }
    fi_log(LOG_INFO, "%s: Adjusted frame_rate=%f, shutter_speed=%d, bit_mode=%d", __PRETTY_FUNCTION__, frame_rate, shutter_speed, bit_mode);

    // Start up the sensor.  This also sets up the FPGA ICM ROI since the way that is done depends
    // on the specific sensor.
    if (!sensor_startup(true_x, true_y, width, height, 0, bit_mode,
                        frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: sensor_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Set up the ICM pipeline. Nothing will happen until we turn on the sensor readout bit.
    if (!fpga_ptr->icmStartup(bit_mode, frame_rate, shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: icm_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    if (!fpga_ptr->icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __PRETTY_FUNCTION__);
        return false;
    }

    fpga_ptr->icmSetLiveArgs(bit_mode, frame_rate, shutter_speed);

    return true;
}

/*!
 * @brief Start capture for PRNU calibration of a LuximaHS sensor
 *
 * @return true if successful, false otherwise
 */
bool LuximaHS::start_prnu_capture()
{
    uint16_t true_x;
    uint16_t true_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;
    uint32_t sensor_gain;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure FPN correction is off while we calibrate.
    FiImageCorrection::fpnDisable();

    // Get the user settings, adjusted to the physical sensor dimensions.
    cs_ptr->get_physical_settings(&true_x, &true_y, &width, &height, &bit_mode,
                                  &frame_rate, &shutter_speed, &sensor_opts);
    if ((bit_mode != (bitmode_t)GIGE_BITMODE_9_0) ||
        (true_x != 0) || (true_y != 0) ||
        (width != get_sensor_max_width()) ||
        (height != get_sensor_max_height()))

    {
        fi_log(LOG_WARNING, "%s: System not setup properly for factory calibration!", __PRETTY_FUNCTION__);

        // Use what we know to be valid settings.
        true_x = 0;
        true_y = 0;
        width = get_sensor_max_width();
        height = get_sensor_max_height();
        bit_mode = (bitmode_t)GIGE_BITMODE_9_0;
    }

    // Make sure we're not binning or anything.
    sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);
    sensor_gain = get_default_gain();
    fi_log(LOG_INFO, "%s: Adjusted frame_rate=%f, shutter_speed=%d, bit_mode=%d", __PRETTY_FUNCTION__, frame_rate, shutter_speed, bit_mode);

    // Start up the sensor.  This also sets up the FPGA ICM ROI since the way that is done depends
    // on the specific sensor.
    if (!sensor_startup(true_x, true_y, width, height, 0, bit_mode,
                        frame_rate, shutter_speed, sensor_opts, sensor_gain))
    {
        fi_log(LOG_WARNING, "%s: sensor_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Set up the ICM pipeline. Nothing will happen until we turn on the sensor readout bit.
    if (!fpga_ptr->icmStartup(bit_mode, frame_rate, shutter_speed))
    {
        fi_log(LOG_WARNING, "%s: icm_startup failed.", __PRETTY_FUNCTION__);
        return false;
    }

    if (!fpga_ptr->icmSensorReadoutEnable(true))
    {
        fi_log(LOG_WARNING, "%s: Failed to turn on sensor readout.", __PRETTY_FUNCTION__);
        return false;
    }

    fpga_ptr->icmSetLiveArgs(bit_mode, frame_rate, shutter_speed);

    return true;
}

void LuximaHS::shutdown()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Shutdown the DACs.
    if (fpga_ptr != nullptr)
    {
        // Shutdown the DACs
        if (!fpga_ptr->i2cWrite(DAC_GLOBAL_I2C_ADDR, (CC_POWERDOWN_ALL << CC_SHIFT) | DAC_GLOBAL, 0, FI_FPGA::I2C_SENSOR, 2))
            fi_log(LOG_ERR, "%s: Unable to turn off DACs.", __PRETTY_FUNCTION__);
        else
            fi_log(LOG_DEBUG, "%s: DAC are turned off.", __PRETTY_FUNCTION__);

        // Reset the external memory interface and power off the sensor board.
        if (!fpga_ptr->regWrite(NC1_FPGA_CONTROL, NC1FC_EMIF_SW_RESET))
            fi_log(LOG_ERR, "%s: Unable to power off sensor board.", __PRETTY_FUNCTION__);
        else
            fi_log(LOG_DEBUG, "%s: Sensor board powered off.", __PRETTY_FUNCTION__);
    }

    Sensor::shutdown();
}

//--------------------------------------------------------------------
// Luxima13HS Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima13HS::Luxima13HS(SENSOR::EEPROM_INFO *eeprom)
    : LuximaHS(
          1280,  /* max width */
          864,  /* max height */
          64,  /* min width */
          64,  /* min height */
          eeprom)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
    config_word = LUX13HS_RECOMMENDED;

    if (fpga_ptr != nullptr)
    {
        // Make sure we are not in sensor test mode.
        fpga_ptr->regWrite(ICM_LUX13HS_TP_CONTROL, ICMX13HSTPC_RESET);
    }
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima13HS::~Luxima13HS()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

bool Luxima13HS::class_init(FiFPGA *fpga_ptr)
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Perform generic initialization.
    if (!LuximaHS::class_init(fpga_ptr))
    {
        fi_log(LOG_DEBUG, "%s: parent class_init() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Initialize the DACs
    if (!dac_init(fpga_ptr))
    {
        fi_log(LOG_ERR, "%s: Unable to initialize the DACs on the sensor board!", __PRETTY_FUNCTION__);
        return false;
    }

    // Bring the sensor out of reset.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_SENSOR_RESET))
    {
        fi_log(LOG_ERR, "%s: Unable to deassert sensor reset!", __PRETTY_FUNCTION__);
        return false;
    }

    // Enable the PLLs.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_SENSOR_PLL_ENA, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable sensor PLLs!", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait a while to let the PLL settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(1));

    // Write the sensor configuration word.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_SERIAL_CONFIG, config_word))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_SERIAL_CONFIG", __PRETTY_FUNCTION__);
        return false;
    }

    // =============================================================================
    // Screw with the LVDS inversion vector.
    // =============================================================================
#if 1
    {
        // uint32_t vec_bl = 0x798a8;   /* 19:0 */
        uint32_t vec_bl = 0x798A8;  /* 19:0 Ch7 not inverted*/
//  uint32_t vec_br = 0xa7376;  /* 39:20 */
//  uint32_t vec_tl = 0x10A32;  /* 59:40 */
//  uint32_t vec_tr = 0x09011;  /* 79:60 */

        fpga_ptr->regWrite(ICM_LUX13HS_LVDS_INV_BL, vec_bl);
//  fpga_ptr->write_reg(ICM_LUX13HS_LVDS_INV_BR, vec_br);
//  fpga_ptr->write_reg(ICM_LUX13HS_LVDS_INV_TL, vec_tl);
//  fpga_ptr->write_reg(ICM_LUX13HS_LVDS_INV_TR, vec_tr);
    }
#endif
    // =============================================================================
    // =============================================================================

    // Adjust the PD_N offset value.
    fpga_ptr->regWrite(ICM_LUX13HS_PD_N_OFFSET, LUX13HS_PD_N_OFFSET);

    // All is well.
    return true;
}

void Luxima13HS::enable_clpout()
{
    // The current thinking is to NOT enable clpout.
#if 0
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Enable the clampout signal.
    fpga_ptr->modify_reg(ICM_LUX13HS_CLPOUT_CONFIG, ICMX13HS_CLPOUT_ENABLE, -1);
#endif
}

bool Luxima13HS::enable_sensor_test()
{
    bool rc(false);
    int tp_row_index;
    int pix_index;
    uint64_t total_pix_index;
    uint32_t pixel_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    extern int sensor_test_type;

    LOG("%s: type=%d", __PRETTY_FUNCTION__, sensor_test_type);

    // Remember that we're in test mode.
    sensor_test_on = true;

    // Switch the FPGA to use the test pattern RAMs. This will also
    // take the test logic out of reset.
    fpga_ptr->regWrite(ICM_LUX13HS_TP_CONTROL, ICMX13HSTPC_ENABLE);

    // Initialize our RAMs.
    pixel_value = 0;
    total_pix_index = 0;
    for (tp_row_index = 0; tp_row_index < MAX_TP_ROW_INDEX; ++tp_row_index)
    {
        int retry;
        uint32_t reg_value = 0;

        // Make sure we are OK to start writing.
        for (retry = 0; retry < MAX_RETRIES; ++retry)
        {
            fpga_ptr->regRead(ICM_LUX13HS_TP_STATUS, &reg_value);
            if ((reg_value & ICMX13HSTPS_BUSY) != ICMX13HSTPS_BUSY)
                break;

            // Wait a while
            using ms = std::chrono::milliseconds;
            std::this_thread::sleep_for(ms(10));
        }
        if (retry == MAX_RETRIES)
        {
            fi_log(LOG_DEBUG, "%s: FIFO is always BUSY!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Sanity check. FIFO should be empty.
        for (retry = 0; retry < MAX_RETRIES; ++retry)
        {
            fpga_ptr->regRead(ICM_LUX13HS_TP_STATUS, &reg_value);
            if (ICMX13HSTPS_GET_FIFO_LEVEL(reg_value) == 0)
                break;

            fi_log(LOG_DEBUG, "%s: %d, FIFO level is not empty at start! [0x%08x]", __PRETTY_FUNCTION__, retry, reg_value);
            fpga_ptr->regModify(ICM_LUX13HS_TP_CONTROL, ICMX13HSTPC_RESET, -1);
            fpga_ptr->regModify(ICM_LUX13HS_TP_CONTROL, 0, ~ICMX13HSTPC_RESET);
        }
        if (retry == MAX_RETRIES)
        {
            fi_log(LOG_DEBUG, "%s: FIFO is never empty! [0x%08x]", __PRETTY_FUNCTION__, reg_value);
            goto bail;
        }

        // Loop until we get the FIFO filled correctly.
        for (retry = 0; retry < MAX_RETRIES; ++retry)
        {
            // Fill up the FIFO.
            fi_log(LOG_DEBUG, "%s: tp_width=%d", __PRETTY_FUNCTION__, sensor_max_width);
            for (pix_index = 0; pix_index < sensor_max_width; ++pix_index)
            {
                // The value to write depends on the test selected.
                switch (sensor_test_type)
                {
                case SENSOR_TEST_USER:
                    // pixel_value = 0x000;
                    pixel_value = 0x25a;
                    break;
                case SENSOR_TEST_INCREMENT:
                    if (pix_index == 0)
                        pixel_value = 0;
                    else
                        ++pixel_value;
                    break;
                case SENSOR_TEST_DECREMENT:
                    if (pix_index == 0)
                        pixel_value = sensor_max_width - 1;
                    else
                        --pixel_value;
                    break;
                case SENSOR_TEST_RANDOM:
                    pixel_value = rand();
                    break;
                case SENSOR_TEST_SHIFT:
                    if ((total_pix_index % TEST_SHIFT_STRIDE) == 0)
                        ++pixel_value;
                    break;
                default:
                    fi_log(LOG_ERR, "Unknown test type [0x%x]", sensor_test_type);
                    goto bail;
                }

                // Make sure we're only writing 10-bits of data.
                pixel_value = pixel_value & 0x3FF;

                // Write the pixel value to the FIFO.
                fpga_ptr->regWrite(ICM_LUX13HS_TP_INIT_FIFO_WRITE, pixel_value);

                // Increment our total pixel index.
                ++total_pix_index;
            }

            // Sanity check. FIFO should be full.
            fpga_ptr->regRead(ICM_LUX13HS_TP_STATUS, &reg_value);
            if (ICMX13HSTPS_GET_FIFO_LEVEL(reg_value) == sensor_max_width)
                break;

            fi_log(LOG_DEBUG, "%s: FIFO level is not full! [0x%08x]", __PRETTY_FUNCTION__, reg_value);
            fpga_ptr->regModify(ICM_LUX13HS_TP_CONTROL, ICMX13HSTPC_RESET, -1);
            fpga_ptr->regModify(ICM_LUX13HS_TP_CONTROL, 0, ~ICMX13HSTPC_RESET);
        }
        if (retry == MAX_RETRIES)
        {
            fi_log(LOG_DEBUG, "%s: FIFO is never filled!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Tell the FPGA to write the FIFO to tp_row_index.
        fpga_ptr->regModify(ICM_LUX13HS_TP_CONTROL, (ICMX13HSTPC_START | (tp_row_index << ICMX13HSTPC_ROW_SEL_SHIFT)), ~ICMX13HSTPC_ROW_SEL_MASK);
    }

    LOG("%s: Initialization complete", __PRETTY_FUNCTION__);

    // All is well.
    rc = true;

bail:
    // If we failed, make sure to put the test logic back into reset.
    if (!rc)
        fpga_ptr->regWrite(ICM_LUX13HS_TP_CONTROL, ICMX13HSTPC_RESET);

    // Return our status.
    return rc;
}

/*!
 * @brief Return the amount of time (in microseconds) the FPGA/Sensor requires to
 *    move a frame, etc. AKA "Frame Overhead".
 */
uint32_t Luxima13HS::get_frame_overhead()
{
    return (uint32_t)(((float)LUX13HS_PD_N_END_CLOCK / (float)sensor_clock_rate_usec()) + 0.5);
}

/*!
 * @brief Return the minimum allowed shutter, in microseconds. To say this another way,
 *    return the fastest shutter period we can support.
 */
uint32_t Luxima13HS::get_sensor_min_shutter()
{
    return (uint32_t)(((float)LUX13HS_TX_N_START / (float)sensor_clock_rate_usec()) + 0.5);
}

void Luxima13HS::hard_reset()
{
    LOG("%s: Under construction =================", __PRETTY_FUNCTION__);

#ifdef ALM_TODO_SENSOR
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Assert the sensor reset.
    if (!fpga_ptr->modify_reg(ICM_LUX13HS_CONTROL,
                              (ICMX13HSC_SENSOR_RESET |
                               ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP |
                               ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT |
                               ICMX13HSC_ALTLVDS_RX_RESET),
                              ~ICMX13HSC_SENSOR_PLL_ENA))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        return;
    }

    // Bring the sensor out of reset.
    // Wait a bit.
    using us = std::chrono::microseconds;
    std::this_thread::sleep_for(us(50));

    if (!fpga_ptr->modify_reg(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_SENSOR_RESET))
    {
        fi_log(LOG_ERR, "%s: Unable to deassert sensor reset!", __PRETTY_FUNCTION__);
        return;
    }

    // Enable the PLLs.
    if (!fpga_ptr->modify_reg(ICM_LUX13HS_CONTROL, ICMX13HSC_SENSOR_PLL_ENA, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable sensor PLLs!", __PRETTY_FUNCTION__);
        return;
    }
#if 0
    // Clear out the readout delay values.
    fpga_ptr->write_reg(ICM_LUX13HS_LVDS_RX_STARTUP_READOUT_DELAY, 0);

    // Bring the sensor out of reset.
    // Wait a bit.
    using us = std::chrono::microseconds;
    std::this_thread::sleep_for(us(5));

    if (!fpga_ptr->modify_reg(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_SENSOR_RESET))
    {
        fi_log(LOG_ERR, "%s: Unable to deassert sensor reset!", __PRETTY_FUNCTION__);
        return;
    }

    // Enable the PLLs.
    if (!fpga_ptr->modify_reg(ICM_LUX13HS_CONTROL, ICMX13HSC_SENSOR_PLL_ENA, -1))
    {
        fi_log(LOG_ERR, "%s: Unable to enable sensor PLLs!", __PRETTY_FUNCTION__);
        return;
    }
#endif
#endif /* ALM_TODO_SENSOR */
}

bool Luxima13HS::is_bit_mode_supported(bitmode_t bit_mode)
{
    bool rc = false;

    switch ((int)bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_MIDDLE_8_BITS:
    case GIGE_BITMODE_8_1:
    case GIGE_BITMODE_UPPER_8_BITS:
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_10_BITS:
    case GIGE_BITMODE_9_0:
        rc = true;
        break;
    default:
        break;
    }

    return rc;
}

/*!
 * @brief Check to see if the passed in sensor_opts parameter is supported.
 * @param sensor_opts   The sensor options, one of GIGE_SENSOPT_xxx.
 * @returns True if the option is supported, false otherwise.
 */
bool Luxima13HS::is_option_supported(uint32_t /* sensor_opts */)
{
    // No binning or subsampling.
    return false;
}


bool Luxima13HS::make_sensor_valid(uint16_t *rtn_startx,
                                   uint16_t *rtn_starty,
                                   uint16_t *rtn_width,
                                   uint16_t *rtn_height,
                                   bitmode_t *rtn_bit_mode,
                                   float *rtn_frame_rate,
                                   uint32_t * /* rtn_shutter_speed */,
                                   uint32_t *rtn_sensor_opts)
{
    uint16_t startx = *rtn_startx;
    uint16_t starty = *rtn_starty;
    uint16_t width = *rtn_width;
    uint16_t height = *rtn_height;
    uint32_t sensor_opts = *rtn_sensor_opts;
    bitmode_t bit_mode = *rtn_bit_mode;
    float max_frame_rate;
    uint32_t any_changes = 0;
    uint32_t frame_time;

    LOG("%s", __PRETTY_FUNCTION__);

    // We support all the bit modes so this should never happen.
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Invalid bit_mode, using 0x%02x", __FUNCTION__, GIGE_BITMODE_11_4);
        bit_mode = (bitmode_t)GIGE_BITMODE_9_2;
        any_changes |= (1 << 0);
    }

    // Don't leave the bit mode setting in one of the old-style values.
    switch (bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
        bit_mode = GIGE_BITMODE_7_0;
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
        bit_mode = GIGE_BITMODE_8_1;
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
        bit_mode = GIGE_BITMODE_9_2;
        break;
    case GIGE_BITMODE_10_BITS:
        bit_mode = GIGE_BITMODE_9_0;
        break;
    default:
        break;
    }

    // We don't support sensor opts.
    if (sensor_opts && (sensor_opts != ((1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT))))
    {
        fi_log(LOG_DEBUG, "%s: sensor_opts are not supported!", __PRETTY_FUNCTION__);
        sensor_opts = ((1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT));
        any_changes |= (1 << 3);
    }

    (void)force_legal_roi(0, &startx, &starty, &width, &height);

    if ((startx != *rtn_startx) ||
        (starty != *rtn_starty) ||
        (width  != *rtn_width) ||
        (height != *rtn_height) ||
        (bit_mode != *rtn_bit_mode) ||
        (sensor_opts != *rtn_sensor_opts))
    {
        // fi_log(LOG_DEBUG, "%s: was: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );
        *rtn_startx = startx;
        *rtn_starty = starty;
        *rtn_width = width;
        *rtn_height = height;
        *rtn_sensor_opts = sensor_opts;
        *rtn_bit_mode = bit_mode;
        // fi_log(LOG_DEBUG, "%s: now: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );

        any_changes |= (1 << 7);
    }

    // The maximum frame rate depends on the height of the readout.
    frame_time = ((LUX13HS_DEFAULT_QROW_TIME * height / 4) + LUX13HS_FRAME_OVERHEAD);
    max_frame_rate = sensor_clock_rate() / frame_time;
    if (*rtn_frame_rate > max_frame_rate)
    {
        *rtn_frame_rate = max_frame_rate;
        any_changes |= (1 << 8);
    }

    // Did we make any changes to the settings?
    if (any_changes != 0)
    {
        //fi_log(LOG_DEBUG, "%s: changed 0x%08x", __FUNCTION__, any_changes );
        return true;
    }
    else
        return false;
}

bool Luxima13HS::update_capture(uint16_t true_startx,
                                uint16_t true_starty,
                                uint16_t roi_width,
                                uint16_t roi_height,
                                fpnmode_t /* fpn_mode */,
                                bitmode_t bit_mode,
                                float frame_rate,
                                uint32_t integration,
                                uint32_t /* sensor_opts */,
                                uint32_t /* sensor_gain */)
{
    int i;
    bool rc(false);
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: %ux%u@(%u,%u)", __PRETTY_FUNCTION__, roi_width, roi_height, true_startx, true_starty);

    // Need to make sure our startx is multiples of 2 and starty is multiples of 4.
    if ((true_startx & 1) || (true_starty & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! startx=0x%x, starty=0x%x", __PRETTY_FUNCTION__, true_startx, true_starty);
        true_startx = (true_startx + 1) & ~1;
        true_starty = (true_starty + 3) & ~3;
    }
    if ((roi_width & 1) || (roi_height & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! width=0x%x, height=0x%x", __PRETTY_FUNCTION__, roi_width, roi_height);
        roi_width = (roi_width + 1) & ~1;
        roi_height = (roi_height + 3) & ~3;
    }
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_DEBUG, "%s: Invalid bit_mode=0x%x", __PRETTY_FUNCTION__, bit_mode);
        bit_mode = GIGE_BITMODE_9_2;
    }

    // Clear the LVDS enable bits and set the reset bits..
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_RESET, ~ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA);

    // Initialize the sensor settings.
    fpga_ptr->icmSetFrameRate(frame_rate);
    fpga_ptr->icmSetShutterSpeed(integration);

    // Initialize the readout delay for the first column.
    set_readout_start();

    // Initialize the timing registers
    reg_value = ICMX13HSCFGB_SET_QROW_TIME(0, LUX13HS_DEFAULT_QROW_TIME);
    reg_value = ICMX13HSCFGB_SET_QROW_COUNT(reg_value, roi_height / 4);
    reg_value = ICMX13HSCFGB_SET_FIRST_QROW(reg_value, true_starty / 4);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_B = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_B, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_B", __PRETTY_FUNCTION__);
        goto bail;
    }

    reg_value = ICMX13HSCFGC_SET_ROW_ADDR_INC_COUNT(0, LUX13HS_ROW_ADDR_INCR_TIME);
    reg_value = ICMX13HSCFGC_SET_ST_READ_BOT_N_COUNT(reg_value, DEFAULT_ROW_CLOCK);
    reg_value = ICMX13HSCFGC_SET_ST_ROW_BOT_N_COUNT(reg_value, DEFAULT_ROW_CLOCK);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_C = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_C, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_C", __PRETTY_FUNCTION__);
        goto bail;
    }

    reg_value = ICMX13HSCFGD_SET_ST_READ_TOP_N_COUNT(0, LUX13HS_ROW_ADDR_INCR_TIME + DEFAULT_ROW_CLOCK);
    reg_value = ICMX13HSCFGD_SET_ST_ROW_TOP_N_COUNT(reg_value, LUX13HS_ROW_ADDR_INCR_TIME + DEFAULT_ROW_CLOCK);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_D = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_D, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_D", __PRETTY_FUNCTION__);
        goto bail;
    }

    if (!fpga_ptr->regWrite(ICM_LUX13HS_ROI_V_CONFIG, roi_height))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_ROI_V_CONFIG", __PRETTY_FUNCTION__);
        goto bail;
    }

    // ------------------------------------------------------------------------
    // This is a little strange, so follow along...
    // - For the start, we want the starting BLOCK and the starting PIXEL within that block.
    // - For the end, we want the ending BLOCK and the number of pixels withing that block.
    // Note the difference: one is an offset and the other is a count!
    // ------------------------------------------------------------------------
    reg_value = ICMX13HSROIH_SET_FIRST_ACTIVE_BLOCK(0, true_startx / BLOCK_SIZE);
    reg_value = ICMX13HSROIH_SET_FIRST_ACTIVE_BLOCK_OFFSET(reg_value, true_startx % BLOCK_SIZE);
    reg_value = ICMX13HSROIH_SET_LAST_ACTIVE_BLOCK_OFFSET(reg_value, ((true_startx + roi_width - 1) / BLOCK_SIZE));
    reg_value = ICMX13HSROIH_SET_LAST_ACTIVE_BLOCK_PIXEL_COUNT(reg_value, (((true_startx + roi_width - 1) % BLOCK_SIZE) + 1));
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_ROI_H_CONFIG = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    {
        fi_log(LOG_DEBUG, "--> startx=%d, roi_width=%d, block pixel count=0x%04X", true_startx, roi_width, (((true_startx + roi_width - 1) % BLOCK_SIZE) + 1));
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_ROI_H_CONFIG, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_ROI_H_CONFIG", __PRETTY_FUNCTION__);
        goto bail;
    }

    if (!set_fpga_bit_mode(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Unable to set the bitmode", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Clear out the readout delay values.
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_STARTUP_READOUT_DELAY, 0);

    // Enable the LVDS lines.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~(ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP | ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Wait a bit to give the PLLs a chance to lock.
    for (i = 0; i < 10; ++i)
    {
        // Take a knee...
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(5));

        // Read the status to see if we've got lock.
        if (!fpga_ptr->regRead(ICM_LUX13HS_STATUS, &reg_value))
        {
            fi_log(LOG_DEBUG, "%s: Unable to read ICM_LUX13HS_STATUS", __PRETTY_FUNCTION__);
            goto bail;
        }
        if ((reg_value & ICMX13HSS_ALTLVDS_RX_PLL_LOCKED) == ICMX13HSS_ALTLVDS_RX_PLL_LOCKED)
            break;
    }

    // Release the reset follwed by enabling the receivers.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_ALTLVDS_RX_RESET))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA, -1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = true;

bail:
    // Do we want to reset the sensor on fail?
    // if (!rc)
    //     hard_reset();
    return rc;
}

bool Luxima13HS::set_fpga_bit_mode(bitmode_t bitmode)
{
    bool rc = false;
    uint32_t value = 0;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // LOG("%s: bitmode=%d", __PRETTY_FUNCTION__, (int)bitmode);

    if (!fpga_ptr->regRead(ICM_LUX13HS_CONTROL, &value))
    {
        fi_log(LOG_ERR, "%s: Unable to read ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Clear the bits so we can set the ones we want.
    value &= ~(ICMX13HSC_BIT_DEPTH_10 | ICMX13HSC_BIT_SEL_MASK);

    // Now figure out the bit mode,
    switch ((unsigned)bitmode)
    {
    case GIGE_BITMODE_9_0:
    case GIGE_BITMODE_10_BITS:
        value |= ICMX13HSC_BIT_DEPTH_10;
        break;
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_LOWER_8_BITS:
        value |= (ICMX13HSC_BIT_DEPTH_8 | ICMX13HSC_BIT_SEL_7_0);
        break;
    case GIGE_BITMODE_8_1:
    case GIGE_BITMODE_MIDDLE_8_BITS:
        value |= (ICMX13HSC_BIT_DEPTH_8 | ICMX13HSC_BIT_SEL_8_1);
        break;
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_UPPER_8_BITS:
    default:
        value |= (ICMX13HSC_BIT_DEPTH_8 | ICMX13HSC_BIT_SEL_9_2);
        break;

    case GIGE_BITMODE_11_0:
    case GIGE_BITMODE_11_2:
    case GIGE_BITMODE_11_4:
    case GIGE_BITMODE_10_1:
    case GIGE_BITMODE_10_3:
        fi_log(LOG_DEBUG, "%s: Invalid bitmode [0x%08x]", __PRETTY_FUNCTION__, bitmode);
        goto bail;
    }

    // Write the settings back to the register.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_CONTROL, value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    rc = true;

bail:
    return rc;
}

bool Luxima13HS::set_integration_time(uint32_t shutter_speed)
{
    uint32_t integration_time;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // LOG("%s: shutter_speed=0x%08x [%u]", __PRETTY_FUNCTION__, shutter_speed, shutter_speed);

    // Figure out the integration time.
    integration_time = sensor_clock_rate_usec() * shutter_speed;

    // Tell the sensor to update it's integration period.
    // Setup the integration pulse to happen at the end of the frame time.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_PD_N_HIGH_COUNT, integration_time))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_PD_N_HIGH_COUNT (ss=%d, time=0x%x)", __PRETTY_FUNCTION__, shutter_speed, integration_time);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_PD_N_LOW_COUNT, LUX13HS_PD_N_END_CLOCK))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_PD_N_LOW_COUNT (ss=%d, time=0x%x)", __PRETTY_FUNCTION__, shutter_speed, integration_time);
        return false;
    }

    // Determine where the TX pulse needs to be. The pulse goes active LOW, so the low count
    // is when the pulse starts and the high count is when the pulse ends.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX_N_LOW_COUNT, LUX13HS_TX_N_START))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX_N_LOW_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX13HS_TX_N_START);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX_N_HIGH_COUNT, LUX13HS_TX_N_DONE))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX_N_HIGH_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX13HS_TX_N_DONE);
        return false;
    }

    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX2_N_LOW_COUNT, LUX13HS_TX2_N_START))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX2_N_LOW_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX13HS_TX2_N_START);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX2_N_HIGH_COUNT, LUX13HS_TX2_N_DONE))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX2_N_HIGH_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX13HS_TX2_N_DONE);
        return false;
    }

    // All is well.
    return true;
}

uint32_t Luxima13HS::get_readout_delay()
{
    uint32_t rdout_delay;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    rdout_delay = regs[GIGE_REG_SENSOR_RDOUT_DELAY].get_value();
    if ((rdout_delay < (uint32_t)get_readout_delay_min()) || (rdout_delay > (uint32_t)get_readout_delay_max()))
    {
        LOG("%s: Bad value [0x%08x], switching to default [0x%08x]", __PRETTY_FUNCTION__, rdout_delay, LUX13HS_READOUT_DELAY_DEFAULT);
        rdout_delay = LUX13HS_READOUT_DELAY_DEFAULT;
    }

    LOG("%s: 0x%08x", __PRETTY_FUNCTION__, rdout_delay);

    return rdout_delay;
}

void Luxima13HS::set_readout_delay(int delay)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    regs[GIGE_REG_SENSOR_RDOUT_DELAY].set_value((uint32_t)delay);
}

void Luxima13HS::set_readout_start()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Set the readout start time.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_A, DEFAULT_READOUT_START))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_A.", __PRETTY_FUNCTION__);
    }

    // Set our readout delay.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_DELAY, get_readout_delay()))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_DELAY", __PRETTY_FUNCTION__);
    }
}

bool Luxima13HS::set_sensor_readout(bool enable)
{
    bool rc;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // LOG("%s: enable=%c", __PRETTY_FUNCTION__, (enable ? 'T' : 'F'));

    // Control the sensor readout.
    if (enable)
        rc = fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_SENSOR_READOUT_ENA, -1);
    else
        rc = fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_SENSOR_READOUT_ENA);

    // Return our result.
    return rc;
}

void Luxima13HS::set_sync_delay(uint32_t delay_amount)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: delay_amount=%u", __PRETTY_FUNCTION__, delay_amount);
    fpga_ptr->regWrite(ICM_LUX13HS_1PPS_DELAY_CNT, delay_amount);
}

void Luxima13HS::set_syncin_enable(uint32_t type)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Figure out what type we are.
    if ((type & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_INTERNAL)
    {
        // Update the control register. Clear the mode (MASTER is 0) and the skip count fields.
        fpga_ptr->regModify(ICM_LUX13HS_CONTROL,
                            0,
                            ~(ICMX13HSC_SYNC_MODE_MASK | ICMX13HSC_1PPF_SKIP_COUNT_MASK));
    }
    else if ((type & GIGE_SYNC_TYPE_MASK) == GIGE_SYNC_TYPE_EXTERNAL)
    {
        // Figure out if we are 1-Pulse/Second or 1-Pulse/Frame.
        if ((type & GIGE_SYNC_IN_PULSE_MASK) == GIGE_SYNC_IN_PPF)
        {
            // Set the 1PPF enable and the skip count.
            fpga_ptr->regModify(ICM_LUX13HS_CONTROL,
                                (ICMX13HSC_SYNC_MODE_1PPF_SLAVE | ICMX13HSC_1PPF_SKIP_COUNT((type & GIGE_SYNC_IN_SKIP_MASK) >> GIGE_SYNC_IN_SKIP_SHIFT)),
                                ~(ICMX13HSC_SYNC_MODE_MASK | ICMX13HSC_1PPF_SKIP_COUNT_MASK));
        }
        else
        {
            // Set the 1PPS enable - no skip count with this mode.
            fpga_ptr->regModify(ICM_LUX13HS_CONTROL,
                                ICMX13HSC_SYNC_MODE_1PPS_SLAVE,
                                ~(ICMX13HSC_SYNC_MODE_MASK | ICMX13HSC_1PPF_SKIP_COUNT_MASK));
        }
    }
}

void Luxima13HS::shutdown()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Shutdown the DACs.
    if (fpga_ptr != nullptr)
    {
        uint32_t reg_value;

        // Put the sensor interface into reset.
        reg_value = ICMX13HSC_ALTLVDS_RX_RESET | ICMX13HSC_SENSOR_RESET |
                    ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT | ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP;
        if (!fpga_ptr->regWrite(ICM_LUX13HS_CONTROL, reg_value))
            fi_log(LOG_ERR, "%s: Unable to put sensor into reset!", __PRETTY_FUNCTION__);
        else
            fi_log(LOG_DEBUG, "%s: Sensor in reset.", __PRETTY_FUNCTION__);
    }
}

bool Luxima13HS::set_frame_rate(float frame_rate, double clock_factor)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    if (!LuximaHS::set_frame_rate(frame_rate, clock_factor))
    {
        fi_log(LOG_DEBUG, "%s: parent set_frame_rate() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we tell FPGA the number of intergrations pulses
    // to use (i.e. the frame rate) if we were in 1PPS sync mode.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_1PPS_FRAME_CNT, CameraServer::get_cs()->frame_rate_to_reg(frame_rate)))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_1PPS_FRAME_CNT.", __PRETTY_FUNCTION__);
        return false;
    }

    return true;
}

bool Luxima13HS::dac_init(FiFPGA *fpga_ptr)
{
    int i;
    bool rc(false);

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Loop through our table of commands.
    for (i = 0; i < (int)(sizeof(lux13hs_dac_settings) / sizeof(lux13hs_dac_settings[0])); ++i)
    {
        uint8_t chip_addr = lux13hs_dac_settings[i].chip;
        uint8_t reg_addr = lux13hs_dac_settings[i].reg;
        uint16_t reg_value = lux13hs_dac_settings[i].val << DT_SHIFT;

        // fi_log(LOG_DEBUG, "%02X[%02X] <= 0x%04X [0x%04X << %d]", chip_addr, reg_addr, reg_value, lux13hs_dac_settings[i].val, DT_SHIFT);

        // Write the DAC setting. NOTE: This is a 2-byte write.
        if (!fpga_ptr->i2cWrite(chip_addr, reg_addr, reg_value, FI_FPGA::I2C_SENSOR, 2))
            goto err_exit;
    }

    // Everything is OK.
    rc = true;

err_exit:
    if (!rc)
        fi_log(LOG_ERR, "%s: I/O error", __PRETTY_FUNCTION__);

    // Return our status.
    return rc;
}

int Luxima13HS::sensor_clock_rate() const
{
    return 133000000;
}

//--------------------------------------------------------------------
// Luxima160 Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima160::Luxima160(SENSOR::EEPROM_INFO *eeprom)
    : LuximaHS(
          4704,  /* max width */
          3424,  /* max height */
          320,  /* min width */
          240,  /* min height */
          eeprom)
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
    config_word = LUX160_RECOMMENDED;
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima160::~Luxima160()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

bool Luxima160::dac_init(FiFPGA *fpga_ptr)
{
    (void)fpga_ptr;
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
    return false;
}

void Luxima160::enable_clpout()
{
}

bool Luxima160::is_bit_mode_supported(bitmode_t bit_mode)
{
    bool rc = false;

    switch ((int)bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
    case GIGE_BITMODE_7_0:
    case GIGE_BITMODE_MIDDLE_8_BITS:
    case GIGE_BITMODE_8_1:
    case GIGE_BITMODE_UPPER_8_BITS:
    case GIGE_BITMODE_9_2:
    case GIGE_BITMODE_10_BITS:
    case GIGE_BITMODE_9_0:
        rc = true;
        break;
    default:
        break;
    }

    return rc;
}

/*!
 * @brief Check to see if the passed in sensor_opts parameter is supported.
 * @param sensor_opts   The sensor options, one of GIGE_SENSOPT_xxx.
 * @returns True if the option is supported, false otherwise.
 */
bool Luxima160::is_option_supported(uint32_t /* sensor_opts */)
{
    // No binning or subsampling.
    return false;
}

int Luxima160::sensor_clock_rate() const
{
    return 100000000;
}

//--------------------------------------------------------------------
// Luxima19HS Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima19HS::Luxima19HS(SENSOR::EEPROM_INFO *eeprom)
    : Luxima13HS(eeprom)
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    sensor_max_width = 1920;
    sensor_max_height = 1080;
    sensor_min_width = 64;
    sensor_min_height = 64;
    vtroi_x = vtroi_y = vtroi_width = vtroi_height = 0;

    config_word = LUX19HS_RECOMMENDED;
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
Luxima19HS::~Luxima19HS()
{
    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);
}

void Luxima19HS::bit_slip()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: Slip the bits", __PRETTY_FUNCTION__);

    // Reset the bitslip logic.
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_RESET_BL, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_RESET_BR, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_RESET_TL, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_RESET_TR, 0xfffff);

    // Slip all channels by one.
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_BL, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_BR, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_TL, 0xfffff);
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_BITSLIP_TR, 0xfffff);
}

bool Luxima19HS::class_init(FiFPGA *fpga_ptr)
{
    bool rc;

    LOG("%s", __PRETTY_FUNCTION__);

    // Perform the base class initialization.
    rc = Luxima13HS::class_init(fpga_ptr);
    if (rc)
    {
        // For the 19HS, we need to adjust the PD_N offset value.
        fpga_ptr->regWrite(ICM_LUX13HS_PD_N_OFFSET, LUX19HS_PD_N_OFFSET);
    }

    // Return the result.
    return rc;
}

bool Luxima19HS::dac_init(FiFPGA *fpga_ptr)
{
    int i;
    bool rc(false);

    fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

    // Loop through our table of commands.
    for (i = 0; i < (int)(sizeof(lux19hs_dac_settings) / sizeof(lux19hs_dac_settings[0])); ++i)
    {
        uint8_t chip_addr = lux19hs_dac_settings[i].chip;
        uint8_t reg_addr = lux19hs_dac_settings[i].reg;
        uint16_t reg_value = lux19hs_dac_settings[i].val << DT_SHIFT;

        // fi_log(LOG_DEBUG, "%02X[%02X] <= 0x%04X [0x%04X << %d]", chip_addr, reg_addr, reg_value, lux19hs_dac_settings[i].val, DT_SHIFT);

        // Write the DAC setting. NOTE: This is a 2-byte write.
        if (!fpga_ptr->i2cWrite(chip_addr, reg_addr, reg_value, FI_FPGA::I2C_SENSOR, 2))
            goto err_exit;
    }

    // Everything is OK.
    rc = true;

err_exit:
    if (!rc)
        fi_log(LOG_ERR, "%s: I/O error", __PRETTY_FUNCTION__);

    // Return our status.
    return rc;
}

void Luxima19HS::enable_clpout()
{
    // Not going to enable clpout for the 19HS.
}

bool Luxima19HS::enable_sensor_test()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    extern int sensor_test_type;

    LOG("%s: sensor_test_type=%d", __PRETTY_FUNCTION__, sensor_test_type);

    // If the type is '5', then we want to do the sensor's built-in test.
    // Otherwise, just pass it on to the base class.
    if (sensor_test_type != 5)
    {
        Luxima13HS::enable_sensor_test();
    }
    else
    {
        // Remember that we're in test mode.
        sensor_test_on = true;

        // Enable the test pattern bit in the control word.
        config_word |= TEST_PATTERN;
        if (!fpga_ptr->regWrite(ICM_LUX13HS_SERIAL_CONFIG, config_word))
        {
            fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_SERIAL_CONFIG", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // All done.
    return true;
}

/*!
 * @brief Return the amount of time (in microseconds) the FPGA/Sensor requires to
 *    move a frame, etc. AKA "Frame Overhead".
 */
uint32_t Luxima19HS::get_frame_overhead()
{
    return (uint32_t)(((float)LUX19HS_PD_N_END_CLOCK / (float)sensor_clock_rate_usec()) + 0.5);
}

uint32_t Luxima19HS::get_readout_delay()
{
    uint32_t rdout_delay;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    rdout_delay = regs[GIGE_REG_SENSOR_RDOUT_DELAY].get_value();
    if ((rdout_delay < (uint32_t)get_readout_delay_min()) || (rdout_delay > (uint32_t)get_readout_delay_max()))
    {
        LOG("%s: Bad value [0x%08x], switching to default [0x%08x]", __PRETTY_FUNCTION__, rdout_delay, LUX19HS_READOUT_DELAY_DEFAULT);
        rdout_delay = LUX19HS_READOUT_DELAY_DEFAULT;
    }

    LOG("%s: 0x%08x", __PRETTY_FUNCTION__, rdout_delay);

    return rdout_delay;
}

int Luxima19HS::get_readout_delay_min()
{
    return LUX19HS_MIN_READOUT_DELAY;
}

int Luxima19HS::get_readout_delay_max()
{
    return LUX19HS_MAX_READOUT_DELAY;
}

/*!
 * @brief Return the minimum allowed shutter, in microseconds. To say this another way,
 *    return the fastest shutter period we can support.
 */
uint32_t Luxima19HS::get_sensor_min_shutter()
{
    return (uint32_t)(((float)LUX19HS_TX_N_START / (float)sensor_clock_rate_usec()) + 0.5);
}

bool Luxima19HS::make_sensor_valid(
    uint16_t *rtn_startx,
    uint16_t *rtn_starty,
    uint16_t *rtn_width,
    uint16_t *rtn_height,
    bitmode_t *rtn_bit_mode,
    float *rtn_frame_rate,
    uint32_t * /* rtn_shutter_speed */,
    uint32_t *rtn_sensor_opts)
{
    uint16_t startx = *rtn_startx;
    uint16_t starty = *rtn_starty;
    uint16_t width = *rtn_width;
    uint16_t height = *rtn_height;
    uint32_t sensor_opts = *rtn_sensor_opts;
    bitmode_t bit_mode = *rtn_bit_mode;
    float max_frame_rate;
    uint32_t any_changes = 0;
    uint32_t frame_time;

    LOG("%s", __PRETTY_FUNCTION__);

    // We support all the bit modes so this should never happen.
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Invalid bit_mode, using 0x%02x", __FUNCTION__, GIGE_BITMODE_11_4);
        bit_mode = (bitmode_t)GIGE_BITMODE_9_2;
        any_changes |= (1 << 0);
    }

    // Don't leave the bit mode setting in one of the old-style values.
    switch (bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
        bit_mode = GIGE_BITMODE_7_0;
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
        bit_mode = GIGE_BITMODE_8_1;
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
        bit_mode = GIGE_BITMODE_9_2;
        break;
    case GIGE_BITMODE_10_BITS:
        bit_mode = GIGE_BITMODE_9_0;
        break;
    default:
        break;
    }

    // We don't support sensor opts.
    if (sensor_opts && (sensor_opts != ((1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT))))
    {
        fi_log(LOG_DEBUG, "%s: sensor_opts are not supported!", __PRETTY_FUNCTION__);
        sensor_opts = ((1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT));
        any_changes |= (1 << 3);
    }

    (void)force_legal_roi(0, &startx, &starty, &width, &height);

    if ((startx != *rtn_startx) ||
        (starty != *rtn_starty) ||
        (width  != *rtn_width) ||
        (height != *rtn_height) ||
        (bit_mode != *rtn_bit_mode) ||
        (sensor_opts != *rtn_sensor_opts))
    {
        // fi_log(LOG_DEBUG, "%s: was: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );
        *rtn_startx = startx;
        *rtn_starty = starty;
        *rtn_width = width;
        *rtn_height = height;
        *rtn_sensor_opts = sensor_opts;
        *rtn_bit_mode = bit_mode;
        // fi_log(LOG_DEBUG, "%s: now: %ux%u@%u,%u 0x%x 0x%08x", __FUNCTION__, *rtn_width, *rtn_height, *rtn_startx, *rtn_starty, *rtn_bit_mode, *rtn_sensor_opts );

        any_changes |= (1 << 7);
    }

    // The maximum frame rate depends on the height of the readout.
    frame_time = ((LUX19HS_DEFAULT_QROW_TIME * height / 4) + LUX19HS_FRAME_OVERHEAD);
    max_frame_rate = sensor_clock_rate() / frame_time;
    if (*rtn_frame_rate > max_frame_rate)
    {
        *rtn_frame_rate = max_frame_rate;
        any_changes |= (1 << 8);
    }

    // Did we make any changes to the settings?
    if (any_changes != 0)
    {
        //fi_log(LOG_DEBUG, "%s: changed 0x%08x", __FUNCTION__, any_changes );
        return true;
    }
    else
        return false;
}

int Luxima19HS::sensor_clock_rate() const
{
    // Use the MCLK here. PCLK is (MCLK * 2), but all these calculations are
    // based on MCLK (so far!).
    return 87500000;
}

bool Luxima19HS::set_integration_time(uint32_t shutter_speed)
{
    uint32_t integration_time;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // LOG("%s: shutter_speed=0x%08x [%u]", __PRETTY_FUNCTION__, shutter_speed, shutter_speed);

    // Figure out the integration time.
    integration_time = sensor_clock_rate_usec() * shutter_speed;

    // Tell the sensor to update it's integration period.
    // Setup the integration pulse to happen at the end of the frame time.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_PD_N_HIGH_COUNT, integration_time))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_PD_N_HIGH_COUNT (ss=%d, time=0x%x)", __PRETTY_FUNCTION__, shutter_speed, integration_time);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_PD_N_LOW_COUNT, LUX19HS_PD_N_END_CLOCK))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_PD_N_LOW_COUNT (ss=%d, time=0x%x)", __PRETTY_FUNCTION__, shutter_speed, integration_time);
        return false;
    }

    // Determine where the TX pulse needs to be. The pulse goes active LOW, so the low count
    // is when the pulse starts and the high count is when the pulse ends.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX_N_LOW_COUNT, LUX19HS_TX_N_START))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX_N_LOW_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX19HS_TX_N_START);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX_N_HIGH_COUNT, LUX19HS_TX_N_DONE))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX_N_HIGH_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX19HS_TX_N_DONE);
        return false;
    }

    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX2_N_LOW_COUNT, LUX19HS_TX2_N_START))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX2_N_LOW_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX19HS_TX2_N_START);
        return false;
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_TX2_N_HIGH_COUNT, LUX19HS_TX2_N_DONE))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_TX2_N_HIGH_COUNT (0x%x)", __PRETTY_FUNCTION__, LUX19HS_TX2_N_DONE);
        return false;
    }

    // All is well.
    return true;
}

bool Luxima19HS::update_capture(
    uint16_t true_startx,
    uint16_t true_starty,
    uint16_t roi_width,
    uint16_t roi_height,
    fpnmode_t /* fpn_mode */,
    bitmode_t bit_mode,
    float frame_rate,
    uint32_t integration,
    uint32_t /* sensor_opts */,
    uint32_t /* sensor_gain */)
{
    int i;
    bool rc(false);
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s: %ux%u@(%u,%u)", __PRETTY_FUNCTION__, roi_width, roi_height, true_startx, true_starty);

    // Need to make sure our startx is multiples of 2 and starty is multiples of 4.
    if ((true_startx & 1) || (true_starty & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! startx=0x%x, starty=0x%x", __PRETTY_FUNCTION__, true_startx, true_starty);
        true_startx = (true_startx + 1) & ~1;
        true_starty = (true_starty + 3) & ~3;
    }
    if ((roi_width & 1) || (roi_height & 3))
    {
        fi_log(LOG_DEBUG, "%s: MISALIGNMENT! width=0x%x, height=0x%x", __PRETTY_FUNCTION__, roi_width, roi_height);
        roi_width = (roi_width + 1) & ~1;
        roi_height = (roi_height + 3) & ~3;
    }
    if (!is_bit_mode_supported(bit_mode))
    {
        fi_log(LOG_DEBUG, "%s: Invalid bit_mode=0x%x", __PRETTY_FUNCTION__, bit_mode);
        bit_mode = GIGE_BITMODE_9_2;
    }

    // Clear the LVDS enable bits and set the reset bits..
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_RESET, ~ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA);

    // Initialize the sensor settings.
    fpga_ptr->icmSetFrameRate(frame_rate);
    fpga_ptr->icmSetShutterSpeed(integration);

    // Initialize the readout delay for the first column.
    set_readout_start();

    // Initialize the timing registers
    reg_value = ICMX19HSCFGB_SET_QROW_TIME(0, LUX19HS_DEFAULT_QROW_TIME);
    reg_value = ICMX19HSCFGB_SET_QROW_COUNT(reg_value, roi_height / 4);
    reg_value = ICMX19HSCFGB_SET_FIRST_QROW(reg_value, true_starty / 4);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_B = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_B, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_B", __PRETTY_FUNCTION__);
        goto bail;
    }

    reg_value = ICMX13HSCFGC_SET_ROW_ADDR_INC_COUNT(0, LUX19HS_ROW_ADDR_INCR_TIME);
    reg_value = ICMX13HSCFGC_SET_ST_READ_BOT_N_COUNT(reg_value, DEFAULT_ROW_CLOCK);
    reg_value = ICMX13HSCFGC_SET_ST_ROW_BOT_N_COUNT(reg_value, DEFAULT_ROW_CLOCK);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_C = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_C, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_C", __PRETTY_FUNCTION__);
        goto bail;
    }

    reg_value = ICMX13HSCFGD_SET_ST_READ_TOP_N_COUNT(0, LUX19HS_ROW_ADDR_INCR_TIME + DEFAULT_ROW_CLOCK);
    reg_value = ICMX13HSCFGD_SET_ST_ROW_TOP_N_COUNT(reg_value, LUX19HS_ROW_ADDR_INCR_TIME + DEFAULT_ROW_CLOCK);
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_READOUT_TIMING_CONFIG_D = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_TIMING_CONFIG_D, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_READOUT_TIMING_CONFIG_D", __PRETTY_FUNCTION__);
        goto bail;
    }

    if (!fpga_ptr->regWrite(ICM_LUX13HS_ROI_V_CONFIG, roi_height))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_ROI_V_CONFIG", __PRETTY_FUNCTION__);
        goto bail;
    }

    // ------------------------------------------------------------------------
    // This is a little strange, so follow along...
    // - For the start, we want the starting BLOCK and the starting PIXEL within that block.
    // - For the end, we want the ending BLOCK and the number of pixels withing that block.
    // Note the difference: one is an offset and the other is a count!
    // ------------------------------------------------------------------------
    reg_value = ICMX13HSROIH_SET_FIRST_ACTIVE_BLOCK(0, true_startx / BLOCK_SIZE);
    reg_value = ICMX13HSROIH_SET_FIRST_ACTIVE_BLOCK_OFFSET(reg_value, true_startx % BLOCK_SIZE);
    reg_value = ICMX13HSROIH_SET_LAST_ACTIVE_BLOCK_OFFSET(reg_value, ((true_startx + roi_width - 1) / BLOCK_SIZE));
    reg_value = ICMX13HSROIH_SET_LAST_ACTIVE_BLOCK_PIXEL_COUNT(reg_value, (((true_startx + roi_width - 1) % BLOCK_SIZE) + 1));
    fi_log(LOG_DEBUG, "%s: ICM_LUX13HS_ROI_H_CONFIG = 0x%08x", __PRETTY_FUNCTION__, reg_value);
    {
        fi_log(LOG_DEBUG, "--> startx=%d, roi_width=%d, block pixel count=0x%04X", true_startx, roi_width, (((true_startx + roi_width - 1) % BLOCK_SIZE) + 1));
    }
    if (!fpga_ptr->regWrite(ICM_LUX13HS_ROI_H_CONFIG, reg_value))
    {
        fi_log(LOG_ERR, "%s: Unable to write ICM_LUX13HS_ROI_H_CONFIG", __PRETTY_FUNCTION__);
        goto bail;
    }

    if (!set_fpga_bit_mode(bit_mode))
    {
        fi_log(LOG_ERR, "%s: Unable to set the bitmode", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Clear out the readout delay values.
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_STARTUP_READOUT_DELAY, 0);

    // Enable the LVDS lines.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~(ICMX13HSC_ALTLVDS_RX_PLL_ARESET_TOP | ICMX13HSC_ALTLVDS_RX_PLL_ARESET_BOT)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Wait a bit to give the PLLs a chance to lock.
    for (i = 0; i < 10; ++i)
    {
        // Take a knee...
        using us = std::chrono::microseconds;
        std::this_thread::sleep_for(us(5));

        // Read the status to see if we've got lock.
        if (!fpga_ptr->regRead(ICM_LUX13HS_STATUS, &reg_value))
        {
            fi_log(LOG_DEBUG, "%s: Unable to read ICM_LUX13HS_STATUS", __PRETTY_FUNCTION__);
            goto bail;
        }
        if ((reg_value & ICMX13HSS_ALTLVDS_RX_PLL_LOCKED) == ICMX13HSS_ALTLVDS_RX_PLL_LOCKED)
            break;
    }

    // Release the reset follwed by enabling the receivers.
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_ALTLVDS_RX_RESET))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (!fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA, -1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_LUX13HS_CONTROL", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Bit slip the channels.
    bit_slip();

    // All is well.
    rc = true;

bail:
    // Do we want to reset the sensor on fail?
    // if (!rc)
    //     hard_reset();
    return rc;
}

uint32_t Luxima19HS::vtrig_control(const char *settings_str, int str_length)
{
    Json::Value requested_settings;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    bool parsingSuccessful;
    bool updateROI;
    uint32_t status;
    uint32_t reg_val;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

//    LOG("%s: settings [%s]", __PRETTY_FUNCTION__, settings_str);

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(settings_str, settings_str + str_length, &requested_settings, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, settings_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        status = GEV_STATUS_INVALID_PARAMETER;
        goto bail;
    }

    // Read any and all ROI settings.
    updateROI = false;
    if (!requested_settings["roi_x"].empty())
    {
        try
        {
            uint16_t offset_x;
            CameraServer *cs_ptr = CameraServer::get_cs();

            // Get some settings we'll need to figure the ROI out.
            cs_ptr->get_sensor_start(&offset_x, nullptr);

            // Get the value sent by the user.
            reg_val = requested_settings["roi_x"].asUInt();

            // Calculate the real x-coordinate.
            vtroi_x = reg_val + offset_x;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_x].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_y"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_y"].asUInt();
            vtroi_y = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_y].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_width"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_width"].asUInt();
            vtroi_width = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_width].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }
    if (!requested_settings["roi_height"].empty())
    {
        try
        {
            reg_val = requested_settings["roi_height"].asUInt();
            vtroi_height = reg_val;
            updateROI = true;
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[roi_height].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // Do we have width/height?
    if (updateROI)
    {
        // Convert the ROI into the FPGA settings.
        reg_val = IL19VT_SET_ROI_H_CONFIG_A((vtroi_x % BLOCK_SIZE), (((vtroi_x + vtroi_width) % BLOCK_SIZE) + 1));
        fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_ROI_H_CONFIG_A, reg_val);

        reg_val = IL19VT_SET_ROI_H_CONFIG_B((vtroi_x / BLOCK_SIZE), ((vtroi_x + vtroi_width) / BLOCK_SIZE));
        fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_ROI_H_CONFIG_B, reg_val);

        reg_val = IL19VT_SET_ROI_V(vtroi_y, (vtroi_y + vtroi_height));
        fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_ROI_V, reg_val);
    }

    // Do we have the threshold settings?
    if (!requested_settings["pix_thresh"].empty() || !requested_settings["area_thresh"].empty())
    {
        uint32_t pix_thresh;
        uint32_t area_thresh;

        // Initialize our thresholds.
        fpga_ptr->regRead(ICM_LUX19HS_VTRIG_THRESHOLDS, &reg_val);

        if (!requested_settings["pix_thresh"].empty())
        {
            try
            {
                pix_thresh = requested_settings["pix_thresh"].asUInt();
            }
            catch (Json::Exception& e)
            {
                fi_log(LOG_DEBUG, "%s: requested_settings[pix_thresh].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
                pix_thresh = 10;
            }

            if ((pix_thresh == 0) || (pix_thresh > 100))
            {
                fi_log(LOG_DEBUG, "%s: Illegal pix_thresh [%u]!", __PRETTY_FUNCTION__, pix_thresh);
                pix_thresh = (reg_val & ILVT_CHANGE_THRESHOLD_MASK);
            }
            else
            {
                // Calculate the actual threshold value.
                pix_thresh = (255 * pix_thresh) / 100;
            }
        }
        else
        {
            // Just reuse the existing value.
            pix_thresh = (reg_val & ILVT_CHANGE_THRESHOLD_MASK);
        }

        if (!requested_settings["area_thresh"].empty())
        {
            try
            {
                area_thresh = requested_settings["area_thresh"].asUInt();
            }
            catch (Json::Exception& e)
            {
                fi_log(LOG_DEBUG, "%s: requested_settings[area_thresh].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
                area_thresh = 10;
            }

            if ((area_thresh == 0) || (area_thresh > 100))
            {
                fi_log(LOG_DEBUG, "%s: Illegal pix_thresh [%u]!", __PRETTY_FUNCTION__, pix_thresh);
                area_thresh = (reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT;
            }
            else
            {
                // Calculate the actual area value.
                area_thresh = (vtroi_width * vtroi_height * area_thresh) / 100;
            }
        }
        else
        {
            // Just reuse the existing value.
            area_thresh = (reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT;
        }

        reg_val = ILVT_SET_THRESHOLD(pix_thresh, area_thresh);
        fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_THRESHOLDS, reg_val);
    }

    // Do we have the interval settings?
    if (!requested_settings["interval"].empty())
    {
        try
        {
            reg_val = requested_settings["interval"].asUInt();
            fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_ROI_BUF_UPDATE_INTERVAL, reg_val);
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[interval].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // Is the control setting available?
    if (!requested_settings["control"].empty())
    {
        try
        {
            reg_val = requested_settings["control"].asUInt();
            fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_CONTROL, reg_val);
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[control].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }
    }

    // To latch these values, we need to write to the status register.
    fpga_ptr->regWrite(ICM_LUX19HS_VTRIG_STATUS, 0);

    // All done.
    status = GEV_STATUS_SUCCESS;

bail:
    return status;
}

uint32_t Luxima19HS::vtrig_info_json(char *dest_ptr, uint32_t max_len)
{
    Json::Value vtrig_json;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;
    uint32_t reg_val;
    uint32_t config_a;
    uint32_t config_b;
    int x;
    int width;
    int height;
    int thresh;
    uint16_t offset_x;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    // Get some settings we'll need to figure the ROI out.
    cs_ptr->get_sensor_start(&offset_x, nullptr);

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // Read the register settings and populate the JSon response.
    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_ROI_H_CONFIG_A, &config_a);
    vtrig_json["roi_ha"] = config_a;
    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_ROI_H_CONFIG_B, &config_b);
    vtrig_json["roi_hb"] = config_b;
    x = ((config_b & IL19VT_FIRST_ACTIVE_PPW_MASK) * BLOCK_SIZE) + (config_a & IL19VT_FIRST_PPW_START_MASK);
    width = (((config_b >> 8) * BLOCK_SIZE) + ((config_a >> 8) - 1)) - x;
    vtrig_json["roi_x"] = x - offset_x;
    vtrig_json["roi_width"] = width;

    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_ROI_V, &reg_val);
    vtrig_json["roi_v"] = reg_val;
    vtrig_json["roi_y"] = (reg_val & IL19VT_FIRST_ROW_MASK);
    height = (reg_val >> ILVT_LAST_ROW_SHIFT) - (reg_val & IL19VT_FIRST_ROW_MASK);
    vtrig_json["roi_height"] = height;

    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_THRESHOLDS, &reg_val);
    vtrig_json["threshold"] = reg_val;
    thresh = (int)((((float)(reg_val & ILVT_CHANGE_THRESHOLD_MASK) * 100.0) / 255.0) + 0.5);
    if (thresh <= 0)
        thresh = 1;
    vtrig_json["pix_thresh"] = thresh;
    thresh = (int)((((float)((reg_val & ILVT_COUNT_THRESHOLD_MASK) >> ILVT_COUNT_THRESHOLD_SHIFT) * 100.0) / (float)(width * height)) + 0.5);
    if (thresh <= 0)
        thresh = 1;
    vtrig_json["area_thresh"] = thresh;

    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_ROI_BUF_UPDATE_INTERVAL, &reg_val);
    vtrig_json["interval"] = reg_val;
    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_CONTROL, &reg_val);
    vtrig_json["control"] = reg_val;
    fpga_ptr->regRead(ICM_LUX19HS_VTRIG_STATUS, &reg_val);
    vtrig_json["status"] = reg_val;

    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 2;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, vtrig_json);

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
        return 0;
    }

    // Copy the string over.
//    LOG("%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
    memcpy(dest_ptr, json_message.c_str(), json_message.length());

    // Return the length, including the null-terminator.
    return json_message.length() + 1;
}

int Luxima19HS::get_capture_fifo_count()
{
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fpga_ptr->regRead(ICM_LUX13HS_STATUS, &reg_value);
    return ((reg_value & ICMX19HSS_CAPSTART_FIFO_LEVEL_MASK) >> ICMX19HSS_CAPSTART_FIFO_LEVEL_SHIFT);
}

void Luxima19HS::reset_capture_fifo()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Toggle the FIFO reset bit.
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX19HSC_CAPSTART_FIFO_RESET, -1);
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX19HSC_CAPSTART_FIFO_RESET);
}

void Luxima19HS::reset_event_fifo()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Toggle the FIFO reset bit.
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX19HSC_EVENT_FIFO_RESET, -1);
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX19HSC_EVENT_FIFO_RESET);
}
