/*!
 * @file LuximaHS_FPN.cpp
 * @brief Functions to support Luxima FPN and PRNU correction within the FPGA.
 *
 * Copyright 2017-2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <fmt/printf.h>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "ct_file.h"
#include "image_correction.h"
#include "luxima_hs.h"
#include "part_mgr.h"
#include "storage.h"

namespace LUXIMA_HS_FPN
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Uncomment to dump intermediate files.
//#define DEBUG_DUMP_FILES 1

#ifdef DEBUG_DUMP_FILES
const char CAL_FRAME_FILE[]         = ".config/Fastec/scratch/fpncal";
const char DEAD_PIX_FRAME_FILE[]    = ".config/Fastec/scratch/deadpix";
const char HOT_PIX_FRAME_FILE[]     = ".config/Fastec/scratch/hotpix";
const char RDOUT_FRAME_FILE[]       = ".config/Fastec/scratch/rdoutdly";
const char PRNU_FRAME_FILE[]        = ".config/Fastec/scratch/prnu";
#endif

// How many frames to we want to grab?
const int FPN_CAPTURE_FRAMES    = 8;
const int FPN_DELAY_FRAMES  = 2;

const int BP_MAX_BAD_PIXELS = 32;
const int BP_LOWER_PIXEL_BOUND  = 4;
const int BP_UPPER_PIXEL_BOUND  = 0xF0;

const int RDOUT_DITHER = 10;

//! Structure to hold coordinates of a defective pixel.
typedef struct DEFECT_ENTRY
{
    uint16_t x;
    uint16_t y;
} DEFECT_ENTRY;

/*!
 * @brief Handy alias to avoid unnecessary typing.
 */
template<class T> using unique_ptr_aligned = std::unique_ptr<T, decltype(free) *>;

} /* End of namespace LUXIMA_HS_FPN */

using namespace LUXIMA_HS_FPN;

/*!
 * @brief The class declaration for Luxima's calibration functions.
 */
class LuximaHSCalibration
{
public:
    LuximaHSCalibration();
    virtual ~LuximaHSCalibration();

    bool delayCal();
    bool runFpnCalibration(bool is_factory);
    bool runPrnuCalibration(const uint8_t *ct_ptr);

private:
    int fpn_part;                           //!< The currently active partition information.
    bool do_factory_cal;                    //!< Is true if we are doing factory calibration.
    CT_GT_FILE::CT_ROI ct_roi;              //!< The ROI used for this calibration capture.
    std::vector<DEFECT_ENTRY> defect_list;

    uint8_t *raw_table;                     //!< An aligned pointer to the raw data.
    unique_ptr_aligned<uint8_t[]> raw_table_base;   //!< A place to store data.

    uint64_t cal_frame_addr;
    uint32_t cal_frame_size;

    int fpnCalc();
    void fpnCalRunFactory();
    void fpnCalRunUser();
    int prnuCalc(const uint8_t *ct_ptr);

    unique_ptr_aligned<uint8_t[]> allocateAligned(size_t alignment, size_t length);
    uint64_t calcMidrow(int delay_val);
    uint64_t frameAddress(int frame_num);
    void deadPixelDetect();
    void hotPixelDetect();
    void roiRead(CT_GT_FILE::CT_ROI *roi);
    void writeDefectMap();
};

/*!
 * @brief The constructor.
 */
LuximaHSCalibration::LuximaHSCalibration() :
    raw_table_base(nullptr, nullptr),
    cal_frame_addr(0),
    cal_frame_size(0)
{
    // Initialize our member variables.
    raw_table = nullptr;
    fpn_part = 0;
    do_factory_cal = false;
}

LuximaHSCalibration::~LuximaHSCalibration()
{
}

/*!
 * @brief An aligned memory allocation function returning a std::unique_ptr.
 *
 * std::unique_ptr takes a deleter "type", so we use decltype(). The
 * decltype(free) gives back a function type of “void (void*)” but we
 * need a pointer to a function. So, we say “decltype(free)*” which
 * gives us “void (*)(void*)”.
 *
 * @param alignment [IN] The address of the allocated memory will be a multiple
 *                       of alignment, which must be a power of two and a multiple
 *                      of sizeof(void *).
 * @param length    [IN] The number of bytes to allocate.
 *
 * @sa https://www.codeproject.com/Articles/820931/Using-std-unique-ptr-RAII-with-malloc-and-free
 * @sa https://embeddedartistry.com/blog/2017/2/23/c-smart-pointers-with-aligned-mallocfree
 */
unique_ptr_aligned<uint8_t[]> LuximaHSCalibration::allocateAligned(size_t alignment, size_t length)
{
    uint8_t *raw = nullptr;

    // Ignore errors for now. The pointer will go back as nullptr.
    posix_memalign((void **)&raw, alignment, length);

    // Return our unique pointer.
    return unique_ptr_aligned<uint8_t[]>(raw, free);
}

/*!
 * @brief Detect "dead" pixels in the sensor.
 *
 * A number (FPN_CAPTURE_FRAMES) of frames are captured with an assumed brightness
 * of ~70% and averaged. Any average pixel below a threshold (BP_LOWER_PIXEL_BOUND)
 * is assumed to be "dead" and added to the defect_list.
 */
void LuximaHSCalibration::deadPixelDetect()
{
    int row;
    int col;
    int frame_idx;
    int pixel_idx;
    bool shutter_was_open;
    uint8_t *raw_table_ptr;
    uint16_t *raw_sum_ptr;
    uint32_t frame_len;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    std::unique_ptr<uint16_t[]> raw_sum_table;

    // We do NOT want the shutter closed. Remember the current state.
    shutter_was_open = fpga_ptr->isShutterOpen();
    if (!shutter_was_open)
    {
        // Open the shutter.
        fpga_ptr->shutterClose(false);
    }

    LOG("%s: shutter_was_open: %c", __PRETTY_FUNCTION__, (shutter_was_open ? 'T' : 'F'));

    // Capture our frames.
    if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
    {
        fi_log(LOG_DEBUG, "%s: unable to capture FPN_CAPTURE_FRAMES", __PRETTY_FUNCTION__);
        goto deadpix_exit;
    }

    // The calibration frame size contains the metadata bytes, and we don't
    // care about those.
    frame_len = cal_frame_size - FRAME_METADATA_HW_BYTES;

    // Allocate our buffer aligned nicely.
    raw_table_base = allocateAligned(4096, frame_len);
    if (raw_table_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of frame buffer failed!", __PRETTY_FUNCTION__);
        goto deadpix_exit;
    }
    raw_table = raw_table_base.get();

    // Allocate our "summation" table.
    raw_sum_table = std::unique_ptr<uint16_t[]>(new uint16_t[frame_len]());
    if (raw_sum_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
        goto deadpix_exit;
    }
    memset(raw_sum_table.get(), 0, sizeof(raw_sum_table[0]) * frame_len);

    // Loop through our frames.
    for (frame_idx = 0; frame_idx < FPN_CAPTURE_FRAMES; ++frame_idx)
    {
        uint32_t read_count;
        uint64_t frame_addr;

        // Read the frame from Image Memory.
        frame_addr = frameAddress(frame_idx);
        read_count = fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, frame_addr, raw_table, frame_len);
        if (read_count != frame_len)
        {
            fi_log(LOG_ERR, "%s: Can't read the raw frame at 0x%" PRIx64 " (0x%x!=0x%x).",
                   __PRETTY_FUNCTION__, frame_addr, read_count, frame_len);
            goto deadpix_exit;
        }

        // ------------------------------------------------------------------------
        // Uncomment constant at top of file.
#ifdef DEBUG_DUMP_FILES
        {
            std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), DEAD_PIX_FRAME_FILE);
            if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
            {
                fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
            }

            std::string tmp_file = fmt::sprintf("%s/dp_img_%03d", tmp_dir, frame_idx);
            int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
            if (fd < 0)
            {
                fi_log(LOG_ERR, "%s: Can't open %s (%s)", __FUNCTION__, tmp_file.c_str(), strerror(errno));
            }
            else
            {
                // Try writing the data.
                write(fd, raw_table, frame_len);
                close(fd);
            }
        }
#endif /* DEBUG_DUMP_FILES */
        // ------------------------------------------------------------------------

        // Add this frame into the summation.
        raw_table_ptr = raw_table;
        raw_sum_ptr = raw_sum_table.get();
        for (pixel_idx = 0; pixel_idx < (int32_t)frame_len; ++pixel_idx)
        {
            // Add in the new pixel value.
            *raw_sum_ptr += (uint16_t)(*raw_table_ptr);

            // Bump our pointers.
            ++raw_table_ptr;
            ++raw_sum_ptr;
        }
    }

    // Walk all the pixels in our summation vector to find any "dead" pixels.
    for (row = 0; row < ct_roi.height; row++)
    {
        for (col = 0; col < ct_roi.width; col++)
        {
            int offset = (row * ct_roi.width) + col;
            uint8_t avg_pixel = (uint8_t)(raw_sum_table[offset] / FPN_CAPTURE_FRAMES);

            // Is this pixel dead?
            if (avg_pixel <= BP_LOWER_PIXEL_BOUND)
            {
                DEFECT_ENTRY entry;

                // Push the defect on our list.
                entry.x = (uint16_t)col;
                entry.y = (uint16_t)row;
                defect_list.push_back(entry);
            }
        }
    }

deadpix_exit:
    // Reset our pointer.
    raw_table = nullptr;
    raw_table_base.reset();

    // Restore the shutter state.
    if (!shutter_was_open)
    {
        // Close the shutter.
        fpga_ptr->shutterClose(true);
    }

    LOG("%s: defect_list has %d entries", __PRETTY_FUNCTION__, defect_list.size());
}

/*!
 * @brief Look for bad "hot" pixels.
 *
 * It is assumed that the raw_table contains the averaged samples of the calibration
 * image. Any pixel found to be above a threshold (BP_UPPER_PIXEL_BOUND) is added
 * to the defect_list.
 */
void LuximaHSCalibration::hotPixelDetect()
{
    int row;
    int col;

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (raw_table == nullptr)
    {
        fi_log(LOG_ERR, "%s: raw_table is NULL!?", __PRETTY_FUNCTION__);
        return;
    }

    // ------------------------------------------------------------------------
    // Uncomment constant at top of file.
#ifdef DEBUG_DUMP_FILES
    {
        uint32_t frame_len;

        // The calibration frame size contains the metadata bytes, and we don't
        // care about those.
        frame_len = cal_frame_size - FRAME_METADATA_HW_BYTES;

        std::string tmp_file = fmt::sprintf("%s/%s", get_home_path(), HOT_PIX_FRAME_FILE);
        int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: Can't open %s (%s)", __FUNCTION__, tmp_file.c_str(), strerror(errno));
        }
        else
        {
            write(fd, raw_table, frame_len);
            close(fd);
        }
    }
#endif /* DEBUG_DUMP_FILES */
    // ------------------------------------------------------------------------

    // Walk all the pixels to find any "hot" pixels.
    for (row = 0; row < ct_roi.height; row++)
    {
        for (col = 0; col < ct_roi.width; col++)
        {
            int offset = (row * ct_roi.width) + col;

            // Is this pixel dead?
            if (raw_table[offset] > BP_UPPER_PIXEL_BOUND)
            {
                DEFECT_ENTRY entry;

                // Push the defect on our list.
                entry.x = (uint16_t)col;
                entry.y = (uint16_t)row;
                defect_list.push_back(entry);
                fi_log(LOG_INFO, "%s: Found defect (0x%02X) at [%d, %d]", __PRETTY_FUNCTION__, raw_table[offset], col, row);
            }
        }
    }

    if (defect_list.size() > BP_MAX_BAD_PIXELS)
        fi_log(LOG_ERR, "%s: defect_list has too many %d entries!", __PRETTY_FUNCTION__, defect_list.size());
    else
        fi_log(LOG_INFO, "%s: defect_list has %d entries", __PRETTY_FUNCTION__, defect_list.size());
}

/*!
 * @brief Write out the defect map.
 * @return true on success, false if failure
 */
void LuximaHSCalibration::writeDefectMap()
{
    int offset;
    FILE *fd = NULL;
    std::string fname;
    std::vector<DEFECT_ENTRY>::iterator it;
    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());
    union
    {
        uint32_t type_32;
        uint8_t  type_8[4];
    } sensor_type;

    LOG("%s", __PRETTY_FUNCTION__);

    fname = sensor_ptr->get_defect_filename(".config/Fastec/scratch");
    if (fname.empty())
        return;

    fd = fopen(fname.c_str(), "w");
    if (fd == NULL)
    {
        fi_log(LOG_ERR, "%s: Can't open %s (%s)", __FUNCTION__, fname.c_str(), strerror(errno));
        return;
    }

    // Make sure we anotate the file if we have too many bad pixels.
    if (defect_list.size() > BP_MAX_BAD_PIXELS)
    {
        fprintf(fd,
                "#=================================#\n"
                "# THERE ARE TOO MANY BAD PIXELS!! #\n"
                "#=================================#\n");
    }

    // Write the human consumable part of the defect file.
    sensor_type.type_32 = sensor_ptr->get_sensor_type();
    fprintf(fd, "%c%c%c%cv1:%d\n",
            sensor_type.type_8[0], sensor_type.type_8[1],
            sensor_type.type_8[2], sensor_type.type_8[3],
            (int)defect_list.size());
    for (it = defect_list.begin(); it != defect_list.end(); ++it)
    {
        uint16_t x = it->x;
        uint16_t y = it->y;

        if (fprintf(fd, "%u,%u\n", x, y) < 0)
        {
            fi_log(LOG_ERR, "%s: Write to %s failed (%s)", __FUNCTION__, fname.c_str(), strerror(errno));
            goto bail;
        }
    }

    // Now write the part to generate the EEPROM data.
    fprintf(fd, "\n\n"
            "##################################\n"
            "Clip the following for the EEPROM info\n"
            "##################################\n");

    // Get the starting offset for the bad pixel information.
    offset = (int)offsetof(SENSOR::EEPROM_INFO, bad_pixels);

    // Now output the commands for writing the bad pixel info
    // to the EEPROM.
    for (it = defect_list.begin(); it != defect_list.end(); ++it)
    {
        uint16_t x = it->x;
        uint16_t y = it->y;

        fprintf(fd, "isbeeprom 0x%02X 0x%02X\n", offset++, (x & 0xFF));
        fprintf(fd, "isbeeprom 0x%02X 0x%02X\n", offset++, ((x >> 8) & 0xFF));
        fprintf(fd, "isbeeprom 0x%02X 0x%02X\n", offset++, (y & 0xFF));
        fprintf(fd, "isbeeprom 0x%02X 0x%02X\n", offset++, ((y >> 8) & 0xFF));
    }

    fi_log(LOG_INFO, "%s: Wrote %d defects to '%s'", __FUNCTION__, defect_list.size(), fname.c_str());

bail:
    if (fd != NULL)
        fclose(fd);
}

/*!
 * @brief Calibrate the readout delay for the Luxima sensor.
 * @return True on success, false otherwise.
 */
bool LuximaHSCalibration::delayCal()
{
    int i;
    int best_delay = 0;
    bool rc = false;
    bool shutter_was_open;
    int64_t best_delta = 0;
    int64_t last_midrow_dif = 0;
    uint64_t last_midrow_value = 0;

    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    LOG("%s", __PRETTY_FUNCTION__);

    // We do NOT want the shutter closed. Remember the current state.
    shutter_was_open = fpga_ptr->isShutterOpen();
    if (!shutter_was_open)
    {
        // Open the shutter.
        fpga_ptr->shutterClose(false);
    }

    //===========================================================================
    // DELAY CALIBRATION BEGIN
    //===========================================================================
    // Clear the LVDS enable bits and set the reset bits..
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_RESET, ~ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA);

    // Clear out the readout delay values.
    fpga_ptr->regWrite(ICM_LUX13HS_LVDS_RX_STARTUP_READOUT_DELAY, 0);

    // Clear the reset bits.
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, 0, ~ICMX13HSC_ALTLVDS_RX_RESET);

    // Set the enable bits.
    fpga_ptr->regModify(ICM_LUX13HS_CONTROL, ICMX13HSC_ALTLVDS_RX_PAR_OUT_ENA, -1);

    // Loop through the various possible delay values. For the color cameras, at least,
    // we want to increment by even values.
    //
    // NOTE: We may want to parameterize the increment value, rather than a hardcoded '2'.
    for (i = sensor_ptr->get_readout_delay_min(); i < sensor_ptr->get_readout_delay_max(); i += 2)
    {
        bool brighter = false;
        uint64_t midrow_val = 0;
        int64_t midrow_dif = 0;
        int64_t dif_delta = 0;

        fi_log(LOG_DEBUG, "%s: Trying delay %d...", __PRETTY_FUNCTION__, i);

        // Set the readout delay.
        if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_DELAY, i))
        {
            fi_log(LOG_DEBUG, "%s: unable to set ICM_LUX13HS_READOUT_DELAY", __PRETTY_FUNCTION__);
            goto cal_exit;
        }

        // Capture our frames.
        if (!fpga_ptr->icmCalibrateCapture(FPN_DELAY_FRAMES))
        {
            fi_log(LOG_DEBUG, "%s: unable to capture FPN_DELAY_FRAMES", __PRETTY_FUNCTION__);
            goto cal_exit;
        }

        // Calculate the midrow column value.
        midrow_val = calcMidrow(i);

        // Was there a difference between this time and last time?
        if (last_midrow_value != 0)
        {
            // What is the difference between this time and last?
            if (midrow_val > last_midrow_value)
                midrow_dif = midrow_val - last_midrow_value;
            else
                midrow_dif = last_midrow_value - midrow_val;
        }

        // What is the delta between the differences? If the difference
        // is 0, then leave the dif_delta at zero too - no change.
        if (midrow_dif != 0)
        {
            // Did we get brighter or darker? Brighter means there are
            // fewer dark columns. Or it could just be the start condition...
            if (midrow_dif > last_midrow_dif)
            {
                brighter = true;
                dif_delta = midrow_dif - last_midrow_dif;
            }
            else
            {
                brighter = false;
                dif_delta = last_midrow_dif - midrow_dif;
            }
        }

        // Could this be a winner?
        if (brighter && (dif_delta > best_delta))
        {
            // This is our best delay so far.
            best_delay = i;
            best_delta = dif_delta;
            fi_log(LOG_DEBUG, "%s: new best_delay(%d)", __PRETTY_FUNCTION__, i);
        }

        // Remember the difference for next time.
        last_midrow_dif = midrow_dif;

        // Remember this midrow value for the next loop.
        last_midrow_value = midrow_val;
        fi_log(LOG_DEBUG, "%s: Delay:%d, midrow:0x%" PRIi64 ", dif:%" PRIi64 ", delta:%" PRIi64 ", brighter:%d", __PRETTY_FUNCTION__, i, midrow_val, midrow_dif, dif_delta, brighter);
    }

    if (best_delay != 0)
    {
        // Save the readout delay.
        fi_log(LOG_DEBUG, "%s: Picking %d as best delay", __PRETTY_FUNCTION__, best_delay);
        sensor_ptr->set_readout_delay(best_delay);
        rc = true;
    }

    // Set the readout delay to our (maybe) new value.
    if (!fpga_ptr->regWrite(ICM_LUX13HS_READOUT_DELAY, sensor_ptr->get_readout_delay()))
    {
        fi_log(LOG_DEBUG, "%s: unable to set ICM_LUX13HS_READOUT_DELAY", __PRETTY_FUNCTION__);
        goto cal_exit;
    }

    //===========================================================================
    // DELAY CALIBRATION END
    //===========================================================================

cal_exit:

    // Restore the shutter state.
    if (!shutter_was_open)
    {
        // Close the shutter.
        fpga_ptr->shutterClose(true);
    }

    LOG("%s: done (%d)", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Run the factory FPN calibration functions.
 *
 * It is assumed that the image is a flat-field, gray image. The mean
 * pixel value should be between 50-75% full scale under sampling
 * conditions.
 *
 * The factory calibration routine automatically configures the camera
 * to run at full ROI, 60 FPS, full exposure (359.9 degrees).
 */
void LuximaHSCalibration::fpnCalRunFactory()
{
    int     ret;
    uint32_t    total_size;
    uint32_t    buf_count;
    uint16_t    width;
    uint16_t    height;
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (sensor_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sensor_ptr is NULL!? Is this a LuximaHS sensor?", __PRETTY_FUNCTION__);
        return;
    }

    // If we are doing a factory calibration, the settings were restored to defaults
    // before starting us but that might have a smaller ROI because of licensing.
    width = sensor_ptr->get_sensor_max_width();
    height = sensor_ptr->get_sensor_max_height();
    ct_roi.origin.x = 0;
    ct_roi.origin.y = 0;
    ct_roi.width = width;
    ct_roi.height = height;

    // Set up capture parameters for factory calibration. This will only be writing to
    // the Unarmed Live View buffers, so the capture buffers don't need to be set yet.
    // The size of the ULV buffers is based on the maximum sensor area without regard
    // to license so we can still capture full-size images even if the license limits
    // us to a smaller size.
    // For factory calibration, we never enable the sensor/column FPN.
    sensor_ptr->start_calibration_capture(true, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);

    // Set up the memory buffer and size that we will capture black frames into.
    part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);

    // Figure out the total size we have for capturing.
    total_size = cal_frame_size * buf_count;

    // The size we got from PM matches the camera settings but it might be adjusted for
    // license restrictions. Factory is always full size so we need to recompute the size.
    cal_frame_size = cs_ptr->calc_frame_buf_size(width, height, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    cal_frame_size = convert_to_HEXWORD(cal_frame_size + FRAME_METADATA_HW_BYTES) << HEXWORD_SHIFT;

    // Now figure out how many buffers we can actually fit.
    buf_count = total_size / cal_frame_size;

    // Make sure we have enough space for the buffers we need to capture. It's
    // OK if we have more buffers than we need.
    if (buf_count < FPN_CAPTURE_FRAMES)
    {
        fi_log(LOG_DEBUG, "%s: Not enough space for capture buffers [buf_count=%d]?!", __PRETTY_FUNCTION__, buf_count);
        goto cal_exit;
    }
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
    fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

    // Calibrate the readout delay offset.
    delayCal();

    // Reset our defect list vector.
    defect_list.clear();

    // Look for "dead" pixels.
    deadPixelDetect();

    // Close the shutter.
    fpga_ptr->shutterClose(true);

    // Capture FPN_CAPTURE frames regardless of the currently-selected mode.
    if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
    {
        fi_log(LOG_ERR, "%s: calibrate_capture() failed!", __PRETTY_FUNCTION__);
        goto cal_exit;
    }

    // Calculate the FPN correction table. This will create /scratch/factory_blkref
    // which we will need to copy.
    if (fpnCalc() < 0)
        goto cal_exit;

    // Look for "hot" pixels in the data we captured.
    hotPixelDetect();

    // Everything is good, tell the script to copy the calibration file to the ro partition
    errno = 0;
    ret = system(CT_GT_FILE::FACTORY_BLKREF_SCRIPT);
    if ((ret < 0) || !WIFEXITED(ret))
    {
        // problem log the error
        fi_log(LOG_ERR, "%s: Cannot save the black reference frame (%d, %d)!", __PRETTY_FUNCTION__,
               errno, WEXITSTATUS(ret));
        goto cal_exit;
    }

    // Save out the defect list.
    writeDefectMap();

cal_exit:
    // Make sure the shutter is open.
    fpga_ptr->shutterClose(false);
}

/*!
 * @brief Run the user FPN calibration functions.
 *
 * It is assumed that the image is a flat-field, gray image. The mean
 * pixel value should be between 50-75% full scale under sampling
 * conditions.
 *
 * For User calibration, leave the file on "scratch" where we constructed it and
 * allow a little slop at the low end so we don't demand that all pixels be
 * above zero in adjusting the VREF.
 */
void LuximaHSCalibration::fpnCalRunUser()
{
    uint32_t    buf_count;
    uint32_t    total_size;
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());

    // Sanity check.
    if (sensor_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sensor_ptr is NULL!? Is this a LuximaHS sensor?", __PRETTY_FUNCTION__);
        return;
    }

    // Remember the user's ROI.
    roiRead(&ct_roi);

    // Set up capture parameters for user calibration. This will only be writing to
    // the Unarmed Live View buffers, so the capture buffers don't need to be set yet.
    sensor_ptr->start_calibration_capture(false, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);

    // Now capture black frames and average them.

    // Get the memory buffer and size that we will capture into.
    part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);
    if (buf_count < FPN_CAPTURE_FRAMES)
        goto cal_exit;

    // The size we got from PM matches the current camera settings but we don't use the same
    // bit mode that the user set. That could result in a different frame size than
    // what we capture as a black frame so recompute for an 8-bit capture.
    total_size = cal_frame_size * buf_count;
    cal_frame_size = cs_ptr->calc_frame_buf_size(ct_roi.width, ct_roi.height, (bitmode_t)GIGE_BITMODE_LOWER_8_BITS);
    cal_frame_size = convert_to_HEXWORD(cal_frame_size + FRAME_METADATA_HW_BYTES) << HEXWORD_SHIFT;
    buf_count = total_size / cal_frame_size;

    // Set the capture buffer layout for our black frames.
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
    fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

    // Close the shutter.
    fpga_ptr->shutterClose(true);

    // Capture FPN_CAPTURE frames regardless of the currently-selected mode.
    if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
        goto cal_exit;

    // Calculate the FPN correction table.
    if (fpnCalc() < 0)
        goto cal_exit;

    // Everything is good. User black reference file is already in the right place.
    errno = 0;

cal_exit:
    // Make sure the shutter is open.
    fpga_ptr->shutterClose(false);
}

/*!
 * @brief Do the FPN calibration work.
 * @param is_factory    [IN] True if we're doing factory FPN calibration, false for user.
 * @returns True on success, false otherwise.
 */
bool LuximaHSCalibration::runFpnCalibration(bool is_factory)
{
    bool rc = false;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();

    LOG("%s: is_factory=%d", __PRETTY_FUNCTION__, is_factory);

    // Remember if we're doing factory or user calibration.
    do_factory_cal = is_factory;

    // Save the currently active partition information.
    fpn_part = part_mgr->getActivePartition();
    if (fpn_part < 0)
    {
        fi_log(LOG_ERR, "%s: No active partition?!", __PRETTY_FUNCTION__);
        fpn_part = 0;
        part_mgr->setActivePartition(0);
        part_mgr->queueCapturePartition(0);
    }

    // Allocate memory for our "raw" histogram.
    raw_table_base = allocateAligned(4096, fpga_ptr->HISTOGRAM_RAW_SIZE);
    if (raw_table_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation for raw histogram failed!", __PRETTY_FUNCTION__);
        goto cal_exit;
    }

    // When we run the calibration, we will restart capture with pixel FPN disabled.
    // For Pixel FPN calibration, we will use GIGE_BITMODE_LOWER_8_BITS.
    // If we are in factory calibration, we will use the full size of the sensor instead of the
    // license-restricted dimensions. This could change the image memory layout but
    // we will set the capture area ourselves and restore when we return to normal capture.
    if (do_factory_cal)
        fpnCalRunFactory();
    else
        fpnCalRunUser();
    rc = true;

cal_exit:

    // All done.
    return rc;
}

/*!
 * @brief Do the PRNU calibration work.
 * @param ct_ptr    [IN] The user black reference table to use for PRNU.
 * @returns True on success, false otherwise.
 */
bool LuximaHSCalibration::runPrnuCalibration(const uint8_t *ct_ptr)
{
    bool rc = false;
    uint32_t    total_size;
    uint32_t    buf_count;
    uint16_t    width;
    uint16_t    height;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    CameraServer *cs_ptr = CameraServer::get_cs();
    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (sensor_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sensor_ptr is NULL!? Is this a LuximaHS sensor?", __PRETTY_FUNCTION__);
        goto cal_exit;
    }

    // If we are doing a PRNU calibration, we want to make sure the settings are
    // how we need them.
    width = sensor_ptr->get_sensor_max_width();
    height = sensor_ptr->get_sensor_max_height();
    ct_roi.origin.x = 0;
    ct_roi.origin.y = 0;
    ct_roi.width = width;
    ct_roi.height = height;

    // Only need to capture images for mono sensors.
    if (!sensor_ptr->is_sensor_color())
    {
        // Save the currently active partition information.
        fpn_part = part_mgr->getActivePartition();
        if (fpn_part < 0)
        {
            fi_log(LOG_ERR, "%s: No active partition?!", __PRETTY_FUNCTION__);
            fpn_part = 0;
            part_mgr->setActivePartition(0);
            part_mgr->queueCapturePartition(0);
        }

        // Set up capture parameters for PRNU calibration. This will only be writing to
        // the calibration buffers, so the capture buffers don't need to be set yet.
        // The size of the buffers is based on the maximum sensor area without regard
        // to license so we can still capture full-size images even if the license limits
        // us to a smaller size.
        //
        // For PRNU calibration, we never enable the sensor/column FPN.
        sensor_ptr->start_prnu_capture();

        // Set up the memory buffer and size that we will capture white frames into.
        part_mgr->getCalibrationBuf(&cal_frame_addr, &cal_frame_size, &buf_count);

        // Figure out the total size we have for capturing.
        total_size = cal_frame_size * buf_count;

        // The size we got from PM matches the camera settings but it might be adjusted for
        // license restrictions. Factory is always full size so we need to recompute the size.
        cal_frame_size = cs_ptr->calc_frame_buf_size(width, height, (bitmode_t)GIGE_BITMODE_10_BITS);
        cal_frame_size = convert_to_HEXWORD(cal_frame_size + FRAME_METADATA_HW_BYTES) << HEXWORD_SHIFT;
        if (cal_frame_size == 0)
        {
            fi_log(LOG_ERR, "%s: cal_frame_size is 0?!!", __PRETTY_FUNCTION__);
            goto cal_exit;
        }

        // Now figure out how many buffers we can actually fit.
        buf_count = total_size / cal_frame_size;

        // Make sure we have enough space for the buffers we need to capture. It's
        // OK if we have more buffers than we need.
        if (buf_count < FPN_CAPTURE_FRAMES)
        {
            fi_log(LOG_DEBUG, "%s: Not enough space for capture buffers [buf_count=%d]?!", __PRETTY_FUNCTION__, buf_count);
            goto cal_exit;
        }
        fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(cal_frame_size));
        fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, buf_count);
        fpga_ptr->icmSetCaptureBuffer(cal_frame_addr);

        // Capture FPN_CAPTURE frames regardless of the currently-selected mode.
        if (!fpga_ptr->icmCalibrateCapture(FPN_CAPTURE_FRAMES))
        {
            fi_log(LOG_ERR, "%s: calibrate_capture() failed!", __PRETTY_FUNCTION__);
            goto cal_exit;
        }
    }

    // Calculate the PRNU table. This will create the FACTORY_GAIN_FILE_RO file.
    if (prnuCalc(ct_ptr) < 0)
        goto cal_exit;

    // All is well.
    rc = true;

cal_exit:
    // All done.
    return rc;
}

/*!
 * @brief Read the current ROI from the camera server.
 *
 * @param roi   Pointer to where we want to put the ROI values.
 *
 * @return 0 on success, <0 on failure.
 */
void LuximaHSCalibration::roiRead(CT_GT_FILE::CT_ROI *roi)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint16_t start_x;
    uint16_t start_y;
    uint16_t width;
    uint16_t height;

    // Sanity check.
    if (roi == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: roi is null?!", __PRETTY_FUNCTION__);
        return;
    }

    // Get the ROI.
    cs_ptr->get_user_roi(&start_x, &start_y, &width, &height);

    // Pass the values to our structure.
    roi->origin.x = start_x;
    roi->origin.y = start_y;
    roi->width = width;
    roi->height = height;
}

/*!
 * @brief Calculate the sum of the average values for the middle 4 columns
 *        of the image. This is used for the delay calibration.
 *
 * @sa LuximaHSCalibration::delayCal
 *
 * @param delay_val     [UNUSED] The readout delay value used.
 * @return The sum of the average values for the middle 4 columns.
 */
uint64_t LuximaHSCalibration::calcMidrow(int delay_val)
{
    int avg_value;
    int width;
    uint8_t threshold;
    uint32_t i;
    uint64_t accum;
    uint64_t frame_addr;
    uint32_t read_count;
    uint32_t offset;
    uint32_t frame_len;
    uint8_t *raw_ptr;
    std::unique_ptr<uint8_t[]> raw_binary;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    (void)delay_val;

    // Initialize
    accum = -1;
    width = (ct_roi.width + 3) & ~3;

    // The calibration frame size contains the metadata bytes, and we don't
    // care about those.
    frame_len = cal_frame_size - FRAME_METADATA_HW_BYTES;

    // Allocate our buffer aligned nicely.
    raw_table_base = allocateAligned(4096, frame_len);
    if (raw_table_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of frame buffer failed!", __PRETTY_FUNCTION__);
        return -1;
    }
    raw_table = raw_table_base.get();

    // Read the frame from Image Memory.
    frame_addr = frameAddress(0);
    read_count = fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, frame_addr, raw_table, frame_len);
    if (read_count != frame_len)
    {
        fi_log(LOG_ERR, "%s: Can't read the raw frame at 0x%" PRIx64 " (0x%x!=0x%x).",
               __PRETTY_FUNCTION__, frame_addr, read_count, frame_len);
        goto mr_exit;
    }

    // ------------------------------------------------------------------------
    // Uncomment constant at top of file.
#ifdef DEBUG_DUMP_FILES
    {
        std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), RDOUT_FRAME_FILE);
        if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
        {
            fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
        }

        std::string tmp_file = fmt::sprintf("%s/rdimg_%03d", tmp_dir, delay_val);
        int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
        }
        else
        {
            // Save out the information.
            write(fd, raw_table, frame_len);
            close(fd);
        }
    }
#endif /* DEBUG_DUMP_FILES */
    // ------------------------------------------------------------------------

    // Allocate our "binary" image.
    raw_binary = std::make_unique<uint8_t[]>(frame_len);

    // Convert our image to an image that's just black and white. First, find
    // the "average" pixel value.
    accum = 0;
    raw_table = raw_table_base.get();
    for (i = 0; i < frame_len; ++i)
    {
        // Used in calculating the overall average black level.
        accum += *raw_table++;
    }
    avg_value = (int)(accum / frame_len);

    // Now that we have the average value, we'll set our threshold offset from that value.
    // Anything over our threshold will be set to "white", and anything below will be "black".
    threshold = avg_value - RDOUT_DITHER;
    fi_log(LOG_INFO, "%s: Got average pixel value = %u [0x%x] (accum=%" PRIu64 ", frame_len=%u", __FUNCTION__, avg_value, avg_value, accum, frame_len);
    raw_ptr = raw_binary.get();
    raw_table = raw_table_base.get();
    for (i = 0; i < frame_len; ++i)
    {
        if (*raw_table >= threshold)
            *raw_ptr = 0xFF;
        else
            *raw_ptr = 0x00;
        ++raw_ptr;
        ++raw_table;
    }

    // ------------------------------------------------------------------------
    // Uncomment constant at top of file.
#ifdef DEBUG_DUMP_FILES
    {
        std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), RDOUT_FRAME_FILE);
        if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
        {
            fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
        }

        std::string tmp_file = fmt::sprintf("%s/rdimg_%03d_mono", tmp_dir, delay_val);
        int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
        }
        else
        {
            // Save out the information.
            write(fd, raw_binary.get(), frame_len);
            close(fd);
        }
    }
#endif /* DEBUG_DUMP_FILES */
    // ------------------------------------------------------------------------

    // Walk down the center columns.
    accum = 0;
    raw_table = raw_table_base.get();
    for (offset = width / 2; offset < frame_len; offset += width)
    {
        // uint32_t *ch_ptr = (uint32_t *)&raw_table[offset];
        // accum += *ch_ptr;

        // We want to straddle the mid-point of the image.
        uint32_t hold_var = 0;
        uint8_t *col_ptr = &raw_table[offset - 2];

        // Copy those center 4 bytes (pixels) into our hold register.
        memcpy(&hold_var, col_ptr, 4);

        // Accumulate our column values.
        accum += hold_var;
    }

mr_exit:
    // Release the memory.
    raw_table = nullptr;
    raw_table_base.reset();

    // Return our result.
    return accum;
}

/*!
 * @brief Calculate the address for the calibration frame specified.
 * @param frame_num     [IN] The number of the frame we are interested in.
 * @return The address of the frame in image memory.
 */
uint64_t LuximaHSCalibration::frameAddress(int frame_num)
{
    return cal_frame_addr + (frame_num * cal_frame_size);
}

/*!
 * @brief Perform the FPN calibration specified (factory or user).
 * @return 0 on success, <0 otherwise.
 */
int LuximaHSCalibration::fpnCalc()
{
    int     frame_idx;
    int     pixel_idx;
    int     file_fd;
    int     retval;
    uint16_t    *raw_sum_ptr;
    uint8_t     *raw_table_ptr;
    uint32_t    framesize;  // Size of the raw frame.
    uint32_t    cur_gain;
    size_t      ct_header_size;
    uint64_t    offset;
    std::string blkref_file;
    CT_GT_FILE::CT_FILE_HDR ct_header;
    std::unique_ptr<uint16_t[]> raw_sum_table;

    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *part_mgr = CameraServer::get_partmgr();
    Sensor *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    cur_gain = regs[ GIGE_REG_SENSOR_GAIN ].get_value();

    // Assume failure.
    retval = -1;

    // Prep the info structure.
    memset(&ct_header, 0, sizeof(ct_header));

    // The calibration frame size contains the metadata bytes, and we don't
    // care about those.
    framesize = cal_frame_size - FRAME_METADATA_HW_BYTES;

    // Allocate our buffer aligned nicely.
    raw_table_base = allocateAligned(4096, framesize);
    if (raw_table_base.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of frame buffer failed!", __PRETTY_FUNCTION__);
        return -1;
    }
    raw_table = raw_table_base.get();

    // Allocate our "summation" table.
    raw_sum_table = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (raw_sum_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }
    memset(raw_sum_table.get(), 0, sizeof(raw_sum_table[0]) * framesize);

    // Loop through our frames.
    for (frame_idx = 0; frame_idx < FPN_CAPTURE_FRAMES; ++frame_idx)
    {
        uint64_t frame_addr;
        uint32_t read_count;

        // Read the frame from Image Memory.
        frame_addr = frameAddress(frame_idx);
        read_count = fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, frame_addr, raw_table, framesize);
        if (read_count != framesize)
        {
            fi_log(LOG_ERR, "%s: Can't read the raw frame at 0x%" PRIx64 " (0x%x!=0x%x).",
                   __PRETTY_FUNCTION__, frame_addr, read_count, framesize);
            goto fpn_exit;
        }

        // ------------------------------------------------------------------------
        // Uncomment constant at top of file.
#ifdef DEBUG_DUMP_FILES
        {
            std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), CAL_FRAME_FILE);
            if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
            {
                fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
            }

            // Open up the black reference file.
            std::string tmp_file = fmt::sprintf("%s/cal_%03d", tmp_dir, frame_idx);
            int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
            if (fd < 0)
            {
                fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
            }
            else
            {
                // Save out the information.
                write(fd, raw_table, framesize);
                close(fd);
            }
        }
#endif /* DEBUG_DUMP_FILES */
        // ------------------------------------------------------------------------

        // Add this frame into the summation.
        raw_table_ptr = raw_table;
        raw_sum_ptr = raw_sum_table.get();
        for (pixel_idx = 0; pixel_idx < (int32_t)framesize; ++pixel_idx)
        {
            // Add in the new pixel value.
            *raw_sum_ptr += (uint16_t)(*raw_table_ptr);

            // Bump our pointers.
            ++raw_table_ptr;
            ++raw_sum_ptr;
        }
    }

    // Now run through our summation table and generate our average table.
    raw_table_ptr = raw_table;
    raw_sum_ptr = raw_sum_table.get();
    offset = 0;
    for (pixel_idx = 0; pixel_idx < (int32_t)framesize; ++pixel_idx)
    {
        // Calculate and save the average.
        *raw_table_ptr = (uint8_t)(*raw_sum_ptr / FPN_CAPTURE_FRAMES);

        // Used in calculating the overall average black level.
        offset += *raw_table_ptr;

        // Bump our pointers.
        ++raw_table_ptr;
        ++raw_sum_ptr;
    }
    offset /= framesize;
    fi_log(LOG_INFO, "%s: Got average black level = %u [0x%x]", __FUNCTION__, (uint32_t)offset, (uint16_t)(offset & 0xFFFF));

    // Write version 1 of the CT header.
    ct_header_size = sizeof(CT_GT_FILE::CT_FILE_HDR);

    ct_header.magic = CT_GT_FILE::CT_MAGIC;
    ct_header.length = framesize;
    ct_header.sensor_params[0] = (cur_gain << 16);

    // Get the ROI used for this calibration.
    ct_header.roi = ct_roi;

    // Make sure the partition ROI reflects 32-bit alignment.
    ct_header.roi.width = (ct_header.roi.width + 3) & ~3;

    // Associate this correction table with the sensor.
    sensor_ptr->get_sensor_ver(ct_header.sensor_id, CT_GT_FILE::SENSOR_ID_LEN);

    // Get the file name we are writing.
    if (do_factory_cal)
    {
        blkref_file = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_BLKREF_FILE_RW);
    }
    else
    {
        blkref_file = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::USER_BLKREF_FILE);
    }

    // Open up the black reference file.
    file_fd = open(blkref_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
    if (file_fd < 0)
    {
        fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, blkref_file.c_str(), strerror(errno));
        goto fpn_exit;
    }
    else
    {
        // Save out the information.
        write(file_fd, &ct_header, ct_header_size);
        write(file_fd, raw_table, framesize);
    }

    // Close the file.
    close(file_fd);

    // If we write it to image memory successfully, we succeeded.
    if (part_mgr->writeCorrectionTable(raw_table, framesize))
        retval = 0;

fpn_exit:
    // Release the memory if we're in error.
    if (retval)
    {
        raw_table = nullptr;
        raw_table_base.reset();
    }

    // Read up the frame from the image memory.
    return retval;
}

/*!
 * @brief Perform the PRNU calibration.
 * @param ct_ptr    [IN] Pointer to the FPN correction table in use.
 * @return 0 on success, <0 otherwise.
 */
int LuximaHSCalibration::prnuCalc(const uint8_t *ct_ptr)
{
    int     file_fd;
    int     retval;
    uint16_t    *avg_ptr;
    uint32_t    framesize;          // Size of the raw frame, in pixels.
    uint32_t    frame_bytecount;    // Size of the raw frame, in bytes.
    std::string prnu_file;
    CT_GT_FILE::GT_FILE_HDR gt_header;
    std::unique_ptr<uint16_t[]> avg_table;
    std::unique_ptr<uint8_t[]> gain_table;

    PartMgr *part_mgr = CameraServer::get_partmgr();
    LuximaHS *sensor_ptr = static_cast<LuximaHS *>(CameraServer::get_sensor());

    LOG("%s", __PRETTY_FUNCTION__);

    // Assume failure.
    retval = -1;

    // Prep the info structure.
    memset(&gt_header, 0, sizeof(gt_header));

    // Since we are dealing with 16-bit values, we don't have to worry about row alignment.
    framesize = ct_roi.width * ct_roi.height;
    frame_bytecount = framesize * 2;
    LOG("%s: frame_bytecount=%d", __PRETTY_FUNCTION__, frame_bytecount);

    // Allocate our "averages" table.
    avg_table = std::unique_ptr<uint16_t[]>(new uint16_t[framesize]());
    if (avg_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of averages buffer failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Allocate our "gain" table.
    gain_table = std::unique_ptr<uint8_t[]>(new uint8_t[framesize]());
    if (gain_table.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: creation of averages buffer failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Are we a color sensor?
    if (sensor_ptr->is_sensor_color())
    {
        // Have the FPGA build up our color image. This will use the red, green, and blue
        // raw TIFF images that have previously been captured.
        if (!FiImageCorrection::prnuColorSum(FPN_CAPTURE_FRAMES, avg_table.get()))
        {
            fi_log(LOG_DEBUG, "%s: prnu_color_sum() failed!", __PRETTY_FUNCTION__);
            goto fpn_exit;
        }
    }
    else
    {
        int pixel_idx;
        uint32_t *raw_sum_ptr;
        std::unique_ptr<uint32_t[]> raw_sum_table;

        // Allocate our "summation" table.
        raw_sum_table = std::unique_ptr<uint32_t[]>(new uint32_t[framesize]());
        if (raw_sum_table.get() == nullptr)
        {
            fi_log(LOG_ERR, "%s: creation of summation buffer failed!", __PRETTY_FUNCTION__);
            goto fpn_exit;
        }
        memset(raw_sum_table.get(), 0, sizeof(raw_sum_table[0]) * framesize);

        // Download the capture images and sum them together.
        FiImageCorrection::prnuMonoSum(FPN_CAPTURE_FRAMES, GIGE_BITMODE_9_0, cal_frame_addr, raw_sum_table.get());

        // Now run through our summation table and generate our average table.
        avg_ptr = avg_table.get();
        raw_sum_ptr = raw_sum_table.get();
        for (pixel_idx = 0; pixel_idx < (int32_t)framesize; ++pixel_idx)
        {
            uint32_t avg_value;

            // Calculate and save the average. Make sure we saturate rather than overflow.
            avg_value = (*raw_sum_ptr / FPN_CAPTURE_FRAMES);
            if (avg_value > 0xFFFF)
                avg_value = 0xFFFF;
            *avg_ptr = (uint16_t)(avg_value & 0xFFFF);

            // Bump our pointers.
            ++avg_ptr;
            ++raw_sum_ptr;
        }
    }

    // ------------------------------------------------------------------------
#ifdef DEBUG_DUMP_FILES
    // I'm cranky with the formatter not indenting properly. Using a do-while false is equivalent to empty braces, so...
    do
    {
        std::string tmp_dir = fmt::sprintf("%s/%s", get_home_path(), PRNU_FRAME_FILE);
        if ((storage::mkpath(tmp_dir.c_str(), 0777) == -1) && (errno != ENOTDIR))
        {
            fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, tmp_dir.c_str(), errno, strerror(errno));
        }

        // Open up the black reference file.
        std::string tmp_file = fmt::sprintf("%s/cal_avg", tmp_dir);
        int fd = open(tmp_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, tmp_file.c_str(), strerror(errno));
        }
        else
        {
            // Save out the information.
            write(fd, avg_table.get(), frame_bytecount);
            close(fd);
        }
    } while (0);
#endif /* DEBUG_DUMP_FILES */
    // ------------------------------------------------------------------------

    // Using the averaged white frame, calculate what the gain table should be.
    if (!FiImageCorrection::prnuGenerate(avg_table.get(), ct_ptr, gain_table.get()))
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr->prnu_gen() failed!", __PRETTY_FUNCTION__);
        goto fpn_exit;
    }

    // Initialze our file header.
    memset(&gt_header, 0, sizeof(gt_header));
    gt_header.magic = CT_GT_FILE::CT_MAGIC;
    gt_header.length = ct_roi.width * ct_roi.height;
    gt_header.roi.width = ct_roi.width;
    gt_header.roi.height = ct_roi.height;
    sensor_ptr->get_sensor_ver(gt_header.sensor_id, CT_GT_FILE::SENSOR_ID_LEN);

    // Get the file name we are writing.
    prnu_file = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::FACTORY_GAIN_FILE_RO);

    // Open up the black reference file.
    file_fd = open(prnu_file.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0666);
    if (file_fd < 0)
    {
        fi_log(LOG_ERR, "%s: Can't open %s: %s.", __PRETTY_FUNCTION__, prnu_file.c_str(), strerror(errno));
        goto fpn_exit;
    }
    else
    {
        // Save out the information.
        write(file_fd, &gt_header, sizeof(gt_header));
        write(file_fd, gain_table.get(), framesize);
    }

    // Close the file.
    close(file_fd);

    // If we write it to image memory successfully, we succeeded.
    if (part_mgr->writeGainTable(gain_table.get(), framesize))
        retval = 0;

fpn_exit:
    // Read up the frame from the image memory.
    return retval;
}

/*!
 * @brief Get the sensor-specific black offset from the CT parameters.
 *
 * @param sensor_type   The sensor type read from the file.
 * @param params    Pointer to buffer containing sensor-specific params.
 *
 * @return true for success, false for failure
 */
uint16_t LuximaHS::get_cal_offset(uint32_t sensor_type, void *params)
{
#ifdef ALM_TODO
    uint16_t ct_offset;

    ct_offset = *((uint16_t *)((uint8_t *)params + LUPA_OFS_IDX));
    return ct_offset;
#else
    (void)sensor_type;
    (void)params;

    return 0;
#endif /* ALM_TODO */
}

/*!
 * @brief Say whether this fpn change requires that the sensor be restarted.
 *
 * If we do sensor+pixel FPN, then whenever GIGE_FPN_FPGA_ENABLE_MASK is set,
 * the sensor FPN is also on.
 *
 * @param old_fpn   The FPN choice used when we last started the sensor.
 * @param new_fpn   The new FPN choice we are changing to.
 *
 * @return true if there's a sensor FPN change , false if not
 */
bool LuximaHS::fpn_needs_restart(fpnmode_t old_fpn, fpnmode_t new_fpn)
{
    bool old_used_sensor_fpn;
    bool new_uses_sensor_fpn;
    fpnmode_t mask;

    mask = GIGE_FPN_SENSOR_ENABLE_MASK;
    if ((old_fpn & mask) != 0)
        old_used_sensor_fpn = true;
    else
        old_used_sensor_fpn = false;

    if ((new_fpn & mask) != 0)
        new_uses_sensor_fpn = true;
    else
        new_uses_sensor_fpn = false;

    if (old_used_sensor_fpn != new_uses_sensor_fpn)
        return true;
    else
        return false;
}

/*!
 * @brief Entry point from the calibration worker thread, this
 *        method performs either factor or user FPN calibration.
 *
 * @param is_factory    [IN] True if we are to do factory FPN, false for user FPN calibration.
 * @param manual_parmas [UNUSED] There are no manual parameters for the Luxima sensors.
 * @param p_rtn_restart [OUT] Retruns true if a restart is required, false otherwise.
 *
 * @returns True if the calibration was successful, false otherwise.
 */
bool LuximaHS::run_calibration(bool is_factory, bool /* manual_params */, bool *p_rtn_restart)
{
    bool rc = false;
    LuximaHSCalibration calib;

    LOG("%s(%d)", __PRETTY_FUNCTION__, is_factory);

    // Run the calibration algorithm.
    rc = calib.runFpnCalibration(is_factory);

    // For LuximaHS, we always need to restart after calibration.
    if (p_rtn_restart != nullptr)
        *p_rtn_restart = true;

    return rc;
}

/*!
 * @brief Entry point from the calibration worker thread, this
 *        method performs the PRNU calibration operations for the Luxima
 *        sensors.
 *
 * @param ct_ptr    [IN] Pointer to the FPN correction table, required by PRNU.
 * @returns True if successfull, false otherwise.
 */
bool LuximaHS::run_prnu(const uint8_t *ct_ptr)
{
    bool rc = false;
    LuximaHSCalibration calib;

    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (ct_ptr != nullptr)
    {
        // Run the calibration algorithm.
        rc = calib.runPrnuCalibration(ct_ptr);
    }

    return rc;
}

