/*!
 * @file main.cxx
 * @brief This file is the main code module for the CameraServer
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <errno.h>
#include <execinfo.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ucontext.h>
#include <unistd.h>
#include <getopt.h>

#include <fastec/fas_log.h>
#include <fastec/fi_exception.h>

#include "camera_server.h"
#include "cs_update.h"
#include "fcp.h"
#include "image_io.h"
#include "part_mgr.h"

#include "override.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Our name in console log.
const char  *OUR_LOG_NAME = "CameraServer";

//! The file where we store our PID.
const char  *PID_FILE = "/var/run/CameraServer.pid";

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

// Our list of signals that we're interested in catching.
// Ignoring the broken pipe signal. We handle that situation in the actual networking code.
static int siglist[] = { SIGINT, SIGQUIT, SIGUSR1, SIGUSR2, SIGTERM, SIGALRM, SIGPIPE };

// Option variables
bool    debug_to_log = false;       //!< TRUE to send PDEBUG output to fi_log.
int     cs_verbose = -1;            //!< Our verbosity level.
bool    sensor_test = false;        //!< TRUE if we want to use the sensor test interface.
int     sensor_test_type = 1;
bool    sigpipe_caught = false;     //!< TRUE if we catch a SIGPIPE.
uint32_t user_ssd_bandwidth = 0;    //!< User supplied SSD bandwidth override.

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

static void cs_sig_handler(int signum);

/*!
 * @brief The following is our signal handler to handle unsolicited signals from the
 * including it's time to close
 *
 * TODO - This should be enhanced once the daemon who controls the
 * network is sending a signal when the network comes up and down.
 *
 * @param signum - what signal is being delivered
 */
static void cs_sig_handler(int signum)
{
    fi_log(LOG_INFO, "%s: signum=%d (%s)", __FUNCTION__, signum, strsignal(signum));

    // Handle the signal which tells us to close
    if ((signum == SIGTERM) || (signum == SIGINT) || (signum == SIGQUIT))
    {
        // Ask the CS to exit.
        fi_log(LOG_NOTICE, "%s: Exiting normally due to signal", __FUNCTION__);
        CameraServer::exit();
    }
    else if (signum == SIGALRM)
    {
        // A timer has expired.
        fi_log(LOG_NOTICE, "%s: Timer expired.", __FUNCTION__);

        // Re-register our handler for this signal.
        signal(signum, cs_sig_handler);
    }
    else if (signum == SIGUSR1)
    {
        CameraServer::saw_signal(1);
    }
    else if (signum == SIGUSR2)
    {
        CameraServer::saw_signal(2);
    }
    else if (signum == SIGPIPE)
    {
        // Remember we got the signal.
        sigpipe_caught = true;
        fi_log(LOG_INFO, "%s: SIGPIPE caught!", __FUNCTION__);

        // Re-register our handler for this signal.
        signal(signum, cs_sig_handler);
    }
    else
    {
        // Don't know why we're here, so let's exit.
        fi_log(LOG_NOTICE, "%s: Exiting abnormally due to signal.", __FUNCTION__);
        CameraServer::exit();
    }
}

void crit_err_hdlr(int sig_num, siginfo_t *info, void *ctx_ptr)
{
#if 0
    void              *array[50];
    void              *caller_address;
    int                size;
    struct ucontext    *uc;

    uc = (struct ucontext *)ctx_ptr;

    /* Get the address at the time the signal was raised from the EIP (x86) */
    caller_address = (void *) uc->uc_mcontext.gregs[REG_RIP];

    fi_log(LOG_ERR, "signal %d (%s), address is %p from %p",
           sig_num, strsignal(sig_num), (void *)info->si_addr,
           (void *)caller_address);

    size = backtrace(array, 50);

    /* overwrite sigaction with caller's address */
    array[1] = caller_address;
    backtrace_symbols_fd(array, size, 2);

#else
    (void)sig_num;
    (void)info;
    (void)ctx_ptr;
#endif

    exit(EXIT_FAILURE);
}

/*!
 * @brief Turn us into a daemon. Errors are considered fatal because they should
 * never happen unless there is something seriously wrong with the OS.
 *
 * @return none
 */
static void daemon_make(void)
{
    int     i;          // Generic loop variable.
    int     num;            // The number of entries in the siglist array.
    pid_t   pid;            // Our PID.
    FILE    *pid_file;      // Pointer to our PID file (where we save our PID).
    struct sigaction sigact;

    // Setup our file mode creation mask so that everyone can
    // access our files, not just the root user.
    umask(0);

    // Ignoring the broken pipe signal. We handle this case in the networking code itself.
    sigact.sa_handler = SIG_IGN;
    sigact.sa_flags = 0;
    if (sigaction(SIGPIPE, &sigact, nullptr) != 0)
    {
        fi_log(LOG_ERR, "%s: sigaction failed with error: %d (%s)", __FUNCTION__, errno, strerror(errno));
    }

    // Setup the signals so that we can catch them and do the appopriate action
    num = sizeof(siglist) / sizeof(siglist[0]);

    // All the signals will go through the same handler.
    sigact.sa_handler = cs_sig_handler;
    sigact.sa_flags = 0;

    // Install the handler.
    for (i = 0; i < num; i++)
    {
        if (sigaction(siglist[i], &sigact, nullptr) != 0)
        {
            fi_log(LOG_ERR, "%s: sigaction failed with error: %d (%s)", __FUNCTION__, errno, strerror(errno));
        }
    }

    // Setup to handle SIGSEGV so we can catch a backtrace.
    sigact.sa_sigaction = crit_err_hdlr;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_RESTART | SA_SIGINFO;
    if (sigaction(SIGSEGV, &sigact, nullptr) != 0)
    {
        // Guess we're not getting our backtrace.
        fi_log(LOG_ERR, "%s: error setting signal handler for %d (%s)", __FUNCTION__, SIGSEGV, strsignal(SIGSEGV));
    }

    // Echo our PID to /var/run note this requires the daemon to be run as root with sudo.
    pid = getpid();
    pid_file = fopen(PID_FILE, "w");
    if (pid_file)
    {
        fprintf(pid_file, "%d\n", (int)pid);
        fclose(pid_file);
    }
    else
    {
        fi_log(LOG_ERR, "%s: Cannot open pid file '%s', errno: %d (%s)",
               __FUNCTION__, PID_FILE, errno, strerror(errno));
    }
}

/*!
 * @brief Show a help message.
 */
static void display_help(void)
{
    printf("\n\nUsage: CameraServer [OPTIONS]\n"
           "Runs the Fastec Imaging Camera Server daemon.\n"
           "\n"
           "           --help              Display this help and exit.\n"
           "-b bw      --bandwidth=bw      Set SSD bandwidth.\n"
           "-s N       --sensor-test=N     Enable the sensor test. default: increment\n"
           "   0=user_pattern\n"
           "   1=incrementing\n"
           "   2=decrementing\n"
           "   3=random.\n"
           "   4=shifted.\n"
           "   5=sensor (if available).\n"
           "-d         --debug-fi_log      Debug to fi_log instead of stdout.\n"
           "-v mask    --verbose=mask      Set verbosity of debug output:\n");
    printf("   0x0001    for specific testing\n"
           "   0x0002    low noise level\n"
           "   0x0004    high noise level\n"
           "   0x0008    trace functions\n"
           "   0x0010    GigE writes\n"
           "   0x0020    GigE reads\n"
           "   0x0040    download debug\n"
           "   0x0080    sensor debug\n"
           "   0x0100    FPGA writes, except LUTs\n"
           "   0x0200    FPGA writes to LUTs\n"
           "   0x1000    FPGA reads, except ICM CRCs and time counters\n"
           "   0x2000    FPGA reads of ICM CRCs\n"
           "   0x4000    FPGA reads of latched time registers SBS_COUNTER and SBUS_COUNTER\n");
    exit(0);
}

/*!
 * @brief Parse the command line options passed to the application.
 *
 * @param argc   Normal application argc.
 * @param argv[] Normal application argv.
 */
static void process_options(int argc, char *argv[])
{
    // Initialize our option variables.
    cs_verbose = 0;
    sensor_test = false;
    user_ssd_bandwidth = 0;

    // Parse the command line parameters.
    while (1)
    {
        int option;
        int option_index = 0;
        static const char *short_options = "b:v:s:d";
        static const struct option long_options[] =
        {
            { "bandwidth", required_argument, 0, 'b' },
            { "verbose", required_argument, 0, 'v' },
            { "debug-fi_log", no_argument, 0, 'd' },
            { "sensor-test", required_argument, 0, 's' },
            { "help", no_argument, 0, 0 },
            { 0, 0, 0, 0 },
        };

        // See if we have any options.
        option = getopt_long(argc, argv, short_options, long_options, &option_index);
        if (option == EOF)
            break;

        // Process the option specified.
        switch (option)
        {
        default:
        case '?':
        case 0:
            display_help();
            break;
        case 's':
            sensor_test = true;
            sensor_test_type = strtoul(optarg, nullptr, 0);
            break;
        case 'd':
            debug_to_log = true;
            break;
        case 'v':
            cs_verbose = strtol(optarg, nullptr, 0);
            break;
        case 'b':
            user_ssd_bandwidth = strtoul(optarg, nullptr, 0);
            break;
        }
    }
}

/*!
 * @brief We want to make sure that only one CS is allowed to run
 *    at a time.
 */
static void check_for_existing_cs(void)
{
    FILE    *pid_file;      // File pointer to an existing PID file.
    struct stat stat_buf;       // Status information structure.

    // Try to open our PID file. If it exists, then there MAY be another CS
    // running.
    pid_file = fopen(PID_FILE, "r");
    if (pid_file)
    {
        char    temp_buf[ MAX_BUFFER_LEN ]; // Temporary buffer.
        char    *result_ptr;

        // Figure out what the pid is of the existing camera server.
        result_ptr = fgets(temp_buf, MAX_BUFFER_LEN, pid_file);
        fclose(pid_file);

        // Make sure we got something.
        if (result_ptr)
        {
            int     pid;            // PID of competing process.

            // Build up the path to the /proc folder to see if this ID exists.
            pid = strtol(temp_buf, nullptr, 0);
            snprintf(temp_buf, MAX_BUFFER_LEN, "/proc/%d", pid);

            // Now see if Linux knows about this guy.
            if (stat(temp_buf, &stat_buf) == 0)
            {
                fi_log(LOG_WARNING, "CameraServer (pid=%d) is already running...exiting this instance.", pid);
                exit(EXIT_SUCCESS);
            }
        }
    }
}

/*!
 * @brief This in the main entry point into the camera server.
 */
int main(int argc, char *argv[])
{
#if 1
    loggerUseStdout(1);
    loggerUseSyslog(1);
#else
    loggerFile("/tmp/cs.log");
    loggerUseStdout(0);
    loggerUseSyslog(0);
#endif

    // Look for a command line option.
    process_options(argc, argv);
    if (cs_verbose != 0)
    {
        fi_log(LOG_INFO, "Started with verbose level: 0x%x", cs_verbose);
        fi_log(LOG_INFO, "Sensor test is%s enabled", sensor_test ? "" : " NOT");
        if (sensor_test)
            fi_log(LOG_INFO, "Sensor test type is: %d", sensor_test_type);
    }

    // Set the logger's verbosity level.
    verbosity = cs_verbose;

    // Check for another version of the CS running. There can be only one!
    check_for_existing_cs();

    // Make us daemon-like.
    daemon_make();

    // Initialize for possible overrides.
    override_init();

    // Check for post-update instructions.
    update_postops();

    try
    {
        // Run the camera server. The run() method will not return until
        // we are done. Allocate the Camera Server object on the stack
        // and go.
        CameraServer cs;
        if (cs.init())
            cs.run();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fprintf(stderr, "Terminating with exception:\n%s\n", e.what().c_str());
    }
    catch (...)
    {
        fprintf(stderr, "Terminating with uncaught exception\n");
    }

    // Get rid of our pid file.
    unlink(PID_FILE);

    // Let everyone know we are done.
    fi_log(LOG_DEBUG, "(%s) shutting down", __FUNCTION__);

    // All done.
    exit(EXIT_SUCCESS);
}
