/*!
 * @file mft.cpp
 * @brief Manage shared usage of the Micro Four Thirds lens.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>

#include <fmt/core.h>
#include <json/json.h>

#include <opencv2/opencv.hpp>

#include <atomic>
#include <chrono>
#include <cmath>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <memory>
#include <mutex>
#include <string>

#include "cs_trace.h"
#include "camera_server.h"
#include "fcp.h"
#include "fpga.h"
#include "fpga_io.h"
#include "fsp.h"
#include "mft.h"
#include "part_mgr.h"

namespace FI_MFT
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Number of bytes in an MFT command structure.
const int CMD_LEN = 4;

//! Index into the command array for parameter #1.
const int PARAM1_INDEX = 2;

//! Index into the command array for parameter #1.
const int PARAM2_INDEX = 3;

//! The number of milliseconds between polling the lens.
const int MFT_POLL_PERIOD = 1000;

//! The name of the settings file.
//! arg0: The home path [see get_home_path()].
const char LENS_SETTINGS_FILE[] = "{0}/.config/Fastec/settings/lens_info.json";

//! @name MFT definitions.
//!
//! @{

// Bit definitions for the lens status data.
#define LS_MF_ENABLE                        (1 << 28)

// Bit definitions for the lens attribute data.
#define LAD_TELECONVERTER_POSSIBLE          (1 << 0)
#define LAD_INTERMEDIATE_RING_POSSIBLE      (1 << 1)
#define LAD_FOCUS_HYSTERESIS_REMOVAL        (1 << 4)
#define LAD_APERTURE_HYSTERESIS_REMOVAL     (1 << 5)
#define LAD_ZOOM_LENS                       (1 << 6)
#define LAD_APERTURE_RING                   (1 << 7)
#define LAD_DISTORTION_CORR_MASK            (0x3 << 8)
#define LAD_DISTORTION_CORR_ON              (0x1 << 8)
#define LAD_DISTORTION_CORR_OFF             (0x2 << 8)
#define LAD_DISTORTION_CORR_ARB             (0x3 << 8)
#define LAD_FOCUS_ENCODER                   (1 << 14)
#define LAD_ZOOM_ENCODER                    (1 << 15)
#define LAD_LENS_RESET_OFF                  (1 << 16)
#define LAD_MOVIE_APERTURE                  (1 << 17)
#define LAD_FOCUS_DEPENDENCE                (1 << 18)
#define LAD_EFFECTIVE_FNO                   (1 << 19)
#define LAD_EXCEPTION_DISTORTION            (1 << 20)
#define LAD_3D_LENS                         (1 << 21)
#define LAD_ELECTRIC_ZOOM                   (1 << 22)
#define LAD_TYPE6                           (1 << 23)
#define LAD_MANUAL_APERTURE                 (1 << 24)
#define LAD_CHROMATIC_ABERATION_DATA        (1 << 25)
#define LAD_ONCHIP_PHASE_DETECTION          (1 << 26)
#define LAD_TARGET_POSITION_COMPLIANT       (1 << 27)
#define LAD_MF_RING                         (1 << 28)
#define LAD_LEN_BASED_IMAGE_STABIL          (1 << 29)
#define LAD_LEN_BASED_IMAGE_STABIL_SW       (1 << 30)

// Bit definitions for the control frequency compliance flag.
#define CFC_30HZ                (1 << 1)
#define CFC_60HZ                (1 << 2)
#define CFC_120HZ               (1 << 3)
#define CFC_240HZ               (1 << 4)
#define CFC_25HZ                (1 << 9)
#define CFC_50HZ                (1 << 10)
#define CFC_100HZ               (1 << 11)
#define CFC_200HZ               (1 << 12)
#define CFC_24HZ                (1 << 17)
#define CFC_48HZ                (1 << 18)
#define CFC_96HZ                (1 << 19)
#define CFC_192HZ               (1 << 20)
#define CFC_WOBBLE_HALF         (1 << 24)
#define CFC_WOBBLE_QUARTER      (1 << 25)
#define CFC_WOBBLE_EIGHTH       (1 << 26)

// SetActLV: Live view lens operation set.
// "Sets operations of the lens synchronized via the
// LSEXP terminal for Live view.
//
// The following define parameter 1: Focus operation mode.
#define FO_WOBBLING             0x01
#define FO_OFFSET_DRIVE         0x02
#define FO_ABSOLUTE_DRIVE       0x03
#define FO_SCAN_DRIVE           0x04
#define FO_DISTANCE_DRIVE       0x05
#define FO_MANUAL_FOCUS         0x06
#define FO_DEFOCUS_DRIVE        0x07
#define FO_NOP                  0xFE
#define FO_STOP                 0xFF

// Or with the above operations bits for operations
// with the zoom encoder.
#define FO_WITH_ZMENC           0x10

// The following define paramter 2: Aperture operation mode
// of the SetActLV command.
#define AO_DRIVE_WITH_HYS_REMOVAL   0x01
#define AO_DRIVE_WOUT_HYS_REMOVAL   0x02
#define AO_NOP                      0xFE
#define AO_STOP                     0xFF

//! @}

//! @name MFT commands.
//!
//! Commands that end in 0, i.e 0xB0, 0xC0, etc. Must be done in low-speed.
//! Others CAN be done in highspeed.
//!
//! @{
const uint8_t SET_INIT_ACT_LENS[]       = { 0x60, 0xF0, 0x00, 0x00 };   // "SetInitActLens"
const uint8_t LENS_STOP2[]              = { 0xB1, 0xFF, 0x00, 0x00 };   // "LensStop2"

const uint8_t ASK_HIGHSPEED[]           = { 0xC0, 0xF4, 0x00, 0x00 };   // "AskHighSpeed"
const uint8_t CHANGE_SPEED[]            = { 0xA0, 0xF5, 0x01, 0x00 };   // "ChangeSpeed"

const uint8_t ASK_APAPX[]               = { 0xB1, 0xDD, 0x00, 0x00 };   // "AskAPapx"
const uint8_t ASK_APE_DATA[]            = { 0xC2, 0xF5, 0x01, 0x00 };   // "AskApeData"
const uint8_t ASK_DATA_M[]              = { 0xC3, 0xF0, 0x00, 0x00 };   // "AskDataM"
const uint8_t ASK_DATA_S[]              = { 0xC0, 0xF9, 0x00, 0x00 };   // "AskDataS"
const uint8_t ASK_DATA_SYNC[]           = { 0xC1, 0xFC, 0x00, 0x00 };   // "AskDataSync"
const uint8_t ASK_FRM_VER[]             = { 0xC1, 0xF8, 0x00, 0x00 };   // "AskFrmVer"
const uint8_t ASK_LDPLS[]               = { 0xB1, 0xAD, 0x00, 0x00 };   // "AskLDpls"
const uint8_t ASK_LDDATA_M[]            = { 0xC2, 0xF5, 0x07, 0x00 };   // "AskLDDataM"
const uint8_t ASK_REF_DATA_M[]          = { 0xC2, 0xF5, 0x06, 0x00 };   // "AskRefDataM"
const uint8_t ASK_STATUS_LV[]           = { 0xC1, 0x80, 0x00, 0x02 };   // "AskStatusLV"
const uint8_t ASK_VER[]                 = { 0xC0, 0xF6, 0x00, 0x00 };   // "AskVer"

const uint8_t CONT_AP_APX_ABS[]         = { 0xB1, 0xD1, 0x00, 0x00 };   // "ContAPapsABS"
const uint8_t CONT_LD_PULSE_ABS[]       = { 0xB1, 0xA2, 0x00, 0x00 };   // "ContLDplsABS"
const uint8_t CONT_LD_DIST_ABS[]        = { 0xB1, 0xA3, 0x00, 0x00 };   // "ContLDdstABS"

const uint8_t CONT_EZ_FLENGTH[]         = { 0xA0, 0xD1, 0x00, 0x00 };   // "ContEZflength": Target focal length goes into P1 & P2.
const uint8_t CHANGE_EZ_SETTING[]       = { 0xA0, 0xD5, 0x00, 0x00 };   // "ChangeEZsetting": Electric zoom operation in P1.

//! @}

//! @name MFT Structures.
//! @{

#pragma pack(1)
typedef struct L_ID
{
    uint16_t    member_id;
    struct
    {
        uint16_t    model_id;
        uint16_t    hardware_ver;
    } lens_id;
} L_ID;

typedef struct ASK_DATA_S_RESP
{
    L_ID member_lens_id;
    char lens_serial_no[12];
    uint16_t lens_firmware_ver;
} ASK_DATA_S_RESP;

typedef struct ASK_DATA_SYNC_RESP
{
    uint32_t compliance_flags;

    uint16_t max_wobbling_amp_15hz;
    uint16_t max_wobbling_amp_30hz;
    uint16_t max_wobbling_amp_60hz;
    uint16_t max_wobbling_amp_120hz;

    uint16_t min_wobbling_amp_15hz;
    uint16_t min_wobbling_amp_30hz;
    uint16_t min_wobbling_amp_60hz;
    uint16_t min_wobbling_amp_120hz;

    uint16_t max_wobbling_disp_15hz;
    uint16_t max_wobbling_disp_30hz;
    uint16_t max_wobbling_disp_60hz;
    uint16_t max_wobbling_disp_120hz;

    uint16_t max_disp_offset_drive_30hz;
    uint16_t max_disp_offset_drive_60hz;
    uint16_t max_disp_offset_drive_120hz;

    uint8_t max_aperture_drive_speed[4];
    uint8_t min_aperture_drive_speed[4];
    uint8_t silent_aperture_drive_speed[4];

    uint16_t W_switching_pt_AV[3];
    uint16_t N_switching_pt_AV[3];
    uint16_t T_switching_pt_AV[3];

    uint16_t max_lsexp_delay;
    uint16_t min_lsexp_delay;
    uint8_t max_ldp_buffers;
    uint8_t min_lsexp_delay_120hz;
    uint16_t min_offset_drive_time;
} ASK_DATA_SYNC_RESP;

typedef struct ASK_HIGHSPEED_RESP
{
    uint16_t    data_comm_amount;
    uint16_t    data_interval;
} ASK_HIGHSPEED_RESP;

typedef struct ASK_REF_DATA_M_RESP
{
    uint32_t lens_attribute_data;
    L_ID    member_lens_id;
    char    lens_name[32];
    uint8_t zoom_divisions;
    uint8_t zoom_components;
    uint8_t focus_divisions;
    uint8_t distortion_zoom;
    uint8_t distortion_focus;
    uint8_t distortion_corr;
    uint16_t focal_length_wide;
    uint16_t focal_length_tele;
    char    lens_serial_no[12];
    uint16_t firmware_ver;
} ASK_REF_DATA_M_RESP;

typedef struct ASK_STATUS_LV_RESP
{
    uint32_t lens_status;
    uint8_t zoom_encoder;
    uint8_t focus_encoder;
    uint16_t object_distance;
    uint16_t current_aperture;
    uint8_t ze_for_distortion;
    uint8_t fe_for_distortion;

    // Note that everything below here is dynamic, and
    // these may or may not be present.
    uint16_t aperture_ring;         // Seems like we get this in any event.
    uint16_t live_af_distance[8];   // I explicitly ask for this.

#if 0
    uint16_t ld_position[LDP_POSITIONS_TO_ACQUIRE]; // Changes based on how many are in the request.
    uint16_t zoom_status;
    uint16_t zoom_speed;
#endif
} ASK_STATUS_LV_RESP;

typedef struct ASK_VER_RESP
{
    uint16_t    mft_ver;
    uint16_t    transmission_speed;
} ASK_VER_RESP;

typedef struct LENS_STOP2_RESP
{
    uint8_t zmenc_dist;
    uint8_t zmenc;
    uint16_t    lens_pulse;
    uint8_t lens_status;
} LENS_STOP2_RESP;

typedef struct SET_INIT_ACT_LENS_RESP
{
    uint8_t distortion_corr_ZE;
    uint8_t zoom_encoder;
    int16_t lens_pulse;
    uint8_t lens_initial_status;
} SET_INIT_ACT_LENS_RESP;

typedef struct WCONT_AP_RESP
{
    uint16_t control_ap;
    int16_t  lens_pulse;
} WCONT_AP_RESP;
#pragma pack(0)

//! @}

/*!
 * @brief The mftPoll class is used to poll the lens for any changes.
 *
 * This object checks for:
 * - Lens connect: On a connection event, this will start the initialization process
 *                 with the control object (mftThread).
 * - Lens disconnect: On a disconnect event, this will start the clean-up process.
 *
 * - Aperture change: Send a notification message to connected clients.
 * - Focus change: Send a notification message to connected clients.
 * - Zoom change: Send a notification message to connected clients.
 */
class mftPoll : public FAS_THREAD::Thread
{
public:
    mftPoll();
    virtual ~mftPoll();

    virtual void run();

    bool is_running()   { return thread_running; }
    void wakeup(uint32_t irq_status);

private:
    //! Enumerate the types of events we're polling for.
    enum POLLING_EVENT
    {
        NO_EVENT,               //!< Nothing happened.
        CONNECT_EVENT,          //!< A lens connected.
        DISCONNECT_EVENT,       //!< A lens disconnected.
        SETTINGS_CHANGE         //!< A setting (aperture, focus, zoom) changed.
    };

private:
    int getChangeEvent();

    void handleConnection();
    void handleDisconnect();
    void handleSettingsChange();

private:
    volatile bool   thread_running;     //!< Flag to indicate if we are running or not.
    std::mutex      irq_mutex;          //!< Used to control changes to irq_status.
    std::condition_variable irq_status; //!< Condition variable we can block on.

    bool lens_connected;                //!< Flag to indicate the connection state.
    LENS_STATUS last_status;            //!< The last information we had from the lens.
};

//! Message types for the lens control processor.
enum MESSAGE_TYPES
{
    GET_LENS_INFO = 0x30000,        //!< Retrieve information about the attached lens.
    GET_LENS_STATUS,                //!< Retrieve the lens status information.
    GET_LENS_STATUS_BLOCKING,       //!< Retrieve the lens status information, blocking for completion.
    CHANGE_APERTURE,                //!< Change the aperture setting.
    CHANGE_FOCUS,                   //!< Change the focus setting.
    CHANGE_POWER,                   //!< Change the power setting.
    CHANGE_ZOOM,                    //!< Change the zoom setting.
    INITIALIZE,                     //!< Perform initialization functions.
    DISCONNECT,                     //!< Perform disconnection functions.
    AUTO_FOCUS                      //!< Perform auto-focus function.
};

//! Message class for GET_LENS_INFO messages.
class getLensInfoMessage : public FAS_THREAD::Message
{
public:
    getLensInfoMessage();
    virtual ~getLensInfoMessage();

    FI_MFT::LENS_INFO request;
};

//! Message class for GET_LENS_STATUS messages.
class getLensStatusMessage : public FAS_THREAD::Message
{
public:
    getLensStatusMessage();
    virtual ~getLensStatusMessage();

    FI_MFT::LENS_STATUS request;
};

//! Message class for GET_LENS_STATUS_BLOCKING messages.
class getLensStatusBlockingMessage : public FAS_THREAD::SyncMessage
{
public:
    getLensStatusBlockingMessage();
    virtual ~getLensStatusBlockingMessage();

    FI_MFT::LENS_STATUS request;
};

//! Message class for CHANGE_APERTURE messages.
//!
//! @note: This is an asynchronous message. There's no reply
//!        so we "fire and forget".
class changeApertureMessage : public FAS_THREAD::Message
{
public:
    changeApertureMessage();
    virtual ~changeApertureMessage();

    // New aperture settings go here.
    float f_number;
};

//! Message class for CHANGE_FOCUS messages.
//!
//! @note: This is an asynchronous message. There's no reply
//!        so we "fire and forget".
class changeFocusMessage : public FAS_THREAD::Message
{
public:
    changeFocusMessage();
    virtual ~changeFocusMessage();

    // New focus settings go here.
    int16_t focus_pulse;
};

//! Message class for CHANGE_POWER messages.
class changePowerMessage : public FAS_THREAD::SyncMessage
{
public:
    changePowerMessage();
    virtual ~changePowerMessage();

    // The enable flag: TRUE power up, FALSE power down.
    bool enable;
};

//! Message class for CHANGE_ZOOM messages.
//!
//! @note: This is an asynchronous message. There's no reply
//!        so we "fire and forget".
class changeZoomMessage : public FAS_THREAD::Message
{
public:
    changeZoomMessage();
    virtual ~changeZoomMessage();

    // The zoom encoder index to change the zoom to.
    int index;
};

//! Message class for INITIALIZE message.
//!
//! @note: This is an asynchronous message. There's no reply
//!        so we "fire and forget".
class initializeControl : public FAS_THREAD::Message
{
public:
    initializeControl();
    virtual ~initializeControl();
};

//! Message class for DISCONNECT message.
//! Note that this is an ASYNCHRONOUS message.
class disconnectControl : public FAS_THREAD::Message
{
public:
    disconnectControl();
    virtual ~disconnectControl();
};

//! Message class for AUTO-FOCUS message.
//! Note that this is an ASYNCHRONOUS message.
class autoFocusMessage : public FAS_THREAD::Message
{
public:
    autoFocusMessage();
    virtual ~autoFocusMessage();
};

/*!
 * @brief The micro four thirds control object. This object is
 *    a singleton and is "local" to this file.
 */
class mftThread : public FAS_THREAD::MsgThread
{
public:
    virtual ~mftThread();
    virtual void messageHandle(FAS_THREAD::Message *message) override;

    // Singletons are not cloneable.
    mftThread(mftThread& other) = delete;

    // Singletons are not assignable.
    void operator=(const mftThread&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static mftThread *factory();

    static bool doConnect();
    static bool doDisconnect();
    static bool lensPresent();

    static bool keepOff();
    static bool lensIsPowered();
    static bool lensIsTalking();

    mftPoll *getPoller()                    { return lens_poller.get(); }
    getLensInfoMessage *getInfoResp()       { return info_resp.release(); }
    getLensStatusMessage *getStatusResp()   { return status_resp.release(); }

    constexpr float getFMinimum()           { return f_min; }
    constexpr float getFMaximum()           { return f_max; }
    constexpr uint16_t getFocusNear()       { return focus_near; }
    constexpr uint16_t getFocusInf()        { return focus_inf; }

    bool isBusy() const;

private:
    mftThread();

    float apexToFnum(uint16_t apex_value);
    uint16_t fnumToApex(float f_num);

    bool getStatusData(LENS_STATUS *status_ptr);
    void lensStop();

    void handleAFRequest(autoFocusMessage *msg);
    int calcClarity();

    void handleInfoRequest(getLensInfoMessage *msg);
    void handleStatusRequest(getLensStatusMessage *msg);
    void handleStatusBlockingRequest(getLensStatusBlockingMessage *msg);
    void handleSetApertureRequest(changeApertureMessage *msg);
    void handleSetFocusRequest(changeFocusMessage *msg);
    void handleSetPowerRequest(changePowerMessage *msg);
    void handleSetZoomRequest(changeZoomMessage *msg);

    void handleInitialize(initializeControl *msg);
    void handleDisconnect(disconnectControl *msg);

    bool sendCommand(const uint8_t *cmd_val);
    bool sendCommandType6(const uint8_t *cmd_val, const uint8_t *data_val, int data_count);

    bool askApApx(uint16_t& apex_value);
    bool askApeData();
    bool askLDpls();
    bool askLDDataM();
    bool askRefDataM();
    bool askStatusLV();
    bool askVer(ASK_VER_RESP& resp);

    bool setAperture(uint16_t apex_value);
    bool setFocus(int16_t pulse_value);
    bool setInitalOps();

    bool accessoryCheck();
    bool lensCheck();
    bool lensPowerup();
    bool powerup();
    bool resetLens();

    bool power(bool enable);
    void reset(bool do_reset);

private:
    static std::unique_ptr<mftThread> mft_proc;
    static std::mutex mft_proc_mutex;

    uint16_t mft_version;           //!< MFT version (without the 'A').
    bool lens_is_OK;                //!< True if we're succesfully talking to the lens.
    bool lens_is_powered;           //!< True if we've applied power to the lens.
    bool force_off;                 //!< Even if the lens is connected, we don't want to apply power.

    uint16_t current_aperture;      //!< Currently selected APEX value.
    float f_min;                    //!< The minimum f-stop.
    float f_max;                    //!< The maximum f-stop.
    uint8_t focus_encoder;          //!< Currently selected focus encoder position.
    uint16_t focus_near;            //!< Closest distance absolute pulse.
    uint16_t focus_pos;             //!< The current lens pulse value.
    uint16_t focus_inf;             //!< Infinite distance absolute pulse.
    bool manual_focus;              //!< True if the camera says we're in manual focus.
    int object_distance;            //!< Currently calculated object distance, in millimeters. -1 means infinity.
    uint8_t zoom_encoder;           //!< Currently selected zoom encoder position.

    std::atomic<bool> busy;         //!< True if we're busy processing a command.

    ASK_REF_DATA_M_RESP lens_info;
    std::unique_ptr<std::vector<uint16_t>> ape_data;
    std::unique_ptr<std::vector<uint16_t>> ld_data;

    std::unique_ptr<getLensInfoMessage> info_resp;
    std::unique_ptr<getLensStatusMessage> status_resp;

    std::unique_ptr<mftPoll> lens_poller;       //!< Our MFT polling thread looking for changes in the lens settings.

    std::unique_ptr<uint8_t[]> raw_image;       //!< Buffer to hold the last raw live image for auto-focus.
    uint32_t raw_len;                           //!< Length of the raw_image buffer.
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<mftThread> mftThread::mft_proc;     //!< Our MFT control object.
std::mutex mftThread::mft_proc_mutex;

} /* End of namespace FI_MFT */

using namespace CS_TRACE;
using namespace FI_MFT;
using namespace FI_FPGA_IO;

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

void mft_dump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Sanity check.
    if (mftThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return;
    }

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(mftThread::factory());
}

void mft_init()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_MFT_PROC;

    // Make sure we have our singleton.
    mftThread::factory();
}

void mft_close()
{
    mftThread *thread_ptr;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Release our singleton.
    thread_ptr = mftThread::factory();
    if (thread_ptr != nullptr)
    {

        // Tell the thread that we are stopping.
        FAS_THREAD::SyncMessage msg;
        msg.req = FAS_THREAD::HALT;

        // Pass it on.
        msg.doSync(thread_ptr);

        // Try to join the thread.
        thread_ptr->join();
    }
}

void mft_irq(uint32_t irq_status)
{
    mftThread *thread_ptr;

//    TP(TRACE_SEL_MFT_PROC, "%s: irq_status=0x%08x", __PRETTY_FUNCTION__, irq_status);

    // Sanity check.
    thread_ptr = mftThread::factory();
    if (thread_ptr == nullptr)
        return;

    // Do we have our poller running?
    if (thread_ptr->getPoller() == nullptr)
        return;

    // Tell the polling object to wakeup.
    thread_ptr->getPoller()->wakeup(irq_status);
}

/*!
 * @brief Request the lens information.
 */
bool mft_info_request(FI_MFT::LENS_INFO& request)
{
    bool rc;
    getLensInfoMessage *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check
    if (mftThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    // This requires a response, so don't even try if the worker
    // thread is already busy.
    if (mftThread::factory()->isBusy())
    {
//        LOG("%s: MFT is busy.", __PRETTY_FUNCTION__);
        return false;
    }

    // Do we have a response?
    msg = mftThread::factory()->getInfoResp();
    if (msg != nullptr)
    {
        // Copy over the output values.
        rc = true;
        request.connected = msg->request.connected;
        request.powered = msg->request.powered;
        request.talking = msg->request.talking;
        request.focal_len_wide = msg->request.focal_len_wide;
        request.focal_len_tele = msg->request.focal_len_tele;
        request.focus_divisions = msg->request.focus_divisions;
        request.zoom_divisions = msg->request.zoom_divisions;
        request.ez_enable = msg->request.ez_enable;
        request.name = msg->request.name;

        // Release the message.
        delete msg;
    }
    else
    {
        rc = false;

        // Allocate our message.
        msg = new getLensInfoMessage();
        if (msg != nullptr)
        {
            // Add the message to the lens control thread.
            mftThread::factory()->messageAdd(msg);
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Request the lens information and format as JSON.
 *
 * @return Length of the JSON string, including the null character.
 *         0 on error.
 */
uint32_t mft_info_json(char *dest_ptr, uint32_t max_len)
{
    LENS_INFO lens_info;
    Json::Value lens_json;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // Get the information.
    if (!mft_info_request(lens_info))
    {
        // This can happen if the worker thread is busy.
        return 0;
    }

    // Convert it to JSON.
    lens_json["connected"] = lens_info.connected;
    if (!lens_info.connected)
        goto li_bail;

    lens_json["powered"] = lens_info.powered;
    if (!lens_info.powered)
        goto li_bail;

    // Since we're connected and powered, we can fill in the rest.
    lens_json["talking"] = lens_info.talking;
    lens_json["focal_len_wide"] = lens_info.focal_len_wide;
    lens_json["focal_len_tele"] = lens_info.focal_len_tele;
    lens_json["focus_divisions"] = lens_info.focus_divisions;
    lens_json["zoom_divisions"] = lens_info.zoom_divisions;
    lens_json["ez_enable"] = lens_info.ez_enable;
    lens_json["name"] = lens_info.name;

li_bail:
    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 5;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, lens_json);

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
        return 0;
    }

    // Copy the string over.
    TP(TRACE_SEL_MFT_PROC, "%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
    memcpy(dest_ptr, json_message.c_str(), json_message.length());

    // Return the length, including the null-terminator.
    return json_message.length() + 1;
}

/*!
 * @brief Request the lens status.
 */
bool mft_status_request(LENS_STATUS& request, bool wait)
{
    bool rc = false;

    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check
    if (mftThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    // See if we're blocking or not.
    if (!wait)
    {
        getLensStatusMessage *msg;

        // Is the worker busy?
        if (mftThread::factory()->isBusy())
        {
            // The worker thread is still working.
            return false;
        }

        // Do we have a response?
        msg = mftThread::factory()->getStatusResp();
        if (msg == nullptr)
        {
            // Assume we need to come back.
            rc = false;

            // Allocate our message.
            msg = new getLensStatusMessage();
            if (msg != nullptr)
            {
                // Add the message to the lens control thread.
                mftThread::factory()->messageAdd(msg);
            }
        }
        else
        {
            // Copy over the output values.
            rc = true;
            request.connected = msg->request.connected;
            request.powered = msg->request.powered;
            request.talking = msg->request.talking;
            request.apex = msg->request.apex;
            request.f_number = msg->request.f_number;
            request.f_minimum = msg->request.f_minimum;
            request.f_maximum = msg->request.f_maximum;
            request.object_distance = msg->request.object_distance;
            request.focus_pos = msg->request.focus_pos;
            request.focus_min = msg->request.focus_min;
            request.focus_near = msg->request.focus_near;
            request.focus_inf = msg->request.focus_inf;
            request.manual_focus = msg->request.manual_focus;
            request.zoom_pos = msg->request.zoom_pos;
            request.focal_len = msg->request.focal_len;

            // Release the memory.
            delete msg;
        }
    }
    else
    {
        getLensStatusBlockingMessage msg;

        // We are blocking.
        msg.doSync(mftThread::factory());

        // Success?
        if (msg.resp == FAS_THREAD::OK)
        {
            // Copy over the output values.
            rc = true;
            request.connected = msg.request.connected;
            request.powered = msg.request.powered;
            request.talking = msg.request.talking;
            request.apex = msg.request.apex;
            request.f_number = msg.request.f_number;
            request.f_minimum = msg.request.f_minimum;
            request.f_maximum = msg.request.f_maximum;
            request.object_distance = msg.request.object_distance;
            request.focus_pos = msg.request.focus_pos;
            request.focus_min = msg.request.focus_min;
            request.focus_near = msg.request.focus_near;
            request.focus_inf = msg.request.focus_inf;
            request.manual_focus = msg.request.manual_focus;
            request.zoom_pos = msg.request.zoom_pos;
            request.focal_len = msg.request.focal_len;
        }
    }

    return rc;
}

/*!
 * @brief Request the lens information and format as JSON.
 *
 * @return Length of the JSON string, including the null character.
 *         0 on error.
 */
uint32_t mft_status_json(char *dest_ptr, uint32_t max_len)
{
    LENS_STATUS lens_status;
    Json::Value lens_json;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // Get the information.
    if (!mft_status_request(lens_status))
    {
        // This can happen if the worker thread is busy.
        return 0;
    }

    // Convert it to JSON.
    lens_json["connected"] = lens_status.connected;
    if (!lens_status.connected)
        goto ls_bail;

    lens_json["powered"] = lens_status.powered;
    if (!lens_status.powered)
        goto ls_bail;

    // Since we're connected and powered, we can fill in the rest.
    lens_json["talking"] = lens_status.talking;
    lens_json["apex"] = lens_status.apex;
    lens_json["f_number"] = lens_status.f_number;
    lens_json["f_minimum"] = lens_status.f_minimum;
    lens_json["f_maximum"] = lens_status.f_maximum;
//    lens_json["object_distance"] = lens_status.object_distance;
    lens_json["focus_pos"] = lens_status.focus_pos;
    lens_json["focus_min"] = lens_status.focus_min;
    lens_json["focus_near"] = lens_status.focus_near;
    lens_json["focus_inf"] = lens_status.focus_inf;
    lens_json["manual_focus"] = lens_status.manual_focus;
    lens_json["zoom_pos"] = lens_status.zoom_pos;
    lens_json["focal_len"] = lens_status.focal_len;

ls_bail:
    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 2;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, lens_json);

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
        return 0;
    }

    // Copy the string over.
    TP(TRACE_SEL_MFT_PROC, "%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
    memcpy(dest_ptr, json_message.c_str(), json_message.length());

    // Return the length, including the null-terminator.
    return json_message.length() + 1;
}

bool mft_set_aperture(float f_number)
{
    changeApertureMessage *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;
    if (f_number < (mftThread::factory()->getFMinimum() - 0.000001))
    {
        fi_log(LOG_DEBUG, "%s: Bad aperture f_number requested [min:%f, max:%f, request:%f]", __PRETTY_FUNCTION__, mftThread::factory()->getFMinimum(), mftThread::factory()->getFMaximum(), f_number);
        f_number = mftThread::factory()->getFMinimum();
    }
    else if (f_number > (mftThread::factory()->getFMaximum() + 0x000001))
    {
        fi_log(LOG_DEBUG, "%s: Bad aperture f_number requested [min:%f, max:%f, request:%f]", __PRETTY_FUNCTION__, mftThread::factory()->getFMinimum(), mftThread::factory()->getFMaximum(), f_number);
        f_number = mftThread::factory()->getFMaximum();
    }

    // Allocate our message.
    msg = new changeApertureMessage();
    if (msg == nullptr)
        return false;

    // Pass on the request.
    msg->f_number = f_number;

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);

    // So far, so good.
    return true;
}

bool mft_set_focus(uint32_t focus_dist)
{
    changeFocusMessage *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;
    if ((focus_dist < mftThread::factory()->getFocusInf()) || (focus_dist > mftThread::factory()->getFocusNear()))
    {
        fi_log(LOG_DEBUG, "%s: Bad focus distance requested [min:%d, max:%d, request:%d]", __PRETTY_FUNCTION__, mftThread::factory()->getFocusInf(), mftThread::factory()->getFocusNear(), focus_dist);
        return false;
    }

    // Allocate our message.
    msg = new changeFocusMessage();
    if (msg == nullptr)
        return false;

    // Pass on the request.
    msg->focus_pulse = (int16_t)(focus_dist);

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);

    // So far, so good.
    return true;
}

bool mft_set_power(bool enable)
{
    bool rc;
    changePowerMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s: enable=%d", __PRETTY_FUNCTION__, enable);

    if (mftThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return false;
    }

    // Pass on the request.
    msg.enable = enable;

    // Send it on.
    msg.doSync(mftThread::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        rc = true;
        TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Set the zoom on the lens.
 *
 * This is a command that can take a really long time to complete. Also,
 * there is no response to this command. For these reasons, we use an
 * asynchronous message to tell our worker thread to do the work. That way
 * we can return quickly and be on our way.
 *
 * @param index     The zoom encoder index to use.
 * @return True for success, false otherwise.
 */
bool mft_set_zoom(int index)
{
    changeZoomMessage *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    if (mftThread::factory() == nullptr)
        return false;

    // Allocate our message.
    msg = new changeZoomMessage();
    if (msg == nullptr)
        return false;

    // Pass on the request.
    msg->index = index;

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);

    // So far, so good.
    return true;
}

void mft_autofocus()
{
    autoFocusMessage *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    if (mftThread::factory() == nullptr)
        return;

    // Allocate our message.
    msg = new autoFocusMessage();
    if (msg == nullptr)
        return;

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);
}

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

getLensInfoMessage::getLensInfoMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)GET_LENS_INFO;

    // Initialize our request structure.
    request.connected = false;
    request.powered = false;
    request.talking = false;
    request.focal_len_wide = 0;
    request.focal_len_tele = 0;
    request.focus_divisions = 0;
    request.zoom_divisions = 0;
    request.ez_enable = false;
    request.name.clear();
}

getLensInfoMessage::~getLensInfoMessage()
{
}

getLensStatusMessage::getLensStatusMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)GET_LENS_STATUS;
    memset(&request, 0, sizeof(request));
}

getLensStatusMessage::~getLensStatusMessage()
{
}

getLensStatusBlockingMessage::getLensStatusBlockingMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)GET_LENS_STATUS_BLOCKING;
    memset(&request, 0, sizeof(request));
}

getLensStatusBlockingMessage::~getLensStatusBlockingMessage()
{
}

changeApertureMessage::changeApertureMessage() :
    Message(),
    f_number(0.0)
{
    req = (FAS_THREAD::REQ_CODE)CHANGE_APERTURE;
}

changeApertureMessage::~changeApertureMessage()
{
}

changeFocusMessage::changeFocusMessage() :
    Message(),
    focus_pulse(0)
{
    req = (FAS_THREAD::REQ_CODE)CHANGE_FOCUS;
}

changeFocusMessage::~changeFocusMessage()
{
}

changePowerMessage::changePowerMessage() :
    SyncMessage(),
    enable(false)
{
    req = (FAS_THREAD::REQ_CODE)CHANGE_POWER;
}

changePowerMessage::~changePowerMessage()
{
}

changeZoomMessage::changeZoomMessage() :
    Message(),
    index(0)
{
    req = (FAS_THREAD::REQ_CODE)CHANGE_ZOOM;
}

changeZoomMessage::~changeZoomMessage()
{
}

initializeControl::initializeControl() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)INITIALIZE;
}

initializeControl::~initializeControl()
{
}

disconnectControl::disconnectControl() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)DISCONNECT;
}

disconnectControl::~disconnectControl()
{
}

autoFocusMessage::autoFocusMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)AUTO_FOCUS;
}

autoFocusMessage::~autoFocusMessage()
{
}

//--------------------------------------------------------------------
// mftThread Implementation
//--------------------------------------------------------------------

mftThread::mftThread() :
    FAS_THREAD::MsgThread(),
    mft_version(0),
    lens_is_OK(false),
    lens_is_powered(false),
    force_off(false),
    current_aperture(0),
    focus_encoder(0),
    focus_pos(0),
    object_distance(0),
    zoom_encoder(0),
    busy(false)
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Initialize our structure members.
    memset(&lens_info, 0, sizeof(lens_info));

    // Make sure reset is asserted and power is disabled.
    reset(true);
    power(false);

    // Start our polling object.
    lens_poller = std::unique_ptr<mftPoll>(new mftPoll());
    if (lens_poller != nullptr)
        lens_poller->start();
}

mftThread::~mftThread()
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure we release the polling object.
    if (lens_poller != nullptr)
    {
        // Close down the polling thread. Since it's a detached
        // thread, we don't have to wait and join.
        lens_poller->stop();
    }
}

bool mftThread::doConnect()
{
    initializeControl *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;

    // Allocate our message.
    msg = new initializeControl();
    if (msg == nullptr)
        return false;

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);

    // So far, so good.
    return true;
}

bool mftThread::doDisconnect()
{
    disconnectControl *msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;

    // Allocate our message.
    msg = new disconnectControl();
    if (msg == nullptr)
        return false;

    // Add the message to the lens control thread.
    mftThread::factory()->messageAdd(msg);

    // So far, so good.
    return true;
}

/*!
 * @brief This is the static method that controls the access to the
 *        singleton instance.
 * @return The pointer to our singleton object.
 */
mftThread *mftThread::factory()
{
    mftThread *rc(nullptr);

    std::lock_guard<std::mutex> lock(mft_proc_mutex);

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // We only allow one thread object.
    if (mft_proc.get() == nullptr)
    {
        // Create our thread. Can't use "std::make_unique" because the
        // constructor is private.
        mft_proc = std::unique_ptr<mftThread>(new mftThread());
        if (mft_proc.get() != nullptr)
            mft_proc.get()->start();

        // Return the pointer to the singleton.
        rc = mft_proc.get();
    }

    // Check to see if we were halted.
    else if (mft_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = mft_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Process the messages for the thread.
 */
void mftThread::messageHandle(FAS_THREAD::Message *message)
{
    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // We are now busy.
    busy = true;

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case GET_LENS_INFO:
            // Ownership of "message" passes here.
            handleInfoRequest(static_cast<getLensInfoMessage *>(message));

            // We're all done with the message. Do not delete it - we are not
            // the owner!
            message = nullptr;
            break;
        case GET_LENS_STATUS:
            // Ownership of "message" passes here.
            handleStatusRequest(static_cast<getLensStatusMessage *>(message));

            // We're all done with the message. Do not delete it - we are not
            // the owner!
            message = nullptr;
            break;
        case GET_LENS_STATUS_BLOCKING:
            handleStatusBlockingRequest(static_cast<getLensStatusBlockingMessage *>(message));
            break;
        case CHANGE_POWER:
            handleSetPowerRequest(static_cast<changePowerMessage *>(message));
            break;

        case CHANGE_APERTURE:
            // Remember, this is an asynchronous message.
            handleSetApertureRequest(static_cast<changeApertureMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;
        case CHANGE_FOCUS:
            // Remember, this is an asynchronous message.
            handleSetFocusRequest(static_cast<changeFocusMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;
        case CHANGE_ZOOM:
            // Remember, this is an asynchronous message.
            handleSetZoomRequest(static_cast<changeZoomMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;
        case INITIALIZE:
            // Remember, this is an asynchronous message.
            handleInitialize(static_cast<initializeControl *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case DISCONNECT:
            handleDisconnect(static_cast<disconnectControl *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case AUTO_FOCUS:
            handleAFRequest(static_cast<autoFocusMessage *>(message));

            // We're all done with the message and it was a one way trip.
            delete message;
            message = nullptr;
            break;

        case FAS_THREAD::HALT:
            handleDisconnect(nullptr);

            // Tell the poller to stop.
            if (lens_poller.get() != nullptr)
            {
                // Tell the poller to stop, and then wake it up so it can respond to the stop request.
                lens_poller->stop();
                lens_poller->wakeup(0);
            }

            // All done.
            message->resp = FAS_THREAD::OK;
            break;

        case FAS_THREAD::DUMP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        if (message != nullptr)
            message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in mftThread:\n%s", e.what().c_str());
        if (message != nullptr)
        {
            message->resp = FAS_THREAD::FAULT;
            message->forward();
        }
    }

    // Not busy.
    busy = false;
}

void mftThread::lensStop()
{
    std::string temp_str;
    std::ofstream json_file;
    const LENS_STOP2_RESP *lens_stop_info;
    Json::Value lens_json;
    Json::StreamWriterBuilder wr_builder;
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s, typeB", __PRETTY_FUNCTION__);

    // Are we not talking to the lens we can't issue a stop command.
    if (!lens_is_OK)
        return;

    // Make sure there's still a lens attached.
    if (!mftThread::lensPresent())
        return;

    // Build up our message to send the stop command to the lens. This will also
    // return information about the current lens position that we want to save so
    // we can restore it when we power the lens back up.
    msg.isTypeB = true;
    msg.max_count = sizeof(lens_stop_info) + 1;
    msg.cmd_val = LENS_STOP2;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return;
    }
    if (msg.typeB_Fault)
    {
        fi_log(LOG_DEBUG, "%s: Got a type-B fault! =========================================", __PRETTY_FUNCTION__);
    }

    // Read the response.
    lens_stop_info = (const LENS_STOP2_RESP *)msg.data_ptr.data();
    if (lens_stop_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to read response!", __PRETTY_FUNCTION__);
        return;
    }

    // Create the settings file.
    temp_str = fmt::format(LENS_SETTINGS_FILE, get_home_path());
    json_file.open(temp_str, std::ios::out | std::ios::trunc);
    if (!json_file.is_open())
    {
        fi_log(LOG_DEBUG, "%s: Unable to create '%s' [%d: %s]", __PRETTY_FUNCTION__, temp_str.c_str(), errno, strerror(errno));
        return;
    }

    // Save the lens ID information.
    lens_json["member_id"] = lens_info.member_lens_id.member_id;
    lens_json["model_id"] = lens_info.member_lens_id.lens_id.model_id;

    // Convert the stop information to JSON.
    lens_json["zmenc_dist"] = lens_stop_info->zmenc_dist;
    lens_json["zmenc"] = lens_stop_info->zmenc;
    lens_json["lens_pulse"] = focus_pos;
    lens_json["lens_status"] = lens_stop_info->lens_status;
    lens_json["apex"] = current_aperture;

    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 5;
    wr_builder.settings_["precisionType"] = "decimal";
    temp_str = Json::writeString(wr_builder, lens_json);

    // Write the file.
    json_file << temp_str;
    json_file.close();

    // Wait to allow the lens to settle.
    using ms = std::chrono::milliseconds;
    std::this_thread::sleep_for(ms(500));
}

void mftThread::handleAFRequest(autoFocusMessage *msg)
{
    int focus;                  // The current focus value.
    int worse_count;            // The number of times we've gotten worse.
    int last_value;             // The last value, to track if we are falling off the edge.
    int max_value;              // The best we've seen so far.
    int step_size;              // How much to move the focus.
    int best_focus;             // The best focus we've found.

    const int MAX_WORSE_COUNT = 6;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    TP(TRACE_SEL_MFT_PROC, "%s: starting...%d to %d", __PRETTY_FUNCTION__, focus_near, focus_inf);

    (void)msg;

    // Step through the focal reange we're supposed to use.
    last_value = 0;
    max_value = -1;
    worse_count = 0;

    // Needs to be negative to go from larger to smaller value.
    step_size = -((focus_near - focus_inf) / lens_info.focus_divisions);

    best_focus = focus_near;
    focus = focus_near;
    while (true)
    {
        int clarity_value;          // Our calculated clarity of the image.

        TP(TRACE_SEL_MFT_PROC, "%s: TOL %d", __PRETTY_FUNCTION__, focus);

        // Adjust the focus.
        while (true)
        {
            int last_pos = focus_pos;

            // Try the new focus value. Note that the lens may not respond if we don't get the
            // value moved to the next "step".
            setFocus(focus);

            // Are we at the maximum point? If we are, then we are done here.
            if (focus <= focus_inf)
                break;

            // Did we move?
            if (last_pos != focus_pos)
                break;

            // Adjust our focus value some more.
            focus += step_size;
        }

        TP(TRACE_SEL_MFT_PROC, "%s: got focus at %d", __PRETTY_FUNCTION__, focus);

        // Take an image and calculate where we are.
        clarity_value = calcClarity();

        // Did it get better?
        if (clarity_value > max_value)
        {
            best_focus = focus_pos;
            max_value = clarity_value;
            TP(TRACE_SEL_MFT_PROC, "%s: best=%d, max_value=%d", __PRETTY_FUNCTION__, best_focus, max_value);
        }

        // Are we getting worse?
        if (clarity_value < last_value)
            ++worse_count;
        else
            worse_count = 0;

        // Is it time to bail?
        if (worse_count > MAX_WORSE_COUNT)
            break;

        // Update our focus value.
        focus += step_size;
        if (focus <= focus_inf)
            break;
    }

    // Set the focus to the best we found.
    TP(TRACE_SEL_MFT_PROC, "%s: best was %d", __PRETTY_FUNCTION__, best_focus);
    setFocus(best_focus);

    TP(TRACE_SEL_MFT_PROC, "%s: done", __PRETTY_FUNCTION__);
}

int mftThread::calcClarity()
{
    int rc(0);
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Did the image size change?
    if (pm_ptr->getFrameSize() != raw_len)
    {
        // Image size has changed.
        raw_image.reset();
    }

    // Do we need a buffer?
    if (raw_image.get() == nullptr)
    {
        // How many bytes do we need.
        raw_len = pm_ptr->getFrameSize();

        // Allocate our buffer.
        raw_image = std::unique_ptr<uint8_t[]>(new uint8_t[raw_len]);
        if (raw_image.get() == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to allocate the raw_image buffer!", FI_FUNCTION);
            return 0;
        }
    }

    // Get an image.
    int jpeg_length;
    if (!fsp_last_raw_image(raw_image.get(), jpeg_length))
    {
        fi_log(LOG_DEBUG, "%s: Unable to get the last raw live image!", FI_FUNCTION);
        return 0;
    }

    // Decode the image so we can import it into OpenCV.
    std::vector<char> img_data(raw_image.get(), raw_image.get() + jpeg_length);

    // Decode the image as a grayscale.
    cv::Mat img_gray = cv::imdecode(cv::Mat(img_data), cv::IMREAD_GRAYSCALE);

    // Run the Laplacian transform on the image.
    cv::Mat img_tran;
    cv::Laplacian(img_gray, img_tran, CV_8UC1);

    // Return the mean.
    rc = (int)cv::mean(img_tran)[0];
    TP(TRACE_SEL_MFT_PROC, "%s: %d", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Retrieve the lens information.
 */
void mftThread::handleInfoRequest(getLensInfoMessage *msg)
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (info_resp.get() != nullptr)
    {
        fi_log(LOG_DEBUG, "%s: The response is not NULL!", __PRETTY_FUNCTION__);
    }

    // Load the requested structure with information.
    msg->request.connected = lensPresent();
    msg->request.powered = lensIsPowered();
    msg->request.talking = lensIsTalking();
    msg->request.focal_len_tele = lens_info.focal_length_tele;
    msg->request.focal_len_wide = lens_info.focal_length_wide;
    msg->request.focus_divisions = lens_info.focus_divisions;
    msg->request.zoom_divisions = lens_info.zoom_divisions;
    msg->request.ez_enable = ((lens_info.lens_attribute_data & LAD_ELECTRIC_ZOOM) == LAD_ELECTRIC_ZOOM);

    msg->request.name.clear();
    msg->request.name.append(lens_info.lens_name);

    // All done.
    msg->resp = FAS_THREAD::OK;

    // Set the pointer to the response. This will release the
    // currently held pointer - although, as noted above, there
    // really should not be a held pointer.
    info_resp.reset(msg);
}

bool mftThread::getStatusData(LENS_STATUS *status_ptr)
{
    int index;

    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Are we connected? Save the status, but if we're not
    // connected to a camera we will just stop here.
    status_ptr->connected = lensPresent();
    if (!status_ptr->connected)
    {
        // All done here, but done successfully.
        return true;
    }

    // Are we powered up? Save the status, but if we're not
    // then we will just stop here.
    status_ptr->powered = lensIsPowered();
    if (!status_ptr->powered)
    {
        // All done here, but done successfully.
        return true;
    }

    // Are we talking to the lens? Save the status, but if we're not
    // talking to a camera we will just stop here.
    status_ptr->talking = lensIsTalking();
    if (!status_ptr->talking)
    {
        // All done here, but done successfully.
        return true;
    }

    // Acquire the lens status data for live view.
    if (!askStatusLV())
    {
        fi_log(LOG_WARNING, "%s: askStatusLV() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Acquire the current focus position.
    if (!askLDpls())
    {
        fi_log(LOG_WARNING, "%s: askLDpls() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Populate the requested information structure.
    status_ptr->zoom_pos = zoom_encoder;

    // The focal length is the 1st component of the APE data ("Focal length at
    // the encoder center position").
//    LOG("%s: zoom_encoder=%d, zoom_divisions=%d, ape_data.size=%d", __PRETTY_FUNCTION__, zoom_encoder, lens_info.zoom_divisions, ape_data->size());

    // Sanity check the zoom encoder value.
    if ((int)zoom_encoder >= (int)ape_data->size())
    {
        fi_log(LOG_DEBUG, "%s: zoom_encoder (%d) is too big (%d)!", __PRETTY_FUNCTION__, zoom_encoder, ape_data->size());
        zoom_encoder = 0;
    }

    // Get the focal length information.
    status_ptr->focal_len = ape_data->at(zoom_encoder);
//    LOG("%s: zoom_encoder=%d, focal_length=%d", __PRETTY_FUNCTION__, status_ptr->zoom_pos, status_ptr->focal_len);

    // We get the min/max f-number from the APE data. The offset to the minimum
    // f-number we want to use (called "Aperture value of actual design f-stop"
    // in the documentation) is the 4th component. The maximum number is the
    // 5th component ("Aperture value of actual design maximum f-number").
    status_ptr->apex = current_aperture;
    status_ptr->f_number = apexToFnum(current_aperture);
    index = (3 * lens_info.zoom_divisions) + zoom_encoder;
    if (index >= (int)ape_data->size())
    {
        fi_log(LOG_DEBUG, "%s: f_minimum index (%d) into ape_data (%d) is too big!", __PRETTY_FUNCTION__, index, ape_data->size());
        index = 0;
    }
    status_ptr->f_minimum = apexToFnum(ape_data->at(index));
    f_min = status_ptr->f_minimum;
    index = (4 * lens_info.zoom_divisions) + zoom_encoder;
    if (index >= (int)ape_data->size())
    {
        fi_log(LOG_DEBUG, "%s: f_maximum index (%d) into ape_data (%d) is too big!", __PRETTY_FUNCTION__, index, ape_data->size());
        index = 0;
    }
    status_ptr->f_maximum = apexToFnum(ape_data->at(index));
    f_max = status_ptr->f_maximum;
    status_ptr->object_distance = object_distance;

    // The "closest photography distance" is the 3rd component of the ld_data.
    index = (2 * lens_info.zoom_divisions) + zoom_encoder;
    if (index >= (int)ld_data->size())
    {
        fi_log(LOG_DEBUG, "%s: Index (%d) into ld_data (%d) is too big!", __PRETTY_FUNCTION__, index, ld_data->size());
        index = 0;
    }
    status_ptr->focus_min = ld_data->at(index);
    status_ptr->focus_pos = focus_pos;
    status_ptr->focus_near = focus_near;
    status_ptr->focus_inf = focus_inf;
    status_ptr->manual_focus = manual_focus;

    // All done.
    return true;
}

/*!
 * @brief Retrieve the lens information.
 */
void mftThread::handleStatusRequest(getLensStatusMessage *msg)
{
    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (status_resp.get() != nullptr)
    {
        fi_log(LOG_DEBUG, "%s: The response is not NULL!", __PRETTY_FUNCTION__);
    }

    // Fill in the data.
    if (getStatusData(&msg->request))
    {
        // Successfully complete.
        msg->resp = FAS_THREAD::OK;
    }
    else
    {
        // There was a problem.
        msg->resp = FAS_THREAD::FAULT;
    }

    // Set the pointer to the response. This will release the
    // currently held pointer - although, as noted above, there
    // really should not be a held pointer.
    status_resp.reset(msg);
}

/*!
 * @brief Retrieve the lens information.
 */
void mftThread::handleStatusBlockingRequest(getLensStatusBlockingMessage *msg)
{
    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Fill in the data.
    if (getStatusData(&msg->request))
    {
        // Successfully complete.
        msg->resp = FAS_THREAD::OK;
    }
    else
    {
        // There was a problem.
        msg->resp = FAS_THREAD::FAULT;
    }
}

/*!
 * @brief Initialize the lens control interface.
 */
void mftThread::handleDisconnect(disconnectControl *msg)
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Tell the lens to stop.
    lensStop();

    // Make sure we assert reset and turn off the power.
    reset(true);
    power(false);

    // We're no longer talking with the lens.
    lens_is_OK = false;

    // If there's no lens found, then make sure the "force_off" flag is reset.
    force_off = mftThread::lensPresent();

    // All done.
    if (msg != nullptr)
        msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Initialize the lens control interface.
 */
void mftThread::handleInitialize(initializeControl *msg)
{
    TP(TRACE_SEL_MFT_PROC, "%s: force_off=%d", __PRETTY_FUNCTION__, force_off);

    // Don't power up the lens if we're forced off.
    if (force_off)
    {
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Make sure reset is asserted.
    reset(true);

    // Power up the lens.
    if (!powerup())
    {
        fi_log(LOG_DEBUG, "%s: Unable to power-up the lens.", __PRETTY_FUNCTION__);

        // ==============================================
        // ALM_TODO:
        // clean up on the way out.
        // ==============================================

        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Get the lens identification data information.
    if (!askRefDataM())
    {
        fi_log(LOG_WARNING, "%s: ask_ref_data_m() failed.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Get the zoom encoder and aperture data.
    if (!askApeData())
    {
        fi_log(LOG_WARNING, "%s: ask_ape_data() failed.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }
    if (!askLDDataM())
    {
        fi_log(LOG_WARNING, "%s: ask_ld_dataM() failed.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Set the Lens initial operation set.
    if (!setInitalOps())
    {
        fi_log(LOG_WARNING, "%s: Setting inital operation set failed!", __PRETTY_FUNCTION__);
        return;
    }

    // Restore the aperture, if we had the old setting saved.
    if (current_aperture)
        setAperture(current_aperture);
    if (focus_pos)
        setFocus(focus_pos);

    // All done. Remember that we're talking just fine.
    lens_is_OK = true;
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Set the aperture of the lens.
 */
void mftThread::handleSetApertureRequest(changeApertureMessage *msg)
{
    uint16_t apex_value;

    TP(TRACE_SEL_MFT_PROC, "%s: f-number=%f", __PRETTY_FUNCTION__, msg->f_number);

    // Convert the f-number to an APEX value.
    apex_value = fnumToApex(msg->f_number);

    // Tell the lens the new value.
    if (!setAperture(apex_value))
    {
        fi_log(LOG_DEBUG, "%s: setAperture() failed.", __PRETTY_FUNCTION__);
        return;
    }
}

/*!
 * @brief Set the focus of the lens.
 */
void mftThread::handleSetFocusRequest(changeFocusMessage *msg)
{
    TP(TRACE_SEL_MFT_PROC, "%s: focus_pulse: 0x%04x", __PRETTY_FUNCTION__, msg->focus_pulse);

    // Set the focus.
    if (!setFocus(msg->focus_pulse))
    {
        fi_log(LOG_DEBUG, "%s: setFocus() failed.", __PRETTY_FUNCTION__);
        return;
    }
}

/*!
 * @brief Enable/disable power to the lens.
 */
void mftThread::handleSetPowerRequest(changePowerMessage *msg)
{
    TP(TRACE_SEL_MFT_PROC, "%s: enable=%c", __PRETTY_FUNCTION__, msg->enable ? 'T' : 'F');

    // Are we powering up or down?
    if (msg->enable)
    {
        // We can power-up the lens now.
        force_off = false;

        // Tell our MFT control object that we seem to be connected.
        mftThread::doConnect();

        // Let folks know that things have changed.
        fcp_notify_apps(GIGE_REG_LENS_STATUS_JSON, mftThread::lensIsTalking());
    }
    else
    {
        // Make sure we don't turn the lens back on later.
        force_off = true;

        // Tell our MFT control object that we want to be disconnected.
        mftThread::doDisconnect();

        // Let folks know that things have changed.
        fcp_notify_apps(GIGE_REG_LENS_STATUS_JSON, mftThread::lensIsTalking());
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief Set the zoom of the lens.
 *
 * Remember that this is an asynchronous message.
 */
void mftThread::handleSetZoomRequest(changeZoomMessage *msg)
{
    uint16_t focal_len_actual;
    uint8_t set_cmd[CMD_LEN];

    TP(TRACE_SEL_MFT_PROC, "%s: index=%d", __PRETTY_FUNCTION__, msg->index);

    // Make sure this is allowed.
    if ((lens_info.lens_attribute_data & LAD_ELECTRIC_ZOOM) == 0)
    {
        fi_log(LOG_DEBUG, "%s: Electric Zoom is NOT allowed!", __PRETTY_FUNCTION__);
        return;
    }
    // Sanity check the zoom encoder value.
    if (msg->index >= (int)ape_data->size())
    {
        fi_log(LOG_DEBUG, "%s: zoom_encoder (%d) is too big (%d)!", __PRETTY_FUNCTION__, msg->index, ape_data->size());
        return;
    }

    // Get the focal length information.
    // The parameter is actually in 0.1mm units.
    focal_len_actual = ape_data->at(msg->index) * 10;
    TP(TRACE_SEL_MFT_PROC, "%s: focal_len_actual=0x%x [%f mm]", __PRETTY_FUNCTION__, focal_len_actual, ((float)focal_len_actual) / 10.0);

    // Initialize our command.
    memcpy(set_cmd, CONT_EZ_FLENGTH, CMD_LEN);

    // Get information from the user.
    set_cmd[PARAM1_INDEX] = focal_len_actual & 0xFF;
    set_cmd[PARAM2_INDEX] = (focal_len_actual >> 8) & 0xFF;

    // Send the command to the lens. This command has no response.
    if (!sendCommand(set_cmd))
    {
        fi_log(LOG_DEBUG, "%s: CONT_EZ_FLENGTH failed!", __PRETTY_FUNCTION__);
        return;
    }
}

//=============================================================================

/*!
 * @brief Convert from an APEX value to an f-number.
 *
 * From BookD-Ver111A_AnnexM-A-data_en-transl_20111027_byOLY_forFastec Imaging.pdf
 * Pg. 22 (pdf), 17 (doc).
 */
float mftThread::apexToFnum(uint16_t apex_value)
{
    return sqrt(pow(2, (double)(apex_value) / 256.0));
}

/*!
 * @brief Convert from an f-number to an APEX value.
 *
 * From BookD-Ver111A_AnnexM-A-data_en-transl_20111027_byOLY_forFastec Imaging.pdf
 * Pg. 22 (pdf), 17 (doc)
 *
 * NOTE: The documentation referenced above is WRONG! The formula to go from
 *   f-number to APEX value is incorrectly formatted. The correct formula
 *   is:
 *      APEX value = log2(f-number^2) * 256
 */
uint16_t mftThread::fnumToApex(float f_num)
{
    uint16_t apex_value;

    // Convert the f-number to an APEX value.
    f_num = std::log2(f_num * f_num);
    apex_value = (uint16_t)(std::round(f_num * 256.0));

    // Return our value.
    return apex_value;
}

bool mftThread::askApApx(uint16_t& apex_value)
{
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s, typeB", __PRETTY_FUNCTION__);

    // Build up our message to read the current aperture setting.
    msg.isTypeB = true;
    msg.max_count = sizeof(apex_value);
    msg.cmd_val = ASK_APAPX;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (msg.typeB_Fault)
    {
        fi_log(LOG_DEBUG, "%s: Got a fault! =========================================", __PRETTY_FUNCTION__);
    }

    // Copy the data over.
    memcpy((uint8_t *)&apex_value, msg.data_ptr.data(), sizeof(apex_value));

    return true;
}

/*!
 * @brief Returns the zoom encoder dependent focal length and aperture data.
 *
 * @note Make SURE you call this AFTER you initialize the lens_info structure!
 *
 * @return True if everything goes well, false otherwise.
 */
bool mftThread::askApeData()
{
    int num_entries;    // There are 5 fields, each with zoom divisions entries.
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // We SHOULD only call this once, but if it does get called multiple times having
    // the unique_ptr SHOULD recycle the prior allocation.
    num_entries = lens_info.zoom_divisions * 5;
    TP(TRACE_SEL_MFT_PROC, "%s: zoom_div=%d, num_entries=%d", __PRETTY_FUNCTION__, lens_info.zoom_divisions, num_entries);
    if (num_entries == 0)
        num_entries = 5;
    ape_data = std::make_unique<std::vector<uint16_t>>(num_entries);

    // Build up our message to read the focal length and aperture data
    // from the lens.
    msg.isTypeB = false;
    msg.max_count = (2 * num_entries);
    msg.cmd_val = ASK_APE_DATA;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_LDDATA_M_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)ape_data->data(), msg.data_ptr.data(), (2 * num_entries));
    TP(TRACE_SEL_MFT_PROC, "%s: read_count=%d", __PRETTY_FUNCTION__, msg.data_ptr.size());

    // All is well.
    return true;
}

/*!
 * @brief Returns the zoom encoder dependent lens control data.
 *
 * @note Make SURE you call this AFTER you initialize the lens_info structure!
 *
 * @return True if everything goes well, false otherwise.
 */
bool mftThread::askLDDataM()
{
    int num_entries;    // There are 3 fields, each with zoom divisions entries.
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // We SHOULD only call this once, but if it does get called multiple times having
    // the unique_ptr SHOULD recycle the prior allocation.
    num_entries = lens_info.zoom_divisions * 3;
    ld_data = std::make_unique<std::vector<uint16_t>>(num_entries);

    // Build up our message to read the zoom encoder information
    // from the lens.
    msg.isTypeB = false;
    msg.max_count = (2 * num_entries);
    msg.cmd_val = ASK_LDDATA_M;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_LDDATA_M_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)ld_data->data(), msg.data_ptr.data(), (2 * num_entries));

    // All done.
    return true;
}

/*!
 * @brief Read the current focus position of the lens.
 * @return True if the read was a success, false otherwise.
 */
bool mftThread::askLDpls()
{
    ioMftReadDataMessage msg;

//    TP(TRACE_SEL_MFT_PROC, "%s, typeB", __PRETTY_FUNCTION__);

    // Build up our message to read the current focus position
    // of the lens.
    msg.isTypeB = true;
    msg.max_count = sizeof(focus_pos);
    msg.cmd_val = ASK_LDPLS;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (msg.typeB_Fault)
    {
        fi_log(LOG_DEBUG, "%s: Got a fault! =========================================", __PRETTY_FUNCTION__);
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_LDDATA_M_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)&focus_pos, msg.data_ptr.data(), sizeof(focus_pos));

    // All done.
    return true;
}

/*!
 * @brief Read lens specific information.
 * @return True if the read was a success, false otherwise.
 */
bool mftThread::askRefDataM()
{
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Build up our message to read the response.
    msg.isTypeB = false;
    msg.max_count = sizeof(lens_info);
    msg.cmd_val = ASK_REF_DATA_M;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_REF_DATA_M_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)&lens_info, msg.data_ptr.data(), sizeof(lens_info));

    // All done.
    return true;
}

/*!
 * @brief Read the current "live" information from the lens.
 *
 * This will retrieve, amoungst other things, the zoom, focus, and
 * current aperture settings.
 *
 * @return True if the read was a success, false otherwise.
 */
bool mftThread::askStatusLV()
{
    ioMftReadDataMessage msg;
    const ASK_STATUS_LV_RESP *resp;

    // Very noisy, so only uncomment when required!
//    LOG("%s", __PRETTY_FUNCTION__);

    // Build up our message to read the response.
    msg.isTypeB = false;
    msg.max_count = sizeof(ASK_STATUS_LV_RESP);
    msg.cmd_val = ASK_STATUS_LV;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_STATUS_LV_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    resp = (const ASK_STATUS_LV_RESP *)msg.data_ptr.data();

    // The only thing we really care about is the zoom encoder and aperture value.
    zoom_encoder = resp->zoom_encoder;
    focus_encoder = resp->focus_encoder;
    current_aperture = resp->current_aperture;

//    LOG("%s: LiveAF=[0x%04x, 0x%04x, 0x%04x, 0x%04x, 0x%04x, 0x%04x, 0x%04x, 0x%04x]", __PRETTY_FUNCTION__,
//        resp.live_af_distance[0], resp.live_af_distance[1], resp.live_af_distance[2], resp.live_af_distance[3],
//        resp.live_af_distance[4], resp.live_af_distance[5], resp.live_af_distance[6], resp.live_af_distance[7]);
    focus_near = ((int)resp->live_af_distance[6]) & 0xFFFF;
    focus_inf = ((int)resp->live_af_distance[0]) & 0xFFFF;
    manual_focus = (resp->lens_status & LS_MF_ENABLE) == LS_MF_ENABLE;

    // For object distance, we'll translate it to millimeters. -1 is INFINITY.
    // The value returned is in units of 5mm, so we need to multiply it up.
    if (resp->object_distance == 0xFFFF)
        object_distance = -1;
    else
        object_distance = resp->object_distance * 5;

    // All done.
    return true;
}

/*!
 * @brief Get the version information from the lens.
 * @param resp  The MFT version information.
 * @return True if the read was a success, false otherwise.
 */
bool mftThread::askVer(ASK_VER_RESP& resp)
{
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Build up our message to read the response.
    msg.isTypeB = false;
    msg.max_count = sizeof(resp);
    msg.cmd_val = ASK_VER;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_VER_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)&resp, msg.data_ptr.data(), sizeof(resp));

    // All done.
    return true;
}

/*!
 * @brief Set the aperture to the value specified by "apex_value".
 *
 * @param apex_value    [IN] The aperture value we want to send to the lens.
 * @return True if the operation was a success, false otherwise.
 */
bool mftThread::setAperture(uint16_t apex_value)
{
    uint16_t control_ap;
    uint8_t set_cmd[CMD_LEN];
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s: apex=0x%04x, typeB", __PRETTY_FUNCTION__, apex_value);

    // Initialize our command.
    memcpy(set_cmd, CONT_AP_APX_ABS, CMD_LEN);

    // Get information from the user.
    set_cmd[PARAM1_INDEX] = apex_value & 0xFF;
    set_cmd[PARAM2_INDEX] = (apex_value >> 8) & 0xFF;

    // Build up our message to read the response.
    msg.isTypeB = true;
    msg.max_count = sizeof(control_ap) + 1;
    msg.cmd_val = set_cmd;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (msg.typeB_Fault)
    {
        fi_log(LOG_DEBUG, "%s: Got a fault! =========================================", __PRETTY_FUNCTION__);
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_DEBUG, "%s: Unable to read response!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data. We only use this for debug purposes.
    memcpy((uint8_t *)&control_ap, msg.data_ptr.data(), sizeof(control_ap));
    TP(TRACE_SEL_MFT_PROC, fmt::format("{0}: sent={1:#04x}, actual={2:#04x}", __PRETTY_FUNCTION__, apex_value, control_ap).c_str());

    // All is well.
    return true;
}

/*!
 * @brief Set the focus to the value specified by "pulse_value".
 * @param pulse_value   [IN] The focus setting we want to send to the lens.
 * @return True if the operation was a success, false otherwise.
 */
bool mftThread::setFocus(int16_t pulse_value)
{
    uint8_t set_cmd[CMD_LEN];
    ioMftReadDataMessage msg;

    TP(TRACE_SEL_MFT_PROC, "%s: focus_pulse: 0x%04x, typeB", __PRETTY_FUNCTION__, pulse_value);

    // Initialize our command.
    memcpy(set_cmd, CONT_LD_PULSE_ABS, CMD_LEN);

    // Get information from the user.
    set_cmd[PARAM1_INDEX] = ((uint16_t)pulse_value) & 0xFF;
    set_cmd[PARAM2_INDEX] = (((uint16_t)pulse_value) >> 8) & 0xFF;

    // Build up our message to read the response.
    msg.isTypeB = true;
    msg.max_count = sizeof(focus_pos) + 1;
    msg.cmd_val = set_cmd;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }
    if (msg.typeB_Fault)
    {
        fi_log(LOG_DEBUG, "%s: Got a fault! =========================================", __PRETTY_FUNCTION__);
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_DEBUG, "%s: Unable to read response!", __PRETTY_FUNCTION__);
        return false;
    }

    // Remember the new focus position.
    memcpy((uint8_t *)&focus_pos, msg.data_ptr.data(), sizeof(focus_pos));
    TP(TRACE_SEL_MFT_PROC, fmt::format("{0}: sent={1:#04x}, actual={2:#04x}", __PRETTY_FUNCTION__, pulse_value, focus_pos).c_str());

    // All is well.
    return true;
}

bool mftThread::setInitalOps()
{
    uint16_t apex_value = 0;
    std::string temp_str;
    std::ifstream json_file;
    LENS_STOP2_RESP lens_stop_info;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Initialize our structure.
    memset(&lens_stop_info, 0, sizeof(lens_stop_info));

    // When we issue a LENS_STOP2 command, we get back information we can pass to the
    // SET_INIT_ACT_LENS command to restore prior settings.
    //
    // Build up the settings file location and see if it exists.
    temp_str = fmt::format(LENS_SETTINGS_FILE, get_home_path());
    json_file.open(temp_str, std::ios::in);
    if (json_file.is_open())
    {
        Json::Value settings_info;
        Json::CharReaderBuilder builder;
        std::unique_ptr<Json::CharReader> reader;
        std::string errInfo;

        // Read the file.
        temp_str.clear();
        json_file >> temp_str;
//        LOG("%s: json_file:%s", __PRETTY_FUNCTION__, temp_str.c_str());

        // Create our reader.
        reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
        if (reader.get() != nullptr)
        {
            bool parsingSuccessful;

            // Parse the information passed in.
            parsingSuccessful = reader->parse(temp_str.c_str(), temp_str.c_str() + temp_str.size(), &settings_info, &errInfo);
            if (parsingSuccessful)
            {
                uint16_t member_id(0);
                uint16_t model_id(0);

                // Get the member/model IDs from the information.
                if (!settings_info["member_id"].isNull())
                    member_id = (uint16_t)(settings_info["member_id"].asUInt() & 0xFFFF);
                if (!settings_info["model_id"].isNull())
                    model_id = (uint16_t)(settings_info["model_id"].asUInt() & 0xFFFF);

                // Only continue if our lens information matches the IDs we just got.
                if ((lens_info.member_lens_id.member_id == member_id) && (lens_info.member_lens_id.lens_id.model_id == model_id))
                {
                    // Initialize our structure.
                    if (!settings_info["zmenc_dist"].isNull())
                        lens_stop_info.zmenc_dist = (uint8_t)(settings_info["zmenc_dist"].asUInt() & 0xFF);
                    if (!settings_info["zmenc"].isNull())
                        lens_stop_info.zmenc = (uint8_t)(settings_info["zmenc"].asUInt() & 0xFF);
                    if (!settings_info["lens_pulse"].isNull())
                        lens_stop_info.lens_pulse = (uint16_t)(settings_info["lens_pulse"].asUInt() & 0xFFFF);
                    if (!settings_info["lens_status"].isNull())
                        lens_stop_info.lens_status = (uint8_t)(settings_info["lens_status"].asUInt() & 0xFF);
                    if (!settings_info["apex"].isNull())
                        apex_value = (uint16_t)(settings_info["apex"].asUInt() & 0xFFFF);
                }
                else
                {
                    fi_log(LOG_DEBUG, "%s: IDs not a match", __PRETTY_FUNCTION__);
                }
            }
            else
            {
                fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, temp_str.c_str(), temp_str.size());
                fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
            }
        }
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: unable to open '%s' [%d: %s]", __PRETTY_FUNCTION__, temp_str.c_str(), errno, strerror(errno));
    }

    // Send the command. Note that there is no response data for this command.
    if (!sendCommandType6(SET_INIT_ACT_LENS, (const uint8_t *)&lens_stop_info, sizeof(lens_stop_info)))
    {
        fi_log(LOG_WARNING, "%s: SET_INIT_ACT_LENS failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Restore the aperture if we have a saved value.
    if (apex_value != 0)
        current_aperture = apex_value;

    // Restore the focus if we have a saved value.
    if (lens_stop_info.lens_pulse != 0)
        focus_pos = lens_stop_info.lens_pulse;

    // All is well.
    return true;
}

// Fig. D4460-02 of MFT Book D.
bool mftThread::accessoryCheck()
{
    bool found(false);
    ioMftAccessoryCheckMessage msg;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
        if (msg.resp == FAS_THREAD::OK)
            found = msg.accessoryFound;
    }

    // Return our status.
    return found;
}

// Fig. D4460-02 of MFT Book D.
bool mftThread::lensCheck()
{
    bool found(false);
    ioMftLensCheckMessage msg;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
        if (msg.resp == FAS_THREAD::OK)
            found = msg.lensFound;
    }

    // Return our status.
    return found;
}

bool mftThread::lensPresent()
{
    bool lens_found = true;         // Assume we find one.
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

//    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Do we have a lens?
    if (!fpga_ptr->regRead(MFT_CONTROL_STATUS, &reg_value))
    {
        fi_log(LOG_WARNING, "%s: Unable to read MFT_CONTROL_STATUS.", __PRETTY_FUNCTION__);
        return false;
    }

    // There is no lens attached if the "not present" bit is set. Even if
    // there is a lens attached, but the user is pressing the switch on the
    // ring, we assume the lens is not attached.
    if (((reg_value & MFT_LSDET_NOT_PRESENT) == MFT_LSDET_NOT_PRESENT) ||
        ((reg_value & MFT_LSLPSW) == MFT_LSLPSW))
    {
        // There's no lens found.
        lens_found = false;
    }

    // Return our status.
//    TP(TRACE_SEL_MFT_PROC, "%s: lens_found=%d", __PRETTY_FUNCTION__, lens_found);
    return lens_found;
}

bool mftThread::isBusy() const
{
    return busy;
}

bool mftThread::keepOff()
{
    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;

    // Return our flag.
    return mftThread::factory()->force_off;
}

bool mftThread::lensIsPowered()
{
    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;

    // Return our flag.
    return mftThread::factory()->lens_is_powered;
}

bool mftThread::lensIsTalking()
{
    // Sanity check.
    if (mftThread::factory() == nullptr)
        return false;

    // Return our flag.
    return mftThread::factory()->lens_is_OK;
}

// This follows the "Check lens attachment (BCPU)" flowchart, Figure D4460-01 of MFT Book D spec.
bool mftThread::lensPowerup()
{
    // Make our lives easier.
    using ms = std::chrono::milliseconds;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Assert reset.
    reset(true);

    // See if there is a lens present?
    if (!mftThread::lensPresent())
    {
        fi_log(LOG_WARNING, "No lens detected!");
        return false;
    }

    // Enable power.
    if (!power(true))
    {
        fi_log(LOG_DEBUG, "%s: power() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait 20 milliseconds.
    std::this_thread::sleep_for(ms(20));

    // Make sure we're out of reset.
    reset(false);

    // Wait 10 milliseconds.
    std::this_thread::sleep_for(ms(20));

    // Check to see if the lens will talk.
    if (lensCheck())
        return true;

    // Nothing found. Disable power on the way out.
    power(false);

    // Nothing found.
    return false;
}

bool mftThread::power(bool enable)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_MFT_PROC, "%s: enable=%c", __PRETTY_FUNCTION__, enable ? 'T' : 'F');

    if (enable)
    {
        if (!fpga_ptr->regModify(MFT_CONTROL_STATUS, MFT_LSPWR_ENABLE, -1))
        {
            fi_log(LOG_WARNING, "%s: Unable to enable power.", __PRETTY_FUNCTION__);
            return false;
        }
    }
    else
    {
        if (!fpga_ptr->regModify(MFT_CONTROL_STATUS, 0, ~MFT_LSPWR_ENABLE))
        {
            fi_log(LOG_WARNING, "%s: Unable to disable power.", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // Remember our powered state.
    lens_is_powered = enable;

    // All is well.
    return true;
}

bool mftThread::powerup()
{
    ASK_VER_RESP ver_resp;
    ASK_HIGHSPEED_RESP hs_resp;
    ioMftReadDataMessage msg;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    using ms = std::chrono::milliseconds;

    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure the FPGA module is out of reset.
    fpga_ptr->regModify(MFT_CONTROL_STATUS, MFT_CONTROL_RESET, -1);
    std::this_thread::sleep_for(ms(1));
    fpga_ptr->regModify(MFT_CONTROL_STATUS, 0, ~MFT_CONTROL_RESET);

    // Make sure we have a lens and give it power.
    if (!lensPowerup())
    {
        fi_log(LOG_WARNING, "%s: Unable to powerup the lens!", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait 20 milliseconds.
    std::this_thread::sleep_for(ms(20));

    // Have the lens reset.
    if (!resetLens())
    {
        fi_log(LOG_WARNING, "%s: resetLens() failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Wait 100 milliseconds.
    std::this_thread::sleep_for(ms(100));

    // Get the version information.
    if (!askVer(ver_resp))
    {
        fi_log(LOG_WARNING, "%s: askVer() failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Remeber the version.
    mft_version = ver_resp.mft_ver;

    // Ask the lens what it is capable of for high-speed communications.
    msg.isTypeB = false;
    msg.max_count = sizeof(ASK_HIGHSPEED_RESP);
    msg.cmd_val = ASK_HIGHSPEED;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Check for an error response, also happens if we didn't actually
    // send the request.
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: ioMftReadDataMessage request failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure we got something.
    if (msg.data_ptr.size() == 0)
    {
        fi_log(LOG_WARNING, "%s: Unable to read ASK_HIGHSPEED_RESP!", __PRETTY_FUNCTION__);
        return false;
    }

    // Copy over the data.
    memcpy((uint8_t *)&hs_resp, msg.data_ptr.data(), sizeof(ASK_HIGHSPEED_RESP));

    // Specify the highspeed comms parameters. There is no
    // response to this command.
    if (!sendCommand(CHANGE_SPEED))
    {
        fi_log(LOG_WARNING, "%s: CHANGE_SPEED failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Give the lens a chance to transition.
    std::this_thread::sleep_for(ms(10));

    // Acquire the lens status data for live view.
    if (!askStatusLV())
    {
        fi_log(LOG_WARNING, "%s: askStatusLV() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // All done.
    return true;
}

void mftThread::reset(bool do_reset)
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_MFT_PROC, "%s: do_reset=%c", __PRETTY_FUNCTION__, do_reset ? 'T' : 'F');

    if (!do_reset)
    {
        // Take the lens out of reset.
        fpga_ptr->regModify(MFT_CONTROL_STATUS, MFT_LSRST_NOT, -1);
    }
    else
    {
        // Assert reset.
        fpga_ptr->regModify(MFT_CONTROL_STATUS, 0, ~MFT_LSRST_NOT);
    }
}

bool mftThread::resetLens()
{
    ioMftResetLensMessage msg;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());
    }

    // Return our status.
    return (msg.resp == FAS_THREAD::OK);
}

// Write a command sequence to the MFT interface.
// This is appropriate for Type A, B, and C/8 commands.
bool mftThread::sendCommand(const uint8_t *cmd_val)
{
    bool rc = false;
    ioMftSendCommandMessage msg;

    // Sanity checks
    if (cmd_val == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: cmd_val is null!", __PRETTY_FUNCTION__);
        goto sc_bail;
    }

    // Setup for our command.
    msg.cmd_val = cmd_val;
    msg.isType6 = false;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());

        // Was it successful?
        rc = (msg.resp == FAS_THREAD::OK);
    }

sc_bail:

    // Return our status.
    return rc;
}

// Write a command sequence to the MFT interface.
// This is appropriate for Type 6 commands
bool mftThread::sendCommandType6(const uint8_t *cmd_val, const uint8_t *data_val, int data_count)
{
    bool rc = false;
    ioMftSendCommandMessage msg;

    // Sanity checks
    if (cmd_val == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: cmd_val is null!", __PRETTY_FUNCTION__);
        goto sc6_bail;
    }
    if (data_val == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: data_val is null!", __PRETTY_FUNCTION__);
        goto sc6_bail;
    }
    if (data_count <= 0)
    {
        fi_log(LOG_DEBUG, "%s: Invalid data_count: %d", __PRETTY_FUNCTION__, data_count);
        goto sc6_bail;
    }

    // Setup for our command.
    msg.cmd_val = cmd_val;
    msg.isType6 = true;
    msg.data_count = data_count;
    msg.data_ptr = data_val;

    // Check to make sure the I/O worker is running.
    if (fpgaIOWorker::isRunning())
    {
        // Send off the request.
        msg.doSync(fpgaIOWorker::factory());

        // Was it successful?
        rc = (msg.resp == FAS_THREAD::OK);
    }

sc6_bail:
    return rc;
}

//============================================================================
// mftPoll class implementation.
//============================================================================

/*!
 * @brief The polling thread constructor. The MFT polling thread
 *    is a "detached" thread. That is, it runs to completion (once started)
 *    and does not expect to "join" with any other thread.
 */
mftPoll::mftPoll() :
    FAS_THREAD::Thread(true),
    thread_running(false),
    lens_connected(false)
{
}

mftPoll::~mftPoll()
{
}

/*!
 * @brief Main thread function to handle the actual polling of the
 *        lens for changes.
 */
void mftPoll::run()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make things easier for ourselves.
    using ms = std::chrono::milliseconds;

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Clear our last status member.
    memset(&last_status, 0, sizeof(last_status));

    // We are now running.
    thread_running = true;

    // Enable the MFT interrupts.
    fpga_ptr->irqEnable(PCIEIRQ_MFT_LSDET | PCIEIRQ_MFT_LSLPSW);

    // Keep going until we're asked to stop.
    while (!stop_request)
    {
        // Anything change?
        switch (getChangeEvent())
        {
        case CONNECT_EVENT:
            handleConnection();
            break;
        case DISCONNECT_EVENT:
            handleDisconnect();
            break;
        case SETTINGS_CHANGE:
            handleSettingsChange();
            break;
        default:
            break;
        }

        {
            // We're going to wait for an interrupt or a timeout. So we don't really care
            // about the return status of "wait_until".
            std::unique_lock<std::mutex> locker(irq_mutex);
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            irq_status.wait_until(locker, now + ms(MFT_POLL_PERIOD));
        }
    }

    // Make sure we turn off the interrupts.
    fpga_ptr->irqDisable(PCIEIRQ_MFT_LSDET | PCIEIRQ_MFT_LSLPSW);

    // We are no longer running.
    thread_running = false;
    fi_log(LOG_DEBUG, "%s: FINISHED", __PRETTY_FUNCTION__);
}

int mftPoll::getChangeEvent()
{
    int event = NO_EVENT;

    // Check for a lens connection/disconnection event.
    if (lens_connected)
    {
        LENS_STATUS old_status;

        // Before checking for a settings change we should make sure we're still connected.
        // We'll treat the pressing of the power switch as a disconnect event.
        if (!mftThread::lensPresent())
        {
            LOG("%s: We might be disconnected", __PRETTY_FUNCTION__);
            event = DISCONNECT_EVENT;
            lens_connected = false;
            goto gce_bail;
        }

        // Is the lens powered?
        if (!mftThread::lensIsPowered())
        {
            // No power, so we can't talk to the lens, so leave.
            goto gce_bail;
        }

        // Save the last setting.
        old_status = last_status;

        // Get status from the lens. We are happy waiting.
        if (!mft_status_request(last_status, true))
        {
            fi_log(LOG_DEBUG, "%s: mft_status_request() failed.", __PRETTY_FUNCTION__);
            goto gce_bail;
        }

        // Check to see if the aperture changed. Note we're using the APEX value, since
        // floats are notoriously difficult to compare.
        if (old_status.apex != last_status.apex)
        {
            // Sanity check the values before we act on them.
            if ((last_status.f_number >= (last_status.f_minimum - 0.000000001)) && (last_status.f_number <= (last_status.f_maximum + 0.000000001)))
            {
                TP(TRACE_SEL_MFT_PROC, "%s: Aperture changed! [%4.2f, %4.2f, %4.2f]", __PRETTY_FUNCTION__, last_status.f_number, last_status.f_minimum, last_status.f_maximum);
                event = SETTINGS_CHANGE;
                goto gce_bail;
            }
        }

        // Check to see if the focus has changed.
        if (old_status.focus_pos != last_status.focus_pos)
        {
            TP(TRACE_SEL_MFT_PROC, "%s: Focus changed! [%d, %d]", __PRETTY_FUNCTION__, last_status.focus_pos, last_status.object_distance);
            event = SETTINGS_CHANGE;
            goto gce_bail;
        }

        // Check to see if the zoom has changed.
        if (old_status.zoom_pos != last_status.zoom_pos)
        {
            TP(TRACE_SEL_MFT_PROC, "%s: Zoom changed! [%d, %d]", __PRETTY_FUNCTION__, last_status.zoom_pos, last_status.focus_min);
            event = SETTINGS_CHANGE;
            goto gce_bail;
        }

        // Check to see if the manual button has changed.
        if (old_status.manual_focus != last_status.manual_focus)
        {
            TP(TRACE_SEL_MFT_PROC, "%s: Manual focus changed! [%d, %d]", __PRETTY_FUNCTION__, last_status.manual_focus, last_status.manual_focus);
            event = SETTINGS_CHANGE;
            goto gce_bail;
        }
    }
    else
    {
        // Are we connected?
        if (mftThread::lensPresent())
        {
            LOG("%s: We might be connected", __PRETTY_FUNCTION__);
            event = CONNECT_EVENT;
            lens_connected = true;
            goto gce_bail;
        }
    }

gce_bail:
    // Return our event.
    return event;
}

void mftPoll::handleConnection()
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Tell our MFT control object that we seem to be connected.
    mftThread::doConnect();

    // Let folks know that things have changed.
    fcp_notify_apps(GIGE_REG_LENS_STATUS_JSON, mftThread::lensIsTalking());
}

void mftPoll::handleDisconnect()
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Tell our MFT control object that we seem to be disconnected.
    mftThread::doDisconnect();

    // Let folks know that things have changed.
    fcp_notify_apps(GIGE_REG_LENS_STATUS_JSON, mftThread::lensIsTalking());
}

void mftPoll::handleSettingsChange()
{
    TP(TRACE_SEL_MFT_PROC, "%s", __PRETTY_FUNCTION__);

    // Let folks know that things have changed.
    fcp_notify_apps(GIGE_REG_LENS_STATUS_JSON, mftThread::lensIsTalking());
}

void mftPoll::wakeup(uint32_t irq_status)
{
    (void)irq_status;
//    TP(TRACE_SEL_MFT_PROC, "%s: irq_status=0x%08x", __PRETTY_FUNCTION__, irq_status);

    // Acquire the processing mutex.
    std::unique_lock<std::mutex> locker(irq_mutex);

    // Wake up anyone waiting.
    this->irq_status.notify_one();
}
