/*!
 * @file mmapper.cxx
 *
 * @brief This file contains all the routines to deal with mapping
 *    files into memory.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "mmapper.h"

//--------------------------------------------------------------------
// Class implementation: mmapper
//--------------------------------------------------------------------

mmapper::mmapper(void)
{
    aligned_len = 0;
    file_len = 0;
    input_fd = -1;
    mmap_ptr = NULL;
}

/*!
 * @brief MMAPper class destructor.
 */
mmapper::~mmapper()
{
    close();
}

void mmapper::close(void)
{
    // Make sure we have something mapped.
    if (mmap_ptr != NULL)
    {
        munmap(mmap_ptr, aligned_len);
        mmap_ptr = NULL;
    }

    // Make sure we have a file opened.
    if (input_fd != -1)
    {
        ::close(input_fd);
    }

    // Reset our values.
    aligned_len = 0;
    file_len = 0;
    input_fd = -1;
}

void *mmapper::open(const char *file_path)
{
    // Open the source file.
    if ((input_fd = ::open(file_path, O_RDONLY)) == -1)
    {
        syslog(LOG_ERR, "%s: Unable to open file (%s), err=%d(%s).\n",
               __PRETTY_FUNCTION__, file_path, errno, strerror(errno));
        return NULL;
    }

    // Figure out how big the file is.
    file_len = (uint32_t)lseek(input_fd, 0, SEEK_END);
    lseek(input_fd, 0, SEEK_SET);

    // Make sure we have a page aligned length.
    aligned_len = ((file_len + 3) >> 2) << 2;

    // Memory map the file. It's important that the current position be a multiple
    // of the page size.
    mmap_ptr = mmap(NULL, aligned_len, PROT_READ, MAP_PRIVATE, input_fd, 0);
    if (mmap_ptr == (uint8_t *)MAP_FAILED)
    {
        // Log the error.
        syslog(LOG_ERR, "%s: mmap failed, err=%d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
        mmap_ptr = NULL;

        // Close the file.
        close();
    }

    // Return our pointer.
    return mmap_ptr;
}
