/*!
 * @file null_sensor.cxx
 * @brief This source file implements the sensor interface so we can
 *    boot and run the CameraServer even when no sensor is attached.
 *
 * Copyright 2010,2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <stdint.h>
#include <syslog.h>

#include <fastec/fi_ioctl.h>

#include "camera_server.h"
#include "null_sensor.h"


//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief This routine is used to bring the sensor up in a controlled, managed fashion.
 *
 * @param startx    Starting x position in pixels, sensor-relative
 * @param starty    Starting y position in pixels
 * @param width     Width in pixels to capture
 * @param height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param shutter   Integration time (exposure) in microseconds
 * @param opts      Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100 )
 *
 * @return true if everything goes well, false otherwise.
 */
bool NullSensor::sensor_startup(
    uint16_t /*startx*/,
    uint16_t /*starty*/,
    uint16_t /*width*/,
    uint16_t /*height*/,
    fpnmode_t /*fpn_mode*/,
    bitmode_t /*bit_mode*/,
    float /*frame_rate*/,
    uint32_t /*shutter*/,
    uint32_t /*opts*/,
    uint32_t /*sensor_gain*/)
{
    return true;
}

/*!
 * @brief Set up the sensor for a specific ROI.
 *
 * If the sensor is enabled, we disable it, set the ROI registers, then reenable it.
 *
 * @param true_startx   Starting x position in pixels, sensor-relative
 * @param true_starty   Starting y position in pixels
 * @param roi_width Width in pixels to capture
 * @param roi_height    Height in pixels to capture
 * @param fpn_mode  FPN setting
 * @param bit_mode  Bit selection
 * @param frame_rate    Frame rate in frames per second
 * @param integration   Integration time (exposure) in microseconds
 * @param sensor_opts   Sensor-specific options
 * @param sensor_gain   Sensor gain in canonical form ( x 100 )
 *
 * @return true if everything is good, otherwise false.
 */
bool NullSensor::update_capture(uint16_t /*true_startx*/,
                                uint16_t /*true_starty*/,
                                uint16_t /*roi_width*/,
                                uint16_t /*roi_height*/,
                                fpnmode_t /*fpn_mode*/,
                                bitmode_t /*bit_mode*/,
                                float /*frame_rate*/,
                                uint32_t /*integration*/,
                                uint32_t /*sensor_opts*/,
                                uint32_t /*sensor_gain*/)
{
    return true;
}

/*!
 */
bool NullSensor::set_fpga_bit_mode(bitmode_t /*bitmode*/)
{
    return true;
}

int NullSensor::set_sensor_gain(uint32_t /*new_sensor_gain*/)
{
    return 1;
}

bool NullSensor::set_sensor_readout(bool /*enable*/)
{
    return true;
}

//--------------------------------------------------------------------
// Public Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Constructor, create a instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
NullSensor::NullSensor(SENSOR::EEPROM_INFO *eeprom)
    : Sensor(
          640,  /* max width */
          480,  /* max height */
          320,  /* min width */
          240,  /* min height */
          eeprom)
{
    PDEBUG(PDEBUG_TRACE, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Destructor. Delete an instance of this sensor subclass.
 *
 * @return true if everything went well, otherwise false
 */
NullSensor::~NullSensor()
{
    PDEBUG(PDEBUG_TRACE, "%s", __PRETTY_FUNCTION__);
}

bool NullSensor::class_init(FiFPGA */*fpga_ptr*/)
{
    return true;
}

bool NullSensor::is_bit_mode_supported(bitmode_t bit_mode)
{
    if ((bit_mode == GIGE_BITMODE_LOWER_8_BITS) ||
        (bit_mode == GIGE_BITMODE_UPPER_8_BITS) ||
        (bit_mode == GIGE_BITMODE_7_0))
        return true;
    else
        return false;
}

bool NullSensor::is_option_supported(uint32_t sensor_opts)
{
    bool rc = true;
    unsigned opt;

    // Check subsampling.
    opt = (sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT;
    switch (opt)
    {
    case 0:
    case 1:
        break;
    default:
        rc = false;
        break;
    }

    // Check binning.
    opt = (sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT;
    switch (opt)
    {
    case 0:
    case 1:
        break;
    default:
        rc = false;
        break;
    }

    return rc;
}

/*!
 * @brief Return the default bit mode we want to use with this sensor.
 */
bitmode_t NullSensor::get_default_bit_mode(void)
{
    return (bitmode_t)GIGE_BITMODE_UPPER_8_BITS;
}

/*!
 * @brief Return the default frame rate we want to use with this sensor.
 */
uint32_t NullSensor::get_default_frame_rate(void)
{
    return 60;
}

uint32_t NullSensor::get_default_gain(void)
{
    return 100;
}

uint32_t NullSensor::get_frame_rate()
{
    return 0;
}

uint32_t NullSensor::IPM_bayer_pattern_start(uint16_t /*startx*/, uint16_t /*starty*/)
{
    return IPMC_SRC_FRM_BAYER_GB;
}

