/*!
 * @file override.cxx
 * @brief This file contains the implementation of parameter overrides.
 *
 * This is the implementation of overrides for hardware settings
 * and parameters. They are read from a file at startup and can
 * be checked when necessary.
 *
 * TODO use the common parse functions instead of duplicated code.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>
#include <syslog.h>
#include <errno.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <fastec/gige_msg.h>

#include <memory>
#include <string>
#include <fmt/printf.h>

#include "override.h"
#include "camera_server.h"
#include "storage.h"

namespace FI_OVERRIDE
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#define LOG_OVERRIDES

#define MAX_LINE_LEN 127        // Maximum length of lines in overrides file.

// The override enable environment variable.
#define OVERRIDE_ENV_VAR    "OVERRIDE_ENABLE"

// Define some constants to indicate where to look for the override file.
#define OVERRIDE_SDCARD     0
#define OVERRIDE_USB        1
#define OVERRIDE_SCRATCH    2

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//! Where to find the override file on the scratch path.
static const char OVERRIDE_SCRATCH_PATH[] = ".config/Fastec/scratch/override.txt";

//! Generic name for the override file. Note the leading '/'!
static const char OVERRIDE_FILE[] = "/override.txt";

//! Flags for what has been overriden.
static uint32_t override_flags = 0;

//! The overrides parsed from the file, only valid if the corresponding flag is set.
static uint32_t framerate;
static uint32_t integration;
static uint32_t seqmode3;
static uint32_t roi_width;
static uint32_t roi_height;
static uint32_t roi_xoffset;
static uint32_t roi_yoffset;
static uint32_t bit_select;
static uint32_t fsp_pkt_delay;
static uint32_t fsp_frame_delay;
static uint32_t static_ip_addr;
static uint32_t static_netmask;
static uint32_t static_gateway;
static bool     corr_mode;
static uint32_t sata_req_size;
static uint32_t sata_timer_period;
static uint32_t sata_max_sectors;
static uint32_t overprovision_pct;
static bool     white_balance_norm;
static bool     log_ssd_response;
static bool     secure_erase_only;
static bool     disable_csi;
static bool     stream_raw;
static uint32_t fan_temp;
static std::string update_server;
static uint32_t fflr_sess_size;

#ifdef INCLUDE_ALCON_VARS
static uint32_t alcon_burst_len;
static uint32_t alcon_ip_addr;
static uint32_t alcon_port;
static uint32_t alcon_max_mtu;
static bool alcon_ext_trig_polarity;
static bool alcon_sync_out_polarity;
static bool alcon_mode;
static uint8_t alcon_mac_addr[6];
#endif

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------
static bool parse_logging(char *arg);
static bool parse_framerate(char *arg);
static bool parse_integration(char *arg);
static bool parse_seqmode3(char *arg);
static bool parse_roi_width(char *arg);
static bool parse_roi_height(char *arg);
static bool parse_roi_xoffset(char *arg);
static bool parse_roi_yoffset(char *arg);
static bool parse_bit_select(char *arg);
static bool parse_corr_mode(char *arg);
static bool parse_fsp_pkt_delay(char *arg);
static bool parse_fsp_frame_delay(char *arg);
static bool parse_static_ip_addr(char *arg);
static bool parse_static_netmask(char *arg);
static bool parse_static_gateway(char *arg);
static bool parse_sata_req_size(char *arg);
static bool parse_sata_timer_period(char *arg);
static bool parse_sata_max_sectors(char *arg);
static bool parse_overprovision_pct(char *arg);
static bool parse_white_balance_norm(char *arg);
static bool parse_log_ssd_response(char *arg);
static bool parse_secure_erase_only(char *arg);
static bool parse_disable_csi(char *arg);
static bool parse_stream_raw(char *arg);
static bool parse_fan_temp(char *arg);
static bool parse_update_server(char *arg);
static bool parse_fflr_sess_size(char *arg);

#ifdef INCLUDE_ALCON_VARS
static bool parse_alcon_mode(char *arg);
static bool parse_alcon_ext_trig_polarity(char *arg);
static bool parse_alcon_sync_out_polarity(char *arg);
static bool parse_alcon_burst_len(char *arg);
static bool parse_alcon_ip_addr(char *arg);
static bool parse_alcon_mac_addr(char *arg);
static bool parse_alcon_port(char *arg);
static bool parse_alcon_max_mtu(char *arg);
#endif

static void log_results();

//--------------------------------------------------------------------
// Parser handler table.
//--------------------------------------------------------------------
static struct
{
    uint32_t        flag;
    const char      *name;
    bool (*parser)(char *);
} override_list[] =
{
    { OVR_LOGGING,          "logging",          parse_logging },
    { OVR_FRAMERATE,        "framerate",        parse_framerate },
    { OVR_INTEGRATION,      "integration",      parse_integration },
    { OVR_SEQMODE3,         "seqmode3",         parse_seqmode3 },
    { OVR_ROI_WIDTH,        "roi_width",        parse_roi_width },
    { OVR_ROI_HEIGHT,       "roi_height",       parse_roi_height },
    { OVR_ROI_X_OFFSET,     "roi_xoffset",      parse_roi_xoffset },
    { OVR_ROI_Y_OFFSET,     "roi_yoffset",      parse_roi_yoffset },
    { OVR_BIT_SELECT,       "bit_select",       parse_bit_select },
    { OVR_CORRECTION_MODE,  "corr_mode",        parse_corr_mode },
    { OVR_FSP_PKT_DELAY,    "fsp_pkt_delay",    parse_fsp_pkt_delay },
    { OVR_FSP_FRAME_DELAY,  "fsp_frame_delay",  parse_fsp_frame_delay },
    { OVR_STATIC_IP_ADDR,   "static_ip_addr",   parse_static_ip_addr },
    { OVR_STATIC_NET_MASK,  "static_net_mask",  parse_static_netmask },
    { OVR_STATIC_GATEWAY,   "static_gateway",   parse_static_gateway },
    { OVR_SATA_REQ_SIZE,        "sata_req_size",        parse_sata_req_size },
    { OVR_SATA_TIMER_PERIOD,    "sata_timer_period",    parse_sata_timer_period },
    { OVR_SATA_MAX_SECTORS,     "sata_max_sectors",     parse_sata_max_sectors },
    { OVR_OVERPROVISION_PCT,    "overprovision_pct",    parse_overprovision_pct },
    { OVR_WHITE_BALANCE_NORM,   "white_balance_norm",   parse_white_balance_norm },
    { OVR_LOG_SSD_RESPONSE,     "log_ssd_response",     parse_log_ssd_response },
    { OVR_SECURE_ERASE_ONLY,    "secure_erase_only",    parse_secure_erase_only },
    { OVR_DISABLE_CSI,          "disable_csi",          parse_disable_csi },
    { OVR_STREAM_RAW,           "stream_raw",           parse_stream_raw },
    { OVR_FAN_TEMPERATURE,      "fan_temp",             parse_fan_temp },
    { OVR_UPDATE_SERVER,        "server_url",           parse_update_server },
    { OVR_FFLR_SESS_SIZE,       "fflr_sess_size",       parse_fflr_sess_size },

#ifdef INCLUDE_ALCON_VARS
    { OVR_ALCON_MODE,       "alcon_mode",       parse_alcon_mode },
    { OVR_ALCON_SYNC_OUT_POL,   "sync_out_pol",     parse_alcon_sync_out_polarity },
    { OVR_ALCON_EXT_TRIG_POL,   "ext_trig_pol",     parse_alcon_ext_trig_polarity },
    { OVR_ALCON_BURST_LEN,  "burst_len",        parse_alcon_burst_len },
    { OVR_ALCON_MAC_ADDR,   "acc_mac_addr",     parse_alcon_mac_addr },
    { OVR_ALCON_IP_ADDR,    "acc_ip_addr",      parse_alcon_ip_addr },
    { OVR_ALCON_PORT,       "acc_port",     parse_alcon_port },
    { OVR_ALCON_MAX_MTU,    "max_mtu",      parse_alcon_max_mtu }
#endif
};
#define NUM_OVERRIDES ( sizeof( override_list ) / sizeof( override_list[0] ))

// @brief Common function to parse a boolean override
//
// @param arg       Argument text string to be parsed as a bool
// @param var_name  The name of the override variable, just for error messages
// @param ovr_id    The OVR_... identifier
// @param p_rtn     Points to bool variable that, on successful return, is filled in with value
//
// @return true if the parse succeeded and a bool value was written at *p_rtn
//
static bool parse_bool(const char *arg, const char *var_name, uint32_t ovr_id, bool *p_rtn)
{
    char *endp;
    uint32_t temp;

    temp = strtoul(arg, &endp, 10);
    if ((endp != NULL) && (*endp == '\0') && ((temp == 0) || (temp == 1)))
    {
        *p_rtn = temp ? true : false;
        override_flags |= ovr_id;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad %s '%s'", __FUNCTION__, var_name, arg);
        return false;
    }
}

// Switch to common function when convenient.
//
// @brief Common function to parse a uint32_t override
//
// @param arg       Argument text string to be parsed as an unsigned int
// @param var_name  The name of the override variable (just for error messages)
// @param ovr_id    The OVR_... identifier
// @param p_rtn     Points to uint32_t variable that, on successful return, is written with value
//
// @return true if the parse succeeded and a unsigned int value was written at *p_rtn
//
static bool parse_u32(const char *arg, const char *var_name, uint32_t ovr_id, uint32_t *p_rtn)
{
    char *endp;
    uint32_t temp;

    temp = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= ovr_id;
        *p_rtn = temp;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad %s '%s'", __FUNCTION__, var_name, arg);
        return false;
    }
}

static bool parse_framerate(char *arg)
{
    char *endp;

    framerate = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_FRAMERATE;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad framerate '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_integration(char *arg)
{
    char *endp;

    integration = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_INTEGRATION;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad integration time '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_seqmode3(char *arg)
{
    char *endp;

    seqmode3 = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_SEQMODE3;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad seqmode3 '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_roi_width(char *arg)
{
    char *endp;

    roi_width = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ROI_WIDTH;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad roi_width '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_roi_height(char *arg)
{
    char *endp;

    roi_height = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ROI_HEIGHT;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad roi_height '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_roi_xoffset(char *arg)
{
    char *endp;

    roi_xoffset = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ROI_X_OFFSET;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad roi_xoffset '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_roi_yoffset(char *arg)
{
    char *endp;

    roi_yoffset = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ROI_Y_OFFSET;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad roi_yoffset '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_bit_select(char *arg)
{
    uint32_t    parsed_value;
    char    *endp;

    parsed_value = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_BIT_SELECT;
        switch (parsed_value)
        {
        case 8:
            bit_select = GIGE_BITMODE_10_BITS;
            break;
        case 2:
            bit_select = GIGE_BITMODE_MIDDLE_8_BITS;
            break;
        case 1:
            bit_select = GIGE_BITMODE_LOWER_8_BITS;
            break;
        default:
        case 4:
            bit_select = GIGE_BITMODE_UPPER_8_BITS;
            break;
        }

        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad bit_select '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_corr_mode(char *arg)
{
    char *endp;

    // This is a boolean value, so it's true if we get a non-zero value.
    corr_mode = strtoul(arg, &endp, 0) != 0;
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_CORRECTION_MODE;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad corr_mode '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_fsp_pkt_delay(char *arg)
{
    char *endp;

    fsp_pkt_delay = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_FSP_PKT_DELAY;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad fsp_pkt_delay '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_fsp_frame_delay(char *arg)
{
    char *endp;

    fsp_frame_delay = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_FSP_FRAME_DELAY;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad fsp_frame_delay '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_static_ip_addr(char *arg)
{
    static_ip_addr = inet_addr(arg);
    if (static_ip_addr != INADDR_NONE)
    {
        override_flags |= OVR_STATIC_IP_ADDR;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad static_ip_addr '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_static_netmask(char *arg)
{
    static_netmask = inet_addr(arg);
    if (static_netmask != INADDR_NONE)
    {
        override_flags |= OVR_STATIC_NET_MASK;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad static_netmask '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_static_gateway(char *arg)
{
    static_gateway = inet_addr(arg);
    if (static_gateway != INADDR_NONE)
    {
        override_flags |= OVR_STATIC_GATEWAY;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad static_gateway '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_sata_req_size(char *arg)
{
    return parse_u32(arg, "sata_request_size", OVR_SATA_REQ_SIZE, &sata_req_size);
}

static bool parse_sata_timer_period(char *arg)
{
    return parse_u32(arg, "sata_timer_period", OVR_SATA_TIMER_PERIOD, &sata_timer_period);
}

static bool parse_sata_max_sectors(char *arg)
{
    return parse_u32(arg, "sata_max_sectors", OVR_SATA_MAX_SECTORS, &sata_max_sectors);
}

static bool parse_overprovision_pct(char *arg)
{
    return parse_u32(arg, "overprovision_pct", OVR_OVERPROVISION_PCT, &overprovision_pct);
}

static bool parse_white_balance_norm(char *arg)
{
    return parse_bool(arg, "white_balance_norm", OVR_WHITE_BALANCE_NORM, &white_balance_norm);
}

static bool parse_log_ssd_response(char *arg)
{
    return parse_bool(arg, "log_ssd_response", OVR_LOG_SSD_RESPONSE, &log_ssd_response);
}

static bool parse_secure_erase_only(char *arg)
{
    return parse_bool(arg, "secure_erase_only", OVR_SECURE_ERASE_ONLY, &secure_erase_only);
}

static bool parse_disable_csi(char *arg)
{
    return parse_bool(arg, "disable_csi", OVR_DISABLE_CSI, &disable_csi);
}

static bool parse_stream_raw(char *arg)
{
    return parse_bool(arg, "stream_raw", OVR_STREAM_RAW, &stream_raw);
}

static bool parse_fan_temp(char *arg)
{
    return parse_u32(arg, "fan_temp", OVR_FAN_TEMPERATURE, &fan_temp);
}

static bool parse_update_server(char *arg)
{
    update_server = std::string(arg);
    if (!update_server.empty())
    {
        override_flags |= OVR_UPDATE_SERVER;
        return true;
    }

    // No override.
    return false;
}

static bool parse_fflr_sess_size(char *arg)
{
    return parse_u32(arg, "fflr_sess_size", OVR_FFLR_SESS_SIZE, &fflr_sess_size);
}

#ifdef INCLUDE_ALCON_VARS
static bool parse_alcon_ext_trig_polarity(char *arg)
{
    char *endp;

    // This is a boolean value, so it's true if we get a non-zero value.
    alcon_ext_trig_polarity = strtoul(arg, &endp, 0) != 0;
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ALCON_EXT_TRIG_POL;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad ext_trig_polarity '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_sync_out_polarity(char *arg)
{
    char *endp;

    // This is a boolean value, so it's true if we get a non-zero value.
    alcon_sync_out_polarity = strtoul(arg, &endp, 0) != 0;
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ALCON_SYNC_OUT_POL;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad sync_out_polarity '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_mac_addr(char *arg)
{

    if (sscanf(arg, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
               &alcon_mac_addr[ 0 ], &alcon_mac_addr[ 1 ], &alcon_mac_addr[ 2 ],
               &alcon_mac_addr[ 3 ], &alcon_mac_addr[ 4 ], &alcon_mac_addr[ 5 ]) == 6)
    {
        override_flags |= OVR_ALCON_MAC_ADDR;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad acc_mac_addr '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_mode(char *arg)
{
    return parse_bool(arg, "alcon_mode", OVR_ALCON_MODE, &alcon_mode);
}

static bool parse_alcon_burst_len(char *arg)
{
    char *endp;

    alcon_burst_len = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ALCON_BURST_LEN;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad alcon_burst_len '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_ip_addr(char *arg)
{
    alcon_ip_addr = inet_addr(arg);
    if (alcon_ip_addr != INADDR_NONE)
    {
        override_flags |= OVR_ALCON_IP_ADDR;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad alcon_ip_addr '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_port(char *arg)
{
    char *endp;

    alcon_port = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ALCON_PORT;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad alcon_port '%s'", __FUNCTION__, arg);
        return false;
    }
}

static bool parse_alcon_max_mtu(char *arg)
{
    char *endp;

    alcon_max_mtu = strtoul(arg, &endp, 0);
    if ((endp != NULL) && (*endp == '\0'))
    {
        override_flags |= OVR_ALCON_MAX_MTU;
        return true;
    }
    else
    {
        syslog(LOG_ERR, "%s: bad alcon_max_mtu '%s'", __FUNCTION__, arg);
        return false;
    }
}
#endif /* INCLUDE_ALCON_VARS */

static bool parse_logging(char */*arg*/)
{
    override_flags |= OVR_LOGGING;
    return true;
}

static const char *flag_to_str(uint32_t flag)
{
    int i;

    for (i = 0; i < (int)NUM_OVERRIDES; ++i)
    {
        if (override_list[i].flag == flag)
            return override_list[i].name;
    }
    return "unknown";
}

static int str_to_index(char *str)
{
    int i;

    for (i = 0; i < (int)NUM_OVERRIDES; ++i)
    {
        if (strcmp(override_list[i].name, str) == 0)
            return i;
    }
    return -1;
}

static bool parse_line(char *buf)
{
    char *save_ptr;
    char *str;
    int  index;

    if (*buf == '#')
        return false;

    str = strtok_r(buf, "= \t", &save_ptr);
    if (str == NULL)
        return false;

    index = str_to_index(str);
    if (index < 0)
    {
        syslog(LOG_ERR, "%s: unrecognized key '%s'", __FUNCTION__, str);
        return false;
    }

    str = strtok_r(NULL, "= \t", &save_ptr);
    return (*override_list[index].parser)(str);
}

static void log_results()
{
#ifdef LOG_OVERRIDES
    if ((override_flags & OVR_LOGGING) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: logging ENABLED");
    if ((override_flags & OVR_FRAMERATE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: framerate=%u", framerate);
    if ((override_flags & OVR_INTEGRATION) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: integration=%u", integration);
    if ((override_flags & OVR_SEQMODE3) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: seqmode3=0x%02x", seqmode3);
    if ((override_flags & OVR_ROI_WIDTH) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: roi_width=0x%04x", roi_width);
    if ((override_flags & OVR_ROI_HEIGHT) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: roi_height=0x%04x", roi_height);
    if ((override_flags & OVR_ROI_X_OFFSET) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: roi_xoffset=0x%04x", roi_xoffset);
    if ((override_flags & OVR_ROI_Y_OFFSET) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: roi_yoffset=0x%04x", roi_yoffset);
    if ((override_flags & OVR_BIT_SELECT) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: bit_select=0x%02x", bit_select);
    if ((override_flags & OVR_CORRECTION_MODE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: corr_mode=%d", corr_mode);
    if ((override_flags & OVR_FSP_PKT_DELAY) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: fsp_pkt_delay=0x%04x", fsp_pkt_delay);
    if ((override_flags & OVR_FSP_FRAME_DELAY) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: fsp_frame_delay=0x%04x", fsp_frame_delay);
    if ((override_flags & OVR_STATIC_IP_ADDR) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: static_ip_addr=0x%08x", static_ip_addr);
    if ((override_flags & OVR_STATIC_NET_MASK) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: static_net_mask=0x%08x", static_netmask);
    if ((override_flags & OVR_STATIC_GATEWAY) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: static_gateway=0x%08x", static_gateway);
    if ((override_flags & OVR_SATA_REQ_SIZE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: sata_req_size=%d", sata_req_size);
    if ((override_flags & OVR_SATA_TIMER_PERIOD) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: sata_timer_period=%d", sata_timer_period);
    if ((override_flags & OVR_SATA_MAX_SECTORS) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: sata_max_sectors=%d", sata_max_sectors);
    if ((override_flags & OVR_OVERPROVISION_PCT) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: overprovision_pct=%d", overprovision_pct);
    if ((override_flags & OVR_WHITE_BALANCE_NORM) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: white_balance_norm=%d", white_balance_norm);
    if ((override_flags & OVR_LOG_SSD_RESPONSE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: log_ssd_response=%d", log_ssd_response);
    if ((override_flags & OVR_SECURE_ERASE_ONLY) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: secure_erase_only=%d", secure_erase_only);
    if ((override_flags & OVR_DISABLE_CSI) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: disable_csi=%d", disable_csi);
    if ((override_flags & OVR_STREAM_RAW) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: stream_raw=%d", stream_raw);
    if ((override_flags & OVR_FAN_TEMPERATURE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: fan_temp=%d", fan_temp);
    if ((override_flags & OVR_UPDATE_SERVER) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: server_url='%s'", update_server.c_str());
    if ((override_flags & OVR_FFLR_SESS_SIZE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: fflr_sess_size=%u", fflr_sess_size);

#ifdef INCLUDE_ALCON_VARS
    if ((override_flags & OVR_ALCON_EXT_TRIG_POL) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: ext_trig_pol=%d", alcon_ext_trig_polarity);
    if ((override_flags & OVR_ALCON_SYNC_OUT_POL) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: sync_out_pol=%d", alcon_sync_out_polarity);
    if ((override_flags & OVR_ALCON_MODE) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: alcon_mode=%d", alcon_mode ? 1 : 0);
    if ((override_flags & OVR_ALCON_BURST_LEN) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: burst_len=0x%02x", alcon_burst_len);
    if ((override_flags & OVR_ALCON_MAC_ADDR) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: acc_mac_addr=%02x:%02x:%02x:%02x:%02x:%02x",
               alcon_mac_addr[0], alcon_mac_addr[1], alcon_mac_addr[2],
               alcon_mac_addr[3], alcon_mac_addr[4], alcon_mac_addr[5]);
    if ((override_flags & OVR_ALCON_IP_ADDR) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: acc_ip_addr=0x%08x", alcon_ip_addr);
    if ((override_flags & OVR_ALCON_PORT) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: acc_port=0x%04x", alcon_port);
    if ((override_flags & OVR_ALCON_MAX_MTU) != 0)
        syslog(LOG_DEBUG, "OVERRIDE: max_mtu=0x%04x", alcon_max_mtu);
#endif /* INCLUDE_ALCON_VARS */

#endif
}

} /* End of namespace FI_OVERRIDE */

using namespace FI_OVERRIDE;

/*!
 * @brief initialize override system by reading values from a file
 */
bool override_init()
{
    FILE *fp;
    char *p;
    char *env_ptr;      // Pointer to the environment variable.

    std::unique_ptr<char[]> buf;
    std::string temp_path;

    // Initialize all flags to "not overridden".
    override_flags = 0;
    fp = NULL;

    // Is there an environment variable? If not, provide a default value.
    env_ptr = getenv(OVERRIDE_ENV_VAR);
    if ((env_ptr == NULL) || (env_ptr[0] == '\0'))
    {
        // With no override environment variable, we only look in
        // the scratch folder.
        temp_path = fmt::sprintf("%s/%s", get_home_path(), OVERRIDE_SCRATCH_PATH);
        fp = fopen(temp_path.c_str(), "r");
        if (fp != NULL)
        {
            syslog(LOG_INFO, "%s: found %s", __FUNCTION__, temp_path.c_str());
        }
    }
    else
    {
        for (unsigned i = OVERRIDE_SDCARD; i <= OVERRIDE_SCRATCH; ++i)
        {
            std::string device;

            switch (i)
            {
            case OVERRIDE_SDCARD:
                // Get the device path for the SDCARD.
                storage::get_media_device(GIGE_MEDIA_SDCARD, device);
                if (device.empty())
                    break;

                // Where is the SDCard mounted?
                storage::get_mount_dir(device, temp_path);
                if (temp_path.empty())
                    break;

                // Build up the path.
                temp_path.append(OVERRIDE_FILE);
                break;
            case OVERRIDE_USB:
                // Get the device path for the USB.
                storage::get_media_device(GIGE_MEDIA_USB, device);
                if (device.empty())
                    break;

                // Where is the SDCard mounted?
                storage::get_mount_dir(device, temp_path);
                if (temp_path.empty())
                    break;

                // Build up the path.
                temp_path.append(OVERRIDE_FILE);
                break;
            case OVERRIDE_SCRATCH:
                temp_path = fmt::sprintf("%s/%s", get_home_path(), OVERRIDE_SCRATCH_PATH);
                break;
            }

            fp = fopen(temp_path.c_str(), "r");
            if (fp != NULL)
            {
                syslog(LOG_INFO, "%s: found %s", __FUNCTION__, temp_path.c_str());
                break;
            }
        }
    }
    if (fp == NULL)
    {
        return false;
    }

    buf = std::unique_ptr<char[]>(new char[MAX_LINE_LEN + 1]());
    if (buf.get() == nullptr)
        return false;

    // Read and parse lines of the file.
    do
    {
        p = fgets(buf.get(), MAX_LINE_LEN, fp);

        if (p != NULL)
        {
            int len;

            buf[MAX_LINE_LEN] = 0;
            len = strlen(buf.get());

            // Warn if it is close, but try to use it anyway.
            if (len >= MAX_LINE_LEN - 3)
                syslog(LOG_ERR, "WARNING %s line too long", temp_path.c_str());

            // Clear newlines, carriage returns, and spaces at the end.
            while ((len > 0) &&
                   ((buf[len - 1] == '\n') ||
                    (buf[len - 1] == '\r') ||
                    (buf[len - 1] == '\t') ||
                    (buf[len - 1] == ' ')))
            {
                buf[len - 1] = 0;
                --len;
            }

            // Skip over initial whitespace.
            p = buf.get();
            while ((*p == ' ') || (*p == '\t'))
                ++p;

            if (*p != 0)
                parse_line(p);
        }

    } while (p != NULL);

    fclose(fp);
    log_results();
    return true;
}

// @brief return an overridden value for a uint32_t variable
//
// @param flag  Which override to retrieve
// @param p_rtn On return, is updated with the new value if overridden
//      but is untouched otherwise.
//
// @return true if a new value was written at *p_rtn
//
bool override_u32(uint32_t flag, uint32_t *p_rtn)
{
    bool rc = false;

    // See if this flag is set. If not, there was no override so say no.
    if ((flag & override_flags) == 0)
        return false;

    switch (flag)
    {
    case OVR_FRAMERATE:
        *p_rtn = framerate;
        rc = true;
        break;
    case OVR_INTEGRATION:
        *p_rtn = integration;
        rc = true;
        break;
    case OVR_SEQMODE3:
        *p_rtn = seqmode3;
        rc = true;
        break;
    case OVR_ROI_WIDTH:
        *p_rtn = roi_width;
        rc = true;
        break;
    case OVR_ROI_HEIGHT:
        *p_rtn = roi_height;
        rc = true;
        break;
    case OVR_ROI_X_OFFSET:
        *p_rtn = roi_xoffset;
        rc = true;
        break;
    case OVR_ROI_Y_OFFSET:
        *p_rtn = roi_yoffset;
        rc = true;
        break;
    case OVR_BIT_SELECT:
        *p_rtn = bit_select;
        rc = true;
        break;
    case OVR_FSP_PKT_DELAY:
        *p_rtn = fsp_pkt_delay;
        rc = true;
        break;
    case OVR_FSP_FRAME_DELAY:
        *p_rtn = fsp_frame_delay;
        rc = true;
        break;
    case OVR_STATIC_IP_ADDR:
        *p_rtn = static_ip_addr;
        rc = true;
        break;
    case OVR_STATIC_NET_MASK:
        *p_rtn = static_netmask;
        rc = true;
        break;
    case OVR_STATIC_GATEWAY:
        *p_rtn = static_gateway;
        rc = true;
        break;
    case OVR_SATA_REQ_SIZE:
        *p_rtn = sata_req_size;
        rc = true;
        break;
    case OVR_SATA_TIMER_PERIOD:
        *p_rtn = sata_timer_period;
        rc = true;
        break;
    case OVR_SATA_MAX_SECTORS:
        *p_rtn = sata_max_sectors;
        rc = true;
        break;
    case OVR_OVERPROVISION_PCT:
        *p_rtn = overprovision_pct;
        rc = true;
        break;
    case OVR_FAN_TEMPERATURE:
        *p_rtn = fan_temp;
        rc = true;
        break;
    case OVR_FFLR_SESS_SIZE:
        *p_rtn = fflr_sess_size;
        rc = true;
        break;

#ifdef INCLUDE_ALCON_VARS
    case OVR_ALCON_BURST_LEN:
        *p_rtn = alcon_burst_len;
        rc = true;
        break;
    case OVR_ALCON_IP_ADDR:
        *p_rtn = alcon_ip_addr;
        rc = true;
        break;
    case OVR_ALCON_PORT:
        *p_rtn = alcon_port;
        rc = true;
        break;
    case OVR_ALCON_MAX_MTU:
        *p_rtn = alcon_max_mtu;
        rc = true;
        break;
#endif
    default:
        syslog(LOG_ERR, "%s: unknown override 0x%x", __FUNCTION__, flag);
        break;
    }

    if (rc && ((override_flags & OVR_LOGGING) != 0))
        syslog(LOG_DEBUG, "OVERRIDING %s=%u (0x%x)", flag_to_str(flag), *p_rtn, *p_rtn);

    return rc;
}

// @brief return an overridden value for a bool variable
//
// @param flag  Which override to retrieve
// @param p_rtn On return, is updated with the new value if overridden
//      but is untouched otherwise.
//
// @return true if a new value was written at *p_rtn
//
bool override_bool(uint32_t flag, bool *p_rtn)
{
    bool rc = false;

    // See if this flag is set. If not, there was no override so say no.
    if ((flag & override_flags) == 0)
        return false;

    switch (flag)
    {

    case OVR_CORRECTION_MODE:
        *p_rtn = corr_mode;
        rc = true;
        break;
    case OVR_WHITE_BALANCE_NORM:
        *p_rtn = white_balance_norm;
        rc = true;
        break;
    case OVR_LOG_SSD_RESPONSE:
        *p_rtn = log_ssd_response;
        rc = true;
        break;
    case OVR_SECURE_ERASE_ONLY:
        *p_rtn = secure_erase_only;
        rc = true;
        break;
    case OVR_DISABLE_CSI:
        *p_rtn = disable_csi;
        rc = true;
        break;
    case OVR_STREAM_RAW:
        *p_rtn = stream_raw;
        rc = true;
        break;
#ifdef INCLUDE_ALCON_VARS
    case OVR_ALCON_MODE:
        *p_rtn = alcon_mode;
        rc = true;
        break;
    case OVR_ALCON_EXT_TRIG_POL:
        *p_rtn = alcon_ext_trig_polarity;
        rc = true;
        break;
    case OVR_ALCON_SYNC_OUT_POL:
        *p_rtn = alcon_sync_out_polarity;
        rc = true;
        break;
#endif
    default:
        syslog(LOG_ERR, "%s: unknown override 0x%x", __FUNCTION__, flag);
        break;
    }

    if (rc && ((override_flags & OVR_LOGGING) != 0))
        syslog(LOG_DEBUG, "OVERRIDING %s=%d", flag_to_str(flag), *p_rtn ? 1 : 0);

    return rc;
}

// @brief return an overridden value for a uint8_t array
//
// Both the override parser and the caller have to have a priori knowledge of the
// length of the array.
//
// @param flag  Which override to retrieve
// @param p_rtn On return, is updated with a pointer to the uint8_t array if overridden
//      but is untouched otherwise.
//
// @return true if a pointer was written at *p_rtn
//
bool override_u8a(uint32_t flag, uint8_t **p_rtn)
{
    bool rc = false;

    // See if this flag is set. If not, there was no override so say no.
    if ((flag & override_flags) == 0)
        return false;

    switch (flag)
    {

#ifdef INCLUDE_ALCON_VARS
    case OVR_ALCON_MAC_ADDR:
        *p_rtn = alcon_mac_addr;
        rc = true;
        break;
#endif
    default:
        syslog(LOG_ERR, "%s: unknown u8* override 0x%x", __FUNCTION__, flag);
        break;
    }

    if (rc && ((override_flags & OVR_LOGGING) != 0))
        syslog(LOG_DEBUG, "OVERRIDING %s=%02x...", flag_to_str(flag), **p_rtn);

    return rc;
}


// @brief return an overridden value for a string variable
//
// @param flag  Which override to retrieve
// @param rtn_str On return, If overridden will be set to the override value,
//      but is untouched otherwise.
//
// @return true if a pointer was written at *p_rtn
//
bool override_string(uint32_t flag, std::string& rtn_str)
{
    bool rc = false;

    // See if this flag is set. If not, there was no override so say no.
    if ((flag & override_flags) == 0)
        return false;

    switch (flag)
    {

    // Place holder as an example
    case OVR_UPDATE_SERVER:
        rtn_str = update_server;
        rc = true;
        break;
    default:
        syslog(LOG_ERR, "%s: unknown string override 0x%x", __FUNCTION__, flag);
        break;
    }

    if (rc && ((override_flags & OVR_UPDATE_SERVER) != 0))
        syslog(LOG_DEBUG, "OVERRIDING %s='%s'", flag_to_str(flag), rtn_str.c_str());

    return rc;
}

