/*!
 * @file part_mgr.cpp
 *
 * @brief This is the source file for the partition manager. It is
 * responsible for controlling everything about partitions. At shutdown
 * the current partition layout gets saved. At init the layout is restored.
 * It is also responsible for talking with the License Manager to restrict
 * what DDR the user can use. In other words this C++ class keeps
 * track of the capture buffers.
 *
 * Copyright 2010 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>

#include <sys/ioctl.h>
#include <sys/time.h>

#include <cmath>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "autosave.h"
#include "bounce.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_correction.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "lic_manager.h"
#include "override.h"
#include "part_mgr.h"
#include "pm_worker.h"
#include "sensor.h"
#include "wb_file.h"

namespace PART_MGR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Developer note: The primary difference between const and constexpr variables is that
// the initialization of a const variable can be deferred until run time. A constexpr
// variable must be initialized at compile time.

//! 2 GB default size.
constexpr uint64_t IMAGE_MEMORY_SIZE_DEFAULT = 0x80000000;

//! How many bytes is a pixel for a processed image. We always assume RGB for worst case calculations.
constexpr int PROC_IMAGE_PIXEL_BYTE_WIDTH       = 3;

//! FasFireLR defaults to a 4 GiByte session length.
constexpr uint32_t FFLR_DEFAULT_SESSION         = (4096 / 256);

// Macros to extract information from the trigger event extra information.
#define EXTRA_SBS( x )          ((x)[0])
#define EXTRA_IRIG_LOCK( x )    (((x)[1]) & (1 << 31))
#define EXTRA_SBUS( x )         (((x)[1]) & 0xFFFFF)

} /* End of namespace PART_MGR */

//--------------------------------------------------------------------
// Static Declarations
//--------------------------------------------------------------------

std::unique_ptr<PartMgr> PartMgr::instance;
std::mutex PartMgr::instance_mutex;

// Make things easier on ourselves.
using namespace PART_MGR;

/*!
 * @brief Constructor, create a instance of the partition manager class.
 *
 * @return true if everything went well, otherwise false
 */
PartMgr::PartMgr() :
    gainTableIsAllowed(true),
    imageMemorySize(IMAGE_MEMORY_SIZE_DEFAULT),
    imageMemorySizeReserved(0),
    partitionSize(0),
    partitionSizeReserved(0),
    imageBufferSize(0),
    imageBufferCount(0),
    correctionTableSize(0),
    pfmdFrameNum(0),
    triggerReqPos(100)
{
    LOG("%s", __PRETTY_FUNCTION__);

    ulvBufferAddr = 0;
    ulvBufferSize = 0;
    ipmOutputBufferAddress = 0;
    ipmOutputBufferSize = 0;
    pmeBufferAddr = 0;
    pmeBufferSize = 0;

    traceControl = 0 |
                   PM_FUNC |
//                   PM_FPGA |
                   PM_MEMORY |
//                   PM_LTR |
//                   PM_STATS |
//                   PM_VERBOSE |
//                   TRACE_0;        // Set our trace level so we'll trace.
                   TRACE_3;        // Set our trace level to turn off tracing.

    memset(capture_comment, 0, sizeof(capture_comment));
    memset(knots, 0, sizeof(knots));

    // Just initialize these with meaningful values.
    sessionLengthMax = 8;
    sessionLength = 0;
}

/*!
 * @brief The PartMgr class destructor, simply invoke the PartMgr::shutdown()
 *        method to make sure everything closes properly.
 *
 * @return True if everything went well, otherwise false.
 */
PartMgr::~PartMgr()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Make sure we are closed.
    shutdown();
}

/*!
 * @brief Return the singleton instance, creating it if need be.
 *
 * This is the static method that controls the access to the singleton
 * instance. On the first run, it creates a singleton object and places it
 * into the static field. On subsequent runs, it returns the client existing
 * object stored in the static field.
 */
PartMgr *PartMgr::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if (instance.get() == nullptr)
    {
        instance = std::unique_ptr<PartMgr>(new PartMgr());
    }

    return instance.get();
}


/*!
 * @brief Initialize the partition manager class, including initial layout of
 *        image memory.
 *
 * @param fpga_ptr  Pointer to instantiation of FPGA class.
 * @param p_sensor  Pointer to an instantiation of a sensor class.
 * @param lic       Pointer to the licensinc options for the camera.
 *
 * @return true if everything is good, otherwise false.
 */
bool PartMgr::classInit(FiFPGA *fpga_ptr, Sensor *sensor_ptr, const LicManager *p_lic)
{
    bitmode_t   max_bit_mode;
    uint16_t    width = 0;
    uint16_t    height = 0;
    int num_partitions;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    width = sensor_ptr->get_sensor_max_width();
    height = sensor_ptr->get_sensor_max_height();
    max_bit_mode = sensor_ptr->get_max_bit_mode();

    // Ask the license manager how much memory there is
    // Memory from the license manager is in Gigabytes, we need to convert to quadwords.
    if (p_lic != nullptr)
    {
        // Convert the licensed amount of memory from MB to bytes.
        imageMemorySize = ((uint64_t)p_lic->memoryMax()) * (1024ULL * 1024ULL);
    }
    else
    {
        imageMemorySize = IMAGE_MEMORY_SIZE_DEFAULT;  // no license manager just use the default amount
    }

    //////////////////// Image Memory Layout Table ///////////////////////////
    // These sizes are based on the max for the current sensor and license.
    // - 2 (ping pong) buffers per IPM channel, 24-bit pixels packed, no meta:  width x height x 3 x 2 x 3
    //   These buffers can hold processed or raw data, depending upon what's asked for.
    //
    // The rest is allocated to capture buffers. For non-LTR mode, it is divided into
    // partitions which have metadata and possibly a correction table.
    //////////////////////////////////////////////////////////////////

    // Allocate space for the unarmed live view buffer. We need a place for the ICM
    // to place the images we ask for. Over allocate for simplicity.
    ulvBufferAddr = 0;
    ulvBufferSize = convert_to_HEXWORD(width * height * PROC_IMAGE_PIXEL_BYTE_WIDTH) << HEXWORD_SHIFT;

    // Tell the FPGA the address and buffer size.
    // This will also start a DMA to clear that area of memory.
    fpga_ptr->icmSetULVBuffers(ulvBufferAddr, ulvBufferSize, NUM_ULV_BUFFERS);

    // Allocate space for the per-frame metadata extraction.
    pmeBufferAddr = ulvBufferAddr + (NUM_ULV_BUFFERS * ulvBufferSize);
    pmeBufferSize = PME_BUFFER_SIZE;

    // Allocate space for the IPM "ping-pong" buffers. Each IPM buffer is,
    // worst case, a processed 24-bit/pixel (3 bytes/pixel) image.
    ipmOutputBufferAddress = pmeBufferAddr + pmeBufferSize;
    ipmOutputBufferSize = convert_to_HEXWORD(width * height * PROC_IMAGE_PIXEL_BYTE_WIDTH) << HEXWORD_SHIFT;
    ipm_proc_reset();

    // Determine the total amount of buffer space we reserve out of the image memory.
    imageMemorySizeReserved = (ulvBufferSize * NUM_ULV_BUFFERS) + pmeBufferSize + (ipmOutputBufferSize * NUM_IPM_PP_BUFFERS);

    //------------------------------------------------------------------------
    // Per-session reserved memory space.
    //
    // Each session requires the following buffers. These required buffers
    // consume some of the session's image memory.
    //------------------------------------------------------------------------

    // Calculate the correction (and gain) table size and how much total is reserved per partition.
    // Note that both the correction table and gain table must start on a hexword boundary! Since we are forcing the
    // size of the table(s) to a hexword granularity, we can ensure that the alignment requirement is met.
    correctionTableSize = convert_to_HEXWORD((uint32_t)width * (uint32_t)height * FI_FPGA::FPGA_CORR_TABLE_BPP) << HEXWORD_SHIFT;
    partitionSizeReserved = PARTITION_METADATA_SIZE + (2 * correctionTableSize);
    partitionSizeReserved = convert_to_HEXWORD(partitionSizeReserved) << HEXWORD_SHIFT;

    // Calculate the maximum session length. This is the number of 256 MiByte "chunks" we can fit
    // per GByte (that would be 4 per gigabyte).
    // We round "up" because, even though that last session isn't a full 256 MiByte, we want the user
    // to be able to make use of as much memory as is available.
    sessionLengthMax = (uint32_t)std::ceil(((double)(imageMemorySize - imageMemorySizeReserved) / (double)(1024.0 * 1024.0 * 1024.0)) * 4.0);

    // Set some defaults. These will be set for real later using GigE registers.
    if (sessionLength == 0)
        sessionLength = sessionLengthMax;
    num_partitions = calcSessionCount(sessionLength);
    if (num_partitions == 0)
        num_partitions = 1;

    // Set up the partition array. We will do this for real later when we init from GigE regs.
    layoutMemory(width, height, max_bit_mode, sessionLength, num_partitions);

    TP(PM_MEMORY, "%s: finished im_size:0x%" PRIx64 " im_rsv:0x%" PRIx64 " num_part:%d part_size:0x%" PRIx64 " part_size_rsv:0x%08x",
       __FUNCTION__,
       imageMemorySize, imageMemorySizeReserved, num_partitions, partitionSize, partitionSizeReserved);
    TP(PM_MEMORY, "%s: finished ulv_buf_addr:0x%" PRIx64 " ulv_buf_size:0x%08x",
       __FUNCTION__, ulvBufferAddr, ulvBufferSize);
    TP(PM_MEMORY, "%s: finished pme_buf_addr:0x%" PRIx64 " pme_buf_size:0x%08x",
       __FUNCTION__, pmeBufferAddr, pmeBufferSize);
    TP(PM_MEMORY, "%s: finished ipmOutputBufferAddress[0]:0x%" PRIx64 " ipmOutputBufferAddress[1]:0x%" PRIx64 " ipmOutputBufferSize:0x%08x",
       __FUNCTION__,
       ipmOutputBufferAddress, ipmOutputBufferAddress + ipmOutputBufferSize, ipmOutputBufferSize);
    TP(PM_MEMORY, "%s: max_session_length:%d", __FUNCTION__, sessionLengthMax);

    return true;
}

/*!
 * @brief Return the address in image memory for the start of the
 *        active partition's correction table.
 */
uint64_t PartMgr::get_active_ct_addr()
{
    uint64_t part_start;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the starting address of the partition.
    part_start = getPartitionStart();
    if (part_start == 0)
    {
        fi_log(LOG_DEBUG, "%s: get_partition_start() failed!", __PRETTY_FUNCTION__);
        return 0;
    }

    // The correction table is located immediately following the
    // partition metadata structure.
    return (part_start + PARTITION_METADATA_SIZE);
}

/*!
 * @brief Get the active partition's correction table address
 *        as a QUADWORD.
 */
uint32_t PartMgr::get_active_ct_addr_qw()
{
    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);
    return convert_to_QWORD(PARTITION_METADATA_SIZE);
}

/*!
 * @brief Return the address in image memory for the start of the
 *        active partition's gain table.
 *
 * The correction_tbl_size value must have a HEXWORD granularity. This
 * ensures that the requirement for the table to align on a HEXWORD
 * boundary will be met.
 */
uint64_t PartMgr::get_active_gt_addr()
{
    uint64_t active_addr(0);
    PART_MGR_WORKER::pmwGetActivePartMessage msg;

    // Retrieve the active partition information structure.
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwGetActivePartMessage failed!", __PRETTY_FUNCTION__);
        goto gt_bail;
    }

    TP(PM_FUNC, "%s: active_gt_addr=0x%" PRIx64 ", sov_off=0x%" PRIx64 " --------------------------", __PRETTY_FUNCTION__,
       (get_active_ct_addr() + correctionTableSize), msg.active_info.capture_buf_start);

    // The gain table immediately follows the correction table.
    active_addr = get_active_ct_addr() + correctionTableSize;
    if (active_addr >= msg.active_info.capture_buf_start)
    {
        // There is no gain table allocated for this partition.
        active_addr = 0;
    }

gt_bail:
    // Return the address.
    return active_addr;
}

int32_t PartMgr::getActivePartition(PART_MGR_WORKER::PartInfo *part_ptr)
{
    int32_t active_part;
    PART_MGR_WORKER::pmwGetActivePartMessage msg;

    // Retrieve the active partition information structure.
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwGetActivePartMessage failed!", __PRETTY_FUNCTION__);
        active_part = -1;
    }
    else
    {
        // Return the active indes.
        active_part = msg.active_index;

        // Does the user want the partition information too?
        if (part_ptr != nullptr)
            memcpy(part_ptr, &msg.active_info, sizeof(PART_MGR_WORKER::PartInfo));
    }

    // Return our result.
    return active_part;
}

int32_t PartMgr::getReviewPartition(PART_MGR_WORKER::PartInfo *part_ptr)
{
    int32_t review_part;
    PART_MGR_WORKER::pmwGetReviewPartMessage msg;

    // Retrieve the review partition information structure.
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwGetReviewPartMessage failed!", __PRETTY_FUNCTION__);
        review_part = -1;
    }
    else
    {
        // Return the review indes.
        review_part = msg.review_index;

        // Does the user want the partition information too?
        if (part_ptr != nullptr)
            memcpy(part_ptr, &msg.review_info, sizeof(PART_MGR_WORKER::PartInfo));
    }

    // Return our result.
    return review_part;
}

/*!
 * @brief Get the address of a buffer for capturing black frames during calibration.
 *
 * Uses the memory allocated for the active partition if in standard mode and the
 * LTR buffers if in LTR mode.
 *
 * @param rtn_addr  Return the image memory address of the capture area start.
 * @param rtn_size  Return the size of each frame in quadwords.
 * @param rtn_count Return the number of frames that can be captured in rtn_addr.
 *
 * @return true for success, false for failure (which must be due to a bug).
 */
void PartMgr::getCalibrationBuf(uint64_t *rtn_addr, uint32_t *rtn_size, uint32_t *rtn_count)
{
    FiFPGA  *fpga_ptr;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    fpga_ptr = CameraServer::get_fpga();

    // Handle things differently for LTR mode.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        *rtn_addr = ltrGetCaptureBufferAddr(rtn_count);
    }
    else
    {
        PART_MGR_WORKER::pmwGetActivePartMessage msg;

        // Retrieve the active partition information structure.
        msg.doSync(PART_MGR_WORKER::pmWorker::factory());
        if (msg.resp != FAS_THREAD::OK)
        {
            fi_log(LOG_DEBUG, "%s: pmwGetActivePartMessage failed!", __PRETTY_FUNCTION__);
        }

        else
        {
            *rtn_addr = msg.active_info.capture_buf_start;
            *rtn_count = imageBufferCount;
        }
    }

    *rtn_size = imageBufferSize;
}

/*!
 * @brief Get the current comment that will go into partition metadata when it is saved.
 *
 * Copies the capture comment into the supplied buf. The result is null-terminated so
 * on return the buffer will contain no more than buf_len-1 characters.
 *
 * @param buf       Buffer to receive comment text.
 * @param buf_len   Size of buffer.
 */
void PartMgr::get_comment(char *buf, unsigned buf_len)
{
    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    memset(buf, 0, buf_len);
    strncpy(buf, capture_comment, buf_len - 1);
}

/*!
 * @brief Set the comment that will go into the partition metadata when it is saved.
 *
 * @param buf   Null-terminated comment text string.
 */
void PartMgr::set_comment(const char *buf)
{
    int i;
    int num_parts;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // How many partitions do we have?
    num_parts = PART_MGR_WORKER::pmWorker::factory()->getNumParts();

    // Don't leave residue in the binary save data.
    memset(capture_comment, 0, sizeof(capture_comment));

    // Retain whatever size we can hold.
    strncpy(capture_comment, buf, sizeof(capture_comment) - 1);

    // Copy it into all the partitions.
    for (i = 0; i < num_parts; ++i)
    {
        PART_MGR_WORKER::PartInfo part_info;

        // Retrieve the partition information.
        if (!getPartInfo(i, part_info))
        {
            fi_log(LOG_DEBUG, "%s: getPartInfo() failed on [%d]!", __PRETTY_FUNCTION__, i);
            break;
        }

        // Update the comment field.
        memset(part_info.metadata.comment, 0, sizeof(part_info.metadata.comment));
        strncpy((char *)part_info.metadata.comment, capture_comment, sizeof(part_info.metadata.comment));

        // Copy it over to the set message and send it on over.
        if (!updatePartInfo(i, part_info))
        {
            fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on [%d]!", __PRETTY_FUNCTION__, i);
            break;
        }
    }
}

/*!
 * @brief Set the knots for the current LUT so they can be stored in metadata.
 *
 * @param knots     Array of knots
 */
void PartMgr::set_knots(const LUT_RGB_FILE *lut_ptr)
{
    int i;
    int num_parts;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retain whatever size we can hold.
    memset(knots, 0, sizeof(knots));
    if (lut_ptr != nullptr)
    {
        if (lut_ptr->hdr.lutDataType == LUT_SERIESNC_COLOR)
        {
            // The first (red) set is starting at 0.
            memcpy(&knots[0], lut_ptr->redKnots, LUT_MAX_KNOT_COUNT * sizeof(struct spline_knot));

            // The second (green) set is immediately following the red.
            memcpy(&knots[LUT_MAX_KNOT_COUNT], lut_ptr->greenKnots, LUT_MAX_KNOT_COUNT * sizeof(struct spline_knot));

            // The last (blue) set is immediately following the green.
            memcpy(&knots[LUT_MAX_KNOT_COUNT * 2], lut_ptr->blueKnots, LUT_MAX_KNOT_COUNT * sizeof(struct spline_knot));
        }
        else
        {
            // For a monochrome image, we just populate the redKnots entry.
            memcpy(knots, lut_ptr->redKnots, LUT_MAX_KNOT_COUNT * sizeof(struct spline_knot));
        }
    }

    // How many partitions do we have?
    num_parts = PART_MGR_WORKER::pmWorker::factory()->getNumParts();

    // Copy it into all the partitions.
    for (i = 0; i < num_parts; ++i)
    {
        PART_MGR_WORKER::PartInfo part_info;

        // Retrieve the partition information.
        if (!getPartInfo(i, part_info))
        {
            fi_log(LOG_DEBUG, "%s: getPartInfo() failed on [%d]!", __PRETTY_FUNCTION__, i);
            break;
        }

        // Update the LUT knots field.
        memcpy(&part_info.metadata.lut_knots, knots, sizeof(part_info.metadata.lut_knots));

        // Copy it over to the set message and send it on over.
        if (!updatePartInfo(i, part_info))
        {
            fi_log(LOG_DEBUG, "%s: updatePartInfo failed on [%d]!", __PRETTY_FUNCTION__, i);
            break;
        }
    }
}

/*!
 * @brief Retrieve a copy of the partitions metadata structure associated with part_num.
 * @return A unique_ptr containing the copy of the active partitions pm_metadata structure.
 */
std::unique_ptr<pm_metadata> PartMgr::getPartitionMetadata(int part_num)
{
    PART_MGR_WORKER::PartInfo part_info;
    std::unique_ptr<pm_metadata> metadata_ptr;

    // Sanity check the partition number.
    if ((part_num < 0) || (part_num >= PART_MGR_WORKER::pmWorker::factory()->getNumParts()))
    {
        // Invalid partition number.
        goto bail;
    }

    // Retrieve the active partition information structure.
    if (getPartInfo(part_num, part_info))
    {
        // Allocate our structure.
        metadata_ptr = std::unique_ptr<pm_metadata>(new pm_metadata);
        if (metadata_ptr.get() == nullptr)
        {
            LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        }
        else
        {
            // Copy the data over.
            memcpy(metadata_ptr.get(), &part_info.metadata, sizeof(pm_metadata));
        }
    }

bail:
    return metadata_ptr;
}

uint32_t PartMgr::getNumFrames()
{
    uint32_t num_frames;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the active partition information structure.
    if (getActivePartition(&part_info) < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        num_frames = 0;
    }
    else
    {
        num_frames = part_info.num_frames;
    }

    return num_frames;
}

uint64_t PartMgr::getPartitionStart()
{
    uint64_t part_start;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the active partition information structure.
    if (getActivePartition(&part_info) < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        part_start = 0;
    }
    else
    {
        part_start = part_info.partition_start;
    }

    return part_start;
}

uint8_t PartMgr::getPartitionState()
{
    uint8_t part_state;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the active partition information structure.
    if (getActivePartition(&part_info) < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        part_state = 0;
    }
    else
    {
        part_state = part_info.partition_state;
    }

    return part_state;
}

uint32_t PartMgr::getTriggerFrame()
{
    uint32_t trigger_frame;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the active partition information structure.
    if (getActivePartition(&part_info) < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        trigger_frame = 0;
    }
    else
    {
        trigger_frame = part_info.trig_frame_num;
    }

    return trigger_frame;
}

bool PartMgr::layoutMemoryFromRegs()
{
    // Get our support classes.
    CameraServer    *cs_ptr = CameraServer::get_cs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PART_MGR_WORKER::PartInfo part_info;

    bool    rtn = false;
    bitmode_t   bit_mode;
    uint16_t    width;
    uint16_t    height;
    uint32_t    sess_len;
    uint32_t    num_parts;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our values.
    cs_ptr->get_user_bit_mode(&bit_mode);
    cs_ptr->get_user_roi(nullptr, nullptr, &width, &height);

    // Reset our counter for the total number of partitions captured.
    finished_part_count = 0;

    // Are we enabling FasFireLR Mode?
    if ((regs[GIGE_REG_LTR_MODE].get_value() & GIGE_LTR_MODE_FFLR) == GIGE_LTR_MODE_FFLR)
    {
        uint32_t ovr_sess_size;

        // Did the session size get overriden?
        if (override_u32(OVR_FFLR_SESS_SIZE, &ovr_sess_size))
        {
            // The override size is in MiBytes. Session length is in 1/4 GiBytes.
            sess_len = ovr_sess_size / 256;
        }
        else
        {
            // FasFireLR mode has a default session length of 4 GiB.
            sess_len = FFLR_DEFAULT_SESSION;
        }

        // Make sure we update the register value.
        regs[GIGE_REG_SESSION_LENGTH].set_value(sess_len);
        fcp_notify_apps(GIGE_REG_SESSION_LENGTH, sess_len);
    }
    else
    {
        // Use the register setting.
        sess_len = regs[GIGE_REG_SESSION_LENGTH].get_value();
    }
    num_parts = calcSessionCount(sess_len);

    // JUL2014 remove GIGE_REG_NUM_PART or make sure it is used consistently.
    regs[GIGE_REG_NUM_PART].set_value(num_parts);    // JUL2014 Reg does not seem to be used.

    // Configure the memory layout: size of frames, number of frames, etc.
    rtn = layoutMemory(width, height, bit_mode, sess_len, num_parts);

    // Fresh layout, so make sure the FPGA capture FIFO is reset.
    fpga_ptr->icmResetCaptureFIFO();

    // Now set the capture buffers in the FPGA
    queueCapturePartition(getActivePartition());

    // For LTR mode, the capture buffer info will be set whenever we start the recording,
    // overwriting what we just set up. But at least we have sensible initial values JIC.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        ltrSetCaptureBufferCount(imageBufferCount);
    }

    // Update registers that depend on the memory layout.
    regs[GIGE_REG_PART_SIZE].set_value(convert_to_QWORD(partitionSize));
    regs[GIGE_REG_FRAME_SIZE].set_value(imageBufferSize);
    regs[GIGE_REG_NUM_FRAMES].set_value(imageBufferCount);

    // Let folks know about this.
    fcp_notify_apps(GIGE_REG_NUM_PART, num_parts);
    fcp_notify_apps(GIGE_REG_PART_SIZE, convert_to_QWORD(partitionSize));
    fcp_notify_apps(GIGE_REG_FRAME_SIZE, imageBufferSize);
    fcp_notify_apps(GIGE_REG_NUM_FRAMES, imageBufferCount);

    // Strictly for the debug message...
    getPartInfo(0, part_info);
    fi_log(LOG_DEBUG, "%s: num_parts:%u part_addr:0x%" PRIx64 " part_size:%" PRIu64 " cap start:0x%" PRIx64 " size:%u cnt:%u",
           __PRETTY_FUNCTION__, PART_MGR_WORKER::pmWorker::factory()->getNumParts(), part_info.partition_start, partitionSize,
           part_info.capture_buf_start, imageBufferSize, imageBufferCount);

    return rtn;
}

void PartMgr::resetFflr()
{
    // Reset the number of partitions finished.
    finished_part_count = 0;

    // Reset the trigger partition value.
    trigger_part = -1;
    fflr_trigger_frame = 0;
}

/*!
 * @brief Public facing method to be used for closing down the partition manager when
 *        the camera server itself is shutting down.
 */
void PartMgr::shutdown()
{
    LOG("%s++", __PRETTY_FUNCTION__);

    // Is our working thread running?
    if (PART_MGR_WORKER::pmWorker::isRunning())
    {
        PART_MGR_WORKER::pmWorker::close();
    }

    LOG("%s--", __PRETTY_FUNCTION__);
}

/*!
 * @brief Lay out memory for the current mode and parameters.
 *
 * @param width     Width of the image, in pixels.
 * @param height    Height of the image.
 * @param bit_mode  Tells us (effectively) bytes per pixel.
 * @param sess_len  In 1/4 GiB increments.
 * @param num_parts The number of partitions we're supporting.
 *
 * @return True if everything is good, otherwise false.
 */
bool PartMgr::layoutMemory(uint16_t width, uint16_t height, bitmode_t bit_mode, uint32_t sess_len, uint32_t num_parts)
{
    // Get our support classes.
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    uint64_t    temp_addr;
    uint64_t    partition_size_avail;     // each partition size - fixed items

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Reset our counter for the total number of partitions captured.
    finished_part_count = 0;

    // Set up the size of the image memory capture buffers based upon the
    // number of partitions. Figure out how big a buffer we need based upon
    // the current settings.
    imageBufferSize = cs_ptr->calc_frame_buf_size(width, height, bit_mode);
    imageBufferSize = convert_to_HEXWORD(imageBufferSize) << HEXWORD_SHIFT;

    // Add the metadata size to it. Even if we are in LTR mode,
    // we still write the per-frame metadata with each frame during capture.
    imageBufferSize += FRAME_METADATA_HW_BYTES;
    imageBufferSize = convert_to_HEXWORD(imageBufferSize) << HEXWORD_SHIFT;

    // Let the FPGA know the frame size.
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_SIZE, convert_to_HEXWORD(imageBufferSize));

    // Sanity check.
    if (num_parts == 0)
    {
        fi_log(LOG_DEBUG, "%s: num_parts == 0?!", __PRETTY_FUNCTION__);
        return false;
    }
    if (imageBufferSize == 0)
    {
        fi_log(LOG_DEBUG, "%s: im_buf_size == 0?!", __PRETTY_FUNCTION__);
        return false;
    }

    // Figure out how many will fit in the partition size.
    if (sess_len < sessionLengthMax)
        partitionSize = ((uint64_t)sess_len) * SESSION_INCR;
    else
    {
        // Sanity check.
        if (sess_len > sessionLengthMax)
        {
            fi_log(LOG_DEBUG, "%s: sess_len [%d] (max=%d) is not right.", __PRETTY_FUNCTION__, sess_len, sessionLengthMax);
            sess_len = sessionLengthMax;
        }
        partitionSize = (imageMemorySize - imageMemorySizeReserved) / num_parts;
    }
    partition_size_avail = partitionSize - partitionSizeReserved;

    // Calculate the number of buffers that will fit in the available space.
    imageBufferCount = (uint32_t)(partition_size_avail / imageBufferSize);

    // We want to make sure that, when loading in from SSD, we don't
    // overflow the end of the partition. Transfers to/from the SSD
    // happen on an SSD_PAGE_LEN boundary...so adjust accordingly.
    //
    // Calculate the start address of the last buffer of the partition.
    temp_addr = (uint64_t)(imageBufferCount - 1) * (uint64_t)imageBufferSize;

    // Calculate the ending address of the buffer when we're
    // transferring from the SSD. We want the ENDING address,
    // not the START address of the next buffer! Round the buffer
    // size up to an SSD_PAGE_LEN boundary. Subtract one to get the
    // ending address and add that length to our starting address.
    temp_addr += ((((imageBufferSize + (FAS_FILESYSTEM::SSD_PAGE_LEN - 1)) / FAS_FILESYSTEM::SSD_PAGE_LEN) * FAS_FILESYSTEM::SSD_PAGE_LEN) - 1);

    // Make sure this will fit.
    if (temp_addr >= partition_size_avail)
    {
        // We're going to slop over, so drop the total number of available buffers.
        imageBufferCount--;
    }

    // Tell the FPGA how many we can capture.
    fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, imageBufferCount);

    // Make sure these are consistent with what we just laid out.
    sessionLength = sess_len;

    // Initialize the partition information.
    PART_MGR_WORKER::pmwInitMessage msg;
    msg.num_parts = num_parts;
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwInitMessage failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // All done.
    return true;
}

/*!
 * @brief Set the active partition to an unlocked one.
 *
 * @param part_num  The number of the partition starting at 0.
 *
 * @return true if everything good, otherwise false. if the partition is locked it will fail.
 */
bool PartMgr::setActivePartition(int part_num)
{
    bool rc(true);
    PART_MGR_WORKER::pmwSetActivePartMessage msg;

    TP(PM_FUNC, "%s: part_num:%d", __PRETTY_FUNCTION__, part_num);

    // Save the new active partition number.
    msg.active_index = part_num;
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: Unable to set the active partition!", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

void PartMgr::queueCapturePartition(int part_num)
{
    PART_MGR_WORKER::PartInfo part_info;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(PM_FUNC, "%s: part_num:%d", __PRETTY_FUNCTION__, part_num);

    // Get the information for the specified partiton.
    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo() failed for %d", __PRETTY_FUNCTION__, part_num);
        return;
    }

    // Now set the capture buffer based upon the partition sent in with the FPGA
    TP(PM_MEMORY, "%s: part_num:%d/0x%09" PRIx64, __PRETTY_FUNCTION__, part_num, part_info.capture_buf_start);
    fpga_ptr->icmSetCaptureBuffer(part_info.capture_buf_start);
}


bool PartMgr::setActiveFrameNum(uint32_t frame_num)
{
    TP(PM_FUNC, "%s: frame_num=%u, num_frames@%d=%u", __PRETTY_FUNCTION__, frame_num, getActivePartition(), getNumFrames());

    // Sanity check the parameter.
    if (frame_num >= getNumFrames())
    {
        fi_log(LOG_ERR, "%s: requested (%d) is too big (%d)!", __PRETTY_FUNCTION__, frame_num, getNumFrames());
        return false;
    }

    // Set the new active frame number.
    pfmdFrameNum = frame_num;
    return true;
}

/*!
 * @brief Set the state of the active partition and send a notification if it changed.
 *
 * part_state   The new partition state.
 */
void PartMgr::setPartitionState(int part_num, uint32_t part_state)
{
    PART_MGR_WORKER::PartInfo part_info;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();

    TP(PM_FUNC, "%s: part=%d, part_state=0x%08x", __PRETTY_FUNCTION__, part_num, part_state);

    // If we're in FasFireLR mode, and we've already triggered, then an ARMED partition is
    // really a TRIGGERED partition.
    if (fpga_ptr->ltrFasFireEnabled() && (trigger_part >= 0) && (part_state == GIGE_PART_STATE_ARMED))
        part_state = GIGE_PART_STATE_TRIGGERED;

    // Get the partition information.
    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo failed for (%d)", __PRETTY_FUNCTION__, part_num);
        return;
    }

    // Did things actually change?
    if (part_info.partition_state == part_state)
        return;

    // Update the state.
    part_info.partition_state = part_state;

    // Update the partition information.
    if (!updatePartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo failed for (%d)", __PRETTY_FUNCTION__, part_num);
        return;
    }

    // Are we erasing the partition?
    if (part_state == GIGE_PART_STATE_EMPTY)
    {
        // Check to see if we need to re-enable gain table support. It may have been disabled if
        // we loaded and older CAP that didn't support gain tables.
        if (!gainTableIsAllowed)
        {
            FiFPGA      *fpga_ptr = CameraServer::get_fpga();

            // Reserve space for the gain table.
            reserveGainTableSpace(true);

            // Make sure the partition manager layout matches our current ROI.
            layoutMemoryFromRegs();

            // Load the correction/gain table information.
            if (FiImageCorrection::fpnIsEnabled())
            {
                // Make sure the IPM is idle.
                if (!fpga_ptr->ipmIsIdle())
                {
                    // Reset the IPM.
                    fpga_ptr->ipmReset();
                }

                // Disable FPN, load the table(s), and then re-enable FPN.
                FiImageCorrection::fpnDisable();
                FiImageCorrection::fpnLoad();
                FiImageCorrection::fpnEnable();
            }
        }
    }

    // Let folks know we've changed the partition state.
    fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);
}

/*!
 * @brief Return an absolute frame number.
 *
 * If "relative", meaning absolute=false, the frame number passed in is relative to the
 * oldest frame in the partition, which can change since it's a circular buffer.
 * If it is "absolute", meaning absolute=true, it is relative to the physical start of
 * the partition; in that case, we just sanity check it and return what was passed.
 *
 * @param part_num  The part number we're interested in.
 * @param frame_num The frame number we're interested in.
 * @param absolute  TRUE if we passed the "absolute" frame_num, or FALSE if we passed a "relative" frame_num.
 *
 * @return The absolute frame number if everything is good, otherwise (uint32_t)(-1).
 */
uint32_t PartMgr::convertToAbsFrameNum(int part_num, uint32_t frame_num, bool absolute)
{
    uint32_t abs_frame;
    uint32_t allocated_frames;
    FiFPGA   *fpga_ptr;

    TP(PM_FUNC, "%s: part_num=%d, frame_num=%d, abs=%c", __PRETTY_FUNCTION__, part_num, frame_num, (absolute ? 'T' : 'F'));

    // Get our support class.
    fpga_ptr = CameraServer::get_fpga();

    // If we aborted the capture, the number of frames can be less than the allocated
    // frame count, but our absolute frame numbers are still modulo the allocated
    // frame count, not the number of frames we report to the user.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        allocated_frames = ltrGetActiveFrameCount();
    else
        allocated_frames = imageBufferCount;

    // Did we receive an "absolute" frame number?
    if (absolute)
    {
        // Just sanity check the frame number.
        if (frame_num < allocated_frames)
            abs_frame = frame_num;
        else
        {
            // We should never be asked to do this.
            fi_log(LOG_WARNING, "%s: part_num=%u absolute=T but frame_num=%u allocated_frames=%u",
                   __FUNCTION__, part_num, frame_num, allocated_frames);

            abs_frame = frame_num % allocated_frames;
        }
    }
    else
    {
        PART_MGR_WORKER::PartInfo part_info;

        // Retrieve the partition information.
        if (!getPartInfo(part_num, part_info))
        {
            fi_log(LOG_DEBUG, "%s: getPartInfo() failed for (%d)", __PRETTY_FUNCTION__, part_num);
            return 0;
        }

        // Check to see if the frame being requested goes beyond (i.e. "wraps")
        // the last frame.
        if ((frame_num + part_info.first_frame_num) < allocated_frames)
        {
            // Nope - we're good. Just add the frame number to our trigger frame
            // position.
            abs_frame = frame_num + part_info.first_frame_num;
        }
        else
        {
            // Yes it does. We need to subtract the total number of frames so
            // that the frame number will wrap around to the beginning again.
            abs_frame = (frame_num + part_info.first_frame_num) - allocated_frames;
        }

        TP(PM_FUNC, "%s: frame_num=%d, first_frame=%d, allocated_frames=%d abs_frame=%d", __PRETTY_FUNCTION__,
           frame_num, part_info.first_frame_num, allocated_frames, abs_frame);
    }

    return abs_frame;
}

/*!
 * @brief Return the address where a specific frame number starts.
 *    If "relative", the frame number passed in is relative to the oldest
 *    frame in the partition, which can change since it's a circular
 *    buffer. If it is "absolute", it is relative to the physical
 *    start of the partition.
 *
 * @param part_num  The part number we're interested in.
 * @param frame_num The frame number we're interested in.
 * @param absolute  TRUE if we passed the "absolute" frame_num,
 *          or FALSE for a "relative" frame_num.
 *
 * @return The frame address if everything is good, otherwise 0.
 *     In LTR mode, the addr is an LBA in sectors from drive start.
*      In normal mode, the addr is an address in image memory.
 */
uint64_t PartMgr::convertFrameNumToAddr(int part_num, uint32_t frame_num, bool absolute)
{
    uint64_t    addr = 0;
    int32_t abs_frame = 0;
    FiFPGA  *fpga_ptr;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s: part_num=%d, frame_num=%d, abs=%c", __PRETTY_FUNCTION__, part_num, frame_num, (absolute ? 'T' : 'F'));

    // Retrieve the partition information.
    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo() failed for (%d)", __PRETTY_FUNCTION__, part_num);
        return 0;
    }

    // Get our support class.
    fpga_ptr = CameraServer::get_fpga();

    // Convert to an absolute frame number so we can compute offset from partition start.
    abs_frame = convertToAbsFrameNum(part_num, frame_num, absolute);

    // Get the offset. We'll add the base next.
    addr = (((uint64_t)abs_frame) * ((uint64_t)part_info.frame_size));
    TP(PM_MEMORY, "%s: abs_frame(%u) * frame_size(%u) = addr(0x%" PRIx64 ")", __PRETTY_FUNCTION__, abs_frame, part_info.frame_size, addr);

    // The address is either an address in image memory or an offset on the SSD.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Sanity check.
        if (abs_frame > (int32_t)(ltrGetActiveFrameCount() - 1))
        {
            fi_log(LOG_DEBUG, "%s: abs_frame(%d) is too big (num_frames=%d), part_num=%d", __PRETTY_FUNCTION__, abs_frame, ltrGetActiveFrameCount(), part_num);
            abs_frame = 0;
        }

        // Now figure out the address, which is an LBA.
        addr += ltrGetSessionStart(part_num);
        TP(PM_MEMORY, "%s: Adding in sesson start addr: 0x%" PRIx64, __PRETTY_FUNCTION__, ltrGetSessionStart(part_num));
    }
    else
    {
        // Sanity check.
        if (abs_frame > (int32_t)(part_info.num_frames - 1))
        {
            fi_log(LOG_DEBUG, "%s: abs_frame(%d) is too big (num_frames=%d), part_num=%d", __PRETTY_FUNCTION__, abs_frame, part_info.num_frames, part_num);
            abs_frame = 0;
        }

        // Figure out the address.
        addr += part_info.capture_buf_start;
        TP(PM_FUNC, "%s: capture_buf_start=0x%" PRIx64, __PRETTY_FUNCTION__, part_info.capture_buf_start);
    }

    TP(PM_MEMORY, "%s: part:%d frame:%d abs_frame:%d first:%d addr:0x%" PRIx64,
       __PRETTY_FUNCTION__, part_num, frame_num, abs_frame, part_info.first_frame_num, addr);

    return addr;
}

/*!
 * @brief return the frame number which matches the address passed in. The address
 *    must be within bounds for this partition and the address must fall on a
 *    frame boundary.
 *
 * @param addr  The address that we want converted to a frame number.
 *      If LTR mode, the address is a sector number (LBA) relative to the start
 *      of the drive. If normal mode, it is aa address in image memory.
 *
 * @return  The frame number if everything is good, otherwise -1.
 *      The frame number is relative to the oldest frame in the
 *      partition, as stored in first_frame_num[].
 */
int32_t PartMgr::convertAddrToFrameNum(uint64_t frame_addr)
{
    int active_part;
    int32_t frame_num = 0;
    int32_t rel_frame_num = 0;
    uint64_t rel_addr = 0;
    uint32_t allocated_frames;
    uint32_t ltr_mode;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s: 0x%" PRIx64, __FUNCTION__, frame_addr);

    {
        FiFPGA  *fpga_ptr;

        // Get our support class and read the LTR mode setting.
        fpga_ptr = CameraServer::get_fpga();
        ltr_mode = fpga_ptr->ltrMode();
    }

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return -1;
    }

    // Sanity check.
    if (part_info.frame_size == 0)
    {
        fi_log(LOG_ERR, "%s: frame_size[ %d ] is 0!?", __PRETTY_FUNCTION__, active_part);
        return -1;
    }

    // Check to see if we're in LTR mode.
    //
    // If we aborted the capture, the number of frames can be less than the allocated
    // frame count, but our absolute frame numbers are still modulo the allocated
    // frame count, not the number of frames we report to the user.
    if ((ltr_mode & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        allocated_frames = ltrGetActiveFrameCount();

        // Figure out which frame this is.
        rel_addr = frame_addr - ltrGetSessionStart(active_part);
    }
    else
    {
        allocated_frames = imageBufferCount;
        rel_addr = frame_addr - part_info.capture_buf_start;
    }
    frame_num = (int32_t)(rel_addr / part_info.frame_size);

    rel_frame_num = frame_num - (int32_t)part_info.first_frame_num;
    if (rel_frame_num < 0)
    {
        rel_frame_num = (int32_t)allocated_frames + rel_frame_num;
    }

    TP(PM_MEMORY, "%s: finish part:%d frame:%d rel_frame:%d first:%d allocated_frames:%d rel_addr:0x%" PRIx64 " addr:0x%" PRIx64,
       __FUNCTION__,  active_part, frame_num, rel_frame_num, part_info.first_frame_num,
       allocated_frames, rel_addr, frame_addr);

    return rel_frame_num;
}

int32_t PartMgr::convertCaptureAddrToFrameNum(uint64_t frame_addr)
{
    int active_part;
    int32_t frame_num = 0;
    int32_t rel_frame_num = 0;
    uint64_t rel_addr = 0;
    uint32_t allocated_frames;
    uint32_t ltr_mode;
    uint32_t icm_frame_count;
    FiFPGA  *fpga_ptr;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return -1;
    }
    TP(PM_FUNC, "%s: 0x%" PRIx64 ", im_buf_count=%u, capture_buf_start=0x%" PRIx64, __PRETTY_FUNCTION__, frame_addr, imageBufferCount, part_info.capture_buf_start);

    // Sanity check.
    if (part_info.frame_size == 0)
    {
        fi_log(LOG_ERR, "%s: frame_size[ %d ] is 0!?", __PRETTY_FUNCTION__, active_part);
        return -1;
    }

    // Get our support class and read the LTR mode setting.
    fpga_ptr = CameraServer::get_fpga();
    ltr_mode = fpga_ptr->ltrMode();

    // How many frames (in total) has the ICM captured?
    fpga_ptr->regRead(ICM_CAPTURE_FRAME_COUNT_LIVE, &icm_frame_count);

    // Get the relative address of the frame (relative to the start).
    rel_addr = frame_addr - part_info.capture_buf_start;

    // Now get the frame number.
    frame_num = (int32_t)(rel_addr / part_info.frame_size);

    // Check to see if we're in LTR mode.
    if ((ltr_mode & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // How many frames are in the IM are we using for LR mode.
        allocated_frames = ltrImageBufferCount;
        if (allocated_frames == 0)
        {
            // Invalid value for the ltrImageBufferCount.
            fi_log(LOG_WARNING, "%s: ltrImageBufferCount is zero!?", __PRETTY_FUNCTION__);
            ltrSetCaptureBufferCount((uint32_t)PartMgr::MAX_IM2SSD_DESCRIPTOR_FIFO_DEPTH);
            allocated_frames =  ltrImageBufferCount;
        }

        // Adjust our frame number appropriately if we've wrapped image memory.
        rel_frame_num = frame_num;
        if (icm_frame_count > allocated_frames)
            rel_frame_num += (allocated_frames * (icm_frame_count / allocated_frames));
        TP(PM_MEMORY, "%s: frame_num=%u, icm_frame_count=%u, allocated=%u, rel_frame_num=%u",
           __PRETTY_FUNCTION__, frame_num, icm_frame_count, allocated_frames, rel_frame_num);

        // Now need to adjust based on the SSD "buffer" size.
        while (rel_frame_num >= (int32_t)ltrGetActiveFrameCount())
            rel_frame_num -= ltrGetActiveFrameCount();
    }
    else
    {
        // How many frames are there in the IM.
        allocated_frames = imageBufferCount;

        rel_frame_num = frame_num - (int32_t)part_info.first_frame_num;
        if (rel_frame_num < 0)
        {
            rel_frame_num = (int32_t)allocated_frames + rel_frame_num;
        }
    }

    TP(PM_MEMORY, "%s: finish part:%d frame:%d rel_frame:%d first:%d allocated_frames:%d rel_addr:0x%" PRIx64 " addr:0x%" PRIx64,
       __FUNCTION__,  active_part, frame_num, rel_frame_num, part_info.first_frame_num,
       allocated_frames, rel_addr, frame_addr);

    return rel_frame_num;
}

uint64_t PartMgr::getCaptureFrameAddr(uint32_t frame_num)
{
    int active_part;
    uint64_t frame_addr;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s: frame_num=%d", __PRETTY_FUNCTION__, frame_num);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Figure out the address.
    frame_addr = (part_info.capture_buf_start + (((uint64_t)frame_num) * part_info.frame_size));
    TP(PM_FUNC, "%s: IM frame addr=0x%" PRIx64, __PRETTY_FUNCTION__, frame_addr);
    return frame_addr;
}

/*!
* @brief Return the current (relative) frame number.
*
* The value returned will depend upon the state of the active partition.
* If the partition is unarmed or completed, '0' will be returned. If it's armed the frame
* number relative to the beginning of the partition will be returned. If it's triggered
* the frame number will be relative to the trigger position.
*
* @return The relative frame number if everything is good otherwise 0.
*/
int32_t PartMgr::getCurrentFrameNum()
{
    int active_part;
    uint32_t mode = 0;
    uint64_t addr = 0;
    int32_t frame_num = 0;
    uint32_t capbuf_count;
    uint32_t ltr_mode;
    FiFPGA  *fpga_ptr;
    PART_MGR_WORKER::PartInfo part_info;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(PM_VERBOSE, "%s", __PRETTY_FUNCTION__);

    // Get our support class and read the LTR mode setting.
    fpga_ptr = CameraServer::get_fpga();
    ltr_mode = fpga_ptr->ltrMode();

    // Get the current frame address from the FPGA. In LTR mode, this
    // will be the frame's LBA. Otherwise this is the image memory address.
    addr = fpga_ptr->icmCurrentFrameAddr();

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return 0;
    }

    // If we are not in an "actively" updating state, ARMED or (post) TRIGGERED,
    // then just return a 0.
    if ((part_info.partition_state != GIGE_PART_STATE_ARMED) &&
        (part_info.partition_state != GIGE_PART_STATE_TRIGGERED))
    {
        return 0;
    }

    // Get the current address from the sensor and make sure it's valid.
    if ((ltr_mode & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        uint64_t start_offset;

        mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

        // Until the first frame is written, the "last frame written" pointer will be zero.
        // We will return frame 0 unless we are appending, in which case we need to show where
        // the append region starts. Our get_ltr_roc_start() will return the right starting
        // point regardless of whether we are in ROC/BROC or Basic mode.
        start_offset = ltrGetSessionStart(active_part);
        if (addr < start_offset)
            addr = ltrGetRocStart();

        // Make sure the address makes sense.
        if (addr < start_offset)
        {
            fi_log(LOG_DEBUG, "%s: addr %" PRIx64 " < partition start %" PRIx64, __FUNCTION__, addr, start_offset);
            return 0;
        }

        // Now figure out the frame number of the trigger frame.
        frame_num = (addr - start_offset) / imageBufferSize;

        // Set the size of the capture buffer for use in later calculations.
        capbuf_count = ltrGetActiveFrameCount();
    }
    else if (fpga_ptr->ltrFasFireEnabled())
    {
        uint64_t total_count;
        uint64_t part_end;

        // Sanity check.
        if (imageBufferSize == 0)
        {
            fi_log(LOG_ERR, "%s: im_buf_size is 0?!", __PRETTY_FUNCTION__);
            return 0;
        }

        // Until we the first frame is captured, the "last frame written" pointer will be zero
        // so we will return frame 0.
        if (addr == 0)
            return 0;

        // Calculate the address of the last frame of the partition.
        part_end = part_info.capture_buf_start + ((part_info.num_frames - 1) * part_info.frame_size);

        // Check to see if the hardware is within the active partition.
        if ((addr < part_info.capture_buf_start) || (addr > part_end))
        {
            // Just peg the frame number on the size of the partition.
            frame_num = part_info.num_frames - 1;
        }
        else
        {
            // Calculate the absolute frame number from the beginning of the partition.
            frame_num = (uint32_t)((addr - part_info.capture_buf_start) / imageBufferSize);
        }

        // If we are in FasFireLR mode, we need to also check to see if we're in ROC/BROC mode.
        mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);

        // Set the size of the capture buffer for use in later calculations.
        capbuf_count = imgSaveWorker::getInstance()->getRecordingFrameCount();
        if (capbuf_count == 0)
        {
            fi_log(LOG_DEBUG, "%s: capbuf_count is zero.", __PRETTY_FUNCTION__);
            return 0;
        }

        // Adjust our frame number by the number of partitions already captured. Keep it
        // within the recording space.
        total_count = ((uint64_t)frame_num) + (finished_part_count * imageBufferCount);
        frame_num = (uint32_t)(total_count % capbuf_count);

        TP(PM_VERBOSE, "%s: frame_num: %u, capbuf_count: %u, part_num: %u, finished_part_count: %u, trigger_part: %d", __PRETTY_FUNCTION__, frame_num, capbuf_count, active_part, finished_part_count, trigger_part);
    }
    else
    {
        // Sanity check.
        if (imageBufferSize == 0)
        {
            fi_log(LOG_ERR, "%s: im_buf_size is 0?!", __PRETTY_FUNCTION__);
            return 0;
        }

        // Until we the first frame is captured, the "last frame written" pointer will be zero
        // so we will return frame 0.
        if (addr < part_info.capture_buf_start)
            return 0;

        // Calculate the absolute frame number from the beginning of the partition.
        frame_num = (uint32_t)((addr - part_info.capture_buf_start) / imageBufferSize);

        // Set the size of the capture buffer for use in later calculations.
        capbuf_count = imageBufferCount;

        TP(PM_VERBOSE, "%s: frame_num: %u, capbuf_count: %u, part_num: %u", __PRETTY_FUNCTION__, frame_num, capbuf_count, active_part);
    }

    // If the "last frame written" hasn't been updated yet, it could still point to a
    // previously-recorded partition. (Actually this is no longer true with the FPGA change
    // to allow setting the register before arming, but we'll do this for safety anyway.)
    if (frame_num >= (int32_t)capbuf_count)
        return 0;

    // Adjust the frame number based on the trigger position, unless we are in ROC/BROC mode.
    if ((part_info.partition_state == GIGE_PART_STATE_TRIGGERED) && (mode == 0))
    {
        int32_t trigger_frame;

        // For FasFireLR mode, we need to worry about WHERE in the capture space we are.
        if (fpga_ptr->ltrFasFireEnabled())
        {
            trigger_frame = fflr_trigger_frame;
        }
        else
        {
            // Use the trigger frame from the partition information.
            trigger_frame = part_info.trig_frame_num;
        }

        // Calculate the relative frame number.
        if (frame_num > trigger_frame)
        {
            TP(PM_VERBOSE, "%s: Triggered, basic mode, frame_num > trigger. Adjust frame_num (%d) - trig_frame_num (%d)", __PRETTY_FUNCTION__, frame_num, trigger_frame);
            frame_num = frame_num - trigger_frame;
        }
        else
        {
            TP(PM_VERBOSE, "%s: Triggered, basic mode, frame_num <= trigger. Adjust frame_num (%d) + (capbuf_count (%u) - trig_frame_num (%d))", __PRETTY_FUNCTION__, frame_num, capbuf_count, trigger_frame);
            frame_num = frame_num + (capbuf_count - trigger_frame);
        }
    }

    // Return the relative frame number.
    TP(PM_VERBOSE, "%s: frame_num=%d", __PRETTY_FUNCTION__, frame_num);
    return frame_num;
}

uint32_t PartMgr::getFrameCount()
{
    uint32_t frame_count;
    FiFPGA  *fpga_ptr;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        frame_count = ltrGetActiveFrameCount();
    else
        frame_count = imageBufferCount;

    // Return our frame count.
    return frame_count;
}

uint64_t PartMgr::ltrGetCaptureBufferAddr(uint32_t *rtn_buf_count)
{
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retrieve the partition information.
    if (!getPartInfo(0, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo failed for the active partition (0)", __PRETTY_FUNCTION__);
        return -1;
    }

    *rtn_buf_count = ltrImageBufferCount;
    return part_info.capture_buf_start;
}

uint32_t PartMgr::getNumPartitions()
{
    return (uint32_t)PART_MGR_WORKER::pmWorker::factory()->getNumParts();
}

bool PartMgr::isReviewPartWrapped()
{
    int review_part;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retrieve the partition information.
    review_part = getReviewPartition();
    if (!getPartInfo(review_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo failed for the review partition (%d)", __PRETTY_FUNCTION__, review_part);
        return -1;
    }

    // We can safely assume the partition wrapped if the frame number for the first frame
    // is non-zero. Of course, there is the posibility that the partition did wrap in such
    // a way as to make the frame number zero again, but that ends up being the same as
    // not wrapping (do the math).
    return (part_info.first_frame_num != 0);
}

uint32_t PartMgr::getReviewPartNumFrames()
{
    PART_MGR_WORKER::PartInfo part_info;
    uint32_t num_frames = 0;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure the request is valid.
    if (getReviewPartition(&part_info) >= 0)
        num_frames = part_info.num_frames;

    return num_frames;
}

uint64_t PartMgr::getReviewPartStartAddr()
{
    PART_MGR_WORKER::PartInfo part_info;
    uint64_t part_addr = 0;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure the request is valid.
    if (getReviewPartition(&part_info) >= 0)
        part_addr = part_info.partition_start;

    return part_addr;
}

uint32_t PartMgr::getReviewPartTrigger()
{
    PART_MGR_WORKER::PartInfo part_info;
    uint32_t trig_frame = 0;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure the request is valid.
    if (getReviewPartition(&part_info) >= 0)
        trig_frame = part_info.trig_frame_num;

    return trig_frame;
}

bool PartMgr::getPartInfo(int part_index, PART_MGR_WORKER::PartInfo& part_info)
{
    bool rc(true);
    PART_MGR_WORKER::pmwGetInfoMessage  msg;

    TP(PM_FUNC, "%s: part_index=%d", __PRETTY_FUNCTION__, part_index);

    // Build up our message.
    msg.part_index = part_index;

    // Send it off.
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwGetInfoMessage failed!", __PRETTY_FUNCTION__);
        rc = false;
    }
    else
    {
        // Copy the data over.
        memcpy(&part_info, &msg.info, sizeof(part_info));
    }

    // Return our result.
    return rc;
}

bool PartMgr::updatePartInfo(int part_index, const PART_MGR_WORKER::PartInfo& part_info)
{
    bool rc(true);
    PART_MGR_WORKER::pmwSetInfoMessage  msg;

    TP(PM_FUNC, "%s: part_index=%d", __PRETTY_FUNCTION__, part_index);

    // Build up our message.
    msg.part_index = part_index;
    memcpy(&msg.info, &part_info, sizeof(msg.info));

    // Send it off.
    msg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (msg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwSetInfoMessage failed!", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

uint32_t PartMgr::getPartitionNumFrames(int part_num)
{
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo() failed for (%d)", __PRETTY_FUNCTION__, part_num);
        return 0;
    }

    return part_info.num_frames;
}

uint32_t PartMgr::getPmSize()
{
    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);
    return PARTITION_METADATA_SIZE;
}

/*!
 * @brief Determine the number of bytes are actually used in the partition.
 *
 * The number of bytes used may be less than the reserved size of the partition. If the
 * capture did not wrap, then it is likely the actual amount of data is less than the
 * reserved size of the partition.
 *
 * @param capture_session   [OPTIONAL] The size of the capture, in SESSION_INCR units.
 * @return The number of bytes actually used in the partition.
 */
uint64_t PartMgr::getReviewPartLength(uint32_t *capture_session)
{
    uint32_t    calculated_len;
    uint64_t    image_length;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Assume that we are using the entire partition.
    image_length = getPartitionSize();

    // Can we shorten the length? Check to see if the partition has wrapped.
    if (!isReviewPartWrapped())
    {
        uint32_t    frame_count;
        uint64_t    capture_len;

        // If the partition didn't wrap, then maybe we can shorten things up a bit.
        // Get the number of frames and figure out how much memory we ACTUALLY
        // consumed in this capture.
        frame_count = getReviewPartNumFrames();
        capture_len = (uint64_t)frame_count * (uint64_t)getFrameSize();

        // Need to include partition overhead (metadata, etc.)
        capture_len += getPartitionSizeReserved();

        // Convert the capture length to a session length (SESSION_INCR units) and determine if
        // we can save some time...
        calculated_len = (uint32_t)(capture_len / SESSION_INCR);
        if (((uint64_t)calculated_len * (uint64_t)SESSION_INCR) < capture_len)
            calculated_len++;

        TP(PM_FUNC, "%s: [NO WRAP] frame_count=%d, calculated_len=%d, session_length=%d", __PRETTY_FUNCTION__, frame_count, calculated_len, getSessionLength());
        if (calculated_len < getSessionLength())
        {
            // Ok...our capture session is less then what the user set, so
            // adjust our length.
            image_length = ((uint64_t)calculated_len) * ((uint64_t)SESSION_INCR);
            TP(PM_FUNC, "%s: [NO WRAP] adjusted image_length=%lld", __PRETTY_FUNCTION__, image_length);
        }
    }
    else
    {
        // The session size is the "normal" size.
        calculated_len = getSessionLength();
        TP(PM_FUNC, "%s: [WRAP] calculated_len=%d", __PRETTY_FUNCTION__, calculated_len);
    }

    // Does the caller want the session length returned?
    if (capture_session != nullptr)
        *capture_session = calculated_len;

    TP(PM_FUNC, "%s: review_part_num=%d, calculated_len=%u, length=%" PRIu64, __PRETTY_FUNCTION__, getReviewPartition(), calculated_len, image_length);

    // Return the length of data actually used in the partition.
    return image_length;
}

/*!
 * @brief Initialize a partition metadata structure from sensor and registers
 *
 * @param meta_ptr          [IN] Pointer to the pm_metadata structure to initialize.
 * @param trigger_setting   [IN] The trig_set value to write into metadata.
 * @param trigger_frames    [IN] The trigger_frames value to write into metadata.
 */
void PartMgr::partitionMetadataInit(pm_metadata *meta_ptr, uint8_t trigger_setting, uint32_t trigger_frames)
{
    // Get our support classes pointer.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer *cs_ptr = CameraServer::get_cs();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    csWorker *p_worker = CameraServer::get_worker();

    uint16_t start_x;
    uint16_t start_y;
    uint16_t width;
    uint16_t height;
    float frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
    bitmode_t bit_mode;
    fpnmode_t fpn_mode;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Cannot initialie a nullptr pm_metadata!", __PRETTY_FUNCTION__);
        return;
    }

    // Gather the information we will need.
    cs_ptr->get_user_settings(&start_x, &start_y, &width, &height, &bit_mode, &frame_rate,
                              &shutter_speed, &sensor_opts);
    cs_ptr->get_user_fpn(&fpn_mode);

    // Set all the metadata we know about
    memset(meta_ptr, 0, sizeof(pm_metadata));

    fcp_get_manufacture_name(meta_ptr->man_name, GIGE_DSC_ACK_VENDOR_B_LEN);
    fcp_get_model_name(meta_ptr->model_name, GIGE_DSC_ACK_MODEL_B_LEN);
    fcp_get_serial_number(meta_ptr->serial_num, GIGE_DSC_ACK_SERIAL_B_LEN);
    fcp_get_device_version_name(meta_ptr->app_ver, GIGE_DSC_ACK_VERSION_B_LEN);
    CsNetwork::getHostname(meta_ptr->camera_name, GIGE_DSC_ACK_USER_NAME_B_LEN);
    p_worker->getMacAddress(meta_ptr->mac_addr);
    meta_ptr->fpga_rev = CameraServer::get_fpga()->getFpgaVersion();

    meta_ptr->roi_x = start_x;
    meta_ptr->roi_y = start_y;
    meta_ptr->roi_width = width;
    meta_ptr->roi_height = height;
    meta_ptr->bit_sel = bit_mode;
    meta_ptr->fps = cs_ptr->frame_rate_to_reg(frame_rate);
    meta_ptr->shutter_speed = fpga_ptr->icmShutterSpeed();

    meta_ptr->sensor_gain = regs[ GIGE_REG_SENSOR_GAIN ].get_value();
    meta_ptr->sensor_type = sensor_ptr->get_sensor_type();
    meta_ptr->sensor_dac = regs[ GIGE_REG_SENSOR_DAC ].get_value();
    meta_ptr->sensor_vrefcol = regs[ GIGE_REG_SENSOR_VREFCOL ].get_value();
    meta_ptr->sensor_opts = sensor_opts;
    meta_ptr->fpn_status = fpn_mode;
    meta_ptr->black_offset = 0;
    if ((fpn_mode & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        meta_ptr->black_offset = FiImageCorrection::fpnGetOffset();

    // Copy the trigger position setting. If frames were requested, mark the
    // trigger setting field as such and put the requested frame count in the
    // 32-bit field. If a percentage was requested, put it in the 8-bit field;
    // In this case, the 32-bit field is officially "undefined" but we can still
    // put our computed post-trigger frame count in it.
    meta_ptr->trig_set = trigger_setting;
    meta_ptr->trig_set_frames = trigger_frames;

    memcpy(meta_ptr->comment, capture_comment, sizeof(capture_comment));
    memcpy(meta_ptr->lut_knots, knots, sizeof(knots));
    meta_ptr->lut_id = regs[ GIGE_REG_LUT ].get_value();

    // Update the metadata color information.
    meta_ptr->matrix_coeff1_red = fpga_ptr->ipmMatrixCoeff1Red();
    meta_ptr->matrix_coeff2_red = fpga_ptr->ipmMatrixCoeff2Red();
    meta_ptr->matrix_coeff3_red = fpga_ptr->ipmMatrixCoeff3Red();
    meta_ptr->matrix_coeff1_grn = fpga_ptr->ipmMatrixCoeff1Green();
    meta_ptr->matrix_coeff2_grn = fpga_ptr->ipmMatrixCoeff2Green();
    meta_ptr->matrix_coeff3_grn = fpga_ptr->ipmMatrixCoeff3Green();
    meta_ptr->matrix_coeff1_blu = fpga_ptr->ipmMatrixCoeff1Blue();
    meta_ptr->matrix_coeff2_blu = fpga_ptr->ipmMatrixCoeff2Blue();
    meta_ptr->matrix_coeff3_blu = fpga_ptr->ipmMatrixCoeff3Blue();
    meta_ptr->dc_offset_red = fpga_ptr->ipmDcOffsetRed();
    meta_ptr->dc_offset_grn = fpga_ptr->ipmDcOffsetGreen();
    meta_ptr->dc_offset_blu = fpga_ptr->ipmDcOffsetBlue();

    meta_ptr->wb_red = fpga_ptr->ipmWbRed();
    meta_ptr->wb_grn = fpga_ptr->ipmWbGreen();
    meta_ptr->wb_blu = fpga_ptr->ipmWbBlue();
    meta_ptr->brightness = fpga_ptr->ipmBrightness();
    meta_ptr->contrast = fpga_ptr->ipmContrast();
    meta_ptr->gamma = fpga_ptr->ipmGamma();
    meta_ptr->red = fpga_ptr->ipmGainRed();
    meta_ptr->green = fpga_ptr->ipmGainGreen();
    meta_ptr->blue = fpga_ptr->ipmGainBlue();
    meta_ptr->as_shot_red = cs_ptr->get_as_shot_red();
    meta_ptr->as_shot_grn = cs_ptr->get_as_shot_grn();
    meta_ptr->as_shot_blu = cs_ptr->get_as_shot_blu();
    meta_ptr->frame_size = imageBufferSize;

    // The correction table follows the partition metadata, and the gain table
    // follows the correction table.
    meta_ptr->corr_off = PARTITION_METADATA_SIZE;
    meta_ptr->corr_len = correctionTableSize;

    // Is there a gain table?
    if (!gainTableIsAllowed)
    {
        // No gain table.
        meta_ptr->gain_off = 0;
        meta_ptr->gain_len = 0;

        // Start of video is after the correction table.
        meta_ptr->sov_off = meta_ptr->corr_off + meta_ptr->corr_len;
    }
    else
    {
        // Add in the gain table.
        meta_ptr->gain_off = meta_ptr->corr_off + meta_ptr->corr_len;
        meta_ptr->gain_len = meta_ptr->corr_len;

        // Start of video is after the gain table.
        meta_ptr->sov_off = meta_ptr->gain_off + meta_ptr->gain_len;
    }
}

/*!
 * @brief Have the FPGA pull the per-frame metadata (PFMD) from the capture buffer
 *    in image memory. Go from "start_frame" to "end_frame", inclusive, or until
 *    MAX_METADATA_XFER_LEN bytes.
 *
 * @return Number of PFMD buffers (i.e. frames) have been moved into pfmd_buffer.
 */
int PartMgr::pme_fpga_fill_buffer(uint32_t start_frame, uint32_t end_frame, int bb_index)
{
    int rc;
    int total_pfmd = 0;     // Number of frames/PFMD blocks we are moving.
    uint64_t pme_buf;
    uint32_t pme_buf_size;
    uint32_t bytes_written;
    FiFPGA *fpga_ptr;

    TP(PM_FUNC, "%s: part=%d, start_frame=%u, end_frame=%u", __PRETTY_FUNCTION__, getReviewPartition(), start_frame, end_frame);

    // Sanity check.
    if (start_frame == end_frame)
    {
        fi_log(LOG_DEBUG, "%s: start_frame (%d) == end_frame!", __PRETTY_FUNCTION__, start_frame);
        return 0;
    }
    if (bb_index < 0)
    {
        fi_log(LOG_DEBUG, "%s: bb_index (%d) is invalid!", __PRETTY_FUNCTION__, bb_index);
        return 0;
    }

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Get our xfer buffer in image memory.
    pme_buf = get_pme_buf_addr();
    pme_buf_size = get_pme_buf_size();

    // Limit the size we transmit in each chunk.
    if (pme_buf_size > MAX_METADATA_XFER_LEN)
        pme_buf_size = MAX_METADATA_XFER_LEN;

    TP(PM_PFMD, "%s: pme_buf=0x%" PRIx64 ", size=%u", __PRETTY_FUNCTION__, pme_buf, pme_buf_size);

    // Ask the FPGA to get the metadata.
    bytes_written = fpga_ptr->pmeStart(getReviewPartition(), start_frame, end_frame, convert_to_HEXWORD(pme_buf), pme_buf_size);
    if (bytes_written == 0)
    {
        fi_log(LOG_ERR, "%s: pme_start() failed!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Wait for the FPGA to finish it up.
    total_pfmd = bytes_written / FRAME_METADATA_HW_BYTES;
    if (fpga_ptr->pmeWaitDone(total_pfmd, nullptr, nullptr) != 0)
    {
        // It didn't work. If it timed out, the error message already printed,
        // but we might have gotten a cancel.
        fpga_ptr->pmeAbort();
        fi_log(LOG_ERR, "%s: pme_wait_done() failed!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Have the FPGA transfer the buffer into our memory. Note that with the
    // PFMD, we assume 32-bit access, so the BYTE order is OK at L2R.
    rc = fpga_ptr->imageDownload(IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_L2R | IM2PCIE_DMA_WORD_ORDER_R2L, 0,
                                 pme_buf, bb_get_avail_addr(bb_index), bytes_written);
    if (rc < 0)
    {
        fi_log(LOG_ERR, "%s: Unable to DMA PFMD buffer!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Return the number of frames of PFMD we've read.
    TP(PM_PFMD, "%s: done, total_pfmd=%u", __PRETTY_FUNCTION__, total_pfmd);
    return total_pfmd;
}

/*!
 * @brief Pull the per-frame metadata (PFMD) from the SSD. Go from "start_frame"
 *    to "end_frame", inclusive, or until MAX_METADATA_XFER_LEN bytes.
 *
 * @return Number of PFMD buffers (i.e. frames) have been moved into pfmd_buffer.
 */
int PartMgr::pme_ltr_fill_buffer(uint32_t start_frame, uint32_t end_frame, uint8_t *pfmd_buffer)
{
    int total_pfmd = 0;     // Number of frames/PFMD blocks we are moving.
    uint32_t bytes_written;

    TP(PM_LTR, "%s: start_frame=%u, end_frame=%u", __PRETTY_FUNCTION__, start_frame, end_frame);

    // Sanity check.
    if (start_frame == end_frame)
    {
        fi_log(LOG_DEBUG, "%s: start_frame (%d) == end_frame!", __PRETTY_FUNCTION__, start_frame);
        return 0;
    }
    if (pfmd_buffer == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: pfmd_buffer is NULL!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Calculate the total number of blocks being request. Adjust as necessary to
    // fit in the space available. The start_/end_frame numbers are inclusive.
    total_pfmd = (int)((end_frame - start_frame) + 1);
    if ((total_pfmd * FRAME_METADATA_HW_BYTES) > (int)(MAX_METADATA_XFER_LEN / 16))
    {
        // Limit the number of frames of PFMD to our buffer size.
        total_pfmd = ((MAX_METADATA_XFER_LEN / 16) / FRAME_METADATA_HW_BYTES) - 1;
    }

    // Read in the metadata. Make sure we get the data in hexwords. While wasteful, the FPGA
    // is limited to hexwords. By matching the FPGA's output here, the user won't have to
    // have a special case (LTR vs. image memory).
    bytes_written = ltrReadPFMD(getActivePartition(), start_frame,
                                start_frame + total_pfmd - 1, pfmd_buffer, MAX_METADATA_XFER_LEN, true);
    if (bytes_written == 0)
    {
        fi_log(LOG_ERR, "%s: Unable to read the PFMD!", __PRETTY_FUNCTION__);
        total_pfmd = 0;
    }

    // Return the number of frames of PFMD we've read.
    TP(PM_LTR, "%s: done, total_pfmd=%u", __PRETTY_FUNCTION__, total_pfmd);
    return total_pfmd;
}


/*!
 * @brief Start a capture, which could be for calibration or a "real" recording.
 *
 * If it is for calibration, we will end with dearm().
 *
 * @return true if everything good, otherwise false.
 */
bool PartMgr::arm()
{
    int active_part;
    bool rc = false;
    bool ltr_mode;
    uint32_t buf_count;
    uint32_t recording_id;
    uint32_t mode = 0;   // ROC or BROC or 0
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition failed.", __PRETTY_FUNCTION__);
        goto arm_bail;
    }
    TP(PM_MEMORY, "%s: active_partition=%d", __PRETTY_FUNCTION__, active_part);

    // In LTR mode, we support ROC/BROC as well.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    if (ltr_mode)
    {
        // Figure out the mode.
        mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
        if ((mode == 0) || (part_info.partition_state == GIGE_PART_STATE_EMPTY))
        {
            // Use the current time as the recording ID.
            recording_id = time(0);
        }
        else
        {
            // If appending, keep using the same recording ID.
            recording_id = getLtrRecordingID();
        }
    }
    else
    {
        // Use the current time as the recording ID.
        recording_id = time(0);
        mode = 0;
    }

    // Set a tag into the PFMD.
    //
    // NOTE: In FasFire, it can take us a frame or two to get this set.
    //       Meanwhile, the FPGA has been using the old value. Really,
    //       this SHOULDN'T be a problem as the tag is only used to determine
    //       the boundary of an LR basic mode recording.
    fpga_ptr->icmSetMetadataTag(recording_id);

    // Handle "basic" mode (LR and IM) set up. This is also the setup for FasFireLR ROC/BROC.
    if (fpga_ptr->ltrFasFireEnabled() || (mode == 0))
    {
        uint8_t pm_trig_set;
        uint32_t pm_trig_frames;
        uint32_t post_trigger_frames;

        // Are we in FasFireLR Mode?
        if (fpga_ptr->ltrFasFireEnabled())
        {
            // The buffer count is the number of frames in the entire recording,
            // not simply the number of frames per capture/partition.
            buf_count = imgSaveWorker::getInstance()->getRecordingFrameCount();
            post_trigger_frames = computePostTrigCount(buf_count, triggerReqPos);

            // Tell the FPGA how many frames are to be recorded after the trigger event.
            fpga_ptr->icmSetPostTriggerFrames(post_trigger_frames);

            // Set the trigger frame number based on the total recording length.
            part_info.trig_frame_num = buf_count - post_trigger_frames;
            part_info.first_frame_num = 0;

            // Update the metadata information for the partition based on what we know at this time.
            part_info.num_frames = imageBufferCount;

            // Are we ROC or BROC?
            if (mode)
            {
                // No post trigger frames in this mode.
                fpga_ptr->icmSetPostTriggerFrames(0);

                // Setup for the partition metadata initialization.
                if (mode == GIGE_LTR_MODE_BROC)
                {
                    pm_trig_frames = regs[ GIGE_REG_BROC_LEN ].get_value();
                    pm_trig_set = 0x7E;
                }
                else /* Must be ROC */
                {
                    pm_trig_frames = 0;
                    pm_trig_set = 0x7F;
                }
            }
            else
            {
                // Basic mode recording.
                if (triggerReqPos & GIGE_TRIGGER_REQ_FRAME_MASK)
                {
                    pm_trig_set = 0x7F;
                    pm_trig_frames = triggerReqPos & GIGE_TRIGGER_REQ_POS_MASK;
                }
                else
                {
                    pm_trig_set = (uint8_t)triggerReqPos; // It is 0-100.
                    pm_trig_frames = post_trigger_frames;
                }
            }
        }
        else
        {
            // Are we in LR mode?
            if (ltr_mode)
            {
                // Layout the capture based on current settings. This may change some things, so
                // we need to do it early on.
                if (!ltrArmSetup(recording_id, mode))
                {
                    fi_log(LOG_DEBUG, "%s: ltrArmSetup() failed!", __PRETTY_FUNCTION__);
                    goto arm_bail;
                }

                // Based on the current layout, how many images can we save?
                buf_count = ltrGetActiveFrameCount();
            }
            else
            {
                // Normal image memory capture.
                buf_count = imageBufferCount;
            }

            // Calculate the number of frames after the trigger event and tell the
            // FPGA that information.
            post_trigger_frames = computePostTrigCount(buf_count, triggerReqPos);
            fpga_ptr->icmSetPostTriggerFrames(post_trigger_frames);

            // Update the metadata information.
            part_info.num_frames = buf_count;
            part_info.trig_frame_num = buf_count - post_trigger_frames;
            part_info.first_frame_num = 0;

            if (triggerReqPos & GIGE_TRIGGER_REQ_FRAME_MASK)
            {
                pm_trig_set = 0x7F;
                pm_trig_frames = triggerReqPos & GIGE_TRIGGER_REQ_POS_MASK;
            }
            else
            {
                pm_trig_set = (uint8_t)triggerReqPos; // It is 0-100.
                pm_trig_frames = post_trigger_frames;
            }
        }

        // Initialize the partition metadata.
        partitionMetadataInit(&part_info.metadata, pm_trig_set, pm_trig_frames);
    }

    // Handle normal LR mode ROC/BROC setup for the first recording in the capture.
    else if (part_info.partition_state == GIGE_PART_STATE_EMPTY)
    {
        // In LR mode, ROC or BROC at this point. Setup the capture space.
        if (!ltrArmSetup(recording_id, mode))
        {
            fi_log(LOG_DEBUG, "%s: ltrArmSetup() failed!", __PRETTY_FUNCTION__);
            goto arm_bail;
        }

        // We are only setting mode!=0 for LTR mode, not handling ROC/BROC to IM yet.
        buf_count = ltrGetActiveFrameCount();

        // Update the metadata information.
        part_info.num_frames = buf_count;
        part_info.trig_frame_num = 0;
        part_info.first_frame_num = 0;

        // No post trigger frames in this mode.
        fpga_ptr->icmSetPostTriggerFrames(0);

        // This will set the time to all zeroes to indicate the trigger hasn't been stored yet.
        if (mode == GIGE_LTR_MODE_BROC)
        {
            uint32_t burst_len = regs[ GIGE_REG_BROC_LEN ].get_value();
            partitionMetadataInit(&part_info.metadata, 0x7E, burst_len);
        }
        else
        {
            partitionMetadataInit(&part_info.metadata, 0x7F, 0);
        }
    }

    // For ROC/BROC, when appending make sure there is still space left.
    else if (ltrIsRocFull())
    {
        fi_log(LOG_WARNING, "%s: Drive is full", __FUNCTION__);
        goto arm_bail;
    }

    // Handle ROC/BROC setup for an appended recording.
    else
    {
        uint32_t trig_time_s;
        uint32_t trig_time_ns;

        // Update the metadata information.

        // The num_frames value here is the number of frames that can fit in the capture.
        // We will update it on capture completion to the number actually stored.
        buf_count = ltrGetActiveFrameCount();
        part_info.num_frames = buf_count;
        part_info.first_frame_num = 0;

        // No post trigger frames in this mode.
        fpga_ptr->icmSetPostTriggerFrames(0);

        // Initializing the metadata will zero the trigger time, but we need
        // to save it. The time for the recording does not change when appending.
        // Also retain the number of frames already captured. There are other ways
        // to do it but this is simple since we're already doing something similar.
        trig_time_s = part_info.metadata.trig_time_s;
        trig_time_ns = part_info.metadata.trig_time_ns;
        buf_count = part_info.metadata.frame_count;

        if (mode == GIGE_LTR_MODE_BROC)
        {
            uint32_t burst_len = regs[GIGE_REG_BROC_LEN].get_value();
            partitionMetadataInit(&part_info.metadata, 0x7E, burst_len);
        }
        else
        {
            partitionMetadataInit(&part_info.metadata, 0x7F, 0);
        }

        part_info.metadata.trig_time_s = trig_time_s;
        part_info.metadata.trig_time_ns = trig_time_ns;
        part_info.metadata.frame_count = buf_count;

        // We need to update the active partitions information.
        if (!ltrRocArmAppend())
        {
            fi_log(LOG_DEBUG, "%s: ltrRocArmAppend() failed!", __PRETTY_FUNCTION__);
            goto arm_bail;
        }
    }

    // Update the active partition's metadata.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for part %d", __PRETTY_FUNCTION__, active_part);
        goto arm_bail;
    }

    // Update our partition state.
    setPartitionState(active_part, GIGE_PART_STATE_ARMED);

    // All is well.
    rc = true;

arm_bail:
    // Return our status.
    return rc;
}

/*!
 * @brief To "update metadata" means that the current FPGA color processing settings
 *        are updated in the pm_metadata structure and copied to image memory.
 *
 * If the optional meta_ptr parameter is used, we will overwrite the existing partition's
 * metadata structure with the values the meta_ptr parameters references. Values are then
 * updated based on settings in the FPGA's register space.
 *
 * @param part_num      [IN] The partition we want to update.
 * @param meta_ptr      [IN, OPTIONAL] A pointer to partition metadata we want to use.
 *
 * @return True if everything was written sucessfully, otherwise false.
 */
bool PartMgr::updateMetadataSettings(int part_num, const pm_metadata *meta_ptr)
{
    bool rc(false);
    pm_metadata *pm_ptr;

    // Get our support class pointer(s).
    CameraServer    *cs_ptr = CameraServer::get_cs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the information associated with this partition number.
    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: pmwGetInfoMessage failed!", __PRETTY_FUNCTION__);
        goto update_bail;
    }
    pm_ptr = &part_info.metadata;

    // Copy the passed in information.
    if (meta_ptr != nullptr)
        memcpy(pm_ptr, meta_ptr, sizeof(pm_metadata));

    pm_ptr->matrix_coeff1_red = fpga_ptr->ipmMatrixCoeff1Red();
    pm_ptr->matrix_coeff2_red = fpga_ptr->ipmMatrixCoeff2Red();
    pm_ptr->matrix_coeff3_red = fpga_ptr->ipmMatrixCoeff3Red();
    pm_ptr->matrix_coeff1_grn = fpga_ptr->ipmMatrixCoeff1Green();
    pm_ptr->matrix_coeff2_grn = fpga_ptr->ipmMatrixCoeff2Green();
    pm_ptr->matrix_coeff3_grn = fpga_ptr->ipmMatrixCoeff3Green();
    pm_ptr->matrix_coeff1_blu = fpga_ptr->ipmMatrixCoeff1Blue();
    pm_ptr->matrix_coeff2_blu = fpga_ptr->ipmMatrixCoeff2Blue();
    pm_ptr->matrix_coeff3_blu = fpga_ptr->ipmMatrixCoeff3Blue();
    pm_ptr->white_bal = regs[GIGE_REG_WHITE_BAL].get_value();
    pm_ptr->wb_red = fpga_ptr->ipmWbRed();
    pm_ptr->wb_grn = fpga_ptr->ipmWbGreen();
    pm_ptr->wb_blu = fpga_ptr->ipmWbBlue();
    pm_ptr->brightness = fpga_ptr->ipmBrightness();
    pm_ptr->contrast = fpga_ptr->ipmContrast();
    pm_ptr->gamma = fpga_ptr->ipmGamma();
    pm_ptr->red = fpga_ptr->ipmGainRed();
    pm_ptr->green = fpga_ptr->ipmGainGreen();
    pm_ptr->blue = fpga_ptr->ipmGainBlue();
    pm_ptr->as_shot_red = cs_ptr->get_as_shot_red();
    pm_ptr->as_shot_grn = cs_ptr->get_as_shot_grn();
    pm_ptr->as_shot_blu = cs_ptr->get_as_shot_blu();

    // Make sure it has the latest comment.
    memcpy(pm_ptr->comment, capture_comment, sizeof(capture_comment));

    // Update the LUT information.
    memcpy(pm_ptr->lut_knots, knots, sizeof(knots));
    pm_ptr->lut_id = regs[GIGE_REG_LUT].get_value();

#if 0
    {
        //---------------------------------------------------------------------------
        int fd;
        std::string meta_path = fmt::sprintf("/tmp/pm_meta_orig_%d.dat", part_num);
        fd = open(meta_path.c_str(), O_CREAT | O_RDWR | O_TRUNC, 0666);
        write(fd, (uint8_t *)pm_ptr, sizeof(struct pm_metadata));
        close(fd);
    }   //---------------------------------------------------------------------------
#endif

    // Save the information back in the worker thread.
    if (!updatePartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d!", __PRETTY_FUNCTION__, part_num);
        goto update_bail;
    }

    // Write the data to the image memory
    if (fpga_ptr->imageMemWrite(part_info.partition_start, (uint8_t *)pm_ptr, sizeof(pm_metadata), false) <= 0)
    {
        fi_log(LOG_ERR, "%s: imageMemWrite failed!", __PRETTY_FUNCTION__);
        goto update_bail;
    }

#if 0
    {
        //---------------------------------------------------------------------------
        int fd;
        uint32_t dma_mod;
        pm_metadata dnload_meta;

        // We are going to transfer the data raw - no formatting.
        dma_mod = IM2PCIE_DMA_GENERIC_DATA;
        dma_mod |= IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
        dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT;

        fpga_ptr->read_image_mem(dma_mod, 0, part_info[part_num].partition_start, (uint8_t *)&dnload_meta, sizeof(dnload_meta));

        fd = open("/tmp/pm_meta_dnload.dat", O_CREAT | O_RDWR | O_TRUNC, 0666);
        write(fd, (uint8_t *)&dnload_meta, sizeof(dnload_meta));
        close(fd);
    }   //---------------------------------------------------------------------------
#endif

    // All is well.
    rc = true;

update_bail:

    // Return our result.
    return rc;
}

/*!
 * @brief Clean up partition metadata based on the app_ver.
 *
 * If we loaded metadata in a CAP file from a previous version of the firmware,
 * some of the fields might not be initialized as expected. This will clean up
 * a pm_metadata structure so any new fields have sensible values.
 *
 * @param meta_ptr  Pointer to the metadata structure.
 */
void PartMgr::cleanup_metadata_version(pm_metadata *meta_ptr)
{
    unsigned major;
    unsigned minor;
    unsigned patch;
    uint32_t version;
    FiFPGA  *fpga_ptr;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No meta_ptr.", __PRETTY_FUNCTION__);
        return;
    }
    if (meta_ptr->fpga_rev != fpga_ptr->getFpgaVersion())
    {
        fi_log(LOG_DEBUG, "%s: FPGA version mismatch, assuming bad data.", __PRETTY_FUNCTION__);

        // Clear the structure.
        memset(meta_ptr, 0, sizeof(pm_metadata));
        return;
    }

    if (sscanf(meta_ptr->app_ver, "%u.%u.%u", &major, &minor, &patch) != 3)
    {
        fi_log(LOG_ERR, "%s: unexpected version %.31s", __FUNCTION__, meta_ptr->app_ver);

        // Force "unknown" to be older than the current version.
        version = 1;
    }
    else if (((major == 0) && (minor == 0) && (patch == 0)) ||
             ((major == 255) && (minor == 255) && (patch == 255)))
    {
        // Development version.
        version = CURRENT_FIRMWARE_VERSION;
        fi_log(LOG_DEBUG, "%s: CAP version %u.%u.%u, assuming CURRENT (0x%08x)", __FUNCTION__,
               major, minor, patch, CURRENT_FIRMWARE_VERSION);
    }
    else
    {
        version = FIRMWARE_VERSION(major, minor, patch);
    }

    if (version != CURRENT_FIRMWARE_VERSION)
        fi_log(LOG_DEBUG, "%s: Restoring from CAP version %u.%u.%u CURRENT=0x%08x", __FUNCTION__,
               major, minor, patch, CURRENT_FIRMWARE_VERSION);

    // Handle a CAP file from version 1.7.5 or earlier.
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(1, 7, 5)))
    {
        memset(((uint8_t *)meta_ptr) + GIGE_PM_META_SIZE_1_7_5, 0,
               sizeof(pm_metadata) - GIGE_PM_META_SIZE_1_7_5);

        // Old PM doesn't have the trigger setting so we will construct one.
        // We stored the trigger frame as an absolute frame number in the PM.
        meta_ptr->trig_set = 0x7f;
        if (meta_ptr->trig_frame < meta_ptr->first_frame)
            meta_ptr->trig_set_frames = meta_ptr->first_frame + meta_ptr->frame_count -
                                        meta_ptr->trig_frame;
        else
            meta_ptr->trig_set_frames = meta_ptr->trig_frame - meta_ptr->first_frame;
    }

    // Handle a CAP file from version 1.9.x or earlier.
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(1, 9, 99)))
    {
        memset(((uint8_t *)meta_ptr) + GIGE_PM_META_SIZE_1_9_99, 0,
               sizeof(pm_metadata) - GIGE_PM_META_SIZE_1_9_99);

        // Old PM doesn't have the sensor options field so fill in default value.
        meta_ptr->sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);
    }

    // Handle a CAP file from version 2.0.4 or earlier.
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 0, 4)))
    {
        memset(((uint8_t *)meta_ptr) + GIGE_PM_META_SIZE_2_0_4, 0,
               sizeof(pm_metadata) - GIGE_PM_META_SIZE_2_0_4);

        meta_ptr->black_offset = 0;
        meta_ptr->sensor_dac = 0;
        meta_ptr->sensor_vrefcol = 0;
        meta_ptr->sensor_gain = 0;

        // We were trying to stuff 10 pounds of sensor gain into an 8-pound bag.
        // Try to convert the old values to something that makes sense.
        if ((meta_ptr->old_sensor_gain == 100) ||
            (meta_ptr->old_sensor_gain == 150) ||
            (meta_ptr->old_sensor_gain == 200) ||
            (meta_ptr->old_sensor_gain == 250))
            meta_ptr->sensor_gain = meta_ptr->old_sensor_gain;
        else if (meta_ptr->old_sensor_gain == (300 - 256))
            meta_ptr->sensor_gain = 300;
        else if (meta_ptr->old_sensor_gain == (400 - 256))
            meta_ptr->sensor_gain = 400;
    }

    // Handle a CAP file from version 2.2.1 or earlier.
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 2, 1)))
    {
        memset(((uint8_t *)meta_ptr) + GIGE_PM_META_SIZE_2_2_1, 0,
               sizeof(pm_metadata) - GIGE_PM_META_SIZE_2_2_1);

        // Old RGB gains were completely invalid so use default.
        meta_ptr->red = 0x100;
        meta_ptr->green = 0x100;
        meta_ptr->blue = 0x100;
        meta_ptr->gamma = meta_ptr->old_gamma;

        // If it is less than 100, chances are good that it was a rollover when
        // the value >255 was written to the 8-bit field.
        if (meta_ptr->gamma < 100)
            meta_ptr->gamma = 100;

        meta_ptr->white_bal = get_wb_option(meta_ptr);

        // No know DC offset values.
        meta_ptr->dc_offset_red = 0;
        meta_ptr->dc_offset_grn = 0;
        meta_ptr->dc_offset_blu = 0;
    }

    // Prior to 2.2.11, the white_bal field wasn't being saved.
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 2, 10)))
    {
        // Try to calculate the white balance setting.
        meta_ptr->white_bal = get_wb_option(meta_ptr);
    }

    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 2, 99)))
    {
        // Make sure to zero out lut_id and lut_knots.
        memset(((uint8_t *)meta_ptr) + GIGE_PM_META_SIZE_2_2_99, 0,
               sizeof(pm_metadata) - GIGE_PM_META_SIZE_2_2_99);
    }


    // Otherwise, nothing to do.
}

/*!
 * @brief Load the partitions metadata information from the image memory.
 *
 * @param part_num  The partition we're interested in.
 *
 * @return true if everything was read sucessfully, otherwise false
 */
bool PartMgr::readImMetadata(int part_num)
{
    uint32_t dma_mod;
    uint32_t meta_trig;
    uint32_t read_count;
    FiFPGA  *fpga_ptr;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the information associated with this partition number.
    if (!getPartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo failed for %d!", __PRETTY_FUNCTION__, part_num);
        return false;
    }

    // Sanity check the partition start information.
    if ((part_info.partition_start + sizeof(pm_metadata)) > (imageMemorySize - imageMemorySizeReserved))
    {
        fi_log(LOG_DEBUG, "%s: Partition (%d) info is bad!", __PRETTY_FUNCTION__, part_num);
        return false;
    }

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();
    if (fpga_ptr == nullptr)
        return false;

    // We are going to transfer the data raw - no formatting.
    dma_mod = IM2PCIE_DMA_GENERIC_DATA;
    dma_mod |= IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
    dma_mod |= IM2PCIE_DMA_8BIT_RAW_FMT;

    // Read the metadata from the image memory. The metadata is at the partition's
    // start - so that's our source address.
    read_count = fpga_ptr->imageMemRead(dma_mod, 0,
                                        part_info.partition_start,
                                        (uint8_t *)&part_info.metadata,
                                        sizeof(pm_metadata));
    if (read_count <= 0)
        return false;

    // Make sure we update the metadata to the latest version.
    cleanup_metadata_version(&part_info.metadata);

    // We may need to adjust some of our settings.
    part_info.num_frames = part_info.metadata.frame_count;
    if (part_info.metadata.skip_count != 0)
    {
        // With a skip count, we need to adjust the metadata so that it
        // looks like we had a start trigger.
        part_info.metadata.first_frame = 0;
        part_info.metadata.trig_frame = 0;
        part_info.metadata.trig_set = 0;
    }
    part_info.first_frame_num = part_info.metadata.first_frame;
    part_info.frame_size = part_info.metadata.frame_size;

    // Now calculate the "relative" trigger offset.
    meta_trig = part_info.metadata.trig_frame;
    if (meta_trig < part_info.first_frame_num)
        part_info.trig_frame_num = part_info.num_frames - (part_info.first_frame_num - meta_trig);
    else
        part_info.trig_frame_num = meta_trig - part_info.first_frame_num;

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(part_num, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d", __PRETTY_FUNCTION__, part_num);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief We had a trigger event. Update our partition metadata.
 *
 * The extra_info parameter is broken down as follows:
 *  Word 0
 *     b0:16    SBS
 *     b17:23   Reserved
 *     b24:29   External marker inputs [5:0]
 *  Word 1
 *     b0:19    SBUS
 *
 * @param trigger_addr  The address of the trigger frame.
 * @param extra_info    The extra information associated with the trigger event.
 *
 * @return
 */
void PartMgr::partitionEventTrigger(uint64_t trigger_addr, const uint32_t *extra_info)
{
    bool ltr_mode;
    uint32_t mode;
    uint32_t trigger_frame;
    FiFPGA  *fpga_ptr;
    struct timespec trig_time;
    PART_MGR_WORKER::PartInfo part_info;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(PM_FUNC, "%s: trigger_addr=0x%" PRIx64, __PRETTY_FUNCTION__, trigger_addr);

    // Sanity check.
    if (imageBufferSize == 0)
    {
        fi_log(LOG_ERR, "%s: im_buf_size is 0?!", __PRETTY_FUNCTION__);
        return;
    }

    // Retrieve the partition information.
    trigger_part = getActivePartition(&part_info);
    if (trigger_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return;
    }

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        ltr_mode = true;
        mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
    }
    else
    {
        ltr_mode = false;
        if (fpga_ptr->ltrFasFireEnabled())
            mode = regs[GIGE_REG_LTR_MODE].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
        else
            mode = 0;
    }

    // For ROC/BROC, only update the trigger time on the first trigger event.
    if (mode != 0)
    {
        // Trigger frame isn't really meaningful.
        trigger_frame = 0;

        // Only calculate the trigger time on the first event.
        if ((part_info.metadata.trig_time_s == 0) && (part_info.metadata.trig_time_ns == 0))
        {
            // Get the time information from our extra information.
            trig_time.tv_sec = EXTRA_SBS(extra_info);
            trig_time.tv_nsec = EXTRA_SBUS(extra_info) * 1000;
        }
        else
        {
            // Just re-use the existing value.
            trig_time.tv_sec = part_info.metadata.trig_time_s;
            trig_time.tv_nsec = part_info.metadata.trig_time_ns;
        }
    }

    // If we're in LTR mode, then we calculate the frame number a bit differently.
    else if (ltr_mode)
    {
        // Now figure out the frame number of the trigger frame.
        trigger_frame = (uint32_t)((trigger_addr - ltrGetSessionStart(trigger_part)) / imageBufferSize);

        // Get the time information from our extra information.
        trig_time.tv_sec = EXTRA_SBS(extra_info);
        trig_time.tv_nsec = EXTRA_SBUS(extra_info) * 1000;
    }
    else
    {
        uint64_t last_addr = get_active_frame_addr(imageBufferCount - 1, true);

        // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // ALM_TODO_AS: The fixes don't really work, but I don't have the required information to
        //              supply anything better at this time. To do this properly will require additional
        //              logic in the end-capture-event handler.
        // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // Sanity check.
        if (trigger_addr < part_info.capture_buf_start)
        {
            // This isn't right.
            fi_log(LOG_DEBUG, "%s: The trigger_addr[0x%" PRIx64 "], is before the capture_buf_start=0x%" PRIx64 " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!",
                   __PRETTY_FUNCTION__, trigger_addr, part_info.capture_buf_start);
            trigger_frame = (uint32_t)(last_addr - (fpga_ptr->icmPostTriggerCount() * (uint64_t)imageBufferSize));
        }
        else if (trigger_addr > last_addr)
        {
            // This isn't right.
            fi_log(LOG_DEBUG, "%s: The trigger_addr[0x%" PRIx64 "], is after the last_addr=0x%" PRIx64 " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!",
                   __PRETTY_FUNCTION__, trigger_addr, last_addr);
            trigger_frame = part_info.num_frames - 1;
        }
        else
        {
            // Normal mode, normal trigger.
            fi_log(LOG_DEBUG, "%s: part=%d, trigger_addr=0x%" PRIx64 ", capture_buf_start=0x%" PRIx64 ", im_buf_size=%u", __PRETTY_FUNCTION__,
                   trigger_part, trigger_addr, part_info.capture_buf_start, imageBufferSize);
            trigger_frame = (uint32_t)((trigger_addr - part_info.capture_buf_start) / imageBufferSize);

            // Are we in FasFireLR mode?
            if (fpga_ptr->ltrFasFireEnabled())
            {
                // What is the trigger frame within the entire recording space.
                fflr_trigger_frame = (trigger_frame + (finished_part_count * imageBufferCount)) % imgSaveWorker::getInstance()->getRecordingFrameCount();
                fi_log(LOG_DEBUG, "%s: part=%d, trigger_frame=%u, fflr_trigger_frame=%u", __PRETTY_FUNCTION__, trigger_part, trigger_frame, fflr_trigger_frame);
            }
        }

        // Get the time information from our extra information.
        trig_time.tv_sec = EXTRA_SBS(extra_info);
        trig_time.tv_nsec = EXTRA_SBUS(extra_info) * 1000;
    }

    fi_log(LOG_DEBUG, "%s: part=%d, trigger_frame=%u", __PRETTY_FUNCTION__, trigger_part, trigger_frame);
    part_info.is_trig_part = true;
    part_info.trig_frame_num = trigger_frame;
    part_info.metadata.trig_frame = trigger_frame;
    part_info.metadata.trig_time_s = trig_time.tv_sec;
    part_info.metadata.trig_time_ns = trig_time.tv_nsec;

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(trigger_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for %d", __PRETTY_FUNCTION__, trigger_part);
    }

    // Update the partition state.
    setPartitionState(trigger_part, GIGE_PART_STATE_TRIGGERED);
}

/*!
 * @brief We need to fake a completion event because the user stopped recording but wants
 *        to retain whatever was already captured. This handles the ROC case.
 *
 * @param mode          ROC vs BROC mode
 * @param last_frame_offset For LTR mode, the offset of the last frame
 *              written to the SSD.
 *
 * @return true if it worked, false if no frames were captured yet.
 */
bool PartMgr::ltrSimulateRocCompletion(uint32_t mode, uint64_t last_frame_offset)
{
    int active_part;
    uint64_t    rel_addr;
    uint32_t    part_frame_num;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s: mode=0x%08x, last_frame_offset=0x%" PRIx64, __PRETTY_FUNCTION__, mode, last_frame_offset);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return false;
    }
    if (mode == 0)
    {
        fi_log(LOG_DEBUG, "%s: Expected non-zero mode!", __PRETTY_FUNCTION__);
        return false;
    }

    // Convert last frame address into a frame number relative to the start of
    // the partition.

    // Figure out which frame this last one is.
    rel_addr = last_frame_offset - ltrGetSessionStart(active_part);
    part_frame_num = (uint32_t)(rel_addr / imageBufferSize);

    // Handle the case where we didn't actually capture anything.
    if (part_frame_num == 0)
    {
        setPartitionState(active_part, GIGE_PART_STATE_EMPTY);
        return false;
    }

    // The number of frames is the main thing that changes as we append recordings.
    part_info.num_frames = part_frame_num;
    part_info.metadata.frame_count = part_info.num_frames;

    // Pretend that the trigger frame is always at the start of the recording.
    part_info.trig_frame_num = 0;

    // These haven't changed and we shouldn't even need to touch them.
    part_info.first_frame_num = 0;
    part_info.metadata.first_frame = part_info.first_frame_num;
    part_info.frame_size = imageBufferSize;
    part_info.metadata.frame_size = part_info.frame_size;

    part_info.partition_state = GIGE_PART_STATE_COMPLETE;

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d!", __PRETTY_FUNCTION__, active_part);
        return false;
    }

    // Finalize the information on the drive, including saving partition metadata.
    ltrFinalizePartitionDone(active_part, part_info.num_frames,
                             part_info.first_frame_num,
                             part_info.trig_frame_num,
                             part_info.metadata.trig_time_s,
                             part_info.metadata.trig_time_ns);

    // Notify after all work is done, including LTR operations.
    fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, part_info.num_frames);
    fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, part_info.trig_frame_num);
    fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);

    // All is well.
    return true;
}

/*!
 * @brief We need to fake a completion event because the user stopped recording but wants
 *        to retain whatever was already captured.
 *
 * @param triggered     TRUE if the trigger event was already reported via
 *              partition_trigger_event(). If not, we are simulating
 *              that too.
 * @param post_trigger_frames   The number of post-trigger frames as read from the FPGA register.
 * @param last_frame_addr   Either the address in image memory of the last frame written
 *              or for LTR mode, the offset of the last frame written to the SSD.
 * @param frames_captured   The frames capture count read from ICM, which latches at 30 bits
 *              so it is only useful for figuring out if we wrapped or if we
 *              failed to capture anything at all.
 * @param trig_info     Contains EXTRA_SBS + EXTRA_SBUS simulation if we
 *              need to simulate an end trigger.
 *
 * @return true if it worked, false if no frames were captured yet.
 */
bool PartMgr::simulateCompletion(bool triggered, uint32_t post_trigger_frames,
                                 uint64_t last_frame_addr, uint32_t frames_captured,
                                 uint32_t *trig_info)
{
    int active_part;
    bool wrap_flag;
    bool ltr_mode;
    FiFPGA *fpga_ptr;
    uint32_t total_capture_space;
    uint64_t rel_addr;
    uint32_t pre_trigger_frames;
    uint32_t meta_trigger_frame;
    uint32_t part_frame_num;
    uint32_t mode = 0;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s: triggered=%d, post_trigger_frames=%d, last_frame_addr=0x%" PRIx64 ", frames_captured=%d",
       __PRETTY_FUNCTION__, triggered, post_trigger_frames, last_frame_addr, frames_captured);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // See if we wrapped by comparing the total frames captured with the size of the session.
    // This is only of interest for non-append modes, where the frames captured start at
    // frame 0 in the partition.
    wrap_flag = (frames_captured > getFrameCount());
    TP(PM_MEMORY, "%s: part=%d, frames_captured=%u, get_frame_count=%u, wrap_flag=%d", __PRETTY_FUNCTION__, active_part, frames_captured, getFrameCount(), wrap_flag);
    TP(PM_MEMORY, "%s: capture_buf_start=0x%" PRIx64, __PRETTY_FUNCTION__, part_info.capture_buf_start);

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();

    // Get our mode information.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        ltr_mode = true;
        mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
        if ((mode != 0) && wrap_flag)
            fi_log(LOG_DEBUG, "%s: Oops, wrapped with mode=0x%08x.", __FUNCTION__, mode);

        // Figure out our total available space.
        total_capture_space = ltrGetActiveFrameCount();
    }
    else
    {
        ltr_mode = false;
        total_capture_space = imageBufferCount;
    }

    // Check special case of no frames captured at all.
    if (frames_captured == 0)
    {
        fi_log(LOG_DEBUG, "%s: stopping capture with frame_count = 0", __FUNCTION__);

        // For non-append mode captures, we just need to restore the partition to an empty state.
        // For append modes, we need to abort and clean up, restoring the previous state.
        if (ltr_mode)
        {
            // We have a null capture.
            ltrNilCapture();

            // Refresh the partition information.
            if (!getPartInfo(active_part, part_info))
            {
                fi_log(LOG_DEBUG, "%s: Unable to refresh the partition information.", __PRETTY_FUNCTION__);
                return false;
            }
        }
        else
        {
            // Mark the partition as empty.
            part_info.partition_state = GIGE_PART_STATE_EMPTY;

            // Send the partition information to the worker thread.
            if (!updatePartInfo(active_part, part_info))
            {
                fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for part %d", __PRETTY_FUNCTION__, active_part);
                return false;
            }
        }

        // The partition is now either EMPTY or COMPLETE.
        fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);
        return (part_info.partition_state == GIGE_PART_STATE_EMPTY) ? false : true;
    }

    // Shortcut to handling for append mode captures.
    if (ltr_mode && (mode != 0))
        return ltrSimulateRocCompletion(mode, last_frame_addr);

    pre_trigger_frames = total_capture_space - post_trigger_frames - 1;

    // Need to set these for the active partition, but can't use existing
    // conversion functions because they assume we've already set up
    // some parameters.
    // first_frame_num[] is relative to start of the partition.
    // num_frames[] needs to be the actual number captured.
    // trig_frame_num[] needs to be relative to first_frame_num[] but if a
    // trigger event was processed already, we saved the frame number relative
    // to the start of the partition.
    // metadata[].trig_frame needs to be relative to the start of the partition,
    // which we need to simulate if there was no trigger yet.

    // Convert last frame address into a frame number relative to the start of
    // the partition.
    if (ltr_mode)
    {
        rel_addr = last_frame_addr - ltrGetSessionStart(active_part);
        part_frame_num = (uint32_t)(rel_addr / part_info.frame_size);
        if (part_frame_num > 0)
            part_frame_num -= 1;
    }
    else
    {
        rel_addr = last_frame_addr - part_info.capture_buf_start;
        part_frame_num = (uint32_t)(rel_addr / part_info.frame_size);
        TP(PM_MEMORY, "%s: rel_addr=0x%" PRIx64 ", last_frame_addr=0x%" PRIx64 ", capture_buf_start=0x%" PRIx64, __PRETTY_FUNCTION__, rel_addr, last_frame_addr, part_info.capture_buf_start);
        TP(PM_MEMORY, "%s: part_frame_num=%u", __PRETTY_FUNCTION__, part_frame_num);
    }

    // Handle the case where we didn't actually capture anything.
    if (!wrap_flag && (part_frame_num == 0))
    {
        setPartitionState(active_part, GIGE_PART_STATE_EMPTY);
        return false;
    }

    // Are we in FasFireLR Mode?
    if (fpga_ptr->ltrFasFireEnabled())
    {
        part_info.num_frames = part_frame_num + 1;
        part_info.first_frame_num = 0;

        // If already triggered, trig_frame_num[] is already
        // set relative to 0 which is also the oldest frame.
        // Otherwise, fake an end trigger.
        if (!triggered)
        {
            fi_log(LOG_DEBUG, "%s: FFLR not triggered, setting trigger to 0.", __PRETTY_FUNCTION__);
            part_info.trig_frame_num = 0;
        }

        // Save the metadata information.
        part_info.metadata.trig_frame = part_info.trig_frame_num;
        part_info.metadata.trig_time_s = EXTRA_SBS(trig_info);
        part_info.metadata.trig_time_ns = EXTRA_SBUS(trig_info) * 1000;
    }
    else
    {
        // Check to see if the partition wrapped.
        if (wrap_flag)
        {
            part_info.num_frames = total_capture_space;
            part_info.first_frame_num = part_frame_num + 1;
            if (part_info.first_frame_num >= total_capture_space)
                part_info.first_frame_num = 0;
            if (triggered)
            {
                // Convert trigger frame number to be relative to where
                // our first frame number is.
                if (part_info.trig_frame_num >= part_info.first_frame_num)
                    part_info.trig_frame_num = part_info.trig_frame_num - part_info.first_frame_num;
                else
                    part_info.trig_frame_num = part_info.num_frames - ((part_frame_num - part_info.trig_frame_num) + 1);
                meta_trigger_frame = part_info.metadata.trig_frame;
            }
            else
            {
                // Fake an end trigger.
                part_info.trig_frame_num = total_capture_space - 1;
                meta_trigger_frame = part_frame_num;
            }
//        LOG("%s: wrap, num_frames=%u, first_frame_num=%u", __PRETTY_FUNCTION__, part_info.num_frames, part_info.first_frame_num);
        }
        else
        {
            part_info.num_frames = part_frame_num + 1;
            part_info.first_frame_num = 0;
//        LOG("%s: no wrap, num_frames=%u", __PRETTY_FUNCTION__, part_info.num_frames);

            // If already triggered, trig_frame_num[] is already
            // set relative to 0 which is also the oldest frame.
            // Otherwise, fake an end trigger.
            if (!triggered)
            {
                part_info.trig_frame_num = part_frame_num;
                meta_trigger_frame = part_frame_num;
            }
            else
            {
                meta_trigger_frame = part_info.trig_frame_num;
            }
        }

        // Set trigger time from extra_trigger_info.
        if (!triggered)
        {
            part_info.metadata.trig_frame = meta_trigger_frame;
            part_info.metadata.trig_time_s = EXTRA_SBS(trig_info);
            part_info.metadata.trig_time_ns = EXTRA_SBUS(trig_info) * 1000;
        }

        // Eliminate the pre-trigger frames that the user didn't request.
        if (part_info.trig_frame_num > pre_trigger_frames)
        {
            uint32_t extra_frames = part_info.trig_frame_num - pre_trigger_frames;
            TP(PM_MEMORY, "%s: trig_frame_num=%u, pre_trigger_frames=%u, extra_frames=%u", __PRETTY_FUNCTION__, part_info.trig_frame_num, pre_trigger_frames, extra_frames);

            part_info.trig_frame_num -= extra_frames;
            part_info.first_frame_num += extra_frames;
            if (part_info.first_frame_num >= total_capture_space)
                part_info.first_frame_num -= total_capture_space;

            // Adjust the number of frames to remove the frames we recorded while waiting for the trigger event.
            part_info.num_frames -= extra_frames;
        }
    }

    part_info.frame_size = imageBufferSize;
    part_info.metadata.frame_count = part_info.num_frames;
    part_info.metadata.first_frame = part_info.first_frame_num;
    part_info.metadata.frame_size = part_info.frame_size;
    part_info.partition_state = GIGE_PART_STATE_COMPLETE;

    // Finalize the information on the drive, including saving partition metadata.
    if (ltr_mode)
    {
        ltrFinalizePartitionDone(active_part, part_info.num_frames,
                                 part_info.first_frame_num,
                                 part_info.trig_frame_num,
                                 part_info.metadata.trig_time_s,
                                 part_info.metadata.trig_time_ns);
    }

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d", __PRETTY_FUNCTION__, active_part);
        return false;
    }

    // Notify after all work is done, including LTR operations.
//    LOG("%s: [%d] num_frames=%d", __PRETTY_FUNCTION__, __LINE__, part_info.num_frames);
    fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, part_info.num_frames);
    fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, part_info.trig_frame_num);
    fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);

    // All is well.
    return true;
}

/*!
 * @brief Handle the FPGA's reported capture end event in ROC/BROC modes.
 *
 * @param mode          ROC or BROC mode
 * @param last_frame_addr   Address of the last frame captured.
 */
void PartMgr::ltrPartitionEventRocDone(uint32_t mode, uint64_t last_frame_addr)
{
    int active_part;
    uint32_t last_frame_num;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return;
    }
    if (mode == 0)
    {
        fi_log(LOG_DEBUG, "%s: Expected non-zero mode!", __PRETTY_FUNCTION__);
        return;
    }

    // Figure out which frame was captured last.
    last_frame_num = (uint32_t)((last_frame_addr - ltrGetSessionStart(active_part)) / imageBufferSize);

    // We don't wrap in ROC/BROC.
    // the number of frames don't match the partition size.
    part_info.first_frame_num = 0;
    part_info.num_frames = last_frame_num + 1;
    part_info.frame_size = imageBufferSize;

    part_info.metadata.frame_count = part_info.num_frames;
    part_info.metadata.first_frame = part_info.first_frame_num;
    part_info.metadata.frame_size = part_info.frame_size;

    part_info.partition_state = GIGE_PART_STATE_COMPLETE;

    // Finalize the information on the drive, including saving partition metadata.
    ltrFinalizePartitionDone(active_part, part_info.num_frames,
                             part_info.first_frame_num,
                             part_info.trig_frame_num,
                             part_info.metadata.trig_time_s,
                             part_info.metadata.trig_time_ns);

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d", __PRETTY_FUNCTION__, active_part);
        return;
    }

    // Notify after all work is done, including LTR operations.
    fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, part_info.num_frames);
    fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, part_info.trig_frame_num);
    fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);
}

/*!
 * @brief A capture end event has happened in the FPGA.
 *
 * @param last_frame_addr   Location of last frame captured in image memory.
 * @param post_trigger_frames   How many frames after the trigger
 * @param frames_captured   Number of frames captured, but latched at 30 bits so it is
 *              only useful to find out (1) did we capture anything at all and
 *              (2) did the buffer wrap.
 *
 */
void PartMgr::partitionEventDone(uint64_t last_frame_addr, uint32_t post_trigger_frames, uint32_t frames_captured)
{
    int active_part;
    bool ltr_mode;
    bool wrap_flag;
    FiFPGA *fpga_ptr;
    uint32_t mode = 0;
    int32_t last_frame_num;
    PART_MGR_WORKER::PartInfo part_info;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return;
    }

    // Increment the count of completed partitions.
    finished_part_count++;

    TP(PM_FUNC, "%s: last_frame_addr=0x%" PRIx64 ", ptf=%d, fc=%d, active_partition=%d, finished_part_count=%u", __PRETTY_FUNCTION__, last_frame_addr, post_trigger_frames, frames_captured, active_part, finished_part_count);

    // Get our FPGA class pointer.
    fpga_ptr = CameraServer::get_fpga();

    // Are we in LR mode?
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // We are in LR mode, determine if it is ROC or BROC.
        ltr_mode = true;
        mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
    }
    else
    {
        // Not in LR mode.
        ltr_mode = false;

        // Are we in FFLR ROC or BROC?
        if (fpga_ptr->ltrFasFireEnabled())
            mode = regs[ GIGE_REG_LTR_MODE ].get_value() & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
    }

    // Did we capture any frames?
    if (frames_captured == 0)
    {
        fi_log(LOG_DEBUG, "%s: capture complete but frame_count = 0", __FUNCTION__);

        // For non-append mode captures, we just need to restore the partition to an empty state.
        // For append modes, we need to abort and clean up, restoring the previous state.
        if (ltr_mode)
        {
            // We have a null capture.
            ltrNilCapture();

            // Refresh the partition information.
            if (!getPartInfo(active_part, part_info))
            {
                fi_log(LOG_DEBUG, "%s: Unable to refresh the partition information.", __PRETTY_FUNCTION__);
                return;
            }
        }
        else
        {
            // Mark the partition as empty.
            part_info.partition_state = GIGE_PART_STATE_EMPTY;

            // Send the updated partition information to the worker thread.
            if (!updatePartInfo(active_part, part_info))
            {
                fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d", __PRETTY_FUNCTION__, active_part);
                return;
            }
        }

        // The partition is now either EMPTY or COMPLETE.
        fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);
        return;
    }

    // Are we in normal LR ROC/BROC mode?
    if (ltr_mode && (mode != 0))
    {
        ltrPartitionEventRocDone(mode, last_frame_addr);
        return;
    }

    // Did we wrap? Note that this should never happen in FasFireLR mode.
    wrap_flag = (frames_captured > getFrameCount());
    if (wrap_flag)
    {
        uint32_t    total_capture_space;

        fi_log(LOG_DEBUG, "%s: wrapped (%u > %u)", __PRETTY_FUNCTION__, frames_captured, getFrameCount());

        // Figure out our total available space.
        if (ltr_mode)
            total_capture_space = ltrGetActiveFrameCount();
        else
            total_capture_space = imageBufferCount;

        // special case begin and end triggers
        if (post_trigger_frames == 0)   // end trigger
        {
            last_frame_num = part_info.trig_frame_num;
            part_info.first_frame_num = last_frame_num + 1;
            if (part_info.first_frame_num >= total_capture_space)
                part_info.first_frame_num = 0;

            // Right now the trigger frame is absolute. We need to make it
            // relative to the first frame.
            part_info.trig_frame_num = total_capture_space - 1;
        }
        else if (post_trigger_frames == total_capture_space)   // begin trigger
        {
            // First frame is the trigger.
            part_info.first_frame_num = part_info.trig_frame_num;

            // Right now the trigger frame is absolute. We need to make it
            // relative to the first frame.
            part_info.trig_frame_num = 0;
        }
        else
        {
            // for everything else the general calculations work
            last_frame_num = part_info.trig_frame_num + post_trigger_frames;
            if (last_frame_num >= (int32_t)total_capture_space)
                last_frame_num = last_frame_num - total_capture_space;
            part_info.first_frame_num = last_frame_num + 1;
            if (part_info.first_frame_num >= total_capture_space)
                part_info.first_frame_num = 0;

            // Make the trigger frame relative to the first
            if (part_info.trig_frame_num >= part_info.first_frame_num)
                part_info.trig_frame_num = part_info.trig_frame_num - part_info.first_frame_num;
            else
                part_info.trig_frame_num = part_info.num_frames - ((last_frame_num - part_info.trig_frame_num) + 1);
        }
        part_info.num_frames = total_capture_space;
    }
    else
    {
        // Are we in FasFireLR Mode?
        if (fpga_ptr->ltrFasFireEnabled())
        {
            // We can't use post_trigger_Frames to calculate the number of frames in this partition.
            // Maybe we can do the calculation in a generic way that works for all modes...
            // "frames_captured" is the number of frames captured in this partition.

            // Looking at the code, frames_captured can sometimes be 0. This is an error condition, but
            // we still need to handle it.
            if (frames_captured == 0)
            {
                frames_captured = convertAddrToFrameNum(last_frame_addr) + 1;
                fi_log(LOG_DEBUG, "%s: FFLR frames_captured was 0, so using last_frame addr gives %d", __PRETTY_FUNCTION__, frames_captured);
            }

            // If the trigger partition is not the active partition, reset the trigger frame number.
            if (trigger_part != active_part)
                part_info.trig_frame_num = 0;
            part_info.first_frame_num = 0;
            part_info.num_frames = frames_captured;
            TP(PM_MEMORY, "%s: FFLR active_part=%d, trig_frame=%d, post_trig=%d, num_frames=%d", __PRETTY_FUNCTION__, active_part, part_info.trig_frame_num, post_trigger_frames, part_info.num_frames);

            // Make sure we have the time stamp setup.
            if ((part_info.metadata.trig_time_s == 0) && (part_info.metadata.trig_time_ns == 0))
            {
                // We want SOME time in the metadata.
                part_info.metadata.trig_time_s = fpga_ptr->timeReadSBS();
                part_info.metadata.trig_time_ns = fpga_ptr->timeReadSBUS() * 1000;
            }
        }
        else
        {
            // We didn't wrap so in this case we didn't fill up the whole partition so
            // the number of frames don't match the partition size.
            last_frame_num = part_info.trig_frame_num + post_trigger_frames;
            part_info.first_frame_num = 0;
            part_info.num_frames = last_frame_num + 1;
            TP(PM_MEMORY, "%s: trig_frame=%d, post_trig=%d, last_frame=%d, num_frames=%d", __PRETTY_FUNCTION__, part_info.trig_frame_num, post_trigger_frames, last_frame_num, part_info.num_frames);
        }
    }

    part_info.frame_size = imageBufferSize;
    part_info.metadata.frame_count = part_info.num_frames;
    part_info.metadata.first_frame = part_info.first_frame_num;
    part_info.metadata.frame_size = part_info.frame_size;
    part_info.partition_state = GIGE_PART_STATE_COMPLETE;

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed on part %d", __PRETTY_FUNCTION__, active_part);
        return;
    }

    // Finalize the information on the drive, including saving partition metadata.
    if (ltr_mode)
    {
        ltrFinalizePartitionDone(active_part, part_info.num_frames,
                                 part_info.first_frame_num,
                                 part_info.trig_frame_num,
                                 part_info.metadata.trig_time_s,
                                 part_info.metadata.trig_time_ns);
    }
    else
    {
        // Write the data to the image memory
        if (fpga_ptr->imageMemWrite(part_info.partition_start, (uint8_t *)&part_info.metadata, sizeof(pm_metadata), false) <= 0)
        {
            fi_log(LOG_ERR, "%s: imageMemWrite failed!", __PRETTY_FUNCTION__);
        }
    }

    // Notify after all work is done, including LTR operations.
    fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, part_info.num_frames);
    fcp_notify_apps(GIGE_REG_PART_TRIG_FRAME, part_info.trig_frame_num);
    fcp_notify_apps(GIGE_REG_PART_STATE, part_info.partition_state);
}

/*!
 * @brief Compute the number of sessions for a session length.
 *
 * Based on the session size and our image memory size,
 * figure out how many sessions we'll support.
 *
 * @param sess_length   The length in .25 GByte increments (1==.25GB).
 *
 * @return the number of sessions we can have of that length or 0 if length is illegal
 */
int PartMgr::calcSessionCount(uint32_t sess_length)
{
    uint32_t    session_count;
    uint64_t    sess_mem;

    TP(PM_FUNC, "%s: sess_length=%d, max_session_length=%d", __PRETTY_FUNCTION__, sess_length, sessionLengthMax);

    // Make sure the value is valid.
    if ((sess_length == 0) || (sess_length > sessionLengthMax))
    {
        fi_log(LOG_DEBUG, "%s: sess_length [%d] (max=%d) is not right.", __PRETTY_FUNCTION__, sess_length, sessionLengthMax);
        return 0;
    }

    // Figure out how many sessions we can divide the memory into.
    session_count = sessionLengthMax / sess_length;

    // Check to see if our sessions will fit into available memory.
    sess_mem = ((uint64_t)session_count) * ((uint64_t)sess_length) * SESSION_INCR;
    TP(PM_FUNC, "%s: sess_mem=%" PRIu64 ", im_size=%" PRIu64 ", im_size_reserved=%" PRIu64 ", avail=%" PRIu64, __PRETTY_FUNCTION__, sess_mem, imageMemorySize, imageMemorySizeReserved, (imageMemorySize - imageMemorySizeReserved));
    if (sess_mem > (imageMemorySize - imageMemorySizeReserved))
    {
        // Can we do anything about this?
        if (session_count > 1)
        {
            // Drop our session by one, as the last one is a runt.
            --session_count;
        }
    }

    TP(PM_FUNC, "%s: session_count=%d", __PRETTY_FUNCTION__, session_count);
    return session_count;
}

/*!
 * @brief Adjust the length of the session in the specified partition.
 *
 * @param adj_length    The new session length in .25 GByte increments (1==.25GB).
 * @param num_frames    [OPTIONAL] The number of frames to be found in the session.
 *
 * @return True on success, false on failure.
 */
bool PartMgr::adjustReviewPartition(uint32_t adj_length, uint32_t num_frames)
{
    bool rc(false);
    int review_part;
    uint64_t adj_size;
    PART_MGR_WORKER::PartInfo part_info;

    // Retrieve the partition information.
    review_part = getReviewPartition(&part_info);
    if (review_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getPartInfo failed for the review partition (%d)", __PRETTY_FUNCTION__, review_part);
        goto arp_bail;
    }
    TP(PM_FUNC, "%s: review_part=%d, adj_length=%d, num_frames=%d", __PRETTY_FUNCTION__, review_part, adj_length, num_frames);

    // Make sure the value is valid.
    if ((adj_length == 0) || (adj_length > sessionLengthMax))
    {
        fi_log(LOG_DEBUG, "%s: adj_length [%d] (max=%d) is not right.", __PRETTY_FUNCTION__, adj_length, sessionLengthMax);
        goto arp_bail;
    }

    // The adjusted session length must be smaller than, or equal to, what
    // the partition "normally" has for a length.
    if (sessionLength < adj_length)
    {
        goto arp_bail;
    }

#ifdef ALM_TODO_IS_IT_AN_ISSUE
    // Worry about state if this is the active partition.
    if (review_part == getActivePartition())
    {
        // Make sure we're in an appropriate state to be adjusting session lengths.
        uint32_t camera_state = CameraServer::get_camera_state();
        if ((camera_state != GIGE_CAMERA_STATE_IDLE) && (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
        {
            fi_log(LOG_DEBUG, "%s: camera_state (%d) isn't right.", __PRETTY_FUNCTION__, camera_state);
            goto arp_bail;
        }
    }
#endif

    // Calculate our adjusted partition size.
    adj_size = (adj_length * SESSION_INCR);

    // Sanity check.
    if (imageBufferSize == 0)
    {
        fi_log(LOG_ERR, "%s: im_buf_size is 0?!", __PRETTY_FUNCTION__);
        goto arp_bail;
    }

    // Now that we've adjust the size of the partition, figure out how many
    // frames can go into it.
    if (num_frames == 0)
    {
        // The number of frames is the maximum number that will fit into the partition.
        part_info.num_frames = (uint32_t)((adj_size - partitionSizeReserved) / imageBufferSize);
    }
    else
    {
        // Make sure the frame count is valid.
        if (num_frames > (uint32_t)((adj_size - partitionSizeReserved) / imageBufferSize))
        {
            fi_log(LOG_DEBUG, "%s: Invalid number of frames (%d > %d)!", __PRETTY_FUNCTION__, num_frames, ((adj_size - partitionSizeReserved) / imageBufferSize));
            num_frames = (uint32_t)((adj_size - partitionSizeReserved) / imageBufferSize);
        }

        // Save the new value.
        part_info.num_frames = num_frames;
    }

    // Send the updated partition information to the worker thread.
    if (!updatePartInfo(review_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for %d", __PRETTY_FUNCTION__, review_part);
        goto arp_bail;
    }

    // All is well.
    rc = true;

arp_bail:
    // All done.
    return rc;
}

/*!
 * @brief Set the trigger position. The high bit says how the rest of the
 *    register should be interpreted. Valid values are 0 for percent
 *    based or 1 for frame based. If it's percent based then values
 *    from 0 to 100 are valid for the rest of the register.  If frame
 *    based the rest of the register is an actual trigger frame number.
 *    In % mode the CS will be responsible for converting it during the
 *    arm process. The default value is % based with a value of 10.
 *
 * @param p_msg     Not used.
 * @param reg_value The new value.
 *
 * @return GEV_STATUS_SUCCESS on success.
 */
int PartMgr::setTriggerPos(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    uint32_t        old_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    old_value = regs[ GIGE_REG_TRIGGER_REQ_POS ].get_value();

    // Save this value in our class for later.
    CameraServer::get_partmgr()->triggerReqPos = reg_value;

    // Save the value into our register cache.
    regs[ GIGE_REG_TRIGGER_REQ_POS ].set_value(reg_value);

    // Let listening apps know about it.
    if (old_value != reg_value)
    {
        fcp_notify_apps(GIGE_REG_TRIGGER_REQ_POS, reg_value);

        // If we adjusted the trigger position, the autosave limits
        // can end up looking strange so force it back to full range.
        fcp_reset_save_limits(CameraServer::get_partmgr()->getFrameCount() - 1);
    }

    return GEV_STATUS_SUCCESS;
}

bool PartMgr::setReviewPartNum(int part_num)
{
    PART_MGR_WORKER::pmwSetReviewPartMessage updateMsg;

    TP(PM_FUNC, "%s: part_num=%d", __PRETTY_FUNCTION__, part_num);

    // All looks good. Save the new partition number.
    updateMsg.review_index = part_num;
    updateMsg.doSync(PART_MGR_WORKER::pmWorker::factory());
    if (updateMsg.resp != FAS_THREAD::OK)
    {
        fi_log(LOG_DEBUG, "%s: pmwSetReviewPartMessage failed!", __PRETTY_FUNCTION__);
        return false;
    }

    return true;
}

/*!
 * @brief Read the per-frame metadata found at the frame specified
 *        by frame_num.
 *
 * @param frame_num [IN] The absolute frame number we're interested in.
 * @param dest_ptr  [IN, OUT] Pointer to where to put the data.
 *
 * @return Number of bytes read.
 */
int PartMgr::getFramePFMD(uint32_t frame_num, uint8_t *dest_ptr)
{
    int read_count;
    uint32_t    pfmd_bytes;
    uint64_t    frame_addr;
    FiFPGA  *fpga_ptr;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Get the address for where we need to read.
    frame_addr = convertFrameNumToAddr(getActivePartition(), frame_num, false);
    TP(PM_MEMORY, "%s: frame_addr=0x%" PRIx64, __PRETTY_FUNCTION__, frame_addr);

    // Get the amount to offset for the PFMD.
    pfmd_bytes = FRAME_METADATA_HW_BYTES;

    // Are we in LTR mode?
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Adjust our address to point to the per-frame metadata.
        frame_addr += (getFrameSize() - pfmd_bytes);
        TP(PM_MEMORY, "%s: pfmd_addr=0x%" PRIx64, __PRETTY_FUNCTION__, frame_addr);

        // Read the metadata from the disk.
        read_count =  ltrReadBytes(dest_ptr, frame_addr, FRAME_METADATA_HW_BYTES);
    }
    else
    {
        // Adjust our address to point to the per-frame metadata.
        frame_addr += (getFrameSize() - pfmd_bytes);
        TP(PM_MEMORY, "%s: pfmd_addr=0x%" PRIx64, __PRETTY_FUNCTION__, frame_addr);

        // Read the metadata associated with that frame. We use big-endian because we
        // don't need to byte-swap data that is interpreted as 32-bit words.
        read_count = fpga_ptr->imageMemRead(IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_L2R | IM2PCIE_DMA_WORD_ORDER_R2L, 0, frame_addr,
                                            dest_ptr, FRAME_METADATA_HW_BYTES);

#ifdef PRINT_PFMD
        {
            extern void dump_pfmd(uint64_t offset, const uint8_t *pfmd_ptr);
            dump_pfmd(frame_addr, dest_ptr);
        }
#endif

    }

    return read_count;
}

uint32_t PartMgr::computePostTrigCount(uint32_t cap_count, uint32_t requested_pos)
{
    uint32_t pre_trigger;
    uint32_t post_trigger;

    TP(PM_FUNC, "%s: cap_count=%d, requested_pos=%d", __PRETTY_FUNCTION__, cap_count, requested_pos);

    if (requested_pos & GIGE_PART_TRIG_FRAMES)
    {
        pre_trigger = requested_pos & GIGE_PART_TRIG_FRAMES_MASK;
    }
    else
    {
        pre_trigger = (uint32_t)(((uint64_t)cap_count * (uint64_t)(requested_pos & GIGE_PART_TRIG_FRAMES_MASK)) / 100UL);
    }

    // Since the fpga wants the number of post trigger frames and what we have so far is the
    // number of pre trigger frames, switch it. Note that the post trigger frames doesn't
    // include the trigger frame so decrement unless it's an end trigger.
    if (pre_trigger >= cap_count)
        post_trigger = 0;
    else
        post_trigger = cap_count - pre_trigger - 1;

    return post_trigger;
}


/*!
 * @brief Write the correction and gain table information out to the image memory
 *        metadata via the FPGA driver.
 *
 * @return  0 on success, <0 on failure.
 */
bool PartMgr::writeCorrectionTable(const uint8_t *table_ptr, uint32_t table_size)
{
    int part_num;
    bool retval = true;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Loop through all the partitions copying the passed correction table into image memory.
    // The CT is always the same position relative to the start of each partition, so
    // ct_addr always refers to the relative address within the partition.
    for (part_num = 0; part_num < (int)getNumPartitions(); part_num++)
    {
        uint64_t ct_addr;   // Address within the partition for the correction table.
        PART_MGR_WORKER::PartInfo part_info;

        // Get the partition information.
        if (!getPartInfo(part_num, part_info))
        {
            fi_log(LOG_DEBUG, "%s: getPartInfo() failed for %d", __PRETTY_FUNCTION__, part_num);
            retval = false;
            break;
        }

        // The address for the correction table is always at a fixed offset
        // relative to the partition start.
        ct_addr = part_info.partition_start + PARTITION_METADATA_SIZE;
//        LOG("%s: [%d] ct_addr=0x%" PRIx64 ", table_ptr=%p, table_size=%d", __PRETTY_FUNCTION__, part_num, ct_addr, (void *)table_ptr, table_size);

        if (!fpga_ptr->imageMemWrite(ct_addr, table_ptr, table_size, false))
        {
            fi_log(LOG_ERR, "%s: Can't write table to IM for #%u - %d(%s)", __FUNCTION__,
                   part_num, errno, strerror(errno));
            retval = false;
            break;
        }
    }

    return retval;
}

/*!
 * @brief Write the gain table information out to the image memory
 *        metadata via the FPGA driver.
 *
 * @return  0 on success, <0 on failure.
 */
bool PartMgr::writeGainTable(uint8_t *table_ptr, uint32_t table_size)
{
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    int      part_num;
    bool     retval = true;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Loop through all the partitions copying the passed gain table into image memory.
    for (part_num = 0; part_num < (int)getNumPartitions(); part_num++)
    {
        uint64_t gt_addr;   // Address within the partition for the gain table.
        PART_MGR_WORKER::PartInfo part_info;

        // Get the partition information.
        if (!getPartInfo(part_num, part_info))
        {
            fi_log(LOG_DEBUG, "%s: getPartInfo() failed for %d", __PRETTY_FUNCTION__, part_num);
            retval = false;
            break;
        }

        // The address for the gain table is always at a fixed offset
        // relative to the partition start.
        gt_addr = part_info.partition_start + PARTITION_METADATA_SIZE + correctionTableSize;
//        LOG("%s: [%d] gt_addr=0x%" PRIx64 ", capture_buf_start=0x%" PRIx64 ", table_addr=%p, table_size=%d", __PRETTY_FUNCTION__, part_num, gt_addr,
//            part_info.capture_buf_start, (void *)table_addr, table_size);

        // Is there space allocated for the gain table?
        if (gt_addr < part_info.capture_buf_start)
        {
            // Write the data.
            if (!fpga_ptr->imageMemWrite(gt_addr, table_ptr, table_size, false))
            {
                fi_log(LOG_ERR, "%s: Can't write table to IM for #%u - %d(%s)", __FUNCTION__,
                       part_num, errno, strerror(errno));
                retval = false;
                break;
            }
        }
    }

    return retval;
}

/*
 * @brief Switch over to using a partition that was loaded from a CAP file
 *
 * After loading a CAP file into the active partition, we need to make sure our
 * local data structures are initialized with the new information.
 * Before calling this function, the metadata must already have been copied from image memory
 * into part_info[active_partition].metadata.
 *
 * @return true for success, false for failure
 */
bool PartMgr::useLoadedPartitionMetadata(const pm_metadata *meta_ptr)
{
    int active_part;
    uint32_t frame_count;
    uint32_t first_frame;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure the video capture area is located in the right spot.
    if (partitionSizeReserved < meta_ptr->sov_off)
    {
        // We need to bump the reserved space out some.
        partitionSizeReserved = meta_ptr->sov_off;
    }
    else if (partitionSizeReserved > meta_ptr->sov_off)
    {
        fi_log(LOG_DEBUG, "%s: The start of video is within the reserved space! psr=0x%x, sov_off=0x%x", __PRETTY_FUNCTION__, partitionSizeReserved, meta_ptr->sov_off);
        return false;
    }

    // Save these so we can restore them later.
    frame_count = part_info.metadata.frame_count;
    first_frame = part_info.metadata.first_frame;

    // This will create a certain amount of mischief by initializing internal data structures.
    layoutMemoryFromRegs();

    // There was a bug such that if a recording terminated ungracefully, the frame size in the
    // pm_metadata was not updated post-capture; then in ROC/BROC, it would forever remain unset.
    // Detect that case and update the frame size now.
    if (part_info.metadata.frame_size == 0)
    {
        fi_log(LOG_WARNING, "%s: Found frame_size=0 in pm_metadata.", __PRETTY_FUNCTION__);
        part_info.metadata.frame_size = imageBufferSize;
    }

    // Make sure the partition state is set to be complete.
    part_info.partition_state = GIGE_PART_STATE_COMPLETE;

    // Among the mischief was changing the "num_frames" variable, so we have to reset it back to
    // what was in the metadata.
    part_info.num_frames = frame_count;
    part_info.first_frame_num = first_frame;

    // Update the partition information in the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for %d", __PRETTY_FUNCTION__, active_part);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief This just reserves space for the table. It doesn't actually do anything else.
 * @param reserve_space     True means to reserve space, false means no gain table.
 */
void PartMgr::reserveGainTableSpace(bool reserve_space)
{
    TP(PM_FUNC, "%s: reserve_space=%d", __PRETTY_FUNCTION__, reserve_space);

    // If we're enabling support for the gain table, we'll need to adjust
    // how much memory we reserve for the partition information.
    if (reserve_space)
    {
        // Each of the table (ct & gt) require the same amount of space.
        partitionSizeReserved = PARTITION_METADATA_SIZE + (2 * correctionTableSize);
        gainTableIsAllowed = true;
    }
    else
    {
        // There is no gain table.
        partitionSizeReserved = PARTITION_METADATA_SIZE + correctionTableSize;
        gainTableIsAllowed = false;
    }

    // Make sure we're on a proper boundary.
    partitionSizeReserved = convert_to_HEXWORD(partitionSizeReserved) << HEXWORD_SHIFT;
}

