/*!
 * @file part_mgr_ltr.cxx
 * @brief This is the source file for the partition manager functions specific
 * to LTR mode.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <string>
#include <vector>

#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_filesystem.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "camera_server.h"
#include "cs_utils.h"
#include "fcp.h"
#include "image_correction.h"
#include "part_mgr.h"
#include "pm_worker.h"
#include "storage.h"

namespace PART_MGR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! How many bytes in this many sectors.
#define SECTORS_TO_BYTES( b )   ((b) * SECTOR_LEN)

//! How many pages it takes to hold these bytes.
#define BYTES_TO_PAGES( b )     (((b) + SSD_PAGE_LEN - 1) / SSD_PAGE_LEN)

//! How many pages it takes to hold these sectors.
#define SECTORS_TO_PAGES( s )   (((s) + SECTORS_PER_PAGE - 1) / SECTORS_PER_PAGE)

//! Number of 32-bit words in a frame metadata block. Note that we
//! are ONLY interested in the actual data in the metadata block.
constexpr int FMETA_SIZE_WORDS = (FRAME_METADATA_SIZE_BYTES / 4);

} /* End of namespace PART_MGR */

// Make things easier on ourselves.
using namespace Fastec;
using namespace FAS_FILESYSTEM;
using namespace PART_MGR;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

// Defined in part_mgr.h
#ifdef PRINT_PFMD
void dump_pfmd(uint64_t offset, const uint8_t *pfmd_ptr)
{
    const uint32_t *data_ptr = (const uint32_t *)pfmd_ptr;

    fi_log(LOG_DEBUG, "%" PRIx64 ": 0x%08x 0x%08x 0x%08x 0x%08x",
           offset, data_ptr[0], data_ptr[1], data_ptr[2], data_ptr[3]);
    fi_log(LOG_DEBUG, "%" PRIx64 ": 0x%08x 0x%08x 0x%08x 0x%08x",
           offset + 16, data_ptr[4], data_ptr[5], data_ptr[6], data_ptr[7]);
}
#endif

std::string PartMgr::flagsToStr(uint8_t flags)
{
    std::string ret_str;

    if ((flags & CAPHD_FLAG_RECONSTRUCTED) == CAPHD_FLAG_RECONSTRUCTED)
    {
        ret_str.append("REC");
    }
    if ((flags & CAPHD_FLAG_APPENDMODE) == CAPHD_FLAG_APPENDMODE)
    {
        if (!ret_str.empty())
            ret_str.append("|");
        ret_str.append("APND");
    }

    return ret_str;
}

std::string PartMgr::stateToStr(uint8_t state)
{
    std::string ret_str;

    switch (state)
    {
    case CAPTUREHD_ALLOCATED:   ret_str.append("Allocated");    break;
    case CAPTUREHD_COMPLETE:    ret_str.append("Complete");     break;
    case CAPTUREHD_CAPTURING:   ret_str.append("Capturing");    break;
    case CAPTUREHD_APPENDING:   ret_str.append("Appending");    break;
    case CAPTUREHD_READY:       ret_str.append("Ready");        break;
    default:    ret_str.append("No capture found");     break;
    }

    return ret_str;
}

std::string PartMgr::typeToStr(uint32_t type)
{
    std::string ret_str;

    ret_str += (char)(type & 0xFF);
    ret_str += (char)((type >> 8) & 0xFF);
    ret_str += (char)((type >> 16) & 0xFF);
    ret_str += (char)((type >> 24) & 0xFF);

    return ret_str;
}

void PartMgr::ltrDisplayPartInfo(const FAS_FILESYSTEM::PARTITION_INFO_VEC& part_info)
{
    FAS_FILESYSTEM::PARTITION_INFO_VEC::const_iterator part_it;

    fi_log(LOG_INFO, "Fastec Filesystem [%s], %u partitions found", ltr_volume_name.c_str(), part_info.size());

    // Walk through the partitions.
    for (part_it = part_info.begin(); part_it != part_info.end(); ++part_it)
    {
        const PARTITION_DATA *part_ptr = &part_it->part;
        CAP_UNION_VEC::const_iterator cap_it;

        fi_log(LOG_INFO, "[%010" PRIu64 "] Type: %s[0x%08x]", part_ptr->lba, typeToStr(part_ptr->type).c_str(), part_ptr->type);

        if (part_ptr->type == FAS_FILESYSTEM::LR_PART_SIG)
        {
            const FAS_FILESYSTEM::LR_PARTITION_HDR *hdr_ptr = &part_ptr->hdr.lr_part;

            fi_log(LOG_INFO, "             Meta=%" PRIu64 ":%" PRIu64 " Image=%" PRIu64 ":%" PRIu64 " FirstCapLBA@%" PRIu64 " Caps=%d",
                   hdr_ptr->fmeta_lba, hdr_ptr->fmeta_sectors, hdr_ptr->image_lba, hdr_ptr->image_sectors, hdr_ptr->capturehd_lba, part_it->caps.size());

            for (cap_it = part_it->caps.begin(); cap_it != part_it->caps.end(); ++cap_it)
            {
                const FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr = &cap_it->hdr.lr_cap;

                fi_log(LOG_INFO, "             [%010" PRIu64 "] Type %s[0x%08x] ID 0x%08X", cap_it->lba, typeToStr(cap_it->type).c_str(),
                       cap_it->type, cap_ptr->recording_id);
                fi_log(LOG_INFO, "                          Part=%" PRIu64 ", Prev=%" PRIu64 ", Next=%" PRIu64,
                       cap_ptr->link.dpart_lba, cap_ptr->link.previous_lba, cap_ptr->link.next_lba);
                fi_log(LOG_INFO, "                          FMeta=%llu:%llu Image=%llu:%llu State=%s[%u]",
                       cap_ptr->fmeta_lba, cap_ptr->fmeta_sectors, cap_ptr->image_lba, cap_ptr->image_sectors,
                       stateToStr(cap_ptr->state).c_str(), cap_ptr->state);
                fi_log(LOG_INFO, "                          Book=%llu:%llu PM=%llu:%llu CT=%llu:%llu",
                       cap_ptr->bk_offset, cap_ptr->bk_bytes, cap_ptr->pm_offset, cap_ptr->pm_bytes, cap_ptr->ct_offset, cap_ptr->ct_bytes);
                fi_log(LOG_INFO, "                          Flags=%s[0x%x] Frames=%u/%u/%u FrameSize=%u",
                       flagsToStr(cap_ptr->flags).c_str(), cap_ptr->flags,
                       cap_ptr->num_frames, cap_ptr->frame_count, cap_ptr->frame_sectors,
                       cap_ptr->frame_size);
            }
        }
    }
}

/*!
 * @brief Establish the layout for the SSD in LTR mode
 *
 * This gives us an initial drive layout so the PM knows the offset to the first frame
 * to be written on the drive, and we will do it without regard to what the
 * real capture parameters will turn out to be. Later we can make adjustments
 * so that the layout is more dynamic but this lets us shoehorn into the
 * existing code with fewer changes.
 *
 * @param really_write      true if we should actually write out the format to the drive,
 *              false to just go through the motions to figure out where the image
 *              region is and initialize memory-based structures.
 * @param num_sectors       Set up the drive to use this many sectors.
 * @param max_ct_size       The space to reserve for a correction table.
 *
 * @return True if everything is good, otherwise false.
 */
bool PartMgr::ltrLayoutDrive(bool really_write, uint64_t num_sectors, uint32_t max_ct_size, const std::string& volume_label)
{
    bool rc;
    std::string dev_path;
    LAYOUT_CAPTURE layout_cap;
    LAYOUT_PARTITION layout_part;
    LAYOUT_PARTITION_VEC layout_vec;

    TP(PM_LTR, "%s: really_write=%d, num_sectors=%" PRIu64 ", max_ct_size=%u", __PRETTY_FUNCTION__, really_write, num_sectors, max_ct_size);

    // The basic LR mode capture has one partition with one capture consuming the entire drive.
    layout_cap.cap_size = 0;
    layout_cap.max_ct_size = max_ct_size;
    layout_cap.signature = FAS_FILESYSTEM::LR_CAPTURE_SIG;
    layout_part.part_size = 0;
    layout_part.cap_info.push_back(layout_cap);
    layout_part.min_frame_size = 0;
    layout_part.signature = FAS_FILESYSTEM::LR_PART_SIG;
    layout_vec.push_back(layout_part);

    // Assume we will fail.
    rc = false;

    // Get our LTR device.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Are we writing this, or just giving it a go?
    if (really_write)
    {
        // Format the device.
        if (!FiFilesystem::getInstance()->format(dev_path, volume_label, num_sectors, layout_vec))
        {
            fi_log(LOG_DEBUG, "format() failed!");
            goto bail;
        }

        // Request an analysis.
        if (!FiFilesystem::getInstance()->analyze(dev_path))
        {
            fi_log(LOG_DEBUG, "analyze() failed.");
            goto bail;
        }

        // We should now have information. Assuming we have a Fastec filesystem.
        if (!FiFilesystem::getInstance()->getDriveInfo(dev_path, ltr_info, num_sectors, ltr_volume_name))
        {
            fi_log(LOG_DEBUG, "getPartitionInfo() failed.");
            goto bail;
        }
    }
    else
    {
        // Create the filesystem format, but don't actually write it to the drive.
        if (!FiFilesystem::getInstance()->tryLayout(dev_path, num_sectors, layout_vec, ltr_info))
        {
            fi_log(LOG_DEBUG, "format() failed!");
            goto bail;
        }
    }

    // Display the information.
//    if ((traceControl & PM_LTR) == PM_LTR)
    ltrDisplayPartInfo(ltr_info);

    // Assume we will succeed.
    rc = true;

bail:
    // Return our status.
    return rc;
}

/*!
 * @brief Set up a capture header so it is ready to be written when starting capture.
 *
 * The capture header must be in the ALLOCATED state, so all the allocations are already filled in.
 * This fills in the information that is specific to this capture.
 *
 * @param recording_id  ID for this recording, to distinguish different ones found on the drive.
 * @param frame_size    Frame size in bytes for this capture.
 * @param frame_count   Number of frames allocated for the capture.
 * @param pm_bytes      Size to reserve for memory partition metadata.
 * @param ct_bytes      Size to reserve for correction table.
 *
 * @return true if everything is good, otherwise false.
 */
bool PartMgr::ltrSetupCaptureHD(uint32_t recording_id, uint32_t frame_size, uint32_t frame_count, uint32_t pm_bytes, uint32_t ct_bytes)
{
    bool rc;
    uint64_t temp_size;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap;
    std::string dev_path;

    fi_log(LOG_DEBUG, "%s: id=0x%08x, frame_size=0x%08x, frame_count=0x%08x, pm_bytes=0x%08x, ct_bytes=0x%08x",
           __PRETTY_FUNCTION__, recording_id, frame_size, frame_count, pm_bytes, ct_bytes);

    // Assume we'll fail.
    rc = false;

    // See if we can find the capture information.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: get_ltr_cap_data() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get a pointer to our header so we can manipulate it easier.
    cap = &cap_info->hdr.lr_cap;

    // Set frame size, number of frames allocated, recording ID.
    cap->frame_size = frame_size;
    cap->frame_sectors = BYTES_TO_SECTORS(frame_size);
    cap->frame_count = frame_count;
    cap->recording_id = recording_id;

    // See if we have room for all the data.
    temp_size = ((uint64_t)frame_size) * ((uint64_t)frame_count);
    temp_size = BYTES_TO_SECTORS(temp_size);
    if (temp_size > cap->image_sectors)
    {
        fi_log(LOG_ERR, "%s: Image area too small (%" PRIu64 " sectors, need %" PRIu64 ")", __FUNCTION__,
               cap->image_sectors, temp_size);
        goto bail;
    }

    // These are not known yet when state is READY or CAPTURING or APPENDING.
    cap->num_frames = 0;
    cap->first_frame_num = ~cap->num_frames;
    cap->trigger_frame_num = ~cap->num_frames;

    // We are now in the ready state.
    cap->state = CAPTUREHD_READY;

    // Save this to disk.
    storage::get_ltr_device(dev_path);
    if (!FiFilesystem::getInstance()->updateCapture(dev_path, *cap_info))
    {
        fi_log(LOG_DEBUG, "%s: updateCapture() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = true;

bail:
    return rc;
}

/*!
 * @brief Write capture information to the drive before starting a capture
 *
 * @param pm_ptr        Memory partition metadata in processor memory.
 * @param ct_addr       Correction table address in IM.
 * @param ct_bytes      Size of correction and gain table to save.
 * @param gt_addr       Gain table address in IM. Size is assumed to be same as the correction table.
 *
 * @return true if everything is good, otherwise false.
 */
bool PartMgr::ltrWriteCaptureTables(const pm_metadata *pm_ptr, uint64_t ct_addr, uint32_t ct_bytes, uint64_t gt_addr)
{
    bool rc = false;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    std::string dev_path;
    std::unique_ptr<std::vector<uint8_t>> pm_xfer_vec = std::make_unique<std::vector<uint8_t>>();
    std::unique_ptr<std::vector<uint8_t>> ct_xfer_vec = std::make_unique<std::vector<uint8_t>>();
    std::unique_ptr<std::vector<uint8_t>> gt_xfer_vec = std::make_unique<std::vector<uint8_t>>();

    TP(PM_LTR, "%s: pm_ptr=%p, ct_addr=0x%" PRIx64 ", ct_bytes=%u, gt_addr=0x%" PRIx64, __PRETTY_FUNCTION__, (void *)pm_ptr, ct_addr, ct_bytes, gt_addr);

    // See if we can find the capture information.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: get_ltr_cap_data() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get our LTR drive.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // We will DMA from image memory so this size must be QWORD aligned.
    ct_bytes = convert_to_QWORD(ct_bytes) * QUADWORD_SIZE;

    // Make sure our buffer is sector aligned.
    // Do we have a pm_metadata structure?
    if (pm_ptr != nullptr)
    {
        // Move the metadata into our vector.
        pm_xfer_vec->resize(sizeof(pm_metadata));
        memcpy(pm_xfer_vec->data(), pm_ptr, sizeof(pm_metadata));

        // Update the metadata.
        if (!FiFilesystem::getInstance()->setCapturePM(dev_path, cap_info->lba, *pm_xfer_vec.get()))
        {
            fi_log(LOG_DEBUG, "%s: setCapturePM() failed.", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update our capture header to reflect that we've successfully written out the metadata.
        cap_info->hdr.lr_cap.pm_bytes = sizeof(pm_metadata);
    }

    // Read the correction table from DDR2 into our local buffer so it can be saved.
    if (ct_bytes != 0)
    {
        int ma_status;
        uint8_t  *buf_ptr;
        uint32_t buf_size;
        FiFPGA *fpga_ptr;
        std::unique_ptr<uint8_t[]> xfer_buf;       // Use the unique_ptr so we don't have to worry about freeing the memory.

        // Make sure our buffer size is on a sector boundary.
        buf_size = SECTORS_TO_BYTES(BYTES_TO_SECTORS(ct_bytes));

        // Setup our buffer. We need an aligned buffer so we can DMA straight into it.
        // We use a unique pointer simply to avoid cleanup issues.
        ma_status = posix_memalign((void **)&buf_ptr, DMA_PAGE_ALIGN, buf_size);
        if ((buf_ptr == nullptr) || (ma_status != 0))
        {
            fi_log(LOG_ERR, "%s: posix_memalign() failed [status=%d]!", __PRETTY_FUNCTION__, ma_status);
            return false;
        }
        xfer_buf.reset(buf_ptr);

        // Get the FPGA object so we can read the correction table.
        fpga_ptr = CameraServer::get_fpga();

        // Read the correction table.
        if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, ct_addr, buf_ptr, ct_bytes) == 0)
        {
            fi_log(LOG_ERR, "%s: Failed to read CT %u bytes at 0x%" PRIx64, __FUNCTION__, ct_bytes, ct_addr);
            goto bail;
        }

        // Put the data into our vector. It would be nice to not have to do this copy, but
        // in order to DMA straight into the vector, we would need to force the alignment.
        // Maybe there is some way to do that, but I don't know what it is, so we are stuck
        // with the extra copy here.
        ct_xfer_vec->resize(ct_bytes);
        memcpy(ct_xfer_vec->data(), buf_ptr, ct_bytes);

        // Is the a gain table?
        if (gt_addr == 0)
        {
            // There is no gain table, so empty out the vector.
            gt_xfer_vec->clear();
        }
        else
        {
            // Read the gain table. The size is assumed to match the correction table.
            if (fpga_ptr->imageMemRead(IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L, 0, gt_addr, buf_ptr, ct_bytes) == 0)
            {
                fi_log(LOG_ERR, "%s: Failed to read GT %u bytes at 0x%" PRIx64, __FUNCTION__, ct_bytes, gt_addr);
                goto bail;
            }

            // Put the data into our vector. It would be nice to not have to do this copy, but
            // in order to DMA straight into the vector, we would need to force the alignment.
            // Maybe there is some way to do that, but I don't know what it is, so we are stuck
            // with the extra copy here.
            gt_xfer_vec->resize(ct_bytes);
            memcpy(gt_xfer_vec->data(), buf_ptr, ct_bytes);
        }

        // Update the capture.
        if (!FiFilesystem::getInstance()->setCaptureInfo(dev_path, cap_info->lba, *ct_xfer_vec.get(), *gt_xfer_vec.get()))
        {
            fi_log(LOG_DEBUG, "%s: setCaptureInfo() failed.", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update our capture header to reflect that we've successfully written out the correction and gain
        // information. There is 1 sector for the header, and then the size of the correction table (sector aligned)
        // and, if we had one, the size of the gain table (sector aligned).
        cap_info->hdr.lr_cap.ct_bytes = SECTORS_TO_BYTES(1) + buf_size;
        if (gt_addr != 0)
            cap_info->hdr.lr_cap.ct_bytes += buf_size;
    }

    // It all worked.
    rc = true;

bail:
    // Return our status.
    return rc;
}

/*!
 * @brief Return the 64-bit time extracted from a frame metadata record.
 *
 * @param metadata_buf  Buffer containing FMETA_SIZE_WORDS words of frame metadata.
 *
 * @return (1M * SBS) + SBUS
 */
constexpr uint64_t PartMgr::fmetaTime(uint32_t *metadata_buf)
{
    return ((uint64_t)metadata_buf[ FMETA_TIME_SECONDS ] * (1000 * 1000)) + (uint64_t)metadata_buf[ FMETA_TIME_USECS ];
}

/*!
 * @brief Read the sector containing metadata for a specific frame.
 *
 * @param cap           Capture header already recovered from the drive.
 * @param sector_buf    Sector buffer for reading metadata, must be big enough for three sectors.
 * @param abs_frame     Absolute frame number of the frame whose metadata is to be read.
 *
 * @return Pointer to frame metadata for abs_frame within sector_buf or NULL if read failed.
 */
uint32_t *PartMgr::ltrGetPFMD(uint32_t *sector_buf, uint32_t abs_frame)
{
    int fd;
    uint32_t pfmd_offset;
    uint64_t frame_offset;
    std::string dev_path;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture header found.", __PRETTY_FUNCTION__);
        return nullptr;
    }

    // Get the offset in bytes. This is to the beginning of the requested frame.
    frame_offset = SECTORS_TO_BYTES(cap_ptr->image_lba) + (((uint64_t)cap_ptr->frame_size) * abs_frame);

    // Get the amount to offset for the PFMD.
    pfmd_offset = FRAME_METADATA_HW_BYTES;

    // Adjust to the PFMD of the requested frame.
    frame_offset += (cap_ptr->frame_size - pfmd_offset);

    // Find the LTR device.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
        return nullptr;
    }

    // Open the disk for reading.
    fd = open(dev_path.c_str(), O_RDONLY | O_LARGEFILE);
    if (fd < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        fi_log(LOG_ERR, "%s: open(%s) for read failed ", __PRETTY_FUNCTION__, dev_path.c_str());
        return nullptr;
    }

    // Seek to the metadata and read it. Only read the actual data in the metadata block.
    // We don't care about the padding.
    if ((lseek64(fd, (off64_t)frame_offset, SEEK_SET) != (off64_t)frame_offset) ||
        (read(fd, sector_buf, FRAME_METADATA_SIZE_BYTES) != FRAME_METADATA_SIZE_BYTES))
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        fi_log(LOG_ERR, "%s: Read(%" PRIu64 ", %u) failed.", __PRETTY_FUNCTION__, frame_offset, FRAME_METADATA_SIZE_BYTES);
        sector_buf = nullptr;
    }

    // Close the disk.
    close(fd);

    // Return the buffer pointer.
    return sector_buf;
}

/*!
 * @brief Analyze the SSD content at startup to see if we left a capture on it.
 *
 * @return true if a capture was found, false if not.
 */
bool PartMgr::ltrStartupAnalyze()
{
    bool rc;
    uint64_t drive_space;
    std::string dev_path;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap;

    TP(PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Assume we'll fail.
    rc = false;

    // Get our device.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we reset the filesystem library in case we have old data laying around.
    FiFilesystem::getInstance()->clearCache(std::string());

    // Analyze the filesystem.
    if (!FiFilesystem::getInstance()->analyze(dev_path))
    {
        fi_log(LOG_DEBUG, "%s: analyze() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Retrieve the drive information.
    if (!FiFilesystem::getInstance()->getDriveInfo(dev_path, ltr_info, drive_space, ltr_volume_name))
    {
        fi_log(LOG_DEBUG, "%s: getDriveInfo() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // For now, we're only interested in the first partition and the first capture.
    // Do some sanity checks.
    if (ltr_info.size() == 0)
    {
        fi_log(LOG_DEBUG, "%s: No partitions found.", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (ltr_info.at(0).part.type != LR_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected type for the first partition: 0x%08x [%s]", ltr_info.at(0).part.type, typeToStr(ltr_info.at(0).part.type).c_str());
        goto bail;
    }
    if (ltr_info.at(0).caps.size() == 0)
    {
        fi_log(LOG_DEBUG, "%s: No captures found.", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (ltr_info.at(0).caps.at(0).type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected type for the first capture: 0x%08x [%s]", ltr_info.at(0).caps.at(0).type, typeToStr(ltr_info.at(0).caps.at(0).type).c_str());
        goto bail;
    }

    // Get a pointer for ease of use.
    cap = &ltr_info.at(0).caps.at(0).hdr.lr_cap;

    // If our current append mode setting does not match the append mode of the capture,
    // we will want to change our current mode to match it.
    if (cap->state == CAPTUREHD_COMPLETE)
    {
        uint32_t mode;
        GIGE_REG_CACHE& regs = CameraServer::get_regs();

        mode = regs[ GIGE_REG_LTR_MODE ].get_value();
        if ((cap->flags & CAPHD_FLAG_APPENDMODE) != 0)
        {
            if ((mode & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_DUAL)) == 0)
            {
                fi_log(LOG_WARNING, "%s: Found mode=0x%08x but capture is APPENDMODE", __PRETTY_FUNCTION__, mode);

                mode |= GIGE_LTR_MODE_ROC;
                regs[ GIGE_REG_LTR_MODE ].set_value(mode);
            }
        }
        else
        {
            if ((mode & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0)
            {
                fi_log(LOG_WARNING, "%s: Found mode=0x%08x but capture is not APPENDMODE", __PRETTY_FUNCTION__, mode);

                mode &= ~(GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC);
                regs[ GIGE_REG_LTR_MODE ].set_value(mode);
            }
        }
    }

    // Display the information.
//    if ((traceControl & PM_LTR) == PM_LTR)
    ltrDisplayPartInfo(ltr_info);

    // All is well.
    rc = true;

bail:
    // Return our status.
    return rc;
}

/*!
 * @brief Load partition metadata from a capture.
 *
 * @param pm_ptr      Buffer to hold PM read from drive.
 *
 * @return true if PM was recovered successfully, false if not
 */
bool PartMgr::ltrRecoverPartMetadata(pm_metadata *pm_ptr)
{
    bool    rc = false;
    uint64_t cap_lba;
    std::string dev_path;
    std::vector<uint8_t> xfer_vec;

    TP(PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get our LR device path.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the capture LBA.
    cap_lba = ltrGetCaptureLBA();

    // Read the metadata from the filesystem.
    if (!FiFilesystem::getInstance()->getCapturePM(dev_path, cap_lba, xfer_vec))
    {
        fi_log(LOG_DEBUG, "%s: getCapturePM() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we got something.
    if (xfer_vec.size() != sizeof(pm_metadata))
    {
        fi_log(LOG_DEBUG, "%s: The metadata structure is the wrong size! [%d != %d]", __PRETTY_FUNCTION__, xfer_vec.size(), sizeof(pm_metadata));
        goto bail;
    }

    // Copy the metadata over.
    memcpy(pm_ptr, xfer_vec.data(), sizeof(pm_metadata));

    // Looks good.
    rc = true;

bail:
    return rc;
}

/*!
 * @brief Load correction table for a capture.
 *
 * @param ct_ptr        Buffer to hold CT read from drive.
 * @param ct_bufsize    Size of CT buffer.
 *
 * @return number of bytes read, which is never more than ct_bufsize.
 */
uint32_t PartMgr::ltrRecoverCorrectionTable(uint8_t *ct_ptr, uint32_t ct_bufsize)
{
    uint32_t ct_len;
    uint64_t cap_lba;
    std::string dev_path;
    std::vector<uint8_t> xfer_vec;

    TP(PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Assume failure.
    ct_len = 0;

    // Get our LR device path.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the capture LBA.
    cap_lba = ltrGetCaptureLBA();

    // Read the metadata from the filesystem.
    if (!FiFilesystem::getInstance()->getCaptureCT(dev_path, cap_lba, xfer_vec))
    {
        fi_log(LOG_DEBUG, "%s: getCaptureCT() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we got something.
    if (xfer_vec.empty())
    {
        fi_log(LOG_DEBUG, "%s: No correction table recovered!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure there's room.
    if (ct_bufsize < xfer_vec.size())
    {
        fi_log(LOG_DEBUG, "%s: There's not enough room [%d < %d]", __PRETTY_FUNCTION__, ct_bufsize, xfer_vec.size());

        // Use the smaller size.
        ct_len = ct_bufsize;
    }
    else
    {
        // Our destination is at least as big as our xfer_vec length, so we'll use
        // the xfer_vec length as our transfer size.
        ct_len = xfer_vec.size();
    }

    // Copy the metadata over.
    if (ct_len > 0)
        memcpy(ct_ptr, xfer_vec.data(), ct_len);

bail:
    // Return the number of bytes we copied.
    return ct_len;
}

/*!
 * @brief Load gain table for a capture.
 *
 * @param gt_ptr        Buffer to hold GT read from drive.
 * @param gt_bufsize    Size of GT buffer.
 *
 * @return number of bytes read, which is never more than gt_bufsize.
 */
uint32_t PartMgr::ltrRecoverGainTable(uint8_t *gt_ptr, uint32_t gt_bufsize)
{
    uint32_t gt_len;
    uint64_t cap_lba;
    std::string dev_path;
    std::vector<uint8_t> xfer_vec;

    TP(PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Assume failure.
    gt_len = 0;

    // Get our LR device path.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: No LTR device found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the capture LBA.
    cap_lba = ltrGetCaptureLBA();

    // Read the metadata from the filesystem.
    if (!FiFilesystem::getInstance()->getCaptureGT(dev_path, cap_lba, xfer_vec))
    {
        fi_log(LOG_DEBUG, "%s: getCaptureGT() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we got something.
    if (xfer_vec.empty())
    {
        fi_log(LOG_DEBUG, "%s: No gain table recovered!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure there's room.
    if (gt_bufsize < xfer_vec.size())
    {
        fi_log(LOG_DEBUG, "%s: There's not enough room [%d < %d]", __PRETTY_FUNCTION__, gt_bufsize, xfer_vec.size());

        // Use the smaller size.
        gt_len = gt_bufsize;
    }
    else
    {
        // Our destination is at least as big as our xfer_vec length, so we'll use
        // the xfer_vec length as our transfer size.
        gt_len = xfer_vec.size();
    }

    // Copy the metadata over.
    if (gt_len > 0)
        memcpy(gt_ptr, xfer_vec.data(), gt_len);

bail:
    // Return the number of bytes we copied.
    return gt_len;
}

uint64_t PartMgr::ltrGetActiveStartLba()
{
    uint64_t image_lba = 0;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get the pointer to the capture data structure.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr != nullptr)
    {
        // Get the LBA for the start of the image area.
        image_lba = cap_ptr->image_lba;
    }

    TP(PM_LTR, "%s: image_lba=%" PRIu64, __PRETTY_FUNCTION__, image_lba);
    return image_lba;
}

uint64_t PartMgr::ltrGetActiveStartFMetaLBA()
{
    uint64_t fmeta_lba = 0;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get the pointer to the capture data structure.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr != nullptr)
    {
        // Get the LBA for the start of the frame metadata area.
        fmeta_lba = cap_ptr->fmeta_lba;
    }

    TP(PM_LTR, "%s: fmeta_lba=%" PRIu64, __PRETTY_FUNCTION__, fmeta_lba);
    return fmeta_lba;
}

uint64_t PartMgr::ltrGetActiveImageSectors()
{
    uint64_t image_sectors = 0;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get the pointer to the capture data structure.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr != nullptr)
    {
        // Get the number of image sectors allocated for this capture.
        image_sectors = cap_ptr->image_sectors;
    }

    TP(PM_VERBOSE, "%s: image_sectors=%" PRIu64, __PRETTY_FUNCTION__, image_sectors);
    return image_sectors;
}

uint64_t PartMgr::ltrGetCaptureLBA()
{
    uint64_t cap_lba = 0;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_data;

    // Get the pointer to the capture data structure.
    cap_data = ltrGetCaptureData();
    if (cap_data != nullptr)
    {
        // Get the LBA for the start of the capture area.
        cap_lba = cap_data->lba;
    }

    TP(PM_LTR, "%s: cap_lba=%" PRIu64, __PRETTY_FUNCTION__, cap_lba);
    return cap_lba;
}

FAS_FILESYSTEM::LR_CAPTURE_HDR *PartMgr::ltrGetCaptureHeader()
{
    FAS_FILESYSTEM::CAPTURE_DATA *cap_data;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr = nullptr;

    // Get the pointer to the capture data structure.
    cap_data = ltrGetCaptureData();
    if (cap_data != nullptr)
    {
        // Get the pointer to the capture header.
        cap_ptr = &cap_data->hdr.lr_cap;
    }

    return cap_ptr;
}

FAS_FILESYSTEM::CAPTURE_DATA *PartMgr::ltrGetCaptureData()
{
    FAS_FILESYSTEM::CAPTURE_DATA *cap_data = nullptr;

    // For now, we're only interested in the first partition and the first capture.
    // Do some sanity checks.
    if (ltr_info.size() == 0)
    {
        // This is OK and normal if we, in fact, do not have an LR recording.
        goto bail;
    }
    if (ltr_info.at(0).part.type != LR_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected type for the first partition: 0x%08x [%s]", ltr_info.at(0).part.type, typeToStr(ltr_info.at(0).part.type).c_str());
        goto bail;
    }
    if (ltr_info.at(0).caps.size() == 0)
    {
        fi_log(LOG_DEBUG, "%s: No captures found.", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (ltr_info.at(0).caps.at(0).type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected type for the first capture: 0x%08x [%s]", ltr_info.at(0).caps.at(0).type, typeToStr(ltr_info.at(0).caps.at(0).type).c_str());
        goto bail;
    }

    // Get the pointer to the capture header.
    cap_data = &ltr_info.at(0).caps.at(0);

bail:
    return cap_data;
}

uint32_t PartMgr::ltrGetSession()
{
    // Right now, we only have one session.
    return 0;
}

/*!
 * @brief Return the session start address as a byte offset.
 */
uint64_t PartMgr::ltrGetSessionStart(uint32_t session)
{
    uint64_t session_start = 0;

    TP(PART_MGR::PM_LTR, "%s: session=%u, start_lba=%" PRIu64, __PRETTY_FUNCTION__, session, ltrGetActiveStartLba());

    // We're only supporting one "session" right now.
    if (session == 0)
    {
        // Convert the LBA to a byte offset.
        session_start = SECTORS_TO_BYTES(ltrGetActiveStartLba());
    }
    else
    {
        // FIXME Not implemented for more than one session
        fi_log(LOG_ERR, "%s: not implemented for session=%u", __FUNCTION__, session);
    }

    return session_start;
}

uint64_t PartMgr::ltrGetSessionSize()
{
    uint64_t sess_size;

    sess_size = SECTORS_TO_BYTES(ltrGetActiveImageSectors());
    TP(PART_MGR::PM_LTR, "%s: active_image_sectors=%" PRIu64 ", sess_size=%" PRIu64, __PRETTY_FUNCTION__, ltrGetActiveImageSectors(), sess_size);

    return sess_size;
}

uint32_t PartMgr::ltrGetActiveFrameCount()
{
    uint32_t frame_count;

    // Sanity check.
    if (imageBufferSize == 0)
    {
        fi_log(LOG_ERR, "%s: im_buf_size == 0?! <<<<<<<<<<<<<<<<<<<<<<<<<<", __PRETTY_FUNCTION__);
        return 1;
    }

    // Given the total available image sectors and how many sectors per image
    // we can calculate the number of frames.
    frame_count = (uint32_t)(ltrGetSessionSize() / imageBufferSize);
    TP(PART_MGR::PM_LTR, "%s: frame_count=%d", __PRETTY_FUNCTION__, frame_count);
    return frame_count;
}


uint64_t PartMgr::ltrGetRocStart()
{
    uint64_t roc_start;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Our start address, in bytes, of the image recording area.
    roc_start = SECTORS_TO_BYTES(ltrGetActiveStartLba());

    // If we're appending, then we need to get to the end of the space.
    if (cap_ptr->state == CAPTUREHD_APPENDING)
    {
        // Move to the "next available location" to start recording.
        roc_start += (((uint64_t)cap_ptr->num_frames) * imageBufferSize);
    }

    // Return the recording start address for ROC/BROC mode.
    TP(PART_MGR::PM_LTR, "%s: start=0x%" PRIx64, __PRETTY_FUNCTION__, roc_start);
    return roc_start;
}

bool PartMgr::ltrIsRocFull()
{
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    return (ltrGetActiveFrameCount() == cap_ptr->num_frames);
}

uint32_t PartMgr::ltrGetRocFrameCount()
{
    uint32_t full_frame_count;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Calculate the number of frames that fit.
    full_frame_count = ltrGetActiveFrameCount();

    // If we're appending, then return the number of frames left
    // in our capture "buffer".
    if (cap_ptr->state == CAPTUREHD_APPENDING)
    {
        // Return the remaining count.
        full_frame_count -= cap_ptr->num_frames;
    }

    // Just return the full count.
    TP(PART_MGR::PM_LTR, "%s: full_frame_count=%u", __PRETTY_FUNCTION__, full_frame_count);
    return full_frame_count;
}

/*!
 * @brief Compute the number of capture buffers available for LTR mode
 *        Sets the member variable ltr_buf_count.
 *
 * @param buf_count The number of buffers that would fit for non-LTR mode.
 */
void PartMgr::ltrSetCaptureBufferCount(uint32_t buf_count)
{
    uint32_t new_count;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // After subtracting two for special buffers, we don't need any more than
    // 4096 buffers since that is how many descriptor FIFOs we have.
    if (buf_count > (PartMgr::MAX_IM2SSD_DESCRIPTOR_FIFO_DEPTH + 2))
        new_count = PartMgr::MAX_IM2SSD_DESCRIPTOR_FIFO_DEPTH + 2;
    else
        new_count = buf_count;

    // We need room for one extra frame where we can load one for processor-driven
    // processing, e.g. when the imem_manager wants to process a JPEG.
    // We also need room to be able to load a sector of metadata.
    new_count -= 2;

    ltrImageBufferCount = new_count;
}

/*!
 * @brief Write post-capture information out to the SSD.
 *
 * @param part_num      As if we could have than one partition for LTR.
 * @param num_frames        Total number of frames captured or retained.
 * @param first_frame_num   Absolute frame number that is oldest.
 * @param trig_frame_num    Trigger frame number relative to first frame number.
 * @param trig_time_s       Trigger time in seconds.
 * @param trig_time_ns      Trigger time ns past the second (but with usec precision).
 *
 * @return true for success, false for failure
 */
bool PartMgr::ltrFinalizePartitionDone(
    uint16_t part_num,
    uint32_t num_frames,
    uint32_t first_frame_num,
    uint32_t trig_frame_num,
    uint32_t trig_time_s,
    uint32_t trig_time_ns
)
{
    bool       rc = true;
    std::unique_ptr<pm_metadata> meta_ptr;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;

    TP(PART_MGR::PM_LTR, "%s: part=%d, num_frames=%u, first=%u, trig=%u", __PRETTY_FUNCTION__,
       part_num, num_frames, first_frame_num, trig_frame_num);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    // If we are appending, handle the ROC/BROC append completion. Otherwise, the
    // first ROC/BROC is the same a normal LTR and has a state of CAPTURING.
    if (cap_info->hdr.lr_cap.state == CAPTUREHD_APPENDING)
        return ltrFinalizeRocPartitionDone(part_num, num_frames);

    // Write out the final capture information.
    // If we are in ROC/BROC mode, this will be the time used for the entire capture.
    cap_info->hdr.lr_cap.num_frames        = num_frames;
    cap_info->hdr.lr_cap.first_frame_num   = first_frame_num;
    cap_info->hdr.lr_cap.trigger_frame_num = trig_frame_num;
    cap_info->hdr.lr_cap.trigger_time_s    = trig_time_s;
    cap_info->hdr.lr_cap.trigger_time_ns   = trig_time_ns;
    cap_info->hdr.lr_cap.state             = CAPTUREHD_COMPLETE;

    // Get the partition metadata.
    meta_ptr = getPartitionMetadata(part_num);
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: getPartitionMetadata() failed for part %d", __PRETTY_FUNCTION__, part_num);
        return false;
    }

    // Update the capture information.
    rc = ltrUpdateCapture(cap_info, meta_ptr.get(), 0, 0, 0);
    if (!rc)
    {
        FiFPGA  *fpga_ptr = CameraServer::get_fpga();

        fpga_ptr->ltrSetSaveError(GIGE_SAVE_STATE_ERR | EIO);
        fi_log(LOG_DEBUG, "%s: ltr_update_capture() failed", __FUNCTION__);
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Write post-capture information out to the SSD.
 *
 * @param part_num      As if we could have than one partition for LTR.
 * @param num_frames        Total number of frames captured or retained.
 *
 * @return true for success, false for failure
 */
bool PartMgr::ltrFinalizeRocPartitionDone(uint16_t part_num, uint32_t num_frames)
{
    bool       rc = true;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    std::unique_ptr<pm_metadata> meta_ptr;

    TP(PART_MGR::PM_LTR, "%s: part_num=%d, num_frames=%u", __PRETTY_FUNCTION__, part_num, num_frames);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    // Write out the final capture information.
    cap_info->hdr.lr_cap.num_frames        = num_frames;
    cap_info->hdr.lr_cap.state             = CAPTUREHD_COMPLETE;

    // Get the partition metadata.
    meta_ptr = getPartitionMetadata(part_num);
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: getPartitionMetadata() failed for part %d", __PRETTY_FUNCTION__, part_num);
        return false;
    }

    // Update the filesystem.
    rc = ltrUpdateCapture(cap_info, meta_ptr.get(), 0, 0, 0);
    if (!rc)
    {
        FiFPGA  *fpga_ptr = CameraServer::get_fpga();

        fpga_ptr->ltrSetSaveError(GIGE_SAVE_STATE_ERR | EIO);
        fi_log(LOG_DEBUG, "%s: ltr_update_capture() failed", __FUNCTION__);
    }
    return rc;
}

/*!
 * @brief Read a range of per-frame metadata into a user-supplied buffer.
 *
 * @param part_num      As if we had multiple partitions for LTR mode.
 * @param start_frame       The first frame whose metadata we want to read.
 * @param end_frame     The last frame (inclusive) whose metadata we want to read.
 * @param buf           The processor buffer into which metadata will be written.
 * @param bufsize       The size in bytes of the passed buffer.
 *
 * @return  The number of bytes of data written to the buffer.
 */
uint32_t PartMgr::ltrReadPFMD(int part_num,
                              uint32_t start_frame,
                              uint32_t end_frame,
                              uint8_t *buf,
                              uint32_t buf_size,
                              bool use_hexwords
                             )
{
    int fd;
    uint32_t loop_frame;
    uint32_t pfmd_offset;
    uint32_t frame_count;
    uint32_t pfmd_length;
    std::string dev_path;

    TP(PART_MGR::PM_LTR, "%s: part_num=%d, start=%u, end=%u, use_hexwords=%c", __PRETTY_FUNCTION__, part_num, start_frame, end_frame, (use_hexwords ? 'T' : 'F'));

    // Statistics
    Fastec::IntervalTimer ltr_frame_timer;
    ltr_frame_timer.start();
    // Statistics

    // How many frames are we talking about?
    frame_count = (end_frame - start_frame) + 1;

    // Are we using hexwords?
    if (use_hexwords)
        pfmd_length = FRAME_METADATA_HW_BYTES;
    else
        pfmd_length = FRAME_METADATA_SIZE_BYTES;

    // Sanity check.
    if ((frame_count * pfmd_length) > buf_size)
    {
        fi_log(LOG_ERR, "%s: buf_size (%u) is too small (%d)", __PRETTY_FUNCTION__, buf_size, (frame_count * pfmd_length));
        return 0;
    }

    // Get the amount to offset for the PFMD.
    pfmd_offset = convert_to_HEXWORD(pfmd_length) << HEXWORD_SHIFT;

    // Find the LTR device.
    storage::get_ltr_device(dev_path);
    if (dev_path.empty())
    {
        fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Open the disk.
    fd = open(dev_path.c_str(), O_RDONLY | O_SYNC | O_LARGEFILE);
    if (fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for read failed - (%d) %s", __FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return 0;
    }

    // Loop through the frames and read up the metadata.
    for (loop_frame = start_frame; loop_frame <= end_frame; ++loop_frame)
    {
        uint64_t frame_offset;

        // Get the offset. The frame numbers are all relative.
        frame_offset = convertFrameNumToAddr(part_num, loop_frame, false);

        // Adjust to the PFMD.
        frame_offset += (getFrameSize() - pfmd_offset);

        // Read the data.
        if (ltrReadBytes(buf, frame_offset, pfmd_length, fd) != (int)pfmd_length)
        {
            fi_log(LOG_ERR, "%s: ltr_read_bytes() failed!", __PRETTY_FUNCTION__);
            break;
        }

        // Move the buffer up.
        buf += pfmd_length;
    }

    // Close the file descriptor.
    close(fd);

    // Statistics
    ltr_frame_timer.stop();
    TP(PART_MGR::PM_STATS, "%s: processing complete: %lf secs.", __PRETTY_FUNCTION__, ltr_frame_timer.lastMeasurement);
    // Statistics

    // Return the number of bytes we read.
    return (frame_count * pfmd_length);
}

/*!
 * @brief Update settings from partition metadata loaded from SSD.
 *
 * This is similar to what is done when loading a CAP file in non-LTR mode, except
 * in that case we could have multiple memory-based partitions.
 *
 * @return true if everything is good, otherwise false.
 */
bool PartMgr::ltrUpdateFromPartMetadata()
{
    // Get our support classes.
    CameraServer    *cs_ptr = CameraServer::get_cs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t        shutter_temp;
    std::unique_ptr<pm_metadata> meta_ptr;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get our partition's metadata.
    meta_ptr = getActivePartMetadata();
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null meta_ptr!", __PRETTY_FUNCTION__);
        return false;
    }

    // Adjust the shutter speed if necessary. We have a history of lying about the shutter speed
    // at 1X instead of hiding the frame overhead. The loaded metadata could have a value that
    // is too high, so adjust it if necessary.
    shutter_temp = meta_ptr->shutter_speed;
    if (!cs_ptr->make_shutter_valid(meta_ptr->fps, &shutter_temp))
    {
        fi_log(LOG_ERR, "%s: adjusted restored shutter %u->%u fps=%u", __FUNCTION__,
               meta_ptr->shutter_speed, shutter_temp, meta_ptr->fps);
        meta_ptr->shutter_speed = shutter_temp;
    }

    // If we can't use the saved ROI, we won't acknowledge this recording's existence.
    // This can fail if you are messing with the license file or using setbw so the
    // frame rate is not supported at boot time but will be fixed later.
    if ((meta_ptr->sensor_type != (uint32_t)sensor_ptr->get_sensor_type()) ||
        !cs_ptr->is_user_roi_valid(meta_ptr->roi_x, meta_ptr->roi_y,
                                   meta_ptr->roi_width, meta_ptr->roi_height,
                                   (bitmode_t)meta_ptr->bit_sel, meta_ptr->fps,
                                   meta_ptr->shutter_speed, meta_ptr->sensor_opts, NULL))
    {
        fi_log(LOG_ERR, "%s: Invalid LTR capture, sensor=0x%08x (%u,%u):%ux%u bits=%u fps=%u",
               __FUNCTION__, meta_ptr->sensor_type,
               meta_ptr->roi_x, meta_ptr->roi_y, meta_ptr->roi_width, meta_ptr->roi_height,
               meta_ptr->bit_sel, meta_ptr->fps);
        fi_log(LOG_ERR, "%s: Can't set ROI to the loaded values!", __FUNCTION__);
        return false;
    }

    // Update our partition's metadata.
    if (!updateActiveMetadata(meta_ptr.get()))
    {
        fi_log(LOG_DEBUG, "%s: Unable to update the active partition's metadata!", __PRETTY_FUNCTION__);
        return false;
    }

    // Update the FCP register database.
    regs[ GIGE_REG_SESSION_LENGTH ] = getSessionLengthMax();
    regs[ GIGE_REG_BIT_MODE ] = meta_ptr->bit_sel;
    regs[ GIGE_REG_OFFSET_X ] = meta_ptr->roi_x;
    regs[ GIGE_REG_OFFSET_Y ] = meta_ptr->roi_y;
    regs[ GIGE_REG_WIDTH ] = meta_ptr->roi_width;
    regs[ GIGE_REG_HEIGHT ] = meta_ptr->roi_height;
    regs[ GIGE_REG_FRAME_RATE ] = meta_ptr->fps;
    regs[ GIGE_REG_SHUTTER_SPEED ] = meta_ptr->shutter_speed;
    regs[ GIGE_REG_FPN ].set_value(meta_ptr->fpn_status) ;

    // Should we reserve space for the gain table?
    reserveGainTableSpace(meta_ptr->gain_off != 0);

    // Update the partition information from what we loaded.
    useLoadedPartitionMetadata(meta_ptr.get());

    // Write the PM metadata to the FPGA's DDR2. We just do this because other code
    // expects a copy of the current partition metadata to be kept there.
    if (fpga_ptr->imageMemWrite(getPartitionStart(), (uint8_t *)meta_ptr.get(), sizeof(pm_metadata), false) <= 0)
        fi_log(LOG_ERR, "%s: write PM to DDR2 failed", __PRETTY_FUNCTION__);

    return true;
}

/*!
 * @brief Analyze the SSD at startup.
 *
 * Look for an existing recording. If there is none, lay out the drive space so we know
 * how many frames can be written.
 *
 * @return true for success, false for failure
 */
bool PartMgr::ltrStartup()
{
    bool rc = false;
    int active_part;
    FiFPGA  *fpga_ptr;
    std::unique_ptr<pm_metadata> meta_ptr;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    meta_ptr = getActivePartMetadata();
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null meta_ptr!", __PRETTY_FUNCTION__);
        goto startup_bail;
    }

    // How big a drive (licensed or real) do we have?
    if (storage::get_ssd_licensed_sector_count() == 0)
    {
        // This should never happen since it should have been checked previously.
        fi_log(LOG_ERR, "%s: no licensed SSD, can't do LR mode", __FUNCTION__);
        goto startup_bail;
    }

    // See what is on the drive, whether we have a recording.
    if (!ltrStartupAnalyze())
    {
        fi_log(LOG_DEBUG, "%s: startup analysis failed!", __PRETTY_FUNCTION__);
        goto startup_bail;
    }
    if (!ltrRecoverPartMetadata(meta_ptr.get()))
    {
        fi_log(LOG_DEBUG, "%s: Unable to recover the partition metadata!", __PRETTY_FUNCTION__);
        goto startup_bail;
    }

    // Update our active partition's metadata.
    if (!updateActiveMetadata(meta_ptr.get()))
    {
        fi_log(LOG_DEBUG, "%s: Unable to update the active partition's metadata!", __PRETTY_FUNCTION__);
        goto startup_bail;
    }

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        goto startup_bail;
    }

    if (cap_ptr->ct_bytes != 0)
    {
        int ma_status;
        uint8_t *ct_buf;
        uint32_t ct_size = 0;
        uint32_t ct_bufsize;

        // If we are going to load CT from a capture, turn off FPN first.
        FiImageCorrection::fpnDisable();

        // We don't know exactly how big a buffer we need until we actually read the table(s).
        // If this is an old style LR recording, then the "ct_bytes" value should be correct
        // for the correction table (there is no gain table in the old LR recordings). On the
        // new recordings, the "ct_bytes" value is the total size of the correction and gain
        // table information.
        ct_bufsize = BYTES_TO_SECTORS(cap_ptr->ct_bytes + 2 * SECTOR_LEN) * SECTOR_LEN;

        // Allocate an aligned buffer so we can DMA directly into, and out of, it.
        ma_status = posix_memalign((void **)&ct_buf, DMA_PAGE_ALIGN, ct_bufsize);
        if ((ct_buf == NULL) || (ma_status != 0))
            fi_log(LOG_ERR, "%s: can't allocate CT read buffer=%u (status=%d)", __FUNCTION__, ct_bufsize, ma_status);
        else
        {
            // Recover the correction table.
            ct_size = ltrRecoverCorrectionTable(ct_buf, ct_bufsize);
            if (ct_size != 0)
                ct_size = fpga_ptr->imageMemWrite(get_active_ct_addr(), ct_buf, ct_size, false);

            // Recover the gain table, if there is one.
            ct_size = ltrRecoverGainTable(ct_buf, ct_bufsize);
            if (ct_size != 0)
                ct_size = fpga_ptr->imageMemWrite(get_active_gt_addr(), ct_buf, ct_size, false);

            // All done with the buffer.
            free(ct_buf);
        }

        // If we couldn't recover a CT, turn off pixel FPN but don't quit trying to recover.
        if (ct_size == 0)
        {
            // Turn off the pixel FPN.
            meta_ptr->fpn_status &= ~GIGE_FPN_FPGA_ENABLE_MASK;

            // Update our active partition's metadata.
            if (!updateActiveMetadata(meta_ptr.get()))
            {
                fi_log(LOG_DEBUG, "%s: Unable to update the active partition's metadata!", __PRETTY_FUNCTION__);
                goto startup_bail;
            }
        }
    }

    // Update our capture settings based on what we find in the partition's metadata.
    if (!ltrUpdateFromPartMetadata())
    {
        fi_log(LOG_DEBUG, "%s: Unable to update our capture settings!", __PRETTY_FUNCTION__);
        goto startup_bail;
    }

    // All is well.
    rc = true;

startup_bail:
    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part >= 0)
    {
        // If we have been successful, then we want to update the partition information.
        if (rc)
        {
            part_info.num_frames      = cap_ptr->num_frames;
            part_info.first_frame_num = cap_ptr->first_frame_num;
            part_info.trig_frame_num  = cap_ptr->trigger_frame_num;
            part_info.partition_state  = GIGE_PART_STATE_COMPLETE;
        }
        else
        {
            // If any of those steps failed, we will just start with a clean drive.
            // This will figure out the size and starting LBA of the image region and the
            // frame metadata region without actually writing anything on the drive.
            rc = ltrLayoutDrive(false, storage::get_ssd_licensed_sector_count(), correctionTableSize, ltr_volume_name);
            part_info.partition_state = GIGE_PART_STATE_EMPTY;
        }

        // Send the partition information to the worker thread.
        if (!updatePartInfo(active_part, part_info))
        {
            fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for part %d", __PRETTY_FUNCTION__, active_part);
            rc = false;
        }
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition failed", __PRETTY_FUNCTION__);
        rc = false;
    }


    TP(PM_LTR, "%s: rc=%d", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Format a drive for LTR use.
 *
 * @return true if successful, false for some kind of bizarre failure.
 */
bool PartMgr::ltrFormat(bool write_it, const std::string& volume_label)
{
    int active_part;
    bool rc = false;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    if (storage::get_ssd_licensed_sector_count() == 0)
    {
        // This should never happen since it should have been checked previously.
        fi_log(LOG_ERR, "%s: no licensed SSD, can't do LR mode", __FUNCTION__);
        return false;
    }

    // Retrieve the partition information.
    active_part = getActivePartition();
    if (active_part >= 0)
    {
        // We always layout the drive for the maximum size correction table that this sensor
        // can use, regardless of the FPN mode. If the CT is not used, we just won't write it out.
        rc = ltrLayoutDrive(write_it, storage::get_ssd_licensed_sector_count(), correctionTableSize, volume_label);
        if (rc)
        {
            // Save the volume label.
            ltr_volume_name = volume_label;

            // Mark the partition as being empty.
            setPartitionState(active_part, GIGE_PART_STATE_EMPTY);
        }
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition failed", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Set the one LTR mode partition to EMPTY and reinit layout information.
 *
 * @return true for success or false for failure.
 */
bool PartMgr::ltrClearPartition()
{
    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);
    return ltrFormat(false, ltr_volume_name);
}

/*!
 * @brief Write out LTR file system headers when starting capture.
 */
bool PartMgr::ltrArmSetup(uint32_t recording_id, uint32_t mode)
{
    bool rc = false;
    uint64_t ct_addr;
    uint64_t gt_addr;
    uint32_t ct_size;
    uint32_t frame_count;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    std::unique_ptr<pm_metadata> meta_ptr;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // The partition metadata was already updated with the current settings so use fpn_status
    // and save a correction table consistent with that setting. The saved correction table size
    // is currently big enough for a full-sized frame, even if the ROI is smaller.
    meta_ptr = getActivePartMetadata();
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartMetadata() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if ((meta_ptr->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
    {
        // The correction/gain table information area contains a header (1 sector)
        // and the correction/gain tables. The tables are the same size. Everything
        // wants to be on a sector boundary.
        ct_size = SECTOR_LEN + (2 * SECTORS_TO_BYTES(BYTES_TO_SECTORS(correctionTableSize)));
        ct_addr = get_active_ct_addr();
        gt_addr = get_active_gt_addr();
    }
    else
    {
        ct_size = 0;
        ct_addr = 0;
        gt_addr = 0;
    }

    // We didn't actually write the format before so time to write it now.
    if (!ltrLayoutDrive(true, storage::get_ssd_licensed_sector_count(), ct_size, ltr_volume_name))
        goto bail;

    // Figure out how many frames we can fit in the capture partition.
    frame_count = (uint32_t)(SECTORS_TO_BYTES(ltrGetActiveImageSectors()) / imageBufferSize);
    LOG("%s: Updating frame count=%d", __PRETTY_FUNCTION__, frame_count);
    if (!ltrSetupCaptureHD(recording_id, imageBufferSize, frame_count, sizeof(pm_metadata), ct_size))
        goto bail;

    // If this is ROC/BROC mode, make sure the append mode flag is set.
    // The capture header state needs to be set before we write it out to the drive.
    if ((mode & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0)
    {
        cap_info->hdr.lr_cap.flags |= CAPHD_FLAG_APPENDMODE;
        cap_info->hdr.lr_cap.trigger_frame_num = 0;

        // Invent a trigger time in case the capture doesn't terminate gracefully.
        cap_info->hdr.lr_cap.trigger_time_s = time(0);
        cap_info->hdr.lr_cap.trigger_time_ns = 0;
    }

    // Even if it is ROC/BROC mode, the state is CAPTURING for the first recording.
    cap_info->hdr.lr_cap.state = CAPTUREHD_CAPTURING;

    // Write out the headers with updated signatures. The size we need to pass for the correction and
    // gain tables is not the combined area size, just the size of each of the tables.
    rc = ltrUpdateCapture(cap_info, meta_ptr.get(), ct_addr, ((ct_size > 0) ? correctionTableSize : 0), gt_addr);
    if (rc)
    {
        FiFPGA *fpga_ptr;

        // Get the FPGA object.
        fpga_ptr = CameraServer::get_fpga();

        // Tell the FPGA how many buffers to use for capturing image data.
        rc = fpga_ptr->regWrite(ICM_CAPTURE_FRAME_BUF_CNT, ltrImageBufferCount);
    }

bail:
    return rc;
}

/*!
 * @brief Update LTR capture header when starting to append a recording into the capture.
 */
bool PartMgr::ltrRocArmAppend()
{
    bool    rc = false;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    std::unique_ptr<pm_metadata> meta_ptr;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    // We'll turn off the RECONSTRUCTED flag now because it just doesn't matter any more.
    // Set the state to APPENDING. The APPENDMODE should already have been set.
    cap_info->hdr.lr_cap.flags &= ~CAPHD_FLAG_RECONSTRUCTED;
    if ((cap_info->hdr.lr_cap.flags & CAPHD_FLAG_APPENDMODE) == 0)
    {
        cap_info->hdr.lr_cap.flags |= CAPHD_FLAG_APPENDMODE;
        fi_log(LOG_DEBUG, "%s: Oops, found APPENDMODE not set but should be.", __FUNCTION__);
    }
    cap_info->hdr.lr_cap.state = CAPTUREHD_APPENDING;

    meta_ptr = getActivePartMetadata();
    if (meta_ptr.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartMetadata() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    rc = ltrUpdateCapture(cap_info, meta_ptr.get(), 0, 0, 0);
    return rc;
}

/*!
 * @brief Get the recording ID that was used for the last capture.
 */
uint32_t PartMgr::getLtrRecordingID()
{
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_ptr;

    // Get a pointer to our capture header.
    cap_ptr = ltrGetCaptureHeader();
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return 0;
    }

    TP(PART_MGR::PM_LTR, "%s: id=0x%08x", __PRETTY_FUNCTION__, cap_ptr->recording_id);
    return cap_ptr->recording_id;
}

/*!
 * @brief Truncate an append-mode recording starting at a particular frame number.
 *
 * @param free_from_frame   First frame number to drop
 *
 * @return true for success, false for failure
 */
bool PartMgr::ltrTruncateFrom(uint32_t free_from_frame)
{
    int active_part;
    bool rc = false;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;
    PART_MGR_WORKER::PartInfo part_info;

    TP(PART_MGR::PM_LTR, "%s: free_from_frame=%u", __PRETTY_FUNCTION__, free_from_frame);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    if ((cap_info->hdr.lr_cap.flags & CAPHD_FLAG_APPENDMODE) == 0)
    {
        fi_log(LOG_DEBUG, "%s: Attempt to truncate in non-append mode.", __FUNCTION__);
        rc = false;
        goto bail;
    }

    // Set the new number of frames.
    cap_info->hdr.lr_cap.num_frames = free_from_frame;

    // We'll turn off the RECONSTRUCTED flag now because it just doesn't matter any more.
    cap_info->hdr.lr_cap.flags &= ~CAPHD_FLAG_RECONSTRUCTED;

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        rc = false;
        goto bail;
    }

    // Update the capture information on the drive.
    if (!ltrUpdateCapture(cap_info, &part_info.metadata, 0, 0, 0))
    {
        fi_log(LOG_DEBUG, "%s: ltr_update_capture() failed!", __PRETTY_FUNCTION__);
        rc = false;
        goto bail;
    }

    // Update our information on the way out.
    part_info.num_frames = cap_info->hdr.lr_cap.num_frames;

    // Send the partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for part %d", __PRETTY_FUNCTION__, active_part);
        rc = false;
        goto bail;
    }

    // All is well.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Read data from the disk.
 */
int PartMgr::ltrReadBytes(uint8_t *dest_ptr, uint64_t src_offset, int byte_count, int fd)
{
    bool close_fd = false;

    TP(PART_MGR::PM_LTR, "%s: src_offset=0x%" PRIx64 ", byte_count=%d", __PRETTY_FUNCTION__, src_offset, byte_count);

    // Is the disk already open?
    if (fd < 0)
    {
        // Find the LTR device.
        std::string dev_path;
        storage::get_ltr_device(dev_path);
        if (dev_path.empty())
        {
            fi_log(LOG_ERR, "%s: No LTR device found!", __PRETTY_FUNCTION__);
            return 0;
        }

        // Open the disk.
        fd = open(dev_path.c_str(), O_RDONLY | O_SYNC | O_LARGEFILE);
        if (fd < 0)
        {
            fi_log(LOG_ERR, "%s: open(%s) for read failed - (%d) %s", __FUNCTION__,
                   dev_path.c_str(), errno, strerror(errno));
            return 0;
        }

        // Don't forget to close the file descriptor.
        close_fd = true;
    }

    // Seek to and read the data.
    if (lseek64(fd, (off64_t)src_offset, SEEK_SET) != (off64_t)src_offset)
    {
        fi_log(LOG_ERR, "%s: lsee64(%" PRIu64 ", %u) failed - (%d) %s", __FUNCTION__, src_offset,
               byte_count, errno, strerror(errno));
        byte_count = 0;
    }
    else if (read(fd, dest_ptr, byte_count) != byte_count)
    {
        fi_log(LOG_ERR, "%s: Read(%" PRIu64 ", %u) failed - (%d) %s", __FUNCTION__, src_offset,
               byte_count, errno, strerror(errno));
        byte_count = 0;
    }

    if (close_fd)
        close(fd);

// Defined in part_mgr.h
#ifdef PRINT_PFMD
    // Display the PFMD we read.
    dump_pfmd(src_offset, dest_ptr);
#endif

    return byte_count;
}

/*!
 * @brief Update capture information on the drive.
 *
 * @note    This could be used to update the capture header and partition metadata after
 *          a capture is complete or to rewrite the capture header before starting to append
 *          in ROC/BROC mode.
 *
 * @param cap_info  [OUT] The filled captured information structure.
 * @param pm_ptr    [IN] Pointer to the partition metadata in processor memory.
 * @param ct_addr   [IN] Address in image memory for the correction table (may be 0).
 * @param ct_bytes  [IN] The size of each of the correction and gain tables, in bytes.
 * @param gt_addr   [IN] Address in image memory for the gain table (PRNU) (may be 0). The size is assumed to match the correction table size.
 * @return true if everything is good, otherwise false.
 */
bool PartMgr::ltrUpdateCapture(FAS_FILESYSTEM::CAPTURE_DATA *cap_info, const pm_metadata *pm_ptr, uint64_t ct_addr, uint32_t ct_bytes, uint64_t gt_addr)
{
    bool rc = false;
    std::string dev_path;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Update the partition metadata on the disk.
    if (!ltrWriteCaptureTables(pm_ptr, ct_addr, ct_bytes, gt_addr))
    {
        fi_log(LOG_DEBUG, "%s: ltr_write_capture_tables() to update the pm_metadata failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Now save our updated capture header.
    storage::get_ltr_device(dev_path);
    if (!FiFilesystem::getInstance()->updateCapture(dev_path, *cap_info))
    {
        fi_log(LOG_DEBUG, "%s: updateCapture() failed.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // It all worked.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Write post-capture when we didn't capture any new frames.
 *
 * In append mode we need to return the partition to its pre-capture state.
 * In non-append mode, we will just mark it as empty.
 *
 * @return true for success, false for failure
 */
bool PartMgr::ltrNilCapture()
{
    int active_part;
    PART_MGR_WORKER::PartInfo part_info;
    FAS_FILESYSTEM::CAPTURE_DATA *cap_info;

    TP(PART_MGR::PM_LTR, "%s", __PRETTY_FUNCTION__);

    // Get a pointer to our capture header.
    cap_info = ltrGetCaptureData();
    if (cap_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: No capture found.", __PRETTY_FUNCTION__);
        return false;
    }

    // If we were appending, we need to get back to our pre-capture state.
    // If the capture state is APPENDING, that means it was previously COMPLETE and
    // all we did to go from COMPLETE to APPENDING was to change the state and write it out;
    // we'll restore that state. We also restore the PM's number of frames, which was set to
    // the available space when we armed.
    if (cap_info->hdr.lr_cap.state != FAS_FILESYSTEM::CAPTUREHD_APPENDING)
    {
        // Otherwise, the capture is empty and we just need to mark the header as such.
        // Just brute force, zap it.
        ltrClearPartition();
        return false;
    }

    // Retrieve the partition information.
    active_part = getActivePartition(&part_info);
    if (active_part < 0)
    {
        fi_log(LOG_DEBUG, "%s: getActivePartition() failed!", __PRETTY_FUNCTION__);
        return false;
    }

    // Update the partition's metadata.
    part_info.partition_state  = GIGE_PART_STATE_COMPLETE;
    part_info.num_frames = cap_info->hdr.lr_cap.num_frames;
    cap_info->hdr.lr_cap.state = FAS_FILESYSTEM::CAPTUREHD_COMPLETE;

    // Send the partition information to the worker thread.
    if (!updatePartInfo(active_part, part_info))
    {
        fi_log(LOG_DEBUG, "%s: updatePartInfo() failed for part %d", __PRETTY_FUNCTION__, active_part);
        return false;
    }

    // Update the capture with the new information.
    return ltrUpdateCapture(cap_info, &part_info.metadata, 0, 0, 0);
}

