/*!
 * @file pb_control.cpp
 * @brief Create a thread to drive the playback of captured
 *    frames.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#ifndef _WIN32
#include <cxxabi.h>
#endif

#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>

#include <chrono>
#include <condition_variable>
#include <iomanip>
#include <mutex>

#include "bounce.h"
#include "camera_server.h"
#include "constants.h"
#include "cs_trace.h"
#include "cs_utils.h"
#include "fcp.h"
#include "fi_internal.h"
#include "jpeg_proc.h"
#include "part_mgr.h"
#include "pb_control.h"
#include "sensor.h"

namespace PB_CONTROL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const uint32_t FRAME_AGEOUT     = 30000;    //!< Number of microseconds, slightly more than half the frame time.
const int POOL_MAX_COUNT        = 4;        //!< Maximum number of buffers we will keep in the pool.

//! The states for our playback machine.
enum PLAYBACK_STATES
{
    PB_IDLE,                    //!< Waiting to be told to start.
    PB_SETUP,                   //!< Initial startup state.
    PB_WAIT,                    //!< Waiting for time to elapse to get the next frame.
    PB_LTR_FRAME_LOAD,              //!< Special state for loading LR mode frames.
    PB_PROCESS_FRAME,               //!< JPEG encode the frame.
    PB_FRAME_READY              //!< Frame is ready, send it to any waiting clients.
};

//! Message types for the Playback Controller.
enum PBC_MESSAGE_TYPES
{
    PBCM_FRAME_REQUEST = 0x30000,       //!< Request a frame from the playback machine.
    PBCM_RELEASE,               //!< Release a buffer.
    PBCM_CONTROL_UPDATE             //!< The playback control information changed.
};

//! Pool buffer management structure.
typedef struct BUF_POOL_REC
{
    struct timespec time_stamp;     //!< Timestamp when buffer was created/filled.
    int use_count;                  //!< Number of clients using this buffer.
    bool aged;                      //!< True if the buffer is too old.

    int frame_num;                  //!< Frame number for the data in this buffer.
    int buf_len;                    //!< Number of encoded bytes in the buffer.
    uint8_t *buf_ptr;               //!< Poiner to the buffer.
} BUF_POOL_REC;

//! Message class for PBCM_FRAME_REQUEST messages.
class pbcFrameMessage : public FAS_THREAD::SyncMessage
{
public:
    pbcFrameMessage();
    virtual ~pbcFrameMessage();

    PB_CONTROL::FRAME_REQUEST request;
};

//! Message class for PBCM_RELEASE messages.
class pbcReleaseMessage : public FAS_THREAD::SyncMessage
{
public:
    pbcReleaseMessage();
    virtual ~pbcReleaseMessage();

    uint8_t *buffer_ptr;    //!< The buffer we want to release.
};

//! Message class for PBCM_CONTROL_UPDATE messages.
class pbcControlUpdateMessage : public FAS_THREAD::Message
{
public:
    explicit pbcControlUpdateMessage(FAS_THREAD::MsgThread *client);
    virtual ~pbcControlUpdateMessage();

    uint32_t control_value; //!< The new playback control information.

private:
    pbcControlUpdateMessage();
};

/*!
 * @brief The playback control object. This object is
 *    a singleton and is "local" to this file.
 *
 * ![Playback Controller States](pb_control.gif)
 */
class pbcThread : public FAS_THREAD::MsgThread
{
public:
    ~pbcThread();

    // Singletons should not be cloneable.
    pbcThread(pbcThread& other) = delete;

    // Singletons should not be assignable.
    void operator=(const pbcThread&) = delete;

    static pbcThread *factory();

    static void close()
    {
        pbcThread *thread_ptr;

        // Release our singleton.
        thread_ptr = pbcThread::factory();
        if (thread_ptr != nullptr)
        {
            // Tell the thread that we are stopping.
            FAS_THREAD::SyncMessage msg;
            msg.req = FAS_THREAD::HALT;

            // Pass it on.
            msg.doSync(thread_ptr);

            // Try to join the thread.
            thread_ptr->join();
        }
    }

    virtual void messageHandle(FAS_THREAD::Message *message) override;
    virtual void run() override;

    void startThread()  { MsgThread::start((FAS_THREAD::THREAD_STARTUP)pbcEntry); }
    bool isIdle()       { return (state == PB_IDLE); }

private:
    pbcThread();

    //! @name Message Processing Methods
    //! @{
    bool handleFrameRequest(pbcFrameMessage *msg);
    void handleReleaseRequest(pbcReleaseMessage *msg);
    void handleControlUpdate(pbcControlUpdateMessage *msg);
    //! @}

    //! @name Utility Methods
    //! @{
    void fill_request(int buf_index, FRAME_REQUEST *request);
    void reclaim_buffers();
    //! @}

    //! @name State Machine Methods
    //! @{
    void idle();
    int setup();
    void wait();
    int ltr_frame_load();
    int process_frame();
    void frame_ready();
    //! @}

    static void pbcEntry(MsgThread *t)
    {
        pbcThread *my_ptr = static_cast<pbcThread *>(t);

        fi_log(LOG_DEBUG, "%s", __PRETTY_FUNCTION__);

        // Sanity check.
        if (my_ptr == nullptr)
        {
            LOG("%s: Can't cast the thread pointer!?", __PRETTY_FUNCTION__);
            return;
        }

        // Execute.
        my_ptr->hold();
        my_ptr->run();
    }

private:
    static std::unique_ptr<pbcThread> pb_control;
    static std::mutex pbc_mutex;

    bool done;              //!< True if the thread is done.
    bool do_single;         //!< True if this is a single playthrough.
    int state;              //!< Current state of the controller.
    uint32_t old_control;   //!< The control setting in use.

    //! Pacing timer measurement for our playback rate.
    Fastec::IntervalTimer pb_timer;

    //! Our request structure to get JPEG encoded frames through the system.
    JPEG_PROC::IPM_REQUEST jpeg_request;

    //! The time the image was encoded. We use this to determine if we can
    //! just return the last image, or if we need to generate a new one.
    struct timespec image_ts;

    //! The frame number of encoded image.
    int last_frame_num;

    //! Address (in either IM or the SSD) for the start of the current frame.
    uint64_t raw_frame_start;

    //! The length of each raw image. This will also include the PFMD area.
    uint32_t raw_frame_length;

    //! @name LR mode related variables.
    //! @{
    bool ltr_mode;      //!< True if LTR mode is enabled.
    int ltr_bb_index;       //!< Bounce buffer index for FPGA processing (LR mode).
    uint64_t ltr_bb_addr;   //!< The address for the LTR bounce buffer (LR mode).
    uint64_t ltr_frame_addr_im; //!< Frame address in image memory.
    uint8_t *ltr_mmap_ptr;  //!< The memory mapped pointer for the bounce buffer.
    //! @}

    BUF_POOL_REC buf_pool[POOL_MAX_COUNT];  //!< The pool of buffers.

    //! Queue of requestors for frame. We don't need to lock access because
    //! only the thread object has access, and it's state machine runs serially.
    std::queue<pbcFrameMessage *> image_requests;
};

} /* End of namespace PB_CONTROL */

using namespace CS_TRACE;
using namespace PB_CONTROL;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<pbcThread> PB_CONTROL::pbcThread::pb_control;
std::mutex PB_CONTROL::pbcThread::pbc_mutex;

//--------------------------------------------------------------------
// Forward Function Prototypes
//--------------------------------------------------------------------
static int pbc_update(uint32_t reg_addr, uint32_t reg_value);

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

//! @name Public facing functions.
//! @{

/*!
 * @brief Shutdown the playback control object.
 */
void pbc_close()
{
    pbcThread::close();
}

bool pbc_busy()
{
    bool rc(false);
    rc = pbcThread::factory()->isIdle() == false;

    return rc;
}

/*!
 * @brief Function interface to generate the FAS_THREAD::DUMP command.
 * @param dump_stream   A FILE pointer for were to write the data.
 */
void pbc_dump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(pbcThread::factory());
}

/*!
 * @brief Create the playback control singleton.
 */
void pbc_init()
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Turn on tracing.
    // traceControl |= TRACE_SEL_PB_CONTROL;

    // Make sure we have our singleton.
    pbcThread::factory();
}

/*!
 * @brief Request a frame during playback, or single-stepping.
 */
bool pbc_request(PB_CONTROL::FRAME_REQUEST *request)
{
    bool rc;
    pbcFrameMessage msg;

    // Sanity check
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request!?", __PRETTY_FUNCTION__);
        return false;
    }
    // FYI: It's OK for frame_buffer to be null.
    TP(TRACE_SEL_PB_CONTROL, "%s: frame_buffer=%p", __PRETTY_FUNCTION__, (void *)request->frame_buffer);

    // Build up our message.
    memcpy(&msg.request, request, sizeof(PB_CONTROL::FRAME_REQUEST));

    // Send it on.
    msg.doSync(pbcThread::factory());
    TP(TRACE_SEL_PB_CONTROL, "%s: result=%d", __PRETTY_FUNCTION__, (int)msg.resp);

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        request->frame_buffer = msg.request.frame_buffer;
        request->frame_len = msg.request.frame_len;
        request->frame_num = msg.request.frame_num;
        rc = true;
        TP(TRACE_SEL_PB_CONTROL, "%s: frame %d, len=%d bytes @ %p", __PRETTY_FUNCTION__, request->frame_num, request->frame_len, (void *)request->frame_buffer);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed. Did we change states?", __PRETTY_FUNCTION__);
        rc = false;
    }

    return rc;
}

/*!
 * @brief Release a JPEG encoded buffer.
 */
void pbc_release_frame(uint8_t *frame_buffer)
{
    pbcReleaseMessage msg;

    // Sanity check.
    if (frame_buffer == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null frame_buffer pointer?", __PRETTY_FUNCTION__);
        return;
    }

    TP(TRACE_SEL_PB_CONTROL, "%s: %p", __PRETTY_FUNCTION__, (void *)frame_buffer);

    // Build up the message.
    msg.buffer_ptr = frame_buffer;

    // Send it on.
    msg.doSync(pbcThread::factory());
}

//! @}
// End of public facing functions.

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

//! @name Message class methods
//! @{

pbcFrameMessage::pbcFrameMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PBCM_FRAME_REQUEST;
    memset(&request, 0, sizeof(PB_CONTROL::FRAME_REQUEST));
}

pbcFrameMessage::~pbcFrameMessage()
{
}

pbcReleaseMessage::pbcReleaseMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PBCM_RELEASE;
    buffer_ptr = nullptr;
}

pbcReleaseMessage::~pbcReleaseMessage()
{
}

pbcControlUpdateMessage::pbcControlUpdateMessage() :
    Message()
{
    req = (FAS_THREAD::REQ_CODE)PBCM_CONTROL_UPDATE;
    control_value = 0;
}

pbcControlUpdateMessage::pbcControlUpdateMessage(FAS_THREAD::MsgThread *client) :
    pbcControlUpdateMessage()
{
    req = (FAS_THREAD::REQ_CODE)PBCM_CONTROL_UPDATE;
    control_value = 0;

    route->push(client);
}

pbcControlUpdateMessage::~pbcControlUpdateMessage()
{
}

//! @}
// End of message class methods.

//--------------------------------------------------------------------
// pbcThread Implementation
//--------------------------------------------------------------------

/*!
 * @brief Object constructor.
 *
 * Ensure that we have a singleton by making this a private method.
 * Only the factory method can create the object now.
 */
pbcThread::pbcThread() :
    FAS_THREAD::MsgThread(),
    done(false),
    do_single(false),
    state(PB_IDLE),
    old_control(0),
    last_frame_num(0),
    raw_frame_start(0),
    raw_frame_length(0),
    ltr_mode(0),
    ltr_bb_index(-1),
    ltr_bb_addr(0),
    ltr_frame_addr_im(0),
    ltr_mmap_ptr(nullptr)
{
    int i;

    TP(TRACE_SEL_PB_CONTROL, "%s", __PRETTY_FUNCTION__);

    // Clear out the request structure.
    memset(&jpeg_request, 0, sizeof(jpeg_request));

    // Initialize our buffer pool.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
        memset(&buf_pool[i], 0, sizeof(BUF_POOL_REC));
}

/*!
 * @brief Object destructor.
 *
 * Ensure that we release all memory we might have gathered
 * during our period of operation.
 */
pbcThread::~pbcThread()
{
    TP(TRACE_SEL_PB_CONTROL, "%s", __PRETTY_FUNCTION__);

    // Make sure we don't have a bounce buffer hanging around.
    if (ltr_bb_index >= 0)
    {
        // Release our bounce buffer.
        bb_release(ltr_bb_index);
        ltr_bb_index = -1;
    }

    // Loop through all waiters, and tell them to go away.
    while (image_requests.size() > 0)
    {
        pbcFrameMessage *msg;

        // Grab the first message off the request queue.
        msg = static_cast<pbcFrameMessage *>(image_requests.front());
        image_requests.pop();

        // Sanity check.
        if (msg != nullptr)
        {
            // Indicate we are closing out the message.
            msg->resp = FAS_THREAD::FAULT;
            msg->forward();
        }
    }
}

/*!
 * @brief The singleton creation method.
 *
 * Only this factory method can create the singleton object, and its
 * purpose is to ensure there is only one playback control object.
 */
pbcThread *pbcThread::factory()
{
    pbcThread *rc(nullptr);

    std::lock_guard<std::mutex> lock(pbc_mutex);

    // We only allow one thread object.
    if (pb_control.get() == nullptr)
    {
        // Create our thread.
        pb_control = std::unique_ptr<pbcThread>(new pbcThread());
        if (pb_control.get() != nullptr)
            pb_control.get()->startThread();

        // Return the pointer to the singleton.
        rc = pb_control.get();
    }

    // Check to see if we were halted.
    else if (pb_control.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = pb_control.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Fulfill a request for a frame of data.
 */
void pbcThread::fill_request(int buf_index, FRAME_REQUEST *request)
{
    TP(TRACE_SEL_PB_CONTROL, "%s++: buf_index=%d, request=%p frame=%d", __PRETTY_FUNCTION__, buf_index, (void *)request, last_frame_num);

    // Satisfy the request right now.
    request->frame_len = buf_pool[buf_index].buf_len;
    request->frame_num = buf_pool[buf_index].frame_num;

    // What to do with the actual data?
    if (request->frame_buffer == nullptr)
    {
        // We'll hand the buffer over. The user is responsible for
        // releasing this when they are done.
        request->frame_buffer = buf_pool[buf_index].buf_ptr;
        buf_pool[buf_index].use_count++;
        TP(TRACE_SEL_PB_CONTROL, "%s: Using buffer %p from pool[%d]", __PRETTY_FUNCTION__, (void *)request->frame_buffer, buf_index);
    }
    else
    {
        // The user wants us to copy the data over. No need to increment the use counter.
        TP(TRACE_SEL_PB_CONTROL, "%s: Using provided buffer %p", __PRETTY_FUNCTION__, (void *)request->frame_buffer);
        memcpy(request->frame_buffer, buf_pool[buf_index].buf_ptr, buf_pool[buf_index].buf_len);
    }

    TP(TRACE_SEL_PB_CONTROL, "%s--: buf_index=%d, use_count=%d, frame_buffer=%p", __PRETTY_FUNCTION__, buf_index, buf_pool[buf_index].use_count, (void *)request->frame_buffer);
}

/*!
 * @brief Age out our ring of buffers.
 */
void pbcThread::reclaim_buffers()
{
    int i;
    uint64_t time_dif;
    uint64_t age_time;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Calculate the age time. This is 3/4th the playback rate.
    age_time = (uint64_t)(regs[GIGE_REG_PB_FRAME_PERIOD].get_value() >> 1);
    age_time = (age_time + (age_time >> 1)) * 1000;
    TP(TRACE_SEL_PB_CONTROL, "%s: age_time=%d", __PRETTY_FUNCTION__, age_time);

    // Loop through all the buffers, and see if we can age any of them.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
    {
        // Does this buffer contain usable data?
        if (!buf_pool[i].aged && (buf_pool[i].buf_ptr != nullptr))
        {
            // Determine the "age" of the buffer.
            time_dif = measuretime(0, &buf_pool[i].time_stamp);
            TP(TRACE_SEL_PB_CONTROL, "%s: pbc_buf=%d, frame=%d, time_dif=%" PRIu64, __PRETTY_FUNCTION__, i, buf_pool[i].frame_num, time_dif);
            if (time_dif > age_time)
            {
                TP(TRACE_SEL_PB_CONTROL, "%s: age out buffer %d", __PRETTY_FUNCTION__, i);

                // Is anyone using this buffer?
                if (buf_pool[i].use_count == 0)
                {
                    // Pass the request on to the JPEG processor.
                    TP(TRACE_SEL_PB_CONTROL, "%s: Release JPEG buffer %d.", __PRETTY_FUNCTION__, i);
                    jpeg_release_frame(buf_pool[i].buf_ptr);

                    // Clear out the buffer.
                    memset(&buf_pool[i], 0, sizeof(BUF_POOL_REC));
                }
                else
                {
                    // Mark it as aged.
                    buf_pool[i].aged = true;
                    TP(TRACE_SEL_PB_CONTROL, "%s: Buf %d count=%d", __PRETTY_FUNCTION__, i, buf_pool[i].use_count);
                }
            }
        }
    }
}

/*!
 * @brief Process the messages for the thread.
 */
void pbcThread::messageHandle(FAS_THREAD::Message *message)
{
    // Sanity check.
    if (message == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null message pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(TRACE_SEL_PB_CONTROL, "%s [%d]: req=0x%x", __PRETTY_FUNCTION__, state, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case PBCM_FRAME_REQUEST:
            if (!handleFrameRequest(static_cast<pbcFrameMessage *>(message)))
                message = nullptr;
            break;
        case PBCM_RELEASE:
            handleReleaseRequest(static_cast<pbcReleaseMessage *>(message));
            break;
        case PBCM_CONTROL_UPDATE:
            handleControlUpdate(static_cast<pbcControlUpdateMessage *>(message));
            delete message;
            message = nullptr;
            break;

        case FAS_THREAD::DUMP:
            *message->stream << "pbcThread Registers:" << std::endl;
            *message->stream << "  control: 0x" << std::setbase(16) << std::setfill('0') << std::setw(8) << old_control << std::endl;
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            halted = true;
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        if (message != nullptr)
            message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in pbcThread:\n%s", e.what().c_str());

        // Make sure we're working with a message.
        if (message != nullptr)
        {
            message->resp = FAS_THREAD::FAULT;
            message->forward();
        }
    }

    TP(TRACE_SEL_PB_CONTROL, "%s: done", __PRETTY_FUNCTION__);
}

/*!
 * @brief Wait for and return the next frame to be played.
 */
bool pbcThread::handleFrameRequest(pbcFrameMessage *msg)
{
    int i;

    TP(TRACE_SEL_PB_CONTROL, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null message pointer?!", __PRETTY_FUNCTION__);
        return true;
    }

    // If we are halted, then fail this message.
    if (halted)
    {
        // All done.
        msg->resp = FAS_THREAD::FAULT;
        return true;
    }

    // Make sure we age out any buffers before we try to use them.
    reclaim_buffers();

    // Loop through all the buffers, and see if we can find one that will work.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
    {
        // Does this buffer contain usable data?
        if (!buf_pool[i].aged)
        {
            if (buf_pool[i].buf_ptr != nullptr)
            {
                break;
            }
            else
            {
                TP(TRACE_SEL_PB_CONTROL, "%s: buf[%d] buf_ptr is null!", __PRETTY_FUNCTION__, i);
            }
        }
        else
        {
            TP(TRACE_SEL_PB_CONTROL, "%s: buf[%d] is aged", __PRETTY_FUNCTION__, i);
        }
    }
    if (i == POOL_MAX_COUNT)
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: No image frame is ready, queue request onto image_requests.", __PRETTY_FUNCTION__);

        // An image frame isn't ready right now. Add the message onto the request queue.
        image_requests.push(msg);
        TP(TRACE_SEL_PB_CONTROL, "%s: image_requests queue: %d", __PRETTY_FUNCTION__, image_requests.size());

        // Indicate that we are not finished with this message just yet.
        return false;
    }

    // Check to see if we only want a "different" frame.
    if ((msg->request.frame_num != -1) && (msg->request.frame_num == buf_pool[i].frame_num))
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: Requested a different frame [%d matches], jpeg_reuse=F, queue onto image_requests. <<<<<<<<<<<<<<<", __PRETTY_FUNCTION__, buf_pool[i].frame_num);

        // An image frame isn't ready right now. Add the message onto the request queue.
        image_requests.push(msg);
        TP(TRACE_SEL_PB_CONTROL, "%s: image_requests queue: %d", __PRETTY_FUNCTION__, image_requests.size());

        // Indicate that we are not finished with this message just yet.
        return false;
    }

    TP(TRACE_SEL_PB_CONTROL, "%s: pbc_buf=%d, jpeg_reuse=T, respond immediately.", __PRETTY_FUNCTION__, i);

    // Satisfy the request.
    msg->request.frame_len = buf_pool[i].buf_len;
    msg->request.frame_num = buf_pool[i].frame_num;

    // What to do with the actual data?
    if (msg->request.frame_buffer == nullptr)
    {
        // We'll hand the buffer over. The user is responsible for
        // releasing this when they are done. Make sure we increment the
        // usage counter.
        msg->request.frame_buffer = buf_pool[i].buf_ptr;
        buf_pool[i].use_count++;
    }
    else
    {
        // The user wants us to copy the data over.
        memcpy(msg->request.frame_buffer, buf_pool[i].buf_ptr, buf_pool[i].buf_len);
    }

    // We're all done.
    msg->resp = FAS_THREAD::OK;
    return true;
}

/*!
 * @brief Release a buffer.
 */
void pbcThread::handleReleaseRequest(pbcReleaseMessage *msg)
{
    int i;

    TP(TRACE_SEL_PB_CONTROL, "%s", __PRETTY_FUNCTION__);

    // Find the buffer for this image.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
    {
        // Is this a match?
        if (buf_pool[i].buf_ptr == msg->buffer_ptr)
            break;
    }
    if (i == POOL_MAX_COUNT)
    {
        fi_log(LOG_DEBUG, "%s: Couldn't find the buffer [%p]!", __PRETTY_FUNCTION__, (void *)msg->buffer_ptr);
        msg->resp = FAS_THREAD::FAULT;
        return;
    }

    // Decrement the use counter.
    buf_pool[i].use_count--;

    // If the count reaches 0, or the buffer has aged out, then we need to free the underlying buffer too.
    if (buf_pool[i].aged || (buf_pool[i].use_count == 0))
    {
        // Pass the request on to the JPEG processor.
        TP(TRACE_SEL_PB_CONTROL, "%s: Release JPEG buffer %d.", __PRETTY_FUNCTION__, i);
        jpeg_release_frame(buf_pool[i].buf_ptr);

        // Clear out the buffer.
        memset(&buf_pool[i], 0, sizeof(BUF_POOL_REC));
    }
    else
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: Buf %d count=%d, aged=%d", __PRETTY_FUNCTION__, i, buf_pool[i].use_count, buf_pool[i].aged);
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

/*!
 * @brief The playback control value has changed.
 */
void pbcThread::handleControlUpdate(pbcControlUpdateMessage *msg)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s: 0x%08x", __PRETTY_FUNCTION__, msg->control_value);

    // Is this a repeat?
    if (msg->control_value == old_control)
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: No change in the control value...exiting.", __PRETTY_FUNCTION__);
        return;
    }

    // Are we getting enabled?
    if ((msg->control_value & GIGE_PBC_ENABLE) == (uint32_t)GIGE_PBC_ENABLE)
    {
        // We are not in single playback mode.
        do_single = false;

        // If we are in IDLE, then it's time to start.
        if (state == PB_IDLE)
        {
            // We are being told to start.
            TP(TRACE_SEL_PB_CONTROL, "%s: Moving from IDLE to SETUP", __PRETTY_FUNCTION__);
            state = PB_SETUP;
        }
        else
        {
            // ALM: This might be OK if, for example, the interval field changed
            //      while we were previously enabled. Does that ever happen?
            LOG("%s: Unexpected state transition from %d [old:0x%08x, new:0x%08x]!",
                __PRETTY_FUNCTION__, state, old_control, msg->control_value);
        }
    }

    // Are we doing a single pass through?
    else if ((msg->control_value & GIGE_PBC_SINGLE_START) == (uint32_t)GIGE_PBC_SINGLE_START)
    {
        // Remember we're doing a single playthrough.
        do_single = true;

        // If we are in IDLE, then it's time to start.
        if (state == PB_IDLE)
        {
            // We are being told to start.
            TP(TRACE_SEL_PB_CONTROL, "%s: Moving from IDLE to SETUP [SINGLE]", __PRETTY_FUNCTION__);
            state = PB_SETUP;
        }
        else
        {
            // ALM: This might be OK if, for example, the interval field changed
            //      while we were previously enabled. Does that ever happen?
            LOG("%s: Unexpected state transition from %d [old:0x%08x, new:0x%08x]!",
                __PRETTY_FUNCTION__, state, old_control, msg->control_value);
        }
    }

    // The playback must be coming to a close. This may be a redundant request.
    else if (state != PB_IDLE)
    {
        // ALM_TODO: Make sure we tell the FPGA to stop if it's in single playout
        //       or DVR, mode.

        // We are being told to stop.
        LOG("%s: Stopping...Do we need to deal with cleanups? Transition from state: %d", __PRETTY_FUNCTION__, state);
        state = PB_IDLE;

        // ALM: Do we want to clear this now, or when it really is in the IDLE state?
        regs[GIGE_REG_PB_STATUS].set_value(0);
        fcp_notify_apps(GIGE_REG_PB_STATUS, 0);
    }

    // Remember the setting.
    old_control = msg->control_value;
    TP(TRACE_SEL_PB_CONTROL, "%s: done", __PRETTY_FUNCTION__);
}

/*!
 * @brief Waiting for something to do.
 */
void pbcThread::idle()
{
    int i;
    uint32_t reg_value;
    FAS_THREAD::Message *message = nullptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s++", __PRETTY_FUNCTION__);

    // On entry, check for any cleanup we might need to do.
    if (ltr_bb_index >= 0)
    {
        // Release our bounce buffer.
        bb_release(ltr_bb_index);
        ltr_bb_index = -1;
    }

    // Loop through all waiters, and tell them to go away.
    TP(TRACE_SEL_PB_CONTROL, "%s: image_requests queue: %d", __PRETTY_FUNCTION__, image_requests.size());
    while (image_requests.size() > 0)
    {
        pbcFrameMessage *msg;

        // Grab the first message off the request queue.
        msg = static_cast<pbcFrameMessage *>(image_requests.front());
        image_requests.pop();

        // Sanity check.
        if (msg != nullptr)
        {
            // Indicate we are closing out the message.
            msg->resp = FAS_THREAD::FAULT;
            msg->forward();
        }
    }

    // Release all buffers still being held.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
    {
        // Is this buffer still in use?
        if (buf_pool[i].use_count != 0)
        {
            // Pass the request on to the JPEG processor.
            TP(TRACE_SEL_PB_CONTROL, "%s: Release JPEG buffer %d.", __PRETTY_FUNCTION__, i);
            jpeg_release_frame(buf_pool[i].buf_ptr);
        }

        // Clear out the buffer.
        memset(&buf_pool[i], 0, sizeof(BUF_POOL_REC));
    }

    // Make sure we have a motion amplification handle and access to the internal library.
    if ((jpeg_request.ma_handle > 0) && (FiInternal::factory() != nullptr))
    {
        // Calculate what the packed size is.
        FiInternal::MA_CLOSE_FN closeHandlePtr = FiInternal::factory()->getMaCloseFuncPtr();
        if (closeHandlePtr != nullptr)
        {
            (*closeHandlePtr)(jpeg_request.ma_handle);
        }
    }
    jpeg_request.ma_handle = 0;

    // Let folks know we're idle.
    regs[GIGE_REG_PB_STATUS].set_value(0);
    fcp_notify_apps(GIGE_REG_PB_STATUS, 0);

    reg_value = (regs[GIGE_REG_PB_CONTROL].get_value() & ~(GIGE_PBC_SINGLE_START | GIGE_PBC_ENABLE));
    pbc_update(GIGE_REG_PB_CONTROL, reg_value);

    // Wait (forever) for a message.
    message = wait_for_message();

    // Sanity check the return value.
    if (message == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected nullptr from wait_for_message", __PRETTY_FUNCTION__);
    }

    // Are we the owner of the message?
    else if (message->owner == this)
    {
        // Handle the disposal of this message.
        ownerMessage(message);
    }
    else
    {
        // Go process the message. Note that the processing of the
        // message may very well mean a change in state.
        messageHandle(message);
    }

    // Are we done?
    if ((queue->size() == 0) && halted)
        done = true;

    TP(TRACE_SEL_PB_CONTROL, "%s--: done=%d", __PRETTY_FUNCTION__, done);
}

int pbcThread::setup()
{
    int rc = -1;
    CameraServer *cs_ptr = CameraServer::get_cs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s", __PRETTY_FUNCTION__);

    // Determine if we're in LTR mode or not.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);

    // Initialize our time structure.
    image_ts.tv_sec = 0;
    image_ts.tv_nsec = 0;

    // Currently there is no frame encoded.
    last_frame_num = -1;

    // Initialize the start of frame.
    TP(TRACE_SEL_PB_CONTROL, "%s: start_frame=0x%0x", __PRETTY_FUNCTION__, regs[GIGE_REG_PB_START_FRAME].get_value());
    raw_frame_start = pm_ptr->get_active_frame_addr(regs[GIGE_REG_PB_START_FRAME].get_value(), false);

    // Get the length of the image. This also includes the PFMD, but that's OK.
    raw_frame_length = pm_ptr->getFrameSize();
    TP(TRACE_SEL_PB_CONTROL, "%s: raw_frame_start=0x%" PRIx64 ", length=%d", __PRETTY_FUNCTION__, raw_frame_start, raw_frame_length);

    // Clear out the request structure.
    memset(&jpeg_request, 0, sizeof(jpeg_request));

    // Read in the user specified JPEG quality setting.
    jpeg_request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();

    // Get our dimensions.
    jpeg_request.width = regs[GIGE_REG_WIDTH].get_value();
    jpeg_request.height = regs[GIGE_REG_HEIGHT].get_value();

    // Processing info is always the same for playback images.
    jpeg_request.dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
    if (sensor_ptr->is_sensor_color())
    {
        jpeg_request.dma_mod |= IM2PCIE_DMA_GENERIC_DATA;
        jpeg_request.proc_info = IMGPROC_TYPE_NV12;
    }
    else
    {
        jpeg_request.dma_mod |= IM2PCIE_DMA_IMAGE_DATA;
        jpeg_request.proc_info = IMGPROC_TYPE_Y8;
    }

    // Are we supposed to use FPN?
    if ((regs[ GIGE_REG_FPN ].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        jpeg_request.proc_info |= IMGPROC_USE_FPN;

    // Allow reuse of frames up to 1/2-ish frame time old.
    jpeg_request.time_limit = FRAME_AGEOUT;

    // Initialize the request address to something.
    jpeg_request.image_addr = 0;

    // Are we in LTR mode and do we need a bounce buffer?
    if (ltr_mode && (ltr_bb_index < 0))
    {
        // Get our buffer for the raw file heading up to the FPGA.
        ltr_bb_index = bb_get_avail();
        ltr_bb_addr = bb_get_avail_addr(ltr_bb_index);

        // Memory map the bounce buffer.
        ltr_mmap_ptr = (uint8_t *)bb_mmap(ltr_bb_index, 0);
        if (ltr_mmap_ptr == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to map ltr_bb_index (%d).", __PRETTY_FUNCTION__, ltr_bb_index);

            // Release the bounce buffer.
            bb_release(ltr_bb_index);
            ltr_bb_index = -1;

            // Indicate an error.
            goto bail;
        }

        // Get a frame address in IM that we can use to store the raw frame.
        ltr_frame_addr_im = pm_ptr->getCaptureFrameAddr(0);
    }

    // Are we going to do motion amplification?
    jpeg_request.ma_handle = 0;
//#define ALM_DO_MOTION_AMPLIFICATION 1
#ifdef ALM_DO_MOTION_AMPLIFICATION
    // Make sure we have access to the internal library.
    if (FiInternal::factory() != nullptr)
    {
        // Calculate what the packed size is.
        FiInternal::MA_GET_HANDLE_FN getHandlePtr = FiInternal::factory()->getMaHandleFuncPtr();
        if (getHandlePtr != nullptr)
        {
            FAS_INTERNAL::FIP_MA_CONFIGURATION configInfo =
            {
                .framerate = (double)regs[GIGE_REG_FRAME_RATE].get_value(),
                .minFreqOfInterest = 10,
                .width = (int)jpeg_request.width,
                .height = (int)jpeg_request.height,
                .factor = 30
            };
            jpeg_request.ma_handle = (*getHandlePtr)(configInfo);
        }
    }
    LOG("%s: got ma_handle=%d", FI_FUNCTION, jpeg_request.ma_handle);
#endif /* ALM_DO_MOTION_AMPLIFICATION */

    // Are we in single playthrough mode.
    if (do_single)
    {
        // Tell the camera server to start polling the DVR status.
        // We start the polling AFTER we've had the FPGA actually start.
        // This prevents (reduces?) start-up timing issues.
        cs_ptr->poll_for_dvr_finish(true);
    }

    // All is well.
    rc = 0;

bail:
    // Return our status.
    return rc;
}

int pbcThread::ltr_frame_load()
{
    int rc = -1;
    uint32_t dma_mod;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(TRACE_SEL_PB_CONTROL, "%s++", __PRETTY_FUNCTION__);

    // Read up the image data.
    if (pm_ptr->ltrReadBytes(ltr_mmap_ptr, raw_frame_start, raw_frame_length) != (int)raw_frame_length)
    {
        fi_log(LOG_ERR, "%s: Unable to read from 0x%" PRIx64 " for %u bytes", __PRETTY_FUNCTION__, raw_frame_start, raw_frame_length);
        goto bail;
    }

    // Now ask the FPGA to put the data into the image memory buffer.
    dma_mod = PCIE2IM_DMA_BYTE_ORDER_R2L | PCIE2IM_DMA_WORD_ORDER_L2R;
    TP(TRACE_SEL_PB_CONTROL, "%s: call FPGA upload_image, dma_mod=0x%08x, len=%u", __PRETTY_FUNCTION__, dma_mod, raw_frame_length);
    rc = fpga_ptr->imageUpload(dma_mod, ltr_bb_addr, ltr_frame_addr_im, raw_frame_length);
    if (rc < 0)
    {
        // Oops! What happened here?!
        fi_log(LOG_ERR, "%s: Unable to DMA data!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = 0;

bail:
    // Return our status.
    TP(TRACE_SEL_PB_CONTROL, "%s--: rc=%d", __PRETTY_FUNCTION__, rc);
    return rc;
}

int pbcThread::process_frame()
{
    int rc;

    TP(TRACE_SEL_PB_CONTROL, "%s++: raw_frame_start=0x%" PRIx64 ", ltr_mode=%c", __PRETTY_FUNCTION__, raw_frame_start, (ltr_mode ? 'T' : 'F'));

    // Are we in LTR mode?
    if (ltr_mode)
    {
        // We already have our IM addres, we just need it in a different format.
        jpeg_request.image_addr = ltr_frame_addr_im;
    }
    else
    {
        // Pass in the image address.
        jpeg_request.image_addr = raw_frame_start;
    }

    // Make sure the JPEG information is reset.
    jpeg_request.jpeg_buffer = nullptr;
    jpeg_request.jpeg_len = 0;

    // Get the frame processed by the IPM.
    rc = jpeg_ipm_request(&jpeg_request);
    if (rc)
    {
        // Move on to transferring the frame.
        rc = 0;
    }
    else
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
        rc = -1;
    }

    // Return our status.
    TP(TRACE_SEL_PB_CONTROL, "%s--: rc=%d", __PRETTY_FUNCTION__, rc);
    return rc;
}

void pbcThread::frame_ready()
{
    int i;
    uint32_t rel_frame;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s++", __PRETTY_FUNCTION__);

    // Find an available entry in our buffer pool.
    for (i = 0; i < POOL_MAX_COUNT; ++i)
    {
        // A count of 0 means this buffer is available.
        if (buf_pool[i].use_count == 0)
            break;
    }
    if (i == POOL_MAX_COUNT)
    {
        fi_log(LOG_DEBUG, "%s: No buffers available! Someone isn't freeing!", __PRETTY_FUNCTION__);
        return;
    }

    // Set the expiration time for this frame.
    clock_gettime(CLOCK_MONOTONIC, &buf_pool[i].time_stamp);

    // What is the frame number?
    last_frame_num = pm_ptr->convertCaptureAddrToFrameNum(raw_frame_start);
    rel_frame = pm_ptr->convertAddrToFrameNum(raw_frame_start);
    buf_pool[i].frame_num = last_frame_num;
    TP(TRACE_SEL_PB_CONTROL, "%s: pbc_buf=%d, frame=%d (%d), image_requests=%d", __PRETTY_FUNCTION__, i, last_frame_num, rel_frame, image_requests.size());

    // Initialize the rest of the fields of the buffer.
    buf_pool[i].buf_len = jpeg_request.jpeg_len;
    buf_pool[i].buf_ptr = jpeg_request.jpeg_buffer;

    // Loop through all waiters, and give them the frame information.
    TP(TRACE_SEL_PB_CONTROL, "%s: image_requests size:%d", __PRETTY_FUNCTION__, image_requests.size());
    while (image_requests.size() > 0)
    {
        pbcFrameMessage *msg;

        // Grab the first message off the request queue.
        msg = image_requests.front();
        image_requests.pop();

        // Sanity check.
        if (msg != nullptr)
        {
            // Fill in the response fields.
            fill_request(i, &msg->request);

            // Return the message.
            msg->resp = FAS_THREAD::OK;
            msg->forward();
        }
    }

    // Update our cache.
    regs[GIGE_REG_PB_LAST_FRAME].set_value(rel_frame);
    regs[GIGE_REG_CSI_RVW_FRAME_NUM].set_value(rel_frame);
    fcp_notify_apps(GIGE_REG_CSI_RVW_FRAME_NUM, rel_frame);

    TP(TRACE_SEL_PB_CONTROL, "%s--: image_requests size:%d", __PRETTY_FUNCTION__, image_requests.size());
}

void pbcThread::wait()
{
    int clip_start;
    int clip_end;
    int play_frame;
    int delay_ms;
    int interval;
    uint32_t rel_frame;
    float time_accum;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s++, image_request size: %d", __PRETTY_FUNCTION__, image_requests.size());

    // Initialize our delay. Based on interval field of control value.
    delay_ms = regs[GIGE_REG_PB_FRAME_PERIOD].get_value();
    time_accum = 0.0;

    do
    {
        int adjust_ms;
        int delay_amt;
        FAS_THREAD::Message *message;

        // Stop the timer and see if we still need to delay.
        pb_timer.snap();
        time_accum += pb_timer.lastMeasurement;

        // Update our delay.
        adjust_ms = ((int)(time_accum * 1000));
        delay_amt = delay_ms - adjust_ms;
        TP(TRACE_SEL_PB_CONTROL, "%s: delay=%d, time_accum=%lf, adjust=%d", __PRETTY_FUNCTION__, delay_amt, time_accum, adjust_ms);

        // Check to see if we went negative. This happens if the time
        // to process the frame took longer than the time requested from
        // frame to frame. A negative value isn't allowed, but a zero means
        // to wait "forever", so we'll use the smallest positive integer.
        if (delay_amt <= 0)
            delay_amt = 1;

        // Wait our playback period, but still handle messages.
        message = wait_for_message(delay_amt);

        // Did we time out?
        if (message == nullptr)
        {
            // Must have timed out.
            TP(TRACE_SEL_PB_CONTROL, "%s: timeout", __PRETTY_FUNCTION__);
            break;
        }

        // Are we the owner of the message?
        else if (message->owner == this)
        {
            // Handle the disposal of this message.
            ownerMessage(message);
        }
        else
        {
            // Go process the message. Note that the processing of the
            // message may very well mean a change in state.
            messageHandle(message);
        }

        TP(TRACE_SEL_PB_CONTROL, "%s: end of while, state=%d, halted=%d", __PRETTY_FUNCTION__, state, halted);

    } while (!halted && (state == PB_WAIT));

    // Are we done?
    if ((queue->size() == 0) && halted)
    {
        done = true;
        goto bail;
    }

    // Did we get a state transition?
    if (state != PB_WAIT)
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: state transitioned to %d. Leaving.", __PRETTY_FUNCTION__, state);
        goto bail;
    }

    // Restart our timer.
    pb_timer.reset();
    pb_timer.start();

    // Where are our boundaries?
    clip_start = static_cast<int>(regs[GIGE_REG_PB_CUT_IN_FRAME].get_value());
    clip_end = static_cast<int>(regs[GIGE_REG_PB_CUT_OUT_FRAME].get_value());
    interval = (old_control >> PBC_INTERVAL_SHIFT) & PBC_INTERVAL_MASK;
    rel_frame = regs[GIGE_REG_PB_LAST_FRAME].get_value();
    TP(TRACE_SEL_PB_CONTROL, "%s: interval=%d, cs=%d, ce=%d last_frame_num=%d(%d)", __PRETTY_FUNCTION__, interval, clip_start, clip_end, last_frame_num, rel_frame);

    // Update our frame start address.
    if ((old_control & GIGE_PBC_REVERSE) == GIGE_PBC_REVERSE)
    {
        // Playing in reverse.
        play_frame = rel_frame - interval;
        if (play_frame < clip_start)
            play_frame = clip_end;
    }
    else
    {
        // Playing forward.
        play_frame = rel_frame + interval;
        if (play_frame > clip_end)
        {
            // Are we doing a single playthrough?
            if (do_single)
            {
                // We are done.
                TP(TRACE_SEL_PB_CONTROL, "%s: single playthough finished. Leaving.", __PRETTY_FUNCTION__);
                state = PB_IDLE;
                goto bail;
            }

            // Update the frame.
            play_frame = clip_start;
        }
    }

    raw_frame_start = pm_ptr->get_active_frame_addr(play_frame, false);
    TP(TRACE_SEL_PB_CONTROL, "%s: play_frame=%d raw_frame_start=0x%" PRIx64, __PRETTY_FUNCTION__, play_frame, raw_frame_start);

    // What is our next state?
    if (ltr_mode)
        state = PB_LTR_FRAME_LOAD;
    else
        state = PB_PROCESS_FRAME;

bail:
    TP(TRACE_SEL_PB_CONTROL, "%s--", __PRETTY_FUNCTION__);
}

void pbcThread::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Keep looping until we're told to stop.
    while (!done)
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: state=%d", __PRETTY_FUNCTION__, state);

        // Process based on the current state.
        switch (state)
        {
        case PB_IDLE:
            // Wait for someone to want to play.
            idle();
            break;
        case PB_SETUP:
            // Initialize ourselves.
            if (setup() < 0)
            {
                TP(TRACE_SEL_PB_CONTROL, "%s: setup() failed, back to idle.", __PRETTY_FUNCTION__);
                state = PB_IDLE;
            }
            else
            {
                // Start timing our playback frame rate.
                pb_timer.start();

                // Where do we go from here?
                if (ltr_mode)
                {
                    // If we're in LTR mode, we need to load up the frame from the SSD.
                    state = PB_LTR_FRAME_LOAD;
                }
                else
                {
                    // Must be "normal". Just process the image.
                    state = PB_PROCESS_FRAME;
                }
            }
            break;
        case PB_LTR_FRAME_LOAD:
            if (ltr_frame_load() < 0)
            {
                TP(TRACE_SEL_PB_CONTROL, "%s: ltr_frame_load() failed, back to idle.", __PRETTY_FUNCTION__);
                state = PB_IDLE;
            }
            else
            {
                // Go process the frame.
                state = PB_PROCESS_FRAME;
            }
            break;
        case PB_PROCESS_FRAME:
            if (process_frame() < 0)
            {
                TP(TRACE_SEL_PB_CONTROL, "%s: process_frame() failed, back to idle.", __PRETTY_FUNCTION__);
                state = PB_IDLE;
            }
            else
            {
                // Go onwards and send the frame to all waiters.
                state = PB_FRAME_READY;
            }
            break;
        case PB_FRAME_READY:
            frame_ready();
            state = PB_WAIT;
            break;
        case PB_WAIT:
            wait();
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state [%d]", __PRETTY_FUNCTION__, state);
            break;
        }
    }
}

//-----------------------------------------------------------------------------
//! @name Register Access Functions.
//! @{

/*!
 * @brief Function to cache the GIGE_REG_PB_xxx writes.
 */
static int pbc_update(uint32_t reg_addr, uint32_t reg_value)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Process based on the register.
    switch (reg_addr)
    {
    case GIGE_REG_PB_CONTROL:
    case GIGE_REG_PB_FRAME_PERIOD:
    case GIGE_REG_PB_CUT_IN_FRAME:
    case GIGE_REG_PB_CUT_OUT_FRAME:
    case GIGE_REG_PB_START_FRAME:
        TP(TRACE_SEL_PB_CONTROL, "%s: addr [0x%08x], val [0x%08x].", __PRETTY_FUNCTION__, reg_addr, reg_value);
        regs[reg_addr].set_value(reg_value);

        // Let folks know about the change in the register settings.
        fcp_notify_apps(reg_addr, reg_value);
        break;
    default:
        fi_log(LOG_ERR, "%s: Unexpected address [0x%08x]!", __PRETTY_FUNCTION__, reg_addr);
        break;
    }

    // All done.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update the playback control register.
 *
 * Make certain all other parameters required for a successful playback
 * are set prior to calling this function!
 *
 * @param p_msg     [UNUSED] The FCP message.
 * @param reg_value The new playback control settings.
 * @returns The status value (GEV_STATUS_SUCCESS, or other).
 */
int pbc_update_control(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    int status = GEV_STATUS_SUCCESS;
    CameraServer *cs_ptr = CameraServer::get_cs();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    TP(TRACE_SEL_PB_CONTROL, "%s: 0x%08x", __PRETTY_FUNCTION__, reg_value);

    // We defined the GIGE_PBC_ bits so they match the FPGA register.
    if ((reg_value & GIGE_PBC_ENABLE) == (uint32_t)GIGE_PBC_ENABLE)
    {
        // Don't enable playback if we are not in review.
        if (CameraServer::get_camera_state() != GIGE_CAMERA_STATE_REVIEW)
        {
            status = GEV_STATUS_WRONG_CONFIG;
            fi_log(LOG_ERR, "%s: Not in REVIEW (%u) writing PB_CONTROL=0x%x", __PRETTY_FUNCTION__,
                   CameraServer::get_camera_state(), reg_value);
        }

        // Are we in DVR mode?
        else
        {
            uint32_t pb_status = GIGE_PBS_BUSY;

            if (cs_ptr->auto_trigger() || cs_ptr->dvr_playback())
            {
                uint32_t pref;

                // Make sure AutoPlay is enabled.
                pref = regs[ GIGE_REG_PREFERENCES ].get_value();
                if ((pref & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY)
                {
                    // Modify the PB_CONTROL register to use DVR mode.
                    // * Make sure the PB_CONTROL[PB_ENABLE] is cleared. Don't want that in DVR mode.
                    reg_value &= ~GIGE_PBC_ENABLE;

                    // * Make sure the PB_CONTROL[PLAYBACK_DIR] is set to 0 (forward).
                    reg_value &= ~GIGE_PBC_REVERSE;

                    // * Make sure the PB_CONTROL[PLAYBACK_SINGLE_START] is set to 1 (DVR mode).
                    reg_value |= GIGE_PBC_SINGLE_START;
                    pb_status |= GIGE_PBS_SINGLEMODE;

                    // DVR Playback mode specific stuff...
                    if (cs_ptr->dvr_playback())
                    {
                        // * Make sure the PB_CONTROL[PLAYBACK_INTERVAL] is set to 1 (no skipping).
                        reg_value &= ~GIGE_PBC_INTERVAL_MASK;
                        reg_value |= (1 << GIGE_PBC_INTERVAL_SHIFT);
                    }

                    // * Make sure the PB_START_PTR is set to PB_CUT_IN_FRAME (not trigger!).
                    pbc_update(GIGE_REG_PB_START_FRAME, regs[GIGE_REG_PB_CUT_IN_FRAME].get_value());
                }
            }

            // Update our register cache.
            regs[GIGE_REG_PB_STATUS].set_value(pb_status);
            fcp_notify_apps(GIGE_REG_PB_STATUS, pb_status);
        }
    }
    else
    {
        // Make sure we stop the single playback mode too.
        reg_value |= GIGE_PBC_SINGLE_ABORT;

        // Tell the camera server to stop polling the DVR status.
        // We want to do that here, so that the poller won't detect
        // the playback stop and automatically put us back into the
        // armed state.
        cs_ptr->poll_for_dvr_finish(false);

        // Everything is fine so far...
        status = GEV_STATUS_SUCCESS;
    }

    // Should we keep going?
    if (status == GEV_STATUS_SUCCESS)
    {
        // Update the control value in our register database.
        status = pbc_update(GIGE_REG_PB_CONTROL, reg_value);
        if ((status == GEV_STATUS_SUCCESS) && (pbcThread::factory() != nullptr))
        {
            pbcControlUpdateMessage *update_msg;

            // Allocate a control message to send to the thread.
            update_msg = new pbcControlUpdateMessage(pbcThread::factory());
            if (update_msg == nullptr)
            {
                fi_log(LOG_ERR, "%s: Unable to allocate control message!", __PRETTY_FUNCTION__);
                return GEV_STATUS_ERROR;
            }

            // Set the new control value.
            update_msg->control_value = reg_value;

            // Send the message on it's way.
            update_msg->forward();
        }
        else
        {
            TP(TRACE_SEL_PB_CONTROL, "%s: pbc_update problem, status = 0x%08x", __PRETTY_FUNCTION__, status);
        }
    }
    else
    {
        TP(TRACE_SEL_PB_CONTROL, "%s: status = 0x%08x", __PRETTY_FUNCTION__, status);
    }

    return status;
}

int pbc_update_frame_period(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    return pbc_update(GIGE_REG_PB_FRAME_PERIOD, reg_value);
}

int pbc_update_cut_in(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    return pbc_update(GIGE_REG_PB_CUT_IN_FRAME, reg_value);
}

int pbc_update_cut_out(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    return pbc_update(GIGE_REG_PB_CUT_OUT_FRAME, reg_value);
}

int pbc_update_start_frame(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
    return pbc_update(GIGE_REG_PB_START_FRAME, reg_value);
}

//! @}
// End of register access functions.
