/*!
 * @file pm_worker.cpp
 * @brief This source file implements the Fastec Imaging Partition Manager's
 *        worker class, pmWorker.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include <chrono>
#include <mutex>

#include "pm_worker.h"

#include "camera_server.h"
#include "fpga.h"
#include "part_mgr.h"

namespace PART_MGR_WORKER
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* End of namespace PART_MGR_WORKER */


using namespace PART_MGR_WORKER;

//--------------------------------------------------------------------
// Static Variable Instantiations
//--------------------------------------------------------------------

std::unique_ptr<pmWorker> pmWorker::worker_proc;
std::mutex pmWorker::worker_mutex;

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

//! @name Message class methods
//! @{

pmwInitMessage::pmwInitMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_INITIALIZE_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    num_parts = 0;
}

pmwInitMessage::~pmwInitMessage()
{
}

pmwGetInfoMessage::pmwGetInfoMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_GET_INFO_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    part_index = -1;
    memset(&info, 0, sizeof(info));
}

pmwGetInfoMessage::~pmwGetInfoMessage()
{
}

pmwSetInfoMessage::pmwSetInfoMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_SET_INFO_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    part_index = -1;
    memset(&info, 0, sizeof(info));
}

pmwSetInfoMessage::~pmwSetInfoMessage()
{
}

pmwGetActivePartMessage::pmwGetActivePartMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_GET_ACTIVE_PART_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    active_index = -1;
    memset(&active_info, 0, sizeof(active_info));
}

pmwGetActivePartMessage::~pmwGetActivePartMessage()
{
}

pmwSetActivePartMessage::pmwSetActivePartMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_SET_ACTIVE_PART_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    active_index = -1;
}

pmwSetActivePartMessage::~pmwSetActivePartMessage()
{
}


pmwGetReviewPartMessage::pmwGetReviewPartMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_GET_REVIEW_PART_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    review_index = -1;
    memset(&review_info, 0, sizeof(review_info));
}

pmwGetReviewPartMessage::~pmwGetReviewPartMessage()
{
}

pmwSetReviewPartMessage::pmwSetReviewPartMessage() :
    SyncMessage()
{
    req = (FAS_THREAD::REQ_CODE)PMW_SET_REVIEW_PART_MESSAGE;
    resp = FAS_THREAD::RESP_CODE::FAULT;
    review_index = -1;
}

pmwSetReviewPartMessage::~pmwSetReviewPartMessage()
{
}

//! @}
// End of message class methods.

//--------------------------------------------------------------------
// fpgaIOWorker class implementation
//--------------------------------------------------------------------

pmWorker::pmWorker() :
    FAS_THREAD::MsgThread(),
    reviewPartition(-1),
    activePartition(-1)
{
    // Reset our partition information vector to a known state.
    part_info.clear();

}

pmWorker::~pmWorker()
{
    LOG("%s", __PRETTY_FUNCTION__);
}

pmWorker *pmWorker::factory()
{
    pmWorker *rc(nullptr);

    std::lock_guard<std::mutex> lock(worker_mutex);

    // We only allow one thread object.
    if (worker_proc.get() == nullptr)
    {
        // Create our thread.
        worker_proc = std::unique_ptr<pmWorker>(new pmWorker());
        if (worker_proc.get() != nullptr)
            worker_proc.get()->start();

        // Return the pointer to the singleton.
        rc = worker_proc.get();
    }

    // Check to see if we were halted.
    else if (worker_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = worker_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Process the messages for the thread.
 */
void pmWorker::messageHandle(FAS_THREAD::Message *message)
{
    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case PMW_INITIALIZE_MESSAGE:
            handleInitRequest(static_cast<pmwInitMessage *>(message));
            break;
        case PMW_GET_INFO_MESSAGE:
            handleGetInfoRequest(static_cast<pmwGetInfoMessage *>(message));
            break;
        case PMW_SET_INFO_MESSAGE:
            handleSetInfoRequest(static_cast<pmwSetInfoMessage *>(message));
            break;
        case PMW_GET_ACTIVE_PART_MESSAGE:
            handleGetActiveRequest(static_cast<pmwGetActivePartMessage *>(message));
            break;
        case PMW_SET_ACTIVE_PART_MESSAGE:
            handleSetActiveRequest(static_cast<pmwSetActivePartMessage *>(message));
            break;
        case PMW_GET_REVIEW_PART_MESSAGE:
            handleGetReviewRequest(static_cast<pmwGetReviewPartMessage *>(message));
            break;
        case PMW_SET_REVIEW_PART_MESSAGE:
            handleSetReviewRequest(static_cast<pmwSetReviewPartMessage *>(message));
            break;

        case FAS_THREAD::DUMP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            // Remember that we've stopped.
            halted = true;

            // All done.
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Do we still have a message?
        if (message != nullptr)
        {
            // Forward on the messsage.
            message->forward();
        }
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in fpgaIOWorker:\n%s", e.what().c_str());

        if (message != nullptr)
        {
            message->resp = FAS_THREAD::FAULT;
            message->forward();
        }
    }
}

/*!
 * @brief Handle processing for the pmwInitMessage message.
 */
void pmWorker::handleInitRequest(pmwInitMessage *msg)
{
    PartMgr *part_mgr;
    FAS_THREAD::RESP_CODE resp = FAS_THREAD::OK;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
        goto init_bail;
    }

    // Are we resetting the information vector?
    if (msg->num_parts == 0)
    {
        activePartition = -1;
        reviewPartition = -1;
        part_info.clear();
        goto init_bail;
    }

    // Get our partition manager object.
    part_mgr = PartMgr::getInstance();
    if (part_mgr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null partition manager?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
        goto init_bail;
    }

    // Walk through the partition structure and initialize it.
    part_info.resize(msg->num_parts);
    for (int i = 0; i < msg->num_parts; i++)
    {
        // Calculate the start address for the partition and make sure
        // we're properly aligned.
        part_info[i].partition_start = part_mgr->get_im_size_reserved() + (part_mgr->getPartitionSize() * i);
        part_info[i].partition_start = (uint64_t)convert_to_HEXWORD(part_info[i].partition_start) << HEXWORD_SHIFT;

        // Calculate the starting location for the image data.
        part_info[i].capture_buf_start = part_info[i].partition_start + part_mgr->getPartitionSizeReserved();
        part_info[i].partition_state = GIGE_PART_STATE_EMPTY;
        part_info[i].num_frames = part_mgr->get_im_buf_count();     // assume we will fill up the whole partition, will be udated when part complete
        part_info[i].frame_size = part_mgr->getFrameSize();         // may get overwritten when partition complete
        part_info[i].first_frame_num = 0;
        part_info[i].is_trig_part = false;
    }

    // Default the active and review partitions.
    activePartition = 0;
    reviewPartition = -1;

init_bail:
    // All done.
    msg->resp = resp;
}

/*!
 * @brief Handle processing for the pmwGetInfoMessage message.
 */
void pmWorker::handleGetInfoRequest(pmwGetInfoMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }

    // Is the requested partition index valid?
    else if ((msg->part_index < 0) || (msg->part_index >= (int)part_info.size()))
    {
        fi_log(LOG_DEBUG, "%s: The requested index (%d) is invalid!", __PRETTY_FUNCTION__, msg->part_index);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Copy over the data.
        memcpy(&msg->info, &part_info[msg->part_index], sizeof(PartInfo));

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}

/*!
 * @brief Handle processing for the pmwSetInfoMessage message.
 */
void pmWorker::handleSetInfoRequest(pmwSetInfoMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }

    // Is the specified partition index valid?
    else if ((msg->part_index < 0) || (msg->part_index >= (int)part_info.size()))
    {
        fi_log(LOG_DEBUG, "%s: The specified index (%d) is invalid!", __PRETTY_FUNCTION__, msg->part_index);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Copy over the data.
        memcpy(&part_info[msg->part_index], &msg->info, sizeof(PartInfo));

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}

/*!
 * @brief Handle processing for the pmwGetActivePartMessage message.
 */
void pmWorker::handleGetActiveRequest(pmwGetActivePartMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Copy over our active partition index.
        msg->active_index = activePartition;

        // Is the active partition actually set? If not, just clear out the return data.
        // Otherwise, copy over the data.
        if (activePartition < 0)
            memset(&msg->active_info, 0, sizeof(PartInfo));
        else
            memcpy(&msg->active_info, &part_info[activePartition], sizeof(PartInfo));

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}


/*!
 * @brief Handle processing for the pmwSetActivePartMessage message.
 */
void pmWorker::handleSetActiveRequest(pmwSetActivePartMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }

    // Is the specified partition index valid?
    else if ((msg->active_index < 0) || (msg->active_index >= (int)part_info.size()))
    {
        fi_log(LOG_DEBUG, "%s: The specified index (%d) is invalid!", __PRETTY_FUNCTION__, msg->active_index);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Save the new active partition.
        activePartition = msg->active_index;

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}

/*!
 * @brief Handle processing for the pmwGetReviewPartMessage message.
 */
void pmWorker::handleGetReviewRequest(pmwGetReviewPartMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Copy over our review partition index.
        msg->review_index = reviewPartition;

        // Is the review partition actually set? If not, just clear out the return data.
        // Otherwise, copy over the data.
        if (reviewPartition < 0)
            memset(&msg->review_info, 0, sizeof(PartInfo));
        else
            memcpy(&msg->review_info, &part_info[reviewPartition], sizeof(PartInfo));

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}


/*!
 * @brief Handle processing for the pmwSetReviewPartMessage message.
 */
void pmWorker::handleSetReviewRequest(pmwSetReviewPartMessage *msg)
{
    FAS_THREAD::RESP_CODE resp;

    // Make sure we have a message.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg?!", __PRETTY_FUNCTION__);
        resp = FAS_THREAD::FAULT;
    }

    // Is the specified partition index valid?
    else if ((msg->review_index < 0) || (msg->review_index >= (int)part_info.size()))
    {
        fi_log(LOG_DEBUG, "%s: The specified index (%d) is invalid!", __PRETTY_FUNCTION__, msg->review_index);
        resp = FAS_THREAD::FAULT;
    }
    // Make sure the partition's state is OK.
    else if ((part_info[msg->review_index].partition_state != GIGE_PART_STATE_COMPLETE) && (part_info[msg->review_index].partition_state != GIGE_PART_STATE_LOCKED))
    {
        fi_log(LOG_DEBUG, "%s: partition [%d] is in the wrong state [0x%x]!", __PRETTY_FUNCTION__, msg->review_index, part_info[msg->review_index].partition_state);
        resp = FAS_THREAD::FAULT;
    }
    else
    {
        // Save the new review partition.
        reviewPartition = msg->review_index;

        // All is well.
        resp = FAS_THREAD::OK;
    }

    // All done.
    msg->resp = resp;
}
