/*!
 * @file save_proc.cpp
 * @brief Manage threaded save processing.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/limits.h>
#include <net/if_arp.h>
#include <time.h>

#include <fastec/fas_thread.h>
#include <fastec/fi_ioctl.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>
#include <fastec/fas_sharedmem.h>

#include <fastec/fas_video.h>
#include <fastec/fi_image.h>
#include <fastec/fi_bmp.h>
#include <fastec/fi_bmp_planar.h>
#include <fastec/fi_jpeg.h>
#include <fastec/fi_image_pkg.h>

#include <atomic>
#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>
#include <queue>
#include <fmt/printf.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_trace.h"
#include "fcp.h"
#include "fi_internal.h"
#include "image_format.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "save_proc.h"
#include "part_mgr.h"
#include "storage.h"

namespace SAVE_PROC
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Maximum number of times we'll retry getting a JPEG buffer.
const int MAX_JPEG_RETRIES = 20;

//! Message types for the save processor.
enum SAVE_MESSAGE_TYPES
{
    SAVE_TO_LOCAL = 0x30000,        //!< Save a sequence of raw images to local media.
    SAVE_FRAME                      //!< Message to process a single frame.
};

//! Message class for SAVE_TO_LOCAL messages.
class SaveLocalMessage : public FAS_THREAD::SyncMessage
{
public:
    SaveLocalMessage();
    virtual ~SaveLocalMessage();

    void assign(SAVE_PROC::SAVE_REQUEST *request);
    const SAVE_PROC::SAVE_REQUEST *getRequest() { return request_info.get(); }

    void setSubfolderCount(int n) { request_info->subfolder_count = n; }

    int status;

private:
    std::unique_ptr<SAVE_PROC::SAVE_REQUEST> request_info;
};

/*!
 * @brief Message class for SAVE_FRAME messages.
 *
 * Note that this is an asynchronous message. It is expected to
 * be routed/forward from one thread to the next as the frame
 * is processed.
 */
class SaveFrameMessage : public FAS_THREAD::Message
{
public:
    SaveFrameMessage();
    virtual ~SaveFrameMessage();

    int status;             //!< 0, or errno on error.

    uint32_t frame_num;         //!< Frame number associated with this message.
    uint32_t start_frame;       //!< Starting (inclusive) frame for the save.
    uint32_t end_frame;         //!< Ending (inclusive) frame for the save.

    uint8_t *buf_ptr;           //!< Pointer to our buffer.
    uint64_t buf_addr;          //!< Physical address of our buffer for DMA. MUST BE PAGE ALIGNED!
    uint64_t image_addr;        //!< Address in Image Memory for the frame to encode.
    uint64_t ltr_im_addr;       //!< Address in Image Memory for our raw LTR frame.

    bool ltr_mode;              //!< True if we're in LTR mode.
    uint32_t proc_info;         //!< IPM processing information.
    uint32_t dma_mod;           //!< DMA options.
    uint32_t dma_dims;          //!< DMA dimensions.
    uint32_t ipm_image_len;     //!< Length of processed image.
    uint64_t ipm_buf_addr;      //!< Buffer address for processed image in image memory.
    uint8_t *jpg_buf_ptr;       //!< Pointer to our JPEG encoded buffer, or nullptr.

    bool shmem_posted;          //!< Flag to indicate if we need to wait for the client to finish.
    uint32_t shmem_index;       //!< Index in the ring buffer of this image.

    //! If we're creating a video, we need a stream writer.
    std::weak_ptr<FAS_VIDEO::FiVideoStream> video_writer;
    uint32_t video_codec;       //!< The video codec being used (i.e. JPEG, BMP, Planar).
    FAS_INTERNAL::FIP_HANDLE pack_handle; //!< Handle to the packed streamer.

    std::string image_path;     //!< Used in the snprintf() as the format string. Just add frame_num.
    std::vector<uint8_t> header; //!< The image header, if any.

    // Some statistics
    std::vector<uint64_t> times_usec;   //!< An array of 'durations' for the various threads.
};

/*!
 * @brief Thread to load images from the SSD to image memory.
 */
class ltrThread : public FAS_THREAD::MsgThread
{
public:
    ltrThread();
    virtual ~ltrThread();

    void setBbAddr(uint64_t bb_addr)  { ltr_bb_addr = bb_addr; }
    void setBbPtr(uint8_t *bb_ptr)    { ltr_bb_ptr = bb_ptr; }

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    bool handleSaveFrameRequest(SaveFrameMessage *msg);
    int ltrFrameLoad(SaveFrameMessage *msg);

    uint64_t ltr_bb_addr;   //!< Physical address of our buffer for DMA to image memory.
    uint8_t *ltr_bb_ptr;    //!< Pointer to the buffer.
};

/*!
 * @brief Thread for performing IPM processing.
 */
class ipmThread : public FAS_THREAD::MsgThread
{
public:
    ipmThread();
    virtual ~ipmThread();

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    bool handleSaveFrameRequest(SaveFrameMessage *msg);
};

/*!
 * @brief Thread for JPEG encoding images.
 */
class jpgThread : public FAS_THREAD::MsgThread
{
public:
    jpgThread();
    virtual ~jpgThread();

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    bool handleSaveFrameRequest(SaveFrameMessage *msg);
};

/*!
 * @brief The shared memory checker thread constructor. The thread
 *    is a "detached" thread. That is, it runs to completion (once started)
 *    and does not expect to "join" with any other thread.
 */
class shmemChecker : public FAS_THREAD::Thread
{
public:
    explicit shmemChecker(FSM_HANDLE shmem_handle = nullptr) :
        FAS_THREAD::Thread(true),
        thread_running(false),
        hSharedMem(shmem_handle)
    {
    }
    ~shmemChecker() = default;

    virtual void run() override;
    bool isRunning()   { return thread_running; }

    void addMessage(SaveFrameMessage *message);
    void setHandle(FSM_HANDLE shmem_handle)     { hSharedMem = shmem_handle; }

private:
    volatile bool thread_running;               //!< Flag to indicate if we are running or not.

    std::queue<SaveFrameMessage *> wait_queue;  //!< Queue of messages waiting to become available.
    std::mutex queue_mutex;                     //!< Protect the wait_queue.
    FSM_HANDLE hSharedMem;                      //!< Handle to the shared memory interface.
};

/*!
 * @brief Thread for shared memory destined images.
 */
class shmemThread : public FAS_THREAD::MsgThread
{
public:
    explicit shmemThread(uint64_t im_base_addr, uint64_t im_length);
    virtual ~shmemThread();

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    bool handleSaveFrameRequest(SaveFrameMessage *msg, bool& forward_msg);

private:
    FSM_HANDLE hSharedMem;                      //!< Handle to the shared memory interface.
    std::unique_ptr<shmemChecker> bufferChecker;  //!< A thread that waits for a buffer to be released.
};

/*!
 * @brief Thread to write the image to disk.
 */
class writeThread : public FAS_THREAD::MsgThread
{
public:
    writeThread();
    virtual ~writeThread();

    constexpr uint64_t getTotalBytes()      { return total_bytes; }

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    int checkSpace(std::weak_ptr<FAS_VIDEO::FiVideoStream> video_writer, uint32_t image_len, uint32_t frame_count, const char *target_path);
    bool handleSaveFrameRequest(SaveFrameMessage *msg);
    bool saveFrame(SaveFrameMessage *msg);
    bool saveVideoFrame(SaveFrameMessage *msg, FAS_VIDEO::FiVideoStream *vid_stream);
    bool saveImageFrame(SaveFrameMessage *msg);
    bool saveToPackedStream(SaveFrameMessage *msg);

    //! Maintiain the total number of bytes written, for statistical purposes.
    uint64_t total_bytes;

    //! A reusable AVI image object.
    std::shared_ptr<FAS_VIDEO::FiImage> videoImage;
};

/*!
 * @brief An asynchronous thread which periodically broadcasts the
 *        progress of the save thread.
 */
class progressThread : public FAS_THREAD::Ticker
{
public:
    progressThread() : FAS_THREAD::Ticker(DEFAULT_UPDATE_TIME)
    {
        progress_value = 0;
        last_progress = 0;
    }

    ~progressThread()
    {
    }

    void setProgress(uint32_t new_value)    { progress_value = new_value; }

    void tick()
    {
//        LOG("%s: last=%u prog=%u", __PRETTY_FUNCTION__, last_progress, progress_value);

        // Has there been a change?
        if (last_progress != progress_value)
        {
            // Remember the current progress.
            last_progress.store(progress_value);

            // Let folks know where we are.
            fcp_notify_apps(GIGE_REG_PROGRESS, progress_value);
        }
    }

private:
    static const int DEFAULT_UPDATE_TIME;

    std::atomic_uint32_t progress_value;
    std::atomic_uint32_t last_progress;
};

//! Default update time in microseconds.
const int progressThread::DEFAULT_UPDATE_TIME = 100000;

/*!
 * @brief The save processor control object. This object is
 *    a singleton and is "local" to this file.
 */
class saveThread : public FAS_THREAD::MsgThread
{
public:
    virtual ~saveThread();
    static saveThread *factory(bool auto_start = true);
    static void reset();

    static bool user_cancel;                //!< Flag to set if the user cancels the operation.

    // Singletons should not be cloneable.
    saveThread(saveThread& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const saveThread&) = delete;

    void cancel();
    void close_stats();
    void open_stats(const char *save_path);
    constexpr void restart()        { user_cancel = false; }

    int status;         //!< 0, or errno on error.

protected:
    void messageHandle(FAS_THREAD::Message *message) override;
    void ownerMessage(FAS_THREAD::Message *message) override;

private:
    saveThread();

    bool buildMessage(SaveFrameMessage *msg);
    bool check_subfolder(SaveFrameMessage *msg);
    bool handleSaveLocalRequest(SaveLocalMessage *msg);

private:
    static std::unique_ptr<saveThread> save_proc;   //!< Our singleton object.
    static std::mutex save_proc_mutex;              //!< Mutex to protect our singleton object.

    bool faulted;                           //!< Flag set to remember if any of the frames failed.
    std::mutex avail_q_mutex;               //!< Used to control changes to the available queue.
    std::condition_variable avail_q_status; //!< Condition variable we can block on.
    std::queue<SaveFrameMessage *> avail_q; //!< Available message structures.

    int bb_index;                           //!< Bounce buffer index.
    uint64_t bb_addr;                       //!< Address for the bounce buffer.
    int buffer_count;                       //!< The number of image buffers we fit in the bounce buffer.
    FILE *stat_fp;                          //!< File to save statistics information to.

    SaveLocalMessage *sl_msg;               //!< The SaveLocalMessage we are processing.
    uint32_t cur_frame;

    std::unique_ptr<ltrThread> ltr_thread;
    std::unique_ptr<ipmThread> ipm_thread;
    std::unique_ptr<jpgThread> jpg_thread;
    std::unique_ptr<progressThread> progress_thread;
    std::unique_ptr<shmemThread> shmem_thread;
    std::unique_ptr<writeThread> write_thread;
};

} /* End of namespace SAVE_PROC */

using namespace CS_TRACE;
using namespace SAVE_PROC;

// #define DO_STATISTICS 1

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<saveThread> saveThread::save_proc;
std::mutex saveThread::save_proc_mutex;
bool saveThread::user_cancel(false);

TRACE_GLOBAL;           //!< Use the global trace control word.

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

void saveProcDump(std::ostream& dump_stream)
{
    FAS_THREAD::SyncMessage msg;

    // Sanity check.
    if (saveThread::factory() == nullptr)
    {
        // Got a request as we're shutting down.
        return;
    }

    // Specify a dump request message.
    msg.req = FAS_THREAD::DUMP;
    msg.stream = std::unique_ptr<std::ostream>(&dump_stream);

    // Send it on.
    msg.doSync(saveThread::factory());
}

void saveProcInit()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_SAVE_PROC;

    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure we have our singleton.
    if (saveThread::factory() == nullptr)
        saveThread::reset();

    // Make sure we are reset and ready to work.
    saveThread::factory()->restart();
}

void saveProcClose()
{
    saveThread *thread_ptr;

    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Release our singleton. Don't start the object if it's already closed.
    thread_ptr = saveThread::factory(false);
    if (thread_ptr != nullptr)
    {
        // Tell the thread that we are stopping.
        FAS_THREAD::SyncMessage msg;
        msg.req = FAS_THREAD::HALT;

        // Pass it on.
        msg.doSync(thread_ptr);

        // Try to join the thread.
        thread_ptr->join();

        // Reset our saver object.
        saveThread::reset();
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: done", __PRETTY_FUNCTION__);
}

void saveProcCancel()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Make sure we still have a singleton.
    if (saveThread::factory() != nullptr)
    {
        // Ask the object to cancel.
        saveThread::factory()->cancel();
    }
}

/*!
 * @brief Entry to threaded saving of a sequence of images.
 */
int saveProcRequest(SAVE_PROC::SAVE_REQUEST *request)
{
    int rc(0);
    std::unique_ptr<SaveLocalMessage> msg;

    // Sanity check
    if (request == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null request!?", __PRETTY_FUNCTION__);
        return EINVAL;
    }
    if (saveThread::factory() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null save_proc. Did you save_proc_init()?", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: start=%d, end=%d, proc_info=0x%x",
       __PRETTY_FUNCTION__, request->start_frame, request->end_frame, request->proc_info);

#ifdef DO_STATISTICS
    save_proc->open_stats("/tmp/save_proc.csv");
#endif /* DO_STATISTICS */

    // Allocate our message.
    msg = std::unique_ptr<SaveLocalMessage>(new SaveLocalMessage());
    if (msg.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Can't allocate SaveLocalMessage!", __PRETTY_FUNCTION__);
        return ENOMEM;
    }

    // Build up our message.
    msg->assign(request);

    // Send it on.
    msg->doSync(saveThread::factory());

#ifdef DO_STATISTICS
    save_proc->close_stats();
#endif /* DO_STATISTICS */

    // Success?
    if (msg->resp == FAS_THREAD::OK)
    {
        // Copy over the output values.
        TP(TRACE_SEL_SAVE_PROC, "%s: All done.", __PRETTY_FUNCTION__);
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: request failed, err=%d(%s)", __PRETTY_FUNCTION__, msg->status, strerror(msg->status));
        rc = msg->status;
    }

    // Return our status.
    return rc;
}

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

SaveLocalMessage::SaveLocalMessage() :
    SyncMessage(),
    status(0)
{
    req = (FAS_THREAD::REQ_CODE)SAVE_TO_LOCAL;
}

SaveLocalMessage::~SaveLocalMessage()
{
}

void SaveLocalMessage::assign(SAVE_PROC::SAVE_REQUEST *request)
{
    // Allocate our request buffer.
    request_info = std::unique_ptr<SAVE_PROC::SAVE_REQUEST>(new SAVE_PROC::SAVE_REQUEST);
    if (request_info.get() == nullptr)
    {
        FI_LOGIC_ERROR("Unable to allocate the request structure!");
        return;
    }

    // Copy over the values.
    request_info->start_frame = request->start_frame;
    request_info->end_frame = request->end_frame;
    request_info->ltr_mode = request->ltr_mode;
    request_info->proc_info = request->proc_info;
    request_info->dma_mod = request->dma_mod;
    request_info->image_height = request->image_height;
    request_info->image_width = request->image_width;
    request_info->video_writer = request->video_writer;
    request_info->video_codec = request->video_codec;
    request_info->pack_handle = request->pack_handle;
    request_info->header_len = request->header_len;
    request_info->header_ptr = request->header_ptr;
    request_info->info_ptr = request->info_ptr;
    request_info->do_subfolders = request->do_subfolders;
    request_info->subfolder_count = request->subfolder_count;
    request_info->using_sharedmem = request->using_sharedmem;

    // Need to treat the std::string differently.
    request_info->path_str.assign(request->path_str);
    LOG("%s: path_str='%s'", __PRETTY_FUNCTION__, request_info->path_str.c_str());
}

SaveFrameMessage::SaveFrameMessage() :
    Message(),
    status(0),
    frame_num(0),
    start_frame(0),
    end_frame(0),
    ltr_im_addr(0),
    ltr_mode(false),
    proc_info(0),
    dma_mod(0),
    dma_dims(0),
    ipm_image_len(0),
    ipm_buf_addr(0),
    shmem_posted(false),
    shmem_index((uint32_t) -1),
    video_codec(0)
{
    req = (FAS_THREAD::REQ_CODE)SAVE_FRAME;
    buf_ptr = nullptr;
    jpg_buf_ptr = nullptr;
    buf_addr = 0;
    image_addr = 0;
    status = 0;
}

SaveFrameMessage::~SaveFrameMessage()
{
}

//--------------------------------------------------------------------
// ltrThread Implementation
//--------------------------------------------------------------------

ltrThread::ltrThread() :
    FAS_THREAD::MsgThread(),
    ltr_bb_addr(0),
    ltr_bb_ptr(nullptr)
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

ltrThread::~ltrThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Process the messages for the thread.
 */
void ltrThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_FRAME:
            // No processing of messages if we're canceling.
            if (saveThread::user_cancel)
                message->resp = FAS_THREAD::FAULT;

            else if (!handleSaveFrameRequest(static_cast<SaveFrameMessage *>(message)))
                message->resp = FAS_THREAD::FAULT;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
        case FAS_THREAD::HALT:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in ltrThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

bool ltrThread::handleSaveFrameRequest(SaveFrameMessage *msg)
{
    int status;
    bool rc(true);
    double elapsed;
    uint64_t elapsed_usecs;
    struct timespec start_time;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    // Get our starting time for the statistics output.
    measuretime(1, &start_time);

    // We need to load the frame from the SSD up into Image Memory so that
    // the IPM can process the frame.
    status = ltrFrameLoad(msg);
    if (status != 0)
    {
        fi_log(LOG_DEBUG, "%s: ltr_frame_load() failed!", __PRETTY_FUNCTION__);
        msg->status = status;
        rc = false;
    }

    // Output our statistics.
    elapsed_usecs = measuretime(0, &start_time);
    elapsed = (double)elapsed_usecs / 1.0e6;
    msg->times_usec.push_back(elapsed_usecs);
    TP(TRACE_SEL_SAVE_PROC, "%s: frame %d [%6.3lf]", __PRETTY_FUNCTION__, msg->frame_num, elapsed);

    // Return our status.
    return rc;
}

int ltrThread::ltrFrameLoad(SaveFrameMessage *msg)
{
    int rc = -1;
    uint32_t raw_length;
    uint32_t dma_mod;
    uint64_t frame_offset;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%u", __PRETTY_FUNCTION__, msg->frame_num);

    // Get the SSD offset in bytes.
    frame_offset = pm_ptr->get_active_frame_addr(msg->frame_num, false);
    TP(TRACE_SEL_SAVE_PROC, "%s: frame_offset=0x%" PRIx64, __PRETTY_FUNCTION__, frame_offset);

    // Get the length of the image. This also includes the PFMD, but that's OK.
    raw_length = pm_ptr->getFrameSize();

    // Read up the image data.
    if (pm_ptr->ltrReadBytes(ltr_bb_ptr, frame_offset, raw_length) != (int)raw_length)
    {
        fi_log(LOG_ERR, "%s: Unable to read from 0x%" PRIx64 " for %u bytes", __PRETTY_FUNCTION__, frame_offset, raw_length);
        goto bail;
    }

    // Now ask the FPGA to put the data into the image memory buffer.
    dma_mod = PCIE2IM_DMA_BYTE_ORDER_R2L | PCIE2IM_DMA_WORD_ORDER_L2R;
    TP(TRACE_SEL_SAVE_PROC, "%s: dma_mod=0x%08x, len=%u", __PRETTY_FUNCTION__, dma_mod, raw_length);
    rc = fpga_ptr->imageUpload(dma_mod, ltr_bb_addr, msg->ltr_im_addr, raw_length);
    if (rc < 0)
    {
        // Oops! What happened here?!
        fi_log(LOG_ERR, "%s: Unable to DMA data!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All is well.
    rc = 0;

bail:
    // Return our status.
    return rc;
}

//--------------------------------------------------------------------
// ipmThread Implementation
//--------------------------------------------------------------------

ipmThread::ipmThread() :
    FAS_THREAD::MsgThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

ipmThread::~ipmThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Process the messages for the thread.
 */
void ipmThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_FRAME:
            // No processing of messages if we're canceling.
            if (saveThread::user_cancel)
                message->resp = FAS_THREAD::FAULT;

            else if (!handleSaveFrameRequest(static_cast<SaveFrameMessage *>(message)))
                message->resp = FAS_THREAD::FAULT;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
        case FAS_THREAD::HALT:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in ipmThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

bool ipmThread::handleSaveFrameRequest(SaveFrameMessage *msg)
{
    uint32_t width;
    uint32_t height;
    uint32_t pc_img_len;
    struct timespec start_time;
    IPM_PROC::IPM_REQUEST ipm;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: part=%d, frame_num=%d, proc_info=0x%x, dma_mod=0x%x, dma_dims=0x%x", __PRETTY_FUNCTION__, pm_ptr->getReviewPartition(), msg->frame_num, msg->proc_info, msg->dma_mod, msg->dma_dims);

    // Get our starting time for the statistics output.
    measuretime(1, &start_time);

    // Get the address of the frame we're going to process.
    if (msg->ltr_mode)
        msg->image_addr = msg->ltr_im_addr;
    else
        msg->image_addr = pm_ptr->getPartitionFrameAddr(pm_ptr->getReviewPartition(), msg->frame_num, false);

    // Build up our IPM request structure.
    ipm.proc_info = msg->proc_info;
    ipm.process_live = false;
    ipm.image_addr = msg->image_addr;

    // How many bytes will the processed frame be in PC memory?
    width = (msg->dma_dims & 0xFFFF);
    height = ((msg->dma_dims >> IPM_FRAME_DIM_Y_SHIFT) & 0xFFFF);
    pc_img_len = get_processed_size(ipm.proc_info, width, height);

    // Process the frame.
    if (!ipmProcRequest(&ipm))
    {
        fi_log(LOG_ERR, "%s: ipm_request for addr 0x%" PRIx64 " failed!", __PRETTY_FUNCTION__, ipm.image_addr);
        msg->status = EFAULT;
        msg->resp = FAS_THREAD::FAULT;
    }
    else
    {
        int rc;
        double elapsed;
        uint64_t elapsed_usecs;

        // Measure our time.
        elapsed_usecs = measuretime(0, &start_time);
        msg->times_usec.push_back(elapsed_usecs);
        elapsed = (double)elapsed_usecs / 1.0e6;

        // Reset our times for the DMA transfer
        measuretime(1, &start_time);

        // We might need the length later.
        msg->ipm_image_len = ipm.image_len;
        msg->ipm_buf_addr = ipm.ipm_buf_addr;

        // Transfer the image data. The IPM always row word aligns it's output data. The "image_len" value
        // is this padded size, so we need to tell the FPGA what the destination size should be.
        rc = fpga_ptr->imageDownload(msg->dma_mod, msg->dma_dims, msg->ipm_buf_addr, msg->buf_addr, msg->ipm_image_len, pc_img_len);

        // Did everything go OK?
        if (rc < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
            msg->status = EIO;
            msg->resp = FAS_THREAD::FAULT;
        }

        // We are done with the IPM buffer so free it.
        ipm_proc_release(ipm.ipm_buf_addr);

        // Measure our time.
        elapsed_usecs = measuretime(0, &start_time);
        msg->times_usec.push_back(elapsed_usecs);

        // Output some statistics.
        elapsed += (double)elapsed_usecs / 1.0e6;
        TP(TRACE_SEL_SAVE_PROC, "%s: frame %d [%6.3lf]", __PRETTY_FUNCTION__, msg->frame_num, elapsed);
    }

    // Return our status.
    return (msg->status == 0);
}

//--------------------------------------------------------------------
// jpgThread Implementation
//--------------------------------------------------------------------

jpgThread::jpgThread() :
    FAS_THREAD::MsgThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

jpgThread::~jpgThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Process the messages for the thread.
 */
void jpgThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_FRAME:
            // No processing of messages if we're canceling.
            if (saveThread::user_cancel)
                message->resp = FAS_THREAD::FAULT;

            else if (!handleSaveFrameRequest(static_cast<SaveFrameMessage *>(message)))
                message->resp = FAS_THREAD::FAULT;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
        case FAS_THREAD::HALT:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in jpgThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

bool jpgThread::handleSaveFrameRequest(SaveFrameMessage *msg)
{
    bool rc(false);
    int retries;
    double elapsed;
    uint64_t elapsed_usecs;
    struct timespec start_time;
    uint32_t uv_offset;  // Number of bytes (HEXWORD alignment) to the UV-plane.
    JPEG_PROC::BUF_REQUEST request;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    // Get our starting time for the statistics output.
    measuretime(1, &start_time);

    // Initialize our request structure with what we know.
    if ((msg->proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
        request.pixel_format = JPEG_UTIL::NV12;
    else
        request.pixel_format = JPEG_UTIL::Y8;
    request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();
    request.jpeg_buffer = nullptr;
    request.jpeg_len = 0;

    // Get our dimensions.
    request.width = msg->dma_dims & 0xFFFF;
    request.height = (msg->dma_dims >> IPM_FRAME_DIM_Y_SHIFT) & 0xFFFF;

    // Calculate our offset to the UV-plane.
    //
    // Note that in the ipmThread::handleSaveFrameRequest() we DMA'd straight from
    // image memory. This means that the UV plan is on a HEXWORD boundary.
    uv_offset = convert_to_HEXWORD(request.width * request.height) << HEXWORD_SHIFT;

    // Now setup the pointers to the data.
    request.src_pixels_Y = msg->buf_ptr;
    request.src_pixels_UV = msg->buf_ptr + uv_offset;

    // Try to get a buffer.
    for (retries = 0; (retries < MAX_JPEG_RETRIES); ++retries)
    {
        // JPEG encode the data.
        jpeg_buf_request(&request);

        // Did it work?
        if ((request.jpeg_len > 0) && (request.jpeg_buffer != nullptr))
        {
            // Save the encoded information.
            msg->jpg_buf_ptr = request.jpeg_buffer;
            msg->ipm_image_len = request.jpeg_len;
            msg->status = 0;

            // All is well.
            rc = true;
            break;
        }

        // The request failed for some reason.
        fi_log(LOG_DEBUG, "%s: [retry %d of %d] jpeg_buf_request() failed!", __PRETTY_FUNCTION__, retries, MAX_JPEG_RETRIES);
        msg->status = ENOMEM;

        // Delay.
        using ms = std::chrono::milliseconds;
        std::this_thread::sleep_for(ms(100));
    }

    // Output our statistics.
    elapsed_usecs = measuretime(0, &start_time);
    elapsed = (double)elapsed_usecs / 1.0e6;
    msg->times_usec.push_back(elapsed_usecs);
    TP(TRACE_SEL_SAVE_PROC, "%s: frame %d [%6.3lf]", __PRETTY_FUNCTION__, msg->frame_num, elapsed);

    // Return our status.
    return rc;
}

//--------------------------------------------------------------------
// shmemChecker Implementation
//--------------------------------------------------------------------

void shmemChecker::run()
{
    fi_log(LOG_DEBUG, "%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // We are now running.
    thread_running = true;

    // Keep going until we're told to stop.
    while (!stop_request)
    {
        // Do we have anything on the queue?
        bool done = wait_queue.empty();

        // Sanity check.
        if (!done && (hSharedMem == nullptr))
        {
            fi_log(LOG_ERR, "%s: Unexpectedly null shared memory handle!", FI_FUNCTION);
            break;
        }

        // Loop until we can do any more.
        while (!done)
        {
            SaveFrameMessage *msg = wait_queue.front();
            fi_log(LOG_DEBUG, "%s: wait_queue: %d, looking at frame %d, index %d", FI_FUNCTION, wait_queue.size(), msg->frame_num, msg->shmem_index);
            if (fsmIsIndexAvailable(hSharedMem, msg->shmem_index))
            {
                // Forward the message.
                msg->forward();

                // Remove this message from the queue.
                {
                    std::unique_lock<std::mutex> locker(queue_mutex);
                    wait_queue.pop();
                }

                // Have we gone through the queue?
                done = wait_queue.empty();
            }
            else
            {
                // We've done everything we can for now.
                done = true;
            }
        }

        // Delay for a bit. Using this method has the add benefit of
        // waking up if someone asks us to stop.
        sleep_ms(50);
    }

    // We need to purge the wait queue.
    fi_log(LOG_DEBUG, "%s: wait_queue=%d", __PRETTY_FUNCTION__, wait_queue.size());
    while (wait_queue.size() != 0)
    {
        SaveFrameMessage *queued_msg = wait_queue.front();
        queued_msg->forward();
        wait_queue.pop();
    }

    // We are no longer running.
    thread_running = false;
}

void shmemChecker::addMessage(SaveFrameMessage *message)
{
    std::unique_lock<std::mutex> locker(queue_mutex);
    wait_queue.push(message);
}

//--------------------------------------------------------------------
// shmemThread Implementation
//--------------------------------------------------------------------

shmemThread::shmemThread(uint64_t im_base_addr, uint64_t im_length) :
    FAS_THREAD::MsgThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Assume things will go badly.
    hSharedMem = CameraServer::get_shmem();
    if (hSharedMem == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unexpectedly no shared memory handle?!", FI_FUNCTION);
        return;
    }

    // Initialize the shared memory control structures.
    fsmServerUpdateMemInfo(hSharedMem, im_base_addr, im_length);

    // Start our checker thread.
    bufferChecker = std::unique_ptr<shmemChecker>(new shmemChecker(hSharedMem));
    if (bufferChecker.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate our buffer checker object!", FI_FUNCTION);
        return;
    }
    bufferChecker->start();
}

shmemThread::~shmemThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Close down our buffer checker thread.
    if (bufferChecker.get() != nullptr)
        bufferChecker->stop();
}

/*!
 * @brief Process the messages for the thread.
 */
void shmemThread::messageHandle(FAS_THREAD::Message *message)
{
    bool forward_msg(true);     // In general, we will want to forward the message.

    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_FRAME:
            // No processing of messages if we're canceling.
            if (saveThread::user_cancel)
                message->resp = FAS_THREAD::FAULT;

            // Process the save frame request.
            else if (!handleSaveFrameRequest(static_cast<SaveFrameMessage *>(message), forward_msg))
                message->resp = FAS_THREAD::FAULT;

            break;

        case FAS_THREAD::HALT:
            fi_log(LOG_DEBUG, "%s: Halting the buffer checker.", FI_FUNCTION);
            if (bufferChecker.get() != nullptr)
                bufferChecker->stop();

            // All done.
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage, if appropriate.
        if (forward_msg)
            message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in shmemThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

bool shmemThread::handleSaveFrameRequest(SaveFrameMessage *msg, bool& forward_msg)
{
    int client_fault;
    bool rc(false);

    // See if the client posted a fault.
    if ((client_fault = fsmGetClientFault(hSharedMem)) != 0)
    {
        // The client is complaining. Log it.
        fi_log(LOG_ERR, "%s: The client sent a fault (%d).", FI_FUNCTION, client_fault);

        // Nothing for us to do, except to acknowledge the fault and move on.
        fsmServerClearFault(hSharedMem);
    }

    // Have we already posted this image to the shared memory ring?
    if (msg->shmem_posted)
    {
        // Sanity check.
        if (msg->shmem_index == -1U)
        {
            fi_log(LOG_DEBUG, "%s: Invalid shared memory ring buffer index.", FI_FUNCTION);
            goto hsfr_bail;
        }

        // Assume this entry will not be available until the client is finished with it.
        TP(TRACE_SEL_SAVE_PROC, "%s: Queuing frame %d/0x%" PRIx64 "/%d", __PRETTY_FUNCTION__, msg->frame_num, msg->buf_addr, msg->shmem_index);
        bufferChecker->addMessage(msg);

        // We are done with the messsage.
        forward_msg = false;
    }
    else
    {
        // Make sure we aren't full.
        if (fsmIsFull(hSharedMem))
        {
            // We aren't going to block the train. Log the condition and move on.
            fi_log(LOG_ERR, "%s: The ring buffer cannot take any more entries!", FI_FUNCTION);
            goto hsfr_bail;
        }

        // Send the information.
        fsmSetInfo(hSharedMem, msg->buf_addr, msg->ipm_image_len, msg->frame_num);

        // All done. This will update the control structure and notify the client if needed.
        msg->shmem_index = fsmWriteDone(hSharedMem);
        TP(TRACE_SEL_SAVE_PROC, "%s: Frame %d/0x%" PRIx64 "put into the ring at %d.", __PRETTY_FUNCTION__, msg->frame_num, msg->buf_addr, msg->shmem_index);

        // Sanity check.
        if (msg->shmem_index == -1U)
        {
            fi_log(LOG_ERR, "%s Unable to post to the shared memory ring buffer!", FI_FUNCTION);
            goto hsfr_bail;
        }

        // Remember that we've posted this image.
        msg->shmem_posted = true;

        // Forward this message onwards.
        forward_msg = true;
    }

    // Everything is well.
    rc = true;

hsfr_bail:
    // Return our result.
    return rc;
}

//--------------------------------------------------------------------
// writeThread Implementation
//--------------------------------------------------------------------

writeThread::writeThread() :
    FAS_THREAD::MsgThread(),
    total_bytes(0)
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
    videoImage.reset();
}

writeThread::~writeThread()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);
    videoImage.reset();
}

/*!
 * @brief Process the messages for the thread.
 */
void writeThread::messageHandle(FAS_THREAD::Message *message)
{
    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_FRAME:
            // No processing of messages if we're canceling.
            if (saveThread::user_cancel)
                message->resp = FAS_THREAD::FAULT;

            else if (!handleSaveFrameRequest(static_cast<SaveFrameMessage *>(message)))
                message->resp = FAS_THREAD::FAULT;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
        case FAS_THREAD::HALT:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in writeThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

/*!
 * @brief Checks the space available on external storage
 *
 * @param image_len Number of bytes per image.
 * @param frame_count Number of frames.
 *
 * @return 0 if there is enough space, one of the errno codes if there is not enough space
 */
int writeThread::checkSpace(std::weak_ptr<FAS_VIDEO::FiVideoStream> video_writer, uint32_t image_len, uint32_t frame_count, const char *target_path)
{
    uint32_t reg_value;
    uint64_t total_required;
    std::size_t pos;
    std::string temp_str;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    std::shared_ptr<FAS_VIDEO::FiVideoStream> avi_stream = video_writer.lock();

    if (avi_stream == nullptr)
        TP(TRACE_SEL_SAVE_PROC, "%s: image_len=0x%08x frame_count=%u target='%s'", __PRETTY_FUNCTION__, image_len, frame_count, target_path);
    else
        TP(TRACE_SEL_SAVE_PROC, "%s: image_len=0x%08x frame_count=%u target='%s'", __PRETTY_FUNCTION__, image_len, frame_count, avi_stream->GetVideoName().c_str());

    // Calculate the total amount of space required.
    total_required = ((uint64_t)image_len) * frame_count;

    // Get the destination information register.
    reg_value = regs[GIGE_REG_IMAGE_DESTINATION].get_value();

    // Round up the total by "a bit". Be more conservative if the format is JPEG
    // (or AVI which contains JPEGs).
    if ((avi_stream != nullptr) || ((reg_value & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG))
        total_required += (total_required >> 3);    // 1/8 is 12.5%
    else
        total_required += (total_required >> 4);    // 1/16 is 6.25%

    // Strip off the filename so we're only checking space on the folder.
    if (avi_stream == nullptr)
        temp_str.assign(target_path);
    else
        temp_str.assign(avi_stream->GetVideoName());
    pos = temp_str.find_last_of('/');

    // Do the check.
    return StreamSave::checkSpace(total_required, temp_str.substr(0, pos).c_str());
}

bool writeThread::handleSaveFrameRequest(SaveFrameMessage *msg)
{
    double elapsed;
    uint64_t elapsed_usecs;
    struct timespec start_time;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    // Get our starting time for the statistics output.
    measuretime(1, &start_time);

    // Save the frame.
    if (!saveFrame(msg))
    {
        fi_log(LOG_ERR, "%s: Error saving frame #%d, %u@%p ", __PRETTY_FUNCTION__,
               msg->frame_num, msg->ipm_image_len, (void *)msg->buf_ptr);
        msg->resp = FAS_THREAD::FAULT;
    }

    // If we are saving JPEGs, then we need to release the encoder's buffer.
    if (msg->jpg_buf_ptr)
    {
        jpeg_release_frame(msg->jpg_buf_ptr);
        msg->jpg_buf_ptr = nullptr;
    }

    // Output our statistics.
    elapsed_usecs = measuretime(0, &start_time);
    elapsed = (double)elapsed_usecs / 1.0e6;
    msg->times_usec.push_back(elapsed_usecs);
    TP(TRACE_SEL_SAVE_PROC, "%s: frame %d [%6.3lf]", __PRETTY_FUNCTION__, msg->frame_num, elapsed);

    // Return our status.
    return (msg->status == 0);
}

/*!
 * @brief Save a single frame
 *
 * @param image_ptr     Pointer to the image data.
 * @param image_size    Number of bytes in the image data.
 *
 * @return true for success, false for failure
 */
bool writeThread::saveFrame(SaveFrameMessage *msg)
{
    bool rc = false;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    // Is this the first time through?
    if (total_bytes == 0)
    {
        int err_code;
        uint32_t frame_count;
        uint32_t image_size;
        uint32_t header_len;

        if (msg->pack_handle)
        {
            uint32_t width;
            uint32_t height;
            std::unique_ptr<pm_metadata> meta_ptr = nullptr;
            PartMgr *part_mgr = CameraServer::get_partmgr();

            // Initialize our lengths.
            image_size = 0;
            header_len = 0;

            // What are the image dimensions being saved?
            width = (msg->dma_dims & 0xFFFF);
            height = ((msg->dma_dims >> IPM_FRAME_DIM_Y_SHIFT) & 0xFFFF);

            // Get a pointer to our metadata.
            if (part_mgr != nullptr)
                meta_ptr = part_mgr->getReviewPartMetadata();

            if ((meta_ptr != nullptr) && (FiInternal::factory() != nullptr))
            {
                FiInternal::GET_PACKED_SIZE_FN funcPtr = FiInternal::factory()->getPackedSizeFuncPtr();
                if (funcPtr != nullptr)
                {
                    image_size = (*funcPtr)(width, height, (IS_8_BIT_MODE(meta_ptr->bit_sel) ? 8 : (IS_10_BIT_MODE(meta_ptr->bit_sel) ? 10 : 12)));
                }
            }
        }
        else
        {
            image_size = msg->ipm_image_len;
            header_len = msg->header.size();
        }

        // Check to see if we think there'll be enough space.
        frame_count = ((msg->end_frame - msg->start_frame) + 1);
        if ((err_code = checkSpace(msg->video_writer, image_size + header_len, frame_count, msg->image_path.c_str())) != 0)
        {
            // Save the status and bail.
            fi_log(LOG_ERR, "%s: Not enough space on disk.", __FUNCTION__);
            msg->status = err_code;
            return false;
        }
    }

    // Do we have a video context?
    std::shared_ptr<FAS_VIDEO::FiVideoStream> vid_stream = msg->video_writer.lock();
    if (vid_stream != nullptr)
    {
        // Save our video frame.
        rc = saveVideoFrame(msg, vid_stream.get());
    }
    else if (msg->pack_handle)
    {
        // Save our packed frame.
        rc = saveToPackedStream(msg);
    }
    else if (!msg->image_path.empty())
    {
        // Just saving a normal image file.
        rc = saveImageFrame(msg);
    }
    else
    {
        // Something ain't right with this boy...
        fi_log(LOG_DEBUG, "%s: Not AVI and no image_path!?", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        rc = false;
    }

    return rc;
}

bool writeThread::saveToPackedStream(SaveFrameMessage *msg)
{
    bool rc = false;
    FiInternal::ADD_TIFF_IMAGE_DATA_FN funcPtr = nullptr;
    FAS_INTERNAL::FIP_STATUS streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_TYPE_ERROR;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }
    if (msg->pack_handle == 0)
    {
        fi_log(LOG_DEBUG, "%s: Pack handle is 0!", __PRETTY_FUNCTION__);
        return false;
    }
    if (FiInternal::factory() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Packed stream not supported!", __PRETTY_FUNCTION__);
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    if (msg->image_path.empty())
    {
        // Something ain't right with this boy...
        fi_log(LOG_DEBUG, "%s: No image_path!?", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        return false;
    }

    // Get the function pointer to add the information.
    funcPtr = FiInternal::factory()->getAddTiffFuncPtr();
    if (funcPtr != nullptr)
    {
        // Add the image to the streamer.
        streamStatus = (*funcPtr)(msg->pack_handle, msg->buf_ptr);
    }

    // Check our stream status.
    if ((streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY) ||
        (streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED))
    {
        // Everything looks OK.
        rc = true;

        // Update the total number of bytes we transferred.
        total_bytes += msg->ipm_image_len;
    }
    else
    {
        // Log the error.
        fi_log(LOG_DEBUG, "%s: Error updating pack streamer, streamStatus:0x%x", __PRETTY_FUNCTION__, (int)streamStatus);
    }

    // All done.
    return rc;
}

bool writeThread::saveImageFrame(SaveFrameMessage *msg)
{
    bool rc = false;
    int fd;
    uint8_t *image_ptr;
    uint32_t image_size;
    uint32_t header_len;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    if (msg->image_path.empty())
    {
        // Something ain't right with this boy...
        fi_log(LOG_DEBUG, "%s: No image_path!?", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        return false;
    }

    // Extract some info for ease of use.
    if (msg->jpg_buf_ptr == nullptr)
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: jpg_buf_ptr is null.", __PRETTY_FUNCTION__);
        image_ptr = msg->buf_ptr;
    }
    else
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: JPG frame.", __PRETTY_FUNCTION__);
        image_ptr = msg->jpg_buf_ptr;
    }
    image_size = msg->ipm_image_len;
    header_len = msg->header.size();

    // Just saving a normal image file.
    fd = open(msg->image_path.c_str(), O_CREAT | O_RDWR | O_TRUNC | O_NOATIME, 0666);
    if (fd == -1)
    {
        msg->status = errno;
        fi_log(LOG_ERR, "%s: Unable to create '%s', err=%d(%s)", __FUNCTION__, msg->image_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Pre-allocate the file. An error of EOPNOTSUPP is possible as not all filesystems (e.g. ExFAT) support this operation.
    if ((fallocate(fd, /*FALLOC_FL_NO_HIDE_STALE*/ 0, 0, header_len + image_size) == -1) && (errno != EOPNOTSUPP))
    {
        // This is a problem, but for now we'll just move on.
        fi_log(LOG_DEBUG, "%s: fallocate() failed: %d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
    }
    if (posix_fadvise(fd, 0, header_len + image_size, POSIX_FADV_DONTNEED))
    {
        fi_log(LOG_ERR, "%s: fadvise() failed: %d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
    }

    // Is there header information?
    if (header_len > 0)
    {
        if (write(fd, msg->header.data(), header_len) < 0)
        {
            msg->status = errno;
            fi_log(LOG_ERR, "%s: Unable to write header, err=%d(%s)", __FUNCTION__, errno, strerror(errno));
            close(fd);
            return false;
        }
    }

    // Write the data.
    if (write(fd, image_ptr, image_size) < 0)
    {
        fi_log(LOG_ERR, "%s: Unable to write buffer, err=%d(%s)", __FUNCTION__, errno, strerror(errno));
        rc = false;
    }
    else
    {
        // All is well.
        rc = true;
    }

    // Flush the file to the drive.
    syncfs(fd);

    // Close the file
    close(fd);

    // Update the total number of bytes we transferred.
    total_bytes += (image_size + header_len);

    return rc;
}

/*!
 * @brief Save a single frame for a video stream.
 *
 * @param msg   Pointer to our SaveFrameMessage object.
 *
 * @return true for success, false for failure
 */
bool writeThread::saveVideoFrame(SaveFrameMessage *msg, FAS_VIDEO::FiVideoStream *vid_stream)
{
    bool rc = false;
    uint8_t *image_ptr;
    uint32_t image_size;
    uint32_t header_len;
    FAS_VIDEO::STREAM_STATUS streamStatus;
    std::unique_ptr<FAS_VIDEO::FiImagePkg> imagePkg;

    // Sanity check.
    if (msg == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        return false;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d", __PRETTY_FUNCTION__, msg->frame_num);

    // Do we have a video context?
    if (vid_stream == nullptr)
    {
        FI_LOGIC_ERROR("msg is NULL!");
        msg->status = EINVAL;
        return false;
    }

    // Extract some info for ease of use.
    if (msg->jpg_buf_ptr == nullptr)
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: jpg_buf_ptr is null.", __PRETTY_FUNCTION__);
        image_ptr = msg->buf_ptr;
    }
    else
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: JPG frame.", __PRETTY_FUNCTION__);
        image_ptr = msg->jpg_buf_ptr;
    }
    image_size = msg->ipm_image_len;
    header_len = msg->header.size();

    // We use an FiImage unique pointer to supply frame data to the stream writer.
    // The FiImage object can attach this externally managed image data frame
    // pointer and it does not take ownership.  It is borrowing the pointer and
    // length for the time it takes to edit or add the image data frame.
    if (msg->jpg_buf_ptr == nullptr)
    {
        // Creat the videoImage object, if we don't already have one.
        if (!videoImage)
        {
            int hdr_size;
            uint32_t fourcc;
            Sensor *sensor_ptr = CameraServer::get_sensor();
            PartMgr *pm_ptr = CameraServer::get_partmgr();

            TP(TRACE_SEL_SAVE_PROC, "%s: Not JPEG [0x%x], so create ImageDataFrame.", __PRETTY_FUNCTION__, msg->video_codec);

            // Because we're using data straight from the FPGA, there's no header associated with it.
            // We create a header here so the video streaming logic knows what to do with the data.
            switch (msg->video_codec)
            {
            case GIGE_STREAM_FORMAT_BMP:
                videoImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
                if (videoImage)
                {
                    // Figure out the FourCC value.
                    if (sensor_ptr->is_sensor_color())
                        fourcc = FAS_VIDEO::FOURCC_RGB;
                    else
                        fourcc = FAS_VIDEO::FOURCC_Y8;

                    TP(TRACE_SEL_SAVE_PROC, "%s: Creating header, fourcc=0x%x", __PRETTY_FUNCTION__, fourcc);
                    hdr_size = videoImage->CreateHeader(fourcc, pm_ptr->getReviewPartMetadata().get());
                    TP(TRACE_SEL_SAVE_PROC, "%s: Header created [size=%d]", __PRETTY_FUNCTION__, hdr_size);
                }
                break;
            case GIGE_STREAM_FORMAT_PLANAR:
                videoImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMPPlanar());
                if (videoImage)
                {
                    // Figure out the FourCC value.
                    if (sensor_ptr->is_sensor_color())
                        fourcc = FAS_VIDEO::FOURCC_NV12;
                    else
                        fourcc = FAS_VIDEO::FOURCC_Y8;

                    TP(TRACE_SEL_SAVE_PROC, "%s: Creating header, fourcc=0x%x", __PRETTY_FUNCTION__, fourcc);
                    hdr_size = videoImage->CreateHeader(fourcc, pm_ptr->getReviewPartMetadata().get());
                    TP(TRACE_SEL_SAVE_PROC, "%s: Header created [size=%d]", __PRETTY_FUNCTION__, hdr_size);
                }
                break;
            default:
                fi_log(LOG_DEBUG, "%s: Unexpected video_codec = 0x%x", __PRETTY_FUNCTION__, msg->video_codec);
                break;
            }
        }
    }
    else if (!videoImage)
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: Create FiJPEG ImageDataFrame.", __PRETTY_FUNCTION__);
        videoImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_JPEG::FiJPEG());
    }

    // Sanity check.
    if (videoImage == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL image data frame for video stream writer!", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        goto vid_bail;
    }

    // Note:  This image package wrapper was created to provide a common
    // object we could use to pass around different types of image data
    // whether it be externally managed by the CameraServer or internally
    // managed by the fas_video library itself in a std::vector form.
    //
    // Create an image package wrapping the image pointer and length.
    imagePkg = FAS_VIDEO::FiImagePkg::Create((uint8_t *)image_ptr, image_size);

    // Attach the created image package into the image object.  It now takes
    // ownership of that unique pointer for the life of videoImage or until
    // another AttachImagePkg is performed on that videoImage.  If the videoImage
    // destructs at the end of this method all is cleaned up because of the use
    // of the unique pointers in this system.
    videoImage->AttachImagePkg(std::move(imagePkg));
    if (!videoImage->IsImageAttached())
    {
        fi_log(LOG_ERR, "%s: Unable to attach image data frame for video stream writer!", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        goto vid_bail;
    }
    TP(TRACE_SEL_SAVE_PROC, "%s: Attached data (%p, %u).", __PRETTY_FUNCTION__, (void *)image_ptr, image_size);

    // Make sure the video stream writer is ready to take a frame.
    if (!vid_stream->IsStreamReady())
    {
        fi_log(LOG_ERR, "%s: Video Stream Writer is not in ready state!", __PRETTY_FUNCTION__);
        msg->status = EINVAL;
        goto vid_bail;
    }

    // Add a frame to the video and write what we've got so far.
    streamStatus = vid_stream->AddVideoFrame(videoImage);
    if ((streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY) &&
        (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED))
    {
        vid_stream->LogStreamStatus();
        msg->status = errno;
        fi_log(LOG_ERR, "%s: AddVideoFrame() failed, err=%d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
    }
    else
    {
        // All is well.
        rc = true;
    }

    // Update the total number of bytes we transferred.
    total_bytes += (image_size + header_len);

vid_bail:
    return rc;
}

//--------------------------------------------------------------------
// saveThread Implementation
//--------------------------------------------------------------------

saveThread::saveThread() :
    FAS_THREAD::MsgThread(),
    status(0),
    faulted(false),
    bb_index(-1),
    bb_addr(0),
    buffer_count(0),
    stat_fp(nullptr),
    sl_msg(nullptr),
    cur_frame(0)
{
    int i;
    bool ltr_mode;
    long sys_page_size;
    uint32_t max_frame_size;
    uint32_t aligned_frame_len;
    uint32_t bb_map_len;
    uint8_t *base_ptr;
    Sensor *sensor_ptr = CameraServer::get_sensor();
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (sensor_ptr == nullptr)
    {
        // Throw an exception.
        FI_LOGIC_ERROR("NULL sensor_ptr!?");
        return;
    }

    // We are not canceling.
    user_cancel = false;

    // Create our working threads.
    ltr_thread = std::unique_ptr<ltrThread>(new ltrThread());
    if (ltr_thread.get() != nullptr)
        ltr_thread->start();
    ipm_thread = std::unique_ptr<ipmThread>(new ipmThread());
    if (ipm_thread.get() != nullptr)
        ipm_thread->start();
    jpg_thread = std::unique_ptr<jpgThread>(new jpgThread());
    if (jpg_thread.get() != nullptr)
        jpg_thread->start();
    write_thread = std::unique_ptr<writeThread>(new writeThread());
    if (write_thread.get() != nullptr)
        write_thread->start();

    // The page size is 4096 - but this is the "correct" way to get it.
    sys_page_size = sysconf(_SC_PAGESIZE);
    if (sys_page_size <= 0)
        sys_page_size = 4096;

    // What is the size of our biggest frame? (W x H) x 3, because color frames are RGB.
    max_frame_size = (sensor_ptr->get_sensor_max_width() * sensor_ptr->get_sensor_max_height()) * 3;

    // How many maximum sized frames fit within a single bounce buffer? Make sure
    // we round the size up so it's a nice even multiple of page size.
    //
    //------------------------------------------------------------------------
    // NOTE: The alignment is REQUIRED for DMA operations!
    //------------------------------------------------------------------------
    aligned_frame_len = ((max_frame_size + (sys_page_size >> 1)) / sys_page_size) * sys_page_size;
    if (aligned_frame_len == 0)
    {
        // Throw an exception.
        FI_LOGIC_ERROR("aligned_frame_len is 0!");
        return;
    }
    TP(TRACE_SEL_SAVE_PROC, "%s: max_frame_size=0x%08x, aligned_frame_len=0x%08x", __PRETTY_FUNCTION__, max_frame_size, aligned_frame_len);

    // Get our buffer for the image data from the IPM.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Memory map the bounce buffer.
    base_ptr = (uint8_t *)bb_mmap(bb_index, 0);
    if (base_ptr == (uint8_t *)MAP_FAILED)
    {
        // Throw an exception.
        FI_RUNTIME_ERROR("bb_mmap() failed: %d(%s)", errno, strerror(errno));
        return;
    }

    // How long is the bounce buffer mmap-ed area?
    bb_map_len = bb_get_mmap_len(bb_index);
    if (bb_map_len == 0)
    {
        // Throw an exception.
        FI_RUNTIME_ERROR("bb_get_mmap_len(%d) return 0!", bb_index);
        return;
    }

    // How many buffers can we fit?
    buffer_count = bb_map_len / aligned_frame_len;
    TP(TRACE_SEL_SAVE_PROC, "%s: buffer_count = %d", __PRETTY_FUNCTION__, buffer_count);

    // If we're in LTR mode, grab the last buffer for transferring images from
    // the SSD to Image Memory.
    ltr_mode = ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    if (ltr_mode)
    {
        // Decrement our buffer count by 1.
        --buffer_count;

        // Pass that last buffer to LTR thread.
        ltr_thread->setBbAddr(bb_addr + (buffer_count * aligned_frame_len));
        ltr_thread->setBbPtr(base_ptr + (buffer_count * aligned_frame_len));
        TP(TRACE_SEL_SAVE_PROC, "%s: LTR mode, new buffer_count = %d", __PRETTY_FUNCTION__, buffer_count);
    }

    // Loop through all our buffers, building up messages.
    for (i = 0; !user_cancel && (i < buffer_count); ++i)
    {
        SaveFrameMessage *save_msg;

        // Allocate a new message.
        save_msg = new SaveFrameMessage();
        if (save_msg == nullptr)
        {
            FI_RUNTIME_ERROR("Unable to allocate SaveFrameMessage(%d)", i);
            return;
        }

        // Initialize the common message data.
        save_msg->owner = this;

        // Initialize the static values.
        save_msg->frame_num = (uint32_t) -1;
        save_msg->buf_ptr = base_ptr + (i * aligned_frame_len);
        save_msg->buf_addr = bb_addr + (i * aligned_frame_len);

        // If we're in LTR mode, grab an image memory buffer.
        if (ltr_mode)
        {
            PartMgr *pm_ptr = CameraServer::get_partmgr();
            save_msg->ltr_im_addr = pm_ptr->getCaptureFrameAddr(i);
            TP(TRACE_SEL_SAVE_PROC, "%s: LTR processing frame %d in IM, frame_addr_hw=0x%" PRIx64, __PRETTY_FUNCTION__, i, save_msg->ltr_im_addr);
        }

        // Create a block for the mutex.
        {
            // Acquire the processing mutex.
            std::unique_lock<std::mutex> locker(avail_q_mutex);

            // Add it to our queue.
            avail_q.push(save_msg);
        }
    }
}

saveThread::~saveThread()
{
    FAS_THREAD::SyncMessage msg;

    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Tell the threads that we are stopping, if they haven't already stopped.
    msg.req = FAS_THREAD::HALT;

    // Clean up our working threads.
    if (ltr_thread.get() != nullptr)
    {
        // Is the thread still running?
        if (!ltr_thread->isHalted())
        {
            // Pass it on. A no-op if already halted.
            msg.doSync(ltr_thread.get());

            // Try to join the thread.
            ltr_thread->join();
        }
    }
    if (ipm_thread != nullptr)
    {
        // Is the thread still running?
        if (!ipm_thread->isHalted())
        {
            // Pass it on. A no-op if already halted.
            msg.doSync(ipm_thread.get());

            // Try to join the thread.
            ipm_thread->join();
        }
    }
    if (jpg_thread != nullptr)
    {
        // Is the thread still running?
        if (!jpg_thread->isHalted())
        {
            // Pass it on. A no-op if already halted.
            msg.doSync(jpg_thread.get());

            // Try to join the thread.
            jpg_thread->join();
        }
    }
    if (progress_thread.get() != nullptr)
    {
        // Try to join the thread.
        progress_thread->stop();
        progress_thread->join();
        progress_thread.reset();
    }
    if (shmem_thread != nullptr)
    {
        // Is the thread still running?
        if (!shmem_thread->isHalted())
        {
            // Pass it on. A no-op if already halted.
            msg.doSync(shmem_thread.get());

            // Try to join the thread.
            shmem_thread->join();
        }
    }
    if (write_thread != nullptr)
    {
        // Is the thread still running?
        if (!write_thread->isHalted())
        {
            // Pass it on. A no-op if already halted.
            msg.doSync(write_thread.get());

            // Try to join the thread.
            write_thread->join();
        }
    }

    // Make sure we have all our buffers back in the available queue.
    while ((int)avail_q.size() < buffer_count)
    {
        fi_log(LOG_DEBUG, "%s: Wait for buffers...", __PRETTY_FUNCTION__);

        std::unique_lock<std::mutex> locker(avail_q_mutex);
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (avail_q_status.wait_until(locker, now + ms(100)) == std::cv_status::timeout)
        {
            fi_log(LOG_ERR, "%s: Can't wait any longer for errant buffers!", __PRETTY_FUNCTION__);
            break;
        }
    }

    // Clean-up. We now have all of the buffers, so no need to lock anything down.
    fi_log(LOG_DEBUG, "%s: Clear out the avail_q.", __PRETTY_FUNCTION__);
    while (!avail_q.empty())
    {
        SaveFrameMessage *save_msg;

        // Get the first message.
        save_msg = avail_q.front();
        avail_q.pop();

        // Delete the message.
        delete save_msg;
    }

    // Release the bounce buffer.
    if (bb_index >= 0)
    {
        bb_release(bb_index);
        bb_index = -1;
    }
}

void saveThread::reset()
{
    std::lock_guard<std::mutex> lock(save_proc_mutex);

    // Check to see if we have a thread object.
    if (save_proc.get() == nullptr)
        return;

    // Check to see if we're halted.
    if (!save_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Saver is not halted!", __PRETTY_FUNCTION__);
    }

    // Reset the saver object.
    save_proc.reset();
}

saveThread *saveThread::factory(bool auto_start)
{
    saveThread *rc(nullptr);

    std::lock_guard<std::mutex> lock(save_proc_mutex);

    // We only allow one thread object.
    if (save_proc.get() == nullptr)
    {
        // Does the caller want us to automatically start the object?
        if (auto_start)
        {
            // Create our thread.
            save_proc = std::unique_ptr<saveThread>(new saveThread());
            if (save_proc.get() != nullptr)
                save_proc.get()->start();

            // Return the pointer to the singleton.
            rc = save_proc.get();
        }
    }

    // Check to see if we were halted.
    else if (save_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = save_proc.get();
    }

    // Return our pointer.
    return rc;
}

bool saveThread::buildMessage(SaveFrameMessage *msg)
{
    std::string temp_buffer;

    // Initialize our message.
    msg->frame_num = cur_frame++;
    msg->start_frame = sl_msg->getRequest()->start_frame;
    msg->end_frame = sl_msg->getRequest()->end_frame;
    msg->image_addr = 0;
    msg->ltr_mode = sl_msg->getRequest()->ltr_mode;
    msg->proc_info = sl_msg->getRequest()->proc_info;
    msg->dma_dims = (sl_msg->getRequest()->image_height << IPM_FRAME_DIM_Y_SHIFT) | sl_msg->getRequest()->image_width;
    msg->dma_mod = sl_msg->getRequest()->dma_mod;
    msg->ipm_image_len = 0;
    msg->ipm_buf_addr = 0;
    msg->jpg_buf_ptr = nullptr;
    msg->video_writer = sl_msg->getRequest()->video_writer;
    msg->video_codec = sl_msg->getRequest()->video_codec;
    msg->pack_handle = sl_msg->getRequest()->pack_handle;
    msg->times_usec.clear();
    msg->shmem_posted = false;
    msg->shmem_index = -1U;

    if (sl_msg->getRequest()->header_len > 0)
    {
        // Reserve the correct number of bytes.
        msg->header.reserve(sl_msg->getRequest()->header_len);
        msg->header.assign(sl_msg->getRequest()->header_ptr, sl_msg->getRequest()->header_ptr + sl_msg->getRequest()->header_len);
    }

    // Is there a destination path for the images?
    if (sl_msg->getRequest()->path_str.empty())
    {
        // Must be part of a container, like an AVI or MP4 file.
        msg->image_path.clear();
    }
    else
    {
        // Check to see if we need to move to another subfolder.
        if (!check_subfolder(msg))
        {
            fi_log(LOG_ERR, "%s: Unable to create the subfolder!", __PRETTY_FUNCTION__);
            return false;
        }

        // Build up the filename for this image.
        if (sl_msg->getRequest()->do_subfolders)
            temp_buffer = fmt::sprintf(sl_msg->getRequest()->path_str.c_str(), sl_msg->getRequest()->subfolder_count, msg->frame_num);
        else
            temp_buffer = fmt::sprintf(sl_msg->getRequest()->path_str.c_str(), msg->frame_num);
        msg->image_path.assign(temp_buffer);
        TP(TRACE_SEL_SAVE_PROC, "%s: image_path='%s'", __PRETTY_FUNCTION__, msg->image_path.c_str());
    }

    // Are we using the shared memory ring buffer? If we are, then we need to create our thread worker.
    if (sl_msg->getRequest()->using_sharedmem && (shmem_thread.get() == nullptr))
    {
        // Allocate our shared memory thread.
        shmem_thread = std::unique_ptr<shmemThread>(new shmemThread(bb_addr, bb_get_mmap_len(bb_index)));
        if (shmem_thread.get() != nullptr)
            shmem_thread->start();
    }

    // Are we using the shared memory interface and do we have a worker thread?
    if (sl_msg->getRequest()->using_sharedmem && (shmem_thread.get() != nullptr))
    {
        // It is important to remember that we are stacking the processing
        // in reverse order. With that in mind, before we can recycle the
        // image, we need to make sure that the shared memory client has released
        // the buffer. We will make one last pass through the shared memory thread
        // to check on that.
        msg->route->push(shmem_thread.get());
    }

    // Build up the route for the message. Note that the route
    // is Last-In-First-Out, or First-In-Last-Out depending on
    // your POV.
    msg->route->push(write_thread.get());

    // Are we using the shared memory interface?
    if (sl_msg->getRequest()->using_sharedmem && (shmem_thread.get() != nullptr))
    {
        // We will want to quickly put the image into the shared memory
        // ring so that the client can process it in parallel with the
        // writer saving the data to the output device.
        msg->route->push(shmem_thread.get());
    }

    // Are we saving JPEGs?
    TP(TRACE_SEL_SAVE_PROC, "%s: proc_info=0x%x", __PRETTY_FUNCTION__, sl_msg->getRequest()->proc_info);
    if (((sl_msg->getRequest()->proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12) ||
        ((sl_msg->getRequest()->proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_Y8))
    {
        // Make sure we're not doing a Planar format. The planar format looks like
        // a JPEG, but we don't want to compress the data.
        if (sl_msg->getRequest()->video_codec != GIGE_STREAM_FORMAT_PLANAR)
        {
            TP(TRACE_SEL_SAVE_PROC, "%s: Doing JPEGs", __PRETTY_FUNCTION__);
            msg->route->push(jpg_thread.get());
        }
    }

    // If we're in LTR mode, we need to push the frame up to the
    // image memory before the IPM can work on it.
    if (sl_msg->getRequest()->ltr_mode)
    {
        TP(TRACE_SEL_SAVE_PROC, "%s: Doing LR mode", __PRETTY_FUNCTION__);
        msg->route->push(ipm_thread.get());
        msg->route->push(ltr_thread.get());
    }
    else
    {
        // Go straight to IPM processing with this frame.
        TP(TRACE_SEL_SAVE_PROC, "%s: Straight to IPM processing", __PRETTY_FUNCTION__);
        msg->route->push(ipm_thread.get());
    }

    // All is well.
    return true;
}

/*!
 * @brief Determine if we need to create a subfolder at this point in the
 *    save process. Create the subfolder if needed.
 *
 * @return true on success, false otherwise.
 */
bool saveThread::check_subfolder(SaveFrameMessage *msg)
{
    int check_val;

    // Are we doing subfolders?
    if (!sl_msg->getRequest()->do_subfolders)
        return true;

    // Do we want to create a new subfolder?
    check_val = (int)((msg->frame_num - sl_msg->getRequest()->start_frame) / MAX_FOLDER_COUNT);
    if (check_val > sl_msg->getRequest()->subfolder_count)
    {
        int status;
        std::size_t pos;
        std::string temp_str;
        std::string temp_buffer;

        // Increment our subfolder number.
        sl_msg->setSubfolderCount(sl_msg->getRequest()->subfolder_count + 1);

        // Build up the filename for this image.
        temp_buffer = fmt::sprintf(sl_msg->getRequest()->path_str.c_str(), sl_msg->getRequest()->subfolder_count, msg->frame_num);

        // Strip off the filename.
        temp_str.assign(temp_buffer);
        pos = temp_str.find_last_of('/');

        // Make our new folder.
        status = storage::mkpath(temp_str.substr(0, pos).c_str(), 0777);
        if (status == -1)
        {
            msg->status = errno;
            fi_log(LOG_ERR, "%s: Can't create '%s', err=%d(%s)", __PRETTY_FUNCTION__, temp_buffer.c_str(), errno, strerror(errno));
            return false;
        }
    }

    // All is well.
    return true;
}

void saveThread::cancel()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Log (once) that we've cancelled.
    if (!user_cancel)
        fi_log(LOG_INFO, "%s", __PRETTY_FUNCTION__);

    // Cancel operations.
    user_cancel = true;

    // Wait to have all our buffers back in the available queue.
    while ((int)avail_q.size() < buffer_count)
    {
        std::unique_lock<std::mutex> locker(avail_q_mutex);
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (avail_q_status.wait_until(locker, now + ms(200)) == std::cv_status::timeout)
        {
            fi_log(LOG_ERR, "%s: Can't wait any longer for errant buffers!", __PRETTY_FUNCTION__);
            break;
        }
    }
}

void saveThread::close_stats()
{
    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    if (stat_fp != nullptr)
    {
        fclose(stat_fp);
        stat_fp = nullptr;
    }
}

void saveThread::open_stats(const char *save_path)
{
    TP(TRACE_SEL_SAVE_PROC, "%s: save_path='%s'", __PRETTY_FUNCTION__, save_path);

    if (stat_fp != nullptr)
    {
        fi_log(LOG_DEBUG, "%s: stat_fp != nullptr?", __PRETTY_FUNCTION__);
    }

    stat_fp = fopen(save_path, "a");
    if (stat_fp == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to open '%s', err=%d(%s)", __PRETTY_FUNCTION__, save_path, errno, strerror(errno));
    }
}

/*!
 * @brief Process the messages for the thread.
 */
void saveThread::messageHandle(FAS_THREAD::Message *message)
{
    // Sanity check.
    if (message == nullptr)
    {
        FI_LOGIC_ERROR("message is NULL!");
        return;
    }

    TP(TRACE_SEL_SAVE_PROC, "%s: req=%d", __PRETTY_FUNCTION__, (int)message->req);

    try
    {
        bool do_forward(true);

        // Parse the message.
        switch ((int)message->req)
        {
        case SAVE_TO_LOCAL:
            // If everything went well, we won't do a forward of the message until
            // we get back all the individual frame processing messages.
            if (handleSaveLocalRequest(static_cast<SaveLocalMessage *>(message)))
            {
                // We will forward the message back in the ownerMessage() method.
                do_forward = false;
            }
            else
            {
                message->resp = FAS_THREAD::FAULT;
            }
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            halted = true;
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        if (do_forward)
            message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in saveThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

/*!
 * @brief Manage the process of a threaded save operation.
 */
bool saveThread::handleSaveLocalRequest(SaveLocalMessage *msg)
{
    bool rc(false);

    TP(TRACE_SEL_SAVE_PROC, "%s", __PRETTY_FUNCTION__);

    // Initialize our variables.
    cur_frame = msg->getRequest()->start_frame;

    // Save the message so we can continue to work on it as required.
    sl_msg = msg;

    // Start our progress thread.
    if (progress_thread.get() == nullptr)
    {
        progress_thread = std::unique_ptr<progressThread>(new progressThread());
        progress_thread->start();
    }

    // Build up messages to process the frames. Keep going as long as:
    // 1. We haven't been asked to stop.
    // 2. We have messages in our available queue.
    // 3. We have more frames to process.
    while (!halted && !user_cancel && !avail_q.empty() && (cur_frame <= msg->getRequest()->end_frame))
    {
        SaveFrameMessage *save_msg;

        // Create a block for the mutex.
        {
            // Acquire the processing mutex.
            std::unique_lock<std::mutex> locker(avail_q_mutex);

            // Get the first available message.
            save_msg = avail_q.front();
            avail_q.pop();
        }

        // Build up the message.
        if (!buildMessage(save_msg))
        {
            fi_log(LOG_ERR, "%s: buildMessage() failed?!", __PRETTY_FUNCTION__);
            goto save_bail;
        }

        // Now the message is setup, send it on it's way.
        save_msg->forward();
    }

    // Did we cancel?
    if (user_cancel)
        rc = false;
    else
        rc = true;

save_bail:
    // Return our status.
    return rc;
}

/*!
 * @brief End of the road for our SAVE_FRAME message.
 */
void saveThread::ownerMessage(FAS_THREAD::Message *msg)
{
    bool queue_it(true);
    SaveFrameMessage *save_msg = static_cast<SaveFrameMessage *>(msg);

    TP(TRACE_SEL_SAVE_PROC, "%s: frame_num=%d resp=%d", __PRETTY_FUNCTION__, (int)save_msg->frame_num, (int)save_msg->resp);

    // Sanity check
    if (sl_msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: NULL sl_msg!?", __PRETTY_FUNCTION__);
        return;
    }

    // Update our status information.
    if (sl_msg->getRequest()->info_ptr)
    {
        // Update our progress information.
        sl_msg->getRequest()->info_ptr->current_frame++;
        if (progress_thread.get() != nullptr)
            progress_thread->setProgress(sl_msg->getRequest()->info_ptr->current_frame - sl_msg->getRequest()->info_ptr->start_frame);
    }

    // Check to see if we had a fault when the user didn't cancel. We ALWAYS have a fault
    // if the user cancelled before we finished the process. No need to log those.
    if ((save_msg->resp == FAS_THREAD::FAULT) && !user_cancel)
    {
        fi_log(LOG_ERR, "%s: Frame %d faulted [status=%d(%s)]!", __PRETTY_FUNCTION__, save_msg->frame_num, save_msg->status, strerror(save_msg->status));

        // Remember that this save is faulted.
        status = save_msg->status;
        faulted = true;
    }

    // ====================================================================
    // Statistics
#ifdef DO_STATISTICS
    if (!faulted && (stat_fp != nullptr))
    {
        fprintf(stat_fp, "%u", save_msg->frame_num);
        for (uint64_t elapsed : save_msg->times_usec)
        {
            fprintf(stat_fp, ",%" PRIu64, elapsed);
        }
        fprintf(stat_fp, "\n");
        fflush(stat_fp);
    }
#endif /* DO_STATISTICS */
    // Statistics
    // ====================================================================

    // Do we have more to do and can we proceed?
    if (!halted && !faulted && !user_cancel && (cur_frame <= sl_msg->getRequest()->end_frame))
    {
        // Build up the message.
        if (!buildMessage(save_msg))
        {
            fi_log(LOG_ERR, "%s: buildMessage() failed?!", __PRETTY_FUNCTION__);
        }
        else
        {
            // Now the message is setup, send it on it's way.
            save_msg->forward();
            queue_it = false;
        }
    }

    // Should we put the message back on the available queue?
    if (queue_it)
    {
        std::weak_ptr<FAS_VIDEO::FiVideoStream> video_writer;

        TP(TRACE_SEL_SAVE_PROC, "%s: Add back onto queue", __PRETTY_FUNCTION__);

        // Grab the video writer (if there is one).
        video_writer = save_msg->video_writer;

        // Create a block for the mutex.
        {
            // Acquire the processing mutex.
            std::unique_lock<std::mutex> locker(avail_q_mutex);

            // Now that we're locked, add this buffer to the available queue.
            avail_q.push(save_msg);

            // Wake up anyone waiting.
            avail_q_status.notify_one();
        }

        // Are we done?
        if (!halted && sl_msg && ((int)avail_q.size() == buffer_count))
        {
            TP(TRACE_SEL_SAVE_PROC, "%s: Looks like we're done. faulted=%c.", __PRETTY_FUNCTION__, (faulted ? 'T' : 'F'));

            // Do we have a video context?
            std::shared_ptr<FAS_VIDEO::FiVideoStream> vid_stream = video_writer.lock();
            if (vid_stream != nullptr)
            {
                int count;
                uint32_t lastFrameEncodedCount;
                const int MAX_COUNT = 100;

                // Wait for the video stream writer to finish.
                lastFrameEncodedCount = 0;
                for (count = 0; (count < MAX_COUNT) && !user_cancel; ++count)
                {
                    // Go ahead and pull out the information from the stream writer.
                    uint32_t framesEncoded = vid_stream->GetFramesEncodedCount();

                    // Get the status of the stream.
                    FAS_VIDEO::STREAM_STATUS streamStatus = vid_stream->IsStreamWriterComplete();
                    TP(TRACE_SEL_SAVE_PROC, "%s: [%d] streamStatus: 0x%x, framesEncoded=%u", __PRETTY_FUNCTION__, count, (int)streamStatus, framesEncoded);

                    // See if we're done.
                    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED)
                    {
                        break;
                    }

                    // Is there a problem?
                    else if ((streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY) &&
                             (streamStatus != FAS_VIDEO::STREAM_STATUS::PLACEHOLDER_WAIT))
                    {
                        // There's a problem with the stream!
                        vid_stream->LogStreamStatus();
                        faulted = true;
                        status = EFAULT;
                        break;
                    }

                    // Have we encoded another frame?
                    if (framesEncoded != lastFrameEncodedCount)
                    {
                        // As long as we are actually encoding frames, we will keep going.
                        count = 0;
                        lastFrameEncodedCount = framesEncoded;
                    }

                    // Delay for a bit.
                    using ms = std::chrono::milliseconds;
                    std::this_thread::sleep_for(ms(50));
                }
                if (count >= MAX_COUNT)
                {
                    fi_log(LOG_ERR, "%s: Unable to wait for the stream writer any longer!", __PRETTY_FUNCTION__);
                    faulted = true;
                    status = ETIMEDOUT;
                }
            }
            else if ((save_msg->pack_handle != 0) && (FiInternal::factory() != nullptr))
            {
                uint32_t frameCount;
                uint32_t framesAdded;
                uint32_t framesEncoded;
                FAS_INTERNAL::FIP_STATUS packStatus = FAS_INTERNAL::NO_IO_STREAM;

                FiInternal::GET_PACK_STATISTICS_FN statsFuncPtr = FiInternal::factory()->getGetStatisticsFuncPtr();
                if (statsFuncPtr != nullptr)
                {
                    // Go ahead and pull out the information from the stream writer.
                    (*statsFuncPtr)(save_msg->pack_handle, frameCount, framesAdded, framesEncoded);
                }

                FiInternal::IS_PACK_STREAMER_FINISHED_FN finishedFuncPtr = FiInternal::factory()->geIsFinishedFuncPtr();
                if (finishedFuncPtr != nullptr)
                {
                    // Pull the current stream status.
                    packStatus = (*finishedFuncPtr)(save_msg->pack_handle);
                }

                // See if we're done.
                if ((packStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY) ||
                    (packStatus == FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED))
                {
                    TP(TRACE_SEL_SAVE_PROC, "%s: All done, packStatus=%d", __PRETTY_FUNCTION__, (int)packStatus);
                }

                else
                {
                    // There is something wrong with the pack writer.
                    fi_log(LOG_DEBUG, "%s: An error occurred when saving the packed stream: framesAdded[%1] framesEncoded[%2] frameCount[%3]", __PRETTY_FUNCTION__, framesAdded, framesEncoded, frameCount);
                    faulted = true;
                    status = EFAULT;
                }
            }
            TP(TRACE_SEL_SAVE_PROC, "%s: Stream writer seems to be done. faulted=%c.", __PRETTY_FUNCTION__, (faulted ? 'T' : 'F'));

            // Update our totals.
            sl_msg->getRequest()->info_ptr->total_bytes = write_thread->getTotalBytes();

            // Close down the original SaveLocal message.
            sl_msg->resp = (faulted ? FAS_THREAD::FAULT : FAS_THREAD::OK);
            sl_msg->status = status;
            sl_msg->forward();
            sl_msg = nullptr;
        }
    }
}
