/*!
 * @file sensor.cxx
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#define _POSIX_C_SOURCE 199309
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <fmt/printf.h>
#include <fastec/fas_log.h>

#include "sensor.h"
#include "lince5m.h"
#include "luxima_hs.h"
#include "null_sensor.h"
#include "ct_file.h"

namespace SENSOR
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The EEPROM address where we find the Sensor ID and other information.
const int SENSOR_I2C_ADDR = 0x50;

} /* End of namespace SENSOR */

// Make things easier for ourselves.
using namespace SENSOR;


//! The path information for the bad pixel map file.
const char *Sensor::DEFECT_MAP_FILE_FMT = "%s/%s/bad_pixels_%c%c%c%c.txt";


//--------------------------------------------------------------------

Sensor::Sensor(uint16_t max_width,
               uint16_t max_height,
               uint16_t min_width,
               uint16_t min_height,
               EEPROM_INFO *eeprom)
    :
    sensor_max_width(max_width),
    sensor_max_height(max_height),
    sensor_min_width(min_width),
    sensor_min_height(min_height),
    sensor_test_on(false)
{
    // Copy the EEPROM info over.
    memcpy(&this->eeprom, eeprom, sizeof(EEPROM_INFO));

    if (IS_SENSOR_COLOR(eeprom->sensor_type))
        sensor_is_color = true;
    else
        sensor_is_color = false;
    LOG("%s: sensor_is_color=%d", __PRETTY_FUNCTION__, sensor_is_color);
}

Sensor::~Sensor()
{
}

/*!
 * @brief Figure out what kind of sensor is on the camera, create the class which encapsulates
 *    everything about the correct type of sensor. Load the correct FPGA software based
 *    upon this sensor.
 *
 * @return The sensor and FPGA with the correct sensor code loaded to it, NULL is an error occurred.
 */
std::unique_ptr<Sensor> Sensor::create_sensor()
{
    EEPROM_INFO eeprom;
    std::unique_ptr<Sensor> sensor;         // Pointer to the generic sensor class.
    uint8_t *data_ptr;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Initialize our EEPROM info.
    memset(&eeprom, 0, sizeof(EEPROM_INFO));
    eeprom.sensor_type = FI_SENSOR_UNKNOWN;

    // Read the sensor board EEPROM.
    data_ptr = (uint8_t *)&eeprom;
    for (uint32_t i = 0; i < sizeof(EEPROM_INFO); i++)
    {
        // Read a byte.
        if (!fpga_ptr->i2cRead(SENSOR_I2C_ADDR, i, data_ptr++, FI_FPGA::I2C_SENSOR))
        {
            LOG("%s: Error reading Sensor board EEPROM!", __PRETTY_FUNCTION__);
            break;
        }
    }

    LOG("%s: sensor_type=0x%08x", __PRETTY_FUNCTION__, eeprom.sensor_type);

    // Create the appropriate sensor class.
    switch (eeprom.sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
    case FI_SENSOR_LINCE5M_MONO:
        sensor = std::unique_ptr<Sensor>(new Lince5M(&eeprom));
        break;
    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX13HS_MONO:
        sensor = std::unique_ptr<Sensor>(new Luxima13HS(&eeprom));
        break;
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX160_MONO:
        sensor = std::unique_ptr<Sensor>(new Luxima160(&eeprom));
        break;
    case FI_SENSOR_LUX19HS_COLOR:
    case FI_SENSOR_LUX19HS_MONO:
        sensor = std::unique_ptr<Sensor>(new Luxima19HS(&eeprom));
        break;
    case FI_SENSOR_LUX2100_COLOR:
    case FI_SENSOR_LUX2100_MONO:
        sensor = std::unique_ptr<Sensor>(new Luxima2100(&eeprom));
        break;

    case FI_SENSOR_LUPA1300_MONO:
    case FI_SENSOR_LUPA1300_COLOR:
    default:
        fi_log(LOG_ERR, "%s: unexpected sensor type 0x%08x", __FUNCTION__, eeprom.sensor_type);
        sensor = std::unique_ptr<Sensor>(new NullSensor(&eeprom));
        break;
    }

    // Make sure we got something!
    if (sensor == NULL)
    {
        fi_log(LOG_ERR, "%s: Couldn't create a sensor class. Error %d(%s)",
               __FUNCTION__, errno, strerror(errno));
    }

    // Return our sensor class pointer.
    return sensor;
}

uint32_t Sensor::get_sensor_type()
{
    uint32_t type;
    EEPROM_INFO eeprom;
    uint8_t type_offset;
    uint8_t *data_ptr;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Initialize our EEPROM info.
    memset(&eeprom, 0, sizeof(EEPROM_INFO));
    eeprom.sensor_type = FI_SENSOR_UNKNOWN;

    // Read the sensor board EEPROM for the type information.
    data_ptr = (uint8_t *)&type;
    type_offset = (uint8_t)offsetof(EEPROM_INFO, sensor_type);

    for (int i = 0; i < 4; ++i)
    {
        // Read a byte.
        if (!fpga_ptr->i2cRead(SENSOR_I2C_ADDR, i + type_offset, data_ptr++, FI_FPGA::I2C_SENSOR))
        {
            LOG("%s: Error reading Sensor board EEPROM!", __PRETTY_FUNCTION__);
            type = 0;
            break;
        }
    }
    LOG("%s: sensor_type=0x%08x", __PRETTY_FUNCTION__, type);

    return type;
}

/*!
 * @brief Load the version of the sensor in from the procfs.
 *
 * @note: The assumption is that the length of the array is at least
 *        GIGE_SENSOR_VER_LEN bytes.
 *
 * @param sensor_ver    Location to put the version information.
 *
 * @return The length of the sensor version, 0 if there was an error.
 */
int Sensor::get_sensor_ver(char *sensor_ver, int buflen)
{
    int  i;

    // Build up our sensor version string.
    snprintf((char *)sensor_ver, buflen, "%c%c%c%c%c%c%c%c%c%c-%c%c%c%c\n",
             eeprom.part_no[5], eeprom.part_no[6],
             eeprom.part_no[7], eeprom.part_no[8],
             eeprom.part_no[9], eeprom.part_no[10],
             eeprom.part_no[11], eeprom.part_no[12],
             eeprom.part_no[13], eeprom.part_no[14],
             eeprom.serial_no[0], eeprom.serial_no[1],
             eeprom.serial_no[2], eeprom.serial_no[3]);

    // Sanity check the string.
    for (i = 0; i < (int)strlen((const char *)sensor_ver); i++)
    {
        if (!isalnum(sensor_ver[i]) && !isspace(sensor_ver[i]) && !ispunct(sensor_ver[i]))
        {
            fi_log(LOG_WARNING, "%s: Bad sensor ID [%s]\n", __FUNCTION__, sensor_ver);

            // We should ONLY have alphanumeric values. This is
            // likely a bogus value. Clear out the array.
            memset(sensor_ver, 0, GIGE_SENSOR_VER_LEN);

            // We can't return nothing, so just make it a space.
            sensor_ver[0] = ' ';
            break;
        }
    }

    // Just to make sure...
    sensor_ver[GIGE_SENSOR_VER_LEN - 1] = 0;

    // Return our string.
    return strlen((const char *)sensor_ver);
}

/*!
 * @brief Get the name of the black reference file to use.
 *
 * The black reference name can have sensor settings encoded into it as
 * an optimization so we can keep multiple black reference files around.
 * What settings are important depend on the sensor.
 *
 * @return pointer to a statically allocated fully-qualified file name.
 */
const char *Sensor::get_blkref_name()
{
    static std::string filename;

    // The default is to always use the same black reference file. We will warn the
    // user if we can't use this one. If we wanted to maintain multiple files as an
    // optimization, we would still need to check for compatibility since we can't
    // expect to encode every possible "issue" into the filename.
    filename = fmt::sprintf("%s/%s", get_home_path(), CT_GT_FILE::USER_BLKREF_FILE);
    return filename.c_str();
}

