/*!
 * @file shmem_stream.cpp
 *
 * @brief This file contains the implementation of the shared
 *        memory streaming function to manage image transfers
 *        to the shared memory ring buffer.
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

//#define ALM_TODO_LIVEMA 1
//#define ALM_TODO_WRITE_TO_SHMEM 1

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include <memory>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>
#include <fastec/fas_sharedmem.h>

#include "bounce.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "part_mgr.h"
#include "fcp.h"
#include "fi_internal.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "shmem_stream.h"
#include "pb_control.h"
#include "threads.h"

namespace FASTEC_SHMEM_STREAM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const double LIVE_DELAY_TIME        = 0.05;         //!< Amount of time between start of frames. Try for 20 FPS.
const uint32_t LIVE_FRAME_AGEOUT    = 30000;        //!< Number of microseconds, slightly more than half the frame time.
const int SS_DELAY_TIME         = 10000;    //!< Amount of time to wait for a single step frame change.
const uint32_t BUFFER_DELAY     = 30000;    //!< Number of milliseconds to wait for a buffer.
const int ERROR_PAUSE           = 5000;     //!< Number of milliseconds to pause after an error.

//! How much will we move in a single file system "write".
//! NOTE: Make sure we fit within a bounce buffer! Nominally,
//!   this means 256 MBytes.
const uint32_t MAX_SINGLE_WRITE = 32 * 1024 * 1024;
const uint32_t MAX_SINGLE_READ = MAX_SINGLE_WRITE;

//! How much will we DMA from image memory to our bounce buffer.
//! NOTE: Make sure that the value is on a "page" boundary (4K should work).
const uint32_t MAX_SINGLE_DMA = MAX_SINGLE_WRITE;

//! The maximum bandwidth we can reasonably support (2000 fps @ 1920x1080, 8bpp).
const uint32_t MAXIMUM_BYTES_PER_SEC = (uint32_t)(2000U * 1920U * 1080U);

//! How many DMA buffers do we want.
const int DMA_BB_COUNT = 8;

//! Uninitialized bounce buffer index.
const int DMA_BB_UNINITIALIZED = -1;

//! Define our return values from waitNewModeOrTimeout().
const int WAIT_RECEIVED_NEW_MODE = 0;
const int WAIT_TIMEOUT = 1;

//! How big is our bounce buffer?
//! NOTE: Make sure the total size fits within the MAXIMUM bounce
//!       buffer size, becaue we're only allocating one.
const uint32_t TOTAL_DMA_BB_SIZE = DMA_BB_COUNT * MAX_SINGLE_DMA;

enum SM_TRACE
{
    SM_FUNC         = 0x00000010,
    SM_STATS        = 0x00000020,
    SM_DOWNLOAD     = 0x00000040,
    SM_LIVE         = 0x00000080,
    SM_STATE        = 0x00000100,
    SM_PLAYBACK     = 0x00000200,
    SM_DETAILS      = 0x00000400,
    SM_DMA_STREAM   = 0x00000800,
};

enum DOWNLOAD_STATES
{
    DNLD_SETUP,
    DNLD_GET_BUFFER,
    DNLD_PROCESS_FRAME,
    DNLD_TRANSFER_FRAME,
    DNLD_ENCODE_FRAME,
    DNLD_SEND_FRAME,
    DNLD_CHECK_MODE
};

enum LIVE_STATES
{
    LIVE_SETUP,
    LIVE_WAIT_FOR_ARM,
    LIVE_PROCESS_FRAMES
};

enum SINGLESTEP_STATES
{
    SS_SETUP,
    SS_WAIT_FOR_FRAME,
    SS_PROCESS_FRAME,
    SS_SEND_FRAME,
    SS_CHECK_MODE
};

enum PLAYBACK_STATES
{
    PB_SETUP,
    PB_GET_FRAME,
    PB_SEND_FRAME
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    SM_FUNC       |
    SM_STATS      |
    SM_DOWNLOAD   |
    SM_LIVE       |
    SM_STATE      |
    SM_PLAYBACK   |
    SM_DETAILS    |
    SM_DMA_STREAM |
    TRACE_0;        // Set our trace level to TRACE_0 to trace.
//    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

//--------------------------------------------------------------------
//! @brief The ShmemStreamWorker class.
//--------------------------------------------------------------------

class ShmemStreamWorker : public FAS_THREAD::Thread
{
public:
    virtual ~ShmemStreamWorker();

    // Singletons should not be cloneable.
    ShmemStreamWorker(ShmemStreamWorker& other) = delete;

    // Singletons should not be assignable.
    void operator=(const ShmemStreamWorker&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static ShmemStreamWorker *factory();

    static int32_t getEndFrameNumber();
    static int32_t getStartFrameNumber();
    static int32_t getSaveEndFrame();
    static int32_t getSaveStartFrame();
    static int32_t getSeqStartFrame(int streaming_mode);
    static uint32_t getStreamFormat();
    static uint32_t getFrameInterval();
    static bool playbackForward();
    static uint32_t getInterframeDelay();

    void run() override;
    void stop() override;

    void createPool();
    constexpr int getBbIndex()          { return bb_index; }
    constexpr uint64_t getBbOffset()    { return bb_addr; }
    uint32_t getBbLen();
    int  setMode(int mode);
    void ssUpdate();
    int  startThread(int mode);

    uint32_t isMaAvail()
    {
#ifdef ALM_TODO_LIVEMA
        if (maIsAvail == nullptr)
            return false;
        if (fipHandle == 0)
            return false;
        return (*maIsAvail)(fipHandle);
#else
        return false;
#endif /* ALM_TODO_LIVEMA */
    }

    bool getFrame(uint8_t *ampFramePtr)
    {
#ifdef ALM_TODO_LIVEMA
        if (maGetFrame == nullptr)
            return false;
        if (fipHandle == 0)
            return false;
        return ((*maGetFrame)(fipHandle, ampFramePtr) == FAS_INTERNAL::FIP_OK);
#else
        (void)ampFramePtr;
        return false;
#endif /* ALM_TODO_LIVEMA */
    }

private:
    ShmemStreamWorker();

    void cleanup();

    bool liveWaitForArm();
    void liveStream();

    int captureLiveFrame();
    void downloadInit(uint32_t *proc_info_ptr, uint32_t *dma_mod_ptr);
    int encodeFrame(uint8_t *image_ptr, JPEG_PROC::BUF_REQUEST *request);
    int getLiveFrame();
    int ipmFrame(uint32_t proc_info);

    int ssCheckMode(int& delay_msecs);
    int ssProcessFrame(JPEG_PROC::IPM_REQUEST *request);
    int ssSetup(uint32_t& dma_mod, uint32_t& proc_info);

    void pbSetup();

    int waitForBuffer(uint32_t msecs);
    int waitNewModeOrTimeout(uint32_t msecs);

    int doDownloadLoop();
    int doLiveLoop();
    int doPlaybackLoop();
    int doSinglestepLoop();

private:
    static std::unique_ptr<ShmemStreamWorker> instance;
    static std::mutex instance_mutex;

    volatile bool   thread_running; //!< true if we are running.

    FSM_HANDLE hSharedMem;          //!< Handle to the shared memory interface.

    int current_mode;               //!< The current mode.
    int initial_mode;               //!< The mode at entry when thread is started.
    int requested_mode;             //!< -1 once mode request has been seen and it's safe to request another.
    std::mutex mode_mutex;                  //!< Used to control changes in modes.
    std::condition_variable mode_status;    //!< Condition variable we can block on.

    int last_frame;                 //!< ID of Last frame processed in playback or single-step
    uint32_t completion_percent;    //!< for reporting completion status.

    int frame_num;                  //!< Frame number being processed.
    uint32_t frame_addr_hw;         //!< Frame address in image memory (hexword addr).
    uint32_t frame_counter;         //!< The last live frame capture counter we read.
    uint32_t ipm_image_len;         //!< Length of processed image.
    uint64_t ipm_buf_addr;          //!< Buffer address for processed image in image memory.

    int bb_index;                   //!< The bounce buffer index for our buffer pool space.
    uint64_t bb_addr;               //!< The physical address of our buffer pool space.
    uint8_t *bb_ptr;                //!< The memory mapped base address of the buffer pool space.

    Fastec::IntervalTimer pb_timer;     //!< Timer for statistics during playback.

#ifdef ALM_TODO_LIVEMA
    uint32_t fipHandle;             //!< Handle to the Fastec Internal Proprietary Motion Amplifier logic.

    // Get the handles to the appropriate functions once.
    FiInternal::MA_ADD_FRAME_FN maAddFrame;
    FiInternal::MA_GET_FRAME_FN maGetFrame;
    FiInternal::MA_IS_AVAIL_FN maIsAvail;
#endif /* ALM_TODO_LIVEMA */
};

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

} /* End of FASTEC_SHMEM_STREAM */

// Make things easier for us.
using namespace FASTEC_SHMEM_STREAM;

// ==============================================================
// Static methods should be defined outside the class.
// ==============================================================

std::unique_ptr<ShmemStreamWorker> ShmemStreamWorker::instance;
std::mutex ShmemStreamWorker::instance_mutex;

ShmemStreamWorker *ShmemStreamWorker::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // Check to see if we have our singleton.
    if (instance.get() == nullptr)
    {
        // Create the one and only instance.
        instance = std::unique_ptr<ShmemStreamWorker>(new ShmemStreamWorker());
    }

    return instance.get();
}

/*!
 * @brief Return the ending frame number we want to stream.
 */
int32_t ShmemStreamWorker::getEndFrameNumber()
{
    int32_t   end_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // The "cut out frame" is one past the last frame.
    end_frame = regs[GIGE_REG_PB_CUT_OUT_FRAME].get_value() - 1;
    return end_frame;
}

/*!
 * @brief Return the begining frame number we want to stream.
 */
int32_t ShmemStreamWorker::getStartFrameNumber()
{
    int32_t   start_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    start_frame = regs[GIGE_REG_PB_CUT_IN_FRAME].get_value();
    return start_frame;
}

/*!
 * @brief Return the ending frame number we want to save.
 */
int32_t ShmemStreamWorker::getSaveEndFrame()
{
    int32_t   end_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    end_frame = regs[GIGE_REG_SAVE_END_NUM].get_value();
    return end_frame;
}

/*!
 * @brief Return the begining frame number we want to save.
 */
int32_t ShmemStreamWorker::getSaveStartFrame()
{
    int32_t   start_frame;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    start_frame = regs[GIGE_REG_SAVE_START_NUM].get_value();
    return start_frame;
}

/*!
 * @brief Return the starting frame number, between GIGE_REG_STREAM_START_NUM
 *    and GIGE_REG_STREAM_END_NUM.
 */
int32_t ShmemStreamWorker::getSeqStartFrame(int streaming_mode)
{
    int32_t seq_start;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    seq_start = regs[GIGE_REG_PB_START_FRAME].get_value();

    if (streaming_mode != GIGE_LOCALSTREAM_SINGLESTEP)
    {
        int32_t start_frame;
        int32_t end_frame;

        // Make sure we're in range.
        start_frame = getStartFrameNumber();
        end_frame = getEndFrameNumber();
        if (seq_start < start_frame)
        {
            seq_start = start_frame;
        }
        else if (seq_start > end_frame)
        {
            seq_start = end_frame;
        }
    }

    // Return our frame number.
    return seq_start;
}

/*!
 * @brief return the type of image we should be transmitting ending frame number we want to stream.
 */
uint32_t ShmemStreamWorker::getStreamFormat()
{
    uint32_t stream_format;
    uint32_t image_dest;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    image_dest = regs[GIGE_REG_IMAGE_DESTINATION].get_value();
    stream_format = image_dest & (GIGE_STREAM_FORMAT_MASK | GIGE_STREAM_FORMAT_RAW);

    return stream_format;
}

/*!
 * @brief Return the interval to skip frames. A '1' means to play every frame,
 *    whereas a '10' would mean to play every 10th frame (i.e. 0, 10, 20...).
 */
uint32_t ShmemStreamWorker::getFrameInterval()
{
    uint32_t   frame_interval;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    frame_interval = (regs[GIGE_REG_PB_CONTROL].get_value() & GIGE_PBC_INTERVAL_MASK) >> GIGE_PBC_INTERVAL_SHIFT;
    return frame_interval;
}

/*!
 * @brief Return the direction the user wants to "play back" the data.
 *
 * @return true if we are playing forward, false for reverse.
 */
bool ShmemStreamWorker::playbackForward()
{
    uint32_t   direction;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    direction = regs[GIGE_REG_PB_CONTROL].get_value() & GIGE_PBC_REVERSE;
    return (direction != GIGE_PBC_REVERSE);
}

/*!
 * @brief Return the number of milliseconds we should delay between frames during playback.
 */
uint32_t ShmemStreamWorker::getInterframeDelay()
{
    uint32_t   rate;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    rate = regs[GIGE_REG_PB_FRAME_PERIOD].get_value();
    return rate;
}

//--------------------------------------------------------------------
// ShmemStreamWorker class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. The worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
ShmemStreamWorker::ShmemStreamWorker() :
    FAS_THREAD::Thread(true),
    requested_mode(0),
    last_frame(-1),
    completion_percent(0),
    frame_num(0),
    frame_addr_hw(0),
    frame_counter(0),
    ipm_image_len(0),
    ipm_buf_addr(0)
{
    // Initialize our member variables.
    thread_running = false;
    initial_mode = current_mode = -1;
    bb_index = DMA_BB_UNINITIALIZED;
    bb_addr = 0;
    bb_ptr = (uint8_t *)MAP_FAILED;

    // Assume things will go badly.
    hSharedMem = CameraServer::get_shmem();
    if (hSharedMem == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unexpectedly no shared memory handle?!", FI_FUNCTION);
        return;
    }

#ifdef ALM_TODO_LIVEMA
    // Initialize our handle.
    fipHandle = 0;

    // Is the internal library available?
    if (FiInternal::factory() != nullptr)
    {
        // Get our function callbacks.
        maAddFrame = FiInternal::factory()->getMaAddFrameFuncPtr();
        maGetFrame = FiInternal::factory()->getMaFrameFuncPtr();
        maIsAvail = FiInternal::factory()->getMaIsAvailFuncPtr();
    }
    else
    {
        // Make sure they are initialized.
        maAddFrame = nullptr;
        maGetFrame = nullptr;
        maIsAvail = nullptr;
    }
#endif /* ALM_TODO_LIVEMA */
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 */
ShmemStreamWorker::~ShmemStreamWorker()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Make sure we have a buffer.
    if (bb_index >= 0)
    {
        // Release the memory.
        bb_release(bb_index);

        // Reset the values.
        bb_index = DMA_BB_UNINITIALIZED;
        bb_addr = 0;
        bb_ptr = (uint8_t *)MAP_FAILED;
    }
}

/*!
 * @brief Start the thread running.
 */
int ShmemStreamWorker::startThread(int mode)
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Are we already running?
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: thread is already running!", __PRETTY_FUNCTION__);
        return GEV_STATUS_SUCCESS;
    }

    // Initalize our thread member variables.
    current_mode = -1;
    initial_mode = mode;

    // Get the thread started.
    start();

    // Wait around for the thread to start.
    wait_for_ready(500);

    // Is the thread running?
    if (thread_running)
    {
        // Success!
        return GEV_STATUS_SUCCESS;
    }

    // That didn't work.
    fi_log(LOG_ERR, "%s: timeout waiting for thread start!", __PRETTY_FUNCTION__);
    return GEV_STATUS_ERROR;
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void ShmemStreamWorker::cleanup()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // We're not running anymore.
    thread_running = false;
}

/*!
 * @brief Create the pool of bounce buffers we'll use.
 */
void ShmemStreamWorker::createPool()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize in case of failure.
    bb_index = DMA_BB_UNINITIALIZED;
    bb_addr = 0;
    bb_ptr = (uint8_t *)MAP_FAILED;

    // Get a bounce buffer to hold our pool.
    bb_index = bb_get_avail();
    bb_addr = bb_get_avail_addr(bb_index);

    // Map the memory.
    bb_ptr = (uint8_t *)bb_mmap(bb_index, 0);
    if (bb_ptr == (uint8_t *)MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        goto create_fail;
    }
    fi_log(LOG_DEBUG, "%s: index=%d, addr=0x%" PRIx64 ", ptr=%p", __PRETTY_FUNCTION__, bb_index, bb_addr, (void *)bb_ptr);

    // All done.
    return;

create_fail:
    // Something didn't work. Clean up on the way out.
    if (bb_index >= 0)
    {
        // Release the memory.
        bb_release(bb_index);

        // Reset the values.
        bb_index = DMA_BB_UNINITIALIZED;
        bb_addr = 0;
        bb_ptr = (uint8_t *)MAP_FAILED;
    }
}

/*!
 * @brief Returns the number of bytes in the bounce buffer. This is useful
 *        for the mmap() calls, etc.
 */
uint32_t ShmemStreamWorker::getBbLen()
{
    if (bb_index >= 0)
        return bb_get_mmap_len(bb_index);
    return 0;
}

/*!
 * @brief Wait up to the specified number of milliseconds for the ring buffer
 *        to free up a slot.
 *
 * @param msecs     [IN] Maximum number of milliseconds to wait.
 * @returns Non-zero for error or timeout (ETIMEDOUT), 0 for success.
 */
int ShmemStreamWorker::waitForBuffer(uint32_t msecs)
{
    int  rc;
    struct timespec ts;

    TP(SM_FUNC, "%s: msecs=%d", FI_FUNCTION, msecs);

    // Wait as specified by caller.
    clock_gettime(CLOCK_REALTIME, &ts);
    timespec_add_ns(&ts, (uint64_t)msecs * NSEC_PER_MSEC);

    // Wait for the specified duration.
    rc = fsmServerWaitUntil(hSharedMem, &ts);

    // Did we timeout (or some other error)?
    if (rc != 0)
    {
        // Flag anything that wasn't a timeout.
        if (rc != ETIMEDOUT)
            fi_log(LOG_ERR, "%s: wait returned error %d", FI_FUNCTION, rc);
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Check to see if we received a mode change. Wait for the number
 *        of milliseconds specified by the 'msecs' parameter.
 *
 * @param msecs     [IN] The number of milliseconds to wait for a change, 0 for none.
 * @return WAIT_RECEIVED_NEW_MODE if there is a change in mode, WAIT_TIMEOUT means no change.
 */
int ShmemStreamWorker::waitNewModeOrTimeout(uint32_t msecs)
{
    int rc = WAIT_TIMEOUT;

// Very Noisy!    TP(SM_FUNC, "%s: msecs=%d", __PRETTY_FUNCTION__, msecs);

    // Do we have a mode request pending?
    if (requested_mode != -1)
    {
        TP(SM_DETAILS, "%s: requested_mode=%d", __PRETTY_FUNCTION__, requested_mode);
        rc = WAIT_RECEIVED_NEW_MODE;
    }
    else if (msecs > 0)
    {
        // Wait until something wakes us up or we time out.
        std::unique_lock<std::mutex> lk(mode_mutex);
        bool mode_change = mode_status.wait_for(lk, std::chrono::milliseconds(msecs), [this] { return (requested_mode != -1); });

        // Did we timeout (or some other error)?
        if (mode_change)
        {
            // We had a mode change.
            rc = WAIT_RECEIVED_NEW_MODE;
        }
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Have the FPGA capture a single frame. Wait until the
 *    FPGA is finished, or we timeout.
 *
 * @return 0 on success, -1 on failure.
 */
int ShmemStreamWorker::captureLiveFrame()
{
    int i;
    int rc = -1;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

#if 0
    // ALM: This doesn't seem necessary - just a time sink. We're going to fill
    //      the buffer with image data, so why fill it with zeros first.
    //
    // Clear the ULV buffer.
    fill_memory(pm_ptr->get_ulv_buf_addr(), 0, pm_ptr->get_ulv_buf_size());
#endif

    // Check to see if the ICM is in IDLE mode - possible if we've bailed out of
    // capture or something earlier.
    if (fpga_ptr->icmIsPipelineIdle(nullptr))
    {
        fi_log(LOG_ERR, "%s: The ICM is idle!", __PRETTY_FUNCTION__);
        goto clf_bail;
    }

    // Tell the ICM to grab a frame.
    if (!fpga_ptr->regModify(ICM_CONTROL, ICMC_CAPTURE_SINGLE, -1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to modify ICM_CONTROL!", __PRETTY_FUNCTION__);
        goto clf_bail;
    }

    // Wait for the capture to complete. Why "100"? Seemed like a nice round
    // number that works - that is, it almost never gets exceeded.
    //
    //-------------------------------------------------------------------------
    // TODO: Rather than polling, it might be better to have a conditional wait
    //       with a timeout.
    //-------------------------------------------------------------------------
    for (i = 0; i < 100; i++)
    {
        uint32_t reg_value;
        struct timespec sleep_req;   // Delay period.

        // Check to see if the ICM has finished.
        if (!fpga_ptr->regRead(ICM_STATUS, &reg_value))
        {
            fi_log(LOG_DEBUG, "%s: read_reg(ICM_STATUS) failed!", __PRETTY_FUNCTION__);
            goto clf_bail;
        }

        // Is the ICM still busy?
        if ((reg_value & ICMS_CAPTURE_SINGLE_BUSY) == 0)
            break;

        // Wait for a little before trying again.
        sleep_req.tv_sec = 0;
        sleep_req.tv_nsec = 1000000;
        nanosleep(&sleep_req, NULL);
    }

    if (i < 100)
    {
        // Everything is OK.
        rc = 0;
        TP(SM_STATS, "%s: Capture in %d msec", __PRETTY_FUNCTION__, i);
    }
    else
    {
        LOG("%s: Capture took too long!", __PRETTY_FUNCTION__);
        goto clf_bail;
    }

    // The FPGA always puts the single capture into the ULV buffer.
    frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_ulv_buf_addr());

clf_bail:
    return rc;
}

void ShmemStreamWorker::liveStream()
{
    int i;
    bool done;
    uint32_t frames_per_buffer;             // The number of frames that will fit in a transfer buffer.
    uint32_t img_size;                      // The number of bytes per frame.
    uint32_t xfer_length;                   // The actual number of bytes to transfer/DMA, on a frame boundary.
    uint64_t im_start_address;              // The starting address of the capture partition.
    uint64_t im_end_address;                // One byte past the end of the capture partition.
    std::mutex list_mutex;                  // Control access to all of the lists.
    std::queue<int>  avail_list;
    std::queue<int>  filled_list;
    std::vector<uint64_t> dma_bb_addr(DMA_BB_COUNT);
    std::vector<uint32_t> amt_to_write(dma_bb_addr.size());    // Sent from DMA thread -> WRITER thread.

    CameraServer *cs_ptr = CameraServer::get_cs();
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(SM_FUNC, "%s: bb_addr=0x" PRIx64, __PRETTY_FUNCTION__);

#ifdef ALM_TODO_LIVEMA
    // Setup for MA.
    if (FiInternal::factory() != nullptr)
    {
        FiInternal::MA_GET_HANDLE_FN maGetHandle = FiInternal::factory()->getMaHandleFuncPtr();
        if (maGetHandle != nullptr)
        {
            GIGE_REG_CACHE& regs = CameraServer::get_regs();

            FAS_INTERNAL::FIP_MA_CONFIGURATION configInfo =
            {
                .framerate = (double)regs[GIGE_REG_FRAME_RATE].get_value(),
                .minFreqOfInterest = 10,
                .width = (int)regs[GIGE_REG_WIDTH].get_value(),
                .height = (int)regs[GIGE_REG_HEIGHT].get_value(),
                .factor = 30
            };

            // Start the MA engine.
            fipHandle = maGetHandle(configInfo);
            if (fipHandle == 0)
            {
                fi_log(LOG_ERR, "%s: Unable to get the MA handle!", FI_FUNCTION);
            }
        }
    }
#endif /* ALM_TODO_LIVEMA */

    // Split the one buffer up into our bounce buffers.
    dma_bb_addr[0] = bb_addr;
    avail_list.push(0);
    for (i = 1; i < DMA_BB_COUNT; ++i)
    {
        dma_bb_addr[i] = dma_bb_addr[i - 1] + MAX_SINGLE_DMA;
        avail_list.push(i);
    }

    // Calculate the frame size, taking into account width, height, and bit depth.
    // NOTE: This does NOT include the PFMD.
    img_size = cs_ptr->calc_frame_buf_size();

    // Round it up to a HEXWORD boundary.
    img_size = convert_to_HEXWORD(img_size) << HEXWORD_SHIFT;

    // Sanity check.
    if (img_size == 0)
    {
        fi_log(LOG_DEBUG, "%s: Unexpected image size of 0!", __PRETTY_FUNCTION__);
        return;
    }

    // Can we support this frame rate?
    float frame_rate;
    cs_ptr->get_frame_rate(&frame_rate);
    if ((uint32_t)((frame_rate * img_size) + 0.5) > MAXIMUM_BYTES_PER_SEC)
    {
        fi_log(LOG_DEBUG, "%s: The frame rate (%f) and image size (%d) are not supported", __PRETTY_FUNCTION__, frame_rate, img_size);
        return;
    }

    // Figure out our actual DMA length.
    frames_per_buffer = MAX_SINGLE_DMA / (img_size + FRAME_METADATA_HW_BYTES);
    xfer_length = frames_per_buffer * (img_size + FRAME_METADATA_HW_BYTES);

    // Where is the image memory?
    im_start_address = pm_ptr->getCaptureFrameAddr(0);
    im_end_address = pm_ptr->getCaptureFrameAddr(pm_ptr->getFrameCount());

    // Make sure the shared memory is in a known state.
    int fsm_status;
    fsm_status = fsmServerReset(hSharedMem);
    if (fsm_status)
    {
        fi_log(LOG_DEBUG, "%s: Unable to reset the shared memory ring buffer [0x%x]", __PRETTY_FUNCTION__, fsm_status);
        shmemStreamControlUpdate(nullptr, GIGE_LOCALSTREAM_STOP);
        return;
    }

    // Not done yet.
    done = false;

    // DMA THREAD ======================================
    // Lambda Tutorial:
    // * We are defining a "lambda function" (LF) called "dma_thread".
    // * We are "capturing" local variables for use inside the LF by
    //   specifying them in the square brackets. These are (by default)
    //   "const copy-by-value" (i.e. read-only).
    //   * By specifying the keyword "mutable", these captured values are
    //     now "copy-by-value", i.e. you can modify them locally.
    // * Add a reference indicator (&) to pass by reference and allow
    //   changes to proceed to the outer scope.
    std::thread dma_thread([this,
                            im_start_address,
                            im_end_address,
                            frames_per_buffer,
                            img_size,
                            xfer_length,
                            dma_bb_addr,
                            &amt_to_write,
                            &list_mutex,
                            &avail_list,
                            &filled_list,
                            &done
                           ]()
                           mutable
                           {
                               int dma_index;
                               uint32_t amt_to_dma;
                               uint32_t dma_mod;
                               uint64_t start_read_address;
                               FiFPGA *fpga_ptr;

                               // Get a pointer to our FPGA class.
                               fpga_ptr = CameraServer::get_fpga();
                               if (fpga_ptr == nullptr)
                               {
                                   fi_log(LOG_ERR, "%s: NULL FPGA pointer?!", __PRETTY_FUNCTION__);
                                   this->stop_request = true;
                                   return;
                               }

                               // We are going to transfer the data raw - no formatting.
                               dma_mod = IM2PCIE_DMA_GENERIC_DATA;
                               dma_mod |= IM2PCIE_DMA_HIGHSPEED_BYPASS;

                               // Initialize our addresses.
                               start_read_address = im_start_address;

                               // Loop until we're done pulling all the data down from the camera.
                               while (!done)
    {
        int retry_count;
        uint32_t waits_for_buf;
        uint64_t newest_frame_address = start_read_address;
        uint64_t ready_address;         // The address for a full buffer.

        // Initialize our address. Don't worry about wrapping, that will be
        // caught when we check the "last frame written" value.
        ready_address = start_read_address + xfer_length;

        // Wait for enough data to make a transfer worthwhile.
        TP(SM_DMA_STREAM, "%s [DMA] start waiting for frames.", FI_FUNCTION);
            for (retry_count = 0; retry_count < 10; ++retry_count)
            {
                uint32_t frame_written;

                // Figure out the latest frame written by the FPGA.
                if (!fpga_ptr->regRead(ICM_LAST_FRAME_WRITTEN_PTR, &frame_written))
                {
                    fi_log(LOG_ERR, "%s: Unable to read LAST_FRAME_WRITTEN register", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

                // Are we waiting for the FPGA to write something?
                if (frame_written != FI_FPGA::LAST_FRAME_INIT)
                {
                    // Convert it to a byte address.
                    newest_frame_address = ((uint64_t)frame_written) << HEXWORD_SHIFT;

                    // Do we have enough to transfer, or did we wrap?
                    if ((newest_frame_address >= ready_address) || (newest_frame_address < start_read_address))
                        break;
                }

                // Wait for frames to be captured. Might as well see if there is a change in mode.
                if (waitNewModeOrTimeout(10) == WAIT_RECEIVED_NEW_MODE)
                {
                    // Mode changed.
                    TP(SM_DMA_STREAM, "%s [DMA] MODE_CHANGE_REQUEST.", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

                // Did the user try to cancel?
                if (this->stop_request)
                {
                    TP(SM_DMA_STREAM, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

                // We are going to need the camera state.
                uint32_t camera_state = CameraServer::get_camera_state();
                if ((camera_state != GIGE_CAMERA_STATE_ARMED) && (camera_state != GIGE_CAMERA_STATE_TRIGGERED))
                {
                    TP(SM_DMA_STREAM, "%s [DMA] Camera state changed [%d].", __PRETTY_FUNCTION__, camera_state);
                    done = true;
                    break;
                }
            }

            // Are we done?
            if (done)
                break;

            TP(SM_DMA_STREAM, "%s [DMA] Waiting for frames %d tries, newest frame addr: 0x%" PRIx64, FI_FUNCTION, retry_count, newest_frame_address);

            // Make sure we got something.
            if (newest_frame_address == start_read_address)
            {
                TP(SM_DMA_STREAM, "%s [DMA] No frames captured.", FI_FUNCTION);
                continue;
            }

            // Figure out how much we can DMA this time.
            // Did we wrap?
            if (newest_frame_address < start_read_address)
            {
                amt_to_dma = (im_end_address - start_read_address);
                TP(SM_DMA_STREAM, "%s [DMA] Wrapping capture.", FI_FUNCTION);
            }
            else
                amt_to_dma = (newest_frame_address - start_read_address);
            if (amt_to_dma > xfer_length)
                amt_to_dma = xfer_length;

            // Make sure there is a buffer available.
            waits_for_buf = 0;
            while (avail_list.empty())
            {
                // Wait a buffer to be released. Might as well see if there is a change in mode.
                if (waitNewModeOrTimeout(1) == WAIT_RECEIVED_NEW_MODE)
                {
                    // Mode changed.
                    TP(SM_DMA_STREAM, "%s [DMA] MODE_CHANGE_REQUEST.", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

                // Did the user try to cancel?
                if (this->stop_request)
                {
                    TP(SM_DMA_STREAM, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

                // Count how many times we wait.
                ++waits_for_buf;
            }

            // Bail if there is a stop request.
            if (done)
            {
                TP(SM_DMA_STREAM, "%s [DMA] MODE CHANGE/STOP REQUEST waiting for buffers.", FI_FUNCTION);
                break;
            }

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);

                // Get the next available buffer index.
                dma_index = avail_list.front();
                avail_list.pop();
            }
            // Release access to the lists.

            // Have the FPGA transfer the data to our transfer buffer.
            TP(SM_DMA_STREAM, "%s [DMA]: waited=%d, buf=%d, im_addr=0x%" PRIx64 ", amt_to_dma=%u", __PRETTY_FUNCTION__, waits_for_buf, dma_index, start_read_address, amt_to_dma);
            int dma_status = fpga_ptr->imageDownload(dma_mod, 0, start_read_address, dma_bb_addr[dma_index], amt_to_dma);
            if (dma_status != 0)
            {
                fi_log(LOG_ERR, "%s: DMA failed [0x%x]. Exiting.", __PRETTY_FUNCTION__, dma_status);
                shmemStreamControlUpdate(nullptr, GIGE_LOCALSTREAM_STOP);
                done = true;
                break;
            }

            // Let the writer know how much data is available.
            amt_to_write[dma_index] = amt_to_dma;

            // Update our image memory address.
            start_read_address += amt_to_dma;

            // Make sure there's room for one more frame.
            if ((start_read_address + (img_size + FRAME_METADATA_HW_BYTES)) >= im_end_address)
            {
                TP(SM_DMA_STREAM, "%s [DMA] Wrapping start address.", FI_FUNCTION);
                start_read_address = im_start_address;
            }

            // Lock access to the lists.
            {
                std::unique_lock<std::mutex> locker(list_mutex);
                filled_list.push(dma_index);
            }
            // Release access to the lists.

            // Did we get a mode change?
            if (waitNewModeOrTimeout(0) == WAIT_RECEIVED_NEW_MODE)
            {
                // Mode changed.
                TP(SM_DMA_STREAM, "%s [DMA] MODE_CHANGE_REQUEST.", __PRETTY_FUNCTION__);
                done = true;
                break;
            }

            // Did the user try to cancel?
            if (this->stop_request)
            {
                TP(SM_DMA_STREAM, "%s [DMA] STOP_REQUEST.", __PRETTY_FUNCTION__);
                done = true;
                break;
            }
        }
                           });
    // DMA THREAD ======================================

    // WRITING THREAD ==================================
    std::thread writer_thread([this,
                               img_size,
                               dma_bb_addr,
                               &amt_to_write,
                               &list_mutex,
                               &avail_list,
                               &filled_list,
                               &done]()
                              mutable
                              {
                                  int write_index = 0;
                                  int32_t frame_counter(0);

                                  // Loop until we've written all the data out to the destination.
                                  while (!done)
                                  {
                                      int frame_index;
                                      int frames_to_transfer;
                                      uint32_t buffer_size;
                                      uint32_t waits_for_buf;
                                      uint64_t bb_addr;

                                      // Make sure there is a buffer available.
                                      waits_for_buf = 0;
                                      while (filled_list.empty())
                                          {
                                              // Wait a buffer to be released. Might as well see if there is a change in mode.
                                              if (waitNewModeOrTimeout(1) == WAIT_RECEIVED_NEW_MODE)
                                                  {
                                                      // Mode changed.
                                                      TP(SM_DMA_STREAM, "%s [WRITER] MODE_CHANGE REQUEST.", __PRETTY_FUNCTION__);
                                                      done = true;
                                                      break;
                                                  }

                                                  // Did the user try to cancel?
                                                  if (this->stop_request)
            {
                TP(SM_DMA_STREAM, "%s [WRITER] STOP_REQUEST.", __PRETTY_FUNCTION__);
                    done = true;
                    break;
                }

    // We are going to need the camera state.
    uint32_t camera_state = CameraServer::get_camera_state();
                            if ((camera_state != GIGE_CAMERA_STATE_ARMED) && (camera_state != GIGE_CAMERA_STATE_TRIGGERED))
{
    TP(SM_DMA_STREAM, "%s [WRITER] Camera state changed [%d].", __PRETTY_FUNCTION__, camera_state);
        done = true;
        break;
    }

    // Count how many times we wait.
    ++waits_for_buf;
                                      }

    // Bail if there is a stop request.
    if (done)
    break;

    // Lock access to the lists.
{
    std::unique_lock<std::mutex> locker(list_mutex);

        // Get the next filled buffer index.
        write_index = filled_list.front();
        filled_list.pop();
    }
    // Release access to the lists.

    // Save the amount of data in a local variable.
    buffer_size = amt_to_write[write_index];
                  amt_to_write[write_index] = 0;

                  // How many frames are we talking?
                  frames_to_transfer = buffer_size / (img_size + FRAME_METADATA_HW_BYTES);

                  // What is the base address.
                  bb_addr = dma_bb_addr[write_index];

                  // Calculate the address based on the offset.
                  uint8_t *image_ptr = this->bb_ptr + (bb_addr - this->bb_addr);

                  // Transfer the frame pointer to the shared memory ring buffer.
                  TP(SM_DMA_STREAM, "%s [WRITER]: waited=%d, buf=%d, size=%u, %d frames, bb_addr=0x%" PRIx64 ", image_ptr=%p", __PRETTY_FUNCTION__, waits_for_buf, write_index, buffer_size, frames_to_transfer, bb_addr, (void *)image_ptr);
                  for (frame_index = 0; frame_index < frames_to_transfer; ++frame_index)
{
#ifdef ALM_TODO_WRITE_TO_SHMEM
    // Make sure there is room.
    if (fsmIsFull(this->hSharedMem))
        {
            fi_log(LOG_ERR, "%s [WRITER]: shared memory ring buffer is full!", FI_FUNCTION);
            break;
        }

        // Put the data into the shared memory space.
        fsmSetInfo(this->hSharedMem, bb_addr + (frame_index * (img_size + FRAME_METADATA_HW_BYTES)), img_size, frame_counter);
        fsmWriteDone(this->hSharedMem);
#endif /* ALM_TODO_WRITE_TO_SHMEM */

#ifdef ALM_TODO_LIVEMA
        // Do we have the internal library, and are we doing LiveMA?
        if ((fipHandle != 0) && (maAddFrame != nullptr))
        {
            // Allocate a buffer to pass to the MA logic.
            uint8_t *buffer_ptr = new uint8_t[img_size];
            if (buffer_ptr == nullptr)
            {
                fi_log(LOG_ERR, "%s [WRITER]: Unable to allocate memory for image %d", FI_FUNCTION, frame_index);
                break;
            }

            // Move the data from the bounce buffer.
            memcpy(buffer_ptr, image_ptr + (frame_index * (img_size + FRAME_METADATA_HW_BYTES)), img_size);

            // Add the frame to the reference stack. The MA logic will free the memory.
            (*maAddFrame)(fipHandle, buffer_ptr);
        }
#endif /* ALM_TODO_LIVEMA */

        // Increment our frame number, checking for a wrap.
        ++frame_counter;
        if (frame_counter < 0)
            frame_counter = 0;
    }

    // Lock access to the lists.
    {
        std::unique_lock<std::mutex> locker(list_mutex);
        avail_list.push(write_index);
    }
    // Release access to the lists.

    // See if there is a change in mode.
    if (waitNewModeOrTimeout(0) == WAIT_RECEIVED_NEW_MODE)
{
    // Mode changed.
    TP(SM_DMA_STREAM, "%s [WRITER] MODE_CHANGE REQUEST.", __PRETTY_FUNCTION__);
        done = true;
        break;
    }

    // Did the user try to cancel?
    if (this->stop_request)
{
    TP(SM_DMA_STREAM, "%s [WRITER] STOP_REQUEST.", __PRETTY_FUNCTION__);
        done = true;
        break;
    }
                                  }
                              });
    // WRITING THREAD ==================================

    // Wait for everyone to finish.
    dma_thread.join();
    writer_thread.join();

#ifdef ALM_TODO_LIVEMA
    // Close the MA engine.
    if (FiInternal::factory() != nullptr)
    {
        FiInternal::MA_CLOSE_FN maClose = FiInternal::factory()->getMaCloseFuncPtr();
        if (maClose != nullptr)
        {
            (*maClose)(fipHandle);
            fipHandle = 0;
        }
    }
#endif /* ALM_TODO_LIVEMA */

    TP(SM_FUNC, "%s: done", __PRETTY_FUNCTION__);
}

bool ShmemStreamWorker::liveWaitForArm()
{
    bool rc;

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Stay here until we have a mode change or the camera is in
    // the ARM/TRIGGER state.
    while (1)
    {
        uint32_t camera_state;

        // We are going to need the camera state.
        camera_state = CameraServer::get_camera_state();
        if ((camera_state == GIGE_CAMERA_STATE_ARMED) || (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
        {
            // We are in the correct state.
            TP(SM_FUNC, "%s: Camera is ARMED/TRIGGERED [%d]", __PRETTY_FUNCTION__, camera_state);
            rc = true;
            break;
        }

        // Wait for either a mode change or time to get the next frame.
        else if (waitNewModeOrTimeout(10) == WAIT_RECEIVED_NEW_MODE)
        {
            // We had a mode change request.
            TP(SM_FUNC, "%s: Got a mode change.", __PRETTY_FUNCTION__);
            rc = false;
            break;
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Ask the FPGA to capture a single frame.
 * @return 0 on success, negative on error.
 */
int ShmemStreamWorker::getLiveFrame()
{
    int rc = -1;
    uint32_t camera_state;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    Fastec::IntervalTimer frame_timer;

    // We are going to need the camera state.
    camera_state = CameraServer::get_camera_state();
    TP(SM_FUNC, "%s: camera_state=%d", __PRETTY_FUNCTION__, camera_state);

    // Statistics
    frame_timer.start();
    // Statistics

    // Live state?
    if (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        // Setup to capture a frame.
        rc = captureLiveFrame();
        if (rc != 0)
        {
            fi_log(LOG_ERR, "%s: Could not capture a live frame!", __PRETTY_FUNCTION__);
            goto glf_bail;
        }
    }

    //===============================================================================================
    //===============================================================================================
    // ALM_TODO:
    // In armed or triggered, go into "hyper-drive" mode and move EVERY frame into the shared
    // memory space.
    //===============================================================================================
    //===============================================================================================
    else if ((camera_state == GIGE_CAMERA_STATE_ARMED) || (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        PartMgr *pm_ptr = CameraServer::get_partmgr();
        uint32_t part_state = pm_ptr->getPartitionState();

        // Determine what the partition state is.
        if ((part_state == GIGE_PART_STATE_ARMED) || (part_state == GIGE_PART_STATE_TRIGGERED))
        {
            uint32_t reg_value;

            // Read the frame counter.
            if (!fpga_ptr->regRead(ICM_CAPTURE_FRAME_COUNT_LIVE, &reg_value))
            {
                fi_log(LOG_ERR, "%s: Unable to read ICM_CAPTURE_FRAME_COUNT_LIVE register", __PRETTY_FUNCTION__);
                goto glf_bail;
            }

            // Have we captured any frames?
            if (reg_value == frame_counter)
            {
                // Go ahead and reuse the reg_value variable, 'cuz we're bailing out anyhow.
                fpga_ptr->regRead(ICM_STATUS, &reg_value);
                fi_log(LOG_DEBUG, "%s: Still at the same frame buffer [0x%08x]. ICM_STATUS=0x%08x", __PRETTY_FUNCTION__, frame_counter, reg_value);
                goto glf_bail;
            }

            // Remember the counter for next time.
            frame_counter = reg_value;

            // Figure out which buffer the ICM used.
            if (!fpga_ptr->regRead(ICM_LAST_FRAME_WRITTEN_PTR, &frame_addr_hw))
            {
                fi_log(LOG_ERR, "%s: Unable to read LAST_FRAME_WRITTEN register", __PRETTY_FUNCTION__);
                goto glf_bail;
            }
        }
        else
        {
            // The partition state is not in the same state we would expect. This could be normal,
            // if for example the capture just completed.
            fi_log(LOG_DEBUG, "%s: wrong part_state [0x%x]", __PRETTY_FUNCTION__, part_state);
            goto glf_bail;
        }
    }
    else
    {
        // Not a valid state.
        goto glf_bail;
    }

    // All done.
    rc = 0;

glf_bail:
    // Statistics
    frame_timer.stop();
    TP(SM_STATS, "%s: frame capture: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
    // Statistics

    return rc;
}

/*!
 * @brief Handle live streaming until the mode changes.
 */
int ShmemStreamWorker::doLiveLoop()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Wait to be in arm/trigger state.
    bool done(false);
    while (!done)
    {
        // We need the camera to actually be armed.
        if (liveWaitForArm())
        {
            // Get to work.
            liveStream();
        }

        // Did the user try to cancel?
        if (this->stop_request)
        {
            TP(SM_DMA_STREAM, "%s: STOP_REQUEST.", __PRETTY_FUNCTION__);
            done = true;
        }
        else if (waitNewModeOrTimeout(0) == WAIT_RECEIVED_NEW_MODE)
        {
            TP(SM_DMA_STREAM, "%s: MODE_CHANGE REQUEST.", __PRETTY_FUNCTION__);
            done = true;
        }
    }

    // All done.
    TP(SM_FUNC, "%s: done, current_mode=%d, requested_mode=%d", __PRETTY_FUNCTION__, current_mode, requested_mode);
    return 0;
}

int ShmemStreamWorker::ssCheckMode(int& delay_msecs)
{
    double rate;
    int ret_state = SS_CHECK_MODE;

    TP(SM_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Update our buffer ID to the next frame we want.
    if (playbackForward())
    {
        frame_num += getFrameInterval();
        if (frame_num > getEndFrameNumber())
        {
            // Wrap our frame number back around to the start.
            frame_num = getStartFrameNumber() + (frame_num - (getEndFrameNumber() + 1));
        }
        TP(SM_PLAYBACK, "%s: playback forward frame=%d", __PRETTY_FUNCTION__, frame_num);
    }
    else
    {
        frame_num -= getFrameInterval();
        if (frame_num < getStartFrameNumber())
        {
            // Wrap our frame number back around to the end.
            frame_num = getEndFrameNumber() - (getStartFrameNumber() - (frame_num + 1));
        }
        TP(SM_PLAYBACK, "%s: playback reverse frame=%d", __PRETTY_FUNCTION__, frame_num);
    }

    // Setup to acquire the next frame.
    ret_state = SS_PROCESS_FRAME;

    // Now stop the timer. All per-frame processing has completed.
    pb_timer.stop();

    // What is the delay we're supposed to have between frames?
    rate = (double)getInterframeDelay() / 1000.0;

    // Determine how much time is left of our interframe delay.
    delay_msecs = (int)((rate - pb_timer.lastMeasurement) * 1000.0);
    if (delay_msecs < 0)
    {
        TP(SM_STATS, "%s: Loop took too long (%lf)!", __PRETTY_FUNCTION__, pb_timer.lastMeasurement);
        delay_msecs = 0;
    }

    return ret_state;
}

int ShmemStreamWorker::ssProcessFrame(JPEG_PROC::IPM_REQUEST *request)
{
    int rc;
    int ret_state;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(SM_PLAYBACK, "%s: frame_num=%d", __PRETTY_FUNCTION__, frame_num);

    // Get the current time so we can calculate how much time each frame
    // takes to proceses.
    pb_timer.start();

    // Get the address of the frame we're going to process.
    frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_active_frame_addr(frame_num, false));

    // Convert to a full address.
    request->image_addr = ((uint64_t)frame_addr_hw) << HEXWORD_SHIFT;

    // Get the memory pointer associated with the buffer.
    request->jpeg_buffer = bb_ptr;

    // Get the frame processed by the IPM.
    rc = jpeg_ipm_request(request);
    if (rc)
    {
        // Move on to transferring the frame.
        ret_state = SS_SEND_FRAME;
    }
    else
    {
        // Go try to acquire another frame.
        fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
        ret_state = SS_CHECK_MODE;
    }

    // Move on.
    return ret_state;
}

int ShmemStreamWorker::ssSetup(uint32_t& dma_mod, uint32_t& proc_info)
{
    int next_state = SS_PROCESS_FRAME;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Sensor  *sensor_ptr = CameraServer::get_sensor();

    TP(SM_PLAYBACK, "%s", __PRETTY_FUNCTION__);

    // Initialize our frame number and IPM buffer ID.
    frame_num = getSeqStartFrame(current_mode);

    // Processing info is always the same for playback images.
    dma_mod = IM2PCIE_DMA_8BIT_RAW_FMT | IM2PCIE_DMA_BYTE_ORDER_R2L | IM2PCIE_DMA_WORD_ORDER_R2L;
    if (sensor_ptr->is_sensor_color())
    {
        // Color images are NV12.
        proc_info = IMGPROC_TYPE_NV12;
    }
    else
    {
        // Mono images are Y8.
        proc_info = IMGPROC_TYPE_Y8;
    }
    proc_info |= IMGPROC_USE_FPN;

    // Only update status and last frame if we're in playback mode.
    if (current_mode == GIGE_LOCALSTREAM_PLAYBACK)
    {
        // Set the status register to indicate we're busy.
        regs[GIGE_REG_PB_STATUS].set_value(GIGE_PBS_BUSY);
    }

    // Initialize pertinent registers incase someone looks.
    TP(SM_PLAYBACK, "%s: start=%u, end=%u", __PRETTY_FUNCTION__, getStartFrameNumber(), getEndFrameNumber());

    // Move on.
    return next_state;
}

void ShmemStreamWorker::pbSetup()
{
    uint32_t pb_control;

    TP(SM_PLAYBACK, "%s: start=%u, end=%u", __PRETTY_FUNCTION__, getStartFrameNumber(), getEndFrameNumber());

    // Make sure we start in the right place.
    pbc_update_start_frame(nullptr, getSeqStartFrame(current_mode));

    // Let folks know we're playing.
    pb_control = GIGE_PBC_ENABLE;
    if (!playbackForward())
        pb_control |= GIGE_PBC_REVERSE;
    if (getFrameInterval() == 0)
        pb_control |= (1 << GIGE_PBC_INTERVAL_SHIFT);
    else
        pb_control |= ((getFrameInterval() << GIGE_PBC_INTERVAL_SHIFT) & GIGE_PBC_INTERVAL_MASK);
    pbc_update_control(nullptr, pb_control);
}

/*!
 * @brief Handle playback until the mode changes.
 *
 * In the playback loop, we wait for a signal or timeout.
 * Then we check whether the mode has changed or there is
 * a new frame to process.
 */
int ShmemStreamWorker::doPlaybackLoop()
{
    int rc = 0;
    int pb_state;
    bool pb_done;
    PB_CONTROL::FRAME_REQUEST frame_request;
    Fastec::IntervalTimer frame_timer;

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    pb_state = PB_SETUP;

    // Keep going until told to stop.
    pb_done = false;
    while (!pb_done)
    {
        int msecs;

        TP(SM_STATE, "%s: state=%d", __PRETTY_FUNCTION__, pb_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (pb_state)
        {
        case PB_SETUP:
            // Initialize the rest of the system.
            pbSetup();

            // Make sure this first time, we ask for the first available frame.
            memset(&frame_request, 0, sizeof(frame_request));
            frame_request.frame_num = -1;

            // Move to the next state.
            pb_state = PB_GET_FRAME;
            break;
        case PB_GET_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // This will block until the playback controller has
            // a new frame.
            if (pbc_request(&frame_request))
            {
                // Got our frame!
                pb_state = PB_SEND_FRAME;
            }
            else
            {
                // Failure here may be because the user cancelled playback.
                // Just in case it was something else, put a little delay in.
                msecs = 10;
            }

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: Get frame %u: %lf secs.", __PRETTY_FUNCTION__, frame_request.frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case PB_SEND_FRAME:
            TP(SM_PLAYBACK, "%s: Send Playback frame %d to client", __PRETTY_FUNCTION__, frame_request.frame_num);

            // Copy the frame over to the shared memory.
            memcpy(bb_ptr, frame_request.frame_buffer, frame_request.frame_len);

            // Release the buffer.
            pbc_release_frame(frame_request.frame_buffer);

            // Clear out some of the request fields. Make sure we keep the frame_num
            // field intact, so that we will only get a new frame (not a repeat).
            frame_request.frame_buffer = nullptr;
            frame_request.frame_len = 0;

            // Back to waiting for a frame.
            pb_state = PB_GET_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, pb_state);
            pb_done = true;
            break;
        }

        // Check for a mode change or wait the specified amount of time to get
        // the next frame.
        if (waitNewModeOrTimeout(msecs) == WAIT_RECEIVED_NEW_MODE)
        {
            // We are done.
            pb_done = true;
            rc = 0;
        }
    }

    // Return our status.
    TP(SM_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

void ShmemStreamWorker::downloadInit(uint32_t *proc_info_ptr, uint32_t *dma_mod_ptr)
{
    bool is_color;
    bitmode_t bitmode;
    uint32_t stream_format;
    uint32_t proc_info;
    uint32_t dma_mod;
    uint32_t format_type;
    std::unique_ptr<pm_metadata> pm_meta_ptr;
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> dnld_header_buffer; // The file header information for the current download operation.

    TP(SM_DOWNLOAD, "%s", __PRETTY_FUNCTION__);

    // Initialize our control variables to figure out what's going on.
    pm_meta_ptr = pm_ptr->getActivePartMetadata();
    stream_format = getStreamFormat();
    format_type = stream_format & GIGE_STREAM_FORMAT_MASK;
    bitmode = regs[GIGE_REG_BIT_MODE].get_value();
    is_color = sensor_ptr->is_sensor_color();

    // This processing info will be needed to set up the IPM.
    proc_info = get_imgproc_type(stream_format, bitmode, is_color);
    if ((proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_JPEG)
    {
        // Specify the more accurate processing type for JPEG.
        if (is_color)
            proc_info = IMGPROC_TYPE_NV12;
        else
            proc_info = IMGPROC_TYPE_Y8;
    }
    if ((regs[GIGE_REG_FPN].get_value() & GIGE_FPN_FPGA_ENABLE_MASK) != 0)
        proc_info |= IMGPROC_USE_FPN;

    // If we're saving bitmaps, then we need to make allowances.
    if (format_type == GIGE_STREAM_FORMAT_BMP)
        proc_info |= IMGPROC_ROW_WD_ALIGN;

    // Initialize our frame number and IPM buffer ID.
    frame_num = getSaveStartFrame();

    // Get the DMA settings we want to use.
    dma_mod = get_stream_dma_mod(stream_format, bitmode, is_color);

    // Pass the values back up.
    *proc_info_ptr = proc_info;
    *dma_mod_ptr = dma_mod;
}

/*!
 * @brief JPEG encode a frame. Relevant information is found in
 *    the ipm_info object.
 *
 * @return 0 on success, negative otherwise.
 */
int ShmemStreamWorker::encodeFrame(uint8_t *image_ptr, JPEG_PROC::BUF_REQUEST *request)
{
    int rc = -1;
    uint32_t uv_offset;  // Number of bytes (HEXWORD alignment) to the UV-plane.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Statistics
    Fastec::IntervalTimer jpeg_timer;
    // Statistics

    TP(SM_FUNC, "%s: image_ptr=%p", __PRETTY_FUNCTION__, (void *)image_ptr);

    // Get our dimensions.
    request->width = regs[GIGE_REG_WIDTH].get_value();
    request->height = regs[GIGE_REG_HEIGHT].get_value();

    // Calculate our length for the Y plane so we can find the start ov the UV plane.
    // We transfered the frame directly from image memory, so the UV plane will be on
    // a HEXWORD boundary.
    uv_offset = convert_to_HEXWORD(request->width * request->height) << HEXWORD_SHIFT;

    // Now setup the pointers to the data.
    request->src_pixels_Y = image_ptr;
    request->src_pixels_UV = image_ptr + uv_offset;

    // ----------------------------------------------------------------------
// #define ALM_SAVE_DATA
#ifdef ALM_SAVE_DATA
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;
        uint32_t yuv_len;

        if (request->pixel_format == JPEG_UTIL::Y8)
            yuv_len  = request->width * request->height;
        else
            yuv_len = uv_offset + ((request->width * request->height) >> 1);

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.raw", file_no);
        // ++file_no;   // Done when we save the encoded data.
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(image_ptr, yuv_len, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_DATA */
    // ----------------------------------------------------------------------

    // Statistics
    jpeg_timer.start();

    // JPEG encode the data.
    if (jpeg_buf_request(request))
        rc = 0;

    // Statistics
    jpeg_timer.stop();
    TP(SM_STATS, "%s: JPEG encoding: %u bytes, %lf secs.", __PRETTY_FUNCTION__, request->jpeg_len, jpeg_timer.lastMeasurement);
    // Statistics

    // ----------------------------------------------------------------------
// #define ALM_SAVE_JPG
#ifdef ALM_SAVE_JPG
    {
        FILE *fd;
        char file_path[PATH_MAX];
        static int file_no = 0;

        snprintf(file_path, PATH_MAX, "/tmp/image_%03d.jpg", file_no);
        ++file_no;
        fd = fopen(file_path, "wb");
        if (fd != NULL)
        {
            fwrite(request->jpeg_buffer, request->jpeg_len, 1, fd);
            fclose(fd);
        }
    }
#endif /* ALM_SAVE_JPG */
    // ----------------------------------------------------------------------

    return rc;
}

/*!
 * @brief Have the FPGA's IPM process the frame we just captured. All
 *    of the relevant information can be found in the ipm_info object.
 *
 * @param proc_info IPM specific processing information (see IMGPROC_XXX).
 * @return 0 on success, negative otherwise.
 */
int ShmemStreamWorker::ipmFrame(uint32_t proc_info)
{
    int rc = -1;
    IPM_PROC::IPM_REQUEST ipm;

    // Clear out our request structure.
    memset(&ipm, 0, sizeof(IPM_PROC::IPM_REQUEST));

    // Initialize the fields.
    ipm.image_addr = ((uint64_t)frame_addr_hw) << HEXWORD_SHIFT;
    ipm.proc_info = proc_info;

    // Get the raw pixel data processed properly through the IPM depending on
    // our current camera state.
    ipm.process_live = false;

    // Process the frame.
    if (!ipmProcRequest(&ipm))
    {
        fi_log(LOG_ERR, "%s: ipm_request for addr 0x%" PRIx64 " failed!", __PRETTY_FUNCTION__, ipm.image_addr);
        ipm_buf_addr = 0;
        ipm_image_len = -1;
        goto if_bail;
    }

    // Grab the output buffer address and the length.
    ipm_buf_addr = ipm.ipm_buf_addr;
    ipm_image_len = ipm.image_len;

    // We are done.
    rc = 0;

if_bail:
    // All done.
    return rc;

}

/*!
 * @brief Handle download until the mode changes.
 *
 * In the download loop, we wait for a signal or timeout.
 * Then we check whether the mode has changed or there is
 * a new buffer available that we can write to.
 *
 * @return 0 on success, or an error code otherwise.
 */
int ShmemStreamWorker::doDownloadLoop()
{
    bool done;
    int rc = 0;
    int download_state;
    int frame_count(0);
    int transfer_count(0);
    int jpeg_bb_index(0);
    uint8_t *jpeg_bb_ptr(nullptr);
    uint64_t jpeg_bb_addr(0);
    uint32_t dma_mod(0);
    uint32_t dma_dims(0);
    uint32_t proc_info;
    uint32_t pc_img_len(0);
    struct timespec ts;
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    PartMgr *pm_ptr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Fastec::IntervalTimer dnld_timer;
    Fastec::IntervalTimer frame_timer;

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    download_state = DNLD_SETUP;

    // Get the current time so we can calculate how much time the download takes.
    dnld_timer.start();

    // Keep going until told to stop.
    done = false;
    while (!done)
    {
        uint32_t msecs;

        TP(SM_STATE, "%s: state=%d", __PRETTY_FUNCTION__, download_state);

        // By default, don't hang around. Just check for a mode change and then carry on.
        msecs = 0;

        // Process based on the current state.
        switch (download_state)
        {
        case DNLD_SETUP:
            // Processing info is always the same for playback images.
            downloadInit(&proc_info, &dma_mod);

            // How many bytes will the processed frame be in PC memory?
            pc_img_len = get_processed_size(proc_info, regs[GIGE_REG_WIDTH].get_value(), regs[GIGE_REG_HEIGHT].get_value());

            // If we're JPEG encoding the image, then we'll need a bounce buffer.
            if ((getStreamFormat() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG)
            {
                // Get a buffer for our image processing.
                jpeg_bb_index = bb_get_avail();
                jpeg_bb_addr = bb_get_avail_addr(jpeg_bb_index);

                // Map the memory.
                jpeg_bb_ptr = (uint8_t *)bb_mmap(jpeg_bb_index, 0);
                if (jpeg_bb_ptr == (uint8_t *)MAP_FAILED)
                {
                    LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
                    break;
                }
            }
            else
            {
                // Make sure these are all reset. Don't want any confusion.
                jpeg_bb_index = DMA_BB_UNINITIALIZED;
                jpeg_bb_addr = 0;
                jpeg_bb_ptr = nullptr;
            }

            // How many frames are we downloading.
            // NOTE: The assumption is that frame numbers go from smallest to largest!
            frame_count = (getSaveEndFrame() - getSaveStartFrame()) + 1;
            if (frame_count <= 0)
            {
                fi_log(LOG_ERR, "%s: Frame numbers are wrong! Is end (%d) before start (%d)?", __PRETTY_FUNCTION__, getSaveEndFrame(), getSaveStartFrame());
                done = true;
                break;
            }

            // We haven't transferred any yet.
            transfer_count = 0;

            // The image dimensions will be constant.
            dma_dims = (regs[GIGE_REG_HEIGHT].get_value() << IPM_FRAME_DIM_Y_SHIFT) | regs[GIGE_REG_WIDTH].get_value();

            // Initialize pertinent registers incase someone looks.
            TP(SM_DOWNLOAD, "%s: cutin=%u start=%d cutout=%u proc=0x%08x dma=0x%08x", __PRETTY_FUNCTION__, getSaveStartFrame(), frame_num, getSaveEndFrame(), proc_info, dma_mod);

            // Move on.
            download_state = DNLD_GET_BUFFER;
            break;
        case DNLD_GET_BUFFER:
            // Wait for a buffer, mode change, or timeout.
            rc = waitForBuffer(BUFFER_DELAY);

            // Move on.
            TP(SM_DOWNLOAD, "%s: Got buffer", __PRETTY_FUNCTION__);
            download_state = DNLD_PROCESS_FRAME;
            break;
        case DNLD_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Get the address of the frame we're going to process.
            frame_addr_hw = convert_to_HEXWORD(pm_ptr->get_active_frame_addr(frame_num, false));
            TP(SM_DOWNLOAD, "%s: frame_addr_hw=0x%08x", __PRETTY_FUNCTION__, frame_addr_hw);

            // Get the frame processed by the IPM.
            rc = ipmFrame(proc_info);
            if (rc == 0)
            {
                // Move on to transferring the frame.
                download_state = DNLD_TRANSFER_FRAME;
            }
            else
            {
                // Go try to acquire another frame.
                fi_log(LOG_ERR, "%s: Could not start frame processing", __PRETTY_FUNCTION__);
                msecs = ERROR_PAUSE;
                download_state = DNLD_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: IPM: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case DNLD_TRANSFER_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Are we JPEG encoding?
            if ((getStreamFormat() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG)
            {
                // Have the FPGA transfer the image into our bounce buffer memory.
                rc = fpga_ptr->imageDownload(dma_mod, dma_dims, ipm_buf_addr,
                                             jpeg_bb_addr, ipm_image_len, pc_img_len);
                if (rc >= 0)
                {
                    // Move on to processing the frame.
                    download_state = DNLD_ENCODE_FRAME;
                    rc = 0;
                }
                else
                {
                    // Go try to acquire another frame.
                    fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
                    msecs = ERROR_PAUSE;
                    download_state = DNLD_CHECK_MODE;
                }
            }
            else
            {
                // Have the FPGA transfer the image straight into the shared memory space.
                rc = fpga_ptr->imageDownload(dma_mod, dma_dims, ipm_buf_addr,
                                             bb_addr, ipm_image_len, pc_img_len);
                if (rc >= 0)
                {
                    // Hand the buffer over the the shared memory interface.
                    download_state = DNLD_SEND_FRAME;
                    rc = 0;
                }
                else
                {
                    // Go try to acquire another frame.
                    fi_log(LOG_ERR, "%s: Unable to DMA frame!", __PRETTY_FUNCTION__);
                    msecs = ERROR_PAUSE;
                    download_state = DNLD_CHECK_MODE;
                }
            }

            // Release our IPM buffer.
            ipm_proc_release(ipm_buf_addr);
            ipm_buf_addr = 0;
            ipm_image_len = 0;

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: DMA: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case DNLD_ENCODE_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // Now that we have the frame in our system memory. Do we want to encode it?
            if ((getStreamFormat() & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_JPEG)
            {
                JPEG_PROC::BUF_REQUEST request;

                // Initialize our request structure with what we know.
                if ((proc_info & IMGPROC_TYPE_MASK) == IMGPROC_TYPE_NV12)
                    request.pixel_format = JPEG_UTIL::NV12;
                else
                    request.pixel_format = JPEG_UTIL::Y8;
                request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();
                request.jpeg_buffer = nullptr;
                request.jpeg_len = 0;

                // JPEG encode the frame. All status information is returned
                // in the request structure, so we can ignore the return value.
                encodeFrame(jpeg_bb_ptr, &request);

                // Wait long enough...but not too long.
                clock_gettime(CLOCK_REALTIME, &ts);
                ts.tv_sec += 2;

                if ((rc == 0) && (request.jpeg_len > 0) && (request.jpeg_buffer != nullptr))
                {
                    // Copy the JPEG encoded data to the shared buffer we got earlier.
                    memcpy(bb_ptr, request.jpeg_buffer, request.jpeg_len);

                    // Release the buffer.
                    jpeg_release_frame(request.jpeg_buffer);

                    // Hand the buffer over the the shared memory interface.
                    download_state = DNLD_SEND_FRAME;
                }
                else
                {
                    // Go try to acquire another frame.
                    fi_log(LOG_WARNING, "%s: Unable to encode the frame! Did we change states?", __PRETTY_FUNCTION__);
                    msecs = ERROR_PAUSE;
                    download_state = DNLD_CHECK_MODE;
                }
            }
            else
            {
                fi_log(LOG_ERR, "%s: Invalid state (DNLD_ENCODE_FRAME) for non-JPEG buffer!", __PRETTY_FUNCTION__);

                // Wait long enough...but not too long.
                clock_gettime(CLOCK_REALTIME, &ts);
                ts.tv_sec += 2;

                // Go try to acquire another frame.
                msecs = ERROR_PAUSE;
                download_state = DNLD_CHECK_MODE;
            }

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: encode: %lf secs.", __PRETTY_FUNCTION__, frame_timer.lastMeasurement);
            // Statistics
            break;
        case DNLD_SEND_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            // One more frame sent.
            transfer_count++;

            // Check for a mode change before getting the next frame.
            download_state = DNLD_CHECK_MODE;

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: Frame %d sent: %lf secs.", __PRETTY_FUNCTION__, frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case DNLD_CHECK_MODE:
            // Are we done?
            if (transfer_count == frame_count)
            {
                // We're done.
                done = true;
                rc = 0;
                break;
            }

            // Move to the next frame.
            frame_num++;

            // Setup to acquire the next frame.
            download_state = DNLD_GET_BUFFER;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, download_state);
            download_state = DNLD_CHECK_MODE;
            break;
        }

        // Check for a mode change.
        if (waitNewModeOrTimeout(msecs) == WAIT_RECEIVED_NEW_MODE)
        {
            // We are done.
            done = true;
            rc = 0;
        }
    }

    // Now stop the timer. All per-frame processing has completed.
    dnld_timer.stop();
    TP(SM_STATS, "%s: download done in %lf secs.", __PRETTY_FUNCTION__, dnld_timer.lastMeasurement);

    // Release the bounce buffer.
    bb_release(jpeg_bb_index);

    // Release the IPM buffer, if we have one.
    if (ipm_buf_addr != 0)
    {
        ipm_proc_release(ipm_buf_addr);
        ipm_buf_addr = 0;
    }

    // Return our status.
    TP(SM_DOWNLOAD, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief Single step mode is a playback mode. The difference
 *    is that the user controls the pace of the frame
 *    updates, not the camera server.
 *
 * In the loop, we wait for a signal or timeout.
 * Then we check whether the mode has changed or there is
 * a new frame to process.
 */
int ShmemStreamWorker::doSinglestepLoop()
{
    int rc = 0;
    int ss_state;
    int cur_frame;
    bool ss_done;
    struct timespec ts;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Fastec::IntervalTimer frame_timer;
    JPEG_PROC::IPM_REQUEST request;

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Initialize our state variable.
    ss_state = SS_SETUP;

    // Keep going until told to stop.
    ss_done = false;
    while (!ss_done)
    {
        uint32_t msecs;

        TP(SM_STATE, "%s: state=%d", __PRETTY_FUNCTION__, ss_state);

        // By default we want to check for a mode change, but we don't want to wait.
        msecs = 0;

        // Process based on the current state.
        switch (ss_state)
        {
        case SS_SETUP:
            // Clear out the result values.
            memset(&request, 0, sizeof(request));
            request.quality = regs[GIGE_REG_JPEG_QUALITY].get_value();

            // There's no time limit on the frame.
            request.time_limit = 0;

            // Get our dimensions.
            request.width = regs[GIGE_REG_WIDTH].get_value();
            request.height = regs[GIGE_REG_HEIGHT].get_value();

            // Initialize the rest of the system.
            ssSetup(request.dma_mod, request.proc_info);

            // Wait for a new frame.
            ss_state = SS_WAIT_FOR_FRAME;
            break;
        case SS_WAIT_FOR_FRAME:
            // See if we should start processing a new frame.
            cur_frame = regs[GIGE_REG_CSI_RVW_FRAME_NUM].get_value();
            if (cur_frame != last_frame)
            {
                TP(SM_PLAYBACK, "%s: new frame %d", __PRETTY_FUNCTION__, cur_frame);

                // Remember the frame we're working on.
                last_frame = frame_num = cur_frame;

                // Move to the right state.
                ss_state = SS_PROCESS_FRAME;
            }
            else
            {
                // Wait for something to happen.
                msecs = SS_DELAY_TIME;
            }
            break;
        case SS_PROCESS_FRAME:
            // Statistics
            frame_timer.start();
            // Statistics

            ss_state = ssProcessFrame(&request);

            // Statistics
            frame_timer.stop();
            TP(SM_STATS, "%s: frame %u processed: %lf secs.", __PRETTY_FUNCTION__, frame_num, frame_timer.lastMeasurement);
            // Statistics
            break;
        case SS_SEND_FRAME:
            TP(SM_PLAYBACK, "%s: Sent frame %d to client", __PRETTY_FUNCTION__, frame_num);

            // Wait long enough...but not too long.
            clock_gettime(CLOCK_REALTIME, &ts);
            ts.tv_sec += 2;

            // Wait for the next frame.
            ss_state = SS_WAIT_FOR_FRAME;
            break;
        default:
            fi_log(LOG_ERR, "%s: Unexpected state! [%d]", __PRETTY_FUNCTION__, ss_state);
            ss_state = SS_WAIT_FOR_FRAME;
            break;
        }

        // Wait for awhile, or until someone wakes us up.
        if (waitNewModeOrTimeout(msecs) == WAIT_RECEIVED_NEW_MODE)
        {
            // Do we have a mode request pending?
            if (requested_mode != -1)
            {
                // We are done.
                ss_done = true;
                rc = 0;
            }
        }
    }

    // Set the status register to indicate we're idle.
    regs[GIGE_REG_PB_STATUS].set_value(0);

    // Return our status.
    TP(SM_FUNC, "%s: Done [%d]", __PRETTY_FUNCTION__, rc);
    return rc;
}

/*!
 * @brief The worker thread.
 */
void ShmemStreamWorker::run()
{
    int rc;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    fi_log(LOG_DEBUG, "%s(%d): PID=%u", __PRETTY_FUNCTION__, initial_mode, get_tid());

    // Don't allow two threads and always start with no mode set.
    if ((current_mode != -1) || thread_running)
    {
        fi_log(LOG_WARNING, "%s: Start fail: running=%d current_mode=%d!", FI_FUNCTION, (thread_running ? 1 : 0), current_mode);
        return;
    }

    // Initialize our settings.
    thread_running = true;
    current_mode = initial_mode;
    requested_mode = -1;

    // Let folks know we are ready.
    notify_ready();

    // Update our register settings.
    regs[GIGE_REG_SMSTREAM_CONTROL].set_value(current_mode);
    fcp_notify_apps(GIGE_REG_SMSTREAM_CONTROL, current_mode);

    // Keep looking until the current mode changes to GIGE_LOCALSTREAM_STOP.
    for (;;)
    {
        int previous_mode;

        TP(SM_STATE, "%s: Top; current_mode == %d", __PRETTY_FUNCTION__, current_mode);

        rc = 0;
        switch (current_mode)
        {
        case GIGE_LOCALSTREAM_LIVE:
            rc = doLiveLoop();
            break;
        case GIGE_LOCALSTREAM_PLAYBACK:
            rc = doPlaybackLoop();
            break;
        case GIGE_LOCALSTREAM_DOWNLOAD:
            rc = doDownloadLoop();
            break;
        case GIGE_LOCALSTREAM_SINGLESTEP:
            rc = doSinglestepLoop();
            break;
        case GIGE_LOCALSTREAM_STOP:
            break;
        }

        // Remember our current mode, so we can tell if we change.
        previous_mode = current_mode;

        // Did we have an error?
        if (rc != 0)
        {
            fi_log(LOG_ERR, "%s: mode %d failed (%d), stopping", __PRETTY_FUNCTION__, current_mode, rc);

            // Set the mode in shared memory.
            current_mode = GIGE_LOCALSTREAM_STOP;
            requested_mode = -1;
        }

        // Are we supposed to change mode?
        else if (requested_mode != -1)
        {
            current_mode = requested_mode;
            requested_mode = -1;
        }

        // Did we change modes?
        if (previous_mode != current_mode)
        {
            // Update our register value.
            regs[GIGE_REG_SMSTREAM_CONTROL].set_value(current_mode);
            fcp_notify_apps(GIGE_REG_SMSTREAM_CONTROL, current_mode);
        }

        // Are we supposed to stop?
        if (current_mode == GIGE_LOCALSTREAM_STOP)
            break;
    }

    // Clean up on the way out.
    cleanup();
}

/*!
 * @brief   Tell the worker thread to enter the requested mode,
 *      starting the thread if necessary.
 *
 * @param   mode_value
 *
 * return GEV_STATUS... value
 */
int ShmemStreamWorker::setMode(int mode_value)
{
    TP(SM_FUNC, "%s: 0x%08x", __PRETTY_FUNCTION__, mode_value);

    // Are we trying to stop the thread?
    if (mode_value == GIGE_LOCALSTREAM_STOP)
    {
        // We want to stop, and we are stopped.
        if (!thread_running)
            return GEV_STATUS_SUCCESS;
    }
    else if (!thread_running)
    {
        // We must be trying to start it.
        return startThread(mode_value);
    }

    // Set the requested mode for the thread to pick up on.
    {
        std::unique_lock<std::mutex> lk(mode_mutex);
        requested_mode = mode_value;
    }
    mode_status.notify_all();

    // All done.
    return GEV_STATUS_SUCCESS;
}

void ShmemStreamWorker::ssUpdate()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Ask the thread to stop.
 */
void ShmemStreamWorker::stop()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Tell the thread to stop.
    FAS_THREAD::Thread::stop();
}

/*!
 * @brief Update shared memory stream control.
 *
 * @param p_msg     Not used.
 * @param reg_value New stream control setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise a GEV_STATUS... error
 */
int shmemStreamControlUpdate(CS_NETWORK::CS_NETWORK_MESSAGE * /*p_msg*/, uint32_t reg_value)
{
#if defined(ALM_TODO_WRITE_TO_SHMEM) || defined(ALM_TODO_LIVEMA)
    uint8_t part_state;
    PartMgr *pm_ptr = CameraServer::get_partmgr();

    TP(SM_FUNC, "%s: 0x%08x", __PRETTY_FUNCTION__, reg_value);

    // If not in the Review state, can download but not playback or single step.
    if ((CameraServer::get_camera_state() != GIGE_CAMERA_STATE_REVIEW) &&
        ((reg_value == GIGE_LOCALSTREAM_PLAYBACK) || (reg_value == GIGE_LOCALSTREAM_SINGLESTEP)))
    {
        fi_log(LOG_ERR, "%s: wrong state for playback", __FUNCTION__);
        return GEV_STATUS_WRONG_CONFIG;
    }

    if ((reg_value != GIGE_LOCALSTREAM_PLAYBACK) &&
        (reg_value != GIGE_LOCALSTREAM_SINGLESTEP) &&
        (reg_value != GIGE_LOCALSTREAM_DOWNLOAD) &&
        (reg_value != GIGE_LOCALSTREAM_LIVE) &&
        (reg_value != GIGE_LOCALSTREAM_STOP))
    {
        fi_log(LOG_ERR, "%s: illegal option %u", __FUNCTION__, reg_value);
        return GEV_STATUS_ERROR;
    }

    // If we don't have a recording in the active partition, STOP and LIVE are the only legal requests.
    if ((part_state = pm_ptr->getPartitionState()) != GIGE_PART_STATE_COMPLETE)
    {
        // Partition state is not COMPLETE, so we can only support live or stop.
        if ((reg_value != GIGE_LOCALSTREAM_LIVE) && (reg_value != GIGE_LOCALSTREAM_STOP))
        {
            fi_log(LOG_ERR, "%s: partition is in the wrong state! [0x%x]", __FUNCTION__, part_state);
            return GEV_STATUS_WRONG_CONFIG;
        }
    }

    // Make sure we are running.
    if (ShmemStreamWorker::factory()->getBbIndex() == DMA_BB_UNINITIALIZED)
        shmemStreamInit();

    // Change the mode { STOP, LIVE, DOWNLOAD, SINGLESTEP, PLAYBACK }
    return ShmemStreamWorker::factory()->setMode((int)reg_value);
#else
    (void)reg_value;
    return GEV_STATUS_NOT_IMPLEMENTED;
#endif /* defined(ALM_TODO_WRITE_TO_SHMEM) || defined(ALM_TODO_LIVEMA) */
}

/*!
 * @brief Single step the local stream interface.
 */
void shmemStreamSsUpdate()
{
    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);
    ShmemStreamWorker::factory()->ssUpdate();
}

/*!
 * @brief CS application startup shared memory initialization
 *
 * @returns non-zero for failure, 0 for success.
 */
int shmemStreamInit()
{
    uint32_t buf_size;
    uint64_t buf_addr;

    TP(SM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Get the worker started.
    ShmemStreamWorker *worker = ShmemStreamWorker::factory();
    if (worker == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to start the ShmemStreamWorker!", FI_FUNCTION);
        return ENOMEM;
    }

    // Create our buffer "pool".
    ShmemStreamWorker::factory()->createPool();

    // Initialize our buffer addresses.
    buf_addr = ShmemStreamWorker::factory()->getBbOffset();
    buf_size = ShmemStreamWorker::factory()->getBbLen();

    // Did we actually get anything?
    if ((buf_addr == 0) || (buf_size == 0))
    {
        fi_log(LOG_ERR, "%s: Unable to allocate our bounce buffer!", FI_FUNCTION);
        return ENOMEM;
    }

    // Initialize the shared memory control structures.
    fsmServerUpdateMemInfo(CameraServer::get_shmem(), buf_addr, buf_size);

    // Make sure the client knows we've shifted the base around.
    fsmServerSetFault(CameraServer::get_shmem(), FSM_SERVER_BASE_CHANGE);

    // All ready.
    return 0;
}

bool shmemMAIsActive()
{
    bool rc(false);

#ifdef ALM_TODO_LIVEMA
    // Do we have a stream worker?
    if (ShmemStreamWorker::factory() != nullptr)
    {
        // Ask if the MA is available.
        rc = ShmemStreamWorker::factory()->isMaAvail();
    }
#endif /* ALM_TODO_LIVEMA */

    // Return our result.
    return rc;
}

bool shmemMAGetFrame(uint8_t *frame_ptr)
{
    bool rc(false);

#ifdef ALM_TODO_LIVEMA
    // Do we have a stream worker?
    if (ShmemStreamWorker::factory() != nullptr)
    {
        // Ask if the MA is available.
        rc = ShmemStreamWorker::factory()->getFrame(frame_ptr);
    }
#else
    (void)frame_ptr;
#endif /* ALM_TODO_LIVEMA */

    return rc;
}
