/*!
 * @file storage.cxx
 *
 * @brief This file contains all the routines to deal with storage
 *    media, such as SDcards, USB and SSDs.
 *
 * Copyright 2012 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#define __USE_XOPEN_EXTENDED 500

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <libudev.h>
#include <mntent.h>
#include <stddef.h>
#include <unistd.h>

#include <blkid/blkid.h>
#include <fmt/printf.h>
#include <linux/types.h>

#include <iostream>
#include <fstream>
#include <filesystem>
#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <fmt/chrono.h>
#include <fmt/printf.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/types.h>

#include <fastec/fas_log.h>
#include <fastec/fas_filesystem.h>
#include <fastec/fas_video.h>

#include "storage.h"

#include "camera_server.h"
#include "constants.h"
#include "cs_trace.h"
#include "formatter.h"
#include "image_io.h"
#include "lic_manager.h"
#include "part_mgr.h"

namespace STORAGE_LOCAL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// How many bytes in a block.
const uint32_t DEFAULT_BLOCK_SIZE = 512;

// Make sure the buffer size is aligned to page boundaries (4096 bytes).
const uint32_t XFER_BUFFER_LEN  = 1024 * 1024;

// The length of our temporary information buffer.
const int INFO_BUFFER_LEN   = 1024;

// The length for our working temporary buffer.
const int MAX_TEMP_BUFFER_LEN = 10240;

// The metadata file name.
const char *METADATA_FILENAME   = "/metadata.txt";

// The various models we know about.
static struct ssd_model_list_t
{
    const char *name;
    uint32_t capabilities;
} ssd_model_list[] =
{
    { "Samsung SSD 960", GIGE_SSD_CAP_SANITIZE },   // NVMe supports secure ERASE, not SANITIZE.
    { "Samsung SSD 970", GIGE_SSD_CAP_SANITIZE },   // NVMe supports secure ERASE, not SANITIZE.
};

//! File operations can take a very long time, so we have
//! a seperate thread to perform the actual task.
class file_action_thread : public FAS_THREAD::Thread
{
private:
    volatile bool   thread_running;         //!< Flag to indicate if we are running or not.

    uint32_t    file_action;                //!< Defines what our job is.
    struct stat file_stat;                  //!< Used to determine if we're file or folder.

    uint32_t    file_offset;                //!< Offset to where the target contents will be found.
    uint32_t    target_offset;              //!< Length of the target_path before adding contents.
    uint32_t    media_sel;

    std::string file_path;                  //!< A local copy of the specified file path.
    std::string target_path;                //!< For copy, the path of the destination.
    std::string media_path;                 //!< For copy, the destination root.

    void    cleanup();

    int do_copy();
    int do_delete();
    int do_save_logs();

public:
    file_action_thread();
    virtual ~file_action_thread();

    virtual void run();
    virtual void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr);
    virtual void stop();

    bool    is_stop_requested()             { return stop_request; }

    bool    is_running()                    { return thread_running; }
    void    set_action(uint32_t action)     { file_action = action; }
    uint32_t    get_action()                { return file_action; }

    void    set_media_path(const char *sel_path, uint32_t path_len);
    void    set_path(const char *sel_path, uint32_t path_len);
    const std::string& get_target_path()    { return target_path; }
    uint32_t    get_target_offset()         { return target_offset; }
    const std::string& get_file_path()      { return file_path; }
    uint32_t    get_file_offset()           { return file_offset; }

    static bool thread_copy_file(const char *target_ptr, const char *source_ptr);

};

//--------------------------------------------------------------------
// Variable Declarations:
//--------------------------------------------------------------------

//! The status of our worker thread.
static uint32_t fa_status;

// Information used to copy files.
static uint8_t  *xfer_buffer;
static long xfer_page_size;

static uint32_t last_frame_count;

static char info_buffer[ INFO_BUFFER_LEN ];

// This holds the JSON media data to be transferred to the user.
std::string json_message;

// Label for a CAP stack "codec".
static constexpr char CAPSTACK_LABEL[] = "CAPSTK";


} /* End of namespace STORAGE_LOCAL */

using namespace STORAGE_LOCAL;
using namespace CS_TRACE;

TRACE_GLOBAL;

// Get our local class variables declared.
std::unique_ptr<storage> storage::local_storage;
std::mutex storage::storage_mutex;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

/*!
 * @brief Determine the size of the files in the folder pointed
 *    to by dir_path. Sort of like a "du".
 *
 * We look for the first image file and get its size, then multiply
 * that by the number of frames which we read from the metadata.
 * This is much faster than walking the whole tree to get the size.
 *
 * @param dir_path  Folder path we're interested in.
 * @param num_frames    So we can estimate using only one file.
 *
 * @return Number of bytes used by the folder contents.
 */
uint64_t storage::calc_stack_size(const char *dir_path, uint32_t num_frames)
{
    FILE *du_file;
    uint64_t totalsize = 0;
    std::string temp_str;

//    fi_log(LOG_DEBUG, "%s: dir_path='%s' num_frames=%u", __PRETTY_FUNCTION__, dir_path, num_frames);

    // Check to see if the user wants to cancel.
    if (storage::factory()->fa_task->is_stop_requested())
    {
        TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
        return 0;
    }

    // Once this starts, there's no stopping.
    temp_str = fmt::sprintf("du -s -b \"`find \"%s\" -type f ! -name \"*.txt\" ! -name \"*.cfg\" ! -name \"*.xml\" ! -name \"*.pm\" -print | head -1`\" > /tmp/size.tmp", dir_path);
    system(temp_str.c_str());

    // Open our file and parse out the info.
    du_file = fopen("/tmp/size.tmp", "r");
    if (!du_file)
        return 0;

    // Read the line from the file.
    if (fgets(info_buffer, INFO_BUFFER_LEN, du_file))
        totalsize = strtoull(info_buffer, NULL, 0);

    // Close our file.
    fclose(du_file);

    // Now finish our calculations - totalsize is currently the
    // size (in bytes) of ONE file. Adjust to be the BYTE count
    // of ALL files.
    totalsize = (totalsize * num_frames);

    // Return our findings.
    return totalsize;
}

/*!
 * @brief Determine the size of the files in the folder pointed
 *    to by dir_path using "du".
 *
 * @param dir_path  Folder path we're interested in.
 *
 * Is like calc_stack_size() but does not depend on knowing the
 * number of frames, so it can be much slower.
 *
 * @return Number of bytes used by the folder contents.
 */
uint64_t storage::calc_dir_size(const char *dir_path)
{
    FILE *du_file;
    std::string temp_str;
    uint64_t    totalsize = 0;

    // Check to see if the user wants to cancel.
    if (storage::factory()->fa_task->is_stop_requested())
    {
        TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
        return 0;
    }

    // Once this starts, there's no stopping and it can take a while.
    {
        std::lock_guard<std::mutex> lock(storage_mutex);

        temp_str = fmt::sprintf("du -s \"%s\" > /tmp/size.tmp", dir_path);
        system(temp_str.c_str());

        // Open our file and parse out the info.
        du_file = fopen("/tmp/size.tmp", "r");
        if (!du_file)
            return 0;

        // Read the line from the file.
        if (fgets(info_buffer, INFO_BUFFER_LEN, du_file))
            totalsize = strtoull(info_buffer, NULL, 0);

        // Close our file.
        fclose(du_file);
    }

    // Now finish our calculations - totalsize is in KBytes, so convert to bytes.
    totalsize *= 1024LL;

    // Return our findings.
    return totalsize;
}

/*!
 * @brief Calculate the approximate capacity of a device.
 *
 * @param mount_path    The mount path we're interested in.
 *
 * @return 0 if successful, one of errno otherwise.
 */
int storage::get_capacity(const char *mount_path, uint64_t *p_total, uint64_t *p_avail, uint64_t *p_inodes)
{
    int err_code;
    struct statvfs  sfs;

    std::lock_guard<std::mutex> lock(storage_mutex);

    // Sanity check.
    if ((p_total == nullptr) || (p_avail == nullptr) || (p_inodes == nullptr))
    {
        fi_log(LOG_DEBUG, "%s: can't pass in nullptrs!", __PRETTY_FUNCTION__);
        err_code = EINVAL;
        goto bail;
    }
    if (strlen(mount_path) == 0)
    {
        err_code = EINVAL;
        goto bail;
    }

    // Get our information. The statvfs() works with a path of any file within the
    // mounted filesystem.
    if (statvfs(mount_path, &sfs) < 0)
    {
        // Save the error.
        err_code = errno;

        // Log the error.
        fi_log(LOG_DEBUG, "%s: statvfs(%s) returned %d:'%s'", __PRETTY_FUNCTION__, mount_path, errno, strerror(errno));
        goto bail;
    }

    // Figure out how many bytes we're talking about.
    *p_total = (uint64_t)sfs.f_bsize * (uint64_t)sfs.f_blocks;
    *p_avail = (uint64_t)sfs.f_bsize * (uint64_t)sfs.f_bfree;

    // Return the number of indoes available.
    *p_inodes = (uint64_t)sfs.f_ffree;
//    fi_log(LOG_DEBUG, "%s: f_files=0x%" PRIx64 ", f_ffree=0x%" PRIx64 ", f_favail=0x%" PRIx64, __PRETTY_FUNCTION__, sfs.f_files, sfs.f_ffree, sfs.f_favail);

    // Everything looks good.
    err_code = 0;

bail:
    // Return our result.
    return err_code;
}

/*!
 * @brief Return a pointer to the file action thread object.
 */
file_action_thread *storage::get_fa_task()
{
    return fa_task.get();
}

/*!
 * @brief Say whether there is enough space to copy a file or stack
 *
 * @param file_path The source path to copy.
 * @param media_ptr The destination media
 *
 * @return true if there's enough space, false if not or if other error.
 */
bool storage::copy_space_available(const char *file_path, const char *media_ptr)
{
    int err_code;
    uint64_t source_size;
    uint64_t total, avail;
    uint64_t inodes;
    struct stat sb;

    // Get the statistics on the source.
    if (stat(file_path, &sb) != 0)
    {
        // If we can't find out, just assume it will work.
        fi_log(LOG_ERR, "%s: Can't stat '%s'", __PRETTY_FUNCTION__, file_path);
        return false;
    }

    // If our source is just a file, get the size of it.
    if (S_ISREG(sb.st_mode))
    {
        source_size = sb.st_mode;
    }
    else if (S_ISDIR(sb.st_mode))
    {
        source_size = storage::calc_dir_size(file_path);
    }
    else
    {
        fi_log(LOG_ERR, "%s: Can't get size of '%s'", __PRETTY_FUNCTION__, file_path);
        return false;
    }

    // If we can't find out space available on the device, go ahead and try anyway.
    if ((err_code = storage::get_capacity(media_ptr, &total, &avail, &inodes)) != 0)
    {
        fi_log(LOG_DEBUG, "%s: get_capacity() failed [%d].", __PRETTY_FUNCTION__, err_code);
        return false;
    }

    // Add 1% to the source size for the size check.
    if ((source_size + (source_size / 100)) >= avail)
    {
        fi_log(LOG_ERR, "%s: no room for %" PRIu64 " bytes, only %" PRIu64 " available", __PRETTY_FUNCTION__, source_size, avail);
        return false;
    }

    return true;
}

void storage::get_internal_ssd_devpath(std::string& devpath)
{
    FILE *bridge_file;
    const int ALTERA_ID = 1172;
    std::string temp_str;
    std::size_t index;

    // Clear out the string.
    devpath.clear();

    // Find the bridge path, if it's up.
    temp_str = fmt::sprintf("lspci -PP -d %d: | cut -d'/' -f1-2 > /tmp/bridge.txt", ALTERA_ID);
    system(temp_str.c_str());

    // Open our file and parse out the info.
    bridge_file = fopen("/tmp/bridge.txt", "r");
    if (!bridge_file)
        return;

    // Read the information.
    info_buffer[0] = 0;
    fgets(info_buffer, INFO_BUFFER_LEN, bridge_file);

    // Close the file.
    fclose(bridge_file);

    // Unfortunately, fgets() tags on a linefeed. Trim that off.
    info_buffer[strlen(info_buffer) - 1] = 0;

    // We need to pad out the values to be consistent with how other PCI
    // paths are displayed.
    temp_str.assign(info_buffer);
    index = temp_str.find('/');
    temp_str.insert(index + 1, "0000:");

    // Build up the path.
    devpath = fmt::sprintf("/devices/pci0000:00/0000:%s", temp_str.c_str());
}

void storage::get_internal_nvme(Json::Value& value)
{
    int i;
    Json::Value dev_tree;
    const LicManager *p_lic;

    // Make sure the return value is cleared.
    value.clear();

    // Get our license pointer.
    p_lic = CameraServer::get_license();
    if ((p_lic == NULL) || (p_lic->ssdMax() == 0))
        return;

    // Get the full NVME device tree information.
    nvme_udev(dev_tree, true);

    // Make sure we got something.
    if (dev_tree["nvme"].empty())
    {
        fi_log(LOG_DEBUG, "%s: No NVME devices found!", __PRETTY_FUNCTION__);
        return;
    }

    // Loop through all the devices, looking for the one in the camera head.
    for (i = 0; ((Json::ArrayIndex)i) < dev_tree["nvme"].size(); ++i)
    {
        Json::Value my_dev = dev_tree["nvme"][i];
        if (my_dev["internal"].asBool() == true)
        {
            value = my_dev;
            break;
        }
    }
}

/*!
 * @brief Return the licensed size of the SSD in sectors.
 *
 * This is for use on A3+ cameras, presumably in LTR mode.
 *
 * @return The number of sectors on the drive or licensed, whichever is smaller.
 */
uint64_t storage::get_ssd_licensed_sector_count()
{
    uint64_t max_licensed = 0;
    uint64_t ssd_sectors = 0;
    int ssd_block = FAS_FILESYSTEM::SECTOR_LEN;
    Json::Value nvme_info;
    const LicManager *p_lic;

    LOG("%s", __PRETTY_FUNCTION__);

    // Get our license pointer.
    p_lic = CameraServer::get_license();
    if ((p_lic == NULL) || ((max_licensed = (uint64_t)p_lic->ssdMax()) == 0))
    {
        fi_log(LOG_ERR, "%s: No SSD license %p,%" PRIu64, __PRETTY_FUNCTION__, (void *)p_lic, (uint64_t)max_licensed);
        return 0;
    }

    // Get the information for the NVME in the head.
    storage::get_internal_nvme(nvme_info);
    if (!nvme_info.empty())
    {
        // Convert to integer.
        ssd_sectors = nvme_info["sectors"].asUInt64();
        ssd_block = nvme_info["block_size"].asInt();
        if (ssd_block == 0)
        {
            fi_log(LOG_DEBUG, "%s: Resetting ssd_block", __PRETTY_FUNCTION__);
            ssd_block = FAS_FILESYSTEM::SECTOR_LEN;
        }
    }

    // Calculate what the maximum licensed sector size is.
    max_licensed *= (uint64_t)((1024 * 1024 * 1024) / ssd_block);
    // fi_log(LOG_DEBUG, "%s: ssd_sectors=%" PRIu64 ", ssd_block=%d, max_licensed=%" PRIu64, __PRETTY_FUNCTION__, ssd_sectors, ssd_block, max_licensed);

    // If the SSD is larger than what's allowed, only allow the licensed value.
    if (ssd_sectors > max_licensed)
    {
        fi_log(LOG_ERR, "%s: License restricts SSD to %" PRIu64 " GB, smaller than actual %" PRIu64 " GB",
               __PRETTY_FUNCTION__, max_licensed / (2 * 1024 * 1024), ssd_sectors / (2 * 1024 * 1024));
        ssd_sectors = max_licensed;
    }

    // Return our findings.
    return ssd_sectors;
}

/*!
 * @brief Scan jhead output to generate XML nodes for a file.
 *
 * @param f_node    The parent node.
 * @param f_path    Path to the jhead output file.
 *
 * @return true on success, false otherwise.
 */
bool storage::add_jhead_info(xmlNodePtr f_node, const char *f_path)
{
    FILE *md_file;

    // Open, but understand it may not exist.
    md_file = fopen(f_path, "r");
    if (!md_file)
        return false;

    // Read the file.
    while (fgets(info_buffer, INFO_BUFFER_LEN, md_file))
    {
        char    *tag_ptr;
        char    *value_ptr;
        char    *end_ptr;

        // The first entry is the tag.
        tag_ptr = info_buffer;

        // Look for the seperator.
        value_ptr = strchr(tag_ptr, ':');

        // Did we find something?
        if (value_ptr == NULL)
            continue;

        // Null terminate at this point and move to the value.
        *value_ptr++ = '\0';

        // Skip over the space after the seperator.
        value_ptr++;

        // Walk the value, to find the last non-space character.
        for (end_ptr = &value_ptr[ strlen(value_ptr) - 1 ]; end_ptr != value_ptr; --end_ptr)
        {
            // Is this a non-space character?
            if (!isspace(*end_ptr))
            {
                // Move back to the last position.
                ++end_ptr;
                break;
            }
        }

        // If we had trailing space, null-terminate.
        if (end_ptr != value_ptr)
            *end_ptr = '\0';

        // Let's look at tags we're interested in.
        if (strstr(tag_ptr, "File date") != NULL)
        {
            // Add the timestamp information.
            xmlNewChild(f_node, NULL, BAD_CAST "time_stamp", BAD_CAST value_ptr);
        }
        else if (strstr(tag_ptr, "Resolution") != NULL)
        {
            char *height_ptr;

            // The value pointer is already at the width, so that's done.
            // We need to move forward to terminate the width string.
            for (height_ptr = value_ptr; !isspace(*height_ptr); ++height_ptr)
                ;

            // Null terminate here.
            *height_ptr = '\0';

            // Now move past the " x ".
            height_ptr += 3;

            // Add the nodes.
            xmlNewChild(f_node, NULL, BAD_CAST "width", BAD_CAST value_ptr);
            xmlNewChild(f_node, NULL, BAD_CAST "height", BAD_CAST height_ptr);
        }
    }

    // Close the file.
    fclose(md_file);

    // All done.
    return true;
}

bool storage::add_capstack_info(xmlNodePtr f_node, std::filesystem::path& file_path)
{
    FILE    *md_file;
    bool    rc = false;
    std::ifstream meta_ifs;
    Json::Value json_value;
    Json::String json_errs;
    Json::CharReaderBuilder rd_builder;

    // Build up what SHOULD be the metamap file path.
    file_path.append("metamap.txt");
//    fi_log(LOG_DEBUG, "%s: Try metamap file '%s'", __PRETTY_FUNCTION__, file_path.c_str());

    // Set exceptions to be thrown on failure.
    meta_ifs.exceptions(std::ifstream::failbit | std::ifstream::badbit);

    try
    {
        // Get our input stream.
        meta_ifs.open(file_path, std::ifstream::in);
    }
    catch (std::ifstream::failure& e)
    {
        fi_log(LOG_DEBUG, "%s: Unable to open metamap file '%s' [%s]", __PRETTY_FUNCTION__, file_path.c_str(), e.what());
        goto bail;
    }

    // Parse the metamap file.
    rd_builder["collectComments"] = false;
    if (!Json::parseFromStream(rd_builder, meta_ifs, &json_value, &json_errs))
    {
        fi_log(LOG_DEBUG, "%s: parseFromStream() failed [%s]", __PRETTY_FUNCTION__, json_errs.c_str());
        goto bail;
    }

    // Open the last metadata file.
    file_path.replace_filename(fmt::sprintf("%07d.txt", json_value["end_cap"].asInt()));
    md_file = fopen(file_path.c_str(), "r");
    if (!md_file)
        return false;

    // Parse the image metadata from the file.
    rc = parse_image_metadata(md_file, f_node, true);

    // Close the file.
    fclose(md_file);

    // If that went OK, then add some of the CAP stack specific details.
    if (rc)
    {
        int total_frames = json_value["tframes"].asInt();
        int metafile_flag = GIGE_FA_META_TXT;     // We know we have the "TXT" metadata file.

        // Add start, end, and total frames.
        xmlNewChild(f_node, NULL, BAD_CAST "start_frame", BAD_CAST "0");
        xmlNewChild(f_node, NULL, BAD_CAST "end_frame", BAD_CAST fmt::format("{}", total_frames - 1).c_str());
        xmlNewChild(f_node, NULL, BAD_CAST "frame_count", BAD_CAST fmt::format("{}", total_frames).c_str());

        // Add the starting and ending CAP indecies.
        xmlNewChild(f_node, NULL, BAD_CAST "start_cap", BAD_CAST fmt::format("{}", json_value["start_cap"].asInt()).c_str());
        xmlNewChild(f_node, NULL, BAD_CAST "end_cap", BAD_CAST fmt::format("{}", json_value["end_cap"].asInt()).c_str());

        // If we're OK, then check to see if the other metadata file(s) exist.
        //
        // We could do the checking logic in a loop (@sa file_action_thread::do_copy()),
        // but then we'd have to associate the vector entry with the flag...doable, but
        // too much work. Just brute force it.

        // Modify our metadata file path to point to a CFG file.
        file_path.replace_extension("cfg");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_CFG;
        }

        // Modify our metadata file path to point to an XML file.
        file_path.replace_extension("xml");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_XML;
        }

        // Modify our metadata file path to point to an PM file.
        file_path.replace_extension("pm");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_PM;
        }

        // Add it to our XML listing.
        snprintf(info_buffer, INFO_BUFFER_LEN, "%d", metafile_flag);
        xmlNewChild(f_node, NULL, BAD_CAST "metafile_count", BAD_CAST info_buffer);
    }

bail:
    // Close our stream.
    if (meta_ifs.is_open())
        meta_ifs.close();

    // Return our result.
    return rc;
}

/*!
 * @brief Parse the Fastec metadata file to provide additional
 *    information nodes.
 *
 * @param f_node    The parent node.
 * @param file_path Path to the metadata file.
 *
 * @return true on success, false otherwise.
 */
bool storage::add_metadata_info(xmlNodePtr f_node, std::string file_path, const std::string *file_ext)
{
    FILE    *md_file;
    bool    rc;
    std::size_t suffix_index;

//    fi_log(LOG_DEBUG, "%s: file_path='%s'", __PRETTY_FUNCTION__, file_path.c_str());

    // Get the length of the file.
    suffix_index = file_path.find_last_of('.');
    if (suffix_index == std::string::npos)
    {
        // We have a malformed file_path value.
        fi_log(LOG_DEBUG, "%s: Unexpectedly cannot find the '.' in [%s]!", __PRETTY_FUNCTION__, file_path.c_str());
        rc = false;
        goto bail;
    }

    // Move past the '.'.
    suffix_index++;

    // Do we have an extension?
    if (file_ext != nullptr)
    {
        std::string test_file(file_path);

        // Modify our metadata file path to point to an "incomplete" file.
        test_file.replace(suffix_index, 3, *file_ext + FAS_VIDEO::FiVideo::GetIncompleteVideoFileString());

        // Can we open it?
        md_file = fopen(test_file.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Add a flag to our XML listing.
            info_buffer[0] = '1';
            info_buffer[1] = '\0';
            xmlNewChild(f_node, NULL, BAD_CAST "incomplete", BAD_CAST info_buffer);
        }
    }

    // Open, but understand it may not exist.
    md_file = fopen(file_path.c_str(), "r");
    if (md_file == nullptr)
    {
        std::filesystem::path test_file(file_path);

        // Could we be looking at a CAP Stack?
        return add_capstack_info(f_node, test_file.remove_filename());
    }

    // Parse the image metadata from the file.
    rc = parse_image_metadata(md_file, f_node);

    // Close the file.
    fclose(md_file);

    // If we're OK, then check to see if the other metadata file(s) exist.
    //
    // We could do the checking logic in a loop (@sa file_action_thread::do_copy()),
    // but then we'd have to associate the vector entry with the flag...doable, but
    // too much work. Just brute force it.
    if (rc)
    {
        int metafile_flag = GIGE_FA_META_TXT;     // We know we have the "TXT" metadata file.

        // Modify our metadata file path to point to a CFG file.
        file_path.replace(suffix_index, 3, "cfg");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_CFG;
        }

        // Modify our metadata file path to point to an XML file.
        file_path.replace(suffix_index, 3, "xml");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_XML;
        }

        // Modify our metadata file path to point to an PM file.
        file_path.replace(suffix_index, 3, "pm");

        // Can we open it?
        md_file = fopen(file_path.c_str(), "r");
        if (md_file)
        {
            // Close the file.
            fclose(md_file);

            // Bump our count.
            metafile_flag |= GIGE_FA_META_PM;
        }

        // Add it to our XML listing.
        snprintf(info_buffer, INFO_BUFFER_LEN, "%d", metafile_flag);
        xmlNewChild(f_node, NULL, BAD_CAST "metafile_count", BAD_CAST info_buffer);
    }

bail:
    return rc;
}

/*!
 * @brief Add folder specific information nodes.
 *
 * @param [in] f_node Our parent node.
 * @param [in] file_path Path to the folder we're interested in.
 */
void storage::get_folder_stats(xmlNodePtr f_node, std::string file_path)
{
    bool got_metadata;
    int path_len;

    // Point to the end of our path.
    path_len = file_path.length();

    // Add our metadata file name.
    file_path.append(METADATA_FILENAME);

    // Add any metadata information we might find in the folder.
    got_metadata = add_metadata_info(f_node, file_path);

    // Restore the original path.
    file_path.resize(path_len);

    // Determine the size of the folder.
    if (got_metadata)
    {
        // Do we know how many frames are in the stack?
        if (last_frame_count > 0)
            sprintf(info_buffer, "%" PRIu64, calc_stack_size(file_path.c_str(), last_frame_count));
        else
            sprintf(info_buffer, "%" PRIu64, calc_dir_size(file_path.c_str()));

        xmlNewChild(f_node, NULL, BAD_CAST "size", BAD_CAST info_buffer);
    }
}

bool storage::parse_image_metadata(FILE *meta_file, xmlNodePtr f_node, bool capstack)
{
    // Assume success.
    bool rc = true;

    // Reset our frame count.
    last_frame_count = 0;

    // Sanity check.
    if (meta_file == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: The meta_file pointer is null!?", __PRETTY_FUNCTION__);
        return false;
    }

    // Read the file.
    while (fgets(info_buffer, INFO_BUFFER_LEN, meta_file))
    {
        char    *tag_ptr;
        char    *value_ptr;
        char    *end_ptr;

        // Check to see if the user wants to cancel.
        if (storage::factory()->fa_task->is_stop_requested())
        {
            TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
            rc = false;
            break;
        }

        // Walk the line, to find the first non-space character.
        for (tag_ptr = info_buffer; *tag_ptr != '\0'; ++tag_ptr)
        {
            // Is this a non-space character?
            if (!isspace(*tag_ptr))
                break;
        }

        // Did we find something?
        if (*tag_ptr == '\0')
            continue;

        // Now find the seperation token '='.
        value_ptr = strchr(tag_ptr, '=');

        // Did we find something?
        if (value_ptr == NULL)
            continue;

        // Null terminate at this point and move to the value.
        *value_ptr++ = '\0';

        // Walk the value, to find the first space character.
        for (end_ptr = &value_ptr[ strlen(value_ptr) - 1 ]; end_ptr != value_ptr; --end_ptr)
        {
            // Is this a non-space character?
            if (!isspace(*end_ptr))
            {
                // Move back to the last position.
                ++end_ptr;
                break;
            }
        }

        // If we had trailing space, null-terminate.
        if (end_ptr != value_ptr)
            *end_ptr = '\0';
        else
        {
            // Move back up one character and then null terminate.
            ++end_ptr;
            *end_ptr = '\0';
        }

        // Let's look at tags we're interested in.
        if ((strcmp(tag_ptr, "width") == 0) ||
            (strcmp(tag_ptr, "height") == 0) ||
            (strcmp(tag_ptr, "fps") == 0) ||
            (strcmp(tag_ptr, "bit_mode") == 0) ||
            (strcmp(tag_ptr, "time_stamp") == 0) ||
            (strcmp(tag_ptr, "raw") == 0) ||
            (strcmp(tag_ptr, "comment") == 0))
        {
            // Add these to our XML listing.
            xmlNewChild(f_node, NULL, BAD_CAST tag_ptr, BAD_CAST value_ptr);
        }

        // The codec tag is different for CAP stacks.
        else if (strcmp(tag_ptr, "codec") == 0)
        {
            if (capstack)
                xmlNewChild(f_node, NULL, BAD_CAST tag_ptr, BAD_CAST CAPSTACK_LABEL);
            else
                xmlNewChild(f_node, NULL, BAD_CAST tag_ptr, BAD_CAST value_ptr);
        }

        // These tags don't apply with CAP stacks.
        else if (!capstack)
        {
            if ((strcmp(tag_ptr, "start_frame") == 0) || (strcmp(tag_ptr, "end_frame") == 0))
            {
                // Add these to our XML listing.
                xmlNewChild(f_node, NULL, BAD_CAST tag_ptr, BAD_CAST value_ptr);
            }
            else if (strcmp(tag_ptr, "frame_count") == 0)
            {
                // Save the frame count.
                last_frame_count = strtoul(value_ptr, NULL, 0);

                // Add it to our XML listing.
                xmlNewChild(f_node, NULL, BAD_CAST tag_ptr, BAD_CAST value_ptr);
            }
        }
    }

    // All done.
    return rc;

}

/*!
 * @brief Add file specific information ndoes.
 *
 * @param f_node    Our parent node.
 * @param f_path    Path to the file we're intereseted in.
 * @param f_ext     Pointer to the extension of the file.
 */
void storage::get_file_stats(xmlNodePtr f_node, std::string file_path, std::string file_ext)
{
    std::string attr_str;
    struct stat64 sb;

    // Get the statistics on the entry.
    stat64(file_path.c_str(), &sb);

    // Add the additional attributes.
    attr_str = std::to_string(((uint64_t)sb.st_size));
    xmlNewChild(f_node, NULL, BAD_CAST "size", BAD_CAST attr_str.c_str());

    // Only add metadata for our video files.
    if ((file_ext.compare("avi") == 0) || (file_ext.compare("mp4") == 0) || (file_ext.compare("cap") == 0))
    {
        // Replace it with the metadata extension.
        file_path.replace(file_path.length() - 3, 3, "txt");

        // Add any metadata information we might find in the folder.
        add_metadata_info(f_node, file_path, &file_ext);
    }
    else if (file_ext.compare("jpg") == 0)
    {
        std::string temp_str;

        // Build up a call to get some information from the JPEG image.
        temp_str = fmt::sprintf("/usr/bin/jhead %s > /tmp/t.txt", file_path.c_str());
        system(temp_str.c_str());

        // Now add any metadata we might have found.
        add_jhead_info(f_node, "/tmp/t.txt");
    }
}

/*!
 * @brief Callback function used by nftw() to copy a tree from one device to another.
 */
static int copy_cb(const char *fpath, const struct stat *sb, int /*typeflag*/, struct FTW * /*ftwbuf*/)
{
    std::string working_target;
    uint32_t    target_off;
    uint32_t    file_off;
    file_action_thread *fa_task;

    TP(TRACE_SEL_STORAGE, "%s: fpath=[%s]", __PRETTY_FUNCTION__, fpath);

    // Get the file action task pointer.
    fa_task = storage::factory()->get_fa_task();
    if (fa_task == nullptr)
    {
        LOG("%s: No task thread!", __PRETTY_FUNCTION__);
        return -1;
    }

    // See if we're being requested to stop.
    if (fa_task->is_stop_requested())
    {
        TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
        errno = 0;
        return -1;
    }

    // Get the target folder.
    working_target = fa_task->get_target_path();
    file_off = fa_task->get_file_offset();
    target_off = fa_task->get_target_offset();

    // Build up our destination path.
    working_target.replace(target_off, std::string::npos, &fpath[file_off]);

    // If the source is a folder, make sure the we make the folder on the target.
    if (S_ISDIR(sb->st_mode))
    {
        if ((storage::mkpath(working_target.c_str(), 0777) == -1) && (errno != ENOTDIR))
        {
            fi_log(LOG_ERR, "%s: Unable to create target '%s'. %d(%s)", __PRETTY_FUNCTION__, working_target.c_str(), errno, strerror(errno));
            return -1;
        }
    }
    else
    {
        bool cap_file;
        std::filesystem::path check_file(fpath);

        // If we are not copying a CAP file for a CAP stack transfer, then we will want to reset the
        // total transfer amount.
        cap_file = (check_file.extension().compare(".cap") == 0);
        if (!storage::factory()->get_capstk_transfer())
        {
            // If we just found a CAP file, then it looks like we're now doing a CAP stack.
            if (cap_file)
            {
                storage::factory()->set_capstk_transfer(true);
            }
            else
            {
                storage::factory()->reset_total_transfer();
            }
        }

        // Copy over the file.
        if (!fa_task->thread_copy_file(working_target.c_str(), fpath))
        {
            fi_log(LOG_ERR, "%s: Unable to copy file '%s'. %d(%s)", __PRETTY_FUNCTION__, fpath, errno, strerror(errno));
            return -1;
        }

        // Update our status. The busy bit is set, we're just incrementing
        // the count at the bottom.
        fa_status++;
    }

    return 0;
}

/*!
 * @brief NFTW callback function, invoked for each element of the tree.
 */
static int unlink_cb(const char *fpath, const struct stat * /*sb*/, int /*typeflag*/, struct FTW * /*ftwbuf*/)
{
    int rv = 0;
    file_action_thread *fa_task;

    TP(TRACE_SEL_STORAGE, "%s: deleting '%s'", __PRETTY_FUNCTION__, fpath);

    // Get the file action task pointer.
    fa_task = storage::factory()->get_fa_task();
    if (fa_task == nullptr)
    {
        LOG("%s: No task thread!", __PRETTY_FUNCTION__);
        return -1;
    }

    // See if we're being requested to stop.
    if (fa_task->is_stop_requested())
    {
        TP(TRACE_SEL_STORAGE, "%s: Stop requested.", __PRETTY_FUNCTION__);
        errno = 0;
        return -1;
    }

    // Remove our current location.
    rv = remove(fpath);
    if (rv)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
    }

    return rv;
}

//--------------------------------------------------------------------
// Class implementation: storage
//--------------------------------------------------------------------

storage::storage()
{
    // Turn on tracing.
//    traceControl |= TRACE_SEL_STORAGE;

    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);

    m_xmlOutputBuffer = nullptr;
    m_xmlOutputBufferLen = 0;

    // Create our file action thread.
    fa_task = std::unique_ptr<file_action_thread>(new file_action_thread());
}

/*!
 * @brief Storage class destructor.
 */
storage::~storage()
{
    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);

    if (m_xmlOutputBuffer != NULL)
    {
        xmlFree(m_xmlOutputBuffer);
        m_xmlOutputBuffer = NULL;
    }
}

/*!
 * @brief Create our singleton.
 */
storage *storage::factory()
{
    std::lock_guard<std::mutex> lock(storage_mutex);

    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);

    // Check to see if we already have our singleton.
    if (local_storage.get() == nullptr)
    {
        // Create our singleton object.
        local_storage = std::unique_ptr<storage>(new storage());

    }

    // Return the class pointer.
    return local_storage.get();
}

void storage::startup_mounts()
{
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // If the SSD is already mounted, then our work is done.
    if (storage::is_mounted(GIGE_MEDIA_SSD))
        return;

    // If we're in LR mode, ignore the mount.
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        return;

    // Go ahead and try to mount the SSD. It's OK if it doesn't work.
    storage::nvme_mount();
}

void storage::stop()
{
    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (local_storage.get() != nullptr)
    {
        // Release the memory.
        local_storage.reset();
    }
}


/*!
 * @brief Get the label used for a device.
 * @param [in]  dev_name    Name of the device we're interested in.
 * @param [out] value       JSON value of the label.
 * @returns True if a label is found, false otherwise.
 */
bool storage::get_label(const char *dev_name, Json::Value& value)
{
    bool rc = false;
    const char *label = nullptr;
    blkid_probe pr;

    // Open the device.
    pr = blkid_new_probe_from_filename(dev_name);
    if (pr == NULL)
    {
        fi_log(LOG_DEBUG, "%s: Couldn't probe: '%s'\n", __PRETTY_FUNCTION__, dev_name);
        return false;
    }

    // Get the partition label, if there is one.
    blkid_do_probe(pr);
    blkid_probe_lookup_value(pr, "LABEL", &label, NULL);

    // Is there a label?
    if (label != NULL)
    {
        value["label"] = label;
        rc = true;
    }
    else
    {
        PartMgr *pm_ptr = CameraServer::get_partmgr();

        // Make sure we have our pointer.
        if (pm_ptr != nullptr)
        {
            // See if we can get the name for the Fastec filesystem. If we get a label,
            // we can use it.
            std::string lr_mode_label = pm_ptr->ltrGetVolumeName();
            if (!lr_mode_label.empty())
            {
                value["label"] = lr_mode_label;
                rc = true;
            }
        }
    }

    blkid_free_probe(pr);
    return rc;
}

/*!
 * @brief Find where a device's mount point, if any.
 * @param [in] device    The device we're interested in.
 * @param [out] mountdir The mountpoint for the device, if found.
 */
void storage::get_mount_dir(const std::string device, std::string& mountdir)
{
    FILE *fp = nullptr;
    std::lock_guard<std::mutex> lock(storage_mutex);

    // Clear the return string.
    mountdir.clear();

    try
    {
        struct mntent fs;

        // Allocate our working buffer.
        std::unique_ptr<char[]> temp_buf = std::make_unique<char[]>(MAX_TEMP_BUFFER_LEN + 1);

        // Open our mount information.
        fp = setmntent("/proc/mounts", "r");   /* read only */
        if (fp == nullptr)
        {
            fi_log(LOG_DEBUG, "/proc/mounts could not open: %s\n", strerror(errno));
            return;
        }

        // Hunt for our device.
        while (getmntent_r(fp, &fs, temp_buf.get(), MAX_TEMP_BUFFER_LEN) != nullptr)
        {
            // Is this the one?
            if ((device.compare(fs.mnt_fsname) == 0) && (fs.mnt_dir != nullptr))
            {
                // Save the mount path.
                mountdir = fs.mnt_dir;
                if (!mountdir.empty())
                    break;
            }
        }
    }
    catch (...)
    {
        fi_log(LOG_DEBUG, "%s: Caught an exception!", __PRETTY_FUNCTION__);
    }

    // Close access to the mount information.
    endmntent(fp);
}

/*!
 * @brief Use the UDEV library to determine if we have any NVME SSDs available.
 */
void storage::nvme_udev(Json::Value& root, bool full_tree)
{
    struct udev *udev;
    struct udev_enumerate *enumerate;
    struct udev_list_entry *devices, *dev_list_entry;
    Json::Value dev_array;
    std::string int_ssd_path;
    std::string dev_path;
    const LicManager *p_lic;

    // Get our license pointer.
    p_lic = CameraServer::get_license();
    if (p_lic == nullptr)
    {
        // No license is bad. Bail.
        fi_log(LOG_DEBUG, "%s: No license manager?!", __PRETTY_FUNCTION__);
        goto nvme_bail;
    }

    // Get the device path to our internall SSD, if there is one.
    int_ssd_path.clear();
    get_internal_ssd_devpath(int_ssd_path);
//    fi_log(LOG_DEBUG, "%s: int_ssd_path [%s]", __PRETTY_FUNCTION__, int_ssd_path.c_str());

    // Create the udev object.
    udev = udev_new();
    if (!udev)
    {
        fi_log(LOG_DEBUG, "%s: Unable to allocate a UDEV!", __PRETTY_FUNCTION__);
        goto nvme_bail;
    }

    // Create a list of the devices "scsi" subsystem.
    enumerate = udev_enumerate_new(udev);
    if (!enumerate)
    {
        fi_log(LOG_DEBUG, "%s: Unable to enumerate a UDEV context!", __PRETTY_FUNCTION__);
        goto nvme_bail;
    }

    udev_enumerate_add_match_subsystem(enumerate, "block");
    udev_enumerate_scan_devices(enumerate);

    // Fillup the device list.
    devices = udev_enumerate_get_list_entry(enumerate);

    // Loop through each device found.
    udev_list_entry_foreach(dev_list_entry, devices)
    {
        const char *path;
        struct udev_device *scsi_dev;
        struct udev_device *nvme;
        Json::Value json_info;

        // Get the filename of the /sys entry for the device
        // and create a udev_device object (scsi_dev) representing it.
        path = udev_list_entry_get_name(dev_list_entry);
        scsi_dev = udev_device_new_from_syspath(udev, path);

        // Note: The returned device is not referenced, so don't
        //   clean it up.
        nvme = udev_device_get_parent_with_subsystem_devtype(
                   scsi_dev,
                   "nvme",
                   NULL);

        if (nvme)
        {
            // Is this an NVME partition?
            if (strncmp(udev_device_get_devtype(scsi_dev), "partition", 9) == 0)
            {
                const char *tmp;
                std::string dev_name;
                std::string mountdir;
                uint64_t disk_size = 0;
                uint64_t dev_size = 0;
                uint64_t avail_size = 0;
                uint64_t inode_count = 0;
                uint32_t block_size = DEFAULT_BLOCK_SIZE;

                // Build up our JSON structure for this device.
                dev_name.assign(udev_device_get_devnode(scsi_dev));
                json_info["name"] = dev_name;
                json_info["type"] = GIGE_MEDIA_SSD;

                // Do we have a label?
                get_label(udev_device_get_devnode(scsi_dev), json_info);

                // How big is the device?
                tmp = udev_device_get_sysattr_value(scsi_dev, "size");
                if (tmp)
                    dev_size = strtoull(tmp, NULL, 10);

                tmp = udev_device_get_sysattr_value(scsi_dev, "queue/logical_block_size");
                if (tmp)
                    block_size = atoi(tmp);

                // Is this the internal SSD?
                dev_path.assign(udev_device_get_devpath(scsi_dev));
//                fi_log(LOG_DEBUG, "%s: dev_path [%s]", __PRETTY_FUNCTION__, dev_path.c_str());
                if (int_ssd_path.empty())
                    json_info["internal"] = false;
                else
                    json_info["internal"] = (dev_path.find(int_ssd_path) != std::string::npos);

                // Are we grabbing all the information we can?
                if (full_tree)
                {
                    json_info["devpath"] = dev_path;
                    json_info["sectors"] = dev_size;
                    json_info["block_size"] = block_size;
                }

                // Find out if and where the device is mounted.
                mountdir.clear();
                get_mount_dir(dev_name, mountdir);
                json_info["mountdir"] = mountdir;

                // Get the device capacity.
                if (mountdir.empty())
                {
                    disk_size = (dev_size * block_size);
                    avail_size = (dev_size * block_size);
                }
                else if (get_capacity(mountdir.c_str(), &disk_size, &avail_size, &inode_count) != 0)
                {
                    fi_log(LOG_DEBUG, "%s[%d]: get_capacity() failed, dev_name=%s, mountdir=%s.", __PRETTY_FUNCTION__, __LINE__, dev_name.c_str(), mountdir.c_str());
                    disk_size = (dev_size * block_size);
                    avail_size = (dev_size * block_size);
                }

                // Set the information into the Json record.
                json_info["devsize"] = disk_size;
                json_info["avail"] = avail_size;

                // Is this an external SSD or are we allowed an internal SSD?
                if ((json_info["internal"].asBool() != true) || (p_lic->ssdMax() > 0))
                {
                    // Add the element to our array.
                    dev_array.append(json_info);
                }
            }
            else
            {
                // It's not a partition. Check to see if it has a partition.
                blkid_probe pr = blkid_new_probe_from_filename(udev_device_get_devnode(scsi_dev));
                if (!pr)
                {
                    fi_log(LOG_DEBUG, "%s: Unable to create a probe!", __PRETTY_FUNCTION__);
                    goto nvme_bail;
                }

                // Get number of partitions
                blkid_partlist ls;
                int nparts = 0;

                ls = blkid_probe_get_partitions(pr);
                if (ls != NULL)
                    nparts = blkid_partlist_numof_partitions(ls);
                blkid_free_probe(pr);

                if (nparts <= 0)
                {
                    const char *tmp;
                    uint64_t disk_size = 0;
                    uint32_t block_size = DEFAULT_BLOCK_SIZE;

                    // Build up our JSON structure for this device.
                    json_info["name"] = udev_device_get_devnode(scsi_dev);
                    json_info["type"] = GIGE_MEDIA_SSD;

                    // Do we have a label?
                    get_label(udev_device_get_devnode(scsi_dev), json_info);

                    // How big is the device?
                    tmp = udev_device_get_sysattr_value(scsi_dev, "size");
                    if (tmp)
                        disk_size = strtoull(tmp, NULL, 10);

                    tmp = udev_device_get_sysattr_value(scsi_dev, "queue/logical_block_size");
                    if (tmp)
                        block_size = atoi(tmp);

                    // Is this the internal SSD?
                    dev_path.assign(udev_device_get_devpath(scsi_dev));
//                    fi_log(LOG_DEBUG, "%s: dev_path [%s]", __PRETTY_FUNCTION__, dev_path.c_str());
                    if (int_ssd_path.empty())
                        json_info["internal"] = false;
                    else
                        json_info["internal"] = (dev_path.find(int_ssd_path) != std::string::npos);

                    // Are we grabbing all the information we can?
                    if (full_tree)
                    {
                        json_info["devpath"] = dev_path;
                        json_info["sectors"] = disk_size;
                        json_info["block_size"] = block_size;
                    }

                    // Get the device capacity.
                    json_info["devsize"] = (disk_size * block_size);
                    json_info["avail"] = (disk_size * block_size);

                    // Is this an external SSD or are we allowed an internal SSD?
                    if ((json_info["internal"].asBool() != true) || (p_lic->ssdMax() > 0))
                    {
                        // Add the element to our array.
                        dev_array.append(json_info);
                    }
                }
            }
        }

        // Release the device object.
        udev_device_unref(scsi_dev);
    }

    /* Free the enumerator object */
    udev_enumerate_unref(enumerate);

    // Release our udev object.
    udev_unref(udev);

nvme_bail:
    // Wrap up the JSON information.
    root["nvme"] = dev_array;
}

/*!
 * @brief Use the UDEV library to determine if we have an SDCARD.
 */
void storage::sdcard_udev(Json::Value& root)
{
    struct udev *udev;
    struct udev_enumerate *enumerate;
    struct udev_list_entry *devices, *dev_list_entry;
    Json::Value dev_array;

    // Create the udev object.
    udev = udev_new();
    if (!udev)
    {
        fi_log(LOG_DEBUG, "%s: Unable to allocate a UDEV!", __PRETTY_FUNCTION__);
        goto sdcard_bail;
    }

    // Create a list of the devices "scsi" subsystem.
    enumerate = udev_enumerate_new(udev);
    if (!enumerate)
    {
        fi_log(LOG_DEBUG, "%s: Unable to enumerate a UDEV context!", __PRETTY_FUNCTION__);
        goto sdcard_bail;
    }

    udev_enumerate_add_match_subsystem(enumerate, "block");
    udev_enumerate_scan_devices(enumerate);

    // Fillup the device list.
    devices = udev_enumerate_get_list_entry(enumerate);

    // Loop through each device found.
    udev_list_entry_foreach(dev_list_entry, devices)
    {
        const char *path;
        struct udev_device *scsi_dev;
        struct udev_device *sdcard;
        Json::Value json_info;

        // Get the filename of the /sys entry for the device
        // and create a udev_device object (scsi_dev) representing it.
        path = udev_list_entry_get_name(dev_list_entry);
        scsi_dev = udev_device_new_from_syspath(udev, path);

        // The device pointed to by scsi_dev contains information about
        // the MMC device. In order to get information about the
        // MMC device, get the parent device with the
        // subsystem/devtype pair of "mmc_host"/NULL. This will
        // be several levels up the tree, but the function will find
        // it.
        //
        // Note: The returned device is not referenced, so don't
        //   clean it up.
        sdcard = udev_device_get_parent_with_subsystem_devtype(scsi_dev, "mmc_host", NULL);
        if (sdcard && (strncmp(udev_device_get_devtype(scsi_dev), "partition", 9) == 0))
        {
            const char *tmp;
            std::string mountdir;
            uint64_t disk_size = 0;
            uint64_t avail_size = 0;
            uint64_t inode_count = 0;
            uint32_t block_size = DEFAULT_BLOCK_SIZE;

            // Build up our JSON structure for this device.
            json_info["name"] = udev_device_get_devnode(scsi_dev);
            json_info["type"] = GIGE_MEDIA_SDCARD;

            // Do we have a label?
            get_label(udev_device_get_devnode(scsi_dev), json_info);

            // How big is the device?
            tmp = udev_device_get_sysattr_value(scsi_dev, "size");
            if (tmp)
                disk_size = strtoull(tmp, NULL, 10);

            // Find out if and where the device is mounted.
            mountdir.clear();
            get_mount_dir(json_info["name"].asString(), mountdir);
            json_info["mountdir"] = mountdir;

            tmp = udev_device_get_sysattr_value(scsi_dev, "queue/logical_block_size");
            if (tmp)
                block_size = atoi(tmp);

            // Get the capacity information.
            if (mountdir.empty())
            {
                disk_size = (disk_size * block_size);
                avail_size = (disk_size * block_size);
            }
            else if (get_capacity(mountdir.c_str(), &disk_size, &avail_size, &inode_count) != 0)
            {
                fi_log(LOG_DEBUG, "%s: get_capacity() failed.", __PRETTY_FUNCTION__);
                disk_size = (disk_size * block_size);
                avail_size = (disk_size * block_size);
            }

            json_info["devsize"] = disk_size;
            json_info["avail"] = avail_size;

            // Add the element to our array.
            dev_array.append(json_info);
        }

        // Release the device object.
        udev_device_unref(scsi_dev);
    }

    /* Free the enumerator object */
    udev_enumerate_unref(enumerate);

    // Release our udev object.
    udev_unref(udev);

sdcard_bail:
    // Wrap up the JSON information.
    root["sdcard"] = dev_array;
}

/*!
 * @brief Use the UDEV library to determine if we have any USB drives connected.
 */
void storage::usb_udev(Json::Value& root)
{
    struct udev *udev;
    struct udev_enumerate *enumerate;
    struct udev_list_entry *devices, *dev_list_entry;
    Json::Value dev_array;

    // Create the udev object.
    udev = udev_new();
    if (!udev)
    {
        fi_log(LOG_DEBUG, "%s: Unable to allocate a UDEV!", __PRETTY_FUNCTION__);
        goto usb_bail;
    }

    // Create a list of the devices "scsi" subsystem.
    enumerate = udev_enumerate_new(udev);
    if (!enumerate)
    {
        fi_log(LOG_DEBUG, "%s: Unable to enumerate a UDEV context!", __PRETTY_FUNCTION__);
        goto usb_bail;
    }

    udev_enumerate_add_match_subsystem(enumerate, "block");
    udev_enumerate_scan_devices(enumerate);

    // Fillup the device list.
    devices = udev_enumerate_get_list_entry(enumerate);

    // Loop through each device found.
    udev_list_entry_foreach(dev_list_entry, devices)
    {
        const char *path;
        struct udev_device *scsi_dev;
        struct udev_device *usb;
        Json::Value json_info;

        // Get the filename of the /sys entry for the device
        // and create a udev_device object (scsi_dev) representing it.
        path = udev_list_entry_get_name(dev_list_entry);
        scsi_dev = udev_device_new_from_syspath(udev, path);

        // The device pointed to by scsi_dev contains information about
        // the usb device. In order to get information about the
        // USB device, get the parent device with the
        // subsystem/devtype pair of "usb"/"usb_device". This will
        // be several levels up the tree, but the function will find
        // it.
        // Note: The returned device is not referenced, so don't
        //   clean it up.
        usb = udev_device_get_parent_with_subsystem_devtype(
                  scsi_dev,
                  "usb",
                  "usb_device");
        if (usb && (strncmp(udev_device_get_devtype(scsi_dev), "partition", 9) == 0))
        {
            const char *tmp;
            std::string mountdir;
            uint64_t disk_size = 0;
            uint64_t avail_size = 0;
            uint64_t inode_count = 0;
            uint32_t block_size = DEFAULT_BLOCK_SIZE;

            // Build up our JSON structure for this device.
            json_info["name"] = udev_device_get_devnode(scsi_dev);
            json_info["type"] = GIGE_MEDIA_USB;

            // Do we have a label?
            if (!get_label(udev_device_get_devnode(scsi_dev), json_info))
            {
                // Use the product information.
                json_info["label"] = udev_device_get_sysattr_value(usb, "product");
            }

            // Find out if and where the device is mounted.
            mountdir.clear();
            get_mount_dir(json_info["name"].asString(), mountdir);
            json_info["mountdir"] = mountdir;

            // How big is the device?
            tmp = udev_device_get_sysattr_value(scsi_dev, "size");
            if (tmp)
                disk_size = strtoull(tmp, NULL, 10);

            tmp = udev_device_get_sysattr_value(scsi_dev, "queue/logical_block_size");
            if (tmp)
                block_size = atoi(tmp);

            // Get the capacity information.
            if (mountdir.empty())
            {
                disk_size = (disk_size * block_size);
                avail_size = (disk_size * block_size);
            }
            else if (get_capacity(mountdir.c_str(), &disk_size, &avail_size, &inode_count) != 0)
            {
                fi_log(LOG_DEBUG, "%s: get_capacity() failed.", __PRETTY_FUNCTION__);
                disk_size = (disk_size * block_size);
                avail_size = (disk_size * block_size);
            }

            json_info["devsize"] = disk_size;
            json_info["avail"] = avail_size;

            // Add the element to our array.
            dev_array.append(json_info);
        }

        // Release the device object.
        udev_device_unref(scsi_dev);
    }

    /* Free the enumerator object */
    udev_enumerate_unref(enumerate);

    // Release our udev object.
    udev_unref(udev);

usb_bail:
    // Wrap up the JSON information.
    root["usb"] = dev_array;
}

uint32_t storage::get_media_json(char *dest_ptr, uint32_t max_len)
{
    uint32_t ret_len = 0;
    Json::Value dev_tree;
    Json::StreamWriterBuilder wr_builder;

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // Do we need to rebuild the JSON data?
    if (json_message.empty())
    {
        // Find all the devices.
        sdcard_udev(dev_tree);
        usb_udev(dev_tree);
        nvme_udev(dev_tree);

        // Create a string.
        wr_builder["commentStyle"] = "All";
        wr_builder["indentation"] = "";
        json_message = Json::writeString(wr_builder, dev_tree);
    }

    // DEBUG: Dump the result.
//    fi_log(LOG_DEBUG, "%s: %s", __PRETTY_FUNCTION__, json_message.c_str());

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        // Copy over what we can.
//        fi_log(LOG_DEBUG, "%s: JSON string [%s] is too long! Only transferring %d bytes", __PRETTY_FUNCTION__, json_message.c_str(), max_len);
        memcpy(dest_ptr, json_message.c_str(), max_len - 1);
        ret_len = max_len - 1;

        // Shorten our string.
        json_message = json_message.substr((max_len - 1), std::string::npos);
    }
    else
    {
        // Copy the string over.
        memcpy(dest_ptr, json_message.c_str(), json_message.length());
        ret_len = json_message.length();

        // Clear the string.
        json_message.clear();
    }

    // Return the length, including the null-terminator.
    return ret_len + 1;
}

uint32_t storage::get_media_type(const std::string device)
{
    int i;
    Json::Value dev_tree;
    Json::Value dev_found;

    // Get the (possible) list of sdcard devices.
    sdcard_udev(dev_tree);
    if (!dev_tree["sdcard"].empty())
    {
        // See if our device is in the array.
        for (i = 0; i < (int)dev_tree["sdcard"].size(); ++i)
        {
            // Extract the entry.
            dev_found = dev_tree["sdcard"][i];

            // Do we have a match?
            if (device.compare(dev_found["name"].asString()) == 0)
            {
                // We have a match!
                return GIGE_MEDIA_SDCARD;
            }
        }
    }

    // Get the (possible) list of usb devices.
    dev_tree.clear();
    usb_udev(dev_tree);
    if (!dev_tree["usb"].empty())
    {
        // See if our device is in the array.
        for (i = 0; i < (int)dev_tree["usb"].size(); ++i)
        {
            // Extract the entry.
            dev_found = dev_tree["usb"][i];

            // Do we have a match?
            if (device.compare(dev_found["name"].asString()) == 0)
            {
                // We have a match!
                return GIGE_MEDIA_USB;
            }
        }
    }

    // Get the (possible) list of NVMe devices.
    dev_tree.clear();
    nvme_udev(dev_tree);
    if (!dev_tree["nvme"].empty())
    {
        // See if our device is in the array.
        for (i = 0; i < (int)dev_tree["nvme"].size(); ++i)
        {
            // Extract the entry.
            dev_found = dev_tree["nvme"][i];

            // Do we have a match?
            if (device.compare(dev_found["name"].asString()) == 0)
            {
                // We have a match!
                return GIGE_MEDIA_SSD;
            }
        }
    }

    // Nothing found.
    return 0;
}

uint32_t storage::get_available_media()
{
    int cur_status;
    uint32_t reg_value;
    const LicManager *p_lic;

    // Check to see if we're in the middle of a format operation.
    if (!formatter::formatDone(cur_status))
        return 0;

    // Get our license pointer.
    p_lic = CameraServer::get_license();

    // Initialize to "nothing is available" value.
    reg_value = 0;

    // Now, walk through each mount point and see if something is there.
    // In the case of the USB, all we want is to know that there is
    // at least one device. There may be more than one.
    if (storage::is_mounted(GIGE_MEDIA_USB))
        reg_value |= (GIGE_MEDIA_USB | GIGE_MEDIA_EXISTS_USB);
    else if (storage::exists(GIGE_MEDIA_USB))
        reg_value |= GIGE_MEDIA_EXISTS_USB;

    if (storage::is_mounted(GIGE_MEDIA_SDCARD))
        reg_value |= (GIGE_MEDIA_SDCARD | GIGE_MEDIA_EXISTS_SDCARD);
    else if (storage::exists(GIGE_MEDIA_SDCARD))
        reg_value |= GIGE_MEDIA_EXISTS_SDCARD;

    // Check to see if we're allowed to have a local SSD.
    // The "exists()" function will report the loal SSD as not
    // even existing if we are currently in LTR mode or
    // if we are not licensed to have it, but it might be
    // automatically mounted even if it isn't licensed.
    if ((p_lic != 0) && (p_lic->ssdMax() > 0))
    {
        if (storage::is_mounted(GIGE_MEDIA_SSD))
            reg_value |= (GIGE_MEDIA_SSD | GIGE_MEDIA_EXISTS_SSD);
        else if (storage::exists(GIGE_MEDIA_SSD))
            reg_value |= GIGE_MEDIA_EXISTS_SSD;
    }

    return reg_value;
}

/*!
 * @brief Determine if at least one type of the specified device is mounted
 *    on the file system.
 *
 * @return True if one of "media_sel" type is found mounted on the file system,
 *     false otherwise.
 */
bool storage::is_mounted(uint32_t media_sel)
{
    std::string device;
    std::string mount_dir;

    // We need to short-circuit this logic if we're looking for a mounted SSD
    // and we're in LTR mode.
    if (media_sel == GIGE_MEDIA_SSD)
    {
        FiFPGA *fpga_ptr = CameraServer::get_fpga();
        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
            return false;
    }

    // Get our destination base path.
    storage::get_media_device(media_sel, device);
    if (device.empty())
    {
        // No device found, so clearly nothing is mounted for the specified type.
        return false;
    }

    // Get the mount point for the device.
    storage::get_mount_dir(device, mount_dir);

    // Is it mounted?
    return !mount_dir.empty();
}

/*!
 * @brief Make a folder if it doesn't already exist.
 *
 * @param path  Folder we're interested in.
 * @param mode  Mode settings for the folder.
 *
 * @return 0 on success, -1 on failure.
 */
int storage::do_mkdir(const char *path, mode_t mode)
{
    struct stat st;
    int     status = 0;

    std::lock_guard<std::mutex> lock(storage_mutex);

    TP(TRACE_SEL_STORAGE, "%s: path='%s'", __PRETTY_FUNCTION__, path);

    // Check to see if the path exists.
    if (stat(path, &st) != 0)
    {
        /* Directory does not exist. EEXIST for race condition */
        if ((mkdir(path, mode) != 0) && (errno != EEXIST))
            status = -1;
    }

    // Make sure it's a directory if it exists.
    else if (!S_ISDIR(st.st_mode))
    {
        errno = ENOTDIR;
        status = -1;
    }

    return status;
}

/*!
 * @brief Ensure all directories in path exist.
 *
 * Algorithm takes the pessimistic view and works top-down to ensure
 * each directory in path exists, rather than optimistically creating
 * the last element and working backwards.
 *
 * @param path  Our path we want to build.
 * @param mode  The permissions setting for the path.
 *
 * @return 0 on success, -1 on failure with errno set appropriately.
 */
int storage::mkpath(const char *path, mode_t mode)
{
    char *pp;
    char *sp;
    int status;
    char *copypath = strndup(path, PATH_MAX);

    TP(TRACE_SEL_STORAGE, "%s: path='%s'", __PRETTY_FUNCTION__, path);

    status = 0;
    pp = copypath;
    while ((status == 0) && ((sp = strchr(pp, '/')) != 0))
    {
        if (sp != pp)
        {
            /* Neither root nor double slash in path */
            *sp = '\0';
            status = storage::do_mkdir(copypath, mode);
            *sp = '/';
        }
        pp = sp + 1;
    }
    if (status == 0)
        status = storage::do_mkdir(path, mode);
    free(copypath);

    return status;
}

/*!
 * @brief Mount the NVMe drive into the filesystem.
 *
 * @return true if everything goes well, false otherwise.
 */
bool storage::nvme_mount()
{
    std::string mount_cmd;
    std::string device;
    std::string mount_dir;

    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    get_media_device(GIGE_MEDIA_SSD, device);
    if (device.empty())
    {
        // This can happen with no USB found.
        fi_log(LOG_ERR, "%s: No NVMe device found", __PRETTY_FUNCTION__);
        return false;
    }

    // Get the mount point for the device. If one exists, then the
    // device is already mounted and our work is done.
    storage::get_mount_dir(device, mount_dir);
    if (!mount_dir.empty())
        return true;

    // Build up our mount command.
    mount_cmd = std::string("/usr/local/sbin/nvme-automount add ") + device;

    // Invoke the command.
    system(mount_cmd.c_str());

    // Return our result.
    return true;
}

/*!
 * @brief Format the device specified by 'device_name', using the format information found in 'format_info'
 *        and 'vol_label'.
 *
 * @param device_name   The name of the device we are interested in.
 * @param vol_label     The label to assign to the device (if applicable).
 * @param format_info   The GIGE_REG_FORMAT_MEDIA information to use.
 *
 * @return
 */
uint32_t storage::media_format(const char *device_name, const char *vol_label, uint32_t format_info)
{
    bool ssd_busy = false;
    std::string command_str;
    uint32_t status = GEV_STATUS_SUCCESS;

    fi_log(LOG_DEBUG, "%s: device='%s', vol_label='%s', format_info=0x%08x", __PRETTY_FUNCTION__, device_name, vol_label, format_info);

    // Are we attempting to format the SSD?
    if ((format_info & GIGE_MEDIA_AVAIL_MASK) == GIGE_MEDIA_SSD)
    {
        FiFPGA *fpga_ptr = CameraServer::get_fpga();

        // Make sure it's not already in use.
        if (imgLoadWorker::getInstance()->isRunning())
        {
            // The SSD is busy.
            ssd_busy = true;
        }

        else if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            uint8_t camera_state = CameraServer::get_camera_state();

            // Don't allow LTR mode formatting if we are using the drive, which means
            // anything other than LIVE or IDLE.
            if ((camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) && (camera_state != GIGE_CAMERA_STATE_IDLE))
            {
                ssd_busy = true;
            }
        }
    }

    // Check to see if the device is open by another process.
    if (!ssd_busy)
        ssd_busy = device_is_busy(device_name);

    // If we're busy, we can't format.
    if (ssd_busy)
        status = GEV_STATUS_BUSY;
    else if (!formatter::formatStart((format_info & GIGE_MEDIA_AVAIL_MASK), (format_info & GIGE_FORMAT_MASK), device_name, vol_label))
        status = GEV_STATUS_INVALID_PARAMETER;

    // Return our status.
    return status;
}

uint32_t storage::media_action(const char *action_str, int str_length)
{
    int action;
    Json::Value requested_action;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    std::string device;
    std::string vol_label;
    bool parsingSuccessful;
    uint32_t status;
    std::string command_str;

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(action_str, action_str + str_length, &requested_action, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, action_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        status = GEV_STATUS_INVALID_PARAMETER;
        goto media_bail;
    }

    // Get our information.
    action = requested_action["action"].asInt();
    device = requested_action["device"].asString();
    vol_label = requested_action["label"].asString();

    // Process the command.
    switch (action)
    {
    case GIGE_MA_EJECT:
        // LOG("%s: GIGE_MA_EJECT", __PRETTY_FUNCTION__);

        // Pass the request through the UDEV system.
        command_str = fmt::sprintf("sync; sync; /usr/bin/udevadm trigger -t devices -c remove --name-match %s", device.c_str());
        TP(TRACE_SEL_STORAGE, "%s: %s", __PRETTY_FUNCTION__, command_str.c_str());
        system(command_str.c_str());

        // We can't "eject" the SDCard, so skip this step if the device is the SDCard.
        if (storage::get_media_type(device) != GIGE_MEDIA_SDCARD)
        {
            // Issue the eject command to remove the device from the system.
            command_str = fmt::sprintf("/usr/bin/eject -F %s", device.c_str());
            TP(TRACE_SEL_STORAGE, "%s: %s", __PRETTY_FUNCTION__, command_str.c_str());
            system(command_str.c_str());
        }

        // All done.
        status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_MA_FORMAT:
        // LOG("%s: GIGE_MA_FORMAT", __PRETTY_FUNCTION__);
        status = media_format(device.c_str(), vol_label.c_str(), (uint32_t)requested_action["format"].asInt());
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected action=%d", __PRETTY_FUNCTION__, action);
        status = GEV_STATUS_INVALID_PARAMETER;
        break;
    }

media_bail:
    // Return our status.
    return status;
}

/*!
 * @brief Forcibly unmount the SSD filesystem.
 *
 * We can't use the storage::umount because it won't even recognize
 * that the SSD exists if the license doesn't allow it.
 *
 * This normally fails in LTR mode because the drive doesn't automount if it's not FAT formatted.
 * If that is not what is expected, use errno to find out what failed.
 *
 * @return true if we successfully unmounted, false if we got an error.
 */
bool storage::force_umount_ssd()
{
    bool rc;

    // Try our kinder, gentler umount
    rc = nvme_umount();
    if (!rc)
    {
        std::string device;

        for (int i = get_ssd_count(); i > 0; i--)
        {
            // Sanity check.
            get_media_device(GIGE_MEDIA_SSD, device, i - 1);
            if (device.empty())
            {
                // This can happen with no USB found.
                fi_log(LOG_ERR, "%s: No SSD found", __PRETTY_FUNCTION__);
                break;
            }

            // Is the device busy?
            if (device_is_busy(device.c_str()))
            {
                fi_log(LOG_DEBUG, "%s: The device (%s) is busy!", __PRETTY_FUNCTION__, device.c_str());
                break;
            }

            // Issue the unmount command.
            if (::umount(device.c_str()) == 0)
                rc = true;
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Unmount an NVMe drive from the filesystem.
 *
 * @return true if everything goes well, false otherwise.
 */
bool storage::nvme_umount()
{
    std::string umount_cmd;
    std::string device;
    std::string mount_dir;

    for (int i = get_ssd_count(); i > 0; i--)
    {
        // Sanity check.
        get_media_device(GIGE_MEDIA_SSD, device, i - 1);
        if (device.empty())
        {
            // This can happen with no USB found.
            fi_log(LOG_ERR, "%s: No NVMe device found", __PRETTY_FUNCTION__);
            return false;
        }

        // Make sure nothing is holding the device open.
        if (device_is_busy(device.c_str()))
        {
            fi_log(LOG_DEBUG, "%s: The device (%s) is busy!", __PRETTY_FUNCTION__, device.c_str());
            return false;
        }

        // Get the mount point for the device. If we can't find one, then the
        // device is already unmounted and our work is done.
        storage::get_mount_dir(device, mount_dir);
        if (!mount_dir.empty())
        {
            // Build up our mount command.
            umount_cmd = std::string("/usr/local/sbin/nvme-automount remove ") + device;
            TP(TRACE_SEL_STORAGE, "%s: umount cmd='%s'", __PRETTY_FUNCTION__, umount_cmd.c_str());

            // Invoke the command.
            system(umount_cmd.c_str());
        }
    }

    // Return our result.
    return true;
}

/*!
 * @brief Check to see if a device is "busy" by checking to see if ANY processes have
 *        the device open.
 *
 * By utilizing the "lsof" command, we can find out if any process has the device of interest
 * (specified by device_name) open. This could be a remote user using Samba to look at the
 * device, an application on the controller, etc.
 *
 * We are not so interested in the actual user, just that there exists one or more users.
 *
 * Error messages are redirected to /dev/null, because we really don't care about errors.
 *
 * @param device_name   The device we are interested in. Presumably something like "/dev/sdb1".
 * @return True if there are users of the device, false if it is not in use.
 */
bool storage::device_is_busy(const char *device_name)
{
    FILE *fp;
    bool dev_busy(false);
    std::string command_str;

    // Check to see if the device is open by another process.
    command_str = fmt::format("/usr/bin/lsof -Fn {0} 2> /dev/null", device_name);
    fp = popen(command_str.c_str(), "r");
    if (fp != NULL)
    {
        std::unique_ptr<char[]> read_buf = std::make_unique<char[]>(MAX_BUFFER_LEN + 1);

        // Point to our unique pointer. We use that so memory cleanup will be easy.
        if (read_buf.get() == nullptr)
        {
            FI_LOGIC_ERROR(__FILE__, __LINE__, "Unable to allocate the read buffer!");
            return GEV_STATUS_ERROR;
        }

        // Read data, if any.
        if (fgets(read_buf.get(), MAX_BUFFER_LEN, fp) != NULL)
        {
            // We read back data, so something is open on the device!
            dev_busy = true;
        }

        // Close the pipe.
        pclose(fp);
    }

    return dev_busy;
}

/*!
 * @brief Determine, by mount point, if the underlying device exists on the system
 *
 * We determine if the requested device is exists by  greping for it in the
 * ouput of the ennumerate_media_devices script. If we get a result back, then the
 * device exists. However, if we are running in LTR mode or the license does not
 * allow an SSD, we do not report the SSD as existing.
 *
 * @return true if the matching device exists, false otherwise.
 */
bool storage::exists(uint32_t media_sel)
{
    std::string device;

    // Get the first device found for the specified media.
    get_media_device(media_sel, device);
    if (device.empty())
    {
        // This can happen with no USB found.
        return false;
    }

    // Figure out what we are looking for
    if (media_sel == GIGE_MEDIA_SSD)
    {
        const LicManager *p_lic;

        // Get our license pointer.
        p_lic = CameraServer::get_license();

        // Are we allowed to have an SSD?
        if ((p_lic == nullptr) || (p_lic->ssdMax() == 0))
        {
            // We must have an SSD, even though one is not licensed!
            return false;
        }
    }

    // We found at least one device.
    return true;
}

/*!
 * @brief Copy a single file from the source_ptr to the target_ptr.
 *
 * @param target_ptr    Fully qualified path for the target image.
 * @param source_ptr    Fully qualified path for the source image.
 * @param buffer    [optional] A memory pool to use for a bounce buffer.
 *
 * @return true if all goes well, false otherwise.
 */
bool storage::stor_copy_file(const char *target_ptr, const char *source_ptr, uint8_t *buffer)
{
    int     ma_status;
    bool    rc;
    int     input, output;
    uint8_t *xfer_ptr;
    ssize_t xfer_len;
    std::unique_ptr<uint8_t[]> xfer_buffer;

    TP(TRACE_SEL_STORAGE, "%s: target='%s', source='%s'", __PRETTY_FUNCTION__, target_ptr, source_ptr);

    // Assume failure.
    rc = false;

    // Open the source file.
    if ((input = open(source_ptr, O_RDONLY)) == -1)
    {
        fi_log(LOG_ERR, "%s: Unable to open file (%s).\n", __PRETTY_FUNCTION__, source_ptr);
        goto bail_1;
    }

    // Open the destination. First try using direct I/O.
    if ((output = open(target_ptr, O_RDWR | O_CREAT | O_TRUNC | O_DIRECT, 0666)) == -1)
    {
        // Nope - fall back to normal, buffered I/O.
        if ((output = open(target_ptr, O_RDWR | O_CREAT | O_TRUNC, 0666)) == -1)
        {
            fi_log(LOG_ERR, "%s: Unable to open file (%s).\n", __PRETTY_FUNCTION__, target_ptr);
            goto bail_2;
        }
    }

    // Have we been provided a buffer?
    if (buffer != nullptr)
    {
        // Use the provided buffer.
        xfer_buffer.reset(buffer);
    }
    else
    {
        // The page size is 4096 - but this is the "correct" way to get it.
        xfer_page_size = sysconf(_SC_PAGESIZE);
        if (xfer_page_size <= 0)
            xfer_page_size = 4096;

        // Allocate our transfer buffer.
        ma_status = posix_memalign((void **)&xfer_ptr, xfer_page_size, XFER_BUFFER_LEN + xfer_page_size);
        if ((xfer_ptr == nullptr) || (ma_status != 0))
        {
            fi_log(LOG_ERR, "%s: Unable to allocate xfer buffer (status=%d).\n", __PRETTY_FUNCTION__, ma_status);
            goto bail_3;
        }
        xfer_buffer.reset(xfer_ptr);
    }

    // Loop through the file, and transfer it to the target.
    while ((xfer_len = read(input, xfer_buffer.get(), XFER_BUFFER_LEN)) > 0)
    {
        ssize_t write_len;
        ssize_t remainder_len;
        file_action_thread *fa_task;

        // Get the file action task pointer.
        fa_task = storage::factory()->get_fa_task();
        if (fa_task && fa_task->is_stop_requested())
        {
            fi_log(LOG_DEBUG, "%s: Stop requested.", __PRETTY_FUNCTION__);
            break;
        }

        // Figure out if we have any non-page aligned data to write.
        remainder_len = xfer_len % xfer_page_size;
        xfer_len = (xfer_len / xfer_page_size) * xfer_page_size;
        xfer_ptr = xfer_buffer.get();

        // See if we're done with the direct I/O.
        if (remainder_len > 0)
        {
            // Write the aligned data out.
            write_len = write(output, xfer_ptr, xfer_len);
            if (write_len != xfer_len)
            {
                fi_log(LOG_ERR, "%s: Unable to write file %s (rtn=%ld err=%d[%s]).", __PRETTY_FUNCTION__, target_ptr, write_len, errno, strerror(errno));
                break;
            }

            // Close the DIRECT I/O FD, and reopen it for standard I/O.
            close(output);
            if ((output = open(target_ptr, O_RDWR | O_APPEND, 0666)) == -1)
            {
                fi_log(LOG_ERR, "%s: Unable to open file (%s), err= %d[%s].", __PRETTY_FUNCTION__, target_ptr, errno, strerror(errno));
                break;
            }

            // Next write will take the remainder length.
            xfer_ptr += xfer_len;
            storage::factory()->total_transfer += xfer_len;
            xfer_len = remainder_len;
        }

        // Write out the data.
        write_len = write(output, xfer_ptr, xfer_len);
        if (write_len != xfer_len)
        {
            fi_log(LOG_ERR, "%s: Unable to write file %s (rtn=%ld, err=%d[%s]).", __PRETTY_FUNCTION__, target_ptr, write_len, errno, strerror(errno));
            break;
        }

        // Update the amount written.
        storage::factory()->total_transfer += xfer_len;

        // Update the status for the big files (i.e. AVI, MP4 or CAP).
        if (storage::factory()->total_transfer > (ssize_t)DL_CHUNK_SIZE)
            fa_status = GIGE_FA_STATUS_BUSY | ((uint32_t)(storage::factory()->total_transfer / DL_CHUNK_SIZE));
    }

    // Were we successful?
    rc = (xfer_len == 0);

    // Did the user give us a buffer?
    if (buffer != nullptr)
    {
        // Release ownership of the user's buffer.
        xfer_buffer.release();
    }

bail_3:
    // If the output file is open, close it.
    if (output > 0)
        close(output);
bail_2:
    // Close the input file.
    close(input);
bail_1:
    // Flush things out to disk.
    sync();

    // Return our status.
    return rc;
}

bool storage::save_path(const char *path_str, uint32_t path_len)
{
    // Is the delete thread active?
    if (storage::factory()->fa_task->is_running())
    {
        fi_log(LOG_ERR, "%s: task (op=0x%x) is still running!", __PRETTY_FUNCTION__, storage::factory()->fa_task->get_action());
        return false;
    }

    // Set the delete path.
    storage::factory()->fa_task->set_path(path_str, path_len);

    // All is well.
    return true;
}

bool storage::operation_in_progress()
{
    // Are we in the middle of a delete?
    if (storage::factory()->fa_task->is_running())
        return true;

    // Nope - nothing going on.
    return false;
}

int storage::reset_media_json(CS_NETWORK::CS_NETWORK_MESSAGE */*p_msg*/, uint32_t /*reg_value*/)
{
    // Reset the buffer.
    json_message.clear();
    return GEV_STATUS_SUCCESS;
}

uint32_t storage::get_status()
{
    // Return the tasks current status.
    return fa_status;
}

bool storage::start_action(uint32_t file_action)
{
    TP(TRACE_SEL_STORAGE, "%s: file_action=0x%X", __PRETTY_FUNCTION__, file_action);

    // Make sure the request is reasonable.
    if (storage::factory()->fa_task->is_running())
    {
        // We're running - action had better be to cancel.
        if (file_action != GIGE_FA_CANCEL)
        {
            fi_log(LOG_ERR, "%s: can't start 0x%x, task (op=0x%x) is still running", __PRETTY_FUNCTION__,
                   file_action, storage::factory()->fa_task->get_action());
            return false;
        }

        // Request a stop.
        storage::factory()->fa_task->stop();

        // All is well.
        return true;
    }

    // Not running..are we trying to cancel?
    if (file_action == GIGE_FA_CANCEL)
        return true;

    // Ok - a normal operation. Save it for use by the worker thread.
    storage::factory()->fa_task->set_action(file_action);

    // Clear our status value.
    fa_status = GIGE_FA_STATUS_BUSY;

    // Start up the task.
    storage::factory()->fa_task->start();

    // All is well.
    return true;
}

uint32_t storage::start_json_fa(const char *action_str, int str_length)
{
    int action;
    uint32_t status;
    std::string dst_path;
    std::string src_path;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    Json::Value requested_action;
    std::string errInfo;
    bool parsingSuccessful;

    TP(TRACE_SEL_STORAGE, "%s: action_str:'%s'", __PRETTY_FUNCTION__, action_str);
    LOG("%s: action_str:'%s'", __PRETTY_FUNCTION__, action_str);

    // Initialize our status value.
    status = GEV_STATUS_INVALID_PARAMETER;

    // Make sure the delete task is IDLE.
    if (storage::operation_in_progress())
    {
        // Return an error indicating we're busy.
        return GEV_STATUS_BUSY;
    }

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    try
    {
        parsingSuccessful = reader->parse(action_str, action_str + str_length, &requested_action, &errInfo);
        if (!parsingSuccessful)
        {
            fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, action_str, str_length);
            fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
            goto fa_bail;
        }

        // Get our information.
        action = requested_action["action"].asInt();

        // Do we have a source path?
        if (!requested_action["srcpath"].empty())
        {
            // Extract the path, to make it easier to use.
            src_path = requested_action["srcpath"].asString();

            // Set the file name we'll be loading from.
            storage::save_path(src_path.c_str(), src_path.length());
        }
        else
        {
            // Clear out the path.
            storage::save_path(nullptr, 0);
        }

        // If we are doing a copy, we should have a destination path.
        switch (action)
        {
        case GIGE_FA_COPY:
        case GIGE_FA_SAVE_LOGS:
            // Do we have a destination path?
            if (requested_action["dstpath"].empty())
            {
                // This is unusual indeed!
                fi_log(LOG_ERR, "%s: No destination path?!", __PRETTY_FUNCTION__);
                goto fa_bail;
            }

            // Get the destination path.
            dst_path = requested_action["dstpath"].asString();
            if (dst_path.length() == 0)
            {
                fi_log(LOG_DEBUG, "%s: Error missing destination path for copy!", __PRETTY_FUNCTION__);
                goto fa_bail;
            }
            else
            {
                struct stat stat_buf;

                // Does the destination already exist?
                if (stat(dst_path.c_str(), &stat_buf) == 0)
                {
                    fi_log(LOG_DEBUG, "%s: Refuse to overwrite existing file [%s]", __PRETTY_FUNCTION__, dst_path.c_str());
                    fa_status = GIGE_FA_STATUS_ERROR | EEXIST;
                    goto fa_bail;
                }
            }

            // Copy the information over.
            storage::factory()->fa_task->set_media_path(dst_path.c_str(), dst_path.length());
            break;
        case GIGE_FA_CANCEL:
        case GIGE_FA_DELETE:
        case GIGE_FA_LISTING:
        case GIGE_FA_LUT_LISTING:
        // Nothing more to do here.
        default:
            break;
        }

        // Get things moving.
        if (!storage::start_action(action))
        {
            status = GEV_STATUS_BUSY;
            goto fa_bail;
        }

        // All is well.
        status = GEV_STATUS_SUCCESS;
    }
    catch (std::exception& e)
    {
        // Note that the standard exceptions "what()" returns a 'const char *'.
        fi_log(LOG_DEBUG, "%s: Unhandled C++ exception: '%s'", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        LOG("%s: Unhandled exception of unknown type", __PRETTY_FUNCTION__);
    }

fa_bail:
    // Return our status.
    return status;
}

/*!
 * @brief Return the length of the XML tree output buffer.
 */
int storage::get_xml_tree_len()
{
    return storage::factory()->m_xmlOutputBufferLen;
}

/*!
 * @brief Return a pointer to the XML tree output buffer.
 */
const char *storage::get_xml_tree_ptr()
{
    return storage::factory()->m_xmlOutputBuffer;
}

/*!
 * @brief Build up an XML representation of a folder.
 *
 * @return 0 on success, errno otherwise.
 */
int storage::buildXmlTree()
{
    int     rc;
    std::string file_path;
    uint64_t    capacity;
    uint64_t    avail;
    uint64_t    inodes;
    std::size_t slash_pos;
    std::string temp_str;
    xmlDocPtr   doc = NULL;         // Document pointer.
    xmlNodePtr  root_node = NULL;       // Node pointer.

    // Get the pointer to the target path.
    file_path = storage::factory()->fa_task->get_file_path();
    if (file_path.empty())
    {
        fi_log(LOG_ERR, "%s: NULL or empty file path!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Make sure our return values are initialized.
    if (m_xmlOutputBuffer != NULL)
    {
        xmlFree(m_xmlOutputBuffer);
        m_xmlOutputBuffer = nullptr;
        m_xmlOutputBufferLen = 0;
    }

    // Make sure the last character is a path divider.
    if (file_path.at(file_path.length() - 1) != '/')
        file_path.append(1, '/');

    // Creates a new document, a node and set it as a root node.
    doc = xmlNewDoc(BAD_CAST "1.0");
    root_node = xmlNewNode(NULL, BAD_CAST "root");
    xmlDocSetRootElement(doc, root_node);

    // Set the specified folder as text for the root node.
    xmlNewProp(root_node, BAD_CAST "target", BAD_CAST file_path.c_str());

    // Strip off the filename or last folder. We really only want the root
    // folder when we are trying to get capacity. Since we just guaranteed
    // a trailing '/', make sure we skip that too.
    temp_str.assign(file_path);
    slash_pos = temp_str.find_last_of('/', temp_str.length() - 2);
    temp_str = temp_str.substr(0, slash_pos);

    // Get the capacity of the target device.
    if (!temp_str.empty() && (get_capacity(temp_str.c_str(), &capacity, &avail, &inodes) == 0))
    {
        temp_str = fmt::sprintf("%" PRIu64, capacity);
        xmlNewProp(root_node, BAD_CAST "capacity", BAD_CAST temp_str.c_str());
        temp_str = fmt::sprintf("%" PRIu64, avail);
        xmlNewProp(root_node, BAD_CAST "available", BAD_CAST temp_str.c_str());
        temp_str = fmt::sprintf("%" PRIu64, inodes);
        xmlNewProp(root_node, BAD_CAST "inodes", BAD_CAST temp_str.c_str());
    }

    try
    {
        // Make sure the path exists.
        if (std::filesystem::exists(file_path))
        {
            // Walk through the contents of the path. But don't go into any subfolders.
            for (auto const& dir_entry : std::filesystem::directory_iterator(file_path))
            {
                xmlNodePtr f_node;      // File/Folder node.

                // Check to see if the user wants to cancel.
                if (storage::factory()->fa_task->is_stop_requested())
                {
                    TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
                    break;
                }

                /*
                * xmlNewChild() creates a new node, which is "attached" as child node
                 * of root_node node.
                 */
                if (dir_entry.is_directory())
                {

                    // Add a node for this folder.
                    f_node = xmlNewChild(root_node, NULL, BAD_CAST "folder", NULL);
                    xmlNewProp(f_node, BAD_CAST "name", BAD_CAST dir_entry.path().filename().c_str());

                    // Go get the stats for the folder.
                    get_folder_stats(f_node, dir_entry.path());
                }
                else
                {
                    std::string file_ext;

                    // Figure out the extension on the file. This will return the extension as ".xxx", so
                    // get rid of that ".".
                    file_ext = dir_entry.path().filename().extension();
                    file_ext.erase(0, 1);
//            LOG("%s: filename='%s', ext='%s'", __PRETTY_FUNCTION__, dir_entry.path().filename().c_str(), file_ext.c_str());

                    // Only add metadata for our video files.
                    if ((file_ext.compare("avi") == 0) ||
                        (file_ext.compare("jpg") == 0) ||
                        (file_ext.compare("mp4") == 0) ||
                        (file_ext.compare("cap") == 0))
                    {
                        // This is a file, so create a file node.
                        f_node = xmlNewChild(root_node, NULL, BAD_CAST "file", NULL);
                        xmlNewProp(f_node, BAD_CAST "name", BAD_CAST dir_entry.path().filename().c_str());
                        get_file_stats(f_node, dir_entry.path(), file_ext);
                    }
                }
            }
        }
    }
    catch (std::filesystem::filesystem_error& e)
    {
        fi_log(LOG_ERR, "%s: Unable to list folder '%s': %s", __PRETTY_FUNCTION__, file_path.c_str(), e.what());
    }


    // Dump document to a buffer.
    xmlDocDumpFormatMemory(doc, (xmlChar **)&m_xmlOutputBuffer, &m_xmlOutputBufferLen, 0);

    // All is well.
    rc = 0;

    // Free the document.
    xmlFreeDoc(doc);

    // Free the global variables that may have been allocated by the parser.
    xmlCleanupParser();

    return rc;
}

/*!
 * @brief Build up an XML representation of the list of LUT files.
 *
 * @return 0 on success, errno otherwise.
 */
int storage::buildLutXmlTree()
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int     rc;
    int     file_index;         // Location in the file path where we want to put the file name.
    std::string file_path;      // The path we're working with.
    DIR     *dir_ptr;
    FILE    *fd;
    xmlDocPtr   doc = NULL;     // Document pointer.
    xmlNodePtr  root_node = NULL;   // Node pointer.
    xmlNodePtr  f_node = NULL;
    struct dirent *dir_info;
    char    small_string[16];
    int     i;
    int     count;
    std::unique_ptr<LUT_RGB_FILE> lf;

    // Make sure our return values are initialized.
    if (m_xmlOutputBuffer != NULL)
    {
        xmlFree(m_xmlOutputBuffer);
        m_xmlOutputBuffer = NULL;
        m_xmlOutputBufferLen = 0;
    }

    // Set the directory name into our local buffer.
    file_path = fmt::sprintf(LUT_FILES_DIR, get_home_path());
    file_index = file_path.length();

    // Creates a new document, a node and set it as a root node.
    doc = xmlNewDoc(BAD_CAST "1.0");
    root_node = xmlNewNode(NULL, BAD_CAST "lut_listing");
    xmlDocSetRootElement(doc, root_node);

    // Set the count of lut files as a property for the root node.
    //  xmlNewProp( root_node, BAD_CAST "lut_count", BAD_CAST num_lut_files );

    // Add the active lut number as a child of the root.
    f_node = xmlNewChild(root_node, NULL, BAD_CAST "active_lut", NULL);
    snprintf(small_string, sizeof(small_string), "%u", regs[GIGE_REG_LUT].get_value() & GIGE_LUT_MASK);
    xmlNewProp(f_node, BAD_CAST "id", BAD_CAST small_string);

    // Open the directory containing the LUT files.
    dir_ptr = opendir(file_path.c_str());
    if (dir_ptr == NULL)
    {
        fi_log(LOG_WARNING, "%s: opendir(%s) failed. %d(%s)", __PRETTY_FUNCTION__, file_path.c_str(), errno, strerror(errno));
        goto bail;
    }

    // Loop through the contents of the folder.
    while ((dir_info = readdir(dir_ptr)) != NULL)
    {
        xmlNodePtr  f_cur_lut = NULL;
        xmlNodePtr  f_sub = NULL;
        unsigned    lut_id;
        int read_remainder;

        // Check to see if the user wants to cancel.
        if (storage::factory()->fa_task->is_stop_requested())
        {
            TP(TRACE_SEL_STORAGE, "%s[%d]: Stop requested.", __PRETTY_FUNCTION__, __LINE__);
            break;
        }

        // Skip over this file if it isn't a LUT.
        if ((dir_info->d_type == DT_REG) && (sscanf(dir_info->d_name, "%u.lut", &lut_id) == 1))
        {
            if (lf.get() == nullptr)
                lf = std::unique_ptr<LUT_RGB_FILE>(new LUT_RGB_FILE);
            if (lf.get() == nullptr)
                continue;
        }
        else
        {
            // Not a matching file.
            continue;
        }

        // Append the name to our path.
        file_path.replace(file_index, std::string::npos, dir_info->d_name);

        // Try to read the file.
        fd = fopen(file_path.c_str(), "rb");
        if (fd == NULL)
            continue;
        count = fread((char *)lf.get(), 1, sizeof(LUT_HEADER), fd);
        if (count != sizeof(LUT_HEADER))
        {
            fi_log(LOG_DEBUG, "%s: header read '%s' failed", __PRETTY_FUNCTION__, file_path.c_str());
            continue;
        }

        // Figure out how much is left in the file.
        if (lf->hdr.lutDataType == LUT_SERIESNC_COLOR)
            read_remainder = sizeof(LUT_RGB_FILE) - count;
        else
            read_remainder = sizeof(LUT_FILE) - count;

        // Read the remainder.
        count = fread((char *)&lf->redKnots, 1, read_remainder, fd);
        fclose(fd);
        if (count != read_remainder)
        {
            fi_log(LOG_DEBUG, "%s: header remainder read '%s' failed", __PRETTY_FUNCTION__, file_path.c_str());
            continue;
        }

        // We were able to read the file so add this node.
        f_cur_lut = xmlNewChild(root_node, NULL, BAD_CAST "lut_info", NULL);
        snprintf(small_string, sizeof(small_string), "%u", lut_id);
        xmlNewProp(f_cur_lut, BAD_CAST "id", BAD_CAST small_string);

        xmlNewChild(f_cur_lut, NULL, BAD_CAST "description", BAD_CAST lf->hdr.description);
        snprintf(small_string, sizeof(small_string), "%u", lf->hdr.lutDataType);
        xmlNewChild(f_cur_lut, NULL, BAD_CAST "lut_data_type", BAD_CAST small_string);

        // Create our "knots" branch.
        f_sub = xmlNewChild(f_cur_lut, NULL, BAD_CAST "knots", NULL);
        i = 0;
        do
        {
            snprintf(small_string, sizeof(small_string), "%u:%u", lf->redKnots[i].x, lf->redKnots[i].y);
            xmlNewChild(f_sub, NULL, BAD_CAST "knot", BAD_CAST small_string);

            ++i;

        } while (((lf->redKnots[i - 1].x < 1023) && (lf->redKnots[i].x != 0)) && (i < LUT_MAX_KNOT_COUNT));

        if (lf->hdr.lutDataType == LUT_SERIESNC_COLOR)
        {
            f_sub = xmlNewChild(f_cur_lut, NULL, BAD_CAST "green_knots", NULL);
            i = 0;
            do
            {
                snprintf(small_string, sizeof(small_string), "%u:%u", lf->greenKnots[i].x, lf->greenKnots[i].y);
                xmlNewChild(f_sub, NULL, BAD_CAST "green_knot", BAD_CAST small_string);

                ++i;

            } while (((lf->greenKnots[i - 1].x < 1023) && (lf->greenKnots[i].x != 0)) && (i < LUT_MAX_KNOT_COUNT));

            f_sub = xmlNewChild(f_cur_lut, NULL, BAD_CAST "blue_knots", NULL);
            i = 0;
            do
            {
                snprintf(small_string, sizeof(small_string), "%u:%u", lf->blueKnots[i].x, lf->blueKnots[i].y);
                xmlNewChild(f_sub, NULL, BAD_CAST "blue_knot", BAD_CAST small_string);

                ++i;

            } while (((lf->blueKnots[i - 1].x < 1023) && (lf->blueKnots[i].x != 0)) && (i < LUT_MAX_KNOT_COUNT));
        }
    }

    // Close our folder.
    closedir(dir_ptr);

bail:
    // Dump document to a buffer.
    xmlDocDumpFormatMemory(doc, (xmlChar **)&m_xmlOutputBuffer, &m_xmlOutputBufferLen, 0);

    {
        FILE *f;
        char temp_path[PATH_MAX];

        snprintf(temp_path, PATH_MAX, "%s/%s", get_home_path(), ".config/Fastec/scratch/lut.xml");
        f = fopen(temp_path, "w");
        if (f != NULL)
        {
            xmlDocDump(f, doc);
            fclose(f);
        }
    }

    // All is well.
    rc = 0;

    // Free the document.
    xmlFreeDoc(doc);

    // Free the global variables that may have been allocated by the parser.
    xmlCleanupParser();

    return rc;
}

/*!
 * @brief Calculate the MD5 checksum for the configuration file
 */
void storage::md5sum_file(const char *file_path, const char *md5_path)
{
    FILE    *fp;
    std::string cmd_str;

    // The file is finished writing, create the checksum file, which we will use to
    // determine validity on a restore of the registers.
    cmd_str = fmt::sprintf("/usr/bin/md5sum %s", file_path);
    fp = popen(cmd_str.c_str(), "r");
    if (fp != NULL)
    {
        std::unique_ptr<char[]> md5_buf_ptr = std::make_unique<char[]>(MAX_BUFFER_LEN + 1);

        // Point to our unique pointer. We use that so memory cleanup will be easy.
        if (md5_buf_ptr.get() == nullptr)
        {
            FI_LOGIC_ERROR(__FILE__, __LINE__, "Unable to allocate the MD5 buffer!");
            return;
        }

        // Null-terminate our string.
        memset(md5_buf_ptr.get(), 0, MAX_BUFFER_LEN + 1);

        // Read data, if any.
        while (fgets(md5_buf_ptr.get(), MAX_BUFFER_LEN, fp) != NULL)
        {
            // Did we get anything?, we need the checksum so that it can be written
            if (*md5_buf_ptr.get() != 0)
                break;
        }

        // Close the pipe.
        pclose(fp);

        // If we got a checksum, put it into the checksum file.
        if (*md5_buf_ptr.get() != 0)
        {
            fp = fopen(md5_path, "w");
            if (fp != NULL)
            {
                fputs(md5_buf_ptr.get(), fp);
                fclose(fp);
            }
            else
            {
                fi_log(LOG_WARNING, "%s: cannot open file %s to save checksum errno:%d(%s) checksum will not be saved",
                       __PRETTY_FUNCTION__, md5_path, errno, strerror(errno));
            }
        }
    }
    else
    {
        fi_log(LOG_WARNING, "%s: cannot create register checksum.", __PRETTY_FUNCTION__);
    }
}

/*!
 *  @brief Retrieve the device path of the NVME SSD used for LR Mode.
 *  @param dev_path The device path for the LR Mode SSD, or empty.
 */
void storage::get_ltr_device(std::string& dev_path)
{
    Json::Value nvme_info;
    std::size_t index;
    static std::string found_internal_ssd;      // Once it's found we don't need to re-find it. The SSD never moves while we're powered up.

    // Have we already found the SSD?
    if (!found_internal_ssd.empty())
    {
        // Copy over to the dev_path.
        dev_path = found_internal_ssd;
        return;
    }

    // Clear out the passed in string.
    dev_path.clear();

    // Get the information for the NVME in the head.
    storage::get_internal_nvme(nvme_info);
    if (nvme_info.empty())
    {
        // No SSD found.
        fi_log(LOG_WARNING, "%s: No SSD found.", __PRETTY_FUNCTION__);
        return;
    }

    // fi_log(LOG_DEBUG, "%s: nvme_info: %s", __PRETTY_FUNCTION__, nvme_info.toStyledString().c_str());

    // Save the name.
    dev_path = nvme_info["name"].asString();

    // Make sure it isn't ending with a partition, i.e. "p1".
    index = dev_path.rfind('p');
    if (index != std::string::npos)
    {
        // Truncate at the found location.
        dev_path.resize(index);
    }

    // Save the found device path.
    found_internal_ssd = dev_path;
}

/*!
 * @brief Get a listing of files, optionally with the specified suffix, in the dir_path.
 *
 * @param dir_path  [IN] Folder to get a listing for.
 * @param suffix    [IN] Optional suffix (e.g. ".tif", don't forget the '.'!), empty if not to be used.
 * @param files     [OUT] Vector of files found.
 */
void storage::get_listing(const std::string& dir_path, const std::string& suffix, std::vector<std::string>& files)
{
    std::lock_guard<std::mutex> lock(storage_mutex);

    // Clear the files vector.
    files.clear();

    try
    {
        // Make sure the path exists.
        if (std::filesystem::exists(dir_path))
        {
            // Walk through the folder.
            for (const auto& file : std::filesystem::directory_iterator(dir_path))
            {
                // Only interested in regular files.
                if (file.is_regular_file())
                {
                    std::string file_path = file.path();

                    // Are we looking for a specific file suffix?
                    if (!suffix.empty())
                    {
                        // Do we have a match?
                        if (suffix.compare(file.path().extension()) != 0)
                        {
                            // No match, so ignore this file.
                            file_path.clear();
                        }
                    }

                    // Do we still have a file?
                    if (!file_path.empty())
                        files.push_back(file_path);
                }
            }
        }
    }
    catch (std::filesystem::filesystem_error& e)
    {
        fi_log(LOG_ERR, "%s: Unable to list folder '%s': %s", __PRETTY_FUNCTION__, dir_path.c_str(), e.what());
        files.clear();
    }

    // If there is more than 1 entry, sort the vector.
    if (files.size() > 1)
        std::sort(files.begin(), files.end());
}

/*!
 * @brief Return the path of the block device for the specified media.
 *    For media which might have multiple device, return the first.
 */
void storage::get_media_device(uint32_t media_sel, std::string& device, int index)
{
    Json::Value dev_tree;
    Json::Value first_dev;

    // Clear out the passed in string.
    device.clear();

    // Sanity check.
    if (index < 0)
    {
        fi_log(LOG_DEBUG, "%s: index (%d) can't be negative!", __PRETTY_FUNCTION__, index);
        return;
    }

    // Get the JSon information for the media specified.
    switch (media_sel)
    {
    case GIGE_MEDIA_SDCARD:
        // Get the (possible) list of devices.
        sdcard_udev(dev_tree);
        if (!dev_tree["sdcard"].empty())
        {
            // Get the first element in the array.
            first_dev = dev_tree["sdcard"][0];
        }
        break;
    case GIGE_MEDIA_USB:
        // Get the (possible) list of devices.
        usb_udev(dev_tree);
        if (!dev_tree["usb"].empty())
        {
            // Get the first element in the array.
            first_dev = dev_tree["usb"][0];
        }
        break;
    case GIGE_MEDIA_SSD:
        // When the caller is asking for the GIGE_MEDIA_SSD, that means
        // the INTERNAL NVME drive. No other will do.
        storage::get_internal_nvme(first_dev);
        break;
    default:
        break;
    }

    // Make a copy of the first entry.
    if (!first_dev.empty())
        device = first_dev["name"].asString();
}

/*!
 * @brief Get the capabilities of the "internal" SSD.
 *
 * @note This only retrieves the capabilities of the NVME SSD that is
 *   in the head of the camera.
 */
uint32_t storage::get_ssd_capabilities()
{
    char ssd_model[ MAX_BUFFER_LEN ];
    FILE *fp;
    uint32_t ret_value = 0;
    std::string ssd_path;
    Json::Value nvme_info;

    // Get the information for the NVME in the head.
    storage::get_internal_nvme(nvme_info);
    if (nvme_info.empty())
    {
        // No SSD found, so no capabilities.
        fi_log(LOG_WARNING, "%s: No SSD found.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Build up our path to the model entry.
    ssd_path = fmt::sprintf("/sys%s/device/model", nvme_info["devpath"].asString());

    // Open our file which contains the SSD model information. This might fail if the drive
    // is formatted as a standard (i.e. NOT LR mode) drive,.
    fp = fopen(ssd_path.c_str(), "r");
    if (!fp)
    {
        // OK...if this is a standard drive then we need to modify the path a little.
        // Try up a level.
        ssd_path = fmt::sprintf("/sys%s/../device/model", nvme_info["devpath"].asString());
        fp = fopen(ssd_path.c_str(), "r");
        if (!fp)
        {
            fi_log(LOG_WARNING, "%s: Cannot open (%s)", __PRETTY_FUNCTION__, ssd_path.c_str());
            return 0;
        }
    }

    // Read the model name.
    if (fgets(ssd_model, MAX_BUFFER_LEN, fp) == NULL)
    {
        fi_log(LOG_WARNING, "%s: Cannot read (%s)", __PRETTY_FUNCTION__, ssd_path.c_str());
        fclose(fp);
        return 0;
    }

    // Close the file.
    fclose(fp);

    // Trim the trailing LF character.
    ssd_model[strlen(ssd_model) - 1] = 0;
//    fi_log(LOG_DEBUG, "%s: ssd_model='%s'", __PRETTY_FUNCTION__, ssd_model);

    // Assume nothing to tell.
    ret_value = 0;

    // Check to see if this is a model that supports some kind of secure erase operation.
    // Note, for these NVME SSDs, we could query the drive directly:
    // # sudo nvme id-ctrl -H <device, i.e. /dev/nvme0n1>
    //
    // Looking at the output for the "fna" bits:
    //
    //  fna     : 0x5
    //    [2:2] : 0x1   Crypto Erase Supported as part of Secure Erase
    //    [1:1] : 0     Crypto Erase Applies to Single Namespace(s)
    //    [0:0] : 0x1   Format Applies to All Namespace(s)
    //
    // We can see that this drive does, in fact, support the Crytpo Erase function.
    for (unsigned i = 0 ; i < sizeof(ssd_model_list) / sizeof(ssd_model_list[0]); ++i)
    {
        if (strncmp(ssd_model, ssd_model_list[i].name, strlen(ssd_model_list[i].name)) == 0)
        {
            ret_value = ssd_model_list[i].capabilities;
            break;
        }
    }

    // Return our result.
    return ret_value;
}

int storage::get_ssd_count()
{
    Json::Value dev_tree;

    // Populate our tree with available NVME drives.
    dev_tree.clear();
    nvme_udev(dev_tree);

    // Return the number of NVME drives.
    return (int)dev_tree["nvme"].size();
}

/*!
 * @brief Check to see if the device specified by the 'target_path' supports files larger than 4GB.
 *
 * @returns 0 if the file system supports large files, an errno code (EFBIG for example) if it doesn't.
 */
int storage::fs_supports_large_file(const std::string target_path)
{
    int err_code;
    FILE *mnt_table;
    std::string fs_name_buf;

    std::lock_guard<std::mutex> lock(storage_mutex);
    TP(TRACE_SEL_STORAGE, "%s: Looking for '%s'", __PRETTY_FUNCTION__, target_path.c_str());

    try
    {
        struct mntent mnt_info;

        // Allocate our working buffer.
        std::unique_ptr<char[]> temp_buf = std::make_unique<char[]>(MAX_TEMP_BUFFER_LEN + 1);

        // Open up the mount table.
        mnt_table = setmntent("/proc/mounts", "r");

        // Loop through all the entries in the file looking for ours.
        while (getmntent_r(mnt_table, &mnt_info, temp_buf.get(), MAX_TEMP_BUFFER_LEN) != nullptr)
        {
            // Do we have a match?
            if (target_path.compare(0, strlen(mnt_info.mnt_dir), mnt_info.mnt_dir) == 0)
            {
                // Get the filesystem name.
                fs_name_buf = mnt_info.mnt_type;

                // All done.
                break;
            }
        }
    }
    catch (...)
    {
        fi_log(LOG_DEBUG, "%s: Caught an exception!", __PRETTY_FUNCTION__);
    }

    // Close access to the mount information.
    endmntent(mnt_table);

    // Did we find a match?
    if (fs_name_buf.empty())
    {
        // There was no such device found.
        err_code = ENXIO;
    }
    else
    {
        // No large files on VFAT systems.
        if (fs_name_buf.compare("vfat") == 0)
            err_code = EFBIG;
        else
            err_code = 0;
    }

    // Return our result.
    return err_code;
}

//--------------------------------------------------------------------
// Class implementation: file_action_thread
//--------------------------------------------------------------------

/*!
 * @brief Copy a single file from the source_ptr to the target_ptr.
 *
 * @param target_ptr Fully qualified path for the target image.
 * @param source_ptr Fully qualified path for the source image.
 *
 * @return true if all goes well, false otherwise.
 */
bool file_action_thread::thread_copy_file(const char *target_ptr, const char *source_ptr)
{
    // LOG("%s: target='%s', source='%s'", __PRETTY_FUNCTION__, target_ptr, source_ptr);

    // Sanity check.
    if (strncasecmp(target_ptr, source_ptr, strlen(target_ptr)) == 0)
    {
        fi_log(LOG_DEBUG, "%s: File [%s] cannot copy over itself!", __PRETTY_FUNCTION__, target_ptr);
        return false;
    }

    // Do the copy.
    return storage::stor_copy_file(target_ptr, source_ptr, xfer_buffer);
}

void file_action_thread::set_media_path(const char *sel_path, uint32_t path_len)
{
    // Sanity check.
    if (path_len >= PATH_MAX)
        path_len = PATH_MAX - 1;

    // Make a local 'safe' copy.
    media_path.clear();
    if (sel_path && (path_len > 0))
        media_path.append(sel_path, path_len);
}

/*!
 * @brief Save the path to the file/folder which is being processed.
 *
 * @param sel_path  Path to the file/folder to be processed.
 * @param path_len  Number of characters in sel_path.
 */
void file_action_thread::set_path(const char *sel_path, uint32_t path_len)
{
    // Sanity check.
    if (path_len >= PATH_MAX)
        path_len = PATH_MAX - 1;

    // Make a local 'safe' copy.
    file_path.clear();
    if (sel_path && (path_len > 0))
        file_path.append(sel_path);
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void file_action_thread::cleanup()
{
    // We're not running anymore.
    thread_running = false;
}

int file_action_thread::do_copy()
{
    int     status;
    size_t  file_index;

    TP(TRACE_SEL_STORAGE, "%s: media_path=[%s]", __PRETTY_FUNCTION__, media_path.c_str());
    LOG("%s: file_path=[%s] media_path=[%s]", __PRETTY_FUNCTION__, file_path.c_str(), media_path.c_str());

    // Build and check our destination base path.
    if (media_path.empty())
    {
        fi_log(LOG_ERR, "%s: No media_path set!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Save the destination.
    target_path = media_path;

    // If our source is a folder, we're good to go. If it's a file,
    // we need to back off the filename before creating the destination path.
    if (S_ISDIR(file_stat.st_mode))
    {
        // We're good - set our index to "not found".
        file_index = std::string::npos;
    }
    else
    {
        // Back off the filename.
        file_index = target_path.find_last_of('/');
        if (file_index != std::string::npos)
            target_path.resize(file_index);
    }

    // Make sure the target folder exists. We may be copying a file, so the
    // mkpath() function might return an error indicating that last element
    // isn't a folder.
    if ((storage::mkpath(target_path.c_str(), 0777) == -1) && (errno != ENOTDIR))
    {
        fi_log(LOG_ERR, "%s: Unable to create the target path '%s'. %d(%s)", __PRETTY_FUNCTION__, target_path.c_str(), errno, strerror(errno));
        status = errno;
        goto bail_1;
    }

    // Make sure there is enough space on the destination to copy this.
    if (!storage::copy_space_available(file_path.c_str(), target_path.c_str()))
    {
        // We already logged the error.
        return ENOSPC;
    }

    // The page size is 4096 - but this is the "correct" way to get it.
    xfer_page_size = sysconf(_SC_PAGESIZE);
    if (xfer_page_size < 0)
        xfer_page_size = 4096;

    // Allocate our transfer buffer.
    status = posix_memalign((void **)&xfer_buffer, xfer_page_size, XFER_BUFFER_LEN + xfer_page_size);
    if ((xfer_buffer == NULL) || (status != 0))
    {
        fi_log(LOG_ERR, "%s: Unable to allocate xfer buffer (status=%d).\n", __PRETTY_FUNCTION__, status);
        status = ENOMEM;
        goto bail_1;
    }

    // Initialize our status.
    fa_status = GIGE_FA_STATUS_BUSY;
    storage::factory()->reset_total_transfer();
    storage::factory()->set_capstk_transfer(false);

    // If we are copying a folder, then walk the source tree and copy it
    // to the target folder. Otherwise, just copy the file.
    if (S_ISDIR(file_stat.st_mode))
    {
        // Restore the folder separator if we are copying a file.
        if (file_index != std::string::npos)
            target_path.append(1, '/');

        // Save our offset values.
        target_offset = target_path.length();
        file_offset = file_path.length();
        TP(TRACE_SEL_STORAGE, "%s: file_path=[%s], target_path=[%s]", __PRETTY_FUNCTION__, file_path.c_str(), target_path.c_str());
        TP(TRACE_SEL_STORAGE, "%s: file_offset=%d, target_offset=%d", __PRETTY_FUNCTION__, file_offset, target_offset);

        // Walk the tree.
        status = nftw(file_path.c_str(), copy_cb, 64, FTW_PHYS);
        if (status == -1)
            status = errno;
    }
    else
    {
        int file_len;
        int target_len;
        struct stat stat_buf;
        std::string file_str(file_path);
        std::string target_str(media_path);
        std::vector<std::string> suffixes = {"cfg", "pm", "txt", "xml"};

        // Get the lengths for ease of use. The lengths are the position after the suffix
        // separator (i.e. the period).
        file_len = file_str.find_last_of('.') + 1;
        target_len = target_str.find_last_of('.') + 1;

        // Copy the file.
        if (!thread_copy_file(target_str.c_str(), file_str.c_str()))
        {
            fi_log(LOG_ERR, "%s: Unable to copy file '%s'. %d(%s)", __PRETTY_FUNCTION__, file_str.c_str(), errno, strerror(errno));
            status = errno;
            if (status == 0)
                status = ENOSPC;
            goto bail_2;
        }

        // Update our status. The busy bit is set, we're just incrementing
        // the count at the bottom.
        fa_status++;

        // Now copy all existing "meta" files.
        for (std::string suffix_str : suffixes)
        {
            // Copy the CFG file. It's OK if the file isn't there.
            //
            // NOTE: We're assuming the suffix is three characters long. If not, there's
            //       going to be a problem.
            file_str.replace(file_len, 3, suffix_str);
            target_str.replace(target_len, 3, suffix_str);

            // Check to see if the CFG file exists.
            if (stat(file_str.c_str(), &stat_buf) == 0)
            {
                // Copy it over.
                if (!thread_copy_file(target_str.c_str(), file_str.c_str()))
                {
                    fi_log(LOG_ERR, "%s: Unable to copy file '%s'. %d(%s)", __PRETTY_FUNCTION__, file_str.c_str(), errno, strerror(errno));
                    status = errno;
                    if (status == 0)
                        status = ENOSPC;
                    goto bail_2;
                }
            }

            // Update our status. The busy bit is set, we're just incrementing
            // the count at the bottom.
            fa_status++;
        }
    }

bail_2:
    // Release our copy buffer.
    free(xfer_buffer);
    xfer_buffer = NULL;

bail_1:
    // Return our status information.
    return status;
}

int file_action_thread::do_delete()
{
    int status;
    int path_len;
    std::string file_suffix;
    std::string file_str(file_path);

    // Get the length for ease of use.
    path_len = file_str.length();

    // Get the suffix.
    file_suffix = file_str.substr(path_len - 3, 3);

    // Are we dealing with a AVI, MP4 or CAP file?
    if ((file_suffix == std::string("avi")) || (file_suffix == std::string("mp4")) || (file_suffix == std::string("cap")))
    {
        // Delete the file.
        status = remove(file_str.c_str());
        if (status)
        {
            fi_log(LOG_ERR, "%s: '%s' %d(%s)", __PRETTY_FUNCTION__, file_str.c_str(), errno, strerror(errno));
            goto bail;
        }

        // Adjust the path length be be the location of the '.'.
        path_len = path_len - 3;

        // Delete the CFG file. It's OK if the file isn't there.
        file_str.replace(path_len, 3, std::string("cfg"));
        remove(file_str.c_str());

        // Delete the TXT file. It's OK if the file isn't there.
        file_str.replace(path_len, 3, std::string("txt"));
        remove(file_str.c_str());

        // Delete the PM file. It's OK if the file isn't there.
        file_str.replace(path_len, 3, std::string("pm"));
        remove(file_str.c_str());

        // Delete the XML file. It's OK if the file isn't there.
        file_str.replace(path_len, 2, std::string("xml"));
        remove(file_str.c_str());

        // Delete the "incomplete" file. It's OK if the file isn't there.
        file_str.replace(path_len, 3, file_suffix + FAS_VIDEO::FiVideo::GetIncompleteVideoFileString());
        remove(file_str.c_str());
    }
    else
    {
        // Walk the tree...
        status = nftw(file_path.c_str(), unlink_cb, 64, FTW_DEPTH | FTW_PHYS);
        if (status == -1)
            status = errno;
    }

bail:
    // Return our status information.
    return status;
}

int file_action_thread::do_save_logs()
{
    const char  *media_ptr;
    std::time_t time_now;
    std::string format_str;
    std::string mount_dir;

    // Determine if we were given a destination path.
    media_sel = (file_action & GIGE_IMAGE_DEST_MASK) >> GIGE_DEST_SHIFT;
    if (media_sel == 0)
    {
        // Our destination should already be set.
        if (media_path.empty())
        {
            fi_log(LOG_ERR, "%s: No destination path specified!", __PRETTY_FUNCTION__);
            return EINVAL;
        }
        media_ptr = media_path.c_str();
    }
    else
    {
        std::string device;

        // Build our destination base path.
        storage::get_media_device(media_sel, device);
        if (device.empty())
        {
            fi_log(LOG_ERR, "%s: Invalid destination media value (0x%x)", __PRETTY_FUNCTION__, media_sel);
            return EINVAL;
        }

        // Get the mount point for the device.
        storage::get_mount_dir(device, mount_dir);
        if (mount_dir.empty())
        {
            fi_log(LOG_ERR, "%s: No mount point for '%s'", __PRETTY_FUNCTION__, device.c_str());
            return EINVAL;
        }

        // Point to the mount folder.
        media_ptr = mount_dir.c_str();
    }

    // Get the current TOD to stamp the log file.
    time_now = std::time(nullptr);

    // Build up the command.
    format_str = fmt::format("tar cjf {0}/log_{1:%G_%m%d}.tar.bz /var/log/journal /home/fascam/.config/Fastec/{{settings,*.conf}}", media_ptr, fmt::localtime(time_now));
    system(format_str.c_str());
    sync();

    // Return all is well.
    return 0;
}

void file_action_thread::run()
{
    int status;

    TP(TRACE_SEL_STORAGE, "%s: PID=%u, file_action=0x%x", __PRETTY_FUNCTION__, get_tid(), file_action);

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Initialize things.
    target_path.clear();

    // We're running.
    thread_running = true;

    // Are we supposed to get a listing?
    if (file_action == GIGE_FA_LISTING)
    {
        // Build the tree.
        status = storage::factory()->buildXmlTree();
        if (status != 0)
        {
            // Set the status.
            fa_status = GIGE_FA_STATUS_ERROR | status;
            fi_log(LOG_ERR, "%s: Unable to build the XML tree.", __PRETTY_FUNCTION__);
        }
        else
        {
            // Set our status appropriately.
            fa_status = GIGE_FA_STATUS_IDLE | storage::factory()->get_xml_tree_len();
        }
    }
    else if (file_action == GIGE_FA_LUT_LISTING)
    {
        // Build the tree.
        status = storage::factory()->buildLutXmlTree();
        if (status != 0)
        {
            // Set the status.
            fa_status = GIGE_FA_STATUS_ERROR | status;
            fi_log(LOG_ERR, "%s: Unable to build the LUT XML tree.", __PRETTY_FUNCTION__);
        }
        else
        {
            // Set our status appropriately.
            fa_status = GIGE_FA_STATUS_IDLE | storage::factory()->get_xml_tree_len();
        }
    }
    else
    {
        // Figure out if we're dealing with a file or folder.
        // Don't really care if we're saving logs.
        if (((file_action & GIGE_FA_OP_MASK) != GIGE_FA_SAVE_LOGS) && stat(file_path.c_str(), &file_stat))
        {
            fi_log(LOG_ERR, "%s: Unable to stat path '%s'. %d(%s)", __PRETTY_FUNCTION__, file_path.c_str(), errno, strerror(errno));
            status = errno;
        }
        else
        {
            // Process based on the file action.
            switch (file_action & GIGE_FA_OP_MASK)
            {
            case GIGE_FA_COPY:
                status = do_copy();
                break;
            case GIGE_FA_DELETE:
                status = do_delete();
                break;
            case GIGE_FA_SAVE_LOGS:
                status = do_save_logs();
                break;
            default:
                status = EFAULT;
                break;
            }
        }

        // Setup our status information.
        if (status)
            fa_status = status | GIGE_FA_STATUS_ERROR;
        else
            fa_status = GIGE_FA_STATUS_IDLE;
    }

    // Clean up on the way out.
    cleanup();
}

file_action_thread::file_action_thread() :
    FAS_THREAD::Thread(true),
    file_action(0),
    file_offset(0),
    target_offset(0),
    media_sel(0)
{
    // Initialize things.
    thread_running = false;

    file_path.clear();
    target_path.clear();
    media_path.clear();

//    traceControl |= TRACE_SEL_STORAGE;
}

file_action_thread::~file_action_thread()
{
    TP(TRACE_SEL_STORAGE, "%s", __PRETTY_FUNCTION__);
}

void file_action_thread::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
        return;
    }

    // Start the thread.
    FAS_THREAD::Thread::start(threadFn);
}

void file_action_thread::stop()
{
    // Issue the standard stop request.
    FAS_THREAD::Thread::stop();
}
