/*!
 * @file telemetry.cpp
 *
 * @brief This file contains all the routines to deal with telemetry
 *    data for the camera.
 *
 * Copyright 2018 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#define __USE_XOPEN_EXTENDED 500

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ftw.h>
#include <inttypes.h>
#include <libudev.h>
#include <math.h>
#include <mntent.h>
#include <stddef.h>
#include <unistd.h>

#include <blkid/blkid.h>
#include <linux/types.h>

#include <memory>
#include <mutex>
#include <string>
#include <vector>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/types.h>

#include <fastec/fas_log.h>
#include <fastec/fi_fpga_regs.h>

#include "telemetry.h"

#include "battery.h"
#include "camera_server.h"
#include "fan_proc.h"
#include "fpga.h"
#include "sensor.h"
#include "storage.h"
#include "tmp451.h"

namespace TELEMETRY_LOCAL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// FPGA External Temperature Sensor
#define FPGA_TMP451_I2C_ADDR        0x4C

// Firefly I2C address.
#define FIREFLY_I2C_ADDR        0x50

// Firefly Registers
#define FIREFLY_VOLTAGE         0x25
#define FIREFLY_TEMP            0x26

// EM2120 I2C address.
#define EM2120_I2C_ADDR         0x40

// EM2120 Registers (of interest)
#define EM2120_STATUS_WORD      0x79
#define EM2120_VOUT         0x8B
#define EM2120_IOUT         0x8C
#define EM2120_TEMP         0x8E

// em2120 bit definitions. "Power not good" is bit3
// in the MS byte.
#define EM2120_POWER_NOT_GOOD       (1 << 3)

// The PLX I2C address.
#define PLX_I2C_ADDR        0x5e

// PLX temperature register.
#define PLX_TEMP_ADDR       0xBAC
#define PLX_TEMP_ENABLE     (1 << 7)

//--------------------------------------------------------------------
// Variable Declarations:
//--------------------------------------------------------------------

// Mutexes to protect items used by main thread, fpga event, and select threads.
std::mutex telemetry_mutex;

} /* End of namespace TELEMETRY_LOCAL */

using namespace TELEMETRY_LOCAL;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
std::unique_ptr<telemetry> telemetry::instance;
std::mutex telemetry::instance_mutex;

//--------------------------------------------------------------------
// Class implementation: telemetry
//--------------------------------------------------------------------

telemetry::telemetry()
{
}

/*!
 * @brief Telemetry class destructor.
 */
telemetry::~telemetry()
{
}

/*!
 * @brief Create our singleton.
 */
telemetry *telemetry::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    if (instance.get() == nullptr)
    {
        instance = std::unique_ptr<telemetry>(new telemetry());
    }

    return instance.get();
}

/*!
 * @brief Read values from the em2120.
 */
double telemetry::get_em2120_reading(int reg_addr, bool mantissa_only)
{
    uint8_t reg_value[2];
    int exponent;   // 'N'
    int mantissa;   // 'Y'
    double result;

    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null.", __PRETTY_FUNCTION__);
        return 0;
    }

    if (!fpga_ptr->i2cRead(EM2120_I2C_ADDR, reg_addr, reg_value, FI_FPGA::I2C_MAIN, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0;
    }

    if (mantissa_only)
    {
        // For mantissa only values, the exponent is presumed to be 2^-13.
        exponent = -13;
        mantissa = (((int)reg_value[1]) << 8) + ((int)reg_value[0]);
    }
    else
    {
        // Extract the information.
        exponent = (int)((reg_value[1] >> 3) & 0x1F);
        if ((exponent & 0x10) == 0x10)
        {
            exponent = -16 + (exponent & 0xF);
        }
        mantissa = ((int)(reg_value[1] & 0x7) << 8) + ((int)reg_value[0]);
        if ((mantissa & 0x400) == 0x400)
        {
            mantissa = -1024 + (mantissa & 0x3FF);
        }
    }

    // Our result is Y * 2^N.
    result = ((double)mantissa) * pow(2.0, (double)exponent);
    return result;
}

double telemetry::get_em2120_temperature()
{
    // Synchronize access to this routine.
    std::unique_lock<std::mutex> locker(telemetry_mutex);

    // Make sure we have a telemetry object.
    telemetry *my_object = telemetry::factory();

    // Return the em2120 temperature.
    return my_object->get_em2120_reading(EM2120_TEMP);
}

bool telemetry::get_fan_stats(Json::Value& dev_tree)
{
    FAN_PROC::STATS_REQUEST request;

    // Read the temperature from the fan controller.
    if (!fan_get_stats(&request))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return false;
    }

    dev_tree["temp_fan_internal"] = request.internal_temp_C;
    dev_tree["temp_fan_external"] = request.external_temp_C;
    dev_tree["fan_tach_reading"] = request.tach_reading;
    dev_tree["fan_pwm_setting"] = request.pwm_setting;
    return true;
}

/*!
 * @brief Read the temperature value from the fan controller.
 */
double telemetry::get_fan_temperature()
{
    double result = 0.0;

    // Read the temperature from the fan controller.
    if (!fan_get_temp(&result))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0.0;
    }

    // Return our result.
    return result;
}

/*!
 * @brief Read the temperature value from the Firefly.
 */
double telemetry::get_firefly_temperature()
{
    double ff_temp = 0.0;
    uint32_t nc1_status;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Sanity check.
    if (fpga_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: fpga_ptr is null.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Check to see if we are an integrated or remote head camera. The integrated
    // head camera does not have a firefly, so we will skip the reading and just
    // return a 0.
    nc1_status = 0;
    if (!fpga_ptr->regRead(NC1_FPGA_STATUS, &nc1_status))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read NC1_FPGA_STATUS.", __PRETTY_FUNCTION__);
        return 0;
    }

    // If we using an optical interface to talk with the head, then we want
    // to query the Firefly device for temperature information.
    CameraServer *cs_ptr = CameraServer::get_cs();
    if (cs_ptr && cs_ptr->optical())
    {
        uint8_t temp[2];

        // Synchronize access to this routine.
        std::unique_lock<std::mutex> locker(telemetry_mutex);

        // Issue the write with no data other than the register address.
        if (!fpga_ptr->i2cWrite(FIREFLY_I2C_ADDR, FIREFLY_TEMP, 0, FI_FPGA::I2C_MAIN, 0, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
            return 0;
        }

        // Now issue a read-only command to get the data.
        if (!fpga_ptr->i2cRead(FIREFLY_I2C_ADDR, FIREFLY_TEMP, temp, FI_FPGA::I2C_MAIN, 2, true, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
            return 0;
        }

        // All is well. The high byte is degrees Celsius, the low byte is the decimal value.
        ff_temp = (((double)temp[0]) + (((double)temp[1]) / 256.0));
    }

    // Return the temp value.
    return ff_temp;
}

double telemetry::get_firefly_voltage()
{
    double ff_volt = 0.0;
    uint32_t nc1_status;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Check to see if we are an integrated or remote head camera. The integrated
    // head camera does not have a firefly, so we will skip the reading and just
    // return a 0.
    nc1_status = 0;
    if (!fpga_ptr->regRead(NC1_FPGA_STATUS, &nc1_status))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read NC1_FPGA_STATUS.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Are we a remote head camera?
    CameraServer *cs_ptr = CameraServer::get_cs();
    if (cs_ptr && cs_ptr->optical())
    {
        uint8_t volt[2];

        // Issue the write with no data other than the register address.
        if (!fpga_ptr->i2cWrite(FIREFLY_I2C_ADDR, FIREFLY_VOLTAGE, 0, FI_FPGA::I2C_MAIN, 0, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
            return 0;
        }

        // Now issue a read-only command to get the data.
        if (!fpga_ptr->i2cRead(FIREFLY_I2C_ADDR, FIREFLY_VOLTAGE, volt, FI_FPGA::I2C_MAIN, 2, true, 1))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
            return 0;
        }

        // All is well. 16-bits of voltage.
        ff_volt = ((((double)volt[0] * 256.0) + ((double)volt[1])) / 1000.0);
    }

    // Return our voltage.
    return ff_volt;
}

double telemetry::get_fpga_internal_temp()
{
    int i;
    uint32_t reg_value;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    double temp_c;

    // Tell the FPGA to start a conversion.
    fpga_ptr->regWrite(TEMP_SENSOR_CONTROL_STATUS, TEMP_SENSOR_CONVERT);

    // Read the value.
    for (i = 0; i < 100; i++)
    {
        struct timespec sleep_req;  // Delay period.

        fpga_ptr->regRead(TEMP_SENSOR_CONTROL_STATUS, &reg_value);
        if ((reg_value & TEMP_SENSOR_BUSY) == 0)
            break;

        // Give the shutter a chance to react.
        sleep_req.tv_sec = 0;
        sleep_req.tv_nsec = 1 * 1000 * 1000;
        nanosleep(&sleep_req, NULL);
    }

    // Sanity check
    if ((reg_value & TEMP_SENSOR_BUSY) == TEMP_SENSOR_BUSY)
    {
        // Conversion didn't finish.
        return 0.0;
    }

    // Get degrees C from the value.
    temp_c = TEMP_SENSOR_TEMPOUT(reg_value);
    return temp_c;
}

/*!
 * @brief Read the temperature sensor for the FPGA.
 */
double telemetry::get_fpga_temperature()
{
    // Synchronize access to this routine.
    std::unique_lock<std::mutex> locker(telemetry_mutex);

    uint8_t oldh, newh, lo_byte;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    /*
     * There is a trick here. We have to read two registers to have the
     * sensor temperature, but we have to beware a conversion could occur
     * between the readings. The datasheet says we should either use
     * the one-shot conversion register, which we don't want to do
     * (disables hardware monitoring) or monitor the busy bit, which is
     * impossible (we can't read the values and monitor that bit at the
     * exact same time). So the solution used here is to read the high
     * byte once, then the low byte, then the high byte again. If the new
     * high byte matches the old one, then we have a valid reading. Else
     * we have to read the low byte again, and now we believe we have a
     * correct reading.
     */
    oldh = lo_byte = newh = 0;
    if (!fpga_ptr->i2cRead(FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPH, &oldh) ||
        !fpga_ptr->i2cRead(FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPL, &lo_byte) ||
        !fpga_ptr->i2cRead(FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPH, &newh))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0;
    }
    if (oldh != newh)
    {
        if (!fpga_ptr->i2cRead(FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPL, &lo_byte))
        {
            fi_log(LOG_DEBUG, "%s: Error doing I2C on lo_byte!", __PRETTY_FUNCTION__);
            return 0;
        }
    }

    // All is well. The high byte is degrees Celsius, the low byte is the decimal value.
    return (((double)newh) + ((double)(lo_byte >> 4) / 16.0));
}

/*!
 * @brief Read the temperature in the PLX device.
 */
double telemetry::get_plx_temperature()
{
    // Synchronize access to this routine.
    std::unique_lock<std::mutex> locker(telemetry_mutex);

    FiFPGA *fpga_ptr = CameraServer::get_fpga();
    uint8_t rd_val[4];
    uint32_t cmd_data;
    uint32_t temp_val;
    double temperature;

    // Build up the command data value to write to the device.
    cmd_data = 0x03003C00 | (PLX_TEMP_ADDR >> 2);

    // Write the temperature enable bit. For PLX that's 4 bytes of data and 4 bytes of command.
    if (!fpga_ptr->i2cWrite(PLX_I2C_ADDR, cmd_data, PLX_TEMP_ENABLE, FI_FPGA::I2C_MAIN, 4, 4))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0.0;
    }

    // Wait for the PLX to sample the temperature sensor.

    // Build up the command data value to read from the device.
    cmd_data = 0x04003C00 | (PLX_TEMP_ADDR >> 2);

    // Write the read command. There's no data, but 4 bytes of command.
    if (!fpga_ptr->i2cWrite(PLX_I2C_ADDR, cmd_data, 0, FI_FPGA::I2C_MAIN, 0, 4))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0.0;
    }

    // Now read the temperature.
    if (!fpga_ptr->i2cRead(PLX_I2C_ADDR, 0, rd_val, FI_FPGA::I2C_MAIN, 4, true, 0))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C read!", __PRETTY_FUNCTION__);
        return 0.0;
    }

    // Build up the command data value to write to the device.
    cmd_data = 0x03003C00 | (PLX_TEMP_ADDR >> 2);

    // Disable the temperature sensor.
    if (!fpga_ptr->i2cWrite(PLX_I2C_ADDR, cmd_data, 0, FI_FPGA::I2C_MAIN, 4, 4))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0.0;
    }

    temp_val = (uint32_t)rd_val[1] | ((uint32_t)(rd_val[0] & 3) << 8);
    temperature = (-4.5646e-11 * pow(temp_val, 4)) + (1.4331e-7 * pow(temp_val, 3)) + (-2.3557e-4 * pow(temp_val, 2)) + (0.32597 * temp_val) + (-53.509);

    // Return the temperature.
    return temperature;
}

/*!
 * @brief Read status from the em2120.
 */
bool telemetry::get_power_good()
{
    uint8_t reg_value[2];

    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    if (!fpga_ptr->i2cRead(EM2120_I2C_ADDR, EM2120_STATUS_WORD, reg_value, FI_FPGA::I2C_MAIN, 2))
    {
        fi_log(LOG_DEBUG, "%s: Error doing I2C!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Extract the information.
    return ((reg_value[1] & EM2120_POWER_NOT_GOOD) == EM2120_POWER_NOT_GOOD);
}

uint32_t telemetry::get_json(char *dest_ptr, uint32_t max_len)
{
    uint32_t reg_value;
    FiFPGA *fpga_ptr;
    Json::Value dev_tree;
    Json::StreamWriterBuilder wr_builder;
    std::string json_message;
    Sensor *sensor_ptr = CameraServer::get_sensor();
    fpga_ptr = CameraServer::get_fpga();

    // Make sure we have a telemetry object.
    telemetry *my_object = telemetry::factory();

    // Clear out the destination.
    memset(dest_ptr, 0, max_len);

    // -------------------------
    // Magic happens here.
    // -------------------------
    {
        // Test until we hook to real values.
        // Update our random temps.
        dev_tree["vext1"] = 0.0;    // ALM_TODO
        dev_tree["vcc"] = 0.0;      // ALM_TODO
        dev_tree["vccp"] = 0.0;     // ALM_TODO
        dev_tree["vccpt"] = 0.0;    // ALM_TODO
        dev_tree["vcceram"] = 0.0;  // ALM_TODO
        dev_tree["adcgnd"] = 0.0;   // ALM_TODO

#if 0
        double new_temp;
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vext0"] = new_temp;   // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vext1"] = new_temp;   // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vcc"] = new_temp; // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vccp"] = new_temp;    // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vccpt"] = new_temp;   // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["vcceram"] = new_temp; // ALM_TODO
        new_temp = (double)(rand() % 330) / 100.0;
        dev_tree["adcgnd"] = new_temp;  // ALM_TODO
#endif

        // Form test until we hook to real values.
        // -------------------------------------------

        // Read our status information.
        fpga_ptr->regRead(NC1_FPGA_STATUS, &reg_value);
        dev_tree["sensor_pwr_ok"] = ((reg_value & NC1FS_ISB_PWR_OK) != NC1FS_ISB_PWR_OK);
        dev_tree["pmbus_alert"] = my_object->get_power_good();

        // Get the temperatures and voltages.
        dev_tree["vext0"] = batteryVoltage();
        dev_tree["temp_a10_int"] = get_fpga_internal_temp();
        dev_tree["temp_a10_ext"] = get_fpga_temperature();
        dev_tree["temp_firefly"] = my_object->get_firefly_temperature();
        dev_tree["firefly_V"] = my_object->get_firefly_voltage();
        dev_tree["temp_sensor"] = sensor_ptr->get_temperature();
        dev_tree["temp_nvme"] = (double)get_nvme_temp();
        dev_tree["temp_em2120"] = my_object->get_em2120_reading(EM2120_TEMP);
        dev_tree["em2120_I"] = my_object->get_em2120_reading(EM2120_IOUT);
        dev_tree["em2120_V"] = my_object->get_em2120_reading(EM2120_VOUT, true);
        dev_tree["temp_pcie"] = get_plx_temperature();

        my_object->get_fan_stats(dev_tree);
    }
    // fi_log(LOG_DEBUG, "%s: json='%s'", __PRETTY_FUNCTION__, dev_tree.toStyledString().c_str());
    //
    // -------------------------
    // Magic ends here.
    // -------------------------

    // Create a string.
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";
    wr_builder.settings_["precision"] = 5;
    wr_builder.settings_["precisionType"] = "decimal";
    json_message = Json::writeString(wr_builder, dev_tree);

    // Make sure we didn't get crazy.
    if (json_message.length() >= max_len)
    {
        fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
        return 0;
    }

    // Copy the string over.
    // fi_log(LOG_DEBUG, "%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
    memcpy(dest_ptr, json_message.c_str(), json_message.length());

    // Return the length, including the null-terminator.
    return json_message.length() + 1;
}

int telemetry::set_adc_channel(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;
    (void)reg_value;

    // ================================================================
    // TODO TODO TODO
    // Tell the FPGA to start an ADC conversion cycle.
    // ================================================================

    return GEV_STATUS_SUCCESS;
}

int telemetry::get_nvme_temp()
{
    int degrees_C = 0;
    FILE *fp;
    std::string device;
    std::string cmd_str;

    // Make sure we have an SSD in our system.
    storage::get_media_device(GIGE_MEDIA_SSD, device);
    if (device.empty())
        return 0;

    // Build up the command.
    cmd_str = fmt::format("smartctl --attributes {0}  | grep 'Temperature Sensor 1:'", device);

    // Have the system run our command.
    fp = popen(cmd_str.c_str(), "r");
    if (fp != NULL)
    {
        char temp_buf[ PATH_MAX ];

        // Read data, if any.
        while (fgets(temp_buf, MAX_BUFFER_LEN, fp) != NULL)
        {
            // Did we get anything?, we need the checksum so that it can be written
            if (temp_buf[0] != 0)
                break;
        }

        // Close the pipe.
        pclose(fp);

        // If we got a checksum, put it into the checksum file.
        if (temp_buf[0] != 0)
        {
            std::string work_str(temp_buf);
            size_t index;

            // Find the location of "Celsius" and then 3 characters
            // back will be the value, as in "23 Celsius".
            index = work_str.find("Celsius") - 3;
            if (index > work_str.length())
            {
                fi_log(LOG_WARNING, "%s: 'Celsius' not found in '%s'", __PRETTY_FUNCTION__, temp_buf);
            }
            else
            {
                // Convert the value to an integer.
                degrees_C = std::stoi(work_str.substr(index, 2));
            }
        }
        else
        {
            fi_log(LOG_WARNING, "%s: smartctl returned nothing.", __PRETTY_FUNCTION__);
        }
    }
    else
    {
        fi_log(LOG_WARNING, "%s: Cannot run smartctl.", __PRETTY_FUNCTION__);
    }

    // Return our result.
    return degrees_C;
}
