/*!
 * @file udp.cxx
 * @brief This file contains the implementation of the cs_udp class.
 *
 * This file contains all the routines to deal with UDP
 * network traffice to allow the CameraServer to communiate with
 * remote devices.
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>

#include <fastec/fas_log.h>

#include "cs_network.h"

namespace CS_NETWORK
{
//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
static NETWORK_INFO local_net_info;     //!< Saved info about our loopback network.

} /* End of namespace CS_NETWORK */

using namespace CS_NETWORK;

/*!
 * @brief The constructor for the cs_udp class.
 */
CsUDP::CsUDP() :
    CsNetwork()
{
}

/*!
 * @brief Initialize UDP for the daemon. Errors should be be considered
 *    fatal because without other processes being able to communicate
 *    with the camera server we can't do anything.
 *
 * TODO - Need to handle the case where the network isn't up yet and
 *    to get a signal later.
 *
 * @return CS_NET_SUCCESS if successful, negative numbers are errors
 *     and should be considered fatal.
 */
int CsUDP::setup()
{
    // Sanity check.
    if (isValid())
    {
        fi_log(LOG_DEBUG, "%s: socket (%d) seems to be having trouble.", __PRETTY_FUNCTION__, sock->getFD());

        // Close the socket.
        close();
    }

    // Create the handle for the FCP UDP port. It's an "internet"
    // DATAGRAM socket (UDP is always datagram).
    sock = std::make_unique<CsSocket>(AF_INET, SOCK_DGRAM);
    if (!isValid())
    {
        fi_log(LOG_ERR, "%s: Error creating UDP FCP socket", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_SOCKET;
    }

    // Create the binding for the FCP port.
    if (!sock->bind(INADDR_ANY, GIGE_SERVER_UDP_PORT))
    {
        fi_log(LOG_ERR, "%s: Error binding to UDP FCP socket using port %d", __PRETTY_FUNCTION__, GIGE_SERVER_UDP_PORT);
        close();
        return CS_NET_ERROR_BIND;
    }

    // Success!
    return CS_NET_SUCCESS;
}

/*!
 * @brief Close down the UDP socket to make sure everything is cleaned up
 *    before shutting down.
 */
void CsUDP::close()
{
    LOG("%s", __PRETTY_FUNCTION__);
    CsNetwork::close();
}

/*!
 * @brief Get the next message from any remote client trying to send us a request.
 *    Since we are using datagrams the client is not waiting.
 *
 * @param p_msg Pointer to where we'll put the received message, if any.
 *
 * @return The size of the message if successful, 0 if no message, <0 if error.
 */
int CsUDP::getMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     fromlen;
    ssize_t msg_size;
    char    *p_net_addr;
    struct sockaddr_in  from;

    // Sanity check.
    if (!isValid())
    {
        LOG("%s: socket_fd < 0", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_FD;
    }
    if (p_msg == nullptr)
    {
        LOG("%s: p_msg is null!", __PRETTY_FUNCTION__);
        return CS_NET_INVALID_PARAM;
    }

    // Clear our receive structures.
    memset(&from, 0, sizeof(from));
    memset(p_msg, 0, sizeof(CS_NETWORK::CS_NETWORK_MESSAGE));

    // Initialize the size of our structure.
    fromlen = sizeof(from);

    // Receive a message from the socket.
    msg_size = recvfrom(sock->getFD(), (char *)&p_msg->message, sizeof(p_msg->message), 0,
                        (struct sockaddr *)&from, (socklen_t *)&fromlen);
    if (msg_size == -1)
    {
        fi_log(LOG_ERR, "%s: Cannot read from UDP client, errno:%d(%s)",
               __PRETTY_FUNCTION__, errno, strerror(errno));
        return CS_NET_ERROR_RECVFROM;
    }

    // Convert the address of the client into ASCII notation so we can
    // more easily manage it. Save the IP address in our message.
    p_net_addr = (char *)inet_ntoa(from.sin_addr);
    strncpy(p_msg->udp_msg.from, p_net_addr, NET_ADDR_LEN);

    // Remember the port the client sent from.
    p_msg->udp_msg.from_port = from.sin_port;

    // Timestamp when we received the message.
    clock_gettime(CLOCK_REALTIME, &p_msg->rcv_time);

// #define UDP_DEBUG 1
#ifdef UDP_DEBUG
    uint16_t cmd = ntohs(p_msg->message.header.req.command);
    uint16_t len = ntohs(p_msg->message.header.req.length);
    uint16_t id = ntohs(p_msg->message.header.req.req_id);

    fi_log(LOG_INFO, "-- UDP message received from IP:%s length:%d rcv:%ld:%ld", p_net_addr, msg_size, p_msg->rcv_time.tv_sec, p_msg->rcv_time.tv_nsec);
    fi_log(LOG_INFO, "-- hdr: cmd=0x%04x, len=0x%04x, req_id=0x%04x", cmd, len, id);
#endif

    // This was a UDP message, after all.
    p_msg->msg_type = CS_NETWORK::MSG_TYPE_UDP;

    // Return the length of the message.
    return msg_size;
}

/*!
 * @brief Send a response back to the remote client for a message request
 *    we have already received.
 *
 * @param p_msg     Message which is destined for the client.
 * @param num_bytes The length of the message to transmit.
 *
 * @return The size of the message if successful, 0 if no message, <0 if error.
 */
int CsUDP::sendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t num_bytes)
{
    int     tolen;
    ssize_t msg_size = 0;
    struct sockaddr_in  to;         // UNIX domain socket structure

// #define UDP_SEND_DEBUG 1
#ifdef UDP_SEND_DEBUG
    uint16_t status = ntohs(p_msg->message.header.resp.status);
    uint16_t cmd = ntohs(p_msg->message.header.resp.command);
    uint16_t len = ntohs(p_msg->message.header.resp.length);
    uint16_t id = ntohs(p_msg->message.header.resp.ack_id);

    LOG("%s", __PRETTY_FUNCTION__);
    fi_log(LOG_INFO, "-- UDP message send to IP:%s length:%d", p_msg->udp_msg.from, num_bytes);
    fi_log(LOG_INFO, "-- hdr: status=0x%04x cmd=0x%04x, len=0x%04x, req_id=0x%04x", status, cmd, len, id);
#endif

    // Sanity check.
    if (!isValid())
        return CS_NET_ERROR_FD;
    if (p_msg == NULL)
        return CS_NET_INVALID_PARAM;

    // Populate the socket.
    memset(&to, 0, sizeof(to));
    to.sin_family = AF_INET;
    inet_aton(p_msg->udp_msg.from, &to.sin_addr);
    to.sin_port = p_msg->udp_msg.from_port;

    // Send the message to the socket.
    tolen = sizeof(to);
    msg_size = ::sendto(sock->getFD(), (char *)&p_msg->message, num_bytes, 0,
                        (struct sockaddr *)&to, tolen);
    if (msg_size == -1)
    {
        fi_log(LOG_ERR, "%s: Cannot send to client:0x%x(%s) port:%d len:%d errno:%d(%s)",
               __PRETTY_FUNCTION__, to.sin_addr.s_addr, p_msg->udp_msg.from, ntohs(to.sin_port), num_bytes,
               errno, strerror(errno));
        return CS_NET_ERROR_SENDTO;
    }

    // Return the size of the message sent.
    return msg_size;
}

//=============================================================================
// cs_lo: The loopback network class.
//=============================================================================

/*!
 * @brief The constructor for the cs_lo class.
 */
CsLocal::CsLocal() :
    CsUDP()
{
}

/*!
 * @brief Close down the loopback socket to make sure everything is cleaned up
 *    before shutting down.
 */
void CsLocal::close()
{
    LOG("%s", __PRETTY_FUNCTION__);
    CsUDP::close();
}

/*!
 * @brief Initialize loopback socket. Errors should be be considered
 *    fatal because without other processes being able to communicate
 *    with the camera server we can't do anything.
 *
 * @return CS_NET_SUCCESS if successful, negative numbers are errors
 *     and should be considered fatal.
 */
int CsLocal::setup()
{
    int rc;
    struct ifreq ifr;       // Interface request structure we use for socket ioctls.

    // Get the basic stuff setup.
    if ((rc = CsUDP::setup()) != CS_NET_SUCCESS)
    {
        fi_log(LOG_ERR, "%s: Unable to setup the local socket", __PRETTY_FUNCTION__);
        return rc;
    }

    // Clear out our structure.
    memset(&ifr, 0, sizeof(ifr));

    // Now find out information about this network.
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, "lo", IFNAMSIZ - 1);

    // Bind the socket as appropriate.
    if (setsockopt(sock->getFD(), SOL_SOCKET, SO_BINDTODEVICE, (void *)&ifr, sizeof(ifr)) < 0)
    {
        fi_log(LOG_ERR, "%s: Unable to bind, errno:%d(%s)", __PRETTY_FUNCTION__, errno, strerror(errno));
        close();
        return CS_NET_ERROR_SOCKET;
    }

    // All done.
    return CS_NET_SUCCESS;
}

/*!
 * @brief Discover information about the network,
 *
 * The information is stored in a static variable to be returned later
 * as requested.
 *
 * @returns on success negative numbers on error
 */
//
int CsLocal::discoverNetwork()
{
    int i;              // Generic loop variable.
    struct ifreq    ifr;        // Interface request structure we use for socket ioctls.
    struct if_nameindex *ifni;      // Pointer to the network interface index/name structure.
    std::string dev_name;

    // Sanity check.
    if (!isValid())
    {
        // Perfectly OK if we don't have a network.
        return CS_NET_ERROR_FD;
    }

    // Get the name of the network interfaces.
    CsNetwork::getDevName(LOCAL, dev_name);

    // Clear out our network information structure.
    memset(&local_net_info, 0, sizeof(local_net_info));

    // Find out the interfaces which are available.
    ifni = if_nameindex();
    if (ifni == nullptr)
    {
        fi_log(LOG_ERR, "%s: Couldn't discover network interfaces", __PRETTY_FUNCTION__);
        return CS_NET_ERROR_FD;
    }

    for (i = 0; ifni[i].if_index; i++)
    {
        // Is this the loopback interface?
        if (dev_name.compare(ifni[i].if_name) == 0)
        {
            // Yup, so save it away.
            strncpy(local_net_info.if_name, ifni[i].if_name, sizeof(local_net_info.if_name));

            // Clear out our structure.
            memset(&ifr, 0, sizeof(ifr));

            // Now find out information about this network.
            ifr.ifr_addr.sa_family = AF_INET;
            strncpy(ifr.ifr_name, local_net_info.if_name, IFNAMSIZ - 1);

            // No MAC address for the local loopback interface.

            // Get the IP address
            refreshIpAddr(&local_net_info);

            // Subnet Mask
            refreshSubnet(&local_net_info);

            // Find out our default gateway
            local_net_info.def_gateway.s_addr = findGateway(&local_net_info);

            // All done.
            break;
        }
    }

    // We want the UDP port to be bound to the hard wired
    // network if it exists.
    if (local_net_info.if_name[0] != '\0')
    {
        // Clear out our structure.
        memset(&ifr, 0, sizeof(ifr));
        strncpy(ifr.ifr_name, local_net_info.if_name, IFNAMSIZ - 1);

        // Bind the socket as appropriate.
        if (setsockopt(sock->getFD(), SOL_SOCKET, SO_BINDTODEVICE, (void *)&ifr, sizeof(ifr)) < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to bind to '%s', errno:%d(%s)",
                   __PRETTY_FUNCTION__, local_net_info.if_name, errno, strerror(errno));
            close();
            return CS_NET_ERROR_SOCKET;
        }
    }

    // Release the interface record(s).
    if_freenameindex(ifni);

    // Return all is well.
    return CS_NET_SUCCESS;
}

/*!
 *
 * @brief Get the default gateway for the first interface
 *
 * @return The default gateweay.
 */
uint32_t CsLocal::getDefGateway()
{
    uint32_t gateway = 0;

    if (local_net_info.if_name[0] != '\0')
    {
        gateway = (uint32_t)local_net_info.def_gateway.s_addr;
    }

    return gateway;
}

/*!
 *
 * @brief Get the IP address for the loopback interface.
 */
uint32_t CsLocal::getIpAddress()
{
    uint32_t ip_addr = 0;

    if (local_net_info.if_name[0] != '\0')
    {
        // Everything is fine. Return the IP address.
        ip_addr = (uint32_t)local_net_info.ip_addr.s_addr;
    }

    return ip_addr;
}

/*!
 *
 * @brief Get the subnet mask for the loopback interface
 *
 * @return the subnet mask as a binary
 */
uint32_t CsLocal::getSubnetMask()
{
    uint32_t subnet = 0;

    // Always give preference to the eth0 interface!
    if (local_net_info.if_name[0] != '\0')
    {
        subnet = (uint32_t)local_net_info.subnet_mask.s_addr;
    }

    return subnet;
}

