/*!
 * @file wb_file.cxx
 *
 * @brief Functions to handle loading white balance coefficients from a file.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>

#include <string>
#include <fmt/printf.h>

#include <fastec/fi_ioctl.h>
#include <fastec/fas_log.h>

#include "camera_server.h"
#include "sensor.h"
#include "wb_file.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#define EPSILON 0.00001     //!< Don't divide by a float smaller than this.
#define AS_SHOT_MAX 15.0    //!< Max value if dividing by EPSILON.

#define NUM_FLOATS   18     //!< Max number of floats we load for each color temperature.
#define NUM_LOADABLE 3      //!< Number of loadable color temperature records.

//--------------------------------------------------------------------
// Variable Declarations:
//--------------------------------------------------------------------

// Initialize our array with place holders. The actual data values will be loaded based on sensor
// type information.
static WBAL_GAINS wb_gains[] =
{
    /* Daylight */  { 0x1000, 0x1000, 0x1000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x10000000, 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000 },
    /* Tungsten */  { 0x1000, 0x1000, 0x1000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x10000000, 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000 },
    /* Fluorescent */   { 0x1000, 0x1000, 0x1000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x10000000, 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000 },
    /* Normal */    { 0x1000, 0x1000, 0x1000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x10000000, 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000 },
    /* Custom */    { 0x1000, 0x1000, 0x1000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x0000, 0x0000, 0x0000, 0x1000, 0x10000000, 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000 },
};

static struct WBFILE
{
    const char *name;
    int tag;
    int8_t status;
    double float_vals[NUM_FLOATS];
} wbfile[ NUM_LOADABLE ] =
{
    { "Daylight",    GIGE_WHITE_BAL_DAYLIGHT,    0, {0, }},
    { "Tungsten",    GIGE_WHITE_BAL_TUNGSTEN,    0, {0, }},
    { "Fluorescent", GIGE_WHITE_BAL_FLOURESCENT, 0, {0, }}
};

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

static uint16_t float_to_4_12(double fval);
static uint32_t float_to_4_28(double fval);
static int read_wb_settings(FILE *fd, bool as_shot, char *buf, int bufsize, int idx, int *p_lineno);
static int convert_to_hex(int idx, bool as_shot);

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

static uint16_t float_to_4_12(double fval)
{
    uint32_t rc;

    int     int_4x12;
    double   float_4x12;

    // Convert from float to 4.12 format.
    float_4x12 = (fval * 4096.0);
    int_4x12 = (int)floor(float_4x12 + 0.5);
    rc = (uint32_t)(int_4x12);
    // LOG("%s: fval=%lf, float_4x12=%lf, int_4x12=%d, rc=0x%04x", __PRETTY_FUNCTION__, fval, float_4x12, int_4x12, rc & 0xFFFF);

    return (uint16_t)(rc & 0xFFFF);
}

static uint32_t float_to_4_28(double fval)
{
    uint32_t value;
    uint32_t fractional;

    value = floor(fval);
    fractional = (uint32_t)floor(((fval - ((double)value)) * pow(2, 28)) + 0.5);
    // LOG("%s: fval=%lf, value=%d, fractional=%d", __PRETTY_FUNCTION__, fval, value, fractional);

    return (value << AS_SHOT_SHIFT) | fractional;
}

static int read_wb_settings(FILE *fd, bool as_shot, char *buf, int bufsize, int idx, int *p_lineno)
{
    int got;

    do
    {
        if (fgets(buf, bufsize, fd) == NULL)
        {
            fi_log(LOG_ERR, "Line %d: no gains for %s\n", *p_lineno, wbfile[idx].name);
            return -1;
        }
        ++*p_lineno;
    } while (buf[0] == '#');

    got = sscanf(buf, "%lf %lf %lf", &wbfile[idx].float_vals[0], &wbfile[idx].float_vals[1],
                 &wbfile[idx].float_vals[2]);
    if (got != 3)
    {
        fi_log(LOG_ERR, "Line %d (%s): didn't find 3 floats for R G B gains, only %d:\n",
               *p_lineno, wbfile[idx].name, got);
        return -1;
    }

    do
    {
        if (fgets(buf, bufsize, fd) == NULL)
        {
            fi_log(LOG_ERR, "Line %d: ran out of data for %s\n", *p_lineno, wbfile[idx].name);
            return -1;
        }
        ++*p_lineno;
    } while (buf[0] == '#');
    got = sscanf(buf, "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
                 &wbfile[idx].float_vals[3],
                 &wbfile[idx].float_vals[4],
                 &wbfile[idx].float_vals[5],
                 &wbfile[idx].float_vals[6],
                 &wbfile[idx].float_vals[7],
                 &wbfile[idx].float_vals[8],
                 &wbfile[idx].float_vals[9],
                 &wbfile[idx].float_vals[10],
                 &wbfile[idx].float_vals[11]);
    if (got != 9)
    {
        fprintf(stderr, "Line %d: didn't find 3 floats for matrix, only %d\n",
                *p_lineno, got);
        return -1;
    }

    do
    {
        if (fgets(buf, bufsize, fd) == NULL)
        {
            fi_log(LOG_ERR, "Line %d: no dc-offset for %s\n", *p_lineno, wbfile[idx].name);
            return -1;
        }
        ++*p_lineno;
    } while (buf[0] == '#');

    got = sscanf(buf, "%lf %lf %lf", &wbfile[idx].float_vals[12], &wbfile[idx].float_vals[13],
                 &wbfile[idx].float_vals[14]);
    if (got != 3)
    {
        fi_log(LOG_ERR, "Line %d (%s): didn't find 3 floats for R G B dc-offsets, only %d:\n",
               *p_lineno, wbfile[idx].name, got);
        return -1;
    }

    if (as_shot)
    {
        do
        {
            if (fgets(buf, bufsize, fd) == NULL)
            {
                fi_log(LOG_ERR, "Line %d: no as-shot for %s\n", *p_lineno, wbfile[idx].name);
                return -1;
            }
            ++*p_lineno;
        } while (buf[0] == '#');

        got = sscanf(buf, "%lf %lf %lf", &wbfile[idx].float_vals[15], &wbfile[idx].float_vals[16],
                     &wbfile[idx].float_vals[17]);
        if (got != 3)
        {
            fi_log(LOG_ERR, "Line %d (%s): didn't find 3 floats for as-shot, only %d:\n",
                   *p_lineno, wbfile[idx].name, got);
            return -1;
        }
    }

    // Mark that we got data.
    wbfile[idx].status = 1;
    return 0;
}

static int convert_to_hex(int idx, bool as_shot)
{
    WBAL_GAINS *pwb = &wb_gains[idx];

    // Color gains
    pwb->red   = float_to_4_12(wbfile[idx].float_vals[0]);
    pwb->green = float_to_4_12(wbfile[idx].float_vals[1]);
    pwb->blue  = float_to_4_12(wbfile[idx].float_vals[2]);

    pwb->matrix_red1 = float_to_4_12(wbfile[idx].float_vals[3]);
    pwb->matrix_red2 = float_to_4_12(wbfile[idx].float_vals[4]);
    pwb->matrix_red3 = float_to_4_12(wbfile[idx].float_vals[5]);
    pwb->matrix_grn1 = float_to_4_12(wbfile[idx].float_vals[6]);
    pwb->matrix_grn2 = float_to_4_12(wbfile[idx].float_vals[7]);
    pwb->matrix_grn3 = float_to_4_12(wbfile[idx].float_vals[8]);
    pwb->matrix_blu1 = float_to_4_12(wbfile[idx].float_vals[9]);
    pwb->matrix_blu2 = float_to_4_12(wbfile[idx].float_vals[10]);
    pwb->matrix_blu3 = float_to_4_12(wbfile[idx].float_vals[11]);

    pwb->dc_offset_red = float_to_4_12(wbfile[idx].float_vals[12]);
    pwb->dc_offset_grn = float_to_4_12(wbfile[idx].float_vals[13]);
    pwb->dc_offset_blu = float_to_4_12(wbfile[idx].float_vals[14]);

    if (as_shot)
    {
        // We were given the AS_SHOT_NEUTRAL values in the file.
        pwb->as_shot_red = float_to_4_28(wbfile[idx].float_vals[15]);
        pwb->as_shot_grn = float_to_4_28(wbfile[idx].float_vals[16]);
        pwb->as_shot_blu = float_to_4_28(wbfile[idx].float_vals[17]);

        PDEBUG(PDEBUG_LOW, "%s: as_shot 0x%08x 0x%08x %08x", __FUNCTION__, pwb->as_shot_red, pwb->as_shot_grn, pwb->as_shot_blu);
    }
    else
    {
        // The "As Shot" values are the multiplicative inverse of the gains.
        if (fabs(wbfile[idx].float_vals[0]) > EPSILON)
            pwb->as_shot_red = float_to_4_28(1.0 / wbfile[idx].float_vals[0]);
        else
            pwb->as_shot_red = float_to_4_28(AS_SHOT_MAX);

        if (fabs(wbfile[idx].float_vals[1]) > EPSILON)
            pwb->as_shot_grn = float_to_4_28(1.0 / wbfile[idx].float_vals[1]);
        else
            pwb->as_shot_grn = float_to_4_28(AS_SHOT_MAX);

        if (fabs(wbfile[idx].float_vals[2]) > EPSILON)
            pwb->as_shot_blu = float_to_4_28(1.0 / wbfile[idx].float_vals[2]);
        else
            pwb->as_shot_blu = float_to_4_28(AS_SHOT_MAX);
    }

    return 0;
}

int load_wb_file(Sensor *sensor_ptr)
{
    FILE     *fd;
    uint32_t sensor_type;
    int      linenum;
    int      bufsize = PATH_MAX + 1;
    int      i, j;
    int      rc = -1;
    std::string buf;
    std::string filename;

    if (! sensor_ptr->is_sensor_color())
        return 0;

    // Initialize local data structure.
    for (i = 0; i < NUM_LOADABLE; ++i)
    {
        wbfile[i].status = 0;
        for (j = 0; j < NUM_FLOATS; ++j)
            wbfile[i].float_vals[j] = 0.0;
    }

    // Construct the filename for this sensor.
    sensor_type = sensor_ptr->get_sensor_type();
    filename = fmt::sprintf("%c%c%c%c_wb.txt", (sensor_type >> 24) & 0xff,
                            (sensor_type >> 16) & 0xff, (sensor_type >> 8) & 0xff, sensor_type & 0xff);

    // Look for the file on the scratch file system first. This is where they would
    // be if we're experimenting with new values.
    buf = fmt::sprintf("%s/.config/Fastec/scratch/%s", get_home_path(), filename);
    fd = fopen(buf.c_str(), "r");
    if (fd != NULL)
        fi_log(LOG_DEBUG, "%s: loading wb from %s", __FUNCTION__, buf.c_str());
    if (fd == NULL)
    {
        // Then look for the file in the settings folder. Historically they have gone
        // here.
        buf = fmt::sprintf("%s/.config/Fastec/settings/%s", get_home_path(), filename);
        fd = fopen(buf.c_str(), "r");
        if (fd != NULL)
            fi_log(LOG_DEBUG, "%s: loading wb from %s", __FUNCTION__, buf.c_str());
    }
    if (fd == NULL)
    {
        // If not found, look for the file in the calibration folder. They really should
        // be here.
        buf = fmt::sprintf("%s/.config/Fastec/calibration/%s", get_home_path(), filename);
        fd = fopen(buf.c_str(), "r");
        if (fd != NULL)
            fi_log(LOG_DEBUG, "%s: loading wb from %s", __FUNCTION__, buf.c_str());
    }
    if (fd == NULL)
    {
        // If not found, look for the file on the fastec filesystem.
        buf = fmt::sprintf("%s/.config/Fastec/%s", get_home_path(), filename);
        fd = fopen(buf.c_str(), "r");
        if (fd != NULL)
            fi_log(LOG_DEBUG, "%s: loading wb from %s", __FUNCTION__, buf.c_str());
    }

    // Could not find it anywhere. We'll use the compiled-in defaults.
    if (fd == NULL)
    {
        fi_log(LOG_ERR, "Can't open WB file %s: %s\n", filename.c_str(), strerror(errno));
        goto bail;
    }

    // Read and process the file.
    rc = 0;
    linenum = 0;
    for (;;)
    {
        std::unique_ptr<char[]> line_buf = std::unique_ptr<char[]>(new char[bufsize + 1]());

        if (fgets(line_buf.get(), bufsize, fd) == NULL)
        {
            if (!feof(fd) && ferror(fd))
                fi_log(LOG_ERR, "%s: read failed: %s\n", __FUNCTION__, strerror(errno));
            break;
        }

        ++linenum;

        // Check for a comment.
        if ((line_buf[0] == '#') || (line_buf[0] == '\n') || (line_buf[0] == '\0'))
            continue;

        // Look for one of the white balance "tags".
        for (i = 0; i < NUM_LOADABLE; ++i)
        {
            // Break out of the loop if we find a known white balance name/tag.
            if (strncmp(line_buf.get(), wbfile[i].name, strlen(wbfile[i].name)) == 0)
                break;
        }

        // Did we find one?
        if (i < NUM_LOADABLE)
        {
            bool as_shot = false;   // Assume no as-shot info.

            // Check for As-Shot information.
            if (strncmp(line_buf.get() + strlen(wbfile[i].name), "+AS", 3) == 0)
                as_shot = true;

            // Read up the white balance settings.
            if (read_wb_settings(fd, as_shot, line_buf.get(), bufsize, i, &linenum))
            {
                rc = -1;
                break;
            }

            if (convert_to_hex(i, as_shot))
            {
                rc = -1;
                break;
            }
        }
    }

bail:
    if (fd != NULL)
        fclose(fd);

    return rc;
}

/*!
 * @brief Retrieve a pointer to the specified white balance gain structure.
 *
 * @param wb_value  The white balance gain structure we're interested in.
 *
 * @return A constant pointer to the requested white balance gain structure, or NULL
 *     if 'wb_value' is out of range.
 */
const WBAL_GAINS *get_wb(int wb_value)
{
    const WBAL_GAINS *wb_ptr = NULL;

    // Make sure we have a safe value.
    if ((wb_value >= GIGE_WHITE_BAL_DAYLIGHT) && (wb_value <= GIGE_WHITE_BAL_CUSTOM))
    {
        // Looks OK. Return the pointer to the structure.
        wb_ptr = &wb_gains[ wb_value ];
    }

    // Return our white balanace gains pointer.
    return wb_ptr;
}

/*!
 * @brief Get the standard light source option
 *
 * @param meta_ptr  Pointer to partition metadata
 *
 * @return One of the GIGE_WHITE_BAL_... values
 */
uint8_t get_wb_option(pm_metadata *meta_ptr)
{
    int wb;

    for (wb = GIGE_WHITE_BAL_DAYLIGHT; wb <= GIGE_WHITE_BAL_NORMAL; ++wb)
    {
        if (wb_gains[ wb ].red != meta_ptr->wb_red)
            continue;
        if (wb_gains[ wb ].green != meta_ptr->wb_grn)
            continue;
        if (wb_gains[ wb ].blue != meta_ptr->wb_blu)
            continue;

        if (wb_gains[ wb ].matrix_red1 != meta_ptr->matrix_coeff1_red)
            continue;
        if (wb_gains[ wb ].matrix_red2 != meta_ptr->matrix_coeff2_red)
            continue;
        if (wb_gains[ wb ].matrix_red3 != meta_ptr->matrix_coeff3_red)
            continue;

        if (wb_gains[ wb ].matrix_grn1 != meta_ptr->matrix_coeff1_grn)
            continue;
        if (wb_gains[ wb ].matrix_grn2 != meta_ptr->matrix_coeff2_grn)
            continue;
        if (wb_gains[ wb ].matrix_grn3 != meta_ptr->matrix_coeff3_grn)
            continue;

        if (wb_gains[ wb ].matrix_blu1 != meta_ptr->matrix_coeff1_blu)
            continue;
        if (wb_gains[ wb ].matrix_blu2 != meta_ptr->matrix_coeff2_blu)
            continue;
        if (wb_gains[ wb ].matrix_blu3 != meta_ptr->matrix_coeff3_blu)
            continue;

        // We don't compare the dc offsets because in the case where
        // we need to infer the WB option from the coefficients, the
        // DC offsets were not being stored.

        // All tests passed so this is a match!
        return wb;
    }

    // None of the standards matched, so call it custom.
    return GIGE_WHITE_BAL_CUSTOM;
}

