/*!
 * @file fcp.cxx
 *
 * @brief This file contains all the routines to deal with GigE
 * FCP messages. It will parse the messages, route them to the
 * correct routines, and then queue the response back.
 *
 * Copyright 2009 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <ctype.h>
#include <fcntl.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <dirent.h>
#include <time.h>
#include <linux/rtc.h>
#include <linux/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/statvfs.h>

#include <ctime>
#include <deque>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

#include <fmt/printf.h>
#include <fmt/format.h>
#include <json/json.h>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/fi_lutfile.h>
#include <fastec/fas_log.h>

#include "autosave.h"
#include "camera_server.h"
#include "cs_network.h"
#include "cs_update.h"
#include "fan_proc.h"
#include "fcp.h"
#include "fsp.h"
#include "grpc_server.h"
#include "image_correction.h"
#include "image_io.h"
#include "lic_manager.h"
#include "mft.h"
#include "override.h"
#include "part_mgr.h"
#include "storage.h"
#include "telemetry.h"
#include "formatter.h"
#include "kernel_timezone.h"
#include "wb_file.h"

extern void fcp_trace(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);

namespace FCP_LOCAL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Our real-time clock device.
const char RTC_DRIVER_PATH[]        = "/dev/rtc0";

// Maximum length of NTP server name
const int MAX_NTP_SERVER_LEN        = 256;

// The script which will update the rc.conf file.
const char NTP_FIX_SCRIPT[]         = ".config/Fastec/scripts/ntp_fix";

// Script to retrieve the NTP server name from the rc.conf file.
const char NTP_GET_NAME[]           = ".config/Fastec/scripts/ntp_name";

// Timezone information
const char TZ_ENV_VAR[]             = "TZ";
const char TZ_FILE_PATH[]           = "/etc/localtime";

// Path length of "/usr/share/zoneinfo/"
const int TZ_PATH_LEN               = 20;

// Size in bytes of buffer for marker list, maximum size of the event marker match list.
const int MARKER_LIST_BUF_SIZE      = (MARKER_LIST_MAX_COUNT *sizeof(uint32_t));

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

time_t lock_time = 0;                       //!< The time that a client locked the system.
time_t fastec_lock_time = 0;                //!< The time that a client locked the system with the Fastec lock.

CS_NETWORK::CS_NETWORK_MESSAGE lock_message;            //!< Message received which locked the system, used for timeout, removing, and keeping the lock.
CS_NETWORK::CS_NETWORK_MESSAGE fastec_lock_message;     //!< Message received which locked the system, used for timeout, removing, and keeping the lock.

// Information about the streaming channel.
time_t stream_time = 0;                     //!< Time that a client started up a stream, or last checked in.
char stream_from[CS_NETWORK::NET_ADDR_LEN];           //!< Network address xxx.xxx.xxx.xxx,inet_ntoa run on it.

uint32_t pp_reg_addr = 0;                   //!< The register address to use for peeking, moreso than poking.

// Variable to hold the content of the RGB Histogram. Its length is always fixed but
// it could vary by the type of sensor.
uint32_t histogram_rgb_size = 0;
uint32_t histogram_rgb_pos = 0;
uint8_t *p_histogram_rgb = NULL;

// Variable to hold the content of the XML folder tree.
uint32_t xml_tree_raw_pos = 0;

// Variables to support the GIGE_REG_HOLD protocol.
bool holding = false;
struct
{
    uint16_t startx;
    uint16_t starty;
    uint16_t width;
    uint16_t height;
    bitmode_t  bit_mode;
    uint32_t frame_rate;
    uint32_t shutter_speed;
    uint32_t sensor_opts;
} hold;

// Variables to control transferring the configuration information.
uint32_t config_pos = 0;

// Variables to control transferring a LUT data file.
uint32_t lut_data_pos = 0;
uint32_t lut_file_size;
int      lut_data_id;

// Variables used to transfer metadata. Note that this same offset is used for readmem of both the
// XML metadata file and the active partition metadata. The expectation, actually the demand, is
// that readmems of different GigE registers will not be interspersed, nor will requests from
// different clients. (Presumably this is protected by a lock.)
uint32_t metadata_pos = 0;

// Variables used to construct and transfer matching metadata list.
uint32_t marker_list_pos    = 0;        // Current position for readmem sequence.
int marker_list_size        = -1;       // Actual size, once match list is constructed.
uint8_t *p_marker_list      = nullptr;  // Buffer to hold match list.

// Items for the event/message unsolicited interface
// Queue of internal applications which want to be notified when an event occurs
std::vector<CS_NETWORK::CS_IPC_MESSAGE> internal_app_event_queue;
int event_msg_id = 0;   // ever incrementing number

// Used to serialize notification events.
std::mutex notifier_mutex;

// The frame number used when reading PFMD.
std::deque<uint32_t> pfmd_frame_requests;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

int handle_discovery(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_forceip(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_packetresend(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_readmem(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_readreg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_writemem(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_writereg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int handle_noncached_write(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_addr, uint32_t reg_value);

int read_config(uint8_t *dest_ptr, uint32_t num_bytes);
int write_config(uint8_t *config_ptr, uint32_t len);
int read_lut_data(uint8_t *dest_ptr, uint32_t num_bytes);
int write_lut_data(uint8_t *config_ptr, uint32_t len);

uint32_t get_config_len(void);
uint32_t get_device_mode(void);
void    get_gige_version(uint16_t *ver_major, uint16_t *ver_minor);
uint32_t get_ifconfig_current(void);
uint32_t get_ifconfig_options(void);
int get_ntp_server(char *ntp_server, int buflen);
int      get_timezone(char *timezone, int buflen);
int get_sensor_ver(char *sensor_ver, int buflen);
uint32_t get_timeout(void);
uint32_t get_file_size(const char *file_path);

int load_file(const char *file_name, uint8_t *dest_ptr, uint32_t byte_count, uint32_t *p_offset);

void    renew_fastec_lock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
void    renew_lock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
void    renew_stream(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
int fcp_kill_stream_thread(void);

bool    update_event(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, bool remove);
int update_load(uint32_t reg_value);
int update_marker_list(uint32_t reg_value);
int update_status(void);
int update_tod(uint32_t reg_value);
bool    validate_access(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);

bool    is_lock_owner(CS_NETWORK::CS_NETWORK_MESSAGE *lock_msg,  CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);
const char *lock_msg_name(CS_NETWORK::CS_NETWORK_MESSAGE *lock_msg);

uint32_t read_temperature(void);
uint32_t fcp_parse_lens_json(const char *settings_str, int str_length);
uint32_t fcp_parse_save_json(const char *save_str, int str_length);
uint32_t fcp_parse_update_json(const char *update_str, int str_length);
bool    fcp_save_limits_okay(uint32_t frame_max);

void    fcp_update_ntp_server(const char *ntp_server);
int fcp_update_timezone(const char *timezone);

int do_hold_operation(uint32_t reg_value);
bool fcp_ltr_mode_select_change(uint32_t reg_value);
int truncate_from(uint32_t free_from);

int fcp_unlock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg);

int fcp_handle_peekpoke(char *json_ptr, uint32_t& num_bytes, bool do_read);

} /* End of namespace FCP_LOCAL */

using namespace FCP_LOCAL;
using namespace CS_NETWORK;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

// #define DEBUG_TIMING 1
#ifdef DEBUG_TIMING

/*!
 * @brief Dump the contents of the specified CS_NETWORK::CS_NETWORK_MESSAGE.
 *
 * @param p_msg Pointer to the message we want to dump.
 */
static void dump_msg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int  temp_val;
    static const char *msg_type_label[] =
    {
        "IPC", "UDP", "FPGA", "Unknown"
    };

    // Make sure we got a message.
    if (p_msg == NULL)
        return;

    // Make sure we have a valid index into our label array.
    temp_val = p_msg->msg_type;
    if (temp_val > MSG_TYPE_FPGA)
        temp_val = 4;

    // Build up the first part of the string.
    // NOTE: MSG_TYPE_xxx is '1' based, so subtract one to get an index into the array.
    fi_log(LOG_INFO, "type: %d (%s) rcv_time: %ld.%ld ",
           p_msg->msg_type, msg_type_label[temp_val - 1],
           p_msg->rcv_time.tv_sec, p_msg->rcv_time.tv_nsec);

    // Now print out the type specific value(s).
    switch (p_msg->msg_type)
    {
    case MSG_TYPE_IPC:
        fi_log(LOG_INFO, "ipc from: '%s'", p_msg->ipc_msg.from);
        break;
    case MSG_TYPE_UDP:
        fi_log(LOG_INFO, "udp from: %s:%d", p_msg->udp_msg.from, p_msg->udp_msg.from_port);
        break;
    case MSG_TYPE_FPGA:
        fi_log(LOG_INFO, "fpga_val: 0x%08x", p_msg->fpga_val);
        break;
    default:
        break;
    }

    // Now see what the request is for.
    uint16_t cmd = ntohs(p_msg->message.header.req.command);
    uint16_t len = ntohs(p_msg->message.header.req.length);
    uint16_t id = ntohs(p_msg->message.header.req.req_id);
    fi_log(LOG_INFO, "hdr: cmd=0x%04x, len=0x%04x, req_id=0x%04x", cmd, len, id);
}

/*!
 * @brief Dump a buffer of size to the console in hex.
 *
 * @param buffer The buffer to dump
 * @param size The size of the buffer to dump
 */
static void dump_mem(uint8_t *buffer, uint16_t size)
{
    int i;

    // dump out the first block of data
    for (i = 0; i < size; i += 16)
    {
        if ((i + 15) > size)
            break;
        PDEBUG(PDEBUG_HIGH, "[dm(0x%x)]:0x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
               i, buffer[i], buffer[i + 1], buffer[i + 2], buffer[i + 3], buffer[i + 4], buffer[i + 5], buffer[i + 6], buffer[i + 7], buffer[i + 8], buffer[i + 9],
               buffer[i + 10], buffer[i + 11], buffer[i + 12], buffer[i + 13], buffer[i + 14], buffer[i + 15]);
    }

}
#else
#define dump_mem( x, y )
#endif /* 0 */

/*!
 * @brief Update our time of day.
 *
 * @param reg_value New time of day setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int FCP_LOCAL::update_tod(uint32_t reg_value)
{
    int rtc_fd;
    int status;
    struct timespec new_time;

    // Since we're setting the TOD, make sure there's no NTP server.
    fcp_update_ntp_server("");

    // Assume success.
    status = GEV_STATUS_SUCCESS;

    // Set the system time.
    new_time.tv_sec = reg_value;
    new_time.tv_nsec = 0;
    clock_settime(CLOCK_REALTIME, &new_time);

    // Now, pass it on to the realtime clock(s).
    rtc_fd = open(RTC_DRIVER_PATH, O_RDWR);
    if (rtc_fd < 0)
    {
        fi_log(LOG_ERR, "%s: Unable to open %s [%d].\n", __FUNCTION__, RTC_DRIVER_PATH, rtc_fd);
        status = GEV_STATUS_ERROR;
    }
    else
    {
        int ret_status;
        time_t  reg_time;
        struct tm   rtc_time;

        // Convert from timespec to tm.
        reg_time = (time_t)reg_value;
        memcpy(&rtc_time, gmtime(&reg_time), sizeof(struct tm));

        // Pass the new time info to the driver.
        ret_status = ioctl(rtc_fd, RTC_SET_TIME, &rtc_time);
        if (ret_status != 0)
            fi_log(LOG_ERR, "%s: Unable to set RTC [%d: %s].\n", __FUNCTION__, errno, strerror(errno));

        // Close the driver.
        close(rtc_fd);
        fi_log(LOG_INFO, "%s: Time updated.", __FUNCTION__);
    }

    // Return our status.
    return status;
}

/*!
 * @brief Update our image loading setting.
 *
 * @param reg_value New image load request.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int FCP_LOCAL::update_load(uint32_t reg_value)
{
    int status;

    // Assume success.
    status = GEV_STATUS_SUCCESS;

    // Make sure we're not running.
    if (imgLoadWorker::getInstance()->isRunning())
    {
        // Only acceptable value is "stop".
        if (reg_value == GIGE_LOAD_STOP)
        {
            // Ask the loader to stop.
            fi_log(LOG_DEBUG, "%s: Stop requested", __FUNCTION__);
            imgLoadWorker::getInstance()->stop();
        }
        else
        {
            // No good. This is an error.
            fi_log(LOG_ERR, "%s: Image Load is already running!", __FUNCTION__);
            status = GEV_STATUS_BUSY;
        }
    }
    else if (reg_value == GIGE_LOAD_START)
    {
        // We have to stop the stream if it's running.
        fcp_stop_stream_thread();

        // Start the loading operation.
        imgLoadWorker::getInstance()->start();
    }

    // Ignore the stop if we've already stopped.
    else if (reg_value != GIGE_LOAD_STOP)
    {
        status = GEV_STATUS_INVALID_PARAMETER;
        fi_log(LOG_INFO, "%s: GIGE_REG_LOAD value %d is wrong", __FUNCTION__, reg_value);
    }

    // Return our status.
    return status;
}

/*!
 * @brief Read the CPU temperature sensor.
 *
 * @return Temperature readings (10ths degree C).
 */
uint32_t FCP_LOCAL::read_temperature(void)
{
    FILE        *fp;
    uint32_t    ret_value = 0;
    char    temp_buf[MAX_BUFFER_LEN + 1];

#define CPU_TEMP_SENSE  "/sys/bus/platform/devices/coretemp.0/hwmon/hwmon2/temp1_input"

    // Open the component side sensor file.
    fp = fopen(CPU_TEMP_SENSE, "r");
    if (!fp)
    {
        fi_log(LOG_WARNING, "%s: Cannot open CPU temp sensor (%s)", __PRETTY_FUNCTION__, CPU_TEMP_SENSE);
        return 0;
    }

    // Read the temperature value.
    if (fgets(temp_buf, MAX_BUFFER_LEN, fp) != NULL)
    {
        // Convert to 10ths degree C. The value read is in 1000ths degree C.
        ret_value = strtoul(temp_buf, NULL, 10) / 100;
    }

    // Close the file.
    fclose(fp);

    // Return our result.
    return ret_value;
}

int FCP_LOCAL::get_ntp_server(char *ntp_server, int buflen)
{
    char temp_buf[PATH_MAX + 1];
    FILE *fp;

    // Sanity check the buffer len.
    if (buflen > MAX_NTP_SERVER_LEN)
        buflen = MAX_NTP_SERVER_LEN;

    // Clear out our return value.
    memset(ntp_server, 0, buflen);

    // Build the command.
    snprintf(temp_buf, PATH_MAX, "%s/%s", get_home_path(), NTP_GET_NAME);

    // Retrieve the NTP server name from the system.
    fp = popen(temp_buf, "r");
    if (fp != NULL)
    {
        // Clear out our buffer.
        memset(temp_buf, 0, MAX_NTP_SERVER_LEN + 1);

        // Read data, if any.
        while (fgets(temp_buf, MAX_NTP_SERVER_LEN, fp) != NULL)
            ;

        // Close the pipe.
        pclose(fp);
    }

    // Make sure we got something.
    if (temp_buf[0] != '\0')
    {
        int str_len;

        // Got the NTP server environment variable - copy it over.
        strncpy(ntp_server, temp_buf, buflen);

        // Kill any trailing line-feed.
        str_len = strlen(ntp_server);
        if (ntp_server[str_len - 1] == '\n')
            ntp_server[str_len - 1] = '\0';
    }

    // Return the length of our string.
    return strlen(ntp_server);
}

/*!
 * @brief Return the string in the TZ environment variable, if set.
 *    Otherwise, return the value linked by /etc/localtime.
 *
 * @return TZ appropriate timezone string.
 */
int FCP_LOCAL::get_timezone(char *timezone, int buflen)
{
    char    *tz_ptr;

    // Clear the return string.
    timezone[0] = '\0';

    // Is there an environment variable?
    tz_ptr = getenv(TZ_ENV_VAR);
    if ((tz_ptr != NULL) && (tz_ptr[0] != '\0'))
    {
        // Got the TZ environment variable - copy it over.
        strncpy(timezone, tz_ptr, buflen);
        timezone[buflen - 1] = '\0';
    }
    else
    {
        int tz_len;
        struct stat64 sb;
        ssize_t r;
        std::unique_ptr<char[]> linkname;

        // Stat the timezone file so we know how big a buffer to allocate.
        if (lstat64(TZ_FILE_PATH, &sb) == -1)
        {
            fi_log(LOG_ERR, "%s: Unable to lstat timezone file.", __PRETTY_FUNCTION__);
            return 0;
        }

        // Allocate our buffer to hold the link name.
        linkname = std::unique_ptr<char[]>(new char[sb.st_size + 1]());
        if (linkname.get() == nullptr)
        {
            LOG_UNLIKELY_ERRNO(__FILE__, __LINE__, errno);
            return 0;
        }

        // Read the link information.
        r = readlink(TZ_FILE_PATH, linkname.get(), sb.st_size + 1);
        if (r == -1)
        {
            LOG_UNLIKELY_ERRNO(__FILE__, __LINE__, errno);
            return 0;
        }

        // Make sure the file name is null terminated.
        linkname[r] = '\0';

        // We can be confident that the path is going to be pointing to
        // "/usr/share/zoneinfo/" and we want everything after that, so
        // just skip over the 20 characters and copy the remainder.
        strncpy(timezone, &linkname[TZ_PATH_LEN], buflen);

        // Did we get anything?
        if (timezone[0] == '\0')
        {
            fi_log(LOG_ERR, "%s: Unable to determine timezone information", __PRETTY_FUNCTION__);
            return 0;
        }

        // Kill any terminating CR.
        tz_len = strlen(timezone);
        if (timezone[tz_len - 1] == '\n')
            timezone[tz_len - 1] = '\0';
    }

    // Return the string.
    return strlen(timezone);
}

/*!
 * @brief Load the version of the sensor in from the procfs.
 *
 * @param sensor_ver    The sensor version name.
 * @param buflen    Length of the buffer to hold the sensor version.
 *
 * @return The length of the sensor version, 0 if there was an error.
 */
int FCP_LOCAL::get_sensor_ver(char *sensor_ver, int buflen)
{
    Sensor *sensor_ptr = CameraServer::get_sensor();
    if (sensor_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: No sensor pointer!", __FUNCTION__);
        return 0;
    }

    return sensor_ptr->get_sensor_ver(sensor_ver, buflen);
}

/*!
 * @brief Get the timeout value from the GIGE register in seconds. If the
 *    register is unavailable returns the default value.
 *
 */
uint32_t FCP_LOCAL::get_timeout(void)
{
    uint32_t    to_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    to_value = regs[ GIGE_REG_HEARTBEAT ].get_value() / 1000;

    // Return our timeout value.
    return to_value;
}

uint32_t FCP_LOCAL::get_file_size(const char *file_path)
{
    struct stat64 file_stat;
    int     fd;

    // Try to open the file.
    fd = open(file_path, O_RDONLY);
    if (fd == -1)
    {
        fi_log(LOG_ERR, "(%s) can't open file %s errno:%d(%s)", __FUNCTION__, file_path, errno, strerror(errno));
        return 0;
    }

    // Find out how big the file is.
    if (fstat64(fd, &file_stat) != 0)
    {
        fi_log(LOG_ERR, "(%s) can't stat file %s errno:%d(%s)", __FUNCTION__, file_path, errno, strerror(errno));
        close(fd);
        return 0;
    }

    // Close our file.
    close(fd);

    // Return the length.
    return (uint32_t)file_stat.st_size;
}

int FCP_LOCAL::load_file(const char *file_name, uint8_t *dest_ptr, uint32_t byte_count, uint32_t *p_offset)
{
    int     fd;
    uint32_t    read_count;

    // Try to open the file.
    fd = open(file_name, O_RDONLY);
    if (fd == -1)
    {
        // Can't open the file.
        fi_log(LOG_ERR, "%s: can't load file %s errno:%d(%s)", __FUNCTION__, file_name, errno, strerror(errno));
        return -1;
    }

    // Try to seek the file to the new position.
    if (lseek(fd, *p_offset, SEEK_SET) == (off_t) -1)
    {
        // Can't seek the file.
        fi_log(LOG_ERR, "%s: can't seek the file, errno:%d(%s)", __FUNCTION__, errno, strerror(errno));
        close(fd);
        return -2;
    }

    // Read the data into our buffer. The amount read may be less
    // than the amount requested. That's OK.
    read_count = read(fd, dest_ptr, byte_count);

    // Update our position.
    *p_offset += read_count;

    // Close our file.
    close(fd);

    // Return the number of bytes read.
    return read_count;
}

int FCP_LOCAL::do_hold_operation(uint32_t reg_value)
{
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint32_t syncin_value;
    uint8_t save_camera_state;
    int     rc = GEV_STATUS_SUCCESS;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // Make sure we're in a good state.
    save_camera_state = CameraServer::get_camera_state();
    if ((save_camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (save_camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
    {
        fi_log(LOG_DEBUG, "%s: Invalid state (%d) for hold!", __PRETTY_FUNCTION__, save_camera_state);
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        FiFPGA      *fpga_ptr = CameraServer::get_fpga();

        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ACCESS_DENIED;
        }
    }

    // Is the client releasing their hold?
    if (reg_value == 0)
    {
        // Check to see if we're actually in a "hold".
        if (!holding)
        {
            // Nothing to see or do here.
            return GEV_STATUS_SUCCESS;
        }

        PDEBUG(PDEBUG_GIGEWRITE, "%s: %ux%u@%u,%u 0x%x %u-fps %u-usecs 0x%08x", __FUNCTION__,
               hold.width, hold.height, hold.startx, hold.starty,
               hold.bit_mode, hold.frame_rate, hold.shutter_speed, hold.sensor_opts);

        // This will "make_capture_valid()" so it will force ROI limits into range,
        // then make sure frame rate works for that ROI, then make sure shutter works
        // for that frame rate. Then it will restart capture with the new settings.
        if (!cs_ptr->modify_capture(hold.startx, hold.starty, hold.width, hold.height,
                                    hold.bit_mode, cs_ptr->reg_to_frame_rate(hold.frame_rate), hold.shutter_speed,
                                    hold.sensor_opts))
        {
            rc = GEV_STATUS_ERROR;
        }

        holding = false;
    }
    else
    {
        float frame_rate;
        // Snapshot the current settings. We want to change only what was modified
        // by writes during the hold so this is our baseline; register writes will
        // now save only these snapshot values.
        holding = true;
        cs_ptr->get_user_settings(&hold.startx, &hold.starty, &hold.width, &hold.height,
                                  &hold.bit_mode, &frame_rate, &hold.shutter_speed,
                                  &hold.sensor_opts);
        hold.frame_rate = cs_ptr->frame_rate_to_reg(frame_rate);
    }
    return rc;
}

/*!
 * @brief Get the device version of the GIGE protocol we are supporting.
 *    if so return it. Otherwise return NULL.
 *
 * @param ver_major - The major number of the version
 * @param ver_minor - The minor number of the version
 */
void FCP_LOCAL::get_gige_version(uint16_t *ver_major, uint16_t *ver_minor)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    GIGE_REG        *reg_ptr;

    // Get our structure pointer.
    reg_ptr = (GIGE_REG *)regs[ GIGE_REG_GIGE_VER ];

    *ver_major = htons(reg_ptr->reg_val_w[1]);
    *ver_minor = htons(reg_ptr->reg_val_w[0]);
}

/*!
 * @brief Get the value of the device mode register.
 *    If the register is available return it. Otherwise return NULL.
 */
uint32_t FCP_LOCAL::get_device_mode(void)
{
    uint32_t    reg_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    reg_value = regs[ GIGE_REG_DEV_MODE ].get_value();
    return htonl(reg_value);
}

/*!
 * @brief Return the ipconfig options GIGE register. If the register is there
 *    return it. Otherwise return NULL.
 */
uint32_t FCP_LOCAL::get_ifconfig_options(void)
{
    uint32_t    reg_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    reg_value = regs[ GIGE_REG_IP_CFG ].get_value();
    return htonl(reg_value);
}

/*!
 * @brief Return the ipconfig current GIGE register. If the register is there
 *    return it. Otherwise return NULL.
 */
uint32_t FCP_LOCAL::get_ifconfig_current(void)
{
    uint32_t    reg_value;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    reg_value = regs[ GIGE_REG_IP_CUR_CFG ].get_value();
    return htonl(reg_value);
}

/*!
 * @brief Truncate a ROC/BROC recording
 *
 * @param free_from Free starting with this frame number.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int FCP_LOCAL::truncate_from(uint32_t free_from)
{
    PartMgr      *p_partmgr = CameraServer::get_partmgr();
    FiFPGA       *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();

    bool    in_ltr_roc = false;
    uint8_t camera_state;
    int     status = GEV_STATUS_SUCCESS;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, free_from);

    // Check the camera state.
    camera_state = CameraServer::get_camera_state();
    if ((camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) &&
        (camera_state != GIGE_CAMERA_STATE_IDLE))
        return GEV_STATUS_INVALID_PARAMETER;

    // This is a little silly but it's the way we do it everywhere else so we'll stick
    // with it for now. We should be able to just look at the GigE cached register instead
    // of using FiFPGA::ltr_mode().
    if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        GIGE_REG_CACHE& regs = CameraServer::get_regs();
        uint32_t mode = regs[ GIGE_REG_LTR_MODE ].get_value();
        if ((mode & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0)
            in_ltr_roc = true;
    }
    if (!in_ltr_roc)
    {
        fi_log(LOG_ERR, "%s: attempting to truncate when not in LTR ROC/BROC", __FUNCTION__);
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // This is mostly to make sure the GOE live stream is stopped.
    if (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        cs_ptr->set_idle_behavior();

    // Don't allow us to create a recording that is less than 2 frames in size.
    if (free_from < 2)
        free_from = 2;

    if (p_partmgr->ltrTruncateFrom(free_from))
    {
        fcp_notify_apps(GIGE_REG_PART_NUM_FRAMES, free_from);
        status = GEV_STATUS_SUCCESS;
    }
    else
    {
        fi_log(LOG_ERR, "%s: ltr_truncate_from(0x%x) failed!", __FUNCTION__, free_from);
        status = GEV_STATUS_ERROR;
    }

    if (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        cs_ptr->set_live_behavior();

    return status;
}

/*!
 * @brief We've received a request, if the system is locked renew the lock
 *
 * @param p_msg The message received for processing.
 *
 * @returns 0 if there were no errors
 */
void FCP_LOCAL::renew_lock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    time_t   cur_time;

    if (lock_time == 0)
        return;  // no lock held.

    if (p_msg->msg_type != lock_message.msg_type)
        return; // mismatch between ipc and udp

    // Get the current time.
    cur_time = time(NULL);

    // Are we doing IPC or GRPC messages?
    if ((lock_message.msg_type == CS_NETWORK::MSG_TYPE_IPC) ||
        (lock_message.msg_type == CS_NETWORK::MSG_TYPE_GRPC))
    {
        // on an internal message don't need to check further.
        lock_time = cur_time;
    }
    else
    {
        if (is_lock_owner(&lock_message, p_msg))
        {
            lock_time = cur_time;
        }
    }
}

/*!
 * @brief We've received a request, if the system is locked by fastec lock, renew the lock
 *
 * @param p_msg The message received for processing.
 *
 * @returns 0 if there were no errors
 */
void FCP_LOCAL::renew_fastec_lock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    time_t   cur_time;

    if (fastec_lock_time == 0)
        return;  // no lock held.

    if (p_msg->msg_type != fastec_lock_message.msg_type)
        return; // mismatch between ipc and udp

    // Get the current time.
    cur_time = time(NULL);

    if (is_lock_owner(&fastec_lock_message, p_msg))
    {
        fastec_lock_time = cur_time;
    }
}

/*!
 * @brief Validate that the message sender has write privilege per the locks
 *
 * @param p_msg The message received for processing.
 *
 * @returns true the sender has the right-of-way
 */
bool FCP_LOCAL::validate_access(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    uint32_t    reg_addr;
    static uint32_t last_reg_addr = 0;

    // Get the register we're interested in.
    reg_addr = ntohl(p_msg->message.data_l[0]);

    // First check the standard lock (CCP).
    // Don't bother if it's an IPC or GRPC client, since those don't need to respect the non-fastec lock.
    if (!((p_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC) || (p_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC)) &&
        (lock_time != 0))
    {
        // If the caller does NOT already own the lock, then
        // access is denied.
        if (!is_lock_owner(&lock_message, p_msg))
        {
            fi_log(LOG_DEBUG, "%s: Remote requestor (%s) is not the owner (%s) [reg=0x%x]", __FUNCTION__, p_msg->udp_msg.from, lock_message.udp_msg.from, reg_addr);
            return false;
        }
    }

    // Special processing based on register...
    switch (reg_addr)
    {
    default:
        break;

    // The FCCP doesn't effect the streaming registers.
    case GIGE_REG_SCP0:
    case GIGE_REG_SCPS0:
    case GIGE_REG_SCPD0:
    case GIGE_REG_SCDA0:
    case GIGE_REG_STREAM_CONTROL:
    case GIGE_REG_STREAM_RETRY:
    case GIGE_REG_STREAM_SEQ_START:
    case GIGE_REG_STREAM_START_NUM:
    case GIGE_REG_STREAM_END_NUM:
    case GIGE_REG_STREAM_FRAME_INTERVAL:
    case GIGE_REG_STREAM_FRAME_DELAY:
    case GIGE_REG_STREAM_FORMAT:
    case GIGE_REG_STREAM_FILEPATH:
        return true;

    // The FCCP doesn't effect the dedicated local "streaming" registers.
    case GIGE_REG_JPEG_FRAME_ADDR:
        return true;
    }

    // Now check the fastec-specific lock (FCCP)
    if (fastec_lock_time != 0)
    {
        bool        access = true;

        // Is this the owner asking for the lock again?
        if (!is_lock_owner(&fastec_lock_message, p_msg))
        {
            fi_log(LOG_DEBUG, "%s: FCCP requestor (%s) is not the owner (%s) [reg=0x%x, last=0x%x]", __FUNCTION__, lock_msg_name(p_msg), lock_msg_name(&fastec_lock_message), reg_addr, last_reg_addr);
            access = false;
        }

        // Is access denied?
        if (!access)
        {
            fi_log(LOG_WARNING, "%s:  Concurrent write access attempted: denied!", __FUNCTION__);
            return false;
        }
    }

    // Remember the last register access.
    last_reg_addr = reg_addr;

    // Access is allowed.
    return true;
}

/*!
 * @brief We've received a request, if the system has a stream going renew it.
 *
 * @param p_msg The message received for processing.
 *
 * @returns 0 if there were no errors
 */
void FCP_LOCAL::renew_stream(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    // Make sure we even have a lock on the stream.
    if (stream_time == 0)
        return;

    // UDP is the only one who can have a stream going.
    if (p_msg->msg_type != CS_NETWORK::MSG_TYPE_UDP)
        return;

    // On a udp message we need to check if it's the same client, if so update the time.
    if (memcmp(p_msg->udp_msg.from, stream_from, sizeof(stream_from)) == 0)
        stream_time = time(NULL);
}

const char *FCP_LOCAL::lock_msg_name(CS_NETWORK::CS_NETWORK_MESSAGE *lock_msg)
{
    const char *ret_value = NULL;

    // If IPC/GRPC, return the socket name.
    if ((lock_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC) || (lock_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC))
    {
        ret_value = lock_msg->ipc_msg.from;
    }

    // For a UDP message, return the IP address.
    else if (lock_msg->msg_type == CS_NETWORK::MSG_TYPE_UDP)
    {
        ret_value = lock_msg->udp_msg.from;
    }

    return ret_value;

}

bool FCP_LOCAL::is_lock_owner(CS_NETWORK::CS_NETWORK_MESSAGE *lock_msg,  CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    bool owner = false;

    // If IPC or GRPC, compare named sockets
    if ((lock_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC) || (lock_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC))
    {
        if (strcmp(p_msg->ipc_msg.from, lock_msg->ipc_msg.from) == 0)
        {
            owner = true;
        }
    }
    else if (lock_msg->msg_type == CS_NETWORK::MSG_TYPE_UDP)
    {
        // On a UDP message we need to check if it's the same client IP
        if (memcmp(p_msg->udp_msg.from, lock_msg->udp_msg.from, sizeof(lock_msg->udp_msg.from)) == 0)
        {
            owner = true;
        }
    }

    return owner;

}

bool FCP_LOCAL::update_event(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, bool remove)
{
    bool rc;
    int elem;

    // Assume failure.
    rc = false;

    if (remove)
    {
        // Removing an app from the notification queue. Look through the whole
        // queue and see if there is an app which matches who is requesting to stop
        // receving notifications. If so then get rid of it. If not then just be silent
        // since we really don't care.
        if ((p_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC) || (p_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC))
        {
            int cnt = internal_app_event_queue.size();
            for (elem = 0; elem < cnt; elem++)
            {
                CS_NETWORK::CS_IPC_MESSAGE& app_ref = internal_app_event_queue[ elem ];
                if (strcmp(app_ref.from, p_msg->ipc_msg.from) == 0)
                {
                    // Found it - remove it from the notifiers queue.
                    internal_app_event_queue.erase(internal_app_event_queue.begin() + elem);
                }
            }

            // All is well.
            rc = true;
        }
    }
    else
    {
        // Adding an app to the notification queue
        if ((p_msg->msg_type == CS_NETWORK::MSG_TYPE_IPC) || (p_msg->msg_type == CS_NETWORK::MSG_TYPE_GRPC))
        {
            // Before we add the app make sure that it's not already in the queue, if so then remove the
            // stale reference
            int cnt = internal_app_event_queue.size();
            for (elem = 0; elem < cnt; elem++)
            {
                CS_NETWORK::CS_IPC_MESSAGE& app_ref = internal_app_event_queue[ elem ];
                if (strcmp(app_ref.from, p_msg->ipc_msg.from) == 0)
                {
                    // Found it - remove it from the notifiers queue.
                    internal_app_event_queue.erase(internal_app_event_queue.begin() + elem);
                }
            }

            // Ok - push it on the event notifiers queue.
            internal_app_event_queue.push_back(p_msg->ipc_msg);

            // All is well.
            rc = true;
        }
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Process the discovery message, by annoucing ourselves on the network.
 *
 * @param p_msg - the message received for processing
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_discovery(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    csWorker *p_worker;

    PDEBUG(PDEBUG_HIGH, "%s: starting", __FUNCTION__);

    if (!p_msg)
        return -1;

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // FORNOW - should do a randomized delay between 0-1 sec and maybe
    // we want to broadcast the response
    p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
    p_msg->message.header.resp.command = htons(GIGE_MSG_DISCOVERY_ACK);

    // see if we are busy with other tasks.
    if (FiImageCorrection::fpnCalibrateActive())
    {
        p_msg->message.header.resp.status = htons(GEV_STATUS_BUSY);
        p_msg->message.header.resp.length = 0;
    }
    else
    {
        // refresh the network just in case something has changed.
        p_worker->discoverNetwork(p_msg);

        p_msg->message.header.resp.status = htons(GEV_STATUS_SUCCESS);

        // FORNOW - We probably will need to revisit this once the
        // camera is further along.
        memset(p_msg->message.data_b, 0, sizeof(p_msg->message.data_b));
        get_gige_version(&p_msg->message.data_w[GIGE_DSC_ACK_PROTO_VER_W], &p_msg->message.data_w[GIGE_DSC_ACK_PROTO_VER_W + 1]); // Version major, minor
        p_msg->message.data_l[GIGE_DSC_ACK_MODE_L] = get_device_mode();

        p_worker->getMacAddress((uint8_t *)(&p_msg->message.data_b[GIGE_DSC_ACK_MAC_B]));
        p_msg->message.data_l[GIGE_DSC_ACK_IP_OPT_L] = get_ifconfig_options();
        p_msg->message.data_l[GIGE_DSC_ACK_IP_CUR_L] = get_ifconfig_current();
        p_msg->message.data_l[GIGE_DSC_ACK_IP_L] = p_worker->getIpAddress(p_msg);
        p_msg->message.data_l[GIGE_DSC_ACK_SUBNET_L] = p_worker->getSubnetMask();
        p_msg->message.data_l[GIGE_DSC_ACK_GATEWAY_L] = p_worker->getDefaultGateway();
        fcp_get_manufacture_name(&p_msg->message.data_b[GIGE_DSC_ACK_VENDOR_B], GIGE_DSC_ACK_VENDOR_B_LEN);
        fcp_get_model_name(&p_msg->message.data_b[GIGE_DSC_ACK_MODEL_B], GIGE_DSC_ACK_MODEL_B_LEN);
        fcp_get_device_version_name(&p_msg->message.data_b[GIGE_DSC_ACK_VERSION_B], GIGE_DSC_ACK_VERSION_B_LEN);
        fcp_get_manufacture_info(&p_msg->message.data_b[GIGE_DSC_ACK_MAN_INFO_B], GIGE_DSC_ACK_MAN_INFO_B_LEN);
        fcp_get_serial_number(&p_msg->message.data_b[GIGE_DSC_ACK_SERIAL_B], GIGE_DSC_ACK_SERIAL_B_LEN);
        CsNetwork::getHostname(&p_msg->message.data_b[GIGE_DSC_ACK_USER_NAME_B], GIGE_DSC_ACK_USER_NAME_B_LEN);

        // Now fill in the length
        p_msg->message.header.resp.length = htons(GIGE_DSC_ACK_B_LEN);
    }

    // Now send the response
    p_worker->sendMsg(p_msg, GIGE_DSC_ACK_B_LEN + sizeof(p_msg->message.header));

    PDEBUG(PDEBUG_HIGH, "(%s) ending", __FUNCTION__);
    return 0;
}

/*!
 * @brief Client has requested that we change our IP, subnet, and possible
 * gateway. Note this function will not return success unless the
 * daemon is running as root.
 *
 * @param p_msg - the message received for processing
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_forceip(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     status;
    uint8_t mac_address[CS_NETWORK::MAC_ADDR_LEN];
    uint8_t our_mac[CS_NETWORK::MAC_ADDR_LEN];
    uint32_t    ip_addr = 0;
    uint32_t    subnet_addr = 0;
    uint32_t    gateway_addr = 0;
    csWorker   *p_worker;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    LOG("%s", __PRETTY_FUNCTION__);

    if (!p_msg)
        return -1;

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // Get the data out which the client has requested.
    memcpy(mac_address, &p_msg->message.data_b[ GIGE_FORCEIP_REQ_MAC_B ], 6);
    ip_addr = p_msg->message.data_l[ GIGE_FORCEIP_REQ_IP_L ];
    subnet_addr = p_msg->message.data_l[ GIGE_FORCEIP_REQ_SUBNET_L ];
    gateway_addr = p_msg->message.data_l[ GIGE_FORCEIP_REQ_GATEWAY_L ];

    fi_log(LOG_DEBUG, "%s: mac:%02x:%02x:%02x:%02x:%02x:%02x ip:0x%x subnet:0x%x gateway:0x%x",
           __FUNCTION__, mac_address[0], mac_address[1], mac_address[2], mac_address[3],
           mac_address[4], mac_address[5], ip_addr, subnet_addr, gateway_addr);

    // Now send the response, status is already filled in.
    p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
    p_msg->message.header.resp.command = htons(GIGE_MSG_FORCEIP_ACK);

    // Make sure that the MAC address is for us since this can be broadcast.
    p_worker->getMacAddress(our_mac);
    if ((our_mac[0] != mac_address[0]) || (our_mac[1] != mac_address[1]) ||
        (our_mac[2] != mac_address[2]) || (our_mac[3] != mac_address[3]) ||
        (our_mac[4] != mac_address[4]) || (our_mac[5] != mac_address[5]))
    {
        fi_log(LOG_INFO, "(%s) mac doesn't match us mac:%02x:%02x:%02x:%02x:%02x:%02x us:%02x:%02x:%02x:%02x:%02x:%02x",
               __FUNCTION__, mac_address[0], mac_address[1], mac_address[2], mac_address[3],
               mac_address[4], mac_address[5], our_mac[0], our_mac[1],
               our_mac[2], our_mac[3], our_mac[4], our_mac[5]);

        // We're done - no need for any further processing.
        return 0;
    }

    // See if we are busy with other tasks and can't respond.
    if (FiImageCorrection::fpnCalibrateActive())
    {
        p_msg->message.header.resp.status = htons(GEV_STATUS_BUSY);
        p_msg->message.header.resp.length = 0;
    }

    // Now see if the message sender owns the appropriate locks
    else if (!validate_access(p_msg))
    {
        p_msg->message.header.resp.status = htons(GEV_STATUS_ACCESS_DENIED);
        p_msg->message.header.resp.length = 0;
    }
    else
    {
        // Sanity check the settings.
        if ((ip_addr == 0) || (subnet_addr == 0) || (gateway_addr == 0))
        {
            fi_log(LOG_INFO, "(%s) invalid ip_addr:0x%x or subnet:0x%x or gateway:0x%x",
                   __FUNCTION__, ip_addr, subnet_addr, gateway_addr);
            p_msg->message.header.resp.status = htons(GEV_STATUS_INVALID_PARAMETER);
            p_msg->message.header.resp.length = 0;
        }
        else
        {
            // Switch to a static IP address.
            status = CsNetwork::startStaticConnection(CsNetwork::ETHER, ip_addr, subnet_addr, gateway_addr);
            if (status == 0)
            {
                uint32_t old_value;

                // Set all the persisent registers
                regs[ GIGE_REG_PERSIST_IP ].set_value(ip_addr);
                regs[ GIGE_REG_PERSIST_SUBNET ].set_value(subnet_addr);
                regs[ GIGE_REG_PERSIST_GATEWAY ].set_value(gateway_addr);

                // Since we've change the IP address we are now in static mode so we need to
                // change the IP current configuration register
                old_value = regs[ GIGE_REG_IP_CUR_CFG ].get_value();
                old_value = (old_value & ~GIGE_IFCONFIG_DHCP_BE) | GIGE_IFCONFIG_PERSISTENT_BE;
                regs[ GIGE_REG_IP_CUR_CFG ].set_value(old_value);

                // Now fill in the status and length
                p_msg->message.header.resp.status = htons(GEV_STATUS_SUCCESS);
                p_msg->message.header.resp.length = htons(GIGE_FORCEIP_ACK_B_LEN);   // 0 length return
            }
            else
            {
                p_msg->message.header.resp.status = htons(GEV_STATUS_ACCESS_DENIED);
                p_msg->message.header.resp.length = 0;
            }
        }
    }

    // Now send the response
    p_worker->sendMsg(p_msg, sizeof(p_msg->message.header));

    return 0;
}

/*!
 * @brief Client has requested to read a register. The register to read is
 *    embedded inside the messsage.
 *
 * @param p_msg The message received for processing.
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_readreg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     i;
    uint16_t    num_regs = 0;
    uint16_t    length = 0;
    uint32_t    reg_value = 0;   // host order value
    uint32_t    reg_value_n = 0; // network order value
    uint16_t    status = GEV_STATUS_SUCCESS;
    int16_t err_reg = -1;
    csWorker   *p_worker;
    FiFPGA  *fpga_ptr;
    Sensor  *sensor_ptr;
    PartMgr *p_partmgr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (!p_msg)
        return -1;

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();
    p_partmgr = CameraServer::get_partmgr();
    sensor_ptr = CameraServer::get_sensor();

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // Figure out how many registers the user would like to read
    length = ntohs(p_msg->message.header.req.length);
    num_regs = length / FCP_REG_SIZE; // each register is 4 long.

    // Now do the work
    for (i = 0; ((i < num_regs) && (err_reg == -1)); i++)
    {
        uint32_t    reg_addr;

        // Get the data out which the client has requested.
        reg_addr = ntohl(p_msg->message.data_l[ i ]);

        PDEBUG(PDEBUG_HIGH, "%s: length:%d num_regs:%d reg_addr:0x%x", __FUNCTION__, length, num_regs, reg_addr);

        // If we are busy doing something which can't be interrupted then bail.
        if (FiImageCorrection::fpnCalibrateActive())
        {
            status = GEV_STATUS_BUSY;
            err_reg = i;
            break;
        }

        // Special case the registers which map to hardware and are not stored
        // in the array of registers because they are transitory and their state
        // is not saved.
        switch (reg_addr)
        {
        case GIGE_REG_FILE_ACTION_STATUS:
            reg_value = storage::get_status();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_IP_ADDR:
            // This value comes back as network byte order
            reg_value = p_worker->getIpAddress();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_SUBNET_MASK:
            // This value comes back as network byte order
            reg_value = p_worker->getSubnetMask();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_DEF_GATEWAY:
            // This value comes back as network byte order
            reg_value = p_worker->getDefaultGateway();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_WIFI_IP_ADDR:
            // This value comes back as network byte order.
            // If DHCP is set for WIFI, query the system,
            // otherwise give the value stored in the array.
            if (regs[ GIGE_REG_WIFI_IP_CUR_CFG ].get_value() & ((uint32_t)GIGE_IFCONFIG_DHCP_BE))
                reg_value = p_worker->getWifiIpAddress();
            else
                reg_value = regs[ GIGE_REG_WIFI_IP_ADDR ].get_value();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_WIFI_SUBNET_MASK:
            // This value comes back as network byte order.
            // If DHCP is set for WIFI, query the system,
            // otherwise give the value stored in the array.
            if (regs[ GIGE_REG_WIFI_IP_CUR_CFG ].get_value() & ((uint32_t)GIGE_IFCONFIG_DHCP_BE))
                reg_value = p_worker->getWifiSubnetMask();
            else
                reg_value = regs[ GIGE_REG_WIFI_SUBNET_MASK ].get_value();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_WIFI_DEF_GATEWAY:
            // This value comes back as network byte order.
            // If DHCP is set for WIFI, query the system,
            // otherwise give the value stored in the array.
            if (regs[ GIGE_REG_WIFI_IP_CUR_CFG ].get_value() & ((uint32_t)GIGE_IFCONFIG_DHCP_BE))
                reg_value = p_worker->getWifiDefaultGateway();
            else
                reg_value = regs[ GIGE_REG_WIFI_DEF_GATEWAY ].get_value();
            reg_value_n = reg_value;
            break;
        case GIGE_REG_WIFI_IP_CUR_CFG:
            if (CsNetwork::getPreferredDhcpMode(CsNetwork::WIFI))
                reg_value = GIGE_IFCONFIG_DHCP_BE;
            else
                reg_value = GIGE_IFCONFIG_PERSISTENT_BE;
            reg_value_n = reg_value;
            break;
        case GIGE_REG_WIFI_AP_CFG:
            reg_value =  CsNetwork::getWifiConfigFromConf();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_WIFI_STATUS:
            // Check to see if the WIFI device is present (and allowed).
            reg_value = CsNetwork::getWifiStatus();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_CAMERA_STATE:
            reg_value = CameraServer::get_camera_state();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_CUR_FRAME_POS:
            // Return the current position for a armed or triggered partition
            reg_value = p_partmgr->getCurrentFrameNum();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_CUR_CAPTURE_COUNT:
            // Return the current count of the number of frames captured.
            reg_value = fpga_ptr->icmCaptureFrameCount();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_FRAME_OVERHEAD:
            reg_value = sensor_ptr->get_frame_overhead();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_PART_STATE:
            reg_value = p_partmgr->getPartitionState();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_PART_NUM_FRAMES:
            reg_value = p_partmgr->getNumFrames();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_PART_TRIG_FRAME:
            reg_value  = p_partmgr->getTriggerFrame();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_ACTIVE_PART:
            reg_value  = p_partmgr->getActivePartition();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_ACTIVE_FRAME_NUM:
            reg_value  = p_partmgr->get_active_frame_num();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_PEAK:
            reg_value = fpga_ptr->ipmRgbHistoPeak();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_RED:
            reg_value = fpga_ptr->ipmRgbHistoSumX_Red();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_GRN:
            reg_value = fpga_ptr->ipmRgbHistoSumX_Green();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_BLU:
            reg_value = fpga_ptr->ipmRgbHistoSumX_Blue();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_L:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_RedL();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_H:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_RedH();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_L:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_GreenL();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_H:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_GreenH();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_L:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_BlueL();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_H:
            reg_value = fpga_ptr->ipmRgbHistoSumXSqrd_BlueH();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_HISTO_RGB_SIZE:
            // Get the histogram info from the sensor.
            histogram_rgb_size = fpga_ptr->HISTOGRAM_RGB_SIZE;

            // if we already have a histogram reset the pointers
            if (!p_histogram_rgb)
                p_histogram_rgb = (uint8_t *)calloc(histogram_rgb_size, 1);
            if (p_histogram_rgb)
            {
                PDEBUG(PDEBUG_HIGH, "%s: allocated 0x%x bytes for rgb histogram buffer:0x%x",
                       __FUNCTION__, histogram_rgb_size, (uint64_t)p_histogram_rgb);
                histogram_rgb_pos = 0;
                reg_value = histogram_rgb_size;
            }
            else
            {
                fi_log(LOG_ERR, "(%s) can't allocate memory for rgb histogram, returning 0 size errno:%d(%s)",
                       __FUNCTION__, errno, strerror(errno));
                reg_value = 0;
            }
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_CONFIG_SIZE:
            reg_value = get_config_len();
            reg_value_n = htonl(reg_value);
            config_pos = 0;
            break;
        case GIGE_REG_SYNC_IN_TIMING:
            // get the sync in timing from the fpga
            reg_value = fpga_ptr->syncInGetTiming();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_LOW_LIGHT_ENABLE:
            // While this register is saved, the real value can change out from under us, say
            // when the camera is armed, so just refresh it.
            reg_value = fpga_ptr->icmIsLowLightEnabled();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_TIMEOFDAY:
            reg_value = time(NULL);
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_SAVE_STATE:
            reg_value = imgSaveWorker::getInstance()->getCompletion();
            reg_value |= (imgSaveWorker::getInstance()->isRunning() ? GIGE_SAVE_STATE_BUSY : GIGE_SAVE_STATE_IDLE);

            // We might need to report an LTR error but don't do it if we have a real save error,
            // and don't interfere with save progress data. If there was an LTR error, it should
            // have been noticed and cleared before we started saving.
            if (((reg_value & (GIGE_SAVE_STATE_ERR | GIGE_SAVE_STATE_IDLE)) == (uint32_t)GIGE_SAVE_STATE_IDLE) &&
                ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE))
            {
                uint32_t ltr_err = fpga_ptr->ltrSaveError();
                if ((ltr_err & GIGE_SAVE_STATE_ERR) != 0)
                    reg_value = ltr_err | GIGE_SAVE_STATE_IDLE;
            }
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_AVAIL_MEDIA:
            reg_value = storage::get_available_media();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_FORMAT_MEDIA:
            // Is the format process done?
            if (!formatter::formatDone((int&)reg_value))
                reg_value = 0;
            else
                reg_value |= GIGE_FORMAT_DONE;

            // Get the network ordered flavor.
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_TEMPERATURE:
            reg_value = read_temperature();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_LOAD:
            // Get the status of the load operation.
            reg_value = imgLoadWorker::getInstance()->getCompletion();
            reg_value |= (imgLoadWorker::getInstance()->isRunning() ? GIGE_LOAD_BUSY : 0);
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_SESSION_LENGTH:
            reg_value  = p_partmgr->getSessionLength();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_BOOT_STATUS:
            reg_value = update_status();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_EXTIO_ACTIVITY:
            reg_value = fpga_ptr->extioGetActivity();
            reg_value_n = htonl(reg_value);
            break;

        case GIGE_REG_CCP:
            // Get the register value.
            reg_value = regs[ GIGE_REG_CCP ].get_value();

            // If the user is reading the config register and someone has it
            // locked, make sure the lock hasn't timed out,in which case show
            // no lock.
            if (reg_value & GIGE_CCP_CA_BE)
            {
                time_t  cur_time;

                // What time is it NOW?
                cur_time = time(NULL);

                // See if the timeout has expired.
                if (lock_time != 0)
                {
                    uint32_t    dtime;

                    // Figure out how much time has elapsed.
                    dtime = (uint32_t)difftime(cur_time, lock_time);

                    // Is it time to release the lock?
                    if (dtime >= get_timeout())
                    {
                        fi_log(LOG_INFO, "(%s): timeout expired on CCP register, breaking lock addr:0x%x", __FUNCTION__, reg_addr);
                        fcp_update_lock_state(NULL, 0);
                        reg_value = 0;
                    }
                }
            }
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_FASTEC_CCP:
            // Get the register value.
            reg_value = regs[ GIGE_REG_FASTEC_CCP ].get_value();
            if (reg_value & GIGE_FASTEC_CCP_CA_BE)
            {
                time_t  cur_time;

                // What time is it NOW?
                cur_time = time(NULL);

                // See if the timeout has expired.
                if (fastec_lock_time != 0)
                {
                    uint32_t    dtime;

                    // Figure out how much time has elapsed.
                    dtime = (uint32_t)difftime(cur_time, fastec_lock_time);

                    // Is it time to release the lock?
                    if (dtime >= get_timeout())
                    {
                        fi_log(LOG_INFO, "(%s): timeout expired on FastecCCP register, breaking lock addr:0x%x",
                               __FUNCTION__, reg_addr);
                        fcp_update_fastec_lock_state(p_msg, 0);
                        reg_value = 0;
                    }
                }
            }
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_IRIG_STATUS:
            reg_value = fpga_ptr->irigStatus();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_TIME_SBS:
            reg_value = fpga_ptr->timeReadSBS();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_TIME_SBUS:
            reg_value = fpga_ptr->timeReadSBUS();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_LTR_SECTOR_COUNT:
            // FIXME Rename this GigE register to alert users of new meaning
            reg_value = p_partmgr->ltrGetActiveImageSectors();
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_XML_METADATA:
        {
            char temp_buf[PATH_MAX];
            snprintf(temp_buf, PATH_MAX, "%s/%s", get_home_path(), DOWNLOAD_XML_METADATA);
            reg_value = get_file_size(temp_buf);
            reg_value_n = htonl(reg_value);
        }
        break;
        case GIGE_REG_MARKER_LIST:
            reg_value = marker_list_size;
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_AUTOSAVE:
            // If the destination is off-camera, the autosave manager doesn't actually do anything,
            // and the available count is not being updated. So we'll invent values that make
            // sense to an outside observer who wants to know how many dirty partitions there are
            // (total - available): The dirty count should be 1 if we are in the AUTOSAVE state or
            // if the partition state is COMPLETE regardless of the state; otherwise dirty is zero.
            if ((imgSaveWorker::getInstance()->getDestInfo() & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
            {
                unsigned total;

                reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
                total = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value);
                reg_value &= ~GIGE_AUTOSAVE_COUNT_MASK;
                if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_AUTOSAVE)
                {
                    reg_value |= (total - 1);
                    reg_value |= GIGE_AUTOSAVE_SAVING;
                }
                else if (p_partmgr->getPartitionState() == GIGE_PART_STATE_COMPLETE)
                    reg_value |= (total - 1);
                else
                    reg_value |= total;
                reg_value_n = htonl(reg_value);
            }
            else
            {
                reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
                reg_value |= (CameraServer::get_autosave()->isRunning() ? GIGE_AUTOSAVE_SAVING : 0);
                reg_value_n = htonl(reg_value);
            }
            break;
        case GIGE_REG_SESSION_SIZE:
            p_partmgr->getReviewPartLength(&reg_value);
            reg_value_n = htonl(reg_value);
            break;
        case GIGE_REG_FAN_CONTROL:
            reg_value = fan_get_status();
            reg_value_n = htonl(reg_value);
            break;
        default:
            // get the register and make sure that the caller asked for
            // one of the binary registers. The string registers are returned
            // different ways.
            if (regs.validate_id(reg_addr))
            {
                if (regs[reg_addr].reg_type() != REG_TYPE_BIN)
                {
                    fi_log(LOG_INFO, "%s: reg reg_addr:0x%x is a string register. Returning error invalid address.",
                           __FUNCTION__, reg_addr);

                    reg_value = 0;
                    reg_value_n = 0;
                    status = GEV_STATUS_INVALID_ADDRESS;
                    err_reg = i;
                }
                else
                {
                    // Get the register value.
                    reg_value = regs[ reg_addr ].get_value();

                    if (regs[reg_addr].reg_access() == REG_ACCESS_WO)
                    {
                        PDEBUG(PDEBUG_LOW, "%s: attempt to read wo register reg_addr:0x%x", __FUNCTION__, reg_addr);
                        status = GEV_STATUS_ACCESS_DENIED;
                        reg_value = 0;
                        err_reg = i;
                    }

                    // If we are supposed to swap the bytes do it.
                    if (regs[reg_addr].reg_swap_mode() == REG_SWAP_HTONL)
                        reg_value_n = htonl(reg_value);
                    else
                        reg_value_n = reg_value;
                }
            }
            else
            {
                // Not one of our "normal" registers. Ask the FPGA to read this address.
                fpga_ptr->regRead(reg_addr, &reg_value);
                reg_value_n = htonl(reg_value);
            }
        }
        p_msg->message.data_l[i] = reg_value_n;

        PDEBUG(PDEBUG_GIGEREAD, "%s: 0x%04x:0x%x=%u", __FUNCTION__, reg_addr, reg_value, reg_value);
    }

    // Now send the response
    p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
    p_msg->message.header.resp.command = htons(GIGE_MSG_READREG_ACK);
    p_msg->message.header.resp.status = htons(status);

    // Now fill in the length
    if (err_reg != -1)
        length = (err_reg * FCP_REG_SIZE);
    p_msg->message.header.resp.length = htons(length);

    PDEBUG(PDEBUG_HIGH, "%s: returning status:0x%x length:%d", __FUNCTION__, status, length);

    fcp_trace(p_msg);

    // Now send the response
    p_worker->sendMsg(p_msg, sizeof(p_msg->message.header) + length);

    return 0;
}

/*!
 * @brief Client has requested to read a chunk of memory. This is usually
 * used to read and transfer the XML fle. Which takes multiple
 * read mem's until the while file has been transferred.
 *
 * @param p_msg - the message received for processin
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_readmem(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     status = GEV_STATUS_SUCCESS;
    uint16_t    num_bytes = 0;
    uint32_t    num_left;
    uint32_t    num_to_copy;
    uint32_t    addr = 0;
    FiFPGA  *fpga_ptr;
    PartMgr *p_partmgr;
    csWorker   *p_worker;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Sensor  *sensor_ptr = CameraServer::get_sensor();

    PDEBUG(PDEBUG_HIGH, "%s: starting", __FUNCTION__);
    if (!p_msg)
        return -1;

    // Get a pointer to our FPGA class.
    fpga_ptr = CameraServer::get_fpga();
    p_partmgr = CameraServer::get_partmgr();

    // Figure out how much to read, the length is fixed so make sure its
    // correct
    addr = ntohl(p_msg->message.data_l[GIGE_READMEM_REQ_ADDR_L]);

    // This is the number of bytes the caller wants returned. If we don't
    // have enough we will just nul fill.
    num_bytes = ntohs(p_msg->message.data_w[GIGE_READMEM_REQ_COUNT_W]);
    if (num_bytes > GIGE_READMEM_B_LEN)
        num_bytes = GIGE_READMEM_B_LEN;
    memset(p_msg->message.data_b, 0, FCP_MSG_DATA_SIZE);

//    fi_log(LOG_DEBUG, "(%s): num_bytes:%d addr:0x%x", __FUNCTION__, num_bytes, addr);

    // Return address asked for
    p_msg->message.data_l[GIGE_READMEM_ACK_ADDR_L] = htonl(addr);

    // see if we are busy and can't process the request
    if (FiImageCorrection::fpnCalibrateActive())
    {
        status = GEV_STATUS_BUSY;
    }
    else if (addr == GIGE_REG_HISTO_RGB)
    {
        // If the position is 0 then we need to get the histogram from the fpga.
        if (histogram_rgb_pos == 0)
        {
            if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_SNAPSHOT)
            {
                fi_log(LOG_ERR, "(%s) can't get rgb histogram in SNAPSHOT state", __FUNCTION__);
                status = GEV_STATUS_ERROR;
            }
            else if (!fpga_ptr->ipmRgbHistogramRead(p_histogram_rgb))
            {
                fi_log(LOG_ERR, "(%s) can't retrieve rgb histogram from fpga, histogram will be empty", __FUNCTION__);
                status = GEV_STATUS_ERROR;
            }
        }

        // Only continue if we're not in error.
        if (status == GEV_STATUS_SUCCESS)
        {
            // Figure out how many bytes are left in the histogram.
            // Make sure that we don't copy more than we are allowed to.
            num_left = histogram_rgb_size - histogram_rgb_pos;
            if (num_left >= num_bytes)
                num_to_copy = num_bytes;
            else
                num_to_copy = num_left;

            // Sanity check - make sure we have some number of bytes to
            // copy and a buffer to copy from.
            if ((num_to_copy > 0) && p_histogram_rgb)
            {
                uint8_t *dest_ptr = (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);
                uint8_t *src_ptr = &p_histogram_rgb[ histogram_rgb_pos ];

                memcpy(dest_ptr, src_ptr, num_to_copy);
                dump_mem(dest_ptr, 16);
                histogram_rgb_pos += num_to_copy;
            }
        }
    }
    else if (addr == GIGE_REG_CONFIGURATION)
    {
        int read_count;

        read_count = read_config((uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]), num_bytes);
        if (read_count < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to read config file (%d).", __FUNCTION__, read_count);
            status = GEV_STATUS_ERROR;
        }
    }
    else if (addr == GIGE_REG_LUT_DATA)
    {
        int read_count;

        read_count = read_lut_data((uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]),
                                   num_bytes);
        if (read_count < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to read LUT data file (%d).", __FUNCTION__, read_count);
            status = GEV_STATUS_ERROR;
        }
    }
    else if (addr == GIGE_REG_FRAME_METADATA)
    {
        int read_count;

        read_count = p_partmgr->getActiveFramePFMD((uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]));
        if (read_count <= 0)
        {
            fi_log(LOG_ERR, "%s: Unable to read frame metadata (%d).", __FUNCTION__, read_count);
            status = GEV_STATUS_ERROR;
        }
    }
    else if (addr == GIGE_REG_FRAME_PFMD)
    {
        // Make sure we have at least one queued request.
        if (pfmd_frame_requests.size() > 0)
        {
            int read_count;

            read_count = p_partmgr->getFramePFMD(pfmd_frame_requests.front(), (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]));
            pfmd_frame_requests.pop_front();
            if (read_count <= 0)
            {
                fi_log(LOG_ERR, "%s: Unable to read frame metadata (%d).", __FUNCTION__, read_count);
                status = GEV_STATUS_ERROR;
            }
        }
    }
    else if (addr == GIGE_REG_FILEPATH)
    {
        const char  *xml_tree;

        // Figure out how many bytes we'll transfer.
        num_left = storage::get_xml_tree_len() - xml_tree_raw_pos;
        if (num_left >= num_bytes)
            num_to_copy = num_bytes;
        else
            num_to_copy = num_left;

        // Get a pointer to our data.
        xml_tree = storage::get_xml_tree_ptr();

        if ((num_to_copy > 0) && xml_tree)
        {
            uint8_t *dest_ptr = (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);
            const uint8_t *src_ptr = (const uint8_t *)&xml_tree[ xml_tree_raw_pos ];

            memcpy(dest_ptr, src_ptr, num_to_copy);
            xml_tree_raw_pos += num_to_copy;
        }
    }
    else if (addr == GIGE_REG_ACTIVE_METADATA)
    {
        uint8_t *dest_ptr = (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);
        std::unique_ptr<pm_metadata> metadata_ptr = p_partmgr->getActivePartMetadata();
        uint8_t *src_ptr = (uint8_t *)metadata_ptr.get();

        // Sanity check.
        if (src_ptr == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to read metadata!", __FUNCTION__);
            status = GEV_STATUS_ERROR;
        }
        else
        {
            // Check to see if we need to reset the metadata position value.
            if (metadata_pos >= sizeof(pm_metadata))
                metadata_pos = 0;

            // Figure out how many bytes are left in the metadata.
            // Make sure that we don't copy more than we are allowed to.
            num_left = sizeof(pm_metadata) - metadata_pos;
            if (num_left >= num_bytes)
                num_to_copy = num_bytes;
            else
                num_to_copy = num_left;

            // Bump our source pointer by our position.
            src_ptr += metadata_pos;

            // Copy over the data.
            memcpy(dest_ptr, src_ptr, num_to_copy);

            // Update our position.
            metadata_pos += num_to_copy;
        }
    }
    else if (addr == GIGE_REG_XML_METADATA)
    {
        char path_buf[PATH_MAX];
        uint8_t *dest_ptr = (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);

        // Make sure our request count is an even multiple of 32-bits.
        num_bytes = ((num_bytes + 3) & ((uint32_t)~3));

        // Make sure the requested amount fits our buffer size.
        if (num_bytes > GIGE_MAX_METADATA_BLOCK)
            num_bytes = GIGE_MAX_METADATA_BLOCK;

        // Clear our buffer.
        memset(dest_ptr, 0, num_bytes);

        // Load the file into our buffer.
        snprintf(path_buf, PATH_MAX, "%s/%s", get_home_path(), DOWNLOAD_XML_METADATA);
        if (load_file(path_buf, dest_ptr, num_bytes, &metadata_pos) < 0)
        {
            fi_log(LOG_ERR, "%s: Unable to copy the metadata.", __FUNCTION__);
            status = GEV_STATUS_ERROR;
        }
    }
    else if (addr == GIGE_REG_MARKER_LIST)
    {
        uint8_t *dest_ptr = (uint8_t *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);

        // Make sure our request count is an even multiple of 32-bits.
        num_bytes = ((num_bytes + 3) & ((uint32_t)~3));

        // Make sure the requested amount fits our buffer size.
        if (num_bytes > GIGE_MAX_METADATA_BLOCK)
            num_bytes = GIGE_MAX_METADATA_BLOCK;

        // Clear our buffer.
        memset(dest_ptr, 0, num_bytes);

        if ((p_marker_list != NULL) && (marker_list_size > 0))
        {
            if (marker_list_size - marker_list_pos > 0)
            {
                if ((marker_list_size - marker_list_pos) < num_bytes)
                    num_bytes = marker_list_size - marker_list_pos;

                memcpy(dest_ptr, p_marker_list + marker_list_pos, num_bytes);
                marker_list_pos += num_bytes;
            }
            else
                num_bytes = 0;
        }
        else
            num_bytes = 0;
    }
    else if (addr == GIGE_REG_SENSOR_VER)
    {
        if (!get_sensor_ver(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), GIGE_SENSOR_VER_LEN))
        {
            fi_log(LOG_ERR, "(%s) can't retrieve sensor version",
                   __FUNCTION__);
            status = GEV_STATUS_ERROR;
        }
    }
    else if (addr == GIGE_REG_USER_DEF_NAME)
    {
        char  new_name[GIGE_REG_USER_DEF_NAME_B_LEN];

        // For this register we always want to get the name from the system
        memset(new_name, 0, sizeof(new_name));
        CsNetwork::getHostname(new_name, sizeof(new_name));
        if (num_bytes > GIGE_REG_USER_DEF_NAME_B_LEN)
            num_bytes = GIGE_REG_USER_DEF_NAME_B_LEN;
        strncpy(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), new_name, GIGE_REG_USER_DEF_NAME_B_LEN);
    }

    else if (addr == GIGE_REG_WIFI_SSID)
    {
        char  new_ssid[GIGE_WIFI_SSID_MAX];
        memset(new_ssid, 0, sizeof(new_ssid));
        CsNetwork::getSsidFromConf(new_ssid, GIGE_WIFI_SSID_MAX);

        if (num_bytes > GIGE_WIFI_SSID_MAX)
            num_bytes = GIGE_WIFI_SSID_MAX;
        strncpy(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), new_ssid, GIGE_WIFI_SSID_MAX);
    }
    else if (addr == GIGE_REG_WIFI_PASSPHRASE)
    {
        char  new_pw[GIGE_WIFI_PASSPHRASE_MAX];
        memset(new_pw, 0, sizeof(new_pw));
        CsNetwork::getWifiPassphraseFromConf(new_pw, GIGE_WIFI_PASSPHRASE_MAX);

        if (num_bytes > GIGE_WIFI_PASSPHRASE_MAX)
            num_bytes = GIGE_WIFI_PASSPHRASE_MAX;
        strncpy(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), new_pw, GIGE_WIFI_PASSPHRASE_MAX);
    }
    else if (addr == GIGE_REG_WIFI_SCAN_RESULTS)
    {
        char  *new_scan = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = GIGE_WIFI_SCAN_RESULTS_READ_SIZE;
        CsNetwork::getWifiScanResults(new_scan, num_bytes);
    }
    else if (addr == GIGE_REG_MEDIA_JSON)
    {
        char  *media_info = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = storage::get_media_json(media_info, GIGE_MEDIA_JSON_MAX_LEN);
    }
    else if (addr == GIGE_REG_FFLR_JSON)
    {
        char  *save_info = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = imgSaveWorker::getInstance()->getLastFflrSaveInfo(save_info, GIGE_FFLR_JSON_LEN);
    }
    else if (addr == GIGE_REG_LOAD_FILE_NAME)
    {
        char  *file_ptr = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);

        // Set the number of bytes we're talking about.
        num_bytes = GIGE_LOAD_FILE_NAME_MAX;

        // Get the file name we'll be loading from.
        imgLoadWorker::getInstance()->getFilename(file_ptr, GIGE_LOAD_FILE_NAME_MAX);
    }
    else if (addr == GIGE_REG_COMMENT)
    {
        char *dest_ptr = (char *) & (p_msg->message.data_b[ GIGE_READMEM_ACK_DATA_B ]);

        if (num_bytes > GIGE_COMMENT_MAX_LEN)
            num_bytes = GIGE_COMMENT_MAX_LEN;

        p_partmgr->get_comment(dest_ptr, num_bytes);
        num_bytes = (uint16_t)strlen(dest_ptr) + 1;
    }
    else if (addr == GIGE_REG_NTP_SERVER)
    {
        // Retrieve the NTP server name.
        get_ntp_server(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), GIGE_NTP_SERVER_LEN);
    }
    else if (addr == GIGE_REG_TIMEZONE)
    {
        // Retrieve the NTP server name.
        get_timezone(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), GIGE_TIMEZONE_LEN);
    }
    else if (addr == GIGE_REG_TELEMETRY_JSON)
    {
        char  *telemetry_info = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = telemetry::get_json(telemetry_info, GIGE_TELEMETRY_JSON_LEN);
    }
    else if (addr == GIGE_REG_LENS_INFO_JSON)
    {
        char  *lens_info = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = mft_info_json(lens_info, GIGE_LENS_INFO_JSON_LEN);
        if (num_bytes == 0)
            status = GEV_STATUS_BUSY;
    }
    else if (addr == GIGE_REG_LENS_STATUS_JSON)
    {
        char  *lens_status = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = mft_status_json(lens_status, GIGE_LENS_STATUS_JSON_LEN);
        if (num_bytes == 0)
            status = GEV_STATUS_BUSY;
    }
    else if (addr == GIGE_REG_VTRIG_SETTINGS_JSON)
    {
        // Pass the request on.
        char  *vtrig_info = &(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]);
        num_bytes = sensor_ptr->vtrig_info_json(vtrig_info, GIGE_VTRIG_SETTINGS_JSON_LEN);
        if (num_bytes == 0)
            status = GEV_STATUS_BUSY;
    }
    else if (addr == GIGE_REG_PEEKPOKE_JSON)
    {
        uint32_t msg_bytes = GIGE_PEEKPOKE_JSON_LEN;

        // Pass the request on.
        status = fcp_handle_peekpoke(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), msg_bytes, true);
        if (status == GEV_STATUS_SUCCESS)
            num_bytes = (uint16_t)msg_bytes;
    }
    else if (regs.validate_id(addr))
    {
        GIGE_REG    *p_gige_reg;

        // Get a pointer to our register.
        p_gige_reg = (GIGE_REG *)regs[ addr ];
        if (p_gige_reg != NULL)
        {
            // This is a regular register read. First thing to do is verify that the
            // register really exists. If so we can process the request. Otherwise
            // it will just fall through to the error condition.
            if (strlen((char *)p_gige_reg->reg_val_b) < num_bytes)
                strncpy(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), (char *)p_gige_reg->reg_val_b, GIGE_READMEM_B_LEN);
            else
                strncpy(&(p_msg->message.data_b[GIGE_READMEM_ACK_DATA_B]), (char *)p_gige_reg->reg_val_b, num_bytes);
            p_msg->message.data_b[num_bytes + GIGE_READMEM_ACK_DATA_B - 1] = 0;
        }
    }
    else
    {
        // Don't need to do anything for this one except log an error and return
        // the error code.
        fi_log(LOG_NOTICE, "%s: invalid register address addr:0x%x", __FUNCTION__, addr);
        status = GEV_STATUS_INVALID_ADDRESS;
    }

    // Add in the length of the address
    num_bytes += sizeof(addr); // length includes the address

    // Now send the response
    p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
    p_msg->message.header.resp.command = htons(GIGE_MSG_READMEM_ACK);
    p_msg->message.header.resp.status = htons(status);

    // Now fill in the length
    if (status == GEV_STATUS_SUCCESS)
        p_msg->message.header.resp.length = htons(num_bytes);
    else
        p_msg->message.header.resp.length = 0;

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // Now send the response
    p_worker->sendMsg(p_msg, sizeof(p_msg->message.header) + num_bytes);

    PDEBUG(PDEBUG_HIGH, "(%s): ending", __FUNCTION__);
    return 0;
}

/*!
 * @brief Client has requested to write a chunk of memory. This is usually
 * used to modify a string register.
 *
 * @param p_msg - the message received for processing
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_writemem(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    // Get class pointers.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    Sensor  *sensor_ptr = CameraServer::get_sensor();

    uint16_t  length = 0;
    uint32_t  addr = 0;
    uint32_t  status = GEV_STATUS_SUCCESS;
    uint32_t  num_bytes = 0;
    csWorker   *p_worker;
    GIGE_REG  *p_gige_reg;

    PDEBUG(PDEBUG_TRACE, "%s: starting", __FUNCTION__);
    if (!p_msg)
        return -1;

    if (!validate_access(p_msg))
    {
        status = GEV_STATUS_ACCESS_DENIED;
        goto writemem_send_ack;
    }

    // Figure out how much to read, the length is fixed so make sure its
    // correct
    length = ntohs(p_msg->message.header.req.length);
    addr = ntohl(p_msg->message.data_l[GIGE_WRITEMEM_REQ_ADDR_L]);
    PDEBUG(PDEBUG_LOW, "%s: length:%d addr:0x%x", __FUNCTION__, length, addr);
    num_bytes = length - sizeof(addr);

    // if the length is invalid, make it valid.
    if (num_bytes > GIGE_READMEM_B_LEN)
        num_bytes = GIGE_READMEM_B_LEN;

    // see if we are busy with other tasks.
    if (FiImageCorrection::fpnCalibrateActive())
    {
        status = GEV_STATUS_BUSY;
    }
    else if (addr == GIGE_REG_USER_DEF_NAME)
    {
        char  new_name[GIGE_REG_USER_DEF_NAME_B_LEN + 1];

        // User defined name is not in the register list because we are getting and
        // updating the linux system.
        // Copy it to a temporary buffer.
        memset(new_name, 0, sizeof(new_name));
        strncpy(new_name, &(p_msg->message.data_b[GIGE_WRITEMEM_REQ_DATA_B]), GIGE_REG_USER_DEF_NAME_B_LEN);
        PDEBUG(PDEBUG_LOW, "%s: new camera name copied", __FUNCTION__);

        // Now export it to the Linux system
        if (CsNetwork::exportCameraName(new_name) != 0)
            status = GEV_STATUS_ERROR;
    }
    else if (addr == GIGE_REG_CONFIGURATION)
    {
        if (write_config((uint8_t *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes))
            status = GEV_STATUS_ERROR;
    }
    else if (addr == GIGE_REG_LUT_DATA)
    {
        if (write_lut_data((uint8_t *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes))
            status = GEV_STATUS_ERROR;
    }
    else if (addr == GIGE_REG_COMMENT)
    {
        PartMgr *p_partmgr = CameraServer::get_partmgr();

        // Make sure what we got is null-terminated and not too long.
        if (num_bytes > GIGE_COMMENT_MAX_LEN)
            num_bytes = GIGE_COMMENT_MAX_LEN;
        p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B + num_bytes - 1 ] = 0;

        // Set the comment for all of the partitions.
        p_partmgr->set_comment((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]));
    }
    else if (addr == GIGE_REG_FILEPATH)
    {
        // Reset our read back position every time we set a new file path.
        xml_tree_raw_pos = 0;

        // Make sure the delete task is IDLE.
        if (storage::operation_in_progress())
        {
            // Return an error indicating we're busy.
            status = GEV_STATUS_BUSY;
        }
        else
        {
            // Set the file name we'll be loading from.
            storage::save_path(&(p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), GIGE_FILEPATH_LEN);
        }
    }
    else if (addr == GIGE_REG_STREAM_FILEPATH)
    {
        // Set the file name/path we'll be streaming from.
        fsp_set_save_path(&(p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), GIGE_STREAM_FILEPATH_LEN);
    }
    else if (addr == GIGE_REG_LOAD_FILE_NAME)
    {
        // Set the file name we'll be loading from.
        imgLoadWorker::getInstance()->setFilename(&(p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), GIGE_LOAD_FILE_NAME_MAX);
    }
    else if (addr == GIGE_REG_NTP_SERVER)
    {
        char  server_name[ GIGE_NTP_SERVER_LEN + 1 ];

        // Save it to a local name to ensure proper length and format.
        memset(server_name, 0, GIGE_NTP_SERVER_LEN + 1);
        if (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ] != 0)
            strncpy(server_name, &(p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), GIGE_NTP_SERVER_LEN);

        // Update the server information.
        fcp_update_ntp_server(server_name);
    }
    else if (addr == GIGE_REG_TIMEZONE)
    {
        char  timezone[ GIGE_TIMEZONE_LEN + 1 ];

        // Save it to a local name to ensure proper length and format.
        memset(timezone, 0, GIGE_TIMEZONE_LEN + 1);
        if (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ] != 0)
            strncpy(timezone, &(p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]),
                    GIGE_TIMEZONE_LEN);

        // Update the server information.
        status = fcp_update_timezone(timezone);
    }
    else if (addr == GIGE_REG_MA_JSON)
    {
        // Pass the request on.
        status = storage::media_action((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_FA_JSON)
    {
        // Reset our read back position every time we set a new file path.
        xml_tree_raw_pos = 0;

        // Pass on the request.
        status = storage::start_json_fa((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_SAVE_JSON)
    {
        // Pass the request on.
        status = fcp_parse_save_json((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_UPDATE_JSON)
    {
        // Pass the request on.
        status = fcp_parse_update_json((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_LENS_SETTINGS_JSON)
    {
        // Pass the request on.
        status = fcp_parse_lens_json((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_VTRIG_SETTINGS_JSON)
    {
        // Pass the request on.
        status = sensor_ptr->vtrig_control((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_PRNU_JSON)
    {
        // Pass the request on.
        status = FiImageCorrection::prnuSetup((const char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes);
    }
    else if (addr == GIGE_REG_PEEKPOKE_JSON)
    {
        // Pass the request on.
        status = fcp_handle_peekpoke((char *) & (p_msg->message.data_b[ GIGE_WRITEMEM_REQ_DATA_B ]), num_bytes, false);
    }
    else if ((p_gige_reg = ((GIGE_REG *)regs[ addr ])) != NULL)
    {
        // See if the address is a valid register, if not do an error.
        // Register address checks out, verify length if not fix it.
        if (num_bytes >= p_gige_reg->reg_size)
            num_bytes = p_gige_reg->reg_size;

        strncpy((char *)p_gige_reg->reg_val_b, &(p_msg->message.data_b[GIGE_WRITEMEM_REQ_DATA_B]),
                num_bytes);
        p_gige_reg->reg_val_b[num_bytes - 1] = 0;
        PDEBUG(PDEBUG_LOW, "(%s): updated register addr:0x%x num_bytes:%d newval:%s",
               __FUNCTION__, addr, num_bytes, (const char *)p_gige_reg->reg_val_b);
    }
    else
    {
        // Don't need to do anything for this one except log an error and return
        // the error code.
        fi_log(LOG_NOTICE, "(%s): invalid register address addr:0x%x",
               __FUNCTION__, addr);
        status = GEV_STATUS_INVALID_ADDRESS;
    }

    // Now send the response
writemem_send_ack:
    p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
    p_msg->message.header.resp.command = htons(GIGE_MSG_WRITEMEM_ACK);
    p_msg->message.header.resp.status = htons(status);

    // fill in the length
    if (status == GEV_STATUS_SUCCESS)
        length = GIGE_WRITEMEM_ACK_B_LEN; // for the index (num_regs)
    else
        length = 0;
    p_msg->message.header.resp.length = htons(length);
    p_msg->message.data_w[GIGE_WRITEMEM_ACK_COUNT_W] = htons(num_bytes);

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // Now send the response
    p_worker->sendMsg(p_msg, sizeof(p_msg->message.header) + length);

    PDEBUG(PDEBUG_HIGH, "(%s)gige_handle_writemem: ending", __FUNCTION__);
    return 0;
}

/*!
 * @brief Client has requested us to retransmit a message already sent. The request
 * should include the request id of the message. The old message can
 * be read from a circular buffer of messages send.
 *
 * TODO - Not yet implemented
 *
 * @param p_msg - the message received for processing
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_packetresend(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)  // the message as it arrived
{
    (void)p_msg;

    PDEBUG(PDEBUG_HIGH, "(%s) starting", __FUNCTION__);
    return 0;  // FORNOW - not implemented.
}

int FCP_LOCAL::handle_noncached_write(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_addr, uint32_t reg_value)
{
    // Get a pointer to our support classes.
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();
    CameraServer    *cs_ptr = CameraServer::get_cs();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    int     status;
    uint8_t camera_state;

    // Get the camera state for use later.
    camera_state = CameraServer::get_camera_state();

    // Assume success.
    status = GEV_STATUS_SUCCESS;

    // Handle registers which are not in the table because they are not saved.
    switch (reg_addr)
    {
    case GIGE_REG_HOLD:
        status = do_hold_operation(reg_value);
        break;
    case GIGE_REG_CS_REGS:
        // Flush our registers to disk
        if (reg_value == GIGE_CS_REGS_FLUSH)
            regs.save();
        else if ((reg_value == GIGE_CS_REGS_LOAD) && (camera_state == GIGE_CAMERA_STATE_IDLE))
        {
            // Restore the base register set.
            regs.init_cache();
            regs.load();
            fcp_validate_registers();

            // Now that we are up and happy, restore the fpn state;
            // FIXME Why on earth do we need to do this?
            // restore_fpn();

            // Time to restart the camera now.
            CameraServer::restart();
        }
        else
        {
            status = GEV_STATUS_WRONG_CONFIG;
        }
        break;
    case GIGE_REG_RESET_TO_FACTORY:
        // need to set all registers back to their factory defaults.
        if (camera_state == GIGE_CAMERA_STATE_IDLE)
        {
            // Remember that we're in a factory reset.
            CameraServer::factory_reset(true);

            // Delete everything in settings, (but not if NFS mounted, it might break things!).
            // On a restart the settings will be restore from the R/O version in NAND.
            if (!CsNetwork::isNfsMount())
            {
                std::string cmd;

                // Clear out the settings folder. When we come back from resetting, we
                // will rebuild using default values.
                cmd = fmt::format("/bin/rm -rf {0}/.config/Fastec/settings/*", get_home_path());
                fi_log(LOG_DEBUG, "%s: GIGE_REG_RESET_TO_FACTOR [%s]", __PRETTY_FUNCTION__, cmd.c_str());
                system(cmd.c_str());
            }
            else
            {
                fi_log(LOG_WARNING, "%s: NFS mounted, therefore not deleting ~/.config/Fastec/settings/* !", __FUNCTION__);
            }

            // Time to restart the camera now.
            CameraServer::restart();
        }
        else
        {
            status = GEV_STATUS_WRONG_CONFIG;
        }
        break;
    case GIGE_REG_FILE_ACTION:
        if (!storage::start_action(reg_value))
            status = GEV_STATUS_BUSY;
        else if (reg_value == GIGE_FA_LUT_LISTING)
        {
            // Reset our read back position because there is no write to GIGE_REG_FILEPATH
            // to do it as in the GIGE_FA_LISTING case.
            xml_tree_raw_pos = 0;
        }
        break;
    case GIGE_REG_ACTIVE_FRAME_NUM:
        if (!p_partmgr->setActiveFrameNum(reg_value))
            status = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_FRAME_PFMD:
        if (reg_value < p_partmgr->getFrameCount())
            pfmd_frame_requests.push_back(reg_value);
        else
            status = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_EVENT_NOTIFY:
        if (!update_event(p_msg, (reg_value == 0)))
            status = GEV_STATUS_INVALID_ADDRESS;
        break;
    case GIGE_REG_GENERATE_RGB_HISTO:
        if (!fpga_ptr->ipmRgbHistoGenerate(reg_value))
            status = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_GENERATE_CUSTOM_WB:
        if (!cs_ptr->generate_custom_white_balance(reg_value))
            status = GEV_STATUS_WRONG_CONFIG;
        break;
    case GIGE_REG_LUT_DATA:
        if ((reg_value == 0) || (reg_value > GIGE_LUT_MAX))
            status = GEV_STATUS_INVALID_PARAMETER;
        else
        {
            lut_data_pos = 0;
            lut_data_id = reg_value;
        }
        break;
    case GIGE_REG_UPDATE_METADATA:
        metadata_pos = 0;
        if (!p_partmgr->updateActiveMetadata())
            status = GEV_STATUS_WRONG_CONFIG;
        break;
    case GIGE_REG_PART_STATE:
        p_partmgr->setPartitionState(p_partmgr->getActivePartition(), reg_value);
        break;
    case GIGE_REG_CHANGE_CAMERA_STATE:
        if (!CameraServer::user_set_camera_state(reg_value))
            status = GEV_STATUS_ERROR;
        break;
    case GIGE_REG_RESET_CAMERA:
        if (reg_value == GIGE_RESET_POWER_DOWN)
        {
            // Make sure the NVME drive is unmounted.
            storage::force_umount_ssd();

            // Power off the system.
            CameraServer::shutdown();
        }
        else if (reg_value == GIGE_RESET_RESTART)
        {
            // Restart the camera.
            CameraServer::restart();
        }
        else
        {
            // Reboot the camera.
            CameraServer::reboot();
        }
        break;
    case GIGE_REG_STREAM_RETRY:
        // Pass the information on to the streamer.
        fsp_set_download_file_retry(reg_value);
        break;
    case GIGE_REG_TIMEOFDAY:
        status = update_tod(reg_value);
        if (status != GEV_STATUS_SUCCESS)
            fi_log(LOG_INFO, "%s: Can't update time of day.", __FUNCTION__);
        break;
    case GIGE_REG_SAVE_STATE:
        // We write GIGE_SAVE_STATE_ERR to this register to clear an error status.
        if (reg_value == GIGE_SAVE_STATE_ERR)
        {
            fpga_ptr->ltrSetSaveError(0);
            imgSaveWorker::getInstance()->clearError();
        }
        // But if we write zero, it means we want to cancel a save operation.
        // Make sure the operation is running before we do that.
        else if ((reg_value == 0) && imgSaveWorker::getInstance()->isRunning())
        {
            AutoSave *as_mgr;

            // Get our autosave manager class.
            as_mgr = CameraServer::get_autosave();
            if (as_mgr->isRunning())
            {
                // Tell the autosave manager to stop.
                as_mgr->stop();
            }
            else
            {
                // We can tell the image save class directly.
                imgSaveWorker::getInstance()->stop();
            }
        }
        break;
    case GIGE_REG_TRUNCATE:
        status = truncate_from(reg_value);
        break;
    case GIGE_REG_LOAD:
        status = update_load(reg_value);
        if (status != GEV_STATUS_SUCCESS)
            fi_log(LOG_INFO, "%s: GIGE_REG_LOAD value 0x%x is wrong", __FUNCTION__, reg_value);
        break;
    case GIGE_REG_ACTIVE_METADATA:
        metadata_pos = 0;
        break;
    case GIGE_REG_XML_METADATA:
        // Make sure the image saver isn't already running.
        if (imgSaveWorker::getInstance()->isRunning())
        {
            // We are BUSY!
            status = GEV_STATUS_BUSY;
        }
        else
        {
            // Reset the file offset for the metadata.
            metadata_pos = 0;

            // Save the XML file.
            imgSaveWorker::getInstance()->setDestInfo(GIGE_IMAGE_SAVE_XML);
            imgSaveWorker::getInstance()->start();
        }
        break;
    case GIGE_REG_MARKER_LIST:
        status = update_marker_list(reg_value);
        if (status != GEV_STATUS_SUCCESS)
            fi_log(LOG_INFO, "%s: Can't update marker list.", __FUNCTION__);
        break;
    case GIGE_REG_FFLR_JSON:
        imgSaveWorker::getInstance()->resetLastFflrSaveInfo();
        break;

    case GIGE_REG_FAN_CONTROL:
        fan_control(reg_value == GIGE_FAN_DISABLE);
        break;

    default:
        // An unexpected address. Ask the FPGA to poke it.
        if (!fpga_ptr->regWrite(reg_addr, reg_value))
        {
            fi_log(LOG_ERR, "%s: Unable to write unexpected address [0x%x]!", __FUNCTION__, reg_addr);
            status = GEV_STATUS_ERROR;
        }
        break;
    }

    return status;
}

/*!
 * @brief Client has requested to write a register. Only the primary application
 *    can write a register.
 *
 * @param p_msg The message received for processing.
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::handle_writereg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int     i;
    int     num_regs;
    int     length;
    int     status;
    int     err_reg;
    uint32_t    reg_addr;
    uint32_t    reg_value_n;
    uint32_t    reg_value;

    // Make sure we got a message.
    if (p_msg == NULL)
    {
        fi_log(LOG_WARNING, "%s: Null message?", __FUNCTION__);
        return -1;
    }

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // Figure out how many registers the user would like to write
    length = ntohs(p_msg->message.header.req.length);
    num_regs = length / (FCP_REG_SIZE + FCP_REG_DATA_SIZE);   // each register is, 4 bytes addr, 4 bytes data.

    // Assume success.
    err_reg = -1;
    status = GEV_STATUS_SUCCESS;

    // See if the requestor has access to write the registers.
    if (!validate_access(p_msg))
    {
        // no access, so short-circuit the follow for-loop by setting num_regs to 0
        status = GEV_STATUS_ACCESS_DENIED;
        num_regs = 0;
        goto writereg_bail;
    }

    // Now do the work
    for (i = 0; i < num_regs; i++)
    {
        // Get the address the client wants changed.
        reg_addr = ntohl(p_msg->message.data_l[ i * 2 ]);
        reg_value_n = p_msg->message.data_l[(i * 2) + 1 ];
        reg_value = ntohl(reg_value_n);

        PDEBUG(PDEBUG_HIGH, "%s: len:%d num:%d addr:0x%x val_N:0x%x val_H:0x%x",
               __FUNCTION__, length, num_regs, reg_addr, reg_value_n, ntohl(reg_value_n));

        // If we are busy doing something which can't be interrupted then bail.
        if (FiImageCorrection::fpnCalibrateActive())
        {
            status = GEV_STATUS_BUSY;
            err_reg = i;
            break;
        }

        // If we are holding some settings until all have been set up, just copy
        // the new value into our hold.
        if (holding)
        {
            bool is_holdable = true;

            switch (reg_addr)
            {
            case GIGE_REG_OFFSET_X:
                hold.startx = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.OFFSET_X=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_OFFSET_Y:
                hold.starty = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.OFFSET_Y=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_WIDTH:
                hold.width = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.WIDTH=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_HEIGHT:
                hold.height = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.HEIGHT=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_FRAME_RATE:
                hold.frame_rate = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.FRAME_RATE=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_SHUTTER_SPEED:
                hold.shutter_speed = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.SHUTTER_SPEED=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_BIT_MODE:
                hold.bit_mode = (bitmode_t)reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.BIT_MODE=%u", __FUNCTION__, reg_value);
                break;
            case GIGE_REG_SENSOR_OPTIONS:
                hold.sensor_opts = reg_value;
                PDEBUG(PDEBUG_GIGEWRITE, "%s: hold.SENSOR_OPTIONS=%u", __FUNCTION__, reg_value);
                break;
            default:
                is_holdable = false;
                break;
            }
            // Writes to holdable registers succeed so just go on to the next register.
            if (is_holdable)
                continue;
        }

        // Check to see if the register is in our cache. Some are not because
        // they go direct to the hardware.
        if (!regs.validate_id(reg_addr))
        {
            // The register is not in our cache, so handle it separately.
            status = handle_noncached_write(p_msg, reg_addr, reg_value);
            if (status != GEV_STATUS_SUCCESS)
            {
                err_reg = i;
                fi_log(LOG_WARNING, "%s: non-cached write failed addr=0x%x val=0x%x status=0x%x!",
                       __FUNCTION__, reg_addr, reg_value, status);
                break;
            }

            // Go back around for the next one.
            continue;
        }

        // Handle registers which are in the table.
        switch (reg_addr)
        {
        default:
            // Make sure we're dealing with binary (not string or memory) registers.
            if (regs[ reg_addr ].reg_type() == REG_TYPE_BIN)
            {
                // Make sure we can write this register.
                if (regs[ reg_addr ].reg_access() == REG_ACCESS_RO)
                {
                    status = GEV_STATUS_ACCESS_DENIED;
                    fi_log(LOG_INFO, "(%s): attempt to modify ro register reg_addr:0x%x",
                           __FUNCTION__, reg_addr);
                    err_reg = i;
                    break;
                }

                // If we are NOT supposed to swap the register bytes
                // then we need to undo it.
                if (regs[ reg_addr ].reg_swap_mode() != REG_SWAP_HTONL)
                    reg_value = reg_value_n;

                // Is there an update function registered?
                if (regs[ reg_addr ].update_is_registered() == NULL)
                {
                    PDEBUG(PDEBUG_GIGEWRITE, "%s: 0x%04x:0x%08x=%u", __FUNCTION__, reg_addr,
                           reg_value, reg_value);

                    // No update registered, so just save the value.
                    regs[ reg_addr ].set_value(reg_value);
                }
                else
                {
                    // Update the register.
                    status = regs[ reg_addr ].reg_update(p_msg, reg_value);
                    if (status != GEV_STATUS_SUCCESS)
                    {
                        err_reg = i;
                        fi_log(LOG_ERR, "%s: update failed writing 0x%x to reg 0x%x status=0x%x",
                               __FUNCTION__, reg_value, reg_addr, status);
                        goto writereg_bail;
                    }
                }
            }
            else
            {
                fi_log(LOG_INFO, "%s: reg_addr:0x%x not a binary register.", __FUNCTION__, reg_addr);
                err_reg = i;
                status = GEV_STATUS_INVALID_ADDRESS;
            }
            break;
        }
    }

writereg_bail:
    // Now send the response, if the caller wants one.
    if (p_msg->message.header.req.flag & GIGE_FLAG_MUST_ACK_BE)
    {
        p_msg->message.header.resp.ack_id = p_msg->message.header.req.req_id;
        p_msg->message.header.resp.command = htons(GIGE_MSG_WRITEREG_ACK);
        p_msg->message.header.resp.status = htons(status);

        // Now fill in the length
        length = FCP_REG_DATA_SIZE; // for the index (num_regs)
        p_msg->message.header.resp.length = htons(length);
        if (err_reg >= 0)
        {
            p_msg->message.data_w[1] = htons(err_reg);
            fi_log(LOG_ERR, "(%s) returning error status:0x%x index:%d",
                   __FUNCTION__, status, err_reg);
        }
        else
        {
            p_msg->message.data_w[1] = htons(num_regs);
            // PDEBUG(PDEBUG_HIGH, "(%s) returning success status:0x%x index:%d",
            //       __FUNCTION__, status, num_regs);
        }

        // Get our worker pointer.
        csWorker   *p_worker;
        p_worker = CameraServer::get_worker();
        if (p_worker == NULL)
            fi_log(LOG_ERR, "%s: worker pointer is NULL!", __PRETTY_FUNCTION__);
        else
            p_worker->sendMsg(p_msg, sizeof(p_msg->message.header) + length);
    }

    // PDEBUG(PDEBUG_HIGH, "(%s) ending", __FUNCTION__);
    return 0;
}

//-----------------------------------------------------------------------------
// Global Functions
//-----------------------------------------------------------------------------

/*!
 * @brief Update our autosave control.
 *
 * @param p_msg     Not used.
 * @param reg_value New autosave setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_autosave_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t    settable_bits;
    uint32_t    old_reg;
    AutoSave    *as_mgr;
    FiFPGA  *fpga_ptr;

    PDEBUG(PDEBUG_GIGEWRITE, "%s: 0x%08x", __FUNCTION__, reg_value);
    LOG("%s: 0x%08x", __FUNCTION__, reg_value);

    // Get our FPGA object.
    fpga_ptr = CameraServer::get_fpga();

    // Get our autosave manager class.
    as_mgr = CameraServer::get_autosave();

    // Get the current settings.
    old_reg = regs[ GIGE_REG_AUTOSAVE ].get_value();

    // Can't go into Autosave if we're in LTR.
    if ((reg_value & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        if ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            // No can do.
            return GEV_STATUS_WRONG_CONFIG;
        }
    }
    else
    {
        // Were we enabled and now we want to disable?
        if ((old_reg & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
        {
            // Reset the capture FIFO.
            LOG("%s: Resetting the capture FIFO", __PRETTY_FUNCTION__);
            fpga_ptr->icmResetCaptureFIFO();
        }
    }

    // Not all of the bits are settable via GigE.
    settable_bits = GIGE_AUTOSAVE_ENABLE | GIGE_AUTOSAVE_USER_ENABLE;
    old_reg &= ~settable_bits;
    old_reg |= (reg_value & settable_bits);

    // We could be turning off autosave while doing an overlapped save. When this happens,
    // the Autosave manager will not rearm but will continue saving. However, as_mgr will
    // not know to update the OVERLAP status until the current save completes. So we will go
    // ahead and clear OVERLAP so the GUIs will know to keep progress dialogs up.
    if ((reg_value & GIGE_AUTOSAVE_ENABLE) == 0)
        old_reg &= ~GIGE_AUTOSAVE_OVERLAP;

    // Save the updated register settings.
    regs[ GIGE_REG_AUTOSAVE ].set_value(old_reg);

    // If we are enabling autosave, then make sure the autosave manager is reset,
    // which will update the register's OVERLAP and total/avail counts.
    if ((reg_value & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
        as_mgr->reset();
    else if ((reg_value & GIGE_AUTOSAVE_QUEUE_UPDATE) == (uint32_t)GIGE_AUTOSAVE_QUEUE_UPDATE)
        as_mgr->queueUpdate();

    // Let folks know it's changed.
    old_reg = regs[ GIGE_REG_AUTOSAVE ].get_value();
    // TODO: Get the autosave manager to update the bit when the running state changes.
    old_reg |= (CameraServer::get_autosave()->isRunning() ? GIGE_AUTOSAVE_SAVING : 0);
    LOG("%s[%d]: notifying GIGE_REG_AUTOSAVE=0x%08x", __PRETTY_FUNCTION__, __LINE__, old_reg);
    fcp_notify_apps(GIGE_REG_AUTOSAVE, old_reg);

    // If autosave is enabled, make sure autoplay and arm-in overwrite are disabled, sending a notification on change.
    if ((reg_value & GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        uint32_t    prefs;

        prefs = regs[ GIGE_REG_PREFERENCES ].get_value();
        if ((prefs & (GIGE_PREF_AUTOPLAY | GIGE_PREF_ARM_IN_OVERWRITE)) != 0)
        {
            prefs &= ~(GIGE_PREF_AUTOPLAY | GIGE_PREF_ARM_IN_OVERWRITE);
            regs[ GIGE_REG_PREFERENCES ].set_value(prefs);
            LOG("%s[%d]: notifying GIGE_REG_AUTOSAVE=0x%08x", __PRETTY_FUNCTION__, __LINE__, prefs);
            fcp_notify_apps(GIGE_REG_AUTOSAVE, prefs);
        }
    }

    // All good.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update our session length register
 *
 * This is the way a UI changes the number of memory partitions.
 * We don't support setting the number of partitions directly, only the session length.
 *
 * @param p_msg     Not used.
 * @param reg_value New stream control setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_session_length_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get support class pointers
    CameraServer    *cs_ptr = CameraServer::get_cs();
    AutoSave        *as_mgr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();
    uint32_t        session_count;
    uint8_t     camera_state;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // Only allowed in IDLE or LIVE_VIDEO state.
    camera_state = CameraServer::get_camera_state();
    if ((camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) && (camera_state != GIGE_CAMERA_STATE_IDLE))
    {
        fi_log(LOG_DEBUG, "%s: Wrong state [%d]!!!", __PRETTY_FUNCTION__, camera_state);
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // Make sure the value is valid.
    if ((reg_value == 0) || (reg_value > p_partmgr->getSessionLengthMax()))
    {
        fi_log(LOG_DEBUG, "%s: reg_value wrong. max=%d", __PRETTY_FUNCTION__, p_partmgr->getSessionLengthMax());
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // Reserve space for the gain table.
    p_partmgr->reserveGainTableSpace(true);

    regs[ GIGE_REG_SESSION_LENGTH ].set_value(reg_value);
    p_partmgr->layoutMemoryFromRegs();

    // If we are using pixel FPN, load the correction tables into DDR2 and the IPM.
    // Update the register value if the CT is non-optimal.
    cs_ptr->load_ct_tables();
    cs_ptr->refresh_ct_status();

    // Set up the autosave class for the new session length. It just needs the session count.
    session_count = p_partmgr->getNumPartitions();
    as_mgr = CameraServer::get_autosave();
    as_mgr->init(session_count);

    // Notify anyone who cares about the change.
    fcp_notify_apps(GIGE_REG_SESSION_LENGTH, reg_value);

    // Make sure the trigger position is in range, if it is specified by frame number.
    fcp_limit_trigger_position(p_partmgr->getFrameCount() - 1);

    // Adjust the save limits so we don't present funny numbers the
    // next time the user sets up autosave.
    fcp_reset_save_limits(p_partmgr->getFrameCount() - 1);

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Setup all the sensor registers which need to be set when
 *    the white balance is set to one of the preset values.
 *
 * @param wb_value  The preset white balance that we want set.
 *
 * @returns 0 if there were no errors
 */
int fcp_set_wb_preset(uint32_t wb_value)
{
    // Get a pointer to our support classes.
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    CameraServer *cs_ptr = CameraServer::get_cs();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_GIGEWRITE, "%s: wb_val:%d", __FUNCTION__, wb_value);

    // If the sensor is not a color sensor then don't go doing any of the preset
    // white balance changes.
    if (!sensor_ptr->is_sensor_color())
    {
        fi_log(LOG_DEBUG, "%s: Not color sensor.", __FUNCTION__);
        return 0; // just ignore the request
    }

    // If the white balance is one of the presets, those need to be restored here.
    // If it's a custom white balance then those will be restored with the individual
    // registers.
    if (wb_value <= GIGE_WHITE_BAL_CUSTOM)
    {
        const WBAL_GAINS *wb_ptr;

        // Retrieve our gain pointer.
        wb_ptr = get_wb(wb_value);
        if (wb_ptr == NULL)
        {
            fi_log(LOG_DEBUG, "%s: wb_value (0x%x) is invalid!", __FUNCTION__, wb_value);
            return 0;
        }

        fpga_ptr->ipmSetWbRed(wb_ptr->red);
        regs[ GIGE_REG_WHITE_BALANCE_RED ].set_value(wb_ptr->red);
        fpga_ptr->ipmSetWbGreen(wb_ptr->green);
        regs[ GIGE_REG_WHITE_BALANCE_GRN ].set_value(wb_ptr->green);
        fpga_ptr->ipmSetWbBlue(wb_ptr->blue);
        regs[ GIGE_REG_WHITE_BALANCE_BLU ].set_value(wb_ptr->blue);

        fpga_ptr->ipmSetMatrixCoeff1Red(wb_ptr->matrix_red1);
        regs[ GIGE_REG_MATRIX_COEFF1_RED ].set_value(wb_ptr->matrix_red1);
        fpga_ptr->ipmSetMatrixCoeff2Red(wb_ptr->matrix_red2);
        regs[ GIGE_REG_MATRIX_COEFF2_RED ].set_value(wb_ptr->matrix_red2);
        fpga_ptr->ipmSetMatrixCoeff3Red(wb_ptr->matrix_red3);
        regs[ GIGE_REG_MATRIX_COEFF3_RED ].set_value(wb_ptr->matrix_red3);

        fpga_ptr->ipmSetMatrixCoeff1Green(wb_ptr->matrix_grn1);
        regs[ GIGE_REG_MATRIX_COEFF1_GRN ].set_value(wb_ptr->matrix_grn1);
        fpga_ptr->ipmSetMatrixCoeff2Green(wb_ptr->matrix_grn2);
        regs[ GIGE_REG_MATRIX_COEFF2_GRN ].set_value(wb_ptr->matrix_grn2);
        fpga_ptr->ipmSetMatrixCoeff3Green(wb_ptr->matrix_grn3);
        regs[ GIGE_REG_MATRIX_COEFF3_GRN ].set_value(wb_ptr->matrix_grn3);

        fpga_ptr->ipmSetMatrixCoeff1Blue(wb_ptr->matrix_blu1);
        regs[ GIGE_REG_MATRIX_COEFF1_BLU ].set_value(wb_ptr->matrix_blu1);
        fpga_ptr->ipmSetMatrixCoeff2Blue(wb_ptr->matrix_blu2);
        regs[ GIGE_REG_MATRIX_COEFF2_BLU ].set_value(wb_ptr->matrix_blu2);
        fpga_ptr->ipmSetMatrixCoeff3Blue(wb_ptr->matrix_blu3);
        regs[ GIGE_REG_MATRIX_COEFF3_BLU ].set_value(wb_ptr->matrix_blu3);

        fpga_ptr->ipmSetDcOffsetRed(wb_ptr->dc_offset_red);
        fpga_ptr->ipmSetDcOffsetGreen(wb_ptr->dc_offset_grn);
        fpga_ptr->ipmSetDcOffsetBlue(wb_ptr->dc_offset_blu);

        cs_ptr->set_as_shot_red(wb_ptr->as_shot_red);
        cs_ptr->set_as_shot_grn(wb_ptr->as_shot_grn);
        cs_ptr->set_as_shot_blu(wb_ptr->as_shot_blu);
    }

    return 0;
}

/*!
 * @brief Update our sync-out settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New sync-out setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_syncout_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int status;
    FiFPGA *fpga_ptr;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    (void)p_msg;

    fi_log(LOG_DEBUG, "%s(0x%08x)", __PRETTY_FUNCTION__, reg_value);

    // Get a pointer to our support classes.
    fpga_ptr = CameraServer::get_fpga();

    // Set our sync-out.
    if (fpga_ptr->syncOutConfigure(reg_value))
    {
        // If we're just being enabled/disabled don't loose the
        // reset of the register settings.
        if ((reg_value == GIGE_SYNC_OUT_DISABLE) || (reg_value == GIGE_SYNC_OUT_ENABLE))
        {
            uint32_t    old_value;

            old_value = regs[ GIGE_REG_SYNC_OUT_SELECT ].get_value();
            old_value = (old_value & ~GIGE_SYNC_OUT_ENABLE) | reg_value;

            // Don't save the channel information on a disable!
            if (reg_value == GIGE_SYNC_OUT_DISABLE)
                old_value &= ~(GIGE_SYNC_OUT_CHANNEL_MASK << GIGE_SYNC_OUT_CHANNEL_SHIFT);

            regs[ GIGE_REG_SYNC_OUT_SELECT ].set_value(old_value);
        }
        else
        {
            regs[ GIGE_REG_SYNC_OUT_SELECT ].set_value(reg_value);
        }

        // All good.
        status = GEV_STATUS_SUCCESS;
    }
    else
    {
        status = GEV_STATUS_WRONG_CONFIG;
    }

    // Return our status.
    return status;
}

/*!
 * @brief Update our sync delay value.
 *
 * @param p_msg     Not used.
 * @param reg_value New sync-in setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_sync_delay(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    FiFPGA  *fpga_ptr;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpga_ptr = CameraServer::get_fpga();

    // Save the new value.
    regs[ GIGE_REG_SYNC_DELAY ].set_value(reg_value);

    // Have the FPGA write it out.
    fpga_ptr->syncInDelay(reg_value);

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update our sync-in settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New sync-in setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_syncin_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    int     status;
    FiFPGA  *fpga_ptr;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%08x)", __FUNCTION__, reg_value);

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpga_ptr = CameraServer::get_fpga();

    // See if we can enable sync-in.
    if (fpga_ptr->syncInEnable(reg_value))
    {
        uint32_t    sync_value;

        switch (reg_value)
        {
        case GIGE_SYNC_TYPE_INTERNAL:
        case GIGE_SYNC_TYPE_EXTERNAL:
        case GIGE_SYNC_TYPE_IRIGB:
        case GIGE_SYNC_TYPE_GPS:
            // If we're just being enabled/disabled don't loose the
            // reset of the register settings.
            sync_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
            sync_value &= ~GIGE_SYNC_TYPE_MASK;

            // ALM_TODO: Don't save the channel information on a disable!
            if (reg_value == GIGE_SYNC_TYPE_INTERNAL)
                sync_value &= ~(0xF << GIGE_SYNC_IN_CHANNEL_SHIFT);

            sync_value |= reg_value;
            regs[ GIGE_REG_SYNC_IN_ENABLE ].set_value(sync_value);
            break;
        default:
            // More than just the enable was being set.
            regs[ GIGE_REG_SYNC_IN_ENABLE ].set_value(reg_value);
            break;
        }

        // All is good.
        status = GEV_STATUS_SUCCESS;
    }
    else
    {
        status = GEV_STATUS_WRONG_CONFIG;
    }

    // Return our status.
    return status;
}

/*!
 * @brief Update the matrix coefficient settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_matrix_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    FiFPGA  *fpga_ptr;
    uint32_t    reg_addr;
    int     status;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpga_ptr = CameraServer::get_fpga();

    // Get our address - this is a common routine for a number of
    // parameters, so read the "reg_addr" to determine which one
    // we're interested in specifically.
    reg_addr = regs.reg_addr();
    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%04x:0x%04x)", __FUNCTION__, reg_addr, reg_value);

    // Assume failure.
    status = GEV_STATUS_INVALID_PARAMETER;

    // Process the specific register.
    switch (reg_addr)
    {
    case GIGE_REG_MATRIX_COEFF1_RED:
        if (fpga_ptr->ipmSetMatrixCoeff1Red(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF2_RED:
        if (fpga_ptr->ipmSetMatrixCoeff2Red(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF3_RED:
        if (fpga_ptr->ipmSetMatrixCoeff3Red(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF1_GRN:
        if (fpga_ptr->ipmSetMatrixCoeff1Green(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF2_GRN:
        if (fpga_ptr->ipmSetMatrixCoeff2Green(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF3_GRN:
        if (fpga_ptr->ipmSetMatrixCoeff3Green(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF1_BLU:
        if (fpga_ptr->ipmSetMatrixCoeff1Blue(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF2_BLU:
        if (fpga_ptr->ipmSetMatrixCoeff2Blue(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_MATRIX_COEFF3_BLU:
        if (fpga_ptr->ipmSetMatrixCoeff3Blue(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_WHITE_BALANCE_RED:
        if (fpga_ptr->ipmSetWbRed(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_WHITE_BALANCE_GRN:
        if (fpga_ptr->ipmSetWbGreen(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    case GIGE_REG_WHITE_BALANCE_BLU:
        if (fpga_ptr->ipmSetWbBlue(reg_value))
            status = GEV_STATUS_SUCCESS;
        break;
    default:
        break;
    }

    // Update our register if we set it OK.
    if (status == GEV_STATUS_SUCCESS)
        regs[ reg_addr ] = reg_value;

    // Return our status.
    return status;
}

int fcp_marker_filter_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t old_value;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%08x)", __FUNCTION__, reg_value);

    // If no change, nothing to do.
    old_value = regs[ GIGE_REG_MARKER_FILTER ].get_value();
    if (old_value == reg_value)
        return GEV_STATUS_SUCCESS;

    // If we are currently running, the markers cannot be changed.
    if (imgSaveWorker::getInstance()->isRunning() && imgSaveWorker::getInstance()->isDoingMarkers())
    {
        if ((reg_value & GIGE_MARKER_FILTER_ENABLE) == (uint32_t)GIGE_MARKER_FILTER_ENABLE)
        {
            fi_log(LOG_ERR, "%s: Marker search is running, can't set new filter!", __FUNCTION__);
            return GEV_STATUS_BUSY;
        }
        else
        {
            // If we need to do it, this is where we would stop it.
            fi_log(LOG_DEBUG, "%s: not stopping marker search ", __FUNCTION__);
            return GEV_STATUS_ERROR;
        }
    }

    marker_list_pos = 0;
    marker_list_size = -1;

    regs[ GIGE_REG_MARKER_FILTER ].set_value(reg_value);
    fcp_notify_apps(GIGE_REG_MARKER_FILTER, reg_value);

    return GEV_STATUS_SUCCESS;
}

int FCP_LOCAL::update_marker_list(uint32_t reg_value)
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();
    uint32_t        marker_filter;

    PDEBUG(PDEBUG_GIGEWRITE, "%s: 0x%08x", __FUNCTION__, reg_value);

    // First see if we have anything to do.
    marker_filter = regs[ GIGE_REG_MARKER_FILTER ].get_value();

    // If not enabled, we are either being asked to stop or to discard
    // what we already did.
    if ((marker_filter & GIGE_MARKER_FILTER_ENABLE) == 0)
    {
        if (imgSaveWorker::getInstance()->isRunning() && imgSaveWorker::getInstance()->isDoingMarkers())
            imgSaveWorker::getInstance()->stop();

        marker_list_pos = 0;
        marker_list_size = -1;

        return GEV_STATUS_SUCCESS;
    }

    // Make sure the image saver and the GOE aren't running because either one
    // of those could use resources we need.
    if (imgSaveWorker::getInstance()->isRunning())
        return GEV_STATUS_BUSY; // We are BUSY!

    // Reset the file offset and the size for the marker data.
    marker_list_pos = 0;
    marker_list_size = -1;

    // Might need to allocate our buffer.
    if (p_marker_list == NULL)
    {
        int ma_status;
        ma_status = posix_memalign((void **)&p_marker_list, DMA_PAGE_ALIGN, MARKER_LIST_BUF_SIZE);
        if ((p_marker_list == NULL) || (ma_status != 0))
            return GEV_STATUS_ERROR;
    }

    // For now we will just run this in the context of the imgsave worker,
    // but this needs refactoring to better manage resource ownership.
    imgSaveWorker::getInstance()->setPartNum(p_partmgr->getActivePartition());
    imgSaveWorker::getInstance()->requestMarkers(marker_filter, p_marker_list, MARKER_LIST_BUF_SIZE,
                                                 &marker_list_size);
    imgSaveWorker::getInstance()->start();

    return GEV_STATUS_SUCCESS;
}

uint32_t FCP_LOCAL::fcp_parse_lens_json(const char *settings_str, int str_length)
{
    Json::Value settings_info;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    bool parsingSuccessful;
    uint32_t status;

    // Initialize our status.
    status = GEV_STATUS_INVALID_PARAMETER;

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(settings_str, settings_str + str_length, &settings_info, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, settings_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        goto parse_bail;
    }
//    LOG("%s: Lens settings='%s'", __PRETTY_FUNCTION__, settings_info.toStyledString().c_str());

    // See if we have a new power setting.
    if (!settings_info["power"].isNull())
    {
        // Ask the controller to change the lens power setting.
        if (!mft_set_power(settings_info["power"].asBool()))
        {
            fi_log(LOG_WARNING, "%s: mft_set_power() failed using [%d] as a power.", __PRETTY_FUNCTION__, settings_info["power"].asBool());
            goto parse_bail;
        }
    }

    // See if we have a new f-number.
    if (!settings_info["f_number"].isNull())
    {
        // Ask the controller to change the aperture.
        if (!mft_set_aperture(settings_info["f_number"].asFloat()))
        {
            fi_log(LOG_WARNING, "%s: mft_set_aperture() failed using [%f] as an f-number.", __PRETTY_FUNCTION__, settings_info["f_number"].asFloat());
            goto parse_bail;
        }
    }

    // See if we have a new focus.
    if (!settings_info["focus_dist"].isNull())
    {
        try
        {
            // Ask the controller to change the focus.
            if (!mft_set_focus(settings_info["focus_dist"].asInt()))
            {
                fi_log(LOG_WARNING, "%s: mft_set_focus() failed using [0x%x] as a focus_dist.", __PRETTY_FUNCTION__, settings_info["focus_dist"].asInt());
                goto parse_bail;
            }
        }
        catch (...)
        {
            fi_log(LOG_DEBUG, "%s: converting JSON for 'focus_dist' faulted!", __PRETTY_FUNCTION__);
            goto parse_bail;
        }
    }

    // See if we have new zoom information.
    if (!settings_info["zoom_enc"].isNull())
    {
        // Ask the controller to change the zoom.
        if (!mft_set_zoom(settings_info["zoom_enc"].asInt()))
        {
            fi_log(LOG_WARNING, "%s: mft_set_zoom() failed using [0x%x] as a zoom_enc.", __PRETTY_FUNCTION__, settings_info["zoom_enc"].asUInt());
            goto parse_bail;
        }
    }

    // See if we have a new focus.
    if (!settings_info["autofocus"].isNull())
    {
        // Ask the controller to run the autofocus algorithm.
        mft_autofocus();
    }

    // So far, all is well.
    status = GEV_STATUS_SUCCESS;

parse_bail:
    // Return our status.
    return status;
}

/*!
 * @brief The client sends the camera server all the parameters required for saving a video encoded
 *        in a JSON formatted string (save_str). This routine parses that string and start the save
 *        process.
 *
 * To start the save process, the client does a "memory write" to GIGE_REG_SAVE_JSON. The possible
 * entries in the JSON encoded string are:
 *
 * Name                 | Description
 * -------------------- | -----------
 * tag                  | GIGE_REG_IMAGE_TAG value
 * frame_start          | GIGE_REG_SAVE_START_NUM value
 * frame_end            | GIGE_REG_SAVE_END_NUM value
 * format               | GIGE_REG_STREAM_FORMAT value
 * dstpath              | Destination root path for saving images (i.e. "/media/FastecSSD")
 * dst_info             | GIGE_REG_IMAGE_DESTINATION value
 * save_metadata_cfg    | Flag "true" or "false" to indicate partition metadata CFG file save.
 * write_buffers        | Flag "true" or "false" to indicate if dirty partitions should be written.
 * save_metadata        | Flag "true" or "false" to indicate partition metadata save.
 *
 * Once all the parameters have been saved into the appropriate places (GIGE_REG_CACHE, or the
 * imgSaveWorker object), the fcp_image_dest_update() routine kicks off the actual
 * save.
 *
 * @param save_str      JSON encoded string holding the save parameters.
 * @param str_length    Number of characters in the string.
 * @return One of GEV_STATUS_ERROR, GEV_STATUS_INVALID_PARAMETER, GEV_STATUS_BUSY,
 *         or GEV_STATUS_SUCCESS.
 */
uint32_t FCP_LOCAL::fcp_parse_save_json(const char *save_str, int str_length)
{
    uint32_t status = GEV_STATUS_INVALID_PARAMETER;

    // Parse the save settings.
    if (imgSaveWorker::getInstance()->parseSaveSettings(save_str, str_length))
    {
        // Are we in FFLR recovery mode?
        if (imgSaveWorker::getInstance()->fflrRecovery())
        {
            // Make sure the save thread isn't already running.
            if (imgSaveWorker::getInstance()->isRunning())
            {
                // This won't work.
                fi_log(LOG_ERR, "%s: Cannot allow multiple instances of imgsave to run!", __PRETTY_FUNCTION__);
                status = GEV_STATUS_ACCESS_DENIED;
            }
            else
            {
                // Start the saver thread to recover the data as requested.
                LOG("%s: FFLR session recovery starting.", __PRETTY_FUNCTION__);
                imgSaveWorker::getInstance()->start();
                status = GEV_STATUS_BUSY;
            }
        }
        else
        {
            // If everything went well, save the JSON data so we can use it as an initializer.
            imgSaveWorker::getInstance()->updateSaveSettings(save_str);
            status = GEV_STATUS_SUCCESS;
        }
    }

    // Return our status.
    return status;
}

uint32_t FCP_LOCAL::fcp_parse_update_json(const char *update_str, int str_length)
{
    Json::Value update_info;
    Json::CharReaderBuilder builder;
    std::unique_ptr<Json::CharReader> reader;
    std::string errInfo;
    std::string device_path;
    std::string package_name;
    std::string update_version;
    bool use_network;
    bool parsingSuccessful;
    uint32_t status;

    // Initialize our status.
    status = GEV_STATUS_INVALID_PARAMETER;

    // Create our reader.
    reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
    if (reader.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
        return GEV_STATUS_ERROR;
    }

    // Parse the information passed in.
    parsingSuccessful = reader->parse(update_str, update_str + str_length, &update_info, &errInfo);
    if (!parsingSuccessful)
    {
        fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, update_str, str_length);
        fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
        goto parse_bail;
    }
    fi_log(LOG_INFO, "%s: update_info='%s'", __PRETTY_FUNCTION__, update_info.toStyledString().c_str());

    // We better have the network information field.
    if (update_info["net"].isNull())
    {
        fi_log(LOG_DEBUG, "%s: No 'net' parameter!", __PRETTY_FUNCTION__);
        goto parse_bail;
    }
    use_network = update_info["net"].asBool();

    // Get the version information.
    if (update_info["version"].isNull())
    {
        fi_log(LOG_DEBUG, "%s: No 'version' parameter!", __PRETTY_FUNCTION__);
        goto parse_bail;
    }
    update_version = update_info["version"].asString();

    // If we're not using the network, then get the base path
    // for where we're going to find the update package.
    if (!use_network)
    {
        // Sanity check. This needs to be here.
        if (update_info["package"].isNull() || update_info["device"].isNull())
        {
            fi_log(LOG_DEBUG, "%s: No 'package' and/or 'device' parameter!", __PRETTY_FUNCTION__);
            goto parse_bail;
        }

        // Get the path and package information.
        device_path = update_info["device"].asString();
        package_name = update_info["package"].asString();
    }

    // Start our update object.
    update_camera(use_network, update_version, device_path, package_name);

    // So far, all is well.
    status = GEV_STATUS_SUCCESS;

parse_bail:
    // Return our status.
    return status;
}

/*!
 * @brief Update the shared preferences register.
 *
 * @param p_msg     Not used.
 * @param reg_value New preference setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_preferences_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get a pointer to our support class.
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    // The GUIs should not be turning this on if Autosave or LTR is enabled, but
    // we'll just make sure.
    if ((reg_value & (GIGE_PREF_AUTOPLAY | GIGE_PREF_ARM_IN_OVERWRITE)) != 0)
    {
        uint32_t        as_value;

        as_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
        if (((as_value & GIGE_AUTOSAVE_ENABLE) != 0) ||
            ((fpga_ptr->ltrMode() & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE))
        {
            return GEV_STATUS_WRONG_CONFIG;
        }
    }

    // Set the external I/O as requested.
    fpga_ptr->extioSetVoltage((reg_value & GIGE_PREF_EXTIO_5V) == GIGE_PREF_EXTIO_5V);

    // Update the cached value and notify any other clients.
    regs[ GIGE_REG_PREFERENCES ].set_value(reg_value);
    fcp_notify_apps(GIGE_REG_PREFERENCES, reg_value);

    // All done.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update our white balance settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New white balance setting.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_wb_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    int     status;
    FiFPGA  *fpga_ptr;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpga_ptr = CameraServer::get_fpga();

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // If it's one of the preset values, call the routine to set the values in the sensor
    // to what they are supposed to be.
    if (reg_value <= GIGE_WHITE_BAL_CUSTOM)
    {
        fcp_set_wb_preset(reg_value);
        regs[ GIGE_REG_WHITE_BAL ].set_value(reg_value);

        // Special handling for custom - we want to reset the RGB gains too.
        fpga_ptr->ipmSetGainRed(GIGE_DEFAULT_RGB_GAIN);
        regs[ GIGE_REG_RED ].set_value(GIGE_DEFAULT_RGB_GAIN);
        fpga_ptr->ipmSetGainGreen(GIGE_DEFAULT_RGB_GAIN);
        regs[ GIGE_REG_GREEN ].set_value(GIGE_DEFAULT_RGB_GAIN);
        fpga_ptr->ipmSetGainBlue(GIGE_DEFAULT_RGB_GAIN);
        regs[ GIGE_REG_BLUE ].set_value(GIGE_DEFAULT_RGB_GAIN);

        // Only reset the BCG if we're doing custom...
        if (reg_value == GIGE_WHITE_BAL_CUSTOM)
        {
            uint32_t gamma_val;

            // Don't use a custom LUT.
            regs[ GIGE_REG_LUT ].set_value(0);

            // Reset the Brightness, Contrast and Gamma to defaults.
            regs[ GIGE_REG_BRIGHTNESS ].set_value(GIGE_DEFAULT_BRIGHTNESS);
            regs[ GIGE_REG_CONTRAST ].set_value(GIGE_DEFAULT_CONTRAST);
            gamma_val = fpga_ptr->ipmDefaultGamma();
            regs[ GIGE_REG_GAMMA ].set_value((gamma_val << 16) | gamma_val);

            fpga_ptr->ipmSetLutValues(regs[ GIGE_REG_BRIGHTNESS ].get_value(),
                                      regs[ GIGE_REG_CONTRAST ].get_value(), gamma_val, gamma_val);
        }

        status = GEV_STATUS_SUCCESS;
    }
    else
    {
        status = GEV_STATUS_INVALID_PARAMETER;
    }

    // Return our status.
    return status;
}

/*!
 * @brief Update the single-step frame number.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_ss_frame_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    PartMgr         *p_partmgr = CameraServer::get_partmgr();

    fi_log(LOG_DEBUG, "%s(%u) from %s", __FUNCTION__, reg_value,
           (p_msg == nullptr) ? "(internal)" : ((p_msg->msg_type == MSG_TYPE_IPC) ? "IPC" : (p_msg->msg_type == MSG_TYPE_GRPC) ? "GRPC" : "UDP"));

    // This only works if we're in REVIEW.
    if (CameraServer::get_camera_state() != GIGE_CAMERA_STATE_REVIEW)
    {
        fi_log(LOG_ERR, "%s: Wrong state [%d]!", __FUNCTION__, CameraServer::get_camera_state());
        return GEV_STATUS_WRONG_CONFIG;
    }

    // Error if requested to step to a frame out of range, but no message
    // because this error can be used to find out when the end is reached.
    if (reg_value >= p_partmgr->getNumFrames())
    {
        fi_log(LOG_DEBUG, "%s: request > num_frames(%d)", __PRETTY_FUNCTION__, p_partmgr->getNumFrames());
        return GEV_STATUS_ERROR;
    }

    // Only update if there's a change.
    if (regs[ GIGE_REG_CSI_RVW_FRAME_NUM ].get_value() != reg_value)
    {
        // Save the new value.
        regs[ GIGE_REG_CSI_RVW_FRAME_NUM ].set_value(reg_value);

        // Notify folks that the frame number has changed.
        fcp_notify_apps(GIGE_REG_CSI_RVW_FRAME_NUM, reg_value);
    }

    // Return our status.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update the low-light settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_low_light_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    CameraServer    *cs_ptr = CameraServer::get_cs();

    int     rc = GEV_STATUS_INVALID_PARAMETER;
    bool    rtn;
    FiFPGA  *fpga_ptr;
    uint32_t    reg_addr;
    uint32_t    temp_value;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    fpga_ptr = CameraServer::get_fpga();

    // Get our address - this is a common routine for a number of
    // parameters, so read the "reg_addr" to determine which one
    // we're interested in specifically.
    reg_addr = regs.reg_addr();
    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%04x:%u)", __FUNCTION__, reg_addr, reg_value);

    // We are enabling/disable low light or setting the
    // low light frame rate.
    if (reg_addr == GIGE_REG_LOW_LIGHT_ENABLE)
    {
        temp_value = regs[ GIGE_REG_LOW_LIGHT_SHUTTER ].get_value();

        // Make sure the shutter speed is in range for this sensor. Otherwise the Camera GUI
        // can get a Catch-22 where it can't fix the shutter speed because it must enable it first
        // and it can't enable it because the speed is illegal.
        if ((temp_value < cs_ptr->get_min_shutter_speed()) ||
            (temp_value > cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE)))
        {
            fi_log(LOG_DEBUG, "%s: fixing bad LOW_LIGHT_SHUTTER %u", __FUNCTION__, temp_value);
            temp_value = cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE);
            regs[ GIGE_REG_LOW_LIGHT_SHUTTER ].set_value(reg_value);
        }
        rtn = fpga_ptr->icmSetLowLightShutterSpeed(reg_value, temp_value);
    }
    else
    {
        temp_value = regs[ GIGE_REG_LOW_LIGHT_ENABLE ].get_value();
        rtn = fpga_ptr->icmSetLowLightShutterSpeed(temp_value, reg_value);
    }

    // Update our register if we were successful.
    if (rtn)
    {
        regs[ reg_addr ] = reg_value;
        fcp_notify_apps(reg_addr, reg_value);

        // Return a good status.
        rc = GEV_STATUS_SUCCESS;
    }

    return rc;
}

/*!
 * @brief Update the color settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_color_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int status;
    uint32_t reg_addr;
    uint32_t lut_id;
    uint32_t new_lut_id;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    (void)p_msg;
    status = GEV_STATUS_ERROR;

    // Get our address - this is a common routine for a number of
    // parameters, so read the "reg_addr" to determine which one
    // we're interested in specifically.
    reg_addr = regs.reg_addr();
    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%04x:%u)", __FUNCTION__, reg_addr, reg_value);

    new_lut_id = lut_id = regs[ GIGE_REG_LUT ].get_value() & GIGE_LUT_MASK;

    switch (reg_addr)
    {
    case GIGE_REG_RED:
        if (reg_value <= 0xFFFF)
        {
            fpga_ptr->ipmSetGainRed((uint16_t)reg_value);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_RED);
            status = GEV_STATUS_SUCCESS;
        }
        break;
    case GIGE_REG_GREEN:
        if (reg_value <= 0xFFFF)
        {
            fpga_ptr->ipmSetGainGreen((uint16_t)reg_value);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_GREEN);
            status = GEV_STATUS_SUCCESS;
        }
        break;
    case GIGE_REG_BLUE:
        if (reg_value <= 0xFFFF)
        {
            fpga_ptr->ipmSetGainBlue((uint16_t)reg_value);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_BLUE);
            status = GEV_STATUS_SUCCESS;
        }
        break;
    case GIGE_REG_BRIGHTNESS:
        if (reg_value <= GIGE_MAX_BRIGHTNESS)
        {
            fpga_ptr->ipmSetBrightness(reg_value);
            new_lut_id = GIGE_LUT_NONE;
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_RED);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_GREEN);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_BLUE);
            status = GEV_STATUS_SUCCESS;
        }
        break;
    case GIGE_REG_CONTRAST:
        if (reg_value <= GIGE_MAX_CONTRAST)
        {
            fpga_ptr->ipmSetContrast(reg_value);
            new_lut_id = GIGE_LUT_NONE;
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_RED);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_GREEN);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_BLUE);
            status = GEV_STATUS_SUCCESS;
        }
        break;
    case GIGE_REG_GAMMA:
    {
        uint32_t new_default = (reg_value & GIGE_GAMMA_SET_DEFAULT);
        uint32_t new_gamma = (reg_value & ~GIGE_GAMMA_SET_DEFAULT);

        // Sanity check the new GAMMA value.
        if ((new_gamma >= GIGE_MIN_GAMMA) && (new_gamma <= GIGE_MAX_GAMMA))
        {
            // Are we setting the default or just the gamma?
            if (new_default)
                fpga_ptr->ipmSetDefaultGamma(new_gamma);
            else
                fpga_ptr->ipmSetGamma(new_gamma);

            // Build up the register value (a combination of the default and the current setting).
            reg_value = (fpga_ptr->ipmDefaultGamma() << 16) | fpga_ptr->ipmGamma();
            new_lut_id = GIGE_LUT_NONE;
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_RED);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_GREEN);
            fpga_ptr->ipmSendLutTableToFpga(GIGE_REG_BLUE);
            status = GEV_STATUS_SUCCESS;
        }
    }
    break;
    default:
        status = GEV_STATUS_INVALID_PARAMETER;
        break;
    }

    // If we were successful, update our register.
    if (status == GEV_STATUS_SUCCESS)
        regs[ reg_addr ] = reg_value;

    if (new_lut_id != lut_id)
    {
        regs[ GIGE_REG_LUT ] = new_lut_id;
        fcp_notify_apps(GIGE_REG_LUT, new_lut_id);
    }

    // Return our status.
    return status;
}

/*!
 * @brief Update the FPN settings.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_fpn_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get pointers to our support classes.
    CameraServer    *cs_ptr = CameraServer::get_cs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    uint32_t    syncin_value;
    bool        manual_params = false;

    int retval = GEV_STATUS_SUCCESS;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%x)", __FUNCTION__, reg_value);

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ACCESS_DENIED;
        }
    }

    if ((reg_value & GIGE_FPN_FPGA_FAC_CALIBRATE_MASK) == GIGE_FPN_FPGA_FAC_CALIBRATE_MASK)
    {
        // Remember the "manual" flag but clear it so it doesn't show up in the notification.
        manual_params = (reg_value & GIGE_FPN_CALIBRATE_MANUAL) ? true : false;
        reg_value &= ~GIGE_FPN_CALIBRATE_MANUAL;

        // Lupa wants factory default settings for the factory calibration operation.
        // Lince wants the application to set up everything so the exposure can be adjusted.
        if (sensor_ptr->factory_needs_reset(manual_params))
        {
            // Set the hardware like IDLE state without notifying anyone.
            cs_ptr->set_idle_behavior();

            // Try to force the camera to a known state.  We try to actually reset everything
            // but there might be some remnants of the old implementation, where some GigE
            // registers were copied from current state variables in the Sensor or FiFPGA classes.
            cs_ptr->reset_registers();

            // This uses the current settings, even if they are license-restricted.
            // The calibration will restart capture with the full sensor area if it needs it.
            cs_ptr->set_live_behavior();
        }

        // Since the calibration is a long process, we want to kick it off on a separate thread
        // and return to the caller. An unsolicited message will go out when it's finished.
        // Note that while the calibration is running, the CS will return busy.
        fcp_notify_apps(GIGE_REG_FPN, reg_value);

        // Make sure all streaming is shutdown.
        if (fcp_kill_stream_thread() != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to stop streaming.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }

        // Start the calibration process.
        else if (FiImageCorrection::fpnCalibrate(true, manual_params) != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to start calibration thread.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }
    }
    else if ((reg_value & GIGE_FPN_PRNU_CALIBRATE) == GIGE_FPN_PRNU_CALIBRATE)
    {
        LOG("%s: GIGE_FPN_PRNU_CALIBRATE...", __PRETTY_FUNCTION__);

        // Since the calibration is a long process, we want to kick it off on a separate thread
        // and return to the caller. An unsolicited message will go out when it's finished.
        // Note that while the calibration is running, the CS will return busy.
        fcp_notify_apps(GIGE_REG_FPN, reg_value);

        // Make sure all streaming is shutdown.
        if (fcp_kill_stream_thread() != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to stop streaming.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }

        // Start the calibration process.
        else if (FiImageCorrection::prnuCalibrate() != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to start calibration thread.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }
    }
    else if ((reg_value & GIGE_FPN_FPGA_CALIBRATE_MASK) == GIGE_FPN_FPGA_CALIBRATE_MASK)
    {
        // Remember the "manual" flag but clear it so it doesn't show up in the notification.
        manual_params = (reg_value & GIGE_FPN_CALIBRATE_MANUAL) ? true : false;
        reg_value &= ~GIGE_FPN_CALIBRATE_MANUAL;

        // Since the calibration is a long process, we want to kick it off on a separate thread
        // and return to the caller. An unsolicited message will go out when it's finished.
        // Note that while the calibration is running, the CS will return busy.
        fcp_notify_apps(GIGE_REG_FPN, reg_value);

        // Make sure all streaming is shutdown.
        if (fcp_kill_stream_thread() != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to stop streaming.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }

        else if (FiImageCorrection::fpnCalibrate(false, manual_params) != 0)
        {
            fi_log(LOG_ERR, "%s: Unable to start calibration thread.", __FUNCTION__);
            retval = GEV_STATUS_ERROR;
        }
    }
    else if ((reg_value & GIGE_FPN_FPGA_LOAD_MASK) == GIGE_FPN_FPGA_LOAD_MASK)
    {
        fi_log(LOG_ERR, "%s: UI requested obsolete GIGE_FPN_FPGA_LOAD_MASK", __FUNCTION__);
    }
    else
    {
        // Make sure only we don't have both enables set.
        if ((reg_value & (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK)) ==
            (GIGE_FPN_FPGA_ENABLE_MASK | GIGE_FPN_SENSOR_ENABLE_MASK))
        {
            retval = GEV_STATUS_INVALID_PARAMETER;
            goto fpn_bail1;
        }

        if (!cs_ptr->set_fpn(reg_value))
            retval = GEV_STATUS_ERROR;

        cs_ptr->refresh_ct_status();
    }

fpn_bail1:
    if (retval == GEV_STATUS_INVALID_PARAMETER)
        fi_log(LOG_ERR, "%s: Unexpected register value [0x%x]", __FUNCTION__, reg_value);

    // Return our result.
    return retval;
}

/*!
 * @brief Update the shutter speed aka integration time.
 *
 * The shutter speed can be adjusted without stopping capture.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_shutter_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    CameraServer *cs_ptr = CameraServer::get_cs();
    uint8_t camera_state;
    float frame_rate;
    uint32_t    syncin_value;
    uint32_t    temp;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // Make sure we're in a good state: IDLE, LIVE_VIDEO, ARMED, TRIGGERED are okay
    // for shutter speed changes.
    camera_state = CameraServer::get_camera_state();
    if ((camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) &&
        (camera_state != GIGE_CAMERA_STATE_ARMED) &&
        (camera_state != GIGE_CAMERA_STATE_TRIGGERED))
        return GEV_STATUS_INVALID_PARAMETER;

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ACCESS_DENIED;
        }
    }

    frame_rate = cs_ptr->reg_to_frame_rate(regs[ GIGE_REG_FRAME_RATE ].get_value());
    temp = cs_ptr->get_max_shutter_speed(frame_rate);
    if (reg_value > temp)
    {
        reg_value = temp;
    }
    temp = regs[ GIGE_REG_MIN_SHUTTER_SPEED ].get_value();
    if (reg_value < temp)
    {
        reg_value = temp;
    }

    regs[ GIGE_REG_SHUTTER_SPEED ].set_value(reg_value);
    fpga_ptr->icmSetShutterSpeed(reg_value);
    fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, reg_value);

    // All done.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Change our current selection for LTR mode.
 *
 * @param reg_value New value for the register.
 * @return True on success, false otherwise.
 */
bool FCP_LOCAL::fcp_ltr_mode_select_change(uint32_t reg_value)
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA  *fpga_ptr = CameraServer::get_fpga();
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    PartMgr *p_partmgr = CameraServer::get_partmgr();
    CameraServer *cs_ptr = CameraServer::get_cs();

    uint32_t    old_value;
    bool    was_updated;
    bool    lr_mode_enable;
    bool    fflr_enable;

    // Get the current setting to compare against.
    old_value = regs[GIGE_REG_LTR_MODE].get_value();
    fi_log(LOG_DEBUG, "%s: reg_value=0x%08x", __FUNCTION__, reg_value);

    // Are we enabling either LR or FFLR mode?
    lr_mode_enable = (reg_value & GIGE_LTR_MODE_ENABLE) == GIGE_LTR_MODE_ENABLE;
    fflr_enable = (reg_value & GIGE_LTR_MODE_FFLR) == GIGE_LTR_MODE_FFLR;

    // Are we enabling LR mode?
    if (lr_mode_enable && ((old_value & GIGE_LTR_MODE_ENABLE) != GIGE_LTR_MODE_ENABLE))
    {
        // Make sure the SSD isn't mounted. We can't use storage::umount
        // because it might not even recognize that the drive exists.
        // We simply want to make sure the drive is not mounted into the file system.
        if (!storage::force_umount_ssd())
        {
            fi_log(LOG_DEBUG, "%s: Unable to unmount the SSD.", __PRETTY_FUNCTION__);
            return false;
        }
    }

    // Set the LR mode.
    fpga_ptr->ltrModeSelect(lr_mode_enable);

    // Are we enabling LR mode?
    if (lr_mode_enable && ((old_value & GIGE_LTR_MODE_ENABLE) != GIGE_LTR_MODE_ENABLE))
    {
        LOG("%s: Enabling LR mode", __PRETTY_FUNCTION__);

        // Make sure we  disable FasFireLR mode.
        fpga_ptr->ltrFasFireDisable();

        // Set LR mode in the register cache.
        regs[GIGE_REG_LTR_MODE].set_value((old_value & ~GIGE_LTR_MODE_FFLR) | GIGE_LTR_MODE_ENABLE);
        regs[GIGE_REG_MIN_WIDTH].set_value(LTR_MIN_WIDTH);
        regs[GIGE_REG_MIN_HEIGHT].set_value(LTR_MIN_HEIGHT);

        // Disable autosave mode when entering LTR mode.
        old_value = regs[GIGE_REG_AUTOSAVE].get_value();
        regs[GIGE_REG_AUTOSAVE].set_value(old_value & ~GIGE_AUTOSAVE_ENABLE);

        // Transition the camera server object to LR mode recording.
        cs_ptr->check_ltr_recording();
    }
    else if (fflr_enable && ((old_value & GIGE_LTR_MODE_FFLR) != GIGE_LTR_MODE_FFLR))
    {
        LOG("%s: Enabling FFLR mode", __PRETTY_FUNCTION__);

        // Set FFLR mode in the register cache.
        regs[GIGE_REG_LTR_MODE].set_value((old_value & ~GIGE_LTR_MODE_ENABLE) | GIGE_LTR_MODE_FFLR);
        regs[GIGE_REG_MIN_WIDTH].set_value(sensor_ptr->get_sensor_min_width());
        regs[GIGE_REG_MIN_HEIGHT].set_value(sensor_ptr->get_sensor_min_height());
    }
    else if (!fflr_enable && !lr_mode_enable)
    {
        LOG("%s: Disabling FFLR and LR modes.", __PRETTY_FUNCTION__);

        // Make sure we  disable FasFireLR mode.
        fpga_ptr->ltrFasFireDisable();

        // Clear the bits.
        regs[GIGE_REG_LTR_MODE].set_value(old_value & ~(GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_FFLR | GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC));
        regs[GIGE_REG_MIN_WIDTH].set_value(sensor_ptr->get_sensor_min_width());
        regs[GIGE_REG_MIN_HEIGHT].set_value(sensor_ptr->get_sensor_min_height());
    }

    // See if our changes imply that ROI and its dependencies need to change too.
    was_updated = cs_ptr->make_capture_regs_valid();
    p_partmgr->reserveGainTableSpace(true);

    // We had to make changes, so relayout memory.
    p_partmgr->layoutMemoryFromRegs();

    // If we are using pixel FPN, load the correction tables into memory and the FPGA.
    // We had to do this in case the registers changed or the memory layout changed.
    cs_ptr->load_ct_tables();
    cs_ptr->refresh_ct_status();

    // If we're enabling FFLR we also have to update the autosave mangager.
    if (fflr_enable)
    {
        uint32_t        session_count;
        AutoSave        *as_mgr;

        // Set up the autosave class for the new session length. It just needs the session count.
        session_count = p_partmgr->getNumPartitions();
        as_mgr = CameraServer::get_autosave();
        as_mgr->init(session_count);
    }

    if (was_updated)
    {
        fcp_notify_apps(GIGE_REG_WIDTH,    regs[ GIGE_REG_WIDTH ].get_value());
        fcp_notify_apps(GIGE_REG_HEIGHT,   regs[ GIGE_REG_HEIGHT ].get_value());
        fcp_notify_apps(GIGE_REG_OFFSET_X, regs[ GIGE_REG_OFFSET_X ].get_value());
        fcp_notify_apps(GIGE_REG_OFFSET_Y, regs[ GIGE_REG_OFFSET_Y ].get_value());
        fcp_notify_apps(GIGE_REG_BIT_MODE, regs[ GIGE_REG_BIT_MODE ].get_value());
        fcp_notify_apps(GIGE_REG_FRAME_RATE,    regs[ GIGE_REG_FRAME_RATE ].get_value());
        fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, regs[ GIGE_REG_SHUTTER_SPEED ].get_value());
    }

    // If the trigger position was specified by frame number, we could end up with
    // a value out of range now. Check and maybe repair it.
    fcp_limit_trigger_position(p_partmgr->getFrameCount() - 1);

    // Any time LTR mode changes, set the save range to the full range so it
    // doesn't risk showing values for some previous session length.
    fcp_reset_save_limits(p_partmgr->getFrameCount() - 1);

    // All is well.
    return true;
}

/*!
 * @brief Do startup initialization for LTR mode
 */
void fcp_ltr_mode_init(void)
{
    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    uint32_t        ltr_mode_reg;

    ltr_mode_reg = regs[ GIGE_REG_LTR_MODE ].get_value();
    if ((ltr_mode_reg & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        fpga_ptr->ltrModeSelect(true);

        // Make sure the SSD isn't mounted. We can't use storage::umount
        // because it might not even recognize that the drive exists.
        // We simply want to make sure the drive is not mounted into the file system.
        storage::force_umount_ssd();
    }
    else
    {
        fpga_ptr->ltrModeSelect(false);
    }
}

/*!
 * @brief Check whether any of the current save limits are bad for the current frame_max
 *
 * @param frame_max The maximum frame number. Legal frame numbers are 0 to frame_max inclusive.
 *
 * @return true if the save limits are okay, false if they should be reset.
 */
bool FCP_LOCAL::fcp_save_limits_okay(uint32_t frame_max)
{
    // Get GIGE registers.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (regs[ GIGE_REG_SAVE_START_NUM ].get_value() > regs[ GIGE_REG_SAVE_END_NUM ].get_value())
        return false;

    if (regs[ GIGE_REG_USER_SAVE_START_NUM ].get_value() > regs[ GIGE_REG_USER_SAVE_END_NUM ].get_value())
        return false;

    if ((regs[ GIGE_REG_USER_SAVE_START_NUM ].get_value() > frame_max)
        || (regs[ GIGE_REG_USER_SAVE_END_NUM ].get_value() > frame_max)
        || (regs[ GIGE_REG_SAVE_START_NUM ].get_value() > frame_max)
        || (regs[ GIGE_REG_SAVE_END_NUM ].get_value() > frame_max))
        return false;

    // Special case after factory reset, where start and end are both zero.
    if ((regs[ GIGE_REG_USER_SAVE_START_NUM ].get_value() == 0)
        && (regs[ GIGE_REG_USER_SAVE_END_NUM ].get_value() == 0))
        return false;
    if ((regs[ GIGE_REG_SAVE_START_NUM ].get_value() == 0)
        && (regs[ GIGE_REG_SAVE_END_NUM ].get_value() == 0))
        return false;

    return true;
}

/*!
 * @brief Reset save start and end frame numbers
 *
 * Whenever the frame count changes, make sure our save start and end values
 * are adjusted so we don't show funny defaults in the autosave dialogs.
 *
 * @param frame_max The maximum frame number. Legal frame numbers are 0 to frame_max inclusive.
 */
void fcp_reset_save_limits(uint32_t new_frame_max)
{
    // Get GIGE registers.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    if (regs[ GIGE_REG_USER_SAVE_START_NUM ].get_value() != 0)
    {
        regs[ GIGE_REG_USER_SAVE_START_NUM ].set_value(0);
        fcp_notify_apps(GIGE_REG_USER_SAVE_START_NUM, 0);
    }
    if (regs[ GIGE_REG_SAVE_START_NUM ].get_value() != 0)
    {
        regs[ GIGE_REG_SAVE_START_NUM ].set_value(0);
        fcp_notify_apps(GIGE_REG_SAVE_START_NUM, 0);
    }
    if (regs[ GIGE_REG_USER_SAVE_END_NUM ].get_value() != new_frame_max)
    {
        regs[ GIGE_REG_USER_SAVE_END_NUM ].set_value(new_frame_max);
        fcp_notify_apps(GIGE_REG_USER_SAVE_END_NUM, new_frame_max);
    }
    if (regs[ GIGE_REG_SAVE_END_NUM ].get_value() != new_frame_max)
    {
        regs[ GIGE_REG_SAVE_END_NUM ].set_value(new_frame_max);
        fcp_notify_apps(GIGE_REG_SAVE_END_NUM, new_frame_max);
    }
}

/*!
 * @brief Make sure the trigger position is valid.
 *
 * Whenever the frame count changes, we need to make sure our trigger position is still
 * valid if it was specified by frame number. If it was specified by percent, it is fine as is.
 *
 * If we do change the trigger position, the PartMgr function will also end up moving the
 * save limits to the ends of the range so they don't look funny once they are translated
 * into trigger-relative values.
 *
 * @param frame_max The maximum frame number. Legal frame numbers are 0 to frame_max inclusive.
 */
void fcp_limit_trigger_position(uint32_t new_frame_max)
{
    // Get GIGE registers.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    uint32_t trig_pos;

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    trig_pos = regs[ GIGE_REG_TRIGGER_REQ_POS ].get_value();
    if (((trig_pos & GIGE_TRIGGER_REQ_FRAME_MASK) == GIGE_TRIGGER_REQ_FRAME_MASK) &&
        ((trig_pos & GIGE_TRIGGER_REQ_POS_MASK) > new_frame_max))
    {
        // Don't set the register in advance because the update function will
        // send a notification if it changes.
        PartMgr::setTriggerPos(NULL, 100);
    }
}

/*!
 * @brief Make sure GigE registers are valid.
 *
 * This is used after loading registers to make sure the settings are mutually consistent
 * and compatible with this hardware.
 *
 * The requested trigger position and save limits will be handled once we have laid out memory
 * for the current mode and session length, which could affect our maximum frame count.
 */
void fcp_validate_registers(void)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer    *cs_ptr = CameraServer::get_cs();
    PartMgr     *p_partmgr = CameraServer::get_partmgr();
    const LicManager *p_lic = CameraServer::get_license();
    uint32_t    image_dest;
    uint32_t    reg_value;
    uint32_t    ltr_mode_reg;

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    // Make sure number of partitions is valid at all. We will fix this later to make sure
    // it is consistent with session length.
    // FIXME Maybe we will just get rid of GIGE_REG_NUM_PART because no one seems to use it.
    reg_value = regs[ GIGE_REG_NUM_PART ].get_value();
    if (reg_value == 0)
        reg_value = 1;
    else if (reg_value > GIGE_MAX_PARTITIONS)
        reg_value = GIGE_MAX_PARTITIONS;
    regs[ GIGE_REG_NUM_PART ].set_value(reg_value);

    // Make sure the LTR mode is consistent with the license.
    ltr_mode_reg = regs[ GIGE_REG_LTR_MODE ].get_value();
    if (p_lic && (p_lic->ssdMax() > 0) &&
        ((p_lic->ltrMode() == LIC_MANAGER::LM_LTR_DUAL) || (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_ONLY)))
    {
        // If dual LTR mode is licensed, allow setting it to anything but if only LTR
        // mode is licensed, force it.
        if (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_ONLY)
            ltr_mode_reg |= GIGE_LTR_MODE_ENABLE;
    }
    else
        ltr_mode_reg &= ~GIGE_LTR_MODE_ENABLE;

    // Don't allow LTR mode if no SSD is licensed.
    if (p_lic && (p_lic->ssdMax() <= 0))
        ltr_mode_reg &= ~GIGE_LTR_MODE_ENABLE;

    // Don't allow LTR or FFLR modes if not licensed.
    if (p_lic && (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_NONE))
        ltr_mode_reg = 0;

    // Set the DUAL bit based on the license, not the restored settings.
    else if (p_lic && (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_DUAL))
        ltr_mode_reg |= GIGE_LTR_MODE_DUAL;
    else
        ltr_mode_reg &= ~GIGE_LTR_MODE_DUAL;

    regs[ GIGE_REG_LTR_MODE ].set_value(ltr_mode_reg);

    if ((ltr_mode_reg & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Disable autosave mode when entering LTR mode.
        reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
        regs[ GIGE_REG_AUTOSAVE ].set_value(reg_value & ~GIGE_AUTOSAVE_ENABLE);

        // We only have one partition in LTR mode.
        regs[ GIGE_REG_NUM_PART ].set_value(1);

        // This is okay. PartMgr set the max session length in its class_init.
        // Otherwise we would do it when we layout the memory.
        regs[ GIGE_REG_SESSION_LENGTH ].set_value(p_partmgr->getSessionLengthMax());

        // Disable autoplay and arm-in overwrite (aka autodelete) if we are in LTR mode.
        reg_value = regs[ GIGE_REG_PREFERENCES ].get_value();
        reg_value &= ~(GIGE_PREF_AUTOPLAY | GIGE_PREF_ARM_IN_OVERWRITE);
        regs[ GIGE_REG_PREFERENCES ].set_value(reg_value);

        regs[ GIGE_REG_MAX_FRAME_RATE ].set_value(cs_ptr->get_max_ltr_frame_rate());
    }

    // Now that we know the LTR mode, we can make sure the minimum ROI is set, which
    // depends on both LTR mode and the sensor.
    regs[ GIGE_REG_MIN_WIDTH ].set_value(cs_ptr->get_min_width());
    regs[ GIGE_REG_MIN_HEIGHT ].set_value(cs_ptr->get_min_height());

    // Log the LTR setting because it helps with making sense of camera logs.
    fi_log(LOG_INFO, "%s: LTR lic=%d, max_ssd=%d, setting mode=0x%x\n", __FUNCTION__,
           p_lic ? p_lic->ltrMode() : -1, p_lic ? p_lic->ssdMax() : -1, ltr_mode_reg);

    // Make sure all the capture parameters are mutually consistent, that they
    // don't violate any limits for the current mode, and that they don't violate
    // any license constraints.
    //      GIGE_REG_OFFSET_X
    //      GIGE_REG_OFFSET_Y
    //      GIGE_REG_WIDTH
    //      GIGE_REG_HEIGHT
    //      GIGE_REG_FRAME_RATE
    //      GIGE_REG_SHUTTER_SPEED
    //      GIGE_REG_BIT_MODE
    //      GIGE_REG_SESSION_LENGTH
    cs_ptr->make_capture_regs_valid();

    // Make sure our image destination is a valid combination.
    // For removable media, we don't require that the target drive be mounted,
    // because the user could still insert media before we need to write to it.
    image_dest = regs[ GIGE_REG_IMAGE_DESTINATION ].get_value();

    // Make sure the format is one of our valid formats
    if (((image_dest & GIGE_STREAM_FORMAT_MASK) == 0) ||
        ((image_dest & GIGE_STREAM_FORMAT_MASK) >= GIGE_STREAM_FORMAT_MAX))
    {
        image_dest &= ~(GIGE_STREAM_FORMAT_MASK | GIGE_STREAM_FORMAT_RAW);
        image_dest |= GIGE_STREAM_FORMAT_JPEG;
    }

    // We can't boot up with the destination set to off-camera. It has to be refreshed
    // when the PC app connects.
    if ((image_dest & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA)
    {
        image_dest &= ~GIGE_IMAGE_DEST_MASK;
        image_dest |= GIGE_IMAGE_DEST_SDCARD;

        // And if AUTOSAVE was enabled, just turn it off since we no longer have
        // the off-camera destination.
        reg_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
        regs[ GIGE_REG_AUTOSAVE ].set_value(reg_value & ~GIGE_AUTOSAVE_ENABLE);
    }

    // Make sure the IMAGE_DEST isn't SSD when we are in LTR mode.
    if ((image_dest & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_SSD)
    {
        if ((ltr_mode_reg & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        {
            image_dest &= ~GIGE_IMAGE_DEST_MASK;
            image_dest |= GIGE_IMAGE_DEST_SDCARD;
            fi_log(LOG_DEBUG, "%s: LTR mode, changed IMAGE_DEST to SD card", __FUNCTION__);
        }
        else if (!storage::exists(GIGE_MEDIA_SSD))
        {
            image_dest &= ~GIGE_IMAGE_DEST_MASK;
            image_dest |= GIGE_IMAGE_DEST_SDCARD;
            fi_log(LOG_DEBUG, "%s: no local SSD present, changed IMAGE_DEST to SD card", __FUNCTION__);
        }
    }
    else if ((image_dest & GIGE_IMAGE_DEST_MASK) == 0)
    {
        image_dest |= GIGE_IMAGE_DEST_SDCARD;
    }

    // Update the register.
    regs[GIGE_REG_IMAGE_DESTINATION].set_value(image_dest);

    // Don't allow a BROC length less than the minimum.
    if (regs[ GIGE_REG_BROC_LEN ].get_value() < GIGE_MIN_BROC_LEN)
        regs[ GIGE_REG_BROC_LEN ].set_value(GIGE_MIN_BROC_LEN);

    // If we loaded a configuration file for a different model, the low-light
    // exposure could be out of range for us. If it's illegal, just use the default/max.
    reg_value = regs[ GIGE_REG_LOW_LIGHT_SHUTTER ].get_value();
    if ((reg_value < cs_ptr->get_min_shutter_speed()) ||
        (reg_value > cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE)))
    {
        fi_log(LOG_DEBUG, "%s: bad LOW_LIGHT_SHUTTER %u, using default", __FUNCTION__, reg_value);
        reg_value = cs_ptr->get_max_shutter_speed((float)LOW_LIGHT_FRAME_RATE);
        regs[ GIGE_REG_LOW_LIGHT_SHUTTER ].set_value(reg_value);
    }
}

/*!
 * @brief Make sure GigE registers related to frame counts are valid.
 *
 * We could not validate these limits until the partition manager laid out memory
 * for the current mode and ROI, which determines the number of frames; that in turn
 * determines what save limits are valid.
 *
 * @param frame_max The maximum frame number, where frames are numbered 0 to frame_max inclusive.
 */
void fcp_validate_frame_limits(uint32_t frame_max)
{
    fcp_limit_trigger_position(frame_max);

    // Make sure the save limits are valid.
    if (!fcp_save_limits_okay(frame_max))
        fcp_reset_save_limits(frame_max);
}

/*!
 * @brief Update the LTR_MODE register and repair the consequences
 *    of doing so, such as updating other registers.
 *
 * @note We already checked whether the license allows writing this
 *   register (ltr == LM_LTR_DUAL).
 *   We require that the camera state be IDLE.
 *   We will update other registers that change as a consequence
 *   in case their updated values should be written to the settings
 *   file. This is the update function for GIGE_REG_LTR_MODE.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_ltr_mode_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    const LicManager *p_lic;
    uint32_t    syncin_value;
    uint32_t    old_reg;
    bool    access_denied = false;

    fi_log(LOG_DEBUG, "%s(0x%08x)", __FUNCTION__, reg_value);

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    old_reg = regs[GIGE_REG_LTR_MODE].get_value();

    // Get our license pointer.
    p_lic = CameraServer::get_license();

    // Don't allow changes to the mode if it is not allowed.
    if (p_lic && (p_lic->ltrMode() == LIC_MANAGER::LM_LTR_NONE))
    {
        access_denied = true;
    }

    // The LTR_MODE_ENABLE can be changed only if the license is DUAL.
    else if (p_lic && (p_lic->ltrMode() != LIC_MANAGER::LM_LTR_DUAL))
    {
        if ((old_reg & GIGE_LTR_MODE_ENABLE) != (reg_value & GIGE_LTR_MODE_ENABLE))
            access_denied = true;
    }

    // The DUAL bit can never be changed.
    if ((old_reg & GIGE_LTR_MODE_DUAL) != (reg_value & GIGE_LTR_MODE_DUAL))
        access_denied = true;

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        FiFPGA      *fpga_ptr = CameraServer::get_fpga();

        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            access_denied = true;
        }
    }

    if (access_denied)
    {
        if (p_lic != nullptr)
        {
            fi_log(LOG_INFO, "(%s): denied attempt to change GIGE_REG_LTR_MODE %08x->%08x ltr=%d\n",
                   __FUNCTION__, old_reg, reg_value, p_lic->ltrMode());
        }
        return GEV_STATUS_ACCESS_DENIED;
    }

    // If both bits are set, we understand that to mean BROC and don't call it an error, just
    // set BROC mode. We should not be asked to do this.
    if ((reg_value & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) == (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC))
        reg_value &= ~GIGE_LTR_MODE_ROC;

    // Did the LTR or FFLR mode bits change?
    if ((reg_value & (GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_FFLR)) != (old_reg & (GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_FFLR)))
    {
        LOG("%s: Changing LR or FFLR enables", __PRETTY_FUNCTION__);

        // Only permit LTR mode writes when in IDLE state.
        if (CameraServer::get_camera_state() != GIGE_CAMERA_STATE_IDLE)
            return GEV_STATUS_BUSY;

        // Update the settings and any other registers that change as a consequence.
        if (!fcp_ltr_mode_select_change(reg_value))
            return GEV_STATUS_BUSY;

        // Save our settings.
        regs.save();
    }

    // Handle a change in the ROC/BROC setting when we are already in LTR mode.
    else if (((reg_value & GIGE_LTR_MODE_ROC)  != (old_reg & GIGE_LTR_MODE_ROC)) ||
             ((reg_value & GIGE_LTR_MODE_BROC) != (old_reg & GIGE_LTR_MODE_BROC)))
    {
        PartMgr *p_partmgr = CameraServer::get_partmgr();

        // No need to be in IDLE state to change ROC/BROC but for now we only allow it in one of the LR modes.
        if ((reg_value & (GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_FFLR)) == 0)
        {
            fi_log(LOG_DEBUG, "%s: Can't set ROC/BROC if we're not in an LR mode.", __PRETTY_FUNCTION__);
            return GEV_STATUS_INVALID_PARAMETER;
        }

        // Are in "normal" LR mode?
        if ((reg_value & GIGE_LTR_MODE_ENABLE) == GIGE_LTR_MODE_ENABLE)
        {
            // If we are changing between an append mode (ROC/BROC) and a non-append mode
            // (basic LTR), set the current partition to empty.
            // There are scenarios where we could allow appending onto a non-append mode
            // capture but then we'd have to explain why it works sometimes and not others.
            // Let's just keep it simple.
            if (((old_reg   & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0) !=
                ((reg_value & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0))
            {
                p_partmgr->ltrClearPartition();
                fcp_notify_apps(GIGE_REG_PART_STATE, GIGE_PART_STATE_EMPTY);
            }
        }

        // Save the value and notify apps about the change.
        regs[ GIGE_REG_LTR_MODE ].set_value(reg_value);
        fcp_notify_apps(GIGE_REG_LTR_MODE, reg_value);
    }

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Update the current frame rate.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_frame_rate_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer    *cs_ptr = CameraServer::get_cs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    Sensor      *sensor_ptr = CameraServer::get_sensor();

    int     rc = GEV_STATUS_SUCCESS;
    uint8_t camera_state;
    bitmode_t   bit_mode;
    uint16_t    true_x;
    uint16_t    true_y;
    uint16_t    startx;
    uint16_t    starty;
    uint16_t    width;
    uint16_t    height;
    uint32_t    sensor_opts;
    uint32_t    shutter_speed;
    uint32_t    syncin_value;
    float frame_rate;
    float old_frame_rate;
    float max_frame_rate;
    bool need_sensor_restart;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(%u)", __FUNCTION__, reg_value);

    // Make sure we're in a good state.
    camera_state = CameraServer::get_camera_state();
    if ((camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
        return GEV_STATUS_INVALID_PARAMETER;

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ACCESS_DENIED;
        }
    }

    // Get the current settings so we can see if this is a valid frame rate.
    cs_ptr->get_user_settings(&startx, &starty, &width, &height, &bit_mode, &frame_rate,
                              &shutter_speed, &sensor_opts);
    old_frame_rate = frame_rate;
    frame_rate = cs_ptr->reg_to_frame_rate(reg_value);

    // Make sure this is a valid frame rate for the current ROI. If not, we'll fix it.
    if (!cs_ptr->is_user_roi_valid(startx, starty, width, height, bit_mode, frame_rate,
                                   shutter_speed, sensor_opts, &max_frame_rate))
    {
        // Whoa! the other settings are invalid?!!! Something is badly broken.
        if (approx_equals(max_frame_rate, 0))
        {
            fi_log(LOG_DEBUG, "%s: Bad ROI (%u,%u)%ux%u %u", __FUNCTION__,
                   startx, starty, width, height, bit_mode);
            return GEV_STATUS_INVALID_PARAMETER;
        }

        // The frame rate was too fast so we need to limit it to the max for this ROI.
        if (!cs_ptr->is_user_roi_valid(startx, starty, width, height, bit_mode, max_frame_rate,
                                       cs_ptr->get_max_shutter_speed(max_frame_rate),
                                       sensor_opts, NULL))
        {
            fi_log(LOG_DEBUG, "%s: Bad ROI (%u,%u)%ux%u %u fps=%f", __FUNCTION__,
                   startx, starty, width, height, bit_mode, max_frame_rate);
            return GEV_STATUS_INVALID_PARAMETER;
        }

        frame_rate = max_frame_rate;
    }

    // See if this change will require that we restart the sensor.
    true_x = cs_ptr->get_true_startx(bit_mode, sensor_opts, startx);
    true_y = cs_ptr->get_true_starty(bit_mode, sensor_opts, starty);
    need_sensor_restart = sensor_ptr->frame_rate_needs_restart(old_frame_rate, frame_rate,
                                                               true_x, true_y, width, height,
                                                               bit_mode, sensor_opts);

    // If the sensor needs to restart, handle it just like we were coming off a HOLD or
    // like an ROI change.
    if (need_sensor_restart)
    {
        if (!cs_ptr->modify_capture(startx, starty, width, height, bit_mode, frame_rate,
                                    shutter_speed, sensor_opts))
            rc = GEV_STATUS_ERROR;
    }
    else
    {
        bool    shutter_changed;

        // Make sure the shutter is valid for this frame rate.
        // If not, the time must have been reduced so the new setting should still be valid now
        // before we change the frame rate, so handle it first.
        shutter_changed = cs_ptr->make_shutter_valid(frame_rate, &shutter_speed);
        if (shutter_changed)
        {
            regs[ GIGE_REG_SHUTTER_SPEED ].set_value(shutter_speed);
            fpga_ptr->icmSetShutterSpeed(shutter_speed);
            fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, shutter_speed);
        }

        reg_value = cs_ptr->frame_rate_to_reg(frame_rate);
        regs[ GIGE_REG_FRAME_RATE ].set_value(reg_value);
        fpga_ptr->icmSetFrameRate(frame_rate);
        fcp_notify_apps(GIGE_REG_FRAME_RATE, reg_value);

        // Update our maximum shutter (longest exposure) based on the new framerate.
        shutter_speed = cs_ptr->get_max_shutter_speed(frame_rate);
        regs[ GIGE_REG_MAX_SHUTTER_SPEED ].set_value(shutter_speed);
        fcp_notify_apps(GIGE_REG_MAX_SHUTTER_SPEED, shutter_speed);

        // If we are in low-light mode, we undid our low-light setting so put it back now.
        if (fpga_ptr->icmIsLowLightEnabled())
        {
            fpga_ptr->icmSetFrameRate((float)LOW_LIGHT_FRAME_RATE);
            fpga_ptr->icmSetShutterSpeed(fpga_ptr->icmLowLightShutterSpeed());
        }
    }

    return rc;
}

/*!
 * @brief Make sure that all the ROI parameters are valid and if so then update them.
 *
 * This is used to change OFFSET_X, OFFSET_Y, WIDTH, HEIGHT, BIT_MODE, FRAME_RATE (sometimes), or
 * SENSOR_OPTIONS.
 * Changing any of those involves stopping capture, changing the parameter, then restarting capture.
 * They can result in a different number of frames so the partition manager is involved.
 * If the change results in an invalid frame rate, it is corrected automatically.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_roi_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get a pointer to our support classes.
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer    *cs_ptr = CameraServer::get_cs();
    Sensor      *sensor_ptr = CameraServer::get_sensor();
    int         rc = GEV_STATUS_SUCCESS;

    uint32_t    reg_addr;
    uint32_t    syncin_value;
    uint8_t save_camera_state;
    bitmode_t   bit_mode;
    uint16_t    startx;
    uint16_t    starty;
    uint16_t    width;
    uint16_t    height;
    float    frame_rate;
    uint32_t    shutter_speed;
    uint32_t    sensor_opts;
    uint32_t    temp;

    // Get our address - this is a common routine for a number of
    // parameters, so read the "reg_addr" to determine which one
    // we're interested in specifically.
    reg_addr = regs.reg_addr();

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%04x:%u)", __FUNCTION__, reg_addr, reg_value);

    // Make sure we're in a good state.
    save_camera_state = CameraServer::get_camera_state();
    if ((save_camera_state != GIGE_CAMERA_STATE_IDLE) &&
        (save_camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
        return GEV_STATUS_INVALID_PARAMETER;

    // We cannot make any changes if sync-in is enabled.
    syncin_value = regs[ GIGE_REG_SYNC_IN_ENABLE ].get_value();
    if ((syncin_value & GIGE_SYNC_TYPE_MASK) != GIGE_SYNC_TYPE_INTERNAL)
    {
        FiFPGA      *fpga_ptr = CameraServer::get_fpga();

        // Are we getting a sync-in signal?
        if (fpga_ptr->syncInGetTiming() == 0)
        {
            fi_log(LOG_DEBUG, "%s: Unable to make changes when sync-in is enabled and no signal present!", __PRETTY_FUNCTION__);
            return GEV_STATUS_ACCESS_DENIED;
        }
    }

    // Which register passed in determines what is being changed and we need to gather the
    // other values so that we can validate the whole ROI.
    cs_ptr->get_user_settings(&startx, &starty, &width, &height, &bit_mode, &frame_rate,
                              &shutter_speed, &sensor_opts);

    // We will do some basic sanity checks here that don't involve parameter interdependencies.
    // Things that involve relationships between parameters will be done later.
    switch (reg_addr)
    {
    case GIGE_REG_OFFSET_X:
        startx = reg_value;
        temp = regs[ GIGE_REG_MIN_WIDTH ].get_value();
        if (startx + temp > cs_ptr->get_lic_width())
            rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_OFFSET_Y:
        starty = reg_value;
        temp = regs[ GIGE_REG_MIN_HEIGHT ].get_value();
        if (starty + temp > cs_ptr->get_lic_height())
            rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_WIDTH:
        width = reg_value;
        if (width > cs_ptr->get_lic_width())
        {
            rc = GEV_STATUS_INVALID_PARAMETER;
        }
        else
        {
            temp = regs[ GIGE_REG_SENSOR_WIDTH ].get_value();
            if (width > temp)
                rc = GEV_STATUS_INVALID_PARAMETER;
        }
        break;
    case GIGE_REG_HEIGHT:
        height = reg_value;
        if (height > cs_ptr->get_lic_height())
        {
            rc = GEV_STATUS_INVALID_PARAMETER;
        }
        else
        {
            temp = regs[ GIGE_REG_SENSOR_HEIGHT ].get_value();
            if (height > temp)
                rc = GEV_STATUS_INVALID_PARAMETER;
        }
        break;
    case GIGE_REG_FRAME_RATE:
        frame_rate = cs_ptr->reg_to_frame_rate(reg_value);
        temp = regs[ GIGE_REG_MAX_FRAME_RATE ].get_value();
        if (frame_rate > (float)temp)
            rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_BIT_MODE:
        bit_mode = (bitmode_t)reg_value;
        if (!sensor_ptr->is_bit_mode_supported(bit_mode))
            rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    case GIGE_REG_SENSOR_OPTIONS:
        sensor_opts = reg_value;
        if (!sensor_ptr->is_option_supported(sensor_opts))
            rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    default:
        fi_log(LOG_ERR, "%s: Unexpected addr [0x%x]", __FUNCTION__, reg_addr);
        rc = GEV_STATUS_INVALID_PARAMETER;
        break;
    }

    if (rc != GEV_STATUS_SUCCESS)
    {
        fi_log(LOG_DEBUG, "%s: failed to set reg 0x%04x to 0x%08x=%u status=0x%x", __FUNCTION__,
               reg_addr, reg_value, reg_value, rc);
        return rc;
    }

    // This will "make_capture_valid()" so it will force ROI limits into range,
    // then make sure frame rate works for that ROI, then make sure shutter works
    // for that frame rate.
    if (!cs_ptr->modify_capture(startx, starty, width, height, bit_mode, frame_rate,
                                shutter_speed, sensor_opts))
        rc = GEV_STATUS_ERROR;

    return rc;
}

/*!
 * @brief Update the sensor gain value, which is sensor-specific
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_sensor_gain_update(CS_NETWORK::CS_NETWORK_MESSAGE * /* p_msg */, uint32_t new_sensor_gain)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int     rc;
    uint32_t old_gain;
    bool    immediate = (new_sensor_gain & GIGE_SENSOR_GAIN_IMMEDIATE) ? true : false;

    fi_log(LOG_DEBUG, "%s: new sensor gain=0x%08x", __PRETTY_FUNCTION__, new_sensor_gain);

    // Get a pointer to our support classes.
    CameraServer *cs_ptr = CameraServer::get_cs();
    Sensor *sensor_ptr = CameraServer::get_sensor();

    // Save away the old value.
    cs_ptr->get_sensor_gain(&old_gain);
    fi_log(LOG_DEBUG, "%s: old_gain=0x%08x", __PRETTY_FUNCTION__, old_gain);

    // For the Lince, we can't set gain on the fly so this just sets what will be used
    // the next time we restart the sensor.
    new_sensor_gain &= ~GIGE_SENSOR_GAIN_IMMEDIATE;
    rc = sensor_ptr->set_sensor_gain(new_sensor_gain);
    fi_log(LOG_DEBUG, "%s: set_sensor_gain=%d", __PRETTY_FUNCTION__, rc);
    if (rc == -1)
    {
        fi_log(LOG_WARNING, "%s: new gain setting [0x%08x] is invalid!", __PRETTY_FUNCTION__, new_sensor_gain);
        return GEV_STATUS_INVALID_PARAMETER;
    }
    else if (rc == 1)
    {
        // No change to the gain value.
        return GEV_STATUS_SUCCESS;
    }
    else if (immediate)   /* && ( rc == 0 ) */
    {
        // Save the new value.
        regs[GIGE_REG_SENSOR_GAIN].set_value(new_sensor_gain);

        // Have the sensor reset with the new value.
        cs_ptr->set_camera_state(GIGE_CAMERA_STATE_IDLE);
        bool setupOK = cs_ptr->setup_fpga_and_sensor();
        cs_ptr->set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO);

        // Did we get an error?
        if (!setupOK)
        {
            fi_log(LOG_WARNING, "%s: setup_fpga_and_sensor() failed", __PRETTY_FUNCTION__);

            // Restore the old value.
            regs[GIGE_REG_SENSOR_GAIN].set_value(old_gain);
            return GEV_STATUS_ERROR;
        }

        // Let folks know we've change the analog gain.
        fcp_notify_apps(GIGE_REG_SENSOR_GAIN, new_sensor_gain);
        return GEV_STATUS_SUCCESS;
    }
    else /*  !immediate && ( rc == 0 ) */
    {
        // For Lince, it will go into effect on the next restart, presumably a calibration,
        // but go ahead and set the register even though it is a lie.
        regs[ GIGE_REG_SENSOR_GAIN ].set_value(new_sensor_gain);
        return GEV_STATUS_SUCCESS;
    }
}

/*!
 * @brief Client has requested to update the packet size register. This also includes
 *    a couple of other fields, such as fire a test packet and if packets should
 *    be split.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS if there were no errors.
 */
int fcp_update_packet_size(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%08x,%u)", __FUNCTION__, reg_value, reg_value & 0xffff);

    // update the register
    regs[ GIGE_REG_SCPS0 ].set_value(reg_value);

    return fsp_set_packet_size(reg_value);
}

/*!
 * @brief Client has requested to update the port number register. This creates
 *    the streaming channel. But it won't start streaming until the device specfic
 *    register GIGE_REG_STREAM start is written to.
 *
 * @param p_msg     Message received from the host.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS if there were no errors.
 */
int fcp_update_stream_port_num(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
//    LOG("%s: %u", __PRETTY_FUNCTION__, reg_value);

    // Tell the streaming object to update it's port number.
    fsp_set_stream_port_num(reg_value);

    // If the update resulted in a port number getting set start up a thread
    if (reg_value != 0)
    {
        // Remember what time things got started.
        stream_time = time(NULL);

        // If this was a UDP message, then save where it came from.
        if (p_msg->msg_type == MSG_TYPE_UDP)
            memcpy(stream_from, p_msg->udp_msg.from, sizeof(stream_from));
    }
    else
    {
        // Make sure we unlock.
        fcp_unlock(p_msg);
    }

    return GEV_STATUS_SUCCESS;
}

int FCP_LOCAL::fcp_unlock(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    int    status = GEV_STATUS_SUCCESS;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

#if 1
    LOG("%s", __PRETTY_FUNCTION__);
    // dump_msg( p_msg );
#endif

    // Does the CCP owner also own the FCCP?
    if (is_lock_owner(&fastec_lock_message, &lock_message))
        fcp_update_fastec_lock_state(p_msg, 0);

    // Remove the CCP lock.
    lock_time = 0;
    memset(&lock_message, 0, sizeof(lock_message));

    // Save the value.
    regs[ GIGE_REG_CCP ].set_value(0);
    fcp_notify_apps(GIGE_REG_CCP, 0);

    // Since no one has the lock, no one SHOULD be streaming right now.
    if (fcp_kill_stream_thread() != 0)
        status = GEV_STATUS_ERROR;

    // When the app disconnects, make sure we are not set up for autosave to path.
    {
        uint32_t as_value;
        uint32_t dest_value;

        as_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
        dest_value = regs[ GIGE_REG_IMAGE_DESTINATION ].get_value();

        if (((dest_value & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA) &&
            ((as_value & GIGE_AUTOSAVE_ENABLE) != 0))
        {
            fi_log(LOG_WARNING, "%s: Release LOCK, turn off AUTOSAVE to path", __FUNCTION__);

            // Go through the update function so that it behaves like the app did it.
            // This should allow for the case where FM disconnects while we are trying
            // to autosave to path.
            fcp_autosave_update(NULL, 0);

            // Send a commmand to take the camera out of AUTOSAVE state because
            // no one else is there to do it.
            if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_AUTOSAVE)
                CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO,
                                                     false, NULL);
        }
    }

    // Break any hold, in case the locker failed to complete the commit.
    holding = false;

    return status;
}

/*!
 * @brief update the state of the lock, this can be a client trying to get a lock
 * or one trying to return the lock.
 *
 * @param p_msg - the incoming message
 * @param reg_value - new value for the register
 *
 * @returns GEV_STATUS_SUCCESS if there were no errors.
 */
int fcp_update_lock_state(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    int    status = GEV_STATUS_SUCCESS;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

#if 0
    fi_log(LOG_INFO, "%s: reg_value:0x%x", __FUNCTION__, reg_value);
    // dump_msg( p_msg );
#endif

    // Is the caller attempting to get the lock?
    if (reg_value == GIGE_CCP_CA_BE)
    {
        time_t      cur_time;
        uint32_t    dtime;

        // What time is it now?
        cur_time = time(NULL);

        // Figure out how much time has elapsed.
        dtime = (uint32_t)difftime(cur_time, lock_time);

        // See if the timeout has occured.
        if ((lock_time != 0) && (dtime < get_timeout()) && !is_lock_owner(&lock_message, p_msg))
        {
            status = GEV_STATUS_ACCESS_DENIED;
            fi_log(LOG_WARNING, "%s: attempt to lock CCP, system already locked", __FUNCTION__);
        }
        else
        {
            // If we are being locked again by the same IP that locked us previously,
            // the app must have dropped off and restarted so it doesn't realize
            // we are still holding its lock. In that case, it also wouldn't know
            // we are set up for autosave to path.
            if ((lock_time != 0) && (dtime < get_timeout()) && is_lock_owner(&lock_message, p_msg))
            {
                uint32_t as_value;
                uint32_t dest_value;

                as_value = regs[ GIGE_REG_AUTOSAVE ].get_value();
                dest_value = regs[ GIGE_REG_IMAGE_DESTINATION ].get_value();
                if (((dest_value & GIGE_IMAGE_DEST_MASK) == GIGE_IMAGE_DEST_OFF_CAMERA) &&
                    ((as_value & GIGE_AUTOSAVE_ENABLE) != 0))
                {
                    fi_log(LOG_WARNING, "%s: Reacquired LOCK, turn off AUTOSAVE to path", __FUNCTION__);
                    // Do it as if the app sent the command to disable autosave.
                    // If we are in the middle of a save, this should act like a cancel.
                    fcp_autosave_update(NULL, 0);

                    // Send a commmand to take the camera out of AUTOSAVE state because
                    // no one else is there to do it.
                    if (CameraServer::get_camera_state() == GIGE_CAMERA_STATE_AUTOSAVE)
                        CameraServer::async_set_camera_state(GIGE_CAMERA_STATE_LIVE_VIDEO,
                                                             false, NULL);
                }
            }

            // Save info about who is locking us.
            lock_time = cur_time;
            memcpy(&lock_message, p_msg, sizeof(lock_message));

            // Save the value.
            regs[ GIGE_REG_CCP ].set_value(reg_value);
            fcp_notify_apps(GIGE_REG_CCP, reg_value);
            PDEBUG(PDEBUG_HIGH, "%s: updated CCP with reg_value:0x%x", __FUNCTION__, reg_value);

            // Break any hold, in case the previous locker timed out without committing.
            holding = false;
        }
    }
    else
    {
        PDEBUG(PDEBUG_HIGH, "%s: updated CCP removed lock reg_value:0x%x", __FUNCTION__, reg_value);

        // Unlock the interface.
        status = fcp_unlock(p_msg);
    }

    return status;
}

/*!
 * @brief update the state of the fastec-specific lock, this can be a client trying to get a lock
 * or one trying to return the lock.  the internal GUI app can set an override bit to force acquisition
 *
 * @param p_msg - the incoming message
 * @param reg_value - new value for the register
 *
 * @returns GEV_STATUS_SUCCESS if there were no errors.
 */
int fcp_update_fastec_lock_state(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    uint32_t    old_fccp;
    uint16_t    status = GEV_STATUS_SUCCESS;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

#if 0
    fi_log(LOG_INFO, "%s: p_msg=%p reg_value:0x%x", __FUNCTION__, (void *)p_msg, reg_value);
    // dump_msg( p_msg );
#endif

    // Get the original value for use later.
    old_fccp = regs[ GIGE_REG_FASTEC_CCP ].get_value();

    // Is the caller attempting to get the lock?
    if (reg_value & GIGE_FASTEC_CCP_CA_BE)
    {
        time_t      cur_time;
        uint32_t    dtime;
        bool        already_owned = false;

        // Sanity check.
        if (p_msg == NULL)
        {
            fi_log(LOG_ERR, "%s: Null msg pointer!", __FUNCTION__);
            return GEV_STATUS_INVALID_PARAMETER;
        }

        // What time is it now?
        cur_time = time(NULL);

        // Figure out how much time has elapsed.
        dtime = (uint32_t)difftime(cur_time, fastec_lock_time);

        already_owned = is_lock_owner(&fastec_lock_message, p_msg);

        // See if the lock is available or can be overriden
        if ((fastec_lock_time != 0) && (dtime < get_timeout()) && !already_owned)
        {
            status = GEV_STATUS_ACCESS_DENIED;
            fi_log(LOG_WARNING, "%s: attempt to lock system already locked.", __FUNCTION__);
        }
        else
        {
            // Save info about who is locking us.
            fastec_lock_time = cur_time;
            memcpy(&fastec_lock_message, p_msg, sizeof(fastec_lock_message));

            PDEBUG(PDEBUG_HIGH, "%s: updated FCCP with lock reg_value:0x%x", __FUNCTION__, reg_value);
            if (holding && !already_owned)
            {
                // Lock owner changed so break the hold.
                holding = false;
                fi_log(LOG_WARNING, "%s: acquiring lock broke HOLD", __FUNCTION__);
            }
        }
    }
    else
    {
        PDEBUG(PDEBUG_HIGH, "%s: updated FCCP removed lock reg_value:0x%x", __FUNCTION__, reg_value);

        // Remove the lock.
        fastec_lock_time = 0;
        memset(&fastec_lock_message, 0, sizeof(fastec_lock_message));

        // Break any hold, in case the locker failed to complete the commit.
        holding = false;
    }

    // See if we need to send out a notice.
    if ((status == GEV_STATUS_SUCCESS) && (old_fccp != reg_value))
    {
        // Save the value.
        regs[ GIGE_REG_FASTEC_CCP ].set_value(reg_value);
        fcp_notify_apps(GIGE_REG_FASTEC_CCP, reg_value);
    }

    return status;
}

/*!
 * @brief Called to process the GIGE_REG_IMAGE_DESTINATION register value.
 *
 * If GIGE_IMAGE_SAVE_DELAY is not set, and we are saving on camera, then
 * this routine will call imgSaveWorker::start() to get the save ball rolling.
 *
 * @param p_msg     [UNUSED] Pointer to the message.
 * @param reg_value The value being written to GIGE_REG_IMAGE_DESTINATION.
 * @return One of the GEV_STATUS_XXX values.
 */
int fcp_image_dest_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    // Get pointers to our support classes.
    Sensor  *sensor_ptr = CameraServer::get_sensor();
    PartMgr *p_partmgr = CameraServer::get_partmgr();
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    int     status = GEV_STATUS_SUCCESS;

    PDEBUG(PDEBUG_GIGEWRITE, "%s(0x%08x)", __FUNCTION__, reg_value);
    LOG("%s(0x%08x)", __FUNCTION__, reg_value);

    // Make sure we're not already running.
    if (imgSaveWorker::getInstance()->isRunning())
    {
        fi_log(LOG_ERR, "%s: Image Save is already running!", __FUNCTION__);
        return GEV_STATUS_BUSY;
    }

    // Do sanity checks on the save destination.
    if (((reg_value & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_DNG) &&
        !sensor_ptr->is_sensor_color())   // No DNGs for monochrome sensor.
    {
        return GEV_STATUS_INVALID_PARAMETER;
    }

    // Save the setting.
    regs[ GIGE_REG_IMAGE_DESTINATION ].set_value(reg_value);

    // Start up the image saving process.
    imgSaveWorker::getInstance()->setDestInfo(reg_value);
    imgSaveWorker::getInstance()->setPartNum(p_partmgr->getActivePartition());

    // Only start if we're supposed to: not save delay and not off camera.
    if (((reg_value & GIGE_IMAGE_SAVE_DELAY) == 0) &&
        ((reg_value & GIGE_IMAGE_DEST_MASK) != GIGE_IMAGE_DEST_OFF_CAMERA))
    {
        // Start up the image save thread and wait for it to show the "running" state.
        imgSaveWorker::getInstance()->start();

        // Is image saver is running now?
        if (!imgSaveWorker::getInstance()->isRunning())
        {
            fi_log(LOG_ERR, "%s: Unable to start the image save thread!", __FUNCTION__);
            status = GEV_STATUS_ERROR;
        }
    }

    // Return our status.
    return status;
}

/*!
 * @brief See if the gige manufactures name register is there
 *    if so return it. Otherwise return NULL.
 *
 * @param name  The returned manufacturers name.
 * @param len   The length of the buffer to hold name.
 */
void fcp_get_manufacture_name(char *name, int len)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (name != NULL)
    {
        strncpy(name, (char *)regs[ GIGE_REG_MAN_NAME ], len);
        name[ len - 1 ] = '\0';
    }
}

/*!
 * @brief See if the gige model name register is there
 *    if so return it. Otherwise return NULL.
 *
 * @param name  The returned model name.
 * @param len   The length of the buffer to hold name.
 */
void fcp_get_model_name(char *name, int len)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (name != NULL)
    {
        strncpy(name, (char *)regs[ GIGE_REG_MODEL_NAME ], len);
        name[ len - 1 ] = '\0';
    }
}

/*!
 * @brief See if the gige device version is there
 *    if so return it as a string. Otherwise return NULL.
 *
 * @param ver   The returned version name.
 * @param len   The length of the buffer to hold ver.
 */
void fcp_get_device_version_name(char *ver, int len)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (ver != NULL)
    {
        strncpy(ver, (char *)regs[ GIGE_REG_DEVICE_VER ], len);
        ver[ len - 1 ] = '\0';
    }
}

time_t fcp_get_lock_time()
{
    return FCP_LOCAL::lock_time;
}

time_t fcp_get_fastec_lock_time()
{
    return FCP_LOCAL::fastec_lock_time;
}

/*!
 * @brief See if the gige manufacture info register is there
 *    if so return it. Otherwise return NULL.
 *
 * @param info  The returned manufactures information.
 * @param len   The length of the buffer to hold info.
 */
void fcp_get_manufacture_info(char *info, int len)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (info != NULL)
    {
        strncpy(info, (char *)regs[ GIGE_REG_MAN_INFO ], len);
        info[ len - 1 ] = '\0';
    }
}

/*!
 * @brief See if the gige serial number register is there
 *    if so return it. Otherwise return NULL.
 *
 * @param ser_num   The returned serial number as a string.
 * @param len       The length of the buffer to hold ser_num.
 */
void fcp_get_serial_number(char *ser_num, int len)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    if (ser_num != NULL)
    {
        // Make sure the length parameter is valid.
        if (len > GIGE_DSC_ACK_SERIAL_B_LEN)
            len = GIGE_DSC_ACK_SERIAL_B_LEN;

        // Copy the serial number string over. Make sure it's
        // null terminated.
        strncpy(ser_num, (char *)regs[ GIGE_REG_SERIAL_NUM ], len);
        ser_num[ len - 1 ] = '\0';
    }
}

int FCP_LOCAL::fcp_handle_peekpoke(char *json_ptr, uint32_t& num_bytes, bool do_read)
{
    int status = GEV_STATUS_SUCCESS;
    uint32_t reg_value;
    std::string json_message;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // Are we reading or writing?
    if (do_read)
    {
        Json::Value reg_json;
        Json::StreamWriterBuilder wr_builder;

        // Clear out the destination.
        memset(json_ptr, 0, num_bytes);

        // Read the register requested.
        fpga_ptr->regRead(pp_reg_addr, &reg_value);

        // Build up our JSON message.
        reg_json["reg"] = (uint64_t)pp_reg_addr;
        reg_json["value"] = (uint64_t)reg_value;

        // Create a string.
        wr_builder["commentStyle"] = "All";
        wr_builder["indentation"] = "";
        wr_builder.settings_["precision"] = 2;
        wr_builder.settings_["precisionType"] = "decimal";
        json_message = Json::writeString(wr_builder, reg_json);

        // Make sure we didn't get crazy.
        if (json_message.length() >= num_bytes)
        {
            fi_log(LOG_ERR, "%s: JSON string [%s] is too long!", __PRETTY_FUNCTION__, json_message.c_str());
            status = GEV_STATUS_ERROR;
            goto pp_bail;
        }

        // Copy the string over.
        //    LOG("%s: json='%s'", __PRETTY_FUNCTION__, json_message.c_str());
        memcpy(json_ptr, json_message.c_str(), json_message.length());
        num_bytes = json_message.length() + 1;
    }
    else
    {
        Json::Value requested_settings;
        Json::CharReaderBuilder builder;
        std::unique_ptr<Json::CharReader> reader;
        std::string errInfo;
        bool do_poke = false;
        bool parsingSuccessful;

//        LOG("%s: settings [%s]", __PRETTY_FUNCTION__, json_ptr);

        // Create our reader.
        reader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
        if (reader.get() == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Unable to create a reader!", __PRETTY_FUNCTION__);
            status = GEV_STATUS_ERROR;
            goto pp_bail;
        }

        // Parse the information passed in.
        parsingSuccessful = reader->parse(json_ptr, json_ptr + num_bytes, &requested_settings, &errInfo);
        if (!parsingSuccessful)
        {
            fi_log(LOG_DEBUG, "%s: Error parsing '%s', str_length=%d", __PRETTY_FUNCTION__, json_ptr, num_bytes);
            fi_log(LOG_DEBUG, "%s: Error='%s'", __PRETTY_FUNCTION__, errInfo.c_str());
            status = GEV_STATUS_INVALID_PARAMETER;
            goto pp_bail;
        }

        // Are we poking?
        if (requested_settings["poke"].empty())
        {
            fi_log(LOG_DEBUG, "%s: No poke parameter! Don't know what we are doing!", __PRETTY_FUNCTION__);
            status = GEV_STATUS_INVALID_PARAMETER;
            goto pp_bail;
        }
        try
        {
            do_poke = requested_settings["poke"].asBool();
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[\"poke\"].asBool() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
        }

        // What is the address?
        if (requested_settings["reg"].empty())
        {
            fi_log(LOG_DEBUG, "%s: No reg parameter! Don't know what we are doing!", __PRETTY_FUNCTION__);
            status = GEV_STATUS_INVALID_PARAMETER;
            goto pp_bail;
        }
        try
        {
            pp_reg_addr = (uint32_t)requested_settings["reg"].asUInt64();
        }
        catch (Json::Exception& e)
        {
            fi_log(LOG_DEBUG, "%s: requested_settings[\"reg\"].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
            status = GEV_STATUS_INVALID_PARAMETER;
            goto pp_bail;
        }

        // If we are peeking, then we are done.
        if (do_poke)
        {
            // What is the value we want to write?
            if (requested_settings["value"].empty())
            {
                fi_log(LOG_DEBUG, "%s: No value parameter! Don't know what we are doing!", __PRETTY_FUNCTION__);
                status = GEV_STATUS_INVALID_PARAMETER;
                goto pp_bail;
            }
            try
            {
                reg_value = (uint32_t)requested_settings["value"].asUInt64();
            }
            catch (Json::Exception& e)
            {
                fi_log(LOG_DEBUG, "%s: requested_settings[\"value\"].asUInt() EXCEPTION: %s", __PRETTY_FUNCTION__, e.what());
                status = GEV_STATUS_INVALID_PARAMETER;
                goto pp_bail;
            }

            // Write out the value.
            fpga_ptr->regWrite(pp_reg_addr, reg_value);
        }
    }

pp_bail:
    // Return our status.
    return status;
}

/*!
 * @brief If a stream thread is running tell it to stop.
 *
 * @returns 0 if there were no errors
 */
int FCP_LOCAL::fcp_kill_stream_thread(void)
{
    int rc;

    // Tell the streamer to stop.
    if ((rc = fsp_control_update(NULL, GIGE_STREAM_STOP)) != GEV_STATUS_SUCCESS)
    {
        fi_log(LOG_ERR, "%s: failed to stop FSP stream thread", __FUNCTION__);
        return rc;
    }

    // Update the port number to reflect that the stream has stopped.
    fsp_set_stream_port_num(0);

    // Reset our stream time.
    stream_time = 0;

    // Tell the streamer to shut down the connection.
    fsp_disconnect();

    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Stop, but don't kill the streaming thread.
 *
 * @return 0 if there were no errors, <0 on error.
 */
int fcp_stop_stream_thread(void)
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Tell the streamer to stop.
    if (fsp_control_update(NULL, GIGE_STREAM_STOP) != GEV_STATUS_SUCCESS)
    {
        fi_log(LOG_ERR, "%s: failed to stop FSP stream thread", __FUNCTION__);
        return -1;
    }

    return 0;
}

/*!
 * @brief Return the size (number of bytes) in the configuration file.
 */
uint32_t FCP_LOCAL::get_config_len(void)
{
    int     reg_fd;
    struct stat64 file_stat;
    std::string file_path;

    // Open the register save file.
    file_path = fmt::sprintf("%s/%s", get_home_path(), GIGE_REG_CACHE::REGISTER_FILE);
    reg_fd = open(file_path.c_str(), O_RDONLY);
    if (!reg_fd)
    {
        fi_log(LOG_WARNING, "%s: cannot open file %s to save registers to errno:%d(%s) registers will not be saved",
               __FUNCTION__, file_path.c_str(), errno, strerror(errno));
        return 0;
    }

    // Find out how big the file is.
    if (fstat64(reg_fd, &file_stat) != 0)
    {
        fi_log(LOG_ERR, "%s: Can't stat file %s errno:%d(%s)", __FUNCTION__, file_path.c_str(), errno, strerror(errno));
        file_stat.st_size = 0;
    }

    // Close our file.
    close(reg_fd);

    // Return the size.
    return file_stat.st_size;
}

/*!
 * @brief Read a block from the configuration file, maintaining the position in the
 *    file by using config_pos.
 *
 * @param dest_ptr Where to put the read data.
 *
 * @return Number of bytes read.
 */
int FCP_LOCAL::read_config(uint8_t *dest_ptr, uint32_t num_bytes)
{
    int reg_fd;
    int read_count;
    std::string file_path;

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    // Open the register save file.
    file_path = fmt::sprintf("%s/%s", get_home_path(), GIGE_REG_CACHE::REGISTER_FILE);
    reg_fd = open(file_path.c_str(), O_RDONLY);
    if (reg_fd < 0)
    {
        fi_log(LOG_WARNING, "%s: cannot open file %s errno:%d(%s)",
               __FUNCTION__, file_path.c_str(), errno, strerror(errno));
        return -1;
    }

    // Seek to the proper location in the file.
    lseek(reg_fd, config_pos, SEEK_SET);

    // Figure out how many bytes we want to read.
    if (num_bytes <= GIGE_CONFIG_MAX_XFER_COUNT)
        read_count = num_bytes;
    else
        read_count = GIGE_CONFIG_MAX_XFER_COUNT;

    // Read up the data...as much as we can.
    read_count = read(reg_fd, dest_ptr, read_count);

    // Close the file.
    close(reg_fd);

    // Update our position information.
    if (read_count > 0)
        config_pos += read_count;

    // Return the number of bytes we read.
    return read_count;
}

/*!
 * @brief Write a new configuration register file.
 *
 * @param config_ptr Pointer to the next chunk of the new configuration.
 * @param len        Size of chunk including control byte.
 *
 * @return -1 for failure or 0 if len bytes were successfully written.
 */
int FCP_LOCAL::write_config(uint8_t *config_ptr, uint32_t len)
{
    int reg_fd;
    int write_count;
    uint8_t control;
    uint32_t open_flags;
    std::string file_path;

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    // Get the control byte.
    control = *config_ptr++;
    len--;

    // Figure out which packet this is.
    open_flags = O_RDWR;
    if (control == GIGE_CONFIG_START_PACKET)
        open_flags |= O_CREAT | O_TRUNC;

    // Open the register save file.
    file_path = fmt::sprintf("%s/%s", get_home_path(), GIGE_REG_CACHE::REGISTER_FILE);
    reg_fd = open(file_path.c_str(), open_flags, 0666);
    if (reg_fd < 0)
    {
        fi_log(LOG_WARNING, "(%s): cannot open file %s to save registers to errno:%d(%s) registers will not be saved",
               __FUNCTION__, file_path.c_str(), errno, strerror(errno));
        return -1;
    }

    // Seek to the proper location in the file.
    lseek(reg_fd, 0, SEEK_END);

    // Write the data.
    write_count = write(reg_fd, config_ptr, len);

    // Close the file.
    close(reg_fd);

    if ((write_count <  0) || ((unsigned)write_count != len))
        return -1;

    // Was this the last packet?
    if (control == GIGE_CONFIG_LAST_PACKET)
    {
        std::string md5_path = fmt::sprintf("%s/%s", get_home_path(), GIGE_REG_CACHE::REGISTER_CHECKSUM_FILE);
        storage::md5sum_file(file_path.c_str(), md5_path.c_str());
    }

    // Return zero for success.
    return 0;
}

/*!
 * @brief Read a block from the current LUT file, maintaining the position in the
 *    file by using lut_data_pos.
 *
 * @param dest_ptr Where to put the read data.
 *
 * @return Number of bytes to return, including offset field plus bytes read.
 */
int FCP_LOCAL::read_lut_data(uint8_t *dest_ptr, uint32_t num_bytes)
{
    int     fd;
    int     read_count;
    uint32_t    offset;
    std::string lut_data_filename;

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    if ((lut_data_id < 1) || (lut_data_id > GIGE_LUT_MAX))
    {
        fi_log(LOG_ERR, "%s: can't read LUT file #%u, out of range", __FUNCTION__, lut_data_id);
        return -1;
    }

    // Build up the file name.
    lut_data_filename = fmt::sprintf(LUT_FILENAME_FMT, get_home_path(), lut_data_id);

    // Open the register save file.
    fd = open(lut_data_filename.c_str(), O_RDONLY);
    if (!fd)
    {
        fi_log(LOG_WARNING, "%s: cannot open file %s errno:%d(%s)",
               __FUNCTION__, lut_data_filename.c_str(), errno, strerror(errno));
        return -1;
    }

    // Seek to the proper location in the file.
    lseek(fd, lut_data_pos, SEEK_SET);

    // Network byte order, high byte first
    offset = htonl(lut_data_pos);
    memcpy(dest_ptr, &offset, sizeof(uint32_t));

    // Figure out how many bytes we want to read.
    if (num_bytes <= GIGE_CONFIG_MAX_XFER_COUNT)
        read_count = num_bytes;
    else
        read_count = GIGE_CONFIG_MAX_XFER_COUNT;

    // Read up the data...as much as we can.
    read_count = read(fd, dest_ptr + sizeof(uint32_t), read_count - sizeof(uint32_t));

    // Close the file.
    close(fd);

    // Update our position information.
    if (read_count > 0)
        lut_data_pos += read_count;

    // Return the number of bytes we read.
    return read_count + sizeof(uint32_t);
}

/*!
 * @brief Write a new LUT file.
 *
 * @param data_ptr Pointer to the next chunk of the new configuration.
 * @param len      Size of chunk including offset.
 *
 * @return -1 for failure or 0 if len bytes were successfully written.
 */
int FCP_LOCAL::write_lut_data(uint8_t *data_ptr, uint32_t len)
{
    int         fd;
    int         write_count;
    uint32_t    open_flags;
    uint32_t    requested_offset;
    std::string temp_path;

    // Get the requested position.
    memcpy(&requested_offset, data_ptr, sizeof(uint32_t));
    requested_offset = ntohl(requested_offset);
    len -= sizeof(uint32_t);
    data_ptr += sizeof(uint32_t);

    PDEBUG(PDEBUG_TRACE, "%s(%u,%u)", __FUNCTION__, requested_offset, len);

    // Saved current offset should match what was requested.
    if (requested_offset != lut_data_pos)
    {
        fi_log(LOG_ERR, "%s: out of sync offset=%u expected %u", __FUNCTION__,
               requested_offset, lut_data_pos);
        return -1;
    }

    // Create the file if it is the first packet.
    open_flags = O_WRONLY;
    if (requested_offset == 0)
    {
        open_flags |= (O_CREAT | O_TRUNC);
    }

    if ((lut_data_id < 1) || (lut_data_id > GIGE_LUT_MAX))
    {
        fi_log(LOG_ERR, "%s: can't write LUT file #%u, out of range", __FUNCTION__, lut_data_id);
        return -1;
    }

    // Open the file.
    temp_path = fmt::sprintf(LUT_TEMP_FILENAME, get_home_path());
    fd = open(temp_path.c_str(), open_flags, 0666);
    if (fd < 0)
    {
        fi_log(LOG_WARNING, "%s: cannot open file %s errno:%d(%s)",
               __FUNCTION__, temp_path.c_str(), errno, strerror(errno));
        return -1;
    }

    // Seek to the proper location in the file.
    lseek(fd, 0, SEEK_END);

    // Write the data.
    write_count = write(fd, data_ptr, len);

    // Close the file.
    close(fd);

    if ((write_count <  0) || ((unsigned)write_count != len))
        return -1;

    // Was this the first block of the LUT file?
    if (lut_data_pos == 0)
    {
        // =========================================================================
        // Remember how big this LUT file SHOULD be.
        //
        LUT_FILE *lf_overlay;

        // Overlay the partial memory block.
        lf_overlay = (LUT_FILE *)data_ptr;

        // ALM_TODO: Change the names to what is proper.
        if (lf_overlay->hdr.lutDataType == LUT_SERIESNC_COLOR)
            lut_file_size = sizeof(LUT_RGB_FILE);
        else
            lut_file_size = sizeof(LUT_FILE);
        // =========================================================================
    }

    // Update the data position.
    lut_data_pos += len;

    // Was this the last packet? If so, rename to the requested file number.
    if (lut_data_pos == lut_file_size)
    {
        std::string lut_data_filename = fmt::sprintf(LUT_FILENAME_FMT, get_home_path(), lut_data_id);

        if (rename(temp_path.c_str(), lut_data_filename.c_str()) != 0)
        {
            fi_log(LOG_ERR, "%s: cannot rename %s to %s, errno:%d(%s)",
                   __FUNCTION__, temp_path.c_str(), lut_data_filename.c_str(), errno, strerror(errno));
            return -1;
        }
    }

    // Return zero for success.
    return 0;
}

/*!
 * @brief Update the user LUT number.
 *
 * @param p_msg     Not used.
 * @param reg_value New value for the register.
 *
 * @returns GEV_STATUS_SUCCESS for success, otherwise an error
 */
int fcp_lut_update(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, uint32_t reg_value)
{
    (void)p_msg;

    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    FiFPGA      *fpga_ptr = CameraServer::get_fpga();
    unsigned        new_lut_number;
    bool        rtn = true;

    new_lut_number = reg_value & GIGE_LUT_MASK;
    if (new_lut_number > GIGE_LUT_MAX)
        return GEV_STATUS_INVALID_PARAMETER;

    if ((reg_value & GIGE_LUT_DEFER) == 0)
    {
        if (new_lut_number == 0)
        {
            uint32_t gamma_val = regs[ GIGE_REG_GAMMA ].get_value();
            fpga_ptr->ipmSetLutValues(regs[ GIGE_REG_BRIGHTNESS ].get_value(),
                                      regs[ GIGE_REG_CONTRAST ].get_value(),
                                      gamma_val & 0xFFFF, (gamma_val >> 16) & 0xFFFF);
        }
        else
        {
            rtn = fpga_ptr->ipmLoadUserLut(new_lut_number);
        }

        regs[ GIGE_REG_LUT ].set_value(reg_value);
        fcp_notify_apps(GIGE_REG_LUT, reg_value);
    }
    else
    {
        // DEFER says to init for a readmem or writemem. We don't actually set the register.
        lut_data_pos = 0;
        lut_data_id = new_lut_number;
    }

    if (rtn)
        // Return a good status.
        return GEV_STATUS_SUCCESS;
    else
        return GEV_STATUS_ERROR;
}

/*!
 * @brief Return the status of the boot.
 */
int FCP_LOCAL::update_status(void)
{
    int rc = GIGE_BOOT_STATUS_OK;
    FiFPGA *fpga_ptr = CameraServer::get_fpga();

    // If the environment variable exists, then the update
    // failed. So all is good if the variable does not exist
    // (ie. getenv() returning NULL is a good thing).
    if (getenv("UPDATE_FAILED") != NULL)
        rc |= GIGE_BOOT_STATUS_UPDATE_FAIL;

    // Did we have a watchdog timeout?
    if (fpga_ptr->watchdogTimeout())
        rc |= GIGE_BOOT_STATUS_WDTIMEOUT;

    return rc;
}

//////////////// Public Functions //////////////////////////////

// Notify any internal apps that something of importance has changed.
// Also notify the external app if they are registers for the message
// channel. There can only be one external app.
void fcp_notify_apps(uint32_t reg_addr, uint32_t reg_value)
{
    CS_NETWORK::CS_NETWORK_MESSAGE  msg;
    int         i, cnt, istat;
    uint32_t        unsol_port = 0;
    uint32_t        unsol_addr = 0;
    struct in_addr  unsol_addr_st;
    csWorker       *p_worker;
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    std::unique_lock<std::mutex> locker(notifier_mutex);

    // Increment our message ID.
    event_msg_id++;

    // Clear out our message structure.
    memset(&msg, 0, sizeof(msg));

    // Build up the message structure.
    msg.message.header.req.code = GIGE_REQ_CODE;
    msg.message.header.req.command = htons(GIGE_MSG_EVENTDATA_CMD);
    msg.message.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg.message.header.req.length = htons(GIGE_EVENTDATA_B_LEN);
    msg.message.header.req.req_id = htons(event_msg_id);
    msg.message.fed.stream_ch = GIGE_NO_STREAM_CHANNEL;
    msg.message.fed.event_id  = htons(reg_addr);
    msg.message.fed.reg_val_l = htonl(reg_value);

    // Get our worker pointer.
    p_worker = CameraServer::get_worker();

    // Get the unsolicited message information...if any.
    unsol_port = regs[ GIGE_REG_MCP ].get_value();
    unsol_addr = regs[ GIGE_REG_MCDA ].get_value();

    // Now that the internal app's have all been notified it's time to let the external app
    // if there is any know.
    if (unsol_port && unsol_addr)
    {
        // Setup the message for an unsolicited UDP message.
        msg.msg_type = CS_NETWORK::MSG_TYPE_UDP;
        unsol_addr_st.s_addr = unsol_addr;
        memset(msg.udp_msg.from, 0, sizeof(msg.udp_msg.from));
        strncpy(msg.udp_msg.from, inet_ntoa(unsol_addr_st), CS_NETWORK::NET_ADDR_LEN);

        // Set our address.
        msg.udp_msg.from_port = htons(unsol_port);

        // Send it off.
        istat = p_worker->sendMsg(&msg, sizeof(msg.message.header) + GIGE_EVENTDATA_B_LEN);
        if (istat < 0)
        {
            // if we got an error sending to a remote client, just log it and leave everything
            // as is.
            fi_log(LOG_ERR, "(%s) send_msg failed to external app", __FUNCTION__);
        }
    }

    // Modify the message for the IPC messages.
    msg.msg_type = MSG_TYPE_IPC;
    msg.message.header.req.length = GIGE_EVENTDATA_B_LEN;

    // Now we need to go through the list of apps which have registers and
    // dispatch the same message to all of them.
    //
    // You CANNOT use an iterator here, in case we have to delete one of the
    // entries. Also, we need to work backwards through the queue for exactly
    // that same reason.
    cnt = internal_app_event_queue.size();
    for (i = cnt - 1; i >= 0; i--)
    {
        // Get a reference to the event queue entry.
        CS_IPC_MESSAGE& app_ref = internal_app_event_queue[ i ];

        // Copy the target's address.
        strncpy(msg.ipc_msg.from, app_ref.from, sizeof(msg.ipc_msg.from));

        // Now send the response
        istat = p_worker->sendMsg(&msg, sizeof(msg.message.header) + GIGE_EVENTDATA_B_LEN);
        if (istat < 0)
        {
            // If we got an error sending the response, that usually means that the client is
            // no longer there, so erase that client.
            fi_log(LOG_ERR, "%s: send_msg failed (%d), removing client:%s from event queue",
                   __FUNCTION__, istat, app_ref.from);
            internal_app_event_queue.erase(internal_app_event_queue.begin() + i);
        }
    }

    // Send the notifications to the GRPC interface.
    grpcNotification(reg_addr, reg_value);
}

/*!
 * @brief Initialize the GIGE interface.
 *
 * @returns 0 if there were no errors
 */
int fcp_init(void)
{
    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    return 0;
}

/*!
 * @brief Send out any startup notifications
 *
 * Once initialization has completed to the point where we are ready to talk FCP.
 * we can send out startup notifications in case anyone failed to notice a restart.
 */
void fcp_startup_notify(void)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();
    CameraServer    *cs_ptr = CameraServer::get_cs();

    PDEBUG(PDEBUG_TRACE, "%s", __FUNCTION__);

    // FIXME These are the notifications that were previously sent at the end of load()
    // or that were sent as part of fcp_validate_registers(). Those were pulled out becase
    // we don't want to send out any notifications until we are ready to talk FCP.

    uint16_t    startx = 0;
    uint16_t    starty = 0;
    uint16_t    width = 0;
    uint16_t    height = 0;
    bitmode_t   bit_mode = (bitmode_t)0;
    float    frame_rate = 0;
    uint32_t    shutter_speed = 0;
    uint32_t    sensor_opts = 0;

    cs_ptr->get_user_settings(&startx, &starty, &width, &height, &bit_mode, &frame_rate,
                              &shutter_speed, &sensor_opts);
    fcp_notify_apps(GIGE_REG_OFFSET_X, (uint32_t)startx);
    fcp_notify_apps(GIGE_REG_OFFSET_Y, (uint32_t)starty);
    fcp_notify_apps(GIGE_REG_WIDTH, (uint32_t)width);
    fcp_notify_apps(GIGE_REG_HEIGHT, (uint32_t)height);
    fcp_notify_apps(GIGE_REG_BIT_MODE, (uint32_t)bit_mode);
    fcp_notify_apps(GIGE_REG_FRAME_RATE, cs_ptr->frame_rate_to_reg(frame_rate));
    fcp_notify_apps(GIGE_REG_SHUTTER_SPEED, shutter_speed);
    fcp_notify_apps(GIGE_REG_SENSOR_OPTIONS, sensor_opts);

    fcp_notify_apps(GIGE_REG_NUM_PART, regs[ GIGE_REG_NUM_PART ].get_value());

    // Make sure the FPN status is accurate for the loaded correction table.
    // We didn't want to do this until we were ready to send notifications.
    cs_ptr->refresh_ct_status();

    // After loading GIGE_REG_IMAGE_DESTINATION, we need to set it in the image saver
    // in case we decide to start autosaving without a GUI doing another set.
    // FIXME This is silly. The image saver should initialize itself.
    imgSaveWorker::getInstance()->setDestInfo(regs[ GIGE_REG_IMAGE_DESTINATION ].get_value());

    // We might have changed the image destination, e.g. if we changed LTR mode.
    fcp_notify_apps(GIGE_REG_IMAGE_DESTINATION, regs[ GIGE_REG_IMAGE_DESTINATION ].get_value());
}

/*!
 * @brief Do the necessary steps to shut down the FCP module.
 *
 * @returns 0 if there were no errors
 */
int fcp_exit(void)
{
    GIGE_REG_CACHE& regs = CameraServer::get_regs();

    LOG("%s: entry", __FUNCTION__);

    // If a streaming thread is running shut it down.
    fcp_kill_stream_thread();

    // Only save the settings if we're not doing a factory reset
    // and our initialization is finished.
    if (!CameraServer::in_factory_reset() && CameraServer::init_is_done())
        regs.save();

    // Clean up the register cache.
    regs.reset();

    // if we have a buffer for the histogram, delete it.
    if (p_histogram_rgb)
    {
        free(p_histogram_rgb);
        p_histogram_rgb = NULL;
    }

    LOG("%s: done", __FUNCTION__);
    return 0;
}

/*!
 * @brief Cleanup locks or streams if the client hasn't checked in
 *
 */
void fcp_cleanup(void)
{
    time_t   cur_time = 0;
    uint32_t  timeout = 0;
    uint32_t  elapsed_lock_time = 0;
    uint32_t  elapsed_fastec_lock_time = 0;
    uint32_t  elapsed_stream_time = 0;

    //  Too much yak.
    //  PDEBUG( PDEBUG_TRACE, "%s", __PRETTY_FUNCTION__ );

    cur_time = time(NULL);
    timeout = get_timeout();

    if (lock_time && (lock_time < cur_time))
        elapsed_lock_time = (uint32_t)difftime(cur_time, lock_time);
    if (fastec_lock_time && (fastec_lock_time < cur_time))
        elapsed_fastec_lock_time = (uint32_t)difftime(cur_time, fastec_lock_time);
    if (stream_time && (stream_time < cur_time))
        elapsed_stream_time = (uint32_t)difftime(cur_time, stream_time);

#if 0
    fi_log(LOG_DEBUG, "(%s) starting timeout:%d lt:%u elt:%d flt:%u, eflt:%d, st:%u est:%d",
           __FUNCTION__, timeout, (uint32_t)lock_time, elapsed_lock_time,
           (uint32_t)fastec_lock_time, elapsed_fastec_lock_time,
           (uint32_t)stream_time, elapsed_stream_time);
#endif

    // see if the timeout has expired
    if ((lock_time != 0) && (elapsed_lock_time >= timeout))
    {
        // Get a character representation of time.
        std::string cur_str = std::ctime(&cur_time);
        std::string lock_str = std::ctime(&lock_time);

        // Trim off the ending newline.
        cur_str.resize(cur_str.size() - 1);
        lock_str.resize(lock_str.size() - 1);

        fi_log(LOG_INFO, "%s: timeout expired on CCP register (elapsed=%u >= timeout=%u), breaking lock",
               __FUNCTION__, elapsed_lock_time, timeout);
        fi_log(LOG_DEBUG, "%s: cur_time: '%s', lock_time: '%s'", __PRETTY_FUNCTION__, cur_str.c_str(), lock_str.c_str());
        fcp_update_lock_state(NULL, 0);
    }

    // see if the fastec lock timeout has expired
    if ((fastec_lock_time != 0) && (elapsed_fastec_lock_time >= timeout))
    {
        fi_log(LOG_INFO, "%s: timeout expired on FastecCCP register (elapsed=%u > timeout=%u), breaking lock",
               __FUNCTION__, elapsed_fastec_lock_time, timeout);
        fcp_update_fastec_lock_state(NULL, 0);
    }

    // if a streaming thread is running shut it down.
    if ((stream_time != 0) && (elapsed_stream_time >= timeout))
    {
        // Stop our stream.
        fi_log(LOG_INFO, "%s: timeout expired on stream thread, killing thread", __FUNCTION__);
        fcp_kill_stream_thread();
    }
}

/*!
 * @brief Reset locks and streaming.
 */
void fcp_lock_reset(void)
{
    LOG("%s", __PRETTY_FUNCTION__);

    // Reset the CCP lock.
    fcp_update_lock_state(NULL, 0);

    // Reset the FastecCCP lock.
    fcp_update_fastec_lock_state(NULL, 0);

    // Stop our stream.
    fcp_kill_stream_thread();
}

/*!
 * @brief Parse a newly arrived GigE message and route it to the correct routine
 *    for processing.
 *
 * @param p_msg - the message received for processin
 * @returns 0 if there were no errors
 */
int fcp_parse(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg)
{
    // Protect acccess to the routine.
    static std::mutex fcp_lock;
    std::unique_lock<std::mutex> locker(fcp_lock);

    uint16_t  cmd;

    if (!p_msg)
        return -1;

    fcp_trace(p_msg);

    // First make sure it really is a FCP message.
    if (p_msg->message.header.req.code != GIGE_REQ_CODE)
    {
        // Is this a NULL message? A NULL message is used to break
        // through firewalls. We just ignore it and continue on our way.
        if (p_msg->message.header.req.code == 0)
        {
            // Ignore NULL messages.
            return 0;
        }
        else
        {
            // This is a wierd message!
            fi_log(LOG_WARNING, "%s: Invalid GigE message arrived, invalid REQ_CODE, expecting:0x%02X got:0x%02X",
                   __FUNCTION__, GIGE_REQ_CODE, p_msg->message.header.req.code);
            return -2;
        }
    }

    // Renew the locks.
    renew_lock(p_msg);
    renew_fastec_lock(p_msg);
    renew_stream(p_msg);

    // Now see what the request is for.
    cmd = ntohs(p_msg->message.header.req.command);

#ifdef DEBUG_TIMING
    {
        uint16_t len = ntohs(p_msg->message.header.req.length);
        uint16_t id = ntohs(p_msg->message.header.req.req_id);
        fi_log(LOG_INFO, "Incoming -- hdr: cmd=0x%04x, len=0x%04x, req_id=0x%04x", cmd, len, id);
    }
#endif

    switch (cmd)
    {
    case GIGE_MSG_DISCOVERY_CMD:
        handle_discovery(p_msg);
        break;
    case GIGE_MSG_FORCEIP_CMD:
        handle_forceip(p_msg);
        break;
    case GIGE_MSG_READREG_CMD:
        handle_readreg(p_msg);
        break;
    case GIGE_MSG_WRITEREG_CMD:
        handle_writereg(p_msg);
        break;
    case GIGE_MSG_READMEM_CMD:
        handle_readmem(p_msg);
        break;
    case GIGE_MSG_WRITEMEM_CMD:
        handle_writemem(p_msg);
        break;
    case GIGE_MSG_PACKETRESEND_CMD:
        handle_packetresend(p_msg);
        break;
    default:
        fi_log(LOG_INFO, "%s: Unknown command type:0x%x", __FUNCTION__, p_msg->message.header.req.command);
        break;
    }

#ifdef DEBUG_TIMING
    {
        struct timespec end_time;
        double  start, stop, elapsed;

        clock_gettime(CLOCK_REALTIME, &end_time);

        start = ((double)p_msg->rcv_time.tv_sec) + (((double)p_msg->rcv_time.tv_nsec) / 1000000000.0);
        stop = ((double)end_time.tv_sec) + (((double)end_time.tv_nsec) / 1000000000.0);
        elapsed = stop - start;

        if (elapsed > ((double)2.0))
        {
            fi_log(LOG_INFO, "%s: Took too long! <<<<<<<<<<<<<<<<", __PRETTY_FUNCTION__);
            dump_msg(p_msg);
        }
        else
        {
            uint16_t len = ntohs(p_msg->message.header.req.length);
            uint16_t id = ntohs(p_msg->message.header.req.req_id);

            cmd = ntohs(p_msg->message.header.req.command);

            fi_log(LOG_INFO, "Finished -- duration=%.6lf", elapsed);
            fi_log(LOG_INFO, "Finished -- hdr: cmd=0x%04x, len=0x%04x, req_id=0x%04x", cmd, len, id);

            // Now print out the type specific value(s).
            switch (p_msg->msg_type)
            {
            case MSG_TYPE_IPC:
                fi_log(LOG_INFO, "Finished -- ipc from: '%s'", p_msg->ipc_msg.from);
                break;
            case MSG_TYPE_UDP:
                fi_log(LOG_INFO, "Finished -- udp from: %s:%d", p_msg->udp_msg.from, p_msg->udp_msg.from_port);
                break;
            case MSG_TYPE_FPGA:
                fi_log(LOG_INFO, "Finished -- fpga_val: 0x%08x", p_msg->fpga_val);
                break;
            default:
                break;
            }
        }

    }
#endif

    return 0;
}

/*!
 * @brief Update the "name" of the NTP server we're using.
 *
 * @param ntp_server    The "name" of the server, an empty string to
 *          indicate there's no server.
 */
void FCP_LOCAL::fcp_update_ntp_server(const char *ntp_server)
{
    char temp_buf[ MAX_NTP_SERVER_LEN ];

    // Sanity check our parameter.
    if (ntp_server == NULL)
    {
        fi_log(LOG_ERR, "%s: Invalid server name passed to function.\n", __FUNCTION__);
        return;
    }

    // Update the NTP variable for the next reboot.
    snprintf(temp_buf, MAX_NTP_SERVER_LEN, "%s/%s %s", get_home_path(), NTP_FIX_SCRIPT, ntp_server);
    system(temp_buf);
}

/*!
 * @brief Update the time zone string
 *
 * @param timezone  A string containing the time zone to use
 */
int FCP_LOCAL::fcp_update_timezone(const char *timezone)
{
    char    tempbuf[ PATH_MAX ];
    const char  *s;
    int     tz_len;

    // Make sure /etc/localtime is removed.
    unlink("/etc/localtime");

    // Locate a semicolon separating the two forms.
    s = strchr(timezone, ';');
    if (s != NULL)
        tz_len = s - timezone;
    else
        tz_len = strlen(timezone);

    // If there is anything before the semicolon, it is a friendly timezone name.
    // See if it is usable on this camera, which means it must be in our zoneinfo database.
    if (tz_len != 0)
    {
        FILE    *fp;
        int     str_len;

        strncpy(tempbuf, "/usr/share/zoneinfo/", sizeof(tempbuf));
        str_len = strlen(tempbuf);
        memcpy(tempbuf + str_len, timezone, tz_len);
        *(tempbuf + str_len + tz_len) = '\0';

        // Try to open the file.
        fp = fopen(tempbuf, "rb");
        if (fp != NULL)
        {
            // Okay, the file exists. Use this timezone.
            fclose(fp);

            // Make our localtime link.
            symlink(tempbuf, "/etc/localtime");

            // Copy the timezone info over to be passed to the environment.
            memcpy(tempbuf, timezone, tz_len);
            tempbuf[tz_len] = '\0';
        }
        else
        {
            // Just use this to signal that we need to use the second form.
            tz_len = 0;
        }
    }

    // Either we had no friendly timezone or else we got one we can't use.
    // We'll have to use the second form.
    if (tz_len == 0)
    {
        // Move past the semicolon if one was found.
        // If there was no semicolon, assume the timezone is just GMT+-hh:mm or similar
        // and try to use whatever they sent.
        if (s != NULL)
            strncpy(tempbuf, s + 1, sizeof(tempbuf));
        else
            strncpy(tempbuf, timezone, sizeof(tempbuf));
    }

    // Update our environment variable.
    setenv(TZ_ENV_VAR, tempbuf, 1);

    // This reinitalizes a library static string with the timezone
    // from the TZ environment variable so localtime(), etc., works.
    tzset();

    // Tell the kernel about the timezone so our FAT filesystems will be right.
    set_kernel_timezone();

    // Let clients know we've changed the timezone.
    fcp_notify_apps(GIGE_REG_TIMEZONE, 0);
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Exit, but in such a way that our parent will know to reboot the system.
 */
void fcp_system_reboot()
{
    // Exit letting folks know we want to reboot.
    exit('R');
}

/*!
 * @brief Exit, but in such a way that our parent will know to restart us.
 */
void fcp_system_restart()
{
    // Exit letting folks know we want to restart.
    exit('r');
}

/*!
 * @brief Exit, but in such a way that our parent will know to shutdown.
 */
void fcp_system_exit()
{
    // Exit letting folks know we want to shutdown.
    exit('s');
}

/*!
 * @brief Say whether the IPM P1 port is in use.
 *
 * This is here because other functions in this file
 * need to check availability of resources so it happens
 * to have access to the information it needs.
 *
 * @return true if busy, false if available
 */
bool ipm_p1_is_busy(void)
{
    if (imgSaveWorker::getInstance()->isRunning())
        return true;
    else
        return false;
}
